/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         BitWord.cpp
// Description:  
// Created:      Sep 8, 1999, 18:30
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//========================================================================
#include "BitWord.hpp"
//========================================================================
#ifdef DEBUG_BIT_WORD
 #define DEBUG_NAMESPACE "BitWord"
#else
 #undef DEBUG_NAMESPACE
#endif
#include "debugMacros.hpp"
//========================================================================
using namespace BK;

long BitWord::InitStatic::_count; // Implicitely initialised to 0L
ulong BitWord::baseValAllSet;
ulong BitWord::baseValAllClear;
ulong BitWord::singleBitMask[sizeof(ulong)*CHAR_BIT];
ulong BitWord::allButOneMask[sizeof(ulong)*CHAR_BIT];
ulong BitWord::pieceMask[sizeof(ulong)*CHAR_BIT][sizeof(ulong)*CHAR_BIT];
ulong BitWord::inversePieceMask[sizeof(ulong)*CHAR_BIT][sizeof(ulong)*CHAR_BIT];

void BitWord::initStatic()
{
 baseValAllSet = makeBaseValAllSet(); 
 baseValAllClear = 0ul;
 for (size_t n = 0; n < sizeof(ulong)*CHAR_BIT; n++)
  {
   singleBitMask[n] = makeSingleBitMask(n); 
   allButOneMask[n] = makeAllButOneMask(n);
  };
 for (size_t begin = 0; begin < sizeof(ulong)*CHAR_BIT; begin++)
  for (size_t end = begin; end < sizeof(ulong)*CHAR_BIT; end++)
   {
    pieceMask[begin][end] = makePieceMask(begin,end);
    inversePieceMask[begin][end] = makeInversePieceMask(begin,end);
   };
}; // void BitWord::initStatic()


ulong BitWord::makeBaseValAllSet()
{
 ulong x;
 ulong y = 1ul;
 while (y) { x = y; y = y << 1; };
 // now only the leftmost bit of x is set
 y = x;
 for (size_t n = 1; n < sizeof(ulong)*CHAR_BIT; n++) { x = x >> 1; y |= x; };
 return y;
};

ulong BitWord::makeSingleBitMask(size_t n)
{
 ulong x;
 ulong y = 1ul;
 while (y) { x = y; y = y << 1; };
 // now only the leftmost bit of x is set
 while (n) { x = x >> 1; n--; };
 return x;
};

ulong BitWord::makeAllButOneMask(size_t n)
{
 ulong result = makeSingleBitMask(n);
 result ^= makeBaseValAllSet();
 return result;
};

ulong BitWord::makePieceMask(size_t begin,size_t end)
{
 return ((makeBaseValAllSet() << begin) >> (sizeInBits() - end + begin - 1)) << (sizeInBits() - end - 1);
}; 

ulong BitWord::makeInversePieceMask(size_t begin,size_t end)
{
 return (makePieceMask(begin,end) ^ makeBaseValAllSet());
}; 


//==========================================================================


