/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         CommandLine.h
// Description:  Parsing command line and printing usage.
// Created:      Jan 25, 2000, 15:22
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//

#ifndef COMMAND_LINE_H
//======================================================================================
#define COMMAND_LINE_H
#include <string.h>
#include <limits.h>
#include <float.h> 
#include <iostream.h>
#include "jargon.h"
#include "BlodKorvDebugFlags.h"
#include "GList.h"
#include "Malloc.h"
//======================================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMMAND_LINE
#define DEBUG_NAMESPACE "CommandLine"
#endif
#include "debugMacros.h"
//======================================================================================
namespace BK 
{

class CommandLine
{
  typedef GList<Malloc,const char*,CommandLine> StrList;
  typedef StrList::Iter StrListIter; 

 public: 
  enum ParamType { String, Long, Double, Flag };
  class Enum
    {
    public:
      Enum(const char* val);
      Enum(const char* val,const Enum* next);
      Enum() {};
      void* operator new(size_t);
      const char* _val;
      const Enum* _next;
      friend class CommandLine;
    };
  class Values : private StrListIter
    {
    public:
      Values();
      operator bool() const; 
      bool next(const char*& str);
      bool next(long& n);
      bool next(double& n); 
      bool next(bool& flag);
    private:
      Values(StrList* lst);
      static bool readInf(const char* str,bool& positive);
      friend class CommandLine; 
    }; // class Values
 public:  
  CommandLine();
  ~CommandLine();
 
  void setHidingChar(char ch);

  // describing arguments

  static const Enum* val(const char* val);
  static const Enum* val(const char* val,const Enum* next);

  void option(const char* optName,const char* cmt);
  void option(const char* optName,ParamType paramType,const char* paramName,const char* cmt);
  
  void option(const char* optName,const Enum* paramType,const char* paramName,const char* cmt);



  void nonoptPrefix(int num,
                    const char* const * names,
                    const char* const * comments);
  void nonoptSuffix(int num,
                    const char* const * names,
                    const char* const * comments);
  void nonoptMiddle(const char* name,const char* comment);

  // parsing arguments  

  void arguments(int argC,const char* const * argV);

  bool parse();
 
  // getting values

  Values values(const char* optName) const;

  const char* nonoptVal(const char* name) const;

  Values nonoptMiddleValues() const;

  // getting help

  ostream& printCommand(ostream& str) const;
  void printUsage(const char* command) const;
  void printHelp(const char* command) const;
  void printError() const;
   
 private:
  struct OptDesc 
  { 
    OptDesc(const char* optName,const char* cmt);
    OptDesc(const char* optName,ParamType paramType,const char* paramName,const char* cmt);   
    OptDesc(const char* optName,const Enum* paramType,const char* paramName,const char* cmt);   

    void addVal(const char* val);

    bool shortOpt;
    const char* name;
    const Enum* possibleValues;
    ParamType ptype;
    const char* parName;
    StrList* values; 
    const char* comment;
  };
  
 public: // made public to please VC++
  typedef GList<Malloc,OptDesc,CommandLine> OptDescList;  
 private:
  bool parseOptions();
  bool parseNonopt();
  OptDescList* findDesc(const char* optName,bool shortOpt) const;     
  static bool strpref(const char* s1,const char* s2);
  static bool checkType(const char* str,ParamType tp,const Enum* possibleValues);
  static int findStr(const char* str,int arrSize,const char* const *arr); 
 private:
  OptDescList* optDescs;
  int argc;
  const char** argv; 
  int nonoptArg;
  int nonoptPrefSize;
  const char* const * nonoptPrefNames;
  const char* const * nonoptPrefComments;
  int nonoptSuffSize;
  const char* const * nonoptSuffNames;
  const char* const * nonoptSuffComments;
  const char* nonoptMiddleName;
  const char* nonoptMiddleComment;
  StrList* nonoptMidValues; 
  bool optParsSuccess;
  char hidingChar;
}; //class CommandLine 


}; // namespace BK


//======================================================================================
#undef DEBUG_NAMESPACE
//=======================================================================================
#endif
