/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         Float.hpp
// Description:  Floating point arithmetics with sane comparisons. 
// Created:      Jun 24, 2002, Alexandre Riazanov,riazanov@cs.man.ac.uk  
//============================================================================
#ifndef FLOAT_H
#define FLOAT_H 
//=============================================================================
#include <math.h>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLOAT
 #define DEBUG_NAMESPACE "Float"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace BK 
{
  class Float
    {
    public:
      Float() : _content(0.0) {}; 
      ~Float() {};
      explicit Float(const float& n) : _content(n) {};
      explicit Float(const long& n) : _content((float)n) {};
      explicit Float(const ulong& n) : _content((float)n) {};
      explicit Float(const int& n) : _content((float)n) {};
      operator float() const { return _content; };
      operator bool() const { return fabsf(_content) > _epsilon; }; 
      bool operator==(const Float& n) const
      {
	return fabsf(_content - n._content) <= _epsilon;
      }; 
      bool operator!=(const Float& n) const
      {
	return !(*this == n);
      };
      bool operator>(const Float& n) const
      {
	return (_content - n._content) > _epsilon; 
      };
      bool operator>=(const Float& n) const
      {
	return !(*this < n);
      };

      bool operator<(const Float& n) const
      {
	return (n._content - _content) > _epsilon; 
      };      

      bool operator<=(const Float& n) const
      {
	return !(*this > n);
      };

      Float operator-() const 
	{
	  return Float(-_content);
	};


      Float operator-(const Float& n) const 
	{
	  return Float(_content - n._content);
	};

      Float& operator-=(const Float& n) 
	{
	  _content -= n._content;
	  return *this;
	};      

      Float operator+(const Float& n) const 
	{
	  return Float(_content + n._content);
	};

      Float& operator+=(const Float& n) 
	{
	  _content += n._content;
	  return *this;
	};

      Float& operator++()
	{
	  ++_content;
	  return *this;
	};

      Float operator*(const Float& n) const 
	{
	  return Float(_content * n._content);
	};

      Float& operator*=(const Float& n) 
	{
	  _content *= n._content;
	  return *this;
	};

      Float operator/(const Float& n) const 
	{
	  return Float(_content / n._content);
	};

    private:
      operator int() const 
	{ 
	  CALL("operator int() const");
	  ICP("ICP0");
	  return 0; 
	};
      operator long() const 
	{ 
	  CALL("operator long() const");
	  ICP("ICP0");
	  return 0L; 
	};
      operator unsigned long() const 
	{ 
	  CALL("operator unsigned long() const");
	  ICP("ICP0");
	  return 0UL; 
	};
    private:
      float _content;
      static const float _epsilon = 0.00000000001;
    }; // class Float

}; // namespace BK 


inline
ostream& operator<<(ostream& str,const BK::Float& n)
{
  return str << (float)n;
}; 



//============================================================================
#endif
