/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         FormatText.h
// Description:  Splits a text into lines and removes extra white space.  
// Created:      Feb 2, 2000, 16:45
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//=================================================================
#ifndef FORMAT_TEXT_H
//=================================================================
#define FORMAT_TEXT_H 
#include <string.h>
#include "jargon.h"
#include "BlodKorvDebugFlags.h"
#include "ClassDesc.h"
#include "Malloc.h"
#include "GList.h"
#include "DestructionMode.h"
//=================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORMAT_TEXT
 #define DEBUG_NAMESPACE "FormatText"
#endif
#include "debugMacros.h" 

//=================================================================
namespace BK 
{
class FormatText
{
 private:
  typedef GList<Malloc,char*,FormatText> StrList;
 public:
  FormatText(const char* str) 
   : words(0),
     currLine(new char[1])
  {
   readWords(str);
  };
  ~FormatText() 
    { 
      if (DestructionMode::isThorough())
	{
	  destroyWords(); 
	  delete [] currLine; 
	};
    };
  const char* nextLine(long lineLength);
 private: 
  FormatText() {};
  void readWords(const char* str);
  void destroyWords();
  bool spaceChar(char ch) { return (ch == ' ') || (ch == '\n') || (ch == '\t'); };
 private:
  StrList* words;
  char* currLine;
}; // class FormatText 

}; // namespace BK

//=================================================================
#endif
