/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         GlobalClock.cpp
// Description:  A handy wrapper for the virtual process timer.
// Created:      Jan 30, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//============================================================================
#include <sys/time.h>
#include <signal.h>
#include "GlobalClock.h"
//============================================================================

using namespace BK;
long GlobalClock::InitStatic::_count; // implicitely initialised to 0L
GlobalClock::AlarmHandler* GlobalClock::_currentAlarmHandler; 
int GlobalClock::_elapsedDeciseconds;



GlobalClock::InitStatic::InitStatic()
{
  if (!_count)
    {
      GlobalClock::_currentAlarmHandler = (AlarmHandler*)0;
      GlobalClock::_elapsedDeciseconds = 0;
      static struct itimerval timerSettings;
      timerSettings.it_interval.tv_sec = 0;
      timerSettings.it_interval.tv_usec = 100000; // 1 decisecond
      timerSettings.it_value.tv_sec = 0;
      timerSettings.it_value.tv_usec = 100000; // 1 decisecond
      signal(SIGVTALRM,GlobalClock::alarmSignalHandler);
      setitimer(ITIMER_VIRTUAL,&timerSettings,(struct itimerval*)0);
    };
  ++_count;
}; //GlobalClock::InitStatic::InitStatic()

GlobalClock::InitStatic::~InitStatic()
{
  --_count;
  if (!_count)
    {
		  
    };
}; // GlobalClock::InitStatic::~InitStatic()

void GlobalClock::alarmSignalHandler(int)
{
  signal(SIGVTALRM,GlobalClock::alarmSignalHandler);
  ++_elapsedDeciseconds;
  if (_currentAlarmHandler)
    {
      if (_currentAlarmHandler->_handlingFunction)
	{
	  _currentAlarmHandler->_handlingFunction(_elapsedDeciseconds);
	};
    };
}; // void GlobalClock::alarmSignalHandler(int)

//============================================================================
