/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         GlobalStopFlag.hpp
// Description:  Soft interrupt.
// Created:      Jun 12, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//============================================================================
#ifndef GLOBAL_STOP_FLAG_H
#define GLOBAL_STOP_FLAG_H
//============================================================================
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
#include "DestructionMode.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_GLOBAL_STOP_FLAG
 #define DEBUG_NAMESPACE "GlobalStopFlag"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace BK 
{

class GlobalStopFlag
{
 public:
  class DefaultException 
    {
    public:
      DefaultException() {};
    };
  class HandlerActivator
    {
    public:
      HandlerActivator(void (*handler)())  : _handler(handler) 
	{
	  _previous = GlobalStopFlag::_currentHandlerActivator;
	  GlobalStopFlag::_currentHandlerActivator = this;
	};
      ~HandlerActivator()
	{
	  CALL("destructor ~HandlerActivator()");
	  if (DestructionMode::isThorough())
	    {
	      ASSERT(GlobalStopFlag::_currentHandlerActivator == this);
	      GlobalStopFlag::_currentHandlerActivator = _previous;
	    };
	};
    private:
      HandlerActivator() {};
      void handle();
    private:
      void (*_handler)();
      HandlerActivator* _previous;
      friend class GlobalStopFlag;
    };
 public:
  static void reset() { _stop = false; };
  static void raise() { _stop = true; };
  static void check()
    {
      CALL("check()");
      if (_stop)
	{
	  _currentHandlerActivator->handle();
	};
    };
 private:
  static bool _stop;
  static HandlerActivator* _currentHandlerActivator;
  friend class HandlerActivator;
}; // class GlobalStopFlag


}; // namespace BK


//============================================================================
#endif
