/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         LinearPolynomial.h
// Description:  Small linear polynomials
//               with integer coefficients.
// Created:      Apr 11, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//===================================================================
#ifndef LINEAR_POLYNOMIAL_H
#define LINEAR_POLYNOMIAL_H
//===================================================================
#ifndef NO_DEBUG_VIS
#include <iostream.h>
#endif
#include "jargon.h"
#include "BlodKorvDebugFlags.h"
#include "LinearCombination.h"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LINEAR_POLYNOMIAL
#define DEBUG_NAMESPACE "LinearPolynomial<CoeffType,MaxNumOfVariables>"
#endif
#include "debugMacros.h"
//===================================================================

namespace BK 
{

template <class CoeffType,ulong MaxNumOfVariables>
class LinearPolynomial
{
 public:
  LinearPolynomial() :
    _monomials(),
    _freeMember((CoeffType)0)
    {
    };
  ~LinearPolynomial() {};
  void reset() { _monomials.reset(); _freeMember = (CoeffType)0; };
  LinearCombination<CoeffType,MaxNumOfVariables>& monomials() 
    { 
      return _monomials; 
    };
  const LinearCombination<CoeffType,MaxNumOfVariables>& monomials() const 
    { 
      return _monomials; 
    };
  CoeffType& freeMember() { return _freeMember; };
  const CoeffType& freeMember() const { return _freeMember; };
  
  bool containsVariableWhichIsNotIn(const LinearPolynomial& lp) const
    {
      return monomials().containsVariableWhichIsNotIn(lp.monomials());
    };
  
 private:
  LinearCombination<CoeffType,MaxNumOfVariables> _monomials;
  CoeffType _freeMember;
}; // class LinearPolynomial<class CoeffType,ulong MaxNumOfVariables>

}; // namespace BK



#ifndef NO_DEBUG_VIS

template <class CoeffType,ulong MaxNumOfVariables>
inline
ostream& operator<<(ostream& str,
		    const BK::LinearPolynomial<CoeffType,MaxNumOfVariables>& lp)
{
  str << lp.freeMember();
  if (!lp.monomials().isZero())
    str << " + " << lp.monomials();
  return str;
}; 
#endif




//===================================================================
#endif
