/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         Malloc.h
// Description:  Interface for "malloc" and "free".
// Created:      Feb 2, 2000, 17:10
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Revised:      Mar 3, 2000, Alexandre Riazanov
//========================================================================
#ifndef MALLOC_H
//=============================================================================
#define MALLOC_H 
#include <stdlib.h> 
#include "jargon.h"
#include "BlodKorvDebugFlags.h"
#include "ClassDesc.h"
#include "ObjDesc.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MALLOC
 #define DEBUG_NAMESPACE "Malloc"
#endif 
#include "debugMacros.h"
//=================================================

namespace BK 
{
class Malloc
{
 public:
  static void* allocate(size_t size) 
  { 
   CALL("allocate(size_t size)");
   #ifndef DEBUG_ALLOC
    return malloc(size); 
   #else
    size_t nsize = size/sizeof(void*);
    if (size != nsize*sizeof(void*)) nsize++;
    ObjDesc* objDesc = (ObjDesc*)(malloc(nsize*sizeof(void*) + _prefSize));
    
    #ifdef DEBUG_ALLOC_OBJ_SIZE
     objDesc->size = size;
    #endif
    #ifdef DEBUG_ALLOC_OBJ_STATUS
     objDesc->status = ObjDesc::Persistent;       
    #endif
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     objDesc->classDesc = ClassDesc::universalClassDesc();
    #endif
    #ifdef DEBUG_ALLOC_EVENT_ID
     objDesc->eventId = Debug::event();
    #endif
    return (void*)(((void**)objDesc) + _nprefSize);
   #endif
  }; // void* allocate(size_t size) 
 
  static void deallocate(void* obj,size_t size) 
  {
   CALL("deallocate(void* obj,size_t size)");
   #ifndef DEBUG_ALLOC
    free(obj);
   #else
    ObjDesc* objDesc = hostObj(obj);
    #ifdef DEBUG_ALLOC_OBJ_SIZE
     ASSERT(objDesc->size == size);
    #endif
    #ifdef DEBUG_ALLOC_OBJ_STATUS
     ASSERT(objDesc->status == ObjDesc::Persistent);   
     objDesc->status = ObjDesc::Recycled;    
    #endif
    #ifdef DEBUG_ALLOC_OBJ_TYPE
     ASSERT(objDesc->classDesc == ClassDesc::universalClassDesc());
    #endif 
    #ifdef DEBUG_ALLOC_EVENT_ID
     objDesc->eventId = Debug::event();
    #endif
    #ifndef DEBUG_ALLOC_NO_REUSE
     free((void*)objDesc);
    #endif
   #endif
  }; //void deallocate(void* obj,size_t size)  


  #ifndef NO_DEBUG
   static void* allocate(size_t size,ClassDesc* classDesc) 
   { 
    void* res;
    #ifndef DEBUG_ALLOC
     res = malloc(size); 
    #else
     size_t nsize = size/sizeof(void*);
     if (size != nsize*sizeof(void*)) nsize++;
     ObjDesc* objDesc = (ObjDesc*)malloc(nsize*sizeof(void*) + _prefSize);        
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      objDesc->size = size;
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      objDesc->status = ObjDesc::Persistent;       
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      objDesc->classDesc = classDesc;
     #endif
     #ifdef DEBUG_ALLOC_EVENT_ID
      objDesc->eventId = Debug::event();
     #endif 
     res = (void*)(((void**)objDesc) + _nprefSize);
    #endif
    return res;
   }; // void* allocate(size_t size,ClassDesc* classDesc) 
 
   static void deallocate(void* obj,size_t size,ClassDesc* classDesc) 
   {
    #ifndef DEBUG_ALLOC
     free(obj);             
    #else
     ObjDesc* objDesc = hostObj(obj);
     #ifdef DEBUG_ALLOC_OBJ_SIZE
      ASSERT(objDesc->size == size);
     #endif
     #ifdef DEBUG_ALLOC_OBJ_STATUS
      ASSERT(objDesc->status == ObjDesc::Persistent);   
      objDesc->status = ObjDesc::Recycled;    
     #endif
     #ifdef DEBUG_ALLOC_OBJ_TYPE
      ASSERT(objDesc->classDesc == classDesc);
     #endif
     #ifdef DEBUG_ALLOC_EVENT_ID
      objDesc->eventId = Debug::event();
     #endif 
     #ifndef DEBUG_ALLOC_NO_REUSE
      free((void*)objDesc);
     #endif
    #endif
   }; //void deallocate(void* obj,size_t size,ClassDesc* classDesc)
  #endif 

  #ifdef DEBUG_ALLOC
   static ObjDesc* hostObj(void* obj) 
   {  
    return (ObjDesc*)(((void**)obj) - _nprefSize);
   };
  #endif
 private:
  #ifdef DEBUG_ALLOC
   static const size_t _prefSize; // = sizeof(ObjDesc);
   static const size_t _nprefSize; // = sizeof(ObjDesc)/sizeof(void*);
  #endif
}; // class Malloc



}; // namespace BK

//================================================
#undef DEBUG_NAMESPACE
//============================================================================
#endif
