/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         Queue.hpp
// Description:  Queues of restricted capacity.
// Created:      Dec 24, 1999, 13:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================================
#ifndef QUEUE_H
//=====================================================================
#define QUEUE_H
#include "jargon.hpp"
//====================================================================
namespace BK 
{
template <class ElemType,unsigned long maxSize>
class Queue
{
 public:
  Queue()
  {
   reset();
  };
  ~Queue() 
  {
    CALL("destructor ~Queue()");
    BK_CORRUPT(hd);
    BK_CORRUPT(tl);
  };
  void init()
  {
    reset();
  };

  void destroy() 
  {
    CALL("destroy()");
    BK_CORRUPT(hd);
    BK_CORRUPT(tl);
  };

  void reset()
  {
   hd = 0;
   tl = 0;
  };
  unsigned long size() const { return tl - hd; };
  bool empty() const { return !(size()); };
  bool nonempty() const { return size(); };
  bool full() const { return size() >= maxSize; };
  ElemType& tail() { return data[tl % maxSize]; }; 
  void enqueue() { tl++; }; 
  const ElemType& head() const { return data[hd % maxSize]; };
  void dequeue() 
   { 
    hd++; 
    if ((hd >= maxSize) && (tl >= maxSize)) 
     {
      hd -= maxSize;
      tl -= maxSize;
     };  
   };    
 private: 
  ElemType data[maxSize];
  ulong hd;
  ulong tl;  
}; // template <class ElemType,unsigned long maxSize> class Queue 

}; // namespace BK


//=====================================================================
#endif


