/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         Tab.hpp
// Description:  Tabulating primitives.
// Created:      Feb 18, 2000, 16:22
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//======================================================================
#ifndef TAB_H
//=======================================================================
#define TAB_H
#if (COMPILER != COMPILER_MS_VISUAL_C_PLUS_PLUS)
 #include <strstream.h>
#else
 #include <strstrea.h>
#endif
#include <string.h>
#include <iostream.h>
#include "jargon.hpp"
#include "DestructionMode.hpp"
//=======================================================================
namespace BK 
{
template <class Alloc> 
class Tab
{
 public:
  class ULong
  {
   public:  
    ULong() { reset(0UL); };
    ULong(ulong n) { reset(n); };
    ~ULong() {};
    void reset(ulong n) 
    {
     ostrstream strstr;    
     strstr << n << '\0';   
     ulong ldiff = 10 - strlen(strstr.str()); // too rigid
     strcpy(str + ldiff,strstr.str());
     for (unsigned i = 0; i < ldiff; i++) str[i] = ' ';
    };
    const char* string() const { return str; };
    static ulong Length() { return 10; }; //	too rigid
   private: 
    char str[10 + 1]; // too rigid
  }; // class ULong
 
  class String
  { 
   public:
    String(const ulong& sz) : _size(sz)
    {      
     _str = Alloc::allocate(sz);
     reset(""); 
    };
    String(const ulong& sz,const char* s) : _size(sz)
    {
     _str = Alloc::allocate(sz); 
     reset(s); 
    };
    ~String() 
    {
     if (DestructionMode::isThorough()) Alloc::deallocate(_str,_size);
    };
    void reset(const char* s) 
    {
     ulong ldiff;
     if (strlen(s) > _size) 
      {
       ldiff = strlen(s) - _size;
       strcpy(_str,s + ldiff);
      }
     else
      {
       ldiff = _size - strlen(s);
       strcpy(_str+ldiff,s);
       for (unsigned i = 0; i < ldiff; i++) _str[i] = ' ';
      };
    };
    const char* string() const { return _str; };
    //static ulong Length() { return _size; };  
   private:
    const ulong _size;
    char* _str; 
 }; // class String 
}; // template <class Alloc> class Tab

}; // namespace BK

template <class Alloc>
ostream& operator<<(ostream& ostr,const BK::Tab<Alloc>::ULong& num)
{
 return ostr << num.string();
};

template <class Alloc>
ostream& operator<<(ostream& ostr,const BK::Tab<Alloc>::String& str)
{
 return ostr << str.string();
};

//========================================================================
#endif
