/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         Timer.hpp
// Description:  Interface for "clock".
// Created:      Feb 16, 2000, 15:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef TIMER_H
//=============================================================================
#define TIMER_H 
#include <ctime>
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TIMER
 #define DEBUG_NAMESPACE "Timer"
#endif
#include "debugMacros.hpp"
//============================================================================
 
namespace BK 
{

class Timer
{
 public:
  Timer() 
  {
   CALL("constructor Timer()"); 
   reset(); 
  };
  ~Timer() 
  {
    CALL("destructor ~Timer()");
    BK_CORRUPT(*this);
  }; 
  void reset()
  {
   stopped = true;
   elapsedSec = 0;
  };
  void start()
  {
   if (stopped)   
    {
     stopped = false;
     startTime = clock();     
    };
  };
  void stop()
  {
   if (!stopped)
    {
     stopped = true;
     clock_t currTime = clock();
     if (currTime > startTime)      
       elapsedSec += (((float)(currTime - startTime))/CLOCKS_PER_SEC);
    };
  };

    
  float elapsed()
  {
   if (stopped) return elapsedSec; 
   clock_t currTime = clock();     
   if (currTime > startTime) 
    return elapsedSec + (((float)(currTime - startTime))/CLOCKS_PER_SEC); 
   return elapsedSec;
  };
 private:   
  bool stopped;
  float elapsedSec;
  clock_t startTime;
}; // class Timer

}; // namespace BK
//======================================================================
#endif
