/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         Tuple.hpp
// Description:  Generic tuples.
// Created:      Feb 26, 2000, 16:40
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef TUPLE_H
//============================================================================
#define TUPLE_H 
#include <iostream>
#include "jargon.hpp"
//============================================================================
namespace BK 
{
template <class Type1,class Type2>
class Tuple2
{
 public:
  Tuple2() {};
  Tuple2(const Type1& e1,const Type2& e2) : el1(e1), el2(e2) {};
  ~Tuple2() {};

 public:
  Type1 el1;
  Type2 el2;
  ostream& output(ostream& str) const 
    {
      return str << '<' << el1 << ',' << el2 << '>';
    };   
  bool operator<(const Tuple2& t)
    {
      return (el1 < t.el1) || ((el1 == t.el1) && (el2 < t.el2));
    };

  bool operator>(const Tuple2& t)
    {
      return (el1 > t.el1) || ((el1 == t.el1) && (el2 > t.el2));
    };

  bool operator<=(const Tuple2& t)
  {
    return (el1 < t.el1) || ((el1 == t.el1) && (el2 <= t.el2));
  };

  bool operator>=(const Tuple2& t)
  {
    return (el1 > t.el1) || ((el1 == t.el1) && (el2 >= t.el2));
  };
  bool operator==(const Tuple2& t)
  {
    return (el1 == t.el1) && (el2 == t.el2);
  };   
}; // template <class Type1,class Type2> class Tuple2



template <class Type1,class Type2,class Type3>
class Tuple3
{
 public:
  Tuple3() {};
  Tuple3(const Type1& e1,const Type2& e2,const Type3& e3)
    : el1(e1), el2(e2), el3(e3)
    {
    };
  ~Tuple3() {};
  Type1 el1;
  Type2 el2;
  Type3 el3;
  ostream& output(ostream& str) const 
    {
      return str << '<' << el1 << ',' << el2 << ',' << el3 << '>';
    };
}; // template <class Type1,class Type2,class Type3> class Tuple3

template <class Type1,class Type2,class Type3,class Type4>
class Tuple4
{
 public:
  Tuple4() {};
  Tuple4(const Type1& e1,const Type2& e2,const Type3& e3,const Type4& e4)
    : el1(e1), el2(e2), el3(e3), el4(e4)
    {
    };
  Type1 el1;
  Type2 el2;
  Type3 el3;
  Type4 el4;
  ostream& output(ostream& str) const 
    {
      return str << '<' << el1 << ',' << el2 << ',' << el3 << ',' << el4 << '>';
    };
}; // template <class Type1,class Type2,class Type3,class Type4> class Tuple4

}; // namespace BK

namespace std
{
template <class Type1,class Type2>
ostream& operator<<(ostream& str,const BK::Tuple2<Type1,Type2>& t)
{
  return t.output(str);
};

template <class Type1,class Type2,class Type3>
ostream& operator<<(ostream& str,const BK::Tuple3<Type1,Type2,Type3>& t)
{
  return t.output(str);
};

template <class Type1,class Type2,class Type3,class Type4>
ostream& operator<<(ostream& str,const BK::Tuple4<Type1,Type2,Type3,Type4>& t)
{
  return t.output(str);
};
};

//=============================================================================
#endif
