/************************************************************************
 *                           Fglob.h                                    *
 *                                                                      *
 *	    External variables used in the FINDER sources               *
 *                                                                      *
 ************************************************************************/

#ifdef MAINFILE
#define GLOBAL
#else
#define GLOBAL extern    
#endif

#include "dynstring.h"

//---------------------- Parsing clauses -------------------------------//

GLOBAL dynstring command_line;    // argv in one dimension
GLOBAL dynstring finder_filename; // For use in SCOTT
GLOBAL boolean over_argmax;                 // flag overlarge functors

GLOBAL boolean no_semantics;        // No model exists: all clauses false!

GLOBAL char varlex[VARMAX][LEXLENGTH];      // Lexicon of variable names
GLOBAL LEXITEM lexicon[LEX_MAX];            // All the defined symbols

GLOBAL KIND sort[ST_MAX];                   // The actual sorts (of course)
GLOBAL FUNCTOR fn[FN_MAX];                  // The defined functions

GLOBAL WFF form[WF_MAX];                    // The well-formed expressions

GLOBAL PARAMETERS setting;                  // The optional settings 

// predefined functions
GLOBAL int id[SZ*SZ+1];                     // Constant function =  
GLOBAL int less[SZ*SZ+1];                   // Constant function <  
GLOBAL int greater[SZ*SZ+1];                // Constant function > 
GLOBAL int difference[SZ*SZ+1];             // Constant function DIF 
GLOBAL int neg[3];                          // Constant function negation

GLOBAL CLAUSE latest_clause;                // Most recently parsed clause

//--------------------------- Models--- --------------------------------//
// theories are indexed 0 thru' num_theories. 
// Theory 0 is the base theory and is only used to start a new theory.

GLOBAL int current_theory;     // index of the current theory being tested
GLOBAL int size[MAXISET_ARRAY];            // clauses in FINDER theory
GLOBAL int updates[MAXISET_ARRAY];         // extensions to base theory

GLOBAL CLAUSE theory[MAXISET_ARRAY][CL_MAX]; // Clauses in each theory
// These are a doubly linked list as well as indexed lists
// (the ordering in the linked list changes during runtime).

GLOBAL clause TH[MAXISET_ARRAY]; // Ptrs to the head of each linked list.

// recording information on the base and previous models
GLOBAL int last_sizes[MAXISET_ARRAY][ST_MAX];
GLOBAL set last_sizmasks[MAXISET_ARRAY][ST_MAX];
GLOBAL int last_vlength[MAXISET_ARRAY];
GLOBAL int last_model[MAXISET_ARRAY][V_LENGTH];

GLOBAL functor FN;  // ptr to head fn; the testable structure
GLOBAL int  first_change;       // Most significant update 
GLOBAL int  input_point; // Offset of next input char 

GLOBAL int theBadness;  // Degree of violation of soft constraints
GLOBAL int theTime;     // Clock ticks when (best) model found
GLOBAL int theWork;     // Number of backtracks when (best) model found
GLOBAL int workNow;     // Number of backtracks at end of search


//------------------ Search flags and parameters -----------------------//

GLOBAL boolean cmd_line_input;       // Using command line for input 
GLOBAL boolean normal_input;         // Incorporate latest clause
GLOBAL boolean piping;               // Structures to be piped 
GLOBAL boolean pipe_open;            // Definitions now being piped 
GLOBAL boolean negative;             // Print models that fail test 
GLOBAL boolean clauses_fail;         // New clauses are always false 
GLOBAL boolean cc_badguy;            // Current clause is a bad guy 
GLOBAL boolean firstmodel;           // No model yet in this search 
GLOBAL boolean single_model;         // Generate one model at most 
GLOBAL boolean forbidden[V_LENGTH];  // Can't bactrack from here 
GLOBAL boolean first_newval;

GLOBAL TRIN tr_parameters;           // Info passed to transref 
GLOBAL TRIN ttr_par;                 // Ditto for better testing 

//------------------------- dummy symbols -----------------------------//

GLOBAL functor dummy_symbol;         // Not really a function symbol 
GLOBAL int zero;                     // A dummy integer, value 0

//============================== END ==================================//

