/*------------------------------------------------------------------------*/
/*
 *                   Type definitions for FINDER 
 */
/*------------------------------------------------------------------------*/


typedef enum   { None, Func, Var, Const } syntype;
/*typedef enum { false, true } boolean;*/
typedef enum   { null, brief, full } prints;
typedef enum   { nul, srt, fcn, fcn0, cls, prm, inc } statement_type;

/*--------------------- a list of cannonical names ------------------------*/
typedef struct s_name_list {
  char *name;
  struct s_name_list *nextname;
} NAME_LIST, *name_list;


/*----------------------------- a sort ------------------------------------*/
typedef struct {
  char *sname;         /* The identifier for this sort */
  int sizmin;          /* Least number of things of this sort */
  int sizmax;          /* Greatest ditto */
  int siz;             /* Current number of such things */
  int sum[(SZ*SZ)+1];  /* Addition is a built-in function */
  int minus[(SZ*SZ)+1]; /* Subtraction is a built-in function */
  name_list canonical; /* List of names, for enumerated sort */
  int Ex[SZ+1];        /* "Exists" is a predicate after all! */
  int len[SZ+1];       /* Name lengths of the elements */
  boolean piped;       /* Sort "siz" comes via the pipe */
  set sizmask;         /* Bit mask for range 0..siz */
} KIND, *kind;


/*----------------------- a model of a functor ----------------------------*/
typedef struct s_functor {
  char *fname;                  /* The identifier for this symbol */
  syntype category;             /* Syntactic type (func, pred, etc) */
  int scope;                    /* For parsing purposes */
  int adicity;                  /* Number of argument places */
  int *table;                   /* Start of matrix for this function */
  int fcc;                      /* Index: start of portion of vector */
  int old_fcc[MAXISET_ARRAY];   /* Was fcc on last good model */
  int fats[ARGMAX];             /* For use in matrix lookup */
  boolean total;                /* Function may not have null value */
  boolean injective;            /* Function is one-one */
  boolean leftin;               /* Each row is one-one */
  boolean rightin;              /* Each column is one-one */
  boolean surjective;           /* Function is onto */
  boolean leftsur;              /* Each row is onto */
  boolean rightsur;             /* Each column is onto */
  boolean comm;                 /* Function is binary and commutative */
  boolean piped;                /* Function values come via the pipe */
  boolean nop;                  /* No decrease in change-order priority */
  boolean cut_it;               /* Prohibit backtracking on function */
  kind arg_type[ARGMAX];        /* The types of the argument(s) */
  kind val_type;                /* The type of this function's value */
  prints    pmode;              /* What kind of print display */
  struct s_functor *nextfunct;  /* Linking the list of "real" functors */
} FUNCTOR, *functor;
        

/*----------------------- a well formed formula ----------------------------*/
typedef struct s_wff {
  functor sym;                /* Main predicate or function symbol */
  struct s_wff *sb[ARGMAX];   /* Offsets of the subformulae */
  int val;                    /* Value currently assigned */
  int *matrix;                /* Start of the relevant value table */
  int *sbval[ARGMAX];         /* Pointers to values of subformulae */
  int wcc;                    /* Offset in info vector for t-testing */
  struct s_wff *nextwff;      /* Linking the list of testables */
} WFF, *wff;


/*-----------------a stack of well formed formulae -------------------------*/
typedef struct s_wff_stack {
  wff w;
  struct s_wff_stack *nxt;
} WFF_STACK, *wff_stack;


/*----------------- clause as a pair of wff lists: prem and conc -----------*/
typedef struct s_clause {
  wff_stack prem;                /* negative literals */
  wff_stack conc;                /* positive literals */
  wff_stack varlist;             /* variables in the clause */
  wff_stack testlist;            /* subformuale in test order */
  int weight;                    /* number of function symbols */
  int varcount;                  /* number of variables */
  int softweight;                /* cost of each violation (0 if hard) */
  int badness;                   /* sum of current violation costs */
  boolean preprocessed;          /* already treated as small clause */
  struct s_clause *nextclause;
  struct s_clause *lastclause;
} CLAUSE, *clause;


/*-------------------------- semi-parsed formulas --------------------------*/
typedef struct s_psw {
  functor main;
  wff actual;
  struct s_psw *leftsub, *rightsub;
} PSW, *psw;


/*---------------------a meaningful word in the language -------------------*/
typedef struct {
  char word[LEXLENGTH];
  boolean iskind;            /* flag if word represents a sort */ 
  int offset;                /* its position in the lexicon */
} LEXITEM;


/*----------------------------PARAMETERS------------------------------------*/

typedef struct {
  int max_models;            /* Limit on number of solutions */
  int maxref;                /* Above which "notop" deletions happen */
  int test_maxref;           /* Same for test of long clauses */
  int max_backtracks;        /* Break condition used in SCOTT */
  int max_invalues;          /* Break condition used in SCOTT */
  int pretest_limit;         /* Above which lazy constraint generation */
  int batches;               /* For determining when to cut and guess */
  int extra_batches;         /* Something or other */
  int target_bound;          /* Only count solutions better than this */
  logical maxtime;           /* Break condition - elapsed time limit */
  boolean Nosecs;            /* No dependency-direction */
  boolean invert_change;     /* Variable values increase in test */
  boolean vars_over_wt;      /* Variable count beats weight in clause order */
  boolean topsoff;           /* Old feature - to be deleted */
  boolean report_breaks;     /* Print reason why search stopped */
  boolean noisy;             /* Beep when a model found */
  boolean sortcut;           /* Don't increase domain size if models found */
  boolean timestamps;        /* Print the time on each model found */
  boolean random;            /* Values to be inserted randomly */
  struct printmode {
    unsigned verbosity;      /* Print job, models, stats to order */
    boolean pretty;          /* Human-readable output */
  } pm;
} PARAMETERS;

/*=========================== END =========================================*/

