
#include <stdio.h>
/*
 * #include <sys/malloc.h>
 */
#include <malloc.h>
#include <time.h>
#include <sys/types.h>
#include <sys/times.h>

/*
 *   Where the error messages go to (usually stderr, or stdout for web use)
 */

#define WARNING_STREAM stdout

/*
* Constant definitions
*/

#define REF_MAX 10000
#define REF_DEFAULT 20
#define EXTRA_DEFAULT 5
#define MAX_REF_SIZE 8
#define STAK_MAX (REF_MAX*MAX_REF_SIZE)
#define MAX_PENALTY 1000000000


/*
 * WSZ (word size) is the number of bits in a "logical" variable.
 * SSZ (set size) is the number of words in a "set".
 * SZ is the maximum domain size.
 */

#ifdef SCOTT
#define WSZ 16
#define SSZ 1
#else
#define WSZ 32
#define SSZ 2
#endif

#define SZ (SSZ*WSZ)

#define V_LENGTH 10000 //original is 2048 // can go up to 100000
#define SURJ_MAX 256
#define RRA 50
#define RRB 30

#define NO_REPORT ((unsigned)1)
#define INT_REPORT ((unsigned)2)
#define FLOAT_REPORT ((unsigned)4)
#define DO_REPORT ((unsigned)8)

#define logical unsigned int

#define FOR(x,y) for( y = x; y; y = y->p )
#define FORR(x,y) FOR( x->cellvals, y )

/*
* Type definitions
*/

/*
*typedef enum {
* false,
* true
*} boolean;
*/

#define boolean char
#define false ((boolean)0)
#define true ((boolean)1)

typedef logical set[SSZ];

typedef enum {
  NOCUT,
  FOUND,
  RECORDED
} cutcode;

typedef enum {
  PERFECT,
  NULL_SPACE,
  BIG_SPACE,
  REF_LIST,
  BUST,
  IBUST,
  NULL_REF,
  SURJES,
  SKIP
} exit_code;

typedef enum {
  SETPOSSES,
  DOSMALLS,
  SEARCHES,
  SOLUTIONS,
  BAD_TESTS,
  SURJ_BACKTRACKS,
  NULL_BACKTRACKS,
  F_SUBSUMED,
  B_SUBSUMED,
  B1_SUBSUMED,
  SOFT_DOUBLE,
  PREREFS,
  REFSIZE,
  TOPLESS,
  LOC_MEMORY,
  MEMORY,
  VECMAX,
  BRANCHES,
  PRE_TIME,
  SEARCH_TIME,
  INVALUES,
  MAX_REPORT
} report_code;

typedef struct IL {
  boolean positive;             /* Represents positive literal */
  short int v, w;               /* A couple of integers */
  struct IL *p;                 /* in a list */
} INTLIST, *intlist;


typedef struct reftype {
  int cardinality;             /* Just what is says */
  boolean notop;               /* Useless extension omitted */
  intlist cellvals;            /* The cell-valuess involved */
  int softweight;              /* Cost if violated if a soft constraint */
  struct reftype *nxtr;        /* Just for stack management */
} REF, *ref;


typedef struct rf_list {        /* A list of refutations  */
  ref rf;
  struct rf_list *nextref;
} REFLIST, *reflist;


typedef struct {
  int blocks;                 /* How many times blocked */
  ref back;                   /* For calculating secondaries */
  reflist cvrefs;             /* The refs involving this c-v */
  reflist posrefs;            /* Positives involving this c-v */
  int vpos;                   /* Offset of its cell */
  int penalty;                /* Soft cost of inserting this value */
} CELLVALUE, *cellvalue;


typedef struct {
  short poss;                 /* How many possible values */
  short vpos;                 /* Vector position of this cell */
  reflist inject;             /* Index to injective function */
  CELLVALUE cvals[SZ];        /* The cell-value structures */
  boolean used;               /* Involved in new refutation */
  boolean has_value;          /* This cell has a value */
  boolean fixed;              /* Has only one possible value */
  int times_picked;           /* How often this cell split upon */
  int least_penalty;          /* Penalty of the best available value */
} CELL, *cell;


typedef struct {
  set valset;
  int last_seen[SZ];
  intlist cellset;
} SURJ;


typedef struct {
  logical report_type;
  char s_report[80];
  int i_report;
  float f_report;
  void (*do_report)();
} REPORT;


typedef struct cng {
  int forkpoint;
  int cutval;
  int icomvec[V_LENGTH];
  int inorder[V_LENGTH];
  set possvals[V_LENGTH];
  boolean active;  
  struct cng *lastone;
} CUTT, *cutt;

#define NEWSLENGTH 1024

typedef struct { REF news[NEWSLENGTH]; }  RBATCH,  *rbatch;
typedef struct { INTLIST news[NEWSLENGTH]; }  ILBATCH, *ilbatch;
typedef struct { REFLIST news[NEWSLENGTH]; }  RLBATCH, *rlbatch;


typedef struct {
  int (*Test)();                /* The principal test function  */
  int (*Onerefs)();             /* Removal of impossible values */
  int (*Smallrefs)();           /* Special treatment for smalls  */
  int vlength;                  /* Local vector length    */
  int batches;                  /* Local stack size / 1000  */
  int extra_batches;            /* Batches after pretest  */
  int maxref;                   /* Maximum refutation size  */
  int maxbak;                   /* Maximum number of backtracks */
  int maxinvals;                /* Maximum number of backtracks */
  int verb;                     /* Statistics verbosity setting */
  int ref_record[RRA][RRB];
  int branches[SZ];
  int max_penalty;              /* Maximum badness allowable */
  boolean Nosecs;               /* Suppress secondary refs  */
  boolean topsoff;              /* Remove redundant ref tops  */
  boolean report_breaks;        /* Print reason for end search   */
  boolean forbidden[V_LENGTH];  /* No backtracking  */
  boolean random;               /* Insert values in random order */
  boolean done;                 /* End-of-search status raised  */
  REPORT  reps[MAX_REPORT+1];   /* Search statistics  */
} TRIN, *trin;



typedef struct {
  trin job;
  CELL vector[V_LENGTH];
  set scomvec[V_LENGTH];
  int icomvec[V_LENGTH];
  int inorder[V_LENGTH];
  int coinorder[V_LENGTH];
  int priority[V_LENGTH];
  int this_cell;
  int susguy;
  int firstchange;
  int forkpoint;
  int maxval;
  int stakmax;
  int unfixt;
  int local_memory;
  int max_batches;
  boolean novect;
  boolean ts;
  boolean cut_recorded;
  boolean noref;
  rbatch stak[REF_MAX];
  ilbatch il_stak[STAK_MAX];
  rlbatch rl_stak[STAK_MAX + REF_MAX];
  intlist il_free;
  reflist rl_free;
  reflist duds;
  ref newref;
  ref stak_ptr;
  SURJ surjection[SURJ_MAX];
  cutt cutptr;
  cutcode cut_flag;
  boolean absolute;
  int best[V_LENGTH];
  int best_penalty;
  int total_penalty;
} TRS, *trs;



/*
*  Global variables
*/

#ifdef TRFILE
#define TR_GLOB
#else
#define TR_GLOB extern
#endif

TR_GLOB trs debugger;
TR_GLOB int begin_time, end_time;

// global variable to handle exit on time_limit
TR_GLOB volatile boolean bug_out;


/*
*  Function prototypes
*/


void transref(trin job);
void init_trin(trin tr);
void clean_up(trs T);
void report_initial(trs T);
trs tr_initial(trin job);
void search(trs T);
int pick_cell(trs T);
int set_fork( trs T, int rc);
void randomise(int arr[], int bound);
void greedy_order(int arr[], int pos, cell c);
void delete_cut(trs T);
reflist get_rl(trs T);
void put_rl(trs T, reflist dead);
intlist get_il(trs T, boolean b, int x, int y, intlist ip);
void put_il(trs T, intlist dead);
int batches_used(trs T);
ref get_ref(trs T);
void put_ref(trs T, ref dead, boolean flag);
void rec_put_il(trs T, intlist ip, ref r, boolean flag);
void more_refs(trs T);
void more_rls(trs T);
void more_ils(trs T);
boolean invalue(trs T, cell c, int v);
void transfer_ref (trs T, ref rf);
void blockval(trs T, cell c, int v, ref rrf, report_code RC);
boolean surjok(trs T);
void find_min_penalty(cell c, trs T);
void set_total_penalty(trs T);
void outvalue(trs T, cell c);
void untransfer_ref (trs T, cell c, ref rf);
void unblockval(trs T, cell c, int v, ref rrf);
void mark_delete(trs T, ref r);
void delete_duds(trs T);
void delete_a_dud(trs T, reflist r);
void setref(trs T);
void reduceref(trs T);
void addref(int x, trs T);
void totsus(trs T);
void make_coherent(trs T);
void no_ref(trs T);
void new_small_ref(int cells[], int vals[], boolean valency[],
		   int softweight, trs T);
void remove_value(trs T, cell c, int x);
intlist get_all_ils(int cells[], int vals[], boolean valency[], int x, trs T);
boolean subsumed(ref the_ref, trs T);
boolean subsumes(ref little_ref, ref big_ref);
void back_sub(trs T, ref the_ref);
boolean identical_ref( ref the_ref, trs T );
boolean is_the_same( ref the_new_ref, ref the_old_ref );
void Ref(int x, trs T);
void record_cut(trs T);
void record_part_cut(int x, trs T);
void complete_cut(int ex, trs T);
boolean prepared(trs T);
void setinx(trs T);
boolean pre_test(trs T);
boolean absfixt(int x);
void surjective(boolean vec[], set vset, trs T);
void injective(boolean vec[], int x, trs T);
void co_priority(int x, int y, trs T);
void ref_size_report();
void Report(trin tr);
void report_branches(trin tr);
void dump_refs(int rows, int columns);
void job_done(trin t, char *s);
int timestamp();
void skipout(char s[], int exitcode);
void add_penalty(int vp, int val, int x, trs T);
void subtract_penalty(int vp, int val, int x, trs T);
void set_best(trs T);

/*
 * Prototypes of functions in sets.c
 */

void init_sets();
void fullset(set s);
void emptyset(set s);
boolean isemptyset(set s);
void singleton(int x, set s);
boolean issingle(set s);
void assignset(set dest, set source);
void intoset(int x, set s);
void outofset(int x, set s);
boolean isinset(int x, set s);
void intersect(set dest, set mask);
