;======================================================================
;
;                        resbase.scm
;
;                   Gandalf version c-2.6 beta
;
; The basic functions of the resolution proof searcher:
; data representation, variables, substitution, unification, subsumption.
;
;                  Copyright (c) Tanel Tammet
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 1, or (at your option)
; any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; The author can be reached at tammet@staff.ttu.ee, tanel@cs.ioc.ee or
; tammet@cs.chalmers.se
;
;
;=======================================================================

;----------------------------------------------------------------------
;
;                          compiler defs
;
;----------------------------------------------------------------------

(define compile-inline
 '(shiftleft shiftright ; logsleft logsright
   composite? non-composite? var? non-composite:-var?
   const? single-occ-lit-var? repeated-occ-lit-var? later-occurrences-var?
   term-may-cont-vars? fun-may-cont-vars? fun-may-cont-repeatedvars?
   pure-fun pure-unsigned-fun fun-depth-strength fun-depth neg-fun? pos-fun?
   complementary-fun? fun-can-subsume? ground-lit-equal?
   clinfo-history-sublist clinfo-clnumber-sublist clinfo-flags-sublist

   get-litinfo-ordernr put-litinfo-ordernr
   not-zero? mymax maxplus

   var-nr y-var? x-var? unassigned-val? touched-var? set-var-touched!
   var-ref set-var-ref! ))
   ; bigger-strength?
   ; bigger-strength-aux?
   ; bigger-strength-aux-aux?))

(define compile-inline-vars
  '(
*max-nr-of-vars-in-clause* *maxvar*  *max-x-var* *max-y-var*
*minconst* *mincurryconst* *maxconst* *unassigned* *var-info-mask* *single-occ-lit-var-mask*
*no-occ-otherlit-var-mask* *single-occ-clause-var-mask*
*no-occ-folllit-var-mask*
*var-nr-mask* *y-var-mask* *z-var-mask* *sgn-mask* *signedfnsymb-mask*
*fnsymb-mask* *fnsymb-shift* *predsymb-mask* *maxvar-mask*
*fun-except-varinf-mask* *repeated-vars-mask* *ground-mask* *strength-mask*
*term-depth-strength-mask* *term-depth-mask* *term-depth-max*
*term-depth-max-shift*
*term-depth-shift* *term-depth-1-shift* *term-size-strength-mask*
*term-size-mask*
*term-size-shift* *term-size-max* *term-size-max-shift* *term-size-1-shift*
*const-number-mask* *const-number-max* *g-and-const-mask*
*cl-strength-mask* *cl-maxvar* *cl-maxvar-shift*
*cl-maxvar-mask* *cl-maxlen* *cl-len-shift* *cl-len-mask*
*cl-rep-predsymbs-mask* *cl-pred-overflow-mask* *cl-predsymbs-area-mask*
*cl-predsymbs-neg-shift* *cl-predsymbs-neg-mask* *cl-predsymbs-pos-max*
*cl-predsymbs-pos-mask*
*litinf-noresolve-mask* *litinf-nofactor-mask* *litinf-isanswer-mask*
*litinf-noparainto-mask* *litinf-noparafromleft-mask*
*litinf-noparafromright-mask*
*litinf-rulelit-mask* *litinf-leftorrulelit-mask*
*litinf-rightimplrulelit-mask* *litinf-reductionrulelit-mask*))



(define compile-stable-vectors '(*var-val-table* *var-touched-table*))


(define compile-uninterned-variables
  '(*occurcheck-flag* *repeatedvars-flag* *var-set-flag* *alt-var-set-flag*
    *xvars-set-flag* *yvars-set-flag* *fac-vars-set*
    *param-use-left-flag* *param-use-right-flag*
    *clause-var-nr* *gcl-maxvar*

    *litinf-nouse-mask* *litinf-nobinrespara-mask*))

;-----------------------------------------------------------------------


; The following two defs are OK for the interpreter,
; UNLESS scmhob.scm is used, defining shiftleft and shiftright x y

; (define (shiftleft x y) (ash x y)) ; interpreter version
; (define (shiftright x y) (ash x (- 0 y))) ; interpreter version

; The following two defs are OK for the compiler

 (define (shiftleft x y) (logsleft x y)) ; compiler version
 (define (shiftright x y) (logsright x y)) ; compiler version

;----------------------------------------------------------------------
;
;                      global limits
;
;----------------------------------------------------------------------

(define *max-nr-of-vars-in-clause* 255) ; could be made three bits bigger
(define *maxvar* 65535)     ;;; max nr of an int encoding a var and
                            ;;; its features at the location.
(define *max-x-var* 256)    ;;; vars larger than *max-x-var* are
                            ;;; yvars (ones in the renamed literals)
                            ;;; or zvars (ones in demodulators)

(define *max-y-var* 512)    ;;; analogous to *max-x-var*. Larger are zvars.


(define *minconst* 65536)   ;;; smallest possible const nr.
(define *mincurryconst* 200000)
(define *maxconst* 99999998) ;;; largest allowed const nr; here it is
                             ;;; smaller than it could be

(define *unassigned* 99999999) ;;; the spec value indicating no assignment:
                               ;;; it must NOT be in a var range!
                                ;;; Here it is *maxconst* + 1

(define *default-time-limit* #f) ; timelimit taken if none given
(define *timelimit* #f) ; no timelimit set here
(define *args-timelimit* #f) ; timelimit given at command line

;--------------------------------------------------------------------
;
;               various locally used global vars
;
;--------------------------------------------------------------------

(define *listofnull* '(())) ; a special constant: single-el list

(define *foundflag* #f)      ; set to #t iff the empty clause is found
(define *sat-found-flag* #f)
(define *answer-cl* '())     ; set to a pure answer-clause if empty is found
(define *answer-history* '()) ; set to history of the answer clause
(define *too-many-vars-flag* #f) ; set when such a clause is generated


(define *occurcheck-flag* #f) ;;; used locally in unify
(define *repeatedvars-flag* #f) ;;; used locally in unify
(define *var-set-flag* #f)    ;;; used locally in subsumption
(define *alt-var-set-flag* #f) ;;; used locally in subsumption, set in case
                               ;;; var is set, but *var-set-flag* is not
(define *clause-var-nr* 0) ;;; the temporary maxvar for a clause
(define *new-var-names* '()) ;;; the tmp var-renaming alist for a clause
(define *xvars-set-flag* #f) ;;; indicates whether any xvar was lately set
(define *yvars-set-flag* #f) ;;; same for yvars
(define *fac-vars-set* #f)  ;;; possibly set to #t by factor-unify-ordinary
(define *cut-clauses* '())  ;;; used in unit-handle-cl!
(define *gcl-maxvar* 0)     ;;; used in clause subsumption
(define *gcl-varsstack* '());;; used in clause subsumption
(define *collected-vars* '()) ;;; used in the version of term subsumption
                              ;;; keeping track of set vars
(define *subsuming-clause* '()) ;;; set by havealready? and unit-handle
                                ;;; in case of success
(define *stopsubsuming-flag* #f) ;;; used by the subsumed-clause-aux?
                                 ;;; to signal stopping to havealready
(define *prohibit-subs-clnr* #f) ;;; clnumber which is prohibited for
                       ;;; subsumption. Used for blocking subs of given cl.
 (define *subs-factor-derived-flag* #f) ;;; used by factor-aux to signal
                                       ;;; a subsuming factor to factor
(define *subsumed-clauses* '()) ;;; used by back-subsume
(define *used-demodulators* '()) ;;; formed during demodulation
(define *ground-check-res* '()) ;;; used in check-discr-tree
(define *compare-ground-litcdr-equal-flag* #f) ;;; used by
         ;;; resolve-unify-ordinary. Should be almost always #f, so that
         ;;; ground lit cdr-s are compared by eq only.
(define *one-step-hyper-phase* #f) ;;; used in hyperresolution
         ;;; to indicate that the one-step phase is going on
(define *noncomputable-ans-funs* '())
(define *noncomputable-ans-preds* '())
;; the following are used for handling discrimination trees
(define *cur-lit* '())      ;;; used in handling the discrimination tree
(define *cur-clause* #f)    ;;; used in handling the discrimination tree
(define *overdouble-discr-flag* #f) ;;; #t iff searching for a clause
      ;;; longer than 2
(define *nonunit-discr-flag* #f) ;;; #t iff searching for a nonunit clause
(define *nonunit-res*  '())   ;;; a list of lists of long candidate clauses

;;; the following block of vars is used by paramodulation
(define *param-tolit* '())
(define *param-eqlit* '())
(define *param-tocl* '())
(define *param-fromcl* '())
(define *param-tofullinf* '()) ;simpleh
(define *param-fromfullinf* '())
(define *param-left-arg* '())
(define *param-right-arg* '())
(define *param-use-left-flag* #f)
(define *param-use-right-flag* #f)
(define *param-toinf* '())
(define *param-frominf* '())
(define *param-from-is-y-flag* #f)
(define *param-yclmaxvar* '())
(define *param-xclmaxvar* '())
(define *param-fromhist* '())
(define *param-tohist* '())
(define *param-fromlitnr* '())
(define *param-tolitnr* '())
;;; end of special paramodulation vars


;--------------------------------------------------------------------
;
;  the bitwise encoding of data in fun-place and pred-place objects
;
;--------------------------------------------------------------------
;
; Variables have the following information encoded:
; single-occurrence-in-lit-flag, no-occurrences-in-other-lits-flag,
; no-future-occurrences-in-later-clauses-flag, variable nr.
;
; *maxvar* is currently 65535 (two bytes of 1-s). No encoded var may be
; bigger than that.
;
; bit positions: (NB!! the actual reperesentation has extra two bits
; 10 after low end (eg, scheme num 1 is actually repr. as 110, 4 as 10010)
;
;
; HIGH 30          24              16              8             1 LOW
;     | , , , , , | , , , , , , , | , , , , , , , | , , , , , , , |
;     |<-   not used (all 0)    ->|s|n|f|<???>|z|y|  <- var nr -> |
;
; s = flag for single occurrence of a var in a given literal. 1: single
;     Mask 32768, shift 15.
; n = flag for no occurrences in other lits of a clause. 1: no occs
;     Mask 16384, shift 14.
; f = flag for no occurrences in FOLLOWING literals of a clause. 1: no occs
;     Mask 8192, shift 13.
; y = flag for yvars: 1: yvar, 0: xvar, 0: zvar (all kept clauses
;      contain only xvars)
; z = flag for zvars: 1: zvar, 0 : xvar or yvar (zvars occur in and only in
;      demodulators).
; var nr = actual variable nr.
;     13 bits, mask 8191, no shift.
;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define *var-info-mask* 57344)
 ; (shiftleft (string->number "111" 2) 13) )

(define *single-occ-lit-var-mask* 32768)
;    (shifleft 1 15) )

(define *no-occ-otherlit-var-mask* 16384)
;    (shiftleft 1 14) )

(define *single-occ-clause-var-mask* 49152)
;    (shiftleft 3 14) )

(define *no-occ-folllit-var-mask* 8192)
;    (shiftleft 1 13) )

(define *var-nr-mask* 8191)
;    (string->number "1111111111111" 2) )

(define *y-var-mask* 256) ; if this is 1, var is an yvar
  ;     (shiftleft 1 8) )


(define *z-var-mask* 512); if this is 1, var is an zvar
  ;   (shiftleft 1 9) )


(define *var-val-table*
        (make-vector (+ 1 (logior *max-nr-of-vars-in-clause*
				  *z-var-mask*))
		     *unassigned*))

(define *var-touched-table*
         (make-vector (+ 1 (logior *max-nr-of-vars-in-clause*
				   *z-var-mask*))

		       #f))

;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; Constants are not encoded: start at *minconst*
; which is currently 65536 (thus, two lower bytes are free for
; variables) and may rise up to 30 bits.
;
; Function-symbol-position objects encode data about the term they are
; leading:
;
; function symbol itself, the var in the term with the max nr,
; term size, term depth, for predicates also the sign.
;
; HIGH 30          24              16              8             1 LOW
;     | , , , , , | , , , , , , , | , , , , , , , | , , , , , , , |
;    sgn| pred/fn symb  |maxvr|r|g| t depth |t size     | cnst nr |
;
; where:
;
; sgn = 0 or 1: 0 means a positive predicate, 1 negative
;        Needs shift 29
;
; pred/fn symb = number of the leading pred/fn symb.
;       Nr of bits 8. Max value thus 255. Needs shift 21.
;
; maxvar in term = the var in the term with the largest number
;             NB! If the actual number is too big, a max nr is taken
;                 (that is: all bits 1)
;       Nr of bits: 3. Max value thus 7. Needs shift 18.
;
; r = contains-repeated-vars flag. 1 means contains, 0 means does not.
;        Needs shift 17.
;
; g = ground/nonground flag. 1 means ground, 0 nonground
;        Needs shift 16.
;
; t depth = the depth of term (x:0, f(x,1):1, f(g(g(x)),a):3)
;           NB! If the depth is too big, the max nr is taken.
;        Nr of bits: 5. Max value thus 31. Needs shift 11.
;
; t size = the size of the term (x:1, c:1, f(x,1):3, g(g(x)):3)
;             NB! if the size is larger than the max accomodatable
;             nr, the max nr is taken (all bits 1-s).
;        Nr of bits: 6. Max value thus 63. Needs shift 5.
;
; cnst nr = the number of constant occurrences in the term
;           (f(x,x):0, f(x,c):1, f(a,x,a):2, f(g(c),a,a):3)
;           NB! If the depth is too big, the max nr is taken.
;        Nr of bits: 5. Max value thus 31. Needs shift 0 (no shift)
;
;                            MASKS
;
; HIGH 30          24              16              8             1 LOW
;     | , , , , , | , , , , , , , | , , , , , , , | , , , , , , , |
;       | |pred/fn symb |maxvr|r|g| t depth |t size     | cnst nr |
;       sgn
;     ->|  |<-  268435456
;       |<-  534773760 >|
;                     ->|       |<-  1966080
;                             ->| |<-  65536
;                               |<--      131071               -->|
;                                 |<--      65535              -->|
;                                           |<-   2047          ->|
;                                                       |<- 31  ->|
;
;-------------------------------------------------------------------

;;(define *signedfnsymb-mask* ; sgn and pred/fn symb
;; ;    (shiftleft (string->number "111111111" 2) 21) ) ; for short ints
;;  (- 0 (shiftleft (string->number "11111111" 2) 21)) ) ; generic

(define *sgn-mask* 268435456)
 ; (shiftleft 1 28) )

(define *signedfnsymb-mask* 534773760) ; pred/fn symb with sign
 ;  (shiftleft (string->number "11111111" 2) 21) )

(define *fnsymb-mask* 534773760); pred/fn symb with sgn
 ;  (shiftleft (string->number "11111111" 2) 21) )

(define *fnsymb-shift* 21)

(define *predsymb-mask* 266338304); pred symb, no sgn
 ; (shiftleft (string->number "1111111" 2) 21) )

(define *maxvar-mask* 1835008); maxvar only
 ; (shiftleft (string->number "111" 2) 18) )

(define *fun-except-varinf-mask* 534904831)
 ;  (string->number "11111111000011111111111111111" 2) )

(define *repeated-vars-mask* 131072) ; 1 means: contains repeated vars,
                                     ;0: doesn't
;  (shiftleft 1 17) )

(define *ground-mask* 65536); ground/nonground only
 ; (shiftleft 1 16) )

(define *strength-mask* 131071); starting with |g| to the end.
 ; (string->number "11111111111111111" 2) )

(define *term-depth-strength-mask* 65535); starting with |t depth| to the end.
 ;  (string->number "1111111111111111" 2) )

(define *term-depth-mask* 63488); only t-depth part
 ; (string->number "1111100000000000" 2) )

(define *term-depth-max* 31)
 ; (string->number "11111" 2) )

(define *term-depth-max-shift* 63488)
 ;  (shiftleft (string->number "11111" 2) 11) )

(define *term-depth-shift* 11)

(define *term-depth-1-shift* 2048)
 ;  (shiftleft 1 11) )

(define *term-size-strength-mask* 2047) ; starting with |t size| to the end.
 ; (string->number "11111111111" 2) )

(define *term-size-mask* 2016) ; only t-size
 ; (string->number "11111100000" 2) )

(define *term-size-shift* 5)

(define *term-size-max* 63)
 ;  (string->number "111111" 2) )

(define *term-size-max-shift* 2016)
 ; (shiftleft (string->number "111111" 2) 5) )

(define *term-size-1-shift* 32)
 ;  (shiftleft 1 5) )

(define *const-number-mask* 31)  ; only |cnst nr|
 ; (string->number "11111" 2) )

(define *const-number-max* 31)
 ;  (string->number "11111" 2) )

(define *g-and-const-mask*
  (logior *ground-mask* *const-number-mask*) )

(define *var-strength*
  *term-size-1-shift*)

(define *const-strength*
  (logior *ground-mask*
     (logior *term-size-1-shift*
             1)))


;--------------------------------------------------------------------
;
;                   the representation of clauses
;                              and
;    bitwise encoding of data in fun-place and pred-place objects
;
;--------------------------------------------------------------------
;
; a clause:
;
; (info literal-1 ... literal-n), where
;
; info:
;
; (strengthint predsymbsint litinfolst weight clnumber history
;  distance ...),
;         where ... is any extra data list
;
; - - - - - strengthint:  - - - - -
;
; HIGH 30          24              16              8             1 LOW
;     | , , , , , | , , , , , , , | , , , , , , , | , , , , , , , |
;       |   maxvar  |    len    |g| t depth |t size     | cnst n  |
;
; maxvar: max var nr in a clause. If max nr, then *max-nr-of-vars-in-clause*
;         is the actually right one.
;         max nr = 63, six bits.
;
; len: length of a clause, starting at 1, max 63
;      six bits
;
; |g|.....| cnst n |: exactly the same as for terms:
;
; g: ground flag: 1: ground, 0:non-ground
; t depth: max literal depth in a clause
; t size: size of a clause
; cnst nr: nr of constants in a clause
;
; - - - - - predsymbsint: - - - - - - -
;
; predsymbsint encodes 13 different predicate symbols (i.e. 26 with signs)
; by bits. It is used for finding whether one clause has a predsymbset
; which is a subset of the predsymbset of another one, etc.
;
; HIGH 30          24              16              8             1 LOW
;     | , , , , , | , , , , , , , | , , , , , , , | , , , , , , , |
;       ->| |<- rep-predsymbs
;         ->| |<- pred-overflow
;             |<-       neg-preds     ->|<-       pos-preds     ->|
;
; rep-predsymbs: 1 if there are lits with the same predsymb; 0 otherwise
; pred-overflow: 1 if there are predsymbs except the 26 encodable ones
; neg-preds: an area for negative predsymb bits
; pos-preds: an area for positive predsymb bits
;
; - - - - - litinfolst: - - - - - - -
;
; a list of integers corresponding to the literals in the clause,
; containing extra info about the literals (eg resolvable/not,
; hyperindex)
;
; may be nil.
;
; - - - - - weight: - - - - - - - - - -
;
; a user-defined weight of the clause. By default its size.
;
; - - - - - clnumber: - - - - - -- - -
;
; the unique number of the clause
;
; - - - - - - history: - - - - - - - -
;
; the derivation history of the clause
;
;--------------------------------------------------------------------

; - - - - - - - - - - - - strengthint - - - - - - - - -

(define *cl-strength-mask* 8388607)  ; starting with | len | to the end.
 ; (string->number "11111111111111111111111" 2) )

(define *cl-maxvar* 63) ; the max number encodable in the cl strengthint
 ; (string->number "111111" 2) )

(define *cl-ground-flag* 65536)
;(shiftleft 1 16)

(define *cl-maxvar-shift* 23)

(define *cl-maxvar-mask* 528482304)
 ; (shiftleft (string->number "111111" 2) 23) )

(define *cl-maxlen* 63)  ; the max cl len encodable in the cl strengthint
 ; (string->number "111111" 2) )

(define *cl-len-shift* 17)

(define *cl-len-mask* 8257536)
 ; (shiftleft (string->number "111111" 2) 17) )

; - - - - - - - - - -- - - - predsymbsint - - - - - - - - - - -

(define *predsymb-bit-vector* ; each pred has an int with 1 bit set
  (make-vector 255 0) )

(define *cl-rep-predsymbs-mask* 134217728); 1 for repeated predsymbs
 ;  (shiftleft 1 27) )

(define *cl-pred-overflow-mask* 67108864); 1 for overflow
 ;  (shiftleft 1 26) )

(define *cl-predsymbs-area-mask* 67108863)
                                    ;both the negative and the positive ones
 ;  (string->number "11111111111111111111111111" 2) )

(define *cl-predsymbs-neg-shift* 13)

(define *cl-predsymbs-neg-mask* 67100672)  ; the negative predsymbs
 ; (shiftleft (string->number "1111111111111" 2) 13) )

(define *cl-predsymbs-pos-max* 4096); the leftmost pos-pred bit set
 ;  (shiftleft 1 12) )

(define *cl-predsymbs-pos-mask* 8191); the positive predsymbs
 ;  (string->number "1111111111111" 2) )

; - - - - - - - - - - - - - other - - - - - - - - - - - - -

(define *kept-cl-nr* 0) ; the number of the last kept clause
(define *last-usable-cl-nr* 0) ; the number of the last clause
                               ; in the "usable" input list - for use
                               ; with set-of-support strategy

(define (clinfo-clnumber-sublist inf)
  (cdddr (cdr inf)) )

(define (clinfo-history-sublist inf)
  (cdddr (cddr inf)) )

(define (clinfo-distance inf)
  (car (cdddr (cdddr inf))) )

(define (clinfo-clnr inf) ;simpleh
  ;(newline)(display "in clinfo-clnr:")(display inf)(newline)
  (car (cdddr (cdr inf))) )


;--------------------------------------------------------------------
;
;         the representation of litinfo of each literal
;                      in the litinfo-lst
;
;--------------------------------------------------------------------
;
; litinfolst contains info-elements for all literals in the clause.
; Each info-element is a list. The first element of the list has
; the following encoding:
;
; HIGH 30          24              16              8             1 LOW
;     | , , , , , | , , , , , , , | , , , , , , , | , , , , , , , |
;                                             |d|w|v|i|r|l|p|a|f|u|
;         |classnr|<-  ordernr  ->|
;
; iff the bit is set, the following holds:
;
; u: not usable for resolving upon
; f: not usable for factoring upon
; a: is answer-literal
; p: not usable for paramodulation into
; l: not usable for paramodulation from the left arg
; r: not usable for paramodulation from the right arg
; i: is intuitionistic input rule literal
; v: is an intuitionistic left-disjunction-rule literal
; w: is an intuitionistic right-implication literal
; d: is an intuitionistic label derived only by reduction
;
; ordernr: universally usable ordernr. For hyperresolution nuclei
;          it shows the pos (1 - ...) in the original input clause
;
; classnr: universally usable classnr. For intuitionistic logic
; it shows the type of the label.
;
;--------------------------------------------------------------------

(define *litinf-noresolve-mask* 1)
(define *litinf-nofactor-mask* 2)
(define *litinf-isanswer-mask* 4)
(define *litinf-noparainto-mask* 8)
(define *litinf-noparafromleft-mask* 16)
(define *litinf-noparafromright-mask* 32)
(define *litinf-rulelit-mask* 64)
(define *litinf-leftorrulelit-mask* 128)
(define *litinf-rightimplrulelit-mask* 256)
(define *litinf-reductionrulelit-mask* 512)

(define *litinf-nouse-mask*
  (logior *litinf-noresolve-mask*
    (logior *litinf-nofactor-mask*
      (logior *litinf-noparainto-mask*
	(logior *litinf-noparafromleft-mask*
                *litinf-noparafromright-mask*)))))

(define *litinf-nobinrespara-mask*
  (logior *litinf-noresolve-mask*
     (logior *litinf-noparainto-mask*
	(logior *litinf-noparafromleft-mask*
                *litinf-noparafromright-mask*))))

(define *litinf-ordernr-mask* (shiftleft (string->number "11111111" 2) 16) )

(define (get-litinfo-ordernr inf)
  (shiftright (logand *litinf-ordernr-mask* inf) 16) )

; put-litinfo-ordernr replaces the ordernr into litinfo

(define (put-litinfo-ordernr nr inf)
  (logior (shiftleft nr 16)
          (logxor (logand *litinf-ordernr-mask* inf) inf) )) ; remove existing

(define *litinf-classnr-mask* (shiftleft (string->number "1111" 2) 24) )

(define (get-litinfo-classnr inf)
  (shiftright (logand *litinf-classnr-mask* inf) 24) )


; put-litinfo-ordernr replaces the classnr into litinfo

(define (put-litinfo-classnr nr inf)
  (logior (shiftleft nr 24)
          (logxor (logand *litinf-classnr-mask* inf) inf) )) ; remove existing




;==================================================================
;
;              the representation of clause lists
;
;=================================================================
;
;
;
; *sos-clauses-list* is just a list of sos-clauses.
;    It is built by adding newly derived and kept clauses
;    to the end (or to the beginning) of the list.
;
; *usable-clauses-list* is just a list of usable clauses.
;    It is built by adding a newly given clause to the
;    end (or to the beginning) of the list.
;
; *subsumption-list* is meant for subsumption tests.
; Its first el is 0 and it continues with a list of sublists
; (1 cl.1 cl1.2 ...) (2 cl2.1 cl2.2) ...
; where each sublist contains clauses of the same length (which
; is the numerical first el of the list) ordered by the
; strength of the clause. There may be gaps in the sublist order.
;
; *discr-tree* is a discrimination tree for literals.
; It includes all unit-clause literals and also all ground literals
; occurring in nonunit clauses. The top level contains only
; POSITIVE predsymbs.
;
; *nonunit-discr-tree* is a discrimination tree for nonunit clauses.
; It is built upon TWO first literals of the clause treated
; as a single literal: ; 0(<lit1>,<lit2>)
;
; *discr-tree-demod-atoms* is a tree for clause demodulation
; It is built upon the selected literal of a clause and contains the whole
; clause in the form: (<clinf> <demodatom> <newatom>)
;
; structure:
;
; each node-level is built of three sublists for funs, consts and vars:
; (funs consts . vars), and in case there are no consts nor vars
; it is just (funs).
;
; level_i:  ( ((f1 level_i+1) (f2 level_i+1) ...)    funs
;             ((c1 level_i+1) (c2 level_i+1) ...)    consts
;             (x1 level_i+1) (x2 level_i+1) ... )  vars
;
;
; UNIT DISCR TREE:
; the last level on each branch is a leaf:
; (positive-lit negative-lit full-unit-cl)
; where the nonexistant ones of the first two are $f and the
; full-unit-cl may be either negative or positive or nonexistent.
;
; NONUNIT DISCR TREE:
; the last level on each branch is a leaf:
; (<full-double-cl> <full-cl1> ... <full-cl-n>)
; where <full-double-cl> is either the full existing double clause or #f:
; in the latter case there are clauses longer than two given as a list
; <full-cl1> ... <full-cl-n>
;--------------------------------------------------------------------

(define *sos-clauses-list* '())
(define *sos-clauses-list-end* '())   ; last pair of *sos-clauses-list
(define *ans-input-sos* '()) ; sos before removing answer-literals

(define *usable-clauses-list* '())
(define *usable-clauses-list-end* '()) ; last pair of *usable-clauses-list*

(define *subsumption-list* '((0)) )

(define *unit-alist* '()) ; alist of all unit clauses, with assoc-keys being
                          ; the pure predicate symbs of units.
(define *unit-alist-list* '()) ; a list of unit-alists
(define *sos-clauses-list-list* '())
(define *usable-clauses-list-list* '())
(define *all-back-subsumed-clauses* '())

(define *discr-tree* '(()) )
(define *nonunit-discr-tree* '(()) )
(define *discr-tree-list* '()) ; a list of several discr trees

(define *discr-tree-demod-terms* '(()) ) ; a special discrimination tree
      ; for demodulating terms

(define *discr-tree-demod-atoms* '(()) ) ; a special discrimination tree
      ; for demodulating atoms


;---------------------------------------------------------------------
;
;            basic predicates, selectors, constructors
;
;---------------------------------------------------------------------


(define (composite? x)
  (not (number? x)) )

(define (non-composite? x)
  (number? x) )

; -- - - - - - - -- - - - non-composite objects - - - - - - - - --

(define (var? x)
   (and (number? x)
	(<= x *maxvar*) ))

(define (non-composite:-var? x)
   (> *maxvar* x))

(define (const? x)
   (and (number? x)
	(> x *maxvar*) ))

(define (single-occ-lit-var? x) ; a single occurrence in a given lit?
  (not-zero? (logand *single-occ-lit-var-mask* x)) )

(define (repeated-occ-lit-var? x); a repeated occ in a given lit?
  (zero? (logand *single-occ-lit-var-mask* x)) )

(define (later-occurrences-var? x)
  (zero? (logand *no-occ-folllit-var-mask* x)) )

;- - - - - -- - - - - - funs, preds, terms, lits  - - - - - - - - - - -

(define (term-may-cont-vars? x)
  (zero? (logand *ground-mask* (car x))) )

(define (fun-may-cont-vars? x)
  (zero? (logand *ground-mask* x)) )

(define (fun-may-cont-repeatedvars? x)
  (not-zero? (logand *repeated-vars-mask* x)) )

(define (pure-fun x) ; with sign  ; check the signs!!
  (logand *signedfnsymb-mask* x) )

(define (pure-unsigned-fun x)
  (logand *predsymb-mask* x) )

(define (fun-depth-strength fun)
  (logand  *term-depth-strength-mask* fun) )

(define (fun-depth fun)
  (logand *term-depth-mask* fun) )

(define (fun-depth-size fun)
  (logand (logior *term-depth-mask* *term-size-mask*) fun) )

(define (fun-real-depth fun)
  (shiftright (fun-depth fun) *term-depth-shift*) )

(define (neg-fun? x)
  (not (zero? (logand *sgn-mask* x))) )

(define (pos-fun? x)
  (zero? (logand *sgn-mask* x)) )

(define (complementary-fun? x)    ; check the signs!
  (- (logand *signedfnsymb-mask* x)) )

(define (fun-can-subsume? sf gf)  ; check the signs!
  (and (eq? (pure-fun sf) (pure-fun gf))
       (>= (logand *strength-mask* sf)
	   (logand *strength-mask* gf) )
       (>= (logand *term-depth-strength-mask* sf)
	   (logand *term-depth-strength-mask* gf) )
       (>= (logand *term-size-strength-mask* sf)
	   (logand *term-size-strength-mask* gf) )
       (>= (logand *const-number-mask* sf)
	   (logand *const-number-mask* gf) )))


; - - - - - - - - - - - -  term equality preds - - - - - - - - - - - -


; ground literals are assumed to be always eq if equal

(define (ground-lit-equal? xlit ylit)
  (composite-equal? xlit ylit) )         ; NB! slow version for debugging
  ; (eq? xlit ylit) )

; check equality of arbitrary terms

(define (term-equal? xterm yterm)
  (if (eq? xterm yterm)
      #t
      (if (or (non-composite? xterm)
	      (non-composite? yterm)
	      (not (eq? (car xterm) (car yterm))) )
	  #f
	  (begin
	    (set! xterm (cdr xterm))
	    (set! yterm (cdr yterm))
	    (do ()
		((null? xterm)
		   (null? yterm) ) ; yterm is used as a succ/fail-flag
	      (let ((xel (car xterm))
		    (yel (car yterm)) )
		(if (or (eq? xel yel)
			(and (composite? xel)
			     (composite? yel)
			     (composite-equal? xel yel) ))
		    (begin (set! xterm (cdr xterm))
			   (set! yterm (cdr yterm)) )
		    (begin (set! xterm '()) ; signal end
			   (set! yterm #t) )))))))) ; signal failure

; check equality of composite terms

(define (composite-equal? xcomp ycomp)
  (if (eq? (car xcomp) (car ycomp))
      (begin
	(set! xcomp (cdr xcomp))
	(set! ycomp (cdr ycomp))
	(do ()
	    ((null? xcomp)
	       (null? ycomp) ) ; ycomp is used as a succ/fail-flag
	    (let ((xel (car xcomp))
		  (yel (car ycomp)) )
	      (if (or (eq? xel yel)
		      (and (composite? xel)
			   (composite? yel)
			   (composite-equal? xel yel) ))
		  (begin (set! xcomp (cdr xcomp))
			 (set! ycomp (cdr ycomp)) )
		  (begin (set! xcomp '()) ; signal end
			 (set! ycomp #t) )))))
      #f))


; check equality of arbitrary terms modulo var-info

(define (term-equal-strong? xterm yterm)
  (if (non-composite? xterm)
      ; xterm is non-composite
      (if (non-composite? yterm)
	  (if (non-composite:-var? xterm)
	      ; xterm is a variable
	      (eq? (logand *var-nr-mask* xterm)
		   (logand *var-nr-mask* yterm) )
	      ; xterm is a constant
	      (eq? xterm yterm) )
	  #f)
      ; xterm is composite
      (and (composite? yterm)
	   (composite-equal-strong? xterm yterm) )))


; check equality of composite terms modulo var-info (2-level deco)

(define (composite-equal-strong? xcomp ycomp)
  (if (eq? (logand *fun-except-varinf-mask* (car xcomp))
	   (logand *fun-except-varinf-mask* (car ycomp)) )
      (begin
	(set! xcomp (cdr xcomp))
	(set! ycomp (cdr ycomp))
	(do ()
	    ((null? xcomp)
	       (null? ycomp) ) ; ycomp is used as a succ/fail-flag
	    (let ((xel (car xcomp))
		  (yel (car ycomp)) )
	      (if (non-composite? xel)
		  ; xel is non-composite
		  (if (non-composite? yel)
		      (if (non-composite:-var? xel)
			  ; xel is a variable
			  (if (eq? (logand *var-nr-mask* xel)
				   (logand *var-nr-mask* yel) )
			      (begin (set! xcomp (cdr xcomp))
				     (set! ycomp (cdr ycomp)) )
			      (begin (set! xcomp '()) ; signal end
				     (set! ycomp #t) ))
			  ; xel is a constant
			  (if (eq? xel yel)
			      (begin (set! xcomp (cdr xcomp))
				     (set! ycomp (cdr ycomp)) )
			      (begin (set! xcomp '()) ; signal end
				     (set! ycomp #t) )))
		      ; yel is composite
		      (begin (set! xcomp '()) ; signal end
			     (set! ycomp #t) ))
		  ; xel is composite
		  (if (and (composite? yel)
			   (composite-equal-strong? xel yel) )
		      (begin (set! xcomp (cdr xcomp))
			     (set! ycomp (cdr ycomp)) )
		      (begin (set! xcomp '()) ; signal end
			     (set! ycomp #t) ))))))
      #f))


; check equality of term lists

(define (term-list-equal? xlst ylst)
  (do ()
      ((null? xlst)
         (null? ylst) ) ; ylst is used as a succ/fail-flag
      (if (term-equal? (car xlst) (car ylst))
	  (begin (set! xlst (cdr xlst))
		 (set! ylst (cdr ylst)) )
	  (begin (set! xlst '()) ; signal end
		 (set! ylst #t) )))) ; signal failure

; check equality of term lists, swapping first two args

(define (swapped-term-list-equal? xlst ylst)
  (and (term-equal? (car xlst) (cadr ylst))
       (term-equal? (cadr xlst) (car ylst))
       (term-list-equal? (cddr xlst) (cddr ylst)) ))


; check equality of term lists modulo var info

(define (term-list-equal-strong? xlst ylst)
  (do ()
      ((null? xlst)
         (null? ylst) ) ; ylst is used as a succ/fail-flag
      (if (term-equal-strong? (car xlst) (car ylst))
	  (begin (set! xlst (cdr xlst))
		 (set! ylst (cdr ylst)) )
	  (begin (set! xlst '()) ; signal end
		 (set! ylst #t) )))) ; signal failure

; check equality of term lists modulo var-info, swapping first two args

(define (swapped-term-list-equal-strong? xlst ylst)
  (and (term-equal-strong? (car xlst) (cadr ylst))
       (term-equal-strong? (cadr xlst) (car ylst))
       (term-list-equal-strong? (cddr xlst) (cddr ylst)) ))

; - - - - - - - - - - - - clauses - - - - - - - - - - - - -- - - -

(define (clause-can-subsume? sclstrength sclpreds gclinfo)
 (let ((gclstrength (car gclinfo)))
    (if (and (>= (logand *cl-strength-mask* sclstrength)
		 (logand *cl-strength-mask* gclstrength) )
	     (>= (logand *strength-mask* sclstrength)
		 (logand *strength-mask* gclstrength) )
	     (>= (logand *term-depth-strength-mask* sclstrength)
		 (logand *term-depth-strength-mask* gclstrength) )
	     (>= (logand *term-size-strength-mask* sclstrength)
		 (logand *term-size-strength-mask* gclstrength)	)
	     (>= (logand *const-number-mask* sclstrength)
		 (logand *const-number-mask* gclstrength) ))
	(let ((gclpreds (cadr gclinfo)))
	  (eq? gclpreds (logand gclpreds sclpreds)) )
	#f)))

; LAST: are all predsymbs encoded in the predsymbs-int of a general
; clause present in the predsymbs-int of a subsumable clause? (i.e.
; is the gpredint subset of spredint?)

(define (pure-clause-may-cont-vars? cl)
  (let ((res #f))
    (do ()
	((or (null? cl) res)
	   res)
      (if (term-may-cont-vars? (car cl))
	  (set! res #t)
	  (set! cl (cdr cl)) ))))

;----------------------------------------------------------------
;
;                 setting and referencing variables
;
;-----------------------------------------------------------------

; all the vars given to the following functions (except var-nr)
; MUST be pure var-nrs without info bits (but y-bit is OK, of course,
; as it is a part of the var-nr)

(define (var-nr var)
  (logand *var-nr-mask* var) )

(define (y-var? var)
  (not (zero? (logand *y-var-mask* var))) )

; x-var? assumes that the var is NOT a z-var.

(define (x-var? var)
  (zero? (logand *y-var-mask* var)) )

(define (unassigned-val? x)
  (eq? x *unassigned*) )

(define (touched-var? var)
  (vector-ref *var-touched-table* var) )

(define (set-var-touched! var val)
  (vector-set! *var-touched-table* var val) )

(define (var-ref var)
  (vector-ref *var-val-table* var) )

(define (set-var-ref! var val)
  (vector-set! *var-val-table* var val) )

; var-val returns the last var in the chain of references or
; the first non-var object.

(define (var-val var)
   (do ((var1 var var2)
        (var2 (var-ref (var-nr var)) (var-ref (var-nr var2))) )
       ((not (var? var2))
	  (if (unassigned-val? var2) var1 var2) )))

; dereference-var sets var to refer to full-built first-level
; decorated value and returns that value. Any vars encountered
; while building a decorated value are also dereferenced.
; In case a touched var is encountered, the nonground term
; structure of its value (immediate reference) is copied to
; avoid structure sharing between nonground terms.

(define (dereference-var var)
  (let ((ref (var-ref var)))
    (if (unassigned-val? ref)
	var
	(if (touched-var? var)
	    (begin
	      (if (non-composite? ref)
		  ref ; reference
		  (dereference-copy-term ref) ))
	    (begin
	      (set-var-touched! var #t)
	      (set! ref (subst-deco-term ref #f))
	      (set-var-ref! var ref)
	      ref)))))

;(define (dereference-var var)
;  (if (touched-var? var)
;      (begin
;	(set! var (var-ref var))
;	(if (non-composite? var)
;	    var ; reference
;	    (dereference-copy-term var) ))
;      (let ((ref (var-ref var)))
;	(set-var-touched! var #t)
;	(if (unassigned-val? ref)
;	    (begin (set-var-ref! var var) var)
;	    (begin (set! ref (subst-deco-term ref #f))
;		   (set-var-ref! var ref)
;		   ref)))))

; term must be a composite

(define (dereference-copy-term term)
  (let ((fun (car term)))
    (if (fun-may-cont-vars? fun)
	(cons fun (map dereference-copy-aux (cdr term)))
	term)))

; term is arbitrary

(define (dereference-copy-aux term)
  (if (non-composite? term)
      term
      (let ((fun (car term)))
	(if (fun-may-cont-vars? fun)
	    (cons fun (map dereference-copy-aux (cdr term)))
	    term))))

(define (clear-x-vars! var)
  (do ()
      ((zero? var))
    (set-var-ref! var *unassigned*)
    (set! var (- var 1)) ))

(define (clear-x-vars-and-touched! var)
  (do ()
      ((zero? var))
    (set-var-ref! var *unassigned*)
    (set-var-touched! var #f)
    (set! var (- var 1)) ))

(define (clear-y-vars! var) ; var is assumed to be an ordinary xvar
   (set! var (logior *y-var-mask* var)) ; it is first made an yvar
   (do ()
      ((eq? *max-x-var* var)) ; a zero for yvars
     (set-var-ref! var *unassigned*)
     (set! var (- var 1)) ))

(define (clear-y-vars-and-touched! var) ; var is assumed to be an ordinary xvar
   (set! var (logior *y-var-mask* var)) ; it is first made an yvar
   (do ()
      ((eq? *max-x-var* var)) ; a zero for yvars
     (set-var-ref! var *unassigned*)
     (set-var-touched! var #f)
     (set! var (- var 1)) ))

(define (clear-z-vars! var) ; var is assumed to be an ordinary xvar
   (set! var (logior *z-var-mask* var)) ; it is first made a zvar
   (do ()
      ((eq? *max-y-var* var)) ; a zero for zvars
     (set-var-ref! var *unassigned*)
     (set! var (- var 1)) ))

(define (clear-z-vars-and-touched! var) ; var is assumed to be an ordinary xvar
   (set! var (logior *z-var-mask* var)) ; it is first made a zvar
   (do ()
      ((eq? *max-y-var* var)) ; a zero for zvars
     (set-var-ref! var *unassigned*)
     (set-var-touched! var #f)
     (set! var (- var 1)) ))

; sets xvars to refer to themselves and sets all of them to be touched

(define (set-x-vars-selfref-touched! var)
  (do ()
      ((zero? var))
    (set-var-ref! var var)
    (set-var-touched! var #t)
    (set! var (- var 1)) ))

; the following three defs are used for the clause subsumption.
; *gcl-maxvar* is local to a current gcl (its maxvar).
; *gcl-varsstack* is local to a current gcl. It is a list (stack) of
; lists of gcl variable values, each list (stack frame) representing a state
; of values at some certain call.

(define (clear-gcl-vars)
  (clear-x-vars! *gcl-maxvar*) )

(define (restore-gcl-vars)
  (do ((var 1 (+ 1 var))
       (frame (car *gcl-varsstack*) (cdr frame)) )
      ((null? frame)
         (set! *gcl-varsstack* (cdr *gcl-varsstack*)) ) ; pop off frame
    (set-var-ref! var (car frame)) ))

(define (save-gcl-vars)
  (let ((res '()))
    (do ((var *gcl-maxvar* (- var 1)))
	((zero? var)
	   (set! *gcl-varsstack* (cons res *gcl-varsstack*)) ) ; push frame
      (set! res (cons (var-ref var) res)) )))

(define (pop-gcl-vars-stack)
  (set! *gcl-varsstack* (cdr *gcl-varsstack*)) ) ; pop off frame

;-------------------------------------------------------------------
;                       substituting
;                           and
;     decorating terms, lits and clauses with extra info
;
;------------------------------------------------------------------


(define (copy-nonground-term term)
  (if (pair? term)
      (if (fun-may-cont-vars? (car term))
	  (cons (car term) (map copy-nonground-term (cdr term)))
	  term)
      term))

; rename-x-to-y-term takes a term and builds a new
; one where x-vars are replaced by y-vars.

(define (rename-x-to-y-term term)
  (if (non-composite? term)
      (if (non-composite:-var? term)
	  (logior *y-var-mask* term)
          term)
      (let ((fun (car term)))
	(if (fun-may-cont-vars? fun)
	    (cons fun (map rename-x-to-y-term (cdr term)))
	    term))))


; rename-y-to-x-term takes a term and builds a new
; one where y-vars are replaced by x-vars. All vars MUST previously
; be y-vars.

(define (rename-y-to-x-term term)
  (if (non-composite? term)
      (if (non-composite:-var? term)
	  (logxor *y-var-mask* term)
          term)
      (let ((fun (car term)))
	(if (fun-may-cont-vars? fun)
	    (cons fun (map rename-y-to-x-term (cdr term)))
	    term))))

; rename-x-to-z-term takes a term and builds a new
; one where x-vars are replaced by z-vars.

(define (rename-x-to-z-term term)
  (if (non-composite? term)
      (if (non-composite:-var? term)
	  (logior *z-var-mask* term)
          term)
      (let ((fun (car term)))
	(if (fun-may-cont-vars? fun)
	    (cons fun (map rename-x-to-z-term (cdr term)))
	    term))))

; rename-x-to-y-term! takes a COMPOSITE term and destructively makes a new
; one where x-vars are replaced by y-vars.

(define (rename-x-to-y-term! cterm)
  (if (fun-may-cont-vars? (car cterm))
      (do ((part (cdr cterm) (cdr part)))
	  ((null? part))
	(let ((el (car part)))
	  (if (composite? el)
	      (rename-x-to-y-term! el)
	      (if (non-composite:-var? el)
		  (set-car! part (logior *y-var-mask* el)) ))))))

; rename-y-to-x-term! takes a COMPOSITE term and destructively makes a new
; one where y-vars are replaced by x-vars.
;
; NB! All the vars in the original must be y-vars.

(define (rename-y-to-x-term! cterm)
  (if (fun-may-cont-vars? (car cterm))
      (do ((part (cdr cterm) (cdr part)))
	  ((null? part))
	(let ((el (car part)))
	  (if (composite? el)
	      (rename-y-to-x-term! el)
	      (if (non-composite:-var? el)
		  (set-car! part (logxor *y-var-mask* el)) ))))))


; There are two levels of info-decoration: first is the level
; where the decoration of vars and the "maxvr" and "r" info
; of terms is not performed: the only-var-oriented deco is "g".
;
; The first-level decoration is useful for newly generated clauses
; before forward-subsumption check, as the properties of vars
; are not important for subsumable clauses. The first level
; works along with substitution.
;
; The second level is full decoration.
;-----------------------------------------------------------------

; subst-deco-term takes a term and uses the standard values-vector
; *var-val-table* to set the values for vars. It performs
; first-level decoration while building a new term.
;
; It also demodulates the term, in case *demodulate-flag* is set.
;
; It uses the flag *demodulate-constants-flag* to check whether
; constants should be demodulated.
; It uses the flag *new-demodulators-flag* to check whether nonchanged
; sould be also demodulated.
;
;
; IMPORTANT: In case *demodulate-flag* = #f, all the other
; demodulation flags are also assumed to be #f.




(define (subst-deco-term term litflag)
  ; (newline)(display "subst-deco-term started: ")
  ; (display term)(display litflag)(newline)
  (if (non-composite? term)
      (if (non-composite:-var? term)
	  ; a var
	  (dereference-var (var-nr term))
          ; a constant
          (if *demodulate-constants-flag*
              (demodulate-constant term)
	      term))
      ; a composite
      (let ((fun (car term)))
	(if (or *new-demodulators-flag*
	        (fun-may-cont-vars? fun) )
	    ; a nonground composite or new demodulators
	    (if (eq? *case-fun-nr* (pure-fun fun))
		(subst-deco-case-term term)
		(let* ((res (cons '() '()))
		       (resend res)
		       (str *ground-mask*) ) ; no args is just ground
		  (do ((part (cdr term) (cdr part)))
		      ((null? part)

		        ;(newline)(display "subst-deco-term ending: ")
		        ;(display res)(newline)

		       (set-car! res (create-new-fun-pos-object fun str))
		       ; args of res are already demodulated (if possible).
		       (cond ((and *demodulate-flag* (not litflag))
			      (demodulate-composite res) )
			     ((and litflag *demodulate-lit-flag*)
			      (demodulate-lit res) )
			     (else
			      res)))
		    (let ((new (subst-deco-term (car part) #f)))
		      (set-cdr! resend (cons new '()))
		      (if (eq? resend res) ; the first arg at hand?
			  ; now comes the case for the first arg
			  (if (non-composite? new)
			      (if (non-composite:-var? new)
				  (set! str *var-strength*)
				  (set! str *const-strength*) )
			      (set! str (car new)) )
			  ; now comes the case for the following args
			  (if (non-composite? new)
			      (if (non-composite:-var? new)
				  (set! str
				    (sum-fun-strengths *var-strength* str) )
				  (set! str
				    (sum-fun-strengths *const-strength* str) ))
			      (set! str
				    (sum-fun-strengths (car new) str) )))
		      (set! resend (cdr resend)) ))))
	    ; a ground composite, no new demodulators
	    term))))


; deco-leading-fun takes a term, assumes args are 1-st level decorated
; and builds a term with the leading fun correctly decorated.
; term must be a composite.

(define (deco-leading-fun term)
  (let* ((fun (car term))
	 (new '())
	 (str *ground-mask*) ) ; no args is just ground
    (do ((part (cdr term) (cdr part)))
	((null? part)
	 (cons (create-new-fun-pos-object fun str) (cdr term)) )
      (set! new (car part))
      (if (eq? part (cdr term)) ; the first arg at hand?
	  ; now comes the case for the first arg
	  (if (non-composite? new)
	      (if (non-composite:-var? new)
		  (set! str *var-strength*)
		  (set! str *const-strength*) )
	      (set! str (car new)) )
	  ; now comes the case for the following args
	  (if (non-composite? new)
	      (if (non-composite:-var? new)
		  (set! str (sum-fun-strengths *var-strength* str))
		  (set! str (sum-fun-strengths *const-strength* str)) )
	      (set! str (sum-fun-strengths (car new) str)) )))))


(define (subst-deco-term-nodemodulate term litflag)
  (let ((tmp1 *demodulate-flag*)
	(tmp2 *demodulate-lit-flag*)
	(tmp3 *demodulate-constants-flag*)
	(tmp4 *new-demodulators-flag*)
	(res '()) )
    (set! *demodulate-flag* #f)
    (set! *demodulate-lit-flag* #f)
    (set! *demodulate-constants-flag* #f)
    (set! *new-demodulators-flag* #f)
    (set! res (subst-deco-term term litflag))
    (set! *demodulate-flag* tmp1)
    (set! *demodulate-lit-flag* tmp2)
    (set! *demodulate-constants-flag* tmp3)
    (set! *new-demodulators-flag* tmp4)
    res))


; subst-deco-case-term takes a term lead by the case function
; and substitutes, decorates and computes the term.
;
; (case <term> <caselist>) where <caselist> is built by an arbitray ternary
; function, eg (c <case1-head> <case1-res>
;                  (c <case2-head> <case2-res> <arbitrary-noncomposite>))
; for two cases.
;
; If the value of case term is matchable by the value of any case-head,
; then the corresponding case-res is chosen.

(define (subst-deco-case-term caseterm)
  (let ((value (subst-deco-term (cadr caseterm) #f))
	(successres #f)
	(tmp '()) )
    (do ((caselist (caddr caseterm) (cadddr caselist)))
	((or (not (composite? caselist))
	     successres)
	  (if successres
	      successres
	      (deco-leading-fun
	       (list *case-fun-nr*
		     value
		     (subst-deco-term (caddr caseterm) #f) ))))
      (set! tmp (subst-deco-term (cadr caselist) #f))
      (set! tmp (subsumed-term-ordinary-collectvars? value tmp))
      (if (car tmp)
	  (begin (set! successres (subst-deco-term (caddr caselist) #f))
		 (clear-varlist (cdr tmp)) )
	  (clear-varlist (cdr tmp)) ))))

(define (clear-varlist lst)
  (do ((lstpart lst (cdr lstpart)))
      ((null? lstpart))
    (set-var-ref! (var-nr (car lstpart)) *unassigned*) ))

; the following function substitutes into term vars
; nd does NOT do any demodulation. It also assumes that ALL vars in the
; term are set.


(define (subst-deco-term-nodemodulate-spec term)
  (if (non-composite? term)
      (if (non-composite:-var? term)
          (dereference-var (var-nr term))
          term)
      ; a composite
      (let ((fun (car term)))
	(if (fun-may-cont-vars? fun)
	    ; a nonground composite
	    (let* ((res (cons '() '()))
		   (resend res)
		   (str *ground-mask*) ) ; no args is just ground
	      (do ((part (cdr term) (cdr part)))
		  ((null? part)
		    (set-car! res (create-new-fun-pos-object fun str))
                    res)
		  (let ((new (subst-deco-term-nodemodulate-spec
                                (car part) )))
		    (set-cdr! resend (cons new '()))
		    (if (eq? resend res) ; the first arg at hand?
			; now comes the case for the first arg
			(if (non-composite? new)
			    (if (non-composite:-var? new)
				(set! str *var-strength*)
				(set! str *const-strength*) )
			    (set! str (car new)) )
			; now comes the case for the following args
			(if (non-composite? new)
			    (if (non-composite:-var? new)
				(set! str
				  (sum-fun-strengths *var-strength* str) )
				(set! str
				  (sum-fun-strengths *const-strength* str) ))
			    (set! str
			      (sum-fun-strengths (car new) str) )))
		    (set! resend (cdr resend)) )))
	    ; a ground composite
            term))))


; subst-deco-replace-term takes a term, a path, a replaceterm.
; It replaces the subterm at path and uses the standard values-vector
; *var-val-table* to set the values for vars. It performs
; first-level decoration while building a new term.
;
; It also demodulates the term, in case *demodulate-flag* is set.
;
; It uses the flag *demodulate-constants-flag* to check whether
; constants should be demodulated.
; It uses the flag *new-demodulators-flag* to check whether nonchanged
; sould be also demodulated.
;
; IMPORTANT: In case *demodulate-flag* = #f, all the other
; demodulation flags are also assumed to be #f.


(define (subst-deco-replace-term term path replterm litflag)
  (cond
   ((null? path)
      (subst-deco-term replterm litflag) )
   ((non-composite? term)
      (if (non-composite:-var? term)
	  ; a var
	  (dereference-var (var-nr term))
          ; a constant
          (if *demodulate-constants-flag*
              (demodulate-constant term)
	      term)))
   ; a composite
   (else
      (let* ((fun (car term))
	     (nr (car path))
	     (res (cons '() '()))
	     (resend res)
	     (new '())
	     (str *ground-mask*) ) ; no args is just ground
      (do ((part (cdr term) (cdr part))
	   (nrpart (- nr 1) (- nrpart 1)) )
	  ((null? part)
	   (set-car! res (create-new-fun-pos-object fun str))
	   ; args of res are already demodulated (if possible).
	   (cond ((and *demodulate-flag* (not litflag))
		     (demodulate-composite res) )
		 ((and litflag *demodulate-lit-flag*)
		     (demodulate-lit res) )
		 (else
		     res)))
	(if (zero? nrpart)
	    (set! new (subst-deco-replace-term
		       (car part) (cdr path) replterm #f))
	    (set! new (subst-deco-term (car part) #f)) )
	(set-cdr! resend (cons new '()))
	(if (eq? resend res) ; the first arg at hand?
	    ; now comes the case for the first arg
	    (if (non-composite? new)
		(if (non-composite:-var? new)
		    (set! str *var-strength*)
		    (set! str *const-strength*) )
		(set! str (car new)) )
	    ; now comes the case for the following args
	    (if (non-composite? new)
		(if (non-composite:-var? new)
		    (set! str (sum-fun-strengths *var-strength* str) )
		    (set! str (sum-fun-strengths *const-strength* str) ))
		(set! str (sum-fun-strengths (car new) str) )))
	(set! resend (cdr resend)) )))))


(define (subst-deco-replace-term-nodemodulate term path replterm litflag)
  (let ((tmp1 *demodulate-flag*)
	(tmp2 *demodulate-lit-flag*)
	(tmp3 *demodulate-constants-flag*)
	(tmp4 *new-demodulators-flag*)
	(res '()) )
    (set! *demodulate-flag* #f)
    (set! *demodulate-lit-flag* #f)
    (set! *demodulate-constants-flag* #f)
    (set! *new-demodulators-flag* #f)
    (set! res (subst-deco-replace-term term path replterm litflag))
    (set! *demodulate-flag* tmp1)
    (set! *demodulate-lit-flag* tmp2)
    (set! *demodulate-constants-flag* tmp3)
    (set! *new-demodulators-flag* tmp4)
    res))

; replace-subterm takes a term, a path in the term and a new subterm
; to be put to the place indicated by the path.
; A path is a list of nrs, (n1, ... nk) where each ni indicates the
; following cxr operation: ca(d1)...(dni)r)
;
; No substitution or demodulation is performed.

; A path is a list of nrs, (n1, ... nk) where each ni indicates the
; following cxr operation: ca(d1)...(dni)r)
;
; It builds a new term with newsub put to the right place.

(define (replace-subterm term path replterm)
  (cond
   ((null? path) replterm)
   ((non-composite? term) term)
   ; a composite
   (else
      (let* ((fun (car term))
	     (nr (car path))
	     (res (cons '() '()))
	     (resend res)
	     (new '())
	     (str *ground-mask*) ) ; no args is just ground
      (do ((part (cdr term) (cdr part))
	   (nrpart (- nr 1) (- nrpart 1)) )
	  ((null? part)
	    (set-car! res (create-new-fun-pos-object fun str))
            res)
	(if (zero? nrpart)
	    (set! new (replace-subterm (car part) (cdr path) replterm))
            (set! new (car part)) )
	(set-cdr! resend (cons new '()))
	(if (eq? resend res) ; the first arg at hand?
	    ; now comes the case for the first arg
	    (if (non-composite? new)
		(if (non-composite:-var? new)
		    (set! str *var-strength*)
		    (set! str *const-strength*) )
		(set! str (car new)) )
	    ; now comes the case for the following args
	    (if (non-composite? new)
		(if (non-composite:-var? new)
		    (set! str (sum-fun-strengths *var-strength* str))
		    (set! str (sum-fun-strengths *const-strength* str)) )
		(set! str (sum-fun-strengths (car new) str)) ))
	(set! resend (cdr resend)) )))))


; subterm-at-path gives a subterm at path

(define (subterm-at-path term path)
  (if (null? path)
      term
      (do ((termpart term (cdr termpart))
	   (nrpart (car path) (- nrpart 1)) )
	  ((zero? nrpart)
	       (subterm-at-path (car termpart) (cdr path)) ))))


; subst-deco-weight-term takes a term and a list of relevant
; funsymbs in the scope of which the term lies and
; uses the standard values-vector
; *var-val-table* to set the values for vars. It performs
; first-level decoration while building a new term and it computes
; the weight of the term. It returns a pair (weight . term).
; If the weight-computation finds a prohibited weight,
; the substitution process is terminated and #f is returned.

(define (subst-deco-weight-term term depth scope)
  (if (non-composite? term)
      (let* ((res (if (non-composite:-var? term)
		     (dereference-var (var-nr term))
		     term) )
	     (weight (compute-term-weight res depth scope)) )
	(if (integer? weight) ; no immediate failure?
	    (cons weight res)
	    #f))
      (let ((fun (car term))
	    (failflag #f) )
	(if (fun-may-cont-vars? fun)
	    (let* ((res (cons '() '()))
		   (resend res)
		   (part (cdr term))
		   (weight 0)
		   (str *ground-mask*) ) ; no args is just ground
	      (do ()
		  ((null? part)
		    (if failflag
			#f
			(begin
			   (set-car! res
				     (create-new-fun-pos-object fun str) )
			   (cons weight res) )))
		(let* ((newpair (subst-deco-weight-term (car part)))
		       (new (cdr newpair))
		       (weight (car newpair)) )
		  (if (integer? weight) ; no immediate failure?
		      ; no immediate failure
		      (begin
			(set-cdr! resend (cons new '()))
			(if (eq? resend res) ; the first arg at hand?
			    ; now comes the case for the first arg
			    (if (non-composite? new)
				(if (non-composite:-var? new)
				    (set! str *var-strength*)
				    (set! str *const-strength*) )
				(set! str (car new)) )
			    ; now comes the case for the following args
			    (if (non-composite? new)
				(if (non-composite:-var? new)
				    (set! str
					  (sum-fun-strengths
					     *var-strength* str) )
				    (set! str
					  (sum-fun-strengths
					     *const-strength* str) ))
				(set! str
				      (sum-fun-strengths (car new) str) )))
			(set! resend (cdr resend))
			(set! part (cdr part)) )
		      ; immediate failure
		      (begin
			(set! part '())
			(set! failflag #t)
		fun)))))))))


(define (compute-term-weight x y z)
   0)

;sum-fun-strengths takes two fun/pred-pos objs or strengths
; (eg var or constant strengths) and composes and adds them
; assuming that they become the args of the same fun term.

(define (sum-fun-strengths fnx fny)
  (logior
     (logand (logand *ground-mask* fnx) (logand *ground-mask* fny))
     (logior
        (mymax
	   (logand *term-depth-mask* fnx)
	   (logand *term-depth-mask* fny) )
        (logior
	   (maxplus *term-size-max-shift*
		    (logand *term-size-strength-mask* fnx)
		    (logand *term-size-strength-mask* fny) )
	   (maxplus *const-number-max*
		    (logand *const-number-mask* fnx)
		    (logand *const-number-mask* fny) )))))

;create-new-fun-pos-object takes a function symb (strength
;unimportant) and the composed/summed strength of its args,
;giving back the fun symb with the right strength (that is,
;argsstrn with depth and size both increased by one.

(define (create-new-fun-pos-object oldfun argsstrn)
  (logior
     (logand *signedfnsymb-mask* oldfun)
     (logior
        (logior
	   (maxplus *term-depth-max-shift*
		    *term-depth-1-shift*
		    (logand *term-depth-mask* argsstrn) )
	   (maxplus *term-size-max-shift*
		    *term-size-1-shift*
		    (logand *term-size-mask* argsstrn) ))
	(logand *g-and-const-mask* argsstrn) )))


; - - - - - - - the second level: full decoration - - - - -


; collect-vars-composite! collects all repeating vars and
; all vars from a composite term and decorates each non-ground
; subterm with a *repeated-vars* flag.
;
; for each composite term it returns a pair (<repeating-vars> . <all-vars>),
;  where all vars are undecorated.
;
; composite is assumed to be a nonground composite

(define (collect-vars-composite! composite)
  (let ((fun (car composite))
	(repeating '())
	(vars '()) )
    (do ((comppart (cdr composite) (cdr comppart)))
	((null? comppart)
	  (if (not (null? repeating))
	      (set-car! composite (logior *repeated-vars-mask* fun)) )
	  (cons repeating vars) )
      (let ((el (car comppart)))
	(if (non-composite? el)
	    (if (non-composite:-var? el)
		(begin
		  (set! el (logand *var-nr-mask* el))
		  (if (not (memq el repeating))
		      (if (memq el vars)
			  (set! repeating (cons el repeating))
			  (set! vars (cons el vars)) ))))
	    (if (term-may-cont-vars? el)
		(let ((elvars (collect-vars-composite! el)))
		  (set! vars (intersection-union
			        (car elvars) (cdr elvars) repeating vars))
		  (set! repeating *intersection-tmp*) )))))))


; intersection-union gives a union of xvars and yvars and
; an enriched union (from xvars, yvars) of xrep and yrep.
; the result is a pair (<repeating-vars> . <all-vars>)
;
; it uses a global variable *intersection-tmp*

(define *intersection-tmp* '())

(define (intersection-union xrep xvars yrep yvars)
  (if (null? yvars)
      (begin (set! *intersection-tmp* xrep) xvars) ; just an optimization
      (let ((tmprep (union xrep yrep))) ; union of given repeating-sets
	(set! *intersection-tmp* tmprep)
	(intersection-union-aux xvars yvars) )))


(define (intersection-union-aux xvars yvars)
  (cond ((null? xvars) yvars)
	((memq (car xvars) yvars)
	   (if (not (memq (car xvars) *intersection-tmp*))
	       (set! *intersection-tmp*
		     (cons (car xvars) *intersection-tmp*) ))
	   (intersection-union-aux (cdr xvars) yvars) )
	(else
           (cons (car xvars) (intersection-union-aux (cdr xvars) yvars)) )))


; decorate-vars-lit! takes a non-ground literal,
; renumbers all vars in a lit
; and decorates them with var information.
;
; it uses the following lists (containing OLD undecorated numbers):
; rep-lit-vars : repeating vars in the given literal
; lit-vars: all vars in the given literal
; other-lit-vars : list of var lists of the other literals
; foll-lit-vars : list of var lists of the following literals
;
; it builds and uses the following renumbering alist:
; *new-var-names* : the list of pairs ((oldvar . newvar) ...)
;   where newvar is DECORATED for the current lit and
;   oldvar is assumed to be undecorated.
; the following var is a current nr of a var in the clause:
; *clause-var-nr*. It should initially be 0.
; if too many vars occur in the clause, it sets the
; *too-many-vars-flag* to #t.

(define (decorate-vars-lit!
	  lit rep-lit-vars lit-vars other-lit-vars foll-lit-vars)
  (if (not (null? lit-vars))
      ; the coming "reverse" is just for beauty: for the vars
      ; of the lit to start with x,y,.. from the BEGINNING of the lit
      (do ((varspart (reverse lit-vars) (cdr varspart))) ;;;; ?!?!? TMP !?!?!
	  ((null? varspart)
	      (decorate-vars-composite! lit *new-var-names*) )
	(let* ((el (car varspart))
	       (renamed (assq el *new-var-names*))
	       (new '()) )
	  (if renamed
	      (set! new (logand *var-nr-mask* (cdr renamed)))
	      (begin
		(set! *clause-var-nr* (+ 1 *clause-var-nr*))
		(if (> *clause-var-nr*
		       *max-nr-of-vars-in-clause*)
		    (set! *too-many-vars-flag* #t) )
		(set! new *clause-var-nr*) ))
	  (if (not (memq el rep-lit-vars))
	      (set! new
		    (logior *single-occ-lit-var-mask* new) ))
	  (if (not (member-memq? el other-lit-vars))
	      (set! new
		    (logior *no-occ-otherlit-var-mask* new) ))
	  (if (not (member-memq? el foll-lit-vars))
	      (set! new
		    (logior *no-occ-folllit-var-mask* new) ))
	  (if renamed
	      (set-cdr! renamed new)
	      (set! *new-var-names*
		    (cons (cons el new) *new-var-names*) ))))))


(define (decorate-vars-composite! composite alist)

  ;(newline)(display "decorate-vars-composite! entered with: ")
  ;(newline)(display composite)(newline)(display alist)(newline)

  (do ((comppart (cdr composite) (cdr comppart)))
      ((null? comppart))
    (let ((el (car comppart)))
      (if (non-composite? el)
	  (if (non-composite:-var? el)
	      (set-car! comppart
			(cdr (assq (logand *var-nr-mask* el) alist)) ))
	  (if (term-may-cont-vars? el)
	      (decorate-vars-composite! el alist) )))))


;-------------------------------------------------------------------
;
;                      unification algorithms
;
;-------------------------------------------------------------------
;
; methods of avoiding an occur check during unification for resolution:
;
; 1. Initially the flag *repeatedvars-flag* is set to the
;    AND of these values encoded in the pred symb position of
;    each literal.
;
; 2. Initially the *occurcheck-flag* is set to #f.
;
;    It is set to #t if an assignment x:=t[y] occurs
;    where t[y] is not ground and x has other occurrences
;    in the given literal (that is: (single-occ-lit-var?  x)  = #t or
;    (repeated-occ-lit-var? x) = #f) and *repeatedvars-flag* = #t.
;
; 3. The occur check for some x:t[y] pair is taken iff all foll. holds:
;    A) *occurcheck-flag* = #t
;    B) t[y] is not ground
;
; methods of avoiding an occur check during unification for factorization:
;
; x:t[y] occur check is performed iff t[y] may contain vars.
;
;--------------------------------------------------------------------
;
; Assumptions:
;
; in the scope of one literal all vars with the same var-nr
; are assumed to be eq (that is, have the same info bits).
;
; resolve-unify-ordinary assumes that xfun and yfun are
; complementary, that is:
; (complementary-fun? (pure-fun xfun) (pure-fun yfun)) gives #t
;
; resolve-unify-ordinary assumes that cdr-s of the complementarily
; equal ground literals are eq.
;
;--------------------------------------------------------------------
;
; side effects:
; the global vars *xvars-set-flag* and *yvars-set-flag*
; are set to #t or #f to indicate whether any vars were set.
;
; Those variables are cleared by resolve-unify-ordinary, thus they
; may have whatever value before entering
;-------------------------------------------------------------------


(define (resolve-unify-ordinary xfun yfun xlit ylit)
  ;(newline)
  ;(display "resolve-unify-ordinary: ")(newline)
  ;(display "xfun: ")(display xfun)(newline)
  ;(display "yfun: ")(display yfun)(newline)
  ;(display "xlit: ")(display xlit)(newline)
  ;(display "ylit: ")(display ylit)(newline)
  (if (fun-may-cont-vars? xfun)
      (if (fun-may-cont-vars? yfun)
	  ; both xlit and ylit are nonground
	  (let ((fail-flag #f))
	    (set! *repeatedvars-flag*
		  (and (fun-may-cont-repeatedvars? xfun)
		       (fun-may-cont-repeatedvars? yfun) ))
	    (set! *occurcheck-flag* #f)
	    (set! *xvars-set-flag* #f)
	    (set! *yvars-set-flag* #f)
	    (do ((xpart (cdr xlit) (cdr xpart))
		 (ypart (cdr ylit) (cdr ypart)) )
		((or fail-flag (null? xpart))
		   (not fail-flag) )
	      (if (not (resolve-unify-ordinary-aux
			(car xpart) (car ypart)))
		  (set! fail-flag #t) )))
	  ; now xlit is nonground and ylit is ground
	  (begin
	    (set! *yvars-set-flag* #f)
	    (if (> (fun-depth-strength xfun)
		   (fun-depth-strength yfun) )
		(begin (set! *xvars-set-flag* #f) #f) ;; immediate failure
		(begin (set! *var-set-flag* #f) ;; blocked for non-follow vars
		       (set! *alt-var-set-flag* #f) ;; alternative
		       (if (subsumed-term-list-ordinary?
			    (cdr ylit) (cdr xlit))
			   (begin (set! *xvars-set-flag* #t) #t) ;;; success
			   ;;; failure
			   (begin (if (or *var-set-flag*
					  *alt-var-set-flag*)
				      (set! *xvars-set-flag* #t)
				      (set! *xvars-set-flag* #f) )
				  #f))))))
      ; now xlit is ground
      (if (fun-may-cont-vars? yfun)
	  ; now xlit is ground and ylit is nonground
	  (begin
	    (set! *xvars-set-flag* #f)
	    (if (> (fun-depth-strength yfun)
		   (fun-depth-strength xfun) )
		(begin (set! *yvars-set-flag* #f) #f) ;; immediate failure
		(begin (set! *var-set-flag* #f) ;; blocked for non-follow vars
		       (set! *alt-var-set-flag* #f) ;; alternative
		       (if (subsumed-term-list-ordinary?
			    (cdr xlit) (cdr ylit))
			   (begin (set! *yvars-set-flag* #t) #t) ;;; success
			   ;;; failure
			   (begin (if (or *var-set-flag*
					  *alt-var-set-flag*)
				      (set! *yvars-set-flag* #t)
				      (set! *yvars-set-flag* #f) )
				  #f)))))
	  ; now both xlit and ylit are ground
	  (begin
	    (set! *xvars-set-flag* #f)
	    (set! *yvars-set-flag* #f)
	    (or (eq? (cdr xlit) (cdr ylit))
		(and *compare-ground-litcdr-equal-flag*
		     (eq? (fun-depth-strength xfun)
			  (fun-depth-strength yfun) )
		     (term-list-equal? (cdr xlit) (cdr ylit)) ))))))


; resolve-unify-continue is used for intuitionistic logic, for
; the case when several literals in xclause have to be unified
; with several literals in yclause. In that case all but the first
; unification are performed by the resolve-unify-continue.

;(define (resolve-unify-continue xfun yfun xlit ylit)
;  (set! *repeatedvars-flag* #t) )

; resolve-unify-cutoff is similar to resolve-unify-ordinary.
;
; The principal difference is that ONLY these variables
; in ylit which have no occurrences in other literals
; (ie have the bit with mask *no-occ-otherlit-var-mask* set)
; are allowed to be set.
; NB! ylit is always assumed to be nonground!
; NB! given xlit may be ground only if all the vars in ylit are local-only;
; otherwise it is WRONG to give a ground xlit as argument.

(define (resolve-unify-cutoff xfun yfun xlit ylit)

  ;(newline)(display "resolve-unify-cutoff called with:")(newline)
  ;(debug-print-lit xlit)(newline)
  ;(debug-print-lit ylit)(newline)

  (if (fun-may-cont-vars? xfun)
      ; both xlit and ylit are nonground
      (let ((fail-flag #f))
	(set! *repeatedvars-flag*
	      (and (fun-may-cont-repeatedvars? xfun)
		   (fun-may-cont-repeatedvars? yfun) ))
	(set! *occurcheck-flag* #f)
	(set! *xvars-set-flag* #f)
	(set! *yvars-set-flag* #f)
	(do ((xpart (cdr xlit) (cdr xpart))
	     (ypart (cdr ylit) (cdr ypart)) )
	    ((or fail-flag (null? xpart))
	     (not fail-flag) )
	  (if (not (resolve-unify-cutoff-aux
		    (car xpart) (car ypart)))
	      (set! fail-flag #t) )))
      ; now xlit is ground.
      ; NB!!! We assume now that all vars in ylit are local-only
      (begin
	(set! *xvars-set-flag* #f)
	(if (> (fun-depth-strength yfun)
	       (fun-depth-strength xfun) )
	    (begin (set! *yvars-set-flag* #f) #f) ;; immediate failure
	    (begin (set! *var-set-flag* #f) ;; blocked for non-follow vars
		   (set! *alt-var-set-flag* #f) ;; alternative
		   (if (subsumed-term-list-ordinary?
			(cdr xlit) (cdr ylit))
		       (begin (set! *yvars-set-flag* #t) #t) ;;; success
		       ;;; failure
		       (begin (if (or *var-set-flag*
				      *alt-var-set-flag*)
				  (set! *yvars-set-flag* #t)
				  (set! *yvars-set-flag* #f) )
			      #f)))))))


(define (factor-unify-ordinary xlit ylit)
  (let ((xfun (car xlit))
	(yfun (car ylit))
	(fail-flag #f) )
    (if (eq? (pure-fun xfun) (pure-fun yfun))
	(if (fun-may-cont-vars? xfun)
	    (if (fun-may-cont-vars? yfun)
		; both xlit and ylit are nonground
		(begin
		  (do ((xpart (cdr xlit) (cdr xpart))
		       (ypart (cdr ylit) (cdr ypart)) )
		      ((or fail-flag (null? xpart)) (not fail-flag))
		    (if (not (factor-unify-ordinary-aux
			      (car xpart) (car ypart)))
			(set! fail-flag #t) )))
		; now xlit is nonground and ylit is ground
		(and (<= (fun-depth-strength xfun)
			 (fun-depth-strength yfun) )
		     (subsumed-term-list-ordinary? (cdr ylit) (cdr xlit)) ))
	    ; now xlit is ground
	    (if (fun-may-cont-vars? yfun)
	        ; now xlit is ground and ylit is nonground
		(and (<= (fun-depth-strength yfun)
			 (fun-depth-strength xfun) )
		     (subsumed-term-list-ordinary? (cdr xlit) (cdr ylit)) )
		; now both xlit and ylit are ground
		(and (eq? xfun yfun)
		     (term-list-equal? (cdr xlit) (cdr ylit)) )))
	#f)))

; resolve-unify-ordinary-aux prefers to substitute into
; vars in ylit.
;
; In case any y-vars are set, the flag *yvars-set-flag* is set to #t
; and in case any x-vars are set, the flag *xvars-set-flag* is set
; to #t.

(define (resolve-unify-ordinary-aux xterm yterm)
  (if (var? xterm) ; those tests can be combined
      (set! xterm (var-val xterm)) )
  (if (var? yterm) ; those tests can be combined
      (set! yterm (var-val yterm)) )
  (cond
    ((non-composite? xterm)
       (cond ((non-composite? yterm)
	        (cond ((eq? xterm yterm) #t)
		      ((non-composite:-var? yterm)
		          ;cld be chcked for dummy
			  (set-var-ref! (var-nr yterm) xterm)
			  (if (x-var? yterm)
			      (set! *xvars-set-flag* #t)
			      (set! *yvars-set-flag* #t) )
			  (if (and *repeatedvars-flag*
                                   (non-composite:-var? xterm)
				   (repeated-occ-lit-var? yterm) )
			      (set! *occurcheck-flag* #t) )
			  #t)
		      ((non-composite:-var? xterm) ;;; yterm is constant
		          (set-var-ref! (var-nr xterm) yterm)
			  (if (x-var? xterm)
			      (set! *xvars-set-flag* #t)
			      (set! *yvars-set-flag* #t) )
			  #t)
		      (else #f) ))
	     ((non-composite:-var? xterm) ;;; now yterm is composite
	        (if (and *occurcheck-flag*
			 (term-may-cont-vars? yterm) )
		    (if (occur-check? xterm yterm)
			#f
			(begin (set-var-ref! (var-nr xterm) yterm)
			       (if (x-var? xterm)
				   (set! *xvars-set-flag* #t)
				   (set! *yvars-set-flag* #t) )
			       #t))
		    (begin (set-var-ref! (var-nr xterm) yterm)
			   (if (x-var? xterm)
			       (set! *xvars-set-flag* #t)
			       (set! *yvars-set-flag* #t) )
			   (if (and *repeatedvars-flag*
                                    (repeated-occ-lit-var? xterm)
				    (term-may-cont-vars? yterm) )
			       (set! *occurcheck-flag* #t) )
			   #t) ))
	     (else ;;; now xterm is a constant and yterm is composite
	        #f)))
    ((non-composite? yterm)  ;;; now xterm is known to be composite
        (if (non-composite:-var? yterm)
	    (if (and *occurcheck-flag*
		     (term-may-cont-vars? xterm) )
		(if (occur-check? yterm xterm)
		    #f
		    (begin (set-var-ref! (var-nr yterm) xterm)
			   (if (x-var? yterm)
			       (set! *xvars-set-flag* #t)
			       (set! *yvars-set-flag* #t) )
			   #t) )
		(begin (set-var-ref! (var-nr yterm) xterm)
		       (if (x-var? yterm)
			   (set! *xvars-set-flag* #t)
			   (set! *yvars-set-flag* #t) )
		       (if (and *repeatedvars-flag*
			        (repeated-occ-lit-var? yterm)
				(term-may-cont-vars? xterm) )
			   (set! *occurcheck-flag* #t) )
		       #t))
	    #f)) ;;; xterm was composite and yterm was a constant
    (else ;;; now both xterm and yterm are composite
        (let ((xfun (car xterm))
	      (yfun (car yterm))
	      (fail-flag #f) )
 	   (if (eq? (pure-fun xfun) (pure-fun yfun))
	       (if (or (fun-may-cont-vars? xfun)
		       (fun-may-cont-vars? yfun) )
		   (do ((xpart (cdr xterm) (cdr xpart))
			(ypart (cdr yterm) (cdr ypart)) )
		       ((or fail-flag (null? xpart)) (not fail-flag))
		       (if (not (resolve-unify-ordinary-aux
				   (car xpart) (car ypart)))
			   (set! fail-flag #t) ))
		   (term-list-equal? (cdr xterm) (cdr yterm)) )
	       #f)))))


; resolve-unify-cutoff-aux prefers to substitute into
; vars in ylit. But the variables in ylit which do not have the
; bit with the mask *no-occ-otherlit-var-mask* set are
; NOT allowed to be set at all!!
;
; In case any y-vars are set, the flag *yvars-set-flag* is set to #t
; and in case any x-vars are set, the flag *xvars-set-flag* is set
; to #t.

(define (resolve-unify-cutoff-aux xterm yterm)
  (if (var? xterm) ; those tests can be combined
      (set! xterm (var-val xterm)) )
  (if (and (var? yterm)
	   (not (zero? (logand *no-occ-otherlit-var-mask* yterm))) )
      (set! yterm (var-val yterm)) )
  (cond
    ((non-composite? xterm)
       (cond ((non-composite? yterm)
	        (cond ((eq? xterm yterm) #t)
		      ((and (non-composite:-var? yterm)
			    (not (zero? (logand *no-occ-otherlit-var-mask*
						yterm))) )
		          ;cld be chcked for dummy
			  (set-var-ref! (var-nr yterm) xterm)
			  (if (x-var? yterm)
			      (set! *xvars-set-flag* #t)
			      (set! *yvars-set-flag* #t) )
			  (if (and *repeatedvars-flag*
                                   (non-composite:-var? xterm)
				   (repeated-occ-lit-var? yterm) )
			      (set! *occurcheck-flag* #t) )
			  #t)
		      ((and (non-composite:-var? xterm) ;;; yterm is constant
			    (not (zero? (logand *no-occ-otherlit-var-mask*
						xterm))) )
		          (set-var-ref! (var-nr xterm) yterm)
			  (if (x-var? xterm)
			      (set! *xvars-set-flag* #t)
			      (set! *yvars-set-flag* #t) )
			  #t)
		      (else #f) ))
	     ((and (non-composite:-var? xterm) ;;; now yterm is composite
		   (not (zero? (logand *no-occ-otherlit-var-mask*
				       xterm))) )
	        (if (and *occurcheck-flag*
			 (term-may-cont-vars? yterm) )
		    (if (occur-check? xterm yterm)
			#f
			(begin (set-var-ref! (var-nr xterm) yterm)
			       (if (x-var? xterm)
				   (set! *xvars-set-flag* #t)
				   (set! *yvars-set-flag* #t) )
			       #t))
		    (begin (set-var-ref! (var-nr xterm) yterm)
			   (if (x-var? xterm)
			       (set! *xvars-set-flag* #t)
			       (set! *yvars-set-flag* #t) )
			   (if (and *repeatedvars-flag*
				    (repeated-occ-lit-var? xterm)
				    (term-may-cont-vars? yterm) )
			       (set! *occurcheck-flag* #t) )
			   #t) ))
	     (else ;;; now xterm is a constant and yterm is composite
	        #f)))
    ((non-composite? yterm)  ;;; now xterm is known to be composite
        (if (and (non-composite:-var? yterm)
		 (not (zero? (logand *no-occ-otherlit-var-mask*
				     yterm))) )
	    (if (and *occurcheck-flag*
		     (term-may-cont-vars? xterm) )
		(if (occur-check? yterm xterm)
		    #f
		    (begin (set-var-ref! (var-nr yterm) xterm)
			   (if (x-var? yterm)
			       (set! *xvars-set-flag* #t)
			       (set! *yvars-set-flag* #t) )
			   #t) )
		(begin (set-var-ref! (var-nr yterm) xterm)
		       (if (x-var? yterm)
			   (set! *xvars-set-flag* #t)
			   (set! *yvars-set-flag* #t) )
		       (if (and *repeatedvars-flag*
				(repeated-occ-lit-var? yterm)
				(term-may-cont-vars? xterm) )
			   (set! *occurcheck-flag* #t) )
		       #t))
	    #f)) ;;; xterm was composite and yterm was a constant
    (else ;;; now both xterm and yterm are composite
        (let ((xfun (car xterm))
	      (yfun (car yterm))
	      (fail-flag #f) )
 	   (if (eq? (pure-fun xfun) (pure-fun yfun))
	       (if (or (fun-may-cont-vars? xfun)
		       (fun-may-cont-vars? yfun) )
		   (do ((xpart (cdr xterm) (cdr xpart))
			(ypart (cdr yterm) (cdr ypart)) )
		       ((or fail-flag (null? xpart)) (not fail-flag))
		       (if (not (resolve-unify-cutoff-aux
				   (car xpart) (car ypart)))
			   (set! fail-flag #t) ))
		   (term-list-equal? (cdr xterm) (cdr yterm)) )
	       #f)))))



(define (factor-unify-ordinary-aux xterm yterm)
  (if (var? xterm) (set! xterm (var-val xterm)))
  (if (var? yterm) (set! yterm (var-val yterm)))
  (cond
    ((non-composite? xterm)
       (cond ((non-composite? yterm)
	        (cond ((eq? xterm yterm) #t)
		      ((non-composite:-var? xterm)
		          (if (and
			        (non-composite:-var? yterm)
			        (eq? (var-nr xterm) (var-nr yterm)) )
			      #t
			      (begin (set! *fac-vars-set* #t)
				     (set-var-ref! (var-nr xterm) yterm)
				     #t)))
		      ((non-composite:-var? yterm) ;;; xterm is constant
		          (set! *fac-vars-set* #t)
		          (set-var-ref! (var-nr yterm) xterm)
			  #t)
		      (else #f) ))
	     ((non-composite:-var? xterm) ;;; now yterm is composite
	        (if (term-may-cont-vars? yterm)
		    (if (factor-occur-check? (var-nr xterm) yterm)
			#f
			(begin
			  (set! *fac-vars-set* #t)
			  (set-var-ref! (var-nr xterm) yterm) #t) )
		    (begin
		      (set! *fac-vars-set* #t)
		      (set-var-ref! (var-nr xterm) yterm) #t) ))
	     (else ;;; now xterm is a constant and yterm is composite
	        #f)))
    ((non-composite? yterm)  ;;; now xterm is known to be composite
        (if (non-composite:-var? yterm)
	    (if (term-may-cont-vars? xterm)
		(if (factor-occur-check? (var-nr yterm) xterm)
		    #f
		    (begin
		      (set! *fac-vars-set* #t)
		      (set-var-ref! (var-nr yterm) xterm) #t) )
		(begin
		  (set! *fac-vars-set* #t)
		  (set-var-ref! (var-nr yterm) xterm) #t) )
	    #f)) ;;; xterm was composite and yterm was a constant
    (else ;;; now both xterm and yterm are composite
        (let ((xfun (car xterm))
	      (yfun (car yterm))
	      (fail-flag #f) )
 	   (if (eq? (pure-fun xfun) (pure-fun yfun))
	       (if (or (fun-may-cont-vars? xfun)
		       (fun-may-cont-vars? yfun) )
		   (do ((xpart (cdr xterm) (cdr xpart))
			(ypart (cdr yterm) (cdr ypart)) )
		       ((or fail-flag (null? xpart)) (not fail-flag))
		       (if (not (factor-unify-ordinary-aux
				   (car xpart) (car ypart)))
			   (set! fail-flag #t) ))
		   (term-list-equal? (cdr xterm) (cdr yterm)) )
	       #f)))))

; occur-check checks for occurrence of var in a composite term
; containing possibly assigned vars
;
; composite MUST really be composite!
; composite is assumed to be non-ground

(define (occur-check? var composite)
  (set! composite (cdr composite))
  (do ()
      ((null? composite)
         (eq? var #t) ) ; var is used as a succ/failure flag here
      (let ((el (car composite)))
	(if (non-composite? el)
	    (if (non-composite:-var? el)
		(begin
		   (set! el (var-val el))
		   (if (or (eq? var el)
			   (and (composite? el)
			        (term-may-cont-vars? el)
				(occur-check? var el) ))
		       (begin (set! composite '()) ; signal end
			      (set! var #t) )      ; signal: occurs
		       (set! composite (cdr composite)) ))
		(set! composite (cdr composite)) )
	    (if (and (term-may-cont-vars? el)
		     (occur-check? var el) )
		(begin (set! composite '()) ; signal end
		       (set! var #t) )      ; signal: occurs
		(set! composite (cdr composite)) )))))

; factor-occur-check checks for occurrence of var in a composite term
; containing possibly assigned vars, modulo varinfo, ie always
; just var-nr-s are considered!
;
; composite MUST really be composite!
; composite is assumed to be non-ground
; var MUST be pure var-nr!

(define (factor-occur-check? var composite)
  (set! composite (cdr composite))
  (do ()
      ((null? composite)
         (eq? var #t) ) ; var is used as a succ/failure flag here
      (let ((el (car composite)))
	(if (non-composite? el)
	    (if (non-composite:-var? el)
		(begin
		   (set! el (var-val el))
		   (if (or (and (var? el) (eq? var (var-nr el)))
			   (and (composite? el)
			        (term-may-cont-vars? el)
				(factor-occur-check? var el) ))
		       (begin (set! composite '()) ; signal end
			      (set! var #t) )      ; signal: occurs
		       (set! composite (cdr composite)) ))
		(set! composite (cdr composite)) )
	    (if (and (term-may-cont-vars? el)
		     (factor-occur-check? var el) )
		(begin (set! composite '()) ; signal end
		       (set! var #t) )      ; signal: occurs
		(set! composite (cdr composite)) )))))

;-------------------------------------------------------------------
;
;                    subsumption algorithms
;
;-------------------------------------------------------------------

; clauseinf is a partial inf (as given by make-clause-info)
; cl is a pure clause; it is assumed to NOT be a unit clause
; NB! It assumes that all vars are initially cleared and it clears up
; vars after itself
;
; In case of successful subsumption it sets the variable
; *subsuming-clause* to the full subsuming clause

;(define *tt1* '())
;(define *tt2* '())

(define (havealready-clause? clauseinf cl)

; - - - - - - - - -
;  (newline)(display "havealready-clause? entered with:")(newline)
;  (print-pure-clause cl)(newline)
; (set! *spec-flag* #f)
;	   (newline)
;	   (do ((i 1 (+ 1 i)))
;	       ((> i 10))
;	     (display i)(display ":")
;	     (display (var-ref i))(display " ") )
;	   (newline)
; - - - - - - - - - -
  (let* ((strengthint (car clauseinf))
	 (predsymbsint (cadr clauseinf))
	 (len (shiftright (logand *cl-len-mask* strengthint)
	                  *cl-len-shift*))
	 (strength (logand *strength-mask* strengthint))
	 (pure-strength (logand *term-depth-strength-mask* strengthint))
	 (foundflag #f)
	 (tmp-subs-list (cdr *subsumption-list*))
	 (tmp #f)
	 (glen 0))
    (set! *havealready-timecounter* (+ 1 *havealready-timecounter*))
    (if (not *classical-synth-stratflag*) (set! cl (remove-ans-literals cl)))
    (set! *gcl-varsstack* '()) ; clear it once more, for any case

    ; binary discrimination tree is checked for clauses of length 2

    ;(set! *tt1*  *nonunit-discr-tree*)
    ;(set! *tt2* cl)

    (set! tmp (check-nonunit-discr-tree *nonunit-discr-tree* cl))

    (cond
      ((and tmp
            (or (not (and *firstlitpref-order-strategy-flag* *sat-strategy-flag*))
                (strong-subsumed-purecl? cl (cdar tmp)) ))
      ; current clause subsumed by a double clause
        (set! *forward-subsumed-double-nr*
	      (+ 1 *forward-subsumed-double-nr*))
	; (display "!")
	(set! foundflag #t)
        (set! *subsuming-clause* (car tmp)) )
      ((and (<= len 2)  ; current clause not subsumed
            (not (and *firstlitpref-order-strategy-flag* *sat-strategy-flag*)) ))
      ((<= len 1))
      (else ; try longer candidates

      ; (newline)(display ". . . . . . . . . . . ")(newline)
      ; (pretty-print *nonunit-res*)(newline)
      ; (for-each
      ;	(lambda(x)
      ; 	  (for-each (lambda(y)(newline)(print-full-clause (cdr y))) x) )
      ; 	*nonunit-res*)
      ; (newline)(display ". . . . . . . . . . . ")(newline)

       ; check the list of candidate lists
	(do ((part *nonunit-res* (cdr part)))
	    ((or (null? part) foundflag)
	     (if foundflag
		 (set! *forward-subsumed-overdouble-nr*
			 (+ 1 *forward-subsumed-overdouble-nr*))))
	  ; check one single candidate list
	  (do ((part1 (car part) (cdr part1)))
	      ((or (null? part1) foundflag))
  	      ; each el is a pair of a numeric timecounter and a clause
	      (if (not (eq? (caar part1) *havealready-timecounter*))
		  (begin
		    ; make timecounter up-to-date to avoid repeated checks
		    (set-car! (car part1) *havealready-timecounter*)
		    (if (and *prohibit-subs-clnr*
			     (eq? (car (clinfo-clnumber-sublist
			     		(cadar part1) ))
			     	  *prohibit-subs-clnr*))
			#f
			(set! foundflag
			  (subsumed-clause-aux?
			   cl (cdar part1)
			   (shiftright (logand *cl-len-mask* (caadar part1))
				       *cl-len-shift*)
			   len
			   strength predsymbsint
			   (cdddr (cdar part1)) )))))))))

;    (if foundflag
;	(begin
;	  (newline)
;	  (display "**********************************")(newline)
;	  (display "subsumer for ")
;	  (print-pure-clause cl)
;	  (display " found: ")
;	  (print-full-clause *subsuming-clause*)
;	  (newline) )
;	(begin
;	  (newline)
;	  (display "---------------------------------")(newline)
;	  (display "subsumer not found for ")
;         (print-pure-clause cl)
;	  (newline) ))

    ; avoid full check with too short clauses

;    (cond ((null? tmp-subs-list))
;	  ((eq? 1 (caar tmp-subs-list)) ; do NOT subsume with units here!
;	      (set! tmp-subs-list (cdr tmp-subs-list)) ))

 ;   (cond ((null? tmp-subs-list))
 ;         ((eq? 2 (caar tmp-subs-list)) ; do NOT subsume with binary clauses!
 ;	      (set! tmp-subs-list (cdr tmp-subs-list)) ))

    ; full check with clauses longer than 2

    ; do not do full check: quit search here!

    foundflag))
;
;    (do ((subs-part tmp-subs-list (cdr subs-part)))
;	((or (null? subs-part) (> (caar subs-part) len)))
;	   (cond
;	      ((not foundflag)) ; do nothing
;	         (newline)
;	     	 (display "the derived clause not subsumed by non-unit")
;	     	 (newline) )
;	      ((>= *display-info-level* 20)  ;80
;	         (newline)
;	     	 (display "the derived clause subsumed by non-unit:")
;	     	 (newline)
;	     	 (print-full-clause *subsuming-clause*)
;	     	 (newline) )
;	      (*display-info-char-flag*
;	         (display "-") ))
; - - - - - - - - -
;	   (newline)(display "havealready? exits with following varvals:")
;	   (newline)
;	   (do ((i 1 (+ 1 i)))
;	       ((> i 10))
;	     (display i)(display ":")
;	     (display (var-ref i))(display " ") )
;	   (newline);(set! *spec-flag* #f)
; - - - - - - - - - -
;	   foundflag)
;      (set! glen (caar subs-part)) ; length of currently considered clauses
; - - - - - - - - -
;      ; (if (and (eq? glen 5)(eq? len 5)) (set! *spec-flag* #t))
;      (newline)(display "starting to look clauses of length: ")(display glen)
;	   (newline)
;	   (do ((i 1 (+ 1 i)))
;	       ((> i 10))
;	     (display i)(display ":")
;	     (display (var-ref i))(display " ") )
;	   (newline)
; - - - - - - - - - -
;
;      (set! *stopsubsuming-flag* #f)
;      (do ((part (cdar subs-part) (cdr part)))
;	  ((or (null? part) *stopsubsuming-flag*)
;	     (if foundflag
;		 (begin
;		   (set! *forward-subsumed-nonunit-nr*
;			 (+ 1 *forward-subsumed-nonunit-nr*))
;		   (set! subs-part *listofnull*) )))
;          (set! foundflag
;                (subsumed-clause-aux?
;                   cl (car part) glen len
;		   strength predsymbsint (cdar part) ))))))


; subsumed-clause? checks whether a fully decorated fullcl gcl subsumes
; a fully decorated fullcl scl. It is also optimized for gcl being

; a unit clause.
; As a side effect subsumed-clause-aux? it may set the value of the variable
; *stopsubsuming-flag* to #t and/or *subsuming-clause* to the subs. clause

(define (subsumed-clause? scl gcl)
  (let* ((len (shiftright (logand *cl-len-mask* (caar scl))
	                  *cl-len-shift*))
	 (glen (shiftright (logand *cl-len-mask* (caar gcl))
	                  *cl-len-shift*)) )

   ;(newline)(display "subsumed-clause? entered with:")(newline)
   ;(print-full-clause scl)(newline)
   ;(print-full-clause gcl)(newline)
   ;(debug-print-clause scl)(newline)
   ;(debug-print-clause gcl)(newline)

   (if (eq? glen 1)
       (let ((glit (cadr gcl))
             (foundflag #f)
             (gmaxvar (give-cl-maxvar (car gcl))) )
         (do ((part (cdr scl) (cdr part)))
             ((or (null? part) foundflag)
                foundflag)
           (if (subsumed-literal? (car part) glit gmaxvar)
               (set! foundflag #t) )))
       (subsumed-clause-aux?
	  (cdr scl) gcl glen len
	  (logand *strength-mask* (caar scl)) (cadar scl) (cdr gcl) ))))


; subsumed-literal? assumes x-vars in glit and assumes they are cleared.
; It cleans them up after itself.

(define (subsumed-literal? slit glit gmaxvar)
  (let* ((gfun (car glit))
         (foundflag #f) )
     (if (fun-can-subsume? (car slit) gfun)
         (if (fun-may-cont-vars? gfun)
             (begin
               (set! *var-set-flag* #f)
               (set! *alt-var-set-flag* #f)
               (if (subsumed-term-list-ordinary? (cdr slit) (cdr glit))
                   (set! foundflag #t)
                   (if (commutative-fun? (pure-fun gfun))
                       (begin (clear-x-vars! gmaxvar)
                              (if (swapped-subsumed-term-list-ordinary?
                                     (cdr slit) (cdr glit))
                                  (set! foundflag #t) ))))
               (if (or *var-set-flag* *alt-var-set-flag*)
                   (begin (clear-x-vars! gmaxvar)
                          (set! *var-set-flag* #f)
                          (set! *alt-var-set-flag* #f) )))
             (or (term-list-equal? (cdr slit) (cdr glit))
                 (and (commutative-fun? (pure-fun gfun))
                      (swapped-term-list-equal? (cdr slit) (cdr glit)) ))))
      foundflag))

; subsumed-clause-aux? is also called directly from havealready?
; It is NOT optimized for gfcl to be unit.


(define (subsumed-clause-aux?
	  cl gfcl glen len strength predsymbsint gcl-fastlits)
  (let* ((ginf (car gfcl))
	 (gstrength (logand *strength-mask* (car ginf)))
	 (gclpreds (cadr ginf))
	 (tmpres '())
	 (foundflag #f) )
    ;(newline)(display "gfcl: ")(debug-print-clause gfcl)(newline)
    ;(newline)(display "gstrength: ")(display gstrength)(newline)
    ;(newline)(display "strength: ")(display strength)(newline)
    (set! *subs-tried-nr* (+ 1 *subs-tried-nr*))
    (if (eq? len glen)
	; clauses with the same length.
	; Here a ground clause cannot subsume a nonground clause
	(if (> gstrength strength)
	    ; a clause with too big a strength reached;
	    ; terminate clauses with this length
	    (begin
	      (set! *subs-length-fail-nr* (+ 1 *subs-length-fail-nr*))
	      (set! *stopsubsuming-flag* #t) )
	    ; gclstrength is not immediately unacceptable
	    (if (not (eq? gclpreds predsymbsint))
		(set! *subs-gclpreds-fail-nr* (+ 1 *subs-gclpreds-fail-nr*))
		; gclpreds is equal to sclpreds
		(if (not (bigger-strength-aux? strength gstrength))
		    (set! *subs-aux-strength-fail-nr*
			  (+ 1 *subs-aux-strength-fail-nr*) )
		    ; full test on strengthints passed OK
		    (if (funlist-can-subsume? cl gcl-fastlits len glen)
			; all OK on funlist level
			(if (fun-may-cont-vars? gstrength)
			    ; gcl is not ground; try regular subsumption
			    (begin
			      (set! *subs-reg-subs-tried-nr*
				    (+ 1 *subs-reg-subs-tried-nr*) )
			      (set! *gcl-maxvar*
				    (give-cl-maxvar ginf) )
                              (if (and *sat-strategy-flag*
                                       *firstlitpref-order-strategy-flag*)
			          (set! tmpres (strong-subsumed-purecl? cl (cdr gfcl)))
			          (set! tmpres (subsumed-pure-clause-aux?
					         cl (cdr gfcl) #f '() #f)))
			      (if (not tmpres)
			          ; subsumption is a failure
				  (set! *subs-reg-subs-fail-nr*
					(+ 1 *subs-reg-subs-fail-nr*) )
				  ; subsumption is a success!!
				  (begin
				    (set! foundflag #t)
				    (set! *subsuming-clause* gfcl)
				    (set! *stopsubsuming-flag* #t)

				    (clear-gcl-vars)
				    (set! *gcl-varsstack* '()) )))
			    ; gcl is ground, thus a success!
			    (begin
			      (set! foundflag #t)
			      (set! *subsuming-clause* gfcl)
			      (set! *stopsubsuming-flag* #t) ))))))
	; clauses with the different length.
	; Here a ground clause may subsume a non-ground clause
	(if (> (logand *term-depth-strength-mask* gstrength)
	       (logand *term-depth-strength-mask* strength) )
	    ; should we terminate clauses with this length?
	    (begin
	      (set! *subs-strength-fail-nr*
		    (+ 1 *subs-strength-fail-nr*) )
	      (if (not (zero? (logand *ground-mask* gstrength)))
		  ;  gcl is ground, thus terminate search
		  (set! *stopsubsuming-flag* #t) ))
	    ; gclstrength is not immediately unacceptable
	    (if (not (eq? gclpreds (logand gclpreds predsymbsint)))
		(set! *subs-gclpreds-fail-nr* (+ 1 *subs-gclpreds-fail-nr*))
		; gclpreds is a subset of sclpreds
		(if (not (bigger-strength-aux-aux? strength gstrength))
		    (set! *subs-aux-strength-fail-nr*
			  (+ 1 *subs-aux-strength-fail-nr*) )
		    ; full test on strengthints passed OK
		    (if (not (funlist-can-subsume? cl gcl-fastlits len glen))
			(set! *subs-funlist-fail-nr*
			      (+ 1 *subs-funlist-fail-nr*) )
			; all OK on funlist level
			(if (fun-may-cont-vars? gstrength)
			    ; gcl is not ground; try regular subsumption
			    (begin
			      (set! *subs-reg-subs-tried-nr*
				    (+ 1 *subs-reg-subs-tried-nr*) )
			      (set! *gcl-maxvar*
				    (give-cl-maxvar ginf) )
			      (if (and *sat-strategy-flag*
                                       *firstlitpref-order-strategy-flag*)
			          (set! tmpres (strong-subsumed-purecl? cl (cdr gfcl)))
			          (set! tmpres (subsumed-pure-clause-aux?
					         cl (cdr gfcl) #f '() #f)))
			      (if (not tmpres)
				  ; failure
				  (set! *subs-reg-subs-fail-nr*
					(+ 1 *subs-reg-subs-fail-nr*) )
				  ; subsumption is a success!!
				  (begin
				    (set! foundflag #t)
				    (set! *subsuming-clause* gfcl)
				    (set! *stopsubsuming-flag* #t) ; stop
				    (clear-gcl-vars)
				    (set! *gcl-varsstack* '()) )))
			    ; gcl is ground, thus a success!
			    (begin
			      (set! foundflag #t)
			      (set! *subsuming-clause* gfcl)
			      (set! *stopsubsuming-flag* #t) )))))))
    foundflag))


(define (strong-subsumed-purecl? scl gcl)
  (let* ((glit (car gcl))
	 (gfun (car glit))
	 (fail-flag #f) )
       (do ((gpart gcl (cdr gpart)))
	   ((or (null? gpart) fail-flag)
              (not fail-flag) )
         (if (not (some? (lambda(lit)
                           (and (eq? gfun (car lit))
                                (term-equal-strong? glit (car lit)) ))

                          scl))
             (set! fail-flag #t) ))))


; back-subsume removes back-subsumed clauses from
; *subsumption-list* and sets destructively their strengthints
; (the part under mask *term-depth-strength-mask* to 0)
;
; clauseinf is a partial inf (as given by make-clause-info)
; cl is a pure clause; it may be a unit clause
; and it is assumed to be sorted and 2-nd level decorated
; NB! It assumes that all vars are initially cleared and it clears up
; vars after itself

(define (back-subsume clauseinf cl)

  ;(newline)(display "back-subsume! entered with:")(newline)
  ;(print-pure-clause cl)(newline)

  (let* ((strengthint (car clauseinf))
	 (predsymbsint (cadr clauseinf))
	 (len (shiftright (logand *cl-len-mask* strengthint)
	                  *cl-len-shift*))
	 (strength (logand *strength-mask* strengthint))
	 (pure-strength (logand *term-depth-strength-mask* strengthint))
	 (clmaxvar (give-cl-maxvar clauseinf))
	 (foundflag #f)
	 (tmp-subs-list (cdr *subsumption-list*))
	 (slen 0)
	 (tmp-part '())
	 (last-part '()) )
    (if (not *classical-synth-stratflag*) (set! cl (remove-ans-literals cl)))
    (set! *gcl-maxvar* clmaxvar)
    (set! *subsumed-clauses* '())
    (set! *gcl-varsstack* '()) ; clear it once more, for any case
    ; first, skip too short clauses
    (do ()
	((or (null? tmp-subs-list)(>= (caar tmp-subs-list) len)))
      (set! tmp-subs-list (cdr tmp-subs-list)) )
    ; loop over OK sublists
    (do ((subs-part tmp-subs-list (cdr subs-part)))
	((null? subs-part)
	   (cond
	      ((null? *subsumed-clauses*)) ; do nothing
	      ((>= *display-info-level* 63)
	         (newline)
		 (display "clauses back-subsumed:")
		 (newline)
		 (for-each (lambda(x)(print-full-clause x)(newline))
			   *subsumed-clauses*)
		 (newline) )
	      (*display-info-char-flag*
	         (display "*") ))
	   (for-each (lambda(x)
		       (set! *back-subsumed-nr* (+ 1 *back-subsumed-nr*))
		       (set-car! (car x)
			  (logxor *term-depth-strength-mask*
		             (logior *term-depth-strength-mask* (caar x)) )))
		     *subsumed-clauses*)
	   (set! *subsumed-clauses* '()) )
      (set! slen (caar subs-part)) ; length of currently considered clauses
      (set! last-part (car subs-part))
      (set! tmp-part (cdr last-part))
      ; loop for OK-strength clauses
      (set! *var-set-flag* #f)
      (set! *alt-var-set-flag* #f)
      (clear-x-vars! clmaxvar) ; for security
      (set! *gcl-varsstack* '()) ; for security
      (if (eq? slen 1)
	  ; back-subsuming unit clauses
	  (let ((purefn (pure-fun (caar cl)))
		(littail (cdar cl)) )
	    (do ((part tmp-part (cdr part)))
		((or (null? part)
		     (> pure-strength
			(logand *term-depth-strength-mask* (caaar part))) ))
	      (let* ((scl (car part))
		     (slit (cadr scl))
		     (sfun (car slit)) )
		(if (and (eq? purefn (pure-fun sfun))
			 (bigger-strength? sfun strengthint) )
		    (begin
		      (if (subsumed-term-list-ordinary?
			   (cdr slit) littail)
			  ; successful subsumption
			  (begin
			    (set-cdr! last-part (cdr part)) ; remove!
			    (set! *subsumed-clauses*
				  (cons scl *subsumed-clauses*) )
			    ; subsumption fails
			    (if (commutative-fun? purefn)
				(begin
				  (if (or *var-set-flag* *alt-var-set-flag*)
				      (begin
					(clear-x-vars! clmaxvar)
					(set! *var-set-flag* #f)
					(set! *alt-var-set-flag* #f) ))
				  (if  (swapped-subsumed-term-list-ordinary?
					(cdr slit) littail)
				       ; successful subsumption
				       (begin
					 (set-cdr! last-part (cdr part))
					 ;remv
					 (set! *subsumed-clauses*
					  (cons scl *subsumed-clauses*) )))))))
		      ; clear up vars
		      (if (or *var-set-flag* *alt-var-set-flag*)
			  (begin
			    (clear-x-vars! clmaxvar)
			    (set! *var-set-flag* #f)
			    (set! *alt-var-set-flag* #f) )))))
	      ; last-part is always the one before part
	      (if (eq? (cdr last-part) part)
		  (set! last-part part) )))
	  ; the clauses to be subsumed are nonunit
	  (do ((part tmp-part (cdr part)))
	      ((or (null? part)
		   (> pure-strength
		      (logand *term-depth-strength-mask* (caaar part))) ))
	    ;(newline)(display "looking at subsumable:")
	    ;(print-full-clause (car part))(newline)

	    (let* ((scl (car part))
		   (sclinf (car scl)) )
	      (if (if (eq? len slen)
		      (bigger-strength? (car sclinf) strength)
		      (bigger-strength-aux? (car sclinf) strength) )
		  ; sclstrength is not immediately unacceptable
		  (if (if (eq? len slen)
			  (eq? predsymbsint (cadr sclinf))
			  (eq? predsymbsint
			       (logand (cadr sclinf) predsymbsint) ))
		      ; gclpreds is a subset of sclpreds or equal
		      ; to sclpreds (if lens are equal), as required
		      (if (funlist-can-subsume? (cdr scl) cl slen len)
			  ; all OK on funlist level
			  (if (fun-may-cont-vars? strengthint)
			      ; gcl is not ground; try regular subsumption
			      (begin
				(if (subsumed-pure-clause-aux?
				     (cdr scl) cl #f '() #f)
				    ; subsumption is a success!!
				    (begin
				      (set-cdr! last-part (cdr part)) ;remv
				      (set! *subsumed-clauses*
					    (cons scl *subsumed-clauses*) )
				      (clear-gcl-vars)
				      (set! *gcl-varsstack* '()) )))
			      ; gcl is ground, thus subsumption is a success!
			      (begin
				(set-cdr! last-part (cdr part)) ; remove
				(set! *subsumed-clauses*
				      (cons scl *subsumed-clauses*) )))))))
	    ; last-part is always the one before part
	    (if (eq? (cdr last-part) part)
		(set! last-part part) ))))))



(define (bigger-strength? sclstrength gclstrength)
  (and (>= (logand *strength-mask* sclstrength)
	   (logand *strength-mask* gclstrength) )
       (>= (logand *term-depth-strength-mask* sclstrength)
	   (logand *term-depth-strength-mask* gclstrength) )
       (>= (logand *term-size-strength-mask* sclstrength)
	   (logand *term-size-strength-mask* gclstrength) )
       (>= (logand *const-number-mask* sclstrength)
	   (logand *const-number-mask* gclstrength) )))


(define (bigger-strength-aux? sclstrength gclstrength)
  (and (>= (logand *term-depth-strength-mask* sclstrength)
	   (logand *term-depth-strength-mask* gclstrength) )
       (>= (logand *term-size-strength-mask* sclstrength)
	   (logand *term-size-strength-mask* gclstrength) )
       (>= (logand *const-number-mask* sclstrength)
	   (logand *const-number-mask* gclstrength) )))

(define (bigger-strength-aux-aux? sclstrength gclstrength)
  (and (>= (logand *term-size-strength-mask* sclstrength)
	   (logand *term-size-strength-mask* gclstrength) )
       (>= (logand *const-number-mask* sclstrength)
	   (logand *const-number-mask* gclstrength) )))

; funlist-can-subsume? takes two ordered (by >-fun-ordering?)
; pure clauses and checks on the pred-symb-position level whether
; the second clause may subsume the first one.
;
; It also takes their lengths, but assumes that glen <= slen.
; It is for use in case clauses may contain ans-literals.
;
; It uses the decoration info of pred-symb-position objects and does not
; enter literals. It checks ground-lit-equal? for ground literals,
; assuming that equal ground literals are always eq.


(define (funlist-can-subsume? scl gcl slen glen)
  (if *classical-synth-stratflag*
      (let* ((glit1 (car gcl))
	     (gfun1 (car glit1)) )
	(if (answer-pred? gfun1)
	    (let* ((slit1 (car scl))
		   (sfun1 (car slit1)) )
	      (if (answer-pred? sfun1)
		  (if (fun-may-cont-vars? gfun1)
		      (if (fun-can-subsume? sfun1 gfun1)
			  (funlist-can-subsume-aux? (cdr scl) (cdr gcl))
			  #f)
		      (if (ground-lit-equal? glit1 slit1)
			  (funlist-can-subsume-aux? (cdr scl) (cdr gcl))
			  #f))
		  #f))
	    (if (answer-pred? (caar scl))
		(if (eqv? slen glen)
		    #f
		    (funlist-can-subsume-aux? (cdr scl) gcl) )
		(funlist-can-subsume-aux? scl gcl) )))
      (funlist-can-subsume-aux? scl gcl) ))


; funlist-can-subsume-aux? takes two ordered (by >-fun-ordering?)
; pure clauses and checks on the pred-symb-position level whether
; the second clause may subsume the first one.
;
; It assumes that the clauses do not contain ans-literals, thus
; it should be used in non-ans proof searches insted of funlist-can
; subsume?.
;
; It uses the decoration info of pred-symb-position objects and does not
; enter literals. It checks ground-lit-equal? for ground literals,
; assuming that equal ground literals are always eq.


(define (funlist-can-subsume-aux? scl gcl)

  ;(newline)
  ;(print-pure-clause scl)(newline)
  ;(print-pure-clause gcl)(newline)

  (let ((spart scl))
    (do ()
	((or (null? spart) (null? gcl))

	  ;(display "res: ")(display (null? gcl))(newline)

          (null? gcl) )
      (let* ((glit (car gcl))
	     (gfun (car glit))
	     (success-flag #f) )
	(set! spart scl)
	(do ()
	    ((or (null? spart) success-flag))
	    (let* ((slit (car spart))
		   (sfun (car slit)) )
	      (if (>-fun-ordering-aux? gfun sfun) ; this & further may be OK?
		  (set! spart '()) ; RETURN failure!
		  (if (fun-may-cont-vars? gfun)
		      (if (fun-can-subsume? sfun gfun) ; this el OK?
			  (begin
			    (set! gcl (cdr gcl))
			    (set! success-flag #t) )
			  (set! spart (cdr spart)) )
		      (if (ground-lit-equal? slit glit)
			  (begin
			    (set! gcl (cdr gcl))
			    (set! success-flag #t) )
			  (set! spart (cdr spart)) )))))))))


(define (>-fun-ordering? xfun yfun)
  (cond
    ((answer-pred? xfun)
      (if (answer-pred? yfun)
	  (if (>= (logand *strength-mask* xfun)
		  (logand *strength-mask* yfun) )
	      #t
	      #f)
	  #t))
    ((answer-pred? yfun) #f)
    ((or (> (logand *strength-mask* xfun) (logand *strength-mask* yfun))
	 (and (eq? (logand *strength-mask* xfun)
		   (logand *strength-mask* yfun))
	      (> (logand *signedfnsymb-mask* xfun)
		 (logand *signedfnsymb-mask* yfun) )))
	  #t)
    (else #f) ))

(define (>-fun-ordering-aux? xfun yfun)
  (if (or (> (logand *strength-mask* xfun) (logand *strength-mask* yfun))
	  (and (eq? (logand *strength-mask* xfun)
		    (logand *strength-mask* yfun))
	       (> (logand *signedfnsymb-mask* xfun)
		  (logand *signedfnsymb-mask* yfun) )))
      #t
      #f))


; subsumed-pure-clause-ans? is for full subsumption check in case
; clauses might contain ans-literals. It calls subsumed-pure-clause-aux?
; and assumes that funlist-can-subsume? has been used before.

(define (subsumed-pure-clause-ans? scl gcl)
  (let* ((glit1 (car gcl)))
    (if (answer-pred? (car glit1))
	; now we assume that sfun1 is also an answer-pred
	(let* ((slit1 (car scl)))
	  (if (not (fun-may-cont-vars? (car glit1)))
	      ; now we assume that sfun1 is also ground and eq
	      (subsumed-pure-clause-aux?
	         (cdr scl) (cdr gcl) #f '() #f)
	      ; now we assume that gfun1 can subsume sfun1
	      (begin
		(set! *var-set-flag* #f) ;; no new vars set yet
		(if (subsumed-term-list-ordinary?
		     (cdr slit1) (cdr glit1) )
		    (if (subsumed-pure-clause-aux?
			   (cdr scl) (cdr gcl) #t '() #f)
			(begin (clear-gcl-vars) #t)
			(begin (clear-gcl-vars) #f) )
		    (begin (if *var-set-flag* (clear-gcl-vars))
			   #f)))))
	; gfun1 was not answer-pred
	(if (answer-pred? (caar scl))
	    (subsumed-pure-clause-aux? (cdr scl) gcl #f '() #f)
	    (subsumed-pure-clause-aux? scl gcl #f '() #f) ))))

;subsumed-pure-clause-aux? takes:
; scl and gcl: two pure clauses
; var-set-before: originally #f
; done-scl-parts: originally ()
; swapped-flag: origianlly #f
;
; variables are assumed to be initially cleared.
;
; NB! in case the subsumption is successful, variable values are
; not cleaned up (ie they stay as set during subsumption).
; In case the subsumption fails, all vars are cleared.


(define (subsumed-pure-clause-aux?
	   scl gcl var-set-before done-scl-parts swapped-flag)

; - - - - - - - - -
;(if #t ; *spec-flag*
;    (begin
;  (newline)(display "subsumed-pure-clause-aux? called with:")
;  (newline)(print-pure-clause scl)
;  (newline)(print-pure-clause gcl)(newline)
;  (newline)(display var-set-before)(newline)(display done-scl-parts)
;  (newline)(display *gcl-varsstack*)(newline)
;	   (newline)
;	   (do ((i 1 (+ 1 i)))
;	       ((> i 10))
;	     (display i)(display ":")
;	     (display (var-ref i))(display " ") )
;	   (newline) ))
; - - - - - - - - - -

  (let* ((glit (car gcl))
	 (gfun (car glit))
	 (success-flag #f) )
    (if  (and *sat-strategy-flag* *firstlitpref-order-strategy-flag*)
       ; very weak subsumption
       (strong-subsumed-purecl? scl gcl)
       ; regular subsumption
       (do ((spart scl (if (null? spart) '() (cdr spart))))
	((or (null? spart) success-flag)
; - - - - - - - - -
;	  (if #t ; *spec-flag*
;           (begin (newline)(display "subsumed-pure-clause-aux? exits with:")
;                  (newline)(display success-flag)(newline)
;	           (display *gcl-varsstack*)(newline)
;	   (newline)
;	   (do ((i 1 (+ 1 i)))
;	       ((> i 10))
;	     (display i)(display ":")
;	     (display (var-ref i))(display " ") )
;	   (newline) ))
; - - - - - - - - - -
	    success-flag)
      (if (or (null? done-scl-parts)
	      (not (memq spart done-scl-parts)) )
	  (let ((slit (car spart)))
	    (if (fun-can-subsume? (car slit) gfun)
		(if (fun-may-cont-vars? gfun)

		    ;;; glit may contain vars

		    (begin
		      (if var-set-before (save-gcl-vars))
		      (set! *var-set-flag* #f) ;; no new vars set yet
		      (set! *alt-var-set-flag* #f) ;; no new vars set
		      (set! swapped-flag #f) ;; no arg-swapping done

		      ;;; try to subsume slit with glit,
		      ;;; possibly with commuted first args.

		      (if (or (subsumed-term-list-ordinary?
			         (cdr slit) (cdr glit))
			      (and (commutative-fun? gfun)
				   (begin
				     (if (or *var-set-flag*
					     *alt-var-set-flag*)
					 (begin (if var-set-before
						    (restore-gcl-vars)
						    (clear-gcl-vars) )
						(set! *var-set-flag* #f)
						(set! *alt-var-set-flag* #f) ))
				     #t)
				   (swapped-subsumed-term-list-ordinary?
				     (cdr slit) (cdr glit) )
				   (begin (set! swapped-flag #t) #t) ))

			  ;;; subsumption of literals successful

			  (if *var-set-flag*

			      ;;; new vars were set when subsuming.
			      ;;; possibly we'll have to also try
			      ;;; the swapped case.

			      (let ((grest (cdr gcl)))
				(if (null? grest)
				    (set! success-flag #t)
				    (if (subsumed-pure-clause-aux?
					   scl grest #t
					   (cons spart done-scl-parts)
					   swapped-flag)
					(set! success-flag #t)
					(if var-set-before
					    (restore-gcl-vars)
					    (clear-gcl-vars) )))

				;;; if full subsumption failed and
				;;; gfun is commutative, we'll try
				;;; the lit subsumption with swapped glit

				(if (and (not success-flag)
					 (commutative-fun? gfun)
					 (not swapped-flag)
            				 (swapped-subsumed-term-list-ordinary?
					    (cdr slit) (cdr glit) ))

               			  ;;; swapped subsumption of literals
				  ;;; successful

			      (if *var-set-flag*

			        ;;; new vars were set during
				;;; swapped subsumption

				(let ((grest (cdr gcl)))
				  (if (null? grest)
				      (set! success-flag #t)
				      (if (subsumed-pure-clause-aux?
					     scl grest #t
					     (cons spart done-scl-parts)
					     swapped-flag)
					  (set! success-flag #t)
					  (if var-set-before
					      (restore-gcl-vars)
					      (clear-gcl-vars) ))))

				;;; no new vars were set when
				;;; doing swapped subsumption

				(let ((grest (cdr gcl))
				      (tmp-var-set-flag *alt-var-set-flag*) )
				  (if (null? grest)
				      (set! success-flag #t)
				      (if (subsumed-pure-clause-aux?
					     scl grest #t
					     (cons spart done-scl-parts)
					     swapped-flag)
					  (set! success-flag #t)
					  (begin
					    (if tmp-var-set-flag
						; one-lit-vars were set
						(if var-set-before
						    (restore-gcl-vars)
						    (clear-gcl-vars) )
						; no vars were set at all
						(if var-set-before
						    (pop-gcl-vars-stack) ))
					    ; do not backtrack: fail now
					    (set! spart '()) )))))))

			      ;;; no new vars were set when subsuming
			      ;;; thus we do not have to backtrack

			      (let ((grest (cdr gcl))
				    (tmp-var-set-flag *alt-var-set-flag*) )
				(if (null? grest)
				    (set! success-flag #t)
				    (if (subsumed-pure-clause-aux?
					   scl grest #t
					   (cons spart done-scl-parts)
					   swapped-flag)
					(set! success-flag #t)
					(begin
					  (if tmp-var-set-flag
					      ; one-lit-vars were set
					      (if var-set-before
						  (restore-gcl-vars)
						  (clear-gcl-vars) )
					      ; no vars were set at all
					      (if var-set-before
						  (pop-gcl-vars-stack) ))
					  ; do not backtrack: fail now
					  (set! spart '()) )))))

			  ;;; subsumption of literals failed,
			  ;;; (glit NOT ground)

			  (if (or *var-set-flag*
				  *alt-var-set-flag*)
			      (if var-set-before
				  (restore-gcl-vars)
				  (clear-gcl-vars) )
			      (if var-set-before
				  (pop-gcl-vars-stack) ))))

		    ;;; the glit is ground

		    (if (ground-lit-equal? glit slit)
			(let ((grest (cdr gcl)))
			  (if (null? grest)
			      (set! success-flag #t)
			      (if (subsumed-pure-clause-aux?
				   scl grest var-set-before
				   (cons spart done-scl-parts)
				   swapped-flag)
				  (set! success-flag #t)
				  ; do not backtrack: fail now
				  (set! spart '()) ))))))))))))


; same-or-nonground-fun? is used with assumption that the subset of ground
; literals in a clause is preceding all nonground literals and is
; ordered by the decorated pred-symb numbers

(define (same-or-nonground-fun? xfun yfun)
  (or (fun-may-cont-vars? yfun)
      (eq? xfun yfun) ))

(define (subsumed-term-ordinary? sterm gterm)

  ;(newline)(display "subsumed-term-ordinary? called with:")(newline)
  ;(debug-print-term sterm)(newline)
  ;(debug-print-term gterm)(newline)

  (cond
    ((non-composite? gterm)
       (if (non-composite:-var? gterm)
	   (let ((gval (var-ref (var-nr gterm))))
	     (cond ((unassigned-val? gval)
		       (set-var-ref! (var-nr gterm) sterm)
		       (if (later-occurrences-var? gterm)
			   (set! *var-set-flag* #t)
			   (set! *alt-var-set-flag* #t) )
		       #t)
		   ((non-composite? sterm)
		      (if (non-composite:-var? sterm)
			  (and (var? gval)
			       (eq? (var-nr sterm) (var-nr gval)) )
			  (eq? sterm gval) ))
		   ; now sterm is composite
		   ((non-composite? gval) #f)
		   ;;; now both gval and sterm are composite
		   ((eq? sterm gval) #t)
		   ((composite-equal-strong? sterm gval) #t)
		   (else #f) ))
	   (eq? sterm gterm) ))
    ((non-composite? sterm) #f)
    (else ;;; now both sterm and gterm are composite
     (let ((gfun (car gterm))
	   (sfun (car sterm)) )
       (if (fun-can-subsume? sfun gfun)
	   (if (fun-may-cont-vars? gfun)
	       (do ((spart (cdr sterm) (cdr spart))
		    (gpart (cdr gterm)  (cdr gpart)) )
		   ((or (null? spart)
			(not (subsumed-term-ordinary?
			       (car spart) (car gpart) )))
		    (null? spart) ))
	       (term-list-equal? (cdr gterm) (cdr sterm)) )
	   #f)))))


; subsumed-term-ordinary-collectvars? is like subsumed-term-ordinary,
; except that it returns a pair (<res> . <vars>) where res is
; the standard result and <vars> is a list of variables which were
; set during subsumption.

(define (subsumed-term-ordinary-collectvars? sterm gterm)
  (set! *collected-vars* '())
  (let ((res (subsumed-term-ordinary-collectvars-aux? sterm gterm)))
    (cons res *collected-vars*) ))

(define (subsumed-term-ordinary-collectvars-aux? sterm gterm)

  ;(newline)(display "subsumed-term-ordinary? called with:")(newline)
  ;(debug-print-term sterm)(newline)
  ;(debug-print-term gterm)(newline)

  (cond
    ((non-composite? gterm)
       (if (non-composite:-var? gterm)
	   (let ((gval (var-ref (var-nr gterm))))
	     (cond ((unassigned-val? gval)
		       (set-var-ref! (var-nr gterm) sterm)
		       (set! *collected-vars* (cons gterm *collected-vars*))
		       (if (later-occurrences-var? gterm)
			   (set! *var-set-flag* #t)
			   (set! *alt-var-set-flag* #t) )
		       #t)
		   ((non-composite? sterm)
		      (if (non-composite:-var? sterm)
			  (and (var? gval)
			       (eq? (var-nr sterm) (var-nr gval)) )
			  (eq? sterm gval) ))
		   ; now sterm is composite
		   ((non-composite? gval) #f)
		   ;;; now both gval and sterm are composite
		   ((eq? sterm gval) #t)
		   ((composite-equal-strong? sterm gval) #t)
		   (else #f) ))
	   (eq? sterm gterm) ))
    ((non-composite? sterm) #f)
    (else ;;; now both sterm and gterm are composite
     (let ((gfun (car gterm))
	   (sfun (car sterm)) )
       (if (fun-can-subsume? sfun gfun)
	   (if (fun-may-cont-vars? gfun)
	       (do ((spart (cdr sterm) (cdr spart))
		    (gpart (cdr gterm)  (cdr gpart)) )
		   ((or (null? spart)
			(not (subsumed-term-ordinary-collectvars-aux?
			       (car spart) (car gpart) )))
		    (null? spart) ))
	       (term-list-equal? (cdr gterm) (cdr sterm)) )
	   #f)))))


(define (subsumed-term-list-ordinary? sterm gterm)
  (do ((spart sterm (cdr spart))
       (gpart gterm (cdr gpart)) )
      ((or (null? spart)
	   (not (subsumed-term-ordinary?
		 (car spart) (car gpart) )))
       (null? spart) )))


(define (swapped-subsumed-term-list-ordinary? sterm gterm)
  (and (subsumed-term-ordinary? (car sterm) (cadr gterm))
       (subsumed-term-ordinary? (cadr sterm) (car gterm))
       (do ((spart (cddr sterm) (cdr spart))
	    (gpart (cddr gterm) (cdr gpart)) )
	   ((or (null? spart)
		(not (subsumed-term-ordinary?
		      (car spart) (car gpart) )))
	    (null? spart) ))))


;=================================================================
;
;              a library of helpful auxiliary funs
;
;================================================================

; ------------ VERY IMPORTANT that the first part is fast and inlined! --

(define (not-zero? x)
  (not (zero? x)) )

(define (mymax x y)
  (if (> x y) x y) )

(define (maxplus max x y)
  (let ((newmaxplusnr (+ x y)))
    (if (> newmaxplusnr max) max newmaxplusnr) ))

; -------------------- now comes possibly slower stuff -------------

; rassq assumes lst is a list of pairs ((ob1 . key1) ...)

(define (rassq key lst)
  (cond ((null? lst) #f)
	((eq? key (cdar lst)) (car lst))
	(else (rassq key (cdr lst))) ))


(define (adjoin x lst)
  (if (memq x lst) lst (cons x lst)) )

(define (union xset yset)
  (cond ((null? xset) yset)
	((memq (car xset) yset)
	   (union (cdr xset) yset) )
	(else
	   (cons (car xset) (union (cdr xset) yset)) )))

(define (list-union lst)
  (if (null? lst)
      '()
      (if (null? (cdr lst))
	  (car lst)
	  (union (car lst) (list-union (cdr lst))) )))

(define (member-memq? x lst)
  (if (null? lst)
      #f
      (if (memq x (car lst))
	  #t
	  (member-memq? x (cdr lst)) )))


(define (assoc-memq x lst)
  (if (null? lst)
      #f
      (if (memq x (caar lst))
	  (car lst)
	  (assoc-memq x (cdr lst)) )))

(define (nth-cdr n lst)
  (if (zero? n) lst (nth-cdr (- n 1) (cdr lst))) )

(define (pos-in-list x lst)
  (pos-in-list-aux x lst 1) )

(define (pos-in-list-aux x lst nr)
  (if (pair? lst)
      (if (eq? (car lst) x) nr (pos-in-list-aux x (cdr lst) (+ 1 nr)))
      -1))

; split-list takes a list and a number (assumed: number <= length of list)
; and returns a pair of the list of the first n els of the list
; (in the original order) and the rest of the list.

(define (split-list l n)
  (let* ((res (cons '() '()))
	 (resend res) )
    (do ((lpart l (cdr lpart)))
	((zero? n)
	   (cons (cdr res) lpart) )
      (set-cdr! resend (cons (car lpart) (cdr resend)))
      (set! resend (cdr resend))
      (set! n (- n 1)) )))

(define (split-if fn lst)
  (let ((yeslst '())
	(nolst '()) )
    (do ((lstpart lst (cdr lstpart)))
	((null? lstpart)
	   (cons (reverse yeslst) (reverse nolst)) )
      (if (fn (car lstpart))
	  (set! yeslst (cons (car lstpart) yeslst))
	  (set! nolst (cons (car lstpart) nolst)) ))))


(define (my-last-pair x)
  (if (pair? x)
      (my-last-pair-aux x)
      x))

(define (my-last-pair-aux x)
  (if (pair? (cdr x))
      (my-last-pair-aux (cdr x))
      x))

(define (my-append! x y)
  (let ((z (my-last-pair x)))
    (if (pair? z)
	(begin (set-cdr! z y) x)
	y)))


(define (but-last lst)
  (cond ((not (pair? lst)) '())
        ((not (pair? (cdr lst))) '())
        (else (cons (car lst) (but-last (cdr lst)))) ))

; destructive map

(define (map! fn lst)
  (if (null? lst)
      lst
      (begin
	(set-car! lst (fn (car lst)))
	(map! fn (cdr lst)) )))

(define (some? fn lst)
  (if (pair? lst)
      (or (fn (car lst))
          (some? fn (cdr lst)) )
      #f))

(define (every? fn lst)
  (if (pair? lst)
      (if (fn (car lst)) (every? fn (cdr lst)) #f)
      #t))

(define (member-if fn lst)
  (if (pair? lst)
      (if (fn (car lst)) lst (member-if fn (cdr lst)))
      #f))

(define (remove-if fn lst)
  (cond ((not (pair? lst)) lst)
        ((fn (car lst))
           (remove-if fn (cdr lst)) )
        (else
           (cons (car lst) (remove-if fn (cdr lst))) )))


(define (filter-map f lst)
  (if (pair? lst)
      (let ((res (f (car lst))))
         (if res
             (cons res (filter-map f (cdr lst)))
             (filter-map f (cdr lst)) ))
      '() ))

(define (copy-list lst)
  (if (null? lst)
      lst
      (cons (car lst) (copy-list (cdr lst))) ))

(define (sum-list lst)
  (if (pair? lst)
      (+ (car lst) (sum-list (cdr lst)))
      0))


(define (remove-one-equal what lst)
  (cond ((null? lst) lst)
	((equal? what (car lst)) (cdr lst))
	(else (cons (car lst) (remove-one-equal what (cdr lst)))) ))

(define (remove-eq el lst)
  (cond ((null? lst) lst)
	((eq? (car lst) el)
	   (remove-eq el (cdr lst)) )
	(else
	   (cons (car lst) (remove-eq el (cdr lst))) )))

(define (remove-at-place place lst)
  (cond ((null? lst) lst)
	((eq? place lst) (cdr lst))
	(else
	   (cons (car lst) (remove-at-place place (cdr lst))) )))

(define (replace-one-eq what with lst)
  (if (null? lst)
      '()
      (if (eq? what (car lst))
	  (cons with (cdr lst))
	  (cons (car lst) (replace-one-eq what with (cdr lst))) )))

(define (replace-nth where nr with)
  (if (eq? 1 nr)
      (cons with (cdr where))
      (cons (car where) (replace-nth (cdr where) (- nr 1) with)) ))

(define (remove-at-place! place lst)
  (if (null? lst)
      '()
      (if (null? (cdr lst))
	  '()
	  (do ((lstpart lst (cdr lstpart)))
	      ((and (pair? (cdr lstpart))
		    (eq? place (cdr lstpart)) )
	       (set-cdr! lstpart (cddr lstpart)) )))))


(define *make-inside-subset-tmp* '())
(define *make-inside-subset-tmp-lst* '())

(define (make-inside-subset lst where)
  (set! *make-inside-subset-tmp* '())
  (set! *make-inside-subset-tmp-lst* lst)
  (make-inside-subset-aux! where)
  *make-inside-subset-tmp*)

(define (make-inside-subset-aux! where)
  (if (not (null? *make-inside-subset-tmp-lst*))
      (if (pair? where)
	  (for-each make-inside-subset-aux! where)
	  (begin
	    (do ((lstpart *make-inside-subset-tmp-lst* (cdr lstpart)))
		((null? lstpart))
	      (if (eq? (car lstpart) where)
		  (begin
		    (set! *make-inside-subset-tmp*
			  (cons (car lstpart)
				*make-inside-subset-tmp*))
		    (set! *make-inside-subset-tmp-lst*
			  (remove-at-place
			    lstpart *make-inside-subset-tmp-lst*)))))))))

;(define (reverse lst)
;  (let ((res '()))
;    (do ()
;	((null? lst) res)
;      (set! res (cons (car lst) res))
;      (set! lst (cdr lst)) )))


;=======================================================================
;
;                          EOF resbase.scm
;
;           Look for: resmain.scm, resio.scm
;
;
;============================ EOF ========================================

