;======================================================================
;
;                          resprops.scm
;
;                   Gandalf version c-2.6 r1
;
;  Sem-type model building procedures.
;
;                  Copyright (c) Tanel Tammet
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 1, or (at your option)
; any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; The author can be reached at tammet@staff.ttu.ee, tanel@cs.ioc.ee or
; tammet@cs.chalmers.se
;
;
;=======================================================================

; ---------- compilation specifics --------------

(define compile-inline
 '(props-eq-pred? props-is-pred? props-pos-pred? props-pure-fun) )

(define *props-dynamic-vectors* #t)

; --------- block for case where *props-dynamic-vectors* is #f -----------

;(define compile-stable-vectors
;'(
;*props-fun-arities-vector*
;*props-fun-originals-vector*
;*props-fun-originalsymb-vector*
;*props-fun-ispred-vector*
;
;*props-cell-cterm-vector*
;*props-cell-occurrences-vector*
;*props-cell-range-vector*
;*props-cell-order-vector*
;*props-fun-basenr-vector*
;
;*props-nterm-vector*
;*props-nterm-originals-vector*
;*props-nterm-matchcells-vector*
;*props-nterm-clauses-vector*
;*props-nterm-fun-basenr-vector*))


; ------ basic objects and counters ----------

(define *props-eq-fun* 0)    ; props eq symbol function
(define *props-domain-size* 0)  ; size of a domain. domain elements are ints from 0...size-1
(define *props-ground-clauses* '())
(define *props-nonground-clauses* '())

(define *props-max-allowed-cells* 50000)
(define *props-max-allowed-clauses* 70000)

(define *props-fun-arities-vector* (make-vector 10000 #f))
(define *props-fun-originals-vector* (make-vector 10000 #f))
(define *props-fun-originalsymb-vector* (make-vector 10000 #f))
(define *props-fun-ispred-vector* (make-vector 10000 #f))

; --------- block for case where *props-dynamic-vectors* is #f -----------

;(define *props-cell-nr* 0)
;
;(define *props-cell-cterm-vector* (make-vector 10000 #f))
;(define *props-cell-occurrences-vector* (make-vector 10000 '()))
;(define *props-cell-range-vector* (make-vector 10000 '()))
;(define *props-cell-order-vector* (make-vector 10000 '()))
;(define *props-fun-basenr-vector* (make-vector 10000 #f))
;
;(define *props-nterm-nr* 0)
;
;(define *props-nterm-vector* (make-vector 10000 '()))
;(define *props-nterm-originals-vector* (make-vector 10000 '()))
;(define *props-nterm-matchcells-vector* (make-vector 10000 '()))
;(define *props-nterm-clauses-vector* (make-vector-10000 '()))
;(define *props-nterm-fun-basenr-vector* (make-vector 10000))

; --------- block for case where *props-dynamic-vectors* is #t -----------

(define *props-cell-nr* 0)

(define *props-cell-cterm-vector* '())
(define *props-cell-occurrences-vector* '())
(define *props-cell-range-vector* '())
(define *props-cell-order-vector* '())
(define *props-fun-basenr-vector* '())

(define *props-nterm-nr* 0)

(define *props-nterm-vector* '())
(define *props-nterm-originals-vector* '())
(define *props-nterm-matchcells-vector* '())
(define *props-nterm-clauses-vector* '())
(define *props-nterm-fun-basenr-vector* '())

; --------------- fine-grained time limits -----------

(define *props-timecheck-counter* 0)

;====================================================================
;
;
;
;====================================================================


; props clause structure:
;
; (literal1 ... literaln)
;
; props literal structure:
;
; (fun info arg1 arg2 ... argn)
;
; where:
;
; fun is a negative integer in case of a negative literal, else positive integer
; info is a list of infoelements
; arg is an integer in range (0 ... size-1) if domain element, else a term

; fun encoding:
;
; HIGH 30          24              16              8             1 LOW
;     | , , , , , | , , , , , , , | , , , , , , , | , , , , , , , |
;    sgn|     spare bits      |c|l|      fun symbol number        |
;
; where:
;
; sgn = 0 or 1: 0 means a positive predicate, 1 negative
;        Needs shift 29
;
; l = 0 or 1: literal/proper term indicator switch:
;   0: proper term
;   1: literal
;
; c = 0 or 1: to distinguish vars from constants in nonground case
;   0: var
;   1: constant
;
; props term structure is analogous to literal
;
; info list structure:
; (parentpointer)
;  parentpointer is a pointer to the superterm.
;
;--------------------------------------------------------------------------------

(define *props-purefun-mask* 65535)
(define *props-isneglit-mask* 268435456)
 ; (shiftleft 1 28)
(define *props-islit-mask* 65536)
 ; (shiftleft 1 16)
(define *props-isconst-mask* 131072)
 ;  (shiftleft 1 17)
(define *props-funnr-mask* 65535)


;------------------------------------------------------------------------
;
;          a top level block
;
;-------------------------------------------------------------------------

(define *props-selection-strategy* 1)

(define *props-debug* #f)
(define *props-small-debug* #f)
(define *props-debug-selectcell* #f)

(define *props-remove-terms-strat* #t)
(define *props-use-nterms-strat* #t)

(define *props-split-sel-count* 0)
(define *props-assign-count* 0)
(define *props-restrict-count* 0)
(define *props-rewrite-count* 0)
(define *props-nterm-assign-count* 0)
(define *props-nterm-restrict-count* 0)
(define *props-nterm-derive-count* 0)
(define *props-clause-count* 0)

(define *props-stop-search-flag* #f)



(define (props-build-model predclauses)
  (props-build-model-aux predclauses) )


(define (props-build-model-0 predclauses)
  (set! *props-selection-strategy* 0)
  (props-build-model-aux predclauses) )


(define (props-build-model-1 predclauses)
  (set! *props-selection-strategy* 1)
  (props-build-model-aux predclauses) )


(define (props-build-model-2 predclauses)
  (set! *props-selection-strategy* 2)
  (props-build-model-aux predclauses) )


(define (props-build-model-3 predclauses)
  (set! *props-selection-strategy* 3)
  (props-build-model-aux predclauses) )


(define (props-build-model-4 predclauses)
  (set! *props-selection-strategy* 4)
  (props-build-model-aux predclauses) )


(define (props-build-model-5 predclauses)
  (set! *props-selection-strategy* 5)
  (props-build-model-aux predclauses) )


(define (props-build-model-6 predclauses)
  (set! *props-selection-strategy* 6)
  (props-build-model-aux predclauses) )



(define (props-build-model-aux predclauses)
  (let ((stopflag #f)
        (domainsize 0)
        (tmp '())
        (tmpnr 0)
        (nongroundpropclauses '())
        (groundpropclauses '())
        (res #f) )

    (cond
      ((eq? *props-selection-strategy* 0)
         (newline)(display "sem-type finite model building: maximise lnh, trivial order") )
      ((eq? *props-selection-strategy* 1)
         (newline)(display "sem-type finite model building: maximise lnh, standard order") )
      ((eq? *props-selection-strategy* 2)
         (newline)(display "sem-type finite model building: maximise lnh, occurrences order") )
      ((eq? *props-selection-strategy* 3)
         (newline)(display "sem-type finite model building: maximise lnh, look-ahead order") )
      ((eq? *props-selection-strategy* 4)
         (newline)(display "sem-type finite model building: pure occurrences order") )
      ((eq? *props-selection-strategy* 5)
         (newline)(display "sem-type finite model building: pure look-ahead order") )
      (else
         (newline)(display "sem-type finite model building: pure look-ahead order") ))
    (newline)

    (set! *props-timecheck-counter* 0)
    (set! *props-stop-search-flag* #f)
    (set! *prop-time-ran-out* #f)
    (set! *given-checkpoint-keptnrs* '())
    (set! *slow-start-time*  (get-check-time 0))
    (store-time-checkpoint! 0)
    (extract-pred-data-to-prop predclauses)

    (if *props-debug*
        (begin
         (newline)
         (display "props-build-model called")(newline)
         (display "predclauses: ")(newline)
         (for-each (lambda(x)(newline)(display x)) predclauses)(newline) ))

    (set! *props-eq-fun* #f) ; will get a number if eq present
    (set! nongroundpropclauses (props-make-nonground-clauses predclauses))

    (if *props-debug*
        (begin
         (newline)
         (display "nonground prop clauses: ")(newline)
         (for-each (lambda(x)(newline)(display x)) nongroundpropclauses)
         (newline)
         (for-each (lambda(x)(newline)(print-props-nonground-clause x)) nongroundpropclauses)
         (newline) ))

    (cond
      ((null? *prop-orig-funs*)
        ; no function symbols: interpret all constants to themselves
        '())
      ; function symbols present: full domain size iteration
     (else
      (do ((domainsize 1 (+ 1 domainsize)))
          ((or stopflag
               *props-stop-search-flag*
               (eq? res #t)
               ;(> domainsize 2)
               (if (> (- (get-check-time 1) *slow-start-time*)
	              *one-pass-time*)
	           (begin
	               (store-time-checkpoint! 110)
	               (set! *prop-time-ran-out* #t)
	               #t)
	           #f))
	      (if (or *prop-time-ran-out* *props-stop-search-flag*)
    	          (begin
	             (if (> *display-info-level* 38)
        	         (begin (newline)
	                        (display "allocated model search time has run out")
		                (newline) ))
	              #f)
                  res))
        (if #t ; (> *display-info-level* 38)
            (begin (newline)
	           (display "starting search for domain size ")(display domainsize)
	           ;(newline)
	           (newline) ))

        (set! *props-split-sel-count* 0)
        (set! *props-assign-count* 0)
        (set! *props-restrict-count* 0)
        (set! *props-rewrite-count* 0)
        (set! *props-nterm-assign-count* 0)
        (set! *props-nterm-restrict-count* 0)
        (set! *props-nterm-derive-count* 0)
        (set! *props-clause-count* 0)

        (set! groundpropclauses
              (props-saturate-clauses nongroundpropclauses domainsize) )

        ;(newline)
        ;(display "clauses saturated:")
        ;(if (pair? groundpropclauses)
        ;    (display (length groundpropclauses))
        ;    (display groundpropclauses) )
        ;(newline)

        (if (not (eq? groundpropclauses #t))
          (begin
           (if (and *props-debug* res)
               (begin
                 (newline)
                 (display "ground prop clauses initially made: ")(newline)
                 ;(for-each (lambda(x)(newline)(display x)) groundpropclauses)
                 ;(newline)
                 (for-each (lambda(x)(newline)(print-props-ground-clause x))
                            groundpropclauses)
                 (newline) ))

           (if (not *props-stop-search-flag*)
               (props-init-indexes domainsize) )

           ;(newline)
           ;(display "indexes inited")
           ;(newline)

           (if (not *props-stop-search-flag*)
               (props-store-initial-occurrences groundpropclauses domainsize) )

           ;(newline)
           ;(display "initial occurrences stored")
           ;(newline)

           (if *props-debug*
               (begin
                (newline)
                (display "ground prop clauses after occ storage: ")(newline)
                ;(for-each (lambda(x)(newline)(display x)) groundpropclauses)
                ;(newline)
                (for-each (lambda(x)(newline)(print-props-ground-clause x)) groundpropclauses)
                (newline)
                (props-print-vectors) ))

           ;(abort)

           ; initialise equality cell ranges if equality present
           (if (not *props-stop-search-flag*)
               (props-init-equality domainsize) )

           ;(newline)
           ;(display "equality inited")
           ;(newline)

           ; propagate all the determined values of cells
           (if (not *props-stop-search-flag*)
               (set! res (props-initial-propagate groundpropclauses domainsize)) )


           ;(newline)
           ;(display "initial propagation done")
           ;(newline)

           ;(newline)
           ;(display "initial propagation returns: ")(display res)
           ;(newline)

           (if (and res *props-debug*)
               (begin
                 (newline)
                 (display "clauses after initial propagation of cells: ")
                 (for-each (lambda(x)(newline)(print-props-ground-clause x)) groundpropclauses)
                 (newline)
                 (props-print-vectors) ))

          ;(newline)(display "initial analysis starts")(newline)

          (if (and (not (eq? res #f))
                   (not *props-stop-search-flag*) )
              (set! res
                   (props-initial-analyze-clauses groundpropclauses domainsize)) )

          ;(newline)
          ;(display "initial analysis done")
          ;(newline)

          ;(newline)
          ;(display "initial analysis returns: ")(display res)
          ;(newline)


          (if (or *props-small-debug* *props-debug*)
               (begin
                 (newline)
                 (display "clauses after initial analysis: ")
                 (for-each (lambda(x)(newline)(print-props-ground-clause x)) groundpropclauses)
                 (newline)
                 (props-print-vectors) ))

           ; make cell measures for selecting a next cell to split
           ; also, order the cells
           (if (not (eq? *props-selection-strategy* 0))
               (begin
                 (props-make-cell-measures! domainsize) ))
                 (props-sort-cells! domainsize)

           ;(newline)
           ;(display "cell measures created")
           ;(newline)

           ;(newline)(display "search starts")(newline)

           ; do splitting search if no immediate contradiction found
           (if (and (not (eq? res #f))
                    (not *props-stop-search-flag*) )
               (set! res
                    (props-split-cells-incremental
                         groundpropclauses domainsize 0 '() 1) ))

           ;(newline)
           ;(display "cells split incr")
           ;(newline)

           (if *props-debug*
               (begin
                (newline)
                (display "clauses after split cells: ")
                (for-each (lambda(x)(newline)(print-props-ground-clause x)) groundpropclauses)
                (newline)
                (display "vectors after split cells: ")
                (newline)
                (props-print-vectors) ))

           (if *prop-time-ran-out*
               (set! res #f) )

           (if (eq? res #t)
               (begin
                 (set! tmp (props-check-model nongroundpropclauses domainsize))
                 (if tmp
                     (begin
                       (newline)
                       (display "**** SATISFIABLE ****")
                        #t)
                      (begin
                        (set! *props-stop-search-flag* #t)
                        (set! res #f) )))
               (props-show-size-end-stats domainsize)  )) ))))))

; props-initial-propagate propagates all the initially determined
; cell values (single range element)

(define (props-initial-propagate clauses domainsize)
  (let ((res '())
        (range '())
        (selcell '())
        (value '())
        (stack (list '()))
        (assignres '())
        (stopflag #f) )

    ;(newline)(display "initial propagation starts")(newline)

    (do ((i 0 (+ 1 i)))
        ((or stopflag (>= i *props-cell-nr*))
           res)

      ;(newline)(display "i: ")(display i)

      (set! range (vector-ref *props-cell-range-vector* i))

      ;(newline)(display "range: ")(display range)

      (if (or (integer? range)
              (null? (cdr range)) )
        (begin
           (if (integer? range)
               (set! value range)
               (set! value (car range)) )
           (set! selcell i)

           (if *props-debug*
               (begin
                 (newline)
                 (display "propagating assignment: cell ")(display selcell)
                 (display " becomes ")(display value)
                 (newline) ))

           (set! assignres (props-propagate-cell! selcell stack domainsize))

           (if *props-debug*
               (begin
                 (newline)
                 (display "assignres: ")(display assignres)
                 (newline)
                 (display "clauses after assignment: ")
                 (for-each (lambda(x)(newline)(print-props-ground-clause x)) clauses)
                 (newline)
                 (display "vectors after assignment: ")(newline)
                 (props-print-vectors) ))

           (cond
              ((eq? assignres #t)
               ; satisfiability found here
                 (set! stopflag #t)
                 (set! res assignres)
                 (newline)
                 ;(display "**** SATISFIABLE ****")(newline)
                 (display "satisfiability found after initial propagation! ")
                 #t)
              ((eq? assignres #f)
               ; unsatisfiability found here: clause set really is unsat.
                 (set! stopflag #t)
                 (set! res #f) )))))))


(define (props-init-equality domainsize)
  (let ((cell 0)
        (vec '()) )
    (if (not (eq? *props-eq-fun* #f))
        (begin
          (set! vec (make-vector 4 '()))
          (vector-set! vec 1 *props-eq-fun*)
          (do ((l 0 (+ 1 l)))
              ((>= l domainsize))
            (vector-set! vec 2 l)
            (do ((r 0 (+ 1 r)))
                ((>= r domainsize))
              (vector-set! vec 3 r)
              (set! cell (props-cterm-index-nr vec domainsize))
              (if (eq? l r)
                  (vector-set! *props-cell-range-vector* cell (list 1))
                  (vector-set! *props-cell-range-vector* cell (list 0)) )))))))


(define (props-initial-analyze-clauses clauses domainsize)
  (let ((stopflag #f)
        (clause '())
        (stack (list '()))
        (res '()) )
      (do ((clausespart clauses (cdr clausespart)))
          ((or stopflag (null? clausespart))
               (if stopflag
                   #f
                   '() ))
        (set! clause (car clausespart))
        (set! res (props-analyze-clause! clause domainsize stack))
        (cond
          ((eq? res #f)
             (set! stopflag #t) )))))


(define (props-split-cells-incremental clauses domainsize maxsplitvalue touched depth)
  (let ((selcell '())
        (selcellorig '())
        (selcellrange '())
        (selcelloccurrences '())
        (selcelllength 0)
        (selcellispredflag #f)
        (value 0)
        (stopflag #f)
        (res '())
        (assignres '())
        (recres '())
        (stack (list '()))
        (tmp '()) )

   (if *props-debug*
       (begin
        (newline)
        (display "props-split-cells-incremental called with depth: ")
        (display depth)
        (newline) ))

   (if (> (- (get-check-time 1) *slow-start-time*)
              *one-pass-time*)
       (begin
          (store-time-checkpoint! 111)
          (set! *prop-time-ran-out* #t)
          (set! *props-stop-search-flag* #t) ))

   (if (not *props-stop-search-flag*)
       (set! selcell (props-select-split-cell clauses domainsize maxsplitvalue depth)) )

   (if (or *props-small-debug* *props-debug*)
       (begin
         (newline)
         (do ((m 0 (+ 1 m)))
             ((>= m depth))
           (display "| ") )
         (display "selcell:")(display selcell)
         ))

   (if (or (not selcell) *props-stop-search-flag*)

       ; either stop or:
       ; nothing to select, hence all cells have values, no contradiction so far,
       ; hence model is found (should be checked specially, though!)

       (if *props-stop-search-flag*
           ; time out: failure
           (begin
              ;(newline)
              ;(display "timeout")
              ;(newline)
              #f)
           ; selcell not found: success!
           (begin
             (props-show-success-stats domainsize depth)
             #t))

       ; a cell was found for splitting

       (begin
        (set! *props-split-sel-count* (+ 1 *props-split-sel-count*))
        (set! selcellorig (vector-ref *props-cell-cterm-vector* selcell))
        (set! selcellrange (vector-ref *props-cell-range-vector* selcell))
        (set! selcelloccurrences (vector-ref *props-cell-occurrences-vector* selcell))
        (set! selcelllength (vector-length selcellorig))
        (set! selcellispredflag (props-term-is-surely-lit? selcellorig))
        ; update maxsplitvalue with the args of the cell
        (if selcellispredflag
            (set! tmp -1)
            (set! tmp (props-cterm-max-arg selcellorig)) )
        (if (>= tmp maxsplitvalue) (set! maxsplitvalue (+ 1 tmp)))
        ; print stuff for debugging
        (if (or *props-small-debug* *props-debug*)
            (begin
              (newline)
              (do ((m 0 (+ 1 m)))
                  ((>= m depth))
                  (display "| ") )
              (display "maxsplitvalue: ")(display maxsplitvalue)))
        ; loop over range of cell
        (set! stopflag #f)
        (set! res '())
        (do ((rpart selcellrange (cdr rpart)))
            ((or (null? rpart) stopflag *props-stop-search-flag*)
               (if stopflag
                   res
                   '()))
         (set! value (car rpart))
         ; use only values which are equal or less than maxsplitvalue
         (if
         (<= value maxsplitvalue)
          (begin
           (if *props-debug*
               (begin
                 (display "clauses before assignment: ")
                 (for-each (lambda(x)(newline)(print-props-ground-clause x)) clauses)
                 (newline)
                 (display "vectors before assignment: ")(newline)
                 (props-print-vectors)
                 (newline) ))
           (if (or *props-small-debug* *props-debug*)
               (begin
                 (newline)
                 (do ((m 0 (+ 1 m)))
                     ((>= m depth))
                    (display "| ") )
                 (display "value in loop: ")(display value)
                 ))

           (set! assignres (props-assign-cell! selcell value stack domainsize))

           ;(newline)
           ;(display "cell assigned:")
           ;(display selcell) (display " value: ")(display value)
           ;(newline)

           (if (or *props-small-debug* *props-debug*)
               (begin
                 (newline)
                 (do ((m 0 (+ 1 m)))
                     ((>= m depth))
                    (display "| ") )
                 (display "assignres: ")(display assignres)
                 ))

           (if *props-debug*
               (begin
                 (newline)
                 (display "clauses after assignment: ")
                 (for-each (lambda(x)(newline)(print-props-ground-clause x)) clauses)
                 (newline)
                 (display "vectors after assignment: ")(newline)
                 (props-print-vectors) ))

           (cond
              ((eq? assignres #t)
               ; satisfiability found here
                 (set! stopflag #t)
                 (set! res assignres)
                 (newline)
                 ;(display "**** SATISFIABLE ****")(newline)
                 (display "satisfiability found after assignment! ")
                 #t)
              ((eq? assignres #f)

                   (if #f ; (or *props-small-debug* *props-debug*)
                       (begin
                        (newline)
                        (display "clauses after assignment: ")
                        (for-each (lambda(x)(newline)(print-props-ground-clause x)) clauses)
                        (newline)
                        (display "vectors after assignment: ")(newline)
                        (props-print-vectors) ))

               ; unsatisfiability found here: no point in recursing, try other values
                 (props-restore-stack! assignres stack) )
              (else
                ; no satisfiability or unsat found after assignment: recurse splitting lower
                 (set! recres (props-split-cells-incremental
                                clauses domainsize
                                (if (and (not selcellispredflag)
                                         (>= value maxsplitvalue) )
                                    (+ 1 value)
                                    maxsplitvalue)
                                (if (not (memq value touched))
                                    (cons value touched)
                                    touched)
                                (+ 1 depth)))

                 (if (or *props-small-debug* *props-debug*)
                     (begin
                       (newline)
                       (do ((m 0 (+ 1 m)))
                           ((>= m depth))
                         (display "| ") )
                       (display "recres: ")(display recres)
                        ))

                 (if (eq? recres #t)
                     ; satisfiability found lower during recursion
                     (begin
                       (set! stopflag #t)
                       (set! res recres) )
                     ; no satisfiability found yet: restore stack
                     (begin
                       ; (props-restore-stack recres)
                       (props-restore-stack! assignres stack) )))) )))))))


(define (props-show-success-stats domainsize depth)
    (if #t
       (begin
          (newline)
          ;(display "**** SATISFIABLE ****")(newline)
          (display "Cell selection found all cells to have assigned values!")(newline)
          (newline)
          (display "model size: ")(display domainsize)
          (newline)
          (display "nr of cells: ")(display *props-cell-nr*)
          (newline)
          (display "model found at split search depth: ")
          (display depth)
          (newline)
          (display "split selections done: ")(display *props-split-sel-count*)
          (newline)
          (display "assignments done: ")(display *props-assign-count*)
          (newline)
          (display "restrictions done: ")(display *props-restrict-count*)
          (newline)
          (display "terms rewritten: ")(display *props-rewrite-count*)
          (newline)
          (display "nterms assigned: ")(display *props-nterm-assign-count*)
          (newline)
          (display "nterms restricted: ")(display *props-nterm-restrict-count*)
          (newline)
          (display "new derivations from nterm: ")(display *props-nterm-derive-count*)
          (newline)
          (display "clauses generated: ")(display *props-clause-count*)
          (newline)

          (newline)
          (display "MODEL STARTS")
          (newline)
          (do ((i 0 (+ 1 i)))
              ((or (>= i *props-cell-nr*)
                   (not (vector? *props-cell-cterm-vector*)) ))
            (print-props-model-term (vector-ref *props-cell-cterm-vector* i))
            (display "=")
            ;(display  (vector-ref *props-cell-cterm-vector* i))
            (if (and #t ;(pair? (cdr (vector-ref *props-cell-cterm-vector* i)))
                     (vector-ref *props-fun-ispred-vector*
                        (logand *props-funnr-mask*
                           (vector-ref
                             (vector-ref *props-cell-cterm-vector* i)
                             1))))
                (begin
                  (if (eq? 1 (car (vector-ref *props-cell-range-vector* i)))
                      (display "t")
                      (display "f") ))
                (display (car (vector-ref *props-cell-range-vector* i))) )
            (newline) )
         (display "MODEL ENDS")
         (newline)

    ;(do ((i 0 (+ 1 i)))
    ;     ((or (>= i *props-fun-counter*)
    ;          (not (vector? *props-fun-originalsymb-vector*))
    ;          (not (vector? *props-fun-arities-vector*))
    ;          (not (vector?  *props-fun-basenr-vector*)) ))
    ;   (display i)(display " : ")
    ;   (display " orig=")(display (vector-ref *props-fun-originalsymb-vector* i))
    ;   (display " orignr=")(display (vector-ref *props-fun-originals-vector* i))
    ;   (display " ispred=")(display (vector-ref *props-fun-ispred-vector* i))
    ;   (display " arity=")(display (vector-ref *props-fun-arities-vector* i))
    ;   (display " base=")(display (vector-ref *props-fun-basenr-vector* i))
    ;   (newline) )
    ; (newline)

          ;(newline)
          ;(props-print-vectors)
          ;(quit)
     )))

(define (props-show-size-end-stats domainsize)
  (if (and (or *props-debug* *props-small-debug*) (not *props-stop-search-flag*))
      (begin
           (newline)
           (display "nr of cells: ")(display *props-cell-nr*)
           (newline)
           (display "split selections done: ")(display *props-split-sel-count*)
           (newline)
           (display "assignments done: ")(display *props-assign-count*)
           (newline)
           (display "restrictions done: ")(display *props-restrict-count*)
           (newline)
           (display "terms rewritten: ")(display *props-rewrite-count*)
           (newline)
           (display "nterms assigned: ")(display *props-nterm-assign-count*)
           (newline)
           (display "nterms restricted: ")(display *props-nterm-restrict-count*)
           (newline)
           (display "new derivations from nterm: ")(display *props-nterm-derive-count*)
           (newline)
           (display "clauses generated: ")(display *props-clause-count*)
           (newline)
       )))

(define (make-touched-list oldtouched newcell domainsize)
  (let ((touched '())
        (touchedlen 0)
        (cellrange '())
        (cterm '())
        (celllen 0)
        (elem 0)
        (stopflag #f) )

   ; use old touched list

   (set! touched oldtouched)
   (set! touchedlen (length touched))

   ; now go through the newcell elements

   (set! cterm (vector-ref *props-cell-cterm-vector* newcell))
   (set! celllen (vector-length cterm))
   (do ((j 2 (+ 1 j)))
       ((or ;stopflag
           (>= touchedlen domainsize)
           (>= j celllen)))
     (set! elem (vector-ref cterm j))
     (if (not (memq elem touched))
         (begin
            (set! touched (cons elem touched))
            (set! touchedlen (+ 1 touchedlen)) )))

   ; finally go through all the existing assignments

   (do ((i 0 (+ 1 i)))
       ((or ;stopflag
           (>= touchedlen domainsize)
           (>= i *props-cell-nr*))
          touched)
    (set! cellrange (vector-ref *props-cell-range-vector* i))
    (if (or (integer? cellrange)
            (null? (cdr cellrange)) )
        (begin
           (if (pair? cellrange)
               (set! cellrange (car cellrange)) )
           (if (not (memq cellrange touched))
               (begin
                  (set! touched (cons cellrange touched))
                  (set! touchedlen (+ 1 touchedlen)) ))
           (set! cterm (vector-ref *props-cell-cterm-vector* i))
           (set! celllen (vector-length cterm))
           (do ((j 2 (+ 1 j)))
               ((or ;stopflag
                    (>= touchedlen domainsize)
                    (>= j celllen)))
             (set! elem (vector-ref cterm j))
             (if (not (memq elem touched))
                 (begin
                  (set! touched (cons elem touched))
                  (set! touchedlen (+ 1 touchedlen)) ))))))))

;---------------------------------------------------------------
;
;         cell selection block
;
;----------------------------------------------------------------


(define (props-select-split-cell clauses domainsize maxsplitvalue depth)

 (let ((i 0))

  (if *props-debug-selectcell*
     (begin
       (newline)(display  "***** props-select-split-cell called ***** ")
       (newline)(display "maxsplitvalue: ")(display maxsplitvalue)
       (display " domainsize: ")(display domainsize)
       (newline)
       (display "ordered cell vectors used in props-select-split-cell: ")
       (newline)
       (do ((j 0 (+ 1 j)))
         ((>= j *props-cell-nr*))
        (set! i (vector-ref *props-cell-order-vector* j))
        (display j)(display " : ")
        (display i) (display " : ")
        (display "orig=")(print-props-ground-term (vector-ref *props-cell-cterm-vector* i))
        (display " cterm=")(display (vector-ref *props-cell-cterm-vector* i))
        (display " range=")(display (vector-ref *props-cell-range-vector* i))
        (display " occs=")(display (length (vector-ref *props-cell-occurrences-vector* i)))
        ;(newline)
        ;(for-each (lambda(x)(print-props-ground-term x)(newline))
        ;          (vector-ref *props-cell-occurrences-vector* i) )
        (newline) )))

       ;(display "cell order: ")(newline)
       ;(do ((i 0 (+ 1 i)))
       ;    ((>= i *props-cell-nr*))
       ; (display i)(display " : ")
       ; (display (vector-ref *props-cell-order-vector* i))
       ; (newline) )

  (cond
    ; 0: maximise lnh, use trivial order (ordering turned off elsewhere)
    ((eq? *props-selection-strategy* 0)
       (if (< maxsplitvalue (- domainsize 1))
           (props-select-split-cell-restricted-initial
                 clauses domainsize maxsplitvalue depth)
           (props-select-split-cell-unrestricted
                 clauses domainsize maxsplitvalue depth) ))
    ; 1: maximise lnh, use standard order
    ((eq? *props-selection-strategy* 1)
       (if (< maxsplitvalue (- domainsize 1))
           (props-select-split-cell-restricted-initial
                 clauses domainsize maxsplitvalue depth)
           (props-select-split-cell-unrestricted
                 clauses domainsize maxsplitvalue depth) ))
   ; 2: maximise lnh, use occurrences order
   ((eq? *props-selection-strategy* 2)
       (if (< maxsplitvalue (- domainsize 1))
           (props-select-split-cell-restricted-initial
               clauses domainsize maxsplitvalue depth)
           (props-select-split-cell-occurrences
               clauses domainsize maxsplitvalue depth) ))
   ; 3: maximise lnh, use look-ahead order
   ((eq? *props-selection-strategy* 3)
       (if (< maxsplitvalue (- domainsize 1))
           (props-select-split-cell-restricted-initial
              clauses domainsize maxsplitvalue depth)
           (if (> (props-open-occ-cell-count) 2)
               (props-select-split-cell-preassign domainsize)
               (props-select-split-cell-unrestricted
                   clauses domainsize maxsplitvalue depth) )))
   ; 4: use pure occurrences order
   ((eq? *props-selection-strategy* 4)
         (props-select-split-cell-occurrences
               clauses domainsize maxsplitvalue depth) )
   ; 5: use pure look-ahead order
   (else
      (if (> (props-open-occ-cell-count) 2)
          (props-select-split-cell-preassign domainsize)
          (props-select-split-cell-unrestricted
                   clauses domainsize maxsplitvalue depth) )))))


(define (props-select-split-cell-restricted-initial-experiment clauses domainsize maxsplitvalue depth)
  (let ((stopflag #f)
        (i 0)
        (tmp '())
        (tmp2 '())
        (nr 0)
        (maxarg -1)
        (bestnr domainsize)
        (foundcell #f)
        (bestcell #f) )

     (if *props-debug-selectcell*
        (begin
          (newline)(display "props-select-split-cell-restricted-initial called:") ))

    ; find an open cell with a least range that has max arg less than maxsplitvalue

    (set! bestcell #f)
    (set! foundcell #f)
    (do ((j 0 (+ 1 j)))
        ((or stopflag (>= j *props-cell-nr*))
            (if bestcell
                bestcell
                (if foundcell
                    foundcell
                    (props-select-split-cell-restricted-secondary
                       clauses domainsize maxsplitvalue depth) )))
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! tmp (vector-ref *props-cell-range-vector* i))
     (if (and (pair? tmp)
              (not (null? (cdr tmp))) )
         (begin
            (set! nr (length tmp))
            (set! tmp2 (vector-ref *props-cell-cterm-vector* i))

            (if (props-term-is-surely-lit? tmp2)
                (set! maxarg -1)
                (set! maxarg (props-cterm-max-arg tmp2)) )

            ; (set! maxarg (props-cterm-max-arg tmp2))

            (if (and (< nr bestnr)
                     ; (< (props-cterm-max-arg tmp2) maxsplitvalue)
                     (< maxarg maxsplitvalue) )
                (begin
                  (if (props-term-is-surely-lit? tmp2)
                      (if (not foundcell) (set! foundcell i))
                      (begin
                        (if (< nr 3) (set! stopflag #t))
                        (set! bestcell i)
                        (set! bestnr nr) )))))))))



(define (props-select-split-cell-restricted-initial clauses domainsize maxsplitvalue depth)
  (let ((stopflag #f)
        (i 0)
        (tmp '())
        (tmp2 '())
        (nr 0)
        (maxarg -1)
        (bestnr domainsize)
        (foundcell #f)
        (bestcell #f) )

     (if *props-debug-selectcell*
        (begin
          (newline)(display "props-select-split-cell-restricted-initial called:") ))

    ; find an open cell with a least range that has max arg less than maxsplitvalue

    (set! bestcell #f)
    (set! foundcell #f)
    (do ((j 0 (+ 1 j)))
        ((or stopflag (>= j *props-cell-nr*))
            (if bestcell
                bestcell
                (if foundcell
                    foundcell
                    (props-select-split-cell-restricted-secondary
                       clauses domainsize maxsplitvalue depth) )))
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! tmp (vector-ref *props-cell-range-vector* i))
     (if (and (pair? tmp)
              (not (null? (cdr tmp))) )
         (begin
            (set! nr (length tmp))
            (set! tmp2 (vector-ref *props-cell-cterm-vector* i))

            ;(if (props-term-is-surely-lit? tmp2)
            ;    (set! maxarg -1)
            ;    (set! maxarg (props-cterm-max-arg tmp2)) )

            (set! maxarg (props-cterm-max-arg tmp2))

            (if (and (< nr bestnr)
                     (< (props-cterm-max-arg tmp2) maxsplitvalue) )
                     ; (< maxarg maxsplitvalue)
                (begin
                  (if (props-term-is-surely-lit? tmp2)
                      (if (not foundcell) (set! foundcell i))
                      (begin
                        (if (< nr 3) (set! stopflag #t))
                        (set! bestcell i)
                        (set! bestnr nr) )))))))))



(define (props-select-split-cell-restricted-secondary clauses domainsize maxsplitvalue depth)
  (let ((stopflag #f)
        (i 0)
        (tmp '())
        (tmp2 '())
        (nr 0)
        (bestnr (+ 2 domainsize))
        (maxarg 0)
        (bestcell #f)
        (foundcell #f) )

     (if *props-debug-selectcell*
        (begin
          (newline)(display "props-select-split-cell-restricted-secondary called:") ))

    ; first find an open cell with a smallest max arg
    (do ((j 0 (+ 1 j)))
        ((or stopflag (>= j *props-cell-nr*)))
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! tmp (vector-ref *props-cell-range-vector* i))
     (if (and (pair? tmp)
              (not (null? (cdr tmp))) )
         (begin
            (set! tmp2 (vector-ref *props-cell-cterm-vector* i))

             ;(if (props-term-is-surely-lit? tmp2)
             ;   (set! nr -1)
             ;   (set! nr (props-cterm-max-arg tmp2)) )
             (set! nr (props-cterm-max-arg tmp2))

            ;(newline)
            ;(display "maxarg of: ")(print-props-ground-term tmp2)(display " is ")
            ;(display nr)
            (if (< nr bestnr)
                (begin
                  (if (< nr 0) (set! stopflag #t))
                  (set! bestnr nr) )))))


    ; now find an open cell with a least range that has this max arg

    (set! maxarg bestnr)

    (if *props-debug-selectcell*
        (begin
         (newline)(display "bestnr found: ")(display bestnr)
         (newline)(display "maxarg found: ")(display maxarg)
         (newline) ))

    (set! bestnr (+ 2 domainsize))
    (set! bestcell #f)
    (set! foundcell #f)
    (set! stopflag #f)
    (do ((j 0 (+ 1 j)))
        ((or stopflag (>= j *props-cell-nr*))
            (if bestcell
                bestcell
                foundcell))
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! tmp (vector-ref *props-cell-range-vector* i))
     (if (and (pair? tmp)
              (not (null? (cdr tmp))) )
              ; (not (null? (vector-ref *props-cell-occurrences-vector* i))) )
         (begin
            (if (not foundcell) (set! foundcell i))
            (set! nr (length tmp))
            (set! tmp2 (vector-ref *props-cell-cterm-vector* i))
            (if (props-term-is-surely-lit? tmp2)
                (set! nr domainsize) )
            (if (and (< nr bestnr)
                     (<= (props-cterm-max-arg tmp2) maxarg) )
                (begin
                  (if (< nr 3) (set! stopflag #t))
                  (set! bestcell i)
                  (set! bestnr nr) )))))))


(define (props-select-split-cell-restricted-secondary-experiment clauses domainsize maxsplitvalue depth)
  (let ((stopflag #f)
        (i 0)
        (tmp '())
        (tmp2 '())
        (nr 0)
        (bestnr (+ 2 domainsize))
        (maxarg 0)
        (bestcell #f)
        (foundcell #f) )

     (if *props-debug-selectcell*
        (begin
          (newline)(display "props-select-split-cell-restricted-secondary called:") ))

    ; first find an open cell with a smallest max arg
    (do ((j 0 (+ 1 j)))
        ((or stopflag (>= j *props-cell-nr*)))
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! tmp (vector-ref *props-cell-range-vector* i))
     (if (and (pair? tmp)
              (not (null? (cdr tmp))) )
         (begin
            (set! tmp2 (vector-ref *props-cell-cterm-vector* i))

             (if (props-term-is-surely-lit? tmp2)
                 (set! nr -1)
                 (set! nr (props-cterm-max-arg tmp2)) )
             ; (set! nr (props-cterm-max-arg tmp2))

            ;(newline)
            ;(display "maxarg of: ")(print-props-ground-term tmp2)(display " is ")
            ;(display nr)
            (if (< nr bestnr)
                (begin
                  (if (< nr 0) (set! stopflag #t))
                  (set! bestnr nr) )))))


    ; now find an open cell with a least range that has this max arg

    (set! maxarg bestnr)

    (if *props-debug-selectcell*
        (begin
         (newline)(display "bestnr found: ")(display bestnr)
         (newline)(display "maxarg found: ")(display maxarg)
         (newline) ))

    (set! bestnr (+ 2 domainsize))
    (set! bestcell #f)
    (set! foundcell #f)
    (set! stopflag #f)
    (do ((j 0 (+ 1 j)))
        ((or stopflag (>= j *props-cell-nr*))
            (if bestcell
                bestcell
                foundcell))
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! tmp (vector-ref *props-cell-range-vector* i))
     (if (and (pair? tmp)
              (not (null? (cdr tmp))) )
              ; (not (null? (vector-ref *props-cell-occurrences-vector* i))) )
         (begin
            (if (not foundcell) (set! foundcell i))
            (set! nr (length tmp))
            (set! tmp2 (vector-ref *props-cell-cterm-vector* i))
            (if (props-term-is-surely-lit? tmp2)
                (set! nr domainsize) )
            (if (and (< nr bestnr)
                     (<= (props-cterm-max-arg tmp2) maxarg) )
                (begin
                  (if (< nr 3) (set! stopflag #t))
                  (set! bestcell i)
                  (set! bestnr nr) )))))))

(define (props-select-split-cell-unrestricted clauses domainsize maxsplitvalue depth)
  (let ((stopflag #f)
        (i 0)
        (tmp '())
        (tmp2 '())
        (nr 0)
        (bestnr (+ 2 domainsize))
        (bestcell #f)
        (foundcell #f) )

    (if *props-debug-selectcell*
       (begin
         (newline)(display "props-select-split-cell-unrestricted called:") ))

    (do ((j 0 (+ 1 j)))
        ((or stopflag (>= j *props-cell-nr*))
            (if bestcell
                bestcell
                foundcell))
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! tmp (vector-ref *props-cell-range-vector* i))
     (if (and (pair? tmp)
              (not (null? (cdr tmp))) )
              ; (not (null? (vector-ref *props-cell-occurrences-vector* i))) )
         (begin
            (if (not foundcell) (set! foundcell i))
            (set! nr (length tmp))
            (set! tmp2 (vector-ref *props-cell-cterm-vector* i))
            (if (and (< nr bestnr)
                     (not (props-term-is-surely-lit? tmp2)) )
                (begin
                  (if (< nr 3) (set! stopflag #t))
                  (set! bestcell i)
                  (set! bestnr nr) )))))))



(define (props-select-split-cell-occurrences clauses domainsize maxsplitvalue depth)
  (let ((stopflag #f)
        (i 0)
        (tmp '())
        (tmp2 '())
        (nr 0)
        (occ '())
        (bestnr -1)
        (bestcell #f)
        (foundcell #f) )

    (if *props-debug-selectcell*
       (begin
         (newline)(display "props-select-split-cell-occurrences called:") ))

    (do ((j 0 (+ 1 j)))
        ((or stopflag (>= j *props-cell-nr*))
            (if bestcell
                bestcell
                foundcell))
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! tmp (vector-ref *props-cell-range-vector* i))
     (set! occ (vector-ref *props-cell-occurrences-vector* i))
     (if (and (pair? tmp)
              (not (null? (cdr tmp)))
              (not (null? occ)) )
         (begin
            (if (not foundcell) (set! foundcell i))
            (set! nr (length occ))
            (if (> nr bestnr)
                (begin
                  (set! bestcell i)
                  (set! bestnr nr) )))))))


(define (props-cterm-max-arg cterm)
  ;(newline)(display "cterm: ")(display cterm)
  (let ((res -1)
        (len (vector-length cterm)) )
    ;(newline)(display "len: ")(display len)
    (do ((i 2 (+ 1 i)))
        ((>= i len)
          res)
      (if (> (vector-ref cterm i) res)
          (set! res (vector-ref cterm i)) ))))


(define *props-cell-measures-vector* '())

(define (props-make-cell-measures! domainsize)
  (let ((stopflag #f)
        (tmp '())
        (nr (- 0 1))
        (bestnr (- 0 1))
        (bestcell #f) )
    (set! *props-cell-measures-vector*
          (make-vector *props-cell-nr* (- 0 1)) )
    (do ((i 0 (+ 1 i)))
        ((>= i *props-cell-nr*))
     (set! nr (length (vector-ref *props-cell-occurrences-vector* i)))
     (vector-set! *props-cell-measures-vector* i nr) )))

(define (props-sort-cells! domainsize)
  (let ((lst '()))
    (set! lst (props-sort-cells-aux!
                 0 (- *props-cell-nr* 1) domainsize))
    (do ((i 0 (+ 1 i))
         (part lst (cdr part)) )
        ((null? part))
      (vector-set! *props-cell-order-vector* i (car part)) )))

(define (props-sort-cells-aux! lower upper domainsize)
 (let ((half 0))
  (cond
    ((> lower upper)
       '())
    ((eq? lower upper)
       (list lower) )
    ((eq? upper (+ 1 lower))
       (if (props-cell-nr-less? lower upper domainsize)
           (list lower upper)
           (list upper lower) ))
    (else
       (set! half (quotient (+ lower upper) 2))
       (props-merge
         (props-sort-cells-aux! lower half domainsize)
         (props-sort-cells-aux! (+ 1 half) upper domainsize)
         domainsize)))))

(define (props-merge lst1 lst2 domainsize)
  (cond
    ((null? lst1)
        lst2)
    ((null? lst2)
        lst1)
    ((props-cell-nr-less? (car lst1) (car lst2) domainsize)
        (cons (car lst1)
              (props-merge (cdr lst1) lst2 domainsize) ))
    (else
        (cons (car lst2)
              (props-merge lst1 (cdr lst2) domainsize) ))))

(define (props-cell-nr-less? cell1 cell2 domainsize)
  (let* ((cterm1 (vector-ref *props-cell-cterm-vector* cell1))
         (cterm2 (vector-ref *props-cell-cterm-vector* cell2))
         (fun1 (vector-ref cterm1 1))
         (fun2 (vector-ref cterm2 1))
         (iseq1 (props-eq-pred? fun1))
         (iseq2 (props-eq-pred? fun2))
         (max1 0)
         (max2 0)
         (sum1 0)
         (sum2 0)
         (occ1 0)
         (occ2 0)
         (arity1 0)
         (arity2 0) )
   (cond
     ;((< cell1 cell2)
     ;   #t)
     ;((> cell1 cell2)
     ;   #f)
     ; - - - - equality - - - - -
     ((and iseq1 iseq2) ; do not care about order of equalities: use orig cell index
        (< cell1 cell2) )
     (iseq2  ; equality is bigger than non-equality
        #t)
     (iseq1
        #f)  ; equality is bigger than non-equality
     ; - - - - maxarg - - - - - -
     ((begin
         (set! max1 (props-cterm-max-arg cterm1))
         (set! max2 (props-cterm-max-arg cterm2))
         #f))
     ((< max1 max2) ; smaller max arg is smaller than bigger max arg
        #t)
     ((> max1 max2)
        #f)
     ; - - - - occurrences - - - -
     ;((begin ; compute occurrences
     ;    (set! occ1 (length (vector-ref *props-cell-occurrences-vector* cell1)))
     ;    (set! occ2 (length (vector-ref *props-cell-occurrences-vector* cell2)))
     ;    #f))
     ;((> occ1 occ2) ; more occurrences is smaller
     ;    #t)
     ;((< occ1 occ2)
     ;    #f)
     ; - - - prefer literals - - - -
     ((and (props-term-is-surely-lit? cterm1)
           (not (props-term-is-surely-lit? cterm2)) )
        #t)
     ((and (not (props-term-is-surely-lit? cterm1))
            (props-term-is-surely-lit? cterm2) )
        #f)
     ; - - - - fun arity  - - - -
     ((begin ; compute symbols
         (set! arity1 (vector-length cterm1))
         (set! arity2 (vector-length cterm2))
         #f))
     ((< arity1 arity2) ; smaller arity is smaller
         #t)
     ((> arity1 arity2)
         #f)
     ; - - - - fun symbol - - - -
     ((< fun1 fun2) ; smaller function nr is smaller
         #t)
     ((> fun1 fun2)
         #f)
     ; - - - argsum - - - -
     ((begin ; compute argsum
         (set! sum1 (props-cterm-sum-arg cterm1))
         (set! sum2 (props-cterm-sum-arg cterm2))
         #f))
     ((< sum1 sum2)  ; smaller sum is smaller
         #t)
     ((> sum1 sum2)
         #f)
     ; - - - default - - -
     (else           ; if all parameters same, use index nr as a measure
       (< cell1 cell2) ))))

(define (props-select-split-cell-preassign domainsize)
  (let ((range '())
        (occ 0)
        (i 0)
        (value '())
        (succvalues 0)
        (tmpassigncount 0)
        (tmprestrictcount 0)
        (minmeasure 0)
        (maxmeasure 0)
        (onemaxmeasure 0)
        (minsuccvalue 1000000)
        (measure 0)
        (bestcell #f)
        (stopflag #f)
        (assignres '())
        (stack '()) )

    (if *props-debug-selectcell*
        (begin
          (display "--- pre-assign comparison starts --- ")
          (newline) ))

    (set! maxmeasure -10000)
    (set! onemaxmeasure -10000)
    (set! minsuccvalue 1000000)
    (do ((j 0 (+ 1 j)))
        ((or (>= j *props-cell-nr*)
             stopflag)
           (if *props-debug-selectcell*
               (begin
                 (display "- - pre-assign comparison selects cell nr: ")
                 (display bestcell)
                 (newline) ))
           bestcell)
     (set! i (vector-ref *props-cell-order-vector* j))
     (set! range (vector-ref *props-cell-range-vector* i))
     (set! occ (vector-ref *props-cell-occurrences-vector* i))
     (if (and (not (null? (cdr range)))
              (not (null? occ)) )
         (begin
            (if *props-debug-selectcell*
                (begin
                  (display "trying cell: ")(display i)(display " : ")
                  (print-props-ground-term (vector-ref *props-cell-cterm-vector* i))
                  (display " : ")(display (vector-ref *props-cell-range-vector* i))
                  (newline) ))

            (set! minmeasure 10000000)
            (set! succvalues 0)
            (do ((part range (cdr part)))
                ((or (null? part) (< minsuccvalue succvalues)))
                ;((null? part))
              (set! value (car part))
              (set! stack (list '()))
              (set! tmpassigncount *props-assign-count*)
              (set! tmprestrictcount *props-restrict-count*)

              (if *props-debug-selectcell*
                  (begin
                    (display "with value: ")(display value)
                    (newline) ))

              (set! assignres (props-assign-cell! i value stack domainsize))
              (cond
                ((not assignres))
                (else
                   (set! succvalues (+ 1 succvalues))
                   (set! measure
                        (+ (* 3 (- *props-assign-count* tmpassigncount))
                           (- *props-restrict-count* tmprestrictcount) ))) )

              (if *props-debug-selectcell*
                  (begin
                    (display "assignres : ")(display assignres)
                    (display " new assigns: ")
                    (display (- *props-assign-count* tmpassigncount))
                    (display " new restricts: ")
                    (display (- *props-restrict-count* tmprestrictcount))
                    (display " measure: ")(display measure)
                    (newline) ))

              (if (and assignres (< measure minmeasure))
                  (begin
                    (set! minmeasure measure) ))

              (set! *props-assign-count* tmpassigncount)
              (set! *props-restrict-count* tmprestrictcount)
              (props-restore-stack! assignres stack) )

            ;(if (not (props-term-is-surely-lit?
            ;           (vector-ref *props-cell-cterm-vector* i)))
            ;    (set! minmeasure (+ minmeasure (- domainsize (length range)))) )

            ; fewer successful values is better!
            (set! minmeasure (- minmeasure succvalues))
            ; check if the cell is best so far
            (cond
               ((eq? succvalues 0) ; all values give contradiction: select that!
                  (if *props-debug-selectcell*
                      (begin
                        (newline)(display "selection found cell with FALSE values!")
                        (newline) ))
                  (set! stopflag #t)
                  (set! bestcell i) )
               ((eq? succvalues 1) ; only one value does not give contradiction!
                  (if *props-debug-selectcell*
                      (begin
                        (newline)(display "selection found cell with one OK value!")
                        (newline) ))
                  (set! minsuccvalue 1)
                  (if (> minmeasure onemaxmeasure)
                      (begin
                        (set! onemaxmeasure minmeasure)
                        (set! bestcell i) )))
               ((> minsuccvalue 1)
                  (if (> minmeasure maxmeasure)
                      (begin
                        (set! maxmeasure minmeasure)
                        (set! bestcell i) ))  )))))))


(define (props-open-occ-cell-count)
  (let ((res 0))
    (do ((j 0 (+ 1 j)))
        ((>= j *props-cell-nr*)
          ; (newline)(display "open: ")(display res)
           res)
     (if (not (null? (cdr (vector-ref *props-cell-range-vector* j))))
         (if (not (null? (vector-ref *props-cell-occurrences-vector* j)))
             (set! res (+ 1 res)) )))))


;------------------------------------------------------------------
;
;        assigning and restricting values
;
;------------------------------------------------------------------

; props-assign-cell! takes a selected cell index,
; a value to be assigned, stack and domainsize.
; it pushes the old cell range to stack, puts in the
; new value and then goes through all the occurrences of
; this cterm, rewriting their parents
; NB! each occurrence is a cterm exactly the same as
; the one which gets a value. Hence, to rewrite, one
; actually rewrites the parent of the occurrence

(define (props-assign-cell! selcell value stack domainsize)
 (let ((tmp '())
       (res '())
       (stopflag #f)
       (occurrences '())
       (origrange '()) )

  (if *props-debug*
      (begin
        (newline)
        (display "props-assign-cell! called on: ")
        (display selcell)(display " ")(display value)
        (newline) ))

  (set! origrange (vector-ref *props-cell-range-vector* selcell))
  (cond
   ((not (memq value origrange))
     ; return false immediately if we try to assign outside range
     #f)
   ((or (integer? origrange)
        (null? (cdr origrange)) )
     ; value has been already set: do nothing
     '())
   ; normal case: assignment inside nonunit range
   (else
    (set! *props-assign-count* (+ 1 *props-assign-count*))
    (props-push-stack!
       (list *props-cell-range-vector* selcell origrange)
       stack)
    (vector-set! *props-cell-range-vector* selcell (list value))
    (set! occurrences (vector-ref *props-cell-occurrences-vector* selcell))
    (set! stopflag #f)

    ;(newline)
    ;(display "occurrences length: ")
    ;(display (length occurrences))
    ;(newline)
    ;(for-each (lambda(x)(props-print-ground-clause x)(newline))
    ;          occurrences)
    ;(newline)

    (do ((occpart occurrences (cdr occpart)))
        ((or stopflag
             *prop-time-ran-out*
             (null? occpart) )
          (if res
              (derive-from-cell-with-nterms selcell stack domainsize)
              #f))

      (set! *props-timecheck-counter* (+ 1 *props-timecheck-counter*))
      (if (> *props-timecheck-counter* 1000)
          (begin
              (set! *props-timecheck-counter* 0)
              (if (> (- (get-check-time 1) *slow-start-time*)
                     *one-pass-time*)
                  (begin
                   ;(store-time-checkpoint! 111)
                   (set! stopflag #t)
                   (set! *prop-time-ran-out* #t)
                   (set! *props-stop-search-flag* #t) ))))


      (set! res (props-rewrite-with-assignment!
                    (car occpart) selcell value stack domainsize))
      (cond
         ((eq? res #f)
             (set! stopflag #t) )
         ((eq? res #t)
             (set! stopflag #t) )))))))



; props-restrict-cell! takes a selected cell index,
; a value to be restricted, stack and domainsize.
; it pushes the old cell range to stack, removes the
; new value.


(define (props-restrict-cell! selcell value stack domainsize)
 (let ((tmp '())
       (origrange '()) )

   (if *props-debug*
      (begin
        (newline)
        (display "props-restrict-cell! called on: ")
        (display selcell)(display " ")(display value)
        (newline) ))

  (set! origrange (vector-ref *props-cell-range-vector* selcell))
  (cond
   ((or (integer? origrange)
        (null? (cdr origrange)) )
     ; value has been already set
     (if (not (integer? origrange))
         (set! origrange (car origrange)) )
     (if (eq? value origrange)
         ; return false immediately
         #f
         ; already restricted with this value
         '()))
   ((not (memq value origrange))
     ; already restricted
     '())
   (else
     ; normal case: now we can really restrict
    (set! *props-restrict-count* (+ 1 *props-restrict-count*))
    (props-push-stack!
       (list *props-cell-range-vector* selcell origrange)
       stack)
    (set! tmp (remove-one-from-list value origrange))
    (vector-set! *props-cell-range-vector* selcell tmp)
    ; now the restriction may have generated a unique possible value:
    ; in that case, propagate the value
    (if (null? (cdr tmp))
        (begin
           (set! tmp (props-propagate-cell! selcell stack domainsize))
           (if tmp
               (derive-from-cell-with-nterms selcell stack domainsize)
               #f))
        '()) ))))


(define (props-propagate-cell! selcell stack domainsize)
 (let ((tmp '())
       (res '())
       (value '())
       (stopflag #f)
       (occurrences '())
       (origrange '()) )

  ;(newline)
  ;(display "props-propagate-cell! called on: ")
  ;(display selcell)
  ;(newline)

  (set! origrange (vector-ref *props-cell-range-vector* selcell))

  ;(display "origrange: ")(display origrange)
  ;(newline)

  (cond
   ((not (or (integer? origrange)
             (null? (cdr origrange)) ))
     ; value has not been determined: do nothing
     '())
   ; normal case: assigned cell
   (else
    (set! value (car origrange))
    (set! occurrences (vector-ref *props-cell-occurrences-vector* selcell))
    (set! stopflag #f)

    ;(newline)
    ;(display "occurrences: ")
    ;(for-each (lambda(x)(print-props-ground-term x)(newline))
    ;          occurrences)
    ;(newline)

    (do ((occpart occurrences (cdr occpart)))
        ((or stopflag
             (null? occpart) )
           res)
      (set! res (props-rewrite-with-assignment!
                    (car occpart) selcell value stack domainsize))
      (cond
         ((eq? res #f)
             (set! stopflag #t) )
         ((eq? res #t)
             (set! stopflag #t) )))))))



;------------------------------------------------------------------------
;
;          rewriting, propagation, clause analysis, stack
;
;-------------------------------------------------------------------------


; props-rewrite-with-assignment!  takes a single occurrence term,
; a cell nr, a value, stack and domainsize.
; It replaces the occurrence in he parent term with the value and then checks
; the parent term:
; if parent term is a clause, something might be inferred;
; if parent term is an ordinary term, maybe after rewriting it became
; a cterm: in that case, maybe it can rewritten too (if it has a unique value)
; else (not unique value) it is stored as a new occurrence
;
; each parent is a pair: (parentvector . vectorindexofocc)

(define (props-rewrite-with-assignment! occurrence selcell value stack domainsize)
  (let ((tmp '())
        (tmpval '())
        (tmpocc '())
        (fun 0)
        (len 0)
        (stopflag #f)
        (parentterm '())
        (parentindex 0)
        (parent '()) )

    ;(newline)
    ;(display "props-rewrite-with assignment called: ")(newline)
    ;(display "occurrence: ")(print-props-ground-term occurrence)
    ;(newline)(display "selcell: ")(display selcell)
    ;(newline)(display "value: ")(display value)
    ;(newline)

    (set! *props-timecheck-counter* (+ 1 *props-timecheck-counter*))
    (if (> *props-timecheck-counter* 1000)
        (begin
              (set! *props-timecheck-counter* 0)
              (if (> (- (get-check-time 1) *slow-start-time*)
                     *one-pass-time*)
                  (begin
                   ;(store-time-checkpoint! 111)
                   (set! stopflag #t)
                   (set! *prop-time-ran-out* #t)
                   (set! *props-stop-search-flag* #t) ))))
  (if (not *prop-time-ran-out*)
   (begin

    (set! *props-rewrite-count* (+ 1 *props-rewrite-count*))
    (set! parent (vector-ref occurrence 0))
    (set! fun (vector-ref occurrence 1))
    (set! parentterm (car parent))
    (set! parentindex (cdr parent))
    ; push the assignment to the parent vec element of occurrence to stack
    (props-push-stack! (list parentterm parentindex occurrence) stack)
    ; different behaviour for clause parents and term parents:
    ; * infer stuff from clause parents: contradiction?
    ; * check rewritten parent: cterm? if yes, try to rewrite
    (if (not (zero? (logand *props-islit-mask* fun)))
        ; clause parent
        (begin

          ;(newline)
          ;(display "clause parent:")(print-props-ground-clause parentterm)
          ;(newline)

          (if (not (zero? (logand *props-isneglit-mask* fun)))
              ; negative fun
              (if (eq? value 0) (set! value 1) (set! value 0)) )
          (vector-set! parentterm parentindex value)
          (if (eq? value 1)
              (begin
                 (remove-term-occurrences! parentterm #t stack domainsize)
                 '())
              ; perhaps clause now contains only boolean falses or is a cterm unit:
              ; check for contradiction (all elems 0-s) and cterm unit clauses
              (props-analyze-clause! parentterm domainsize stack) ))
        ; non-clause term parent
        (begin

          ;(newline)
          ;(display "non-clause parent: ")(print-props-ground-term parentterm)
          ;(newline)

          (vector-set! parentterm parentindex value)

          ;(newline)
          ;(display "after rewr non-clause parent: ")(print-props-ground-term parentterm)
          ;(newline)

          (set! tmp (props-cterm-index-nr parentterm domainsize))
          (if tmp
              ; parentterm is now a cterm
              (begin
                ; rewrite parentterm if we can
                (set! tmpval (vector-ref *props-cell-range-vector* tmp))

                ;(newline)
                ;(display "tmp: ")(display tmp)(newline)
                ;(display "tmpval: ")(display tmpval)
                ;(newline)

                (cond ((integer? tmpval))
                      ((null? (cdr tmpval))
                         (set! tmpval (car tmpval)) ))
                (if (integer? tmpval)
                    ; unique value: now we have parentterm as a new cterm and
                    ; we can indeed rewrite the parentterm (unique value known)
                    (begin
                       ; recursive call
                       (props-rewrite-with-assignment!
                          parentterm tmp tmpval stack domainsize) )
                    ; not unique value: add an occurrence
                    (begin
                      (set! tmpocc (vector-ref *props-cell-occurrences-vector* tmp))

                      ;(newline)
                      ;(display "tmpocc: ")(newline)
                      ;(for-each (lambda(x)(print-props-ground-term x)(newline)) tmpocc)
                      ;(newline)

                      ; first push occurrence list to stack
                      (props-push-stack!
                         (list *props-cell-occurrences-vector* tmp tmpocc)
                         stack)
                      ; now add the new occurrence
                      (vector-set! *props-cell-occurrences-vector* tmp
                          (cons parentterm tmpocc) )

                      ;(newline)
                      ;(display "tmpocc after addition: ")(newline)
                      ;(for-each (lambda(x)(print-props-ground-term x)(newline))
                      ;  (cons parentterm tmpocc))
                      ;(newline)

                      ; parentterm might be a single literal in a clause:
                      ; check that out
                      (if (props-is-pred? (vector-ref parentterm 1))
                          ; parent of a parent is a clause: analyze it
                          (props-analyze-clause!
                             (car (vector-ref parentterm 0))
                             domainsize stack)
                          ; parentterm might be an argument of an equality,
                          ; which is a unit lit in a clause, or parentterm
                          ; might be a single nterm in a clause: analyse it!
                          (begin
                             (set! tmp (car (vector-ref parentterm 0)))
                             ; tmp is now the containing term of a parentterm
                             (if (props-is-pred? (vector-ref tmp 1))
                                 ; tmp is a literal
                                 (props-analyze-clause!
                                     (car (vector-ref tmp 0))
                                     domainsize stack)
                                 ; or it might be an arg of an equality
                                 (if (props-eq-pred?
                                        (vector-ref (car (vector-ref tmp 0)) 1) )
                                     (props-analyze-clause!
                                         (car (vector-ref (car (vector-ref tmp 0)) 0))
                                          domainsize stack)
                                     '() )) )))))
              ; parent is not a cterm: however, it might
              ; be a single equality pred in a clause or
              ; a single nterm in a clause: check that out
              (if (props-is-pred? (vector-ref parentterm 1))
                  (props-analyze-clause!
                             (car (vector-ref parentterm 0))
                             domainsize stack)
                  ; or it might be a near-term equality!
                  (if (props-eq-pred? (vector-ref (car (vector-ref parentterm 0)) 1))
                      (props-analyze-clause!
                             (car (vector-ref (car (vector-ref parentterm 0)) 0))
                             domainsize stack)
                       '() )) )))))))



(define (props-analyze-clause! clause domainsize stack)

 (if *props-debug*
     (begin
      (newline)
      (display "props-analyze-clause! called on: ")
      (print-props-ground-clause clause)
      (newline) ))

 (let ((falseflag #t)
       (stopflag #f)
       (tautflag #f)
       (foundlit #f)
       (foundres #f)
       (oldfoundlit #f)
       (oldfoundres #f)
       (foundposflag #f)
       (oldfoundposflag #f)
       (elem #f)
       (len 0)
       (foundtype #f) ; 1,-1: unit 2,-2: eq , 3,-3: nterm unit, 4,-4: nterm eq
       (lst '())
       (tmpposflag #f)
       (tmp '()) )
  (if (eq? 1 (vector-ref clause 0))
   ; analysing a known tautology: stop
    '()
   ; analysing a not-known tautology
   (begin
    (set! len (vector-length clause))
    (do ((i 2 (+ i 1)))
        ((or stopflag (>= i len))

           (if *props-debug*
               (begin
                (newline)
                (display "clause analysis about to end: ")(newline)
                (display "falseflag: ")(display falseflag)(newline)
                (display "stopflag: ")(display stopflag)(newline)
                (display "foundtype: ")(display foundtype)(newline)
                (display "foundres: ")(display foundres)(newline) ))

           (cond
             (tautflag
                (props-push-stack!
                    (list clause 0 (vector-ref clause 0))
                    stack)
                (vector-set! clause 0 1)
                (remove-term-occurrences! clause #t stack domainsize)
                '())
             (falseflag
                ; contradiction found!!!
                #f)
             (stopflag '())
             ((not foundres) '())
             ((eq? foundtype 1)

               ;(newline)
               ;(display "clause analysis generated value from: ")
               ;(print-props-ground-clause clause)
               ;(newline)

               (if foundposflag
                   (props-assign-cell! foundres 1 stack domainsize)
                   (props-assign-cell! foundres 0 stack domainsize) ))
             ((eq? foundtype 2)

               ;(newline)
               ;(display "clause analysis generated equality assignment from: ")
               ;(print-props-ground-clause clause)
               ;(newline)

               (if (caddr foundres)
                   (props-assign-cell!
                         (car foundres) (cadr foundres) stack domainsize)
                   (props-restrict-cell!
                         (car foundres) (cadr foundres) stack domainsize) ))
             ((eq? foundtype 3)

               ;(newline)
               ;(display "clause analysis generated value from: ")
               ;(print-props-ground-clause clause)
               ;(newline)

               (if foundposflag
                   (props-assign-nterm!
                        (car foundres) (cadr foundres) (caddr foundres)
                        1 stack domainsize)
                   (props-assign-nterm!
                        (car foundres) (cadr foundres) (caddr foundres)
                        0 stack domainsize) ))

             ((eq? foundtype 4)

               ;(newline)
               ;(display "clause analysis generated equality assignment from: ")
               ;(print-props-ground-clause clause)
               ;(newline)

               (set! tmp (car foundres))
               (if (caddr foundres)
                   (props-assign-nterm!
                         (car tmp) (cadr tmp) (caddr tmp)
                         (cadr foundres) stack domainsize)
                   (props-restrict-nterm!
                         (car tmp) (cadr tmp) (caddr tmp)
                         (cadr foundres) stack domainsize) ))

             (else '()) ))
      (set! elem (vector-ref clause i))
      (cond
        ; integer case: stopped if 1 found. 0 ignored.
        ((integer? elem)
          (if (eq? elem 1)
              (begin
                 (set! stopflag #t)
                 (set! tautflag #t)
                 (set! falseflag #f) )))
        ; unitlit case: only tried if
        ; nothing found or found of the same type
        ((and (or falseflag (eq? foundtype 1))
              (begin
                 (set! foundres (props-cterm-index-nr elem domainsize))
                 foundres))
            (set! falseflag #f)
            (set! foundposflag
                  (zero? (logand *props-isneglit-mask*
                                  (vector-ref elem 1) )))
            (if (and oldfoundres
                     (or (not (eq? oldfoundres foundres))
                         (not (eq? oldfoundposflag foundposflag)) ))
               (set! stopflag #t)
               ; here we actually found an OK unit
               (begin
                    (set! foundtype 1)
                    (set! oldfoundres foundres)
                    (set! oldfoundposflag foundposflag) )))
        ; eqlit case: only tried if
        ; nothing found or found of the same type
        ((and (or falseflag (eq? foundtype 2))
              (props-eq-pred? (vector-ref elem 1))
              (begin
                 (set! foundres
                       (props-check-eq-assign elem domainsize))
                 foundres))
            (set! falseflag #f)

            ;(newline)
            ;(display "interesting eqlit found in clause: ")
            ;(print-props-ground-term elem)
            ;(newline)
            ;(newline)
            ;(display "eqlitres is: ")
            ;(display eqlitres)
            ;(newline)

            (if (and oldfoundres
                     (not (equal? oldfoundres  foundres)))
                (set! stopflag #t)
                ; here we actually found an OK eqlit
                (begin
                     (set! foundtype 2)
                     (set! oldfoundres foundres) )))
        ; nterm unitlit case: tried always!
        ((and (not (props-eq-pred? (vector-ref elem 1)))
              (begin
                 (set! tmp (props-check-nterm-index-nr elem domainsize))
                 tmp))
            (set! tmpposflag
                  (zero? (logand *props-isneglit-mask*
                                  (vector-ref elem 1) )))
            ; check if we have a value for it!
            ; result format: (nterm_index cterm_pos cterm_index)

            (set! lst (vector-ref *props-nterm-vector* (car tmp)))
            (set! lst (assq (caddr tmp) lst))
            (if (and lst (null? (cddr lst)))
                ; a range found with a single value:
                ; we may get a tautology (then we set vars)
                ; or the literal may be false (then we set nothing)
                (begin
                  (set! lst (cadr lst))
                  ; lst is now the value
                  ;(newline)(display "nterm in clause has value ")
                  (if tmpposflag
                      (if (eq? 1 lst)
                          (begin
                             ;(display "tautology")
                             (set! stopflag #t)
                             (set! tautflag #t)
                             (set! falseflag #f) ))
                      (if (eq? 0 lst)
                          (begin
                             ;(display "tautology")
                             (set! stopflag #t)
                             (set! tautflag #t)
                             (set! falseflag #f) )) ))
               ; no range found or no single value
               (begin
                 ; now check if some lit has been found earlier: if yes,
                 ; do not store the foundres for further assigning if unit
                 (if falseflag
                     ; here we actually found an OK nterm unit
                     (begin
                        (set! falseflag #f)
                        (set! foundtype 3)
                        (set! foundposflag tmpposflag)
                        (set! foundres tmp) )
                     ; we have found stuff before: no units, sorry!
                     (begin
                        (set! stopflag #t)
                        (set! falseflag #f) ))) ))
        ; nterm eqlit case: only tried if
        ; nothing found
        ((and falseflag
              ;(begin (newline)(display "checking type 4")(newline) #t)
              (props-eq-pred? (vector-ref elem 1))
              (begin
                 (set! foundres (props-check-nterm-eq-assign elem domainsize))
                 ;(begin (newline)(display "foundres: ")(display foundres)(newline) #t)
                 foundres))
            (set! falseflag #f)
            ; here we actually found an OK nterm eqlit
            (set! foundtype 4) )
         ; we found a non-integer lit which either does not
         ; fall under interesting categories or we found several
         ; different of these
         (else
            (set! falseflag #f)
            (set! stopflag #t) )))))))



(define (props-analyze-clause-old! clause domainsize stack)

  (if *props-debug*
      (begin
       (newline)
       (display "props-analyze-clause! called on: ")
       (print-props-ground-clause clause)
       (newline) ))

  (let ((falseflag #t)
        (stopflag #f)
        (tautflag #f)
        (foundlit #f)
        (foundres #f)
        (oldfoundlit #f)
        (oldfoundres #f)
        (foundposflag #f)
        (oldfoundposflag #f)
        (elem #f)
        (len 0)
        (foundtype #f) ; 1,-1: unit 2,-2: eq , 3,-3: nterm unit, 4,-4: nterm eq
        (tmp '()) )
    (set! len (vector-length clause))
    (do ((i 2 (+ i 1)))
        ((or stopflag (>= i len))

           (if *props-debug*
               (begin
                (newline)
                (display "clause analysis about to end: ")(newline)
                (display "falseflag: ")(display falseflag)(newline)
                (display "stopflag: ")(display stopflag)(newline)
                (display "foundtype: ")(display foundtype)(newline)
                (display "foundres: ")(display foundres)(newline) ))

           (cond
             (tautflag
                (remove-term-occurrences! clause #t stack domainsize)
                '())
             (falseflag
                ; contradiction found!!!
                #f)
             (stopflag '())
             ((not foundres) '())
             ((eq? foundtype 1)

               ;(newline)
               ;(display "clause analysis generated value from: ")
               ;(print-props-ground-clause clause)
               ;(newline)

               (if foundposflag
                   (props-assign-cell! foundres 1 stack domainsize)
                   (props-assign-cell! foundres 0 stack domainsize) ))
             ((eq? foundtype 2)

               ;(newline)
               ;(display "clause analysis generated equality assignment from: ")
               ;(print-props-ground-clause clause)
               ;(newline)

               (if (caddr foundres)
                   (props-assign-cell!
                         (car foundres) (cadr foundres) stack domainsize)
                   (props-restrict-cell!
                         (car foundres) (cadr foundres) stack domainsize) ))
             ((eq? foundtype 3)

               ;(newline)
               ;(display "clause analysis generated value from: ")
               ;(print-props-ground-clause clause)
               ;(newline)

               (if foundposflag
                   (props-assign-nterm!
                        (car foundres) (cadr foundres) (caddr foundres)
                        1 stack domainsize)
                   (props-assign-nterm!
                        (car foundres) (cadr foundres) (caddr foundres)
                        0 stack domainsize) ))

             ((eq? foundtype 4)

               ;(newline)
               ;(display "clause analysis generated equality assignment from: ")
               ;(print-props-ground-clause clause)
               ;(newline)

               (set! tmp (car foundres))
               (if (caddr foundres)
                   (props-assign-nterm!
                         (car tmp) (cadr tmp) (caddr tmp)
                         (cadr foundres) stack domainsize)
                   (props-restrict-nterm!
                         (car tmp) (cadr tmp) (caddr tmp)
                         (cadr foundres) stack domainsize) ))

             (else '()) ))
      (set! elem (vector-ref clause i))
      (cond
        ; integer case: stopped if 1 found. 0 ignored.
        ((integer? elem)
          (if (eq? elem 1)
              (begin
                 (set! stopflag #t)
                 (set! tautflag #t)
                 (set! falseflag #f) )))
        ; unitlit case: only tried if
        ; nothing found or found of the same type
        ((and (or falseflag (eq? foundtype 1))
              (begin
                 (set! foundres (props-cterm-index-nr elem domainsize))
                 foundres))
            (set! falseflag #f)
            (set! foundposflag
                  (zero? (logand *props-isneglit-mask*
                                  (vector-ref elem 1) )))
            (if (and oldfoundres
                     (or (not (eq? oldfoundres foundres))
                         (not (eq? oldfoundposflag foundposflag)) ))
               (set! stopflag #t)
               ; here we actually found an OK unit
               (begin
                    (set! foundtype 1)
                    (set! oldfoundres foundres)
                    (set! oldfoundposflag foundposflag) )))
        ; eqlit case: only tried if
        ; nothing found or found of the same type
        ((and (or falseflag (eq? foundtype 2))
              (props-eq-pred? (vector-ref elem 1))
              (begin
                 (set! foundres
                       (props-check-eq-assign elem domainsize))
                 foundres))
            (set! falseflag #f)

            ;(newline)
            ;(display "interesting eqlit found in clause: ")
            ;(print-props-ground-term elem)
            ;(newline)
            ;(newline)
            ;(display "eqlitres is: ")
            ;(display eqlitres)
            ;(newline)

            (if (and oldfoundres
                     (not (equal? oldfoundres  foundres)))
                (set! stopflag #t)
                ; here we actually found an OK eqlit
                (begin
                     (set! foundtype 2)
                     (set! oldfoundres foundres) )))
        ; nterm unitlit case: only tried if
        ; nothing found
        ((and falseflag
              (not (props-eq-pred? (vector-ref elem 1)))
              (begin
                 (set! foundres (props-check-nterm-index-nr elem domainsize))
                 foundres))
            (set! falseflag #f)
            (set! foundposflag
                  (zero? (logand *props-isneglit-mask*
                                  (vector-ref elem 1) )))
            ; here we actually found an OK nterm unit

            (set! foundtype 3)
            (set! foundlit elem) )
        ; nterm eqlit case: only tried if
        ; nothing found
        ((and falseflag
              ;(begin (newline)(display "checking type 4")(newline) #t)
              (props-eq-pred? (vector-ref elem 1))
              (begin
                 (set! foundres (props-check-nterm-eq-assign elem domainsize))
                 ;(begin (newline)(display "foundres: ")(display foundres)(newline) #t)
                 foundres))
            (set! falseflag #f)
            ; here we actually found an OK nterm eqlit
            (set! foundtype 4) )
         ; we found a non-integer lit which either does not
         ; fall under interesting categories or we found several
         ; different of these
         (else
            (set! falseflag #f)
            (set! stopflag #t) )))))

(define (remove-term-occurrences-old! term clauseflag stack domainsize)
  '() )

(define (remove-term-occurrences! term clauseflag stack domainsize)
 ; (newline)(display "remove-term-occurrences called! ")(newline)
 (if (not *props-remove-terms-strat*)
  '()
  (let ((len (vector-length term))
        (nr '())
        (range '())
        (occs '())
        (place '())
        (bigsub #f) )
    (do ((i 2 (+ 1 i)))
        ((>= i len)
           (if (and (not bigsub) (not clauseflag))
               (begin
                 ;(newline)(display "about to remove: ")
                 ;(print-props-ground-term term)(newline)
                 (set! nr (props-cterm-index-nr term domainsize))
                 ;(display "nr: ")(display nr)(newline)
                 (set! range (vector-ref *props-cell-range-vector* nr))
                 ;(display "range: ")(display range)(newline)
                 (cond
                   ((null? (cdr range)))
                   ((begin (set! occs (vector-ref *props-cell-occurrences-vector* nr))
                           ;(display "occs: ")(display (length occs))(newline)
                           #f))
                   ((null? occs))
                   ((eq? term (car occs))
                      ;(newline)(display "removing: ")(display nr)
                      (props-push-stack!
                          (list *props-cell-occurrences-vector* nr occs)
                          stack)
                      (vector-set! *props-cell-occurrences-vector* nr (cdr occs)) )
                   (else
                       (set! place (props-pre-memq term occs))
                       (if place
                           (begin
                            ;(newline)(display "removing:")(display nr)
                            (props-push-stack! (list place 1 (cdr place)) stack)
                            (set-cdr! place (cddr place)) )))))))
      (set! place (vector-ref term i))
      (if (not (integer? place))
          (begin
            (set! bigsub #t)
            (remove-term-occurrences! place #f stack domainsize)   ))))))

(define (props-pre-memq el lst)
  (let ((res #f)
        (tmp lst)
        (part lst) )
    (do ()
        ((or res (null? tmp))
             res)
      (set! tmp (cdr part))
      (if (and (not (null? tmp))
               (eq? el (car tmp)) )
          (set! res part) )
      (set! part tmp) )))


(define (props-check-eq-assign eqterm domainsize)
  (let ((tmp '())
        (a1 (vector-ref eqterm 2))
        (a2 (vector-ref eqterm 3)) )
    (if (integer? a1)
        (if (not (integer? a2))
            ; a1 is integer, a2 is not integer
            (begin
              (set! tmp (props-cterm-index-nr a2 domainsize))
              (if tmp
                  (if (props-pos-pred? (vector-ref eqterm 1))
                      (list tmp a1 #t)
                      (list tmp a1 #f) )
                  #f))
            #f)
        (if (integer? a2)
            ; a2 is integer, a1 is not integer
            (begin
              (set! tmp (props-cterm-index-nr a1 domainsize))
              (if tmp
                  (if (props-pos-pred? (vector-ref eqterm 1))
                      (list tmp a2 #t)
                      (list tmp a2 #f) )
                  #f))
            #f) )))



(define (props-push-stack! assignment stack)
  (set-cdr! stack (cons assignment (cdr stack))) )

(define (props-restore-stack! assignments stack)
  (let ((tmp '()))
    (do ((stpart stack (cdr stpart)))
        ((null? stpart)
           (set-cdr! stack '()) )
      (set! tmp (car stpart))
      (if (pair? tmp)
          (if (vector? (car tmp))
              (vector-set! (car tmp) (cadr tmp) (caddr tmp))
              (if (pair? (car tmp))
                  (if (eq? 0 (cadr tmp))
                      (set-car! (car tmp) (caddr tmp))
                      (set-cdr! (car tmp) (caddr tmp)) )))))))






;------------------------------------------------------------------------
;
;          a block for indexing and retrieval functions
;
;-------------------------------------------------------------------------


(define (props-init-indexes domainsize)
  (let ((curbase 0)
        (nr 0)
        (base 0)
        (tmp '())
        (tmp2 '())
        (tmp3 '())
        (matchnr #f)
        (lst '())
        (allchecked #f)
        (varvaluations '())
        (stopflag #f)
        (arity 0))
  ; init funsymb base indexes for celltable
  (if *props-dynamic-vectors* (set! *props-fun-basenr-vector* (make-vector 10000 #f)))
  (do ((i 0 (+ 1 i)))
      ((>= i *props-fun-counter*))
   (vector-set! *props-fun-basenr-vector* i curbase)
   (set! arity (vector-ref *props-fun-arities-vector* i))
   (set! nr (int-power domainsize arity))
   (set! curbase (+ nr curbase)) )
  ; now make the celltables
  ; curbase right now is the actual nr of cells
  (set! *props-cell-nr* curbase)
  (if (> *props-cell-nr* *props-max-allowed-cells*)
      (set! *props-stop-search-flag* #t) )

  (if (not *props-stop-search-flag*)
   (begin
    (if *props-dynamic-vectors*
       (begin
         (set! *props-cell-cterm-vector* (make-vector curbase #f))
         (set! *props-cell-occurrences-vector* (make-vector curbase '()))
         (set! *props-cell-range-vector* (make-vector curbase '()))
         (set! *props-cell-order-vector* (make-vector curbase '()) )))
    ; fill the originals vector (cterm)
    (do ((i 0 (+ 1 i)))
        ((>= i *props-fun-counter*))
       (set! base (vector-ref *props-fun-basenr-vector* i))
       (set! arity (vector-ref *props-fun-arities-vector* i))
       (set! varvaluations (make-list arity 0))
       (set! allchecked #f)
       ; loop over all the valuations for vars
       (do ((j 0 (+ 1 j)))
           (allchecked)
         (set! tmp (make-vector (+ 2 arity) 0))
         (vector-set! tmp 0 '())
         (vector-set! tmp 1 i)
         (do ((k 2 (+ 1 k))
              (part varvaluations (cdr part)) )
             ((>= k (vector-length tmp)))
           (vector-set! tmp k (car part)) )

         ;(newline)
         ;(display tmp)(display " : ")
         ;(display (props-cterm-index-nr tmp domainsize)) ; (display " : ")
         ;(newline)

         (vector-set! *props-cell-cterm-vector*
                      (props-cterm-index-nr tmp domainsize)
                      tmp)
         (set! allchecked
              (or (null? varvaluations)
                  (props-valuation-list-nextvalues!
                      varvaluations domainsize)))))
    ; create initial values for cell ranges and occurrences and types and order
    (do ((i 0 (+ i 1)))
        ((>= i *props-cell-nr*))
      (set! tmp
            (props-make-initial-range
                 (vector-ref
                      (vector-ref *props-cell-cterm-vector*  i)
                      1)
                 domainsize))
      (vector-set! *props-cell-order-vector* i i)
      (vector-set! *props-cell-range-vector* i tmp)
      (vector-set! *props-cell-occurrences-vector* i '()) )
    ; create the nterm basenr and nterm vectors
    (if *props-dynamic-vectors* (set! *props-nterm-fun-basenr-vector* (make-vector 10000)))
    (set! curbase 0)
    (do ((i 0 (+ 1 i)))
        ((>= i *props-fun-counter*))
     (vector-set! *props-nterm-fun-basenr-vector* i curbase)
     (set! arity (vector-ref *props-fun-arities-vector* i))
     (set! nr (int-power (+ 1 domainsize) arity))
     (if (or (> nr 10000000) (< nr 0))
         (report-error "Too many possible combinations for model finder.") )
     (set! curbase (+ nr curbase)) )
    (set! *props-nterm-nr* curbase)
    (if *props-dynamic-vectors*
        (begin
         (set! *props-nterm-vector* (make-vector *props-nterm-nr* '()))
         (set! *props-nterm-originals-vector* (make-vector *props-nterm-nr* '()))
         (set! *props-nterm-matchcells-vector* (make-vector *props-nterm-nr* '()))
         (set! *props-nterm-clauses-vector* (make-vector *props-nterm-nr* '())) ))
    (set! tmp2 (vector-length *props-nterm-vector*))
    (do ((i 0 (+ 1 i)))
        ((>= i tmp2))
      (vector-set! *props-nterm-vector* i '())
      (vector-set! *props-nterm-originals-vector* i '())
      (vector-set! *props-nterm-originals-vector* i '()) )
    ; fill the originals vector (nterm)
    (do ((i 0 (+ 1 i)))
        ((or #f ;(not *props-debug*)
            (>= i *props-fun-counter*)))
       (set! base (vector-ref *props-fun-basenr-vector* i))
       (set! arity (vector-ref *props-fun-arities-vector* i))
       (set! varvaluations (make-list arity 0))
       (set! allchecked #f)
       ; loop over all the valuations for vars
       (do ((j 0 (+ 1 j)))
           (allchecked)
         (set! tmp (make-vector (+ 2 arity) 0))
         (vector-set! tmp 0 '())
         (vector-set! tmp 1 i)
         (set! tmp3 (make-vector (+ 2 arity) 0))
         (vector-set! tmp3 0 '())
         (vector-set! tmp3 1 i)

         ;(newline)(display "about to fill.")(newline)

         (do ((k 2 (+ 1 k))
              (part varvaluations (cdr part)) )
             ((>= k (vector-length tmp)))
           (if (eq? domainsize (car part))
               (vector-set! tmp k tmp3)
               (vector-set! tmp k (car part)) ))

        ; (newline)
        ; (display tmp)(display " : ")
        ; (display (props-check-nterm-index-nr tmp domainsize)) ; (display " : ")
        ; (newline)

         (set! tmp2 (props-check-nterm-index-nr tmp domainsize))
            ;returns:  (nterm_index cterm_pos cterm_index)
         (if tmp2
             (begin
                 (vector-set! tmp (cadr tmp2) domainsize)
                 (vector-set! *props-nterm-originals-vector*
                     (car tmp2)
                     tmp)
                 (set! lst '())
                 (do ((i 0 (+ 1 i)))
                     ((>= i domainsize))
                   (vector-set! tmp (cadr tmp2) i)
                   (set! matchnr (props-cterm-index-nr tmp domainsize))
                   (set! lst (cons matchnr lst)) )
                 (vector-set! *props-nterm-matchcells-vector* (car tmp2) lst)
                 (vector-set! tmp (cadr tmp2) domainsize) ))
         (set! allchecked
              (or (null? varvaluations)
                  (props-valuation-list-nextvalues!
                      varvaluations (+ domainsize 1)))) ))
    ; finished
    ))))

(define (props-store-initial-occurrences clauses domainsize)
  (let ((clause '())
        (len 0)
        (lit '())
        (occ '())
        (tmp2 '())
        (tmp3 '())
        (tmp '()) )
    (do ((clausespart clauses (cdr clausespart)))
        ((null? clausespart))
      (set! clause (car clausespart))
      (set! len (vector-length clause))
      (do ((i 2 (+ 1 i)))
          ((>= i len))
        (set! lit (vector-ref clause i))
        (if (vector? lit)
            (begin
              (set! tmp (props-cterm-index-nr lit domainsize))
              (if tmp
                  ; literal is cterm, put into occ list
                  (begin
                    (set! occ (vector-ref *props-cell-occurrences-vector* tmp))
                    (vector-set! *props-cell-occurrences-vector*
                                 tmp
                                 (cons lit occ) ))
                  ; literal is not a cterm, recursed over for cterm subterms
                  (begin
                    ; first test if literal is nterm and clause is non-unit:
                    ; try to add to nterm clause list
                    (if (> len 3)
                        (begin
                          (set! tmp (props-check-nterm-index-nr lit domainsize))
                          ;(newline)(display "tmp: ")(display tmp)
                          (if tmp
                             (begin
                               (set! tmp2
                                  (vector-ref *props-nterm-clauses-vector* (car tmp)) )
                               (set! tmp3 (assq (caddr tmp) tmp2))
                               (if tmp3
                                  (set-cdr! tmp3 (cons clause (cdr tmp3)))
                                  (vector-set! *props-nterm-clauses-vector* (car tmp)
                                        (cons (list (caddr tmp) clause)
                                              tmp2) ))))))
                    ; recurse over
                    (props-store-subterm-occurrences lit domainsize)  ))))))))

(define (props-store-subterm-occurrences term domainsize)
  (let ((len 0)
        (sterm '())
        (occ '())
        (tmp '()) )
      (set! len (vector-length term))
      (do ((i 2 (+ 1 i)))
          ((>= i len))
        (set! sterm (vector-ref term i))
        (if (vector? sterm)
            (begin
              (set! tmp (props-cterm-index-nr sterm domainsize))
              (if tmp
                  ; sterm is cterm, put into occ list
                  (begin
                    (set! occ (vector-ref *props-cell-occurrences-vector* tmp))
                    (vector-set! *props-cell-occurrences-vector*
                                 tmp
                                 (cons sterm occ) ))
                  ; sterm is not a cterm, recursed over for cterm subterms
                  (begin
                    (props-store-subterm-occurrences sterm domainsize) )))))))


(define (props-make-initial-range fun domainsize)
  (let ((ispred #f)
        (tmp '()) )
    (set! ispred (vector-ref *props-fun-ispred-vector* fun))
    (if ispred
        (list 0 1)
        (do ((i (- domainsize 1) (- i 1)))
            ((< i 0)
               tmp)
          (set! tmp (cons i tmp)) ))))

(define (props-cterm-fun-basenr fun)
  (vector-ref *props-fun-basenr-vector*
              (logand *props-funnr-mask* fun) ))


(define (props-cterm-index-nr term domainsize)
  (let ((part term)
        (stopflag #f)
        (len (vector-length term))
        (pwr 1)
        (tmp 0)
        (nr 0) )

    ;(newline)
    ;(display term)(display " :: ")(display (props-cterm-fun-basenr (vector-ref term 1)))
    ;(newline)

    (do ((i 2 (+ 1 i)))
        ((or stopflag
            (>= i len) )
          (if stopflag
              nr
              (+ (props-cterm-fun-basenr (vector-ref term 1))
                 nr)))
      (set! tmp (vector-ref term i))
      (if (not (integer? tmp))
          (begin
             (set! stopflag #t)
             (set! nr #f) )
          (begin
             (set! nr (+ (* tmp pwr) nr))
             (set! pwr (* pwr domainsize)) )))))


;------------------------------------------------------
;
;      nterm (near term of cterm) handling procedures
;
;-------------------------------------------------------

(define (props-nterm-fun-basenr fun)
  (vector-ref *props-nterm-fun-basenr-vector*
              (logand *props-funnr-mask* fun) ))

; props-nterm-index-nr takes an nterm with a KNOWN contained
; cterm pointed by cellpos and computes its index.
; It does not check that at that pos we indeed have a cterm.

(define (props-nterm-index-nr term cellpos domainsize)
  (let ((part term)
        (stopflag #f)
        (len (vector-length term))
        (pwr 1)
        (tmp 0)
        (nr 0) )

    ;(newline)
    ;(display term)(display " :: ")(display (props-nterm-fun-basenr (vector-ref term 1)))
    ;(newline)

    (do ((i 2 (+ 1 i)))
        ((or stopflag
            (>= i len) )
          (if stopflag
              nr
              (+ (props-nterm-fun-basenr (vector-ref term 1))
                 nr)))
      (if (eq? i cellpos)
          (set! tmp domainsize)
          (set! tmp (vector-ref term i)) )
      (if (not (integer? tmp))
          (begin
             (set! stopflag #t)
             (set! nr #f) )
          (begin
             (set! nr (+ (* tmp pwr) nr))
             (set! pwr (* pwr (+ 1 domainsize))) )))))

; props-check-nterm-index-nr takes a term and a domainsize:
; it checks whether term is a near term (one arg cell, rest integers)
; If not, #f is returned. Else a list (nterm_index cterm_pos cterm_index) is
; returned.


(define (props-check-nterm-index-nr term domainsize)
 (if (not *props-use-nterms-strat*)
   #f
  (let ((stopflag #f)
        (len (vector-length term))
        (pwr 1)
        (tmp 0)
        (cval #f)
        (cpos #f)
        (nr 0) )

    ;(newline)
    ;(display term)(display " :: ")(display (props-nterm-fun-basenr (vector-ref term 1)))
    ;(newline)

    (do ((i 2 (+ 1 i)))
        ((or stopflag
            (>= i len) )
          (if (or stopflag (not cval))
              #f
              (list (+ (props-nterm-fun-basenr (vector-ref term 1))
                        nr)
                    cpos
                    cval)))
      (set! tmp (vector-ref term i))
      (if (not (integer? tmp))
          ; complex subterm
          (if cval
             ; there was a cterm term before
             (set! stopflag #t)
             ; no cterms were found before
             (begin
                (set! cval (props-cterm-index-nr tmp domainsize))
                (if  cval
                     (begin
                       (set! cpos i)
                       (set! nr (+ (* domainsize pwr) nr))
                       (set! pwr (* pwr (+ 1 domainsize))) )
                     (set! stopflag #t) )))
          ; integer subterm
          (begin
             (set! nr (+ (* tmp pwr) nr))
             (set! pwr (* pwr (+ 1 domainsize))) ))))) )





(define (props-check-nterm-eq-assign eqterm domainsize)
 (if (not *props-use-nterms-strat*)
  #f
  (let ((tmp '())
        (a1 (vector-ref eqterm 2))
        (a2 (vector-ref eqterm 3)) )
    (if (integer? a1)
        (if (not (integer? a2))
            ; a1 is integer, a2 is not integer
            (begin
              (set! tmp (props-check-nterm-index-nr a2 domainsize))
              (if tmp
                  (if (props-pos-pred? (vector-ref eqterm 1))
                      (list tmp a1 #t)
                      (list tmp a1 #f) )
                  #f))
            #f)
        (if (integer? a2)
            ; a2 is integer, a1 is not integer
            (begin
              (set! tmp (props-check-nterm-index-nr a1 domainsize))
              (if tmp
                  (if (props-pos-pred? (vector-ref eqterm 1))
                      (list tmp a2 #t)
                      (list tmp a2 #f) )
                  #f))
            #f) ))) )



(define (props-assign-nterm! ntermnr pos ctermnr value stack domainsize)
 (if (not *props-use-nterms-strat*)
  '()
  (let ((tmp '())
        (lst '())
        (clauses '())
        (res #f)
        (celem #f) )

    (if *props-debug*
        (begin
          (newline)
          (display "props-assign-nterm! called on: ")(newline)
          (display "ntermnr: ")(display ntermnr)(newline)
          (display "ctermnr: ")(display ctermnr)(newline)
          (display "value: ")(display value)(newline) ))

    (set! lst (vector-ref *props-nterm-vector* ntermnr))
    (set! celem (assq ctermnr lst))
    (cond
      ; no values for ctermnr: push old lst to stack, then add elem to lst
      ((not celem)
         (set! *props-nterm-assign-count* (+ 1 *props-nterm-assign-count*))
         (props-push-stack!
             (list *props-nterm-vector* ntermnr lst)
              stack)
         (vector-set!
               *props-nterm-vector* ntermnr
              (cons (cons ctermnr (list value))
                    lst))
         (set! res (derive-from-nterm-with-cells ntermnr pos ctermnr stack domainsize))
         (if #f ;res
             (begin
              (set! clauses (assq ctermnr (vector-ref *props-nterm-clauses-vector* ntermnr)))
              (if clauses
                  (do ((part (cdr clauses) (cdr part)))
                      ((or (null? part) (not res))
                        ;(newline)(display "special res: ")(display res)
                        res)
                    (set! tmp (props-analyze-clause! (car part) domainsize stack))
                    (if (not tmp) (set! res #f)) )
                  res)
              res)
             res))
      ; values already present, but new value out of range: contradiction!
      ((not (memq value (cdr celem)))
         #f)
      ; value present in range as a single element: do nothing
      ((null? (cddr celem))
         '() )
      ; now we have a value present in a non-empty value list:
      ; push the value list to stack, then change the value list!
      (else
        (set! *props-nterm-assign-count* (+ 1 *props-nterm-assign-count*))
        ; first push celem cdr to stack
        (props-push-stack!
             (list celem 1 (cdr celem))
              stack)
        (set-cdr! celem (list value))
        (derive-from-nterm-with-cells ntermnr pos ctermnr stack domainsize) ) )))  )


(define (props-restrict-nterm! ntermnr pos ctermnr value stack domainsize)
 (if (not *props-use-nterms-strat*)
  '()
  (let ((tmp '())
        (lst '())
        (celem #f) )

    (if *props-debug*
        (begin
         (newline)
         (display "props-restrict-nterm! called on: ")(newline)
         (display "ntermnr: ")(display ntermnr)(newline)
         (display "ctermnr: ")(display ctermnr)(newline)
         (display "value: ")(display value)(newline) ))

    (set! lst (vector-ref *props-nterm-vector* ntermnr))
    (set! celem (assq ctermnr lst))
    (cond
      ; no values for ctermnr: push old lst to stack, then add elem to lst
      ((not celem)
         (set! *props-nterm-restrict-count* (+ 1 *props-nterm-restrict-count*))
         (set! tmp (props-make-initial-range
                      (vector-ref
                          (vector-ref *props-nterm-originals-vector* ntermnr)
                          1)
                      domainsize))
         (set! tmp (remove-one-from-list value tmp))
         (if (null? tmp)
             ; contradiction
             #f
             ; range is not empty: OK to set
             (begin
                 (props-push-stack!
                     (list *props-nterm-vector* ntermnr lst)
                      stack)
                 (vector-set!
                     *props-nterm-vector* ntermnr
                      (cons (cons ctermnr tmp)
                            lst))
                 (derive-from-nterm-with-cells ntermnr pos ctermnr stack domainsize) )))
      ; value not present in range: do nothing
      ((not (memq value (cdr celem)))
         '() )
      ; value present in range as a single element: contradiction
      ((null? (cddr celem))
         #f)
      ; now we have a value present in a non-empty, non-unit value list:
      ; push the value list to stack, then change the value list!
      (else
        (set! *props-nterm-restrict-count* (+ 1 *props-nterm-restrict-count*))
        ; first push celem cdr to stack
        (props-push-stack!
             (list celem 1 (cdr celem))
              stack)
        (set-cdr! celem (remove-one-from-list value (cdr celem)))
        (derive-from-nterm-with-cells ntermnr pos ctermnr stack domainsize) ) )))  )



;(define *props-fun-basenr-vector* '())
;(define *props-cell-nr* 0)
;(define *props-cell-cterm-vector* '())
;(define *props-cell-occurrences-vector* '())
;(define *props-cell-range-vector* '())


;(define *props-nterm-nr* 0)
;(define *props-nterm-vector* '())
;(define *props-nterm-originals-vector* '())
;(define *props-nterm-fun-basenr-vector* '())


(define (derive-from-cell-with-nterms cell stack domainsize)
  (let ((stopflag #f)
        (nindex '())
        (range '())
        (origcterm '())
        (elem '())
        (res '())
        (lst '())
        (len 0) )
    (set! range (vector-ref *props-cell-range-vector* cell))
    (set! origcterm (vector-ref *props-cell-cterm-vector* cell))
    (set! len (vector-length origcterm))
    (do ((i 2 (+ i 1)))
        ((or stopflag (>= i len))
           (if stopflag
               #f
               '()) )
      (set! nindex (props-nterm-index-nr origcterm i domainsize))
      (set! lst (vector-ref *props-nterm-vector* nindex))
      (do ((part lst (cdr part)))
          ((or stopflag (null? part)))
        (set! elem (car part))
        (if (props-no-intersection range (cdr elem))
            ; now we can derive something
            (begin
              (set! *props-nterm-derive-count* (+ 1 *props-nterm-derive-count*))
              (set! res (props-restrict-cell!
                          (car elem) (vector-ref origcterm i)
                          stack domainsize))
              (if (not res) (set! stopflag #t)) ))))))


; derive-from-nterm-with-cells should attempt to
; generate restrictions for nterm ntermindex, with all
; corresponding cells that have cell nr cell at position pos


(define (derive-from-nterm-with-cells ntermindex pos cell stack domainsize)
  (let ((stopflag #f)
        (elem '())
        (range '())
        (matchcells '())
        (matchrange '())
        (matchcell '())
        (res '()) )

    ;(newline)
    ;(display "derive-from-nterm-with-cells called: ")(newline)
    ;(display "ntermindex: ")(display ntermindex)(newline)
    ;(display "pos: ")(display pos)(newline)
    ;(display "cell: ")(display cell)(newline)

    (set! elem
          (assq cell (vector-ref *props-nterm-vector* ntermindex)) )

    (if (and elem
            (not (null? (cdr elem))) )
        (begin
          (set! range (cdr elem))
          (set! matchcells (vector-ref *props-nterm-matchcells-vector* ntermindex))

          ;(newline)
          ;(display "elem: ")(display elem)(newline)
          ;(display "range: ")(display range)(newline)
          ;(display "matchcells: ")(display matchcells)(newline)

          (do ((part matchcells (cdr part)))
              ((or stopflag (null? part))
                  (if stopflag
                      #f
                      '()))
            (set! matchcell (car part))
            (set! matchrange (vector-ref *props-cell-range-vector* matchcell))
            (if (props-no-intersection range matchrange)
              ; now we can derive something
              (begin

                ;(display "matchcell: ")(display matchcell)(newline)
                ;(display "matchrange: ")(display matchrange)(newline)

                (set! *props-nterm-derive-count* (+ 1 *props-nterm-derive-count*))
                (set! res (props-restrict-cell!
                             cell
                             (vector-ref
                                 (vector-ref *props-cell-cterm-vector* matchcell)
                                 pos)
                             stack
                             domainsize))
                (if (not res) (set! stopflag #t)) )))))))



;(define (nterm-cellarg-pos nterm)
;  (let ((res #f)
;        (len (vector-length nterm)) )
;    (do ((i 2 (+ 1 i)))
;        ((or res (>= i len))
;           res)
;     (if (not (integer? (vector-ref nterm i)))
;         (set! res i) ))))



;------------------------------------------------------------------------
;
;          a block for making ground clauses from prepared nonground
;
;-------------------------------------------------------------------------

(define *props-ground-clauses* '())

(define (props-saturate-clauses clauselst domainsize)
  (let ((satres #f)
        (unsatisfiable #f) )
    (set! *props-ground-clauses* '())
    (do ((lstpart clauselst (cdr lstpart)))
        ((or (null? lstpart)
             unsatisfiable
             *props-stop-search-flag*
              (> (- (get-check-time 1) *slow-start-time*)
	         *one-pass-time*) )
	   (if unsatisfiable
	       #t
	       *props-ground-clauses*) )
      (set! unsatisfiable
           (props-saturate-clause! (car lstpart) domainsize) ))))


(define (props-saturate-clause! clause domainsize)
  (let ((varslist (vector-ref clause 0))
        (varvaluations '())
        (unsatisfiable #f)
        (allchecked #f)
        (propclause '()) )

    ;(newline)(display "clause to be saturated: ")(newline)
    ;(print-props-nonground-clause clause)
    ;(newline)(display "domain size: ")(display domainsize)(newline)

    (if (not (null? varslist))
        (begin
          (set! varvaluations
               (map (lambda(x) 0) varslist) )))

    ; loop over all the valuations for vars
    (do ()
        ((or unsatisfiable
             allchecked
             *props-stop-search-flag*)
	   (if *props-stop-search-flag*
	       #f
	       unsatisfiable) )

	   ;(if (and limitlocalclausesnr
	   ;         (> clausesgenerated limitlocalclausesnr) )
	   ;    (set! *prop-clauses-generation-limited-flag* #t) )
	   ;(if (and (not unsatisfiable)(not allchecked))
           ;    (set! *prop-time-ran-out* #t) )
           ;unsatisfiable

      (if (> (- (get-check-time 1) *slow-start-time*)
	     *one-pass-time*)
	  (set! *props-stop-search-flag* #t) )

      (set-pred-var-values varslist varvaluations)
      (set! propclause (props-make-ground-clause clause))
      (cond
        ((not propclause)
           (set! unsatisfiable #t) )
        ((not (null? propclause))
          (begin
            (set! *props-clause-count* (+ 1 *props-clause-count*))
            (if (> *props-clause-count* *props-max-allowed-clauses*)
                (set! *props-stop-search-flag* #t) )
            (set! *props-ground-clauses*
                 (cons propclause *props-ground-clauses*) ))))
      ;(newline)(display "valuations: ")(display varvaluations)
      (set! allchecked
           (or (null? varvaluations)
               (props-valuation-list-nextvalues!
                   varvaluations domainsize) )))))


(define (props-make-ground-clause term)
  (let ((tmp '())
        (lst '())
        (lstlen 0)
        (sameflag #f)
        (tautflag #f)
        (falseflag #f)
        (vec '())
        (len 0) )

     ; first make a list

     (set! len (vector-length term))
     (set! vec '())
     (do ((i 2 (+ 1 i)))
         ((or tautflag (>= i len)))
       (set! tmp (vector-ref term i))
       (if (integer? tmp)
           (set! tmp (var-ref tmp))
           (set! tmp (props-make-ground-term tmp vec i)) )
       (set! sameflag #f)
       (set! tautflag #f)
       (set! falseflag #f)
       (cond
        ((eq? #t tmp) (set! tautflag #t))
        ((eq? #f tmp) (set! falseflag #t)) )
       (do ((part lst (cdr part)))
           ((or tautflag falseflag (null? part)))
         (if (props-equal-terms-nopolarity (car lst) tmp)
             (if (eq? (logand *props-isneglit-mask* (vector-ref (car lst) 1))
                      (logand *props-isneglit-mask* (vector-ref tmp 1)) )
                 (set! sameflag #t)
                 (set! tautflag #t) )))
       (if (and (not sameflag)(not tautflag)(not falseflag))
           (set! lst (cons tmp lst)) ))

     ; second, make a clause of non-repeated literals (if not tautology)
     ; in lst

     (cond
       (tautflag
           '() )
       ((null? lst)
           #f)
       (else
           (set! lstlen (length lst))
           (set! len (+ 2 lstlen))
           (set! vec (make-vector len '()))
           (do ((i 2 (+ 1 i))
                (part lst (cdr part)) )
               ((>= i len)
                  (vector-set! vec 0 '())
                  (vector-set! vec 1 '())
                  vec)
            (set! tmp (car part))
            (vector-set! tmp 0 (cons vec i))
            (vector-set! vec i tmp) ))) ))


(define (props-make-ground-term term parent index)
  (let ((tmp '())
        (vec '())
        (fun 0)
        (len 0) )
    (cond
      ((integer? term)
         (var-ref term) )
      (else
         (set! len (vector-length term))
         (set! vec (make-vector len '()))
         (do ((i 2 (+ 1 i)))
             ((>= i len)
                 (vector-set! vec 0 (cons parent index))
                 (set! fun (vector-ref term 1))
                 (vector-set! vec 1 fun)
                 (if (props-eq-pred? (vector-ref vec 1))
                     (begin
                       (set! tmp (props-eq-lit-value-check vec))
                       (if (null? tmp)
                           vec
                           tmp))
                     vec))
           (set! tmp (vector-ref term i))
           (if (integer? tmp)
               (set! tmp (var-ref tmp))
               (set! tmp (props-make-ground-term tmp vec i)) )
           (vector-set! vec i tmp) )))))

; props-eq-lit-value-check assumes the lit is an eq lit
; and returns:
; #t iff lit is surely true
; #f iff lit is surely false
; '() else

(define (props-eq-lit-value-check term)
  (let ((a1 (vector-ref term 2))
        (a2 (vector-ref term 3)) )
   (cond
     ((props-equal-terms-nopolarity a1 a2)
       (if (props-pos-pred? (vector-ref term 1))
           #t
           #f))
     ((and (integer? a1) (integer? a2)
           (not (eq? a1 a2)) )
       (if (props-pos-pred? (vector-ref term 1))
           #f
           #t))
     (else '()) )))


; props-eq-lit-fast-value-check assumes the lit is an eq lit
; and returns:
; #t iff lit is surely true
; #f iff lit is surely false
; '() else

(define (props-eq-lit-fast-value-check term)
  (let ((a1 (vector-ref term 2))
        (a2 (vector-ref term 3)) )
   (if (and (integer? a1) (integer? a2))
       (if (eq? a1 a2)
           (if (props-pos-pred? (vector-ref term 1))
               #t
               #f)
           (if (props-pos-pred? (vector-ref term 1))
               #f
               #t))
       '() )))


(define (props-valuation-list-nextvalues! valuationlist domainsize)
    (if (and (pair? valuationlist)
             (>= (car valuationlist) (- domainsize 1)) )
        (begin
              (if (null? (cdr valuationlist))
                   #t
                   (begin
                     (set-car! valuationlist 0)
                     (props-valuation-list-nextvalues!
                         (cdr valuationlist) domainsize) )))
        (begin (set-car! valuationlist (+ 1 (car valuationlist)))
               #f)))


;------------------------------------------------------------------------
;
;          a block for ground clause and term handling without cterms
;
;-------------------------------------------------------------------------

(define (props-equal-terms-nopolarity term1 term2)
  (let ((len 0)
        (stopflag #f) )
    (cond
      ((number? term1)
         (eq? term1 term2) )
      ((number? term2) #f)
      ((not (eq? (logand *props-funnr-mask* (vector-ref term1 1))
                 (logand *props-funnr-mask* (vector-ref term2 1)) ))
         #f)
      (else
        (set! len (vector-length term1))
        (do ((i 2 (+ i 1)))
            ((or stopflag (>= i len))
                (not stopflag) )
          (if (not (props-equal-terms-nopolarity
                         (vector-ref term1 i) (vector-ref term2 i) ))
              (set! stopflag #t) ))))))


;------------------------------------------------------------------------
;
;     a block for making prepared nonground clauses suitable for grounding
;
;-------------------------------------------------------------------------

(define *props-pred-nr-table* '())
(define *props-fun-nr-table* '())
(define *props-const-nr-table* '())
(define *props-fun-counter* 0)


(define (props-make-nonground-clauses clauses)
  (let ((res '()))
    (set! *props-pred-nr-table* '())
    (set! *props-fun-nr-table* '())
    (set! *props-const-nr-table* '())
    (set! *props-fun-counter* 0)
    (map
      (lambda(cl)(props-make-nonground-clause cl))
      clauses)))

(define *props-nonground-vars-tmp* '())

(define (props-make-nonground-clause cl)
  (let ((tmp '())
        (vec '())
        (len (length (cdr cl))) )
    (set! *props-nonground-vars-tmp* '())
    (set! vec (make-vector (+ 2 len) '()))
    (set! cl (cdr cl))
    (do ((i 2 (+ 1 i)))
        ((null? cl)
           (vector-set! vec 0 *props-nonground-vars-tmp*)
            vec)
      (set! tmp (props-make-nonground-lit (car cl)))
      (set! cl (cdr cl))
      (vector-set! vec i tmp) )))


(define (props-make-nonground-lit lit)
  (let ((tmp '())
        (vec '())
        (lpart (cdr lit))
        (len (length lit)) )
    (set! vec (make-vector (+ 1 len) '()))
    (do ((i 2 (+ 1 i)))
        ((null? lpart)
            (set! tmp (props-make-pred-nr (car lit) (- len 1)))
            (vector-set! vec 1 tmp)
            vec)
      (set! tmp (props-make-nonground-term (car lpart)))
      (set! lpart (cdr lpart))
      (vector-set! vec i tmp) )))


(define (props-make-nonground-term term)
  (let ((tmp '())
        (vec '())
        (tpart term)
        (len 0) )
    (cond
      ((var? term)
         (if (not (memq (props-var-nr term)
                        *props-nonground-vars-tmp*))
             (set! *props-nonground-vars-tmp*
                   (cons (props-var-nr term) *props-nonground-vars-tmp*) ))
          (props-var-nr term) )
      ((not (composite? term))
         (set! vec (make-vector 2 '()))
         (vector-set! vec 1 (props-make-const-nr term))
         vec)
      (else
         (set! len (length term))
         (set! tpart (cdr term))
         (set! vec (make-vector (+ 1 len) '()))
         (do ((i 2 (+ 1 i)))
             ((null? tpart)
                 (set! tmp (props-make-fun-nr (car term) (- len 1)))
                 (vector-set! vec 1 tmp)
                 vec)
           (set! tmp (props-make-nonground-term (car tpart)))
           (set! tpart (cdr tpart))
           (vector-set! vec i tmp) )))))


(define (props-var-nr term)
  (var-nr term) )

(define (props-make-const-nr term)
  (let ((tmp '()))
    (set! tmp (assq term *props-const-nr-table*))
    (if tmp
        ;(logior *props-isconst-mask* (cdr tmp))
        (cdr tmp)
        (begin
          (set! tmp *props-fun-counter*)
          (set! *props-const-nr-table*
                (cons (cons term tmp)
                      *props-const-nr-table*))
          (vector-set! *props-fun-arities-vector* *props-fun-counter* 0)
          (vector-set! *props-fun-originals-vector* *props-fun-counter* term)
          (vector-set! *props-fun-originalsymb-vector*
                       *props-fun-counter*
                       (car (rassq term *const-num-table*)) )
          (set! *props-fun-counter* (+ 1 *props-fun-counter*))
          ; (logior *props-isconst-mask* tmp)
          tmp))))

(define (props-make-fun-nr term arity)
  (let ((tmp '())
        (tmp2 '())
        (pterm (pure-fun term)) )
    (set! tmp (assq pterm *props-fun-nr-table*))
    (if tmp
        (cdr tmp)
        (begin
          (set! tmp2 (rassq pterm *fun-num-table*))
          (set! tmp *props-fun-counter*)
          (set! *props-fun-nr-table*
                (cons (cons pterm tmp)
                      *props-fun-nr-table*))
          (vector-set! *props-fun-arities-vector* *props-fun-counter* arity)
          (vector-set! *props-fun-originals-vector* *props-fun-counter* pterm)
          (vector-set! *props-fun-originalsymb-vector*
                       *props-fun-counter*
                       (car tmp2) )
          (vector-set! *props-fun-ispred-vector* *props-fun-counter* #f)
          (set! *props-fun-counter* (+ 1 *props-fun-counter*))
          tmp))))


(define (props-make-pred-nr term arity)
  (let ((tmp '())
        (tmp2 '())
        (polarity #t)
        (pterm (pure-unsigned-fun term)) )
    (if (neg-fun? term)
        (set! polarity #f)
        (set! polarity #t) )
    (set! tmp (assq pterm *props-pred-nr-table*))
    (if tmp
        (if polarity
           (logior *props-islit-mask* (cdr tmp))
           (logior *props-islit-mask* (logior *sgn-mask* (cdr tmp))) )
        (begin
          (if (eq? pterm *eq-pred-nr*) (set! *props-eq-fun* *props-fun-counter*))
          (set! tmp2 (rassq pterm *pred-num-table*))
          (set! tmp *props-fun-counter*)
          (set! *props-pred-nr-table*
                (cons (cons pterm tmp)
                      *props-pred-nr-table*))
          (vector-set! *props-fun-arities-vector* *props-fun-counter* arity)
          (vector-set! *props-fun-originals-vector* *props-fun-counter* pterm)
          (vector-set! *props-fun-originalsymb-vector*
                       *props-fun-counter*
                       (car tmp2) )
          (vector-set! *props-fun-ispred-vector* *props-fun-counter* #t)
          (set! *props-fun-counter* (+ 1 *props-fun-counter*))
          (if polarity
              (logior *props-islit-mask* tmp)
              (logior *props-islit-mask* (logior *sgn-mask* tmp)) )))))



;------------------------------------------------------------------------
;
;          printing
;
;-------------------------------------------------------------------------

(define (print-props-nonground-clause clause)
  (display (vector-ref clause 0))
  (display " : ")
  ;(display clause)
  ;(newline)
  (do ((i 2 (+ 1 i)))
      ((>= i (vector-length clause)))
   (print-props-nonground-term (vector-ref clause i))
   (if (< i (- (vector-length clause) 1))
       (display " | ")
       (display ".") )))


(define (print-props-ground-clause clause)
 (if (not (vector? clause))
     (display clause)
     (begin
       (display (vector-ref clause 0))
       (display " : ")
       (do ((i 2 (+ 1 i)))
           ((>= i (vector-length clause)))
        (if (number? (vector-ref clause i))
            (display (vector-ref clause i))
            (begin
              (if (and
      	             (not (integer? (vector-ref clause i)))
                     (or (not (eq? clause (car (vector-ref (vector-ref clause i) 0))))
                         (not (eq? (vector-ref (car (vector-ref (vector-ref clause i) 0))
                                               (cdr (vector-ref (vector-ref clause i) 0)) )
                                   (vector-ref clause i) ))))
                  (begin
                    (newline)(display "clause literal does not point to clause")(newline)
                    (quit) ))
              (print-props-ground-term (vector-ref clause i)) ))
        (if (< i (- (vector-length clause) 1))
            (display " | ")
            (display ".") )))))


(define (print-props-nonground-term term)
  ;(newline)
  ;(display "term: ")(display term)(newline)
  (let ((tmp '()))
    (cond
      ((var? term)
        (display "v")
        (display term) )
      ((integer? term)
        (if (not (zero? (logand *props-isconst-mask* term)))
            (display "c") )
        (display (logand *props-funnr-mask* term)) )
      (else
        (if (not (zero? (logand *props-isneglit-mask* (vector-ref term 1))))
            (display "-") )
        (if (not (zero? (logand *props-islit-mask* (vector-ref term 1))))
            (display "l")
            (display "f") )
        (display (logand *props-funnr-mask* (vector-ref term 1)))
        ;(display ":")
        ;(display (length (cadr term)))
        (display "(")
        (do ((i 2 (+ 1 i)))
      	    ((>= i (vector-length term))
      	       (display ")") )
          (print-props-nonground-term (vector-ref term i))
          (if (< i (- (vector-length term) 1))
              (display ",") ))))))


(define (print-props-ground-term term)
  (let ((tmp '()))
    (cond
      ;((var? term)
      ;  (display "v")
      ;  (display term) )
      ((integer? term)
        (if (not (zero? (logand *props-isconst-mask* term)))
            (display "c") )
        (display (logand *props-funnr-mask* term)) )
      (else
        ;(if (not (zero? (logand *props-isneglit-mask* (vector-ref term 1))))
        ;    (display "-") )
        ;(cond
        ;  ((props-eq-pred? (vector-ref term 1))
        ;     (display "e") )
        ;  ((vector-ref *props-fun-ispred-vector*
        ;              (logand *props-funnr-mask* (vector-ref term 1)) )
        ;     (display "l") )
        ;  ;((not (zero? (logand *props-islit-mask* (vector-ref term 1))))
        ;  ;  (display "l") )
        ;  (else
        ;    (display "f") ))
        (display (logand *props-funnr-mask* (vector-ref term 1)))
        ;(display ":")
        ;(display (length (cadr term)))
        (display "(")
        (do ((i 2 (+ 1 i)))
      	    ((>= i (vector-length term))
      	       (display ")") )
      	  (if (and
      	        (not (integer? (vector-ref term i)))
      	        (or (not (eq? term (car (vector-ref (vector-ref term i) 0))))
                    (not (eq? (vector-ref (car (vector-ref (vector-ref term i) 0))
                                          (cdr (vector-ref (vector-ref term i) 0)) )
                              (vector-ref term i) ))))
      	      (begin
      	        (newline)(display "term does not point to superterm")(newline)
      	        (quit) ))
          (print-props-ground-term (vector-ref term i))
          (if (< i (- (vector-length term) 1))
              (display ",") ))))))



(define (print-props-model-term term)
  (let ((tmp '()))
    (cond
      ;((var? term)
      ;  (display "v")
      ;  (display term) )
      ((integer? term)
        (if (not (zero? (logand *props-isconst-mask* term)))
            (display "c") )
        (display (logand *props-funnr-mask* term))
        (display (vector-ref *props-fun-originalsymb-vector*
                             (logand *props-funnr-mask* term))) )

      (else
        (if (not (zero? (logand *props-isneglit-mask* (vector-ref term 1))))
            (display "-") )
        ;(cond
        ;  ((props-eq-pred? (vector-ref term 1))
        ;     (display "e") )
        ;  ((vector-ref *props-fun-ispred-vector*
        ;              (logand *props-funnr-mask* (vector-ref term 1)) )
        ;     (display "l") )
        ;  ;((not (zero? (logand *props-islit-mask* (vector-ref term 1))))
        ;  ;  (display "l") )
        ;  (else
        ;    (display "f") ))
        ;(display (logand *props-funnr-mask* (vector-ref term 1)))
        (if (< (vector-ref term 1) 70000)
            (display (vector-ref *props-fun-originalsymb-vector*
                             (logand *props-funnr-mask*
                               (vector-ref term 1) )))
            (display (vector-ref *props-fun-originalsymb-vector*
                             (logand *props-funnr-mask*
                               (vector-ref term 1) ))))
        ;(display ":")
        ;(display (length (cadr term)))
        (display "(")
        (do ((i 2 (+ 1 i)))
      	    ((>= i (vector-length term))
      	       (display ")") )
      	  (if (and
      	        (not (integer? (vector-ref term i)))
      	        (or (not (eq? term (car (vector-ref (vector-ref term i) 0))))
                    (not (eq? (vector-ref (car (vector-ref (vector-ref term i) 0))
                                          (cdr (vector-ref (vector-ref term i) 0)) )
                              (vector-ref term i) ))))
      	      (begin
      	        (newline)(display "term does not point to superterm")(newline)
      	        (quit) ))
          (print-props-ground-term (vector-ref term i))
          (if (< i (- (vector-length term) 1))
              (display ",") ))))))

(define (props-print-vectors)
   (let ()
     (newline)
     (display "function vectors: ")
     (newline)
     (do ((i 0 (+ 1 i)))
         ((or (>= i *props-fun-counter*)
              (not (vector? *props-fun-originalsymb-vector*))
              (not (vector? *props-fun-arities-vector*))
              (not (vector?  *props-fun-basenr-vector*)) ))
       (display i)(display " : ")
       (display " orig=")(display (vector-ref *props-fun-originalsymb-vector* i))
       (display " orignr=")(display (vector-ref *props-fun-originals-vector* i))
       (display " ispred=")(display (vector-ref *props-fun-ispred-vector* i))
       (display " arity=")(display (vector-ref *props-fun-arities-vector* i))
       (display " base=")(display (vector-ref *props-fun-basenr-vector* i))
       (newline) )
     (newline)
     (display "cell vectors: ")
     (newline)
     (do ((i 0 (+ 1 i)))
         ((or (>= i *props-cell-nr*)
              (not (vector? *props-cell-cterm-vector*)) ))
       (display i)(display " : ")
       (display "orig=")(print-props-ground-term (vector-ref *props-cell-cterm-vector* i))
       (display " cterm=")(display (vector-ref *props-cell-cterm-vector* i))
       (display " range=")(display (vector-ref *props-cell-range-vector* i))
       (display " occs=")(display (length (vector-ref *props-cell-occurrences-vector* i)))
       ;(newline)
       ;(for-each (lambda(x)(print-props-ground-term x)(newline))
       ;          (vector-ref *props-cell-occurrences-vector* i) )
       (newline) )
      (newline)
     (display "nterm vectors: ")
     (newline)
     (do ((i 0 (+ 1 i)))
         ((>= i *props-nterm-nr*))
       (if (vector? (vector-ref *props-nterm-originals-vector* i))
           (begin
             (display i)(display " : ")
             (display "orig=")
             (print-props-ground-term
                        (vector-ref *props-nterm-originals-vector* i))
             (display " values=")
             (display (vector-ref *props-nterm-vector* i))
             (display " matchcells=")
             (display (vector-ref *props-nterm-matchcells-vector* i))
             (newline) )))
       ))




;------------------------------------------------------------------------
;
;          model testing
;
;-------------------------------------------------------------------------

; use the cell table prepared by the model finder,
; take nonground clauses as input,
; compute all the ground clauses and then analyze each clause:
; a clause must be a tautology.
; if some clause is not a tautology, our model may be erroneous.

(define (props-check-model clauses domainsize)
  (let ((res '())
        (stopflag #f) )
    (do ((part clauses (cdr part)))
        ((or (null? part) stopflag)
           (if stopflag
               (begin
                 (newline)
                 (display "NB! wrong model found, not accepted, search stopped.")
                 (newline)
                 #f)
                #t))
     (set! res (props-check-saturate-clause (car part) domainsize))
     (if (not (eq? res #t)) (set! stopflag #t)) )))


(define (props-check-saturate-clause clause domainsize)
  (let ((varslist (vector-ref clause 0))
        (varvaluations '())
        (badflag #f)
        (allchecked #f)
        (propclause '()) )

    ;(newline)(display "checking nonprop clause: ")(newline)
    ;(print-props-nonground-clause clause)
    ;(newline)(display "domain size: ")(display domainsize)(newline)

    (if (not (null? varslist))
        (begin
          (set! varvaluations
               (map (lambda(x) 0) varslist) )))

    ; loop over all the valuations for vars
    (do ()
        ((or badflag
             allchecked)
           (if badflag
               #f
               #t) )
      (set-pred-var-values varslist varvaluations)
      (set! propclause (props-check-make-ground-clause clause domainsize))
      (if (not (eq? #t propclause))
          (set! badflag #t) )
      ;(newline)(display "valuations: ")(display varvaluations)
      (set! allchecked
           (or (null? varvaluations)
               (props-valuation-list-nextvalues!
                   varvaluations domainsize) )))))


(define (props-check-make-ground-clause term domainsize)
  (let ((tmp '())
        (res '())
        (tautflag #f)
        (vec '())
        (len 0) )

     ;(newline)(display "props-check-make-ground-clause called: ")(display term)

     (set! len (vector-length term))
     (do ((i 2 (+ 1 i)))
         ((or tautflag (>= i len))
             (if (not tautflag)
                  #f
                  #t))
       (set! tmp (vector-ref term i))

       ;(newline)(display "checking literal: ")(display tmp)

       (if (integer? tmp)
           (set! res (var-ref tmp))
           (set! res (props-check-make-ground-term tmp domainsize)) )

       ;(newline)(display "literal check gave: ")(display res)

       (cond
         ((eq? res 1)
            (if (props-pos-pred? (vector-ref tmp 1))
                (set! tautflag #t) ))
         ((eq? res 0)
            (if (not (props-pos-pred? (vector-ref tmp 1)))
                (set! tautflag #t) )) ))))

(define (props-check-make-ground-term term domainsize)
  (let ((tmp '())
        (vec '())
        (fun 0)
        (len 0) )

    ;(newline)(display "props-check-make-ground-term called: ")(display term)

    (cond
      ((integer? term)
         (var-ref term) )
      (else
         (set! len (vector-length term))
         (set! vec (make-vector len '()))
         (do ((i 2 (+ 1 i)))
             ((>= i len)
                 (vector-set! vec 0 '())
                 (set! fun (vector-ref term 1))
                 (vector-set! vec 1 fun)

                 ;(newline)(display "compute grounded term: ")(display vec)

                 (set! tmp
                      (props-check-compute-term vec domainsize) )

                 ;(newline)(display "term value: ")(display tmp)

                 tmp)

                 ;(if (props-eq-pred? (vector-ref vec 1))
                 ;    (begin
                 ;      (set! tmp (props-eq-lit-value-check vec))
                 ;      (if (null? tmp)
                 ;          vec
                 ;          tmp))
                 ;     vec)

           (set! tmp (vector-ref term i))
           (if (integer? tmp)
               (set! tmp (var-ref tmp))
               (set! tmp (props-check-make-ground-term tmp domainsize)) )
           (vector-set! vec i tmp) )))))


(define (props-check-compute-term term domainsize)
  (let ((res '())
        (cell '()) )
    (set! cell (props-cterm-index-nr term domainsize))
    (if cell
        (car (vector-ref *props-cell-range-vector* cell))
        term)))

;------------------------------------------------------------------------
;
;          utilities
;
;-------------------------------------------------------------------------


(define (props-eq-pred? prednr)
   (eq? (logand *props-purefun-mask* prednr)
        *props-eq-fun*) )

(define (props-is-pred? prednr)
  (not (zero? (logand *props-islit-mask* prednr))) )

(define (props-term-is-surely-lit? term)
  (let ((fun (vector-ref term 1)))
    (vector-ref *props-fun-ispred-vector* fun) ))

(define (props-pos-pred? prednr)
   (zero? (logand *props-isneglit-mask* prednr)) )

 (define (props-pure-fun funnr)
   (logand *props-purefun-mask* funnr) )

(define (int-power nr exp)
   (case exp
     ((0) 1)
     ((1) nr)
     ((2)(* nr nr))
     ((3)(* nr (* nr nr)))
     ((4)(* nr (* nr (* nr nr))))
     ((5)(* nr (* nr (* nr (* nr nr)))))
     ((6)(* nr (* nr (* nr (* nr (* nr nr))))))
     ((7)(* nr (* nr (* nr (* nr (* nr (* nr nr)))))))
     (else (slow-int-power nr exp)) ))

(define (slow-int-power nr exp)
 (let ((res 1))
  (do ((i exp (- i 1)))
      ((< i 1)
        res)
    (set! res (* res nr)) )))

(define (props-cterm-max-arg cterm)
  (let ((len (vector-length cterm))
        (res -1)
        (tmp -1) )
    (do ((i 2 (+ 1 i)))
        ((>= i len)
           res)
     (set! tmp (vector-ref cterm i))
     (if (> tmp res) (set! res tmp)) )))


(define (props-cterm-sum-arg cterm)
  (let ((len (vector-length cterm))
        (res 0)
        (tmp 0) )
    (do ((i 2 (+ 1 i)))
        ((>= i len)
           res)
     (set! tmp (vector-ref cterm i))
     (set! res (+ res tmp)) )))


;(define (int-power nr exp)
;  (if (< exp 1)
;      1
;      (* nr (int-power nr (- exp 1))) ))

(define (make-list nr elem)
  (if (< nr 1)
      '()
      (cons elem (make-list (- nr 1) elem)) ))

(define (remove-one-from-list elem lst)
  (cond
   ((null? lst)
      lst)
   ((eq? elem (car lst))
      (cdr lst) )
   (else
      (cons (car lst) (remove-one-from-list elem (cdr lst))) )))

(define (props-no-intersection alst blst)
 (let ((stopflag #f))
   (do ((part alst (cdr part)))
       ((or stopflag (null? part))
            (not stopflag))
    (if (memq (car part) blst)
        (set! stopflag #t) ))))
