;======================================================================
;
;                          resstrat.scm
;
;                   Gandalf version c-2.6 r1
;
;             Special strategies, strategy selection
;
;                  Copyright (c) Tanel Tammet
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 1, or (at your option)
; any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; The author can be reached at tammet@staff.ttu.ee, tanel@cs.ioc.ee or
; tammet@cs.chalmers.se
;
;
;=======================================================================


;=====================================================================

; Hyper-contradiction tries to produce an immediate contradiction from cl
; using (generalised) hyperresolution.
; If that fails, nothing is produced.
;
; cl is always longer than one literal (otherwise a contradiction
; should have been found earlier). If cl contains ground literals,
; it is not checked (ground literals should have been uniit-deleted).
;
; called in handle-derived-inf-cl-pair-aux
; iff *hyper-contradiction-strategy* is #t

(define (hyper-contradiction? fullcl)
  (if (and (not *foundflag*)
           (not (empty-cl? (cdr fullcl)))
	   (not (unit-cl? (cdr fullcl)))
	   (every? (lambda(lit)(term-may-cont-vars? lit)) (cdr fullcl)) )
      (begin
	(if (>= *display-info-level* 70)
	    (begin
	      (newline)
	      (display "hyper-contradiction? entered: ")
	      (print-full-clause fullcl)
	      (newline) ))
	(set!  *unit-clash-flag* #t) ; ONLY #t during hyper-contradiction
	(let* ((ycl (map rename-x-to-y-term (cdr fullcl)))
	       (yinf (car fullcl))
	       (ylitinf (caddr yinf))
	       (one-step-results (one-step-for-hyper (cons yinf ycl))) )
	  ; loop for all partial (one-step) results
	  (do ((results-part one-step-results (cdr results-part)))
	      ((or (null? results-part)
		   *foundflag*
		   *stop-fast-loop-flag*
	           (and (if  (and (not *stop-fast-loop-flag*)
				  (> (- (get-check-time 63) *slow-start-time*)
				     *one-pass-time*))
			     (begin
			       (store-time-checkpoint! 63)
			       (set! *stop-fast-loop-flag* #t)
			       #t)))))
	    (hyper-contradiction? (car results-part)) ))
	(set! *unit-clash-flag* #f) ; ONLY #t during hyper-contradiction
	*foundflag*)
      #f))

; full-hyper-contradiction? allows fullcl to contain also ground
; literals


(define (full-hyper-contradiction? fullcl)
    (if (and (not *foundflag*)
           (not (empty-cl? (cdr fullcl)))
	   (not (unit-cl? (cdr fullcl))) )
      (begin
	(if (>= *display-info-level* 70)
	    (begin
	      (newline)
	      (display "full-hyper-contradiction? entered: ")
	      (print-full-clause fullcl)
	      (newline) ))
	(set!  *unit-clash-flag* #t) ; ONLY #t during hyper-contradiction
	(let* ((ycl (map rename-x-to-y-term (cdr fullcl)))
	       (yinf (car fullcl))
	       (ylitinf (caddr yinf))
	       (one-step-results (one-step-for-hyper (cons yinf ycl))) )
	  ; loop for all partial (one-step) results
	  (do ((results-part one-step-results (cdr results-part)))
	      ((or (null? results-part)
		   *foundflag*
		   *stop-fast-loop-flag*
		   (and (if  (and (not *stop-fast-loop-flag*)
				  (> (- (get-check-time 62) *slow-start-time*)
				     *one-pass-time*))
			     (begin
			       (store-time-checkpoint! 62)
			       (set! *stop-fast-loop-flag* #t)
			       #t)))))
	    (hyper-contradiction? (car results-part)) ))
	(set! *unit-clash-flag* #f) ; ONLY #t during hyper-contradiction
	*foundflag*)
      #f))

; start-binary-hypercheck-strategy first checks all the existing
; clauses for immediate hyper-contradiction and then allows the
; ordinary binary resolution to continue with full immediate check
; of each derived clause


(define (start-binary-hypercheck-strategy)
  (let ()
  ; first for usable
    (do ((part *usable-clauses-list* (cdr part)))
	((or (null? part)
	     *foundflag*
	     *stop-fast-loop-flag*))
      (if (and (pair? (car part))
	       (not (unit-cl? (car part))) )
	  (full-hyper-contradiction? (car part)) ))
    ; second for sos
    (do ((part *sos-clauses-list*  (cdr part)))
	((or (null? part)
	     *foundflag*
	     *stop-fast-loop-flag*))
      (if (and (pair? (car part))
	       (not (unit-cl? (car part))) )
	  (full-hyper-contradiction? (car part))  ))
    ; set the full-check strategy
    (set! *hyper-contradiction-strategy* #t) ))



; start-backward-hyper-analysis retains only fully negative
; clauses in usable, all others are moved to sos.
; Only nonunit clauses are then retained in sos (unit clauses
; removed - they are left only in *unit-alist*).
; All non-answer literals are made resolvable (no ordering!)

(define (start-backward-hyper-analysis)

;(newline)
;(display "units:")
;(newline)
;(for-each (lambda(y)
;	    (newline)
;	    (for-each
;	     (lambda(x) (if (pair? x) (print-clause x))(newline))
;	     (cdr y) ))
;	  *unit-alist*)


;(newline)
;(display "usable:")
;(newline)
; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))
;*usable-clauses-list*)

;(newline)
;(display "sos:")
;(newline)
;(for-each (lambda(x)(if (pair? x) (print-clause x))(newline))
;*sos-clauses-list*)

  (let ((newsos '())
	(newusable '())
	(cl '()) )

    ; set strategic flags
    (init-binres-strat)
    (set! *hyper-contradiction-strategy* #t)

    (set! *unit-strategy-flag* #f)
    ; (set! *factors-stratflag* #f)
    (set! *factors-stratflag* #t)
    (set! *ordered-factor-flag* #f)
    (set! *order-strategy-flag* #f)
    (set! *poslitpref-order-strategy-flag* #t)
    (set! *name-order-strategy-flag* #f)
    ; (set! *paramodulation-allowed-flag* #f)
    (set! *paramodulation-allowed-flag* #t)
    (set! *remove-identical-lit-stratflag* #t)

    (set! *sos-preference-cur-ratio* 2)

    ; clear nonunit subsumption stuff

    (set! *nonunit-discr-tree* (list '()))

;(trace havealready-clause? )
;(trace check-nonunit-discr-tree)
;(trace   one-step-for-hyper)
;(trace hyper-contradiction?)
;(trace given-clause-slow-loop)
;(trace usable-clauses-fast-loop)
;(trace two-clause-binres-loop)
;(trace one-step-for-hyper)
;(trace resolve-unify-ordinary)
;(trace make-poslit-preferred!)
;(trace mixed-cl?)
;(trace make-all-resolvable!)
;(trace handle-derived-inf-cl-pair-aux )
;(trace handle-derived-inf-cl-pair)
;(trace insert-clause-discr-tree!)

;(set! *display-info-level* 200)
;(trace full-unit-cutoff!)
;(trace full-unit-cutoff-double! )

    ; filter usable
    (do ((part *usable-clauses-list* (cdr part)))
	((or *foundflag*
	     (null? part) ))
      (set! cl (car part))
      (if (pair? cl)(make-poslit-preferred! cl))
      (cond ((not (pair? cl)))
	    ((negative-cl? (cdr cl))
  	         (hyper-contradiction? cl) ; test for immediate contradiction!
		 (set! newsos (cons cl newsos)) )
	     ; ((unit-cl? (cdr cl))) ; remove nonnegative units
	     ; ((unit-cl? (cdr cl))) ; remove nonnegative units
	     (else
	        (set! newusable (cons cl newusable)) )))
    (set! *usable-clauses-list* newusable)


    ; filter sos

    (do ((part *sos-clauses-list* (cdr part)))
	((or *foundflag*
	     (null? part)))
      (set! cl (car part))
      (if (pair? cl)(make-poslit-preferred! cl))
      (cond
        ((not (pair? cl)))
	((negative-cl? (cdr cl))
	    (hyper-contradiction? cl)
            (set! newsos (cons cl newsos)) )
	(else
	    (set! *usable-clauses-list*
		  (cons cl *usable-clauses-list*) ))))

    (set! *sos-clauses-list* newsos)

    (if (not *int-logic-flag*)
	(set! *sos-clauses-list-end* (my-last-pair *sos-clauses-list*)) )
    ; (set! *usable-clauses-list* (cons '() *usable-clauses-list*))
    (set! *usable-clauses-list-end* (my-last-pair *usable-clauses-list*))


;(newline)
;(display "usable:")
;(newline)
; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))
;  *usable-clauses-list*)

;(newline)
;(display "sos:")
;(newline)
; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))
; *sos-clauses-list*)

    ; start search all over again

    ))

; make-all-resolvable! changes litinfolst of a clause destructively,
; making all non-answer-literals resolvable.

(define (make-all-resolvable! fullcl)
  (let ((tmp 0)
	(x 0) )
    (do ((lit-part (caddr (car fullcl)) (cdr lit-part)))
	((null? lit-part))
      (set! x (car lit-part))
      (if (zero? (logand *litinf-isanswer-mask* (car x)))
	  (begin
	    (set! tmp (logxor *litinf-nouse-mask*
			  (logior *litinf-nouse-mask* (car x)) ))
	    (set-car! x tmp) )))))

; make-poslit-preferred! changes litinfolst of a clause destructively,
; making positive literals preferred.


(define (make-poslit-preferred! fullcl)
  (if (not (mixed-cl? (cdr fullcl)))
      ; non-mixed clauses are fully cuttable
      (make-all-resolvable! fullcl)
      ; we have a mixed clause
      (do ((info-part (caddr (car fullcl)) (cdr info-part))
	   (lit-part (cdr fullcl) (cdr lit-part)) )
	  ((null? info-part))
	  (if (zero? (logand *litinf-isanswer-mask* (caar info-part)))
	      (if (neg-fun? (caar lit-part))
		  ; negative literals prohibited
		  (set-car!
		   (car info-part)
		   (logior *litinf-nouse-mask* (caar info-part)) )
		  ; positive literals allowed
		  (set-car!
		   (car info-part)
		   (logxor *litinf-nouse-mask*
			   (logior *litinf-nouse-mask*
				   (caar info-part) ))))))))


;----------------------------------------------------------------
;
;         splitting for sat
;
;----------------------------------------------------------------


; - - - - - - - - - - - - - - - - - - - - - -
;
; Splitting flat clauses
;
; - - - - - - - - - - - - - - - - - - - - - -

; split-flat-clause-set does a full splitting of a clause set,
; internally recursively splitting splitting results of
; successfully split clauses
;
; NB! internally the new predicate symbols are also
; added to the lists:
; *prop-orig-preds-enum*: ((orignr . flatnr) ...)
; *prop-orig-preds*: ((origname . arity) ...)

(define *prop-const-min* 80000)
;(define *splitted-clause-set-len* 0)

(define (split-flat-clause-set clauses)
  (let ((res '())
        (tmp '()) )
    ;(set! *splitted-clause-set-len* (length clauses))
    ;(newline)
    ;(display "Clauses to split: ")(display *splitted-clause-set-len*)
    ;(newline)
    (for-each
       (lambda(clause)

          ;(newline)
          ;(display clause)

          (if (null? (cddr clause))
              (set! res (cons clause res))
              (set! res (append (split-flat-clause-set-aux (list clause)) res)) ))
        clauses)
     res))


; split-flat-clause-set-aux does recursive splitting of
; a clause set: splits are done again and again until no new
; splits are produced

(define (split-flat-clause-set-aux clauses)
  (let ((res '())
        (tmp '()) )
    (for-each
       (lambda(clause)
          (if (null? (cddr clause))
              (set! res (cons clause res))
              (set! res (append (split-flat-clause clause) res)) ))
        clauses)
    ;(newline)(display "res: ")(display res)
    ;(newline)
    (if (<= (length res) (length clauses))
        res
        (split-flat-clause-set-aux res) )))

; split-flat-clause splits a clause into components, and then
; builds a list of new correct clauses from these splits,
; containing also the split-joining clauses
;
; input list: (() (10 1 5) (20 1 4) (30 10 20) (40 20 10))
; intermediate result:
;  (() (() ((10 20) (30 10 20) (40 20 10))) ((1) ((5 1) (10 1 5)) ((4 1) (20 1 4))))
; which contains two sublists:
; (() ((10 20) (30 10 20) (40 20 10)))    - contains one split component only
; ((1) ((5 1) (10 1 5)) ((4 1) (20 1 4))) - contains two nonground components over (1)
; the result is used to create the split clauses:
; ((split1)  (split2))              - ground splits combining clause
; ((-split1) (30 10 20) (40 20 10)) - first ground split component
; ((-split2) (split3 1) (split4 1)  - nonground split combining clause
; ((-split3 1) (10 1 5))             - nonground split 1
; ((-split4 1) (20 1 4))             - nonground split 2

(define (split-flat-clause clause)
  (let ((res '())
        (splitcomponents '())
        (gsplit '())
        (ngsplit '())
        (ngsplitcomponents '())
        (ngsplitvars '())
        (newclause '())
        (splitliteral '())
        (splittedclauses '())
        (gjoinclause '())
        (ngjoinclause '())
        (ngjoinclauses '())
        (joinclauses '())
        (origclause clause) )
    (set! splitcomponents
          (split-flat-clause-into-components clause))
    ; splitcomponents is a list of ground splits, each
    ; being a list of nonground splits
    ; loop first over ground splits
    (do ((gsplits (cdr splitcomponents) (cdr gsplits)))
        ((null? gsplits)
          ;(newline)(display "gjoinclause: ")(display gjoinclause)
          ;(newline)(display "ngjoinclauses: ")(display ngjoinclauses)
          ;(newline)(display "splittedclauses: ")(display splittedclauses)
          ;(newline)
          (if (not (null? gjoinclause))
              (set! res (cons gjoinclause
                              (append ngjoinclauses  splittedclauses) ))
              (set! res (append ngjoinclauses splittedclauses)) )
          (set! res (map (lambda(x)(cons '() x)) res))
          res)
      (if (not (null? (cddr splitcomponents)))
          ; ground splits exist
          (begin
            (set! gjoinclause (cons (make-flat-split-literal '()) gjoinclause))
            (set! ngjoinclause
                  (list (cons (- 0 (caar gjoinclause)) (cdar gjoinclause))) ))
          ; no ground splits
          (begin
            (set! gjoinclause '())
            (set! ngjoinclause '()) ))
      (set! gsplit (car gsplits))
      (set! ngsplitcomponents (cdr gsplit))
      (set! ngsplitvars (car gsplit))
      ; loop over nonground splits
      (do ((ngsplits ngsplitcomponents (cdr ngsplits)))
          ((null? ngsplits)
             (if (not (null? ngjoinclause))
                 (set! ngjoinclauses (cons ngjoinclause ngjoinclauses)) ))
        (if (not (null? (cdr ngsplitcomponents)))
            ; nonground splits exist
            (begin
              (set! ngsplit (cdar ngsplits))
              (set! splitliteral (make-flat-split-literal ngsplitvars))
              (set! newclause (cons splitliteral ngsplit))
              (set! splittedclauses (cons newclause splittedclauses))
              (set! ngjoinclause
                   (cons (cons (- 0 (car splitliteral)) (cdr splitliteral))
                      ngjoinclause)))
            ; no nonground splits
            (if (not (null? gjoinclause))
                ; ground splits exist
                (begin
                  (set! ngsplit (cdar ngsplits))
                  (set! ngjoinclause '())
                  (set! splitliteral
                       (cons (- 0 (caar gjoinclause)) (cdar gjoinclause)) )
                  (set! newclause (cons splitliteral ngsplit))
                  (set! splittedclauses (cons newclause splittedclauses)) )
                ; ground splits also do not exist
                (begin
                  (set! ngsplit (cdar ngsplits))
                  (set! ngjoinclause '())
                  (set! splittedclauses (cons ngsplit splittedclauses)) )))))))


(define *flat-split-literal-nr* 1000)

(define (make-flat-split-literal vars)
  (if (null? vars) (set! vars (list *prop-const-min*)))
  (set! *flat-split-literal-nr* (+ 1 *flat-split-literal-nr*))
  (set!  *prop-orig-preds-enum*
         (cons (cons *flat-split-literal-nr* *flat-split-literal-nr*)
               *prop-orig-preds-enum*))
  (set! *prop-orig-preds*
        (cons (cons *flat-split-literal-nr* (length vars))
               *prop-orig-preds*))
  (cons *flat-split-literal-nr* vars) )


; split a single flat clause, outputting a list of split clauses
;
; the result is a list  of ground splits, each element (ground split)
; being actually a list of nonground splits inside this ground-split
; component. Each list of nonground splits is prefixed by a list of
; vars split over in a non-ground way.
; example:
; input list: (() (10 1 5) (20 1 4) (30 10 20) (40 20 10))
; result: (() (() ((10 20) (30 10 20) (40 20 10))) ((1) ((5 1) (10 1 5)) ((4 1) (20 1 4))))
; which contains two sublists:
; (() ((10 20) (30 10 20) (40 20 10)))    - contains one split component only
; ((1) ((5 1) (10 1 5)) ((4 1) (20 1 4))) - contains two nonground components over (1)
; the result is used to create the split clauses:
; ((split1)  (split2))              - ground splits combining clause
; ((-split1) (30 10 20) (40 20 10)) - first ground split component
; ((-split2) (split3 1) (split4 1)  - nonground split combining clause
; ((-split3 1) (10 1 5))             - nonground split 1
; ((-split4 1) (20 1 4))             - nonground split 2

(define (split-flat-clause-into-components clause)
  (let ((res '())
        (splitcomponents '())
        (origclause clause)
        (clausevars '())
        (joinvars '())
        (clausewithvars '()) )

    ;(newline)
    ;(display "clause: ")(display clause)

    (set! clausevars (get-fullclause-varslist clause))
    (set! clause (cdr clause))
    (set! clausewithvars
          (map (lambda(x)
                  (list (give-flat-lit-varset x) x) )
                clause))

    ;(newline)(display "with vars: ")(display clausewithvars)(newline)

    ; try to create splittable components

    (set! splitcomponents
          (make-flat-clause-splitcomponents clausewithvars joinvars))

    ;(newline)(display "initial ground split gives: ")(newline)
    ;(display splitcomponents)(newline)

    (set! splitcomponents (make-flat-clause-joingroundsplits splitcomponents))

    ;(newline)(display "joining ground literals gives: ")(newline)
    ;(display splitcomponents)(newline)

    (set! splitcomponents
          (map (lambda(x)
                 (make-flat-clause-nongroundsplits x) )
               splitcomponents))
    (set! splitcomponents (cons '() splitcomponents))

    ;(newline)(display "nonground splitting gives: ")(newline)
    ;(display splitcomponents)(newline)

    (cond
      ; empty or unit clause
      ((or (null? clause) (null? (cdr clause)))
        (set! res (list origclause)) )
      (else
        (set! res (list origclause)) ))
    ; return result set
    ;res
    splitcomponents))

; make-flat-clause-splitcomponents takes a clause split (incorrectly!)
; into single literal always and combines the splits to make a correct
; split.
; The second argument (joinvars) is a list of vars which are allowed
; to exist in SEVERAL splits. Ie, if joinvars is not empty, the function
; does non-ground splitting.
; It only uses the variable lists (sets) which are at the beginning
; of each component:
; ((varlist12 lit1 lit2) (varlist34 lit3 lit4))
; the resulting structure is the same.

(define (make-flat-clause-splitcomponents clausewithvars joinvars)
  (let ((tmp '())
        (compvars1 '())
        (compvars2 '())
        (complits1 '())
        (complits2 '())
        (newcomp '())
        (split '())
        (res '())
        (joinflag #t) )
    (set! split clausewithvars)
    ; initially all the literals are separated, we start joining them
    ; and all the joins MUST be done to keep the algorithm correct.
    ; loop over joining operations until no more new joins are possible:
    ; fresh joining is invoked from scratch if a join is constructed
    (do ()
        ((not joinflag)
            (set! res split) )
       (set! joinflag #f)
       ; inside the outer loop, try to find a first joinable component
       (do ((splitpart split (cdr splitpart)))
           ((or (null? splitpart) joinflag))
          (set! compvars1 (caar splitpart))
          (set! complits1 (cdar splitpart))
          ; look for the rest of the clause to find a component to join with
          (do ((splitrest (cdr splitpart) (cdr splitrest)))
              ((or (null? splitrest) joinflag))
            (set! compvars2 (caar splitrest))
            (set! complits2 (cdar splitrest))
            ; joinable or not?
            (if (some? (lambda(var)
                         (and (not (memq var joinvars))
                              (memq var compvars2) ))
                       compvars1)
                (begin
                  ; joinable
                  (set! joinflag #t)
                  (set! newcomp
                       (cons (union compvars1 compvars2)
                             (append complits1 complits2) ))
                  (set! split
                      (filter-map
                        (lambda(comp)
                          (if (or (eq? comp (car splitpart))
                                  (eq? comp (car splitrest)) )
                              #f
                              comp))
                         split))
                  (set! split (cons newcomp split)) )))))
     split))


; make-flat-clause-joingroundsplits takes a split and joins ground
; components with the first non-ground component of the split
; (if it exists) or into one big component:
; no real point in splitting into ground components anyway

(define (make-flat-clause-joingroundsplits split)
  ;(newline)(display "make-flat-clause-joingroundsplits entered: ")
  ;(display split)(newline)
  (let ((tmp '())
        (maincomp '())
        (origmaincomp '())
        (newcomp '())
        (res '()) )
    (set! maincomp
          (member-if (lambda(comp) (not (null? (car comp))))
                     split))
    (if maincomp
        (set! maincomp (car maincomp))
        (set! maincomp (car split)) )
    (set! origmaincomp maincomp)
    (set! res
          (filter-map
            (lambda(x)
              (cond
                ((eq? x origmaincomp)  #f)
                ((null? (car x))
                    (set! maincomp (append maincomp (cdr x)))
                    #f)
                (else x)))
            split))
    (set! res (cons maincomp res))
    ;(newline)(display "make-flat-clause-joingroundsplits returns: ")
    ;(display res)(newline)
    res))


; make-flat-clause-nongroundsplits takes a ready-made split
; (list of variables at the beginning of the list of lits)
; and creates a list of splits of the clause, which are non-ground
; splits, and precedes all the splits with a list of vars which are
; split over. Example:
; input:
; ((x y z) (p x y) (r y z))
; result:
; ((y) ((x y) (p x y)) ((y z) (r y z)) )
; Such splits can be later transformed into clauses like:
; ((p x y) | (split y)) AND ((r y z) | (-split y)))

(define (make-flat-clause-nongroundsplits split)
  ;(newline)(display "make-flat-clause-nongroundsplits entered: ")(display split)
  ;(newline)
  (let ((tmp '())
        (res '())
        (vars '())
        (clause '())
        (clausevars '())
        (varcombinations '())
        (varcombinationsnr 0)
        (clausewithvars '())
        (splitfoundflag #f)
        (splitvars '()) )
    (if (or (null? split) ; test clause length
            (null? (cdr split))
            (null? (cddr split))
            (null? (car split)) ; test nr of vars in a clause
            (null? (cdar split))
            (null? (cddar split)) )
        (begin
          (list '() split) )
        (begin
          (set! res (list '() split))
          ; make a list of splitcomponents, not joined
          (set! clausevars (car split))
          (set! clause (cdr split))
          (set! clausewithvars
               (map (lambda(x)
                       (list (give-flat-lit-varset x) x) )
                     clause))
          ; create varcombination lists of different length
          (set! varcombinations
                (make-flat-nongroundsplit-varcombinations clausevars) )
          ; loop over the varcombinations in clause
          (do ((varspart varcombinations (cdr varspart)))
              ((or (null? varspart) splitfoundflag)
                 res)

            ;(newline)(display "varspart: ")(display varspart)(newline)

            (set! splitvars (car varspart))
            (set! tmp
                 (make-flat-clause-splitcomponents clausewithvars splitvars) )

            ;(newline)(display tmp)(newline)

            (set! tmp
                 (make-flat-clause-joingroundsplits tmp) )

            ;(newline)(display tmp)(newline)

            (if (and (not (null? (cdr tmp)))
                     (good-flat-split? splitvars clausevars tmp) )
                (begin
                   (set! splitfoundflag #t)
                   (set! res (cons splitvars tmp)) )))))))

; ((p x u) (r x v))
; (x u v) -> ((x) (u) (v)) : OK is (x)
; ((p x y u) (r x y v))
; (x y u v) -> ((x) (y) (u) (v) (x y) (x u) (x v) (y u) (y v) (u v)) : OK is (x y)
; max combinations length is vars length - 2


(define (make-flat-nongroundsplit-varcombinations-justatry clausevars)
  (let ((clausevarslen (length clausevars))
        (res '()) )
    (cond
      ((eqv? 0 clausevarslen)
        (set! res '()) )
      ((eqv? 1 clausevarslen)
        (set! res '()) )
      ((eqv? 2 clausevarslen)
        (set! res '()) )
      ((eqv? 3 clausevarslen)
        (set! res (map (lambda(x)(list x)) clausevars)) )
      ((eqv? 4 clausevarslen)
        (set! res
              (append
                 (map (lambda(x)(list x)) clausevars)
                 (make-subsets-of-len-2 clausevars) )))
      ((and (< 4 clausevarslen)
            (>= 6 clausevarslen) )
        (set! res
              (append
                 (map (lambda(x)(list x)) clausevars)
                 (append
                      (make-subsets-of-len-2 clausevars)
                      (make-subsets-of-len-3 clausevars) ))))
      (else
         (set! res (map (lambda(x)(list x)) clausevars)) ))
      res))

(define (make-flat-nongroundsplit-varcombinations clausevars)
  (let ((clausevarslen (length clausevars))
        (res '()) )
    (cond
      ((eqv? 0 clausevarslen)
        (set! res '()) )
      ((eqv? 1 clausevarslen)
        (set! res '()) )
      ((eqv? 2 clausevarslen)
        (set! res '()) )
      ((eqv? 3 clausevarslen)
        (set! res (map (lambda(x)(list x)) clausevars)) )
      ((eqv? 4 clausevarslen)
        (set! res
              (append
                 (map (lambda(x)(list x)) clausevars)
                 (make-subsets-of-len-2 clausevars) )))
      ((and (< 4 clausevarslen)
            (>= 6 clausevarslen) )
        (set! res
              (append
                 (map (lambda(x)(list x)) clausevars)
                 (append
                      (make-subsets-of-len-2 clausevars)
                      (make-subsets-of-len-3 clausevars) ))))
      ((and (< 6 clausevarslen)
            (>= 8 clausevarslen) )
        (set! res
              (append
                 (map (lambda(x)(list x)) clausevars)
                 (append
                      (make-subsets-of-len-2 clausevars)
                      (append
                          (make-subsets-of-len-3 clausevars)
                          (make-subsets-of-len-4 clausevars) )))))
     ((and (< 8 clausevarslen)
           (>= 20 clausevarslen) )
         (if #t ;(< *splitted-clause-set-len* 500)
             (set! res
              (append
                 (map (lambda(x)(list x)) clausevars)
                 (append
                      (make-subsets-of-len-2 clausevars)
                      (append
                          (make-subsets-of-len-3 clausevars)
                          (make-subsets-of-len-4 clausevars) ))))
             (set! res
              (append
                 (map (lambda(x)(list x)) clausevars)
                 (append
                      (make-subsets-of-len-2 clausevars) )))))
      ((and (< 20 clausevarslen)
            (>= 40 clausevarslen) )
         (if #t ;(< *splitted-clause-set-len* 500)
             (set! res
              (append
                 (map (lambda(x)(list x)) clausevars)
                 (make-subsets-of-len-2 clausevars) ))
             (set! res (map (lambda(x)(list x)) clausevars)) ))
      (else
         (set! res (map (lambda(x)(list x)) clausevars)) ))
    res))

; (x y z) -> ((x y)(x z)(y z))

(define (make-subsets-of-len-2 lst)
  (let ((res '()))
    (do ((lstpart1 lst (cdr lstpart1)))
        ((null? lstpart1)
           res)
      (do ((lstpart2 (cdr lstpart1) (cdr lstpart2)))
          ((null? lstpart2))
        (set! res (cons (list (car lstpart1) (car lstpart2)) res)) ))))

; (x y z u) -> ((x y z)(x y u)(y z u))

(define (make-subsets-of-len-3 lst)
  (let ((tmp '())
        (res '()))
    (do ((lstpart1 lst (cdr lstpart1)))
        ((null? lstpart1)
           res)
      (do ((lstpart2 (cdr lstpart1) (cdr lstpart2)))
          ((null? lstpart2))
        (do ((lstpart3 (cdr lstpart2) (cdr lstpart3)))
            ((null? lstpart3))
          (set! res
                (cons (list (car lstpart1) (car lstpart2) (car lstpart3))
                      res)))))))

(define (make-subsets-of-len-4 lst)
  (let ((res '())
        (tmp '()) )
    (do ((lstpart1 lst (cdr lstpart1)))
        ((null? lstpart1)
           res)
      (do ((lstpart2 (cdr lstpart1) (cdr lstpart2)))
          ((null? lstpart2))
       (do ((lstpart3 (cdr lstpart2) (cdr lstpart3)))
           ((null? lstpart3))
         (set! tmp (list (car lstpart1) (car lstpart2) (car lstpart3)))
         (do ((lstpart4 (cdr lstpart3) (cdr lstpart4)))
             ((null? lstpart4))
           (set! res
                 (cons (cons (car lstpart4) tmp)
                       res))))))))

(define (make-subsets-of-len-5 lst)
  (let ((res '())
        (tmp '()) )
    (do ((lstpart1 lst (cdr lstpart1)))
        ((null? lstpart1)
           res)
      (do ((lstpart2 (cdr lstpart1) (cdr lstpart2)))
          ((null? lstpart2))
       (do ((lstpart3 (cdr lstpart2) (cdr lstpart3)))
           ((null? lstpart3))
         (set! tmp (list (car lstpart1) (car lstpart2) (car lstpart3)))
         (do ((lstpart4 (cdr lstpart3) (cdr lstpart4)))
             ((null? lstpart4))
           (set! tmp (cons (car lstpart4) tmp))
           (do ((lstpart5 (cdr lstpart4) (cdr lstpart5)))
               ((null? lstpart5))
             (set! res
                  (cons (cons (car lstpart5) tmp)
                        res)))))))))

; good-flat-split? will accept only such splits,
; where each component contains less vars than the original clause

(define (good-flat-split? splitvars clausevars splits)
  ;(newline)(display "good-flat-split called: ")(newline)
  ;(display splitvars)(newline)(display clausevars)(newline)(display splits)(newline)
  (let ((clausevarsnr (length clausevars)))
    (every?
       (lambda(split)
          (< (length (car split)) clausevarsnr) )
       splits)))


(define (give-flat-lit-varset lit)
   (get-fullclause-varslist (list '() lit)) )


;-----------------------------------------------------------------------------------------
;
;      splitting sos clauses
;
;------------------------------------------------------------------------------------------



(define (pseudo-split-input-clause clause)
  (let ((res '())
        (bestlen 99999)
        (bestsplit '())
        (splitcomponents '())
        (gsplit '())
        (ngsplit '())
        (ngsplitcomponents '())
        (ngsplitvars '())
        (newclause '())
        (splitliteral '())
        (splittedclauses '())
        (gjoinclause '())
        (ngjoinclause '())
        (ngjoinclauses '())
        (joinclauses '())
        (tmp '())
        (origclause clause) )
    (set! splitcomponents
          (split-input-clause-into-components clause))
    ; splitcomponents is a list of ground splits, each
    ; being a list of nonground splits
    ; loop first over ground splits
    (do ((gsplits (cdr splitcomponents) (cdr gsplits)))
        ((null? gsplits)
          ;(newline)(display "gjoinclause: ")(display gjoinclause)
          ;(newline)(display "ngjoinclauses: ")(display ngjoinclauses)
          ;(newline)(display "splittedclauses: ")(display splittedclauses)
          ;(newline)
          (if (not (null? gjoinclause))
              (set! res (cons gjoinclause
                              (append ngjoinclauses  splittedclauses) ))
              (set! res (append ngjoinclauses splittedclauses)) )
          (set! res (map (lambda(x)(cons '() x)) res))
          res)
      (if (not (null? (cddr splitcomponents)))
          ; ground splits exist
          (begin
            (set! tmp (make-input-split-literal '()))
            (set! gjoinclause (cons (cons (polarize-pred-nr (car tmp) #f)
                                          (cdr tmp) )
                                    gjoinclause))
            (set! ngjoinclause
                  (list (cons (car tmp)
                              (cdar gjoinclause) )) ))
          ; no ground splits
          (begin
            (set! gjoinclause '())
            (set! ngjoinclause '()) ))
      (set! gsplit (car gsplits))
      (set! ngsplitcomponents (cdr gsplit))
      (set! ngsplitvars (car gsplit))
      ; show temporary stuff
      ;(display "temporary results in pseudo-split-input-clause: ")
      ;(newline)
      ;(display "gsplit: ")(display gsplit)(newline)
      ;(display "ngsplitcomponents: ")(display ngsplitcomponents)(newline)
      ;(display "ngsplitvars: ")(display ngsplitvars)(newline)

      ; loop over nonground splits
      (set! bestlen 99999)
      (set! bestsplit '())
      (do ((ngsplits ngsplitcomponents (cdr ngsplits)))
          ((null? ngsplits)
             (if (not (null? ngjoinclause))
                 (set! ngjoinclauses (cons ngjoinclause ngjoinclauses)) ))
        (if (not (null? (cdr ngsplitcomponents)))
            ; nonground splits exist
            (begin
              (set! ngsplit (cdar ngsplits))
              (set! splitliteral (make-input-split-literal ngsplitvars))
              (set! newclause (cons splitliteral ngsplit))
              (set! splittedclauses (cons newclause splittedclauses))
              (set! ngjoinclause
                   (cons (cons (polarize-pred-nr (car splitliteral) #f)
                               (cdr splitliteral))
                      ngjoinclause)))
            ; no nonground splits
            (if (not (null? gjoinclause))
                ; ground splits exist
                (begin
                  (set! ngsplit (cdar ngsplits))
                  (set! ngjoinclause '())
                  (set! splitliteral
                       (cons (reverse-polarize-pred-nr (caar gjoinclause))
                             (cdar gjoinclause)) )
                  (set! newclause (cons splitliteral ngsplit))
                  (set! splittedclauses (cons newclause splittedclauses)) )
                ; ground splits also do not exist
                (begin
                  (set! ngsplit (cdar ngsplits))
                  (set! ngjoinclause '())
                  (set! splittedclauses (cons ngsplit splittedclauses)) )))))))







; NB! !!! the predsymb and constant clashes should
; be checked for and avoided!!!

(define *input-split-constant-nr* 70000)

; warning:
; should not do any splitting if
; (> *pred-num* 100)

(define (make-input-split-literal vars)
  (let ((tmp0 '())
        (tmp1 '()) )
    ;(newline)
    ;(display "vars: ")(display vars)
    ;(newline)
    ; create split predicates if necessary
    (set! tmp0
         (assoc (string->symbol "$spltprd0") *pred-num-table*) )
    (set! tmp1
         (assoc (string->symbol "$spltprd1") *pred-num-table*) )
    (if tmp0
        (set! tmp0 (cdr tmp0))
        (begin
          (set! *pred-num* (+ 1 *pred-num*))
          (set! tmp0 (shiftleft *pred-num* *fnsymb-shift*))
          (set! *pred-num-table*
               (cons (cons (string->symbol "$spltprd0") tmp0)
                      *pred-num-table*))))
    (if tmp1
        (set! tmp1 (cdr tmp1))
        (begin
          (set! *pred-num* (+ 1 *pred-num*))
          (set! tmp1 (shiftleft *pred-num* *fnsymb-shift*))
          (set! *pred-num-table*
               (cons (cons (string->symbol "$spltprd1") tmp1)
                      *pred-num-table*))))

    ; select a new constant
    (do ((nr (+ 1 *input-split-constant-nr*) (+ 1 nr)))
        ((not (rassq nr *const-num-table*))
          (set! *input-split-constant-nr* nr) ))
    (set! *const-num-table*
         (cons (cons (string->symbol
                       (string-append "$spltcnst"
                         (number->string
                           (- *input-split-constant-nr* 70000) )))
                      *input-split-constant-nr*)
                *const-num-table*))
    ; finally, create the literal
    (cond
      ((> *pred-num* *max-nr-of-preds*)
         (report-error "Splitting created too many predicate symbs in the clause set" #t) )
      ((null? vars)
         (list tmp0 *input-split-constant-nr*) )
      ((null? (cdr vars))
         (cons tmp1 (cons *input-split-constant-nr* vars)) )
      (else
         (newline)
         (display "error in sos splitting, aborting!")
         (newline)
         (quit) ))))


(define (make-input-clause-nongroundsplits split)
  ;(newline)(display "make-input-clause-nongroundsplits entered: ")
  ;(newline)
  ;(display split)
  ;(newline)
  (let ((tmp '())
        (res '())
        (vars '())
        (clause '())
        (clausevars '())
        (varcombinations '())
        (varcombinationsnr 0)
        (clausewithvars '())
        (splitfoundflag #f)
        (splitvars '()) )
    (if (or (null? split) ; test clause length
            (null? (cdr split))
            (null? (cddr split))
            (null? (car split)) ; test nr of vars in a clause
            (null? (cdar split))
            (null? (cddar split)) )
        (begin
          (list '() split) )
        (begin
          (set! res (list '() split))
          ; make a list of splitcomponents, not joined
          (set! clausevars (car split))
          (set! clause (cdr split))
          (set! clausewithvars
               (map (lambda(x)
                       (list (give-flat-lit-varset x) x) )
                     clause))
          ; create varcombination lists of different length
          (set! varcombinations
                (make-input-nongroundsplit-varcombinations clausevars) )
          ; loop over the varcombinations in clause
          (do ((varspart varcombinations (cdr varspart)))
              ((or (null? varspart) splitfoundflag)
                 res)

            ;(newline)(display "varspart: ")(display varspart)(newline)

            (set! splitvars (car varspart))
            (set! tmp
                 (make-flat-clause-splitcomponents clausewithvars splitvars) )

            ;(newline)
            ;(display "in loop make-input-clause-splitcomponents: ")
            ;(newline)
            ;(display tmp)
            ;(newline)

            (set! tmp
                 (make-flat-clause-joingroundsplits tmp) )

            ;(newline)
            ;(display "in loop make-flat-clause-joingroundsplits: ")
            ;(newline)
            ;(display tmp)
            ;(newline)

            (if (and (not (null? (cdr tmp)))
                     (good-flat-split? splitvars clausevars tmp) )
                (begin
                   (set! splitfoundflag #t)
                   (set! res (cons splitvars tmp)) )))))))

; ((p x u) (r x v))
; (x u v) -> ((x) (u) (v)) : OK is (x)
; ((p x y u) (r x y v))
; (x y u v) -> ((x) (y) (u) (v) (x y) (x u) (x v) (y u) (y v) (u v)) : OK is (x y)
; max combinations length is vars length - 2

(define (make-input-nongroundsplit-varcombinations clausevars)
  (let ((clausevarslen (length clausevars))
        (res '()) )
    (cond
      ((eqv? 0 clausevarslen)
        (set! res '()) )
      ((eqv? 1 clausevarslen)
        (set! res '()) )
      ((eqv? 2 clausevarslen)
        (set! res '()) )
      (else
        (set! res (map (lambda(x)(list x)) clausevars)) ))
    res))


(define (split-input-clause-into-components clause)
  (let ((res '())
        (splitcomponents '())
        (origclause clause)
        (clausevars '())
        (joinvars '())
        (clausewithvars '()) )

    ;(newline)
    ;(display "clause: ")(display clause)

    (set! clausevars (get-fullclause-varslist clause))
    (set! clause (cdr clause))
    (set! clausewithvars
          (map (lambda(x)
                  (list (give-flat-lit-varset x) x) )
                clause))

    ;(newline)(display "with vars: ")(display clausewithvars)(newline)

    ; try to create splittable components

    (set! splitcomponents
          (make-flat-clause-splitcomponents clausewithvars joinvars))

    ;(newline)(display "initial ground split gives: ")(newline)
    ;(display splitcomponents)(newline)

    (set! splitcomponents (make-flat-clause-joingroundsplits splitcomponents))

    ;(newline)(display "joining ground literals gives: ")(newline)
    ;(display splitcomponents)(newline)

    (set! splitcomponents
          (map (lambda(x)
                 (make-input-clause-nongroundsplits x) )
               splitcomponents))
    (set! splitcomponents (cons '() splitcomponents))

    ;(newline)(display "nonground splitting gives: ")(newline)
    ;(display splitcomponents)(newline)

    (cond
      ; empty or unit clause
      ((or (null? clause) (null? (cdr clause)))
        (set! res (list origclause)) )
      (else
        (set! res (list origclause)) ))
    ; return result set
    ;res
    splitcomponents))

; NB! The following is not used

(define (find-good-splittable-sos-clause sos)
  (let ((res '())
        (preres '())
	(tmp '())
	(soslen 0)
	(splittablenr 0)
	(nongsplittablenr 0)
	(purenegsplittablenr 0)
	(splittablemaxlen 0)
	(purenegnr 0)
	(stopflag #f)
        (cl '()) )
    ; - - - -
    ; first collect all splittable clauses into preres
    (do ((part sos (cdr part)))
        ((null? part))
      (set! cl (car part))
      ; collect data about sos
      (set! soslen (+ 1 soslen))
      (if (every? (lambda(lit)(neg-fun? (car lit))) cl)
	  (set! purenegnr (+ 1 purenegnr)) )
      ; check if splittable
      (if (not (null? (cdr cl)))
	(begin
	   (set! tmp (split-full-clause cl))
	   (if (not (null? (cdr tmp)))
	       (begin
		 (if (> *display-info-level* 70)
		     (begin
		       (newline)
		       (display "splittable clause: ")
		       (print-pure-clause cl)
		       (newline) ))
		 (set! splittablenr (+ 1 splittablenr))
                 (if (some? (lambda(el)(not (null? el))) tmp)
                     (set! nongsplittablenr (+ 1 nongsplittablenr)) )
		 (if (> (length cl) splittablemaxlen)
                     (set! splittablemaxlen (length cl)) )
		 (if (every? (lambda(lit)(neg-fun? (car lit))) cl)
	             (set! purenegsplittablenr (+ 1 purenegsplittablenr)) )
		 (set! preres (cons (car part) preres)) )))))
    ; - - - -
    ; now find a good splitting candidate if available:
    ; return it, if found, return #f if no good candidate found

    (newline)
    (display "splitting preres is: ")
    (for-each (lambda(x)(display x)) preres)
    (newline)

    ; select the best split:
    ; if only one split, select this
    ; if only one nonground split, select this
    ; if several nonground splits or all ground, then:
    ;   - select the pure neg splittable if available
    ;  - if again no single preference, select longest

    (set! stopflag #f)
    (set! res '())
    (do ((rpart preres (cdr rpart)))
        ((or (null? rpart)
	     stopflag)
	   (if (or (not res) (null? res))
	       #f
	       res))
      (set! cl (car rpart))
      (cond
        ((null? (cdr preres))
	   (set! res cl)
	   (set! stopflag #t) )
        ((eqv? nongsplittablenr 1)
            (if (some? (lambda(x)(fun-may-cont-vars? (car x)))
                       cl)
                (begin
		  (set! res cl)
		  (set! stopflag #t) )))
        ((eqv? purenegnr 1)

            ;(if (eqv? nongsplittablenr 1)
            ;    (begin
            )))))


(define (split-full-clause cl)
  (let ((tmp '())
        (clinf (car cl))
	(purecl (cdr cl))
	(splitted '())
	(res '()) )
    (set! splitted
          (map
	    (lambda(x)
	      (list (get-lit-split-varlist x) x))
	    purecl))
    (set! res (make-clause-splitcomponents splitted '()))
    res))

(define *lit-split-varlist-tmp* '())

(define (get-lit-split-varlist lit)
   (set! *lit-split-varlist-tmp* '())
   (get-lit-split-varlist-aux! lit)
   *lit-split-varlist-tmp*)


(define (get-lit-split-varlist-aux! term)
  (let ((tmp '()))
    (cond
      ((pair? term)
         (do ((part (cdr term) (cdr part)))
	     ((null? part))
	   (get-lit-split-varlist-aux! (car part)) ))
      ((var? term)
          (set! tmp (var-nr term))
	  (if (not (memq tmp *lit-split-varlist-tmp*))
	      (set! *lit-split-varlist-tmp*
	            (cons tmp *lit-split-varlist-tmp*) ))))))



; make-clause-splitcomponents is similar to make-flat-clause-splitcomponents
;
; It takes a clause split (incorrectly!)
; into single literal always and combines the splits to make a correct
; split.
; The second argument (joinvars) is a list of vars which are allowed
; to exist in SEVERAL splits. Ie, if joinvars is not empty, the function
; does non-ground splitting.
; It only uses the variable lists (sets) which are at the beginning
; of each component:
; ((varlist12 lit1 lit2) (varlist34 lit3 lit4))
; the resulting structure is the same.

(define (make-clause-splitcomponents clausewithvars joinvars)
  (let ((tmp '())
        (compvars1 '())
        (compvars2 '())
        (complits1 '())
        (complits2 '())
        (newcomp '())
        (split '())
        (res '())
        (joinflag #t) )
    (set! split clausewithvars)
    ; initially all the literals are separated, we start joining them
    ; and all the joins MUST be done to keep the algorithm correct.
    ; loop over joining operations until no more new joins are possible:
    ; fresh joining is invoked from scratch if a join is constructed
    (do ()
        ((not joinflag)
            (set! res split) )
       (set! joinflag #f)
       ; inside the outer loop, try to find a first joinable component
       (do ((splitpart split (cdr splitpart)))
           ((or (null? splitpart) joinflag))
          (set! compvars1 (caar splitpart))
          (set! complits1 (cdar splitpart))
          ; look for the rest of the clause to find a component to join with
          (do ((splitrest (cdr splitpart) (cdr splitrest)))
              ((or (null? splitrest) joinflag))
            (set! compvars2 (caar splitrest))
            (set! complits2 (cdar splitrest))
            ; joinable or not?
            (if (some? (lambda(var)
                         (and (not (memq var joinvars))
                              (memq var compvars2) ))
                       compvars1)
                (begin
                  ; joinable
                  (set! joinflag #t)
                  (set! newcomp
                       (cons (union compvars1 compvars2)
                             (append complits1 complits2) ))
                  (set! split
                      (filter-map
                        (lambda(comp)
                          (if (or (eq? comp (car splitpart))
                                  (eq? comp (car splitrest)) )
                              #f
                              comp))
                         split))
                  (set! split (cons newcomp split)) )))))
     split))


; =========================  END ==================================