/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/**
 * File options.cpp.
 * Implements Vampire options.
 *
 * @since 06/06/2001 Manchester, completely rewritten
 */

#include <string>
#include <cstdlib>
#include <iostream>
#include <climits>

#include "assert.hpp"
#include "Options.hpp"
#include "Random.hpp"


Options* options = 0;


/** 
 * Initialize options to the default values.
 *
 * @since 10/07/2003 Manchester, _normalize added
 */
Options::Options ()
  :
  _mode (Opt::MODE, VAMPIRE, "mode"),
  _tab (Opt::TAB, 0, "tab"),
  _inputFile (Opt::INPUT_FILE, 0, "input_file"),
  _randomSeed (Opt::RANDOM_SEED, Random::seed(), "random_seed"),
  
  // input/output options
  _statistics (Opt::STATISTICS, STATISTICS_DETAILED, "statistics"),
  _inputSyntax (Opt::INPUT_SYNTAX, INPUT_TPTP, "input_syntax"),
  _outputSyntax (Opt::OUTPUT_SYNTAX, NATIVE, "output_syntax"),
  _silent (Opt::SILENT, false, "silent"),
  _showGen (Opt::SHOW_GEN, false, "show_generated"),
  _showPassive (Opt::SHOW_PASSIVE, false, "show_passive"),
  _showActive (Opt::SHOW_ACTIVE, false, "show_active"),
  _showPrepro (Opt::SHOW_PREPRO, false, "show_prepro"),
  _showProfile (Opt::SHOW_PROFILE, false, "show_profile"),
  _showInput (Opt::SHOW_INPUT, false, "show_input"),
  _showOptions (Opt::SHOW_OPTIONS, false, "show_options"),
  _showWeightLimitChanges (Opt::SHOW_WEIGHT_LIMIT_CHANGES, false, "show_weight_limit_changes"),
  // preprocessor options
  _removeUnusedDefs (Opt::REMOVE_UNUSED_DEF, true, "remove_unused_defs"),
  _arityCheck (Opt::ARITY_CHECK, true, "arity_check"),
  _noFreeVars (Opt::NO_FREE_VARS, true, "no_free_vars"),
  _miniscope (Opt::MINISCOPE, true, "miniscope"),
  _outputEqualityInfix (Opt::OUTPUT_EQUALITY_INFIX, false, "output_equality_infix"),
  _include (Opt::INCLUDE, 0, "include"),
  // Kernel's options
  _mainAlg (Opt::MAIN_ALG, LRS, "man_alg"),
  _splitting (Opt::SPLITTING, SPLIT_OFF, "splitting"),
  _splittingStyle (Opt::SPLITTING_STYLE, PARALLEL, "splitting_style"),
  _splittingWithNaming (Opt::SPLITTING_WITH_NAMING, true, "splitting_with_naming"),
  _selection (Opt::SELECTION, P_4, "selection"),
  _literalComparisonMode (Opt::LITERAL_COMPARISON_MODE, NORMAL, "literal_comparison_mode"),
  _inheritedNegativeSelection (Opt::INHERITED_NEGATIVE_SELECTION, false, "inherited_negative_selection"),
  _forwardSubsumptionResolution (Opt::FORWARD_SUBSUMPTION_RESOLUTION, true, "forward_subsumption_resolution"),
  _forwardDemodulation (Opt::FORWARD_DEMODULATION, DEMODULATION_ALL, "forward_demodulation"),
  _fdOnSplittingBranches (Opt::FD_ON_SPLITTING_BRANCHES, true, "fd_on_splitting_branches"),
  _backwardDemodulation (Opt::BACKWARD_DEMODULATION, DEMODULATION_ALL, "backward_demodulation"),
  _bdOnSplittingBranches (Opt::BD_ON_SPLITTING_BRANCHES, true, "bd_on_splitting_branches"),
  _forwardSubsumption (Opt::FORWARD_SUBSUMPTION, SUBSUMPTION_MULTISET, "forward_subsumption"),
  _backwardSubsumption (Opt::BACKWARD_SUBSUMPTION, SUBSUMPTION_MULTISET, "backward_subsumption"),
  _orphanMurder (Opt::ORPHAN_MURDER, false, "orphan_murder"),
  _elimDef (Opt::ELIM_DEF, 0, "elim_def"),
  _lrsFirstTimeCheck (Opt::LRS_FIRST_TIME_CHECK, 5, "lrs_first_time_check"),
  _lrsFirstMemoryCheck (Opt::LRS_FIRST_MEMORY_CHECK, 100, "lrs_first_memory_check"),
  _maxInferenceDepth (Opt::MAX_INFERENCE_DEPTH, 2147483647, "max_inference_depth"),
  _symbolPrecedence (Opt::SYMBOL_PRECEDENCE, BY_ARITY, "symbol_precedence"),
  _headerPrecedenceKinky (Opt::HEADER_PRECEDENCE_KINKY, false, "header_precedence_kinky"),
  _noInferencesBetweenRules (Opt::NO_INFERENCES_BETWEEN_RULES, false, "no_inferences_between_rules"),
  _timeLimit (Opt::TIME_LIMIT, LONG_MAX, "time_limit"),
  _memoryLimit (Opt::MEMORY_LIMIT, 300, "memory_limit"),
  _allocationBufferSize (Opt::ALLOCATION_BUFFER_SIZE, 10, "allocation_buffer_size"),
  _allocationBufferPrice (Opt::ALLOCATION_BUFFER_PRICE, 2.0, "allocation_buffer_price"),
  _inequalitySplitting (Opt::INEQUALITY_SPLITTING, 3, "inequality_splitting"),
  _maxSkolem (Opt::MAX_SKOLEM, 4096, "max_skolem"),
  _maxActive (Opt::MAX_ACTIVE, LONG_MAX, "max_active"),
  _maxPassive (Opt::MAX_PASSIVE, LONG_MAX, "max_passive"),
  _maxWeight (Opt::MAX_WEIGHT, LONG_MAX, "max_weight"),
  _normalize(Opt::NORMALIZE, false, "normalize"),
  _ageWeightRatio (Opt::AGE_WEIGHT_RATIO, 5, "age_weight_ratio"),
  _proof (Opt::PROOF, PROOF_FULL, "proof"),
  _testId (Opt::TEST_ID, "unspecified_test", "test_id"),
  _testNumber (Opt::TEST_NUMBER, Random::integer(), "test_number"),
  _paramodulation (Opt::PARAMODULATION, true, "paramodulation"),
  _symbolWeightByArity (Opt::SYMBOL_WEIGHT_BY_ARITY, 0, "symbol_weight_by_arity"),
  _termWeightingScheme (Opt::TERM_WEIGHTING_SCHEME, TWS_UNIFORM, "term_weighting_scheme"),
  _simplificationOrdering (Opt::SIMPLIFICATION_ORDERING, KBO, "simplification_ordering"),
  _nongoalWeightCoefficient (Opt::NONGOAL_WEIGHT_COEFFICIENT, 1.0, "nongoal_weight_coefficient" ),
  _selectedEqualityWeightCoefficient (Opt::SELECTED_EQUALITY_WEIGHT_COEFFICIENT, 1.0, "selected_equality_weight_coefficient" ),
  _sos (Opt::SOS, false, "sos")
{
} // Options::Options


// correct options using the command line arguments
void Options::correct (int argc, char* argv [])
{
  TRACER ( "Options::correct" );

  char** last = argv + argc;
  char** current = argv + 1;

  while (current != last) {
    Opt* opt = find (*current); // find option by name
    if (opt) { // found
      current++; // advance current, it should now point to the value of the option
      if (current == last) { // wrong, no value
        cerr << "Error in command line after " << current[-1] << "\n";
	throw MyException ("Terminated by error in command line");
      }
      // value found
      if (! opt->set (*current)) { // option not set successfully
        cerr << "Error in command line: a bad value ("
             << current[0] << ") for "
             << current[-1] << "\n";
	throw MyException ("Terminated by error in command line");
      }
      // now option is set successfully, go to the next option
      current++;
    }
    // option not found, should be the input file name
    else if (current+1 != last) { // there is something beyond the file name
      cerr << "Error in command line: " << *current << " is not a valid option\n";
      throw MyException ("Terminated by error in command line");
    }
    else { // the last option, should be the file name then
      inputFile (*current);

      const long maxTime = LONG_MAX/10;
      if (timeLimit() > maxTime) {
	timeLimit (maxTime);
      }

      // changing time to deciseconds
      timeLimit (timeLimit()*10);
      return;
    }
  }
  
  throw MyException ("input file name missing");
} // Options::correct


bool Option<bool>::set (const char* val)
{
  if (! strcmp (val,"on")) {
    _value = true;
  }
  else if (! strcmp (val,"off")) {
    _value = false;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<bool>::set


/**
 * Output the option in a stream in the form name=value.
 *
 * @since 27/05/2003 Manchester
 * @param str the output stream
 */
void Opt::output (ostream& str) const
{
  str << _name << " = ";
  toStream (str);
  str << '\n';
} // Opt::output


/**
 * Output the option in a stream in the XML syntax.
 *
 * @since 28/06/2003 Manchester
 * @param str the output stream
 */
void Opt::toXML (ostream& str) const
{
  str << " <option name='"
      << _name 
      << "' value='";
  toStream (str);
  str << "'/>\n";
} // Opt::toXML


// 27/05/2003 Manchester
void Opt::tabulate (ostream& str) const
{
  str << "option("
      << _name 
      << ',';
  if (quotes()) {
    str << '\'';
    toStream (str);
    str << '\'';
  }
  else {
    toStream (str);
  }
  str << ").\n";
} // Opt::tabulate


// 02/01/2003 Manchester
void Option<bool>::toStream (ostream& str) const
{
  str << (_value ? "on" : "off");
} // Option<bool>::toStream


// 27/05/2003 Manchester
bool Option<bool>::quotes () const
{
  return false;
} // Option<bool>::quotes


bool Option<const char*>::set (const char* val)
{
  _value = val;

  _changed = true;
  return true;
} // Option<const char*>::set


// 27/05/2003 Manchester
bool Option<const char*>::quotes () const
{
  return true;
} // Option<const char*>::quotes


// 02/01/2003 Manchester
void Option<const char*>::toStream (ostream& str) const
{
  str << _value;
} // Option<const char*>::toStream


bool Option<int>::set (const char* val)
{
  char* endptr = 0;
  _value = (int)(strtol (val, &endptr, 10));

  if (*endptr) { // error returned by strtol
    return false;
  }

  _changed = true;
  return true;
} // Option<int>::set


// 27/05/2003 Manchester
bool Option<int>::quotes () const
{
  return false;
} // Option<int>::quotes


// 02/01/2003 Manchester
void Option<int>::toStream (ostream& str) const
{
  str << _value;
} // Option<int>::toStream


bool Option<long>::set (const char* val)
{
  char* endptr = 0;
  _value = strtol (val, &endptr, 10);

  if (*endptr) { // error returned by strtol
    return false;
  }

  _changed = true;
  return true;
} // Option<long>::set


// 27/05/2003 Manchester
bool Option<long>::quotes () const
{
  return false;
} // Option<long>::quotes


// 02/01/2003 Manchester
void Option<long>::toStream (ostream& str) const
{
  str << _value;
} // Option<long>::toStream


bool Option<float>::set (const char* val)
{
  char* endptr = 0;
  _value = (float)(strtod (val, &endptr));

  if (*endptr) { // error returned by strtol
    return false;
  }

  _changed = true;
  return true;
} // Option<float>::set


// 27/05/2003 Manchester
bool Option<float>::quotes () const
{
  return false;
} // Option<float>::quotes


// 02/01/2003 Manchester
void Option<float>::toStream (ostream& str) const
{
  str << _value;
} // Option<float>::toStream


/**
 * Change mode of Vampire. The mode change can lead to 
 * the change of other options.
 *
 * @param val the value, can be one of "vampire", "kif", "casc-N",
 *    where N is in [18.19]
 * @return true if the parameter is one of the admissible values
 * @since 16/07/2003 flight Novosibirsk-Moscow, CASC-19 mode added
 */
bool Option<Options::Mode>::set (const char* val)
{
  if (_changed) {
    throw MyException ("--mode cannot be set twice");
  }

  if (! strcmp (val,"vampire")) {
    ASS (_value == Options::VAMPIRE);
    // nothing to do since vampire is the default mode anyhow
  }
  else if (! strcmp (val,"kif")) {
    _value = Options::KIF;
    options->kif();
  }
  else if (! strcmp (val,"casc-18")) {
    _value = Options::CASC_18;
    options->casc18();
  }
  else if (! strcmp (val,"casc-19")) {
    _value = Options::CASC_19;
    options->casc19();
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Mode>::set


// 27/05/2003 Manchester
bool Option<Options::Mode>::quotes () const
{
  return true;
} // Option<Options::Mode>::quotes


/**
 * Print the string representation of the mode of Vampire.
 *
 * @since 02/01/2003 Manchester
 * @since 16/07/2003 flight Novosibirsk-Moscow, CASC-19 mode added
 * @since 21/07/2003 Manchester, CASC-18 mode added
 */
void Option<Options::Mode>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::VAMPIRE:
      str << "vampire";
      break;
    case Options::KIF:
      str << "kif";
      break;
    case Options::CASC_18:
      str << "casc_18";
      break;
    case Options::CASC_19:
      str << "casc_19";
      break;
    }
} // Option<Options::Mode>::toStream


// 02/01/2003, Manchester
bool Option<Options::Statistics>::set (const char* val)
{
  if (! strcmp (val,"off")) {
    _value = Options::STATISTICS_OFF;
  }
  else if (! strcmp (val,"brief")) {
    _value = Options::STATISTICS_BRIEF;
  }
  else if (! strcmp (val,"detailed")) {
    _value = Options::STATISTICS_DETAILED;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Statistics>::set


// 27/05/2003 Manchester
bool Option<Options::Statistics>::quotes () const
{
  return false;
} // Option<Options::Statistics>::quotes


// 02/01/2003 Manchester
void Option<Options::Statistics>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::STATISTICS_OFF:
      str << "off";
      break;
    case Options::STATISTICS_BRIEF:
      str << "brief";
      break;
    case Options::STATISTICS_DETAILED:
      str << "detailed";
      break;
    }
} // Option<Options::Statistics>::toStream


// 25/12/2003 Manchester, changed by integrating former LRS
bool Option<Options::MainAlg>::set (const char* val)
{
  if (! strcmp (val,"otter")) {
    _value = Options::OTTER;
  }
  else if (! strcmp (val,"discount")) {
    _value = Options::DISCOUNT;
  }
  else if (! strcmp (val,"lrs")) {
    _value = Options::LRS;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<MainAlg>::set


// 27/05/2003 Manchester
bool Option<Options::MainAlg>::quotes () const
{
  return false;
} // Option<Options::MainAlg>::quotes


// 02/01/2003 Manchester
void Option<Options::MainAlg>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::OTTER:
      str << "otter";
      break;
    case Options::DISCOUNT:
      str << "discount";
      break;
    case Options::LRS:
      str << "lrs";
      break;
    }
} // Option<Options::MainAlg>::toStream


// 29/12/2003 Manchester
bool Option<Options::Subsumption>::set (const char* val)
{
  if (! strcmp (val,"off")) {
    _value = Options::SUBSUMPTION_OFF;
  }
  else if (! strcmp (val,"multiset")) {
    _value = Options::SUBSUMPTION_MULTISET;
  }
  else if (! strcmp (val,"optimized")) {
    _value = Options::SUBSUMPTION_OPTIMIZED;
  }
  else if (! strcmp (val,"set")) {
    _value = Options::SUBSUMPTION_SET;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Options::Subsumption>::set


// 27/05/2003 Manchester
bool Option<Options::Subsumption>::quotes () const
{
  return false;
} // Option<Options::Subsumption>::quotes


// 02/01/2003 Manchester
void Option<Options::Subsumption>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::SUBSUMPTION_OFF:
      str << "off";
      break;
    case Options::SUBSUMPTION_MULTISET:
      str << "multiset";
      break;
    case Options::SUBSUMPTION_OPTIMIZED:
      str << "optimized";
      break;
    case Options::SUBSUMPTION_SET:
      str << "set";
      break;
    }
} // Option<Options::Subsumption>::toStream


// 24/10/2002 Manchester
bool Option<Options::Proof>::set (const char* val)
{
  if (! strcmp (val,"off")) {
    _value = Options::PROOF_OFF;
  }
  else if (! strcmp (val,"condensed")) {
    _value = Options::PROOF_CONDENSED;
  }
  else if (! strcmp (val,"full")) {
    _value = Options::PROOF_FULL;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Options::Proof>::set


// 27/05/2003 Manchester
bool Option<Options::Proof>::quotes () const
{
  return false;
} // Option<Options::Proof>::quotes


// 02/01/2003 Manchester
void Option<Options::Proof>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::PROOF_OFF:
      str << "off";
      break;
    case Options::PROOF_CONDENSED:
      str << "condensed";
      break;
    case Options::PROOF_FULL:
      str << "full";
      break;
    }
} // Option<Options::Proof>::toStream


// 02/01/2003 Manchester symbolic names introduced
bool Option<Options::SymbolPrecedence>::set (const char* val)
{
  if (! strcmp (val,"occurrence")) {
    _value = Options::BY_OCCURRENCE;
  }
  else if (! strcmp (val,"arity")) {
    _value = Options::BY_ARITY;
  }
  else if (! strcmp (val,"reverse_arity")) {
    _value = Options::BY_REVERSE_ARITY;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Options::SymbolPrecedence>::set


// 27/05/2003 Manchester
bool Option<Options::SymbolPrecedence>::quotes () const
{
  return false;
} // Option<Options::SymbolPrecedence>::quotes


// 02/01/2003 Manchester
void Option<Options::SymbolPrecedence>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::BY_OCCURRENCE:
      str << "occurrence";
      break;
    case Options::BY_ARITY:
      str << "arity";
      break;
    case Options::BY_REVERSE_ARITY:
      str << "reverse_arity";
      break;
    }
} // Option<Options::SymbolPrecedence>::toStream


bool Option<Options::TermWeightingScheme>::set (const char* val)
{
  if (! strcmp (val,"uniform")) {
    _value = Options::TWS_UNIFORM;
  }
  else if (! strcmp (val,"nonuniform_constant")) {
    _value = Options::TWS_NONUNIFORM_CONSTANT;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Options::TermWeightingScheme>::set


// 27/05/2003 Manchester
bool Option<Options::TermWeightingScheme>::quotes () const
{
  return false;
} // Option<Options::TermWeightingScheme>::quotes


// 02/01/2003 Manchester
void Option<Options::TermWeightingScheme>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::TWS_UNIFORM:
      str << "uniform";
      break;
    case Options::TWS_NONUNIFORM_CONSTANT:
      str << "nonuniform_constant";
      break;
    }
} // Option<Options::TermWeightingScheme>::toStream


bool Option<Options::Demodulation>::set (const char* val)
{
  if (! strcmp (val,"off")) {
    _value = Options::DEMODULATION_OFF;
  }
  else if (! strcmp (val,"preordered")) {
    _value = Options::DEMODULATION_PREORDERED;
  }
  else if (! strcmp (val,"all")) {
    _value = Options::DEMODULATION_ALL;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Options::Demodulation>::set


// 27/05/2003 Manchester
bool Option<Options::Demodulation>::quotes () const
{
  return false;
} // Option<Options::Demodulation>::quotes


// 02/01/2003 Manchester
void Option<Options::Demodulation>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::DEMODULATION_OFF:
      str << "off";
      break;
    case Options::DEMODULATION_PREORDERED:
      str << "preordered";
      break;
    case Options::DEMODULATION_ALL:
      str << "all";
      break;
    }
} // Option<Options::Demodulation>::toStream


// 02/01/2003 Manchester, changed to symbolic names
bool Option<Options::LiteralComparisonMode>::set (const char* val)
{
  if (! strcmp (val,"normal")) {
    _value = Options::NORMAL;
  }
  else if (! strcmp (val,"kinky")) {
    _value = Options::KINKY;
  }
  else if (! strcmp (val,"predicate")) {
    _value = Options::PREDICATE;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<LiteralComparisonMode>::set


// 27/05/2003 Manchester
bool Option<Options::LiteralComparisonMode>::quotes () const
{
  return false;
} // Option<Options::LiteralComparisonMode>::quotes


// 02/01/2003 Manchester
void Option<Options::LiteralComparisonMode>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::NORMAL:
      str << "normal";
      break;
    case Options::KINKY:
      str << "kinky";
      break;
    case Options::PREDICATE:
      str << "predicate";
      break;
    }
} // Option<Options::LiteralComparisonMode>::toStream


/**
 * Set selection to a new value.
 *
 * @param val string representation of the value
 * @since ???
 * @since 01/07/2003 Manchester, selections 0, 8-15 added
 * @since 07/07/2003 Manchester, selections -9 to -15, 1010 t0 1014, -1012, -1014
 */
bool Option<Options::Selection>::set (const char* val)
{
  if (! strcmp (val,"0")) {
    _value = Options::P_0;
  }
  else if (! strcmp (val,"1")) {
    _value = Options::P_1;
  }
  else if (! strcmp (val,"2")) {
    _value = Options::P_2;
  }
  else if (! strcmp (val,"3")) {
    _value = Options::P_3;
  }
  else if (! strcmp (val,"4")) {
    _value = Options::P_4;
  }
  else if (! strcmp (val,"5")) {
    _value = Options::P_5;
  }
  else if (! strcmp (val,"6")) {
    _value = Options::P_6;
  }
  else if (! strcmp (val,"7")) {
    _value = Options::P_7;
  }
  else if (! strcmp (val,"8")) {
    _value = Options::P_8;
  }
  else if (! strcmp (val,"9")) {
    _value = Options::P_9;
  }
  else if (! strcmp (val,"10")) {
    _value = Options::P_10;
  }
  else if (! strcmp (val,"11")) {
    _value = Options::P_11;
  }
  else if (! strcmp (val,"12")) {
    _value = Options::P_12;
  }
  else if (! strcmp (val,"13")) {
    _value = Options::P_13;
  }
  else if (! strcmp (val,"14")) {
    _value = Options::P_14;
  }
  else if (! strcmp (val,"15")) {
    _value = Options::P_15;
  }
  else if (! strcmp (val,"-2")) {
    _value = Options::N_2;
  }
  else if (! strcmp (val,"-3")) {
    _value = Options::N_3;
  }
  else if (! strcmp (val,"-4")) {
    _value = Options::N_4;
  }
  else if (! strcmp (val,"-5")) {
    _value = Options::N_5;
  }
  else if (! strcmp (val,"-6")) {
    _value = Options::N_6;
  }
  else if (! strcmp (val,"-7")) {
    _value = Options::N_7;
  }
  else if (! strcmp (val,"-9")) {
    _value = Options::N_9;
  }
  else if (! strcmp (val,"-10")) {
    _value = Options::N_10;
  }
  else if (! strcmp (val,"-11")) {
    _value = Options::N_11;
  }
  else if (! strcmp (val,"-12")) {
    _value = Options::N_12;
  }
  else if (! strcmp (val,"-13")) {
    _value = Options::N_13;
  }
  else if (! strcmp (val,"-14")) {
    _value = Options::N_14;
  }
  else if (! strcmp (val,"-15")) {
    _value = Options::N_15;
  }
  else if (! strcmp (val,"1002")) {
    _value = Options::P_1002;
  }
  else if (! strcmp (val,"1003")) {
    _value = Options::P_1003;
  }
  else if (! strcmp (val,"1004")) {
    _value = Options::P_1004;
  }
  else if (! strcmp (val,"1005")) {
    _value = Options::P_1005;
  }
  else if (! strcmp (val,"1006")) {
    _value = Options::P_1006;
  }
  else if (! strcmp (val,"1007")) {
    _value = Options::P_1007;
  }
  else if (! strcmp (val,"1008")) {
    _value = Options::P_1008;
  }
  else if (! strcmp (val,"1009")) {
    _value = Options::P_1009;
  }
  else if (! strcmp (val,"1010")) {
    _value = Options::P_1010;
  }
  else if (! strcmp (val,"1011")) {
    _value = Options::P_1011;
  }
  else if (! strcmp (val,"1012")) {
    _value = Options::P_1012;
  }
  else if (! strcmp (val,"1013")) {
    _value = Options::P_1013;
  }
  else if (! strcmp (val,"1014")) {
    _value = Options::P_1014;
  }
  else if (! strcmp (val,"-1005")) {
    _value = Options::N_1005;
  }
  else if (! strcmp (val,"-1009")) {
    _value = Options::N_1009;
  }
  else if (! strcmp (val,"-1012")) {
    _value = Options::N_1012;
  }
  else if (! strcmp (val,"-1014")) {
    _value = Options::N_1014;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Selection>::set


// 27/05/2003 Manchester
bool Option<Options::Selection>::quotes () const
{
  return false;
} // Option<Options::Selection>::quotes


/**
 * Print selection to stream
 *
 * @param str the stream
 * @since 02/01/2003 Manchester
 * @since 01/07/2003 Manchester selections 0, 8-15 added
 * @since 07/07/2003 Manchester changed to use integer number of the selection
 */
void Option<Options::Selection>::toStream (ostream& str) const
{
  str << (int)_value;
} // Option<Options::Selection>::toStream


// 25/12/2002 Manchester, changed
bool Option<Options::Splitting>::set (const char* val)
{
  if (! strcmp (val,"off")) {
    _value = Options::SPLIT_OFF;
  }
  else if (! strcmp (val,"prepro_only")) {
    _value = Options::SPLIT_PREPRO;
  }
  else if (! strcmp (val,"on")) {
    _value = Options::SPLIT_ON;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<Splitting>::set


// 27/05/2003 Manchester
bool Option<Options::Splitting>::quotes () const
{
  return false;
} // Option<Options::Splitting>::quotes


// 02/01/2003 Manchester
void Option<Options::Splitting>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::SPLIT_OFF:
      str << "off";
      break;
    case Options::SPLIT_PREPRO:
      str << "prepro_only";
      break;
    case Options::SPLIT_ON:
      str << "on";
      break;
    }
} // Option<Options::Splitting>::toStream


// 25/12/2002 Manchester, names of the options changed 
bool Option<Options::SplittingStyle>::set (const char* val)
{
  if (! strcmp (val,"blocking")) {
    _value = Options::BLOCKING;
  }
  else if (! strcmp (val,"parallel")) {
    _value = Options::PARALLEL;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<SplittingStyle>::set


// 27/05/2003 Manchester
bool Option<Options::SplittingStyle>::quotes () const
{
  return false;
} // Option<Options::SplittingStyle>::quotes


// 02/01/2003 Manchester
void Option<Options::SplittingStyle>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::BLOCKING:
      str << "blocking";
      break;
    case Options::PARALLEL:
      str << "parallel";
      break;
    }
} // Option<Options::SplittingStyle>::toStream


bool Option<Options::SimplificationOrdering>::set (const char* val)
{
  if (! strcmp (val,"standard_KBO")) {
    _value = Options::KBO;
  }
  else if (! strcmp (val,"nonrecursive_KBO")) {
    _value = Options::KBO_NONREC;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<SimplificationOrdering>::set


// 27/05/2003 Manchester
bool Option<Options::SimplificationOrdering>::quotes () const
{
  return false;
} // Option<Options::SimplificationOrdering>::quotes


// 02/01/2003 Manchester
void Option<Options::SimplificationOrdering>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::KBO:
      str << "standard_KBO";
      break;
    case Options::KBO_NONREC:
      str << "nonrecursive_KBO";
      break;
    }
} // Option<Options::SimplificationOrdering>::toStream


bool Option<Options::InputSyntax>::set (const char* val)
{
  if (! strcmp (val,"tptp")) {
    _value = Options::INPUT_TPTP;
  }
  else if (! strcmp (val,"kif")) {
    _value = Options::INPUT_KIF;
  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<InputSyntax>::set


// 27/05/2003 Manchester
bool Option<Options::InputSyntax>::quotes () const
{
  return false;
} // Option<Options::InputSyntax>::quotes


// 02/01/2003 Manchester
void Option<Options::InputSyntax>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case Options::INPUT_TPTP:
      str << "tptp";
      break;
    case Options::INPUT_KIF:
      str << "kif";
      break;
    }
} // Option<Options::InputSyntax>::toStream


bool Option<OutputSyntax>::set (const char* val)
{
  if (! strcmp (val,"tptp")) {
    _value = TPTP;
  }
  else if (! strcmp (val,"kif")) {
    _value = KIF;
  }
  else if (! strcmp (val,"xml")) {
    _value = XML;
  }
  else if (! strcmp (val,"native")) {
    _value = NATIVE;
  }
  else if (! strcmp (val,"latex")) {
    _value = LATEX;
  }
  //  else if (! strcmp (val,"otter")) {
  //    _value = Options::SYNTAX_OTTER;
  //  }
  else { // wrong value
    return false;
  }

  _changed = true;
  return true;
} // Option<OutputSyntax>::set


// 27/05/2003 Manchester
bool Option<OutputSyntax>::quotes () const
{
  return false;
} // Option<OutputSyntax>::quotes


// 02/01/2003 Manchester
void Option<OutputSyntax>::toStream (ostream& str) const
{
  switch (_value) 
    {
    case TPTP:
      str << "tptp";
      break;
    case KIF:
      str << "kif";
      break;
    case XML:
      str << "xml";
      break;
    case NATIVE:
      str << "native";
      break;
    case LATEX:
      str << "latex";
      break;
    }
} // Option<OutputSyntax>::toStream

/** 
 * Find an option by its name.
 *
 * @since 10/07/2003 Manchester, "normalize" added.
 */
Opt* Options::find (const char* name) 
{
  if (*name != '-') {
    return 0;
  }
  name++;
  if (*name != '-') { // short name, like -t
    // now name is whatever follows '-'
    switch (*name) {
      case 'T': 
        if (! strcmp(name,"T") ) {
          return &_testId;
        }
        return 0;

      case 'm': 
        if (! strcmp(name,"m") ) {
          return &_memoryLimit;
        }
        return 0;

      case 'p': 
        if (! strcmp(name,"p") ) {
          return &_proof;
        }
        return 0;

      case 't': 
        if (! strcmp(name,"t") ) {
          return &_timeLimit;
        }
        return 0;

      case 'w': 
        if (! strcmp(name,"w") ) {
          return &_maxWeight;
        }
        return 0;

    default:
      // no short names so far
      return 0;
    }
  }

  name ++;
  // now name is whatever follows '--'

  switch (*name) {
    case 'a': 
      if (! strcmp(name,"age_weight_ratio") ) {
        return &_ageWeightRatio;
      }
      if (! strcmp(name,"arity_check") ) {
        return &_arityCheck;
      }
      if (! strcmp(name,"allocation_buffer_size") ) {
        return &_allocationBufferSize;
      }
      if (! strcmp(name,"allocation_buffer_price") ) {
        return &_allocationBufferPrice;
      }
      return 0;

    case 'b': 
      if (! strcmp(name,"backward_demodulation") ) {
        return &_backwardDemodulation;
      }
      if (! strcmp(name,"backward_subsumption") ) {
        return &_backwardSubsumption;
      }
      if (! strcmp(name,"bd_on_splitting_branches") ) {
        return &_bdOnSplittingBranches;
      }
      return 0;

    case 'c': 
      return 0;

    case 'd': 
      return 0;

    case 'e': 
      if (! strcmp(name,"elim_def") ) {
        return &_elimDef;
      }
      return 0;

    case 'f': 
      if (! strcmp(name,"fd_on_splitting_branches") ) {
        return &_fdOnSplittingBranches;
      }
      if (! strcmp(name,"forward_demodulation") ) {
        return &_forwardDemodulation;
      }
      if (! strcmp(name,"forward_subsumption") ) {
        return &_forwardSubsumption;
      }
      if (! strcmp(name,"forward_subsumption_resolution") ) {
        return &_forwardSubsumptionResolution;
      }
      return 0;

    case 'h': 
      if (! strcmp(name,"header_precedence_kinky") ) {
        return &_headerPrecedenceKinky;
      }
      return 0;

    case 'i': 
      if (! strcmp(name,"include") ) {
        return &_include;
      }
      if (! strcmp(name,"inherited_negative_selection") ) {
        return &_inheritedNegativeSelection;
      }
      if (! strcmp(name,"input_syntax") ) {
        return &_inputSyntax;
      }
      if (! strcmp(name,"inequality_splitting") ) {
	return &_inequalitySplitting;
      }
      if (! strcmp(name,"input_file") ) {
        return &_inputFile;
      }
      return 0;

    case 'k': 
      return 0;

    case 'l': 
      if (! strcmp(name,"literal_comparison_mode") ) {
        return &_literalComparisonMode;
      }
      if (! strcmp(name,"lrs_first_time_check") ) {
        return &_lrsFirstTimeCheck;
      }
      if (! strcmp(name,"lrs_first_memory_check") ) {
        return &_lrsFirstMemoryCheck;
      }
      return 0;

    case 'm':
      if (! strcmp(name,"main_alg") ) {
        return &_mainAlg;
      }
      if (! strcmp(name,"max_active") ) {
        return &_maxActive;
      }
      if (! strcmp(name,"max_passive") ) {
        return &_maxPassive;
      }
      if (! strcmp(name,"max_skolem") ) {
        return &_maxSkolem;
      }
      if (! strcmp(name,"max_weight") ) {
        return &_maxWeight;
      }
      if (! strcmp(name,"memory_limit") ) {
        return &_memoryLimit;
      }
      if (! strcmp(name,"miniscope") ) {
        return &_miniscope;
      }
      if (! strcmp(name,"mode") ) {
        return &_mode;
      }
      if (! strcmp(name,"max_inference_depth") ) {
        return &_maxInferenceDepth;
      }
      return 0;

    case 'n':
      if (! strcmp(name,"normalize") ) {
        return &_normalize;
      }
      if (! strcmp(name,"nongoal_weight_coefficient") ) {
        return &_nongoalWeightCoefficient;
      }
      if (! strcmp(name,"no_free_vars") ) {
        return &_noFreeVars;
      }
      if (! strcmp(name,"no_inferences_between_rules") ) {
        return &_noInferencesBetweenRules;
      }
      return 0;

    case 'o':
      if (! strcmp(name,"output_equality_infix") ) {
        return &_outputEqualityInfix;
      }
      if (! strcmp(name,"orphan_murder") ) {
        return &_orphanMurder;
      }
      if (! strcmp(name,"output_syntax") ) {
        return &_outputSyntax;
      }
      return 0;

    case 'p':
      if (! strcmp(name,"proof") ) {
        return &_proof;
      }
      if (! strcmp(name,"paramodulation") ) {
        return &_paramodulation;
      }
      return 0;

    case 'r':
      if (! strcmp(name,"random_seed") ) {
        return &_randomSeed;
      }
      if (! strcmp(name,"remove_unused_defs") ) {
        return &_removeUnusedDefs;
      }
      return 0;

    case 's':
      switch (name[1]) {
      case 'e':
	if (! strcmp(name,"selection") ) {
	  return &_selection;
	}
	if (! strcmp(name,"selected_equality_weight_coefficient") ) {
	  return &_selectedEqualityWeightCoefficient;
	}
	return 0;

      case 'h':
	if (! strcmp(name,"show_generated") ) {
	  return &_showGen;
	}
	if (! strcmp(name,"show_passive") ) {
	  return &_showPassive;
	}
	if (! strcmp(name,"show_active") ) {
	  return &_showActive;
	}
	if (! strcmp(name,"show_options") ) {
	  return &_showOptions;
	}
	if (! strcmp(name,"show_profile") ) {
	  return &_showProfile;
	}
	if (! strcmp(name,"show_prepro") ) {
	  return &_showPrepro;
	}
	if (! strcmp(name,"show_input") ) {
	  return &_showInput;
	}
	if (! strcmp(name,"show_weight_limit_changes") ) {
	  return &_showWeightLimitChanges;
	}
	return 0;

      case 'i':
	if (! strcmp(name,"silent") ) {
	  return &_silent;
	}
	if (! strcmp(name,"simplification_ordering") ) {
	  return &_simplificationOrdering;
	}
	return 0;

      case 'o':
	if (! strcmp(name,"sos") ) {
	  return &_sos;
	}
	return 0;

      case 'p':
        if (! strcmp(name,"splitting") ) {
          return &_splitting;
        }
	if (! strcmp(name,"splitting_with_naming") ) {
	  return &_splittingWithNaming;
	}
	if (! strcmp(name,"splitting_style") ) {
	  return &_splittingStyle;
	}
	return 0;

      case 't':
        if (! strcmp(name,"statistics") ) {
          return &_statistics;
	}
	return 0;

      case 'y':
	if (! strcmp(name,"symbol_precedence") ) {
	  return &_symbolPrecedence;
	}
	if (! strcmp(name,"symbol_weight_by_arity") ) {
	  return &_symbolWeightByArity;
	}
	return 0;

      default:
	return 0;
      }

    case 't':
      if (! strcmp(name,"tab") ) {
        return &_tab;
      }
      if (! strcmp(name,"term_weighting_scheme") ) {
        return &_termWeightingScheme;
      }
      if (! strcmp(name,"test_id") ) {
        return &_testId;
      }
      if (! strcmp(name,"test_number") ) {
        return &_testNumber;
      }
      if (! strcmp(name,"time_limit") ) {
        return &_timeLimit;
      }
      return 0;

    default:
      return 0;
  }
} // Options::find


// 25/12/2003 Manchester, changed passing MainAlg
Options* Options::lrs (Selection sel)
{
  Options* options = new Options;
  options->mainAlg (LRS);
  options->selection (sel);

  return options;
} // Options::lrs (Selection selection)


Options* Options::lrs (Selection sel, LiteralComparisonMode lcm)
{
  Options* options = lrs (sel);
  options->literalComparisonMode (lcm);

  if (lcm == KINKY) {
    options->headerPrecedenceKinky (true);
  }

  return options;
} // Options::lrs (Selection selection)


// 25/12/2003 Manchester, changed passing MainAlg
Options* Options::discount (Selection sel)
{
  Options* options = new Options;
  options->mainAlg (DISCOUNT);
  options->selection (sel);

  return options;
} // Options::discount (Selection selection)


// 25/12/2003 Manchester, changed passing MainAlg
Options* Options::discount (Selection sel, LiteralComparisonMode lcm)
{
  Options* options = discount (sel);
  options->literalComparisonMode (lcm);

  if (lcm == KINKY) {
    options->headerPrecedenceKinky (true);
  }

  return options;
} // Options::discount (Selection selection)


// 25/12/2003 Manchester, changed passing MainAlg
Options* Options::otter (Selection sel)
{
  Options* options = new Options;
  options->mainAlg (OTTER);
  options->selection (sel);

  return options;
} // Options::otter (Selection selection)


/** 
 * Set KIF-specific options.
 *
 * @since 13/09/2002
 */
void Options::kif ()
{
  // preprocessing options
  removeUnusedDefs (false); // everything should be fast
  arityCheck (false); // arities in KIF are variable
  noFreeVars (false); // queries may contain free variables
  miniscope (false);

  // the only proof-search option
  nongoalWeightCoefficient (2.0);

  // KIF-specific
  sos (true);
  silent (true);
  backwardSubsumption (SUBSUMPTION_MULTISET);
  tab (0);
  inputSyntax (INPUT_KIF);
  outputSyntax (::KIF);
} // Options::kif ()


/** 
 * Set CASC-19-specific options.
 *
 * @since 17/07/2003 Manchester
 */
void Options::casc19 ()
{
  // preprocessing options
  arityCheck (false);

  // CASC-19-specific
  tab (0);
  normalize(true);
  showOptions(false);
  miniscope(false);
  splitting(SPLIT_PREPRO);
  // outputSyntax(XML);
  // proof(PROOF_OFF);
  statistics(STATISTICS_BRIEF);
} // Options::casc19 ()


/** 
 * Set CASC-18-specific options.
 *
 * @since 21/07/2003 Manchester
 */
void Options::casc18 ()
{
  // preprocessing options
  arityCheck (false);

  // CASC-18-specific
  tab (0);
  normalize(true);
  showOptions(false);
  miniscope(false);
  splitting(SPLIT_PREPRO);
  // outputSyntax(XML);
  // proof(PROOF_OFF);
  statistics(STATISTICS_BRIEF);
} // Options::casc18 ()


// 25/12/2002 Manchester, changed due to change in the splitting options
void Options::fullSplitting ()
{
  splitting (SPLIT_ON);
  splittingStyle (PARALLEL);
  splittingWithNaming (true);
  fdOnSplittingBranches (true);
  bdOnSplittingBranches (true);
} // Options::fullSplitting


// give the include file name using its relative name
// WARNING: the relative name must begin and terminate with "'"
// because of the TPTP syntax
string Options::includeFileName (char* relativeName)
{
  ASS (*relativeName == '\'');

  // skip the "'" character
  relativeName++; 

  // find the ending "'" character and remove it
  char* p = relativeName + 1;
  while (*p != '\'') {
    ASS (*p != 0);
    p++;
  }
  *p = 0;

  if (*relativeName == '/') { // absolute name
    return relativeName;
  }

  // relativeName is relative. 
  // Use the conventions of Vampire: 
  // (a) first search the value of "include"
  const char* dir = include ();
  if (! dir) { // include undefined
    // (b) search the value of the environment variable TPTP_DIR
    dir = getenv("TPTP_DIR");
    if (! dir) { // the environment variable is undefined as well
      dir = ".";
    }
  }
  // now dir is the directory to search

  string result (dir);
  result += "/";
  result += relativeName;

  return result;
} // Options::includeFileName (const char* name)


/**
 * Output options to a stream.
 * @param str the stream
 * @since 02/01/2003 Manchester
 * @since 28/06/2003 Manchester, changed to treat XML output as well
 * @since 10/07/2003 Manchester, "normalize" added.
 */
void Options::output (ostream& str) const
{
  if (! _showOptions.value()) {
    return;
  }
  if (outputSyntax() == XML) {
    toXML(str);
    return;
  }

  str << "=========== Options ==========\n";

  // general options

  _mode.output(str);
  _tab.output(str);
  _inputFile.output(str);
  _randomSeed.output(str);

  // input/output options

  _statistics.output(str);
  _inputSyntax.output(str);
  _outputSyntax.output(str);
  _silent.output(str);
  _showGen.output(str);  
  _showPassive.output(str);  
  _showActive.output(str);
  _showPrepro.output(str);
  _showProfile.output(str);
  _showInput.output(str);
  _showOptions.output(str);
  _showWeightLimitChanges.output(str);  

  // preprocessor options

  _removeUnusedDefs.output(str);
  _arityCheck.output(str);
  _noFreeVars.output(str);
  _miniscope.output(str);
  _outputEqualityInfix.output(str);
  _include.output(str);

  // Kernel's options

  _mainAlg.output(str);
  _splitting.output(str);
  _splittingStyle.output(str);
  _splittingWithNaming.output(str);
  _selection.output(str);
  _literalComparisonMode.output(str);
  _inheritedNegativeSelection.output(str);
  _forwardSubsumptionResolution.output(str);
  _forwardDemodulation.output(str);
  _fdOnSplittingBranches.output(str);
  _backwardDemodulation.output(str);
  _bdOnSplittingBranches.output(str);
  _forwardSubsumption.output(str);
  _backwardSubsumption.output(str);
  _orphanMurder.output(str);
  _elimDef.output(str);
  _lrsFirstTimeCheck.output(str);
  _lrsFirstMemoryCheck.output(str);
  _maxInferenceDepth.output(str);
  _symbolPrecedence.output(str);
  _headerPrecedenceKinky.output(str);
  _noInferencesBetweenRules.output(str);
  _timeLimit.output(str);
  _memoryLimit.output(str);
  _allocationBufferSize.output(str);
  _allocationBufferPrice.output(str);
  _inequalitySplitting.output(str);
  _maxSkolem.output(str);
  _maxActive.output(str);
  _maxPassive.output(str);
  _maxWeight.output(str);
  _normalize.output(str);
  _ageWeightRatio.output(str);
  _proof.output(str);
  _testId.output(str);
  _testNumber.output(str);
  _paramodulation.output(str);
  _symbolWeightByArity.output(str);
  _termWeightingScheme.output(str);
  _simplificationOrdering.output(str);
  _nongoalWeightCoefficient.output(str);
  _selectedEqualityWeightCoefficient.output(str);
  _sos.output(str);

  str << "======= End of options =======\n";
} // Options::output (ostream& str) const


/**
 * Output options to a stream in the XML format.
 *
 * @param str the stream
 * @since 28/06/2003 Manchester
 * @since 10/07/2003 Manchester, "normalize" added.
 */
void Options::toXML (ostream& str) const
{
  str << "<options test_id='" 
      << testId()
      << "' test_number='" 
      << testNumber()
      << "'>\n";

  // general options
  _mode.toXML(str);
  _tab.toXML(str);
  _inputFile.toXML(str);
  _randomSeed.toXML(str);

  // input/output options

  _statistics.toXML(str);
  _inputSyntax.toXML(str);
  _outputSyntax.toXML(str);
  _silent.toXML(str);
  _showGen.toXML(str);  
  _showPassive.toXML(str);  
  _showActive.toXML(str);
  _showPrepro.toXML(str);
  _showProfile.toXML(str);
  _showInput.toXML(str);
  _showOptions.toXML(str);
  _showWeightLimitChanges.toXML(str);  

  // preprocessor options

  _removeUnusedDefs.toXML(str);
  _arityCheck.toXML(str);
  _noFreeVars.toXML(str);
  _miniscope.toXML(str);
  _outputEqualityInfix.toXML(str);
  _include.toXML(str);

  // Kernel's options

  _mainAlg.toXML(str);
  _splitting.toXML(str);
  _splittingStyle.toXML(str);
  _splittingWithNaming.toXML(str);
  _selection.toXML(str);
  _literalComparisonMode.toXML(str);
  _inheritedNegativeSelection.toXML(str);
  _forwardSubsumptionResolution.toXML(str);
  _forwardDemodulation.toXML(str);
  _fdOnSplittingBranches.toXML(str);
  _backwardDemodulation.toXML(str);
  _bdOnSplittingBranches.toXML(str);
  _forwardSubsumption.toXML(str);
  _backwardSubsumption.toXML(str);
  _orphanMurder.toXML(str);
  _elimDef.toXML(str);
  _lrsFirstTimeCheck.toXML(str);
  _lrsFirstMemoryCheck.toXML(str);
  _maxInferenceDepth.toXML(str);
  _symbolPrecedence.toXML(str);
  _headerPrecedenceKinky.toXML(str);
  _noInferencesBetweenRules.toXML(str);
  _timeLimit.toXML(str);
  _memoryLimit.toXML(str);
  _allocationBufferSize.toXML(str);
  _allocationBufferPrice.toXML(str);
  _inequalitySplitting.toXML(str);
  _maxSkolem.toXML(str);
  _maxActive.toXML(str);
  _maxPassive.toXML(str);
  _maxWeight.toXML(str);
  _normalize.toXML(str);
  _ageWeightRatio.toXML(str);
  _proof.toXML(str);
  _testId.toXML(str);
  _testNumber.toXML(str);
  _paramodulation.toXML(str);
  _symbolWeightByArity.toXML(str);
  _termWeightingScheme.toXML(str);
  _simplificationOrdering.toXML(str);
  _nongoalWeightCoefficient.toXML(str);
  _selectedEqualityWeightCoefficient.toXML(str);
  _sos.toXML(str);

  str << "</options>\n";
} // Options::toXML (ostream& str) const


/**
 * Tabulate options in the Prolog format to a stream.
 * @since 27/05/2003 Manchester
 * @since 10/07/2003 Manchester, "normalize" added
 */
void Options::tabulate (ostream& str) const
{
  // general options

  _mode.tabulate(str);
  _tab.tabulate(str);
  _inputFile.tabulate(str);
  _randomSeed.tabulate(str);

  // input/output options

  _statistics.tabulate(str);
  _inputSyntax.tabulate(str);
  _outputSyntax.tabulate(str);
  _silent.tabulate(str);
  _showGen.tabulate(str);  
  _showPassive.tabulate(str);  
  _showActive.tabulate(str);
  _showPrepro.tabulate(str);
  _showProfile.tabulate(str);
  _showInput.tabulate(str);
  _showOptions.tabulate(str);
  _showWeightLimitChanges.tabulate(str);  

  // preprocessor options

  _removeUnusedDefs.tabulate(str);
  _arityCheck.tabulate(str);
  _noFreeVars.tabulate(str);
  _miniscope.tabulate(str);
  _outputEqualityInfix.tabulate(str);
  _include.tabulate(str);

  // Kernel's options

  _mainAlg.tabulate(str);
  _splitting.tabulate(str);
  _splittingStyle.tabulate(str);
  _splittingWithNaming.tabulate(str);
  _selection.tabulate(str);
  _literalComparisonMode.tabulate(str);
  _inheritedNegativeSelection.tabulate(str);
  _forwardSubsumptionResolution.tabulate(str);
  _forwardDemodulation.tabulate(str);
  _fdOnSplittingBranches.tabulate(str);
  _backwardDemodulation.tabulate(str);
  _bdOnSplittingBranches.tabulate(str);
  _forwardSubsumption.tabulate(str);
  _backwardSubsumption.tabulate(str);
  _orphanMurder.tabulate(str);
  _elimDef.tabulate(str);
  _lrsFirstTimeCheck.tabulate(str);
  _lrsFirstMemoryCheck.tabulate(str);
  _maxInferenceDepth.tabulate(str);
  _symbolPrecedence.tabulate(str);
  _headerPrecedenceKinky.tabulate(str);
  _noInferencesBetweenRules.tabulate(str);
  _timeLimit.tabulate(str);
  _memoryLimit.tabulate(str);
  _allocationBufferSize.tabulate(str);
  _allocationBufferPrice.tabulate(str);
  _inequalitySplitting.tabulate(str);
  _maxSkolem.tabulate(str);
  _maxActive.tabulate(str);
  _maxPassive.tabulate(str);
  _maxWeight.tabulate(str);
  _normalize.tabulate(str);
  _ageWeightRatio.tabulate(str);
  _proof.tabulate(str);
  _testId.tabulate(str);
  _testNumber.tabulate(str);
  _paramodulation.tabulate(str);
  _symbolWeightByArity.tabulate(str);
  _termWeightingScheme.tabulate(str);
  _simplificationOrdering.tabulate(str);
  _nongoalWeightCoefficient.tabulate(str);
  _selectedEqualityWeightCoefficient.tabulate(str);
  _sos.tabulate(str);
} // Options::tabulate (ostream& str) const


/**
 * Return the problem name computed from the input file name. Returns
 * null pointer if the input file name is not specified or is wrong;
 *
 * @since 07/07/2003 Manchester, bug fixed (when the problem name contained
 *        more than one ".", the name was calculated wrongly
 */ 
const char* Options::problemName () const
{
  const char* inputFileName = _inputFile.value();

  if (inputFileName == 0) {
    return 0;
  }

  int length = strlen(inputFileName) - 1; 

  int b = length - 1; 
  while (b >= 0 && inputFileName[b] != '/') {
    b--;
  }
  b++;

  int e = length-1; 
  while (e >= b && inputFileName[e] != '.') {
    e--;
  }
  if (e < b) {
    e = length-1;
  }

  // now b points at the beginning of the name and
  // e at the character after the name
  // should be enough for most TPTP names
#define MAX_PROBLEM_NAME_LENGTH 30
  static char problemName [MAX_PROBLEM_NAME_LENGTH]; 
  ASS(e - b + 1 <= MAX_PROBLEM_NAME_LENGTH);

  strncpy(problemName,inputFileName + b,e-b); 
  problemName[e-b] = '\0';

  return problemName;
} // const char* Options::problemName () const
