/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
#ifndef __Problem__
#define __Problem__


#include "Unit.hpp"


class FSymbol;
class Options;


// ****************** class Problem, definition *******************


class Problem {
 public:
  // constructors & destructors
  explicit Problem ( const UnitList& );
  ~Problem ();
  void preprocess (Options&);

  /**
   * Return the units
   */
  UnitChain& units () { 
    return _units; 
  }

  // properties
  bool hasFormulas () const { return _hasFormulas; }

  // miscellaneous
  int removeUnusedDefinitions (); // the result is the number of removed definitions
  int removeEqualityAxioms ();
  int split ();
  void ennf ();
  int removeIff ();
  int miniscope ();
  void flatten ();
  void rectify ();
  void skolemize (); 
  void clausify ();
  void normalize ();   // normalize and sort clauses/formulas in the input
  void addUnit (Unit u); // for incremental building
  void removeAllConjectures (); // for work with KIF
  const VarList& originalAnswerVariables () const
    { return _originalAnswerVariables; }
  const Literal& answerLiteral () const { return _answerLiteral; }

 private:

  // structure
  UnitChain _units;
  bool _hasFormulas;
  VarList _originalAnswerVariables;
  Literal _answerLiteral;

  // auxiliary functions
  bool hasFreeVars (Formula&) const;
  int findDefinitions ();
  void clausify (FormulaList formulas, 
		 LiteralList literals,
		 Unit parent);
};


#endif // __Problem__
