/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
//  file Refutation.hpp
//  implements output of refutations
//  


#ifndef __Refutation__
#  define __Refutation__


// class ostream;
class Output;


#include "List.hpp"
#include "Unit.hpp"
#include "VampireKernel.hpp"


class Refutation
{
 public:
  // constructors/destructors
  Refutation (const VampireKernel& kernel, 
	      const Problem& problem);
  Unit condense () const; // return refutation in condensed form
  /**
   * Return the goal unit of this refutation.
   *
   * @since 28/06/2003 Manchester
   */
  Unit goal () const { return _goal; }
  const VampireKernel& kernel () const;
  const Problem& problem () const;

 private:
  class Map; // defined in Refutation.cpp

  const VampireKernel& _kernel;
  const Problem& _problem;
  Unit _goal;

  Unit collect (const VampireKernel::Clause*, Map& map);
  UnitList collect (const VampireKernel::Clause::Ancestor* ancestors, Map& map);
  Unit condense (Unit root, Map& condenced) const; 
  UnitList condense (UnitList units, Map& condensed) const; 
}; // class Refutation


// **************** class Refutation, implementation ***********************


inline
const VampireKernel& Refutation::kernel () const
{
  return _kernel;
} // Refutation::kernel


inline
const Problem& Refutation::problem () const
{
  return _problem;
} // Refutation::problem


#endif // __Refutation__
