
/******************************************************************
*											*
*     Header file to be included by all SOFTIE SCOTT files.		*
*											*
*											*
******************************************************************/


// #define SOFTIE_IN_OTTER_MODE  // compile as OTTER but retain runtime stats dump etc.

#include "header.h"   	/* OTTER declarations and standard include files */


/* ----------------- SEMANTIC GUIDANCE PARAMETERS ------------------ */

/*** number of given clauses after which model does not get updated anymore ***/
#define UPDATE_STOP 250

/*** maximum number of seconds allowed per SFINDER model search ***/
#define SFINDER_TIME_LIMIT 2

/*** cutoff number of function symbols for clause to be considered short ***/
#define SHORT_CLAUSE_CUTOFF 4

/*** SFINDER sort cardinality, default is 3 ***/
#define SFINDER_SORT_CARDINALITY 3


/* ----------------- INTERNAL PARAMETERS ------------------ */

/*** maximum number of distinct constants/functions/predicates allowed ***/
#define MAX_CONSTANTS 222
#define MAX_FUNCTIONS 222
#define MAX_PREDICATES 222

/*** maximum number of literals per clause ***/
#define MAX_NUMBER_LITERALS 333

/* ----------------- SYMBOLS AND TYPE SUBSTITUTIONS ------------------ */

/*** symbol types ***/
#define CONSTANT 1
#define FUNCTION 2
#define PREDICATE 3

/*** sign of literals ***/
#define POS 1
#define NEG 0

/*** domain types ***/
#define BOOL 1
#define THING 2

/* SFINDER's boolean type, also defined in Fvntr.h */
#define boolean char
#define false ((boolean)0)
#define true ((boolean)1)



/* ----------------- functions in SS_init.c ------------------ */

void init_softie_scott();								// initialize SFINDER model
void write_sfinder_constant(struct term *constant, unsigned char type);	// called from init_softie_scott() and write_sfinder_complex(...)
void write_sfinder_complex(struct term *compl, unsigned char type);	// called from init_softie_scott() and itself
BOOLEAN is_used(unsigned char symbol, char type);				// called from write_sfinder_... functions
int get_count(unsigned char type);		// return distinct total count of symbols type


/* ----------------- functions in SS_clauses.c ------------------ */

void convert_to_sfinder_clause(struct clause *cl, int softweight, dynstring *str);		// convert OTTER clause to SFINDER input string
void init_buckets();
void convert_to_sfinder_term(struct term *t, unsigned char complex, dynstring *str);	// convert OTTER term to SFINDER input string
void convert_to_sfinder_complex(struct term *t, unsigned char complex, dynstring *str);	// convert OTTER complex to SFINDER input string


/* ----------------- functions in SS_test.c ------------------ */

boolean test_in_model(struct clause *cl);	// test clause in model without updating it
boolean update_model(struct clause *cl);	// update the model with given clause
boolean still_update_model();			// find out if update stop is reached


/* ----------------- functions from SS_stats.c ------------------ */

void print_initial_softie_stats();  // print a series of initial stats of the problem

// helper functions
int get_clause_count(struct list *clause_list);
int get_number_short_clauses(struct list *clause_list);
void count_shorts(struct term *current_term);
boolean symbol_short_counted(unsigned short sym_num);


// the following provide runtime stats for visualization with the program anim

int model_clauses;		// number of clauses on which current model is based
int false_given;			// number of clauses false in current model
void init_dump_file();		// creates blank dump file
void close_dump_file();		// closes dump file
void update_dump_stats();	// updates stats variables to be written to dump file
void dump_runtime_data();	// writes selected data to a file once per selected given clause to be used to visualize process using the tool anim
void output_soft_stats(FILE *fp);


// the following variables and functions are for statistics determination

#define MAX_SYMBOLS 133

int symbol_count;

struct {
	int symbol;
	int count;
} symbols[MAX_SYMBOLS];

int sym_index;

int short_tested_syms_index;
unsigned short short_tested_symbols[MAX_SYMBOLS];

int get_UL_symbol_counts();
int get_percentage();

void count_constant(struct term *current_term);
void count_complex(struct term *current_term);
boolean is_counted(int sym_num);
void remove_from_bucket(struct clause *cl);
void empty_bucket_zero();
void put_in_bucket(struct clause *cl);
void retest_list(struct list *lp);
void print_buckets();


/* ----------------- functions from FINDER.h ------------------ */

int found( const dynstring *s, int add_if_true, int seek_model );
boolean init_finder(char s[]);
void set_no_semantics();
