(module connection
	(include "dctp.sch")
	(export (clause-connections clauses) (clear-clause-connections clauses))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm")))

;;;*************************************************************************
;;--------------------------------
;; The connection graph functions
;;--------------------------------

(define (clause-connections clauses)
  (if (null? clauses)
      #t
      (if (subgoals-connections (clause-subgoals (car clauses)) (cdr clauses))
	  (let (u-connection)
	    (set! u-connection (contains-unit-connection (subgoal-links (car (clause-subgoals (car clauses))))))
	    (if (and u-connection (unit-clause (car clauses)))
		(begin
		  (if (not silent)
		      (begin
			(display "Found unit connection between clauses ")
			(display-clause (car clauses))
			(display " and  ")
			(display-clause (subgoal-clause u-connection))
			(newline)))
		  #f)
		(clause-connections (cdr clauses))))
	  (begin
	    (if (not silent)
		(begin
		  (display "Empty clause found in input set")
		  (newline)))
	    #f))))

(define (subgoals-connections subgoals clauses)
  (if (null? subgoals)
      #t
      (if (eq? falsum (car subgoals))
	  #f
	  (if (subgoal-connections (car subgoals) clauses)
	      (subgoals-connections (cdr subgoals) clauses)
		#f))))

(define (subgoal-connections subgoal clauses)
  (if (null? clauses)
      #t
      (if (eq? falsum (car (clause-subgoals (car clauses))))
	  #f
	  (begin
	    (subgoal-clause-connections subgoal (clause-subgoals (car clauses)))
	    (subgoal-connections subgoal (cdr clauses))))))

(define (subgoal-clause-connections subgoal subgoals)
  (if (null? subgoals)
      #t
      (let ((other-subgoal (car subgoals)))
	(if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign other-subgoal)))
		 (eq? (subgoal-pred subgoal) (subgoal-pred other-subgoal))
		 (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms other-subgoal)))
	    (begin
	      (unbind-vars)
	      (if (> debug-level 4)
		  (begin
		    (display "Found connection between subgoals ")
		    (display-subgoal subgoal)
		    (display " and  ")
		    (display-subgoal other-subgoal)
		    (newline)))
	      (set-subgoal-links subgoal (cons other-subgoal (subgoal-links subgoal)))
	      (set-subgoal-links other-subgoal (cons subgoal (subgoal-links other-subgoal))))
	    (if (and (eq? (subgoal-pred subgoal) equality-pred)
		     (not (eq? (subgoal-sign subgoal) (subgoal-sign other-subgoal)))
		     (eq? (subgoal-pred subgoal) (subgoal-pred other-subgoal))
		     (unify-lists-stop (subgoal-terms subgoal) 
				       (list (cadr (subgoal-terms other-subgoal)) (car (subgoal-terms other-subgoal)))))
		(begin
		  (unbind-vars)
		  (if (> debug-level 4)
		      (begin
			(display "Found symmetric connection between eq-subgoals ")
			(display-subgoal subgoal)
			(display " and  ")
			(display-subgoal other-subgoal)
			(newline)))
		  (set-subgoal-links subgoal (cons other-subgoal (subgoal-links subgoal)))
		  (set-subgoal-links other-subgoal (cons subgoal (subgoal-links other-subgoal))))))
	(subgoal-clause-connections subgoal (cdr subgoals)))))

(define (subgoal-clause-connections-internal subgoal renamed subgoals)
  (if (null? subgoals)
      #t
      (let ((other-subgoal (car subgoals)))
	(if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign other-subgoal)))
		 (eq? (subgoal-pred subgoal) (subgoal-pred other-subgoal))
		 (unify-lists-stop (subgoal-terms renamed) (subgoal-terms other-subgoal)))
	    (begin
	      (unbind-vars)
	      ;;;(print "Found internal link in clause " (subgoal-depth other-subgoal) " (" (subgoal-depth subgoal)  ")")
	      (set-subgoal-links subgoal (cons other-subgoal (subgoal-links subgoal)))
	      (set-subgoal-links other-subgoal (cons subgoal (subgoal-links other-subgoal)))))
	(subgoal-clause-connections subgoal (cdr subgoals)))))

(define (contains-unit-connection connections)
  (if (pair? connections)
      (if (unit-clause (subgoal-clause (car connections)))
	  (car connections)
	  (contains-unit-connection (cdr connections)))
      #f))

(define (clear-clause-connections clauses)
  (if (pair? clauses)
      (begin
	(clear-subgoals-connections (clause-subgoals (car clauses)))
	(clear-clause-connections (cdr clauses)))))

(define (clear-subgoals-connections subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-links (car subgoals) '())
	(clear-subgoals-connections (cdr subgoals)))))