(module dctp
	(main call-dctp) ;; this is the main procedure
	(include "dctp.sch")
	(export (variables-in-subgoals subgoals vars) (clause-pointers-in-subgoals subgoals clause) (clause-subsumed clause)
		(unrename-vars))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm") (pruning "pruning.scm") 
		(read-tptp input "input.scm") (index "index.scm") (parser "parser.scm") (select "select.scm") 
		(signals "signals.scm") (path "path.scm") (link "link.scm") (connection "connection.scm") 
		(initial-path "initial-path.scm") (normal-form "normal-form.scm") (cputime "cputime.scm")))

(define (call-dctp argv) 
  (set-signal-handlers)
  (dctp (cdr argv)))

(define (dctp args)
  (let ((initial-subgoal #f))
    (set! start (getcputime))
    (init-index-constants)
    (read-tptp (car (reverse (get-flags args))))
    (if (= nr-all-negatives 0)
	(begin
	  (set! end (getcputime))
	  (set! elapsed (/fl (exact->inexact (- end start)) (exact->inexact 100.0)))
	  (print "!!!!!!!!!!!!!!!! Exit status is success (MODEL found in " elapsed " seconds) !!!!!!!!!!!!!!!!")
	  (exit 0)))
    (if (clause-connections (formula-clauses formula))
	(begin
	  (select-initial-path (formula-clauses formula))
	  (initialize-prover)
	  (if dewey-numbers
	      (print "initial active path:"))
	  (set! initial-subgoal (initial-active-path (formula-clauses formula) #f))
	  (if initial-subgoal
	      (begin
		(init-link-bias)
		(trail-stop)
		(make-initial-link-list initial-subgoal)
		(trail-erase)
		(if limit-term-depth
		    (begin
		      (set! allowed-term-depth min-relevant-depth)
		      (display "limiting term complexity to ")
		      (display allowed-term-depth)
		      (newline)))
		(if (> debug-level 5)
		    (begin
		      (print-predicates-indexes predicate-list)
		      (display "initial links: ")
		      (newline)
		      (display-link-vector)
		      (newline)))
		(if test-mode
		    (begin
		      (print "entering test mode")
		      (set! debug-level 9)
		      (test-subsumption-for-units (formula-clauses formula)))
		    (begin
		      (if (not silent) (print "Starting proof"))
		      (if (solve-initial-subgoal initial-subgoal)
			  (begin
			    (if (> debug-level 4)
				(print-predicates-indexes predicate-list))
			    (set! end (getcputime))
			    (set! elapsed (/fl (exact->inexact (- end start)) (exact->inexact 100.0)))
			    (print "**************** Exit status is success (PROOF found in " elapsed " seconds) ****************"))
			  (begin
			    (set! end (getcputime))
			    (set! elapsed (/fl (exact->inexact (- end start)) (exact->inexact 100.0)))
			    (print "!!!!!!!!!!!!!!!! Exit status is success (MODEL found in " elapsed " seconds) !!!!!!!!!!!!!!!!")))
		      (print-statistics)
		      (exit 0))))
	      (begin
		(print "Empty input set.")
		(if error-on-empty
		    (begin
		      (print "Exit status is ERROR (empty input set)")
		      (exit 1))
		    (begin
		      (set! end (getcputime))
		      (set! elapsed (/fl (exact->inexact (- end start)) (exact->inexact 100.0)))
		      (print "!!!!!!!!!!!!!!!! Exit status is success (MODEL found in " elapsed " seconds) !!!!!!!!!!!!!!!!")
		      (exit 0))))))
	(begin
	  (print "complementary unit pair in input set.")
	  (set! end (getcputime))
	  (set! elapsed (/fl (exact->inexact (- end start)) (exact->inexact 100.0)))
	  (print "**************** Exit status is success (PROOF found in " elapsed " seconds) ****************")
	  (exit 0)))))

(define (solve-subgoals subgoals rel-vector)
  ;; iterative version, recursive version had to be abandoned to save stack space
  (let (subgoal-rel-vector (subgoal (select-subgoal subgoals #f)))
    (if subgoal
	(set! subgoal-rel-vector (solve-subgoal subgoal)))
    (do ()
	((or (not subgoal) (not subgoal-rel-vector) (and pruning (not (subgoal-relevant subgoal subgoal-rel-vector)))) #t)
      (merge-rel-vectors rel-vector subgoal-rel-vector (subgoal-depth subgoal))
      (set! subgoal (select-subgoal subgoals #f))
      (if subgoal
	  (set! subgoal-rel-vector (solve-subgoal subgoal))))
    (if subgoal 
	(if subgoal-rel-vector
	    (if (and pruning (not (subgoal-relevant subgoal subgoal-rel-vector)))
		(begin
		  (if dewey-numbers
		      (begin
			(display "pruning clause ") 
			(display-clause (subgoal-clause subgoal))
			(newline)))
		  (set! nr-prunings (+ nr-prunings 1))
		  (cons subgoal-rel-vector #t)))
	    (cons subgoal-rel-vector #f))
	(cons rel-vector #f))))

(define (clear-clause clause)
  (clear-subgoals (clause-subgoals clause))
  (vector-fill! clause #f))

(define (clear-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(if (not (eq? (car subgoals) falsum)) (vector-fill! (car subgoals) #f))
	(clear-subgoals (cdr subgoals)))))

(define (solve-initial-subgoal subgoal)
  (let ((rel-vector (new-rel-vector 0)))
    (if (ground-closed subgoal (subgoal-predecessor subgoal) rel-vector)
	rel-vector
	(begin
	  (if toggle-inferences
	      (toggle-selection))
	  (try-link-with-subgoal subgoal rel-vector)))))

(define (solve-subgoal subgoal)
  ;; NB: As of now, it is assumed that all links in the link list are unifiable.
  ;;     If they are not, solve-subgoal will return with an unsound #t
  ;;     Furthermore, the selected link is not trailed as used, the tagging is done
  ;;     manually.
  (let ((tmp-result #f) (solved #f)
	(rel-vector (new-rel-vector (subgoal-depth subgoal))))
    (if dewey-numbers
	(begin
	  (newline)
	  (display "solving subgoal ") 
	  (display-subgoal subgoal) (display " [")
	  (display (clause-unsimplified (subgoal-clause subgoal))) (display "] on depth ")
	  (display (subgoal-depth subgoal)) (newline)))
    (if (eq? falsum subgoal)
	(begin
	  (if dewey-numbers
	      (print "empty clause deduced by unit simplification!"))
	  rel-vector)
	(begin
	  (if (and top-down-lemmas (simplified-unit subgoal))
	      (begin
		(set-subgoal-unlinkable subgoal #t)
		(set! tmp-result (make-top-down-lemma subgoal))))
	  (if (and pruning tmp-result)
	      (begin
		(mark-rel-vector rel-vector tmp-result)
		(if dewey-numbers (print "new lemma causes pruning to level " tmp-result))
		(if (> debug-level 4)
		    (begin
		      (display "referenced levels:")
		      (display-rel-vector rel-vector)
		      (newline)))
		rel-vector)
	      (if (subgoal-simplifiable subgoal)
		  (begin
		    (mark-rel-vector rel-vector (subgoal-depth subgoal))
		    rel-vector)
		  (if (ground-closed subgoal (subgoal-predecessor subgoal) rel-vector)
		      rel-vector
		      (begin
			;;(check-for-local-vars subgoal)
			;;(set! total-subgoals-solved (+ total-subgoals-solved 1))
			(insert-subgoal-in-gr-index subgoal)
			(if (> debug-level 5) (print-indexes (subgoal-pred subgoal) 'gr))
			(if (and index-info (> successful-inferences 0) (= 0 (remainder successful-inferences index-info)))
			    (begin
			      (print "index statistics after " successful-inferences " linkings on depth " (subgoal-depth subgoal))
			      (print "each field has format index-type entries total-paths empty-paths avg-depth avg-branch max-branch")
			      (predicate-index-statistics predicate-list)))
			(if (and variant-info (> successful-inferences 0) (= 0 (remainder successful-inferences variant-info)))
			    (begin
			      (print "variant index statistics after " successful-inferences " linkings on depth " (subgoal-depth subgoal))
			      (print "each field has format index-type entries total-paths empty-paths avg-depth avg-branch max-branch")
			      (variant-index-statistics (formula-clauses formula))))
			(if (not (subgoal-unlinkable subgoal))
			    (begin
			      (if (> debug-level 4)
				  (print "inserting path links"))
			      (insert-links subgoal (potential-linked-lits subgoal))
			      (if (> debug-level 4)
				  (print "inserting lemma links"))
			      (insert-links subgoal (potential-linked-lemmas subgoal))
			      (add-path-literal subgoal)
			      (set! solved (try-link-with-subgoal subgoal rel-vector))
			      (delete-path-literal subgoal))
			    (set! solved (try-link-with-subgoal subgoal rel-vector)))
			(set-subgoal-next subgoal #f)
			(mark-links-deletable (subgoal-links subgoal) subgoal)
			(set-subgoal-links subgoal '())
			(delete-subgoal-from-gr-index subgoal)
			(set! tmp-result #f)
			(if (and foldup solved (not (subgoal-simplified subgoal)) (subgoal-relevant subgoal solved))
			    (begin
			      (set! tmp-result (foldup-subgoal subgoal solved))
			      (if (and dewey-numbers (not tmp-result))
				  (begin
				    (display "referenced levels:")
				    (display-rel-vector solved)
				    (newline)))))
			(if (and pruning tmp-result)
			    (begin
			      (set! rel-vector (new-rel-vector (subgoal-depth subgoal)))
			      (mark-rel-vector rel-vector tmp-result)
			      (if dewey-numbers (print "new lemma causes pruning to level " tmp-result))
			      rel-vector)
			    solved)))))))))

(define (try-link-with-subgoal subgoal rel-vector)
  (let ((link-cell  '())
	used-link-cells result-rel-vector subgoals-pair)
    (do ((terminate #f))
	(terminate #t)
      ;; premature abort after defined max nr. of inferences
      (if (and inference-limit (> total-inferences inference-limit))
	  (begin
	    (set! end (getcputime))
	    (set! elapsed (/fl (exact->inexact (- end start)) (exact->inexact 100.0)))
	    (print "Exit status is FAIL (inference limit exceeded after " elapsed " seconds)")
	    (if tptp-lemmas (print-tptp-lemmas subgoal))
	    (print-statistics)
	    (exit 0)))
      (set! used-link-cells link-cell)
      (set! link-cell  (select-link-cell search-start-index 1))
      ;;(set! link-cell  (select-link-cell 0 1))
      (if (pair? link-cell)
	  (begin 
	    (set-cdr! link-cell used-link-cells)
	    (set! used-link-cells link-cell)
	    (set-link-active (car link-cell) #f)
	    ;;(set! links-tagged (+ links-tagged 1))
	    (set! subgoals-pair (linking-step (car link-cell) subgoal rel-vector))))
      (if (or (not (pair? link-cell)) (pair? (car subgoals-pair)))
	  (set! terminate #t)))
    (if (not (pair? link-cell))
	(begin 
	  (if (not silent) (print "Affirmation state reached."))
	  #f)
	(let (solved-pair-a solved-pair-b 
			    (rel-vector-b (new-rel-vector (+ 1 (subgoal-depth (caar subgoals-pair))))))
	  (if variant-index (insert-clause-in-index (subgoal-clause (caar subgoals-pair))))
	  (set! solved-pair-b (cons rel-vector-b #f)) ;;create a dummy result for the second clause
	  (if (pair? (cdr subgoals-pair)) ;; if two new clauses have been created
	      (begin
		(if variant-index (insert-clause-in-index (subgoal-clause (cadr subgoals-pair))))
		(insert-subgoal-in-gr-index (subgoal-predecessor (cadr subgoals-pair)))
		(if (not (subgoal-unlinkable (subgoal-predecessor (cadr subgoals-pair)))) ;; create the links
		    (begin                                                                ;; for the upper linked subgoal
		      (insert-links (subgoal-predecessor (cadr subgoals-pair)) 
				    (potential-linked-lits (subgoal-predecessor (cadr subgoals-pair))))
		      (insert-links (subgoal-predecessor (cadr subgoals-pair)) 
				    (potential-linked-lemmas (subgoal-predecessor (cadr subgoals-pair))))
		      (add-path-literal (subgoal-predecessor (cadr subgoals-pair)))
		      (set! solved-pair-b (solve-subgoals (cdr subgoals-pair) rel-vector-b))
		      (delete-path-literal (subgoal-predecessor (cadr subgoals-pair))))     ;; from the path list
		    (set! solved-pair-b (solve-subgoals (cdr subgoals-pair) rel-vector-b)))
		(set-subgoal-next (subgoal-predecessor (cadr subgoals-pair)) #f)
		(mark-links-deletable (subgoal-links (subgoal-predecessor (cadr subgoals-pair)))
				      (subgoal-predecessor (cadr subgoals-pair)))
		(set-subgoal-links (subgoal-predecessor (cadr subgoals-pair)) '())
		(if context-lemmas (delete-clause-foldup-lits (subgoal-clause (cadr subgoals-pair))))
		(if variant-index (delete-clause-from-index (subgoal-clause (cadr subgoals-pair))))
		(delete-subgoal-from-gr-index (subgoal-predecessor (cadr subgoals-pair)))))
	  ;; the following is new: once the lower subgoals have been solved, they are no longer of significance
	  ;; regardless whether they have been pruned or not. the only thing that matters is the relevance of 
	  ;; the upper linked literal, which is guaranteed for all cases except when the lower linked literal
	  ;; has been simplified. but if the upper linked literal was not used, the upper clause can be pruned
	  ;; in any case
	  (if (or (not (car solved-pair-b)) ;; if the second clause (cdr subgoals-pair) could not be solved or
		  (and (pair? (cdr subgoals-pair)) ;; there was a lower clause and for its solution
		       (not                             ;; the upper linked literal was not used
			(subgoal-relevant (subgoal-predecessor (cadr subgoals-pair)) (car solved-pair-b)))))
	      (set! result-rel-vector (car solved-pair-b)) ;; pass back solely the result of the lower clause
	      (let ((tmp-result #f)) ;; otherwise solve the remaining subgoals of the upper clause
		(if (and foldup
			 (pair? (cdr subgoals-pair))
			 (not (subgoal-simplified (subgoal-predecessor (cadr subgoals-pair)))))
		    (begin
		      (set! tmp-result (foldup-subgoal (subgoal-predecessor (cadr subgoals-pair)) (car solved-pair-b)))
		      (if (and dewey-numbers (not tmp-result))
			  (begin
			    (display "referenced levels:")
			    (display-rel-vector (car solved-pair-b))
			    (newline)))))
		(if (and pruning tmp-result)
		    (begin
		      (set! result-rel-vector 
			    (new-rel-vector (subgoal-depth (subgoal-predecessor (cadr subgoals-pair)))))
		      (mark-rel-vector result-rel-vector tmp-result)
		      (if dewey-numbers (print "new lemma causes pruning to level " tmp-result)))
		    (begin
		      (set! solved-pair-a (solve-subgoals (car subgoals-pair) rel-vector))
		      (if (or (not (car solved-pair-a)) (cdr solved-pair-a))
			  ;; if the upper clause (car subgoals-pair) could not be solved or has been pruned away
			  (set! result-rel-vector (car solved-pair-a))
			  (set! result-rel-vector (merge-rel-vectors (car solved-pair-a)
								     (car solved-pair-b)
								     (subgoal-depth (caar subgoals-pair)))))))))
	  (if context-lemmas (delete-clause-foldup-lits (subgoal-clause (caar subgoals-pair))))
	  (if variant-index (delete-clause-from-index (subgoal-clause (caar subgoals-pair))))
	  (if (pair? (cdr subgoals-pair)) (clear-clause (subgoal-clause (cadr subgoals-pair))))
	  (clear-clause (subgoal-clause (caar subgoals-pair)))
	  (restore-link-cells used-link-cells)
	  result-rel-vector))))

(define (ground-closed subgoal predsubgoal rel-vector)
  (let (index-sg (index-end (ground-closed-with-index subgoal)))
    (if (and index-end (or (not (clause-initial (subgoal-clause subgoal))) 
			   (unit-subgoal subgoal)))
	(begin
	  (set! index-sg (car (vector-ref index-end 1)))
	  (if (or dewey-numbers (> debug-level 4))
	      (begin
		(display "GR-reduction-step for subgoal ")
		(display-subgoal subgoal)
		(display " on depth ") 
		(display (subgoal-depth subgoal))
		(if (subgoal-clause index-sg)
		    (begin
		      (display " with path literal ") 
		      (display-subgoal index-sg)
		      (display " on depth ") 
		      (display (subgoal-depth index-sg)))
		    (begin
		      (display " with context lemma ") 
		      (display-subgoal index-sg)
		      (display " on depth ") 
		      (display (subgoal-number index-sg))))
		(newline)))
	  (set! successful-reduction-steps (+ successful-reduction-steps 1))
	  (mark-rel-vector rel-vector (subgoal-depth subgoal))
	  (if (subgoal-clause index-sg)
	      (mark-rel-vector rel-vector (subgoal-depth index-sg))
	      (begin
		(set! context-reduction-steps (+ context-reduction-steps 1))
		(merge-rel-vectors rel-vector (subgoal-links index-sg) (subgoal-number index-sg))))
	  rel-vector)
	#f)))

(define (unrename-vars)
  (if (trail-content)
      (begin (set-variable-renaming (caar trail) #f) (trail-previous) (unrename-vars))
      (begin (trail-previous) #f)))

(define (trail-erase)
 (if (trail-content)
     (begin (trail-previous) (trail-erase))
     (begin (trail-previous) #f)))

(define (linking-step link subgoal rel-vector)
  (let ((clausea (subgoal-clause (link-lita link)))
	(clauseb (subgoal-clause (link-litb link)))
	(depth (subgoal-depth subgoal))
	(new-subgoals (cons '() '())))
    (if (unify-lists-stop (subgoal-terms (link-lita link)) (subgoal-terms (link-litb link)))
	(begin
	  (set! total-inferences (+ total-inferences 1))
	  (if dewey-start
	      (if (>= total-inferences dewey-start)
		  (begin
		    (set! dewey-numbers #t)
		    (set! dewey-start #f))))
	  (if dewey-numbers
	      (begin
		(display "linking-step with link depths ") 
		(display (subgoal-depth (link-lita link)))
		(display " and ") 
		(display (subgoal-depth (link-litb link)))
		(display " below subgoal ")
		(display-subgoal subgoal)
		(newline)
		(print "+++++++++++++++++++++++++++++++++++++")
		(display-clauses (list clausea clauseb))
		(newline)
		(print "+++++++++++++++++++++++++++++++++++++")))
	  (if (or tautologies
		  (and (non-tautological clausea) (non-tautological clauseb)))
	      (let (variant-a variant-b (regular-a #f) (regular-b #f) (subsumed-a #f) (subsumed-b #f) 
			      (unsimplifieds-a #f) (unsimplifieds-b #f))
		(set-linked-clauses-tested clausea clauseb)
		(set! variant-a (clause-index-variant? clausea))
		(if (not variant-a)
		    (begin
		      (set! subsumed-a (clause-subsumed clausea))
		      (if (not subsumed-a)
			  (begin
			    (set! regular-a  (or (not regularity) (ground-regular clausea)))
			    (if (and regular-a full-variants)
				(begin
				  (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea)))
				  (compute-current-weights unsimplifieds-a)
				  (set! variant-a (clause-variant-subsumed unsimplifieds-a))))))))
		(set! variant-b (clause-index-variant? clauseb))
		(if (not variant-b)
		    (begin
		      (set! subsumed-b (clause-subsumed clauseb))
		      (if (not subsumed-b)
			  (begin
			    (set! regular-b  (or (not regularity) (ground-regular clauseb)))
			    (if (and regular-b full-variants)
				(begin
				  (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals clauseb)))
				  (compute-current-weights unsimplifieds-b)
				  (set! variant-b (clause-variant-subsumed unsimplifieds-b))))))))
		(unset-linked-clauses clausea clauseb)
		(if (and couples (non-unit-clause clausea) (non-unit-clause clauseb)
			 (not (eq? variant-a clausea)) (not (eq? variant-b clauseb))
			 (or variant-a variant-b subsumed-a subsumed-b (not regular-a) (not regular-b)))
		    (begin 
		      (set! coupled-eliminations (+ 1 coupled-eliminations))
		      (if dewey-numbers
			  (begin
			    (display "coupled elimination") 
			    (newline))))
		    (let (copya copyb copylita copylitb (grclosed-a #f))
		      (if (and (not (or subsumed-a variant-a)) regular-a)
			  (begin
			    (set! copya (renamed-copy clausea unsimplifieds-a))
			    (set! copylita (subgoal-in-copy (link-lita link) (clause-subgoals copya)))
			    (if (and copylita (vector? copylita) (not (eq? copylita falsum))
				     (ground-closed copylita subgoal rel-vector))
				(set! grclosed-a #t))
			    (set! depth (+ depth 1))
			    (set-clause-depth copya depth)
			    (path-vector-set depth copya)
			    (set-clause-path copya subgoal)
			    (set-subgoal-next subgoal copya) 
			    (set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clausea)
									      (clause-unsimplified copya))))
			    (set-car! new-subgoals (clause-subgoals copya))))
		      (if (and (not (or subsumed-b variant-b)) regular-b (not grclosed-a))
			  (begin
			    (set! copyb (renamed-copy clauseb unsimplifieds-b))
			    (set! copylitb (subgoal-in-copy (link-litb link) (clause-subgoals copyb)))
			    (set! depth (+ depth 1))
			    (set-clause-depth copyb depth)
			    (path-vector-set depth copyb)
			    (set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clauseb)
									      (clause-unsimplified copyb))))
			    (if (= depth (+ (subgoal-depth subgoal) 1))
				(begin
				  (set-clause-path copyb subgoal)
				  (set-subgoal-next subgoal copyb)
				  (set-car! new-subgoals (clause-subgoals copyb)))
				(begin
				  (set-clause-path copyb copylita)
				  (set-subgoal-next copylita copyb)
				  (set-subgoal-selected copylita #t)
				  ;; this should cause no harm: at this point we are sure to add two clauses...
				  ;; this might cause harm, rel-vector-wise. The lower subgoal must be left unselected
				  ;; so it can be actively solved by a GR-closure, thus making copylita relevant!
				  ;;(set-subgoal-selected copylitb #t)
				  (if unit-on-top
				      (if (and simplify (unit-clause copya) (not (subgoal-simplified copylitb)))
					  (begin 
					    (if (or dewey-numbers (> debug-level 4))
						(begin
						  (display "simplifying subgoal ")
						  (display-subgoal copylitb)
						  (display " linked to unit ")
						  (display-subgoal copylita)
						  (newline)))
					    (set-clause-unsimplified copyb (- (clause-unsimplified copyb) 1))
					    (set-subgoal-simplified copylitb #t)))
				      (if (and simplify (unit-clause copyb) (not (subgoal-simplified copylita)))
					  (begin 
					    (if (or dewey-numbers (> debug-level 4))
						(begin
						  (display "simplifying subgoal ")
						  (display-subgoal copylita)
						  (display " linked to unit ")
						  (display-subgoal copylitb)
						  (newline)))
					    (set-clause-unsimplified copya (- (clause-unsimplified copya) 1))
					    (set-subgoal-simplified copylita #t))))
				  (set-cdr! new-subgoals (clause-subgoals copyb))))))))
		(if (pair? (car new-subgoals))
		    (begin
		      (set! successful-inferences (+ successful-inferences 1))
		      ;;(mark-rel-vector rel-vector (clause-depth (clause-predecessor (subgoal-clause (caar new-subgoals)))))
		      (set! tableau-clauses (+ tableau-clauses 1))
		      (if (> depth max-depth) (set! max-depth depth))
		      (if (pair? (cdr new-subgoals))
			  (begin
			    ;;(mark-rel-vector rel-vector (clause-depth (clause-predecessor (subgoal-clause (cadr new-subgoals)))))
			    (set! tableau-clauses (+ tableau-clauses 1))
			    (if dewey-numbers
				(begin
				  (display "attached new clauses on depths ") 
				  (display (- depth 1)) (display " and ") (display depth)
				  (newline))))
			  (begin
			    (if dewey-numbers
				(begin
				  (display "attached new clause on depth ") 
				  (display depth)
				  (newline)))))
		      (if dewey-numbers
			  (begin
			    (display "new subgoals: ")
			    (display-subgoal-list (car new-subgoals))
			    (if (pair? (cdr new-subgoals)) (display " and "))
			    (display-subgoal-list (cdr new-subgoals))
			    (newline))))))
	      (set-link-active link 'delete)) ;; tautological links can be discarded
	  (unbind-vars)
	  new-subgoals)
	#f)))

(define (unsimplifiable-subgoals subgoals)
  (if (pair? subgoals)
      (if (or (subgoal-simplified (car subgoals)) (match-subgoal-with-index (car subgoals)))
          (unsimplifiable-subgoals (cdr subgoals))
          (cons (car subgoals) (unsimplifiable-subgoals (cdr subgoals))))
      '()))     

(define (compute-current-weights subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-misc (car subgoals) (current-subgoal-weight (car subgoals)))
	(compute-current-weights (cdr subgoals)))))

(define (renamed-copy clause unsimplifieds)
  (let ((copy (new-clause)))
    (trail-stop)
    (if unsimplifieds
	(set-clause-subgoals copy (renamed-subgoals unsimplifieds))
	(set-clause-subgoals copy (renamed-subgoals (unsimplifiable-subgoals (clause-subgoals clause)))))
    (unrename-vars)
    (set-clause-unsimplified copy (length (clause-subgoals copy)))
    (if (null? (clause-subgoals copy))
	(set-clause-subgoals copy (list falsum)))
    (set-clause-variables copy (variables-in-subgoals (clause-subgoals copy) '()))
    (set-clause-nr-variables copy (length (clause-variables copy)))
    (clause-pointers-in-subgoals (clause-subgoals copy) copy)
    (set-clause-input copy (clause-input clause))
    (set-clause-initial copy #f)
    (set-clause-relevant copy (clause-relevant clause))
    (set-clause-predecessor copy clause)
    copy))

(define (variables-in-subgoals subgoals vars)
  (if (null? subgoals)
      vars
      (variables-in-subgoals (cdr subgoals) (free-vars-in-termlist (subgoal-terms (car subgoals)) vars))))
  

(define (free-vars-in-termlist terms vars)
  ;; finds all free variables in terms and inserts them into vars exactly once
  ;;(print vars)
  (if (not (null? terms))
      (begin
	(free-vars-in-termlist (cdr terms) (free-vars-in-term (ref (car terms)) vars)))
      vars))

(define (free-vars-in-term term vars)          
  ;; checks if a term is a variable.
  ;; if yes, it is added to the variable list vars if it is not already in there
  ;; otherwise if it is a proper functor, the argument list of term is searched for
  ;; free variables
  (if (free-variable term)
      (begin
	(find-or-insert-var term vars))
      (if (not (constant term))
	  (free-vars-in-termlist (cdr term) vars)
	  vars)))

(define (find-or-insert-var term vars)
  (if (pair? vars)
      (begin
	(if (not (eq? term (car vars)))
	    (if (pair? (cdr vars))
		(find-or-insert-var term (cdr vars))
		(set-cdr! vars (list term))))
	vars)
      (list term)))
      
(define (clause-pointers-in-subgoals subgoals clause)
  (if (pair? subgoals)
      (begin
	(set-subgoal-clause (car subgoals) clause)
	(clause-pointers-in-subgoals (cdr subgoals) clause))))

(define (subgoal-in-copy subgoal copy-subgoals)
  (if (null? copy-subgoals)
      #f
      (if (eq? (subgoal-number subgoal) (subgoal-number (car copy-subgoals)))
	  (car copy-subgoals)
	  (subgoal-in-copy subgoal (cdr copy-subgoals)))))

(define (indexq obj list)
  (if (pair? list)
      (if (eq? obj (car list))
	  0
	  (let ((value (indexq obj (cdr list))))
	    (if value
		(+ 1 value)
		value)))
      #f))

(define (clause-index-variant? clause)
  ;; Right now does the variant test only for clauses that are descended from the same input clause
  (let ((index-result #f))
    (let (index-end)
      (set! index-end (clause-in-index? clause))
      (if (vector? index-end)
	  (if (= 0 (vector-ref index-end 0))
	      (begin
		(if (> debug-level 5)
			(print "variant not found: " index-end))
		#f)
	      (begin
		(if dewey-numbers
		    (begin
		      (display "delete variant ") 
		      (display-clause clause)
		      (display " of ") 
		      (display-clause (car (vector-ref index-end 1)))
		      (if (> (vector-ref index-end 0) 1)
			  (begin
			    (display " (strange variant number: ")
			    (display (vector-ref index-end 0))
			    (display ")")))
		      (newline)))
		(set! nr-variants (+ nr-variants 1))
		(set! index-result (car (vector-ref index-end 1)))
		index-result))
	  (begin
	    (if (> debug-level 4)
		(begin 
		  (display "index-search failed for ")
		  (display-clause clause) (display " with index ")
		  (newline)
		  (print-index-entries (cdr (input-variant-index (clause-input clause))) "(" '())
		  (newline)))
	    #f)))))

(define (get-path-clause subgoal depth)
  (if subgoal
      (if (= depth (subgoal-depth subgoal))
	  (subgoal-clause subgoal)
	  (if (< depth (subgoal-depth subgoal))
	      (get-path-clause (subgoal-predecessor subgoal) depth)
	      #f))
      #f))

(define (get-top-depth subgoal)
  (if subgoal
      (if (subgoal-predecessor subgoal)
	  (get-top-depth (subgoal-predecessor subgoal))
	  (subgoal-depth subgoal))
      #f))

(define (clause-subsumed clause)
  (if subsumption
      (if (or (and unit-subsume simplify) (non-unit-clause clause))
	  (let ((subsumed (subgoals-subsumed-by-index (clause-subgoals clause))))
	    (if subsumed
		(begin
		  (if (or (> debug-level 4) dewey-numbers)
		      (begin
			(display "Subsumption deletion of ")
			(display-clause clause)
			(display " by unit ")
			(display-clause (subgoal-clause subsumed))
			(newline)))
		  (set! nr-subsumption-deletions (+ nr-subsumption-deletions 1))))
	    subsumed)
	  #f)
      #f))

(define (ground-regular clause)
  ;; Right now does the variant test only for clauses that are descended from the same input clause
  (if (index-regular-subgoals (clause-subgoals clause))
      #t
      (begin
	(set! regularity-fails (+ regularity-fails 1))
	#f)))

(define (index-regular-subgoals subgoals)
  (if (pair? subgoals)
      (if (ground-regular-subgoal? (car subgoals))
          (index-regular-subgoals (cdr subgoals))
          (begin 
            #f))
      #t))

 (define (check-subgoals-regularity subgoals subgoal)
  (if (pair? subgoals)
      (if (check-subgoal-regularity (car subgoals) subgoal)
	  (check-subgoals-regularity (cdr subgoals) subgoal)
	  #f)
      #t))

(define (check-subgoal-regularity a b)
  (if (and (not (subgoal-selected a))
	   (not (eq? a b))
	   (eq? (subgoal-sign a) (subgoal-sign b))
	   (eq? (subgoal-pred a) (subgoal-pred b))
	   (ground-and-equal-termlists (subgoal-terms a) (subgoal-terms b)))
      #f
      #t))

(define (ground-and-equal-termlists k l)
  (if (pair? k)
      (if (ground-and-equal-terms (ref (car k)) (ref (car l)))
	  (ground-and-equal-termlists (cdr k) (cdr l))
	  #f)
      #t))

(define (ground-and-equal-terms s t)
  (if (eq? s t)
      #t
      (if (variable s)
	  #f
	  (if (and (non-variable t) (eq? (car s) (car t)))
	      (ground-and-equal-termlists (cdr s) (cdr t))
	      #f))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Junk functions, various
;; ------------------------------------------------------------

(define (print-predicates-indexes predicates)
  (if (pair? predicates)
      (begin
	(print-indexes (car predicates) 'gr)
	(print-indexes (car predicates) 'unit)
	(print-predicates-indexes (cdr predicates)))))

(define (test-subsumption-for-units clauses)
  (if (pair? clauses)
      (begin
	(display-clause (car clauses))
	(if (unit-clause (car clauses))
	    (display " (unit)")
	    (if (non-unit-clause (car clauses))
		(display " (non-unit)")))
	(newline)
	(clause-subsumed (car clauses))
	(test-subsumption-for-units (cdr clauses)))))

(define (get-path-subgoal subgoal depth)
  (if (and subgoal (>= (subgoal-depth subgoal) depth))
      (if (= (subgoal-depth subgoal) depth)
	  subgoal
	  (get-path-subgoal (subgoal-predecessor subgoal) depth))
      #f))

(define (variables-in-other-subgoals subgoals subgoal vars)
  (if (null? subgoals)
      vars
      (if (or (eq? subgoal (car subgoals)) (subgoal-simplified (car subgoals)))
	  (variables-in-other-subgoals (cdr subgoals) subgoal vars)
	  (variables-in-other-subgoals (cdr subgoals) subgoal (free-vars-in-termlist (subgoal-terms (car subgoals)) vars)))))
  
(define (check-for-local-vars subgoal)
  (let (subgoal-vars other-vars)
    (set! subgoal-vars (free-vars-in-termlist (subgoal-terms subgoal) '()))
    (set! other-vars (variables-in-other-subgoals (clause-subgoals (subgoal-clause subgoal)) subgoal '()))
    (if (and (not (simplified-unit subgoal)) (contains-local-var subgoal-vars other-vars))
	(set! subgoals-with-local-var (+ subgoals-with-local-var 1)))))

(define (contains-local-var subgoal-vars other-vars)
  (if (pair? subgoal-vars)
      (if (memq (car subgoal-vars) other-vars)
	  (contains-local-var (cdr subgoal-vars) other-vars)
	  #t)
      #f))

(define (toggle-selection)
  (set! complexity-measure (not complexity-measure))
  (set! min-sg-select (not min-sg-select)))


(define (free-variables? vars)
  (if (pair? vars)
      (if (free-variable (car vars))
	  (free-variables? (cdr vars))
	  #f)
      #t))

(define (disjoint-renamed? vars)
  (if (pair? vars)
      (if (bound-variable (car vars))
	  (if (free-variable (variable-binding (car vars)))
	      (let ((result (disjoint-renamed? (cdr vars))))
		(if result
		    (if (memq (variable-binding (car vars)) result)
			#f
			(cons (variable-binding (car vars)) result))
		    #f))
	      #f)
	  #f)
      '()))
		 
(define (variant-varlist? vars)
  (if (pair? vars)
      (if (free-variable (ref (car vars)))
	  (let ((result (variant-varlist? (cdr vars))))
	    (if result
		(if (memq (ref (car vars)) result)
		    #f
		    (cons (ref (car vars)) result))
		#f))
	  #f)
      '()))
		 
