;;;*************************************************************************
;; ------------------------
;; GENERAL GLOBAL VARIABLES
;; ------------------------
(module globals
	(include "dctp.sch")
	(import (pruning "pruning.scm") (path "path.scm"))
	(export dewey-numbers trail-root trail filename formula norel max-depth proof-output successful-inferences
		successful-reduction-steps nr-reduction-steps nr-prunings total-inferences initial-length silent
		total-successful-inferences total-links links-tagged links-untagged unit-subsume context-reduction-steps
		max-variables regularity nr-variants foldup foldup-root debug-level link-list full-variants
		min-rel-vec-length bit-vector-length rel-vec-indexing-table regularity-fails pruning links-pushed 
		links-popped tautologies tautology-fails global-index-bindings nr-subsumption-deletions 
		test-mode subsumption input-root bit-mask-table nr-foldups nr-simplifications nr-top-down-lemmas simplify 
		min-sg-select coupled-eliminations backward-subsumptions back-subsume top-down-lemmas couples dewey-start 
		unit-on-top index-info variant-index variant-info nr-context-foldups link-reuses context-lemmas link-vector 
		link-vector-size search-start-index complexity-measure link-info error-on-empty maxint tableau-clauses 
		relevant-terms-ground sg-bias term-bias groundable horn max-term-depth least-initial-links input-term-depth
		allowed-term-depth current-depth limit-term-depth min-relevant-depth current-complexity complexity 
		max-complexity link-from-conjecture total-subgoals-solved subgoals-with-local-var initial-links 
		initial-relevant-links toggle-inferences falsum variant-subsumptions variant-subsumption-tests
		potential-inputs potential-subgoals foreign-variants false-potentials var-counter potential-fails 
		full-fails single-match-fails length-fails signature-fails inference-limit tptp-lemmas start end elapsed)
	(export (get-flags args) (initialize-prover)))

(define debug-level 0)
(define dewey-numbers #f)
(define silent #f)
(define dewey-start #f)
(define trail-root (list (cons #f #f)))
(define trail trail-root)
(define link-vector #f)
(define link-vector-size 5000)
(define search-start-index 0)
(define filename #f)
(define formula (make-vector 3))
(define norel #f)
(define max-depth 0)
(define proof-output #f)
(define inference-limit #f)
(define successful-inferences 0)
(define successful-reduction-steps 0)
(define context-reduction-steps 0)
(define nr-reduction-steps 0)
(define pruning #t)
(define nr-prunings 0)
(define tableau-clauses 0)
(define regularity-fails 0)
(define total-inferences 0)
(define total-successful-inferences 0)
(define total-links 0)
(define links-tagged 0)
(define links-untagged 0)
(define links-pushed 0)
(define links-popped 0)
(define max-variables 0) 
(define regularity #t)
(define variant-index #t)
(define subsumption #t)
(define back-subsume #t)
(define unit-subsume #t)
(define complexity-measure #f)
(define couples #t)
(define tautologies #f)
(define tautology-fails 0)
(define nr-variants 0)
(define variant-subsumptions 0)
(define coupled-eliminations 0)
(define nr-subsumption-deletions 0)
(define backward-subsumptions 0)
(define nr-foldups 0)
(define nr-context-foldups 0)
(define nr-top-down-lemmas 0)
(define nr-simplifications 0)
(define foldup #t)
(define context-lemmas #t)
(define simplify #t)
(define top-down-lemmas #t)
(define tptp-lemmas #f)
(define full-variants #t)
(define foldup-root (make-vector 10 #f))
(define link-list '())
(define link-reuses 0)
(define initial-length 0)
(define min-rel-vec-length 8)
(define bit-vector-length 30)
(define rel-vec-indexing-table (make-vector 1000 #f))
(define bit-mask-table #f)
(define maxint 1)
(define global-index-bindings #f)
(define test-mode #f)
(define input-root #f)
(define min-sg-select #f)
(define unit-on-top #f)
(define least-initial-links #f)
(define index-info #f)
(define variant-info #f)
(define link-info #f)
(define error-on-empty #t)
(define relevant-terms-ground #t)
(define groundable #f)
(define horn #f)
(define sg-bias #f)
(define term-bias #f)
(define max-term-depth 0)
(define current-depth 0)
(define input-term-depth 0)
(define allowed-term-depth 0)
(define limit-term-depth #f)
(define min-relevant-depth #f)
(define max-complexity 0)
(define current-complexity 0)
(define complexity 0)
(define link-from-conjecture #t)
(define total-subgoals-solved 0)
(define subgoals-with-local-var 0)
(define initial-links 0)
(define	initial-relevant-links 0)
(define toggle-inferences #f)
(define falsum #f)
(define var-counter 1)
(define variant-subsumption-tests 0)
(define potential-inputs 0)
(define potential-subgoals 0)
(define foreign-variants 0)
(define false-potentials 0)

(define potential-fails 0)
(define full-fails 0)
(define single-match-fails 0)
(define length-fails 0)
(define signature-fails 0)

(define start 0)
(define end 0)
(define elapsed #f)

;;;*************************************************************************
;; ---------------------------
;; UPDATE OF DEFAULTS SETTINGS
;; ---------------------------

(define (get-flags args)
  (if (and (not (null? args))
	   (string? (car args))
	   (char=? (string-ref (car args) 0) #\-))
      (cond
       ((string=? (car args) "-debug") (set! debug-level (string->integer (cadr args))) (get-flags (cddr args)))
       ((string=? (car args) "-dewey") (set! dewey-numbers #t) (get-flags (cdr args)))
       ((string=? (car args) "-silent") (if (not dewey-numbers) (set! silent #t)) (get-flags (cdr args)))
       ((string=? (car args) "-dewey-start") 
	(set! dewey-numbers #f) 
	(set! silent #f)
	(set! dewey-start (string->integer (cadr args))) (get-flags (cddr args)))
       ((string=? (car args) "-noreg") (set! regularity #f) (get-flags (cdr args)))
       ((string=? (car args) "-norel") (set! norel #t) (get-flags (cdr args)))
       ((string=? (car args) "-noprune") (set! pruning #f) (get-flags (cdr args)))
       ((string=? (car args) "-nosubsume") (set! subsumption #f) (get-flags (cdr args)))
       ((string=? (car args) "-noback") (set! back-subsume #f) (get-flags (cdr args)))
       ((string=? (car args) "-nounitsubsume") (set! unit-subsume #f) (get-flags (cdr args)))
       ((string=? (car args) "-nosimplify") (set! simplify #f) (get-flags (cdr args)))
       ((string=? (car args) "-nofoldup") (set! foldup #f) (get-flags (cdr args)))
       ((string=? (car args) "-nocontext") (set! context-lemmas #f) (get-flags (cdr args)))
       ((string=? (car args) "-nolemmas") (set! top-down-lemmas #f) (get-flags (cdr args)))
       ((string=? (car args) "-novariants") (set! full-variants #f) (get-flags (cdr args)))
       ((string=? (car args) "-nocoupled") (set! couples #f) (get-flags (cdr args)))
       ((string=? (car args) "-taut") (set! tautologies #t) (get-flags (cdr args)))
       ((string=? (car args) "-minselect") (set! min-sg-select #t) (get-flags (cdr args)))
       ((string=? (car args) "-maxselect") (set! min-sg-select #f) (get-flags (cdr args)))
       ((string=? (car args) "-topunits") (set! unit-on-top #t) (get-flags (cdr args)))
       ((string=? (car args) "-sparse") (set! least-initial-links #t) (get-flags (cdr args)))
       ((string=? (car args) "-instantiated") (set! complexity-measure #f) (get-flags (cdr args)))
       ((string=? (car args) "-complexity") (set! complexity-measure #t) (get-flags (cdr args)))
       ((string=? (car args) "-linkup") (set! link-from-conjecture #f) (get-flags (cdr args)))
       ((string=? (car args) "-termlimit") (set! limit-term-depth #t) (get-flags (cdr args)))
       ((string=? (car args) "-toggle") (set! toggle-inferences #t) (get-flags (cdr args)))
       ((string=? (car args) "-termbias") (set! term-bias (string->integer (cadr args))) (get-flags (cddr args)))
       ((string=? (car args) "-sgbias") (set! sg-bias (string->integer (cadr args))) (get-flags (cddr args)))
       ((string=? (car args) "-inference-limit") (set! inference-limit (string->integer (cadr args))) (get-flags (cddr args)))
       ((string=? (car args) "-tptp-lemmas") (set! tptp-lemmas #t) (get-flags (cdr args)))
       ((string=? (car args) "-indexinfo") (set! index-info (string->integer (cadr args))) (get-flags (cddr args)))
       ((string=? (car args) "-linkinfo") (set! link-info (string->integer (cadr args))) (get-flags (cddr args)))
       ((string=? (car args) "-variantinfo") (set! variant-info (string->integer (cadr args))) (get-flags (cddr args)))
       ((string=? (car args) "-test") (set! test-mode #t) (get-flags (cdr args)))
       ;;((string=? (car args) "-switchbias") (set! switch-level (string->integer (cadr args))) (get-flags (cddr args)))
       (else (print "Abort: error in parameter list") (exit 1)))
      args))

;;;*************************************************************************
;; ---------------------
;; PROVER INITIALIZATION
;; ---------------------
 
(define (initialize-prover)
  (do ((i 0 (+ i 1)))
      ((= i (- bit-vector-length 1)) #t)
    (set! maxint (bit-or maxint (bit-lsh 1 i))))
  (set! link-vector (make-vector link-vector-size '()))
  (set! search-start-index (- link-vector-size 1))
  ;;(set! allowed-term-depth input-term-depth)
  (set! min-relevant-depth maxint)
  (set! allowed-term-depth maxint)
  (initialize-table rel-vec-indexing-table 0)
  (initialize-bit-mask-table)
  (initialize-path-literal-vectors)
  (let ((bot (new-predsym)))
    (set-predsym-arity bot 0)
    (set-predsym-number bot -1)
    (set-predsym-weight bot 0)
    (set-predsym-symbol bot (string->symbol "false"))
    (set! falsum (new-subgoal))
    (set-subgoal-selected falsum #f)
    (set-subgoal-simplified falsum #f)
    (set-subgoal-pred falsum bot)
    (set-subgoal-terms falsum '())
    (set-subgoal-sign falsum '+)))

;;;*************************************************************************
;; -------------
;; FOR DEBUGGING
;; -------------

(define l1 #f)
(define l2 #f)
(define cl #f)
(define d #f)
(define i 0)
