(module index
	(include "dctp.sch")
	(export (insert-subgoal-in-index subgoal index-type) (delete-subgoal-from-index subgoal index-type) 
		(insert-subgoal-in-gr-index subgoal) (delete-subgoal-from-gr-index subgoal)
		(ground-regular-subgoal? subgoal) (subgoals-in-index? subgoals index) 
		(ground-closed-with-index subgoal) (ground-termlist k) (print-indexes predsym index-type) 
		(insert-clause-in-index clause) (delete-clause-from-index clause) (clause-in-index? clause)
		(delete-foldup-from-gr-index subgoal) (insert-foldup-in-gr-index subgoal)
		(print-index-entries index p paren-stack) (subgoals-subsumed-by-index subgoals)
		(match-trail-with-index index) (match-subgoal-with-index subgoal) (init-index-constants)
		(predicate-index-statistics predicates) (variant-index-statistics clauses) (clear-term-trail) 
		(ground-irregular-subgoal? subgoal) (subgoal-in-gr-index subgoal) (subgoal-subsumed-by-index subgoal))
	(import (globals "globals.scm") (display "display.scm") (unification "unification.scm") 
		(ref-equal pruning "pruning.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define local-index-vars 0)

(define term-trail-root (list (cons #f #f)))
(define term-trail term-trail-root)

(define pos-sign-constant #f)
(define neg-sign-constant #f)
(define clause-end #f)

(define nr-index-paths 0)
(define nr-index-entries 0)
(define nr-empty-paths 0)
(define total-depth 0)
(define total-empty-depth 0)
(define total-branching 0)
(define nr-branching 0)
(define max-branching 0)

;;;*************************************************************************
;; ------------------------------------------------------------
;; Index constant init function
;; ------------------------------------------------------------

(define (init-index-constants)
  (set! clause-end (new-functor))
  (set-functor-arity clause-end 0)
  (set-functor-symbol clause-end 'clause-end)
  (set-functor-number clause-end -3)
  (set! clause-end (list (list clause-end)))
  (set! pos-sign-constant (new-functor))
  (set-functor-arity pos-sign-constant 0)
  (set-functor-symbol pos-sign-constant '+)
  (set-functor-number pos-sign-constant -1)
  (set! pos-sign-constant (list (list pos-sign-constant)))
  (set! neg-sign-constant (new-functor))
  (set-functor-arity neg-sign-constant 0)
  (set-functor-symbol neg-sign-constant '-)
  (set-functor-number neg-sign-constant -2)
  (set! neg-sign-constant (list (list neg-sign-constant))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Term trail functions
;; ------------------------------------------------------------

(define (term-trail-erase)
 (if (term-trail-content)
     (begin (term-trail-previous) (term-trail-erase))
     (begin (term-trail-previous) #f)))


(define (clear-term-trail)
  (do ((i 0 (+ i 1)))
      ((eq? term-trail term-trail-root) (print "required " i " iterations to clear term trail"))
    (term-trail-erase)))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Index search functions
;; ------------------------------------------------------------

(define (clause-in-index? clause)
  (let (index-end)
    (trail-stop)
    (set! local-index-vars 0)
    (set! index-end (subgoals-in-index? (clause-subgoals clause) (cdr (input-variant-index (clause-input clause)))))
    (unbind-vars)
    index-end))

(define (subgoals-in-index? subgoals index)
  (if (pair? subgoals)
      (let (sign-index)
	(if (eq? '+ (subgoal-sign (car subgoals)))
	    (set! sign-index (term-in-index? (car pos-sign-constant) index))
	    (set! sign-index (term-in-index? (car neg-sign-constant) index)))
	(if sign-index
	    (let ((new-index (term-in-index? (cons (subgoal-pred (car subgoals)) 
						   (subgoal-terms (car subgoals))) sign-index)))
	      (if new-index
		  (subgoals-in-index? (cdr subgoals) new-index)
		  new-index))
	    sign-index))
      (term-in-index? (car clause-end) index)))

;;; This function seems to remain unchanged with ordering...
(define (termlist-in-index? terms index)
  (if (pair? terms)
      (let ((new-index (term-in-index? (ref (car terms)) index)))
	(if new-index
	    (termlist-in-index? (cdr terms) new-index)
	    (begin 
	      (if (> debug-level 8)
		  (print "subsearch failed for term " (ref (car terms))))
	      new-index)))
      index))

(define (term-in-index? term index)
  (if (pair? index)
      (if (index-variable term)
	  (if (index-variable (and-content index)) 
	      (if (= term (and-content index)) ;; So we found the proper variable
		  (and-next index)             ;; Search on one level down
		  (if (> term (and-content index)) ;; Found variable bigger than searched variable?
		      (term-in-index? term (or-next index)) ;; We can go on searching
		      #f))                                  ;; Otherwise abort
	      #f)                                           ;; Type mismatch. Abort (Variables are stored in the front).
	  (if (variable term)
	      (begin
		(set! local-index-vars (+ local-index-vars 1)) ;; A free variable is simply bound to the next unused
		(bind term local-index-vars)                   ;; index variable
		(term-in-index? (ref term) index))             ;; The search is repeated for that binding
	      (if (eq? (car term) (and-content index)) ;; So we found the proper functor
		  (if (constant term)                  ;; Is it a constant?
		      (and-next index)                 ;; Yes. Search on one level down
		      (termlist-in-index? (cdr term) (and-next index))) ;; No, check the argument terms
		  (if (or (index-variable (and-content index)) ;; Functors are stored after vars, go on
			  (< (functor-number (car term)) (functor-number (and-content index)))) ;; Found functor smaller than
		                                                                                ;; searched functor?
		      (term-in-index? term (or-next index))
		      #f))))
      #f)) ;; most likely we hit the end of the or-cell-list


;;;*************************************************************************
;; ------------------------------------------------------------
;; Index insertion functions
;; ------------------------------------------------------------

(define (insert-object-in-index pred-index object)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; more terms to insert?
	(begin  ;; it is important here that empty lists are never pushed on the term-trail!
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (if (> debug-level 8)
	      (if (variable (ref (car current-terms)))
		  (print "index inserting variable " (ref (car current-terms)))
		  (if (index-variable (ref (car current-terms)))
		      (print "index inserting index-v  " (ref (car current-terms)))
		      (print "index inserting functor  " (functor->list (ref (car current-terms)))))))
	  (insert-term-in-index (ref (car current-terms)) pred-index object));; insert the next term
	(if (vector? (or-next pred-index)) ;; if this is not the end of a new path
	    (if (memq object (vector-ref (or-next pred-index) 1)) ;; if object is already stored here
		#f                                            ;; it cant be stored a second time
		(begin ;; otherwise append it to the stored objects of the path
		  (vector-set! (or-next pred-index) 0 (+ (vector-ref (or-next pred-index) 0) 1))
		  (vector-set! (or-next pred-index) 1 (cons object (vector-ref (or-next pred-index) 1)))
		  object))
	    (if (null? (or-next pred-index))
		(begin  ;; make a new storage vector and place object in it
		  (set-or-next pred-index (make-vector 2))
		  (vector-set! (or-next pred-index) 0 1)
		  (vector-set! (or-next pred-index) 1 (list object))
		  object)
		#f))))) ;; apparently our terms ended somewhere in the middle of a path, abort

(define (functor->list term)
  (cons (vector-ref (car term) 0)
	(cons (vector-ref (car term) 1)
	      (cons (vector-ref (car term) 2)
		    (cons (vector-ref (car term) 3) '())))))

;; this function has been taken out of insert-object-in-index because the latter will always pop new terms
;; off the trail which would be incompatible with shallow backtracking within the or-cell-list
(define (insert-term-in-index term pred-index object)
  (if (pair? (or-next pred-index))                                  ;; have we reached the end of the or-cell-list?
      (if (index-variable term) 
	  (if (index-variable (and-content (or-next pred-index)))
	      (if (= term (and-content (or-next pred-index)))       ;; have we found the proper variable?
		  (if (insert-object-in-index (or-current (or-next pred-index)) object) ;; try to insert the rest of the
		      object
		      #f)
		  (if (> term (and-content (or-next pred-index)))
		      (insert-term-in-index term (or-next pred-index) object) ;; shallow backtracking over the index vars
		      (insert-new-cell-and-continue term pred-index object))) ;; the searched var is not there, must insert
	      (insert-new-cell-and-continue term pred-index object))          ;; same here, we already reached the functors
	  (if (variable term)
	      (begin
		(set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
		(bind term local-index-vars)                          ;; index variable
		(insert-term-in-index (ref term) pred-index object))  ;; shallow backtracking with the binding
	      (if (eq? (car term) (and-content (or-next pred-index))) ;; found the functor
		  (begin
		    (if (> (functor-arity (car term)) 0)
			(push-term-trail (cdr term)))      ;; push the args for non-constants
		    (if (insert-object-in-index (or-current (or-next pred-index)) object) ;; try to insert the rest of the
			object
			#f))
		  (if (or (index-variable (and-content (or-next pred-index))) ;; Functors are stored after vars, go on
			  (< (functor-number (car term)) (functor-number (and-content (or-next pred-index))))) 
		                                                             ;;; Found functor smaller than searched functor?
		      (insert-term-in-index term (or-next pred-index) object)   ;; shallow backtracking over the functors
		      (insert-new-cell-and-continue term pred-index object))))) ;; the searched functor is not there, insert
      (insert-new-cell-and-continue term pred-index object))) ;; insert at the end of the or-cell-list

;; this function is called whenever a term could not be found in the index
;; and new memory has to be allocated
(define (insert-new-cell-and-continue term pred-index object)
  (let ((new-cell (new-or-and-cell)))
    (set-or-next new-cell (or-next pred-index)) ;; link the new cell betwenn pred-index and pred-index's
    (set-or-next pred-index new-cell)           ;; old successor (or simply after pred-index)
    (if (index-variable term)
	(set-and-content new-cell term)                             ;; set the and-content of the new cell to the variable
	(if (variable term)
	    (begin
	      (set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
	      (bind term local-index-vars)                          ;; index variable
	      (set! term (ref term))
	      (set-and-content new-cell local-index-vars))
	    (set-and-content new-cell (car term))))  ;; or set the and-content of the new cell to the functor symbol
    (if (and (not (index-variable term)) (not (constant term)))  ;; if term is a complex function, get the args on the trail
	    (push-term-trail (cdr term)))
    (if (insert-object-in-index (or-current new-cell) object) ;; continue the insertion of the trailed terms
	object
	#f)))


(define (insert-subgoal-in-index subgoal index-type)
  (let (index index-end)
    (if (eq? index-type 'path)
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-path-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-path-index (subgoal-pred subgoal))))
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-unit-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-unit-index (subgoal-pred subgoal)))))
    (if (> debug-level 4)
	(begin
	  (display "inserting subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (and  (> debug-level 6) (> (car index) 0))
	      (begin
		(display "current index content: ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (insert-object-in-index index subgoal))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	(set-car! index (+ (car index) 1))
	(begin
	  (display "subgoal ")
	  (display-subgoal subgoal)
	  (display " already in index or bad subgoal -- insertion of subgoal failed!")
	  (newline)
	  #f))))

(define (insert-clause-in-index clause)
  (let ((index (input-variant-index (clause-input clause))) index-end)
    (if (> debug-level 4)
	(begin
	  (display "inserting ")
	  (display-clause clause)
	  (newline)))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-subgoal-termlists (clause-subgoals clause))
    (set! index-end (insert-object-in-index index clause))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	(set-car! index (+ (car index) 1))
	(begin
	  (display-clause clause)
	  (display " already in index or bad clause -- insertion of clause failed!")
	  (newline)
	  #f))))

(define (push-subgoal-termlists subgoals)
  (if (pair? subgoals)
      (begin
	(push-subgoal-termlists (cdr subgoals))
	(push-term-trail (list (cons (subgoal-pred (car subgoals)) (subgoal-terms (car subgoals)))))
	(if (eq? '+ (subgoal-sign (car subgoals)))
	    (push-term-trail pos-sign-constant)
	    (push-term-trail neg-sign-constant)))
      (begin 
	(if (> debug-level 8)
	    (print "pushing " clause-end))
	(push-term-trail clause-end))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Index deletion functions
;; ------------------------------------------------------------

(define (delete-object-from-index pred-index object)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; more terms to delete?
	(begin 
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (if (> debug-level 8)
	      (if (variable (ref (car current-terms)))
		  (print "index deleting variable " (ref (car current-terms)))
		  (if (index-variable (ref (car current-terms)))
		      (print "index deleting index-v  " (ref (car current-terms)))
		      (print "index deleting functor  " (functor->list (ref (car current-terms)))))))
	  (delete-term-from-index (ref (car current-terms)) pred-index object)) ;; delete the next term
	(if (and (vector? (or-next pred-index)) (> (vector-ref (or-next pred-index) 0) 0)) ;; if this is a proper path end
	    (let (new-objects)
	      (set! new-objects (remq! object (vector-ref (or-next pred-index) 1)))
	      (if new-objects           ;; if object could be deleted
		  (if (pair? new-objects)
		      (begin                 ;; adjust object vector
			(vector-set! (or-next pred-index) 0 (- (vector-ref (or-next pred-index) 0) 1))
			(vector-set! (or-next pred-index) 1 new-objects)
			object)
		      (set-or-next pred-index '())) ;; if we deleted the last of the path's objects, kill the path
		  #f))
	    #f))))  ;; our terms ended not at a proper path end, abort

(define (delete-term-from-index term pred-index object)
  (if (pair? (or-next pred-index))                                  ;; have we reached the end of the or-cell-list?
      (if (index-variable term)                                     
	  (if (index-variable (and-content (or-next pred-index)))
	      (if (= term (and-content (or-next pred-index)))       ;; have we found the proper variable?
		  (if (delete-object-from-index (or-current (or-next pred-index)) object) ;; try to delete the rest of the
		      (begin                                                            ;; trailed terms
			(if (null? (and-next (or-next pred-index))) ;; if all lower and-nodes were deleted, delete
			    (set-or-next pred-index (or-next (or-next pred-index)))) ;; this and-node as well
			object)
		      #f)
		  (if (> term (and-content (or-next pred-index)))
		      (delete-term-from-index term (or-next pred-index) object) ;; shallow backtracking over the index vars
		      #f)) ;; the searched var is not there, abort
	      #f)          ;; same here, we already reached the functors
	  (if (variable term)
	      (begin
		(set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
		(bind term local-index-vars)                          ;; index variable
		(delete-term-from-index (ref term) pred-index object))  ;; shallow backtracking with the binding
	      (if (eq? (car term) (and-content (or-next pred-index)))  ;; found the functor
		  (begin
		    (if (> (functor-arity (car term)) 0)
			(push-term-trail (cdr term)))      ;; push the args for non-constants
		    (if (delete-object-from-index (or-current (or-next pred-index)) object) ;; try to delete the rest of the
			(begin                                                            ;; trailed terms
			  (if (null? (and-next (or-next pred-index))) ;; if all lower and-nodes were deleted, delete
			      (set-or-next pred-index (or-next (or-next pred-index)))) ;; this and-node as well
			  object)
			#f))
		  (if (or (index-variable (and-content (or-next pred-index))) ;; Functors are stored after vars, go on
			  (< (functor-number (car term)) (functor-number (and-content (or-next pred-index))))) 
		                                                             ;;; Found functor smaller than searched functor?
		      (delete-term-from-index term (or-next pred-index) object)   ;; shallow backtracking over the functors
		      #f)))) ;; the searched functor is not there, abort
      #f)) ;; term not found in or-cell-list, abort

(define (delete-subgoal-from-index subgoal index-type)
  (let (index index-end)
    (if (eq? index-type 'path)
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-path-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-path-index (subgoal-pred subgoal))))
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-unit-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-unit-index (subgoal-pred subgoal)))))
    (if (> debug-level 4)
	(begin
	  (display "deleting subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (delete-object-from-index index subgoal))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	(set-car! index (- (car index) 1))
	(begin
	  (display "subgoal ")
	  (display-subgoal subgoal)
	  (display " not found in index -- deletion of subgoal failed!")
	  (newline)
	  #f))))

(define (delete-clause-from-index clause)
  (let (index index-end new-clauses)
    (if (> debug-level 3)
	(begin
	  (display "deleting ")
	  (display-clause clause)
	  (newline)))
    (set! index (input-variant-index (clause-input clause)))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-subgoal-termlists (clause-subgoals clause))
    (set! index-end (delete-object-from-index index clause))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	(set-car! index (- (car index) 1))
	(begin
	  (display-clause clause)
 	  (display " not found in index -- deletion of clause failed!")
	  (newline)
	  #f))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Index ground regularity functions
;; ------------------------------------------------------------

(define (ground-regular-subgoal? subgoal)
  (if (and (not (subgoal-simplified subgoal)) (ground-termlist (subgoal-terms subgoal)))
      (let (index index-end)
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
	(set! index-end (termlist-in-index? (subgoal-terms subgoal) (cdr index)))
	(if (and (vector? index-end) (> (vector-ref index-end 0) 0)         ;; no regularity fails with
		 (not (subgoal-unlinkable (car (vector-ref index-end 1))))) ;; back-subsumed literals!
	    (begin
	      (if dewey-numbers
		  (begin
		    (display "regularity check failed with ") 
		    (display-subgoal (car (vector-ref index-end 1)))
		    (display " of depth ") 
		    (display (subgoal-depth (car (vector-ref index-end 1))))
		    (newline)))
	      #f)
	    #t))
      #t))

(define (ground-irregular-subgoal? subgoal)
  (if (and (not (subgoal-simplified subgoal)) (ground-termlist (subgoal-terms subgoal)))
      (let (index index-end)
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
	(set! index-end (termlist-in-index? (subgoal-terms subgoal) (cdr index)))
	(if (and (vector? index-end) (> (vector-ref index-end 0) 0)         ;; no regularity fails with
		 (not (subgoal-unlinkable (car (vector-ref index-end 1))))) ;; back-subsumed literals!
	    (begin
	      (if dewey-numbers
		  (begin
		    (display "regularity check failed with ") 
		    (display-subgoal (car (vector-ref index-end 1)))
		    (display " of depth ") 
		    (display (subgoal-depth (car (vector-ref index-end 1))))
		    (newline)))
	      (car (vector-ref index-end 1)))
	    #f))
      #f))

(define (ground-termlist k)
  (if (pair? k)
      (if (ground-term (ref (car k)))
	  (ground-termlist (cdr k))
	  #f)
      #t))

(define (ground-term s)
  (if (variable s)
      #f
      (ground-termlist (cdr s))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; GR-index functions
;; ------------------------------------------------------------

(define (ground-closed-with-index subgoal)
  (let (index index-end)
    (if (> debug-level 4)
	(begin
	  (display "gr-searching for subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (if (eq? '- (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (set! index-end (termlist-in-gr-index? (subgoal-terms subgoal) (cdr index)))
    (if (or (not (vector? index-end)) (= (vector-ref index-end 0) 0))
	(begin
	  (if (eq? '- (subgoal-sign subgoal))
	      (set! index (predsym-pos-path-index (subgoal-pred subgoal)))
	      (set! index (predsym-neg-path-index (subgoal-pred subgoal))))
	  (set! index-end (termlist-in-gr-index? (subgoal-terms subgoal) (cdr index)))
	  (if (or (not (vector? index-end)) (= (vector-ref index-end 0) 0))
	      (set! index-end #f))))
    index-end))

(define (termlist-in-gr-index? terms index)
  (if (pair? terms)
      (let ((new-index (term-in-gr-index? (ref (car terms)) index)))
	(if new-index
	    (termlist-in-gr-index? (cdr terms) new-index)
	    (begin 
	      (if (> debug-level 8)
		  (print "subsearch failed for term " (ref (car terms))))
	      new-index)))
      index))

(define (term-in-gr-index? term index)
  (if (pair? index)
      (if (variable term)
	  (if (index-variable (and-content index)) ;; So we found the one variable
	      (and-next index)             ;; Search on one level down
	      #f)                          ;; Type mismatch. Abort (Variables are stored in the front).
	  (if (eq? (car term) (and-content index)) ;; So we found the proper functor
	      (if (constant term)              ;; Yes. Is it a constant?
		  (and-next index)             ;; Yes. Search on one level down
		  (termlist-in-gr-index? (cdr term) (and-next index))) ;; No, check the argument terms
	      (if (or (index-variable (and-content index)) ;; Functors are stored after vars, go on
		      (< (functor-number (car term)) (functor-number (and-content index)))) ;; Found functor smaller than
		                                                                            ;; searched functor?
		  (term-in-gr-index? term (or-next index))
		  #f)))
      #f)) ;; we hit the end of the or-cell-list

(define (insert-object-in-gr-index pred-index object)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; more terms to insert?
	(begin 
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (insert-term-in-gr-index (ref (car current-terms)) pred-index object)) ;; insert the next term
	(if (vector? (or-next pred-index)) ;; if this is not the end of a new path
	    (begin
	      (if (memq object (vector-ref (or-next pred-index) 1)) ;; if object is already stored here
		  #f                                            ;; it cant be stored a second time
		  (begin ;; otherwise append it to the stored objects of the path
		    (vector-set! (or-next pred-index) 0 (+ (vector-ref (or-next pred-index) 0) 1))
		    (vector-set! (or-next pred-index) 1 (cons object (vector-ref (or-next pred-index) 1)))
		    object)))
	    (if (null? (or-next pred-index))
		(begin  ;; make a new storage vector and place object in it
		  (set-or-next pred-index (make-vector 2))
		  (vector-set! (or-next pred-index) 0 1)
		  (vector-set! (or-next pred-index) 1 (list object))
		  object)
		#f))))) ;; apparently our terms ended somewhere in the middle of a path, abort

;; this function has been taken out of insert-object-in-gr-index because the latter will always pop new terms
;; off the trail which would be incompatible with shallow backtracking within the or-cell-list
(define (insert-term-in-gr-index term pred-index object)
  (if (pair? (or-next pred-index))                                  ;; have we reached the end of the or-cell-list?
      (if (variable term)                                     
	  (if (index-variable (and-content (or-next pred-index)))
	      (if (insert-object-in-gr-index (or-current (or-next pred-index)) object) ;; try to insert the rest of the
		  object
		  #f)
	      (insert-new-gr-cell-and-continue term pred-index object))          ;; same here, we already reached the functors
	  (if (eq? (car term) (and-content (or-next pred-index))) ;; found the functor
	      (begin
		(if (> (functor-arity (car term)) 0)
		    (push-term-trail (cdr term)))      ;; push the args for non-constants
		(if (insert-object-in-gr-index (or-current (or-next pred-index)) object) ;; try to insert the rest of the
		    object
		    #f))
	      (if (or (index-variable (and-content (or-next pred-index))) ;; Functors are stored after vars, go on
		      (< (functor-number (car term)) (functor-number (and-content (or-next pred-index))))) 
		                                                             ;;; Found functor smaller than searched functor?
		  (insert-term-in-gr-index term (or-next pred-index) object)   ;; shallow backtracking over the functors
		  (insert-new-gr-cell-and-continue term pred-index object)))) ;; the searched functor is not there, insert
      (insert-new-gr-cell-and-continue term pred-index object))) ;; insert at the end of the or-cell-list

;; this function is called whenever a term could not be found in the index
;; and new memory has to be allocated
(define (insert-new-gr-cell-and-continue term pred-index object)
  (let ((new-cell (new-or-and-cell)))
    (set-or-next new-cell (or-next pred-index)) ;; link the new cell betwenn pred-index and pred-index's
    (set-or-next pred-index new-cell)           ;; old successor (or simply after pred-index)
    (if (variable term)
	(set-and-content new-cell 1)            ;; set the and-content of the new cell to the one variable
	(set-and-content new-cell (car term)))  ;; or the functor symbol
    (if (and (not (variable term)) (not (constant term)))  ;; if term is a complex function, get the args on the trail
	(push-term-trail (cdr term)))
    (if (insert-object-in-gr-index (or-current new-cell) object) ;; continue the insertion of the trailed terms
	object
	#f)))

(define (insert-subgoal-in-gr-index subgoal)
  (let (index index-end)
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (if (> debug-level 4)
	(begin
	  (display "gr-inserting subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (and  (> debug-level 6) (> (car index) 0))
	      (begin
		(display "current index content: ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (insert-object-in-gr-index index subgoal))
    (term-trail-erase)
    (if index-end
	(set-car! index (+ (car index) 1))
	(begin
	  (display "subgoal already in gr-index or bad subgoal -- insertion of subgoal ")
	  (display-subgoal subgoal)
	  (display " failed!")
	  (newline)
	  #f))))

(define (delete-subgoal-from-gr-index subgoal)
  (let (index index-end)
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (if (> debug-level 4)
	(begin
	  (display "gr-deleting subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (delete-object-from-gr-index index subgoal))
    (term-trail-erase)
    (if index-end
	(set-car! index (- (car index) 1))
	(begin
	  (print "subgoal not found in gr-index -- deletion of subgoal ")
	  (display-subgoal subgoal)
	  (display " failed!")
	  (newline)
	  #f))))

(define (insert-foldup-in-gr-index subgoal)
  (let (index index-end)
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (if (> debug-level 4)
	(begin
	  (display "gr-inserting context lemma ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (and  (> debug-level 6) (> (car index) 0))
	      (begin
		(display "current index content: ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (insert-object-in-gr-index index subgoal))
    (term-trail-erase)
    (if index-end
	(set-car! index (+ (car index) 1))
	(begin
	  (print "context lemma in gr-index or bad context lemma -- insertion of context lemma failed!")
	  #f))))

(define (delete-foldup-from-gr-index subgoal)
  (let (index index-end)
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (if (> debug-level 4)
	(begin
	  (display "gr-deleting context lemma ")
	  (display-subgoal subgoal)
	  (newline)))
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (delete-object-from-gr-index index subgoal))
    (term-trail-erase)
    (if index-end
	(set-car! index (- (car index) 1))
	(begin
	  (print "subgoal not found in gr-index -- deletion of subgoal failed!")
	  #f))))

(define (delete-object-from-gr-index pred-index object)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; more terms to delete?
	(begin 
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (delete-term-from-gr-index (ref (car current-terms)) pred-index object)) ;; delete the next term
	(if (and (vector? (or-next pred-index)) (> (vector-ref (or-next pred-index) 0) 0)) ;; if this is a proper path end
	    (let (new-objects)
	      (set! new-objects (remq! object (vector-ref (or-next pred-index) 1)))
	      (if new-objects           ;; if object could be deleted
		  (begin                 ;; adjust object vector
		    (vector-set! (or-next pred-index) 0 (- (vector-ref (or-next pred-index) 0) 1))
		    (vector-set! (or-next pred-index) 1 new-objects)
		    object)
		  #f))
	    #f))))  ;; our terms ended not at a proper path end, abort

(define (delete-term-from-gr-index term pred-index object)
  (if (pair? (or-next pred-index))                                  ;; have we reached the end of the or-cell-list?
      (if (variable term)                                     
	  (if (index-variable (and-content (or-next pred-index)))   ;; have we found the one variable?
	      (if (delete-object-from-gr-index (or-current (or-next pred-index)) object) ;; try to delete the rest of the
		  (begin                                                                 ;; trailed terms
		    (if (null? (and-next (or-next pred-index))) ;; if all lower and-nodes were deleted, delete
			(set-or-next pred-index (or-next (or-next pred-index)))) ;; this and-node as well
		    object)
		  #f)
	      #f)          ;; same here, we already reached the functors
	  (if (eq? (car term) (and-content (or-next pred-index)))  ;; found the functor
	      (begin
		(if (> (functor-arity (car term)) 0)
		    (push-term-trail (cdr term)))      ;; push the args for non-constants
		(if (delete-object-from-gr-index (or-current (or-next pred-index)) object) ;; try to delete the rest of the
		    (begin                                                                 ;; trailed terms
		      (if (null? (and-next (or-next pred-index))) ;; if all lower and-nodes were deleted, delete
			  (set-or-next pred-index (or-next (or-next pred-index)))) ;; this and-node as well
		      object)
		    #f))
	      (if (or (index-variable (and-content (or-next pred-index))) ;; Functors are stored after vars, go on
		      (< (functor-number (car term)) (functor-number (and-content (or-next pred-index))))) 
		                                                             ;;; Found functor smaller than searched functor?
		  (delete-term-from-gr-index term (or-next pred-index) object)   ;; shallow backtracking over the functors
		  #f))) ;; the searched functor is not there, abort
      #f)) ;; term not found in or-cell-list, abort

;;;*************************************************************************
;; ------------------------------------------------------------
;; index print functions
;; ------------------------------------------------------------

(define (print-indexes predsym index-type)
  (if (or (and (eq? index-type 'path) 
	       (or (> (car (predsym-pos-path-index predsym)) 0) (> (car (predsym-neg-path-index predsym)) 0)))
	  (and (eq? index-type 'gr) 
	       (or (> (car (predsym-pos-gr-index predsym)) 0) (> (car (predsym-neg-gr-index predsym)) 0)))
	  (and (eq? index-type 'unit) 
	       (or (> (car (predsym-pos-unit-index predsym)) 0) (> (car (predsym-neg-unit-index predsym)) 0))))
      (begin
	(if (eq? index-type 'path)
	    (display "path ")
	    (if (eq? index-type 'gr)
		(display "gr   ")
		(display "unit ")))
	(display "index entries for predsym ")
	(display (predsym-symbol predsym))
	(display " ( ")
	(if (eq? index-type 'path)
	    (display (car (predsym-pos-path-index predsym)))
	    (if (eq? index-type 'gr)
		(display (car (predsym-pos-gr-index predsym)))
		(display (car (predsym-pos-unit-index predsym)))))
	(display " pos, ")
	(if (eq? index-type 'path)
	    (display (car (predsym-neg-path-index predsym)))
	    (if (eq? index-type 'gr)
		(display (car (predsym-neg-gr-index predsym)))
		(display (car (predsym-neg-unit-index predsym)))))
	(display " neg ):")
	(newline)
	(if (eq? index-type 'path)
	    (begin
	      (print-index-entries (cdr (predsym-pos-path-index predsym)) 
				   (string-append "+" (symbol->string (predsym-symbol predsym)) "(") '())
	      (print-index-entries (cdr (predsym-neg-path-index predsym)) 
				   (string-append "-" (symbol->string (predsym-symbol predsym)) "(") '()))
	    (if (eq? index-type 'gr)
		(begin
		  (print-index-entries (cdr (predsym-pos-gr-index predsym)) 
				       (string-append "+" (symbol->string (predsym-symbol predsym)) "(") '())
		  (print-index-entries (cdr (predsym-neg-gr-index predsym)) 
				       (string-append "-" (symbol->string (predsym-symbol predsym)) "(") '()))
		(begin
		  (print-index-entries (cdr (predsym-pos-unit-index predsym)) 
				       (string-append "+" (symbol->string (predsym-symbol predsym)) "(") '())
		  (print-index-entries (cdr (predsym-neg-unit-index predsym)) 
				       (string-append "-" (symbol->string (predsym-symbol predsym)) "(") '())))))))


(define (print-index-entries index p paren-stack)
  (if (pair? index)
      (begin
	(if (index-variable (and-content index))
	    (let ((parens (decrement-paren-stack paren-stack)))
	      (if (pair? (and-next index))
		  (print-index-entries (and-next index) 
				       (string-append p "X" (number->string (and-content index)) (car parens) ",")
				       (cdr parens))
		  (print-index-entries (and-next index) 
				       (string-append p "X" (number->string (and-content index)) (car parens))
				       (cdr parens)))
	      (print-index-entries (or-next index) p paren-stack))
	    (if (index-constant (and-content index))
		(let ((parens (decrement-paren-stack paren-stack)))
		  (if (pair? (and-next index))
		      (print-index-entries (and-next index) 
					   (string-append p (symbol->string (functor-symbol (and-content index))) 
							  (car parens) ",")
					   (cdr parens))
		      (print-index-entries (and-next index) 
					   (string-append p (symbol->string (functor-symbol (and-content index))) 
							  (car parens))
					   (cdr parens)))
		  (print-index-entries (or-next index) p paren-stack))
		(begin
		  (print-index-entries (and-next index) 
				       (string-append p (symbol->string (functor-symbol (and-content index))) "(")
				       (cons (functor-arity (and-content index)) paren-stack))
		  (print-index-entries (or-next index) p paren-stack)))))
      (if (vector? index)
	  (if (> (vector-ref index 0) 0)
	      (begin
		(display p)
		(display ") [")
		(display (vector-ref index 0))
		(display "] ")
		(if (= (subgoal-size) (vector-length (car (vector-ref index 1))))
		       (display-subgoal-list (vector-ref index 1))
		       (display-clauses (vector-ref index 1)))
		(newline))))))

(define (decrement-paren-stack paren-stack)
  (if (pair? paren-stack)
      (if (> (car paren-stack) 1)
	  (cons "" (cons (- (car paren-stack) 1) (cdr paren-stack)))
	  (let ((new-paren (decrement-paren-stack (cdr paren-stack))))
	    (cons (string-append ")" (car new-paren)) (cdr new-paren))))
      (cons  "" '())))


;;;*************************************************************************
;; ------------------------------------------------------------
;; index statistics functions
;; ------------------------------------------------------------

(define (variant-index-statistics clauses)
  (if (pair? clauses)
      (begin
	(display "clause ") 
	(display (input-number (clause-input (car clauses))))
	(display ": ")
	(display-index-statistics (input-variant-index (clause-input (car clauses))))
	(newline)
	(variant-index-statistics (cdr clauses)))))

(define (predicate-index-statistics predicates)
  (if (pair? predicates)
      (begin
	(print-all-index-statistics (car predicates))
	(predicate-index-statistics (cdr predicates)))
      (newline)))
		
(define (print-all-index-statistics predsym)
  (display (predsym-symbol predsym))
  (display ":")
  (display " (p-g ")
  (display-index-statistics (predsym-pos-gr-index predsym))
  (display ")")		    
  (display " (n-g ")	    
  (display-index-statistics (predsym-neg-gr-index predsym))
  (display ")")		    
  (display " (p-u ")	    
  (display-index-statistics (predsym-pos-unit-index predsym))
  (display ")")		    
  (display " (n-u ")	    
  (display-index-statistics (predsym-neg-unit-index predsym))
  (display ")")
  (newline))

(define (display-index-statistics index)
  (set! nr-index-paths 0)
  (set! nr-index-entries 0)
  (set! nr-empty-paths 0)
  (set! total-depth 0)
  (set! total-empty-depth 0)
  (set! total-branching 0)
  (set! nr-branching 0)
  (set! max-branching 0)
  (if (pair? (cdr index))
      (begin
	(set! total-branching (+ total-branching (length index)))
	(set! max-branching total-branching)
	(set! nr-branching (+ 1 nr-branching))))
  (get-index-statistics (cdr index) 0)
  (if (not (= (car index) nr-index-entries))
      (begin
	(display (car index))
	(display "/")))
  (display nr-index-entries)
  (display " ")
  (display nr-index-paths)
  (display " ")
  (display nr-empty-paths)
  (display " ")
  (display (if (> nr-index-paths 0) (quotient total-depth nr-index-paths) 0))
  (display " ")
  (display (if (> nr-branching 0) (quotient total-branching nr-branching) 0))
  (display " ")
  (display max-branching))

(define (get-index-statistics index depth)
  (if (pair? index)
      (begin
	(if (pair? (and-next index))
	    (begin
	      (set! total-branching (+ total-branching (length (and-next index))))
	      (if (> (length (and-next index)) max-branching) (set! max-branching (length (and-next index))))
	      (set! nr-branching (+ 1 nr-branching))))
	(get-index-statistics (and-next index) (+ 1 depth))
	(get-index-statistics (or-next index) depth))
      (if (vector? index)
	  (begin
	    (set! nr-index-paths (+ 1 nr-index-paths))
	    (set! total-depth (+ depth total-depth))
	    (if (= 0 (vector-ref index 0)) 
		(begin
		  (set! total-empty-depth (+ depth total-empty-depth))
		  (set! nr-empty-paths (+ 1 nr-empty-paths)))
		(begin 
		  (set! nr-index-entries (+ (length (vector-ref index 1)) nr-index-entries))))))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; index matching functions
;; ------------------------------------------------------------

(define (match-trail-with-index index)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; it is important here that empty lists are never pushed on the term-trail!
	(let (result)
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (set! result (match-term-with-index (ref (car current-terms)) index)) ;; insert the next term
	  (if result
	      result
	      (begin
		(if (pair? (cdr current-terms)) ;; in case of matching failure, restore the trail
		      (pop-term-trail))         ;; to its original state to enable backtracking
		(push-term-trail current-terms)
		#f)))
	(begin
	  (if (vector? index) ;; no more terms to match, check for a unit at the index-end
	    (if (> (vector-ref index 0) 0)
		(begin
		  (if (> debug-level 6)
		      (begin
			(display "Found valid index-end ")
			(display-subgoal-list (vector-ref index 1))
			(newline)))
		  (car (vector-ref index 1)))
		(begin ;; the case of reaching an empty index-end should be very rare, only with propositionals
		  (if (> debug-level 6)
		      (begin
			(display "Found empty index-end ")
			(newline)))
		  #f))
	    #f)))))

(define (match-term-with-index term index)
  (if (pair? index)
      (if (index-variable (and-content index))
	  (if (free-index-variable (and-content index))
	      (begin
		(bind-index-variable (and-content index) term)
		(let ((result (match-trail-with-index (and-next index))))
		  (unbind-index-variable (and-content index))
		  (if result
		      result
		      (match-term-with-index term (or-next index)))))
	      (begin
		(if (ref-equal term (index-variable-binding (and-content index)))
		    (begin
		      (let ((result (match-trail-with-index (and-next index))))
			(if result
			    result
			    (match-term-with-index term (or-next index)))))
		    (begin 
		      (match-term-with-index term (or-next index))))))
	  (if (variable term)
	      ;; a variable that has not already been matched above with an index-variable must fail immediately
	      ;; at this point it is clear that (and-content) is a functor, this cannot work
	      #f
	      (if (eq? (car term) (and-content index)) ;; we came across the proper functor
		  (if (constant term)
		      (match-trail-with-index (and-next index))
		      (let (result)
			(push-term-trail (cdr term))  ;; try matching functor args
			(set! result (match-trail-with-index (and-next index)))
			(if (not result)      ;; pop functor args off the trail to enable backtracking
			    (pop-term-trail))
			result))
		  (if (< (functor-number (car term)) (functor-number (and-content index)))
		                                                             ;;; Found functor smaller than searched functor?
		      (match-term-with-index term (or-next index))   ;; shallow backtracking over the functors
		      #f)))) ;; the searched functor is not there, abort
      #f)) ;; term not found in or-cell-list, abort

(define (match-terms-with-index terms rest-terms index)
  (if (pair? terms)
      (let ((term (ref (car terms))))
	(if (pair? index)
	    (if (index-variable (and-content index))
		(if (free-index-variable (and-content index))
		    (begin
		      (bind-index-variable (and-content index) term)
		      (let ((result (match-terms-with-index (cdr terms) rest-terms (and-next index))))
			(unbind-index-variable (and-content index))
			(if result
			    result
			    (match-terms-with-index terms rest-terms (or-next index)))))
		    (begin
		      (if (ref-equal term (index-variable-binding (and-content index)))
			  (begin
			    (let ((result (match-terms-with-index (cdr terms) rest-terms (and-next index))))
			      (if result
				  result
				  (match-terms-with-index terms rest-terms (or-next index)))))
			  (begin 
			    (match-terms-with-index terms rest-terms (or-next index))))))
		(if (variable term)
		    ;; a variable that has not already been matched above with an index-variable must fail!
		    (begin 
		      (match-terms-with-index terms rest-terms (or-next index)))
		    (if (eq? (car term) (and-content index))       ; the functor case
			(begin
			  (if (constant term)
			      (let ((result (match-terms-with-index (cdr terms) rest-terms (and-next index))))
				(if result
				    result
				    (match-terms-with-index terms rest-terms (or-next index))))
			      ;; try matching functor args
			      (let ((result (match-terms-with-index (cdr term) (cons (cdr terms) rest-terms) (and-next index))))
				(if result
				    result
				    (match-terms-with-index terms rest-terms (or-next index))))))
			(if (< (functor-number (car term)) (functor-number (and-content index)))
			    (match-terms-with-index terms rest-terms (or-next index))
			    #f))))
	    #f))
      (if (pair? rest-terms)
	  (begin
	    (match-terms-with-index (car rest-terms) (cdr rest-terms) index))
	  (begin
	    (if (vector? index)
		(if (> (vector-ref index 0) 0)
		    (begin
		      (if (> debug-level 2)
			  (begin
			    (display "Found valid index-end ")
			    (display-subgoal-list (vector-ref index 1))
			    (newline)))
		      (get-input-unit-from-subgoals (vector-ref index 1)))
		    (begin
		      (if (> debug-level 2)
			  (begin
			    (display "Found empty index-end ")
			  (newline)))
		      #f))
		(begin 
		  (if (> debug-level 8)
		      (print "term/index length mismatch"))
		  #f))))))

(define (free-index-variable i)
  (not (global-binding-ref i)))

(define (bind-index-variable i term)
  (global-binding-set i term))

(define (unbind-index-variable i)
  (global-binding-set i #f))

(define (index-variable-binding i)
  (global-binding-ref i))

(define (get-input-unit-from-subgoals subgoals)
  (if (pair? subgoals)
      (if (and (unit-subgoal (car subgoals)) (clause-initial (subgoal-clause (car subgoals))))
	  (car subgoals)
	  (get-input-unit-from-subgoals (cdr subgoals)))
      #f))

(define (global-binding-ref i)
  (if (< i (vector-length global-index-bindings))
      (vector-ref global-index-bindings i)
      (let ((new-binding (make-vector (* 2 (vector-length global-index-bindings)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length global-index-bindings)) #f)
	  (vector-set! new-binding j (vector-ref global-index-bindings j)))
	(set! global-index-bindings new-binding)
	#f)))
	
(define (global-binding-set i term)
  (if (< i (vector-length global-index-bindings))
      (vector-set! global-index-bindings i term)
      (let ((new-binding (make-vector (* 2 (vector-length global-index-bindings)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length global-index-bindings)) #f)
	  (vector-set! new-binding j (vector-ref global-index-bindings j)))
	(set! global-index-bindings new-binding)
	(vector-set! global-index-bindings i term)
	#t)))
	
(define (subgoals-subsumed-by-index subgoals)
  (if (pair? subgoals)
      (if (subgoal-simplified (car subgoals))
	  (subgoals-subsumed-by-index (cdr subgoals))
	  (let (index index-end)
	    (if (> debug-level 4)
		(begin
		  (display "Subsumption check of ")
		  (display-subgoal (car subgoals))
		  (newline)))
	    (if (eq? '+ (subgoal-sign (car subgoals)))
		(set! index (predsym-pos-unit-index (subgoal-pred (car subgoals))))
		(set! index (predsym-neg-unit-index (subgoal-pred (car subgoals)))))
	    (if (> debug-level 8)
		(begin
		  (print "index entries: ")
		  (print-index-entries (cdr index) "(" '())
		  (newline)))
	    ;;(term-trail-stop)
	    ;;(push-term-trail (subgoal-terms (car subgoals)))
	    ;;(set! index-end (match-trail-with-index (cdr index)))
	    ;;(term-trail-erase)
	    (set! index-end (match-terms-with-index (subgoal-terms (car subgoals)) '() (cdr index)))
	    (if index-end 
		index-end
		(subgoals-subsumed-by-index (cdr subgoals)))))
      #f))

(define (subgoal-subsumed-by-index subgoal)
  (let (index index-end)
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-unit-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-unit-index (subgoal-pred subgoal))))
    (if (> debug-level 5)
	(begin
	  (display "matching subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (> debug-level 6)
	      (begin
		(display "current index content: ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (set! index-end (match-terms-with-index (subgoal-terms subgoal) '() (cdr index)))
    index-end))

(define (match-subgoal-with-index subgoal)
  (let (index index-end)
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-neg-unit-index (subgoal-pred subgoal)))
	(set! index (predsym-pos-unit-index (subgoal-pred subgoal))))
    (if (> debug-level 5)
	(begin
	  (display "matching subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (> debug-level 6)
	      (begin
		(display "current index content: ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    ;;(term-trail-stop)
    ;;(push-term-trail (subgoal-terms subgoal))
    ;;(set! index-end (match-trail-with-index (cdr index)))
    ;;(term-trail-erase)
    (set! index-end (match-terms-with-index (subgoal-terms subgoal) '() (cdr index)))
    index-end))


;;;*************************************************************************
;; ------------------------------------------------------------
;; path variant functions
;; ------------------------------------------------------------

(define (subgoal-in-gr-index subgoal)
  (let (index index-end)
    (if (> debug-level 5)
	(begin
	  (display "gr-searching for subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (set! index-end (termlist-in-gr-index? (subgoal-terms subgoal) (cdr index)))
    (if (or (not (vector? index-end)) (= (vector-ref index-end 0) 0))
	(set! index-end #f)
	(set! index-end (vector-ref index-end 1)))
    index-end))

