(module initial-path
	(include "dctp.sch")
	(export (select-initial-path clauses) (initial-active-path clauses predsg) (make-initial-link-list subgoal))
	(import (globals "globals.scm") (index "index.scm") (unification "unification.scm") (link "link.scm") 
		(display "display.scm") (select "select.scm") (path "path.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define all-clauses-selected #f)
(define one-new-clause-selected #f)
(define one-new-subgoal-selected #f)

;;;*************************************************************************
;; ------------------------------------------------------------
;; Initial path functions
;; ------------------------------------------------------------

(define (initial-active-path clauses predsg)
  (if (pair? clauses)
      (let (sg)
	(reset-subgoal-links (clause-subgoals (car clauses)))
	(set! initial-length (+ initial-length 1))
	(if (clause-relevant (car clauses))
	    (set-relevant-predicate-weights (clause-subgoals (car clauses))))
	(compute-subgoal-weights (clause-subgoals (car clauses)))
	(set! sg (get-selected-subgoal (clause-subgoals (car clauses))))
	;;(set! sg (select-subgoal (clause-subgoals (car clauses)) #f))
	(unmark-subgoals (clause-subgoals (car clauses)))
	(if (unit-clause (car clauses))
	    (begin 
	      (insert-clause-in-index (car clauses))
	      (insert-subgoal-in-gr-index sg)
	      (insert-subgoal-in-index sg 'unit)))
	(set-clause-path (car clauses) predsg)
	(if dewey-numbers
	    (begin
	      (display-subgoal sg)
	      ;;(display " ")
	      ;;(display-dewey-number sg)
	      (newline)))
	(if predsg
	    (begin
	      (set-subgoal-next predsg (car clauses))
	      (set-clause-depth (car clauses) (+ 1 (subgoal-depth predsg))))
	    (begin
	      (set! input-root sg)
	      (set-clause-depth (car clauses) (- (length clauses)))))
	(if (pair? (cdr clauses))
	    (initial-active-path (cdr clauses) sg)
	    sg))
      #f))

(define (set-relevant-predicate-weights subgoals)
  (if (pair? subgoals)
      (begin
	(set-predsym-weight (subgoal-pred (car subgoals)) 0)
	(set-relevant-termlist-weights (subgoal-terms (car subgoals)))
	(set-relevant-predicate-weights (cdr subgoals)))))
  
(define (set-relevant-termlist-weights terms)
  (if (pair? terms)
      (begin
	(set-relevant-term-weights (car terms))
	(set-relevant-termlist-weights (cdr terms)))))

(define (set-relevant-term-weights term)
  (if (variable term)
      (set! relevant-terms-ground #f)
      (if (constant term)
          (set-functor-weight (car term) 0)
	  (begin
	    (set-functor-weight (car term) 1)
	    (set-relevant-termlist-weights (cdr term))))))

(define (reset-subgoal-links subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-links (car subgoals) '())
	(reset-subgoal-links (cdr subgoals)))))

(define (make-initial-link-list subgoal)
  (if subgoal
      (begin
	(make-initial-link-list (subgoal-predecessor subgoal))
	(insert-links subgoal (potential-linked-lits subgoal))
	(if (clause-relevant (subgoal-clause subgoal))
	    (set! initial-relevant-links (+ initial-relevant-links (length (subgoal-links subgoal)))))
	(set! initial-links (+ initial-links (length (subgoal-links subgoal))))
	(add-path-literal subgoal))
      #t))

(define (clause-selected clause)
  (subgoals-selected (clause-subgoals clause)))

(define (subgoals-selected subgoals)
  (if (pair? subgoals)
      (if (subgoal-selected (car subgoals))
	  #t
	  (subgoals-selected (cdr subgoals)))
      #f))

(define (select-initial-path clauses)
  ;;(display "marking relevant clauses...")
  (mark-relevant-clauses clauses)
  ;;(display "done")
  ;;(newline)
  (do ()
      (all-clauses-selected #t)
    ;;(display "loop for selecting initial subgoals...")
    ;;(newline)
    (set! all-clauses-selected #t)
    (set! one-new-clause-selected #f)
    (set! one-new-subgoal-selected #f)
    (mark-subgoals-in-clauses clauses)
    ;;(display "done")
    ;;(newline)
    (if (and (not all-clauses-selected) (not one-new-subgoal-selected))
	(begin
	  (mark-unselected-clauses clauses)
	  (set! all-clauses-selected #t))))
  (reset-link-statistics))

(define (mark-relevant-clauses clauses)
  (if (pair? clauses)
      (if link-from-conjecture
	  (begin
	    (mark-relevant-clauses (cdr clauses))
	    (if (and (clause-relevant (car clauses)) (not (clause-selected (car clauses))))
		(mark-subgoal-in-clause (car clauses))))
	  (begin
	    (if (and (clause-relevant (car clauses)) (not (clause-selected (car clauses))))
		(mark-subgoal-in-clause (car clauses)))
	    (mark-relevant-clauses (cdr clauses))))
      #t))

(define (mark-unselected-clauses clauses)
  (if (pair? clauses)
      (if link-from-conjecture
	  (begin
	    (mark-unselected-clauses (cdr clauses))
	    (if (not (clause-selected (car clauses)))
		(begin
		  ;;(display-clause (car clauses))
		  ;;(newline)
		  ;;(display-links-of-subgoals (clause-subgoals (car clauses)))
		  (mark-subgoal-in-clause (car clauses)))))
	  (begin
	    (if (not (clause-selected (car clauses)))
		(begin
		  ;;(display-clause (car clauses))
		  ;;(newline)
		  ;;(display-links-of-subgoals (clause-subgoals (car clauses)))
		  (mark-subgoal-in-clause (car clauses))))
	    (mark-unselected-clauses (cdr clauses))))
      #t))

(define (mark-subgoals-in-clauses clauses)
  (if (pair? clauses)
      (if link-from-conjecture
	  (begin
	    (mark-subgoals-in-clauses (cdr clauses))
	    (if (clause-selected (car clauses))
		(mark-subgoal-in-clause (car clauses))
		(set! all-clauses-selected #f)))
	  (begin
	    (if (clause-selected (car clauses))
		(mark-subgoal-in-clause (car clauses))
		(set! all-clauses-selected #f))
	    (mark-subgoals-in-clauses (cdr clauses))))
      #t))

(define (mark-subgoal-in-clause clause)
  (let (sg)
    (if least-initial-links
	(set! sg (least-linked-subgoal (clause-subgoals clause)))
	(set! sg (widest-linked-subgoal (clause-subgoals clause))))
    (if sg
	(begin
	  (set! one-new-subgoal-selected #t)
	  (if (clause-selected clause) 
	      (set-subgoal-selected sg 'semi-selected)
	      (begin
		;;(print "marked clause " (input-number (clause-input clause)) "(" (clause-depth clause) ")")
		(set-subgoal-selected sg #t)))
	  (mark-linked-subgoals (best-of-links (subgoal-links sg) sg))))))

(define (widest-linked-subgoal subgoals)
  (if (null? subgoals)
      #f
      (if (subgoal-selected (car subgoals))
	  (widest-linked-subgoal (cdr subgoals))
	  (let ((sg (widest-linked-subgoal (cdr subgoals))))
	    (if (and sg (> (length (subgoal-links sg)) (length (subgoal-links (car subgoals)))))
		sg
		(car subgoals))))))

(define (least-linked-subgoal subgoals)
  (if (null? subgoals)
      #f
      (if (subgoal-selected (car subgoals))
	  (least-linked-subgoal (cdr subgoals))
	  (let ((sg (least-linked-subgoal (cdr subgoals))))
	    (if (and sg (<= (length (subgoal-links sg)) (length (subgoal-links (car subgoals)))))
		sg
		(car subgoals))))))

(define (best-of-links links subgoal)
  (if (null? links)
      '()
      (if (null? (cdr links))
	  links
	  (let (c1 c2
		(r (best-of-links (cdr links) subgoal)))
	    (set! c1 (compute-link-complexity subgoal (car r)))
	    (set! c2 (compute-link-complexity subgoal (car links)))
	    (if (eq? (subgoal-clause (car r)) (subgoal-clause (car links)))
		(if (> c1 c2)
		    (cons (car links) (cdr r))
		    r)
		(cons (car links) r))))))

(define (mark-following-links links)
  (if (pair? links)
      (begin
	(if (not (clause-selected (subgoal-clause (car links))))
	    (begin
	      (set-subgoal-selected (car links) #t)
	      (let (sg)
		(if least-initial-links
		    (set! sg (least-linked-subgoal (clause-subgoals (subgoal-clause (car links)))))
		    (set! sg (widest-linked-subgoal (clause-subgoals (subgoal-clause (car links))))))
		(if sg
		    (mark-following-links (best-of-links (subgoal-links sg) sg))))))
	(mark-following-links (cdr links)))))

(define (mark-linked-subgoals links)
  (if (pair? links)
      (begin
	(if (not (clause-selected (subgoal-clause (car links))))
	    (begin
	      (set-subgoal-selected (car links) #t)
	      ;;(print "marked clause " (input-number (clause-input (subgoal-clause (car links)))) "(" 
	      ;;(clause-depth (subgoal-clause (car links))) ")")
	      (set! one-new-clause-selected #t))
	(mark-linked-subgoals (cdr links))))))

(define (get-selected-subgoal subgoals)
  (if (null? subgoals)
      #f
      (if (eq? (subgoal-selected (car subgoals)) #t)
	  (car subgoals)
	  (get-selected-subgoal (cdr subgoals)))))

(define (unmark-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-selected (car subgoals) #f)
	(unmark-subgoals (cdr subgoals)))))

(define (display-links-of-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(display-subgoal-list (subgoal-links (car subgoals)))
	(newline)
	(display-links-of-subgoals (cdr subgoals)))))

