(module normal-form
	(include "dctp.sch")
	(export (renamed-subgoals subgoals) (clause-variant-subsumed clause) (non-equivalent-termlists k l)
		(renamed-termlist terms) (sign-predsym< sga sgb) (set-linked-clauses-tested clausea clauseb)
		(unset-linked-clauses clausea clauseb))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm")
		(index "index.scm")))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define current-class 1)
(define path-clause-test-vector (make-vector 5000 #f))
(define tested-path-clauses '())

;;;*************************************************************************
;; ------------------------------------------------------------
;; Normal form functions
;; ------------------------------------------------------------

(define (non-equivalent-termlists k l)
  (let (result)
    (set! var-counter 0)
    (trail-stop)
    (set! result (non-equivalent-termlists-1 k l))
    (if (not result) (unbind-vars-and-undo-var-counter))
; 	(begin
; 	  (display "equivalent: ")
; 	  (display-ref-list k)
; 	  (display " and ")
; 	  (display-ref-list l)
; 	  (newline)
; 	  (unbind-vars-and-undo-var-counter)
; 	  (display "unbound:    ")
; 	  (display-subgoal-terms k)
; 	  (display " and ")
; 	  (display-subgoal-terms l)
; 	  (newline)))
    result))

(define (non-equivalent-termlists-1 k l)
  (if (pair? k)
      (let ((result (non-equivalent-terms (ref (car k)) (ref (car l)))))
	(if result
	    result
	    (non-equivalent-termlists-1 (cdr k) (cdr l))))
      #f))

(define (non-equivalent-terms s t)
  (if (eq? s t)
      #f
      (if (and (variable s) (variable t))
	  (begin
	    (bind s var-counter)
	    (bind t var-counter)
	    (set! var-counter (+ var-counter 2))
	    #f)
	  (if (and (non-variable s) (non-variable t) (eq? (car s) (car t)))
	      (non-equivalent-termlists-1 (cdr s) (cdr t))
	      (begin
		(unbind-vars-and-undo-var-counter)
		#t)))))

(define (non-equivalent-subgoals a b)
  (if (eq? (subgoal-sign a) (subgoal-sign b))
      (if (eq? (subgoal-pred a) (subgoal-pred b))
	  (non-equivalent-termlists (subgoal-terms a) (subgoal-terms b))
	  #t)
      #t))

(define (renamed-subgoals subgoals)
  (if (pair? subgoals)
;       (let (index-end)
; 	(if (not (subgoal-simplified (car subgoals)))
; 	    (begin
; 	      (set! index-end (match-subgoal-with-index (car subgoals)))
; 	      (if index-end
; 		  (begin
; 		    (if (or dewey-numbers (> debug-level 4))
; 			(begin
; 			  (display "simplifying subgoal ")
; 			  (display-subgoal (car subgoals))
; 			  ;;(display " of depth ")
; 			  ;;(display (subgoal-depth (car subgoals)))
; 			  (display " matching unit ")
; 			  (display-clause (subgoal-clause index-end))
; 			  (newline)))
; 		    (set! nr-simplifications (+ nr-simplifications 1))))))
; 	(if (or (subgoal-simplified (car subgoals)) index-end)
; 	    (renamed-subgoals (cdr subgoals))
      (let ((sg (new-subgoal)) (new-subgoals (renamed-subgoals (cdr subgoals))))
	(set-subgoal-sign sg (subgoal-sign (car subgoals)))
	(set-subgoal-pred sg (subgoal-pred (car subgoals)))
	(set-subgoal-unlinkable sg (subgoal-unlinkable (car subgoals)))
	(set-subgoal-selected sg #f)
	(set-subgoal-simplified sg #f)
	(set-subgoal-links sg '())
	(set-subgoal-terms sg (renamed-termlist (subgoal-terms (car subgoals))))
	(set-subgoal-number sg (subgoal-number (car subgoals)))
	(set-subgoal-weight sg (subgoal-misc (car subgoals)))
	(set-subgoal-misc (car subgoals) #f)
	(cons sg new-subgoals))
      '()))
	    
(define (renamed-termlist terms)
  (if (pair? terms)
      (cons (renamed-term (semi-ref (car terms))) (renamed-termlist (cdr terms)))
      '()))

(define (renamed-term term)
  (if (variable term)
      (if (variable-renaming term)
	  (variable-renaming term)
	  (let ((renaming (new-variable)))
	    (push-trail term)
	    (set-variable-renaming term renaming)
	    (set-variable-name renaming (variable-name term))
	    (if (bound-variable term)
		(set-variable-binding renaming (renamed-term (variable-binding term))))
	    renaming))
      (if (constant term)
	  term
	  (begin
	    (if (= (functor-arity (car term)) 0) (print "copying constant " term))
	    (cons (car term) (renamed-termlist (cdr term)))))))

; (define (semi-ref term)
; ;; if term is a <bound_variable> bound to another variable, its binding is returned, 
; ;; otherwise term is returned
;  (if (or (non-variable term) (free-variable term) (domain-element term)
; 	 (and (variable term) (non-variable (variable-binding term))))
;      term
;      (semi-ref (car term))))

(define (semi-ref term)
  (if (and (bound-variable term) (variable (variable-binding term)))
      (semi-ref (car term))
      term))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variant subsumption functions
;; ------------------------------------------------------------

(define (clause-variant-subsumed unsimplifieds)
  (if (pair? unsimplifieds)
      (let ((result (test-potential-path-variants unsimplifieds (length unsimplifieds) unsimplifieds)))
	(unset-tested-path-clauses)
	(if result
	    (begin
	      (if dewey-numbers
		  (begin
		    (display "variant ") 
		    (display-clause (subgoal-clause (car unsimplifieds)))
		    (display " subsumed by ") 
		    (display-clause result)
		    (newline)))
	      (set! variant-subsumptions (+ variant-subsumptions 1))))
	result)
      #f))

(define (test-potential-path-variants subgoals clause-length unsimplifieds)
  (if (pair? subgoals)
      (let ((result (subgoal-in-gr-index (car subgoals))))
	(if result
	    (set! result (test-path-clauses (car subgoals) result clause-length unsimplifieds)))
	(if result
	    result
	    (test-potential-path-variants (cdr subgoals) clause-length unsimplifieds)))
      #f))

(define (test-path-clauses subgoal path-subgoals clause-length unsimplifieds)
  (if (pair? path-subgoals)
      (if (or (= (clause-unsimplified (subgoal-clause (car path-subgoals))) 1)
	      (path-subgoal-clause-tested (car path-subgoals)))
	  (test-path-clauses subgoal (cdr path-subgoals) clause-length unsimplifieds)
;	  (begin
; 	    (if dewey-numbers
; 		(begin
; 		  (display "vs-test with ")
; 		  (display-clause (subgoal-clause (car path-subgoals)))
; 		  (display " ")))
; 	    (if (non-equivalent-termlists (subgoal-terms (car path-subgoals)) (subgoal-terms subgoal))
; 		(begin
; 		  ;;(if dewey-numbers (print "fail: false potential"))
; 		  (set! potential-fails (+ potential-fails 1))
; 		  (test-path-clauses subgoal (cdr path-subgoals) clause-length))
		(begin
		  (set! variant-subsumption-tests (+ variant-subsumption-tests 1))
		  (set-path-subgoal-clause-tested (car path-subgoals))
		  (if (>= clause-length (clause-unsimplified (subgoal-clause (car path-subgoals))))
		      (let ((c-subgoals (remq subgoal unsimplifieds))
			    (p-subgoals (unsimplifieds-minus-subgoal (car path-subgoals) 
								     (clause-subgoals (subgoal-clause (car path-subgoals))))))
			(if (predicates-match-individually c-subgoals p-subgoals)
			    (if (subgoals-match-individually c-subgoals p-subgoals)
				(begin
				  (set! var-counter 1)
				  (if (permutation-variant-subsumed (clause-subgoals (subgoal-clause subgoal)) 
								    (clause-subgoals (subgoal-clause (car path-subgoals))))
				      (subgoal-clause (car path-subgoals))
				      (begin
					;;(if dewey-numbers (print "fail: full test"))
					(set! full-fails (+ full-fails 1))
					(test-path-clauses subgoal (cdr path-subgoals) clause-length unsimplifieds))))
				(begin 
				  ;;(if dewey-numbers (print "fail: single match"))
				  (set! single-match-fails (+ single-match-fails 1))
				  (test-path-clauses subgoal (cdr path-subgoals) clause-length unsimplifieds)))
			    (begin 
			      ;;(if dewey-numbers (print "fail: signature match"))
			      (set! signature-fails (+ signature-fails 1))
			      (test-path-clauses subgoal (cdr path-subgoals) clause-length unsimplifieds))))
		      (begin 
			;;(if dewey-numbers (print "fail: length"))
			(set! length-fails (+ length-fails 1))
			(test-path-clauses subgoal (cdr path-subgoals) clause-length unsimplifieds)))));;)
      #f))

(define (unsimplifieds-minus-subgoal subgoal subgoals)
  (if (pair? subgoals)
      (if (not (or (subgoal-simplified (car subgoals)) (eq? subgoal (car subgoals))))
	  (cons (car subgoals) (unsimplifieds-minus-subgoal subgoal (cdr subgoals)))
	  (unsimplifieds-minus-subgoal subgoal (cdr subgoals)))
      '()))

(define (predicates-match-individually subgoals path-subgoals)
  (if (pair? path-subgoals)
      (let ((result (predicate-matches-individually subgoals (car path-subgoals))))
	(if result
	    (predicates-match-individually result (cdr path-subgoals))
	    #f))
      #t))

(define (predicate-matches-individually subgoals path-subgoal)
  (let ((valid-subgoals (do ((sg-pointer subgoals (cdr sg-pointer)))
			    ((or (null? sg-pointer) (sign-predsym>= (car sg-pointer) path-subgoal))
			     (if (and (pair? sg-pointer) (sign-predsym= (car sg-pointer) path-subgoal))
				 sg-pointer
				 '())))))
    (if (pair? valid-subgoals)
	(cdr valid-subgoals)
	#f)))

(define (subgoals-match-individually subgoals path-subgoals)
  (if (pair? path-subgoals)
      (let ((result (subgoal-matches-individually subgoals (car path-subgoals))))
	(if result
	    (subgoals-match-individually result (cdr path-subgoals))
	    #f))
      #t))

(define (subgoal-matches-individually subgoals path-subgoal)
  (let ((valid-subgoals (do ((sg-pointer subgoals (cdr sg-pointer)))
			    ((or (null? sg-pointer) (sign-predsym>= (car sg-pointer) path-subgoal))
			     (if (and (pair? sg-pointer) (sign-predsym= (car sg-pointer) path-subgoal))
				 sg-pointer
				 '())))))
    (if (pair? valid-subgoals)
	(if (= (subgoal-weight path-subgoal) (subgoal-misc (car valid-subgoals)))
	    (let ((result (non-equivalent-termlists (subgoal-terms path-subgoal) (subgoal-terms (car valid-subgoals)))))
	      (if result
		  (begin
		    (set! result (subgoal-matches-individually (cdr valid-subgoals) path-subgoal))
		    (if result
			(cons (car valid-subgoals) result)
			#f))
		  (cdr valid-subgoals)))
	    (let ((result (subgoal-matches-individually (cdr valid-subgoals) path-subgoal)))
		    (if result
			(cons (car valid-subgoals) result)
			#f)))
	#f)))

(define (permutation-variant-subsumed subgoals path-subgoals)
  (if (pair? path-subgoals)
      (do ((success #f) (fail #f) (possible-matches subgoals) (rest-subgoals '()) (v #f))
	  ((or success fail) success)
	(set! v (get-globally-matching-subgoal possible-matches (car path-subgoals)))
	(if v
	    (begin
	      (set! rest-subgoals (remq v subgoals))
	      (if (permutation-variant-subsumed rest-subgoals (cdr path-subgoals))
		  (set! success #t)
		  (set! possible-matches (remq v possible-matches)))
	      (unbind-vars-and-undo-var-counter))
	    (set! fail #t)))
      #t))
		    
(define (get-globally-matching-subgoal possible-matches path-subgoal)
  (if (pair? possible-matches)
      (let ((result (globally-non-equivalent-subgoals path-subgoal (car possible-matches))))
	(if result
	    (if (sign-predsym> (car possible-matches) path-subgoal)
		#f
		(get-globally-matching-subgoal (cdr possible-matches) path-subgoal))
	    (car possible-matches)))
      #f))

(define (globally-non-equivalent-subgoals a b)
  (if (eq? (subgoal-sign a) (subgoal-sign b))
      (if (eq? (subgoal-pred a) (subgoal-pred b))
	  (begin
	    (trail-stop)
	    (non-equivalent-termlists-1 (subgoal-terms a) (subgoal-terms b)))
	  (if (> (predsym-number (subgoal-pred a)) (predsym-number (subgoal-pred b)))
	      1
	      -1))
      (if (eq? '+ (subgoal-sign a))
	  -1
	  1)))

(define (sign-predsym< sga sgb)
  (if (and (eq? (subgoal-sign sga) '+) (eq? (subgoal-sign sgb) '-))
      #t
      (if (and (eq? (subgoal-sign sga) (subgoal-sign sgb)) 
	       (< (predsym-number (subgoal-pred sga)) (predsym-number (subgoal-pred sgb))))
	  #t
	  #f)))

(define (sign-predsym> sga sgb)
  (if (and (eq? (subgoal-sign sga) '-) (eq? (subgoal-sign sgb) '+))
      #t
      (if (and (eq? (subgoal-sign sga) (subgoal-sign sgb))
	       (> (predsym-number (subgoal-pred sga)) (predsym-number (subgoal-pred sgb))))
	  #t
	  #f)))

(define (sign-predsym>= sga sgb)
  (if (and (eq? (subgoal-sign sga) '-) (eq? (subgoal-sign sgb) '+))
      #t
      (if (and (eq? (subgoal-sign sga) (subgoal-sign sgb))
	       (>= (predsym-number (subgoal-pred sga)) (predsym-number (subgoal-pred sgb))))
	  #t
	  #f)))

(define (sign-predsym= sga sgb)
  (if (and (eq? (subgoal-sign sga) (subgoal-sign sgb)) 
	   (= (predsym-number (subgoal-pred sga)) (predsym-number (subgoal-pred sgb))))
      #t
      #f))

(define (set-linked-clauses-tested clausea clauseb)
  (let ((deptha (clause-depth clausea)) (depthb (clause-depth clauseb)))
    (if (>= deptha 0)
	(path-clause-tested-set deptha #t))
    (if (>= depthb 0)
	(path-clause-tested-set depthb #t))))

(define (unset-linked-clauses clausea clauseb)
  (let ((deptha (clause-depth clausea)) (depthb (clause-depth clauseb)))
    (if (>= deptha 0)
	(path-clause-tested-set deptha #f))
    (if (>= depthb 0)
	(path-clause-tested-set depthb #f))))

(define (path-subgoal-clause-tested subgoal)
  (let ((depth (subgoal-depth subgoal)))
    (if (>= depth 0) 
	(path-clause-tested-ref depth)
	#t)))

(define (set-path-subgoal-clause-tested subgoal)
  (let ((depth (subgoal-depth subgoal)))
    (if (>= depth 0) 
	(if (vector-ref path-clause-test-vector depth)
	    #f
	    (begin
	      (path-clause-tested-set depth #t)
	      (set! tested-path-clauses (cons depth tested-path-clauses))
	      #t))
	#f)))

(define (unset-tested-path-clauses)
  (undo-vector-settings tested-path-clauses)
  (set! tested-path-clauses '()))

(define (undo-vector-settings path-clauses)
  (if (pair? path-clauses)
      (begin
	(undo-vector-settings (cdr path-clauses))
	(set-cdr! path-clauses '())
	(path-clause-tested-set (car path-clauses) #f))))

(define (path-clause-tested-ref i)
  (if (< i (vector-length path-clause-test-vector))
      (vector-ref path-clause-test-vector i)
      (let ((new-test-vector (make-vector (* 2 (vector-length path-clause-test-vector)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length path-clause-test-vector)) #f)
	  (vector-set! new-test-vector j (vector-ref path-clause-test-vector j)))
	(set! path-clause-test-vector new-test-vector)
	#f)))
	
(define (path-clause-tested-set i value)
  (if (< i (vector-length path-clause-test-vector))
      (vector-set! path-clause-test-vector i value)
      (let ((new-test-vector (make-vector (* 2 (vector-length path-clause-test-vector)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length path-clause-test-vector)) #f)
	  (vector-set! new-test-vector j (vector-ref path-clause-test-vector j)))
	(set! path-clause-test-vector new-test-vector)
	(vector-set! path-clause-test-vector i value)
	#t)))

