;; -------------------------------------------------------
;; PARSER FOR TPTP INPUT FILES
;; -------------------------------------------------------
(module parser
	(include "dctp.sch")
	(export tptp-parser tptp-scanner (display-statistics))
	(export nr-clauses nr-non-horn nr-units nr-literals nr-eq-literals nr-predicates nr-propositionals 
		nr-functors nr-constants nr-variables nr-singleton-variables nr-axioms nr-lemmas nr-hypotheses
		nr-conjectures  nr-all-negatives predicate-list models)
	(import (globals "globals.scm") (formula globals "globals.scm")	(index  "index.scm") 
		(sign-predsym< normal-form "normal-form.scm") (display-clause display "display.scm")
		(non-tautological pruning "pruning.scm") (unrename-vars dctp  "dctp.scm")))

(define current-clause '())
(define current-input '())
(define current-clause-nr 0)
(define current-clause-variables '())
(define current-clause-var-occurrences '())
(define current-variable '())
(define current-literal '())
(define current-predicate '())
(define current-pred-nr 1)
(define current-functor '())
(define current-signature '())
(define current-functor-nr 1)
(define current-model #f)
(define current-mapping #f)
(define current-class 1)
(define predicate-list '())
(define functor-list '())
(define models '())
(define propositional-list '())
(define constant-list '())
(define current-positive-subgoals 0)
(define current-negative-subgoals 0)
(define nr-clauses 0)               
(define nr-non-horn 0)              
(define nr-units 0)                 
(define nr-literals 0)              
(define nr-eq-literals 0)           
(define max-clause-length 0)        
(define max-clause-variables 0)        
(define nr-predicates 0)            
(define nr-propositionals 0)        
(define max-pred-arity 0)           
(define min-pred-arity 0)           
(define nr-functors 0)              
(define nr-constants 0)             
(define max-func-arity 0)           
(define min-func-arity 0)           
(define nr-variables 0)             
(define nr-singleton-variables 0)   
(define nr-axioms 0)   
(define nr-lemmas 0)   
(define nr-hypotheses 0)   
(define nr-conjectures 0)   
(define nr-all-negatives 0)   

(define tptp-scanner
  (regular-grammar
   ((lowercase (in (#\a #\z)))
    (uppercase (in (#\A #\Z)))
    (numeric (in (#\0 #\9)))
    (special (in #\+ #\- #\_))
    (alphanum (or lowercase uppercase numeric special)))
   ("input_clause" (display-for-debug) 'input_clause)
   ("model_clause" (display-for-debug) 'model_clause)
   ("domain_definition" (display-for-debug) 'domain_definition)
   ("predicate_map" (display-for-debug) 'predicate_map)
   ("functor_map" (display-for-debug) 'functor_map)
   ("equal" (display-for-debug)
	      (cons 'symbol (the-symbol)))
   ((: uppercase (* alphanum)) ;; (print (the-string))
			     (cons 'variable (string->symbol (the-string))))
   ((: (or lowercase numeric) (* alphanum)) ;; (print (the-string))
                             (cons 'symbol (string->symbol (the-string))))
   ("++" (display-for-debug) 'pos_sign)
   ("--" (display-for-debug) 'neg_sign)
   (#\( (display-for-debug) 'openpar)
   (#\) (display-for-debug) 'closepar)
   (#\[ (display-for-debug) 'openbracket)
   (#\] (display-for-debug) 'closebracket)
   (#\, (display-for-debug) 'comma)
   (#\. (display-for-debug) 'dot)

   ((eol (: #\% (* all))) (display-for-debug) (ignore))

   ((+ (in #\space #\tab #\newline)) (display-for-debug) (ignore))))

(define (variable-exists? variable variables)
  (if (null? variables)
      #f
      (if (eq? variable (variable-name (car variables)))
	  #t
	  (variable-exists? variable (cdr variables)))))

(define (get-variable variable variables)
  (if (null? variables)
      #f
      (if (eq? variable (variable-name (car variables)))
	  (car variables)
	  (get-variable variable (cdr variables)))))

(define (increment-occurences variable variables occurrences)
  (if (null? variables)
      #f
      (if (eq? variable (variable-name (car variables)))
	  (set-car! occurrences (+ (car occurrences) 1))
	  (increment-occurences variable (cdr variables) (cdr occurrences)))))

(define (set-clause-of-litlist litlist copy)
  (if (null? litlist)
      #t
      (begin
	(set-subgoal-clause (car litlist) copy)
	(set-clause-of-litlist (cdr litlist) copy))))

(define (compute-clause-complexity clause)
  (- (subgoals-complexity (clause-subgoals clause)) (length (clause-variables clause))))

(define (subgoals-complexity subgoals)
  (if (null? subgoals)
      0
      (+ (subgoal-complexity (car subgoals)) (subgoals-complexity (cdr subgoals)))))

(define (term-complexity term depth)
  (if (> depth input-term-depth)
      (set! input-term-depth depth))
  (if (or (variable term) (constant term)) 
      1
      (+ 1 (termlist-complexity (cdr term) (+ depth 1)))))

(define (termlist-complexity terms depth)
  (if (null? terms)
      0
      (+ (term-complexity (car terms) depth) (termlist-complexity (cdr terms) depth))))

(define (subgoal-complexity subgoal)
  (let ((terms (subgoal-terms subgoal)))
    (termlist-complexity terms 1)))

(define tptp-parser
  (lalr-grammar
   (variable symbol input_clause domain_definition predicate_map functor_map pos_sign neg_sign openpar closepar 
	     openbracket closebracket comma dot) 
   (formula ((clauselist)
                          (set! nr-clauses current-clause-nr)
			  (set! nr-predicates (length predicate-list))
			  (set! nr-propositionals (length propositional-list))
			  (set! nr-functors (length functor-list))
			  (set! nr-constants (length constant-list))
			  (set! max-pred-arity (get-max-pred-arity predicate-list))
			  (set! min-pred-arity (get-min-pred-arity predicate-list))
			  (set! max-func-arity (get-max-func-arity functor-list))
			  (set! min-func-arity (get-min-func-arity functor-list))
			  (if (= nr-all-negatives 0)
			      (if (not silent) (print "no all negative clause in formula")))
			  (if (and (not norel) (= nr-clauses nr-conjectures))
			      (if (not silent) (print "all clauses of type conjecture ... restricting relevance to all-negative clauses")))
			  (if (and (not norel) (= nr-conjectures 0))
			      (if (not silent) (print "no conjecture type clause ... restricting relevance to all-negative clauses")))
			  (if (or norel (= nr-clauses nr-conjectures) (= nr-conjectures 0))
			      (set-all-negatives-relevant clauselist))
			  (if (= nr-propositionals nr-predicates)
			      (begin
				(if (not silent) (print "input formula is propositional"))
				(set! groundable #t))
			      (if (= max-func-arity 0)
				  (begin
				    (if (not silent) (print "input formula is in Bernays-Schnfinkel class"))
				    (set! groundable #t))))
			  (if (= 0 nr-non-horn)
			      (begin
				(if (not silent) (print "input formula is Horn"))
				(set! horn #t)))
			  (set-clause-depths clauselist)
			  (set! global-index-bindings (make-vector (+ nr-variables 1) #f))
			  (set-formula-clauses formula clauselist)))
   (clauselist
    (() '())
    ((clauselist clause) 
     (if (or tautologies (non-tautological clause))
	 (begin
	   (if (all-negative (clause-subgoals clause))
	       (set! nr-all-negatives (+ nr-all-negatives 1)))
	   (insert-new-clause-in-clauselist clause clauselist))
	 (begin
	   (if (not silent) 
	       (begin 
		 (display "deleted tautological input ")
		 (display-clause clause)
		 (newline)))
	   clauselist))))
   (clause
    ((input_clause openpar clausename comma type comma literals closepar dot)
     (set! current-input (new-input))
     (set! current-clause (new-clause))
     (set-clause-input current-clause current-input)
     (set-input-name current-input clausename)
     (set-input-type current-input type)
     (if (and (not norel) (eq? type 'conjecture))
	 (set-clause-relevant current-clause #t))
     (set! current-clause-nr (+ current-clause-nr 1))
     (set-input-number current-input current-clause-nr)
     (set-clause-subgoals current-clause literals)
     (set-input-length current-input (length literals))
     (set-clause-unsimplified current-clause (length literals))
     (set! nr-literals (+ nr-literals (input-length current-input)))
     (if (> (input-length current-input) max-clause-length)
	 (set! max-clause-length (input-length current-input)))
     (if (= (input-length current-input) 1)
	 (begin
	   (set! nr-units (+ nr-units 1))))
     (set-clause-of-litlist literals current-clause)
     (set-input-copies current-input (list current-clause))
     (set-input-all-copies current-input (input-copies current-input))
     (set-input-variant-index current-input (cons 0 '()))
     (set-clause-variables current-clause current-clause-variables)
     (set! nr-variables (+ nr-variables (length current-clause-variables)))
     (if (> (length current-clause-variables) max-clause-variables)
	 (set! max-clause-variables (length current-clause-variables)))
     (set! nr-singleton-variables 
	   (+ nr-singleton-variables 
	      (get-nr-current-singletons current-clause-var-occurrences)))
     (set-clause-nr-variables current-clause (length current-clause-variables))
     (set-clause-initial current-clause #t)
     (set-input-complexity current-input
			    (compute-clause-complexity current-clause))
     (set! current-clause-variables '())
     (set! current-clause-var-occurrences '())
     (if (> current-positive-subgoals 1)
	 (set! nr-non-horn (+ nr-non-horn 1)))
     (set! current-positive-subgoals 0) (set! current-negative-subgoals 0)
     (set-input-signature current-input current-signature)
     (set! current-signature '())
     current-clause))
   (literals
    ((openbracket litlist closebracket) litlist))
   (litlist
    ;;((litlist comma literal) (append litlist (list literal)))
    ((litlist comma literal) (insert-literal literal litlist))
    ((literal) (list literal)))
   (literal
    ((pos_sign atom)
     (set! current-literal (new-subgoal))
     (set-subgoal-sign current-literal '+)
     (set-subgoal-pred current-literal (car atom))
     (set-subgoal-terms current-literal (cdr atom))
     (set-subgoal-links current-literal '())
     (set-subgoal-selected current-literal #f)
     (set-subgoal-next current-literal #f)
     (set! current-positive-subgoals (+ current-positive-subgoals 1))
     (set-subgoal-number current-literal 
              (+ current-negative-subgoals current-positive-subgoals))
     current-literal)
    ((neg_sign atom)
     (set! current-literal (new-subgoal))
     (set-subgoal-sign current-literal '-)
     (set-subgoal-pred current-literal (car atom))
     (set-subgoal-terms current-literal (cdr atom))
     (set-subgoal-links current-literal '())
     (set-subgoal-selected current-literal #f)
     (set-subgoal-next current-literal #f)
     (set! current-negative-subgoals (+ current-negative-subgoals 1))
     (set-subgoal-number current-literal 
              (+ current-negative-subgoals current-positive-subgoals))
     current-literal))
   (atom
    ((symbol) (set! nr-propositionals (+ nr-propositionals 1)) 
	      ;; (print symbol)
	      (if (not (get-predsym symbol 0 predicate-list))
		  (begin
		    (set! current-predicate (new-predsym))
		    (set-predsym-pos-path-index current-predicate (cons 0 '()))
		    (set-predsym-neg-path-index current-predicate (cons 0 '()))
		    (set-predsym-pos-gr-index current-predicate (cons 0 '()))
		    (set-predsym-neg-gr-index current-predicate (cons 0 '()))
		    (set-predsym-pos-unit-index current-predicate (cons 0 '()))
		    (set-predsym-neg-unit-index current-predicate (cons 0 '()))
		    (set-predsym-symbol current-predicate symbol)
		    (set-predsym-arity  current-predicate 0)
		    (set-predsym-number current-predicate current-pred-nr)
		    (set-predsym-weight current-predicate 1)
		    (set! current-pred-nr (+ current-pred-nr 1))
		    (set! predicate-list (cons current-predicate predicate-list)))
		  (set! current-predicate (get-predsym symbol 0 predicate-list)))
	      (if (not (get-signature symbol 0 current-signature))
		    (set! current-signature (cons current-predicate 
						  current-signature)))
	      (if (not (memq symbol propositional-list))
		  (set! propositional-list (cons symbol propositional-list)))
	      (cons current-predicate '()))
    ((symbol parameters) 
     ;; (print symbol)
     (if (eq? symbol 'equal)
	 (set! nr-eq-literals (+ nr-eq-literals 1)))
     (if (not (get-predsym symbol (length parameters) predicate-list))
	 (begin
	   (set! current-predicate (new-predsym))
	   (set-predsym-pos-path-index current-predicate (cons 0 '())) 
	   (set-predsym-neg-path-index current-predicate (cons 0 '()))
	   (set-predsym-pos-gr-index current-predicate (cons 0 '())) 
	   (set-predsym-neg-gr-index current-predicate (cons 0 '()))
	   (set-predsym-pos-unit-index current-predicate (cons 0 '())) 
	   (set-predsym-neg-unit-index current-predicate (cons 0 '()))
	   (set-predsym-symbol current-predicate symbol)
	   (set-predsym-arity  current-predicate (length parameters))
	   (set-predsym-number current-predicate current-pred-nr)
           (if (eq? symbol 'equal)
	       (set-predsym-weight current-predicate 1)
	       (set-predsym-weight current-predicate 1))
	   (set! current-pred-nr (+ current-pred-nr 1))
	   (set! predicate-list (cons current-predicate predicate-list)))
	 (set! current-predicate (get-predsym symbol (length parameters) predicate-list)))
     (if (not (get-signature symbol (length parameters) current-signature))
	 (set! current-signature (cons current-predicate current-signature)))
     (cons current-predicate parameters)))
   (term
    ((openpar term closepar) term)
    ((symbol) 
     (if (not (get-functor symbol 0 functor-list))
	 (begin
	   (set! current-functor (new-functor))
	   (set-functor-symbol current-functor symbol)
	   (set-functor-arity  current-functor 0)
	   (set-functor-number current-functor current-functor-nr)
	   (set-functor-weight current-functor 1)
	   (set! current-functor-nr (+ current-functor-nr 1))
	   (set! functor-list (cons current-functor functor-list)))
	 (set! current-functor (get-functor symbol 0 functor-list)))
     (if (not (get-signature symbol 0 current-signature))
	 (set! current-signature (cons current-functor current-signature)))
     (if (not (assq symbol constant-list))
	 (set! constant-list (cons (cons symbol (list current-functor)) constant-list)))
     (cdr (assq symbol constant-list)))
    ((variable)
     (if (variable-exists? variable current-clause-variables)
	 (begin (set! current-variable 
		      (get-variable variable current-clause-variables))
		(increment-occurences variable current-clause-variables 
				      current-clause-var-occurrences))
	 (begin
	   (set! current-variable (new-variable))
	   (set-variable-name current-variable variable)
	   (set! current-clause-var-occurrences
		 (cons 1 current-clause-var-occurrences))
	   (set! current-clause-variables 
		 (append! current-clause-variables (list current-variable)))))
      current-variable)
    ((symbol parameters) 
     (if (not (get-functor symbol (length parameters) functor-list))
	 (begin
	   (set! current-functor (new-functor))
	   (set-functor-symbol current-functor symbol)
	   (set-functor-arity  current-functor (length parameters))
	   (set-functor-number current-functor current-functor-nr)
	   (set-functor-weight current-functor 2)
	   (set! current-functor-nr (+ current-functor-nr 1))
	   (set! functor-list (cons current-functor functor-list)))
	 (set! current-functor (get-functor symbol (length parameters) functor-list)))
     (if (not (get-signature symbol (length parameters) current-signature))
	 (set! current-signature (cons current-functor current-signature)))
     (append (list current-functor) parameters)))
   (parameters
    ((openpar termlist closepar) termlist))
   (termlist
    ((termlist comma term) (append termlist (list term)))
    ((term) (list term)))
   (type
    ((symbol) 
     (cond ((eq? symbol (string->symbol "axiom"))
            (set! nr-axioms (+ nr-axioms 1)) 'axiom)
	   ((eq? symbol (string->symbol "lemma"))
	    (set! nr-lemmas (+ nr-lemmas 1)) 'lemma)
	   ((eq? symbol (string->symbol "hypothesis"))
	    (set! nr-hypotheses (+ nr-hypotheses 1)) 'hypothesis)
	   ((eq? symbol (string->symbol "conjecture"))
	    (set! nr-conjectures (+ nr-conjectures 1)) 'conjecture))
	   ))
   (clausenamepars
    ((openpar clausenametl closepar) clausenametl))
   (clausenametl
    ((clausenametl comma clausename) (append clausenametl (list clausename)))
    ((clausename) (list clausename)))
   (clausename
    ((openpar clausename closepar) clausename)
    ((symbol) symbol)
    ((variable) variable)
    ((symbol clausenamepars) 
     (append (list symbol) clausenamepars)))))


(define (insert-new-clause-in-clauselist clause clauses)
  (if (null? clauses)
      (list clause) ;; initialize previously empty list
      (begin
	(if (not (clause< clause (car clauses)))
	  (begin
	    (set-cdr! clauses (cons (car clauses) (cdr clauses)))
	    (set-car! clauses clause))
	  (if (null? (cdr clauses))
	      (set-cdr! clauses (list clause))
	      (insert-new-clause-in-clauselist clause (cdr clauses))))
	clauses)))

(define (clause< c d)
  (if (= (type-rating c) (type-rating d))
      (if (= (input-length (clause-input c)) (input-length (clause-input d)))
	  (if (= (input-complexity (clause-input c)) (input-complexity (clause-input d)))
	      (if (< (clause-nr-variables c) (clause-nr-variables d))
		  #t
		  #f)
	      (if (< (input-complexity (clause-input c)) (input-complexity (clause-input d)))
		  #t
		  #f))
	  (if (< (input-length (clause-input c)) (input-length (clause-input d)))
	      #t
	      #f))
      (if (< (type-rating c) (type-rating d))
	  #t
	  #f)))

(define (type-rating clause)
  (if norel
      (if (all-negative (clause-subgoals clause))
	  0
	  1)
      (let ((type (input-type (clause-input clause))))
	(cond
	 ((eq? type 'conjecture) 0)
	 ((eq? type 'lemma) 1)
	 (else 2)))))

(define (set-all-negatives-relevant clauses)
  (if (pair? clauses)
      (begin
	(if (all-negative (clause-subgoals (car clauses)))
	    (set-clause-relevant (car clauses) #t)
	    (set-clause-relevant (car clauses) #f))
	(set-all-negatives-relevant (cdr clauses)))))

(define (all-negative subgoals)
  (if (null? subgoals)
      #t
      (if (eq? '- (subgoal-sign (car subgoals)))
	  (all-negative (cdr subgoals))
	  #f)))

(define (set-clause-depths clauses)
  (if (null? clauses)
      0
      (begin
	(set-clause-depth (car clauses) (- (set-clause-depths (cdr clauses)) 1))
	(clause-depth (car clauses)))))

(define (get-nr-current-singletons occurrences-list)
  (if (null? occurrences-list)
      0
      (if (= (car occurrences-list) 1)
	  (+ (get-nr-current-singletons (cdr occurrences-list)) 1)
	  (get-nr-current-singletons (cdr occurrences-list)))))

(define (get-max-pred-arity predlist)
  (if (null? predlist)
      '-
      (let ((tmp (get-max-pred-arity (cdr predlist))))
	(if (or (eq? tmp '-) (< tmp (predsym-arity (car predlist))))
	    (predsym-arity (car predlist))
	    tmp))))

(define (get-min-pred-arity predlist)
  (if (null? predlist)
      '-
      (let ((tmp (get-min-pred-arity (cdr predlist))))
	(if (or (eq? tmp '-) (> tmp (predsym-arity (car predlist))))
	    (predsym-arity (car predlist))
	    tmp))))

(define (get-max-func-arity funclist)
  (if (null? funclist)
      '-
      (let ((tmp (get-max-func-arity (cdr funclist))))
	(if (or (eq? tmp '-) (< tmp (functor-arity (car funclist))))
	    (functor-arity (car funclist))
	    tmp))))

(define (get-min-func-arity funclist)
  (if (null? funclist)
      '-
      (let ((tmp (get-min-func-arity (cdr funclist))))
	(if (or (eq? tmp '-) (> tmp (functor-arity (car funclist))))
	    (functor-arity (car funclist))
	    tmp))))

(define (get-predsym symbol arity predicate-list)
  (if (null? predicate-list)
      #f
      (if (and (eq? symbol (predsym-symbol (car predicate-list)))
	       (eq? arity (predsym-arity (car predicate-list))))
	  (car predicate-list)
	  (get-predsym symbol arity (cdr predicate-list)))))

(define (get-functor symbol arity functor-list)
  (if (null? functor-list)
      #f
      (if (and (eq? symbol (functor-symbol (car functor-list)))
	       (eq? arity (functor-arity (car functor-list))))
	  (car functor-list)
	  (get-functor symbol arity (cdr functor-list)))))

(define (get-signature symbol arity signature-list)
  (if (null? signature-list)
      #f
      (if (and (eq? symbol (signature-symbol (car signature-list)))
	       (eq? arity (signature-arity (car signature-list))))
	  (car signature-list)
	  (get-signature symbol arity (cdr signature-list)))))


; (define (insert-literal literal lits)
;   (if (pair? lits)
;       (if (eq? (subgoal-sign literal) (subgoal-sign (car lits)))
; 	  (if (eq? (subgoal-pred literal) (subgoal-pred (car lits)))
; 	      (cons (car lits) (insert-literal literal (cdr lits)))
; 	      (if (> (predsym-number (subgoal-pred literal)) (predsym-number (subgoal-pred (car lits))))
; 		  (cons (car lits) (insert-literal literal (cdr lits)))
; 		  (cons literal lits)))
; 	  (if (eq? (subgoal-sign literal) '-) 
; 	      (cons (car lits) (insert-literal literal (cdr lits)))
; 	      (cons literal lits)))
;       (cons literal lits)))

(define (insert-literal literal lits)
  (if (pair? lits)
      (if (sign-predsym< literal (car lits))
	  (cons literal lits)
	  (cons (car lits) (insert-literal literal (cdr lits))))
      (cons literal lits)))


(define (display-statistics)
  (display #"Syntax:\t")
  (display #"Number of clauses:\t")
  (display nr-clauses)
  (display " ( ")
  (display nr-non-horn)
  (display " non-Horn; ")
  (display nr-units)
  (display " unit)")
  (newline)
  (display #"\t")
  (display #"Number of literals:\t")
  (display nr-literals)
  (display " ( ")
  (display nr-eq-literals)
  (display " equality)")
  (newline)
  (display #"\t")
  (display #"Maximal clause size:\t")
  (display max-clause-length)
  (newline)
  (display #"\t")
  (display #"Number of predicates:\t")
  (display nr-predicates)
  (display " ( ")
  (display nr-propositionals)
  (display " propositional; ")
  (display min-pred-arity)
  (display "-")
  (display max-pred-arity)
  (display " arity)")
  (newline)
  (display #"\t")
  (display #"Number of functors:\t")
  (display nr-functors)
  (display " ( ")
  (display nr-constants)
  (display " constant; ")
  (display min-func-arity)
  (display "-")
  (display max-func-arity)
  (display " arity)")
  (newline)
  (display #"\t")
  (display #"Number of variables:\t")
  (display nr-variables)
  (display " ( ")
  (display nr-singleton-variables)
  (display " singleton)")
  (newline)
  (display #"\t")
  (display #"Maximal term depth:\t")
  (display input-term-depth)
  (newline))

(define (display-for-debug)
  ;;(display (the-string))
  #t)

(define (reset-parser)
  (set! current-clause '())
  (set! current-clause-nr 0)
  (set! current-clause-variables '())
  (set! current-clause-var-occurrences '())
  (set! current-variable '())
  (set! current-literal '())
  (set! current-predicate '())
  (set! current-pred-nr 1)
  (set! current-functor '())
  (set! current-functor-nr 1) 
  (set! current-signature '())
  (set! predicate-list '())
  (set! functor-list '())
  (set! models '())
  (set! propositional-list '())
  (set! constant-list '())
  (set! current-positive-subgoals 0)
  (set! nr-clauses 0)
  (set! nr-non-horn 0)
  (set! nr-units 0)
  (set! nr-literals 0)
  (set! nr-eq-literals 0)
  (set! max-clause-length 0)
  (set! nr-predicates 0)
  (set! nr-propositionals 0)
  (set! max-pred-arity 0)
  (set! min-pred-arity 0)
  (set! nr-functors 0)
  (set! nr-constants 0)
  (set! max-func-arity 0)
  (set! min-func-arity 0)
  (set! nr-variables 0)
  (set! nr-singleton-variables 0))
