(module select
	(include "dctp.sch")
	(export linked-terms-ground)
	(export (select-subgoal subgoals max-subgoal) (compute-subgoal-weights subgoals)
		(termlist-symbol-count terms depth) (compute-link-complexity sg predsg) (current-subgoal-weight subgoal))
	(import (globals "globals.scm") (unification "unification.scm")))

(define linked-terms-ground #t)

(define (term-instantiation term)
; computes instantiation value of a term as follows:
;           |  0                          if t is a new variable
; t-i(t) =  {  1                          if t is a constant or a non-new variable
;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
;
 (cond ((not term) 1000)
       ((variable term) (bind term #f) 0)
       ((constant term) 1000)
       (else (+ 500 (quotient (termlist-instantiation (cdr term)) 
                              (* 2 (functor-arity (car term))))))))

(define (termlist-instantiation termlist)
 (if (null? (cdr termlist))
     (term-instantiation (ref-ext (car termlist)))
     (+ (term-instantiation (ref-ext (car termlist)))
        (termlist-instantiation (cdr termlist)))))

(define (ref-ext ext-term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
 (if (or (not ext-term) (non-variable ext-term) (free-variable ext-term))
     ext-term
     (ref-ext (car ext-term))))

(define (subgoal-instantiation subgoal)
  (if (null? (subgoal-terms subgoal))
      1000
      (quotient (termlist-instantiation (subgoal-terms subgoal))
		(functor-arity (subgoal-pred subgoal)))))

(define (compute-subgoal-weights subgoals)
  (if (pair? subgoals)
      (if complexity-measure
	  (begin
	    (set-subgoal-weight (car subgoals) (termlist-symbol-count (subgoal-terms (car subgoals)) 1))
	    (compute-subgoal-weights (cdr subgoals)))
	  (begin
	    (trail-stop)
	    (set-subgoal-weight (car subgoals) (subgoal-instantiation (car subgoals)))
	    (unbind-vars)
	    (compute-subgoal-weights (cdr subgoals))))))
	  
(define (current-subgoal-weight subgoal)
  (if complexity-measure
      (termlist-symbol-count (subgoal-terms subgoal) 1)
      (let (weight)
	(trail-stop)
	(set! weight (subgoal-instantiation subgoal))
	(unbind-vars)
	weight)))
	  
(define (select-subgoal subgoals max-subgoal)
  (if (pair? subgoals)
      (if (subgoal-selected (car subgoals))
	  (select-subgoal (cdr subgoals) max-subgoal)
	  (if (or (not max-subgoal) 
		  ;;(and (not min-sg-select) (> (subgoal-weight (car subgoals)) (subgoal-weight max-subgoal)))
		  ;;(and min-sg-select (< (subgoal-weight (car subgoals)) (subgoal-weight max-subgoal))))
		  (and (not min-sg-select) (subgoal-> (car subgoals) max-subgoal))
		  (and min-sg-select (subgoal-< (car subgoals) max-subgoal)))
	      (select-subgoal (cdr subgoals) (car subgoals))
	      (select-subgoal (cdr subgoals) max-subgoal)))
      (if max-subgoal
	  (begin
	    (set-subgoal-selected max-subgoal #t)
	    max-subgoal)
	  #f)))

(define (subgoal-> sga sgb)
  (if (= (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))
      (if (= (subgoal-weight sga) (subgoal-weight sgb))
	  (if (eq? (subgoal-sign sga) (subgoal-sign sgb))
	      (if (= (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb))) 
		  (terms-earlier-instantiated (subgoal-terms sga) (subgoal-terms sgb))
		  (> (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb))))
	      (if (eq? (subgoal-sign sga) '-)
		  #t
		  #f))
	  (> (subgoal-weight sga) (subgoal-weight sgb)))
      (< (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))))

(define (subgoal-< sga sgb)
  (if (= (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))
      (if (= (subgoal-weight sga) (subgoal-weight sgb))
	  (if (eq? (subgoal-sign sga) (subgoal-sign sgb))
	      (if (= (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb))) 
		  (terms-earlier-instantiated (subgoal-terms sgb) (subgoal-terms sga))
		  (< (predsym-arity (subgoal-pred sga)) (predsym-arity (subgoal-pred sgb))))
	      (if (eq? (subgoal-sign sga) '+)
		  #t
		  #f))
	  (< (subgoal-weight sga) (subgoal-weight sgb)))
      (< (predsym-weight (subgoal-pred sga)) (predsym-weight (subgoal-pred sgb)))))

(define (terms-earlier-instantiated terms-a terms-b)
  (if (pair? terms-a)
      (let ((result (term-more-instantiated (ref (car terms-a)) (ref (car terms-b)))))
	(if (eq? result 'eq-instantiated)
	    (terms-earlier-instantiated (cdr terms-a) (cdr terms-b))
	    result))
      #f))

(define (term-more-instantiated k l)
  (if (variable k)
      (if (variable l)
	  'eq-instantiated
	  #f)
      (if (variable l)
	  #t
	  (if (= (functor-weight (car k)) (functor-weight (car l)))
	      (if (= (functor-arity (car k)) (functor-arity (car l)))
		  (if (constant k)
		      'eq-instantiated
		      (terms-earlier-instantiated (cdr k) (cdr l)))
		  (if (< (functor-arity (car k)) (functor-arity (car l)))
		      #t
		      #f))
	      (< (functor-weight (car k)) (functor-weight (car l)))))))


(define (termlist-symbol-count terms depth)
  (if (pair? terms)
      (+ (term-symbol-count (ref (car terms)) depth) (termlist-symbol-count (cdr terms) depth))
      (begin
	(if (> current-depth max-term-depth)
	    (set! max-term-depth current-depth))
	0)))

(define (term-symbol-count term depth)
  (if (> depth current-depth)
      (set! current-depth depth))
  (if (variable term)
      (begin
	(set! complexity (+ complexity 1))
	(set! linked-terms-ground #f)
	1)
      (if (constant term)
	  (begin
	    (set! complexity (+ complexity 1))
	    0)
	  (begin
	    (set! complexity (+ complexity 1))
	    (+  (* 2 depth depth) (termlist-symbol-count (cdr term) (+ depth 1)))))))

(define (compute-link-complexity sg predsg)
  (+ (compute-link-clause-complexity (clause-subgoals (subgoal-clause sg)) sg)
     (compute-link-clause-complexity (clause-subgoals (subgoal-clause predsg)) predsg)))

(define (compute-link-clause-complexity subgoals linksg)
  (if (pair? subgoals)
      (if (or (subgoal-simplified (car subgoals)) (eq? (car subgoals) linksg))
          (compute-link-clause-complexity (cdr subgoals) linksg)
	  (+ (bit-lsh (termlist-symbol-count (subgoal-terms (car subgoals)) 1) (predsym-weight (subgoal-pred (car subgoals))))
	     (compute-link-clause-complexity (cdr subgoals) linksg)))
      0))

; (define (compute-link-clause-complexity subgoals linksg)
;   (if (pair? subgoals)
;       (if (or (subgoal-simplified (car subgoals)) (eq? (car subgoals) linksg))
; 	  (compute-link-clause-complexity (cdr subgoals) linksg)
; 	  (+ (bit-lsh (termlist-symbol-count (subgoal-terms (car subgoals))) (predsym-weight (subgoal-pred (car subgoals))))
; 	     (compute-link-clause-complexity (cdr subgoals) linksg)))
;       0))
