(module dctp
	(main call-dctp) ;; this is the main procedure
	(include "dctp.sch")
	(export (variables-in-subgoals subgoals vars) (clause-pointers-in-subgoals subgoals clause) (clause-subsumed clause)
		(unrename-vars) (ground-closed subgoal predsubgoal rel-vector) (renamed-copy clause unsimplifieds)
		(subgoal-in-copy subgoal copy-subgoals) (free-vars-in-term term vars) (free-vars-in-termlist terms vars)
		(eliminate-var-disequations subgoals) (set-clauses-input clauses input) (ground-regular clause)
		(set-clause-pointers-in-clauses clauses) (compute-current-weights subgoals) (clause-index-variant? clause)
		(renamed-and-splitted-copy clause unsimplifieds lower-upper special-sg depth) (clauses-path-vector-set clauses)
		(unrename-vars-returning-renamings) (check-for-local-vars subgoal) (display-local-instances instances))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm") (pruning "pruning.scm") 
		(read-tptp input "input.scm") (index "index.scm") (preprocess "preprocess.scm") (select "select.scm") 
		(signals "signals.scm") (path "path.scm") (link "link.scm") (connection "connection.scm") 
		(initial-path "initial-path.scm") (normal-form "normal-form.scm") (cputime "cputime.scm") (kbo "kbo.scm")
		(parser "parser.scm") (splitting "splitting.scm") (disagreement "disagreement.scm")
		(rewrite-subgoal rewrite "rewrite.scm") (subsumption "subsumption.scm")))

(define (call-dctp argv) 
  (set-signal-handlers)
  (dctp (cdr argv)))

(define (dctp args)
  (let ((initial-subgoal #f) host)
    (set! start (getcputime))
    (pre-parser-initialize)
    (set! filename (car (reverse (get-flags args))))
    (set! preprocessing #t)
    (read-tptp filename)
    (set! host (getenv "HOST"))
    (if (not silent)
	(begin
	  (display "running dctp on problem file ")
	  (display (basename filename))
	  (if host
	      (begin
		(display " on host ")
		(display host)))
	  (newline)))
    (if equality 
	(clauses-check-orientation (formula-clauses formula))
	(clauses-orientate-disequations (formula-clauses formula)))
    (initialize-prover)
    (if (not (or (preprocess-formula) preprocess-only))
	(begin
	  (if (not silent)
	      (begin
		(get-formula-statistics)
		(display "problem statistics for file ")
		(display filename)
		(newline)
		(display-formula-statistics)))
	  (if preprocess
	      (print "complementary unit pair generated by formula preprocessing.")
	      (print "complementary unit pair found in input."))
	  (set! end (getcputime))
	  (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
	  (print "**************** Exit status is success (PROOF found in " elapsed " seconds) ****************")
	  (exit 0)))
    (set! preprocessing #f)
    (set-formula-clauses formula (sort-preprocessed-clauses (formula-clauses formula)))
    (get-formula-statistics)
    (check-if-path-indexes-empty)
    ;;(initialize-prover)
    (if preprocess-only
	(begin
	  (display-tptp-clauses (formula-clauses formula))
	  (exit 0)))
    (if (not silent)
	(begin
	  (display "problem statistics for file ")
	  (display filename)
	  (newline)
	  (display-formula-statistics)))
    (if (= nr-all-negatives 0)
	(begin
	  (set! end (getcputime))
	  (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
	  (print "!!!!!!!!!!!!!!!! Exit status is success (MODEL found in " elapsed " seconds) !!!!!!!!!!!!!!!!")
	  (exit 0)))
    (if dynamic-split
	(set! split #t)
	(set! split #f))
    ;;(select-initial-path (formula-clauses formula))
    (if t-closure (select-closure-constant))
    (if dewey-numbers
	(print "initial active path:"))
    (set! initial-subgoal (initial-active-path (formula-clauses formula) #f))
    (if initial-subgoal
	(begin
	  (init-link-bias)
	  (trail-stop)
	  (if (make-initial-link-list initial-subgoal)
	      (begin
		(trail-erase)
		(if (> debug-level 4)
		    (begin
		      (print "done with initial links")))
		(if limit-term-depth
		    (begin
		      (set! allowed-term-depth min-relevant-depth)
		      (display "limiting term complexity to ")
		      (display allowed-term-depth)
		      (newline)))
		(if (> debug-level 4)
		    (begin
		      (print-predicates-indexes predicate-list)
		      (display "initial links: ")
		      (newline)
		      (display-link-vector link-vector)
		      (newline)
		      (display "initial eq-links: ")
		      (newline)
		      (display-link-vector eq-link-vector)
		      (newline)))
		(if test-mode
		    (begin
		      (print "entering test mode")
		      (set! debug-level 9))
		    (begin
		      (if (not silent) (print "Starting proof"))
		      (if (solve-initial-subgoal initial-subgoal)
			  (begin
			    (if (> debug-level 4)
				(print-predicates-indexes predicate-list))
			    (set! end (getcputime))
			    (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
			    (print "**************** Exit status is success (PROOF found in " elapsed " seconds) ****************"))
			  (begin
			    (set! end (getcputime))
			    (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
			    (print "!!!!!!!!!!!!!!!! Exit status is success (MODEL found in " elapsed " seconds) !!!!!!!!!!!!!!!!")))
		      (print-statistics)
		      (exit 0))))
	      (begin
		(trail-erase)
		(print "equality closure in input set.")
		(set! end (getcputime))
		(set! elapsed (/fl (exact->inexact (- end start)) 100.0))
		(print "**************** Exit status is success (PROOF found in " elapsed " seconds) ****************")
		(exit 0))))
	(begin
	  (print "Empty input set.")
	  (if error-on-empty
	      (begin
		(print "Exit status is ERROR (empty input set)")
		(exit 1))
	      (begin
		(set! end (getcputime))
		(set! elapsed (/fl (exact->inexact (- end start)) 100.0))
		(print "!!!!!!!!!!!!!!!! Exit status is success (MODEL found in " elapsed " seconds) !!!!!!!!!!!!!!!!")
		(exit 0)))))))

(define (solve-subgoals subgoals rel-vector)
  ;; iterative version, recursive version had to be abandoned to save stack space
  (let (subgoal-rel-vector (subgoal (select-subgoal subgoals)))
;     (if (eq? (subgoal-depth (car subgoals)) 1115)
; 	(begin
; 	  (display "solving and inserting clause ")
; 	  (display-clause (subgoal-clause (car subgoals)))
; 	  (newline)))
    (if variant-index (insert-clause-in-index (subgoal-clause (car subgoals))))
    (if subgoal
	(set! subgoal-rel-vector (solve-subgoal subgoal)))
    (do ()
	((or (not subgoal) (not subgoal-rel-vector) (and pruning (not (subgoal-relevant subgoal subgoal-rel-vector)))) #t)
      (merge-rel-vectors rel-vector subgoal-rel-vector (subgoal-depth subgoal))
      (set! subgoal (select-subgoal subgoals))
      (if subgoal
	  (set! subgoal-rel-vector (solve-subgoal subgoal))))
    (if subgoal 
	(if (and subgoal-rel-vector pruning (not (subgoal-relevant subgoal subgoal-rel-vector)))
	    (begin
	      (if dewey-numbers
		  (begin
		    (display "pruning clause ") 
		    (display-clause (subgoal-clause subgoal))
		    (newline)))
	      (clear-local-instances-of-subgoals subgoals)
	      (set! nr-prunings (+ nr-prunings 1))))
	(begin
	  (if (= 1 (clause-unsimplified (subgoal-clause (car subgoals))))
	      (let ((tmp-result #f) (sg (get-unsimplified-subgoal subgoals)))
		(if (and top-down-lemmas (subgoal? sg) (not (subgoal-unlinkable sg)))
		    (begin
		      (if (or dewey-numbers (> debug-level 4))
			  (begin
			    (display "leftover lemma ")
			    (display-subgoal sg)
			    (display " found after solving ")
			    (display-clause (subgoal-clause (car subgoals)))
			    (newline)))
		      (set-subgoal-unlinkable sg #t)
		      (set! tmp-result (make-top-down-lemma sg))
		      (if (and pruning tmp-result)
			  (begin
			    (set! rel-vector (new-rel-vector (subgoal-depth sg)))
			    (mark-rel-vector rel-vector tmp-result)
			    (if dewey-numbers (print "new lemma causes pruning to level " tmp-result))
			    (if (> debug-level 4)
				(begin
				  (display "referenced levels:")
				  (display-rel-vector rel-vector)
				  (newline)))))))))
	  (set! subgoal-rel-vector rel-vector)))
;     (if (eq? (subgoal-depth (car subgoals)) 1115)
; 	(begin
; 	  (display "removing clause ")
; 	  (display-clause (subgoal-clause (car subgoals)))
; 	  (newline)))
    (if (eq? (subgoal-clause (car subgoals)) (input-in-tableau (clause-input (subgoal-clause (car subgoals)))))
	(set-input-in-tableau (clause-input (subgoal-clause (car subgoals))) #f))
    (if context-lemmas (delete-clause-foldup-lits (subgoal-clause (car subgoals))))
    (if variant-index (delete-clause-from-index (subgoal-clause (car subgoals))))
    (clear-clause (subgoal-clause (car subgoals)))
    subgoal-rel-vector))

(define (get-unsimplified-subgoal subgoals)
  (if (pair? subgoals)
      (if (subgoal-simplified (car subgoals))
	  (get-unsimplified-subgoal (cdr subgoals))
	  (car subgoals))
      #f))

(define (clear-clause clause)
  (clear-subgoals (clause-subgoals clause))
  (path-vector-set (clause-depth clause) #f)
  (vector-fill! clause #f))

(define (clear-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(if (not (eq? (car subgoals) falsum)) (vector-fill! (car subgoals) #f))
	(clear-subgoals (cdr subgoals)))))

(define (clear-local-instances-of-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	;; this doesnt't look too good. if problems ever occur here use "let" to introduce temp vars
	(if (subgoal? (subgoal-local-instances (car subgoals)))
	    (begin
	      (if dewey-numbers
		  (begin
		    (display "deleting subgoal ")
		    (display-subgoal (car subgoals))
		    (display " on depth ")
		    (display (subgoal-depth (car subgoals)))
		    (display " from instances of subgoal ")
		    (display-subgoal (subgoal-local-instances (car subgoals)))
		    (display " on depth ")
		    (display (subgoal-depth (subgoal-local-instances (car subgoals))))
		    (newline)))
	      (set-subgoal-local-instances (subgoal-local-instances (car subgoals))
					   (remq! (assq (car subgoals) (subgoal-local-instances (subgoal-local-instances (car subgoals))))
						  (subgoal-local-instances (subgoal-local-instances (car subgoals)))))))
	(clear-local-instances-of-subgoals (cdr subgoals)))))

(define (solve-initial-subgoal subgoal)
  (let ((rel-vector (new-rel-vector 0)))
    (if (ground-closed subgoal (subgoal-predecessor subgoal) rel-vector)
	rel-vector
	(begin
	  (try-link-with-subgoal subgoal rel-vector)))))

(define (solve-subgoal subgoal)
  ;; NB: As of now, it is assumed that all links in the link list are unifiable.
  ;;     If they are not, solve-subgoal will return with an unsound #t
  (let ((tmp-result #f) (solved #f) (gr-inserted #f) (ground #f)
	(rel-vector (new-rel-vector (subgoal-depth subgoal))))
    (if (> (subgoal-depth subgoal) max-depth) (set! max-depth (subgoal-depth subgoal)))
    (if dewey-numbers
	(begin
	  (newline)
	  (if (subgoal-next subgoal)
	      (display "descending over subgoal ")
	      (display "solving subgoal "))
	  (display-subgoal subgoal) (display " [")
	  (display (clause-unsimplified (subgoal-clause subgoal))) (display "] on depth ")
	  (display (subgoal-depth subgoal)) (newline)))
    (if (eq? falsum subgoal)
	(begin
	  (if dewey-numbers
	      (print "empty clause deduced by unit simplification!"))
	  rel-vector)
	(let ((rewritten #f) (index-variant #f) unsimplifiables)
	  (if (and equality ordered-eq (or full-rewrite (and eq-rewrite (eq-subgoal? subgoal))))
	      (begin
		(delete-clause-from-index (subgoal-clause subgoal))
		(set! rewritten (rewrite-subgoal subgoal))
		;;(if rewritten (set! index-variant (clause-index-variant? (subgoal-clause (subgoal)))))
		(if rewritten 
		    (begin
		      (set! unsimplifiables (unsimplifiable-subgoals (clause-subgoals (subgoal-clause subgoal))))
		      (set-subgoal-weight subgoal (current-subgoal-weight subgoal))))
		(insert-clause-in-index (subgoal-clause subgoal))))
	  (if (or (clause-subsumed (subgoal-clause subgoal))
 		  (and rewritten full-variants (set-current-weights unsimplifiables) (clause-variant-subsumed unsimplifiables)))
	      (begin
		(set! rewritten #t)
		(set-subgoal-unlinkable subgoal #t)
		(set-subgoals-selected (clause-subgoals (subgoal-clause subgoal))))
	      (set! rewritten #f))
	  (if (and (not (subgoal-unlinkable subgoal)) top-down-lemmas (simplified-unit subgoal))
	      (begin
		(set-subgoal-unlinkable subgoal #t)
		(set! tmp-result (make-top-down-lemma subgoal))))
	  (if (and local-vars (not equality) (not tmp-result) (not (subgoal-unlinkable subgoal)) (not (subgoal-local subgoal)) (not (subgoal-ground subgoal))
		   (exists-simplified-side-subgoal subgoal (clause-subgoals (subgoal-clause subgoal))))
	      (check-for-local-vars subgoal))
	  (if (and (> debug-level 5) (subgoal-ground subgoal) (not (subgoal-local subgoal)))
	      (begin
		(display "subgoal ")
		(display-subgoal subgoal)
		(display " ground but not local!")
		(newline)))
	  (if (and pruning tmp-result)
	      (begin
		(mark-rel-vector rel-vector tmp-result)
		(if dewey-numbers (print "new lemma causes pruning to level " tmp-result))
		(if (> debug-level 4)
		    (begin
		      (display "referenced levels:")
		      (display-rel-vector rel-vector)
		      (newline)))
		rel-vector)
	      (if (and (not rewritten) (subgoal-simplifiable subgoal)) ;; simplification is useless once the clause has been
		  (begin                                               ;; made redundant
		    (mark-rel-vector rel-vector (subgoal-depth subgoal))
		    rel-vector)
		  ;; Beware: Ground closures can happen even if the subgoal has been made a simplification lemma:
		  ;; Just consider the case of path subgoal p(X,Y) and new lemma -p(X,X). The lemma does not simplify
		  ;; the path subgoal yet it is ground closed.
		  ;; In the other case the current clause is subsumed after the rewriting of subgoal in which case
		  ;; subgoal must simply be ignored (all subgoals of the current clause were marked as solved, a closure of
		  ;; this subgoal would mean an incorrect fold-up of the predecessor subgoal)
		  (if (and (not rewritten) (ground-closed subgoal (subgoal-predecessor subgoal) rel-vector))
		      rel-vector
		      (begin
			(if (not (subgoal-unlinkable subgoal))
			    (begin
			      (if (and dewey-numbers (not equality) local-vars (not (subgoal-ground subgoal)))
				  (if (subgoal-local subgoal)
				      (begin
					(display "LOCAL subgoal ")
					(display-subgoal subgoal)
					(display " on depth ")
					(display (subgoal-depth subgoal))
					(newline))
				      (if (subgoal-semi-local subgoal)
					  (begin
					    (display "semi-local subgoal ")
					    (display-subgoal subgoal)
					    (display " on depth ")
					    (display (subgoal-depth subgoal))
					    (newline)))))
			      (insert-subgoal-in-gr-index subgoal)
			      (set! ground (and regularity (subgoal-ground subgoal)))
			      (if ground
				  (insert-subgoal-in-reg-index subgoal))
			      (set! gr-inserted #t)))
			(if (> debug-level 5) (print-indexes (subgoal-pred subgoal) 'gr))
			(if (and index-info (> successful-inferences 0) (= 0 (remainder successful-inferences index-info)))
			    (begin
			      (print "index statistics after " successful-inferences " linkings on depth " (subgoal-depth subgoal))
			      (print "each field has format index-type entries total-paths empty-paths avg-depth avg-branch max-branch")
			      (predicate-index-statistics predicate-list)))
			(if (and variant-info (> successful-inferences 0) (= 0 (remainder successful-inferences variant-info)))
			    (begin
			      (print "variant index statistics after " successful-inferences " linkings on depth " (subgoal-depth subgoal))
			      (print "each field has format index-type entries total-paths empty-paths avg-depth avg-branch max-branch")
			      (variant-index-statistics (formula-clauses formula))))
			(if (not (subgoal-unlinkable subgoal))
			    (begin
			      (if (and (not disagreement) equality (eq-subgoal? subgoal))
				  (if (equation? subgoal)
				      (begin
					(if (> debug-level 4) (print "inserting subgoal eq-links"))
					(if ordered-eq
					    (insert-eq-links-for-new-equation subgoal subgoal)
					    (insert-eq-links-for-new-equation subgoal (subgoal-predecessor subgoal)))))
				    (if (and equality disagreement (eq-subgoal? subgoal))
					(if (disequation? subgoal)
					    (begin
					      (if (> debug-level 4)
						  (print "inserting decomposition link"))
					      (insert-decomposition-link subgoal)
					      (if (> debug-level 4)
						  (print "inserting path disagreement links"))
					      (disequation-insert-links subgoal (potential-linked-lits subgoal))
					      (if (> debug-level 4)
						  (print "inserting lemma disagreement links"))
					      (disequation-insert-links subgoal (potential-linked-lemmas subgoal)))
					    (begin
					      (if (> debug-level 4)
						  (print "inserting path disagreement links"))
					      (equation-insert-links subgoal (potential-linked-lits subgoal))
					      (if (> debug-level 4)
						  (print "inserting lemma disagreement links"))
					      (equation-insert-links subgoal (potential-linked-lemmas subgoal))))
					(begin
; 					  (if (> debug-level 4)
; 					      (print "inserting path links"))
; 					  (insert-links subgoal (potential-linked-lits subgoal))
; 					  (if (> debug-level 4)
; 					      (print "inserting lemma links"))
; 					  (insert-links subgoal (potential-linked-lemmas subgoal)))))
					  (if (> debug-level 4)
					      (print "inserting path and lemma links"))
					  (generate-links-for-subgoal subgoal))))
			      ;; NOTA BENE: Sign selection does not preserve completeness. The condition below
			      ;; must be replaced by (if (and (not disagreement) equality)...) only.
			      ;;(if (and (not disagreement) equality (or ordered-eq (disequation? subgoal)
								       ;;(and (not (eq-subgoal? subgoal)) 
									    ;;(eq? (subgoal-sign subgoal) selected-sign))))
			      (if (and (not disagreement) equality)
				  (begin
				    (if (> debug-level 4)
					(print "inserting path eq-links"))
				    (insert-eq-links subgoal (positive-eq-lits))
				    (if (> debug-level 4)
					(print "inserting lemma eq-links"))
				    (insert-eq-links subgoal (positive-eq-lemmas))))
			      (if (and equality (not disagreement) (disequation? subgoal))
				  (begin
				    (if (> debug-level 4)
					(print "checking for reflexivity link"))
				    (insert-reflexivity-link subgoal)))
			      ;;(if factoring (insert-factoring-links subgoal))
; 			      (if (not equality)
; 				  (let ((index-links (unify-subgoal-with-index subgoal 'path)))
; 				    (check-links-with-index-links subgoal (subgoal-links subgoal) index-links)))
			      (if (equation? subgoal) (add-path-literal subgoal))
; 			      (if (eq? (subgoal-depth subgoal) 1115)
; 				  (begin
; 				    (display "inserting subgoal ")
; 				    (display-subgoal subgoal)
; 				    (display " on depth ")
; 				    (display (subgoal-depth subgoal))
; 				    (display " into path index")
; 				    (newline)))
			      (insert-subgoal-in-index subgoal 'path)
			      (if (subgoal-next subgoal)
				  (set! solved (solve-subgoals (clause-subgoals (subgoal-next subgoal)) rel-vector))
				  (set! solved (try-link-with-subgoal subgoal rel-vector)))
; 			    (if (eq? (subgoal-depth subgoal) 1115)
; 				(begin
; 				  (display "removing subgoal ")
; 				  (display-subgoal subgoal)
; 				  (display " on depth ")
; 				  (display (subgoal-depth subgoal))
; 				  (display " from path index")
; 				  (newline)))
			      (delete-subgoal-from-index subgoal 'path)
			      (if (equation? subgoal) (delete-path-literal subgoal)))
			    (if (subgoal-next subgoal)
				(set! solved (solve-subgoals (clause-subgoals (subgoal-next subgoal)) rel-vector))
				(set! solved (try-link-with-subgoal subgoal rel-vector))))
			(set-subgoal-next subgoal #f)
			(mark-links-deletable (subgoal-links subgoal) subgoal)
			(set-subgoal-links subgoal '())
			;; NB: replaced by new booelan local var gr-inserted
			;; The problem was that an inserted subgoal could be declared unlinkable
			;; a posteriori and thus would not be removed properly from the GR-index
			;;(if (not (subgoal-unlinkable subgoal)) 
; 			(if (eq? (subgoal-depth subgoal) 1115)
; 			    (begin
; 			      (display "removing subgoal ")
; 			      (display-subgoal subgoal)
; 			      (display " on depth ")
; 			      (display (subgoal-depth subgoal))
; 			      (display " from other indexes")
; 			      (newline)))
			(if gr-inserted 
			    (begin 
			      (delete-subgoal-from-gr-index subgoal)
			      (if ground
				  (delete-subgoal-from-reg-index subgoal))))
			(set! tmp-result #f)
			(if (and foldup solved (not (subgoal-simplified subgoal)) 
				 (subgoal-relevant subgoal solved))
			    (begin
			      (if (and (> debug-level 5) (ground-closed-with-index subgoal))
				  (begin
				    (display "subgoal a posteriori ground closed")
				    (newline)))
			      (set! tmp-result (foldup-subgoal subgoal solved))
			      (if (and dewey-numbers (not tmp-result))
				  (begin
				    (display "no pruning by new lemma, referenced levels:")
				    (display-rel-vector solved)
				    (newline)))))
; 			(if (eq? (subgoal-depth subgoal) 1115)
; 			    (begin
; 			      (display "done removing subgoal ")
; 			      (display-subgoal subgoal)
; 			      (display " on depth ")
; 			      (display (subgoal-depth subgoal))
; 			      (display " from indexes")
; 			      (newline)))
			(if (and pruning tmp-result)
			    (begin
			      (if (integer? tmp-result)
				  (begin
				    (set! rel-vector (new-rel-vector (subgoal-depth subgoal)))
				    (mark-rel-vector rel-vector tmp-result)
				    (if dewey-numbers (print "new lemma causes pruning to level " tmp-result)))
				  (begin
				    (set! rel-vector tmp-result)
				    (if dewey-numbers (print "new context lemma causes pruning to c-point"))))
			      rel-vector)
			    solved)))))))))

(define (try-link-with-subgoal subgoal rel-vector)
  (let ((link-cell  '())
	used-link-cells result-rel-vector new-subgoals)
    (do ((terminate #f))
	(terminate #t)
      ;; premature abort after defined max nr. of inferences
      (if (and inference-limit (> (+ total-eq-inferences total-inferences) inference-limit))
	  (begin
	    (set! end (getcputime))
	    (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
	    (print "Exit status is FAIL (inference limit exceeded after " elapsed " seconds)")
	    ;;(if tptp-lemmas (print-tptp-lemmas subgoal))
	    (print-statistics)
	    (exit 0)))
      (set! used-link-cells link-cell)
      (set! link-cell  (select-appropriate-link-cell))
      (if (pair? link-cell)
	  (begin 
	    (set-cdr! link-cell used-link-cells)
	    (set! used-link-cells link-cell)
	    (set-link-active (car link-cell) #f)
	    (if (eq-link? (car link-cell))
		(if (link-litb (car link-cell))
		    (if (and variant-pruning (clause-initial (subgoal-clause (link-lita (car link-cell))))
			     (input-in-tableau (clause-input (subgoal-clause (link-lita (car link-cell))))))
			(set! new-subgoals #f)
			(if (and variant-pruning (clause-initial (subgoal-clause (link-litb (car link-cell))))
				 (input-in-tableau (clause-input (subgoal-clause (link-litb (car link-cell))))))
			    (set! new-subgoals #f)
			    (set! new-subgoals (eq-linking-step (car link-cell) subgoal rel-vector))))
		    (set! new-subgoals (reflexivity-linking-step (car link-cell) subgoal rel-vector)))
		(if (disagreement-link? (car link-cell))
		    (set! new-subgoals (disagreement-linking-step (car link-cell) subgoal rel-vector))
		    (if (and variant-pruning (clause-initial (subgoal-clause (link-lita (car link-cell))))
			     (input-in-tableau (clause-input (subgoal-clause (link-lita (car link-cell))))))
			(set! new-subgoals #f)
			(if (and variant-pruning (clause-initial (subgoal-clause (link-litb (car link-cell))))
				 (input-in-tableau (clause-input (subgoal-clause (link-litb (car link-cell))))))
			    (set! new-subgoals #f)
			    (set! new-subgoals (linking-step (car link-cell) subgoal rel-vector))))))))
      (if (or (not (pair? link-cell)) new-subgoals)
	  (set! terminate #t)))
    (if (not (pair? link-cell))
	(begin 
	  (if (not silent) (print "Saturation state reached."))
	  (if (> debug-level 4)
	      (begin
		(print-predicates-indexes predicate-list)
		(display "final links: ")
		(newline)
		(display-link-vector link-vector)
		(newline)
		(display "final eq-links: ")
		(newline)
		(display-link-vector eq-link-vector)
		(newline)))
	  #f)
	(if (eq? falsum (car new-subgoals))
	    (begin
	      (if dewey-numbers
		  (print "empty clause deduced by unit simplification!"))
	      rel-vector)
	    (begin
	      (set! result-rel-vector (solve-subgoals new-subgoals rel-vector))
	      (restore-link-cells used-link-cells)
	      result-rel-vector)))))

(define (ground-closed subgoal predsubgoal rel-vector)
  ;; The predecessor subgoal must exist and it must not be in a clause in the initial active path!
  ;; t-closure is now also handled with the help of the gr-index
  (let ((index-sg (ground-closed-with-index subgoal)))
    (if (and index-sg (or (not (clause-initial (subgoal-clause subgoal))) 
			   (unit-subgoal subgoal)))
	(begin
	  (if (or dewey-numbers (> debug-level 4) (> (subgoal-depth index-sg) (subgoal-depth subgoal)))
	      (begin
		(display "GR-reduction-step for subgoal ")
		(display-subgoal subgoal)
		(display " on depth ") 
		(display (subgoal-depth subgoal))
		(if (context-lemma? index-sg)
		      (display " with context lemma ") 
		      (display " with path literal "))
		(display-subgoal index-sg)
		(if (context-lemma? index-sg)
		    (begin
		      (display " (")
		      (display-rel-vector (lemma-rel-vector index-sg))
		      (display ") ")))
		(display " on depth ") 
		(display (subgoal-depth index-sg))
		(newline)))
	  (if (and (not equality) local-vars (not (context-lemma? index-sg)) (subgoal-semi-local index-sg) (not (subgoal-ground index-sg)))
	      (begin
;	      (let (local-instances)
; 		(trail-stop)
; 		(set! local-instances (gr-unify-lists-gathering-local-bindings (subgoal-terms index-sg) (subgoal-terms subgoal)))
; 		(if local-instances
; 		    (begin
; 		      (if (pair? local-instances)
; 			  (begin
			    ;;(trail-stop)
			    ;;(set! local-instances (renamed-copy-of-assoc-list local-instances))
			    ;;(unrename-vars)
		(if dewey-numbers
		    (begin
		      (display "recording instantiation ")
		      (display-ref-list (subgoal-terms subgoal))
		      (display " for semi-local path subgoal ")
		      (display-subgoal index-sg)
		      (display " on depth ")
		      (display (subgoal-depth index-sg))
		      (newline)))
		(set-subgoal-local-instances subgoal index-sg)
		(if (pair? (subgoal-local-instances index-sg))
		    (set-subgoal-local-instances index-sg (cons (cons subgoal (subgoal-terms subgoal)) (subgoal-local-instances index-sg)))
		    (set-subgoal-local-instances index-sg (list (cons subgoal (subgoal-terms subgoal)))))
		(if dewey-numbers
		    (begin
		      (display "Subgoal now used with ")
		      (display (length (subgoal-local-instances index-sg)))
		      (display " instantiations")
		      (newline)))
		;;(unbind-vars)
		(if dewey-numbers
		    (begin
		      (display "local instances: ")
		      (display-local-instances (subgoal-local-instances index-sg))
		      (newline)))))
; 		    (begin
; 		      (display "gr-unification of subgoal ")
; 		      (display-subgoal subgoal)
; 		      (display " on depth ") 
; 		      (display (subgoal-depth subgoal))
; 		      (display " and path subgoal ") 
; 		      (display-subgoal index-sg)
; 		      (display " on depth ") 
; 		      (display (subgoal-depth index-sg))
; 		      (display " failed !!!!")
; 		      (newline)))))
	  (set! successful-reduction-steps (+ successful-reduction-steps 1))
	  (mark-rel-vector rel-vector (subgoal-depth subgoal))
	  (if (context-lemma? index-sg)
	      (begin
		(set! context-reduction-steps (+ context-reduction-steps 1))
		(merge-rel-vectors rel-vector (lemma-rel-vector index-sg) (subgoal-depth index-sg)))
	      (mark-rel-vector rel-vector (subgoal-depth index-sg)))
	  rel-vector)
	#f)))

(define (set-subgoals-selected subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-selected (car subgoals) #t)
	(set-subgoals-selected (cdr subgoals)))))

(define (unrename-vars)
  (if (trail-content)
      (begin 
	;; if the current trail variable is a local variable and has just been made as a copy
	;; and is still local after the copying process is complete, mark the semi-local flag of the subgoal
	;; this variable had been introduced in to record that subgoal has a local variable
	(if (and local-vars (local-variable (caar trail)) (subgoal? (variable-renaming (caar trail))))
	    (set-subgoal-semi-local (variable-renaming (caar trail)) #t))
	(set-variable-renaming (caar trail) #f)
	(trail-previous)
	(unrename-vars))
      (begin (trail-previous) #f)))

(define (unrename-vars-returning-renamings)
  (if (trail-content)
      (let ((renaming (variable-renaming (caar trail))))
	(set-variable-renaming (caar trail) #f) (trail-previous) (cons renaming (unrename-vars-returning-renamings)))
      (begin (trail-previous) '())))

(define (trail-erase)
 (if (trail-content)
     (begin (trail-previous) (trail-erase))
     (begin (trail-previous) #f)))

(define (linking-step link subgoal rel-vector)
  (let ((clausea (subgoal-clause (link-lita link)))
	(clauseb (subgoal-clause (link-litb link)))
	(depth (subgoal-depth subgoal))
	(copya #f) (copyb #f) copylita
	(new-subgoals #f))
    (if (unify-lists-stop (subgoal-terms (link-lita link)) (subgoal-terms (link-litb link)))
	(begin
	  (set! total-inferences (+ total-inferences 1))
	  (if dewey-start
	      (if (>= total-inferences dewey-start)
		  (begin
		    (set! dewey-numbers #t)
		    (set! dewey-start #f))))
	  (if dewey-numbers
	      (begin
		(display "linking-step with link depths ") 
		(display (subgoal-depth (link-lita link)))
		(display " and ") 
		(if (and (vector? (link-litb link)) (vector? (subgoal-clause (link-litb link))))
		    (display (subgoal-depth (link-litb link)))
		    (begin
		      (display "strange b-subgoal ")))
		;;(display (link-litb link))))
		(display " below subgoal ")
		(display-subgoal subgoal)
; 		(display " on depth ")
; 		(display (subgoal-depth subgoal))
		(newline)
		(print "+++++++++++++++++++++++++++++++++++++")
		(display-clauses (list clausea clauseb))
		(newline)
		(print "+++++++++++++++++++++++++++++++++++++")))
; 	  (if (context-lemma? (link-lita link))
; 	      (begin
; 		(display "subgoal a ")
; 		(display-subgoal (link-lita link))
; 		(display " of depth ")
; 		(display (subgoal-depth (link-lita link)))
; 		(display " is a context lemma")
; 		(newline)))
; 	  (if (context-lemma? (link-litb link))
; 	      (begin
; 		(display "subgoal b ")
; 		(display-subgoal (link-litb link))
; 		(display " of depth ")
; 		(display (subgoal-depth (link-litb link)))
; 		(display " is a context lemma")
; 		(newline)))
	  (if (and (or tautologies (and (non-tautological clausea) (non-tautological clauseb)))
		   (or (not (and equality (or (expansion-order-violation (clause-subgoals clausea)) (expansion-order-violation (clause-subgoals clauseb))))))
		   (or equality (not purity) (and (not (pure-clause (link-lita link))) (not (pure-clause (link-litb link))))))
	      (let ((variant-a #f) (variant-b #f) (regular-a #f) (regular-b #f) (subsumed-a #f) (subsumed-b #f)
		    (keep-a #f) (keep-b #f) unsimplifieds-a unsimplifieds-b)
;		(if full-variants (set-linked-clauses-tested clausea clauseb))
		(if (and (not (clause-initial clausea)) (or (subgoal-ground (link-lita link)) (and local-vars (subgoal-local (link-lita link)))))
		    (set! variant-a clausea)
		    (set! variant-a (clause-index-variant? clausea)))
		(if (not variant-a)
		    (begin
		      (set! subsumed-a (clause-subsumed clausea))
		      (if (not subsumed-a)
			  (begin
			    (set! regular-a  (ground-regular clausea))))))
; 			    (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea)))
; 			    (compute-current-weights unsimplifieds-a)
; 			    (if (and regular-a full-variants)
; 				(begin
; 				  (set! variant-a (clause-variant-subsumed unsimplifieds-a))
; 				  (if (and path-subsumption (not variant-a))
; 				      (set! subsumed-a (clause-full-subsumed-by-path-clause (link-lita link) subgoal)))))))))
		(if (not (and couples (non-unit-clause clausea) (not (eq? variant-a clausea))
			      (or variant-a (not regular-a) subsumed-a)))
		    (begin
		      (if (and (not (clause-initial clauseb)) (or (subgoal-ground (link-litb link)) (and local-vars (subgoal-local (link-litb link)))))
			  (set! variant-b clauseb)
			  (set! variant-b (clause-index-variant? clauseb)))
		      (if (not variant-b)
			  (begin
			    (set! subsumed-b (clause-subsumed clauseb))
			    (if (not subsumed-b)
				(begin
				  (set! regular-b  (ground-regular clauseb)))))))
; 			    (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals clauseb)))
; 			    (compute-current-weights unsimplifieds-b)
; 			    (if (and regular-b full-variants)
; 				(begin
; 				  (set! variant-b (clause-variant-subsumed unsimplifieds-b))
; 				  (if (and path-subsumption (not variant-b))
; 				      (set! subsumed-b (clause-full-subsumed-by-path-clause (link-litb link) subgoal))))))))))
		      (set! variant-b #t))
;		(if full-variants (unset-linked-clauses clausea clauseb))
		;; FIXME: This code is supposed to speed up the computation of clause deletions.
		;; first, coupled elimination must be checked immediately after the tests of clausea
		;; then, the expensive tests such as path subsumption need to be done separately!
		;; then, after the variant subsumption test for clausea another test for coupled elimination needs to be done
		(if (not (and couples (non-unit-clause clausea) (non-unit-clause clauseb)
			      (not (eq? variant-a clausea)) (not (eq? variant-b clauseb))
			      (or variant-a variant-b subsumed-a subsumed-b (not regular-a) (not regular-b))))
		    (begin
		      (if full-variants (set-linked-clauses-tested clausea clauseb))
		      (if (not (or variant-a subsumed-a (not regular-a)))
			  (begin
			    (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea)))
			    (compute-current-weights unsimplifieds-a)
			    (if (and regular-a full-variants)
				(begin
				  (set! variant-a (clause-variant-subsumed unsimplifieds-a))
				  (if (and path-subsumption (not variant-a))
				      (begin
					(set! subsumed-a (clause-path-subsumed unsimplifieds-a))))))))
		      (if (not (or variant-b subsumed-b (not regular-b)))
			  (begin
			    (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals clauseb)))
			    (compute-current-weights unsimplifieds-b)
			    (if (and regular-b full-variants)
				(begin
				  (set! variant-b (clause-variant-subsumed unsimplifieds-b))
				  (if (and path-subsumption (not variant-b))
				      (begin
					(set! subsumed-b (clause-path-subsumed unsimplifieds-b))))))))
		      (if full-variants (unset-linked-clauses clausea clauseb))))
		(if (and couples (non-unit-clause clausea) (non-unit-clause clauseb)
			 (not (eq? variant-a clausea)) (not (eq? variant-b clauseb))
			 (or variant-a variant-b subsumed-a subsumed-b (not regular-a) (not regular-b)))
		    (begin 
		      (set! coupled-eliminations (+ 1 coupled-eliminations))
		      (if dewey-numbers
			  (begin
			    (display "coupled elimination") 
			    (newline))))
			(let ((delete-a (or subsumed-a variant-a (not regular-a))) (delete-b (or subsumed-b variant-b (not regular-b))))
			  (if (and path-closure (not delete-a) (not delete-b))
			      (let ((path-closed-a (ground-closed-with-index (link-lita link)))
				    (path-closed-b (ground-closed-with-index (link-litb link))))
				(if path-closed-a
				    (if path-closed-b
					(if (> (length unsimplifieds-a) (length unsimplifieds-b))
					    (set! delete-a #t)
					    (set! delete-b #t))
					(set! delete-b #t))
				    (if path-closed-b
					(set! delete-a #t)))))
			  (if (not delete-a)
			      (begin
				(if (not unsimplifieds-a)
				    (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea))))
				(set! copya (renamed-and-splitted-copy clausea unsimplifieds-a 'upper 
								       (link-lita link) (+ depth 1)))
				(set! depth (+ depth (length copya)))
				(clauses-path-vector-set copya)
				(set-clause-path (car copya) subgoal)
				(set-subgoal-next subgoal (car copya))
				(set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clausea)
										  (length unsimplifieds-a))))
				(if (and (clause-initial clausea)
					 (varlist-renamed? (clause-variables clausea) (clause-variables clausea) '()))
				    (set-input-in-tableau (clause-input clausea) (car copya)))
				(set! new-subgoals (clause-subgoals (car copya)))))
			  (if (not delete-b)
			      (begin
				(if (not unsimplifieds-b)
				    (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals clauseb))))
				(set! copyb (renamed-and-splitted-copy clauseb unsimplifieds-b 'lower 
								       (link-litb link) (+ depth 1)))
				(set! depth (+ depth (length copyb)))
				(clauses-path-vector-set copyb)
				(set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clauseb)
										  (length unsimplifieds-b))))
				(if (and (clause-initial clauseb)
					 (varlist-renamed? (clause-variables clauseb) (clause-variables clauseb) '()))
				    (set-input-in-tableau (clause-input clauseb) (car copyb)))
				(if (not new-subgoals)
				    (begin
				      (set-clause-path (car copyb) subgoal)
				      (set-subgoal-next subgoal (car copyb))
				      (set! new-subgoals (clause-subgoals (car copyb))))
				    (let ((copylita (subgoal-in-copy (link-lita link) (clause-subgoals (car copya)))))
; 				      (if (not (eq? (car copya) (subgoal-clause copylita)))
; 					  (begin
; 					    (display "WARNING: lower ")
; 					    (display-clause (car copyb))
; 					    (display " not properly linked to ")
; 					    (display-clause (car copya))
; 					    (display " by subgoal ")
; 					    (if (subgoal? copylita)
; 						(begin
; 						  (display-subgoal copylita)
; 						  (display " of depth ")
; 						  (display (subgoal-depth copylita)))
; 						(display copylita))
; 					    (newline)))
				      (set-clause-path (car copyb) copylita)
				      (set-subgoal-next copylita (car copyb))))))))
		(if new-subgoals
		    (begin
		      (set! successful-inferences (+ successful-inferences 1))
		      (if copya
			  (set! tableau-clauses (+ tableau-clauses (length copya))))
		      (if copyb
			  (set! tableau-clauses (+ tableau-clauses (length copyb))))
		      (if (> depth (+ 1 (subgoal-depth subgoal)))
			  (begin
			    (if dewey-numbers
				(begin
				  (display "attached new clauses on depths ") 
				  (display (+ (subgoal-depth subgoal) 1)) (display " to ") (display depth)
				  (newline))))
			  (begin
			    (if dewey-numbers
				(begin
				  (display "attached new clause on depth ") 
				  (display depth)
				  (newline)))))
		      (if dewey-numbers
			  (begin
			    (display "new clauses: ")
			    (if copya 
				(display-clauses copya)
				(display-clauses copyb))
			    (if (and copya copyb)
				(begin
				  (display " and ")
				  (display-clauses copyb)))
			    (newline))))))
	      (set-link-active link 'delete)) ;; tautological links can be discarded
	  (unbind-vars)
	  new-subgoals)
	#f)))

(define (compute-current-weights subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-misc (car subgoals) (current-subgoal-weight (car subgoals)))
	(compute-current-weights (cdr subgoals)))))

(define (set-current-weights subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-misc (car subgoals) (subgoal-weight (car subgoals)))
	(set-current-weights (cdr subgoals)))
      #t))

(define (renamed-copy clause unsimplifieds)
  (let ((copy (new-clause)))
    (set-clause-foldup-lits copy '())
    (trail-stop)
    (if unsimplifieds
	(set-clause-subgoals copy (renamed-subgoals unsimplifieds))
	(begin
	  (set! unsimplifieds (unsimplifiable-subgoals (clause-subgoals clause)))
	  (compute-current-weights unsimplifieds)
	  (set-clause-subgoals copy (renamed-subgoals unsimplifieds))))
    (unrename-vars)
    (if equality (set-clause-subgoals copy (eliminate-var-disequations (clause-subgoals copy))))
    (if (and equality (not ordered-eq))
	(subgoals-orientate-disequations (clause-subgoals copy)))
    (set-clause-unsimplified copy (length (clause-subgoals copy)))
    (if (null? (clause-subgoals copy))
	(set-clause-subgoals copy (list falsum)))
    (set-clause-variables copy (variables-in-subgoals (clause-subgoals copy) '()))
    (clause-pointers-in-subgoals (clause-subgoals copy) copy)
    (set-clause-input copy (clause-input clause))
    (set-clause-initial copy #f)
    (if (not norel) (set-clause-relevant copy (clause-relevant clause)))
    copy))

(define (renamed-and-splitted-copy clause unsimplifieds lower-upper special-sg depth)
  (let (new-subgoals (splitted '()))
    (trail-stop)
    (if unsimplifieds
	(set! new-subgoals (renamed-subgoals unsimplifieds))
	(begin
	  (set! unsimplifieds (unsimplifiable-subgoals (clause-subgoals clause)))
	  (compute-current-weights unsimplifieds)
	  (set! new-subgoals (renamed-subgoals unsimplifieds))))
    (unrename-vars)
    (if equality (set! new-subgoals (eliminate-var-disequations new-subgoals)))
    (if (and equality (not ordered-eq))
	(subgoals-orientate-disequations new-subgoals))
    (if (null? new-subgoals)
	(set! new-subgoals (list falsum)))
    (if (and split (not (eq? (car new-subgoals) falsum)))
	(set! splitted (split-subgoals new-subgoals '())))
    (if (and split (> (length splitted) 1) (not (and (= (length splitted) 2) (one-list-is-ground splitted))))
	(let ((special-copy (subgoal-in-copy special-sg new-subgoals)))
	  (if (or dewey-numbers (> debug-level 4))
	      (begin
		(display "subgoals ")
		(display-subgoal-list new-subgoals)
		(display " can be splitted into ")
		(display (length splitted))
		(display " disjoint subsets")
		(newline)))
	  (if (eq? lower-upper 'upper)
	      (set! splitted (move-special-split-list-to-rear special-copy splitted))
	      (set! splitted (move-special-split-list-to-front special-copy splitted)))
	  (set! splitted (split-lists-to-clauses splitted depth #f))
	  (set-clauses-input splitted (clause-input clause))
	  (if (not norel) (set-clauses-relevant splitted clause))
	  (set-clause-pointers-in-clauses splitted)
	  (set! nr-splitted (+ nr-splitted (- (length splitted) 1)))
	  splitted)
	(let ((copy (new-clause)))
	  (set-clause-foldup-lits copy '())
	  (set-clause-subgoals copy new-subgoals)
	  (set-clause-unsimplified copy (length (clause-subgoals copy)))
	  (set-clause-variables copy (variables-in-subgoals (clause-subgoals copy) '()))
	  (clause-pointers-in-subgoals (clause-subgoals copy) copy)
	  (set-clause-input copy (clause-input clause))
	  (set-clause-depth copy depth)
	  (if (not norel) (set-clause-relevant copy (clause-relevant clause)))
	  ;;(set-clause-initial copy #f)
	  ;;(if (not norel) (set-clause-relevant copy (clause-relevant clause)))
	  (list copy)))))

(define (check-subgoals-for-refs subgoals)
  (if (pair? subgoals)
      (begin
	(check-termlist-for-refs (subgoal-terms (car subgoals)))
	(check-subgoals-for-refs (cdr subgoals)))))

(define (set-clauses-relevant clauses original)
  (if (pair? clauses)
      (begin
	(set-clause-relevant (car clauses) (clause-relevant original))
	(set-clauses-relevant (cdr clauses) original))))

(define (set-clauses-input clauses input)
  (if (pair? clauses)
      (begin
	(set-clause-input (car clauses) input)
	(set-clauses-input (cdr clauses) input))))

(define (set-clause-pointers-in-clauses clauses)
  (if (pair? clauses)
      (begin
	(clause-pointers-in-subgoals (clause-subgoals (car clauses)) (car clauses))
	(set-clause-pointers-in-clauses (cdr clauses)))))

(define (variables-in-subgoals subgoals vars)
  (if (null? subgoals)
      vars
      (variables-in-subgoals (cdr subgoals) (free-vars-in-termlist (subgoal-terms (car subgoals)) vars))))
  
(define (rewrite-new-subgoals subgoals)
  (if (pair? subgoals)
      (let ((rewritten #f))
	(set! rewritten (rewrite-subgoal (car subgoals)))
	(if rewritten
	    (begin
	      (display "rewrote subgoal to ")
	      (display-subgoal (car subgoals))
	      (newline)
	      (set-subgoal-weight (car subgoals) (current-subgoal-weight (car subgoals)))
	      (rewrite-new-subgoals (cdr subgoals))
	      #t)
	    (rewrite-new-subgoals (cdr subgoals))))
      #f))

(define (free-vars-in-termlist terms vars)
  ;; finds all free variables in terms and inserts them into vars exactly once
  ;;(print vars)
  (if (pair? terms)
      (begin
	(free-vars-in-termlist (cdr terms) (free-vars-in-term (ref (car terms)) vars)))
      vars))

(define (free-vars-in-term term vars)          
  ;; checks if a term is a variable.
  ;; if yes, it is added to the variable list vars if it is not already in there
  ;; otherwise if it is a proper functor, the argument list of term is searched for
  ;; free variables
  (if (variable term)
      (if (memq term vars)
	  vars
	  (cons term vars))
      (if (constant term)
	  vars
	  (free-vars-in-termlist (cdr term) vars))))

(define (find-or-insert-var term vars)
  (if (pair? vars)
      (begin
	(if (not (eq? term (car vars)))
	    (if (pair? (cdr vars))
		(find-or-insert-var term (cdr vars))
		(set-cdr! vars (list term))))
	vars)
      (list term)))
      
(define (clause-pointers-in-subgoals subgoals clause)
  (if (pair? subgoals)
      (begin
	(set-subgoal-clause (car subgoals) clause)
	(clause-pointers-in-subgoals (cdr subgoals) clause))))

(define (subgoal-in-copy subgoal copy-subgoals)
  (subgoal-misc subgoal))

(define (indexq obj list)
  (if (pair? list)
      (if (eq? obj (car list))
	  0
	  (let ((value (indexq obj (cdr list))))
	    (if value
		(+ 1 value)
		value)))
      #f))

(define (clause-index-variant? clause)
  ;; Right now does the variant test only for clauses that are descended from the same input clause
  (let ((index-result #f))
    (let (index-end)
      (set! index-end (clause-in-index? clause))
      (if (vector? index-end)
	  (if (= 0 (vector-ref index-end 0))
	      (begin
		(if (> debug-level 5)
			(print "variant not found: " index-end))
		#f)
	      (begin
		(if dewey-numbers
		    (begin
		      (display "delete variant ") 
		      (display-clause clause)
		      (display " of ") 
		      (display-clause (car (vector-ref index-end 1)))
		      (if (> (vector-ref index-end 0) 1)
			  (begin
			    (display " (strange variant number: ")
			    (display (vector-ref index-end 0))
			    (display ")")))
		      (newline)))
		(set! nr-variants (+ nr-variants 1))
		(set! index-result (car (vector-ref index-end 1)))
		index-result))
	  (begin
	    (if (> debug-level 4)
		(begin 
		  (display "index-search failed for ")
		  (display-clause clause) (display " with index ")
		  (newline)
		  (print-index-entries (cdr (input-variant-index (clause-input clause))) "(" '())
		  (newline)))
	    #f)))))

(define (get-path-clause subgoal depth)
  (if subgoal
      (if (= depth (subgoal-depth subgoal))
	  (subgoal-clause subgoal)
	  (if (< depth (subgoal-depth subgoal))
	      (get-path-clause (subgoal-predecessor subgoal) depth)
	      #f))
      #f))

(define (get-top-depth subgoal)
  (if subgoal
      (if (subgoal-predecessor subgoal)
	  (get-top-depth (subgoal-predecessor subgoal))
	  (subgoal-depth subgoal))
      #f))

(define (clause-subsumed clause)
  (if subsumption
      (if (or (and unit-subsume simplify) (non-unit-clause clause))
	  (let ((subsumed (subgoals-subsumed-by-index (clause-subgoals clause))))
	    (if subsumed
		(begin
		  (if (and (or (> debug-level 4) dewey-numbers) (not preprocessing))
		      (begin
			(display "Subsumption deletion of ")
			(display-clause clause)
			(display " by unit ")
			(display-clause (subgoal-clause subsumed))
			(newline)))
		  (set! nr-subsumption-deletions (+ nr-subsumption-deletions 1)))
		(if (and (not equality) context-units context-lemmas (not preprocessing))
		    (begin
		      (set! subsumed (subgoals-subsumed-by-context-index (clause-subgoals clause)))
		      (if subsumed
			  (if (and (or (> debug-level 4) dewey-numbers) (not preprocessing))
			      (begin
				(display "Subsumption deletion of ")
				(display-clause clause)
				(display " by context-lemma ")
				(display-clause (subgoal-clause subsumed))
				(newline)))
			  (set!  nr-context-subsumptions (+ nr-context-subsumptions 1))))))
	    subsumed)
	  #f)
      #f))

(define (ground-regular clause)
  (if (or (not regularity) (index-regular-subgoals (clause-subgoals clause)))
      #t
      (begin
	(set! regularity-fails (+ regularity-fails 1))
	#f)))

(define (index-regular-subgoals subgoals)
  (if (pair? subgoals)
      (if (ground-regular-subgoal? (car subgoals))
          (index-regular-subgoals (cdr subgoals))
          (begin 
            #f))
      #t))

; (define (context-regular clause linked-lit other-clause-exists)
;   (if (or (not regularity) (not context-units) (context-regular-subgoals (clause-subgoals clause) linked-lit other-clause-exists))
;       #t
;       (begin
; 	(set! regularity-fails (+ regularity-fails 1))
; 	#f)))

; (define (context-regular-subgoals subgoals linked-lit other-clause-exists)
;   (if (pair? subgoals)
;       (let ((irregular (context-irregular-subgoal? (car subgoals))))
; 	(if irregular
; 	    (let ((closed (ground-closed-with-index (car subgoals))))
; 	      (if (or (and (not closed) (not (eq? (car subgoals) linked-lit)))
; 		      ;;(and (not closed) (eq? (car subgoals) linked-lit) other-clause-exists)
; 		      (and closed (>= (subgoal-depth closed) (subgoal-depth irregular))))
; 		  (begin
; 		    (if dewey-numbers
; 			(begin
; 			  (display "regularity fail of subgoal ") 
; 			  (display-subgoal (car subgoals))
; 			  (display " of depth ") 
; 			  (display (subgoal-depth (car subgoals)))
; 			  (display " by c-lemma ") 
; 			  (display-subgoal irregular)
; 			  (display " of depth ") 
; 			  (display (subgoal-depth irregular))
; 			  (newline)))
; 		    #f)
; 		  (context-regular-subgoals (cdr subgoals) linked-lit other-clause-exists)))
; 	    (context-regular-subgoals (cdr subgoals) linked-lit other-clause-exists)))
;       #t))


(define (ground-and-equal-termlists k l)
  (if (pair? k)
      (if (ground-and-equal-terms (ref (car k)) (ref (car l)))
	  (ground-and-equal-termlists (cdr k) (cdr l))
	  #f)
      #t))

(define (ground-and-equal-terms s t)
  (if (eq? s t)
      #t
      (if (variable s)
	  #f
	  (if (and (non-variable t) (eq? (car s) (car t)))
	      (ground-and-equal-termlists (cdr s) (cdr t))
	      #f))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Junk functions, various
;; ------------------------------------------------------------

(define (print-predicates-indexes predicates)
  (if (pair? predicates)
      (begin
	(print-indexes (car predicates) 'gr)
	(print-indexes (car predicates) 'unit)
	(print-predicates-indexes (cdr predicates)))))

(define (test-subsumption-for-units clauses)
  (if (pair? clauses)
      (begin
	(display-clause (car clauses))
	(if (unit-clause (car clauses))
	    (display " (unit)")
	    (if (non-unit-clause (car clauses))
		(display " (non-unit)")))
	(newline)
	(clause-subsumed (car clauses))
	(test-subsumption-for-units (cdr clauses)))))

(define (get-path-subgoal subgoal depth)
  (if (and subgoal (>= (subgoal-depth subgoal) depth))
      (if (= (subgoal-depth subgoal) depth)
	  subgoal
	  (get-path-subgoal (subgoal-predecessor subgoal) depth))
      #f))

(define (variables-in-unsimplified-subgoals subgoals vars)
  (if (pair? subgoals)
      (if (subgoal-simplified (car subgoals))
	  (variables-in-unsimplified-subgoals (cdr subgoals) vars)
	  (variables-in-unsimplified-subgoals (cdr subgoals) (free-vars-in-termlist (subgoal-terms (car subgoals)) vars)))
      vars))
  
(define (check-for-local-vars subgoal)
  (let (subgoal-vars other-vars)
    (set-subgoal-local subgoal #t)
    (set-subgoal-semi-local subgoal #f)
    (set! subgoal-vars (free-vars-in-termlist (subgoal-terms subgoal) '()))
    (if (pair? subgoal-vars)
	(begin
	  (set! other-vars (variables-in-unsimplified-subgoals (remq subgoal (clause-subgoals (subgoal-clause subgoal))) '()))
	  (if (and (not (simplified-unit subgoal)) (contains-local-var subgoal subgoal-vars other-vars))
	      (if dewey-numbers
		  (if (subgoal-local subgoal)
		      (begin
			(display "LOCAL subgoal ")
			(display-subgoal subgoal)
			(display " in ")
			(display-clause (subgoal-clause subgoal))
			(newline))
		      (begin
			(display "subgoal ")
			(display-subgoal subgoal)
			(display " with local var in ")
			(display-clause (subgoal-clause subgoal))
			(newline)))))))))

(define (contains-local-var subgoal subgoal-vars other-vars)
  (if (pair? subgoal-vars)
      (if (memq (car subgoal-vars) other-vars)
	  (begin
	    (if (local-variable (car subgoal-vars))
		(set-variable-name (car subgoal-vars) (car (variable-name (car subgoal-vars)))))
	    (set-subgoal-local subgoal #f)
	    (contains-local-var subgoal (cdr subgoal-vars) other-vars))
	  (begin
	    (if (not (local-variable (car subgoal-vars)))
		(set-variable-name (car subgoal-vars) (cons (variable-name (car subgoal-vars)) 'l)))
	    (set-subgoal-semi-local subgoal #t)
	    (contains-local-var subgoal (cdr subgoal-vars) other-vars)
	    #t))
      #f))

(define (local-subgoal subgoal-vars other-vars)
  (if (pair? subgoal-vars)
      (if (not (local-variable (car subgoal-vars)))
	  #f
	  (local-subgoal (cdr subgoal-vars) other-vars))
      #t))

(define (free-variables? vars)
  (if (pair? vars)
      (if (free-variable (car vars))
	  (free-variables? (cdr vars))
	  #f)
      #t))

(define (disjoint-renamed? vars)
  (if (pair? vars)
      (if (bound-variable (car vars))
	  (if (free-variable (variable-binding (car vars)))
	      (let ((result (disjoint-renamed? (cdr vars))))
		(if result
		    (if (memq (variable-binding (car vars)) result)
			#f
			(cons (variable-binding (car vars)) result))
		    #f))
	      #f)
	  #f)
      '()))
		 
(define (variant-varlist? vars)
  (if (pair? vars)
      (if (free-variable (ref (car vars)))
	  (let ((result (variant-varlist? (cdr vars))))
	    (if result
		(if (memq (ref (car vars)) result)
		    #f
		    (cons (ref (car vars)) result))
		#f))
	  #f)
      '()))
		 
;;;*************************************************************************
;; ------------------------------------------------------------
;; Equality functions
;; ------------------------------------------------------------

(define (eq-linking-step link subgoal rel-vector)
  (let ((clausea (subgoal-clause (link-lita link)))
	(clauseb (subgoal-clause (link-litb link)))
	(original-a (subgoal-clause (link-lita link)))
	(depth (subgoal-depth subgoal))
	(copya #f) (copyb #f) copylita copylitb replace by replaced-saved new-disequation eq-sg
	(disequation-added #f)
	(rewritable #f)
	(new-subgoals #f))
    (if (eq? (subgoal-clause (link-lita link)) (subgoal-clause (link-litb link)))
	(begin
	  (print "creating renamed copy of overlapping clause")
	  (set! clausea (renamed-copy clausea #f))
	  (set-clause-depth clausea (clause-depth original-a))
	  (set! eq-sg (subgoal-in-copy (link-lita link) (clause-subgoals clausea))))
	(set! eq-sg (link-lita link)))
    ;; NOTA BENE: ref-fing the sides of the equation is necessary as eager elimination may introduce bound variables
    ;; into tableau terms!
    (if (link-left-to-right link)
	(begin
	  (set! replace (ref (car (subgoal-terms eq-sg))))
	  (set! by (ref (cadr (subgoal-terms eq-sg)))))
	(begin
	  (set! replace (ref (cadr (subgoal-terms eq-sg))))
	  (set! by (ref (car (subgoal-terms eq-sg))))))
    (if dewey-numbers
	(begin
	  (display "eq-linking-step with link depths ") 
	  (display (subgoal-depth (link-lita link)))
	  (display " and ") 
	  (display (subgoal-depth (link-litb link)))
	  (display " below subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (display "with equation ") 
	  (display-subgoal (link-lita link))
	  (if (link-left-to-right link)
	      (display " (->) ") 
	      (display " (<-) "))
	  (display " into subgoal ") 
	  (display-overlapped-subgoal (link-litb link) (link-position link))
	  (newline)
	  (display "of ")
	  (if (not (non-tautological clauseb))
	      (display "tautological "))
	  (display-clause clauseb)
	  (newline)
	  (check-termlist-for-refs (subgoal-terms (link-lita link)))
	  (newline)))
    (if (unify-terms-stop replace (ref (car (link-position link))))
	(begin
; 	  (if (and ordered-eq full-rewrite
; 		   (or (not (clause-initial clauseb)) (unit-clause clauseb))
; 		   (unit-subgoal (link-lita link))
; 		   (or (subgoal-ordered-eq (link-lita link)) (kbo> replace by))
; 		   (varlist-untouched? (clause-variables clauseb))
; 		   (or (not (top-level? (ref (car (link-position link))) (subgoal-terms (link-litb link))))
; 		       (not (varlist-renamed? (clause-variables clausea) (clause-variables clausea) '()))))
; 	      (begin
; 		(if dewey-numbers
; 		    (begin
; 		      (display "rewrite-condition fulfilled, rewriting")
; 		      (newline)))
; 		(set! rewritable #t)))
	  (set! total-eq-inferences (+ total-eq-inferences 1))
	  (if (not (and simplify (unit-subgoal (link-lita link))))
	      (begin
		(set! new-disequation (make-new-negated-eq-sg eq-sg (link-litb link)))
		(set-subgoal-expansion-lit new-disequation #t)))
	  ;; actually apply the equation
	  (set! replaced-saved (car (link-position link)))
	  (set-car! (link-position link) by)
	  (if dewey-start
	      (if (>= total-inferences dewey-start)
		  (begin
		    (set! dewey-numbers #t)
		    (set! dewey-start #f))))
	  (if dewey-numbers
	      (begin
		(print "+++++++++++++++++++++++++++++++++++++")
		(display-clauses (list clausea clauseb))
		(newline)
		(print "+++++++++++++++++++++++++++++++++++++")))
	  (if (and (or tautologies (and (non-tautological clausea))) ;; (non-tautological clauseb)))
		   (or (not (and equality (or (expansion-order-violation (clause-subgoals clausea)) (expansion-order-violation (clause-subgoals clauseb)))))))

	      (let ((variant-a #f) (variant-b #f) (regular-a #f) (regular-b #f) (subsumed-a #f) (subsumed-b #f) 
			      (unsimplifieds-a #f) (unsimplifieds-b #f))
		(if full-variants (set-linked-clauses-tested original-a clauseb))
		(if (and (or (not (clause-initial original-a))
			     (unit-clause original-a)) ;; checking for input units is OK this way
			 (varlist-renamed? (clause-variables clausea) (clause-variables clausea) '()))
		    (begin
		      (if dewey-numbers
			  (begin
			    (display "deleting renamed variant ")
			    (display-clause original-a)
			    (newline)))
		      (set! variant-a original-a))
		    (set! variant-a (clause-index-variant? clausea)))
		(if (not variant-a)
		    (begin
		      (set! subsumed-a (clause-subsumed clausea))
		      (if (not subsumed-a)
			  (begin
			    (set! regular-a  (ground-regular clausea))
			    (if (and regular-a full-variants); (non-unit-clause clausea))
				(begin
				  (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea)))
				  (compute-current-weights unsimplifieds-a)
				  ;;(set! variant-a (clause-variant-subsumed unsimplifieds-a))))))))
				  (set! variant-a (clause-variant-subsumed unsimplifieds-a))
				  (if (and path-subsumption (not variant-a))
				      (set! subsumed-a (clause-path-subsumed unsimplifieds-a)))))))))
		;; I doubt this is correct... even with eager elimination the following can happen:
		;; unit-lemma  X0 == a
		;; overlapping equation X0 == X1 v p(X0) v q(X1)
		;; into p(a) of p(a) v r(a)
		;; results to X0 == a v p(X0) v q(a) (subsumed)
		;; and        X0 =! a v p(X0) v r(a) (disequation simplified before eager elimination)
		;; the second clause is a valid new tableau clause!
		;;(if (or (and eager (not (unit-subgoal (link-lita link)))
			     ;;(variable (ref by))
			     ;;(not (occurs-term (ref by) (ref replace)))) no checking for renamings in the overlapped clause!
			;;(and ordered-eq (not (subgoal-ordered-eq eq-sg)) (kbo> (ref by) (ref replace))))
		;; check for ordering violations in the overlapping equations after unification
		;;(if (and ordered-eq (not (subgoal-ordered-eq eq-sg)) (kbo> (ref by) (ref replace)))
		(if (or (and eager (not (unit-subgoal (link-lita link)))
			     (not (and subsumption (subgoal-subsumed-by-index eq-sg)))
			     (variable (ref by))
			     (not (occurs-term (ref by) (ref replace))))
			(and ordered-eq (not (subgoal-ordered-eq eq-sg)) (kbo> (ref by) (ref replace))))
		    (begin
		      (if dewey-numbers
			  (begin
			    (display "deleting overlapped instance ")
			    (display-clause clauseb)
			    (newline)))
		      (set! variant-b clauseb))
		    (set! variant-b (clause-index-variant? clauseb)))
		(if (not (and simplify (unit-subgoal (link-lita link))))
		    (begin
		      (set-clause-subgoals clauseb (insert-literal new-disequation (clause-subgoals clauseb)))
		      (set! disequation-added #t)))
		(if (not variant-b)
		    (begin
		      ;; temporarily add the new disequation to the instantiated paramodulated clause
		      ;; this is done only here so as not to disturb the clause-variant-index-test
		      (set! subsumed-b (clause-subsumed clauseb))
		      (if (not subsumed-b)
			  (begin
			    (set! regular-b  (ground-regular clauseb))
			    (if (and regular-b full-variants); (non-unit-clause clauseb))
				(begin
				  (set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals clauseb)))
				  (compute-current-weights unsimplifieds-b)
				  ;;(set! variant-b (clause-variant-subsumed unsimplifieds-b))))))))
				  (set! variant-b (clause-variant-subsumed unsimplifieds-b))
				  (if (and path-subsumption (not variant-b))
				      (set! subsumed-b (clause-path-subsumed unsimplifieds-b)))))))))
		(if full-variants (unset-linked-clauses original-a clauseb))
		(if (or (and couples (non-unit-clause clausea) (non-unit-clause clauseb)
			     (not (eq? variant-a clausea)) (not (eq? variant-b clauseb))
			     (or variant-a variant-b subsumed-a subsumed-b (not regular-a) (not regular-b))
			     (or (not (variable (ref by))) (occurs-term (ref by) (ref replace))))
			(not (non-tautological clauseb)))
		    (begin 
		      (set! coupled-eliminations (+ 1 coupled-eliminations))
		      (if dewey-numbers
			  (begin
			    (display "coupled eq-elimination") 
			    (newline))))
		    (begin
		      (if (and (not (or subsumed-a variant-a)) regular-a)
			  (begin
			    (if (not unsimplifieds-a)
				(set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea))))
			    (set! copya (renamed-and-splitted-copy clausea unsimplifieds-a 'upper 
								   (link-lita link) (+ depth 1)))
			    (set! depth (+ depth (length copya)))
			    (clauses-path-vector-set copya)
			    (set-clause-path (car copya) subgoal)
			    (set-subgoal-next subgoal (car copya))
			    (set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clausea)
									      (length unsimplifieds-a))))
			    (set! new-subgoals (clause-subgoals (car copya)))))
		      (if (and (not (or subsumed-b variant-b)) regular-b
			       (not (and path-closure new-subgoals (ground-closed-with-index (link-lita link)))))
			  (begin
			    (if (varlist-untouched? (clause-variables clauseb))
				(set! nr-e-variants (+ nr-e-variants 1)))
			    (if (not unsimplifieds-b)
				(set! unsimplifieds-b  (unsimplifiable-subgoals (clause-subgoals clauseb))))
			    (if (and path-closure new-subgoals disequation-added)
				(let ((closeb #f))
				  (set! closeb (ground-closed-with-index new-disequation))
				  (if closeb
				      (begin
					(set! depth (subgoal-depth subgoal))
					(set-subgoal-next subgoal #f)
					(set! new-subgoals #f)
					(if dewey-numbers
					    (begin
					      (display "disequation ")
					      (display-subgoal new-disequation)
					      (display " below depth ")
					      (display (subgoal-depth subgoal))
					      (display " closed with ")
					      (if (subgoal-clause closeb)
						  (if (subgoal-unlinkable closeb)
						      (display "unlinkable path subgoal ")
						      (display "subgoal "))
						  (display "lemma "))
					      (display-subgoal closeb)
					      (display " on depth ")
					      (display (subgoal-depth closeb))
					      (newline)))))))
; 			    (if (not (non-tautological clauseb))
; 				(begin
; 				  (display "producing paramodulated instance of tautological ")
; 				  (display-clause clauseb)
; 				  (newline)))
			    (set! copyb (renamed-and-splitted-copy clauseb unsimplifieds-b 'lower 
								   (link-litb link) (+ depth 1)))
			    (set! copylitb (subgoal-in-copy (link-litb link) (clause-subgoals (car copyb))))
			    (if (and copylitb ordered-eq (eq-subgoal? copylitb))
				(begin
				  (check-orientation copylitb)))
			    (set! depth (+ depth (length copyb)))
; 			    (if (eq? depth 1115) 
; 				(begin
; 				  (print "clause on depth 1115 is paramodulated instance (" (length copyb) ")")
; 				  (check-termlist-for-refs (subgoal-terms copylitb))))
			    (clauses-path-vector-set copyb)
			    (set! nr-simplifications (+ nr-simplifications (+ 1 (- (clause-unsimplified clauseb)
										   (length unsimplifieds-b)))))
			    ;; FIXME: If the old literal is destructively rewritten, all index information for
			    ;; the literal must be destroyed as well! Is the rewriting of units actually possible?
			    ;; FIXME even more: replacing a global copy with a branch local rewritten copy cannot
			    ;; be right. This significantly reduces the equational theory available on other branches
			    ;; in the non-unit case and is therefore incomplete!
			    ;;(if (link-rewriting link)
			    (if (and unit-eq (link-rewriting link))
				(begin
				  (if dewey-numbers
				      (begin
					(display "discarding overlapped original after destructive rewrite") 
					(newline)))
				  (delete-subgoals-links (clause-subgoals clauseb) (link-litb link))
				  (set! nr-rewrites (+ nr-rewrites 1))))
			    (if (not new-subgoals)
				(begin
				  (set-clause-path (car copyb) subgoal)
				  (set-subgoal-next subgoal (car copyb))
				  (set! new-subgoals (clause-subgoals (car copyb))))
				(let ((copylita (subgoal-in-copy (link-lita link) (clause-subgoals (car copya)))))
				  (set-clause-path (car copyb) copylita)
				  (set-subgoal-next copylita (car copyb))))))))
		(if new-subgoals
		    (begin
		      (set! successful-eq-inferences (+ successful-eq-inferences 1))
		      (if copya
			  (set! tableau-clauses (+ tableau-clauses (length copya))))
		      (if copyb
			  (set! tableau-clauses (+ tableau-clauses (length copyb))))
		      (if (> depth (+ 1 (subgoal-depth subgoal)))
			  (begin
			    (if dewey-numbers
				(begin
				  (display "eq-linking attached new clauses on depths ") 
				  (display (+ (subgoal-depth subgoal) 1)) (display " to ") (display depth)
				  (newline))))
			  (begin
			    (if dewey-numbers
				(begin
				  (display "eq-linking attached new clause on depth ") 
				  (display depth)
				  (newline)))))
		      (if dewey-numbers
			  (begin
			    (display "new clauses: ")
			    (if copya 
				(display-clauses copya)
				(display-clauses copyb))
			    (if (and copya copyb)
				(begin
				  (display " and ")
				  (display-clauses copyb)))
			    (newline))))))
	      (set-link-active link 'delete)) ;; tautological links can be discarded
	  ;; undo the effects of the paramodulation step for the original clauses
	  (set-car! (link-position link) replaced-saved)
	  (if disequation-added
	      (set-clause-subgoals clauseb (delete-element-from-list new-disequation (clause-subgoals clauseb) 
								     (clause-subgoals clauseb))))
; 	  (if (and (eq? (subgoal-depth (link-lita link)) 1115) (eq? (subgoal-depth (link-litb link)) 1144))
; 	      (begin
; 		(display "overlapping subgoal instantiated to ") 
; 		(display-subgoal (link-lita link))
; 		(newline)))
	  (unbind-vars)
; 	  (if (and (eq? (subgoal-depth (link-lita link)) 1115) (eq? (subgoal-depth (link-litb link)) 1144))
; 	      (begin
; 		(display "overlapping subgoal restored to ") 
; 		(display-subgoal (link-lita link))
; 		(newline)))
;  	  (if (and (eq? (subgoal-depth (link-lita link)) 1115) (eq? (subgoal-depth (link-litb link)) 1144))
; 	      (begin
; 		(display "overlapped subgoal restored to ") 
; 		(display-subgoal (link-litb link))
; 		(newline)))
	  new-subgoals)
	#f)))

(define (reflexivity-linking-step link subgoal rel-vector)
  (let ((clausea (subgoal-clause (link-lita link)))
	(depth (subgoal-depth subgoal))
	(copya #f) copylita left-side right-side
	(new-subgoals #f))
    (if (disequation? (link-lita link))
	(begin
	  (set! left-side (car (subgoal-terms (link-lita link))))
	  (set! right-side (cadr (subgoal-terms (link-lita link)))))
	(begin
	  (display "WARNING: reflexivity linking applied to non-disequation!")
	  (newline)))
    (if dewey-numbers
	(begin
	  (display "reflexivity-linking-step with link depth ") 
	  (display (subgoal-depth (link-lita link)))
	  (display " below subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (display "with disequation ") 
	  (display-subgoal (link-lita link))
	  (newline)
	  (display "of clause ")
	  (display-clause clausea)
	  (newline)))
    (if (unify-terms-stop left-side right-side)
	(begin
	  (set-subgoal-simplified (link-lita link) #t)
	  (set! total-refl-inferences (+ total-refl-inferences 1))
	  (if dewey-numbers
	      (begin
		(print "+++++++++++++++++++++++++++++++++++++")
		(display-clause clausea)
		(newline)
		(print "+++++++++++++++++++++++++++++++++++++")))
	  (if (and (or tautologies (non-tautological clausea))
		   (or (not (and equality (expansion-order-violation (clause-subgoals clausea))))))
	      (let ((variant-a #f) (regular-a #f) (subsumed-a #f) (unsimplifieds-a #f))
		(if full-variants (set-clause-tested clausea))
		(if (and (or (not (clause-initial clausea))
			     (unit-clause clausea)) ;; checking for input units is OK this way
			 (varlist-renamed? (clause-variables clausea) (clause-variables clausea) '()))
		    (set! variant-a clausea)
		    (set! variant-a (clause-index-variant? clausea)))
		(if (not variant-a)
		    (begin
		      (set! subsumed-a (clause-subsumed clausea))
		      (if (not subsumed-a)
			  (begin
			    (set! regular-a  (ground-regular clausea))
			    (if (and regular-a full-variants)
				(begin
				  (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea)))
				  (compute-current-weights unsimplifieds-a)
				  (set! variant-a (clause-variant-subsumed unsimplifieds-a))
				  (if (and path-subsumption (not variant-a))
				      (set! subsumed-a (clause-path-subsumed unsimplifieds-a)))))))))
		(if full-variants (unset-clause-tested clausea))
		(if (and (not (or subsumed-a variant-a)) regular-a)
		    (begin
		      (if (not unsimplifieds-a)
			  (set! unsimplifieds-a  (unsimplifiable-subgoals (clause-subgoals clausea))))
		      (set! copya (renamed-and-splitted-copy clausea unsimplifieds-a 'upper 
							     (link-lita link) (+ depth 1)))
		      (set! depth (+ depth (length copya)))
		      (clauses-path-vector-set copya)
		      (set-clause-path (car copya) subgoal)
		      (set-subgoal-next subgoal (car copya))
		      (set! nr-simplifications (+ nr-simplifications (- (clause-unsimplified clausea)
									(length unsimplifieds-a))))
		      (set! new-subgoals (clause-subgoals (car copya)))))
		(if new-subgoals
		    (begin
		      (set! successful-refl-inferences (+ successful-refl-inferences 1))
		      (set! tableau-clauses (+ tableau-clauses (length copya)))
		      (begin
			(if dewey-numbers
			    (begin
			      (display "attached new clause on depth ") 
			      (display depth)
			      (newline)))
		      (if dewey-numbers
			  (begin
			    (display "new clause: ")
			    (display-clauses copya)
			    (newline))))))
	      (set-link-active link 'delete))) ;; tautological links can be discarded
	  ;; undo the effects of the reflexivity linking step for the original clause
	  (set-subgoal-simplified (link-lita link) #f)
	  (unbind-vars)
	  new-subgoals)
	#f)))

(define (eliminate-var-disequations subgoals)
  (if eager
      (if (pair? subgoals)
	  (if (disequation? (car subgoals))
	      (let ((left (ref (car (subgoal-terms (car subgoals)))))
		    (right (ref (cadr (subgoal-terms (car subgoals))))))
		(if (and simplify (eq? left right))
		    (eliminate-var-disequations (cdr subgoals))
		    (if (and (variable left) (not (occurs-term left right)))
			(begin
			  (set-car! left right) ;; destructively bind left var
			  (set! nr-eager (+ nr-eager 1))
			  (eliminate-var-disequations (cdr subgoals)))
			(if (and (variable right) (not (occurs-term right left)))
			    (begin
			      (set-car! right left) ;; destructively bind right var
			      (set! nr-eager (+ nr-eager 1))
			      (eliminate-var-disequations (cdr subgoals)))
			    (cons (car subgoals) (eliminate-var-disequations (cdr subgoals)))))))
	      (cons (car subgoals) (eliminate-var-disequations (cdr subgoals))))
	  '())
      subgoals))

(define (free-vars-in-varlist vars)
  ;; this function expects vars to be a list of originally free variables, it will not follow ref's
  (if (pair? vars)
      (if (free-variable (car vars))
	  (cons (car vars) (free-vars-in-varlist (cdr vars)))
	  (free-vars-in-varlist (cdr vars)))
      '()))

(define (clauses-path-vector-set clauses)
  (if (pair? clauses)
      (begin
	(path-vector-set (clause-depth (car clauses)) (car clauses))
	(clauses-path-vector-set (cdr clauses)))))

(define (clauses-valid-subgoals clauses)
  (if (pair? clauses)
      (+ (clause-valid-subgoals (clause-subgoals (car clauses))) (clauses-valid-subgoals (cdr clauses)))
      0))

(define (clause-valid-subgoals subgoals)
  (if (pair? subgoals)
      (if (eq? (car subgoals) falsum)
	  0
	  (if (splitlit (car subgoals))
	      (clause-valid-subgoals (cdr subgoals))
	      (+ 1 (clause-valid-subgoals (cdr subgoals)))))
      0))

(define (set-subgoals-misc-to-weight subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-misc (car subgoals) (subgoal-weight (car subgoals)))
	(set-subgoals-misc-to-weight (cdr subgoals)))))

(define (check-subgoals-next subgoals)
  (if (pair? subgoals)
      (if (subgoal-next (car subgoals))
	  (+ 1 (check-subgoals-next (cdr subgoals)))
	  (check-subgoals-next (cdr subgoals)))
      0))

(define (exists-simplified-side-subgoal subgoal subgoals)
  (if (pair? subgoals)
      (if (and (not (eq? subgoal (car subgoals)))
	       (not (subgoal-ground (car subgoals)))
	       (subgoal-simplified (car subgoals)))
	  #t
	  (exists-simplified-side-subgoal subgoal (cdr subgoals)))
      #f))
      
(define (renamed-copy-of-assoc-list alist)
  (if (pair? alist)
      (cons (cons (caar alist) (renamed-term (cdar alist) 'copy)) (renamed-copy-of-assoc-list (cdr alist)))
      '()))

(define (display-local-instances instances)
  (if (pair? instances)
      (begin
	(display "[")
	(display-ref-list (cdar instances))
	(display "] ")
	(display-local-instances (cdr instances)))))
