(module display
	(include "dctp.sch")
	(export (display-link-vector link-vector) (display-current-link-list l) (display-subgoal subgoal)
		(display-branch-subgoals subgoal) (display-clause cl) 
		(display-subgoal-list subgoal-list) (display-clauses clauses) (display-link link) (display-ref term) 
		(display-subgoal-terms terms) (display-ref-list terms) (print-statistics) (print-tptp-lemmas subgoal) 
		(display-overlapped-subgoal subgoal position) (print-tptp-lemma subgoal) (print-version)
		(display-tptp-clauses clauses) (display-subgoal-lists subgoal-lists))
	(import (globals "globals.scm") (unification "unification.scm") (index "index.scm")
		(display-formula-statistics preprocess "preprocess.scm")))

;;--------------------------------
;; Display functions
;;--------------------------------
(define (display-ref term)
  (if (index-variable term)
      (if (positive? term)
	  (if (free-index-variable term)
	      (begin
		(display "X")
		(display term))
	      (display-ref (index-variable-binding term)))
	  (begin
	    (display "L")
	    (display (- term))))
      (if (variable term)
	  (if (local-variable term)
	      (begin
		(display (car (variable-name term)))
		(display "l"))
	      (display (variable-name term)))
	  (begin (display (functor-symbol (car term))) 
		 (if (pair? (cdr term))
		     (begin (display "(") 
			    (display-ref-list (cdr term))
			    (display ")")))))))

(define (display-ref-list terms)
 (if (pair? terms)
     (begin (display-ref (ref (car terms)))
            (if (pair? (cdr terms))
	        (begin (display ",") (display-ref-list (cdr terms)))))
     '#\ ))

(define (display-overlapped-ref term position)
  (if (index-variable term)
      (begin
	(display "X")
	(display term))
      (if (variable term)
	  (display (variable-name term))
	  (begin (display (functor-symbol (car term))) 
		 (if (pair? (cdr term))
		     (begin (display "(") 
			    (display-overlapped-ref-list (cdr term) position)
			    (display ")")))))))

(define (display-overlapped-ref-list terms position)
 (if (pair? terms)
     (begin 
       (if (eq? position terms) (display "{"))
       (display-overlapped-ref (ref (car terms)) position)
       (if (eq? position terms) (display "}"))
       (if (pair? (cdr terms))
	   (begin (display ",") (display-overlapped-ref-list (cdr terms) position))))
     '#\ ))

(define (display-subgoal-lists subgoal-lists)
  (if (pair? subgoal-lists)
      (begin
	(display-subgoal-list (car subgoal-lists))
	(if (pair? (cdr subgoal-lists))
	    (begin
	      (display ",")
	      (display-subgoal-lists (cdr subgoal-lists)))))))

(define (display-subgoal-list subgoal-list)
  (trail-stop)
  (bind-termlist-to-index-variables (variables-in-subgoals subgoal-list '()) 1)
  (display-subgoal-list1 subgoal-list)
  (unbind-vars))

(define (display-subgoal-list1 subgoal-list)
 (if (pair? subgoal-list)
     (begin (display-subgoal1 (car subgoal-list))
	    (if (pair? (cdr subgoal-list))
		(display ","))
            (display-subgoal-list1 (cdr subgoal-list)))))

(define (display-subgoal subgoal)
  (trail-stop)
  (bind-termlist-to-index-variables (free-vars-in-termlist (subgoal-terms subgoal) '()) 1)
  (display-subgoal1 subgoal)
  (unbind-vars))

(define (display-overlapped-subgoal subgoal position)
  (trail-stop)
  (bind-termlist-to-index-variables (free-vars-in-termlist (subgoal-terms subgoal) '()) 1)
  (display-overlapped-subgoal1 subgoal position)
  (unbind-vars))

(define (display-subgoal-terms terms)
  (trail-stop)
  (bind-termlist-to-index-variables (free-vars-in-termlist terms '()) 1)
  (display-ref-list terms)
  (unbind-vars))

(define (display-subgoal1 subgoal)
  (if (subgoal-selected subgoal) (display "["))
  (if (eq? (subgoal-pred subgoal) equality-pred)
      (begin
	(display-ref (ref (car (subgoal-terms subgoal))))
	(if (eq? (subgoal-sign subgoal) '+)
	    (display " ==")
	    (display " !="))
	(if (subgoal-ordered-eq subgoal)
	    (display "> ")
	    (display " "))
	(display-ref (ref (cadr (subgoal-terms subgoal)))))
      (begin
	(display (subgoal-sign subgoal)) 
	(display (predsym-symbol (subgoal-pred subgoal))) 
	(if (and (eq? (subgoal-pred subgoal) equality-pred) (subgoal-ordered-eq subgoal))
	    (display ">"))
	(if (> (predsym-arity (subgoal-pred subgoal)) 0) 
	    (begin (display "(")
		   (display-ref-list (subgoal-terms subgoal)) 
		   (display ")")))))
  (if (subgoal-simplified subgoal) (display "*"))
  (if (subgoal-selected subgoal) (display "]")))

(define (display-overlapped-subgoal1 subgoal position)
  (if (subgoal-selected subgoal) (display "["))
  (if (eq? (subgoal-pred subgoal) equality-pred)
      (begin
	(if (eq? position (subgoal-terms subgoal)) (display "{"))
	(display-overlapped-ref (ref (car (subgoal-terms subgoal))) position)
	(if (eq? position (subgoal-terms subgoal)) (display "}"))
	(if (eq? (subgoal-sign subgoal) '+)
	    (display " ==")
	    (display " !="))
	(if (subgoal-ordered-eq subgoal)
	    (display "> ")
	    (display " "))
	(if (eq? position (cdr (subgoal-terms subgoal))) (display "{"))
	(display-overlapped-ref (ref (cadr (subgoal-terms subgoal))) position)
	(if (eq? position (cdr (subgoal-terms subgoal))) (display "}")))
      (begin
	(display (subgoal-sign subgoal)) 
	(display (predsym-symbol (subgoal-pred subgoal))) 
	(if (and (eq? (subgoal-pred subgoal) equality-pred) (subgoal-ordered-eq subgoal))
	    (display ">"))
	(if (> (predsym-arity (subgoal-pred subgoal)) 0) 
	    (begin (display "(")
		   (display-overlapped-ref-list (subgoal-terms subgoal) position) 
		   (display ")")))))
  (if (subgoal-simplified subgoal) (display "*"))
  (if (subgoal-selected subgoal) (display "]")))

(define (display-clauses clauses)
  (trail-stop)
  (bind-termlist-to-index-variables (clauselist-variables clauses) 1)
  (display-clauses1 clauses)
  (unbind-vars))

(define (display-clauses1 clauses)
 (if (pair? clauses)
     (begin (display-clause1 (car clauses))
	    (if (pair? (cdr clauses)) (newline))
            (display-clauses1 (cdr clauses)))))

(define (display-clause cl)
  (trail-stop)
  (bind-termlist-to-index-variables (clause-variables cl) 1)
  (display-clause1 cl)
  (unbind-vars))

(define (display-clause1 cl)
 (display "clause ") (display (input-number (clause-input cl)))
 (display "(") (display (clause-depth cl)) (display ")")
 (display "[") (display (clause-unsimplified cl)) (display "]")
 (display ": ") 
 (display-subgoal-list1 (clause-subgoals cl)))

(define (display-link-vector link-vector)
  (do ((i 0 (+ i 1)))
      ((= i link-vector-size) (newline))
    (if (pair? (vector-ref link-vector i))
	(display-current-link-list (vector-ref link-vector i)))))

(define (display-link-list l)
  (if (pair? l)
      (begin
	(display-link (car l))
	(if (pair? (cdr l)) (display " , "))
	(display-link-list (cdr l)))
      (newline)))

(define (display-current-link-list l)
  (if (pair? l)
      (begin
	(if (not (eq? 'delete (link-active (car l))))
	    (begin
	      (display-link (car l))
	      (if (pair? (cdr l)) (display " , "))))
	(display-current-link-list (cdr l)))
      (newline)))

(define (display-active-link-list l)
  (if (pair? l)
      (begin
	(if (eq? #t (link-active (car l)))
	    (begin
	      (display-link (car l))
	      (if (pair? (cdr l)) (display " , "))))
	(display-active-link-list (cdr l)))
      (newline)))

(define (display-link link)
  (display "<")
  (if (link-lita link)
      (begin
	(display-subgoal (link-lita link))
	(display "(")
	(display (subgoal-depth (link-lita link)))
	(display "),"))
      (display "#f,"))
  (if (link-litb link)
      (if (pair? (link-litb link))
	  (begin
	    (display-subgoal (cdr (link-litb link)))
	    (display "(")
	    (display "factor")
	    (display "),"))
	  (begin
	    (if (eq-link? link)
		(display-overlapped-subgoal (link-litb link) (link-position link))
		(display-subgoal (link-litb link)))
	    (display "(")
	    (display (subgoal-depth (link-litb link)))
	    (display "),")))
      (if (eq-link? link)
	  (display "reflexivity,")
	  (display "#f,")))
  (display (link-weight link))
  (display ",")
  (display (link-active link))
  (display ",")
  (display (link-depth link))
  (if (eq-link? link)
      (if (link-left-to-right link)
	  (display ",->")
	  (display ",<-")))
  (display ">"))

(define (display-branch-subgoals subgoal)
  (if subgoal
      (begin
	(display-branch-subgoals (subgoal-predecessor subgoal))
	(if (subgoal-predecessor subgoal)
	    (display "-"))
	(display-subgoal subgoal))))

(define (bind-termlist-to-index-variables terms value)
  (if (pair? terms)
      (bind-termlist-to-index-variables (cdr terms) (bind-term-to-index-variables (ref (car terms)) value))
      value))

(define (bind-term-to-index-variables term value)
  ;;(print "binding " term " to " value)
  (if (index-variable term)
      value
      (if (variable term)
	  (begin
	    (if (local-variable term)
		(bind term (- value))
		(bind term value))
	    (+ value 1))
	  (if (constant term)
	      value
	      (bind-termlist-to-index-variables (cdr term) value)))))

(define (clauselist-variables clauses)
  (if (pair? clauses)
      (append (clause-variables (car clauses)) (clauselist-variables (cdr clauses)))
      '()))

(define (variables-in-subgoals subgoals vars)
  (if (null? subgoals)
      vars
      (variables-in-subgoals (cdr subgoals) (free-vars-in-termlist (subgoal-terms (car subgoals)) vars))))

(define (free-vars-in-termlist terms vars)
  ;; finds all free variables in terms and inserts them into vars exactly once
  ;;(print vars)
  (if (pair? terms)
      (begin
	(free-vars-in-termlist (cdr terms) (free-vars-in-term (ref (car terms)) vars)))
      vars))

(define (free-vars-in-term term vars)          
  ;; checks if a term is a variable.
  ;; if yes, it is added to the variable list vars if it is not already in there
  ;; otherwise if it is a proper functor, the argument list of term is searched for
  ;; free variables
  (if (not (index-variable term))
      (if (free-variable term)
	  (begin
	    (find-or-insert-var term vars))
	  (if (not (constant term))
	      (free-vars-in-termlist (cdr term) vars)
	      vars))))

(define (find-or-insert-var term vars)
  (if (pair? vars)
      (begin
	(if (not (eq? term (car vars)))
	    (if (pair? (cdr vars))
		(find-or-insert-var term (cdr vars))
		(set-cdr! vars (list term))))
	vars)
      (list term)))

;;;*************************************************************************
;; ----------
;; Statistics
;; ----------

(define (print-statistics)	
  (if (not silent)
      (begin
	(print "Proof search statistics:")
	(print "========================")
	(print "Length initial path:        " initial-length)
	(print "Linking steps (succ/fail):  " successful-inferences " / " (- total-inferences successful-inferences))
	(if (or factoring pre-factoring)
	    (print "Factoring (succ/fail/links):" successful-factorings " / " (- total-factorings successful-factorings) 
		   " / " total-factoring-links))
	(if equality
	    (begin
	      (print "Eq-linkings (succ/fail)  :  " successful-eq-inferences " / " (- total-eq-inferences successful-eq-inferences))
	      (print "Reflexivities (succ/fail):  " successful-refl-inferences " / " (- total-refl-inferences successful-refl-inferences))
	      (print "E-variants/Order viol.:     " nr-e-variants " / " expansion-order-violations)))
	(print "Twin-clause inferences:     " twin-clause-inferences)
	(if eager
	    (print "Eager eliminations:         " nr-eager))
	(if rewrite-steps
	    (print "Rewrites (subgoal/path):    " total-rewrites " / " nr-rewrites))
	(if context-lemmas 
	    (print "GR-closures total/c-foldup: " successful-reduction-steps " / " context-reduction-steps)
	    (print "GR-closures:                " successful-reduction-steps))
	(if equality
	    (print "Links non-eq/eq:            " total-links " / " total-eq-links "(" total-rewrite-links ")")
	    (print "Links generated/deleted:    " total-links " / " links-deleted))
	(if sweep-links
	    (print "Links sweeped vector/all:   " vector-links-sweeped "(" nr-vector-sweeps ") / " links-sweeped "(" nr-sweeps ")" ))
	(if pruning 
	    (print "Clauses in tableau/pruned:  " tableau-clauses " / " nr-prunings)
	    (print "Clauses in tableau:         " tableau-clauses))
	(if regularity 
	    (if tautologies
		(print "Ground regularity fails:    " regularity-fails)
		(print "Regularity/Tautology fails: " regularity-fails " / " tautology-fails " (" tautological-resolvents ")"))
	    (if (not tautologies) (print "Tautology fails:            " tautology-fails)))
	(if contract 
	    (print "Renamings/Contractions:     " total-renamings " / " nr-contractions)
	    (print "Renamings detected       :  " total-renamings) )
	(print "Variants deleted/subsumed:  " nr-variants " / " variant-subsumptions) 
	(print "Clauses by splitting:       " nr-splitted)
	(print "Variant subsumption tests:  " variant-subsumption-tests)
	(print "Fails ful/mat/sig/len:      " full-fails " / " single-match-fails " / " signature-fails " / " length-fails)
	(if pre-subsume
	    (print "Pre subsumption tests:      " pre-subsumption-tests))
	(if path-subsumption
	    (begin
	      (print "Path subsumption tests:     " path-subsumption-tests)
	      (print "Fails ful/mat/sig/len:      " full-complete-fails " / " full-single-fails " / " full-signature-fails " / " full-length-fails)))
	(print "Subsumed unit/path/back:    " nr-subsumption-deletions " / " path-subsumptions " / " backward-subsumptions)
	(if couples (print "Coupled eliminations:       " coupled-eliminations))
	(if simplify 
	    (if top-down-lemmas
		(print "Simplifications/lemmas:     " nr-simplifications " / " nr-top-down-lemmas " ("  nr-on-the-fly-lemmas ")")
		(print "Simplifications:            " nr-simplifications))
	    (if top-down-lemmas (print "Top-down-lemmas:            " nr-top-down-lemmas " ("  nr-on-the-fly-lemmas ")")))
	(if foldup 
	    (if context-lemmas
		(print "Foldups root/context:       " nr-foldups " / " nr-context-foldups)
		(print "Root foldups:               " nr-foldups))
	    (if context-lemmas (print "Context foldups:            " nr-context-foldups)))
	(if (and (not equality) context-units)
	    (print "Context links/subsumptions: " context-links " / " nr-context-subsumptions))
	(print "Maximum term depth/compl:   " max-term-depth " / " max-complexity)
	(print "Relevance vectors:          " rel-vectors-allocated)
 	(print "Unifications/failed:        " nr-unifications " / " nr-failed-unifications)
	(if forall-unify
	    (print "Terms copied:               " nr-copied-terms))
	(print "Maximum tableau depth:      " max-depth))))


;;;*************************************************************************
;; ------------
;; Lemma output
;; ------------

(define (print-tptp-lemmas subgoal)
  (if subgoal
      (begin
	(if (and (eq? 'lemma (input-type (clause-input (subgoal-clause subgoal))))
		 (not (subgoal-unlinkable subgoal)))
	    (begin
	      (display "input_clause(")
	      (display (input-name (clause-input (subgoal-clause subgoal))))
	      (display ",lemma,[")
	      (if (eq? (subgoal-sign subgoal) '+)
		  (display "++ ")
		  (display "-- "))
	      (if (eq? (predsym-symbol (subgoal-pred subgoal)) 'equal)
		  (display "equal")
		  (display (predsym-symbol (subgoal-pred subgoal))))
	      (display "(")
	      (display-subgoal-terms (subgoal-terms subgoal))
	      (display ")]).")
	      (newline)))
	(print-tptp-lemmas (subgoal-predecessor subgoal)))))

(define (print-tptp-lemma subgoal)
  (if subgoal
      (begin
	(if (and (eq? 'lemma (input-type (clause-input (subgoal-clause subgoal))))
		 (not (subgoal-unlinkable subgoal)))
	    (begin
	      (display "input_clause(")
	      (display (input-name (clause-input (subgoal-clause subgoal))))
	      (display ",lemma,[")
	      (if (eq? (subgoal-sign subgoal) '+)
		  (display "++ ")
		  (display "-- "))
	      (if (eq? (predsym-symbol (subgoal-pred subgoal)) 'equal)
		  (display "equal")
		  (display (predsym-symbol (subgoal-pred subgoal))))
	      (display "(")
	      (display-subgoal-terms (subgoal-terms subgoal))
	      (display ")]).")
	      (newline))))))

;;;*************************************************************************
;; -------------------
;; Version information
;; -------------------

(define (print-version)
  (display "DCTP version 1.31 CASC-J2")
  (newline))

;;;*************************************************************************
;; -------------------
;; Version information
;; -------------------

(define (display-tptp-clauses clauses)
  (display-formula-statistics)
  (newline)
  (if (and equality setheo-out)
      (begin
	(display "% insertion of reflexivity axiom for scheme-setheo")
	(newline)
	(display "input_clause(reflexivity,axiom,[++ equal(X, X)]).")
	(newline)
	(newline)))
  (display-tptp-clauselist clauses 1)
  (display "% output: done")
  (newline))

(define (display-tptp-clauselist clauses number)
  (if (pair? clauses)
      (begin
	(trail-stop)
	(bind-termlist-to-index-variables (variables-in-subgoals (clause-subgoals (car clauses)) '()) 1)
	(display "input_clause(clause_")
	(display number)
	(display ",")
	(if all-conjectures
	    (display "conjecture")
	    (display (string-downcase (symbol->string (input-type (clause-input (car clauses)))))))
	(display ",")
	(newline)
	(display "    [")
	(display-tptp-subgoals (clause-subgoals (car clauses)))
	(display "]).")
	(newline)
	(newline)
	(unbind-vars)
	(display-tptp-clauselist (cdr clauses) (+ number 1)))))

(define (display-tptp-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(if (eq? (subgoal-sign (car subgoals)) '+)
	    (display "++ ")
	    (display "-- "))
	(if (eq? (predsym-symbol (subgoal-pred (car subgoals))) 'equal)
	    (display "equal")
	    (display (predsym-symbol (subgoal-pred (car subgoals)))))
	(if (pair? (subgoal-terms (car subgoals)))
	    (begin
	      (display "(")
	      (display-ref-list (subgoal-terms (car subgoals)))
	      (display ")")))
	(if (pair? (cdr subgoals))
	    (begin
	      (display ",")
	      (newline)
	      (display "     ")
	      (display-tptp-subgoals (cdr subgoals)))))))
