(module index
	(include "dctp.sch")
	(export (insert-subgoal-in-index subgoal index-type) (delete-subgoal-from-index subgoal index-type) 
		(insert-subgoal-in-gr-index subgoal) (delete-subgoal-from-gr-index subgoal)
		(subgoals-in-index? subgoals index) 
		(ground-closed-with-index subgoal) (ground-termlist k) (print-indexes predsym index-type) 
		(insert-clause-in-index clause) (delete-clause-from-index clause) (clause-in-index? clause)
		(print-index-entries index p paren-stack) (subgoals-subsumed-by-index subgoals)
		(match-terms-with-complete-index terms rest-terms index) (match-subgoal-with-index subgoal action-type)
		(predicate-index-statistics predicates) (variant-index-statistics clauses) (clear-term-trail) 
		(ground-irregular-subgoal? subgoal) (subgoals-subsumed-by-gr-index subgoals)  (init-index-constants)
		(insert-unit-in-indexes unit-sg) (delete-unit-from-indexes unit-sg) (subgoals-subsumed-by-context-index subgoals)
		(unify-subgoal-with-index subgoal type)
		(subgoal-in-path-index subgoal) (unify-subgoal-with-gr-index subgoal type)
		(forall-irregular-with-index subgoal) (display-index-statistics index) (free-index-variable i) (index-variable-binding i)
		(subsume-index-with-subgoal subgoal type) (match-index-with-subgoal subgoal type) (subgoal-subsumed-by-path-index subgoal)
		(match-terms-with-index terms rest-terms index))
	(import (globals "globals.scm") (display "display.scm") (unification "unification.scm") 
		(ref-equal pruning "pruning.scm") (delete-element-from-list link "link.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define local-index-vars 0)

(define term-trail-root (list (cons #f #f)))
(define term-trail term-trail-root)

(define pos-sign-constant #f)
(define neg-sign-constant #f)
(define clause-end #f)

(define nr-index-paths 0)
(define nr-index-entries 0)
(define nr-empty-paths 0)
(define total-depth 0)
(define total-empty-depth 0)
(define total-branching 0)
(define nr-branching 0)
(define max-branching 0)

(define rewrite-index  (cons 0 '()))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Index constant init function
;; ------------------------------------------------------------

(define (init-index-constants)
  (set! clause-end (new-functor))
  (set-functor-arity clause-end 0)
  (set-functor-symbol clause-end 'clause-end)
  (set-functor-number clause-end -3)
  (set! clause-end (list (list clause-end)))
  (set! pos-sign-constant (new-functor))
  (set-functor-arity pos-sign-constant 0)
  (set-functor-symbol pos-sign-constant '+)
  (set-functor-number pos-sign-constant -1)
  (set! pos-sign-constant (list (list pos-sign-constant)))
  (set! neg-sign-constant (new-functor))
  (set-functor-arity neg-sign-constant 0)
  (set-functor-symbol neg-sign-constant '-)
  (set-functor-number neg-sign-constant -2)
  (set! neg-sign-constant (list (list neg-sign-constant))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Term trail functions
;; ------------------------------------------------------------

(define (term-trail-erase)
 (if (term-trail-content)
     (begin (term-trail-previous) (term-trail-erase))
     (begin (term-trail-previous) #f)))


(define (clear-term-trail)
  (do ((i 0 (+ i 1)))
      ((eq? term-trail term-trail-root) (print "required " i " iterations to clear term trail"))
    (term-trail-erase)))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Multiple index insert and delete of units
;; ------------------------------------------------------------

(define (insert-unit-in-indexes unit-sg)
  (insert-subgoal-in-index unit-sg 'unit)
  ;;(if (or (subgoal-ground unit-sg) (ground-termlist (subgoal-terms unit-sg)))
  ;;(if (ground-termlist (subgoal-terms unit-sg))
      ;;(set-subgoal-ground unit-sg #t)
      ;;(set-subgoal-ground unit-sg #f))
  (if gr-lemmas 
      (begin 
	(insert-subgoal-in-gr-index unit-sg)
	;;(if (and regularity (subgoal-ground unit-sg))
	(if (and regularity (subgoal-ground unit-sg))
	      (insert-subgoal-in-index unit-sg 'reg)))))

(define (delete-unit-from-indexes unit-sg)
  (delete-subgoal-from-index unit-sg 'unit)
  (if gr-lemmas
      (begin
	(delete-subgoal-from-gr-index unit-sg)
	(if (and regularity (subgoal-ground unit-sg))
	    (delete-subgoal-from-index unit-sg 'reg)))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Index search functions
;; ------------------------------------------------------------

(define (clause-in-index? clause)
  (let (index-end)
    (trail-stop)
    (set! local-index-vars 0)
    (set! index-end (subgoals-in-index? (clause-subgoals clause) (cdr (input-variant-index (clause-input clause)))))
    (unbind-vars)
    index-end))

(define (subgoals-in-index? subgoals index)
  (if (pair? subgoals)
      (let (sign-index)
	(if (eq? '+ (subgoal-sign (car subgoals)))
	    (set! sign-index (term-in-index? (car pos-sign-constant) index))
	    (set! sign-index (term-in-index? (car neg-sign-constant) index)))
	(if sign-index
	    (let ((new-index (term-in-index? (cons (subgoal-pred (car subgoals)) 
						   (subgoal-terms (car subgoals))) sign-index)))
	      (if new-index
		  (subgoals-in-index? (cdr subgoals) new-index)
		  new-index))
	    sign-index))
      (term-in-index? (car clause-end) index)))

;;; This function seems to remain unchanged with ordering...
(define (termlist-in-index? terms index)
  (if (pair? terms)
      (let ((new-index (term-in-index? (ref (car terms)) index)))
	(if new-index
	    (termlist-in-index? (cdr terms) new-index)
	    (begin 
	      (if (> debug-level 8)
		  (print "subsearch failed for term " (ref (car terms))))
	      new-index)))
      index))

(define (term-in-index? term index)
  (if (pair? index)
      (if (index-variable term)
	  (if (index-variable (and-content index)) 
	      (if (= term (and-content index)) ;; So we found the proper variable
		  (and-next index)             ;; Search on one level down
		  (if (> term (and-content index)) ;; Found variable bigger than searched variable?
		      (term-in-index? term (or-next index)) ;; We can go on searching
		      #f))                                  ;; Otherwise abort
	      #f)                                           ;; Type mismatch. Abort (Variables are stored in the front).
	  (if (variable term)
	      (begin
		(set! local-index-vars (+ local-index-vars 1)) ;; A free variable is simply bound to the next unused
		(bind term local-index-vars)                   ;; index variable
		(term-in-index? (ref term) index))             ;; The search is repeated for that binding
	      (if (eq? (car term) (and-content index)) ;; So we found the proper functor
		  (if (constant term)                  ;; Is it a constant?
		      (and-next index)                 ;; Yes. Search on one level down
		      (termlist-in-index? (cdr term) (and-next index))) ;; No, check the argument terms
		  (if (or (index-variable (and-content index)) ;; Functors are stored after vars, go on
			  (< (functor-number (car term)) (functor-number (and-content index)))) ;; Found functor smaller than
		                                                                                ;; searched functor?
		      (term-in-index? term (or-next index))
		      #f))))
      #f)) ;; most likely we hit the end of the or-cell-list


;;;*************************************************************************
;; ------------------------------------------------------------
;; Index insertion functions
;; ------------------------------------------------------------

(define (insert-object-in-index pred-index object)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; more terms to insert?
	(begin  ;; it is important here that empty lists are never pushed on the term-trail!
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (if (> debug-level 8)
	      (if (variable (ref (car current-terms)))
		  (print "index inserting variable " (ref (car current-terms)))
		  (if (index-variable (ref (car current-terms)))
		      (print "index inserting index-v  " (ref (car current-terms)))
		      (print "index inserting functor  " (functor->list (ref (car current-terms)))))))
	  (insert-term-in-index (ref (car current-terms)) pred-index object));; insert the next term
	(if (vector? (or-next pred-index)) ;; (not (memq object (vector-ref (or-next pred-index) 1)))) ;; if this is not the end of a new path
	    (begin ;; append it to the stored objects of the path
	      (vector-set! (or-next pred-index) 0 (+ (vector-ref (or-next pred-index) 0) 1))
	      (vector-set! (or-next pred-index) 1 (cons object (vector-ref (or-next pred-index) 1)))
	      object)
	    (if (null? (or-next pred-index))
		(begin  ;; make a new storage vector and place object in it
		  (set-or-next pred-index (make-vector 2))
		  (vector-set! (or-next pred-index) 0 1)
		  (vector-set! (or-next pred-index) 1 (list object))
		  object)
		#f))))) ;; apparently our terms ended somewhere in the middle of a path, abort

(define (functor->list term)
  (cons (vector-ref (car term) 0)
	(cons (vector-ref (car term) 1)
	      (cons (vector-ref (car term) 2)
		    (cons (vector-ref (car term) 3) '())))))

;; this function has been taken out of insert-object-in-index because the latter will always pop new terms
;; off the trail which would be incompatible with shallow backtracking within the or-cell-list
(define (insert-term-in-index term pred-index object)
  (if (pair? (or-next pred-index))                                  ;; have we reached the end of the or-cell-list?
      (if (index-variable term) 
	  (if (index-variable (and-content (or-next pred-index)))
	      (if (= term (and-content (or-next pred-index)))       ;; have we found the proper variable?
		  (if (insert-object-in-index (or-current (or-next pred-index)) object) ;; try to insert the rest of the
		      object
		      #f)
		  (if (> term (and-content (or-next pred-index)))
		      (insert-term-in-index term (or-next pred-index) object) ;; shallow backtracking over the index vars
		      (insert-new-cell-and-continue term pred-index object))) ;; the searched var is not there, must insert
	      (insert-new-cell-and-continue term pred-index object))          ;; same here, we already reached the functors
	  (if (variable term)
	      (begin
		(set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
		(bind term local-index-vars)                          ;; index variable
		(insert-term-in-index (ref term) pred-index object))  ;; shallow backtracking with the binding
	      (if (eq? (car term) (and-content (or-next pred-index))) ;; found the functor
		  (begin
		    (if (> (functor-arity (car term)) 0)
			(push-term-trail (cdr term)))      ;; push the args for non-constants
		    (if (insert-object-in-index (or-current (or-next pred-index)) object) ;; try to insert the rest of the
			object
			#f))
		  (if (or (index-variable (and-content (or-next pred-index))) ;; Functors are stored after vars, go on
			  (< (functor-number (car term)) (functor-number (and-content (or-next pred-index))))) 
		                                                             ;;; Found functor smaller than searched functor?
		      (insert-term-in-index term (or-next pred-index) object)   ;; shallow backtracking over the functors
		      (insert-new-cell-and-continue term pred-index object))))) ;; the searched functor is not there, insert
      (insert-new-cell-and-continue term pred-index object))) ;; insert at the end of the or-cell-list

;; this function is called whenever a term could not be found in the index
;; and new memory has to be allocated
(define (insert-new-cell-and-continue term pred-index object)
  (let ((new-cell (new-or-and-cell)))
    (set-or-next new-cell (or-next pred-index)) ;; link the new cell betwenn pred-index and pred-index's
    (set-or-next pred-index new-cell)           ;; old successor (or simply after pred-index)
    (if (index-variable term)
	(set-and-content new-cell term)                             ;; set the and-content of the new cell to the variable
	(if (variable term)
	    (begin
	      (set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
	      (bind term local-index-vars)                          ;; index variable
	      (set! term (ref term))
	      (set-and-content new-cell local-index-vars))
	    (set-and-content new-cell (car term))))  ;; or set the and-content of the new cell to the functor symbol
    (if (and (not (index-variable term)) (not (constant term)))  ;; if term is a complex function, get the args on the trail
	    (push-term-trail (cdr term)))
    (if (insert-object-in-index (or-current new-cell) object) ;; continue the insertion of the trailed terms
	object
	#f)))


(define (insert-subgoal-in-index subgoal index-type)
  (let (index index-end)
    (cond 
     ((eq? index-type 'context)
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-context-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-context-index (subgoal-pred subgoal)))))
     ((eq? index-type 'path)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-pos-path-index (subgoal-pred subgoal)))
	  (set! index (predsym-neg-path-index (subgoal-pred subgoal)))))
     ((eq? index-type 'purity)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-pos-purity-index (subgoal-pred subgoal)))
	  (set! index (predsym-neg-purity-index (subgoal-pred subgoal)))))
     ((eq? index-type 'unit)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-pos-unit-index (subgoal-pred subgoal)))
	  (set! index (predsym-neg-unit-index (subgoal-pred subgoal)))))
     ((eq? index-type 'reg)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-pos-reg-index (subgoal-pred subgoal)))
	  (set! index (predsym-neg-reg-index (subgoal-pred subgoal))))))
    (if (> debug-level 4)
	(begin
	  (display "inserting subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (and  (> debug-level 6) (> (car index) 0))
	      (begin
		(display "current index content: ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (insert-object-in-index index subgoal))
    (term-trail-erase)
    (unbind-vars)
    (if (and index-end equality (eq-subgoal? subgoal))
	;; Insert the symmetric version of the equation as well
	(begin
	  (set-car! index (+ (car index) 1))
	  (if (> debug-level 4)
	      (begin
		(display "inserting symmetric variant of subgoal ")
		(display-subgoal subgoal)
		(newline)))
	  (trail-stop)
	  (set! local-index-vars 0)
	  (term-trail-stop)
	  (push-term-trail (list (cadr (subgoal-terms subgoal)) (car (subgoal-terms subgoal))))
	  (set! index-end (insert-object-in-index index subgoal))
	  (term-trail-erase)
	  (unbind-vars)
	  (if (> debug-level 4)
	      (begin
		(display "insertion of symmetric variant done.")
		(display-subgoal subgoal)
		(newline)))
	  ))
    (if index-end
	(set-car! index (+ (car index) 1))
	(begin
	  (display "subgoal ")
	  (display-subgoal subgoal)
	  (display " already in index or bad subgoal -- insertion of subgoal failed!")
	  (newline)
	  #f))))

(define (insert-clause-in-index clause)
  (let ((index (input-variant-index (clause-input clause))) index-end)
    (if (> debug-level 4)
	(begin
	  (display "inserting ")
	  (display-clause clause)
	  (newline)))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-subgoal-termlists (clause-subgoals clause))
    (set! index-end (insert-object-in-index index clause))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	(set-car! index (+ (car index) 1))
	(begin
	  (display-clause clause)
	  (display " already in index or bad clause -- insertion of clause failed!")
	  (newline)
	  #f))))

(define (push-subgoal-termlists subgoals)
  (if (pair? subgoals)
      (begin
	(push-subgoal-termlists (cdr subgoals))
	(push-term-trail (list (cons (subgoal-pred (car subgoals)) (subgoal-terms (car subgoals)))))
	(if (eq? '+ (subgoal-sign (car subgoals)))
	    (push-term-trail pos-sign-constant)
	    (push-term-trail neg-sign-constant)))
      (begin 
	(if (> debug-level 8)
	    (print "pushing " clause-end))
	(push-term-trail clause-end))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Index deletion functions
;; ------------------------------------------------------------

(define (delete-object-from-index pred-index object)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; more terms to delete?
	(begin 
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (if (> debug-level 8)
	      (if (variable (ref (car current-terms)))
		  (print "index deleting variable " (ref (car current-terms)))
		  (if (index-variable (ref (car current-terms)))
		      (print "index deleting index-v  " (ref (car current-terms)))
		      (print "index deleting functor  " (functor->list (ref (car current-terms)))))))
	  (delete-term-from-index (ref (car current-terms)) pred-index object)) ;; delete the next term
	(if (and (vector? (or-next pred-index)) (> (vector-ref (or-next pred-index) 0) 0)) ;; if this is a proper path end
	    (let (new-objects)
	      (set! new-objects (delete-element-from-list object (vector-ref (or-next pred-index) 1)
							  (vector-ref (or-next pred-index) 1)))
	      (if new-objects           ;; if object could be deleted
		  (if (pair? new-objects)
		      (begin                 ;; adjust object vector
			(vector-set! (or-next pred-index) 0 (- (vector-ref (or-next pred-index) 0) 1))
			(vector-set! (or-next pred-index) 1 new-objects)
			object)
		      (set-or-next pred-index '())) ;; if we deleted the last of the path's objects, kill the path
		  #f))
	    #f))))  ;; our terms ended not at a proper path end, abort

(define (delete-term-from-index term pred-index object)
  (if (pair? (or-next pred-index))                                  ;; have we reached the end of the or-cell-list?
      (if (index-variable term)                                     
	  (if (index-variable (and-content (or-next pred-index)))
	      (if (= term (and-content (or-next pred-index)))       ;; have we found the proper variable?
		  (if (delete-object-from-index (or-current (or-next pred-index)) object) ;; try to delete the rest of the
		      (begin                                                            ;; trailed terms
			(if (null? (and-next (or-next pred-index))) ;; if all lower and-nodes were deleted, delete
			    (set-or-next pred-index (or-next (or-next pred-index)))) ;; this and-node as well
			object)
		      #f)
		  (if (> term (and-content (or-next pred-index)))
		      (delete-term-from-index term (or-next pred-index) object) ;; shallow backtracking over the index vars
		      #f)) ;; the searched var is not there, abort
	      #f)          ;; same here, we already reached the functors
	  (if (variable term)
	      (begin
		(set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
		(bind term local-index-vars)                          ;; index variable
		(delete-term-from-index (ref term) pred-index object))  ;; shallow backtracking with the binding
	      (if (eq? (car term) (and-content (or-next pred-index)))  ;; found the functor
		  (begin
		    (if (> (functor-arity (car term)) 0)
			(push-term-trail (cdr term)))      ;; push the args for non-constants
		    (if (delete-object-from-index (or-current (or-next pred-index)) object) ;; try to delete the rest of the
			(begin                                                            ;; trailed terms
			  (if (null? (and-next (or-next pred-index))) ;; if all lower and-nodes were deleted, delete
			      (set-or-next pred-index (or-next (or-next pred-index)))) ;; this and-node as well
			  object)
			#f))
		  (if (or (index-variable (and-content (or-next pred-index))) ;; Functors are stored after vars, go on
			  (< (functor-number (car term)) (functor-number (and-content (or-next pred-index))))) 
		                                                             ;;; Found functor smaller than searched functor?
		      (delete-term-from-index term (or-next pred-index) object)   ;; shallow backtracking over the functors
		      #f)))) ;; the searched functor is not there, abort
      #f)) ;; term not found in or-cell-list, abort

(define (delete-subgoal-from-index subgoal index-type)
  (let (index index-end)
    (cond 
     ((eq? index-type 'context)
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-context-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-context-index (subgoal-pred subgoal)))))
     ((eq? index-type 'path)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-pos-path-index (subgoal-pred subgoal)))
	  (set! index (predsym-neg-path-index (subgoal-pred subgoal)))))
     ((eq? index-type 'purity)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-pos-purity-index (subgoal-pred subgoal)))
	  (set! index (predsym-neg-purity-index (subgoal-pred subgoal)))))
     ((eq? index-type 'unit)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-pos-unit-index (subgoal-pred subgoal)))
	  (set! index (predsym-neg-unit-index (subgoal-pred subgoal)))))
     ((eq? index-type 'reg)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-pos-reg-index (subgoal-pred subgoal)))
	  (set! index (predsym-neg-reg-index (subgoal-pred subgoal))))))
    (if (> debug-level 4)
	(begin
	  (display "deleting subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (delete-object-from-index index subgoal))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	;; Delete the symmetric version of the equation as well
	(if (and equality (eq-subgoal? subgoal))
	    (begin
	      (set-car! index (- (car index) 1))
	      (if (> debug-level 4)
		  (begin
		    (display "deleting symmetric variant of subgoal ")
		    (display-subgoal subgoal)
		    (newline)))
	      (trail-stop)
	      (set! local-index-vars 0)
	      (term-trail-stop)
	      (push-term-trail (list (cadr (subgoal-terms subgoal)) (car (subgoal-terms subgoal))))
	      (set! index-end (delete-object-from-index index subgoal))
	      (term-trail-erase)
	      (unbind-vars)))
	(begin
	  (display "subgoal ")
	  (display-subgoal subgoal)
	  (display " not found in ")
	  (display index-type)
	  (display " index -- deletion of subgoal failed!")
	  (newline)))
    (if index-end
	(set-car! index (- (car index) 1))
	(begin
	  (display "symmetric variant of subgoal ")
	  (display-subgoal subgoal)
	  (display " not found in ")
	  (display index-type)
	  (display " index -- deletion of subgoal failed!")
	  (newline)
	  #f))))

(define (delete-clause-from-index clause)
  (let (index index-end new-clauses)
    (if (> debug-level 3)
	(begin
	  (display "deleting ")
	  (display-clause clause)
	  (newline)))
    (set! index (input-variant-index (clause-input clause)))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-subgoal-termlists (clause-subgoals clause))
    (set! index-end (delete-object-from-index index clause))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	(set-car! index (- (car index) 1))
	(begin
	  (display-clause clause)
 	  (display " not found in index -- deletion of clause failed!")
	  (newline)
	  #f))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Index ground regularity functions
;; ------------------------------------------------------------

(define (ground-irregular-subgoal? subgoal)
  (if (and regularity (not (subgoal-simplified subgoal)))
      (let (index index-end)
	(if (eq? '+ (subgoal-sign subgoal))
	    (set! index (predsym-pos-reg-index (subgoal-pred subgoal)))
	    (set! index (predsym-neg-reg-index (subgoal-pred subgoal))))
	(set! index-end (match-terms-with-index (subgoal-terms subgoal) '() (cdr index)))
	(if index-end
	    (set! index-end (do ((i index-end (cdr i))) ((or (not (subgoal-unlinkable (car i))) (null? i)) i))))
	(if (and index-end (not (null? index-end)))     ;; no regularity fails with
	    (begin                                      ;; back-subsumed literals!
	      (if dewey-numbers
		  (begin
		    (if (context-lemma? (car index-end))
			(display "regularity fail by c-lemma ") 
			(display "regularity fail by "))
		    (display-subgoal (car index-end))
		    (display " of depth ") 
		    (display (subgoal-depth (car index-end)))
		    (newline)))
	      (car index-end))
	    #f))
      #f))

(define (ground-termlist k)
  (if (pair? k)
      (if (ground-term (ref (car k)))
	  (ground-termlist (cdr k))
	  #f)
      #t))

(define (ground-term s)
  (if (variable s)
      #f
      (ground-termlist (cdr s))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; GR-index functions
;; ------------------------------------------------------------

(define (ground-closed-with-index subgoal)
  (let (index index-end p-subgoal i-subgoal)
    (if (> debug-level 6)
	(begin
	  (display "gr-searching for subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (set! index-end (unify-subgoal-with-gr-index subgoal 'gr))
    (set! p-subgoal (extreme-path-literal-in-lists index-end #f global-closure))
    ;; try the path index containing the context lemmas
    (set! index-end (unify-subgoal-with-gr-index subgoal 'context))
    (set! i-subgoal (extreme-path-literal-in-lists index-end #f global-closure))
    (if p-subgoal
	(if i-subgoal
	    (if global-closure
		(if (< (subgoal-depth p-subgoal) (subgoal-depth i-subgoal))
		    p-subgoal
		    i-subgoal)
		(if (> (subgoal-depth p-subgoal) (subgoal-depth i-subgoal))
		    p-subgoal
		    i-subgoal))
	    p-subgoal)
	i-subgoal)))

;; these functions deliver the topmost or lowermost subgoal from the returns
;; of a unify-with-index call depending on the top value.
(define (extreme-path-literal-in-lists subgoal-lists top-sg top)
  (if (pair? subgoal-lists)
      (let ((current-top (extreme-path-literal-in-list (car subgoal-lists) top-sg top)))
	(if (and current-top (>= (subgoal-depth current-top) 0)
		 (or (and top (or (not top-sg) (< (subgoal-depth current-top) (subgoal-depth top-sg))))
		     (and (not top) (or (not top-sg) (> (subgoal-depth current-top) (subgoal-depth top-sg))))))
	    (extreme-path-literal-in-lists (cdr subgoal-lists) current-top top)
	    (extreme-path-literal-in-lists (cdr subgoal-lists) top-sg top)))
      top-sg))

(define (extreme-path-literal-in-list subgoals top-sg top)
  (if (pair? subgoals)
      (if (and (>= (subgoal-depth (car subgoals)) 0)
	       (or (and top (or (not top-sg) (< (subgoal-depth (car subgoals)) (subgoal-depth top-sg))))
		   (and (not top) (or (not top-sg) (> (subgoal-depth (car subgoals)) (subgoal-depth top-sg))))))
	  (extreme-path-literal-in-list (cdr subgoals) (car subgoals) top)
	  (extreme-path-literal-in-list (cdr subgoals) top-sg top))
      top-sg))


(define (forall-irregular-with-index subgoal)
  (let (index index-end p-subgoal i-subgoal)
    (if (> debug-level 6)
	(begin
	  (display "gr-searching for subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (set! index-end (termlist-in-gr-index? (subgoal-terms subgoal) (cdr index)))
    (if (and (vector? index-end) (> (vector-ref index-end 0) 0))
	(if global-closure
	    (set! p-subgoal (last (vector-ref index-end 1)))
	    (set! p-subgoal (car (vector-ref index-end 1))))
	(set! p-subgoal #f))
    ;; try the path index containing the context lemmas
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-context-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-context-index (subgoal-pred subgoal))))
    (set! index-end (termlist-in-gr-index? (subgoal-terms subgoal) (cdr index)))
    (if (and (vector? index-end) (> (vector-ref index-end 0) 0))
	(if global-closure
	    (set! i-subgoal (last (vector-ref index-end 1)))
	    (set! i-subgoal (car (vector-ref index-end 1))))
	(set! i-subgoal #f))
    (if p-subgoal
	(if i-subgoal
	    (if global-closure
		(if (< (subgoal-depth p-subgoal) (subgoal-depth i-subgoal))
		    p-subgoal
		    i-subgoal)
		(if (> (subgoal-depth p-subgoal) (subgoal-depth i-subgoal))
		    p-subgoal
		    i-subgoal))
	    p-subgoal)
	i-subgoal)))

(define (last l)
  (if (pair? l)
      (if (pair? (cdr l))
	  (last (cdr l))
	  (car l))
      '()))

(define (termlist-in-gr-index? terms index)
  (if (pair? terms)
      (let (new-index)
	(if (variable (ref (car terms)))
	    ;; if we do t-closure, search for t instead of any variable
	    (if (and local-vars (local-variable (ref (car terms))))
		(set! new-index (term-in-gr-index? (ref (car terms)) index)) ;; check the local variable
		(set! new-index (term-in-gr-index? dollar-constant index)))
	    (set! new-index (term-in-gr-index? (ref (car terms)) index)))
	(if new-index
	    (termlist-in-gr-index? (cdr terms) new-index)
	    (begin 
	      (if (> debug-level 8)
		  (print "subsearch failed for term " (ref (car terms))))
	      new-index)))
      index))

(define (term-in-gr-index? term index)
  (if (pair? index)
      (if (variable term)
	  (if (index-variable (and-content index)) ;; So we found the one variable
	      (and-next index)             ;; Search on one level down
	      #f)                          ;; Type mismatch. Abort (Variables are stored in the front).
	  (if (eq? (car term) (and-content index)) ;; So we found the proper functor
	      (if (constant term)              ;; Yes. Is it a constant?
		  (and-next index)             ;; Yes. Search on one level down
		  (termlist-in-gr-index? (cdr term) (and-next index))) ;; No, check the argument terms
	      (if (or (index-variable (and-content index)) ;; Functors are stored after vars, go on
		      (< (functor-number (car term)) (functor-number (and-content index)))) ;; Found functor smaller than
		                                                                            ;; searched functor?
		  (term-in-gr-index? term (or-next index))
		  #f)))
      #f)) ;; we hit the end of the or-cell-list

(define (insert-object-in-gr-index pred-index object)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; more terms to insert?
	(begin 
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (if (variable (ref (car current-terms)))
	       ;; if we do t-closure, insert t instead of any variable
	      (if (and local-vars (local-variable (ref (car current-terms))))
		  (insert-term-in-gr-index (ref (car current-terms)) pred-index object) ;; insert the local variable
		  (insert-term-in-gr-index selected-constant pred-index object)) ;; insert the selected constant
	      (insert-term-in-gr-index (ref (car current-terms)) pred-index object))) ;; insert the next term
	  (if (vector? (or-next pred-index)) ;; if this is not the end of a new path
	      (begin ;; append it to the stored objects of the path
		(vector-set! (or-next pred-index) 0 (+ (vector-ref (or-next pred-index) 0) 1))
		(vector-set! (or-next pred-index) 1 (cons object (vector-ref (or-next pred-index) 1)))
		object)
	      (if (null? (or-next pred-index))
		  (begin  ;; make a new storage vector and place object in it
		    (set-or-next pred-index (make-vector 2))
		    (vector-set! (or-next pred-index) 0 1)
		    (vector-set! (or-next pred-index) 1 (list object))
		    object)
		  #f))))) ;; apparently our terms ended somewhere in the middle of a path, abort

;; this function has been taken out of insert-object-in-gr-index because the latter will always pop new terms
;; off the trail which would be incompatible with shallow backtracking within the or-cell-list
(define (insert-term-in-gr-index term pred-index object)
  (if (pair? (or-next pred-index))                                  ;; have we reached the end of the or-cell-list?
      (if (index-variable term) 
	  (if (index-variable (and-content (or-next pred-index)))
	      (if (= term (and-content (or-next pred-index)))       ;; have we found the proper variable?
		  (if (insert-object-in-gr-index (or-current (or-next pred-index)) object) ;; try to insert the rest of the
		      object
		      #f)
		  (if (> term (and-content (or-next pred-index)))
		      (insert-term-in-gr-index term (or-next pred-index) object) ;; shallow backtracking over the index vars
		      (insert-new-gr-cell-and-continue term pred-index object))) ;; the searched var is not there, must insert
	      (insert-new-gr-cell-and-continue term pred-index object))          ;; same here, we already reached the functors
	  (if (variable term) ;; this means we have a local variable here!
	      (if (and local-vars (local-variable term))
		  (begin
		    (set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
		    (bind term local-index-vars)                          ;; index variable
		    (insert-term-in-gr-index (ref term) pred-index object))  ;; shallow backtracking with the binding
		  (insert-term-in-gr-index selected-constant pred-index object))
	      (if (eq? (car term) (and-content (or-next pred-index))) ;; found the functor
		  (begin
		    (if (> (functor-arity (car term)) 0)
			(push-term-trail (cdr term)))      ;; push the args for non-constants
		    (if (insert-object-in-gr-index (or-current (or-next pred-index)) object) ;; try to insert the rest of the
			object
			#f))
		  (if (or (index-variable (and-content (or-next pred-index))) ;; Functors are stored after vars, go on
			  (< (functor-number (car term)) (functor-number (and-content (or-next pred-index))))) 
		                                                             ;;; Found functor smaller than searched functor?
		      (insert-term-in-gr-index term (or-next pred-index) object)   ;; shallow backtracking over the functors
		      (insert-new-gr-cell-and-continue term pred-index object))))) ;; the searched functor is not there, insert
	  (insert-new-gr-cell-and-continue term pred-index object))) ;; insert at the end of the or-cell-list

;; this function is called whenever a term could not be found in the index
;; and new memory has to be allocated
(define (insert-new-gr-cell-and-continue term pred-index object)
  (let ((new-cell (new-or-and-cell)))
    (set-or-next new-cell (or-next pred-index)) ;; link the new cell between pred-index and pred-index's
    (set-or-next pred-index new-cell)           ;; old successor (or simply after pred-index)
    (if (index-variable term)
	(set-and-content new-cell term)                             ;; set the and-content of the new cell to the variable
	(if (variable term)
	    (if (and local-vars (local-variable term))
		(begin
		  (set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
		  (bind term local-index-vars)                          ;; index variable
		  (set! term (ref term))
		  (set-and-content new-cell local-index-vars))
		(set-and-content new-cell (car selected-constant)))
	    (set-and-content new-cell (car term))))  ;; or set the and-content of the new cell to the functor symbol
    ;;(if (complex term)  ;; if term is a complex function, get the args on the trail
    (if (and (not (index-variable term)) (not (variable term)) (not (constant term)))  ;; if term is a complex function, get the args on the trail
	(push-term-trail (cdr term)))
    (if (insert-object-in-gr-index (or-current new-cell) object) ;; continue the insertion of the trailed terms
	object
	#f)))

(define (insert-subgoal-in-gr-index subgoal)
  (let (index index-end)
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (if (> debug-level 4)
	(begin
	  (display "gr-inserting subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (and  (> debug-level 6) (> (car index) 0))
	      (begin
		(display "current index content: ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (insert-object-in-gr-index index subgoal))
    (term-trail-erase)
    (unbind-vars)
    (if (and index-end equality (eq-subgoal? subgoal))
	;; Insert the symmetric version of the equation as well
	(begin
	  (set-car! index (+ (car index) 1))
	  (if (> debug-level 4)
	      (begin
		(display "gr-inserting symmetric variant of subgoal ")
		(display-subgoal subgoal)
		(newline)))
	  (trail-stop)
	  (set! local-index-vars 0)
	  (term-trail-stop)
	  (push-term-trail (list (cadr (subgoal-terms subgoal)) (car (subgoal-terms subgoal))))
	  (set! index-end (insert-object-in-gr-index index subgoal))
	  (term-trail-erase)
	  (unbind-vars)))
  (if (not (subgoal-sign subgoal))
	(begin
	  (display "inserting defunct subgoal!!!")
	  (newline)))
    (if index-end
	(set-car! index (+ (car index) 1))
	(begin
	  (display "subgoal already in gr-index or bad subgoal -- insertion of subgoal ")
	  (display-subgoal subgoal)
	  (display " failed!")
	  (newline)
	  #f))))

(define (delete-subgoal-from-gr-index subgoal)
  (let (index index-end)
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-gr-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-gr-index (subgoal-pred subgoal))))
    (if (> debug-level 4)
	(begin
	  (display "gr-deleting subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (delete-object-from-gr-index index subgoal))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	  ;; Delete the symmetric version of the equation as well
	(if (and equality (eq-subgoal? subgoal))
	    (begin
	      (set-car! index (- (car index) 1))
	      (if (> debug-level 4)
		  (begin
		    (display "gr-deleting symmetric variant of subgoal ")
		    (display-subgoal subgoal)
		    (newline)))
	      (trail-stop)
	      (set! local-index-vars 0)
	      (term-trail-stop)
	      (push-term-trail (list (cadr (subgoal-terms subgoal)) (car (subgoal-terms subgoal))))
	      (set! index-end (delete-object-from-gr-index index subgoal))
	      (term-trail-erase)
	      (unbind-vars))))
    (if index-end
	(set-car! index (- (car index) 1))
	(begin
	  (display "subgoal not found in gr-index -- deletion of subgoal ")
	  (display-subgoal subgoal)
	  (display " failed!")
	  (newline)
	  #f))))

(define (delete-object-from-gr-index pred-index object)
  (let ((current-terms (pop-term-trail)))
    (if (pair? current-terms) ;; more terms to delete?
	(begin 
	  (if (pair? (cdr current-terms)) (push-term-trail (cdr current-terms))) ;; push the rest on the trail again
	  (if (variable (ref (car current-terms)))
	      (if (and local-vars (local-variable (ref (car current-terms))))
		  (delete-term-from-gr-index (ref (car current-terms)) pred-index object) ;; insert the local variable
		  (delete-term-from-gr-index selected-constant pred-index object)) ;; insert the selected constant
	      (delete-term-from-gr-index (ref (car current-terms)) pred-index object))) ;; delete the next term
	(if (and (vector? (or-next pred-index)) (> (vector-ref (or-next pred-index) 0) 0)) ;; if this is a proper path end
	    (let (new-objects)
	      (set! new-objects (delete-element-from-list object (vector-ref (or-next pred-index) 1) 
							  (vector-ref (or-next pred-index) 1)))
	      (if new-objects           ;; if object could be deleted
		  (if (pair? new-objects)
		      (begin                 ;; adjust object vector
			(vector-set! (or-next pred-index) 0 (- (vector-ref (or-next pred-index) 0) 1))
			(vector-set! (or-next pred-index) 1 new-objects)
			object)
		      (set-or-next pred-index '())) ;; if we deleted the last of the path's objects, kill the path
		  #f))
	    #f))))  ;; our terms ended not at a proper path end, abort

(define (delete-term-from-gr-index term pred-index object)
  (if (pair? (or-next pred-index))                                  ;; have we reached the end of the or-cell-list?
      (if (index-variable term)                                     
	  (if (index-variable (and-content (or-next pred-index)))
	      (if (= term (and-content (or-next pred-index)))       ;; have we found the proper variable?
		  (if (delete-object-from-gr-index (or-current (or-next pred-index)) object) ;; try to delete the rest of the
		      (begin                                                            ;; trailed terms
			(if (null? (and-next (or-next pred-index))) ;; if all lower and-nodes were deleted, delete
			    (set-or-next pred-index (or-next (or-next pred-index)))) ;; this and-node as well
			object)
		      #f)
		  (if (> term (and-content (or-next pred-index)))
		      (delete-term-from-gr-index term (or-next pred-index) object) ;; shallow backtracking over the index vars
		      #f)) ;; the searched var is not there, abort
	      #f)          ;; same here, we already reached the functors
	  (if (variable term)
	      (if (and local-vars (local-variable term))
		  (begin
		    (set! local-index-vars (+ local-index-vars 1))        ;; A free variable is simply bound to the next unused
		    (bind term local-index-vars)                          ;; index variable
		    (delete-term-from-gr-index (ref term) pred-index object))  ;; shallow backtracking with the binding
		  (delete-term-from-gr-index selected-constant pred-index object))
	      (if (eq? (car term) (and-content (or-next pred-index)))  ;; found the functor
		  (begin
		    (if (> (functor-arity (car term)) 0)
			(push-term-trail (cdr term)))      ;; push the args for non-constants
		    (if (delete-object-from-gr-index (or-current (or-next pred-index)) object) ;; try to delete the rest of the
			(begin                                                                 ;; trailed terms
			  (if (null? (and-next (or-next pred-index))) ;; if all lower and-nodes were deleted, delete
			      (set-or-next pred-index (or-next (or-next pred-index)))) ;; this and-node as well
			  object)
			#f))
		  (if (or (index-variable (and-content (or-next pred-index))) ;; Functors are stored after vars, go on
			  (< (functor-number (car term)) (functor-number (and-content (or-next pred-index))))) 
		                                                             ;;; Found functor smaller than searched functor?
		      (delete-term-from-gr-index term (or-next pred-index) object)   ;; shallow backtracking over the functors
		      #f)))) ;; the searched functor is not there, abort
      #f)) ;; term not found in or-cell-list, abort

;;;*************************************************************************
;; ------------------------------------------------------------
;; index print functions
;; ------------------------------------------------------------

(define (print-indexes predsym index-type)
  (if (or (and (eq? index-type 'context) 
	       (or (> (car (predsym-pos-context-index predsym)) 0) (> (car (predsym-neg-context-index predsym)) 0)))
	  (and (eq? index-type 'gr) 
	       (or (> (car (predsym-pos-gr-index predsym)) 0) (> (car (predsym-neg-gr-index predsym)) 0)))
	  (and (eq? index-type 'unit) 
	       (or (> (car (predsym-pos-unit-index predsym)) 0) (> (car (predsym-neg-unit-index predsym)) 0))))
      (begin
	(if (eq? index-type 'context)
	    (display "context ")
	    (if (eq? index-type 'gr)
		(display "gr   ")
		(display "unit ")))
	(display "index entries for predsym ")
	(display (predsym-symbol predsym))
	(display " ( ")
	(if (eq? index-type 'context)
	    (display (car (predsym-pos-context-index predsym)))
	    (if (eq? index-type 'gr)
		(display (car (predsym-pos-gr-index predsym)))
		(display (car (predsym-pos-unit-index predsym)))))
	(display " pos, ")
	(if (eq? index-type 'context)
	    (display (car (predsym-neg-context-index predsym)))
	    (if (eq? index-type 'gr)
		(display (car (predsym-neg-gr-index predsym)))
		(display (car (predsym-neg-unit-index predsym)))))
	(display " neg ):")
	(newline)
	(if (eq? index-type 'context)
	    (begin
	      (print-index-entries (cdr (predsym-pos-context-index predsym)) 
				   (string-append "+" (symbol->string (predsym-symbol predsym)) "(") '())
	      (print-index-entries (cdr (predsym-neg-context-index predsym)) 
				   (string-append "-" (symbol->string (predsym-symbol predsym)) "(") '()))
	    (if (eq? index-type 'gr)
		(begin
		  (print-index-entries (cdr (predsym-pos-gr-index predsym)) 
				       (string-append "+" (symbol->string (predsym-symbol predsym)) "(") '())
		  (print-index-entries (cdr (predsym-neg-gr-index predsym)) 
				       (string-append "-" (symbol->string (predsym-symbol predsym)) "(") '()))
		(begin
		  (print-index-entries (cdr (predsym-pos-unit-index predsym)) 
				       (string-append "+" (symbol->string (predsym-symbol predsym)) "(") '())
		  (print-index-entries (cdr (predsym-neg-unit-index predsym)) 
				       (string-append "-" (symbol->string (predsym-symbol predsym)) "(") '())))))))


(define (print-index-entries index p paren-stack)
  (if (pair? index)
      (begin
	(if (index-variable (and-content index))
	    (let ((parens (decrement-paren-stack paren-stack)))
	      (if (pair? (and-next index))
		  (print-index-entries (and-next index) 
				       (string-append p "X" (number->string (and-content index)) (car parens) ",")
				       (cdr parens))
		  (print-index-entries (and-next index) 
				       (string-append p "X" (number->string (and-content index)) (car parens))
				       (cdr parens)))
	      (print-index-entries (or-next index) p paren-stack))
	    (if (index-constant (and-content index))
		(let ((parens (decrement-paren-stack paren-stack)))
		  (if (pair? (and-next index))
		      (print-index-entries (and-next index) 
					   (string-append p (symbol->string (functor-symbol (and-content index))) 
							  (car parens) ",")
					   (cdr parens))
		      (print-index-entries (and-next index) 
					   (string-append p (symbol->string (functor-symbol (and-content index))) 
							  (car parens))
					   (cdr parens)))
		  (print-index-entries (or-next index) p paren-stack))
		(begin
		  (print-index-entries (and-next index) 
				       (string-append p (symbol->string (functor-symbol (and-content index))) "(")
				       (cons (functor-arity (and-content index)) paren-stack))
		  (print-index-entries (or-next index) p paren-stack)))))
      (if (vector? index)
	  (if (> (vector-ref index 0) 0)
	      (begin
		(display p)
		(display ") [")
		(display (vector-ref index 0))
		(display "] ")
		(if (= (subgoal-size) (vector-length (car (vector-ref index 1))))
		       (display-subgoal-list (vector-ref index 1))
		       (display-clauses (vector-ref index 1)))
		(newline))))))

(define (decrement-paren-stack paren-stack)
  (if (pair? paren-stack)
      (if (> (car paren-stack) 1)
	  (cons "" (cons (- (car paren-stack) 1) (cdr paren-stack)))
	  (let ((new-paren (decrement-paren-stack (cdr paren-stack))))
	    (cons (string-append ")" (car new-paren)) (cdr new-paren))))
      (cons  "" '())))


;;;*************************************************************************
;; ------------------------------------------------------------
;; index statistics functions
;; ------------------------------------------------------------

(define (variant-index-statistics clauses)
  (if (pair? clauses)
      (begin
	(display "clause ") 
	(display (input-number (clause-input (car clauses))))
	(display ": ")
	(display-index-statistics (input-variant-index (clause-input (car clauses))))
	(newline)
	(variant-index-statistics (cdr clauses)))))

(define (predicate-index-statistics predicates)
  (if (pair? predicates)
      (begin
	(print-all-index-statistics (car predicates))
	(predicate-index-statistics (cdr predicates)))
      (newline)))
		
(define (print-all-index-statistics predsym)
  (display (predsym-symbol predsym))
  (display ":")
  (display " (p-g ")
  (display-index-statistics (predsym-pos-gr-index predsym))
  (display ")")		    
  (display " (n-g ")	    
  (display-index-statistics (predsym-neg-gr-index predsym))
  (display ")")		    
  (display " (p-r ")
  (display-index-statistics (predsym-pos-reg-index predsym))
  (display ")")		    
  (display " (n-r ")	    
  (display-index-statistics (predsym-neg-reg-index predsym))
  (display ")")		    
  (display " (p-u ")	    
  (display-index-statistics (predsym-pos-unit-index predsym))
  (display ")")		    
  (display " (n-u ")	    
  (display-index-statistics (predsym-neg-unit-index predsym))
  (display ")")
  (display " (p-p ")	    
  (display-index-statistics (predsym-pos-path-index predsym))
  (display ")")		    
  (display " (n-p ")	    
  (display-index-statistics (predsym-neg-path-index predsym))
  (display ")")
  (display " (p-c ")	    
  (display-index-statistics (predsym-pos-context-index predsym))
  (display ")")		    
  (display " (n-c ")	    
  (display-index-statistics (predsym-neg-context-index predsym))
  (display ")")
  (newline))

(define (display-index-statistics index)
  (set! nr-index-paths 0)
  (set! nr-index-entries 0)
  (set! nr-empty-paths 0)
  (set! total-depth 0)
  (set! total-empty-depth 0)
  (set! total-branching 0)
  (set! nr-branching 0)
  (set! max-branching 0)
  (if (pair? (cdr index))
      (begin
	(set! total-branching (+ total-branching (length index)))
	(set! max-branching total-branching)
	(set! nr-branching (+ 1 nr-branching))))
  (get-index-statistics (cdr index) 0)
  (if (not (= (car index) nr-index-entries))
      (begin
	(display (car index))
	(display "/")))
  (display nr-index-entries)
  (display " ")
  (display nr-index-paths)
  (display " ")
  (display nr-empty-paths)
  (display " ")
  (display (if (> nr-index-paths 0) (quotient total-depth nr-index-paths) 0))
  (display " ")
  (display (if (> nr-branching 0) (quotient total-branching nr-branching) 0))
  (display " ")
  (display max-branching))

(define (get-index-statistics index depth)
  (if (pair? index)
      (begin
	(if (pair? (and-next index))
	    (begin
	      (set! total-branching (+ total-branching (length (and-next index))))
	      (if (> (length (and-next index)) max-branching) (set! max-branching (length (and-next index))))
	      (set! nr-branching (+ 1 nr-branching))))
	(get-index-statistics (and-next index) (+ 1 depth))
	(get-index-statistics (or-next index) depth))
      (if (vector? index)
	  (begin
	    (set! nr-index-paths (+ 1 nr-index-paths))
	    (set! total-depth (+ depth total-depth))
	    (if (= 0 (vector-ref index 0)) 
		(begin
		  (set! total-empty-depth (+ depth total-empty-depth))
		  (set! nr-empty-paths (+ 1 nr-empty-paths)))
		(begin 
		  (set! nr-index-entries (+ (length (vector-ref index 1)) nr-index-entries))))))))


;;;*************************************************************************
;; ------------------------------------------------------------
;; index matching functions
;; ------------------------------------------------------------

(define (match-terms-with-index terms rest-terms index)
  (if (pair? terms)
      (let ((term (ref (car terms))))
	(if (pair? index)
	    (if (index-variable (and-content index))
		(if (free-index-variable (and-content index))
		    (begin
		      (bind-index-variable (and-content index) term)
		      (let ((result (match-terms-with-index (cdr terms) rest-terms (and-next index))))
			(unbind-index-variable (and-content index))
			(if result
			    result
			    (match-terms-with-index terms rest-terms (or-next index)))))
		    (begin
		      (if (ref-equal term (ref (index-variable-binding (and-content index))))
			  (begin
			    (let ((result (match-terms-with-index (cdr terms) rest-terms (and-next index))))
			      (if result
				  result
				  (match-terms-with-index terms rest-terms (or-next index)))))
			  (begin 
			    (match-terms-with-index terms rest-terms (or-next index))))))
		(if (variable term)
		    ;; a variable that has not already been matched above with an index-variable must fail!
		    (begin 
		      (match-terms-with-index terms rest-terms (or-next index)))
		    (if (eq? (car term) (and-content index))       ; the functor case
			(begin
			  (if (constant term)
			      (let ((result (match-terms-with-index (cdr terms) rest-terms (and-next index))))
				(if result
				    result
				    (match-terms-with-index terms rest-terms (or-next index))))
			      ;; try matching functor args
			      (let ((result (match-terms-with-index (cdr term) (cons (cdr terms) rest-terms) (and-next index))))
				(if result
				    result
				    (match-terms-with-index terms rest-terms (or-next index))))))
			(if (< (functor-number (car term)) (functor-number (and-content index)))
			    (match-terms-with-index terms rest-terms (or-next index))
			    #f))))
	    #f))
      (if (pair? rest-terms)
	  (begin
	    (match-terms-with-index (car rest-terms) (cdr rest-terms) index))
	  (begin
	    (if (vector? index)
		(if (> (vector-ref index 0) 0)
		    (begin
		      (if (> debug-level 6)
			  (begin
			    (display "Found valid index-end ")
			    (display-subgoal-list (vector-ref index 1))
			    (newline)))
		      (vector-ref index 1))
		    (begin
		      (if (> debug-level 6)
			  (begin
			    (display "Found empty index-end ")
			  (newline)))
		      #f))
		(begin 
		  (if (> debug-level 8)
		      (print "term/index length mismatch"))
		  #f))))))

(define (match-terms-with-complete-index terms rest-terms index)
  (if (pair? terms)
      (let ((term (ref (car terms))))
	(if (pair? index)
	    (if (index-variable (and-content index))
		(if (free-index-variable (and-content index))
		    (begin
		      (bind-index-variable (and-content index) term)
		      (let ((result (match-terms-with-complete-index (cdr terms) rest-terms (and-next index))))
			(unbind-index-variable (and-content index))
			(if (pair? result)
			    (append result (match-terms-with-complete-index terms rest-terms (or-next index)))
			    (match-terms-with-complete-index terms rest-terms (or-next index)))))
		    (begin
		      (if (ref-equal term (ref (index-variable-binding (and-content index))))
			  (begin
			    (let ((result (match-terms-with-complete-index (cdr terms) rest-terms (and-next index))))
			      (if (pair? result)
				  (append result (match-terms-with-complete-index terms rest-terms (or-next index)))
				  (match-terms-with-complete-index terms rest-terms (or-next index)))))
			  (begin 
			    (match-terms-with-complete-index terms rest-terms (or-next index))))))
		(if (variable term)
		    ;; a variable that has not already been matched above with an index-variable must fail!
		    (begin 
		      (match-terms-with-complete-index terms rest-terms (or-next index)))
		    (if (eq? (car term) (and-content index))       ; the functor case
			(begin
			  (if (constant term)
			      (let ((result (match-terms-with-complete-index (cdr terms) rest-terms (and-next index))))
				(if (pair? result)
				    result
				    (match-terms-with-complete-index terms rest-terms (or-next index))))
			      ;; try matching functor args
			      (let ((result (match-terms-with-complete-index (cdr term) (cons (cdr terms) rest-terms) (and-next index))))
				(if (pair? result)
				    result
				    (match-terms-with-complete-index terms rest-terms (or-next index))))))
			(if (< (functor-number (car term)) (functor-number (and-content index)))
			    (match-terms-with-complete-index terms rest-terms (or-next index))
			    '()))))
	    '()))
      (if (pair? rest-terms)
	  (begin
	    (match-terms-with-complete-index (car rest-terms) (cdr rest-terms) index))
	  (begin
	    (if (vector? index)
		(if (> (vector-ref index 0) 0)
		    (begin
		      (if (> debug-level 6)
			  (begin
			    (display "Found valid index-end ")
			    (display-subgoal-list (vector-ref index 1))
			    (newline)))
		      (vector-ref index 1))
		    (begin
		      (if (> debug-level 6)
			  (begin
			    (display "Found empty index-end ")
			  (newline)))
		      '()))
		(begin 
		  (if (> debug-level 8)
		      (print "term/index length mismatch"))
		  '()))))))

(define (free-index-variable i)
  (not (global-binding-ref i)))

(define (bound-index-variable i)
  (global-binding-ref i))

(define (bind-index-variable i term)
;   (display "binding index var X")
;   (display i)
;   (display " to ")
;   (display-ref (ref term))
;   (newline)
  (global-binding-set i (ref term)))

(define (unbind-index-variable i)
;   (display "unbinding index var X")
;   (display i)
;   (newline)
  (global-binding-set i #f))

(define (index-variable-binding i)
  (global-binding-ref i))

;; FIXME: obsolete
; (define (get-input-unit-from-subgoals subgoals)
;   (if (pair? subgoals)
;       ;;(if (and (unit-subgoal (car subgoals)) (clause-initial (subgoal-clause (car subgoals))))
;       (if (and (unit-subgoal (car subgoals)) (< (clause-depth (subgoal-clause (car subgoals))) 0))
; 	  (car subgoals)
; 	  (get-input-unit-from-subgoals (cdr subgoals)))
;       #f))

(define (global-binding-ref i)
  (if (< i (vector-length global-index-bindings))
      (vector-ref global-index-bindings i)
      (let ((new-binding (make-vector (* 2 (vector-length global-index-bindings)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length global-index-bindings)) #f)
	  (vector-set! new-binding j (vector-ref global-index-bindings j)))
	(set! global-index-bindings new-binding)
	#f)))
	
(define (global-binding-set i term)
  (if (< i (vector-length global-index-bindings))
      (vector-set! global-index-bindings i term)
      (let ((new-binding (make-vector (* 2 (vector-length global-index-bindings)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length global-index-bindings)) #f)
	  (vector-set! new-binding j (vector-ref global-index-bindings j)))
	(set! global-index-bindings new-binding)
	(vector-set! global-index-bindings i term)
	#t)))
	
(define (subgoals-subsumed-by-index subgoals)
  (if (pair? subgoals)
      (if (subgoal-simplified (car subgoals))
	  (subgoals-subsumed-by-index (cdr subgoals))
	  (let (index index-end)
	    (if (> debug-level 6)
		(begin
		  (display "Subsumption check of ")
		  (display-subgoal (car subgoals))
		  (newline)))
	    (if (eq? '+ (subgoal-sign (car subgoals)))
		(set! index (predsym-pos-unit-index (subgoal-pred (car subgoals))))
		(set! index (predsym-neg-unit-index (subgoal-pred (car subgoals)))))
	    (if (> debug-level 8)
		(begin
		  (print "index entries: ")
		  (print-index-entries (cdr index) "(" '())
		  (newline)))
	    (set! index-end (match-terms-with-index (subgoal-terms (car subgoals)) '() (cdr index)))
	    (if index-end 
		(car index-end)
		(subgoals-subsumed-by-index (cdr subgoals)))))
      #f))

(define (subgoals-subsumed-by-context-index subgoals)
  (if (pair? subgoals)
      (if (subgoal-simplified (car subgoals))
	  (subgoals-subsumed-by-context-index (cdr subgoals))
	  (let (index index-end)
	    (if (> debug-level 6)
		(begin
		  (display "Context subsumption check of ")
		  (display-subgoal (car subgoals))
		  (newline)))
	    (if (eq? '+ (subgoal-sign (car subgoals)))
		(set! index (predsym-pos-context-index (subgoal-pred (car subgoals))))
		(set! index (predsym-neg-context-index (subgoal-pred (car subgoals)))))
	    (if (> debug-level 8)
		(begin
		  (print "index entries: ")
		  (print-index-entries (cdr index) "(" '())
		  (newline)))
	    (set! index-end (match-terms-with-index (subgoal-terms (car subgoals)) '() (cdr index)))
	    (if index-end 
		(car index-end)
		(subgoals-subsumed-by-context-index (cdr subgoals)))))
      #f))

(define (subgoals-subsumed-by-gr-index subgoals)
  (if (pair? subgoals)
      (if (subgoal-simplified (car subgoals))
	  (subgoals-subsumed-by-gr-index (cdr subgoals))
	  (let (index index-end)
	    (if (> debug-level 6)
		(begin
		  (display "Extended regularity check of ")
		  (display-subgoal (car subgoals))
		  (newline)))
	    (if (eq? '+ (subgoal-sign (car subgoals)))
		(set! index (predsym-pos-gr-index (subgoal-pred (car subgoals))))
		(set! index (predsym-neg-gr-index (subgoal-pred (car subgoals)))))
	    (if (> debug-level 8)
		(begin
		  (print "index entries: ")
		  (print-index-entries (cdr index) "(" '())
		  (newline)))
	    (set! index-end (match-terms-with-index (subgoal-terms (car subgoals)) '() (cdr index)))
	    (if index-end 
		(car index-end)
		(subgoals-subsumed-by-gr-index (cdr subgoals)))))
      #f))

(define (match-subgoal-with-index subgoal action-type)
  (let (index index-end)
    (if (or (and (eq? action-type 'simplify) (eq? '+ (subgoal-sign subgoal))) (and (eq? action-type 'subsume) (eq? '- (subgoal-sign subgoal))))
	(set! index (predsym-neg-unit-index (subgoal-pred subgoal)))
	(set! index (predsym-pos-unit-index (subgoal-pred subgoal))))
    (if (> debug-level 5)
	(begin
	  (display "matching subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (> debug-level 6)
	      (begin
		(display "current index content: ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (set! index-end (match-terms-with-index (subgoal-terms subgoal) '() (cdr index)))
    (if index-end
	(car index-end)
	index-end)))


;;;*************************************************************************
;; ------------------------------------------------------------
;; path subsumption pre-filtering function
;; ------------------------------------------------------------

(define (subgoal-subsumed-by-path-index subgoal)
  (let (index index-end)
    (if (> debug-level 5)
	(begin
	  (display "path index-searching for subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-path-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-path-index (subgoal-pred subgoal))))
    (set! index-end (match-terms-with-complete-index (subgoal-terms subgoal) '() (cdr index)))
    (if (not (pair? index-end))
	(set! index-end '()))
    index-end))

(define (subgoal-in-path-index subgoal)
  (let (index index-end)
    (if (> debug-level 5)
	(begin
	  (display "gr-searching for subgoal ")
	  (display-subgoal subgoal)
	  (newline)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set! index (predsym-pos-path-index (subgoal-pred subgoal)))
	(set! index (predsym-neg-path-index (subgoal-pred subgoal))))
    (trail-stop)
    (set! local-index-vars 0)
    (set! index-end (termlist-in-index? (subgoal-terms subgoal) (cdr index)))
    (unbind-vars)
    (if (or (not (vector? index-end)) (= (vector-ref index-end 0) 0))
	(set! index-end #f)
	(set! index-end (vector-ref index-end 1)))
    index-end))


;;;*************************************************************************
;; ------------------------------------------------------------
;; rewrite-index functions
;; ------------------------------------------------------------

(define (match-term-with-rewrite-index? term)
  (if (> debug-level 5)
      (begin
	(display "matching rewritable term ")
	(display-ref term)
	(newline)
	(if (> debug-level 6)
	    (begin
	      (display "current index content: ") 
	      (newline)
	      (print-index-entries (cdr rewrite-index) "(" '())))))
  (match-terms-with-index (list term) '() (cdr rewrite-index)))

(define (insert-equation-in-rewrite-index subgoal)
  (let (index index-end)
    (if (> debug-level 4)
	(begin
	  (display "inserting rewrite-equation ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (and  (> debug-level 6) (> (car rewrite-index) 0))
	      (begin
		(display "current rewrite-index content: ") 
		(newline)
		(print-index-entries (cdr rewrite-index) "(" '())))))
    (trail-stop)
    (set! local-index-vars 0)
    (term-trail-stop)
    (push-term-trail (subgoal-terms subgoal))
    (set! index-end (insert-object-in-index rewrite-index subgoal))
    (term-trail-erase)
    (unbind-vars)
    (if index-end
	(begin
	  (set-car! rewrite-index (+ (car rewrite-index) 1))
	  ;; Insert the symmetric version of the equation as well
	  (if (and equality (eq-subgoal? subgoal) (not (subgoal-ordered-eq subgoal)))
		   ;;(not (ref-equal (cadr (subgoal-terms subgoal)) (car (subgoal-terms subgoal)))))
	      (begin
		(if (> debug-level 4)
		    (begin
		      (display "inserting symmetric variant of equation ")
		      (display-subgoal subgoal)
		      (newline)))
		(trail-stop)
		(set! local-index-vars 0)
		(term-trail-stop)
		(push-term-trail (list (cadr (subgoal-terms subgoal)) (car (subgoal-terms subgoal))))
		(set! index-end (insert-object-in-index rewrite-index subgoal))
		(term-trail-erase)
		(unbind-vars)))))
    (if index-end
	(set-car! rewrite-index (+ (car rewrite-index) 1))
	(begin
	  (display "subgoal ")
	  (display-subgoal subgoal)
	  (display " already in rewrite-index or bad subgoal -- insertion of subgoal failed!")
	  (newline)
	  #f))))


;; ------------------------------------------------------------
;; index unification functions
;; ------------------------------------------------------------

(define (unify-subgoal-with-index subgoal type)
  (let (index result)
    (cond
     ((eq? type 'unit)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-unit-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-unit-index (subgoal-pred subgoal)))))
     ((eq? type 'purity)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-purity-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-purity-index (subgoal-pred subgoal)))))
     ((eq? type 'context)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-context-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-context-index (subgoal-pred subgoal)))))
     ((eq? type 'path)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-path-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-path-index (subgoal-pred subgoal))))))
    (if (> debug-level 5)
	(begin
	  (display "unifying subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (> debug-level 6)
	      (begin
		(display "current index content of ")
		(if (eq? '+ (subgoal-sign subgoal)) (display "-") (display "+"))
		(display (predsym-symbol (subgoal-pred subgoal)))
		(display ": ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (set! result (unify-terms-with-index (subgoal-terms subgoal) '() (cdr index) '() 'standard))
    (if result
	result
	(begin
	  (if (= (car index) 0)
	      '()
	      (begin
		(print "index unification error in non-empty index!")
		'()))))))

;; FIXME: this appears to be overly complicated! as ref also derefences index variables, part of the below code is redundant.
;; a bound index-variable cannot occur in term.
;; FIXME: maybe this can be done better... why not bind a shared variable to selected/dollar-constant before repeating the recursion?
(define (unify-terms-with-index terms rest-terms index prior-results type)
  (if (pair? terms)
      (let ((result '()) (term (ref (car terms))))
	(if (pair? index)
	    (cond
	     ((and (eq? type 'gr) (variable term))
	      (if (and local-vars (local-variable term))
		  (begin
		    (if (index-variable (and-content index))
			(if (free-index-variable (and-content index))
			    (begin
			      (bind-index-variable (and-content index) term)
			      (set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
			      (unbind-index-variable (and-content index)))
			    (begin
			      (if (ref-equal term (ref (index-variable-binding (and-content index)))) ;; if the index-variable already has been
				  ;; bound to the local var
				  (set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
				  (if (not (index-occurs-term term (ref (index-variable-binding (and-content index)))))
				      (begin
					(bind-var term (ref (index-variable-binding (and-content index))))
					(set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
					(unbind-var term))
				      (set! result prior-results)))))
			(if (index-constant (and-content index))
			    (begin
			      (bind-var term (cons (and-content index) '()))
			      (set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
			      (unbind-var term))
			    ;; enter the complex case:
			    ;; the current index content is a proper functor. So we need to call a different routine
			    ;; that gathers up all unifiable index paths, does on the fly occurs checking and
			    ;; returns all unifying subgoals while iteratively building the unification term for the variable
			    (let ((built-term (list (and-content index))))
			      (set! result (unify-terms-with-index-building-instantiation term built-term 
											  (list (cons (functor-arity (and-content index)) built-term))
											  (cdr terms) rest-terms (and-next index) prior-results type)))))
		    ;; result isn't set here as the concatenation of the shallow backtracking is the final result that is passed back
		    (unify-terms-with-index terms rest-terms (or-next index) result type))
		  (unify-terms-with-index (cons selected-constant (cdr terms)) rest-terms index prior-results type)))
	     ((variable term)
	      (if (index-variable (and-content index))
		  (if (free-index-variable (and-content index))
		      (begin
			(bind-index-variable (and-content index) term)
			(set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
			(unbind-index-variable (and-content index)))
		      (begin
			(if (ref-equal term (ref (index-variable-binding (and-content index))))
			    (set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
			    (if (not (index-occurs-term term (ref (index-variable-binding (and-content index)))))
				(begin
				  (bind-var term (ref (index-variable-binding (and-content index))))
				  (set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
				  (unbind-var term))
				(set! result prior-results)))))
		  (if (index-constant (and-content index))
		      (begin
			(bind-var term (cons (and-content index) '()))
			(set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
			(unbind-var term))
		      ;; enter the complex case:
		      ;; the current index content is a proper functor. So we need to call a different routine
		      ;; that gathers up all unifiable index paths, does on the fly occurs checking and
		      ;; returns all unifying subgoals while iteratively building the unification term for the variable
		      (let ((built-term (list (and-content index))))
			(set! result (unify-terms-with-index-building-instantiation term built-term 
										    (list (cons (functor-arity (and-content index)) built-term))
										    (cdr terms) rest-terms (and-next index) prior-results type)))))
	      ;; result isn't set here as the concatenation of the shallow backtracking is the final result that is passed back
	      (unify-terms-with-index terms rest-terms (or-next index) result type))
	     ((index-variable term)
	      (if (free-index-variable term)
		  (begin
		    (if (index-variable (and-content index))
			(if (free-index-variable (and-content index))
			    (if (eq? term (and-content index))
				(set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
				(begin
				  (bind-index-variable (and-content index) term)
				  (set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
				  (unbind-index-variable (and-content index))))
			    (if (not (index-occurs-term term (ref (index-variable-binding (and-content index)))))
				(begin
				  (bind-index-variable term (ref (index-variable-binding (and-content index))))
				  (set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
				  (unbind-index-variable term))
				(set! result prior-results)))
			(if (index-constant  (and-content index))
			    (begin
			      (bind-index-variable term (cons (and-content index) '()))
			      (set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
			      (unbind-index-variable term))
			    ;; here comes another instance of the complex case:
			    ;; actually the same as above, only the variable to be instantiated here is an index-variable
			    (let ((built-term (list (and-content index))))
			      (set! result (unify-terms-with-index-building-instantiation term built-term 
											  (list (cons (functor-arity (and-content index)) built-term)) 
											  (cdr terms) rest-terms (and-next index) prior-results type)))))
		    (unify-terms-with-index terms rest-terms (or-next index) result type))
		  ;; the else case: term is a bound index variable, so the function call is repeated with the explicit binding
		  ;; in this case no shallow backtracking is required, as it is performed in the repeat call
		  ;; FIXME: This is the unlikely case. just for the record, we place a print statement here
		  (begin
		    (print "Encountered bound index variable in index unification term!")
		    (unify-terms-with-index (cons (ref (index-variable-binding term)) (cdr terms)) rest-terms index prior-results type))))
	     ((or (constant term) (complex term))
	      (if (index-variable (and-content index))
		  (begin
		    (if (free-index-variable (and-content index))
			(begin
			  (if (or (constant term) (not (index-occurs-term (and-content index) term)))
			      (begin
				(bind-index-variable (and-content index) term)
				(set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
				(unbind-index-variable (and-content index)))
			      (begin
				(set! result prior-results))))
			(begin
			  (trail-stop)
			  (if (index-unify-terms term (ref (index-variable-binding (and-content index))) type)
			      (begin
				(set! result (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type))
				(index-unbind-vars))
			      (set! result prior-results))))
		    (unify-terms-with-index terms rest-terms (or-next index) result type))
		  (if (eq? (car term) (and-content index))       ; functor symbol match, no shallow backtracking required
		      (if (constant term)
			  (unify-terms-with-index (cdr terms) rest-terms (and-next index) prior-results type)
			  ;; try matching functor args
			  (unify-terms-with-index (cdr term) (cons (cdr terms) rest-terms) (and-next index) prior-results type))
		      (if (< (functor-number (car term)) (functor-number (and-content index)))
			  (unify-terms-with-index terms rest-terms (or-next index) prior-results type)
			  prior-results))))
	     (else
	      (print "index unification error: unidentified term type")
	      (unify-terms-with-index terms rest-terms (or-next index) prior-results type)))
	    prior-results))
      (if (pair? rest-terms)
	  (begin
	    (unify-terms-with-index (car rest-terms) (cdr rest-terms) index prior-results type))
	  (begin
	    (if (vector? index)
		(if (> (vector-ref index 0) 0)
		    (begin
		      (if (> debug-level 6)
			  (begin
			    (display "Found valid index-end ")
			    (display-subgoal-list (vector-ref index 1))
			    (newline)))
		      (cons (vector-ref index 1) prior-results))
		    (begin
		      (if (> debug-level 6)
			  (begin
			    (display "Found empty index-end ")
			    (newline)))
		      '()))
		(begin 
		  (if (> debug-level 6)
		      (print "term/index length mismatch"))
		  #f))))))

	      
(define (index-unify-lists k l type)
;; unifies termlists k and l.
;; If unification succeeds, #t is returned, else #f.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     #t
     (if (index-unify-terms (ref (car k)) (ref (car l)) type)
	 (index-unify-lists (cdr k) (cdr l) type)
	 #f)))

(define (index-unify-terms s t type)          
  ;; unifies terms s and t. 
  ;; If unification succeeds, #t is returned, else #f.
  ;; Side effect: terms and trail destructively modified.
  ;; Before every unification, one has to call (trail-stop) in order
  ;; to mark the trail position for the stored substitution
  ;; the following is an add-on for the 'gr-case: if either s or t are shared variables, they are bound to the appropriate constant
  ;; and the call to index-unify-terms is redone.
  (if (and (eq? type 'gr) (or (and local-vars (or (shared-variable s) (shared-variable t)))
			      (and (not local-vars) (or (variable s) (variable t)))))
      (begin
	(if (or (and local-vars (shared-variable s))
		(and (not local-vars) (variable s)))
	    (bind s selected-constant))
	(if (or (and local-vars (shared-variable t))
		(and (not local-vars) (variable t)))
	    (bind t selected-constant))
	(index-unify-terms (ref s) (ref t) type))
      (if (eq? s t)
	  #t
	  (if (variable s)
	      (if (variable t) 
		  (bind s t)
		  (if (index-variable t)
		      (if (free-index-variable t)
			  (index-bind t s)
			  ;; FIXME: again, this cannot happen. for bound index variables, the binding is passed, not the variable
			  (begin
			    (print "encountered bound index-variable in index-unify-terms!")
			    (if (index-occurs-termlist s (ref (index-variable-binding t)))
				(index-unbind-vars)
				(bind s t))))
		      (if (index-occurs-termlist s (cdr t))
			  (index-unbind-vars)
			  (bind s t))))
	      (if (index-variable s)
		  (if (free-index-variable s)
		      (if (index-occurs-term s t)
			  (index-unbind-vars)
			  (index-bind s t))
		      (index-unify-terms (ref (index-variable-binding s)) t type))
		  (if (variable t) ;; at this point, s can be neither a variable nor an index-variable
		      (if (index-occurs-termlist t (cdr s))
			  (index-unbind-vars)
			  (bind t s))
		      (if (index-variable t)
			  (if (free-index-variable t)
			      (if (index-occurs-termlist t (cdr s))
				  (index-unbind-vars)
				  (index-bind t s))
			      (index-unify-terms s (ref (index-variable-binding t)) type))
			  (if (eq? (car s) (car t))
			      (index-unify-lists (cdr s) (cdr t) type)
			      (index-unbind-vars)))))))))

(define (index-occurs-term var term)
  (if (eq? var term)
      #t
      (if (variable term)
	  #f
	  (if (index-variable term)
	      (if (bound-index-variable term)
		  (index-occurs-term var (ref (index-variable-binding term)))
		  #f)
	      (if (constant term)
		  #f
		  (index-occurs-termlist var (cdr term)))))))

(define (index-occurs-termlist index-var terms)
  (if (pair? terms)
      (if (index-occurs-term index-var (ref (car terms)))
	  #t
	  (index-occurs-termlist index-var (cdr terms)))
      #f))

(define (index-unbind-vars)
;; makes free all variables (both normal and index) back to the previous trailstop
 (if (trail-content)
     (begin
       (if (index-variable (trail-content))
	   (unbind-index-variable (trail-content))
	   (set-car! (trail-content) '()))
       (trail-previous) (index-unbind-vars))
     (begin (trail-previous) #f)))

(define (index-bind var term)
;; binds index-variable var to term destructively, next trail position is used
  (begin (bind-index-variable var term)
	 (trail-next)
	 (set-car! (car trail) var)
	 #t))

(define (unify-terms-with-index-building-instantiation var built-term counter-pointers terms rest-terms index prior-results type)
  (if (pair? counter-pointers)
      (if (= 0 (caar counter-pointers))
	  (unify-terms-with-index-building-instantiation var built-term (cdr counter-pointers) terms rest-terms index prior-results type)
	  (if (pair? index)
	      (let ((result prior-results) (counter (caar counter-pointers)) (pointer (cdar counter-pointers))
		    (counter-pointer (car counter-pointers)))
		(if (> debug-level 6)
		    (begin
		      (display "entering term building with initial term ")
		      (display-ref built-term)
		      (newline)))
		(set-car! counter-pointer (- (car counter-pointer) 1)) ;; decrement remaining arguments of current index-side functor
		;; the following case distinction requires occurs checks in two of the cases. should an occurs check fail, nothing
		;; is done, the above manipulations of the built term are undone and the passed result is the pre-set prior-results
		;; the main difference between all the following cases is of course the type of subterm appended at pointer
		(if (index-variable (and-content index)) ;; occurs-check required here
		    (if (free-index-variable (and-content index))
			(if (not (eq? (and-content index) var)) ;; direct occurs check
			    (begin                      ;; now add the current variable to the built term and proceed
			      (set-cdr! pointer (list (and-content index)))
			      (set-cdr! counter-pointer (cdr pointer)) ;; cdr down the append pointer for the next argument
			      (set! result (unify-terms-with-index-building-instantiation var built-term counter-pointers 
											  terms rest-terms (and-next index) prior-results type))))
			(if (not (index-occurs-term var (ref (index-variable-binding (and-content index))))) ;; recursive occurs-check for binding
			    (begin                      ;; now add the current variable to the built term and proceed
			      (set-cdr! pointer (list (ref (index-variable-binding (and-content index)))))
			      (set-cdr! counter-pointer (cdr pointer)) ;; cdr down the append pointer for the next argument
			      (set! result (unify-terms-with-index-building-instantiation var built-term counter-pointers 
											  terms rest-terms (and-next index) prior-results type)))))
		    (if (index-constant (and-content index))
			(begin                      ;; now add the current constant to the built term and proceed
			  (set-cdr! pointer (list (list (and-content index))))
			  (set-cdr! counter-pointer (cdr pointer)) ;; cdr down the append pointer for the next argument
			  (set! result (unify-terms-with-index-building-instantiation var built-term counter-pointers 
										      terms rest-terms (and-next index) prior-results type)))
			;;(let ((new-counter-pointer (cons (functor-arity (and-content index)) (list (list (and-content index))))))
			(let (new-counter-pointer new-pointer)
			  ;; the most complex case: push a new counter-pointer pair and add the new functor
			  (set! new-pointer (list (and-content index)))
			  (set! new-counter-pointer (cons (functor-arity (and-content index)) new-pointer))
			  (if (> debug-level 6)
			      (begin
				(display "pushing new functor ")
				(display (functor-symbol (and-content index)))
				(newline)))
			  (set-cdr! pointer (list new-pointer))
			  (set-cdr! counter-pointer (cdr pointer)) ;; cdr down the append pointer for the next argument
			  (set! result (unify-terms-with-index-building-instantiation var built-term (cons new-counter-pointer counter-pointers)
										      terms rest-terms (and-next index) prior-results type)))))
		(set-car! counter-pointer (+ (car counter-pointer) 1)) ;; undo term appendage
		(set-cdr! counter-pointer pointer)
		(set-cdr! pointer '())
		(unify-terms-with-index-building-instantiation var built-term counter-pointers terms rest-terms (or-next index) result type))
	      (begin
		(if (> debug-level 6)
		    (begin
		      (print "aborting recursion due to null-index")))
		prior-results)))
      (let ((result '()))
	(if (> debug-level 6)
	    (begin
	      (display "binding variable to term ")
	      (display-ref built-term)
	      (newline)))
	(if (index-variable var)
	    (bind-index-variable var built-term)
	    (bind-var var built-term))
	(set! result (unify-terms-with-index terms rest-terms index prior-results type))
	(if (index-variable var)
	    (unbind-index-variable var)
	    (unbind-var var))
	result)))

;;;*************************************************************************
;; ------------------------------------------------------------
;; inverse index matching functions: term variables are instantiated only
;; ------------------------------------------------------------


;; function for index simplification
(define (match-index-with-subgoal subgoal type)
  (let (index result)
    (cond
     ((eq? type 'unit)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-unit-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-unit-index (subgoal-pred subgoal)))))
     ((eq? type 'purity)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-purity-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-purity-index (subgoal-pred subgoal)))))
     ((eq? type 'path)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-path-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-path-index (subgoal-pred subgoal))))))
    (if (> debug-level 5)
	(begin
	  (display "matching subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (> debug-level 6)
	      (begin
		(display "current index content of ")
		(if (eq? '+ (subgoal-sign subgoal)) (display "-") (display "+"))
		(display (predsym-symbol (subgoal-pred subgoal)))
		(display ": ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (set! result (match-index-with-terms (subgoal-terms subgoal) '() (cdr index) '()))
    (if result
	result
	(begin
	  (if (= (car index) 0)
	      '()
	      (begin
		(print "index matching error in non-empty index!")
		'()))))))

;; function for index subsumption
(define (subsume-index-with-subgoal subgoal type)
  (let (index result)
    (cond
     ((eq? type 'unit)
      (if (eq? '- (subgoal-sign subgoal))
	  (set! index (predsym-neg-unit-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-unit-index (subgoal-pred subgoal)))))
     ((eq? type 'purity)
      (if (eq? '- (subgoal-sign subgoal))
	  (set! index (predsym-neg-purity-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-purity-index (subgoal-pred subgoal)))))
     ((eq? type 'path)
      (if (eq? '- (subgoal-sign subgoal))
	  (set! index (predsym-neg-path-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-path-index (subgoal-pred subgoal))))))
    (if (> debug-level 5)
	(begin
	  (display "matching subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (> debug-level 6)
	      (begin
		(display "current index content of ")
		(if (eq? '+ (subgoal-sign subgoal)) (display "-") (display "+"))
		(display (predsym-symbol (subgoal-pred subgoal)))
		(display ": ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (set! result (match-index-with-terms (subgoal-terms subgoal) '() (cdr index) '()))
    (if result
	result
	(begin
	  (if (= (car index) 0)
	      '()
	      (begin
		(print "index matching error in non-empty index!")
		'()))))))

(define (match-index-with-terms terms rest-terms index prior-results)
  (if (pair? terms)
      (let ((result '()) (term (ref (car terms))))
	(if (pair? index)
	    (cond
	     ((variable term)
	      (if (index-variable (and-content index))
		  (begin
		    (bind-var term (and-content index))
		    (set! result (match-index-with-terms (cdr terms) rest-terms (and-next index) prior-results))
		    (unbind-var term))
		  (if (index-constant (and-content index))
		      (begin
			(bind-var term (cons (and-content index) '()))
			(set! result (match-index-with-terms (cdr terms) rest-terms (and-next index) prior-results))
			(unbind-var term))
		      ;; enter the complex case:
		      ;; the current index content is a proper functor. So we need to call a different routine
		      ;; that gathers up all unifiable index paths, does on the fly occurs checking and
		      ;; returns all unifying subgoals while iteratively building the unification term for the variable
		      (let ((built-term (list (and-content index))))
			(set! result (match-index-with-terms-building-instantiation term built-term 
										    (list (cons (functor-arity (and-content index)) built-term))
										    (cdr terms) rest-terms (and-next index) prior-results)))))
	      ;; result isn't set here as the concatenation of the shallow backtracking is the final result that is passed back
	      (match-index-with-terms terms rest-terms (or-next index) result))
	     ((index-variable term)
	      (if (eq? term (and-content index))
		  (set! result (match-index-with-terms (cdr terms) rest-terms (and-next index) prior-results)))
	      (match-index-with-terms terms rest-terms (or-next index) result))
	     ((index-variable (and-content index))
	      (match-index-with-terms terms rest-terms (or-next index) prior-results))
	     ((or (constant term) (complex term))
	      (if (eq? (car term) (and-content index))       ; functor symbol match, no shallow backtracking required
		  (if (constant term)
		      (match-index-with-terms (cdr terms) rest-terms (and-next index) prior-results)
		      ;; try matching functor args
		      (match-index-with-terms (cdr term) (cons (cdr terms) rest-terms) (and-next index) prior-results))
		  (if (< (functor-number (car term)) (functor-number (and-content index)))
		      (match-index-with-terms terms rest-terms (or-next index) prior-results)
		      prior-results)))
	     (else
	      (print "index unification error: unidentified term type")
	      (match-index-with-terms terms rest-terms (or-next index) prior-results)))
	    prior-results))
      (if (pair? rest-terms)
	  (begin
	    (match-index-with-terms (car rest-terms) (cdr rest-terms) index prior-results))
	  (begin
	    (if (vector? index)
		(if (> (vector-ref index 0) 0)
		    (begin
		      (if (> debug-level 6)
			  (begin
			    (display "Found valid index-end ")
			    (display-subgoal-list (vector-ref index 1))
			    (newline)))
		      (cons (vector-ref index 1) prior-results))
		    (begin
		      (if (> debug-level 6)
			  (begin
			    (display "Found empty index-end ")
			    (newline)))
		      '()))
		(begin 
		  (if (> debug-level 6)
		      (print "term/index length mismatch"))
		  #f))))))

(define (match-index-with-terms-building-instantiation var built-term counter-pointers terms rest-terms index prior-results)
  (if (pair? counter-pointers)
      (if (= 0 (caar counter-pointers))
	  (match-index-with-terms-building-instantiation var built-term (cdr counter-pointers) terms rest-terms index prior-results)
	  (if (pair? index)
	      (let ((result prior-results) (counter (caar counter-pointers)) (pointer (cdar counter-pointers))
		    (counter-pointer (car counter-pointers)))
		(if (> debug-level 6)
		    (begin
		      (display "entering term building with initial term ")
		      (display-ref built-term)
		      (newline)))
		(set-car! counter-pointer (- (car counter-pointer) 1)) ;; decrement remaining arguments of current index-side functor
		(if (index-variable (and-content index)) ;; occurs-check required here
		    (begin                      ;; now add the current variable to the built term and proceed
		      (set-cdr! pointer (list (and-content index)))
		      (set-cdr! counter-pointer (cdr pointer)) ;; cdr down the append pointer for the next argument
		      (set! result (match-index-with-terms-building-instantiation var built-term counter-pointers 
										  terms rest-terms (and-next index) prior-results)))
		    (if (index-constant (and-content index))
			(begin                      ;; now add the current constant to the built term and proceed
			  (set-cdr! pointer (list (list (and-content index))))
			  (set-cdr! counter-pointer (cdr pointer)) ;; cdr down the append pointer for the next argument
			  (set! result (match-index-with-terms-building-instantiation var built-term counter-pointers 
										      terms rest-terms (and-next index) prior-results)))
			(let (new-counter-pointer new-pointer)
			  ;; the most complex case: push a new counter-pointer pair and add the new functor
			  (set! new-pointer (list (and-content index)))
			  (set! new-counter-pointer (cons (functor-arity (and-content index)) new-pointer))
			  (if (> debug-level 6)
			      (begin
				(display "pushing new functor ")
				(display (functor-symbol (and-content index)))
				(newline)))
			  (set-cdr! pointer (list new-pointer))
			  (set-cdr! counter-pointer (cdr pointer)) ;; cdr down the append pointer for the next argument
			  (set! result (match-index-with-terms-building-instantiation var built-term (cons new-counter-pointer counter-pointers)
										      terms rest-terms (and-next index) prior-results)))))
		(set-car! counter-pointer (+ (car counter-pointer) 1)) ;; undo term appendage
		(set-cdr! counter-pointer pointer)
		(set-cdr! pointer '())
		(match-index-with-terms-building-instantiation var built-term counter-pointers terms rest-terms (or-next index) result))
	      (begin
		(if (> debug-level 6)
		    (begin
		      (print "aborting recursion due to null-index")))
		prior-results)))
      (let ((result '()))
	(if (> debug-level 6)
	    (begin
	      (display "binding variable to term ")
	      (display-ref built-term)
	      (newline)))
	(if (index-variable var)
	    (bind-index-variable var built-term)
	    (bind-var var built-term))
	(set! result (match-index-with-terms terms rest-terms index prior-results))
	(if (index-variable var)
	    (unbind-index-variable var)
	    (unbind-var var))
	result)))

;; ------------------------------------------------------------
;; GR-index unification functions
;; ------------------------------------------------------------

(define (unify-subgoal-with-gr-index subgoal type)
  (let (index result)
    (cond
     ((eq? type 'context)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-context-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-context-index (subgoal-pred subgoal)))))
     ((eq? type 'gr)
      (if (eq? '+ (subgoal-sign subgoal))
	  (set! index (predsym-neg-gr-index (subgoal-pred subgoal)))
	  (set! index (predsym-pos-gr-index (subgoal-pred subgoal))))))
    (if (> debug-level 5)
	(begin
	  (display "unifying subgoal ")
	  (display-subgoal subgoal)
	  (newline)
	  (if (> debug-level 6)
	      (begin
		(display "current index content of ")
		(if (eq? '+ (subgoal-sign subgoal)) (display "-") (display "+"))
		(display (predsym-symbol (subgoal-pred subgoal)))
		(display ": ") 
		(newline)
		(print-index-entries (cdr index) "(" '())))))
    (set! result (unify-terms-with-index (subgoal-terms subgoal) '() (cdr index) '() 'gr))
    (if result
	result
	(begin
	  (if (= (car index) 0)
	      '()
	      (begin
		(print "index unification error in non-empty index!")
		'()))))))

