(module link
	(include "dctp.sch")
	(export link-count eq-link-count link-inc eq-link-inc)
	(export (make-link subgoal predsubgoal)  (set-alternation-rate value) (insert-link link weight)
		(mark-links-deletable links subgoal) (print-link-statistics) (generate-links-for-subgoal subgoal)
		(restore-link-cells link-cells)  (select-link-cell index depth) (reset-link-statistics) (init-link-bias) 
		(insert-eq-links subgoal eq-lits) (insert-eq-links-for-new-equation eq-lit predsubgoal)
		(make-new-negated-eq-sg eq-lit subgoal) (insert-eq-lemma-links lemma-sg predsubgoal) 
		(delete-subgoals-links subgoals path-sg) (varlist-untouched? vars) (top-level? term sg-terms)
		(select-appropriate-link-cell) (delete-element-from-list element list pred-list) (recompute-link-lists)
		(insert-reflexivity-link subgoal) (set-eq-alternation-rate value)
		(generate-links-for-lemma lemma) (generate-links-for-context-lemma lemma))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm")	(index "index.scm") 
		(select "select.scm") (subgoal-simplifiable pruning "pruning.scm") (subgoal-in-list pruning "pruning.scm") 
		(simplify-subgoals-on-the-fly pruning "pruning.scm") (renamed-copy dctp "dctp.scm") (variant-varlist? dctp "dctp.scm")
		(subgoal-in-copy dctp "dctp.scm") (kbo> kbo "kbo.scm") (varlist-renamed? normal-form "normal-form.scm")
		(nr-eq-literals preprocess "preprocess.scm") (nr-non-eq-literals preprocess "preprocess.scm")
		(unrenamed-termlist-copy rewrite "rewrite.scm") (insert-literal parser "parser.scm")
		(positive-eq-lits path "path.scm") (positive-eq-lemmas path "path.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define total-weight 0)
(define min-weight 100000)
(define max-weight 0)
(define total-select 0)
(define min-select 100000)
(define max-select 0)
(define deleted-select 0)
(define new-links 0)
(define new-selects 0)

(define max-sg-weight 0)
(define open1 0)
(define delete1 #f)
(define open2 0)
(define delete2 #f)
(define index-end1 #f)
(define index-end2 #f)
(define immediately-solved 0)

(define unit-offset 10)
(define active-links-exist #f)
(define active-eq-links-exist #f)

(define link-count 0)
(define eq-link-count 0)
(define link-inc 0)
(define eq-link-inc 0)

(define alternate-count 0)
(define alternate-index 0)
(define alternation-rate 5)
(define eq-alternate-count 0)
(define eq-alternate-index 0)
(define eq-alternation-rate 50)

(define turn-for-optimised #f)

(define all-links '())
(define last-link '())
(define all-eq-links '())
(define last-eq-link '())
(define max-all-links 10000)
(define length-all-links 0)
(define max-vector-links 10000)
(define length-vector-links 0)
(define max-all-eq-links 10000)
(define length-all-eq-links 0)
(define max-vector-eq-links 10000)
(define length-vector-eq-links 0)
(define critical-all-links 70)

(define (set-alternation-rate value)
  (set! alternation-rate value))

(define (set-eq-alternation-rate value)
  (set! eq-alternation-rate value))

;;;*************************************************************************
;; ------------------------------------------------------------
;; The link functions
;; ------------------------------------------------------------

(define (reset-link-statistics)
  (set! total-select 0)
  (set! min-select 100000)
  (set! max-select 0)
  (set! deleted-select 0)
  (set! new-links 0)
  (set! new-selects 0)
  (set! total-weight 0)
  (set! min-weight 100000)
  (set! max-weight 0))

(define (generate-links-for-subgoal subgoal)
  (let ((linked-subgoal-lists (unify-subgoal-with-index subgoal 'path)))
    (generate-links-with-lists subgoal linked-subgoal-lists)
    (if (and (not equality) context-lemmas context-units)
	(begin
	  (set! linked-subgoal-lists (unify-subgoal-with-index subgoal 'context))
	  (generate-links-with-lists subgoal linked-subgoal-lists)))
    (set! linked-subgoal-lists (unify-subgoal-with-index subgoal 'unit))
    (generate-links-with-lists subgoal linked-subgoal-lists)))

(define (generate-links-with-lists subgoal linked-subgoal-lists)
  (if (pair? linked-subgoal-lists)
      (begin
	(generate-links-with-list subgoal (car linked-subgoal-lists))
	(generate-links-with-lists subgoal (cdr linked-subgoal-lists)))))

(define (generate-links-with-list subgoal linked-subgoals)
  (if (pair? linked-subgoals)
      (begin
	(if (or (not (context-lemma? (car linked-subgoals))) (subgoal-local (car linked-subgoals)) (subgoal-ground (car linked-subgoals)))
	    (generate-link subgoal (car linked-subgoals)))
	(generate-links-with-list subgoal (cdr linked-subgoals)))))  

(define (generate-link subgoal linked-sg)
  (if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign linked-sg)))
	   (eq? (subgoal-pred subgoal) (subgoal-pred linked-sg))
	   (not (subgoal-unlinkable linked-sg)))
      (begin
	(if (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms linked-sg))
	    (let (path-closed-a path-closed-b)
; 	      (if (and forall-unify (not (subgoal-ground (link-lita link))) (not (subgoal-ground (link-litb link))))
; 		  (copy-trailed-bindings))
	      (set! linked-terms-ground #t)
	      (set! max-sg-weight 0)
	      (set! current-depth 0)
	      (set! current-complexity 0)
	      (set! immediately-solved 0)
	      (set! index-end1 (clause-in-index? (subgoal-clause subgoal)))
	      ;;(if propagate (trail-stop))
	      (if (and (vector? index-end1) (> (vector-ref index-end1 0) 0)) ;; if the first clause would become a variant
	      ;;(if (and (not (clause-initial (subgoal-clause subgoal))) (variant-varlist? (clause-variables (subgoal-clause subgoal))))
		  (begin
		    (set! delete1 #t)
		    (set! open1 0))
		  (if (and subsumption
			   (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause subgoal)))
			   (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause subgoal))))
; 			       (and (not equality) context-units context-lemmas 
; 				    (subgoals-subsumed-by-context-index (clause-subgoals (subgoal-clause subgoal))))))
		      (begin ;; or if it were subsumed
			(set! delete1 #t)
			(set! open1 0))
		      (if (index-irregular-subgoals (clause-subgoals (subgoal-clause subgoal)))
			  (begin
			    (set! delete1 #t)
			    (set! open1 0))
			  (begin
			    (set! delete1 #f)
			    (set! open1 (get-open-subgoals (clause-subgoals (subgoal-clause subgoal)) subgoal subgoal))))))
	      (set! index-end2 (clause-in-index? (subgoal-clause linked-sg)))
	      (if (and (vector? index-end2) (> (vector-ref index-end2 0) 0)) ;; if the second clause would become a variant
	      ;;(if (and (not (clause-initial (subgoal-clause linked-sg))) (variant-varlist? (clause-variables (subgoal-clause linked-sg))))
		  (begin
		    (set! delete2 #t)
		    (set! open2 0))
		  (if (and subsumption
			   (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause linked-sg)))
			   (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause linked-sg))))
; 			       (and (not equality) context-units context-lemmas 
; 				    (subgoals-subsumed-by-context-index (clause-subgoals (subgoal-clause linked-sg))))))
		      (begin ;; or if it were subsumed
			(set! delete2 #t)
			(set! open2 0))
		      (if (index-irregular-subgoals (clause-subgoals (subgoal-clause linked-sg)))
			  (begin
			    (set! delete2 #t)
			    (set! open2 0))
			  (begin
			    (set! delete2 #f)
			    (set! open2 (get-open-subgoals (clause-subgoals (subgoal-clause linked-sg)) 
							   linked-sg subgoal))))))
	      (if (not (or (and delete1 delete2)
			   (and couples (non-unit-clause (subgoal-clause subgoal)) 
				(non-unit-clause (subgoal-clause linked-sg))
				(or (not (vector? index-end1))
				    (= (vector-ref index-end1 0) 0)
				    (not (eq? (car (vector-ref index-end1 1)) (subgoal-clause subgoal))))
				(or (not (vector? index-end2))
				    (= (vector-ref index-end2 0) 0)
				    (not (eq? (car (vector-ref index-end2 1)) (subgoal-clause linked-sg))))
				(or delete1 delete2))))
		  (let ((link (make-link subgoal linked-sg)))
		    (set-subgoal-links subgoal (cons link (subgoal-links subgoal)))
		    (set-subgoal-links linked-sg (cons link (subgoal-links linked-sg)))
		    (set-link-weight link (compute-link-weight subgoal linked-sg open1 open2 max-sg-weight))
		    (set-link-depth link current-complexity)
		    (if (< current-complexity min-relevant-depth)
			(set! min-relevant-depth current-complexity))
		    (if (> debug-level 4)
			(begin
			  (display "inserting link between depths ")
			  (display (subgoal-depth (link-lita link)))
			  (display " and ")
			  (display (subgoal-depth (link-litb link)))
			  (newline)))
		    (insert-link link (link-weight link))
		    (if (or (context-lemma? subgoal) (context-lemma? linked-sg)) (set! context-links (+ context-links 1)))
		    (if (and link-info (= 0 (remainder total-links link-info)))
			(begin
			  (display "link statistics after ")
			  (display total-links)
			  (display " links (average min max):")
			  (newline)
			  (print-link-statistics)))))
	      ;;(if propagate (untrail-subgoals))
 	      (unbind-vars))))))

(define (print-link-statistics)
  (display "weight: ")
  (display (quotient total-weight new-links))
  (display " ")
  (display min-weight)
  (display " ")
  (display max-weight)
  (display "    select: ")
  (if (> new-selects 0)
      (display (quotient total-select new-selects))
      (display "0"))
  (display " ")
  (display min-select)
  (display " ")
  (display max-select)
  (display " deleted ")
  (display deleted-select)
  (newline)
  (set! total-weight 0)
  (set! min-weight 1000000)
  (set! max-weight 0)
  (set! total-select 0)
  (set! min-select 1000000)
  (set! max-select 0)
  (set! deleted-select 0)
  (set! new-selects 0)
  (set! new-links 0))

(define (insert-link link weight)
  (let (index appropriate-link-vector appropriate-link-vector-size)
    (if (eq-link? link)
	(begin
	  (if alternate
	      (if (pair? all-eq-links)
		  (begin
		    (set-cdr! last-eq-link (list link))
		    (set! last-eq-link (cdr last-eq-link))
		    (set! length-all-eq-links (+ length-all-eq-links 1))
		    (if (and sweep-links (> length-all-eq-links max-all-eq-links))
			(let ((old-length-all-eq-links length-all-eq-links))
			  (set! nr-sweeps (+ nr-sweeps 1))
			  (if dewey-numbers
			      (begin
				(display "starting sweep of eq-link list for deletable links. length of eq-link list before sweep: ")
				(display length-all-eq-links)
				(display " ( ")
				(display (length all-eq-links))
				(display " )")
				(newline)))
			  (sweep-all-eq-links)
			  (if dewey-numbers
			      (begin
				(display "length of eq-link list after sweep: ")
				(display length-all-eq-links)
				(display " ( ")
				(display (length all-eq-links))
				(display " )")
				(newline)))
			  (if (> length-all-eq-links (quotient (* old-length-all-eq-links critical-all-links) 100))
			      (begin
				(set! max-all-eq-links (quotient (* max-all-eq-links 14) 10))
				(if dewey-numbers
				    (begin
				      (display "maximum length of eq-link list set to ")
				      (display max-all-eq-links)
				      (newline))))))))
		  (begin
		    (set! length-all-eq-links 1)
		    (set! all-eq-links (list link))
		    (set! last-eq-link all-eq-links))))
	  (set! appropriate-link-vector eq-link-vector)
	  (set! appropriate-link-vector-size eq-link-vector-size)
	  (set! length-vector-eq-links (+ length-vector-eq-links 1))
	  (if (and sweep-links (> length-vector-eq-links max-vector-eq-links))
	      (let ((old-length-vector-eq-links length-vector-eq-links))
		(set! nr-vector-sweeps (+ nr-vector-sweeps 1))
		(if dewey-numbers
		    (begin
		      (display "starting sweep of eq-link vector for deletable links. length of eq-link vector  before sweep: ")
		      (display length-vector-eq-links)
		      (newline)))
		(sweep-link-vector appropriate-link-vector appropriate-link-vector-size 'non-eq)
		(if dewey-numbers
		    (begin
		      (display "entries in eq-link vector after sweep: ")
		      (display length-vector-eq-links)
		      (newline)))
		(if (> length-vector-eq-links (quotient (* old-length-vector-eq-links critical-all-links) 100))
		    (begin
		      (set! max-vector-eq-links (quotient (* max-vector-eq-links 14) 10))
		      (if dewey-numbers
			  (begin
			    (display "maximum entries in eq-link vector set to ")
			    (display max-vector-eq-links)
			    (newline))))))))
	(begin
	  (if alternate
	      (if (pair? all-links)
		  (begin
		    (set-cdr! last-link (list link))
		    (set! last-link (cdr last-link))
		    (set! length-all-links (+ length-all-links 1))
		    (if (and sweep-links (> length-all-links max-all-links))
			(let ((old-length-all-links length-all-links))
			  (set! nr-sweeps (+ nr-sweeps 1))
			  (if dewey-numbers
			      (begin
				(display "starting sweep of link list for deletable links. length of link list before sweep: ")
				(display length-all-links)
				(display " ( ")
				(display (length all-links))
				(display " )")
				(newline)))
			  (sweep-all-links)
			  (if dewey-numbers
			      (begin
				(display "length of link list after sweep: ")
				(display length-all-links)
				(display " ( ")
				(display (length all-links))
				(display " )")
				(newline)))
			  (if (> length-all-links (quotient (* old-length-all-links critical-all-links) 100))
			      (begin
				(set! max-all-links (quotient (* max-all-links 14) 10))
				(if dewey-numbers
				    (begin
				      (display "maximum length of link list set to ")
				      (display max-all-links)
				      (newline))))))))
		  (begin
		    (set! length-all-links 1)
		    (set! all-links (list link))
		    (set! last-link all-links))))
	  (set! appropriate-link-vector link-vector)
	  (set! appropriate-link-vector-size link-vector-size)
	  (set! length-vector-links (+ length-vector-links 1))
	  (if (and sweep-links (> length-vector-links max-vector-links))
	      (let ((old-length-vector-links length-vector-links))
		(set! nr-vector-sweeps (+ nr-vector-sweeps 1))
		(if dewey-numbers
		    (begin
		      (display "starting sweep of link vector for deletable links. length of link vector  before sweep: ")
		      (display length-vector-links)
		      (newline)))
		(sweep-link-vector appropriate-link-vector appropriate-link-vector-size 'non-eq)
		(if dewey-numbers
		    (begin
		      (display "entries in link vector after sweep: ")
		      (display length-vector-links)
		      (newline)))
		(if (> length-vector-links (quotient (* old-length-vector-links critical-all-links) 100))
		    (begin
		      (set! max-vector-links (quotient (* max-vector-links 14) 10))
		      (if dewey-numbers
			  (begin
			    (display "maximum entries in link vector set to ")
			    (display max-vector-links)
			    (newline)))))))))
    ;;(if propagate (insert-link-in-propagation-index link))
    (if (or (>= weight appropriate-link-vector-size) (< weight 0))
	(set! index (- appropriate-link-vector-size 1))
	(set! index weight))
    (if (eq-link? link)
	(if (and (< index eq-search-start-index) (<= (link-depth link) allowed-eq-term-depth))
	    (set! eq-search-start-index index)
	    (set! active-eq-links-exist #t))
	(if (and (< index search-start-index) (<= (link-depth link) allowed-term-depth))
	    (set! search-start-index index)
	    (set! active-links-exist #t)))
    (if (= index (- appropriate-link-vector-size 1))
	(if (null? (vector-ref appropriate-link-vector index))
	    (vector-set! appropriate-link-vector index (list link))
	    (linear-insert-link link weight (vector-ref appropriate-link-vector index)))
	(vector-set! appropriate-link-vector index (cons link (vector-ref appropriate-link-vector index))))))

(define (linear-insert-link link weight link-list)
  (if (or (<= weight (link-weight (car link-list)))
	  (and (= weight (link-weight (car link-list)))
	       (> (lower-depth link) (lower-depth (car link-list)))))
      (begin
	(set-cdr! link-list (cons (car link-list) (cdr link-list)))
	(set-car! link-list link))
      (if (pair? (cdr link-list))
	  (linear-insert-link link weight (cdr link-list))
	  (set-cdr! link-list (list link)))))

(define (lower-depth link)
  (if (not (link-litb link))
      (subgoal-depth (link-lita link))
      (if (> (subgoal-depth (link-lita link)) (subgoal-depth (link-litb link)))
	  (subgoal-depth (link-lita link))
	  (subgoal-depth (link-litb link)))))

(define (restore-link-cells link-cells)
  (if (pair? link-cells)
      (let ((rest-cells (cdr link-cells)))
	(if (not (link-active (car link-cells))) ;; do not restore deletable links
	    (begin
	      (restore-link-cell link-cells)))
	(restore-link-cells rest-cells))))

(define (restore-link-cell link-cell)
  (let (index appropriate-link-vector appropriate-link-vector-size)
    (if (eq-link? (car link-cell))
	(begin
	  (set! appropriate-link-vector eq-link-vector)
	  (set! appropriate-link-vector-size eq-link-vector-size)
	  (set! length-vector-eq-links (+ length-vector-eq-links 1)))
	(begin
	  (set! appropriate-link-vector link-vector)
	  (set! appropriate-link-vector-size link-vector-size)
	  (set! length-vector-links (+ length-vector-links 1))))
    (if (or (>= (link-weight (car link-cell)) appropriate-link-vector-size) (< (link-weight (car link-cell)) 0))
	(set! index (- appropriate-link-vector-size 1))
	(set! index (link-weight (car link-cell))))
    (if (eq-link? (car link-cell))
	(if (and (< index eq-search-start-index) (<= (link-depth (car link-cell)) allowed-eq-term-depth))
	    (set! eq-search-start-index index)
	    (set! active-eq-links-exist #t))
	(if (and (< index search-start-index) (<= (link-depth (car link-cell)) allowed-term-depth))
	    (set! search-start-index index)
	    (set! active-links-exist #t)))
    (set-link-active (car link-cell) #t)
    (if (= index (- appropriate-link-vector-size 1))
	(if (null? (vector-ref appropriate-link-vector index))
	    (begin
	      (vector-set! appropriate-link-vector index link-cell)
	      (set-cdr! link-cell '()))
	    (linear-insert-link (car link-cell) (link-weight (car link-cell)) (vector-ref appropriate-link-vector index)))
	(begin
	  (set-cdr! link-cell (vector-ref appropriate-link-vector index))
	  (vector-set! appropriate-link-vector index link-cell)))))

(define (make-link subgoal predsubgoal)
  (let ((link (new-link)))
    (set! total-links (+ total-links 1))
    (set! new-links (+ new-links 1))
    (if unit-on-top
	(if (and (unit-subgoal predsubgoal) (non-unit-subgoal subgoal))
	    (begin
	      (set-link-lita link predsubgoal)
	      (set-link-litb link subgoal))
	    (begin
	      (set-link-lita link subgoal)
	      (set-link-litb link predsubgoal)))
	(if (and (unit-subgoal subgoal) (non-unit-subgoal predsubgoal))
	    (begin
	      (set-link-lita link predsubgoal)
	      (set-link-litb link subgoal))
	    (begin
	      (set-link-lita link subgoal)
	      (set-link-litb link predsubgoal))))
    (set-link-active link #t)
    link))
  
(define (make-eq-link eq-lit subgoal terms eq-term)
  (let ((link (new-eq-link)))
    (set! total-eq-links (+ total-eq-links 1))
    (set! new-links (+ new-links 1))
    (set-link-lita link eq-lit)
    (set-link-litb link subgoal)
    (set-link-active link #t)
    (set-link-position link terms)
    (set-link-left-to-right link (if (eq? eq-term (car (subgoal-terms eq-lit))) #t #f))
    link))

(define (link-relevant link)
  (or (and (subgoal? (link-lita link)) (clause-relevant (subgoal-clause (link-lita link))))
      (and (subgoal? (link-litb link)) (clause-relevant (subgoal-clause (link-litb link))))))

(define (compute-link-weight subgoal predsubgoal open1 open2 complexity)
  (let (weight)
    (set! weight (- (+ (* sg-bias (+ open1 open2)) (* (+ open1 open2) term-bias complexity)) (* gr-closed-weight immediately-solved)))
    ;;(set! weight (+ (* sg-bias (+ open1 open2)) (* (+ open1 open2) term-bias complexity)))
    (if strong-relevance
	(begin
	  (if (and predsubgoal (not (or norel (clause-relevant (subgoal-clause predsubgoal)))))
	      (set! weight (bit-lsh weight 1)))
	  (if (not (or norel (clause-relevant (subgoal-clause subgoal))))
	      (set! weight (bit-lsh weight 1)))))
    (if (not limit-term-depth)
	(set! weight (+ weight current-depth)))
    ;; Nota bene: predsubgoal may be empty for decomposition or reflexivity links
    (if (and (non-unit-subgoal subgoal) predsubgoal (non-unit-subgoal predsubgoal) (> sg-bias 0))
	(set! weight (+ unit-offset weight)))
    (if (not (and (or (not predsubgoal) (clause-initial (subgoal-clause predsubgoal))) (clause-initial (subgoal-clause subgoal))))
	(set! weight (* weight init-bias)))
;     (if (and (> open1 0) (> open2 0))
;  	(set! weight (bit-lsh weight 1)))
    (if (< weight 2)
	(set! weight 2))
    (if (= (+ open1 open2) immediately-solved)
	(set! weight 1))
    (if (and (= open1 0) (= open2 0))
	(set! weight 0))
    (if (< weight 0) (set! weight maxint))
    (if link-info
	(begin
	  (if (< weight min-weight) (set! min-weight weight))
	  (if (> weight max-weight) (set! max-weight weight))
	  (set! total-weight (+ total-weight weight))))
    weight))

; (define (select-appropriate-link-cell)
;   (let (cell)
;     (if (= link-count eq-link-count)
; 	(begin
; 	  (set! link-count 0)
; 	  (set! eq-link-count 0)))
;     (if (or (and (> nr-non-eq-literals nr-eq-literals) (>= eq-link-count link-count))
; 	    (> eq-link-count link-count))
; 	(begin
; 	  (set! cell (select-link-cell search-start-index 1))
; 	  (if cell
; 	      (set! link-count (+ link-count link-inc))
; 	      (set! cell (select-eq-link-cell eq-search-start-index 1))))
; 	(begin
; 	  (set! cell (select-eq-link-cell eq-search-start-index 1))
; 	  (if cell
; 	      (set! eq-link-count (+ eq-link-count eq-link-inc))
; 	      (set! cell (select-link-cell search-start-index 1)))))
;     cell))

(define (select-appropriate-link-cell)
  (let ((link-cell #f))
    (if (= link-count eq-link-count)
	(begin
	  (set! link-count 0)
	  (set! eq-link-count 0)))
    (if (or (and (> nr-non-eq-literals nr-eq-literals) (>= eq-link-count link-count))
	    (> eq-link-count link-count))
	(begin
	  (if (and local-optimise turn-for-optimised)
	      (begin
		(set! link-cell (local-link-check))
		(set! turn-for-optimised #f)))
	  (if (not link-cell)
	      (if (>= alternate-count alternation-rate)
		  (begin
		    (set! alternate-index (+ alternate-index 1))
		    (set! alternate-count 0)
		    (set! turn-for-optimised #t)
		    (set! link-cell (get-oldest-link-cell)))
		  (begin
		    (if alternate (set! alternate-count (+ alternate-count 1)))
		    (set! link-cell (select-link-cell search-start-index 1)))))
	  (if link-cell
	      (set! link-count (+ link-count link-inc))
	      (set! link-cell (select-eq-link-cell eq-search-start-index 1))))
	(begin
	  (if (>= eq-alternate-count eq-alternation-rate)
	      (begin
		(set! eq-alternate-index (+ eq-alternate-index 1))
		(set! eq-alternate-count 0)
		(set! link-cell (get-oldest-eq-link-cell)))
	      (begin
		(if alternate (set! eq-alternate-count (+ eq-alternate-count 1)))
		(set! link-cell (select-eq-link-cell eq-search-start-index 1))))
	  (if link-cell
	      (set! eq-link-count (+ eq-link-count eq-link-inc))
	      (set! link-cell (select-link-cell search-start-index 1)))))
    link-cell))

(define (select-link-cell index depth)
  ;;(print "searching for link at index " index)
  (if (>= index link-vector-size)
      (if (and limit-term-depth active-links-exist)
	  (begin
	    (set! allowed-term-depth (+ allowed-term-depth 1))
	    (set! active-links-exist #f)
	    (set! search-start-index 0)         ;; next searches must start from scratch
	    (if dewey-numbers
		(begin
		  (display "allowed link term complexity increased to ")
		  (display allowed-term-depth)
		  (newline)))
	    (select-link-cell 0 (+ depth 1)))
	  #f)      ;; once we get past the last link-vector entry, there are no more links: affirmation state
      (let ((link-cell (extract-active-link-from-list (vector-ref link-vector index) (vector-ref link-vector index) index)))
	(if (not link-cell)
	    (begin       ;; if the current link-vector entry contains no active link-cells
	      (set! depth (+ depth 1))
	      (set! index (+ index 1))
	      (do ()     ;; look for the next non-empty link-vector entry
		  ((or (>= index link-vector-size) (pair? (vector-ref link-vector index))) #t)
		(set! depth (+ depth 1))
		(set! index (+ index 1)))
	      ;;(print "link search unsuccessful, continuing in vector " index " with depth " depth)
	      (select-link-cell index (+ depth 1))) ;; then search through that entry
	    (begin
	      (set! total-select (+ total-select depth))
	      (set! new-selects (+ new-selects 1))
	      (if (< depth min-select) (set! min-select depth))
	      (if (> depth max-select) (set! max-select depth))
	      ;; if we selected the only link-cell in the entry
	      (if (null? (vector-ref link-vector index))
		  (set! search-start-index (+ index 1))    ;; next search begins one entry down the vector
		  (set! search-start-index index))         ;; next search might be successful in this entry
	      (set-cdr! link-cell '())
	      ;;(display "found link ")
	      ;;(display-link (car link-cell))
	      ;;(display " in vector ")
	      ;;(display index)
	      ;;(newline)
	      link-cell)))))

(define (extract-active-link-from-list list pred-list index)
  (if (pair? list)
      (if (eq? (link-active (car list)) #t)
	  (if (<= (link-depth (car list)) allowed-term-depth)
	      (begin
		(set! length-vector-links (- length-vector-links 1))
		(if (eq? list (vector-ref link-vector index))
		    (begin
		      (vector-set! link-vector index (cdr list))
		      list)
		    (begin
		      (set-cdr! pred-list (cdr list))
		      list)))
	      (begin
		(set! active-links-exist #t)
		(extract-active-link-from-list (cdr list) list index)))
	  (begin
	    (set! length-vector-links (- length-vector-links 1))
	    (if (eq? list (vector-ref link-vector index)) ;; on the fly deletion of deletable links. as non-active links are extracted,
		(begin                                    ;; anything not active #t still here must be deletable
		  (vector-set! link-vector index (cdr list))
		  (extract-active-link-from-list (vector-ref link-vector index) (vector-ref link-vector index) index))
		(begin
		  (set-cdr! pred-list (cdr list))
		  (extract-active-link-from-list (cdr pred-list) pred-list index)))))
      #f))

(define (select-eq-link-cell index depth)
  (if (>= index eq-link-vector-size)
      (if (and limit-term-depth active-eq-links-exist)
	  (begin
	    (set! allowed-eq-term-depth (+ allowed-eq-term-depth 1))
	    (set! active-eq-links-exist #f)
	    (set! eq-search-start-index 0)         ;; next searches must start from scratch
	    (if dewey-numbers
		(begin
		  (display "allowed eq-link term complexity increased to ")
		  (display allowed-eq-term-depth)
		  (newline)))
	    (select-eq-link-cell 0 (+ depth 1)))
	  #f)      ;; once we get past the last eq-link-vector entry, there are no more links: affirmation state
      (let ((link-cell (extract-active-eq-link-from-list (vector-ref eq-link-vector index) (vector-ref eq-link-vector index) index)))
	(if (not link-cell)
	    (begin       ;; if the current eq-link-vector entry contains no active link-cells
	      (set! depth (+ depth 1))
	      (set! index (+ index 1))
	      (do ()     ;; look for the next non-empty eq-link-vector entry
		  ((or (>= index eq-link-vector-size) (pair? (vector-ref eq-link-vector index))) #t)
		(set! depth (+ depth 1))
		(set! index (+ index 1)))
	      (select-eq-link-cell index (+ depth 1))) ;; then search through that entry
	    (begin
	      (set! total-select (+ total-select depth))
	      (set! new-selects (+ new-selects 1))
	      (if (< depth min-select) (set! min-select depth))
	      (if (> depth max-select) (set! max-select depth))
	      ;; if we selected the only link-cell in the entry
	      (if (null? (vector-ref eq-link-vector index))
		  (set! eq-search-start-index (+ index 1))    ;; next search begins one entry down the vector
		  (set! eq-search-start-index index))         ;; next search might be successful in this entry
	      (set-cdr! link-cell '())
	      link-cell)))))

(define (extract-active-eq-link-from-list list pred-list index)
  (if (pair? list)
      (if (eq? (link-active (car list)) #t)
	  (if (<= (link-depth (car list)) allowed-eq-term-depth)
	      (begin
		(set! length-vector-eq-links (- length-vector-eq-links 1))
		(if (eq? list (vector-ref eq-link-vector index))
		    (begin
		      (vector-set! eq-link-vector index (cdr list))
		      list)
		    (begin
		      (set-cdr! pred-list (cdr list))
		      list)))
	      (begin
		(set! active-eq-links-exist #t)
		(extract-active-eq-link-from-list (cdr list) list index)))
	  (begin
	    (set! length-vector-links (- length-vector-links 1))
	    (if (eq? list (vector-ref eq-link-vector index)) ;; on the fly deletion of deletable links
		(begin
		  (vector-set! eq-link-vector index (cdr list))
		  (extract-active-eq-link-from-list (vector-ref eq-link-vector index) (vector-ref eq-link-vector index) index))
		(begin
		  (set-cdr! pred-list (cdr list))
		  (extract-active-eq-link-from-list (cdr pred-list) pred-list index)))))
      #f))

(define (mark-links-deletable links subgoal)
  (if (pair? links)
      (begin 
	(if (not (eq? (link-active (car links)) 'delete))
	    (begin
	      (set-link-active (car links) 'delete)
	      (set! links-deleted (+ links-deleted 1))))
	(if (and (link-lita (car links))
		 (or (not (eq? (link-lita (car links)) subgoal))
		     (not (link-litb (car links)))
		     (pair? (link-litb (car links))))) ;; what does this mean? disagreement linking?
	    (delete-link-from-list (car links) (link-lita (car links)))
; 	    (if (and (link-litb (car links)) (not (eq? (link-litb (car links)) subgoal)))
; 		(delete-link-from-list (car links) (link-litb (car links)))))
	    (if (and (link-litb (car links)) (not (pair? (link-litb (car links))))
		     (not (eq? (link-litb (car links)) subgoal)))
		(delete-link-from-list (car links) (link-litb (car links)))))
	(set-link-lita (car links) #f)
	(set-link-litb (car links) #f)
	(set-link-depth (car links) #f)
	(mark-links-deletable (cdr links) subgoal)
	(set-cdr! links '()))))

(define (delete-link-from-list link subgoal)
  (set-subgoal-links subgoal (delete-element-from-list link (subgoal-links subgoal) (subgoal-links subgoal))))

(define (delete-element-from-list element list pred-list)
  (if (pair? list)
      (if (eq? (car list) element)
          (if (eq? list pred-list)
              (cdr list)
              (begin
                (set-cdr! pred-list (cdr list))
                (cdr list)))
          (if (delete-element-from-list element (cdr list) list)
              list
              #f))  
      #f))

;; FIXME: Suboptimal! GR-closed literals do not matter in effect. How about a new (weighted) scheme, where really
;; open subgoals value 3, GR-closed subgoals rate 1 and simplified subgoals rate 0? Subgoal and term bias will
;; have to be modified accordingly. Also, a special treatment for the linked subgoals is required.
(define (get-open-subgoals subgoals link-sg bottom-sg)
  (if (pair? subgoals)
      (if (or (subgoal-simplified (car subgoals))
	      (match-subgoal-with-index (car subgoals) 'simplify))
	  (get-open-subgoals (cdr subgoals) link-sg bottom-sg)
	  (let (sg-weight (gr-closed (ground-closed-with-index (car subgoals)))
			  (duplicate (subgoal-in-list (car subgoals) (cdr subgoals))))
	    (if duplicate
		(begin
		  (if (eq? (car subgoals) link-sg)
		      (get-open-subgoals (cdr subgoals) duplicate bottom-sg)
		      (get-open-subgoals (cdr subgoals) link-sg bottom-sg)))
		(if (not (and gr-closed (closing-literal-above bottom-sg (list gr-closed))))
		    (begin
		      (set! complexity 0)
		      (set! sg-weight (bit-lsh (termlist-symbol-count (subgoal-terms (car subgoals)) 1) 
					       (predsym-weight (subgoal-pred (car subgoals)))))
		      (if (> complexity current-complexity)
			  (set! current-complexity complexity))
		      (if (> complexity max-complexity)
			  (set! max-complexity complexity))
		      (if (> sg-weight max-sg-weight)
			  (set! max-sg-weight sg-weight))
		      (if (eq? (car subgoals) link-sg)
			  (begin
			    (set! immediately-solved (+ immediately-solved 1))
			    (+ gr-closed-weight (get-open-subgoals (cdr subgoals) link-sg bottom-sg)))
			  (begin
			    ;; push open subgoal of link onto trail for later insertion in propagation index
			    ;;(trail-subgoal (car subgoals))
			    (+ open-weight (get-open-subgoals (cdr subgoals) link-sg bottom-sg)))))
		    (if (and (not (= ground-weight gr-closed-weight)) 
			     (or (and (not equality) (subgoal-ground (car subgoals))) (ground-termlist (subgoal-terms (car subgoals)))))
			     ;;(ground-termlist (subgoal-terms (car subgoals))))
			(begin
			  (set! immediately-solved (+ immediately-solved 1))
			  (+ ground-weight (get-open-subgoals (cdr subgoals) link-sg bottom-sg)))
			(begin
			  (set! immediately-solved (+ immediately-solved 1))
			  (+ gr-closed-weight (get-open-subgoals (cdr subgoals) link-sg bottom-sg))))))))
      0))


(define (get-open-subgoals-eq subgoals eq-lit)
  ;; special version of the above function for obtaining the number of new subgoals for eq-links
  ;; only checks whether a subgoal is the equation part of the paramodulation step, as in a paramodulation step
  ;; all subgoals remain open except the instantiated equation
  (if (pair? subgoals)
      (if (or (subgoal-simplified (car subgoals))
	      (match-subgoal-with-index (car subgoals) 'simplify))
	  (get-open-subgoals-eq (cdr subgoals) eq-lit)
	  (let (sg-weight (gr-closed (ground-closed-with-index (car subgoals))))
	    (if (not (and gr-closed (closing-literal-above (car subgoals) (list gr-closed))))
		(begin
		  (set! complexity 0)
		  (set! sg-weight (bit-lsh (termlist-symbol-count (subgoal-terms (car subgoals)) 1) 
					   (predsym-weight (subgoal-pred (car subgoals)))))
		  (if (> complexity current-complexity)
		      (set! current-complexity complexity))
		  (if (> complexity max-complexity)
		      (set! max-complexity complexity))
		  (if (> sg-weight max-sg-weight)
		      (set! max-sg-weight sg-weight))
		  (if (eq? (car subgoals) eq-lit)
		      (begin
			(set! immediately-solved (+ immediately-solved 1))
			(+ gr-closed-weight (get-open-subgoals-eq (cdr subgoals) eq-lit)))
		      (+ open-weight (get-open-subgoals-eq (cdr subgoals) eq-lit))))
		(if (and (not (= ground-weight gr-closed-weight)) (ground-termlist (subgoal-terms (car subgoals))))
			(begin 
			  (set! immediately-solved (+ immediately-solved 1))
			  (+ ground-weight (get-open-subgoals-eq (cdr subgoals) eq-lit)))
			(begin
			  (set! immediately-solved (+ immediately-solved 1))
			  (+ gr-closed-weight (get-open-subgoals-eq (cdr subgoals) eq-lit)))))))
      0))

(define (generate-links-for-lemma lemma)
;   (newline)
;   (display "generating links for new lemma ")
;   (display-subgoal lemma)
;   (newline)
;   (newline)
  (let ((linked-unit-lists (unify-subgoal-with-index lemma 'unit)))
    (if (and (pair? linked-unit-lists) (pair? (car linked-unit-lists)))
	(begin
	  (if dewey-numbers
	      (begin
		(display "Found complementary unit for new lemma: ")
		(display-clause (subgoal-clause (caar linked-unit-lists)))
		(newline)))
	  (subgoal-depth (caar linked-unit-lists)))
	(let ((pruning-depth #f) subsumed-unit-lists simplified-path-lists top-simplified backsubsumed-lists linked-subgoal-lists)
	  (set! simplified-path-lists (match-index-with-subgoal lemma 'path))
	  (if (pair? simplified-path-lists)
	      (begin
		;; FIXME: what about simplified context lemmas?
		(set! top-simplified (topmost-path-literal-in-lists simplified-path-lists #f))
		(if top-simplified
		    (set! pruning-depth (subgoal-depth top-simplified)))))
	  (if (and back-subsume (not (context-lemma? lemma)))
	      (begin
		(set! subsumed-unit-lists (subsume-index-with-subgoal lemma 'unit))
		(backsubsume-lists subsumed-unit-lists lemma)
		(set! backsubsumed-lists (subsume-index-with-subgoal lemma 'path))
		(backsubsume-lists backsubsumed-lists lemma)))
	  (if equality
	      (begin
		(generate-eq-links-for-lemma lemma (positive-eq-lits) pruning-depth)
		(generate-eq-links-for-lemma lemma (positive-eq-lemmas) pruning-depth)))
	  (set! linked-subgoal-lists (unify-subgoal-with-index lemma 'path))
	  (generate-lemma-links-with-lists lemma linked-subgoal-lists pruning-depth)
	  pruning-depth))))

(define (generate-links-for-context-lemma lemma)
  ;; NOTA BENE: as the entire technique of subsuming with context lemmas is as of now only done
  ;; in the non-equality case, non eq-links need to be generated here
  (let ((top-closed #f) (pruning-depth #f) linked-subgoal-lists p-subgoal i-subgoal)
    (if (not global-closure)
	(begin
	  ;; find potential closing subgoal for lemma above c-point
	  ;; this can happen when global closure is not used
	  (set! p-subgoal (unify-subgoal-with-gr-index lemma 'gr))
	  (set! p-subgoal (topmost-path-literal-in-lists p-subgoal #f))
	  (set! i-subgoal (unify-subgoal-with-gr-index lemma 'context))
	  (set! i-subgoal (topmost-path-literal-in-lists i-subgoal #f))
	  (if p-subgoal
	      (if i-subgoal
		  (if (< (subgoal-depth p-subgoal) (subgoal-depth i-subgoal))
		      (set! top-closed p-subgoal)
		      (set! top-closed i-subgoal))
		  (set! top-closed p-subgoal))
	      (set! top-closed i-subgoal))
	  (if (and top-closed (< (subgoal-depth top-closed) (subgoal-depth lemma)))
	      (set! pruning-depth (subgoal-depth lemma)))))
    (set! linked-subgoal-lists (unify-subgoal-with-index lemma 'path))
    (generate-links-with-lists lemma linked-subgoal-lists)
    (set! linked-subgoal-lists (unify-subgoal-with-index lemma 'context))
    (generate-links-with-lists lemma linked-subgoal-lists)
    (set! linked-subgoal-lists (unify-subgoal-with-index lemma 'unit))
    (generate-links-with-lists lemma linked-subgoal-lists)
    (if pruning-depth
	(lemma-rel-vector lemma)
	#f)))

(define (generate-eq-links-for-lemma lemma eq-lits pruning-depth)
  (if (pair? eq-lits)
      (begin
	(if (and (or (not pruning-depth) (<= (subgoal-depth (car eq-lits)) pruning-depth)) 
		 (not (subgoal-unlinkable (car eq-lits))))
	    (insert-eq-links lemma (list (car eq-lits))))
	(generate-eq-links-for-lemma lemma (cdr eq-lits) pruning-depth))))

(define (generate-lemma-links-with-lists lemma linked-subgoal-lists pruning-depth)
  (if (pair? linked-subgoal-lists)
      (begin
	(generate-lemma-links-with-list lemma (car linked-subgoal-lists) pruning-depth)
	(generate-lemma-links-with-lists lemma (cdr linked-subgoal-lists) pruning-depth))))

(define (generate-lemma-links-with-list lemma linked-subgoals pruning-depth)
  (if (pair? linked-subgoals)
      (begin
	(if (and (not (subgoal-unlinkable (car linked-subgoals)))
		 (not (and pruning-depth (> (subgoal-depth (car linked-subgoals)) pruning-depth))))
	    (generate-lemma-link lemma (car linked-subgoals)))
	(generate-lemma-links-with-list lemma (cdr linked-subgoals) pruning-depth))))  

;; FIXME: Might it not be a lot faster if the links are computed top to bottom instead of the other way round?
;; We have to compute a lot of links that might become redundant just a bit later on when a complementary path literal
;; is found further up. Check with the integrated refinements such as back-subsumption, but there should be no problem.
(define (generate-lemma-link lemma-sg predsubgoal)
  (simplify-on-the-fly predsubgoal)
  (if (not (unify-lists-stop (subgoal-terms lemma-sg) (subgoal-terms predsubgoal)))
      (begin
	(display "Warning! Subgoals do not unify!")
	(newline)
	(display-subgoal lemma-sg)
	(display " of depth ")
	(display (subgoal-depth lemma-sg))
	(newline)
	(display-subgoal predsubgoal)
	(display " of depth ")
	(display (subgoal-depth predsubgoal))
	(newline)))
  (set! linked-terms-ground #t)
  (set! max-sg-weight 0)
  (set! current-depth 0)
  (set! current-complexity 0)
  (set! immediately-solved 0)
  ;; the instantiated lemma is of course subsumed by its own uninstanitated version
  (set! delete1 #t)
  (set! open1 0)
  (if (and subsumption
	   (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause predsubgoal)))
	   (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause predsubgoal))))
; 	       (and (not equality) context-units context-lemmas 
; 		    (subgoals-subsumed-by-context-index (clause-subgoals (subgoal-clause predsubgoal))))))
      (begin ;; or if it were subsumed
	(set! delete2 #t)
	(set! open2 0))
      (begin
	(set! delete2 #f)
	(set! open2 (get-open-subgoals (clause-subgoals (subgoal-clause predsubgoal)) predsubgoal predsubgoal))))
  (unbind-vars)
  (if (> debug-level 8)
      (begin
	(display "Found unifying subgoal ")
	(display-subgoal predsubgoal)
	(display " of depth ")
	(display (subgoal-depth predsubgoal))
	(newline)))
  (if (not (and delete1 delete2))
      (let ((link (make-link lemma-sg predsubgoal)))
	(set-subgoal-links predsubgoal (cons link (subgoal-links predsubgoal)))
	(set-subgoal-links lemma-sg (cons link (subgoal-links lemma-sg)))
	(set-link-weight link (compute-link-weight lemma-sg predsubgoal 0 open2 max-sg-weight))
	(set-link-depth link current-complexity)
	(if (> debug-level 4)
	    (begin
	      (display "inserting link ")
	      (newline)))
	(insert-link link (link-weight link))
	(if (or (context-lemma? lemma-sg) (context-lemma? predsubgoal)) (set! context-links (+ context-links 1)))
	(if (and link-info (= 0 (remainder total-links link-info)))
	    (begin
	      (print "link statistics after " total-links " links ")
	      (print-link-statistics))))))

(define (topmost-path-literal-in-lists subgoal-lists top-sg)
  (if (pair? subgoal-lists)
      (let ((current-top (topmost-path-literal-in-list (car subgoal-lists) top-sg)))
	(if (and current-top (>= (subgoal-depth current-top) 0)
		 (or (not top-sg) (< (subgoal-depth current-top) (subgoal-depth top-sg))))
	    (topmost-path-literal-in-lists (cdr subgoal-lists) current-top)
	    (topmost-path-literal-in-lists (cdr subgoal-lists) top-sg)))
      top-sg))

(define (topmost-path-literal-in-list subgoals top-sg)
  (if (pair? subgoals)
      (if (and (>= (subgoal-depth (car subgoals)) 0)
	       (or (not top-sg) (< (subgoal-depth (car subgoals)) (subgoal-depth top-sg))))
	  (topmost-path-literal-in-list (cdr subgoals) (car subgoals))
	  (topmost-path-literal-in-list (cdr subgoals) top-sg))
      top-sg))

(define (backsubsume-lists subgoal-lists lemma)
  (if (pair? subgoal-lists)
      (begin
	(backsubsume-list (car subgoal-lists) lemma)
	(backsubsume-lists (cdr subgoal-lists) lemma))))

(define (backsubsume-list subgoals lemma)
  (if (pair? subgoals)
      (begin
	(if (not (eq? (car subgoals) lemma))
	    (backsubsume-subgoal (car subgoals)))
	(backsubsume-list (cdr subgoals) lemma))))

(define (backsubsume-subgoal subgoal)
  (if (and (not (subgoal-unlinkable subgoal))
	   (or unit-subsume (non-unit-clause (subgoal-clause subgoal))))
      (begin
	(if dewey-numbers
	    (begin
	      (display "back subsuming with new lemma: ")
	      (display-clause (subgoal-clause subgoal))
	      (newline)))
	(set! backward-subsumptions (+ 1 backward-subsumptions))
	(delete-subgoals-links (clause-subgoals (subgoal-clause subgoal)) subgoal))))
	

;; FIXME: Wouldn't all back-subsumptions (for which this is nearly explicitly used) require the
;; the actual subgoal to be removed from the gr-index (and path-subgoal-lists?)
(define (delete-subgoals-links subgoals path-sg)
  (if (pair? subgoals)
      (begin
	(mark-links-deletable (subgoal-links (car subgoals)) (car subgoals))
	(set-subgoal-selected (car subgoals) #t)
	(set-subgoal-unlinkable (car subgoals) #t)
	(delete-subgoals-links (cdr subgoals) path-sg))))

(define (index-irregular-subgoals subgoals)
  (if (and regularity (pair? subgoals))
      (let ((result (ground-irregular-subgoal? (car subgoals))))
	(if result
	    result
	    (index-irregular-subgoals (cdr subgoals))))
      #f))

(define (init-link-bias)
  (if (and groundable (or (not sg-bias) (not term-bias)))
      (begin
	(set! sg-bias 4)
	(set! term-bias 1))
      (begin
	(if (not sg-bias)
	    (set! sg-bias 4))
	(if (not term-bias)
	    (set! term-bias 4))))
  (if horn
      (set! unit-offset 100)))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Equality functions
;; ------------------------------------------------------------

;; NOTA BENE: This is sufficient for the inital path only. If we solve a new subgoal which is an equation,
;; we need to create all the eq-links of the new equation with ALL path-literals! This is handled by function
;; insert-eq-links-for-new-equation which is called additionally for new equation subgoals.
(define (insert-eq-links subgoal eq-lits)
  ;; check for eq-links of subgoal with every eq-lit
  (if (pair? eq-lits)
      (if (subgoal-unlinkable (car eq-lits))
	  (insert-eq-links subgoal (cdr eq-lits))
	  (let ((eq-lit (car eq-lits)) eq-copy)
	    (if (eq? (subgoal-clause subgoal) (subgoal-clause eq-lit))
		(begin
		  (if (> debug-level 4)
		      (begin
			(display "creating renamed copy of clause ")
			(display-clause (subgoal-clause eq-lit))
			(newline)))
		  (set! eq-copy (renamed-copy (subgoal-clause eq-lit) #f))
		  (set-clause-depth eq-copy (clause-depth (subgoal-clause eq-lit)))
		  (set-clause-path eq-copy 'copy)
		  (set! eq-lit (subgoal-in-copy eq-lit (clause-subgoals eq-copy)))))
	    (if (not (eq? eq-lit falsum))
		;; side selection is incomplete!
		(if (and ordered-eq
			 (eq? (subgoal-pred subgoal) equality-pred)
			 (subgoal-ordered-eq subgoal))
			 ;;(and (not ordered-eq) (eq-subgoal? subgoal)))
		    (begin ;; if we have an ordered equation, overlapping is required into the maximal side only
		      (insert-eq-links-subterm subgoal eq-lit (subgoal-terms subgoal)
					       (ref (car (subgoal-terms eq-lit))))
		      ;; otherwise, overlapping is needed for all argument terms
		      (if (or (not ordered-eq) (not (subgoal-ordered-eq eq-lit)))
			  (insert-eq-links-subterm subgoal eq-lit (subgoal-terms subgoal)
						    (ref (cadr (subgoal-terms eq-lit)))))
		      (insert-eq-links subgoal (cdr eq-lits)))
		    (begin ;; usually we need to overlap into all argument terms
		      (insert-eq-links-subterms subgoal eq-lit (subgoal-terms subgoal) (ref (car (subgoal-terms eq-lit))))
		      ;; if we link with an ordered equation, overlapping is required with the maximal side only
		      (if (or (not ordered-eq) (not (subgoal-ordered-eq eq-lit)))
			  (insert-eq-links-subterms subgoal eq-lit (subgoal-terms subgoal) (ref (cadr (subgoal-terms eq-lit)))))
		      (insert-eq-links subgoal (cdr eq-lits)))))))))

(define (insert-eq-links-subterms subgoal eq-lit sg-terms eq-term)
  (if (pair? sg-terms)
      (begin
	(if (non-variable (ref (car sg-terms)))
	    ;; no need to paramodulate into variables!
	    (begin
	      (check-for-eq-link subgoal eq-lit sg-terms eq-term) 
	      (if (complex (ref (car sg-terms)))
		  (insert-eq-links-subterms subgoal eq-lit (cdr (ref (car sg-terms))) eq-term))))
	(insert-eq-links-subterms subgoal eq-lit (cdr sg-terms) eq-term))))

(define (insert-eq-links-subterm subgoal eq-lit terms eq-term)
  (if (pair? terms)
      (begin
	;; check term (and possibly its subterms) for an eq-link with eq-term
	(if (non-variable (ref (car terms)))
	    ;; no need to paramodulate into variables!
	    (begin
	      (check-for-eq-link subgoal eq-lit terms eq-term) 
	      (if (complex (ref (car terms))) ;; recursively check the arguments of term
		  (insert-eq-links-subterms subgoal eq-lit (cdr (ref (car terms))) eq-term)))))))

(define (check-for-eq-link subgoal eq-lit terms eq-term)
  (if (unify-terms-stop eq-term (ref (car terms)))
      (let ((save-term (car terms))
	    (rewritable #f)
	    (ordered-after-instantiation #f)
	    (replacing-side (get-other-eq-side eq-lit eq-term))
	    negated-plus-subgoals neg-eq-lit bottom-sg)
	;; here we check the rewrite condition: if the unit(!) equation is orientated after unification and the eq-term 
	;; can be matched into the overlapped subgoal and either at least one variable in the eq-term is instantiated
	;; or the overlapping is not on top level, then we have a rewrite step.
	(if (> (subgoal-depth subgoal) (subgoal-depth eq-lit))
	    (set! bottom-sg subgoal)
	    (set! bottom-sg eq-lit))
	(if (> debug-level 4)
	    (begin
	      (display "checking possible eq-link of ")
	      (display-subgoal eq-lit)
	      (if (eq? eq-term (ref (car (subgoal-terms eq-lit))))
		  (display " (->) ") 
		  (display " (<-) "))
	      (display " into subgoal ") 
	      (display-overlapped-subgoal subgoal terms)
	      (newline)))
	(if (not (and simplify (unit-subgoal eq-lit)))
	    (set! neg-eq-lit (make-new-negated-eq-sg eq-lit subgoal)))
	(if ordered-eq
	    (set! ordered-after-instantiation (or (subgoal-ordered-eq eq-lit) 
						  (kbo> (ref eq-term) (ref replacing-side)))))
	(if (and ordered-eq full-rewrite
		 (or (not (clause-initial (subgoal-clause subgoal))) (unit-subgoal subgoal))
		 (unit-subgoal eq-lit)
		 ordered-after-instantiation
		 (varlist-untouched? (clause-variables (subgoal-clause subgoal)))
		 (or (not (top-level? (ref (car terms)) (subgoal-terms subgoal)))
		     (not (varlist-renamed? (clause-variables (subgoal-clause eq-lit)) 
					    (clause-variables (subgoal-clause eq-lit)) '()))))
	    (set! rewritable #t))
	(set! linked-terms-ground #t)
	(set! max-sg-weight 0)
	(set! current-depth 0)
	(set! current-complexity 0)
	(set! immediately-solved 0)
	;; here we have to omit a number of the link redundancy checks as the same checks are done for
	;; original links and a posteriori lemma links!
	(if (and subsumption
		 (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause eq-lit)))
		 (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause eq-lit))))
	    (begin ;; or if it were subsumed
	      (if (> debug-level 4)
		  (begin
		    (display "overlapping equation subsumed")
		    (newline)))
	      (set! delete1 #t)
	      (set! open1 0))
	    (begin
	      (set! delete1 #f)
	      (set! open1 (get-open-subgoals (clause-subgoals (subgoal-clause eq-lit)) eq-lit bottom-sg))))
	;; Now the hairy bit begins! Temporarily replace term by replacing side
	(if (and ordered-eq (not (subgoal-ordered-eq eq-lit))
		 (kbo> (ref replacing-side) (ref eq-term))) ;; if the overlapping equation is ordered the wrong way
	    (begin                                          ;; after unification discard the paramodulated clause
	      (if (> debug-level 4)
		  (begin
		    (display "for eq-link of ")
		    (display-subgoal eq-lit)
		    (if (eq? eq-term (ref (car (subgoal-terms eq-lit))))
			(display " (->) ") 
			(display " (<-) "))
		    (display "overlapped clause deleted due to wrong orientation after instantiation") 
		    (newline)))
	      (set! delete2 #t)
	      (set! open2 0))
	    (if (and eager (not (unit-subgoal eq-lit))
		     (not (and subsumption (match-subgoal-with-index eq-lit 'subsume)))
		     (variable (ref replacing-side)) 
		     (not (occurs-term (ref replacing-side) (ref eq-term))))
		(begin
		  (if (> debug-level 4)
		      (begin
			(display "overlapped clause variant deleted due to variable replacing term")
			(newline)))
		  (set! delete2 #t)
		  (set! open2 0))
		(begin
		  (set-car! terms replacing-side)
		  (if (and simplify (unit-subgoal eq-lit))
		      (set! negated-plus-subgoals (clause-subgoals (subgoal-clause subgoal)))
		      (set! negated-plus-subgoals (cons neg-eq-lit (clause-subgoals (subgoal-clause subgoal)))))
		  (if (and subsumption
			   (subgoals-subsumed-by-index negated-plus-subgoals))
		      (begin ;; or if it were subsumed
			(if (> debug-level 4)
			    (begin
			      (display "overlapped clause subsumed")
			      (newline)))
			(set! delete2 #t)
			(set! open2 0))
		      (begin
			(set! delete2 #f)
			(if (and simplify (unit-subgoal eq-lit))
			    (set! open2 (get-open-subgoals negated-plus-subgoals #f bottom-sg))
			    (set! open2 (get-open-subgoals negated-plus-subgoals neg-eq-lit bottom-sg))))))))
	;; undo the temporary paramodulation
	(set-car! terms save-term)
	(unbind-vars)
	(if (not (and delete1 delete2))
	    (let ((link (make-eq-link eq-lit subgoal terms eq-term)))
	      (if (not (eq? (clause-path (subgoal-clause eq-lit)) 'copy))
		  (set-subgoal-links eq-lit (cons link (subgoal-links eq-lit))))
	      (set-subgoal-links subgoal (cons link (subgoal-links subgoal)))
	      (set-link-weight link (compute-link-weight eq-lit subgoal open1 open2 max-sg-weight))
	      (if rewritable 
		  (begin
		    (set! total-rewrite-links (+ total-rewrite-links 1))
		    (set-link-rewriting link #t)
		    (set-link-weight link 2)))
	      (if (and ordered-eq (not ordered-after-instantiation))
		  (set-link-weight link (bit-lsh (link-weight link) 1)))
	      (set-link-depth link current-complexity)
	      (if (< current-complexity min-relevant-depth)
		  (set! min-relevant-depth current-complexity))
	      (if (> debug-level 4)
		  (begin
		    (display "inserting eq-link with ")
		    (display-subgoal eq-lit)
		    (display " overlapping into ")
		    (display-subgoal subgoal)
		    (newline)))
	      (insert-link link (link-weight link))
	      (if (and link-info (= 0 (remainder total-links link-info)))
		  (begin
		    (display "link statistics after ")
		    (display total-links)
		    (display " links (average min max):")
		    (newline)
		    (print-link-statistics))))))))

;; this function is supposed to provide for all the required eq-links for a new positive eq-literal on the path or
;; a new positive eq-lemma
;; FIXME: insert-lemma-links must be changed to recognize symmetric path closure. The same holds true for subgoal-simplifiable!
(define (insert-eq-links-for-new-equation eq-lit predsubgoal)
  (if predsubgoal
      (begin
	;; NOTA BENE: Sign selection does not preserve completeness. The condition below
	;; must be replaced by (if (not (subgoal-unlinkable predsubgoal))...) only.
	;;(if (and (not (subgoal-unlinkable predsubgoal))
		 ;;(or ordered-eq (disequation? predsubgoal)
		     ;;(and (not (eq-subgoal? predsubgoal)) (eq? (subgoal-sign predsubgoal) selected-sign))))
	(if (not (subgoal-unlinkable predsubgoal))
	    (let (eq-copy)
	      (if (eq? (subgoal-clause predsubgoal) (subgoal-clause eq-lit))
		  (begin
		    (if (> debug-level 4)
			(begin
			  (display "creating renamed copy of clause ")
			  (display-clause (subgoal-clause eq-lit))
			  (newline)))
		    (set! eq-copy (renamed-copy (subgoal-clause eq-lit) #f))
		    (set-clause-depth eq-copy (clause-depth (subgoal-clause eq-lit)))
		    (set-clause-path eq-copy 'copy)
		    (set! eq-lit (subgoal-in-copy eq-lit (clause-subgoals eq-copy)))))
	      (if (> debug-level 4)
		  (begin
		    (display "checking ")
		    (display-subgoal eq-lit)
		    (display " with path subgoal ")
		    (display-subgoal predsubgoal)
		    (newline)))
	      (if (and ordered-eq (subgoal-ordered-eq predsubgoal))
		  ;; side selection is still incomplete!
		  ;;(and (not ordered-eq) (eq-subgoal? predsubgoal)))
		  (begin
		    (insert-eq-links-subterm predsubgoal eq-lit (subgoal-terms predsubgoal)
					     (ref (car (subgoal-terms eq-lit))))
		    (if (or (not ordered-eq) (not (subgoal-ordered-eq eq-lit)))
			(insert-eq-links-subterm predsubgoal eq-lit (subgoal-terms predsubgoal)
						 (ref (cadr (subgoal-terms eq-lit))))))
		  (begin
		    (insert-eq-links-subterms predsubgoal eq-lit (subgoal-terms predsubgoal) (ref (car (subgoal-terms eq-lit))))
		    (if (not (and ordered-eq (subgoal-ordered-eq eq-lit)))
			(insert-eq-links-subterms predsubgoal eq-lit (subgoal-terms predsubgoal) 
						  (ref (cadr (subgoal-terms eq-lit)))))))))
	(insert-eq-links-for-new-equation eq-lit (subgoal-predecessor predsubgoal)))))
		       
(define (get-other-eq-side eq-lit eq-term)
  (if (eq? eq-term (car (subgoal-terms eq-lit)))
      (cadr (subgoal-terms eq-lit))
      (car (subgoal-terms eq-lit))))

(define (make-new-negated-eq-sg eq-lit subgoal)
  ;; just creates a loose copy of eq-lit for a variety of checks, no variable-renaming takes place!
  (let ((new-sg (new-subgoal)))
    (set-subgoal-pred new-sg (subgoal-pred eq-lit))
    (set-subgoal-sign new-sg '-)
    (set-subgoal-terms new-sg (subgoal-terms eq-lit))
    (set-subgoal-weight new-sg (current-subgoal-weight eq-lit))
    (set-subgoal-ordered-eq new-sg (subgoal-ordered-eq eq-lit))
    (set-subgoal-clause new-sg (subgoal-clause subgoal))
    new-sg))
		       
(define (insert-eq-lemma-links lemma-sg predsubgoal)
  (if predsubgoal
      (begin
	(simplify-on-the-fly predsubgoal)
	(if (not (subgoal-unlinkable predsubgoal))
	    (begin
	      (if (> debug-level 8)
		  (begin
		    (display "Found unifying eq-subgoal ")
		    (display-subgoal predsubgoal)
		    (display " of depth ")
		    (display (subgoal-depth predsubgoal))
		    (newline)))
	      (if (and (unit-clause (subgoal-clause predsubgoal)) ;; if we find a unifiable unit we are done anyway
		       (not (subgoal-simplified predsubgoal))
		       (not (eq? (subgoal-sign lemma-sg) (subgoal-sign predsubgoal)))
		       (eq? (subgoal-pred lemma-sg) (subgoal-pred predsubgoal))
		       (or (unify-lists-stop (subgoal-terms lemma-sg) (subgoal-terms predsubgoal))
			   (unify-lists-stop (list (cadr (subgoal-terms lemma-sg)) (car (subgoal-terms lemma-sg))) 
					     (subgoal-terms predsubgoal)))) ;; check other order of args for symmetry!
		  (begin
		    (unbind-vars)
		    (if dewey-numbers
			(begin
			  (display "Found complementary unit for new eq-lemma: ")
			  (display-clause (subgoal-clause predsubgoal))
			  (display " of length ")
			  (display (clause-unsimplified (subgoal-clause predsubgoal)))
			  (newline)))
		    (subgoal-depth predsubgoal))
		  (if (and (subgoal-simplifiable predsubgoal) ;; if we can simplify a path subgoal with the new lemma...
			   ;; we must not allow pruning up to non-unit input clauses unless they have been totally
			   (or (not (clause-initial (subgoal-clause predsubgoal)))       ;; simplified
			       (= 0 (clause-unsimplified (subgoal-clause predsubgoal)))))
		      (let (tmp-result)
			(set! tmp-result (insert-eq-lemma-links lemma-sg (subgoal-predecessor predsubgoal)))
			;; we must not allow pruning up to non-unit input clauses!
			(if tmp-result
			    tmp-result
			    (subgoal-depth predsubgoal)))
		      (begin
			;; create all necessary eq-links
			(if (and (eq? (subgoal-sign lemma-sg) '+) ;; overlap lemma equation into all path literals
				 (not (eq? lemma-sg predsubgoal))) ;; do not overlap lemma in itself twice!
			    (overlap-with-eq-lemma lemma-sg predsubgoal))
			;; NOTA BENE: Sign selection does not preserve completeness. The condition below
			;; must be replaced by (if (equation? predsubgoal)...)
			;;(if (and (or ordered-eq (disequation? lemma-sg)
				     ;;(and (not (eq-subgoal? lemma-sg)) (eq? (subgoal-sign lemma-sg) selected-sign)))
				 ;;(equation? predsubgoal))
			(if (equation? predsubgoal)
			    (overlap-into-eq-lemma lemma-sg predsubgoal))
			(insert-eq-lemma-links lemma-sg (subgoal-predecessor predsubgoal))))))
	    (begin
	      (if (and back-subsume (not (subgoal-unlinkable predsubgoal))
		       (or unit-subsume (non-unit-clause (subgoal-clause predsubgoal))))
		  (let (tmp)
		    ;;(if regularity ;; strong back subsumption test
			(set! tmp (match-subgoal-with-index predsubgoal 'subsume))
			;;(set! tmp (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause predsubgoal)))))
		    (if (and tmp (not (eq? tmp predsubgoal)))
			(begin ;; back subsumption
			  (if dewey-numbers
			      (begin
				(display "back subsuming with new eq-lemma: ")
				(display-clause (subgoal-clause predsubgoal))
				(newline)))
			  (set! backward-subsumptions (+ 1 backward-subsumptions))
			  (delete-subgoals-links (clause-subgoals (subgoal-clause predsubgoal)) predsubgoal)))))
	      (insert-eq-lemma-links lemma-sg (subgoal-predecessor predsubgoal)))))
      (if (and (disequation? lemma-sg)
	       (unify-terms-stop (car (subgoal-terms lemma-sg)) (cadr (subgoal-terms lemma-sg))))
	  (begin
	    (unbind-vars)
	    (if dewey-numbers
		(begin
		  (display "Reflexivity link in new eq-lemma ")
		  (display-clause (subgoal-clause lemma-sg))
		  (display " closes tableau")
		  (newline)))
	    (subgoal-depth lemma-sg))
	  #f)))

(define (overlap-into-eq-lemma eq-lemma subgoal)
  (let (eq-copy)
    (if (eq? (subgoal-clause eq-lemma) (subgoal-clause subgoal))
	(begin
	  (if (> debug-level 4)
	      (begin
		(display "creating renamed copy of eq-clause ")
		(display-clause (subgoal-clause subgoal))
		(newline)))
	  (set! eq-copy (renamed-copy (subgoal-clause subgoal) #f))
	  (set-clause-depth eq-copy (clause-depth (subgoal-clause subgoal)))
	  (set-clause-path eq-copy 'copy)
	  (set! subgoal (subgoal-in-copy subgoal (clause-subgoals eq-copy)))))
    (if (equation? subgoal)
	(if (or (and ordered-eq
		     (eq? (subgoal-pred eq-lemma) equality-pred)
		     (subgoal-ordered-eq eq-lemma))
		(not ordered-eq))
	    (begin ;; if we have an ordered equation, overlapping is required into the maximal side only
	      (insert-eq-links-subterm eq-lemma subgoal (subgoal-terms eq-lemma)
				       (ref (car (subgoal-terms subgoal))))
	      ;; otherwise, overlapping is needed for all argument terms
	      (if (or (not ordered-eq) (not (subgoal-ordered-eq subgoal)))
		  (insert-eq-links-subterm eq-lemma subgoal (subgoal-terms eq-lemma)
					    (ref (cadr (subgoal-terms subgoal))))))
	    (begin ;; usually we need to overlap into all argument terms
	      (insert-eq-links-subterms eq-lemma subgoal (subgoal-terms eq-lemma) (ref (car (subgoal-terms subgoal))))
	      ;; if we link with an ordered equation, overlapping is required with the maximal side only
	      (if (or (not ordered-eq) (not (subgoal-ordered-eq subgoal)))
		  (insert-eq-links-subterms eq-lemma subgoal (subgoal-terms eq-lemma) (ref (cadr (subgoal-terms subgoal))))))))))

(define (overlap-with-eq-lemma eq-lemma predsubgoal)
  ;; NOTA BENE: Sign selection does not preserve completeness. The condition below
  ;; must be replaced by (if (not (subgoal-unlinkable predsubgoal))...) only.
  ;;(if (and (not (subgoal-unlinkable predsubgoal))
	   ;;(or ordered-eq (disequation? predsubgoal)
	       ;;(and (not (eq-subgoal? predsubgoal)) (eq? (subgoal-sign predsubgoal) selected-sign))))
  (if (not (subgoal-unlinkable predsubgoal))
      (begin
	(if (> debug-level 4)
	    (begin
	      (display "overlapping eq-lemma ")
	      (display-subgoal eq-lemma)
	      (display " into path subgoal ")
	      (display-subgoal predsubgoal)
	      (newline)))
	(if (or (and ordered-eq (subgoal-ordered-eq predsubgoal))
		(and (not ordered-eq) (eq-subgoal? predsubgoal)))
	    (begin
	      (insert-eq-links-subterm predsubgoal eq-lemma (subgoal-terms predsubgoal)
				       (ref (car (subgoal-terms eq-lemma))))
	      (if (or (not ordered-eq) (not (subgoal-ordered-eq eq-lemma)))
		  (insert-eq-links-subterm predsubgoal eq-lemma (subgoal-terms predsubgoal)
					   (ref (cadr (subgoal-terms eq-lemma))))))
	    (begin
	      (insert-eq-links-subterms predsubgoal eq-lemma (subgoal-terms predsubgoal) (ref (car (subgoal-terms eq-lemma))))
	      (if (not (and ordered-eq (subgoal-ordered-eq eq-lemma)))
		  (insert-eq-links-subterms predsubgoal eq-lemma (subgoal-terms predsubgoal) 
					    (ref (cadr (subgoal-terms eq-lemma))))))))))

(define (varlist-untouched? vars)
  (if (pair? vars)
      (if (free-variable (car vars))
	  (varlist-untouched? (cdr vars))
	  #f)
      #t))

(define (varlist-instantiated? vars)
  (if (pair? vars)
      (if (variable (ref (car vars)))
	  (varlist-instantiated? (cdr vars))
	  #t)
      #f))

(define (top-level? term sg-terms)
  (if (pair? sg-terms)
      (if (eq? term (ref (car sg-terms)))
	  #t
	  (top-level? term (cdr sg-terms)))
      #f))

(define (closing-literal-above subgoal closing-subgoals)
  (if (pair? closing-subgoals)
      (if (< (subgoal-depth (car closing-subgoals)) (subgoal-depth subgoal))
	  #t
	  (closing-literal-above subgoal (cdr closing-subgoals)))
      #f))

(define (insert-reflexivity-link subgoal)
  (if (unify-terms-stop (car (subgoal-terms subgoal)) (cadr (subgoal-terms subgoal)))
      (begin
	(set-subgoal-simplified subgoal #t)
	(set! linked-terms-ground #t)
	(set! max-sg-weight 0)
	(set! current-depth 0)
	(set! current-complexity 0)
	(set! immediately-solved 0)
	(set! index-end1 (clause-in-index? (subgoal-clause subgoal)))
	(if (and (vector? index-end1) (> (vector-ref index-end1 0) 0)) ;; if the first clause would become a variant
	    (begin
	      (set! delete1 #t)
	      (set! open1 0))
	    (if (and subsumption
		     (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause subgoal)))
		     (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause subgoal))))
		(begin ;; or if it were subsumed
		  (set! delete1 #t)
		  (set! open1 0))
		(begin
		  (set! delete1 #f)
		  (set! open1 (get-open-subgoals (clause-subgoals (subgoal-clause subgoal)) subgoal subgoal)))))
	(set-subgoal-simplified subgoal #f)
	(unbind-vars)
	(if (not delete1)
	    (let ((link (make-eq-link subgoal #f #f #f)))
	      (set-subgoal-links subgoal (cons link (subgoal-links subgoal)))
	      (set-link-weight link (compute-link-weight subgoal #f open1 0 max-sg-weight))
	      (set-link-depth link current-complexity)
	      (if (< current-complexity min-relevant-depth)
		  (set! min-relevant-depth current-complexity))
	      (if (> debug-level 4)
		  (begin
		    (display "inserting reflexivity link at depth ")
		    (display (subgoal-depth (link-lita link)))
		    (newline)))
	      ;; preference of reflexivity linkings in the absence of eager elimination
	      (if (not eager) (set-link-weight link 1))
	      (insert-link link (link-weight link))
	      (if (and link-info (= 0 (remainder total-links link-info)))
		  (begin
		    (display "link statistics after ")
		    (display total-links)
		    (display " links (average min max):")
		    (newline)
		    (print-link-statistics))))))
      (if (> debug-level 4)
	  (begin
	    (display "disequation sides not unifiable")
	    (newline)))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Functions for on-the-fly-simplification
;; ------------------------------------------------------------

(define (simplify-on-the-fly subgoal)
;   (display "simplifying side-literals of ")
;   (display-subgoal subgoal)
;   (newline)
  (if (and side-simplification simplify (simplify-subgoals-on-the-fly (clause-subgoals (subgoal-clause subgoal)) subgoal))
      (begin
	(if (> debug-level 8)
	    (begin
	      (display "Recomputing ")
	      (display (length (subgoal-links subgoal)))
	      (display " links for subgoal ")
	      (display-subgoal subgoal)
	      (display " of depth ")
	      (display (subgoal-depth subgoal))
	      (newline)))
	(recompute-links (subgoal-links subgoal)))))

(define (recompute-links links)
  (if (pair? links)
      (begin
	(if (not (eq? (link-active (car links)) 'delete))
	    (if (eq-link? (car links))
		(if (link-litb (car links))
		    (recompute-eq-link (car links))
		    (recompute-reflexivity-link (car links)))
		(recompute-link (car links))))
	(recompute-links (cdr links)))))

(define (recompute-link-lists)
  (recompute-regular-links all-links)
  (recompute-eq-links all-eq-links))

(define (recompute-regular-links links)
  (if (pair? links)
      (begin
	(if (not (eq? (link-active (car links)) 'delete))
	    (recompute-link (car links)))
	(recompute-regular-links (cdr links)))))

(define (recompute-eq-links links)
  (if (pair? links)
      (begin
	(if (not (eq? (link-active (car links)) 'delete))
	    (if (link-litb (car links))
		(recompute-eq-link (car links))
		(recompute-reflexivity-link (car links))))
	(recompute-eq-links (cdr links)))))

(define (recompute-link link)
  (let (link-cell bottom-sg)
    (if (> debug-level 8)
	(begin
	  (display "Recomputing ")
	  (if (link-active link)
	      (display "active")
	      (display "inactive"))
	  (display " link ")
	  (display-link link)
	  (newline)))
    (if (link-active link)
	(set! link-cell (extract-link-cell link)))
    (if (> (subgoal-depth (link-lita link)) (subgoal-depth (link-litb link)))
	(set! bottom-sg (link-lita link))
	(set! bottom-sg (link-litb link)))
    (unify-lists-stop (subgoal-terms (link-lita link)) (subgoal-terms (link-litb link)))
    (set! linked-terms-ground #t)
    (set! max-sg-weight 0)
    (set! current-depth 0)
    (set! current-complexity 0)
    (set! immediately-solved 0)
    (if (and subsumption
	     (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause (link-lita link))))
	     (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause (link-lita link)))))
	(begin ;; or if it were subsumed
	  (set! delete1 #t)
	  (set! open1 0))
	(begin
	  (set! delete1 #f)
	  (set! open1 (get-open-subgoals (clause-subgoals (subgoal-clause (link-lita link))) (link-lita link) bottom-sg))))
    (if (and subsumption
	     (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause (link-litb link))))
	     (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause (link-litb link)))))
	(begin ;; or if it were subsumed
	  (set! delete2 #t)
	  (set! open2 0))
	(begin
	  (set! delete2 #f)
	  (set! open2 (get-open-subgoals (clause-subgoals (subgoal-clause (link-litb link))) (link-litb link) bottom-sg))))
    (unbind-vars)
    (if (and delete1 delete2)
	(mark-links-deletable (list link) #f)
	(begin
	  (set-link-weight link (compute-link-weight (link-lita link) (link-litb link) open1 open2 max-sg-weight))
	  (set-link-depth link current-complexity)
	  (if (link-active link)
	      (restore-link-cell link-cell))))))
    


(define (recompute-eq-link link)
  (let (link-cell clausea replace by replaced-saved new-disequation eq-sg negated-plus-subgoals bottom-sg
		  (ordered-after-instantiation #f));; (rewritable #f))
    (if (> debug-level 8)
	(begin
	  (display "Recomputing ")
	  (if (link-active link)
	      (if (eq? (link-active link) 'delete)
		  (display "deletable")
		  (display "active"))
	      (display "inactive"))
	  (display " eq-link  ")
	  (display-link link)
	  (newline)))
    (if (> (subgoal-depth (link-lita link)) (subgoal-depth (link-litb link)))
	(set! bottom-sg (link-lita link))
	(set! bottom-sg (link-litb link)))
    (if (link-active link)
	(set! link-cell (extract-link-cell link)))
    (if (eq? (subgoal-clause (link-lita link)) (subgoal-clause (link-litb link)))
	(begin
	  (set! clausea (renamed-copy (subgoal-clause (link-lita link)) #f))
	  (set-clause-depth clausea (clause-depth (subgoal-clause (link-lita link))))
	  (set! eq-sg (subgoal-in-copy (link-lita link) (clause-subgoals clausea))))
	(set! eq-sg (link-lita link)))
    (if (link-left-to-right link)
	(begin
	  (set! replace (car (subgoal-terms eq-sg)))
	  (set! by (cadr (subgoal-terms eq-sg))))
	(begin
	  (set! replace (cadr (subgoal-terms eq-sg)))
	  (set! by (car (subgoal-terms eq-sg)))))
    (unify-terms-stop replace (ref (car (link-position link))))
    (if ordered-eq
	(set! ordered-after-instantiation (or (subgoal-ordered-eq (link-lita link)) 
					      (kbo> (ref replace) (ref by)))))
; 	(if (and ordered-eq full-rewrite
; 		 (or (not (clause-initial (subgoal-clause (link-litb link)))) (unit-subgoal (link-litb link)))
; 		 (unit-subgoal eq-sg)
; 		 ordered-after-instantiation
; 		 (varlist-untouched? (clause-variables (subgoal-clause (link-litb link))))
; 		 (or (not (top-level? (ref (car (link-position link))) (subgoal-terms (link-litb link))))
; 		     (not (varlist-renamed? (clause-variables (subgoal-clause eq-sg)) 
; 					    (clause-variables (subgoal-clause eq-sg)) '()))))
; 	    (set! rewritable #t))
    (set! new-disequation (make-new-negated-eq-sg eq-sg (link-litb link)))
    (set! replaced-saved (car (link-position link)))
    (set! linked-terms-ground #t)
    (set! max-sg-weight 0)
    (set! current-depth 0)
    (set! current-complexity 0)
    (set! immediately-solved 0)
    (if (and subsumption
	     (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause (link-lita link))))
	     (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause (link-lita link)))))
	(begin ;; or if it were subsumed
	  (if (> debug-level 4)
	      (begin
		(display "overlapping equation subsumed")
		(newline)))
	  (set! delete1 #t)
	  (set! open1 0))
	(begin
	  (set! delete1 #f)
	  (set! open1 (get-open-subgoals (clause-subgoals (subgoal-clause (link-lita link))) (link-lita link) bottom-sg))))
    ;; Now the hairy bit begins! Temporarily replace term by replacing side
    (if (and ordered-eq (not (subgoal-ordered-eq (link-lita link)))
	     (kbo> (ref by) (ref replace))) ;; if the overlapping equation is ordered the wrong way
	(begin                              ;; after unification discard the paramodulated clause
	  (if (> debug-level 4)
	      (begin
		(display "ignoring eq-link of ")
		(display-subgoal (link-lita link))
		(if (eq? replace (ref (car (subgoal-terms (link-lita link)))))
		    (display " (->) ") 
		    (display " (<-) "))
		(display "overlapped clause deleted due to wrong orientation after instantiation") 
		(newline)))
	  (set! delete2 #t)
	  (set! open2 0))
	(if (and (variable (ref by)) (not (occurs-term (ref by) (ref replace))))
	    (begin
	      (if (> debug-level 4)
		  (begin
		    (display "overlapped clause variant deleted due to variable replacing term")
		    (newline)))
	      (set! delete2 #t)
	      (set! open2 0))
	    (begin
	      (set-car! (link-position link) by)
	      (set! negated-plus-subgoals (cons new-disequation (clause-subgoals (subgoal-clause (link-litb link)))))
	      (if (and subsumption
		       (subgoals-subsumed-by-index negated-plus-subgoals))
		  (begin ;; or if it were subsumed
		    (if (> debug-level 4)
			(begin
			  (display "overlapped clause subsumed")
			  (newline)))
		    (set! delete2 #t)
		    (set! open2 0))
		  (begin
		    (set! delete2 #f)
		    (set! open2 (get-open-subgoals negated-plus-subgoals #f bottom-sg)))))))
    ;; undo the temporary paramodulation
    (set-car! (link-position link) replaced-saved)
    (unbind-vars)
    (if (and delete1 delete2)
	(mark-links-deletable (list link) #f)
	(begin
	  (set-link-weight link (compute-link-weight (link-lita link) (link-litb link) open1 open2 max-sg-weight))
	  (if (and ordered-eq (not ordered-after-instantiation))
	      (set-link-weight link (bit-lsh (link-weight link) 1)))
	  (if (link-rewriting link)
	      (set-link-weight link 2))
	  (set-link-depth link current-complexity)
	  (if (link-active link)
	      (restore-link-cell link-cell))))))

(define (recompute-reflexivity-link link)
  (let (link-cell)
    (if (> debug-level 8)
	(begin
	  (display "Recomputing ")
	  (if (link-active link)
	      (display "active")
	      (display "inactive"))
	  (display " reflexivity link ")
	  (display-link link)
	  (newline)))
    (if (link-active link)
	(set! link-cell (extract-link-cell link)))
    (unify-terms-stop (car (subgoal-terms (link-lita link))) (cadr (subgoal-terms (link-lita link))))
    (set! linked-terms-ground #t)
    (set! max-sg-weight 0)
    (set! current-depth 0)
    (set! current-complexity 0)
    (set! immediately-solved 0)
    (if (and subsumption
	     (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause (link-lita link))))
	     (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause (link-lita link)))))
	(begin ;; or if it were subsumed
	  (set! delete1 #t)
	  (set! open1 0))
	(begin
	  (set! delete1 #f)
	  (set! open1 (get-open-subgoals (clause-subgoals (subgoal-clause (link-lita link))) (link-lita link) (link-lita link)))))
    (unbind-vars)
    (if delete1
	(mark-links-deletable (list link) #f)
	(begin
	  (set-link-weight link (compute-link-weight (link-lita link) #f open1 0 max-sg-weight))
	  (set-link-depth link current-complexity)
	  (if (link-active link)
	      (restore-link-cell link-cell))))))
    
(define (extract-link-cell link)
  (let (index appropriate-link-vector appropriate-link-vector-size)
    (if (eq-link? link)
	(begin
	  (set! appropriate-link-vector eq-link-vector)
	  (set! appropriate-link-vector-size eq-link-vector-size))
	(begin
	  (set! appropriate-link-vector link-vector)
	  (set! appropriate-link-vector-size link-vector-size)))
    (if (or (>= (link-weight link) appropriate-link-vector-size) (< (link-weight link) 0))
	(set! index (- appropriate-link-vector-size 1))
	(set! index (link-weight link)))
    (extract-link-from-list link (vector-ref appropriate-link-vector index) (vector-ref appropriate-link-vector index) index)))

(define (extract-link-from-list link list pred-list index)
  (if (pair? list)
      (if (eq-link? link)
	  (if (eq? (car list) link)
	      (if (eq? list (vector-ref eq-link-vector index))
		  (begin
		    (vector-set! eq-link-vector index (cdr list))
		    list)
		  (begin
		    (set-cdr! pred-list (cdr list))
		    list))
	      (extract-link-from-list link (cdr list) list index))
	  (if (eq? (car list) link)
	      (if (eq? list (vector-ref link-vector index))
		  (begin
		    (vector-set! link-vector index (cdr list))
		    list)
		  (begin
		    (set-cdr! pred-list (cdr list))
		    list))
	      (extract-link-from-list link (cdr list) list index)))
      #f))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Functions for the complete link list
;; ------------------------------------------------------------

(define (get-oldest-link-cell)
  (let ((link (get-first-active-link all-links all-links)))
    ;;(display "all-links: ")
    ;;(display-current-link-list all-links)
    (if link
	(begin
	  ;;(display "found active link ")
	  ;;(display-link link)
	  ;;(newline)
	  (extract-specified-link-cell link))
	#f)))

(define (get-first-active-link links pred-links)
  (if (pair? links)
      (if (eq? (link-active (car links)) 'delete)
	  (begin
	    (set! length-all-links (- length-all-links 1))
	    (if (eq? links all-links)
		(begin
		  (if (eq? last-link links) (set! last-link (cdr links)))
		  (set! all-links (cdr links))
		  (get-first-active-link (cdr links) (cdr links)))
		(begin
		  (set-cdr! pred-links (cdr links))
		  (if (null? (cdr links)) (set! last-link pred-links))
		  (get-first-active-link (cdr links) pred-links))))
	  (if (not (link-active (car links)))
	      (get-first-active-link (cdr links) links)
	      (car links)))
      #f))

(define (get-oldest-eq-link-cell)
  (let ((link (get-first-active-eq-link all-eq-links all-eq-links)))
    (if link
	(extract-specified-link-cell link)
	#f)))

(define (get-first-active-eq-link links pred-links)
  (if (pair? links)
      (if (eq? (link-active (car links)) 'delete)
	  (begin
	    (set! length-all-eq-links (- length-all-eq-links 1))
	    (if (eq? links all-eq-links)
		(begin
		  (if (eq? last-eq-link links) (set! last-eq-link (cdr links)))
		  (set! all-eq-links (cdr links))
		  (get-first-active-eq-link (cdr links) (cdr links)))
		(begin
		  (set-cdr! pred-links (cdr links))
		  (if (null? (cdr links)) (set! last-eq-link pred-links))
		  (get-first-active-eq-link (cdr links) pred-links))))
	  (if (not (link-active (car links)))
	      (get-first-active-eq-link (cdr links) links)
	      (car links)))
      #f))

(define (extract-specified-link-cell link)
  (let ((weight (link-weight link))
	index appropriate-link-vector-size)
    (if (eq-link? link)
	(set! appropriate-link-vector-size eq-link-vector-size)
	(set! appropriate-link-vector-size link-vector-size))
    (if (or (>= weight appropriate-link-vector-size) (< weight 0))
	(set! index (- appropriate-link-vector-size 1))
	(set! index weight))
    (if (eq-link? link)
	(extract-specified-eq-link-from-list link (vector-ref eq-link-vector index) (vector-ref eq-link-vector index) index)
	(extract-specified-link-from-list link (vector-ref link-vector index) (vector-ref link-vector index) index))))

(define (extract-specified-link-from-list link list pred-list index)
  (if (pair? list)
      (if (eq? (link-active (car list)) #t)
	  (if (eq? link (car list))
	      (begin
		(set! length-vector-links (- length-vector-links 1))
		(if (eq? list (vector-ref link-vector index))
		    (begin
		      (vector-set! link-vector index (cdr list))
		      (set-cdr! list '())
		      list)
		    (begin
		      (set-cdr! pred-list (cdr list))
		      (set-cdr! list '())
		      list)))
	      (begin
		(set! active-links-exist #t)
		(extract-specified-link-from-list link (cdr list) list index)))
	  (begin
	    (set! length-vector-links (- length-vector-links 1)) 
	    (if (eq? list (vector-ref link-vector index)) ;; on the fly deletion of deletable links
		(begin
		  (vector-set! link-vector index (cdr list))
		  (extract-specified-link-from-list link (vector-ref link-vector index) (vector-ref link-vector index) index))
		(begin
		  (set-cdr! pred-list (cdr list))
		  (extract-specified-link-from-list link (cdr pred-list) pred-list index)))))
      #f))

(define (extract-specified-eq-link-from-list link list pred-list index)
  (if (pair? list)
      (if (eq? (link-active (car list)) #t)
	  (if (eq? link (car list))
	      (begin
		(set! length-vector-eq-links (- length-vector-eq-links 1))
		(if (eq? list (vector-ref eq-link-vector index))
		    (begin
		      (vector-set! eq-link-vector index (cdr list))
		      (set-cdr! list '())
		      list)
		    (begin
		      (set-cdr! pred-list (cdr list))
		      (set-cdr! list '())
		      list)))
	      (begin
		(set! active-eq-links-exist #t)
		(extract-specified-eq-link-from-list link (cdr list) list index)))
	  (begin
	    (set! length-vector-eq-links (- length-vector-eq-links 1)) 
	    (if (eq? list (vector-ref eq-link-vector index)) ;; on the fly deletion of deletable links. as non-active links are extracted,
		(begin                                    ;; anything not active #t still here must be deletable
		  (vector-set! eq-link-vector index (cdr list))
		  (extract-specified-eq-link-from-list link (vector-ref eq-link-vector index) 
						       (vector-ref eq-link-vector index) index))
		(begin
		  (set-cdr! pred-list (cdr list))
		  (extract-specified-eq-link-from-list link (cdr pred-list) pred-list index)))))
      #f))


(define (sweep-all-links)
  (set! length-all-links 0)
  (set! all-links (sweep-link-list all-links 'all))
  (if (not (pair? all-links))
      (set! last-link all-links)))
	
(define (sweep-all-eq-links)
  (set! length-all-eq-links 0)
  (set! all-eq-links (sweep-link-list all-eq-links 'all))
  (if (not (pair? all-eq-links))
      (set! last-eq-link all-eq-links)))
	
(define (sweep-link-vector vector size type)
  (let (index)
    (if (eq? type 'eq)
	(set! length-vector-eq-links 0)
	(set! length-vector-links 0))
    (do ((index 0 (+ index 1)))
	((>= index size) #t)
      (vector-set! vector index (sweep-link-list (vector-ref vector index) 'vector)))))

; (define (sweep-link-list links type)
;   (if (pair? links)
;       (if (eq? (link-active (car links)) 'delete)
; 	  (begin
; 	    (if dewey-numbers
; 		(begin
; 		  (if (eq? type 'all)
; 		      (display "Sweeping link ")
; 		      (display "Sweeping vector link "))
; 		  (display-link (car links))
; 		  (newline)))
; 	    (if (eq? type 'all)
; 		(set! links-sweeped (+ links-sweeped 1))
; 		(set! vector-links-sweeped (+ vector-links-sweeped 1)))
; 	    (sweep-link-list (cdr links) type))
; 	  (begin
; 	    (if (eq-link? (car links))
; 		(if (eq? type 'all)
; 		    (set! length-all-eq-links (+ length-all-eq-links 1))
; 		    (set! length-vector-eq-links (+ length-vector-eq-links 1)))
; 		(if (eq? type 'all)
; 		    (set! length-all-links (+ length-all-links 1))
; 		    (set! length-vector-links (+ length-vector-links 1))))
; 	    (set-cdr! links (sweep-link-list (cdr links) type))
; 	    (if (and (eq? type 'all) (not (pair? (cdr links))))
; 		(if (eq-link? (car links))
; 		    (set! last-eq-link links)
; 		    (set! last-link links)))
; 	    links))
;       '()))

(define (sweep-link-list links type)
  (let ((new-links '()))
    (do ((l links (cdr l)))
	((or (not (pair? l)) (not (eq? (link-active (car l)) 'delete)))
	 (set! new-links l))
      (begin
	(if dewey-numbers
	    (begin
	      (if (eq? type 'all)
		  (display "Sweeping link ")
		  (display "Sweeping vector link "))
	      (display-link (car l))
	      (newline)))
	(if (eq? type 'all)
	    (set! links-sweeped (+ links-sweeped 1))
	    (set! vector-links-sweeped (+ vector-links-sweeped 1)))))
    (if (pair? new-links)
	(do ((l (cdr new-links) (cdr l))
	     (new-l new-links))
	    ((not (pair? l))
	     (begin
	       (set-cdr! new-l '())
	       (if (eq? type 'all)
		   (if (eq-link? (car new-links))
		       (set! last-eq-link new-l)
		       (set! last-link new-l)))
	       new-links))
	  (begin
	    (if (eq? (link-active (car l)) 'delete)
		(begin
		  (if dewey-numbers
		      (begin
			(if (eq? type 'all)
			    (display "Sweeping link ")
			    (display "Sweeping vector link "))
			(display-link (car links))
			(newline)))
		  (if (eq? type 'all)
		      (set! links-sweeped (+ links-sweeped 1))
		      (set! vector-links-sweeped (+ vector-links-sweeped 1))))
		(begin
		  (set-cdr! new-l l)
		  (set! new-l (cdr new-l))
		  (if (eq-link? (car new-links))
		      (if (eq? type 'all)
			  (set! length-all-eq-links (+ length-all-eq-links 1))
			  (set! length-vector-eq-links (+ length-vector-eq-links 1)))
		      (if (eq? type 'all)
			  (set! length-all-links (+ length-all-links 1))
			  (set! length-vector-links (+ length-vector-links 1))))))))
	new-links)))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Functions for locally optimised link selection
;; ------------------------------------------------------------

(define (local-link-check)
  (check-for-locally-optimised-links all-links))

(define (check-for-locally-optimised-links links)
  (if (pair? links)
      (if (or (eq? (link-active (car links)) 'delete)
	      (not (find-locally-optimal-link (car links))))
	  (check-for-locally-optimised-links (cdr links))
	  (extract-specified-link-cell (car links)))
      #f))

(define (find-locally-optimal-link link)
  (let (link-cell index-variant)
    (if (> debug-level 8)
	(begin
	  (display "Locally checking ")
	  (if (link-active link)
	      (display "active")
	      (display "inactive"))
	  (display " link ")
	  (display-link link)
	  (newline)))
    (if (link-active link)
	(begin
	  (unify-lists-stop (subgoal-terms (link-lita link)) (subgoal-terms (link-litb link)))
	  (set! index-variant (clause-in-index? (subgoal-clause (link-lita link))))
	  (if (or (and (vector? index-variant) (> (vector-ref index-variant 0) 0))
		  (and subsumption
		       (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause (link-lita link))))
		       (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause (link-lita link)))))
		  (and regularity (index-irregular-subgoals (clause-subgoals (subgoal-clause (link-lita link))))))
	      (begin ;; or if it were subsumed
		(set! delete1 #t)
		(set! open1 0))
	      (begin
		(set! delete1 #f)
		(set! open1 (get-local-open-subgoals (clause-subgoals (subgoal-clause (link-lita link))) (link-lita link)))))
	  (set! index-variant (clause-in-index? (subgoal-clause (link-litb link))))
	  (if (or (and (vector? index-variant) (> (vector-ref index-variant 0) 0))
		  (and subsumption
		       (or (and unit-subsume simplify) (non-unit-clause (subgoal-clause (link-litb link))))
		       (subgoals-subsumed-by-index (clause-subgoals (subgoal-clause (link-litb link)))))
		  (and regularity (index-irregular-subgoals (clause-subgoals (subgoal-clause (link-litb link))))))
	      (begin ;; or if it were subsumed
		(set! delete2 #t)
		(set! open2 0))
	      (begin
		(set! delete2 #f)
		(set! open2 (get-local-open-subgoals (clause-subgoals (subgoal-clause (link-litb link))) (link-litb link)))))
	  (unbind-vars)
	  (if (not (and delete1 delete2))
	      (if (and (= open1 0) (= open2 0))
		  (begin
		    (if dewey-numbers
			(begin
			  (display "found locally optimal link: ")
			  (display-link link)
			  (newline)))
		    link)))))))


(define (get-local-open-subgoals subgoals link-sg)
  (if (pair? subgoals)
      (if (or (subgoal-simplified (car subgoals))
	      (match-subgoal-with-index (car subgoals) 'simplify))
	  (get-local-open-subgoals (cdr subgoals) link-sg)
	  (let (sg-weight (gr-closed (ground-closed-with-index (car subgoals)))
			  (duplicate (subgoal-in-list (car subgoals) (cdr subgoals))))
	    (if duplicate
		(begin
		  (if (eq? (car subgoals) link-sg)
		      (get-local-open-subgoals (cdr subgoals) duplicate)
		      (get-local-open-subgoals (cdr subgoals) link-sg)))
		(if (and (not gr-closed) (not (eq? (car subgoals) link-sg)))
		    (+ 1 (get-local-open-subgoals (cdr subgoals) link-sg))
		    (get-local-open-subgoals (cdr subgoals) link-sg)))))
      0))
