(module pruning
	(include "dctp.sch")
	(export (merge-rel-vectors rel-vector-a rel-vector-b depth) (mark-rel-vector rel-vector depth) 
		(subgoal-relevant subgoal rel-vector) (new-rel-vector depth) (initialize-table table offset) 
		(non-tautological clause) (ref-equal s t) (initialize-bit-mask-table)  (path-vector-ref i)
		(is-context-free rel-vector depth) (display-rel-vector r) (make-top-down-lemma subgoal) 
		(subgoal-simplifiable subgoal) (foldup-subgoal subgoal rel-vector) (delete-clause-foldup-lits clause)
		(path-vector-set i clause) (simplified-unit subgoal) (expansion-order-violation subgoals)
		(non-tautological-resolvent clausea clauseb linked-sga linked-sgb)
		(simplify-subgoals-on-the-fly subgoals subgoal) (subgoal-in-list subgoal subgoals))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm") (kbo "kbo.scm")
		(index "index.scm") (link "link.scm")  (dctp "dctp.scm") (select "select.scm") (path "path.scm")
		(renamed-termlist normal-form "normal-form.scm") (display-local-instances dctp "dctp.scm")))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define path-vector (make-vector 5000 #f))
(define on-the-fly-lemmas '())
(define simplification-prefix-string "simplification_lemma_")
(define foldup-prefix-string "foldup_lemma_")

;;;*************************************************************************
;; ------------------------------------------------------------
;; Path vector access
;; ------------------------------------------------------------

(define (path-vector-ref i)
  (if (< i (vector-length path-vector))
      (vector-ref path-vector i)
      (let ((new-path-vector (make-vector (* 2 (vector-length path-vector)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length path-vector)) #f)
	  (vector-set! new-path-vector j (vector-ref path-vector j)))
	(set! path-vector new-path-vector)
	#f)))
	
(define (path-vector-set i clause)
  (if (< i (vector-length path-vector))
      (vector-set! path-vector i clause)
      (let ((new-path-vector (make-vector (* 2 (vector-length path-vector)) #f)))
	(do ((j 0 (+ j 1)))
	    ((= j (vector-length path-vector)) #f)
	  (vector-set! new-path-vector j (vector-ref path-vector j)))
	(set! path-vector new-path-vector)
	(vector-set! path-vector i clause)
	#t)))
	
;;;*************************************************************************
;; ------------------------------------------------------------
;; Term equality
;; ------------------------------------------------------------

(define (ref-equals k l)
;; checks whether termlists k and l are equal 
 (if (null? k)
     #t
     (if (ref-equal (ref (car k)) (ref (car l)))
	 (ref-equals (cdr k) (cdr l))
	 #f)))

(define (ref-equal s t)          
;; checks whether terms s and t are equal 
 (if (eq? s t)
     #t
     (if (or (index-variable s) (index-variable t) (variable s) (variable t))
         #f
         (and (eq? (car s) (car t)) 
	      (ref-equals (cdr s) (cdr t))))))

;;;*************************************************************************
;; -------
;; Pruning
;; -------

(define (merge-rel-vectors rel-vector-a rel-vector-b depth)
  (if (and rel-vector-a rel-vector-b)
      (if (eq? #t rel-vector-a)
	  rel-vector-b
	  (if (eq? #t rel-vector-b)
	      rel-vector-a
	      (begin
		(merge-boolean-vectors rel-vector-a rel-vector-b (rel-vec-index depth))
		rel-vector-a)))
      #f))

(define (merge-boolean-vectors vec1 vec2 index)
 (if (>= index 0)
     (begin (vector-set! vec1 index (bit-or (vector-ref vec1 index)
					    (vector-ref vec2 index)))
	    (merge-boolean-vectors vec1 vec2 (- index 1)))))

(define (mark-rel-vector rel-vector depth)
  (if (>= depth 0)
      (vector-set! rel-vector (rel-vec-index depth)
		   (bit-or (rel-vec-position depth) (vector-ref rel-vector (rel-vec-index depth))))))

(define (subgoal-relevant subgoal rel-vector)
  (let (depth result)
    (set! depth (subgoal-depth subgoal))
    (if (>= depth 0)
	(begin
	  (set! result (bit-and (rel-vec-position depth) (vector-ref rel-vector (rel-vec-index depth))))
	  (if (= 0 result)
	      #f
	      #t))
	#t)))

(define (new-rel-vector depth)
  (set! rel-vectors-allocated (+ rel-vectors-allocated 1))
  (make-vector (max min-rel-vec-length (+ 2 (rel-vec-index depth))) 0))

(define (copy-rel-vector rel-vector)
  (let ((new-vector (make-vector (vector-length rel-vector))))
    (set! rel-vectors-allocated (+ rel-vectors-allocated 1))
    (do ((i 0 (+ i 1)))
	((>= i (vector-length new-vector)) new-vector)
      (vector-set! new-vector i (vector-ref rel-vector i)))))

(define (initialize-table table offset)
  (do ((i offset (+ i 1)))
      ((>= i (vector-length table)) #t)
    (vector-set! table i 
		 (cons (quotient i bit-vector-length) (bit-lsh 1 (- bit-vector-length 1 (remainder i bit-vector-length)))))))

(define (rel-vec-index depth)
  (if (>= depth (vector-length rel-vec-indexing-table))
      (let ((new-table (make-vector (* 2 (vector-length rel-vec-indexing-table)) #f)))
	(do ((i 0 (+ i 1)))
	    ((>= i (vector-length rel-vec-indexing-table)) #t)
	  (vector-set! new-table i (vector-ref rel-vec-indexing-table i)))
	(initialize-table new-table (vector-length rel-vec-indexing-table))
	(set! rel-vec-indexing-table new-table)))
  (car (vector-ref rel-vec-indexing-table depth)))

(define (rel-vec-position depth)
  (if (>= depth (vector-length rel-vec-indexing-table))
      (let ((new-table (make-vector (* 2 (vector-length rel-vec-indexing-table)) #f)))
	(do ((i 0 (+ i 1)))
	    ((>= i (vector-length rel-vec-indexing-table)) #t)
	  (vector-set! new-table i (vector-ref rel-vec-indexing-table i)))
	(initialize-table new-table (vector-length rel-vec-indexing-table))
	(set! rel-vec-indexing-table new-table)))
  (cdr (vector-ref rel-vec-indexing-table depth)))

(define (initialize-bit-mask-table)
  (let ((mask (bit-lsh 1 (- bit-vector-length 1))))
    (set! bit-mask-table (make-vector bit-vector-length 0))
    (do ((i 1 (+ i 1)))
	((= i bit-vector-length) #t)
      (vector-set! bit-mask-table i mask)
      (set! mask (bit-or mask (bit-rsh mask 1))))))

(define (is-context-free rel-vector depth)
  (let ((rel-vector-index 0))
    (if (do ((i 0 (+ i 1)))
	    ((or (not (= (vector-ref rel-vector i) 0)) (< depth bit-vector-length)) (< depth bit-vector-length))
	  (set! rel-vector-index (+ rel-vector-index 1))
	  (set! depth (- depth bit-vector-length)))
	(begin
	  ;;(display depth)
	  ;;(display " ")
	  ;;(display (vector-ref bit-mask-table depth))
	  ;;(display " ")
	  (= 0 (bit-and (vector-ref rel-vector rel-vector-index) (vector-ref bit-mask-table depth))))
	#f)))

(define (foldup-depth r depth)
  (let ((l 0) 
	(depth-offset (quotient depth bit-vector-length)) 
	(depth-index (remainder depth bit-vector-length)) 
	tmp)
    (set! tmp (vector-ref r depth-offset)) ;; save the rel-vector entry of depth and blank out levels depth and below
    (vector-set! r depth-offset (bit-and (vector-ref r depth-offset) (vector-ref bit-mask-table depth-index)))
    (set! l (do ((i depth-offset (- i 1))) ;; search for the first non-zero rel-vector element
		    ((or (< i 0) (not (= (vector-ref r i) 0))) i)))
    (if (< l 0) 
	(begin
	  (vector-set! r depth-offset tmp) ;; restore levels depth and below
	  l)
	(let (index)
	  (set! index (do ((j 0 (+ j 1))) ;; from the right end of the bit-vector, search the first non-zero bit
			  ((or (>= j bit-vector-length) (not (= 0 (bit-and (vector-ref r l) (bit-lsh 1 j))))) j)))
	  (vector-set! r depth-offset tmp)
	  (+ (- bit-vector-length index 1) (* bit-vector-length l))))))
    
(define (copy-up-to rel-vector depth) ;; depth is the first level not copied!
  (let ((new-vector (make-vector (vector-length rel-vector) 0))
	(depth-offset (quotient depth bit-vector-length)) 
	(depth-index (remainder depth bit-vector-length)))
    (set! rel-vectors-allocated (+ rel-vectors-allocated 1))
    (do ((i 0 (+ i 1)))
	((> i depth-offset) new-vector)
      (vector-set! new-vector i (vector-ref rel-vector i)))
    (vector-set! new-vector depth-offset (bit-and (vector-ref rel-vector depth-offset) 
						  (vector-ref bit-mask-table depth-index)))
    new-vector))

(define (clear-depth-and-below r depth)
  (let ((depth-offset (quotient depth bit-vector-length)) 
	(depth-index (remainder depth bit-vector-length)))
    (vector-set! r depth-offset (bit-and (vector-ref r depth-offset) (vector-ref bit-mask-table depth-index)))
    (do ((i (+ 1 depth-offset) (+ i 1)))
	((>= i (vector-length r)) #t)
      (vector-set! r i 0))))

(define (display-rel-vector r)
  (do ((i 0 (+ i 1)))
      ((>= i (vector-length r)) #t)
    (if (not (= (vector-ref r i) 0))
	(do ((j 1 (+ j 1)))
	    ((> j bit-vector-length) #t)
	  (if (not (= 0 (bit-and (vector-ref r i) (bit-lsh 1 (- bit-vector-length j)))))
	      (begin
		(display " ")
		(display (+ (- j 1) (* bit-vector-length i)))))))))



;;;*************************************************************************
;; ---------------------
;; Purity deletion
;; ---------------------

; (define (pure-clause linked-sg)
;   (if (pure-subgoals (clause-subgoals (subgoal-clause linked-sg)) linked-sg)
;       (begin
; 	(set! purity-fails (+ purity-fails 1))
; 	(if (or dewey-numbers (> debug-level 4))
; 	    (begin
; 	      (display "rejected pure instance ")
; 	      (display-clause (subgoal-clause linked-sg))
; 	      (newline)))
; 	#t)
;       #f))

; (define (pure-subgoals subgoals linked-sg)
;   (if (pair? subgoals)
;       (if (and (not (eq? linked-sg (car subgoals)))
; 	       (not (pair? (unify-subgoal-with-index (car subgoals) 'purity))))
; 	  (car subgoals)
; 	  (pure-subgoals (cdr subgoals) linked-sg))
;       #f))

;;;*************************************************************************
;; ---------------------
;; Tautology elimination
;; ---------------------

(define (non-tautological clause)
  (if (non-tautological-subgoals (clause-subgoals clause))
      #t
      (begin
	(set! tautology-fails (+ tautology-fails 1))
	(if (or dewey-numbers (> debug-level 4))
	    (begin
	      (display "rejected tautological ")
	      (display-clause clause)
	      (newline)))
	#f)))

(define (non-tautological-subgoals subgoals)
  (if (and (pair? subgoals) (eq? (subgoal-sign (car subgoals)) `+))  ;; as positive subgoals are stored first, no need to search in negative rest of clause
      (and (not (eq? verum (car subgoals)))             ;; Verum is tautological, too
	   (not (tautology-check (car subgoals) (cdr subgoals)))  
	   (non-tautological-subgoals (cdr subgoals)))
      #t))

(define (tautology-check subgoal subgoals)
  ;; tautological equations need not be checked, these are caught by reflexivity subsumption
  (if (pair? subgoals)
      (if (and (not (eq? (subgoal-sign subgoal) (subgoal-sign (car subgoals))))
	       (eq? (subgoal-pred subgoal) (subgoal-pred (car subgoals)))
	       (ref-equals (subgoal-terms subgoal) (subgoal-terms (car subgoals))))
	  #t
	  (tautology-check subgoal (cdr subgoals)))
      #f))

(define (non-tautological-resolvent clausea clauseb linked-sga linked-sgb)
  (if (non-tautological-subgoalsa (clause-subgoals clausea) (clause-subgoals clauseb) linked-sga linked-sgb)
      #t
      (begin
	(set! tautology-fails (+ tautology-fails 1))
	(set! tautological-resolvents (+ tautological-resolvents 1))
	(if (or dewey-numbers (> debug-level 4))
	    (begin
	      (display "rejected tautological resolvent ")
	      (display-clause clausea)
	      (display " and ")
	      (display-clause clauseb)
	      (newline)))
	#f)))

(define (non-tautological-subgoalsa subgoalsa subgoalsb linked-sga linked-sgb)
  (if (pair? subgoalsa)
      (and (or (eq? (car subgoalsa) linked-sga) (not (linked-tautology-check (car subgoalsa) subgoalsb linked-sgb)))
	   (non-tautological-subgoalsa (cdr subgoalsa) subgoalsb linked-sga linked-sgb))
      #t))

(define (linked-tautology-check subgoal subgoals linked-sg)
  (if (pair? subgoals)
      (if (and (not (eq? (car subgoals) linked-sg))
	       (not (eq? (subgoal-sign subgoal) (subgoal-sign (car subgoals))))
	       (eq? (subgoal-pred subgoal) (subgoal-pred (car subgoals)))
	       (ref-equals (subgoal-terms subgoal) (subgoal-terms (car subgoals))))
	  #t
	  (tautology-check subgoal (cdr subgoals)))
      (if (eq? #t (subgoal-pred subgoal))
	  #t
	  #f)))

(define (expansion-order-violation subgoals)
  (if (pair? subgoals)
      (if (and (subgoal-expansion-lit (car subgoals))
	       (not (subgoal-ordered-eq (car subgoals)))
	       (kbo> (cadr (subgoal-terms (car subgoals))) (car (subgoal-terms (car subgoals)))))
	  (begin
	    (if dewey-numbers
		(begin
		  (display "order violating instantiation of expansion literal ")
		  (display-subgoal (car subgoals))
		  (newline)))
	    (set! expansion-order-violations (+ expansion-order-violations 1))
	    #t)
	  (expansion-order-violation (cdr subgoals)))
      #f))
		  

;;;*************************************************************************
;; -------------------------
;; Foldup and simplification
;; -------------------------

(define (subgoal-simplifiable subgoal)
  (if simplify
      (if (subgoal-simplified subgoal)
	  #t
	  (let (index-end)
	    (set! index-end (match-subgoal-with-index subgoal 'simplify))
	    (if index-end
		(begin
		  (if (or dewey-numbers (> debug-level 4))
		      (begin
			(display "simplifying subgoal ")
			(display-subgoal subgoal)
			(display " of depth ")
			(display (subgoal-depth subgoal))
			(display " matching unit ")
			(display-clause (subgoal-clause index-end))
			(newline)))
		  (set-subgoal-simplified subgoal #t)
		  (if (subgoal-clause subgoal)
		      (set-clause-unsimplified (subgoal-clause subgoal) (- (clause-unsimplified (subgoal-clause subgoal)) 1)))
		  (set! nr-simplifications (+ nr-simplifications 1))
		  #t)
		#f)))
      #f))

(define (simplify-subgoals-on-the-fly subgoals subgoal)
  (if (pair? subgoals)
      (let ((success #f))
	(if (and (or (not (eq? (car subgoals) subgoal)) (clause-initial (subgoal-clause (car subgoals))))
		 (not (subgoal-simplified (car subgoals))))
	    (let (index-end)
	      (set! index-end (match-subgoal-with-index (car subgoals) 'simplify))
	      (if index-end
		  (begin
		    (if (or dewey-numbers (> debug-level 4))
			(begin
			  (display "on-the fly simplification of subgoal ")
			  (display-subgoal (car subgoals))
			  (display " of depth ")
			  (display (subgoal-depth (car subgoals)))
			  (display " in clause of length ")
			  (display (clause-unsimplified (subgoal-clause (car subgoals))))
			  (display " matching unit ")
			  (display-clause (subgoal-clause index-end))
			  (newline)))
		    (set-subgoal-simplified (car subgoals) #t)
		    (set! success #t)
		    (set-clause-unsimplified (subgoal-clause (car subgoals)) 
					     (- (clause-unsimplified (subgoal-clause (car subgoals))) 1))
		    (if (= 1 (clause-unsimplified (subgoal-clause (car subgoals))))
			(begin
			  (mark-links-deletable (subgoal-links subgoal) subgoal)
			  (if (not (subgoal-simplified subgoal)) (set-subgoal-unlinkable subgoal #t))
			  (set! on-the-fly-lemmas (cons (unsimplified-subgoal (clause-subgoals (subgoal-clause subgoal)))
							on-the-fly-lemmas))))
		    (set! nr-simplifications (+ nr-simplifications 1))))))
	(if success
	    (begin
	      (simplify-subgoals-on-the-fly (cdr subgoals) subgoal)
	      #t)
	    (simplify-subgoals-on-the-fly (cdr subgoals) subgoal)))
      #f))

(define (unsimplified-subgoal subgoals)
  (if (pair? subgoals)
      (if (subgoal-simplified (car subgoals))
	  (unsimplified-subgoal (cdr subgoals))
	  (car subgoals))
      #f))

(define (root-foldup subgoal)
  (let ((foldup-clause (new-clause)) (single-var (new-variable)) (foldup-input (new-input)))
    (if (and (not equality) local-vars)
	(set-variable-name single-var (list "X"))
	(set-variable-name single-var "X"))
    (set-clause-subgoals foldup-clause (list (foldup-subgoal-copy subgoal single-var)))
    (set-clause-variables foldup-clause (variables-in-subgoals (clause-subgoals foldup-clause) '()))
    (clause-pointers-in-subgoals (clause-subgoals foldup-clause) foldup-clause)
    (set-clause-unsimplified foldup-clause 1)
    (set-input-type foldup-input 'lemma)         
    (set-clause-input foldup-clause foldup-input)
    (set-clause-initial foldup-clause #f)
    (set-clause-relevant foldup-clause (clause-relevant (subgoal-clause subgoal)))
    (set-subgoal-next (car (clause-subgoals foldup-clause)) (subgoal-clause input-root))
    (set-clause-path (subgoal-clause input-root) (car (clause-subgoals foldup-clause)))
    (set-clause-depth foldup-clause (- (subgoal-depth input-root) 1))
    (set-input-number foldup-input (clause-depth foldup-clause))       
    (set! input-root (car (clause-subgoals foldup-clause)))
    (set! nr-foldups (+ nr-foldups 1))
    (set-input-name foldup-input (string-append foldup-prefix-string (integer->string nr-foldups)))         
    (insert-unit-in-indexes (car (clause-subgoals foldup-clause)))
    (if (and local-vars (subgoal-semi-local subgoal)) (unbind-vars))
    (if (match-subgoal-with-index subgoal 'simplify)
	(begin
	  (set-subgoal-simplified subgoal #t)
	  (set-clause-unsimplified (subgoal-clause subgoal) (- (clause-unsimplified (subgoal-clause subgoal)) 1))
	  (if (or dewey-numbers (> debug-level 4))
	      (begin
		(display "foldup lemma ")
		(display-subgoal (car (clause-subgoals foldup-clause)))
		(display " simplifies generating clause  ")
		(display-clause (subgoal-clause subgoal))
		(newline)))
	  (set! nr-simplifications (+ nr-simplifications 1))))
    (set-input-variant-index foldup-input (cons 0 '()))
    ;;(insert-clause-in-index foldup-clause)
    ;;(subgoal-simplifiable subgoal)
    (if (or dewey-numbers (> debug-level 4))
	(begin
	  (display "folding up subgoal ")
	  (display-subgoal subgoal)
	  (display "(")
	  (display (clause-depth foldup-clause))
	  (display ")")
	  (display " of depth ")
	  (display (subgoal-depth subgoal))
	  (display " to root as ")
	  (display-subgoal (car (clause-subgoals foldup-clause)))
	  (newline)))
    (if tptp-lemmas (print-tptp-lemma (car (clause-subgoals foldup-clause))))
    ;;(insert-subgoal-in-index (car (clause-subgoals foldup-clause)) 'path)
    (if (equation? (car (clause-subgoals foldup-clause))) (add-unit-lemma (car (clause-subgoals foldup-clause))))
    (if re-evaluation (recompute-link-lists))
    (let (result)
      (if (and equality (eq? (subgoal-pred (car (clause-subgoals foldup-clause))) equality-pred))
	  (set! result (insert-eq-lemma-links (car (clause-subgoals foldup-clause)) (subgoal-predecessor subgoal)))
	  ;;(set! result (insert-lemma-links (car (clause-subgoals foldup-clause)) (subgoal-predecessor subgoal))))
	  (set! result (generate-links-for-lemma (car (clause-subgoals foldup-clause)))))
      (if (and (or (not result) (>= result 0)) (pair? on-the-fly-lemmas) (subgoal? (car on-the-fly-lemmas)))
	  (if result
	      (let ((on-the-fly-result (make-on-the-fly-lemma subgoal)))
		(if on-the-fly-result (set! result (min result on-the-fly-result))))
	      (set! result (make-on-the-fly-lemma subgoal))))
      result)))

(define (foldup-subgoal-copy subgoal single-var)
  (let ((sg (new-subgoal)))
    (if (eq? '+ (subgoal-sign subgoal))
	(set-subgoal-sign sg '-)
	(set-subgoal-sign sg '+))
    (set-subgoal-local sg #t)
    (set-subgoal-pred sg (subgoal-pred subgoal))
    (trail-stop)
    (set-subgoal-terms sg (foldup-termlist-copy sg (subgoal-terms subgoal) single-var))
    (unrename-vars)
    (set-subgoal-selected sg #f)
    (set-subgoal-ordered-eq sg (subgoal-ordered-eq subgoal))
    (set-subgoal-links sg '())
    sg))

(define (foldup-termlist-copy subgoal terms single-var)
  (if (null? terms)
      '()
      (cons (foldup-term-copy subgoal (ref (car terms)) single-var) (foldup-termlist-copy subgoal (cdr terms) single-var))))

(define (foldup-term-copy subgoal term single-var)
  (if (variable term)
      (if (and local-vars (local-variable term))
	  (if (variable-renaming term)
	      (variable-renaming term)
	      (let ((renaming (new-variable)))
		(push-trail term)
		(set-variable-renaming term renaming)
		(set-variable-name renaming (variable-name term))
		;; use of variable-renaming field of new var to allow checking for local-sg's
		renaming))
	  (begin
	    (if local-vars (set-subgoal-local subgoal #f))
	    (if (not (eq? selected-constant dollar-constant))
		selected-constant
		single-var)))
      (if (constant term)
	  (if (eq? term dollar-constant)
	      (begin
		(if local-vars (set-subgoal-local subgoal #f))
		single-var)
	      term)
	  (cons (car term) (foldup-termlist-copy subgoal (cdr term) single-var)))))

(define (make-top-down-lemma subgoal)
  (let ((lemma-clause (new-clause)) (lemma-input (new-input)))
    (set-clause-subgoals lemma-clause (list (lemma-subgoal-copy subgoal)))
    (set-clause-variables lemma-clause (variables-in-subgoals (clause-subgoals lemma-clause) '()))
    (clause-pointers-in-subgoals (clause-subgoals lemma-clause) lemma-clause)
    (set-clause-unsimplified lemma-clause 1)
    (set-input-type lemma-input 'lemma)         
    (set-clause-input lemma-clause lemma-input)
    (set-clause-initial lemma-clause #f)
    (set-clause-relevant lemma-clause (clause-relevant (subgoal-clause subgoal)))
    (set-subgoal-next (car (clause-subgoals lemma-clause)) (subgoal-clause input-root))
    (set-clause-path (subgoal-clause input-root) (car (clause-subgoals lemma-clause)))
    (set-clause-depth lemma-clause (- (subgoal-depth input-root) 1))
    (set-input-number lemma-input (clause-depth lemma-clause))       
    (set! input-root (car (clause-subgoals lemma-clause)))
    (set! nr-top-down-lemmas (+ nr-top-down-lemmas 1))
    (set-input-variant-index lemma-input (cons 0 '()))
    (set-input-name lemma-input (string-append simplification-prefix-string (integer->string nr-top-down-lemmas)))         
    (insert-unit-in-indexes (car (clause-subgoals lemma-clause)))
    ;;(insert-clause-in-index lemma-clause)
    (if (or dewey-numbers (> debug-level 4))
	(begin
	  (display "making simplification lemma ")
	  (display-subgoal subgoal)
	  (display "(")
	  (display (clause-depth lemma-clause))
	  (display ")")
	  (display " on depth ")
	  (display (subgoal-depth subgoal))
	  (newline)))
    (if tptp-lemmas (print-tptp-lemma (car (clause-subgoals lemma-clause))))
    ;;(insert-subgoal-in-index (car (clause-subgoals lemma-clause)) 'path)
    (if (equation? (car (clause-subgoals lemma-clause))) (add-unit-lemma (car (clause-subgoals lemma-clause))))
    (if re-evaluation (recompute-link-lists))
    (let (result)
      (if (and equality (eq? (subgoal-pred (car (clause-subgoals lemma-clause))) equality-pred))
	  (set! result (insert-eq-lemma-links (car (clause-subgoals lemma-clause)) (subgoal-predecessor subgoal)))
	  ;;(set! result (insert-lemma-links (car (clause-subgoals lemma-clause)) (subgoal-predecessor subgoal))))
	  (set! result (generate-links-for-lemma (car (clause-subgoals lemma-clause)))))
      (if (and (or (not result) (>= result 0)) (pair? on-the-fly-lemmas) (subgoal? (car on-the-fly-lemmas)))
	  (if result
	      (let ((on-the-fly-result (make-on-the-fly-lemma subgoal)))
		(if on-the-fly-result (set! result (min result on-the-fly-result))))
	      (set! result (make-on-the-fly-lemma subgoal))))
      result)))

(define (lemma-subgoal-copy subgoal)
  (let ((sg (new-subgoal)))
    (set-subgoal-sign sg (subgoal-sign subgoal))
    (set-subgoal-pred sg (subgoal-pred subgoal))
    (if local-vars (check-for-local-vars subgoal))
    (trail-stop)
    (set-subgoal-terms sg (renamed-termlist (subgoal-terms subgoal) 'copy))
    (unrename-vars)
    (set-subgoal-selected sg #f)
    (set-subgoal-simplified sg #f)
    (set-subgoal-links sg '())
    (set-subgoal-ordered-eq sg (subgoal-ordered-eq subgoal))
    sg))

(define (simplified-unit subgoal)
  (if (subgoal-simplified subgoal)
      #f
      (if (and (= (clause-unsimplified (subgoal-clause subgoal)) 1) (or unit-subsume (not (unit-clause (subgoal-clause subgoal)))))
      ;;(if (and (= (clause-unsimplified (subgoal-clause subgoal)) 1) (not (unit-clause (subgoal-clause subgoal))))
	  #t
	  #f)))

(define (foldup-subgoal subgoal rel-vector)
;   (if (and local-vars (not equality) (subgoal-semi-local subgoal))
;       (begin
; 	(display "trying to unify potential fold up subgoal ")
; 	(display-subgoal subgoal)
; 	(display " with ")
; 	(display (length (subgoal-local-instances subgoal)))
; 	(display " different instantiations: ")
; 	(display-local-instances (subgoal-local-instances subgoal))
; 	(newline)))
  (if (or (not local-vars) (not (subgoal-semi-local subgoal)) (not (trail-stop))
	  (unify-local-instances (subgoal-terms subgoal) (subgoal-local-instances subgoal)))
      (let (first-level)
	(set! first-level (foldup-depth rel-vector (subgoal-depth subgoal)))
	(if (< first-level 0)
	    (let ((result (root-foldup subgoal)))
	      ;;(if (and local-vars (subgoal-semi-local subgoal)) (unbind-vars))
	      result)
	    (let ((result #f))
	      (if context-lemmas (set! result (context-foldup subgoal (+ first-level 1) rel-vector)))
	      (if (and local-vars (subgoal-semi-local subgoal)) (unbind-vars))
	      result)))
      (begin
	;;(unbind-vars)
	(if dewey-numbers
	    (begin
	      (display "not folding up subgoal ")
	      (display-subgoal subgoal)
	      (display " with ")
	      (display (length (subgoal-local-instances subgoal)))
	      (display " different instantiations: ")
	      (display-local-instances (subgoal-local-instances subgoal))
	      (newline)))
	#f)))

(define (context-foldup subgoal depth rel-vector)
  (let (sg path-clause new-vector (foldup-clause (new-clause)) (foldup-input (new-input)))
    (set! sg (foldup-subgoal-copy subgoal dollar-constant))
    (set-subgoal-ground sg (subgoal-ground subgoal))
    (set-subgoal-local sg (or (subgoal-ground subgoal) (subgoal-local subgoal)))
    (set-clause-subgoals foldup-clause (list sg))
    (set-clause-variables foldup-clause (variables-in-subgoals (clause-subgoals foldup-clause) '()))
    (clause-pointers-in-subgoals (clause-subgoals foldup-clause) foldup-clause)
    (set-clause-unsimplified foldup-clause 1)
    (set-clause-input foldup-clause foldup-input)
    (set-clause-initial foldup-clause #f)
    (set-clause-relevant foldup-clause (clause-relevant (subgoal-clause subgoal)))
    (set-clause-depth foldup-clause depth)
    (set-input-type foldup-input 'clemma)         
    (set-input-variant-index foldup-input (cons 0 '()))
    (set-input-number foldup-input 'clemma)       
    (set! new-vector (copy-up-to rel-vector depth))
    (set-lemma-rel-vector sg new-vector)
    (insert-subgoal-in-index sg 'context)
    (if dewey-numbers
	(begin
	  (if (and context-units local-vars (not equality) (subgoal-local sg))
	      (display "folding up local context subgoal ")
	      (display "folding up context subgoal "))
	  (display-subgoal subgoal)
	  (display " of depth ")
	  (display (subgoal-depth subgoal))
	  (display " to depth ")
	  (display depth)
	  (display " as ")
	  (display-subgoal sg)
	  (newline)))
    (set! path-clause (path-vector-ref depth))
    (if (clause-foldup-lits path-clause)
	(set-clause-foldup-lits path-clause (cons sg (clause-foldup-lits path-clause)))
	(set-clause-foldup-lits path-clause (cons sg '())))
    (set! nr-context-foldups (+ nr-context-foldups 1))
    (if (and (or (subgoal-ground sg) (subgoal-local sg)) context-units (not equality))
	(generate-links-for-context-lemma sg))))

(define (delete-clause-foldup-lits clause)
  (if (pair? (clause-foldup-lits clause))
      (begin
; 	(display "deleting context lemmas of clause ")
; 	(display-clause clause)
; 	(display " : ")
; 	(display-subgoal-list (clause-foldup-lits clause))
; 	(newline)
	(delete-foldup-lits (clause-foldup-lits clause))
	(set-clause-foldup-lits clause #f))))

(define (delete-foldup-lits subgoals)
  (if (pair? subgoals)
      (begin
; 	(display "deleting context lemmas ")
; 	(display-subgoal (car subgoals))
; 	(if (pair? (subgoal-links (car subgoals)))
; 	    (begin
; 	      (display " disabling ")
; 	      (display (length (subgoal-links (car subgoals))))
; 	      (display " links")))
; 	(newline)	      
	(mark-links-deletable (subgoal-links (car subgoals)) (car subgoals))
	(delete-subgoal-from-index (car subgoals) 'context)
	(set-lemma-rel-vector (car subgoals) #f)
	;;(vector-fill! (car subgoals) #f)
	(set-subgoal-clause (car subgoals) (subgoal-depth (car subgoals)))
	(set-car! subgoals #f)
	;;(set-car! subgoals 'foo)
	(delete-foldup-lits (cdr subgoals)))))

(define (make-on-the-fly-lemma subgoal)
  (let (lemma (lemma-clause (new-clause)) (lemma-input (new-input)))
    (set! lemma (car on-the-fly-lemmas))
    (set! on-the-fly-lemmas (cdr on-the-fly-lemmas))
    (set-clause-subgoals lemma-clause (list (lemma-subgoal-copy lemma)))
    (set-clause-variables lemma-clause (variables-in-subgoals (clause-subgoals lemma-clause) '()))
    (clause-pointers-in-subgoals (clause-subgoals lemma-clause) lemma-clause)
    (set-clause-unsimplified lemma-clause 1)
    (set-input-type lemma-input 'lemma)         
    (set-clause-input lemma-clause lemma-input)
    (set-clause-initial lemma-clause #f)
    (set-clause-relevant lemma-clause (clause-relevant (subgoal-clause lemma)))
    (set-subgoal-ground (car (clause-subgoals lemma-clause)) (subgoal-ground subgoal))
    (set-subgoal-local (car (clause-subgoals lemma-clause)) #t)
    (set-subgoal-next (car (clause-subgoals lemma-clause)) (subgoal-clause input-root))
    (set-clause-path (subgoal-clause input-root) (car (clause-subgoals lemma-clause)))
    (set-clause-depth lemma-clause (- (subgoal-depth input-root) 1))
    (set-input-number lemma-input (clause-depth lemma-clause))       
    (set! input-root (car (clause-subgoals lemma-clause)))
    (set! nr-top-down-lemmas (+ nr-top-down-lemmas 1))
    (set! nr-on-the-fly-lemmas (+ nr-on-the-fly-lemmas 1))
    (set-input-name lemma-input (string-append simplification-prefix-string (integer->string nr-top-down-lemmas)))         
    (insert-unit-in-indexes (car (clause-subgoals lemma-clause)))
    (set-input-variant-index lemma-input (cons 0 '()))
    ;;(insert-clause-in-index lemma-clause)
    (if (or dewey-numbers (> debug-level 4))
	(begin
	  (display "making on-the-fly lemma ")
	  (display-subgoal lemma)
	  (display "(")
	  (display (clause-depth lemma-clause))
	  (display ")")
	  (display " on depth ")
	  (display (subgoal-depth lemma))
	  (newline)))
    (if tptp-lemmas (print-tptp-lemma (car (clause-subgoals lemma-clause))))
    ;;(insert-subgoal-in-index (car (clause-subgoals lemma-clause)) 'path)
    (if (equation? (car (clause-subgoals lemma-clause))) (add-unit-lemma (car (clause-subgoals lemma-clause))))
    (let (result)
      (if (and equality (eq? (subgoal-pred (car (clause-subgoals lemma-clause))) equality-pred))
	  (set! result (insert-eq-lemma-links (car (clause-subgoals lemma-clause)) (subgoal-predecessor subgoal)))
	  ;;(set! result (insert-lemma-links (car (clause-subgoals lemma-clause)) (subgoal-predecessor subgoal))))
	  (set! result (generate-links-for-lemma (car (clause-subgoals lemma-clause)))))
      (if (and (or (not result) (>= result 0)) (pair? on-the-fly-lemmas) (subgoal? (car on-the-fly-lemmas)))
	  (if result
	      (let ((on-the-fly-result (make-on-the-fly-lemma subgoal)))
		(if on-the-fly-result (set! result (min result on-the-fly-result))))
	      (set! result (make-on-the-fly-lemma subgoal))))
      result)))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Finding duplicate subgoals
;; ------------------------------------------------------------

(define (subgoal-in-list subgoal subgoals)
  (if (pair? subgoals)
      (if (and (eq? (subgoal-sign subgoal) (subgoal-sign (car subgoals)))
	       (eq? (subgoal-pred subgoal) (subgoal-pred (car subgoals))))
	  (if (and (not (eq-subgoal? subgoal))
		   (ref-equals (subgoal-terms subgoal) (subgoal-terms (car subgoals))))
	      (car subgoals)
	      ;; symmetric test for ref-equality for eq-literals
	      (if (and (eq-subgoal? subgoal)
		       (or (and (ref-equal (ref (car (subgoal-terms subgoal))) (ref (car (subgoal-terms (car subgoals)))))
				(ref-equal (ref (cadr (subgoal-terms subgoal))) (ref (cadr (subgoal-terms (car subgoals))))))
			   (and (ref-equal (ref (car (subgoal-terms subgoal))) (ref (cadr (subgoal-terms (car subgoals)))))
				(ref-equal (ref (cadr (subgoal-terms subgoal))) (ref (car (subgoal-terms (car subgoals))))))))
		  (car subgoals)
		  (subgoal-in-list subgoal (cdr subgoals))))
	      (subgoal-in-list subgoal (cdr subgoals)))
      #f))

;;;*************************************************************************
;; ------------------------------------------------------------
;; unifying local instances
;; ------------------------------------------------------------

(define (unify-local-instances subgoal-termlist instances)
  (if (pair? instances)
      (if (gr-unify-instances subgoal-termlist (cdar instances))
	  (unify-local-instances subgoal-termlist (cdr instances))
	  #f)
      #t))

(define (unify-local-instance-of-variables subgoal-termlist instance-termlist)
  (if (pair? instance-termlist)
      (if (gr-unify-instances subgoal-termlist (car instance-termlist))
	  (unify-local-instance-of-variables subgoal-termlist (cdr instance-termlist)))
      #t))

(define (gr-unify-instances k l)
;; gr-unifies termlists k and l.
 (if (null? k)
     #t
     (if (gr-unify-instance-terms (ref (car k)) (ref (car l)))
	 (gr-unify-instances (cdr k) (cdr l))
	 #f)))

(define (gr-unify-instance-terms s t)          
  (if (shared-variable s) (bind s dollar-constant))
  (if (shared-variable t) (bind t dollar-constant))
  (if (eq? s t)
      #t
      (if (variable s)
	  (if (variable t)
	      (bind s t)
	      (if (occurs-list s (cdr t))
		  (unbind-vars)
		  (bind s t)))
	  (if (variable t)
	      (if (occurs-list t (cdr s))
		  (unbind-vars)
		  (bind t s))
	      (if (eq? (car s) (car t))
		  (gr-unify-instances (cdr s) (cdr t))
		  (unbind-vars))))))
