(module signals
	(include "dctp.sch")
	(export sigint sigxcpu sigbus sigsegv (xcpu arg) (interrupt arg) (bus-error arg) (seg-fault arg) (set-signal-handlers))
	(import (print-statistics display "display.scm") (cputime "cputime.scm") (start globals "globals.scm") 
		(end globals "globals.scm") (elapsed globals "globals.scm")))

;;;*************************************************************************
;; ---------------
;; Signal handling
;; ---------------

(define sigint 2)
(define sigxcpu 30)
(define sigbus 10)
(define sigsegv 11)

(define (set-signal-handlers)
  (signal sigxcpu xcpu)
  (signal sigint interrupt)
  (signal sigbus bus-error)
  (signal sigsegv seg-fault))

(define (xcpu arg)
  (set! end (getcputime))
  (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
  (print "Exit status is FAIL (CPU limit exceeded after " elapsed " seconds)")
  (print-statistics)
  (exit 0))

(define (interrupt arg)
  (set! end (getcputime))
  (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
  (print "Exit status is ERROR (external interrupt after " elapsed " seconds)")
  (print-statistics)
  (exit arg))

(define (bus-error arg)
  (set! end (getcputime))
  (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
  (print "Exit status is ERROR (bus error after " elapsed " seconds)")
  (print-statistics)
  (exit arg))

(define (seg-fault arg)
  (set! end (getcputime))
  (set! elapsed (/fl (exact->inexact (- end start)) 100.0))
  (print "Exit status is ERROR (segmentation violation after " elapsed " seconds)")
  (print-statistics)
  (exit arg))

