(module normal-form
	(include "dctp.sch")
	(export (renamed-subgoals subgoals) (non-equivalent-termlists k l) (renamed-termlist terms copy-local-bindings)
		(varlist-renamed? vars all-vars renamings) (unsimplifiable-subgoals subgoals)
		(non-equivalent-termlists-1 k l) (renamed-term term copy-local-bindings) (check-termlist-for-refs terms))
	(import (globals "globals.scm") (unification "unification.scm") (display "display.scm") (kbo "kbo.scm")
		(index "index.scm") (subgoal-in-list pruning "pruning.scm") (current-subgoal-weight select "select.scm")
		(rewrite-subgoal rewrite "rewrite.scm") (insert-literal parser "parser.scm")))


;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define current-class 1)
(define current-renamed-sg #f)

;;;*************************************************************************
;; ------------------------------------------------------------
;; Normal form functions
;; ------------------------------------------------------------

(define (non-equivalent-termlists k l)
  (let (result)
    (set! var-counter 0)
    (trail-stop)
    (set! result (non-equivalent-termlists-1 k l))
    (if (not result) (unbind-vars-and-undo-var-counter))
; 	(begin
; 	  (display "equivalent: ")
; 	  (display-ref-list k)
; 	  (display " and ")
; 	  (display-ref-list l)
; 	  (newline)
; 	  (unbind-vars-and-undo-var-counter)
; 	  (display "unbound:    ")
; 	  (display-subgoal-terms k)
; 	  (display " and ")
; 	  (display-subgoal-terms l)
; 	  (newline)))
    result))

(define (non-equivalent-termlists-1 k l)
  (if (pair? k)
      (let ((result (non-equivalent-terms (ref (car k)) (ref (car l)))))
	(if result
	    result
	    (non-equivalent-termlists-1 (cdr k) (cdr l))))
      #f))

(define (non-equivalent-terms s t)
  (if (eq? s t)
      #f
      (if (and (variable s) (variable t))
	  (begin
	    (bind s var-counter)
	    (bind t var-counter)
	    (set! var-counter (+ var-counter 2))
	    #f)
	  (if (and (non-variable s) (non-variable t) (eq? (car s) (car t)))
	      (non-equivalent-termlists-1 (cdr s) (cdr t))
	      (begin
		(unbind-vars-and-undo-var-counter)
		#t)))))

(define (non-equivalent-subgoals a b)
  (if (eq? (subgoal-sign a) (subgoal-sign b))
      (if (eq? (subgoal-pred a) (subgoal-pred b))
	  (non-equivalent-termlists (subgoal-terms a) (subgoal-terms b))
	  #t)
      #t))

(define (renamed-subgoals subgoals)
  (if (pair? subgoals)
      (let ((sg (new-subgoal)) (new-subgoals (renamed-subgoals (cdr subgoals))) same-sg (rewritten #f))
	(set! same-sg (subgoal-in-list (car subgoals) (cdr subgoals)))
	(if (and contract same-sg) 
	    (begin
	      (if (> debug-level 4)
		  (begin
		    (display "deleting multiple occurrence of subgoal ")
		    (display-subgoal (car subgoals))
		    (newline)))
	      (set! nr-contractions (+ nr-contractions 1))
	      (set-subgoal-misc (car subgoals) (subgoal-misc same-sg))
	      new-subgoals)
	    (begin
	      ;; use of module-global variable to enable detection of non-local variables
	      (set! current-renamed-sg sg)
	      (set-subgoal-ground sg #t)
	      (if local-vars
		  (set-subgoal-local sg #t)
		  (set-subgoal-local sg #f))
	      (set-subgoal-sign sg (subgoal-sign (car subgoals)))
	      (set-subgoal-pred sg (subgoal-pred (car subgoals)))
	      (set-subgoal-split-lit sg (subgoal-split-lit (car subgoals)))
	      (set-subgoal-expansion-lit sg (subgoal-expansion-lit (car subgoals)))
	      (set-subgoal-unlinkable sg (subgoal-unlinkable (car subgoals)))
	      (set-subgoal-selected sg #f)
	      (set-subgoal-simplified sg #f)
	      (set-subgoal-links sg '())
	      (set-subgoal-terms sg (renamed-termlist (subgoal-terms (car subgoals)) 'nocopy))
	      (set! current-renamed-sg #f)
              ;;(if (and equality ordered-eq (or full-rewrite (and eq-rewrite (eq-subgoal? sg))))
	      ;;(set! rewritten (rewrite-subgoal sg)))
	      (if (and (not rewritten) (integer? (subgoal-misc (car subgoals))))
		  (set-subgoal-weight sg (subgoal-misc (car subgoals)))
		  (set-subgoal-weight sg (current-subgoal-weight sg)))
	      (set-subgoal-misc (car subgoals) sg)
	      (set-subgoal-ordered-eq sg (subgoal-ordered-eq (car subgoals)))
; 	      (if (and (subgoal-expansion-lit sg) (kbo> (cadr (subgoal-terms sg)) (car (subgoal-terms sg))))
; 		  (begin
; 		    (display "instantiation of expansion literal produced order violating")
; 		    (if (subgoal-ordered-eq sg)
; 			(display " ordered"))
; 		    (display " subgoal ")
; 		    (display-subgoal sg)
; 		    (newline)))
	      (if (or (and ordered-eq (eq-subgoal? sg) (or rewritten (not (subgoal-ordered-eq sg))))
		      (and (or (not ordered-eq) disagreement) (disequation? sg)))
		  (check-orientation sg))
	      ;;(cons sg new-subgoals))))
	      (insert-literal sg new-subgoals))))
      '()))

(define (renamed-termlist terms copy-local-bindings)
  (if (pair? terms)
      (if (or (eq? 'copy copy-local-bindings) (shared-variable (car terms)))
	  (cons (renamed-term (ref (car terms)) 'copy) (renamed-termlist (cdr terms) copy-local-bindings))
	  (cons (renamed-term (car terms) copy-local-bindings) (renamed-termlist (cdr terms) copy-local-bindings)))
      '()))

(define (renamed-term term copy-local-bindings)
  (if (variable term)
      (begin
	(if current-renamed-sg (set-subgoal-ground current-renamed-sg #f))
	(if (variable-renaming term)
	    (begin
	      (if (and current-renamed-sg (not (eq? (variable-renaming (variable-renaming term)) current-renamed-sg)))
		  (begin
		    (if (and local-vars (local-variable (variable-renaming term)))
			(set-variable-name (variable-renaming term) (car (variable-name (variable-renaming term)))))
		    (set-subgoal-local (variable-renaming (variable-renaming term)) #f)
		    (set-subgoal-local current-renamed-sg #f)))
	      (variable-renaming term))
	    (let ((renaming (new-variable)))
	      (set-variable-renaming term renaming)
	      (push-trail term)
	      ;; if current-renamed-sg is set, we are in the standard subgoal copy mode. all variables
	      ;; initially are considered local.
	      ;; if current-renamed-sg is not set, we produce a one-on-one copy of term local to local, shared to shared
	      (if (and local-vars current-renamed-sg (not (local-variable term)))
		  (set-variable-name renaming (cons (variable-name term) '()))
		  (set-variable-name renaming (variable-name term)))
	      ;; use of variable-renaming field of new var to allow checking for local-sg's
	      (if current-renamed-sg
		  (begin
		    (push-trail renaming)
		    (set-variable-renaming renaming current-renamed-sg)))
	      ;;(if (bound-variable term)
		  ;;(set-variable-binding renaming (renamed-term (variable-binding term))))
	      renaming)))
      (if (constant term)
	  term
	  (begin
	    (if (= (functor-arity (car term)) 0) (print "copying constant " term))
	    (cons (car term) (renamed-termlist (cdr term) copy-local-bindings))))))

; (define (semi-ref term)
; ;; if term is a <bound_variable> bound to another variable, its binding is returned, 
; ;; otherwise term is returned
;  (if (or (non-variable term) (free-variable term) (domain-element term)
; 	 (and (variable term) (non-variable (variable-binding term))))
;      term
;      (semi-ref (car term))))

(define (semi-ref term)
  (if (and (bound-variable term) (variable (variable-binding term)))
      (semi-ref (car term))
      term))

(define (unsimplifiable-subgoals subgoals)
  (if (pair? subgoals)
      (if (and simplify
	       (or (subgoal-simplified (car subgoals)) (match-subgoal-with-index (car subgoals))))
	  (begin
	    (if (and (> debug-level 4) (not (subgoal-simplified (car subgoals))))
		(begin
		  (display "simplified subgoal instance ")
		  (display-subgoal (car subgoals))
		  (newline)))
	    (set-subgoal-misc (car subgoals) #f)
	    (unsimplifiable-subgoals (cdr subgoals)))
	  (cons (car subgoals) (unsimplifiable-subgoals (cdr subgoals))))
      '()))     


;;;*************************************************************************
;; ------------------------------------------------------------
;; New check for variable renamings
;; ------------------------------------------------------------

(define (varlist-renamed? vars all-vars renamings)
  (if (pair? vars)
      (if (free-variable (car vars)) ;; if the current variable has not been bound at all
	  (varlist-renamed? (cdr vars) all-vars renamings)
	  (if (and (variable (ref (car vars))) ;; if the current variable has been bound to another variable
		   (not (memq (ref (car vars)) all-vars))  ;; that is not one from the var's own clause
		   (not (memq (ref (car vars)) renamings)));; or has been used to rename another var from the clause
	      (varlist-renamed? (cdr vars) all-vars (cons (ref (car vars)) renamings))
	      #f))
      (begin
	(set! total-renamings (+ total-renamings 1))
	#t)))

(define (check-termlist-for-refs terms)
  (if (pair? terms)
      (begin
	(check-term-for-refs (car terms))
	(check-termlist-for-refs (cdr terms)))))

(define (check-term-for-refs term)
  (if (variable term)
      (if (bound-variable term)
	  (begin
	    (display "term contains variable ")
	    (display (variable-name term))
	    (display " bound to ")
	    (display-ref (ref term))
	    (newline)))
      (check-termlist-for-refs (cdr term))))
