;; -------------------------------------------------------
;; PARSER FOR TPTP INPUT FILES
;; -------------------------------------------------------
(module parser
	(include "dctp.sch")
	(export tptp-parser tptp-scanner nr-initial-predicates
		(variables-in-cleared-subgoals subgoals vars)
		(insert-new-clause-in-clauselist clause clauses) (insert-literal literal lits))
	(import (globals "globals.scm")	(index  "index.scm") (preprocess "preprocess.scm") (select "select.scm")
		(sign-predsym< subsumption "subsumption.scm") (display "display.scm") (normal-form "normal-form.scm")
		(display-ref display "display.scm") (non-tautological pruning "pruning.scm") (unrename-vars dctp "dctp.scm")
		(free-vars-in-term dctp "dctp.scm") (eliminate-var-disequations dctp  "dctp.scm") (kbo "kbo.scm")
		(unification "unification.scm") (initial-path "initial-path.scm")
		(compute-current-weights dctp "dctp.scm") (variables-in-subgoals dctp "dctp.scm") (clause-subsumed dctp "dctp.scm")
		(subsumption "subsumption.scm")))

(define predicate-list '())
(define functor-list '())
(define constant-list '())
(define current-clause-variables '())
(define current-clause-nr 0)
(define current-clause '())
(define current-input '())
(define current-variable '())
(define current-literal '())
(define current-predicate '())
(define current-pred-nr 1)
(define nr-initial-predicates 0)
(define current-functor '())
(define current-functor-nr 1)
(define current-positive-subgoals 0)
(define current-negative-subgoals 0)
(define max-clause-variables 0)        

(define tptp-scanner
  (regular-grammar
   ((lowercase (in (#\a #\z)))
    (uppercase (in (#\A #\Z)))
    (numeric (in (#\0 #\9)))
    (special (in #\+ #\- #\_))
    (underscore (in #\_))
    (alphanum (or lowercase uppercase numeric special)))
   ("input_clause" (display-for-debug) 'input_clause)
   ("model_clause" (display-for-debug) 'model_clause)
   ("domain_definition" (display-for-debug) 'domain_definition)
   ("predicate_map" (display-for-debug) 'predicate_map)
   ("functor_map" (display-for-debug) 'functor_map)
   ("equal" (display-for-debug)
	      (cons 'symbol (the-symbol)))
   ((: uppercase (* alphanum)) (cons 'variable (string->symbol (the-string))))
   ((: (or lowercase numeric underscore) (* alphanum)) (cons 'symbol (string->symbol (the-string))))
   ("++" (display-for-debug) 'pos_sign)
   ("--" (display-for-debug) 'neg_sign)
   (#\( (display-for-debug) 'openpar)
   (#\) (display-for-debug) 'closepar)
   (#\[ (display-for-debug) 'openbracket)
   (#\] (display-for-debug) 'closebracket)
   (#\, (display-for-debug) 'comma)
   (#\. (display-for-debug) 'dot)

   ((eol (: #\% (* all))) (display-for-debug) (ignore))

   ((+ (in #\space #\tab #\newline)) (display-for-debug) (ignore))))

(define (variable-exists? variable variables)
  (if (null? variables)
      #f
      (if (eq? variable (variable-name (car variables)))
	  #t
	  (variable-exists? variable (cdr variables)))))

(define (get-variable variable variables)
  (if (null? variables)
      #f
      (if (eq? variable (variable-name (car variables)))
	  (car variables)
	  (get-variable variable (cdr variables)))))

(define (increment-occurences variable variables occurrences)
  (if (null? variables)
      #f
      (if (eq? variable (variable-name (car variables)))
	  (set-car! occurrences (+ (car occurrences) 1))
	  (increment-occurences variable (cdr variables) (cdr occurrences)))))

(define (set-clause-of-litlist litlist copy)
  (if (null? litlist)
      #t
      (begin
	(set-subgoal-clause (car litlist) copy)
	(set-clause-of-litlist (cdr litlist) copy))))

(define (compute-clause-complexity clause)
  (- (subgoals-complexity (clause-subgoals clause)) (length (clause-variables clause))))

(define (subgoals-complexity subgoals)
  (if (null? subgoals)
      0
      (+ (subgoal-complexity (car subgoals)) (subgoals-complexity (cdr subgoals)))))

(define (term-complexity term depth)
  (if (> depth input-term-depth)
      (set! input-term-depth depth))
  (if (or (variable term) (constant term)) 
      1
      (+ 1 (termlist-complexity (cdr term) (+ depth 1)))))

(define (termlist-complexity terms depth)
  (if (null? terms)
      0
      (+ (term-complexity (car terms) depth) (termlist-complexity (cdr terms) depth))))

(define (subgoal-complexity subgoal)
  (let ((terms (subgoal-terms subgoal)))
    (termlist-complexity terms 1)))

(define tptp-parser
  (lalr-grammar
   (variable symbol input_clause domain_definition predicate_map functor_map pos_sign neg_sign openpar closepar 
	     openbracket closebracket comma dot) 
   (formula ((clauselist)
	     (compute-kbo-precedences functor-list)
	     (set-clause-depths clauselist)
	     (set! global-index-bindings (make-vector (+ nr-variables 1) #f))
	     (set! nr-initial-predicates (- current-pred-nr 1))
	     (set-formula-clauses formula clauselist)))
; 	     (if equality 
; 		 (begin 
; 		   (add-reflexivity-axiom)
; 		   (if (or dewey-numbers (> debug-level 4))
; 		       (begin
; 			 (print "reflexivity axiom added to indexes")))))))
   (clauselist
    (() '())
    ((clauselist clause) 
     (if equality 
	 (begin
	   (set-clause-subgoals clause (eliminate-var-disequations (clause-subgoals clause)))
	   (if (null? (clause-subgoals clause))
	       (set-clause-subgoals clause (list falsum)))
	   (set-clause-unsimplified clause (length (clause-subgoals clause)))
	   (set-clause-variables clause (variables-in-cleared-subgoals (clause-subgoals clause) '()))))
     (if pre-factoring
	 (insert-clause-and-factors-in-clauselist clause clauselist)
	 (if (and (or tautologies (non-tautological clause)) (or (not subsumption) (not (clause-subsumed clause))))
	 ;;(if (or tautologies (non-tautological clause))
	     (begin
	       (if (and preprocess (unit-clause clause) (not (eq? falsum (car (clause-subgoals clause)))))
		   (insert-unit-in-indexes (car (clause-subgoals clause))))
	       (insert-subgoals-in-path-index (clause-subgoals clause))
	       (insert-new-clause-in-clauselist clause clauselist))
	     (begin
	       (if dewey-numbers
		   (begin 
		     (display "deleted tautological or subsumed input ")
		     (display-clause clause)
		     (newline)))
	       clauselist)))))
   (clause
    ((input_clause openpar clausename comma type comma literals closepar dot)
     (set! current-input (new-input))
     (set! current-clause (new-clause))
     (set-clause-foldup-lits current-clause '())
     (set-clause-input current-clause current-input)
     (set-input-name current-input clausename)
     (set-input-type current-input type)
     (if (and (not norel) (eq? type 'conjecture))
	 (set-clause-relevant current-clause #t))
     (set! current-clause-nr (+ current-clause-nr 1))
     (set-input-number current-input current-clause-nr)
     (set-clause-subgoals current-clause literals)
     (set-subgoal-weights literals)
     (set-clause-unsimplified current-clause (length literals))
     (set-clause-of-litlist literals current-clause)
     (set-input-variant-index current-input (cons 0 '()))
     (set-clause-variables current-clause current-clause-variables)
     (if (> (length current-clause-variables) max-clause-variables)
	 (set! max-clause-variables (length current-clause-variables)))
     (set-clause-initial current-clause #t)
     (set! current-clause-variables '())
     (set! current-positive-subgoals 0) (set! current-negative-subgoals 0)
     current-clause))
   (literals
    ((openbracket litlist closebracket) litlist))
   (litlist
    ((litlist comma literal) (insert-literal literal litlist))
    ((literal) (list literal)))
   (literal
    ((pos_sign atom)
     (set! current-literal (new-subgoal))
     (set-subgoal-sign current-literal '+)
     (set-subgoal-pred current-literal (car atom))
     (set-subgoal-terms current-literal (cdr atom))
     (set-subgoal-links current-literal '())
     (set-subgoal-selected current-literal #f)
     (set-subgoal-next current-literal #f)
     (set! current-positive-subgoals (+ current-positive-subgoals 1))
     current-literal)
    ((neg_sign atom)
     (set! current-literal (new-subgoal))
     (set-subgoal-sign current-literal '-)
     (set-subgoal-pred current-literal (car atom))
     (set-subgoal-terms current-literal (cdr atom))
     (set-subgoal-links current-literal '())
     (set-subgoal-selected current-literal #f)
     (set-subgoal-next current-literal #f)
     (set! current-negative-subgoals (+ current-negative-subgoals 1))
     current-literal))
   (atom
    ((symbol) (set! nr-propositionals (+ nr-propositionals 1)) 
	      (if (not (get-predsym symbol 0 predicate-list))
		  (begin
		    (set! current-predicate (new-predsym))
		    (set-predsym-pos-context-index current-predicate (cons 0 '()))
		    (set-predsym-neg-context-index current-predicate (cons 0 '()))
		    (set-predsym-pos-gr-index current-predicate (cons 0 '()))
		    (set-predsym-neg-gr-index current-predicate (cons 0 '()))
		    (set-predsym-pos-unit-index current-predicate (cons 0 '()))
		    (set-predsym-neg-unit-index current-predicate (cons 0 '()))
		    (set-predsym-pos-reg-index current-predicate (cons 0 '()))
		    (set-predsym-neg-reg-index current-predicate (cons 0 '()))
		    (set-predsym-pos-path-index current-predicate (cons 0 '()))
		    (set-predsym-neg-path-index current-predicate (cons 0 '()))
		    (set-predsym-pos-purity-index current-predicate (cons 0 '()))
		    (set-predsym-neg-purity-index current-predicate (cons 0 '()))
		    (set-predsym-symbol current-predicate symbol)
		    (set-predsym-arity  current-predicate 0)
		    (set-predsym-number current-predicate current-pred-nr)
		    (set-predsym-weight current-predicate 1)
		    (set! current-pred-nr (+ current-pred-nr 1))
		    (set! predicate-list (cons current-predicate predicate-list)))
		  (begin
		    (set! current-predicate (get-predsym symbol 0 predicate-list))))
; 		    (if (not (= (predsym-arity current-predicate) 0))
; 			(begin
; 			  (display "parser error: predicate symbol ")
; 			  (display (predsym-symbol current-predicate))
; 			  (display " introduced with arity ")
; 			  (display (predsym-arity current-predicate))
; 			  (display " but reused with arity ")
; 			  (display 0)
; 			  (newline)
; 			  (exit 1)))))
	      (cons current-predicate '()))
    ((symbol parameters) 
;      (if (eq? symbol 'equal)
; 	 (set! equality #t))
     (if (not (get-predsym symbol (length parameters) predicate-list))
	 (begin
	   (set! current-predicate (new-predsym))
	   (set-predsym-pos-context-index current-predicate (cons 0 '())) 
	   (set-predsym-neg-context-index current-predicate (cons 0 '()))
	   (set-predsym-pos-gr-index current-predicate (cons 0 '())) 
	   (set-predsym-neg-gr-index current-predicate (cons 0 '()))
	   (set-predsym-pos-unit-index current-predicate (cons 0 '())) 
	   (set-predsym-neg-unit-index current-predicate (cons 0 '()))
	   (set-predsym-pos-reg-index current-predicate (cons 0 '())) 
	   (set-predsym-neg-reg-index current-predicate (cons 0 '()))
	   (set-predsym-pos-path-index current-predicate (cons 0 '()))
	   (set-predsym-neg-path-index current-predicate (cons 0 '()))
	   (set-predsym-pos-purity-index current-predicate (cons 0 '()))
	   (set-predsym-neg-purity-index current-predicate (cons 0 '()))
	   (set-predsym-symbol current-predicate symbol)
	   (set-predsym-arity  current-predicate (length parameters))
	   (set-predsym-number current-predicate current-pred-nr)
           (if (eq? symbol 'equal)
	       (begin
		 (set-predsym-weight current-predicate eq-weight)
		 (set! equality-pred current-predicate)
		 (if (not equality)
		     (begin
		       (add-reflexivity-axiom)
		       (if (or dewey-numbers (> debug-level 4))
			   (begin
			     (print "reflexivity axiom added to indexes")))
		       (set! equality #t))))
	       (set-predsym-weight current-predicate 1))
	   (set! current-pred-nr (+ current-pred-nr 1))
	   (set! predicate-list (cons current-predicate predicate-list)))
	 (set! current-predicate (get-predsym symbol (length parameters) predicate-list)))
     (cons current-predicate parameters)))
   (term
    ((openpar term closepar) term)
    ((symbol) 
     (if (not (get-functor symbol 0 functor-list))
	 (begin
	   (set! current-functor (new-functor))
	   (set-functor-symbol current-functor symbol)
	   (set-functor-arity  current-functor 0)
	   (set-functor-number current-functor current-functor-nr)
	   (set-functor-weight current-functor 1)
	   (set! current-functor-nr (+ current-functor-nr 1))
	   (set! functor-list (cons current-functor functor-list)))
	 (set! current-functor (get-functor symbol 0 functor-list)))
     (if (not (assq symbol constant-list))
	 (set! constant-list (cons (cons symbol (list current-functor)) constant-list)))
     (cdr (assq symbol constant-list)))
    ((variable)
     (if (variable-exists? variable current-clause-variables)
	 (set! current-variable (get-variable variable current-clause-variables))
	 (begin
	   (set! current-variable (new-variable))
	   (set-variable-name current-variable variable)
	   (set! current-clause-variables (append current-clause-variables (list current-variable)))))
      current-variable)
    ((symbol parameters) 
     (if (not (get-functor symbol (length parameters) functor-list))
	 (begin
	   (set! current-functor (new-functor))
	   (set-functor-symbol current-functor symbol)
	   (set-functor-arity  current-functor (length parameters))
	   (set-functor-number current-functor current-functor-nr)
	   (set-functor-weight current-functor 2)
	   (set! current-functor-nr (+ current-functor-nr 1))
	   (set! functor-list (cons current-functor functor-list)))
	 (set! current-functor (get-functor symbol (length parameters) functor-list)))
     (append (list current-functor) parameters)))
   (parameters
    ((openpar termlist closepar) termlist))
   (termlist
    ((termlist comma term) (append termlist (list term)))
    ((term) (list term)))
   (type
    ((symbol) 
     (cond ((eq? symbol (string->symbol "axiom")) 'axiom)
	   ((eq? symbol (string->symbol "lemma")) 'lemma)
	   ((eq? symbol (string->symbol "hypothesis")) 'hypothesis)
	   ((eq? symbol (string->symbol "conjecture")) 'conjecture))))
   (clausenamepars
    ((openpar clausenametl closepar) clausenametl))
   (clausenametl
    ((clausenametl comma clausename) (append clausenametl (list clausename)))
    ((clausename) (list clausename)))
   (clausename
    ((openpar clausename closepar) clausename)
    ((symbol) symbol)
    ((variable) variable)
    ((symbol clausenamepars) 
     (append (list symbol) clausenamepars)))))


(define (insert-new-clause-in-clauselist clause clauses)
  (if (null? clauses)
      (begin
	;;(insert-subgoals-in-path-index (clause-subgoals clause))
	(list clause)) ;; initialize previously empty list
      (begin
	(if (not (clause< clause (car clauses)))
	  (begin
	    (set-cdr! clauses (cons (car clauses) (cdr clauses)))
	    (set-car! clauses clause))
	    ;;(insert-subgoals-in-path-index (clause-subgoals clause)))
	  (if (null? (cdr clauses))
	      (begin
		;;(insert-subgoals-in-path-index (clause-subgoals clause))
		(set-cdr! clauses (list clause)))
	      (insert-new-clause-in-clauselist clause (cdr clauses))))
	clauses)))

(define (clause< c d)
  (if (= (type-rating c) (type-rating d))
      (if (= (length (clause-subgoals c)) (length (clause-subgoals d)))
	  (let ((clause-complexity-c (compute-clause-complexity c))
		(clause-complexity-d (compute-clause-complexity d)))
	    (if (= clause-complexity-c clause-complexity-d)
	      (if (< (length (clause-variables c)) (length (clause-variables d)))
		  #t
		  #f)
	      (if (< clause-complexity-c clause-complexity-d)
		  #t
		  #f)))
	  (if (< (length (clause-subgoals c)) (length (clause-subgoals d)))
	      #t
	      #f))
      (if (< (type-rating c) (type-rating d))
	  #t
	  #f)))

(define (type-rating clause)
  (if norel
      (if (all-negative (clause-subgoals clause))
	  0
	  1)
      (let ((type (input-type (clause-input clause))))
	(cond
	 ((eq? type 'conjecture) 0)
	 ((eq? type 'lemma) 1)
	 (else 2)))))


(define (set-clause-depths clauses)
  (if (null? clauses)
      0
      (begin
	(set-clause-depth (car clauses) (- (set-clause-depths (cdr clauses)) 1))
	(clause-depth (car clauses)))))

(define (get-nr-current-singletons occurrences-list)
  (if (null? occurrences-list)
      0
      (if (= (car occurrences-list) 1)
	  (+ (get-nr-current-singletons (cdr occurrences-list)) 1)
	  (get-nr-current-singletons (cdr occurrences-list)))))

(define (get-max-pred-arity predlist)
  (if (null? predlist)
      '-
      (let ((tmp (get-max-pred-arity (cdr predlist))))
	(if (or (eq? tmp '-) (< tmp (predsym-arity (car predlist))))
	    (predsym-arity (car predlist))
	    tmp))))

(define (get-min-pred-arity predlist)
  (if (null? predlist)
      '-
      (let ((tmp (get-min-pred-arity (cdr predlist))))
	(if (or (eq? tmp '-) (> tmp (predsym-arity (car predlist))))
	    (predsym-arity (car predlist))
	    tmp))))

(define (get-max-func-arity funclist)
  (if (null? funclist)
      '-
      (let ((tmp (get-max-func-arity (cdr funclist))))
	(if (or (eq? tmp '-) (< tmp (functor-arity (car funclist))))
	    (functor-arity (car funclist))
	    tmp))))

(define (get-min-func-arity funclist)
  (if (null? funclist)
      '-
      (let ((tmp (get-min-func-arity (cdr funclist))))
	(if (or (eq? tmp '-) (> tmp (functor-arity (car funclist))))
	    (functor-arity (car funclist))
	    tmp))))

(define (get-predsym symbol arity predicate-list)
  (if (null? predicate-list)
      #f
      (if (and (eq? symbol (predsym-symbol (car predicate-list)))
	       (eq? arity (predsym-arity (car predicate-list))))
	  (car predicate-list)
	  (get-predsym symbol arity (cdr predicate-list)))))

(define (get-functor symbol arity functor-list)
  (if (null? functor-list)
      #f
      (if (and (eq? symbol (functor-symbol (car functor-list)))
	       (eq? arity (functor-arity (car functor-list))))
	  (car functor-list)
	  (get-functor symbol arity (cdr functor-list)))))

(define (get-signature symbol arity signature-list)
  (if (null? signature-list)
      #f
      (if (and (eq? symbol (signature-symbol (car signature-list)))
	       (eq? arity (signature-arity (car signature-list))))
	  (car signature-list)
	  (get-signature symbol arity (cdr signature-list)))))


(define (insert-literal literal lits)
  (if (pair? lits)
      (if (sign-predsym< literal (car lits))
	  (cons literal lits)
	  (cons (car lits) (insert-literal literal (cdr lits))))
      (cons literal lits)))

(define (display-for-debug)
  ;;(display (the-string))
  #t)

(define (variables-in-cleared-subgoals subgoals vars)
  (if (null? subgoals)
      vars
      (variables-in-cleared-subgoals (cdr subgoals) (free-vars-in-cleared-termlist (subgoal-terms (car subgoals)) vars))))
  

(define (free-vars-in-cleared-termlist terms vars)
  ;; finds all free variables in terms and inserts them into vars exactly once
  ;;(print vars)
  (if (not (null? terms))
      (begin
	(set-car! terms (ref (car terms)))
	(free-vars-in-cleared-termlist (cdr terms) (free-vars-in-term (car terms) vars)))
      vars))

(define (set-subgoal-weights subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-weight (car subgoals) (current-subgoal-weight (car subgoals)))
	(set-subgoal-weights (cdr subgoals)))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Preprocess factoring functions
;; ------------------------------------------------------------

(define (insert-clause-and-factors-in-clauselist clause clauses)
  (if (or tautologies (non-tautological clause))
      (if (not (clause-full-subsumed-by-clauselist clause clauses))
	  (begin
	    (set! clauses (backsubsume-clauses-with-new-clause clause clauses))
	    (insert-subgoals-in-path-index (clause-subgoals clause))
	    (if (unit-clause clause)
		(begin
		  (if (and preprocess  (not (eq? falsum (car (clause-subgoals clause)))))
		      (insert-unit-in-indexes (car (clause-subgoals clause))))
		  (insert-new-clause-in-clauselist clause clauses))
		(insert-subgoals-factors-in-clauselist (clause-subgoals clause)
						       (insert-new-clause-in-clauselist clause clauses))))
	  (begin
	    (if dewey-numbers 
		(begin 
		  (display "deleted subsumed input ")
		  (display-clause clause)
		  (newline)))
	    clauses))
      (begin
	(if dewey-numbers
	    (begin 
	      (display "deleted tautological input ")
	      (display-clause clause)
	      (newline)))
	clauses)))

(define (insert-subgoals-factors-in-clauselist subgoals clauses)
  (if (and (pair? subgoals) (not (null? (cdr subgoals))))
      (insert-subgoals-factors-in-clauselist (cdr subgoals) 
					     (insert-subgoal-factors-in-clauselist (car subgoals) (cdr subgoals) clauses))
      clauses))

(define (insert-subgoal-factors-in-clauselist subgoal rest-subgoals clauses)
  (if (and (pair? rest-subgoals) (sign-predsym= subgoal (car rest-subgoals)))
      (begin
	(if (eq-subgoal? subgoal)
	    (let ((left (ref (car (subgoal-terms subgoal))))
		  (right (ref (cadr (subgoal-terms subgoal)))))
	      (set-subgoal-terms subgoal (list right left))
	      (if (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms (car rest-subgoals)))
		  (let ((new-factor (renamed-factor (subgoal-clause subgoal))))
		    (unbind-vars)
		    (set! clauses (insert-clause-and-factors-in-clauselist new-factor clauses))))
	      (set-subgoal-terms subgoal (list left right))))
	(if (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms (car rest-subgoals)))
	    (let (new-factor)
	      (set! new-factor (renamed-factor (subgoal-clause subgoal)))
	      (unbind-vars)
	      (if dewey-numbers
		  (begin
		    (display "subgoal ")
		    (display-subgoal subgoal)
		    (display " generated new factor ")
		    (display-clause new-factor)
		    (display " of clause ")
		    (display-clause (subgoal-clause subgoal))
		    (newline)))
	      (set! clauses (insert-clause-and-factors-in-clauselist new-factor clauses))))
	(insert-subgoal-factors-in-clauselist subgoal (cdr rest-subgoals) clauses))
      clauses))

(define (renamed-factor clause)
  (let ((factor (new-clause)) (factor-input (new-input)) (subgoals (clause-subgoals clause)))
    (set-clause-foldup-lits factor '())
    (trail-stop)
    (compute-current-weights subgoals)
    (set-clause-subgoals factor (renamed-subgoals subgoals))
    (unrename-vars)
    (if equality (set-clause-subgoals factor (eliminate-var-disequations (clause-subgoals factor))))
    (if (and equality (not ordered-eq))
	(subgoals-orientate-disequations (clause-subgoals factor)))
    (set-clause-unsimplified factor (length (clause-subgoals factor)))
    (set! total-factorings (+ total-factorings 1))
    (set! successful-factorings (+ successful-factorings 1))
    (if (null? (clause-subgoals factor))
	(set-clause-subgoals factor (list falsum)))
    (set-clause-variables factor (variables-in-subgoals (clause-subgoals factor) '()))
    (set-subgoals-pointers (clause-subgoals factor) factor)
    (clear-subgoals-misc (clause-subgoals factor))
    (clear-subgoals-misc (clause-subgoals clause))
    (set-clause-input factor factor-input) 
    (set! current-clause-nr (+ current-clause-nr 1))
    (set-input-number factor-input current-clause-nr)
    (set-input-name factor-input (input-name (clause-input clause)))
    (set-input-type factor-input (input-type (clause-input clause)))
    (set-input-variant-index factor-input (cons 0 '()))
    (set-clause-initial factor #t)
    (if (not norel) (set-clause-relevant factor (clause-relevant clause)))
    factor))

(define (set-subgoals-pointers subgoals clause)
  (if (pair? subgoals)
      (begin
	(set-subgoal-clause (car subgoals) clause)
	(set-subgoals-pointers (cdr subgoals) clause))))

(define (clear-subgoals-misc subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-misc (car subgoals) #f)
	(clear-subgoals-misc (cdr subgoals)))))
