(module preprocess
	(include "dctp.sch")
	(export nr-clauses nr-non-horn nr-units nr-literals nr-eq-literals nr-predicates nr-propositionals 
		nr-functors nr-constants nr-variables nr-singleton-variables nr-axioms nr-lemmas nr-hypotheses
		nr-conjectures  nr-all-negatives predicate-list nr-non-eq-literals)
	(export (preprocess-formula) (get-formula-statistics) (display-formula-statistics) (all-negative subgoals) (check-if-path-indexes-empty)
		(select-closure-constant) (insert-subgoals-in-path-index subgoals) (delete-subgoals-from-path-index subgoals)
		(sort-preprocessed-clauses clauses) (check-links-with-index-links subgoal links index-links))
	(import (globals "globals.scm") (link-count link "link.scm") (eq-link-count link "link.scm")
		(link-inc link "link.scm") (eq-link-inc link "link.scm") (delete-element-from-list link "link.scm")
		(index "index.scm") (eliminate-var-disequations dctp "dctp.scm") (dctp "dctp.scm") (kbo> kbo "kbo.scm")
		(check-orientation kbo "kbo.scm") (kbo-precedence kbo "kbo.scm") (connection "connection.scm")
		(display "display.scm")	(non-tautological pruning "pruning.scm") (subgoal-in-list pruning "pruning.scm")
		(parser "parser.scm") (renamed-subgoals normal-form "normal-form.scm") (splitting "splitting.scm")
		(subsumption "subsumption.scm") (unification "unification.scm") (kbo-precedences kbo "kbo.scm")
		(varlist-renamed? normal-form "normal-form.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables of global use
;; ------------------------------------------------------------

(define current-clause-variables '())
(define current-singleton-variables '())
(define predicate-list '())
(define propositional-list '())
(define functor-list '())
(define constant-list '())
(define current-positive-subgoals 0)
(define nr-clauses 0)               
(define nr-non-horn 0)              
(define nr-units 0)
(define nr-literals 0)              
(define nr-eq-literals 0)           
(define nr-non-eq-literals 0)           
(define max-clause-length 0)        
(define max-clause-variables 0)        
(define nr-predicates 0)            
(define nr-propositionals 0)        
(define max-pred-arity 0)           
(define min-pred-arity 0)           
(define nr-functors 0)              
(define nr-constants 0)             
(define max-func-arity 0)           
(define min-func-arity 0)           
(define nr-variables 0)             
(define nr-singleton-variables 0)   

(define nr-axioms 0)   
(define nr-lemmas 0)   
(define nr-hypotheses 0)   
(define nr-conjectures 0)   
(define nr-all-negatives 0)   

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define formula-changed #f)
(define processing-conjecture #f)
(define clause-rewritten #f)
(define term-rewritten #f)

;;;*************************************************************************
;; ------------------------------------------------------------
;; Formula preprocessing functions
;; ------------------------------------------------------------

(define (preprocess-formula)
  (if (and preprocess (not silent)) 
      (let ((l (length (formula-clauses formula))))
	(display "starting preprocessing loop for ")
	(display l)
	(display " clauses")
	(newline)))
  (set! formula-changed #f)
  (if (and equality (= 0 (car (predsym-pos-path-index equality-pred))))
      ;; if all positive equality literals have been deleted, equality mode can be switched off
      ;; disequations become pure
      (set! equality #f))
  (if (preprocess-clauses (formula-clauses formula) (formula-clauses formula))
      (begin
	;;(print "preprocessing loop done")
	(if formula-changed
	    (preprocess-formula)
	    (begin
	      ;;(if preprocess (set-formula-clauses formula (sort-preprocessed-clauses (formula-clauses formula))))
	      #t)))
      #f))

(define (preprocess-clauses clauses prev-clauses)
  (if (pair? clauses)
      ;;(display "preprocessing ")
      ;;(display-clause  (car clauses))
      ;;(newline)
      (if (preprocess-clause clauses)
	  (begin
	    (if (car clauses)
		(preprocess-clauses (cdr clauses) clauses)
		;; below: the current clause must be deleted from the clause list
		(if (eq? clauses prev-clauses) ;; if we deleted the first clause
		    (if (null? (cdr clauses))  ;; and we also deleted the only clause
			(begin
			  (set-formula-clauses formula '())
			  #t)
			(begin
			  (set-car! clauses (cadr clauses))
			  (set-cdr! clauses (cddr clauses))
			  (preprocess-clauses clauses clauses)))
		    (begin
		      (set-cdr! prev-clauses (cdr clauses))
		      (preprocess-clauses (cdr prev-clauses) prev-clauses)))))
	  #f)
      #t))

(define (preprocess-clause clauses)
  (if preprocess
      (let ((clause (car clauses)) old-length)
	;;(display "preprocessing ")
	;;(display-clause clause)
	;;(newline)
	(set! old-length (length (clause-subgoals clause)))
	(if (and equality eager) (set-clause-subgoals clause (eliminate-var-disequations (clause-subgoals clause))))
	(set-clause-subgoals clause (delete-redundant-subgoals (clause-subgoals clause)))
	(set-clause-unsimplified clause (length (clause-subgoals clause)))
	(set-clause-variables clause (variables-in-cleared-subgoals (clause-subgoals clause) '()))
	(if (not (eq? (length (clause-subgoals clause)) old-length))
	    (begin
	      (set! formula-changed #t)))
	(if (null? (clause-subgoals clause))
	    (begin
	      (if dewey-numbers
		  (begin 
		    (display "empty clause deduced by preprocessing")
		    (newline)))
	      (set-clause-subgoals clause (list falsum))
	      (set-clause-variables clause '())
	      (set-clause-unsimplified clause 0)
	      #f)
	    (begin
	      (if (and (> old-length 1) (unit-clause clause))
		  (begin
		    (if dewey-numbers 
			(begin 
			  (display "generated new unit ")
			  (display-clause clause)
			  (newline)))
		    (insert-unit-in-indexes (car (clause-subgoals clause)))))
	      (if (or (unit-clause clause)
		      (not equality))
		  (get-subgoals-links (clause-subgoals clause)))
	      ;; now all the previous preprocessing steps are uniformly fitted into one cond form
	      (cond
	       ((and (unit-clause clause) (has-unit-link (car (clause-subgoals clause))))
		(if (not silent)
		    (begin
		      (display "found unit link between input subgoals ")
		      (display-subgoal (car (clause-subgoals clause)))
		      (display " and ")
		      (display-subgoal (has-unit-link (car (clause-subgoals clause))))
		      (newline)))
		#f)
	       ((let ((subsumed (clause-subsumed clause))) (and subsumption subsumed (not (eq? (subgoal-clause subsumed) clause))))
		(if dewey-numbers 
		    (begin 
		      (display "deleted subsumed clause ")
		      (display-clause clause)
		      (newline)))
		(set! formula-changed #t)
		(delete-subgoals-from-path-index (clause-subgoals clause))
		;;(set-formula-clauses formula (remq! clause (formula-clauses formula)))
		(set-car! clauses #f)
		(if (unit-clause clause)
		    (delete-unit-from-indexes (car (clause-subgoals clause))))
		#t)
	       ((not (or tautologies (non-tautological clause)))
		(if dewey-numbers 
		    (begin 
		      (display "deleted tautological clause ")
		      (display-clause clause)
		      (newline)))
		(set! formula-changed #t)
		(delete-subgoals-from-path-index (clause-subgoals clause))
		;;(set-formula-clauses formula (remq! clause (formula-clauses formula)))
		(set-car! clauses #f)
		(if (unit-clause clause)
		    (delete-unit-from-indexes (car (clause-subgoals clause))))
		#t)
	       ((and (not equality) (pure-clause clause))
		(if dewey-numbers 
		    (begin 
		      (display "deleted pure clause ")
		      (display-clause clause)
		      (newline)))
		(set! formula-changed #t)
		(delete-subgoals-from-path-index (clause-subgoals clause))
		;;(set-formula-clauses formula (remq! clause (formula-clauses formula)))
		(set-car! clauses #f)
		(if (unit-clause clause)
		    (delete-unit-from-indexes (car (clause-subgoals clause))))
		#t)
	       ((and pre-subsume (let ((full-subsumed (clause-full-subsumed-by-clauselist clause (formula-clauses formula))))
				   (if full-subsumed
				       (if dewey-numbers 
					   (begin 
					     (display "deleted clause ")
					     (display-clause clause)
					     (display " subsumed by ")
					     (display-clause full-subsumed)
					     (newline))))
				   full-subsumed))
		(set! formula-changed #t)
		(delete-subgoals-from-path-index (clause-subgoals clause))
		;;(set-formula-clauses formula (remq! clause (formula-clauses formula)))
		(set-car! clauses #f)
		(if (unit-clause clause)
		    (delete-unit-from-indexes (car (clause-subgoals clause))))
		#t)
	       ((and res-isol (not equality) (subgoals-isolated (clause-subgoals clause))
		     (or (not isol-limit) (<= (length (clause-subgoals clause)) isol-limit)))
		(let ((isolated (subgoals-isolated (clause-subgoals clause))) link)
		  (set! link (caar (subgoal-links isolated)))
		  (if (eq? (input-type (clause-input (subgoal-clause link))) 'conjecture)
		      (set! processing-conjecture #t))
		  (if dewey-numbers
		      (begin 
			(display "found isolated subgoal ")
			(display-subgoal isolated)
			(display " in ")
			(display-clause clause)
			(display " connected to clause ")
			(display-clause (subgoal-clause link))
			(newline)))
		  (if (resolve-isolated-link isolated link)
		      (begin
			(set! processing-conjecture #f)
			#t)
		      #f)))
	       ((and isol (not equality) (subgoals-isolated (clause-subgoals clause)))
		(let ((isolated (subgoals-isolated (clause-subgoals clause))) link)
		  (set! link (caar (subgoal-links isolated)))
		  (if (eq? (input-type (clause-input (subgoal-clause link))) 'conjecture)
		      (set! processing-conjecture #t))
		  (if dewey-numbers
		      (begin 
			(display "found isolated subgoal ")
			(display-subgoal isolated)
			(display " in ")
			(display-clause clause)
			(display " connected to clause ")
			(display-clause (subgoal-clause link))
			(newline)))
		  (propagate-isolated-unifier isolated link)
		  (set! processing-conjecture #f)
		  #t))
	       ((and equality demod (unit-clause clause) (equation? (car (clause-subgoals clause))) (is-demodulator clause))
		;; FIXME: The preprocessing needs to keep track of the number of
		;; equality literals in the input. if e.g. demodulation kills the only equation
		;; in the problem, equality can then be set to #f, purity can be applied etc.
		(if (eq? (input-type (clause-input clause)) 'conjecture)
		    (set! processing-conjecture #t))
		(set-subgoal-ordered-eq (car (clause-subgoals clause)) #t)
		(if (and (not (is-left-demodulator clause)) (is-right-demodulator clause))
		    (set-subgoal-terms (car (clause-subgoals clause))
				       (list (cadr (subgoal-terms (car (clause-subgoals clause))))
					     (car (subgoal-terms (car (clause-subgoals clause)))))))
		(if dewey-numbers 
		    (begin 
		      (if (and (is-left-demodulator clause) (is-right-demodulator clause))
			  (display "applying symmectric demodulator ")
			  (display "applying demodulator "))
		      (display-clause clause)
		      (newline)))
		(try-rewrite-clauses (car (clause-subgoals clause)) (formula-clauses formula))
		(set! formula-changed #t)
		(delete-unit-from-indexes (car (clause-subgoals clause)))
		(delete-subgoal-from-index (car (clause-subgoals clause)) 'path)
		(set-subgoal-terms (car (clause-subgoals clause))
				   (list (cadr (subgoal-terms (car (clause-subgoals clause))))
					 (cadr (subgoal-terms (car (clause-subgoals clause))))))
		(insert-subgoal-in-index (car (clause-subgoals clause)) 'path)
		(insert-unit-in-indexes (car (clause-subgoals clause)))
		(set! processing-conjecture #f)
		#t)
	       ((and rewrite-steps (unit-clause clause) (equation? (car (clause-subgoals clause))))
		(if (or dewey-numbers (> debug-level 5))
		    (begin 
		      (display "rewriting with unit equation ")
		      (display-clause clause)
		      (newline)))
		(try-rewrite-clauses (car (clause-subgoals clause)) (formula-clauses formula))
		#t)
	       ((and split (not (eq? (car (clause-subgoals clause)) falsum)))
		(let ((splitted '()))
		  (set! splitted (split-subgoals (clause-subgoals clause) '()))
		  (if (and split (> (length splitted) 1) (not (and (= (length splitted) 2) (one-list-is-ground splitted))))
		      (begin
			(if (or dewey-numbers (> debug-level 4))
			    (begin
			      (display "subgoals ")
			      (display-subgoal-list (clause-subgoals clause))
			      (display " can be split into ")
			      (display (length splitted))
			      (display " disjoint subsets")
			      (newline)))
			(delete-subgoals-from-path-index (clause-subgoals clause))
			(set! splitted (split-lists-to-clauses splitted (clause-depth clause) #f))
			(set-clauses-input splitted (clause-input clause))
			(set-clause-pointers-in-clauses splitted)
			;;(insert-clauses-in-path-index splitted)
			(if (or dewey-numbers (> debug-level 4))
			    (begin
			      (display "new splitted clauses:")
			      (newline)
			      (display-clauses splitted)
			      (newline)))
			(set! nr-splitted (+ nr-splitted (- (length splitted) 1)))
			(set-formula-clauses formula (insert-splitted-clauses-in-clauselist splitted (formula-clauses formula)))
			(set-car! clauses #f)
			(set! formula-changed #t))
		      (if (not (eq? (length (clause-subgoals clause)) old-length))
			  (set! formula-changed #t)))
		  #t))
	       (else
		#t)))))
      (if (unit-clause (car clauses))
	  (let ((clause (car clauses)) unit-link)
	    (set-subgoal-links (car (clause-subgoals clause)) (unify-subgoal-with-index (car (clause-subgoals clause)) 'path))
	    (set! unit-link (has-unit-link (car (clause-subgoals clause))))
	    (if unit-link
		(begin
		  (if (not silent)
		      (begin
			(display "found unit link between input subgoals ")
			(display-subgoal (car (clause-subgoals clause)))
			(display " and ")
			(display-subgoal unit-link)
			(newline)))
		  #f)
		#t))
	  #t)))

(define (subgoals-isolated subgoals)
  (if (pair? subgoals)
      (if (and (pair? (subgoal-links (car subgoals)))         ;; if the subgoal has one list of links
               (null? (cdr (subgoal-links (car subgoals))))   ;; exactly one
               (pair? (car (subgoal-links (car subgoals))))   ;; if this list has one link
               (null? (cdar (subgoal-links (car subgoals))))  ;; exactly one
	       (not (subgoal-split-lit (car subgoals))))      ;; and it's not a split literal
          ;; NOTA BENE: internal links need not be checked here, as a subgoal with exactly one internal link would not have survived
          ;; the purity test
          (car subgoals)
          (subgoals-isolated (cdr subgoals)))
      #f))


(define (propagate-isolated-unifier subgoal other-subgoal)
  (let ((other-clause (subgoal-clause other-subgoal)))
    (if (unit-subgoal subgoal)
	(delete-unit-from-indexes subgoal))
    (delete-subgoals-from-path-index (clause-subgoals (subgoal-clause subgoal)))
    (if (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms other-subgoal))
	(if (varlist-renamed? (clause-variables (subgoal-clause subgoal)) (clause-variables (subgoal-clause subgoal)) '())
	    (unbind-vars)
	    (let ((old-subgoals (clause-subgoals (subgoal-clause subgoal))))
	      (trail-stop)
	      (set-clause-subgoals (subgoal-clause subgoal) (renamed-subgoals old-subgoals))
	      (unrename-vars)
	      (unbind-vars)
	      (set! formula-changed #t)
	      (set-clause-unsimplified (subgoal-clause subgoal) (length (clause-subgoals (subgoal-clause subgoal))))
	      (set-clause-variables (subgoal-clause subgoal) (variables-in-subgoals (clause-subgoals (subgoal-clause subgoal)) '()))
	      (clause-pointers-in-subgoals (clause-subgoals (subgoal-clause subgoal)) (subgoal-clause subgoal))
	      (if processing-conjecture
		  (set-input-type (clause-input (subgoal-clause subgoal)) 'conjecture))
	      (if dewey-numbers
		    (begin
		      (display "applied isolated unifier to obtain ")
		      (display-clause (subgoal-clause subgoal))
		      (newline))))))
    (insert-subgoals-in-path-index (clause-subgoals (subgoal-clause subgoal)))
    (if (unit-clause (subgoal-clause subgoal))
	(insert-unit-in-indexes (car (clause-subgoals (subgoal-clause subgoal)))))
    #t))

(define (resolve-isolated-link subgoal other-subgoal)
  (let ((other-clause (subgoal-clause other-subgoal)) (falsified #f))
    (if (unit-subgoal subgoal)
	(delete-unit-from-indexes subgoal))
    (delete-subgoals-from-path-index (clause-subgoals (subgoal-clause subgoal)))
    (if (unify-lists-stop (subgoal-terms subgoal) (subgoal-terms other-subgoal))
	(let ((inserted-subgoals (remq other-subgoal (clause-subgoals other-clause)))
	      new-subgoals ancestors)
	  (trail-stop)
	  (set! new-subgoals (renamed-subgoals (append inserted-subgoals (remq subgoal (clause-subgoals (subgoal-clause subgoal))))))
	  (unrename-vars)
	  (unbind-vars)
	  (if (subgoal-next subgoal)
	      (set! ancestors (cons other-subgoal (cons subgoal (subgoal-next subgoal))))
	      (set! ancestors (list other-subgoal subgoal)))
	  (if (common-subgoal-in-ancestors (cons other-subgoal (subgoal-next other-subgoal)) (cdr ancestors))
	      (begin
		(if dewey-numbers
		    (begin
		      (display "rejecting res-isol as subgoal ")
		      (display-subgoal subgoal)
		      (display " already resolved with ancestor of subgoal ")
		      (display-subgoal other-subgoal)
		      (newline)))
		(if (unit-clause (subgoal-clause subgoal))
		    (insert-unit-in-indexes (car (clause-subgoals (subgoal-clause subgoal))))))
	      (begin
		(set! formula-changed #t)
		(set-ancestors-in-definition inserted-subgoals ancestors)
		(if (subgoal-next other-subgoal)
		    (set-subgoal-next other-subgoal (append ancestors (subgoal-next other-subgoal)))
		    (set-subgoal-next other-subgoal ancestors))
		(set-clause-subgoals (subgoal-clause subgoal) new-subgoals)
		(set-clause-unsimplified (subgoal-clause subgoal) (length (clause-subgoals (subgoal-clause subgoal))))
		(if (null? (clause-subgoals (subgoal-clause subgoal)))
		    (begin
		      (set! falsified #t)
		      (set-clause-subgoals (subgoal-clause subgoal) (list falsum))))
		(set-clause-variables (subgoal-clause subgoal) (variables-in-subgoals (clause-subgoals (subgoal-clause subgoal)) '()))
		(clause-pointers-in-subgoals (clause-subgoals (subgoal-clause subgoal)) (subgoal-clause subgoal))
		(set-clause-initial (subgoal-clause subgoal) (clause-initial other-clause))
		(if processing-conjecture
		    (set-input-type (clause-input (subgoal-clause subgoal)) 'conjecture))
		(if (unit-clause (subgoal-clause subgoal))
		    (insert-unit-in-indexes (car (clause-subgoals (subgoal-clause subgoal)))))
		(if dewey-numbers
		    (begin
		      (display "resolved isolated connection to ")
		      (display-clause (subgoal-clause subgoal))
		      (newline))))))
	(begin
	  (if (unit-clause (subgoal-clause subgoal)) (insert-unit-in-indexes (car (clause-subgoals (subgoal-clause subgoal)))))))
    (if falsified
	(begin
	  (if dewey-numbers
	      (begin
		(display "resolving isolated connection produced empty clause")
		(newline)))
	  #f)
	(begin
	  (insert-subgoals-in-path-index (clause-subgoals (subgoal-clause subgoal)))
	  #t))))

(define (definition-list-subsumes-ancestors definition ancestors)
  (if (pair? definition)
      (let ((renamed-definition (subgoal-misc (car definition))))
	(if (definition-subsumes-ancestors renamed-definition ancestors)
	    #t
	    (definition-list-subsumes-ancestors (cdr definition) ancestors)))
      #f))
      
(define (definition-subsumes-ancestors definition-sg ancestors)
  (if (pair? ancestors)
      (if (and (eq? (subgoal-sign definition-sg) (subgoal-sign (car ancestors)))
	       (eq? (subgoal-pred definition-sg) (subgoal-pred (car ancestors)))
	       (match-lists-stop (subgoal-terms definition-sg) (subgoal-terms (car ancestors))))
	  (begin
	    (unbind-vars)
	    (if dewey-numbers
		(begin
		  (display "rejecting res-isol as defining subgoal ")
		  (display-subgoal definition-sg)
		  (display " subsumes resolved subgoal ")
		  (display-subgoal (car ancestors))
		  (newline)))
	    (car ancestors))
	  (definition-subsumes-ancestors definition-sg (cdr ancestors)))
      #f))
	
(define (ancestors-subsume-definition ancestors definition)
  (if (pair? ancestors)
      (begin
	(if (null? instantiatable-vars)
	    (set! instantiatable-vars (free-vars-in-termlist (subgoal-terms (car ancestors)) '())))
	(if (resolved-subsumes-definition (car ancestors) definition)
	    #t
	    (begin
	      (set! instantiatable-vars '())
	      (ancestors-subsume-definition (cdr ancestors) definition))))
      #f))
      
(define (resolved-subsumes-definition resolved-sg definition)
  (if (pair? definition)
      (let ((definition-sg (subgoal-misc (car definition))))
      (if (and (eq? (subgoal-sign resolved-sg) (subgoal-sign definition-sg))
	       (eq? (subgoal-pred resolved-sg) (subgoal-pred definition-sg))
	       (match-lists-stop (subgoal-terms resolved-sg) (subgoal-terms definition-sg)))
	  (begin
	    (unbind-vars)
	    (if dewey-numbers
		(begin
		  (display "rejecting res-isol as defining subgoal ")
		  (display-subgoal definition-sg)
		  (display " subsumed by resolved subgoal ")
		  (display-subgoal resolved-sg)
		  (newline)))
	    definition-sg)
	  (resolved-subsumes-definition resolved-sg (cdr definition))))
      #f))
	
(define (set-ancestors-in-definition inserted-subgoals ancestors)
  (if (pair? inserted-subgoals)
      (begin
	(if (subgoal-next (car inserted-subgoals))
	    (set-subgoal-next (subgoal-misc (car inserted-subgoals)) (append ancestors (subgoal-next (car inserted-subgoals))))
	    (set-subgoal-next (subgoal-misc (car inserted-subgoals)) ancestors))
	(set-ancestors-in-definition (cdr inserted-subgoals) ancestors))))

(define (set-ancestors-in-copies subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-next (subgoal-misc (car subgoals)) (subgoal-next (car subgoals)))
	(set-ancestors-in-copies (cdr subgoals)))))

(define (common-subgoal-in-ancestors ancestors-a ancestors-b)
  (if (pair? ancestors-a)
      (or (memq (car ancestors-a) ancestors-b)
	  (common-subgoal-in-ancestors (cdr ancestors-a) ancestors-b))
      #f))

(define (replace-clause-in-clauselist old-clause new-cl clauses)
  (if (pair? clauses)
      (if (eq? old-clause (car clauses))
	  (set-car! clauses new-cl)
	  (replace-clause-in-clauselist old-clause new-cl (cdr clauses)))))

(define (varlist-instantiated? variables)
  (if (pair? variables)
      (if (free-variable (car variables))
	  (varlist-instantiated? (cdr variables))
	  #t)
      #f))

(define (pure-clause clause)
  (if (and (eq? (car (clause-subgoals clause)) falsum)
	   (pair? (clause-subgoals clause))        ;; these two lines check for clause length 1
	   (null? (cdr (clause-subgoals clause))))
      #f
      (pure-subgoals (clause-subgoals clause))))

(define (pure-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(if (contains-external-link (car subgoals) (subgoal-links (car subgoals)))
	    (pure-subgoals (cdr subgoals))
	    (begin
	      (if dewey-numbers 
		  (begin
		    (display "found pure subgoal ")
		    (display-subgoal (car subgoals))
		    (display " with links ")
		    (display-subgoal-lists (subgoal-links (car subgoals)))
		    (newline)))
	      (set! formula-changed #t)
	      (car subgoals))))
      #f))

(define (delete-redundant-subgoals subgoals)
  (if (pair? subgoals)
      (let ((matching-unit #f))
	(if simplify (set! matching-unit (match-subgoal-with-index (car subgoals))))
	(if (and simplify matching-unit)
	    (begin
	      (if dewey-numbers 
		  (begin 
		    (display "delete simplifiable subgoal ")
		    (display-subgoal (car subgoals))
		    (newline)))
	      (set! formula-changed #t)
	      (if (eq? (input-type (clause-input (subgoal-clause matching-unit))) 'conjecture)
		  (set-input-type (clause-input (subgoal-clause (car subgoals))) 'conjecture))
	      (delete-subgoal-from-index (car subgoals) 'path)
	      (delete-redundant-subgoals (cdr subgoals)))
	    (let (same-sg (new-subgoals (delete-redundant-subgoals (cdr subgoals))))
	      (set! same-sg (subgoal-in-list (car subgoals) new-subgoals))
	      (if (and contract same-sg)
		  (begin
		    (if dewey-numbers
			(begin
			  (display "deleting multiple occurrence of subgoal ")
			  (display-subgoal (car subgoals))
			  (newline)))
		    (set! nr-contractions (+ nr-contractions 1))
		    (delete-subgoal-from-index (car subgoals) 'path)
		    (set! formula-changed #t)
		    new-subgoals)
		  (cons (car subgoals) new-subgoals)))))
      '()))

(define (get-formula-statistics)
  (reset-formula-statistics)
  (get-clauses-statistics (formula-clauses formula))
  (if (= nr-clauses 0)
      (begin
	(if (not silent) (print "all clauses deleted by preprocessing")))
      (begin
	(if (= nr-all-negatives 0)
	    (if (not silent) (print "no all negative clause in formula")))
	(if (and (not norel) (= nr-clauses nr-conjectures))
	    (if (not silent) (print "all clauses of type conjecture ... restricting relevance to all-negative clauses")))
	(if (and (not norel) (= nr-conjectures 0))
	    (if (not silent) (print "no conjecture type clause ... restricting relevance to all-negative clauses")))
	(if (or norel (= nr-clauses nr-conjectures) (= nr-conjectures 0))
	    (set-all-negatives-relevant (formula-clauses formula)))
	(if (= nr-propositionals nr-predicates)
	    (begin
	      (if (not silent) (print "input formula is propositional"))
	      (set! groundable #t))
	    (if (= max-func-arity 0)
		(begin
		  (if (not silent) (print "input formula is in Bernays-Schnfinkel class"))
		  (set! groundable #t))))
	(if (= 0 nr-non-horn)
	    (begin
	      (if (not silent) (print "input formula is Horn"))
	      (set! horn #t)))
	(if (= nr-eq-literals 0)
	    (set! equality #f))
	(if (and equality (= nr-eq-literals nr-literals) (= nr-clauses nr-units))
	    (begin
	      (if (not silent) (print "input formula is in unit equality class"))
	      (set! unit-eq #t)
	      (set! full-variants #f)))))
  (set! link-inc nr-eq-literals)
  (set! eq-link-inc (* eq-bias nr-non-eq-literals))
  (set! link-count link-inc)
  (set! eq-link-count eq-link-inc))

(define (get-clauses-statistics clauses)
  (if (pair? clauses)
      (begin
	(cond ((eq? (input-type (clause-input (car clauses))) 'axiom)
	       (set! nr-axioms (+ nr-axioms 1)))
	      ((eq? (input-type (clause-input (car clauses))) 'lemma)
	       (set! nr-lemmas (+ nr-lemmas 1)))
	      ((eq? (input-type (clause-input (car clauses))) 'hypothesis)
	       (set! nr-hypotheses (+ nr-hypotheses 1)))
	      ((eq? (input-type (clause-input (car clauses))) 'conjecture)
	       (set! nr-conjectures (+ nr-conjectures 1))))
	(set! nr-clauses (+ nr-clauses 1))
	(if (unit-clause (car clauses))
	    (set! nr-units (+ nr-units 1)))
	(set! current-positive-subgoals 0)
	(if (> (length (clause-subgoals (car clauses))) max-clause-length)
	    (set! max-clause-length (length (clause-subgoals (car clauses)))))
	(set! nr-variables (+ nr-variables (length (clause-variables (car clauses)))))
	(set! current-clause-variables '())
	(set! current-singleton-variables '())
	(get-subgoals-statistics (clause-subgoals (car clauses)))
	(set! nr-singleton-variables (+ nr-singleton-variables (length current-singleton-variables)))
; 	(if (> (length current-singleton-variables) 0)
; 	    (begin
; 	      (display "singleton var ")
; 	      (display current-singleton-variables)
; 	      (display " found in ")
; 	      (display-clause (car clauses))
; 	      (newline)))
; 	(if (not (= (length current-clause-variables) (length (clause-variables (car clauses)))))
; 	    (print "mismatching clause lists!"))
	(if (> current-positive-subgoals 1)
	    (set! nr-non-horn (+ nr-non-horn 1)))
	(if (= current-positive-subgoals 0)
	    (set! nr-all-negatives (+ nr-all-negatives 1)))
	(get-clauses-statistics (cdr clauses)))))

(define (get-subgoals-statistics subgoals)
  (if (pair? subgoals)
      (begin
	(set! nr-literals (+ nr-literals 1))
	(if (eq? (subgoal-sign (car subgoals)) '+)
	    (set! current-positive-subgoals (+ current-positive-subgoals 1)))
	(if (eq? (subgoal-pred (car subgoals)) equality-pred)
	    (set! nr-eq-literals (+ nr-eq-literals 1))
	    (set! nr-non-eq-literals (+ nr-non-eq-literals 1)))
	(if (not (memq (subgoal-pred (car subgoals)) predicate-list))
	    (begin
	      (set! predicate-list (cons (subgoal-pred (car subgoals)) predicate-list))
	      (set! nr-predicates (+ nr-predicates 1))
	      ;;(set-predsym-number (subgoal-pred (car subgoals)) nr-predicates)
	      (if (= (predsym-arity (subgoal-pred (car subgoals))) 0)
		  (begin
		    (set! propositional-list (cons (subgoal-pred (car subgoals)) propositional-list))
		    (set! nr-propositionals (+ nr-propositionals 1))))))
	(if (or (< (predsym-arity (subgoal-pred (car subgoals))) min-pred-arity) 
		(and (= min-pred-arity 0) (= nr-propositionals 0)))
	    (set! min-pred-arity (predsym-arity (subgoal-pred (car subgoals)))))
	(if (> (predsym-arity (subgoal-pred (car subgoals))) max-pred-arity)
	    (set! max-pred-arity (predsym-arity (subgoal-pred (car subgoals)))))
	(get-termlist-statistics (subgoal-terms (car subgoals)))
	(get-subgoals-statistics (cdr subgoals)))))

(define (get-termlist-statistics terms)
  (if (pair? terms)
      (begin
	(if (variable (ref (car terms)))
	    (if (memq (ref (car terms)) current-clause-variables)
		(set! current-singleton-variables (remq! (ref (car terms)) current-singleton-variables))
		(begin
		  (set! current-clause-variables (cons (ref (car terms)) current-clause-variables))
		  (set! current-singleton-variables (cons (ref (car terms)) current-singleton-variables))))
	    (begin
	      (if (or (< (functor-arity (car (ref (car terms)))) min-func-arity) 
		      (and (= min-func-arity 0) (= nr-constants 0)))
		  (set! min-func-arity (functor-arity (car (ref (car terms))))))
	      (if (> (functor-arity (car (ref (car terms)))) max-func-arity)
		  (set! max-func-arity (functor-arity (car (ref (car terms))))))
	      (if (constant (ref (car terms)))
		  (if (not (assq (car (ref (car terms))) functor-list))
		      (begin
			(set! functor-list (cons (cons (car (ref (car terms))) 1) functor-list))
			(set! constant-list (cons (cons (car (ref (car terms))) 1) constant-list))
			(set! nr-functors (+ nr-functors 1))
			(set! nr-constants (+ nr-constants 1)))
		      (let ((functor (assq (car (ref (car terms))) functor-list))
			    (constant (assq (car (ref (car terms))) constant-list)))
			(set-cdr! functor (+ (cdr functor) 1))
			(set-cdr! constant (+ (cdr constant) 1))))
		  (begin
		    (if (not (assq (car (ref (car terms))) functor-list))
		      (begin
			(set! functor-list (cons (cons (car (ref (car terms))) 1) functor-list))
			(set! nr-functors (+ nr-functors 1)))
		      (let ((functor (assq (car (ref (car terms))) functor-list)))
			(set-cdr! functor (+ (cdr functor) 1))))
		    (get-termlist-statistics (cdr (ref (car terms))))))))
	(get-termlist-statistics (cdr terms)))))

(define (select-closure-constant)
  (select-constant constant-list #f)
  ;;(if (pair? constant-list)
      ;;(set! selected-constant (list (caar constant-list))))
  (if (or #t dewey-numbers (> debug-level 4))
      (begin
	(if (and selected-constant (not silent))
	    (begin
	      (vector-set! kbo-precedences (- (functor-number (car selected-constant)) 1) 0)
	      (display "constant selected for t-closures: ")
	      (display-ref selected-constant))
	    (display "no constant in input"))
	(newline))))

(define (select-constant constant-list max-constant)
  (if (pair? constant-list)
      (if (or (not max-constant)
	      (> (cdar constant-list) (cdr max-constant)))
	  (select-constant (cdr constant-list) (car constant-list))
	  (select-constant (cdr constant-list) max-constant))
      (if max-constant
	  (set! selected-constant (list (car max-constant))))))

(define (reset-formula-statistics)
  (set! nr-axioms 0)   
  (set! nr-lemmas 0)   
  (set! nr-hypotheses 0)   
  (set! nr-conjectures 0)   
  (set! nr-all-negatives 0)   
  (set! current-clause-variables '())
  (set! current-singleton-variables '())
  (set! predicate-list '())
  (set! functor-list '())
  (set! propositional-list '())
  (set! constant-list '())
  (set! current-positive-subgoals 0)
  (set! nr-clauses 0)
  (set! nr-non-horn 0)
  (set! nr-units 0)
  (set! nr-literals 0)
  (set! nr-eq-literals 0)
  (set! max-clause-length 0)
  (set! nr-predicates 0)
  (set! nr-propositionals 0)
  (set! max-pred-arity 0)
  (set! min-pred-arity 0)
  (set! nr-functors 0)
  (set! nr-constants 0)
  (set! max-func-arity 0)
  (set! min-func-arity 0)
  (set! nr-variables 0)
  (set! nr-singleton-variables 0))

(define (set-all-negatives-relevant clauses)
  (if (pair? clauses)
      (begin
	(if (all-negative (clause-subgoals (car clauses)))
	    (set-clause-relevant (car clauses) #t)
	    (set-clause-relevant (car clauses) #f))
	(set-all-negatives-relevant (cdr clauses)))))

(define (all-negative subgoals)
  (if (null? subgoals)
      #t
      (if (eq? '- (subgoal-sign (car subgoals)))
	  (all-negative (cdr subgoals))
	  #f)))

(define (display-formula-statistics)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\tSyntax :\t")
  (display #"Number of clauses :\t")
  (display nr-clauses)
  (display " ( ")
  (display nr-non-horn)
  (display " non-Horn; ")
  (display nr-units)
  (display " unit)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Number of literals :\t")
  (display nr-literals)
  (display " ( ")
  (display nr-eq-literals)
  (display " equality)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Maximal clause size :\t")
  (display max-clause-length)
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Number of predicates :\t")
  (display nr-predicates)
  (display " ( ")
  (display nr-propositionals)
  (display " propositional; ")
  (display min-pred-arity)
  (display "-")
  (display max-pred-arity)
  (display " arity)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Number of functors :\t")
  (display nr-functors)
  (display " ( ")
  (display nr-constants)
  (display " constant; ")
  (display min-func-arity)
  (display "-")
  (display max-func-arity)
  (display " arity)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Number of variables :\t")
  (display nr-variables)
  (display " ( ")
  (display nr-singleton-variables)
  (display " singleton)")
  (newline)
  (if preprocess-only
      (display "% ")
      (display "  "))
  (display #"\t\t\t")
  (display #"Maximal term depth :\t")
  (display input-term-depth)
  (newline))

(define (var-in-list? var list)
  (if (pair? list)
      (if (eq? var (car list))
	  list
	  (var-in-list? var (cdr list)))
      #f))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Demodulation functions
;; ------------------------------------------------------------

(define (is-demodulator eq-unit)
  (or (is-right-demodulator eq-unit) (is-left-demodulator eq-unit)))

;; FIXME: Demodulators are not restricted to complex terms! Change that
(define (is-left-demodulator eq-unit)
  (let ((left (ref (car (subgoal-terms (car (clause-subgoals eq-unit))))))
	(right (ref (cadr (subgoal-terms (car (clause-subgoals eq-unit)))))))
    (and (complex left)
	 (pairwise-different-variable-args (cdr left) '())
	 (not (functor-occurs-in-term (car left) right))
	 (< (simple-term-complexity left) (* demod-limit (simple-term-complexity right))))))

(define (is-right-demodulator eq-unit)
  (let ((left (ref (car (subgoal-terms (car (clause-subgoals eq-unit))))))
	(right (ref (cadr (subgoal-terms (car (clause-subgoals eq-unit)))))))
    (and (complex right)
	 (pairwise-different-variable-args (cdr right) '())
	 (not (functor-occurs-in-term (car right) left))
	 (< (simple-term-complexity left) (* demod-limit (simple-term-complexity right))))))

(define (pairwise-different-variable-args terms arg-vars)
  (if (pair? terms)
      (if (and (variable (ref (car terms))) (not (memq (ref (car terms)) arg-vars)))
	  (pairwise-different-variable-args (cdr terms) (cons (ref (car terms)) arg-vars))
	  #f)
      #t))

(define (functor-occurs-in-term functor term)
  (if (or (index-variable term) (variable term))
      #f
      (if (eq? functor (car term))
	  #t
	  (if (complex term)
	      (functor-occurs-in-termlist functor (cdr term))
	      #f))))

(define (functor-occurs-in-termlist functor terms)
  (if (pair? terms)
      (if (functor-occurs-in-term functor (ref (car terms)))
	  #t
	  (functor-occurs-in-termlist functor (cdr terms)))
      #f))

(define (simple-term-complexity term)
  (if (or (variable term) (constant term)) 
      1
      (+ 1 (simple-termlist-complexity (cdr term)))))

(define (simple-termlist-complexity terms)
  (if (null? terms)
      0
      (+ (simple-term-complexity (car terms)) (simple-termlist-complexity (cdr terms)))))

(define (insert-splitted-clauses-in-clauselist clauses clauselist)
  (if (pair? clauses)
      (begin
	(insert-subgoals-in-path-index (clause-subgoals (car clauses)))
	(insert-new-clause-in-clauselist (car clauses) (insert-splitted-clauses-in-clauselist (cdr clauses) clauselist)))
      clauselist))

(define (clear-clauses-path-info clauses)
   (if (pair? clauses)
       (begin
	 (set-clause-path (car clauses) #f)
	 (clear-subgoals-next-info (clause-subgoals (car clauses)))
	 (clear-clauses-path-info (cdr clauses)))))  

(define (clear-subgoals-next-info subgoals)
   (if (pair? subgoals)
       (begin
	 (if (subgoal-next (car subgoals))
	     (begin
	       (set-subgoal-links (car subgoals) (list (subgoal-next (car subgoals))))
	       (set-subgoal-next (car subgoals) #f)))
	 (clear-subgoals-next-info (cdr subgoals)))))  

;;;*************************************************************************
;; ------------------------------------------------------------
;; Functions for preprocess rewriting
;; ------------------------------------------------------------

(define instantiatable-vars '())

(define (match-termlists k l)
  ;; tries to match termlists k and l.
  ;; If matching succeeds, #t is returned, else #f.
  ;; Side effect: termlists and trail destructively modified.
  ;; Before every matching, (trail-stop) must be called in order
  ;; to mark the trail position for the stored substitution
  ;; Also, instantiatable-vars must be set to the list of variables occurring in k
   (if (null? k)
     #t
     (if (match-terms (ref (car k)) (ref (car l)))
	 (match-termlists (cdr k) (cdr l))
	 #f)))

(define (match-terms s t)
  ;; tries to match terms s and t.
  ;; If matching succeeds, #t is returned, else #f.
  ;; Side effect: terms and trail destructively modified.
  ;; Before every unification, (trail-stop) must be called in order
  ;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
	 (if (memq s instantiatable-vars)
	     (bind s t)
	     (unbind-vars))
	 (if (variable t)
	     (unbind-vars)
	     (if (eq? (car s) (car t))
	         (match-termlists (cdr s) (cdr t))
                 (unbind-vars))))))

(define (clear-instantiations)
  ;; deletes all instantiations to the previous trailstop, thus making
  ;; those instantiations irreversible
 (if (trail-content)
     (begin (trail-previous) (clear-instantiations))
     (begin (trail-previous) #f)))


(define (try-rewrite-clauses eq-sg clauses)
  (if (pair? clauses)
      (begin
	(if (> debug-level 5)
	    (begin 
	      (display "try rewriting ")
	      (display-clause (car clauses))
	      (newline)))
	(if (not (eq? (car clauses) (subgoal-clause eq-sg)))
	    (begin
	      (if (unit-clause (car clauses))
		  (delete-subgoal-from-index (car (clause-subgoals (car clauses))) 'unit))
	      (delete-subgoals-from-path-index (clause-subgoals (car clauses)))
	      (set! clause-rewritten #f)
	      (try-rewrite-subgoals eq-sg (clause-subgoals (car clauses)))
	      (if (and clause-rewritten processing-conjecture)
		  (set-input-type (clause-input (car clauses)) 'conjecture))
	      (insert-subgoals-in-path-index (clause-subgoals (car clauses)))
	      (if (unit-clause (car clauses))
		  (insert-subgoal-in-index (car (clause-subgoals (car clauses))) 'unit))))
	;;(if clause-rewritten
	    ;;(try-rewrite-clauses eq-sg clauses)
	    (try-rewrite-clauses eq-sg (cdr clauses)))))

(define (try-rewrite-subgoals eq-sg subgoals)
  (if (pair? subgoals)
      (begin
	(try-rewrite-terms eq-sg (subgoal-terms (car subgoals)))
	(if (eq-subgoal? (car subgoals)) (check-orientation (car subgoals)))
	(try-rewrite-subgoals eq-sg (cdr subgoals)))))

(define (try-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (begin
	(set! term-rewritten #f)
	(if (not (variable (ref (car terms))))
	    (if (subgoal-ordered-eq eq-sg)
		(try-ordered-rewrite-term eq-sg terms)
		(begin
		  (try-left-rewrite-term eq-sg terms)
		  (try-right-rewrite-term eq-sg terms))))
	(if term-rewritten
	    (try-rewrite-terms eq-sg terms)
	    (try-rewrite-terms eq-sg (cdr terms))))))

(define (try-ordered-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (begin
	(if (not (variable (ref (car terms))))
	    (try-ordered-rewrite-term eq-sg terms))
	(try-ordered-rewrite-terms eq-sg (cdr terms)))))

(define (try-left-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (begin
	(if (not (variable (ref (car terms))))
	    (try-left-rewrite-term eq-sg terms))
	(try-left-rewrite-terms eq-sg (cdr terms)))))

(define (try-right-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (begin
	(if (not (variable (ref (car terms))))
	    (try-right-rewrite-term eq-sg terms))
	(try-right-rewrite-terms eq-sg (cdr terms)))))

(define (try-ordered-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (car (subgoal-terms eq-sg))) (ref (car terms)))
      (begin
	(if dewey-numbers
	    (begin
	      (display "using equation ")
	      (display-subgoal eq-sg)
	      (display " for rewriting ")
	      (display-ref (ref (car terms)))))
	(set-car! terms (unrenamed-term-copy (ref (cadr (subgoal-terms eq-sg)))))
	(if dewey-numbers
	    (begin
	      (display " to ")
	      (display-ref (ref (car terms)))
	      (newline)))
	(unbind-vars)
	(set! total-rewrites (+ total-rewrites 1))
	(set! clause-rewritten #t)
	(set! term-rewritten #t)
	(set! formula-changed #t))
      (if (complex (ref (car terms)))
	  (try-ordered-rewrite-terms eq-sg (cdr (ref (car terms)))))))

(define (try-left-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (car (subgoal-terms eq-sg))) (ref (car terms)))
      (if (kbo> (ref (car (subgoal-terms eq-sg))) (ref (cadr (subgoal-terms eq-sg))))
	  (begin
	    (if dewey-numbers
		(begin
		  (display "using equation ")
		  (display-subgoal eq-sg)
		  (display " for rewriting ")
		  (display-ref (ref (car terms)))))
	    (set-car! terms (unrenamed-term-copy (ref (cadr (subgoal-terms eq-sg)))))
	    (if dewey-numbers
		(begin
		  (display " to ")
		  (display-ref (ref (car terms)))
		  (newline)))
	    (unbind-vars)
	    (set! total-rewrites (+ total-rewrites 1))
	    (set! clause-rewritten #t)
	    (set! term-rewritten #t)
	    (set! formula-changed #t))
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-left-rewrite-terms eq-sg (cdr (ref (car terms)))))))
      (begin
	(if (complex (ref (car terms)))
	    (try-left-rewrite-terms eq-sg (cdr (ref (car terms))))))))

(define (try-right-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (cadr (subgoal-terms eq-sg))) (ref (car terms)))
      (if (kbo> (ref (cadr (subgoal-terms eq-sg))) (ref (car (subgoal-terms eq-sg))))
	  (begin
	    (if dewey-numbers
		(begin
		  (display "using equation ")
		  (display-subgoal eq-sg)
		  (display " for rewriting ")
		  (display-ref (ref (car terms)))))
	    (set-car! terms (unrenamed-term-copy (ref (car (subgoal-terms eq-sg)))))
	    (if dewey-numbers
		(begin
		  (display " to ")
		  (display-ref (ref (car terms)))
		  (newline)))
	    (unbind-vars)
	    (set! total-rewrites (+ total-rewrites 1))
	    (set! clause-rewritten #t)
	    (set! term-rewritten #t)
	    (set! formula-changed #t))
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-right-rewrite-terms eq-sg (cdr (ref (car terms)))))))
      (begin
	(if (complex (ref (car terms)))
	    (try-right-rewrite-terms eq-sg (cdr (ref (car terms))))))))

(define (unrenamed-term-copy term)
  (if (variable term)
      term
      (if (constant term)
	  (cons (car term) '())
	  (cons (car term) (unrenamed-termlist-copy (cdr term))))))

(define (unrenamed-termlist-copy terms)
  (if (pair? terms)
      (cons (unrenamed-term-copy (ref (car terms))) (unrenamed-termlist-copy (cdr terms)))
      '()))

(define (contains-external-link subgoal index-links)
  (if (pair? index-links)
      (or (contains-external-link-in-list subgoal (car index-links))
	  (contains-external-link subgoal (cdr index-links)))
      #f))

(define (contains-external-link-in-list subgoal linked-subgoals)
  (if (pair? linked-subgoals)
      (if (eq? (subgoal-clause subgoal) (subgoal-clause (car linked-subgoals)))
	  (contains-external-link-in-list subgoal (cdr linked-subgoals))
	  (car linked-subgoals))
      #f))

(define (contains-internal-link subgoal index-links)
  (if (pair? index-links)
      (or (contains-internal-link-in-list subgoal (car index-links))
	  (contains-internal-link subgoal (cdr index-links)))
      #f))

(define (contains-internal-link-in-list subgoal linked-subgoals)
  (if (pair? linked-subgoals)
      (if (eq? (subgoal-clause subgoal) (subgoal-clause (car linked-subgoals)))
	  (car linked-subgoals)
	  (contains-internal-link-in-list subgoal (cdr linked-subgoals)))
      #f))

(define (isomorphic-links? subgoal)
  (let ((index-links (unify-subgoal-with-index subgoal 'path)))
    ;;(display "checking ")
    ;;(display-subgoal subgoal)
    ;;(display ": ")
    ;;(display-subgoal-lists index-links)
    ;;(newline)
    (check-links-with-index-links subgoal (subgoal-links subgoal) index-links)))

(define (check-links-of-clauses clauses)
  (if (pair? clauses)
      (begin
	(check-links-of-subgoals (clause-subgoals (car clauses)))
	(check-links-of-clauses (cdr clauses)))))

(define (check-links-of-subgoals subgoals)
  (if (pair? subgoals)
      (begin
	(isomorphic-links? (car subgoals))
	(check-links-of-subgoals (cdr subgoals)))))

(define (check-links-with-index-links subgoal links index-links)
  (if (pair? links)
      (let ((link (link-litb (car links))) (rest-index-links (remove-link-from-link-lists (link-litb (car links)) index-links)))
	(if rest-index-links
	    (check-links-with-index-links subgoal (cdr links) rest-index-links)
	    (begin
	      (display "link ")
	      (display-subgoal link)
	      (display " (")
	      (display (subgoal-depth link))
	      (display ") of subgoal ")
	      (display-subgoal subgoal)
	      (display " not in ")
	      (display-subgoal-lists index-links)
	      (newline)
	      #f)))))

;       (if (and (pair? index-links) (contains-external-link subgoal index-links))
; 	  (begin
; 	    (display "indexed link ")
; 	    (display-subgoal (contains-external-link subgoal index-links))
; 	    (display " of subgoal ")
; 	    (display-subgoal subgoal)
; 	    (display " not in connection graph")
; 	    (newline)
; 	    #f)
; 	  (begin
; 	    ;;(print "consistent")
; 	    #t))))
	  
(define (remove-link-from-link-lists link index-links)
  (if (pair? index-links)
      (if (memq link (car index-links))
	  (cons (remq link (car index-links)) (cdr index-links))
	  (let ((rest (remove-link-from-link-lists link (cdr index-links))))
	    (if rest
		(cons (car index-links) rest)
		#f)))
      #f))	    

(define (insert-clauses-in-path-index clauses)
  (if (pair? clauses)
      (begin
	(insert-subgoals-in-path-index (clause-subgoals (car clauses)))
	(insert-clauses-in-path-index (cdr clauses)))))

(define (insert-subgoals-in-path-index subgoals)
  (if (pair? subgoals)
      (begin
	(insert-subgoal-in-index (car subgoals) 'path)
	(insert-subgoals-in-path-index (cdr subgoals)))))

(define (delete-clauses-from-path-index clauses)
  (if (pair? clauses)
      (begin
	(delete-subgoals-from-path-index (clause-subgoals (car clauses)))
	(delete-clauses-from-path-index (cdr clauses)))))

(define (delete-subgoals-from-path-index subgoals)
  (if (pair? subgoals)
      (begin
	(delete-subgoal-from-index (car subgoals) 'path)
	(delete-subgoals-from-path-index (cdr subgoals)))))

(define (get-subgoals-links subgoals)
  (if (pair? subgoals)
      (begin
	(set-subgoal-links (car subgoals) (unify-subgoal-with-index (car subgoals) 'path))
	(get-subgoals-links (cdr subgoals)))))
	

(define (has-unit-link subgoal)
  (contains-unit-link subgoal (subgoal-links subgoal)))

(define (contains-unit-link subgoal index-links)
  (if (pair? index-links)
      (or (contains-unit-link-in-list subgoal (car index-links))
	  (contains-unit-link subgoal (cdr index-links)))
      #f))

(define (contains-unit-link-in-list subgoal linked-subgoals)
  (if (pair? linked-subgoals)
      (if (unit-clause (subgoal-clause (car linked-subgoals)))
	  (car linked-subgoals)
	  (contains-unit-link-in-list subgoal (cdr linked-subgoals)))
      #f))

(define (sort-preprocessed-clauses clauses)
  (if (pair? clauses)
      (begin
	(delete-subgoals-from-path-index (clause-subgoals (car clauses)))
	(insert-new-clause-in-clauselist (car clauses) (sort-preprocessed-clauses (cdr clauses))))
      '()))

(define (check-if-path-indexes-empty)
  (check-if-path-index-empty predicate-list))

(define (check-if-path-index-empty predicates)
  (if (pair? predicates)
      (let ((pred (car predicates)))
	(if (> (car (predsym-pos-path-index pred)) 0)
	    (begin
	      (display "WARNING: pos-path-index of predicate symbol ")
	      (display (predsym-symbol pred))
	      (display " still contains ")
	      (display (car (predsym-pos-path-index pred)))
	      (display " entries! ")
	      (display-index-statistics (predsym-pos-path-index pred))
	      (newline)))
	(if (> (car (predsym-neg-path-index pred)) 0)
	    (begin
	      (display "WARNING: neg-path-index of predicate symbol ")
	      (display (predsym-symbol pred))
	      (display " still contains ")
	      (display (car (predsym-neg-path-index pred)))
	      (display " entries! ")
	      (display-index-statistics (predsym-neg-path-index pred))
	      (newline)))
	(check-if-path-index-empty (cdr predicates)))))
	
