(module rewrite
	(include "dctp.sch")
	(export (rewrite-subgoal subgoal) (unrenamed-termlist-copy terms) (match-termlists k l))
	(export instantiatable-vars)
	(import (globals "globals.scm")(kbo> kbo "kbo.scm") (unification "unification.scm") (display "display.scm")
		(check-orientation kbo "kbo.scm") (varlist-renamed? normal-form "normal-form.scm")
		(unit-eq-lits path "path.scm") (subgoal-simplifiable pruning "pruning.scm")
		(index "index.scm") (delete-subgoals-from-path-index preprocess "preprocess.scm") 
		(insert-subgoals-in-path-index preprocess "preprocess.scm")))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Variables global to this module
;; ------------------------------------------------------------

(define subgoal-rewritten #f)
(define subgoal-rewritten-at-least-once #f)
(define instantiatable-vars '())

;;;*************************************************************************
;; ------------------------------------------------------------
;; Functions for subgoal normalization
;; ------------------------------------------------------------

(define (rewrite-subgoal subgoal)
  (set! subgoal-rewritten-at-least-once #f)
  (if (not (or (subgoal-simplifiable subgoal) (ground-closed-with-index subgoal)))
      (internal-rewrite-subgoal subgoal))
  (if subgoal-rewritten-at-least-once
      (begin
	(if dewey-numbers
	    (begin
	      (display "subgoal rewritten to ")
	      (display-subgoal subgoal)
	      (newline)))
	#t)
      #f))
     
(define (internal-rewrite-subgoal subgoal)
  (set! subgoal-rewritten #f)
  (rewrite-subgoal-with-equations subgoal (unit-eq-lits))
  (if subgoal-rewritten
      (begin
	(internal-rewrite-subgoal subgoal))))

(define (rewrite-subgoal-with-equations subgoal equations)
  (if (pair? equations)
      (let ((rewritten #f))
	(set! rewritten (rewrite-subgoal-with-equation subgoal (car equations)))
	(if rewritten
	    (begin
	      ;;(display "subgoal rewritten to ")
	      ;;(display-subgoal subgoal) (display " [")
	      ;;(display (clause-unsimplified (subgoal-clause subgoal))) (display "] on depth ")
	      ;;(display (subgoal-depth subgoal))
	      ;;(newline)
	      (set! subgoal-rewritten-at-least-once #t)
	      (set! subgoal-rewritten #t)
	      (if (or (subgoal-simplifiable subgoal) (ground-closed-with-index subgoal))
		  (set! subgoal-rewritten #f)
		  (rewrite-subgoal-with-equations subgoal (cdr equations))))
	    (rewrite-subgoal-with-equations subgoal (cdr equations))))))
		
(define (rewrite-subgoal-with-equation subgoal equation)
  (if (equation? subgoal)
      (if (rewrite-equation-with-equation subgoal equation)
	  (begin
	    (check-orientation subgoal)
	    #t)
	  #f)
      (if (try-rewrite-terms equation (subgoal-terms subgoal))
	  (begin
	    (if (eq-subgoal? subgoal) (check-orientation subgoal))
	    #t)
	  #f)))
	  

(define (rewrite-equation-with-equation subgoal equation)
  (let ((left-rewritten #f)(right-rewritten #f))
    (if (subgoal-ordered-eq subgoal)
	(begin
	  (set! left-rewritten (try-rewrite-top-level-term equation (subgoal-terms subgoal) 
							   (clause-variables (subgoal-clause equation))))
	  (set! right-rewritten (try-rewrite-terms equation (cdr (subgoal-terms subgoal))))
	  (or  left-rewritten right-rewritten))
	(begin
	  (set! left-rewritten (try-rewrite-top-level-term equation (subgoal-terms subgoal)
							   (clause-variables (subgoal-clause equation))))
	  (set! right-rewritten (try-rewrite-top-level-term equation (cdr (subgoal-terms subgoal))
							   (clause-variables (subgoal-clause equation))))
	  (or  left-rewritten right-rewritten)))))

;;(or (not (top-level? (ref (car terms)) (subgoal-terms subgoal)))
;;(not (varlist-renamed? (clause-variables (subgoal-clause eq-lit)) 
;;(clause-variables (subgoal-clause eq-lit)) '()))))

(define (try-rewrite-top-level-term eq-sg terms vars)
  (if (subgoal-ordered-eq eq-sg)
      (try-ordered-rewrite-top-term eq-sg terms vars)
      (let ((left-rewritten (try-left-rewrite-top-term eq-sg terms vars))
	    (right-rewritten (try-right-rewrite-top-term eq-sg terms vars)))
	(or left-rewritten right-rewritten))))

;;;*************************************************************************
;; ------------------------------------------------------------
;; Functions for preprocess rewriting
;; ------------------------------------------------------------

(define (match-termlists k l)
  ;; tries to match termlists k and l.
  ;; If matching succeeds, #t is returned, else #f.
  ;; Side effect: termlists and trail destructively modified.
  ;; Before every matching, (trail-stop) must be called in order
  ;; to mark the trail position for the stored substitution
  ;; Also, instantiatable-vars must be set to the list of variables occurring in k
   (if (null? k)
     #t
     (if (match-terms (ref (car k)) (ref (car l)))
	 (match-termlists (cdr k) (cdr l))
	 #f)))

(define (match-terms s t)
  ;; tries to match terms s and t.
  ;; If matching succeeds, #t is returned, else #f.
  ;; Side effect: terms and trail destructively modified.
  ;; Before every unification, (trail-stop) must be called in order
  ;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
	 (if (memq s instantiatable-vars)
	     (bind s t)
	     (unbind-vars))
	 (if (variable t)
	     (unbind-vars)
	     (if (eq? (car s) (car t))
	         (match-termlists (cdr s) (cdr t))
                 (unbind-vars))))))

(define (clear-instantiations)
  ;; deletes all instantiations to the previous trailstop, thus making
  ;; those instantiations irreversible
 (if (trail-content)
     (begin (trail-previous) (clear-instantiations))
     (begin (trail-previous) #f)))


(define (try-rewrite-clauses eq-sg clauses)
  (if (pair? clauses)
      (begin
	(if (> debug-level 5)
	    (begin 
	      (display "try rewriting ")
	      (display-clause (car clauses))
	      (newline)))
	(if (not (eq? (car clauses) (subgoal-clause eq-sg)))
	    (begin
	      (if (unit-clause (car clauses))
		  (delete-subgoal-from-index (car (clause-subgoals (car clauses))) 'unit))
	      (delete-subgoals-from-path-index (clause-subgoals (car clauses)))
	      (try-rewrite-subgoals eq-sg (clause-subgoals (car clauses)))
	      (insert-subgoals-in-path-index (clause-subgoals (car clauses)))
	      (if (unit-clause (car clauses))
		  (insert-subgoal-in-index (car (clause-subgoals (car clauses))) 'unit))))
	(try-rewrite-clauses eq-sg (cdr clauses)))))

(define (try-rewrite-subgoals eq-sg subgoals)
  (if (pair? subgoals)
      (begin
	(try-rewrite-terms eq-sg (subgoal-terms (car subgoals)))
	(if (eq-subgoal? (car subgoals)) (check-orientation (car subgoals)))
	(try-rewrite-subgoals eq-sg (cdr subgoals)))))

(define (try-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (let ((rest-rewritten (try-rewrite-terms eq-sg (cdr terms))))
	(if (subgoal-ordered-eq eq-sg)
	    (or (try-ordered-rewrite-term eq-sg terms) rest-rewritten)
	    (let ((left-rewritten (try-left-rewrite-term eq-sg terms))
		  (right-rewritten (try-right-rewrite-term eq-sg terms)))
	      (or left-rewritten right-rewritten rest-rewritten))))
      #f))

(define (try-ordered-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (let ((rest-rewritten (try-ordered-rewrite-terms eq-sg (cdr terms))))
	(or (try-ordered-rewrite-term eq-sg terms) rest-rewritten))
      #f))

(define (try-left-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (let ((rest-rewritten (try-left-rewrite-terms eq-sg (cdr terms))))
	(or (try-left-rewrite-term eq-sg terms) rest-rewritten))
      #f))

(define (try-right-rewrite-terms eq-sg terms)
  (if (pair? terms)
      (let ((rest-rewritten (try-right-rewrite-terms eq-sg (cdr terms))))
	(or (try-right-rewrite-term eq-sg terms) rest-rewritten))
      #f))

(define (try-ordered-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (car (subgoal-terms eq-sg))) (ref (car terms)))
      (begin
	(if dewey-numbers
	    (begin
	      (display "using equation ")
	      (display-subgoal eq-sg)
	      (display " for rewriting ")
	      (display-ref (ref (car terms)))))
	(set-car! terms (unrenamed-term-copy (ref (cadr (subgoal-terms eq-sg)))))
	(if dewey-numbers
	    (begin
	      (display " to ")
	      (display-ref (ref (car terms)))
	      (newline)))
	(unbind-vars)
	(set! total-rewrites (+ total-rewrites 1))
	#t)
      (if (complex (ref (car terms)))
	  (try-ordered-rewrite-terms eq-sg (cdr (ref (car terms))))
	  #f)))

(define (try-left-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (car (subgoal-terms eq-sg))) (ref (car terms)))
      (if (kbo> (ref (car (subgoal-terms eq-sg))) (ref (cadr (subgoal-terms eq-sg))))
	  (begin
	    (if dewey-numbers
		(begin
		  (display "using equation ")
		  (display-subgoal eq-sg)
		  (display " for rewriting ")
		  (display-ref (ref (car terms)))))
	    (set-car! terms (unrenamed-term-copy (ref (cadr (subgoal-terms eq-sg)))))
	    (if dewey-numbers
		(begin
		  (display " to ")
		  (display-ref (ref (car terms)))
		  (newline)))
	    (unbind-vars)
	    (set! total-rewrites (+ total-rewrites 1))
	    #t)
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-left-rewrite-terms eq-sg (cdr (ref (car terms))))
		#f)))
      (if (complex (ref (car terms)))
	  (try-left-rewrite-terms eq-sg (cdr (ref (car terms))))
	  #f)))

(define (try-right-rewrite-term eq-sg terms)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (cadr (subgoal-terms eq-sg))) (ref (car terms)))
      (if (kbo> (ref (cadr (subgoal-terms eq-sg))) (ref (car (subgoal-terms eq-sg))))
	  (begin
	    (if dewey-numbers
		(begin
		  (display "using equation ")
		  (display-subgoal eq-sg)
		  (display " for rewriting ")
		  (display-ref (ref (car terms)))))
	    (set-car! terms (unrenamed-term-copy (ref (car (subgoal-terms eq-sg)))))
	    (if dewey-numbers
		(begin
		  (display " to ")
		  (display-ref (ref (car terms)))
		  (newline)))
	    (unbind-vars)
	    (set! total-rewrites (+ total-rewrites 1))
	    #t)
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-right-rewrite-terms eq-sg (cdr (ref (car terms))))
		#f)))
      (if (complex (ref (car terms)))
	  (try-right-rewrite-terms eq-sg (cdr (ref (car terms))))
	  #f)))

(define (unrenamed-term-copy term)
  (if (variable term)
      term
      (if (constant term)
	  (cons (car term) '())
	  (cons (car term) (unrenamed-termlist-copy (cdr term))))))

(define (unrenamed-termlist-copy terms)
  (if (pair? terms)
      (cons (unrenamed-term-copy (ref (car terms))) (unrenamed-termlist-copy (cdr terms)))
      '()))


(define (try-ordered-rewrite-top-term eq-sg terms vars)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (car (subgoal-terms eq-sg))) (ref (car terms)))
      (if (varlist-renamed? vars vars '())
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-ordered-rewrite-terms eq-sg (cdr (ref (car terms))))
		#f))
	  (begin
	    (if dewey-numbers
		(begin
		  (display "using equation ")
		  (display-subgoal eq-sg)
		  (display " for rewriting ")
		  (display-ref (ref (car terms)))))
	    (set-car! terms (unrenamed-term-copy (ref (cadr (subgoal-terms eq-sg)))))
	    (if dewey-numbers
		(begin
		  (display " to ")
		  (display-ref (ref (car terms)))
		  (newline)))
	    (unbind-vars)
	    (set! total-rewrites (+ total-rewrites 1))
	    #t))
      (if (complex (ref (car terms)))
	  (try-ordered-rewrite-terms eq-sg (cdr (ref (car terms))))
	  #f)))

(define (try-left-rewrite-top-term eq-sg terms vars)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (car (subgoal-terms eq-sg))) (ref (car terms)))
      (if (varlist-renamed? vars vars '())
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-left-rewrite-terms eq-sg (cdr (ref (car terms))))
		#f))
	  (if (kbo> (ref (car (subgoal-terms eq-sg))) (ref (cadr (subgoal-terms eq-sg))))
	      (begin
		(if dewey-numbers
		    (begin
		      (display "using equation ")
		      (display-subgoal eq-sg)
		      (display " for rewriting ")
		      (display-ref (ref (car terms)))))
		(set-car! terms (unrenamed-term-copy (ref (cadr (subgoal-terms eq-sg)))))
		(if dewey-numbers
		    (begin
		      (display " to ")
		      (display-ref (ref (car terms)))
		      (newline)))
		(unbind-vars)
		(set! total-rewrites (+ total-rewrites 1))
		#t)
	      (begin
		(unbind-vars)
		(if (complex (ref (car terms)))
		    (try-left-rewrite-terms eq-sg (cdr (ref (car terms))))
		    #f))))
      (if (complex (ref (car terms)))
	  (try-left-rewrite-terms eq-sg (cdr (ref (car terms))))
	  #f)))

(define (try-right-rewrite-top-term eq-sg terms vars)
  (set! instantiatable-vars (clause-variables (subgoal-clause eq-sg)))
  (if (match-terms-stop (ref (cadr (subgoal-terms eq-sg))) (ref (car terms)))
      (if (varlist-renamed? vars vars '())
	  (begin
	    (unbind-vars)
	    (if (complex (ref (car terms)))
		(try-right-rewrite-terms eq-sg (cdr (ref (car terms))))
		#f))
	  (if (kbo> (ref (cadr (subgoal-terms eq-sg))) (ref (car (subgoal-terms eq-sg))))
	      (begin
		(if dewey-numbers
		    (begin
		      (display "using equation ")
		      (display-subgoal eq-sg)
		      (display " for rewriting ")
		      (display-ref (ref (car terms)))))
		(set-car! terms (unrenamed-term-copy (ref (car (subgoal-terms eq-sg)))))
		(if dewey-numbers
		    (begin
		      (display " to ")
		      (display-ref (ref (car terms)))
		      (newline)))
		(unbind-vars)
		(set! total-rewrites (+ total-rewrites 1))
		#t)
	      (begin
		(unbind-vars)
		(if (complex (ref (car terms)))
		    (try-right-rewrite-terms eq-sg (cdr (ref (car terms))))
		    #f))))
      (if (complex (ref (car terms)))
	  (try-right-rewrite-terms eq-sg (cdr (ref (car terms))))
	  #f)))
