;; -------------------------------------------------------
;; DATA STRUCTURES AND META SYMBOLS (using <...> notation)
;; -------------------------------------------------------
;; <any> ::= any Scheme object
;; <symbol> ::= Scheme symbol
;; <integer> ::= integer
;; <natnum> ::= natural number (0,1,2,3,...)
;; <posint> ::= positive integer (1,2,3,...)
;; <arity> ::= <natnum>
;; <weight> ::= <natnum>

;; meta-commented! ;; <funcsymb> ::= #( <symbol> <arity> <weight> ... )
;; <funcsymb> ::= <symbol>
;; funcsymb will later be expanded to full functionality by vector use.

;; <free_variable> ::= < () . <any> )
;; [variables are distinguished by their addresses, i.e., CONS-cells,
;; so two variables ( () . X ) and ( () . X ) may be treated as different.
;; Normally <any> will be a symbol, the print name of the variable

;; <constant> ::= ( <funcsymb> )
;; <atom> ::= <constant> | <free_variable> 

;; by simultaneous induction:
;; <terms> ::= () | ( <term> . <terms> )
;; <term> ::= <atom> | ( <funcsymb> . <terms> )
;; <bound_variable> ::= ( <term> . <any> ) | ( <bound_variable> . <any> )
;; so a free variable is bound by setting its car to the respective term
;; <variable> ::= <free_variable> | <bound_variable>

;; <predsymb> ::= <funcsymb>
;; <sign> ::= + | -
;; <link> ::= ( <clause> . <posint> )
;; a link is a pair of a clause pointer and the 
;; subgoal index of the connected literal in that clause

;; <selected> ::= #f | #t
;; <path> ::= <subgoal> | ()
;; path is simply a pointer to the predecessor subgoal in the tableau.

;; <subgoal> ::= 
;; #( <sign> <predsymb> <terms> <links> <clause-copy> <selected> <next>) 

;; <eq-subgoal> ::= 
;; #( <sign> <predsymb> <terms> <links> <clause-copy> <selected> <next>
;;    <other-subgoals> ) 

;; <subgoals> ::= #( <subgoal> ... <subgoal> <other-subgoals> )
;; <other-subgoals> ::= #f | <subgoals>

;; <clause> ::=
;; #( <clause-copies> <generic-subgoals> <length> <nr-variables> <variables> 
;;    <constraints> <clause-type> <number> <name> <all-copies> <complexity> )

;; <clause-type> ::= axiom | hypothesis | conjecture

;; <clause-copy> ::=
;; #( <subgoals> <variables> <path> <clause> <constraints> <depth> 
;;    <entry-inference> <free-depth> <eq-trail> )

;; <formula> ::=
;; #( <clauses> <statistics> <lemmas> ) 

;; The global variable trail:
;; used for backtracking, old values of modified variables are stored there;
;; in general, trail is a cyclic structure 
;; <trail-stop> ::= ( #f . <prev-trail> )
;; <trail-point> ::= <trail-stop> | ( <any> . <prev-trail> ) 
;; <prev-trail> is the trail whose cadr is <trail-point>
;; <trail> ::= ( <trailpt> ) | ( <trailpt> . <trail> )
