; ;;;*************************************************************************
; ;; ----------------------------------
; ;; DISAGREEMENT UNIFICATION PROCEDURE
; ;; ----------------------------------

; (define (d-unify-lists k l d-pairs)
; ;; Disagreement unifies termlists k and l.
; ;; unification always succeeds, list of disagreement pairs is returned
; ;; Side effect: termlists and trail destructively modified.
; ;; Returns the list of disagreement pairs.
; ;; Before every unification, one has to call (trail-stop) in order
; ;; to mark the trail position for the stored substitution
;  (if (null? k)
;      d-pairs
;      (begin (append! (d-unify-terms (ref (car k)) (ref (car l)) '())
; 		     d-pairs)
; 	    (d-unify-lists (cdr k) (cdr l) d-pairs))))

; (define (d-unify-terms s t d-pairs)          
; ;; Disagreement unifies variable-disjoint terms s and t. 
; ;; Side effect: terms and trail destructively modified.
; ;; Returns the singleton list containing the new disagreement pair 
; ;; or otherwise '()
;  (if (not (ref-equal s t))
;      (if (variable s)
; 	 (bind s t) ; forced  substitution
;          (if (variable t)
; 	     (bind t s) ; forced  substitution
;              (set! d-pairs (cons (list t s))))))
;  d-pairs)

; (define (top-unifiable s t)
; ;; Checks whether two variable-disjoint terms are top-unifiable
;  (or (variable s) (variable t) (eq? (car s) (car t))))

(define (subgoals->eq-subgoals subgoals d-pairs clause-copy sgs)
 (set! sgs (get-subgoals (+ (vector-length subgoals) (length d-pairs)) #f))
 (fill-vector subgoals sgs (- (vector-length subgoals) 2))
 (insert-neq-subgoals d-pairs
		      (- (vector-length subgoals) 1) sgs clause-copy #f)
 sgs)

(define (insert-neq-subgoals d-pairs index sgs clause-copy sg)
;; Transfers d-pairs into the standard subgoals format 
 (set! sg (get-eq-subgoal #f))
 (set-subgoal-terms sg (car d-pairs))
 (set-subgoal-copy sg clause-copy)
 (set-subgoal-selected sg #f)
 (set-subgoal-next sg #f)
 (set-subgoal-number sg index)
; (set-subgoal-foldup-predecessor sg 'fu)
 (set-subgoal-time-stamp sg 0)
 (vector-set! sgs index sg)
 (set! d-pairs (cdr d-pairs))
 (if (pair? d-pairs)
     (insert-neq-subgoals d-pairs (+ index 1) sgs clause-copy #f)))

(define (fill-vector vec1 vec2 index)
;; fills vec2 with contents of vec1 until index position
 (if (>= index 0)
     (begin (vector-set! vec2 index (vector-ref vec1 index))
	    (fill-vector vec1 vec2 (- index 1)))))

(define (get-subgoals length sgs)
;; Gets a new list of subgoals in standard format; 
;; length should be the number of subgoals in the clause plus 1;
;; sgs may be anything, since it is only used as a local variable
 (if (eq? '() (vector-ref subgoals-stack length))
     (set! sgs (make-vector length))
     (begin (set! sgs (car (vector-ref subgoals-stack length)))
	    (vector-set! subgoals-stack length
			 (cdr (vector-ref subgoals-stack length)))))
 sgs)

(define (insert-vectors vec length index)
 (if (>= index 0)
     (begin (vector-set! vec index (make-vector length))
	    (insert-vectors vec length (- index 1)))))

(define (free-subgoals length subgoals)
;; When a list of subgoals is no more needed on backtracking, it may be freed
;; by putting it on the subgoals-stack at the respective length position
 (vector-set! subgoals-stack length
	      (cons subgoals (vector-ref subgoals-stack length))))

(define (get-eq-subgoal sg)
;; Gets a new eq-subgoal in standard format for subgoals; 
;; sg may be anything, since it is only used as a local variable
 (if (eq? '() eq-subgoal-stack)
     (begin (set! sg (make-vector 10))
	    (set-subgoal-sign sg '-)
	    (set-subgoal-pred sg 'equal)
	    (set-subgoal-links sg 'equality-links))
     (begin (set! sg (car eq-subgoal-stack))
	    (set! eq-subgoal-stack (cdr eq-subgoal-stack))))
 sg)

(define (free-eq-subgoal sg)
;; When an eq-subgoal is no more needed on backtracking, it may be freed
;; by putting it on the eq-subgoal-stack
  (set! eq-subgoal-stack (cons sg eq-subgoal-stack)))
	    
(define eq-subgoal-stack '())
(define subgoals-stack (make-vector 100 '()))
(define d-pairs '((a b) (c d)))
(define sg1 '#(- p '(a b c) links c1 sel next nr fu ts))
(define sg2 '#(- q '(u c) links c1 sel next nr fu ts))
(define subgoals (vector sg1 sg2 #f))
(define sgs 0)
(define sg 0)
(define clause-copy 1)




