(module display
	(include "setheo.sch")
	(export (display-inferences depth) (display-proof proof-output)
	        (display-subgoal subgoal)
	        (display-subgoals subgoals offset)
	        (display-clause clause)
	        (display-ref term)
	        (display-ref-list terms)
	        (display-var-name var)
	        (display-substitution trail)
	        (display-comp-subgoal subgoal))
	(import (globals "globals.scm")
		(unification "unification.scm")))

(define (display-substitution trail)
; displays the substitutions back to the previous trail stop
 (if (caar trail)
     (begin (display-substitution (cdar trail)) 
            (newline)
            (display (cdr (caar trail))) (display '/) 
            (display (caaar trail)))))
     
(define (display-var-name term)
  (if (vector? (var-name term))
      (let ((var (memq term variable-names)) nr)
	(display "X")
	(if (eq? variable-names '())
	    (begin (set! nr 1) (set! variable-names (list term nr)))
	    (if var 
		(set! nr (cadr var))
		(begin (set! nr (+ 1 (cadr variable-names)))
		  (set! variable-names (cons term (cons nr variable-names))))))
	(display nr))
      (display (var-name term))))

(define (display-ref term)
 (if (free-variable term)
     (if (vector? (var-name term))
	 (let ((var (memq term variable-names)) nr)
	   (display "X")
	   (if (eq? variable-names '())
	       (begin (set! nr 1) (set! variable-names (list term nr)))
	       (if var 
		   (set! nr (cadr var))
		   (begin (set! nr (+ 1 (cadr variable-names)))
		     (set! variable-names (cons term (cons nr variable-names))))))
	   (display nr))
	 (display (var-name term)))
     (if (variable term)
         (display-ref (ref term))
         (begin (display (functor-symbol (car term))) 
                (if (pair? (cdr term))
		    (begin (display "(") 
			   (display-ref-list (cdr term))
			   (display ")")))))))

(define (display-ref-list terms)
 (if (pair? terms)
     (begin (display-ref (ref (car terms)))
            (if (pair? (cdr terms))
	        (begin (display ",") (display-ref-list (cdr terms)))))
     '#\ ))

(define (display-subgoal-list subgoal-list)
 (if (pair? subgoal-list)
     (begin (display-subgoal (car subgoal-list)) (newline)
            (display-subgoal-list (cdr subgoal-list)))))

(define (display-eq-subgoal subgoal tag)
 (display-ref (ref (car (subgoal-terms subgoal))))
 (if (eq? '- (subgoal-sign subgoal))
     (if tag (display " != ") (display " = "))
     (if tag (display " = ") (display " != ")))
 (display-ref (ref (cadr (subgoal-terms subgoal)))))

(define (display-subgoal subgoal)
; (let (v1 v2 v3)
 (if (and infix (eq? eq-predicate (subgoal-pred subgoal)))
     (display-eq-subgoal subgoal #t)
     (begin (if (eq? '+ (subgoal-sign subgoal))
		(display "++")
		(display "--"))
	    (if (eq? (predsym-symbol (subgoal-pred subgoal)) 'equal)
		(display "equal")
		(display (predsym-symbol (subgoal-pred subgoal))))
	    (if (> (predsym-arity (subgoal-pred subgoal)) 0) 
		(begin (display "(")
		       (display-ref-list (subgoal-terms subgoal)) 
		       (display ")"))))))


(define (display-comp-subgoal subgoal)
  (if (and infix (eq? eq-predicate (subgoal-pred subgoal)))
     (display-eq-subgoal subgoal #f)
     (begin (if (eq? '- (subgoal-sign subgoal))
		(display "++")
		(display "--"))
	    (if (eq? (predsym-symbol (subgoal-pred subgoal)) 'equal)
		(display "equal")
		(display (predsym-symbol (subgoal-pred subgoal))))
	    (if (> (predsym-arity (subgoal-pred subgoal)) 0) 
		(begin (display "(")
		       (display-ref-list (subgoal-terms subgoal)) 
		       (display ")"))))))

(define (display-subgoals subgoals offset)
 (if (< offset (- (vector-length subgoals) 1))
     (begin (display-subgoal (vector-ref subgoals offset))
	    (set! offset (+ 1 offset))
	    (if (< offset (- (vector-length subgoals) 1))
		(begin (display " | ")
		       (display-subgoals subgoals offset))
		(display " ")))))

(define (display-copies all-copies new-copies)
 (if (not (eq? all-copies new-copies))
     (begin (display-subgoals (copy-subgoals (car all-copies)) 0)
;            (display-copy-constraints (copy-subgoals (car all-copies)))
	    (newline) (display "        ")
            (display-copies (cdr all-copies) new-copies))))

(define (display-clauses clauses)
 (if (pair? clauses)
     (begin (display-clause (car clauses)) (newline)
            (display-clauses (cdr clauses)))))

(define (display-clause cl)
 (newline) (display "clause ")
 (display (clause-number cl)) (display ": ") 
 (display-subgoals (clause-generic cl) 0) 
; (if (not (eq? (clause-copies cl) (clause-all-copies cl)))
;     (begin (newline) (display "copies: ") 
;            (display-copies (clause-all-copies cl) (clause-copies cl))))
 )

;(define (display-copy-constraints subgoals)
; (display-constraints 
;  (term-variables (subgoals-termlists-appended subgoals 0))))

(define (display-proof proof-output)
 (if (pair? proof-output)
     (begin (display-proof-step (caar proof-output) (cdar proof-output))
	    (display-proof (cdr proof-output)))
     (print "-----")))

(define (display-proof-step sg entry)
 (if sg
     (begin
       (display "Subgoal ") (display (subgoal-number sg)) (display ": ")
       (display-subgoal sg) (newline)
       (display "Clause ") (display (subgoal-clause-number entry))
       (display "(") (display (subgoal-number entry)) (display "): ")
       (display-subgoals (copy-subgoals (subgoal-copy entry)) 0)
;		(- (vector-length (copy-subgoals (subgoal-copy entry))) 1))
       (newline))
     (begin (display "Start clause: ") (display-clause entry) (newline)))
 (newline))
       
(define (display-proof-clauses clauses)
 (if (pair? clauses)
     (begin (if (not (eq? (clause-all-copies (car clauses))
                          (clause-copies (car clauses))))
                (display-clause (car clauses)))
            (display-proof-clauses (cdr clauses)))
     (newline)))

(define (display-trail-content trail)
; displays the contents of the entire trail
 (begin (if (cdar trail)
            (display-trail-content (cdar trail)))
        (print (caar trail) "")))

(define (display-inferences depth)
 (display "Lev. ") (display depth)
 (display ": inferences: succ")
 (if sgindex (display " match"))
 (display " fail")
 (if constraints (display " constr"))
 (display ": ")
 (display successful-inferences) 
 (if sgindex
     (begin (display " ") (display successful-matchings)))
 (display " ")
 (display failed-inferences)
 (if constraints
     (begin (display " ") (display constraint-fails)))
 (newline)
 (set! total-successful-inferences (+ total-successful-inferences 
                                      successful-inferences))
 (set! total-successful-matchings (+ total-successful-matchings
				     successful-matchings))
 (set! total-failed-inferences (+ total-failed-inferences 
				  failed-inferences))
 (set! total-constraint-fails (+ total-constraint-fails constraint-fails))
 (set! successful-inferences 0) (set! successful-matchings 0)
 (set!  failed-inferences 0) (set! constraint-fails 0)
)


(define (display-links subgoal)
 (newline) (display-links-subgoal (subgoal-links subgoal)))

(define (display-links-subgoal links)
 (if (pair? links)
     (if (eq? 'reduction (car links))
         (display-links-subgoal (cdr links))
         (begin (display (clause-number (caar links))) 
                (display " ") (display (cdar links))
                (newline) (display-links-subgoal (cdr links))))))

(define (display-constraints variables)
 (if (pair? variables)
     (begin (display (var-name (car variables)))
	    (if (not (free-variable (car variables)))
		(begin (display "/") (display-ref (car variables))))
	    (display ":  ")
	    (display-constraints1 (var-constraints (car variables)))
	    (newline)
	    (display-constraints (cdr variables)))
     ""))

(define (display-constraints1 subconstraint)
  (if subconstraint
      (begin (display-ref (vector-ref subconstraint 0))
	     (display "[")
             (display (vector-ref (vector-ref subconstraint 1) 0))
	     (display ",")
             (display (vector-ref (vector-ref subconstraint 1) 1))
	     (display "]  ")
	     (display-constraints1 (vector-ref subconstraint 2)))))
 
(define (display-v-trail point)
 (display (vector-ref (car point) 1))
 (display " ")
 (display (vector-ref (car point) 2))
 (if (eq? point v-trail)
     ""
     (begin (display " | ")
	    (display-v-trail (cdr point)))))






