(define (merge-sort unsorted-list)
 (set! c 0)
 (merge-sort-divide unsorted-list (length unsorted-list)))

(define (merge-sort-divide unsorted up2l)
; up2l must be <= (length list)
 (if (> up2l 2)
     (begin
       (let* ((l2 (quotient up2l 2)) (l1 (- up2l l2)))
	 (merge-sort-merge (merge-sort-divide unsorted l1)
			   (merge-sort-divide (list-tail unsorted l1) l2))))
     (if (= up2l 2)
	 (if (merge-sort->= (car unsorted) (cadr unsorted))
	     (list (car unsorted) (cadr unsorted))
	     (list (cadr unsorted) (car unsorted)))
	 (list (car unsorted)))))

(define (merge-sort-merge sorted1 sorted2)
; sorted1 and sorted2 must be non-empty
 (if (merge-sort->= (car sorted1) (car sorted2))
     (begin (merge-sort-merge1 (cdr sorted1) sorted2 sorted1)
            sorted1)
     (begin (merge-sort-merge1 (cdr sorted2) sorted1 sorted2)
            sorted2)))

(define (merge-sort-merge1 sorted1 sorted2 sorted1pred)
; sorted2 must be non-empty and sorted1 = (list-tail sorted1pred 1)
 (if (pair? sorted1)
     (if (merge-sort->= (car sorted1) (car sorted2))
         (merge-sort-merge1 (cdr sorted1) sorted2 sorted1)
         (begin (set-cdr! sorted1pred sorted2)
		(merge-sort-merge1 (cdr sorted2) sorted1 sorted2)))
     (set-cdr! sorted1pred sorted2)))
         
(define c 0)
(define (merge-sort->= m n) (set! c (+ c 1)) (>= m n))

(define (make-random-list length mod)
 (if (> length 0)
     (cons (random mod) (make-random-list (- length 1) mod))
     '()))

(define (random mod)
 (set! current-random1 (- (* current-random1 459782513) 1))
 (set! current-random2 (- (* current-random2 459782513) 1))
 (if (>= current-random1 current-random2)
     (modulo (abs (- current-random1 current-random2)) mod)
     (modulo (abs (+ (abs (- current-random1 current-random2)) 1)) mod)))
 
(define current-random1 344956781)
(define current-random2 444956787)

(define (test-merge-sort list-length)
 (merge-sort (make-random-list list-length 1000)) c)

; ---------------------------------------------
; Fully functional linear program for fibonacci
;
;             /
;            / (1, 1)            if n = 1
; fib2(n) = <
;            \ (r1 + r2, r1)     otherwise
;             \ where (r1, r2) = fib2(n-1)
;
; fib(n) = r1 where (r1, r2) = fib2(n)

(define (fib n) (car (fib2 n)))

(define (fib2 n)
 (if (= n 1) 
     '(1 1)
     (let ((r (fib2 (- n 1))))
          (list (+ (car r) (cadr r)) (car r)))))

; In Prolog
;
; fib2(1,1,1).
; (fib2(N,R1+R2,R1) :- fib2(N-1,R1,R2).)
; fib2(N,R,R1) :- N1 is N-1, fib2(N1,R1,R2), R is R1 + R2.
;
; fib(N,R1) :- fib2(N,R1,R2).
; ---------------------------------------------

;fib2(1) = [1,1]
;fib2(2) = [2,1]
;fib2(3) = [3,2]
;fib2(4) = [5,3]
;fib2(5) = [8,5]
;fib2(6) = [13,8]
