;; -------------------------------------------------------
;; PARSER FOR TPTP INPUT FILES
;; -------------------------------------------------------
(module parser
	(include "setheo.sch")
	(export tptp-parser tptp-scanner (display-statistics))
	(export max-clause-length max-func-arity max-pred-arity nr-clauses
		nr-non-horn nr-units nr-literals nr-eq-literals nr-predicates
		nr-propositionals nr-functors nr-constants nr-variables
		nr-singleton-variables nr-axioms nr-hypotheses
		nr-conjectures predicate-list)
	(import (globals "globals.scm")
		;(formula globals "globals.scm")
		;(nr-lemmas globals "globals.scm")
		;(sgindex globals "globals.scm") 
		(make-clause-copy copy "copy.scm")))


(define current-clause '())
(define current-clause-nr 0)
(define current-clause-variables '())
(define current-clause-var-occurrences '())
(define current-variable '())
(define current-literal '())
(define current-predicate '())
(define current-pred-nr 1)
(define current-functor '())
(define current-signature '())
(define current-functor-nr 1)
(define predicate-list '())
(define functor-list '())
(define propositional-list '())
(define constant-list '())
(define current-positive-subgoals 0)
(define current-negative-subgoals 0)
(define nr-clauses 0)               
(define nr-non-horn 0)              
(define nr-units 0)                 
(define nr-literals 0)              
(define nr-eq-literals 0)           
(define max-clause-length 0)        
(define nr-predicates 0)            
(define nr-propositionals 0)        
(define max-pred-arity 0)           
(define min-pred-arity 0)           
(define nr-functors 0)              
(define nr-constants 0)             
(define max-func-arity 0)           
(define min-func-arity 0)           
(define nr-variables 0)             
(define nr-singleton-variables 0)   
(define nr-axioms 0)   
(define nr-hypotheses 0)   
(define nr-conjectures 0)   

(define tptp-scanner
  (regular-grammar
   ((lowercase (in (#\a #\z)))
    (uppercase (in (#\A #\Z)))
    (numeric (in (#\0 #\9)))
    (special (in #\+ #\- #\_))
    (alphanum (or lowercase uppercase numeric special)))
   ("input_clause" (display-for-debug) 'input_clause)
;;   ("axiom" (display-for-debug) 
;;    (set! nr-axioms (+ nr-axioms 1)) (cons 'type (the-symbol))) 
;;   ("hypothesis" (display-for-debug) 
;;    (set! nr-hypotheses (+ nr-hypotheses 1)) (cons 'type (the-symbol))) 
;;   ("conjecture" (display-for-debug) 
;;    (set! nr-conjectures (+ nr-conjectures 1)) (cons 'type (the-symbol))) 
   ("equal" (display-for-debug)
	      (cons 'symbol (the-symbol)))
   ((: uppercase (* alphanum)) ;; (print (the-string))
			     (cons 'variable (string->symbol (the-string))))
   ((: (or lowercase numeric) (* alphanum)) ;; (print (the-string))
                             (cons 'symbol (string->symbol (the-string))))
   ("++" (display-for-debug) 'pos_sign)
   ("--" (display-for-debug) 'neg_sign)
   (#\( (display-for-debug) 'openpar)
   (#\) (display-for-debug) 'closepar)
   (#\[ (display-for-debug) 'openbracket)
   (#\] (display-for-debug) 'closebracket)
   (#\, (display-for-debug) 'colon)
   (#\. (display-for-debug) 'dot)

;   ((#\%) (context 'comment) (display-for-debug) (ignore))
;   ((context ('comment) (eol (* all))) (display-for-debug) (ignore) (context))

   ((eol (: #\% (* all))) (display-for-debug) (ignore))

   ((+ (in #\space #\tab #\newline)) (display-for-debug) (ignore))))

(define (variable-exists? variable variables)
  (if (null? variables)
      #f
      (if (eq? variable (var-name (car variables)))
	  #t
	  (variable-exists? variable (cdr variables)))))

(define (get-variable variable variables)
  (if (null? variables)
      #f
      (if (eq? variable (var-name (car variables)))
	  (car variables)
	  (get-variable variable (cdr variables)))))

(define (increment-occurences variable variables occurrences)
  (if (null? variables)
      #f
      (if (eq? variable (var-name (car variables)))
	  (set-car! occurrences (+ (car occurrences) 1))
	  (increment-occurences variable (cdr variables) (cdr occurrences)))))

(define (set-copy-of-litlist litlist copy)
  (if (null? litlist)
      #t
      (begin
	(set-subgoal-copy (car litlist) copy)
	(set-copy-of-litlist (cdr litlist) copy))))

(define (compute-clause-complexity clause)
  (let ((cx 0)
	(len (clause-length clause)))
    (do ((i 0 (+ i 1)))
	((= i len) (- cx (clause-nr-variables clause)))
      (set! cx (+ cx (subgoal-complexity (vector-ref (clause-generic clause) i)))))))

(define (term-complexity term)
 (if (or (variable term) (constant term)) 
     1
     (+ 1 (termlist-complexity (cdr term)))))
   
(define (termlist-complexity terms)
  (if (null? terms)
      0
      (+ (term-complexity (car terms)) (termlist-complexity (cdr terms)))))

(define (subgoal-complexity subgoal)
  (let ((terms (subgoal-terms subgoal)))
    (termlist-complexity terms)))

(define tptp-parser
  (lalr-grammar
   (variable symbol input_clause pos_sign neg_sign openpar closepar 
	     openbracket closebracket colon dot) 
   (formula ((clauselist) (set! nr-clauses current-clause-nr)
			  (set! nr-predicates (length predicate-list))
			  (set! nr-propositionals (length propositional-list))
			  (set! nr-functors (length functor-list))
			  (set! nr-constants (length constant-list))
			  (set! max-pred-arity (get-max-pred-arity predicate-list))
			  (set! min-pred-arity (get-min-pred-arity predicate-list))
			  (set! max-func-arity (get-max-func-arity functor-list))
			  (set! min-func-arity (get-min-func-arity functor-list))
			  (set-formula-clauses formula (reverse clauselist))))
   (clauselist
    (() '())
    ((clauselist clause) (cons clause clauselist)))
   (clause
    ((input_clause openpar clausename colon type colon literals closepar dot)
     (set! current-clause (make-vector 12))
     (set-clause-name current-clause clausename)
     (set-clause-type current-clause type)
     (set! current-clause-nr (+ current-clause-nr 1))
     (set-clause-number current-clause current-clause-nr)
     (set-clause-generic current-clause 
			 (list->vector (append literals (list #f))))
     (set-clause-length current-clause (length literals))
     (set! nr-literals (+ nr-literals (clause-length current-clause)))
     (if (> (clause-length current-clause) max-clause-length)
	 (set! max-clause-length (clause-length current-clause)))
     (if (= (clause-length current-clause) 1)
	 (set! nr-units (+ nr-units 1)))
     (set-copy-of-litlist literals current-clause)
     (set-clause-copies current-clause 
			(list (make-clause-copy (clause-generic current-clause) 
						(clause-length current-clause) 
						'() current-clause '())))
     (set-clause-all-copies current-clause (clause-copies current-clause))
     (set-clause-variables current-clause current-clause-variables)
     (set! nr-variables (+ nr-variables (length current-clause-variables)))
     (set! nr-singleton-variables 
	   (+ nr-singleton-variables 
	      (get-nr-current-singletons current-clause-var-occurrences)))
     (set-clause-nr-variables current-clause (length current-clause-variables))
     (set-clause-complexity current-clause 
			    (compute-clause-complexity current-clause))
     (set! current-clause-variables '())
     (set! current-clause-var-occurrences '())
     (if (> current-positive-subgoals 1)
	 (set! nr-non-horn (+ nr-non-horn 1)))
     (set! current-positive-subgoals 0) (set! current-negative-subgoals 0)
     (set-clause-signature current-clause current-signature)
     (set! current-signature '())
     (if (and (not eq-clause) (is-reflection current-clause))
 	 (set! eq-clause current-clause))
     current-clause))
   (literals
    ((openbracket litlist closebracket) litlist))
   (litlist
    ((litlist colon literal) (append litlist (list literal)))
    ((literal) (list literal)))
   (literal
    ((pos_sign atom)
     (set! current-literal (make-vector 8))
     (set-subgoal-sign current-literal '+)
     (set-subgoal-pred current-literal (car atom))
     (set-subgoal-terms current-literal (cdr atom))
     (set-subgoal-links current-literal '())
     (set-subgoal-selected current-literal #f)
     (set-subgoal-next current-literal #f)
     (set! current-positive-subgoals (+ current-positive-subgoals 1))
     (set-subgoal-number current-literal 
              (+ current-negative-subgoals current-positive-subgoals))
     current-literal)
    ((neg_sign atom)
     (set! current-literal (make-vector 8))
     (set-subgoal-sign current-literal '-)
     (set-subgoal-pred current-literal (car atom))
     (set-subgoal-terms current-literal (cdr atom))
     (set-subgoal-links current-literal '())
     (set-subgoal-selected current-literal #f)
     (set-subgoal-next current-literal #f)
     (set! current-negative-subgoals (+ current-negative-subgoals 1))
     (set-subgoal-number current-literal 
              (+ current-negative-subgoals current-positive-subgoals))
     current-literal))
   (atom
    ((symbol) (set! nr-propositionals (+ nr-propositionals 1)) 
	      ;; (print symbol)
	      (if (not (get-predsym symbol 0 predicate-list))
		  (begin
		    (if (or sgindex failinfo)
			(begin (set! current-predicate (make-vector 10))
			 (set-predsym-poslinks current-predicate '())
			 (set-predsym-neglinks current-predicate '())
			 (set-predsym-pos-index current-predicate #f)
			 (set-predsym-neg-index current-predicate #f)
			 (set-predsym-pos-infindex current-predicate #f)
			 (set-predsym-neg-infindex current-predicate #f))
			(set! current-predicate (make-vector 4)))
		    (set-predsym-symbol current-predicate symbol)
		    (set-predsym-arity  current-predicate 0)
		    (set-predsym-number current-predicate current-pred-nr)
		    (set-predsym-weight current-predicate 1)
		    (set! current-pred-nr (+ current-pred-nr 1))
		    (set! predicate-list (cons current-predicate predicate-list)))
		  (set! current-predicate (get-predsym symbol 0 predicate-list)))
	      (if (not (get-signature symbol 0 current-signature))
		    (set! current-signature (cons current-predicate 
						  current-signature)))
	      (if (not (member symbol propositional-list))
		  (set! propositional-list (cons symbol propositional-list)))
	      (cons current-predicate '()))
    ((symbol parameters) 
     ;; (print symbol)
     (if (eq? symbol 'equal)
	 (set! nr-eq-literals (+ nr-eq-literals 1)))
     (if (not (get-predsym symbol (length parameters) predicate-list))
	 (begin
	   (if (or sgindex failinfo)
	       (begin (set! current-predicate (make-vector 10))
		(set-predsym-poslinks current-predicate '())
		(set-predsym-neglinks current-predicate '())
	        (set-predsym-pos-index current-predicate (cons '() 0)) 
		(set-predsym-neg-index current-predicate (cons '() 0))
	        (set-predsym-pos-infindex current-predicate (cons '() 0))
	        (set-predsym-neg-infindex current-predicate (cons '() 0)))
	       (set! current-predicate (make-vector 4)))
	   (set-predsym-symbol current-predicate symbol)
	   (set-predsym-arity  current-predicate (length parameters))
	   (set-predsym-number current-predicate current-pred-nr)
           (set-predsym-weight current-predicate 1)
	   (set! current-pred-nr (+ current-pred-nr 1))
	   (if (and equality (eq? symbol 'equal))
	       (set! eq-predicate current-predicate))
	   (set! predicate-list (cons current-predicate predicate-list)))
	 (set! current-predicate (get-predsym symbol (length parameters) predicate-list)))
     (if (not (get-signature symbol (length parameters) current-signature))
	 (set! current-signature (cons current-predicate current-signature)))
     (cons current-predicate parameters)))
   (term
    ((openpar term closepar) term)
    ((symbol) 
     (if (not (get-functor symbol 0 functor-list))
	 (begin
	   (set! current-functor (make-vector 5))
	   (set-functor-symbol current-functor symbol)
	   (set-functor-arity  current-functor 0)
	   (set-functor-number current-functor current-functor-nr)
	   (set-functor-weight current-functor 1)
	   (set-functor-frequency current-functor 1)
	   (set! current-functor-nr (+ current-functor-nr 1))
	   (set! functor-list (cons current-functor functor-list)))
	 (begin
	   (set! current-functor (get-functor symbol 0 functor-list))
	   (set-functor-frequency current-functor
			(+ 1 (functor-frequency current-functor)))
	   ))
     (if (not (get-signature symbol 0 current-signature))
	 (set! current-signature (cons current-functor current-signature)))
     (if (not (member symbol constant-list))
	 (set! constant-list (cons symbol constant-list)))
     (list current-functor))
    ((variable)
     (if (variable-exists? variable current-clause-variables)
	 (begin (set! current-variable 
		      (get-variable variable current-clause-variables))
		(increment-occurences variable current-clause-variables 
				      current-clause-var-occurrences))
	 (begin
	   (set! current-variable (cons '() (cons #f variable)))
	   (set! current-clause-var-occurrences
		 (cons 1 current-clause-var-occurrences))
	   (set! current-clause-variables 
		 (cons current-variable current-clause-variables))))
      current-variable)
    ((symbol parameters) 
     (if (not (get-functor symbol (length parameters) functor-list))
	 (begin
	   (set! current-functor (make-vector 4))
	   (set-functor-symbol current-functor symbol)
	   (set-functor-arity  current-functor (length parameters))
	   (set-functor-number current-functor current-functor-nr)
	   (set-functor-weight current-functor 1)
	   (set! current-functor-nr (+ current-functor-nr 1))
	   (set! functor-list (cons current-functor functor-list)))
	 (set! current-functor (get-functor symbol (length parameters) functor-list)))
     (if (not (get-signature symbol (length parameters) current-signature))
	 (set! current-signature (cons current-functor current-signature)))
     (append (list current-functor) parameters)))
   (parameters
    ((openpar termlist closepar) termlist))
   (termlist
    ((termlist colon term) (append termlist (list term)))
    ((term) (list term)))
   (type
    ((symbol) 
     (cond ((eq? symbol (string->symbol "axiom"))
            (set! nr-axioms (+ nr-axioms 1)) 'axiom)
	   ((eq? symbol (string->symbol "lemma"))
	    (set! nr-lemmas (+ nr-lemmas 1)) 'lemma)
	   ((eq? symbol (string->symbol "hypothesis"))
	    (set! nr-hypotheses (+ nr-hypotheses 1)) 'hypothesis)
	   ((eq? symbol (string->symbol "conjecture"))
	    (set! nr-conjectures (+ nr-conjectures 1)) 'conjecture))
	   ))
   (clausenamepars
    ((openpar clausenametl closepar) clausenametl))
   (clausenametl
    ((clausenametl colon clausename) (append clausenametl (list clausename)))
    ((clausename) (list clausename)))
   (clausename
    ((openpar clausename closepar) clausename)
    ((symbol) symbol)
    ((variable) variable)
    ((symbol clausenamepars) 
     (append (list symbol) clausenamepars)))))


(define (get-nr-current-singletons occurrences-list)
  (if (null? occurrences-list)
      0
      (if (= (car occurrences-list) 1)
	  (+ (get-nr-current-singletons (cdr occurrences-list)) 1)
	  (get-nr-current-singletons (cdr occurrences-list)))))

(define (get-max-pred-arity predlist)
  (if (null? predlist)
      '-
      (let ((tmp (get-max-pred-arity (cdr predlist))))
	(if (or (eq? tmp '-) (< tmp (predsym-arity (car predlist))))
	    (predsym-arity (car predlist))
	    tmp))))

(define (get-min-pred-arity predlist)
  (if (null? predlist)
      '-
      (let ((tmp (get-min-pred-arity (cdr predlist))))
	(if (or (eq? tmp '-) (> tmp (predsym-arity (car predlist))))
	    (predsym-arity (car predlist))
	    tmp))))

(define (get-max-func-arity funclist)
  (if (null? funclist)
      '-
      (let ((tmp (get-max-func-arity (cdr funclist))))
	(if (or (eq? tmp '-) (< tmp (functor-arity (car funclist))))
	    (functor-arity (car funclist))
	    tmp))))

(define (get-min-func-arity funclist)
  (if (null? funclist)
      '-
      (let ((tmp (get-min-func-arity (cdr funclist))))
	(if (or (eq? tmp '-) (> tmp (functor-arity (car funclist))))
	    (functor-arity (car funclist))
	    tmp))))

(define (get-predsym symbol arity predicate-list)
  (if (null? predicate-list)
      #f
      (if (and (eq? symbol (predsym-symbol (car predicate-list)))
	       (eq? arity (predsym-arity (car predicate-list))))
	  (car predicate-list)
	  (get-predsym symbol arity (cdr predicate-list)))))

(define (get-functor symbol arity functor-list)
  (if (null? functor-list)
      #f
      (if (and (eq? symbol (functor-symbol (car functor-list)))
	       (eq? arity (functor-arity (car functor-list))))
	  (car functor-list)
	  (get-functor symbol arity (cdr functor-list)))))

(define (get-signature symbol arity signature-list)
  (if (null? signature-list)
      #f
      (if (and (eq? symbol (signature-symbol (car signature-list)))
	       (eq? arity (signature-arity (car signature-list))))
	  (car signature-list)
	  (get-signature symbol arity (cdr signature-list)))))

(define (display-statistics)
  (display #"Syntax:\t")
  (display #"Number of clauses:\t")
  (display nr-clauses)
  (display " ( ")
  (display nr-non-horn)
  (display " non-Horn; ")
  (display nr-units)
  (display " unit)")
  (newline)
  (display #"\t")
  (display #"Number of literals:\t")
  (display nr-literals)
  (display " ( ")
  (display nr-eq-literals)
  (display " equality)")
  (newline)
  (display #"\t")
  (display #"Maximal clause size:\t")
  (display max-clause-length)
  (newline)
  (display #"\t")
  (display #"Number of predicates:\t")
  (display nr-predicates)
  (display " ( ")
  (display nr-propositionals)
  (display " propositional; ")
  (display min-pred-arity)
  (display "-")
  (display max-pred-arity)
  (display " arity)")
  (newline)
  (display #"\t")
  (display #"Number of functors:\t")
  (display nr-functors)
  (display " ( ")
  (display nr-constants)
  (display " constant; ")
  (display min-func-arity)
  (display "-")
  (display max-func-arity)
  (display " arity)")
  (newline)
  (display #"\t")
  (display #"Number of variables:\t")
  (display nr-variables)
  (display " ( ")
  (display nr-singleton-variables)
  (display " singleton)")
  (newline))

(define (display-for-debug)
  ;;(display (the-string))
  #t)

(define (reset-parser)
  (set! current-clause '())
  (set! current-clause-nr 0)
  (set! current-clause-variables '())
  (set! current-clause-var-occurrences '())
  (set! current-variable '())
  (set! current-literal '())
  (set! current-predicate '())
  (set! current-pred-nr 1)
  (set! current-functor '())
  (set! current-functor-nr 1) 
  (set! current-signature '())
  (set! predicate-list '())
  (set! functor-list '())
  (set! propositional-list '())
  (set! constant-list '())
  (set! current-positive-subgoals 0)
  (set! nr-clauses 0)
  (set! nr-non-horn 0)
  (set! nr-units 0)
  (set! nr-literals 0)
  (set! nr-eq-literals 0)
  (set! max-clause-length 0)
  (set! nr-predicates 0)
  (set! nr-propositionals 0)
  (set! max-pred-arity 0)
  (set! min-pred-arity 0)
  (set! nr-functors 0)
  (set! nr-constants 0)
  (set! max-func-arity 0)
  (set! min-func-arity 0)
  (set! nr-variables 0)
  (set! nr-singleton-variables 0))

(define (is-reflection clause)
 (and equality
      (= (clause-length clause) 1)
      (let ((sg (vector-ref (clause-generic clause) 0)) s t)
	(and (eq? (predsym-symbol (subgoal-pred sg)) 'equal)
	     (eq? (subgoal-sign sg) '+)
;      (print "PASSED")
	     (begin (set! s (car (subgoal-terms sg)))
		    (set! t (cadr (subgoal-terms sg)))
		    #t)
	     (not (vector? (car s)))
	     (eq? s t)))))
