;;;***********************************************************************
;; ----------------------
;; EQUAL PREDECESSOR FAIL
;; ----------------------
(module pruning
	(include "setheo.sch")
	(export
	 (no-equal-predecessor sg sg-terms predsg) 
	 (make-regularity-constraints-clause entry subgoals predsg index max-index) 
	 (ref-equal s t)
;	 (make-folded-vs-subgoals-constraints sg subgoals index max-index)
	 )
	(import
	 (globals "globals.scm")
	 (unification "unification.scm")
	 (constraints "constraints.scm")
	 (display "display.scm")
	 ))

(define (no-equal-predecessor sg sg-terms predsg)
 (if 
  (and predsg (not (free-variable-in sg-terms)))
  (if (and (eq? (subgoal-sign sg) (subgoal-sign predsg))
           (eq? (subgoal-pred sg) (subgoal-pred predsg))
           (ref-equals sg-terms (subgoal-terms predsg)))
      #f
      (no-equal-predecessor sg sg-terms (subgoal-predecessor predsg)))
  #t))

(define (ref-equals k l)
;; checks whether termlists k and l are equal 
 (if (null? k)
     #t
     (if (ref-equal (ref (car k)) (ref (car l)))
	 (ref-equals (cdr k) (cdr l))
	 #f)))

(define (ref-equal s t)          
;; checks whether terms s and t are equal 
 (if (eq? s t)
     #t
     (if (or (variable s) (variable t))
         #f
         (and (eq? (car s) (car t)) 
              (ref-equals (cdr s) (cdr t))))))

(define (free-variable-in terms)
 (if (pair? terms)
     (if (free-variable (car terms))
         #t
         (free-variable-in (cdr terms)))
     #f))
 
;;;***********************************************************************
;; ----------------------
;; REGULARITY CONSTRAINTS
;; ----------------------

(define (make-regularity-constraints-clause entry subgoals predsg index
					    max-index)
; generates regularity-constraints for all subgoals/literals of a clause
; output: #t if none of the generated constraints is violated, otherwise #f
; side effects on the constraints are stored on the value trail
;  (print "make-regularity-constraints-clause entered:")
;  (display-subgoal (vector-ref subgoals index))
;  (display-subgoal entry)
;  (newline)
(if (= index max-index) ; constraints for all subgoals in clause generated
    #t
    (if (or (and (eq? entry (vector-ref subgoals index)) ; head literal
		 regularity-only-for-subgoals)
	    (let ((result 
		   (make-regularity-constraints (vector-ref subgoals index) 
		       	(subgoal-terms (vector-ref subgoals index)) predsg))
		  )
	;         (display "RESULT: ") (display result) (newline)
	      result))
	(make-regularity-constraints-clause entry subgoals predsg (+ 1 index)
					    max-index)
	#f))) ; some regularity constraint is violated

(define (make-regularity-constraints sg sg-terms predsg)
; (let (l1 l2)
;    (newline) (print "make-regularity-constraints entered:" )
;    (set! l1 sg) (set! l2 predsg)
;    (display-subgoal sg) 
;    (if predsg (begin (newline) (display-subgoal predsg))))
;  (repl)
 (if predsg 
     (if (and (eq? (subgoal-sign sg) (subgoal-sign predsg)) 
	      (eq? (subgoal-pred sg) (subgoal-pred predsg))
	      (or (not (make-disequation-constraint sg-terms
						    (subgoal-terms predsg)))
	          (and (eq? eq-predicate (subgoal-pred sg))
		       (not (make-disequation-constraint (reverse sg-terms)
						     (subgoal-terms predsg))))))
	 #f
	 (if strong-regularity
	     (begin ;(print "foldup")
	       (make-regularity-constraints sg sg-terms
					    (subgoal-foldup-predecessor predsg)))
	     (begin ;(print "no foldup")
	       (make-regularity-constraints sg sg-terms
					    (subgoal-predecessor predsg)))))
     #t))

; (define (make-folded-vs-subgoals-constraints sg subgoals index max-index)
; ; subgoals := #f | #(subgoal ... subgoal subgoals)
; ; outputs #f if one of the constraints is already violated
; ; side effects on the constraints are stored on the value trail
;  (if (< index max-index)
;      (if
;       (and (not (subgoal-selected (vector-ref subgoals index)))
; 	   (not (eq? (subgoal-sign sg) ; since subgoal sign not yet changed
; 		     (subgoal-sign (vector-ref subgoals index))))
; 	   (eq? (subgoal-pred sg) (subgoal-pred (vector-ref subgoals index)))
; 	   (not (make-disequation-constraint (subgoal-terms sg)
; 				  (subgoal-terms (vector-ref subgoals index)))))
;       #f
;       (make-folded-vs-subgoals-constraints sg subgoals (+ index 1) max-index))))

; (define (make-unit-subsumption-constraint subgoal unit-terms constraints)
; ; generates subsumption constraint for a subgoal whose subgoal list unifies
; ; (with unifier sigma) with unit-terms. 
; ; Side effect: integrates constraint with variables not in subgoal made
; ; universal into the constraints of the variables of the subgoal.
; ; Output: #f if sigma is the empty substitution, otherwise #t
;  (if (unify-lists-stop (subgoal-terms subgoal) unit-terms)
;      (if (trail-content) ; constraint has length > 0
; 	 (begin (set! constraints   ; new header is made
; 		      (cons (vector 0 #f #f) constraints))
; 		(extract-constraint-static 1 (car constraints) 
; 					   constraints (cdr constraints))
; 		(trail-previous) #t)
; 	 (begin (trail-previous) #f)) ; constraint is immediately violated
;      #t)) ; term lists not unifiable: no constraint generated

; (define (extract-constraint-static length c-header constraints previous)
; ; makes free all variables back to the previous trailstop
; ; and pushes the bindings as subconstraints on the list constraints
; ; (caar trail) is current variable, (car constraints) is new subconstraint
; ; previous is the previous subconstraint of the respective variable
;  (set! constraints
;        (cons (vector (ref (var-binding (caar trail))) c-header previous)
; 	     constraints))
;  (set-car! (caar trail) '()) ; var. binding from protocol unification undone 
;  (trail-previous) ; moves to previous trail-point 
;  (if (trail-content) ; next binding to be processed
;      (extract-constraint-static (+ length 1) c-header constraints 
; 				(cdr constraints))
;      ; below: all bindings processed; now constraint length inserted in header
;      (vector-set! c-header 0 (+ (vector-ref c-header 0) length))
;      ))






