(module setheo
	(main call-setheo)    ;; this is the main procedure
	(include "setheo.sch")
	(import (solve "solve.scm")
		(pruning "pruning.scm")
		(unification "unification.scm")
		(constraints "constraints.scm")
		(copy "copy.scm")
		(globals "globals.scm")
		(misc "misc.scm")
		(connection-graph "connection-graph.scm")
		(parser "parser.scm")
		(input "input.scm")
		(index "index.scm")
		(display "display.scm")
		(grounding "grounding.scm")))
      
(define (call-setheo argv) 
   (setheo (cdr argv)))

(define (setheo args)
 (update-defaults args)
 (vector-set! foldup-root 9 0)
 (set! filename (car (last args)))
 (if (not (file-exists? filename))
     (begin (print "ERROR: file does not exist") (exit 1)))
 (read-tptp filename)
 (set-formula-clauses formula (remove-lemmas (formula-clauses formula) '()))
 (if (null? (formula-clauses formula))
     (begin (newline) (display "     ***** INPUT IS EMPTY")
     (if error-on-empty (print " *****") (print ": SATISFIABLE *****"))
     (exit 0)))
 (if (< max-lemmas 0)
     (set! max-lemmas (min 100
	   (max 10 (quotient (* (- nr-clauses nr-lemmas) (- max-lemmas)) 100)))))
 (set! printlemmas max-lemmas)
 (if (< uselemmas 0)
     (set! uselemmas (min 100
	   (max 10 (quotient (* (- nr-clauses nr-lemmas) (- uselemmas)) 100)))))
 (if sgindex (in-lemma-index-lemmas (formula-lemmas formula)))
 (if lemmas
     (let ((old-lemmas lemmas))
       (set! lemmas #f)
       (in-lemma-index-lemmas (formula-clauses formula))
       (set! lemmas old-lemmas)))
 (if (and (> maxinput-lemmas 0) (> nr-lemmas 0))
     (insert-new-clauses (reverse lemmas) maxinput-lemmas))
 (if clause-reordering (sort-clauses (formula-clauses formula)))
 (if equality (if (= nr-eq-literals 0)
                  (set! equality #f)
                  (begin
		    (set! subgoals-stack
			  (make-vector
			   (+ (max max-pred-arity max-func-arity)
			      max-clause-length 3) '()))
		    (display "Equality handling") (newline))))
; (print predicate-list) (print eq-predicate)
 (make-connection-graph (formula-clauses formula) (formula-clauses formula))
 (if test-grounding (grounding (formula-clauses formula)))
 (if equality (make-eq-links (formula-clauses formula)))
; (display "length eq-links: ") (print (length eq-links))
; (set! start-clauses (formula-clauses formula))
 (if (or norel (all-negated-conjectures (formula-clauses formula)))
     (begin (newline) (display "all negated are conjectures")
     (set! start-clauses (negated-clauses (formula-clauses formula)))
     )
     (begin (newline) (display "some negated are no conjectures")
     (set! start-clauses (conjecture-clauses (formula-clauses formula)))
     ))
 (if signature-bound
     (begin (update-pred-numbers predicate-list)
            (set! sig-vector (make-vector (+ nr-functors nr-predicates 1) #f))
            (set! sig-trail (make-vector (+ nr-functors nr-predicates)))
            (if (> nr-axioms 0)
                (begin (free-signature-hyp-con (formula-clauses formula))
                   (remove-free-signature-clauses (formula-clauses formula))))
     ))
 (newline)
; (if eq-clause (display-clause eq-clause))
 (search-proof start-clauses)
)

;;;*************************************************************************
;; ----------------------------------------------
;; START CLAUSE SELECTION AND ITERATIVE DEEPENING
;; ----------------------------------------------

(define (all-negated-conjectures clauses)
 (if (pair? clauses)
     (if (or (has-positive-literal (clause-generic (car clauses))
                                   (- (clause-length (car clauses)) 1))
             (eq? (clause-type (car clauses)) 'conjecture))
         (all-negated-conjectures (cdr clauses))
         #f)
     #t))
         
(define (has-positive-literal subgoals index)
 (if (>= index 0)
     (if (eq? (subgoal-sign (vector-ref subgoals index)) '+) 
         #t
         (has-positive-literal subgoals (- index 1)))
     #f))

(define (conjecture-clauses clauses)
 (if (pair? clauses)
     (if (eq? (clause-type (car clauses)) 'conjecture)
         (cons (car clauses) (conjecture-clauses (cdr clauses)))
         (conjecture-clauses (cdr clauses)))
     '()))

(define (negated-clauses clauses)
 (if (pair? clauses)
     (if (not (has-positive-literal (clause-generic (car clauses))
                                    (- (clause-length (car clauses)) 1)))
         (cons (car clauses) (negated-clauses (cdr clauses)))
         (negated-clauses (cdr clauses)))
     '()))

(define (free-signature-hyp-con clauses)
 (if (pair? clauses)
     (begin (if (or (eq? (clause-type (car clauses)) 'conjecture)
                    (eq? (clause-type (car clauses)) 'hypothesis))
                (free-signature-weights (clause-signature (car clauses))))
            (free-signature-hyp-con (cdr clauses)))))

(define (remove-free-signature-clauses clauses)
 (if (pair? clauses)
     (begin (set-clause-signature (car clauses)
                    (remove-free-signature (clause-signature (car clauses))))
            (remove-free-signature-clauses (cdr clauses)))))

(define (remove-free-signature sigs)
 (if (pair? sigs)
     (if (> (signature-weight (car sigs)) 0)
         (cons (car sigs) (remove-free-signature (cdr sigs)))
         (remove-free-signature (cdr sigs)))
     '()))

(define (remove-lemmas clauses lemmas)
 (if (pair? clauses)
     (if (eq? (clause-type (car clauses)) 'lemma)
	 (remove-lemmas (cdr clauses) (cons (car clauses) lemmas))
	 (cons (car clauses) (remove-lemmas (cdr clauses) lemmas)))
     (begin (set-formula-lemmas formula lemmas)
	    '())))

(define (search-proof clauses) 
 (if (pair? clauses)
     (begin ;(display-clause (car clauses))
            (extension #f (car clauses) #f depth 0 2 #f)
            (search-proof (cdr clauses)))
     (begin (display-inferences depth)
       (if bound-fail
	   (begin (set! bound-fail #f)
; 		       (if (and failinfo (> failinfo-pruning 0))
; 			   (begin (display "# of failinfo-prunings: ")
; 				  (print failinfo-pruning)
; 				  (set! failinfo-pruning 0)))
             (if (= depth final-depth)
		 (begin (newline)
		   (display "     ***** BOUND REACHED *****")
		   (newline) ; (exit 0)
		   (if lemmas (begin (newline); (print "Best unit lemmas:")
				(display-ulemmas (reverse! lemmas) 1)))
		   (newline))
		 (begin (set! depth (+ depth 1))
			(if ulemmas
			    (begin ; (delete-ulemmas predicate-list)
			      ; (print "ULEMMAS WILL BE INSERTED!")
			      (insert-new-clauses (reverse lemmas)
						  uselemmas)))
			(search-proof start-clauses))))
	   (begin (newline) 
	     (if (and extension-limit (> extension-steps extension-limit))
		 (begin (display "     ***** EXTENSION LIMIT REACHED *****")
		   (if lemmas (begin (newline) (newline)
;				     (print "Best unit lemmas:")
				(display-ulemmas (reverse! lemmas) 1))))
		 (display "     ***** FORMULA IS SATISFIABLE *****"))
	     (newline) (newline)
	     (display "Total successful/failed inferences(constraints): ") 
	     (display total-successful-inferences) (display "/")
	     (display total-failed-inferences) (display " (")
	     (display total-constraint-fails) (display ")") (newline)
;		   (exit 0)
	     #f)))))

;;;END *********************************************************************

(define (delete-ulemmas predicates)
  (if (pair? predicates)
      (begin (set-predsym-poslinks (car predicates) '())
             (set-predsym-neglinks (car predicates) '())
	     (delete-ulemmas (cdr predicates)))))

	     