;;;*************************************************************************
;; --------------------------------------
;; MODEL ELIMINATION PROOF PROCEDURE
;; --------------------------------------
(module solve
	(include "setheo.sch")
	(export ;(solve sg links depth inference nr-sgs)
		(extension sg clause entry depth inference nr-sgs eq-tag)
		(ref-int int-term) 
		(free-signature-weights sigs)
		(update-pred-numbers preds)
		(insert-lemma sg sign-tag)
		(insert-new-clauses lemmas nr)
		(display-ulemmas lemmas nr)
		)
	(import (globals "globals.scm") (unification "unification.scm")
		(parser "parser.scm") (copy "copy.scm")
		(pruning "pruning.scm") (constraints "constraints.scm")
		(display "display.scm")
		(comp-in-lemma-index-subgoal index "index.scm")
		(find-negmatching-lemma index "index.scm")
		(in-infindex-subgoal index "index.scm")
		(is-in-infindex-subgoal? index "index.scm")
		(in-lemma-index-subgoal index "index.scm")
		(print-predicate-lemma-indices index "index.scm")
		(add-unit-links connection-graph "connection-graph.scm")
		))

(define (solve sg links depth inference nr-sgs copies)
 (if 
  (and (or (not extension-limit) (<= extension-steps extension-limit))
       (not (null? links)))
  (if (and sgindex copies
           (begin ; insert unit-lemma if last subgoal
	     (if (= nr-sgs 1)
		 (begin
		   ;(display-subgoal sg) (display " inserted, nr-sgs= ")
		   ;(display nr-sgs) (newline)
		   (set! glob-inf inference)
		   (set! lemmatype 0)
		   (in-lemma-index-subgoal sg)
		   ))
	     #t)
;           (or (is-in-infindex-negsubgoal? sg)
;	       (let ((result
		      (find-negmatching-lemma sg)
; 		      ))
; 		 (if result (comp-in-lemma-index-subgoal sg))
; 		 result))
; 	   (begin (newline) (display "Lemma matching on ")
; 		  (display-subgoal sg) (display " (")
; 		  (display (subgoal-clause-number sg)) (display ".")
; 		  (display (subgoal-number sg)) (display ")")
; 		  #t)
      )
      (let ((old-global-subgoals global-subgoals) (new-inf inference)
            new-sg new-links new-depth (new-nr-sgs (- nr-sgs 1)))
;         (if proof-output
; 	    (set! proof-output (cons (list (subgoal-clause-number sg)
; 					   (subgoal-number sg)
; 					   (- depth (subgoal-depth sg))
; 					   0)
; 				     proof-output)))
	(set! successful-inferences (+ successful-inferences 1))
	(set! successful-matchings (+ successful-matchings 1))
	(set! nr-matchings (+ 1 nr-matchings))
        (next-subgoal-call new-inf new-sg new-links new-depth 
			 old-global-subgoals new-nr-sgs)
; 	        (newline) (display "Backtracking lemma step on ")
; 	(display (subgoal-clause-number sg)) (display ".")
; 	(display (subgoal-number sg)) (display ")")
; 	(if proof-output (set! proof-output (cdr proof-output)))
	(set! nr-matchings (- nr-matchings 1)))

      (begin 
	(if (eq? (car links) 'reduction) 
	    (if copies
		(if foldup
		    (reduction-steps sg (subgoal-foldup-predecessor sg)
				     inference nr-sgs)
		    (reduction-steps sg (subgoal-predecessor sg)
				     inference nr-sgs)))
	    (if (and equality (if (and units-double antilemmas) (not copies) #t)
		     (or (not (pair? (car links)))
			 (begin
			   (set! current-subgoal ; clause is not reflection
				 (vector-ref (copy-subgoals
				       (car (clause-copies (caar links))))
				       (cdar links)))
			   (or (not (eq? eq-predicate
					 (subgoal-pred current-subgoal)))
			       (eq? '- (subgoal-sign current-subgoal))
			       (not (ref-equal
				     (car (subgoal-terms current-subgoal))
				     (cadr (subgoal-terms current-subgoal))))))))
		(if (and (eq? (subgoal-pred sg) eq-predicate)
			 (eq? (subgoal-sign sg) '-))
		    (if (eq? (car links) 'decomposition)
			(if (and (complex (ref (car (subgoal-terms sg))))
				 (complex (ref (cadr (subgoal-terms sg))))
				 (eq? (car (ref (car (subgoal-terms sg))))
				      (car (ref (cadr (subgoal-terms sg)))))
				 )
			    (extension sg eq-clause ; decomposition
			       (vector-ref
			        (copy-subgoals (car (clause-copies eq-clause)))
			        0)
			       depth inference nr-sgs 3))
			(begin 
			  (extension sg (caar links) ; d-extension for ~eq-subgoal
			   (vector-ref 
			    (copy-subgoals (car (clause-copies (caar links))))
			    (cdar links))
			   depth inference nr-sgs 1)
			  (if antilemma-test
			      (begin
				(if (pair? actual-copy)
				    (set! copies (cons actual-copy copies)))
				(extension sg (caar links) ; then symmetric case
				  (vector-ref (copy-subgoals
					       (car (clause-copies (caar links))))
					      (cdar links))
				  depth inference nr-sgs 2)))))
		    (extension sg (caar links) ; d-extension for other literal
		      (vector-ref 
		       (copy-subgoals (car (clause-copies (caar links))))
		       (cdar links))
		      depth inference nr-sgs 0))
		(if (and (not (eq? (car links) 'decomposition)) copies
		    (if (and units-double antilemmas equality)
			(= 1 (clause-length (caar links))) #t))
		    (extension sg (caar links) ; standard extension step
			       (vector-ref 
				(copy-subgoals (car (clause-copies (caar links))))
				(cdar links))
			       depth inference nr-sgs #f))))
	(if antilemma-test
	    (begin (if (and copies (pair? actual-copy))
		       (set! copies (cons actual-copy copies)))
	      (solve sg (cdr links) depth inference nr-sgs copies))
	    (begin (set! constraint-fails (+ constraint-fails 1))
		   (free-copies copies)))))
  (free-copies copies)))

(define (free-copies copies)
 (if (pair? copies)
     (begin (set-clause-copies (copy-clause (caar copies)) (car copies))
	    (free-copies (cdr copies)))))

(define (reduction-steps sg predsg inference nr-sgs)
 (if 
  predsg
  (begin
   (if 
    (and (not (eq? (subgoal-sign sg) (subgoal-sign predsg)))
         (eq? (subgoal-pred sg) (subgoal-pred predsg))
         (or (not matings-pruning) ; begin of matings pruning
             (> (subgoal-clause-number sg)
                (subgoal-clause-number (subgoal-next predsg)))
             (and (= (subgoal-clause-number sg)
                     (subgoal-clause-number (subgoal-next predsg)))
                  (>= (subgoal-number sg)
                     (subgoal-number (subgoal-next predsg))))) ; end mat. pr.
         (or (and (unify-lists-stop (subgoal-terms sg) (subgoal-terms predsg))
	          (or (and (not constraints)
			   (begin
			     (if (or foldup index-foldup)
				 (begin (v-trail-stop)
					(update-time-stamp predsg inference)))
			     #t))
  		      (begin (v-trail-stop) ; backtracking marker
			(set! actual-c-stack c-stack)
			(if (propagate-substitution) ; constr. not viol.
			    (begin
			      (if (or foldup index-foldup)
				  (update-time-stamp predsg inference))
			      #t);     below: shallow constraint and unfication
			    (begin (set! c-stack actual-c-stack) ; backtracking
				(undo-v-trail) (unbind-vars) 
				(set! constraint-fails (+ constraint-fails 1))
				#f))))) ; next line ends shallow backtracking
             (begin (set! failed-inferences (+ failed-inferences 1)) #f)))
    (let ((old-global-subgoals global-subgoals) (new-inf inference)
          (old-c-stack actual-c-stack) new-sg new-links new-depth)
     (set! successful-inferences (+ successful-inferences 1))
     (set! successful-reduction-steps (+ successful-reduction-steps 1))
     (set! nr-reduction-steps (+ nr-reduction-steps 1))
     (if failinfo (set-subgoal-selected sg 1))
     (next-subgoal-call new-inf new-sg new-links new-depth
			old-global-subgoals (- nr-sgs 1))
     ;; deep backtracking:
     (set! nr-reduction-steps (- nr-reduction-steps 1))
     (if (or constraints foldup index-foldup) (undo-v-trail))
     (if constraints (set! c-stack old-c-stack))
     (if antilemmas
	 (begin (set! antilemma-test #t)
		(set! antilemma-test (extract-disequation-constraint-relevant)))
	 (unbind-vars))
     (if debug (begin
		 (display "Backtracking reduction step on ")
		 (display-subgoal sg) (display " (")
		 (display (subgoal-clause-number sg)) (display ".")
		 (display (subgoal-number sg)) (display ")")
		 (display " on depth ")
		 (display (- depth (subgoal-depth sg))) (newline)
;		 (if proof-output (set! proof-output (cdr proof-output)))
		 ))
      ))
   (if (or (not antilemmas) antilemma-test)
       (if foldup
	   (reduction-steps sg (subgoal-foldup-predecessor predsg)
			    inference nr-sgs)
	   (reduction-steps sg (subgoal-predecessor predsg)
			    inference nr-sgs))
       (begin (set! constraint-fails (+ constraint-fails 1)))))))
     
(define (extension sg clause entry r-depth inference nr-sgs eq-tag)
 (if (and (or (not extension-limit) (<= extension-steps extension-limit))
          (check-bounds clause r-depth)
          (or (not signature-bound)
              (begin (set! old-free-sig free-sig)
                     (set! old-sig-trail-position sig-trail-position)
                     (if (sig-bound (clause-signature clause))
                         #t
                         (begin (set! free-sig old-free-sig) 
                                (set! bound-fail #t) #f))))
          (or (begin (set! actual-depth r-depth) (set! d-pairs '())
                     (set! actual-copy (car (clause-copies clause)))
		     (set! old-copy-subgoals (copy-subgoals actual-copy))
		     (not sg)) ; top clause entered: no unification/regularity
              (and
	       (cond ((or (not eq-tag) (if (<= r-depth 0)
					   (begin (set! bound-fail #t) #t)
					   #f))
		      (extension-unify-lists-stop (subgoal-terms entry) 
						      (subgoal-terms sg)))
		     ((= eq-tag 0) (trail-stop) ; d-extension from not ~eq-literal
		      (set! d-pairs (d-unify-lists (subgoal-terms entry) 
						   (subgoal-terms sg) '()))
		      (if (and units-double antilemmas
			       (= (clause-length clause) 1) (eq? d-pairs '()))
			  (begin (unbind-vars) #f) #t))
		     ((= eq-tag 3) (trail-stop) ; decomposition
		      (set! d-pairs
			    (d-unify-lists (cdr (ref (cadr (subgoal-terms sg))))
			       (cdr (ref (car (subgoal-terms sg)))) '())) #t)
		     ((= eq-tag 1) (trail-stop) ; d-extension from ~eq-literal
		      (set! d-pairs
			    (d-unify-eq-terms (car (subgoal-terms entry))
					      (ref (car (subgoal-terms sg))) '()))
		      (if d-pairs
			  (begin
			    (set! d-pairs
				  (d-unify-terms (ref (cadr (subgoal-terms entry)))
				     (ref (cadr (subgoal-terms sg))) d-pairs))
			    (if (and units-double antilemmas
				     (= (clause-length clause) 1) (eq? d-pairs '()))
				(begin (unbind-vars) #f) #t))
			  (begin (unbind-vars) #f)))
;		     ((and (= eq-tag 2) (begin (print "PASSED") #f)))
		     ((= eq-tag 2) (trail-stop) ; sym. d-extension from ~eq-literal 
		      (set! d-pairs (d-unify-eq-terms (cadr (subgoal-terms entry))
					   (ref (car (subgoal-terms sg))) '()))
		      (if d-pairs
			  (begin
			    (set! d-pairs
				  (d-unify-terms (ref (car (subgoal-terms entry)))
				     (ref (cadr (subgoal-terms sg))) d-pairs)) #t)
			  (begin (unbind-vars) #f))))
	       (begin (if (not (eq? d-pairs '()))
			  (set-copy-subgoals actual-copy
			       (subgoals->eq-subgoals (copy-subgoals actual-copy)
						      d-pairs actual-copy #f)))
			  #t)
	       (or (and (not constraints)
			(begin (if (or foldup index-foldup) (v-trail-stop)) #t))
		   (begin (v-trail-stop) ; backtracking marker
		     (set! actual-c-stack c-stack)
		     (if (and (propagate-substitution) ; constr. not viol.
			      (regularity entry (copy-subgoals actual-copy) sg
				 (+ (clause-length clause) (length d-pairs))))
			 #t ; below: shallow constraint and unfication
			 (begin (set! c-stack actual-c-stack) ; backtracking
				(set-copy-subgoals actual-copy old-copy-subgoals)
				(undo-v-trail) (unbind-vars) 
				(set! constraint-fails (+ constraint-fails 1))
				#f)))))
	      (begin (if signature-bound ; shallow signature backtracking
			 (begin (untrail-sig old-sig-trail-position)
				(set! free-sig old-free-sig)))
		     #f)))
     (let ((copy actual-copy) (old-global-subgoals global-subgoals)
	   (old-copies (clause-copies clause)) (new-inf inference)
	   (old-clause-subgoals old-copy-subgoals)
	   (clause-length (clause-length clause))
           new-sg new-links new-depth (old-local-free-sig old-free-sig)
           (old-local-sig-trail-position old-sig-trail-position)
           (old-c-stack actual-c-stack) old-subgoal-terms)
      (set! variable-names '())
      (if debug
	  (begin (display "EXTENSION STEP SUCCESSFUL: ")
	     (if sg (begin (display "subgoal: ") (display-subgoal sg) (newline)))
	     (display "free depth is: ") (display r-depth) 
	     (display-clause clause) (display " is entered clause") (newline)
	     (display " eq-tag is: ") (display eq-tag) (newline)))
      (if proof-output (set! proof-output
	  (cons (cons sg (if entry entry clause)) proof-output)))
      (set! extension-steps (+ extension-steps 1))
      (set! successful-inferences (+ successful-inferences 1))
      (if (and failinfo (= clause-length clause 1) (eq? d-pairs '()))
	  (set-subgoal-selected sg 1))
      (set-copy-path copy sg)
      (if (and eq-tag sg)
	  (begin (set! old-subgoal-terms (subgoal-terms entry))
		 (set-subgoal-terms entry (subgoal-terms sg))))
      (if (not (eq? d-pairs '()))
	  (begin (if debug (begin (display "New clause after extension: ")
		     (display-subgoals (copy-subgoals copy) 0) (newline)))
		 (set! clause-length (+ clause-length (length d-pairs)))))
      (if foldup
	  (if sg
	      (set-subgoal-foldup-predecessors (copy-subgoals copy)
				(- clause-length 1) sg) 
	      (set-subgoal-foldup-predecessors (copy-subgoals copy)
        		(- clause-length 1) foldup-root)))
      (new-clause-copy clause)
      (set-clause-copies clause (cdr (clause-copies clause)))
      (set! new-inf (+ new-inf clause-length))
      (if sg (begin (set! new-inf (- new-inf 1))
                    (set-subgoal-next sg entry)))
      (set! inf new-inf)
      (if (or (not sg) (> clause-length 1))
          (begin 
            (if (eq? eq-tag 3)
		(if (and decomposition-cost (> r-depth decomposition-cost))
		    (set-copy-depth copy (- r-depth 1))
		    (set-copy-depth copy r-depth))
		(set-copy-depth copy (- r-depth 1)))
            (set-copy-inference copy new-inf)
            (if weighted-depth
                (set-copy-free-depth100 copy 
                   (- (+ (* 100 (copy-depth copy)) wd1)
                      (* wd2 (- clause-length 1)))))
	    (if (or (not sg) (> clause-length 1))
	        (begin (vector-set! (copy-subgoals copy) 
			            clause-length global-subgoals)
                       (set! global-subgoals (copy-subgoals copy))))
            (if sg 
                (set-subgoal-selected entry #t)
	        (set! all-global-subgoals global-subgoals))))
      (next-subgoal-call new-inf new-sg new-links new-depth old-global-subgoals
			 (+ (- nr-sgs 2) clause-length))
      ;; deep backtracking:
      (if (and debug sg)
	  (begin
		  (display "Backtracking extension step on ")
		  (display-subgoal sg) (display " (")
		  (display (subgoal-clause-number sg)) (display ".")
		  (display (subgoal-number sg)) (display ")")
		  (display " on depth ")
		  (display (- depth r-depth)) (newline)))
       (if proof-output (set! proof-output (cdr proof-output)))
      (if sg 
          (begin (if eq-tag (set-subgoal-terms entry old-subgoal-terms))
	         (set-subgoal-selected entry #f)
		 (set-copy-subgoals copy old-clause-subgoals)
                 (if (or constraints foldup index-foldup) (undo-v-trail))
                 (if constraints (set! c-stack old-c-stack))
                 (if (not (and antilemmas (= clause-length 1)))
		     (unbind-vars))))
      (if signature-bound (begin (untrail-sig old-local-sig-trail-position)
				 (set! free-sig old-local-free-sig)))
      (if antilemmas (set! antilemma-test #t))
;      (if debug (print "PASSED 1"))
      (set! actual-copy #f)
      (if (not (and sg antilemmas (= clause-length 1)))
	  (set-clause-copies clause old-copies)
	  (begin (set! actual-copy copy)
;		 (if debug (print "PASSED 1a"))
	    (set! antilemma-test (extract-disequation-constraint-relevant))
	    (if antilemma-test
		(set! actual-copy old-copies)
		(set-clause-copies clause old-copies))
;	    (if debug (print "PASSED 2"))
	    )))
      ;; end of shallow backtracking:
      (set! failed-inferences (+ failed-inferences 1))))
 
;;;*************************************************************************
;; -----------------
;; SUBGOAL SELECTION
;; -----------------

(define (select-subgoal subgoals index max-index inference)
; subgoals := #f | #(subgoal ... subgoal subgoals)
 (if 
  (= index max-index)
   (begin
    (set! glob-inf (+ glob-inf
     (clause-length (copy-clause (subgoal-copy (vector-ref subgoals 0))))))
    (if (and failinfo (subgoal-predecessor (vector-ref subgoals 0)))
	(set-subgoal-selected (subgoal-predecessor (vector-ref subgoals 0))
			      2))
    (if (and subgoal-reordering current-subgoal)
       (if (and global-reordering 
                (vector? (vector-ref subgoals max-index))
                (> (subgoal-depth current-subgoal) 0))
           (select-subgoal (vector-ref subgoals max-index) 0
                (- (vector-length (vector-ref subgoals max-index)) 1)
		inference)
           current-subgoal)
       (begin
        ; foldup:
	(if (and (or foldup index-foldup)
		 (vector? (vector-ref global-subgoals max-index)))
	    (foldup-subgoal
	     (subgoal-predecessor (vector-ref global-subgoals 0))
	     (subgoal-predecessor (subgoal-predecessor
				   (vector-ref global-subgoals 0)))
	     (copy-inference (subgoal-copy (vector-ref global-subgoals 0)))
	      ))
	(set! global-subgoals (vector-ref subgoals max-index))
	(if (vector? global-subgoals)
	    (select-subgoal global-subgoals 0 
			    (- (vector-length global-subgoals) 1)
			    inference)
	    (begin  (display-inferences depth)
	            (newline) (display "     ***** PROOF WITH ") 
		    (display (+ inf 1)) (display " INFERENCES ")
                    (if (> nr-reduction-steps 0) 
                        (begin (display "(") (display nr-reduction-steps) 
                               (display " RED.) ")))
                    (if (> nr-matchings 0) 
                        (begin (display "(") (display nr-matchings) 
                               (display " LEM.) ")))
		    (display "FOUND *****") (newline) (newline)
		    (display "Total successful/failed inferences")
		    (display ": ") 
		    (display total-successful-inferences)
                    (if (> successful-reduction-steps 0)
                        (begin (display " (RED: ") 
		               (display successful-reduction-steps)
                               (display ")")))
                    (if (> total-successful-matchings 0)
                        (begin (display " (LEM: ") 
		               (display total-successful-matchings)
                               (display ")")))
                    (display "/")
		    (display total-failed-inferences) 
;		    (if constraints 
;			(begin (display " (") 
;			       (display total-constraint-fails) (display ")")))
		    (newline)
                    (if proof-output
                        (begin (newline) (newline) (display "PROOF")
                           (newline) (display "-----") (newline)
                           (display-proof (reverse! proof-output))))
                    ;(print-predicate-lemma-indices predicate-list)
		    (newline)
		    (exit 0) #f)))))
  (if (subgoal-selected (vector-ref subgoals index))
      (select-subgoal subgoals (+ index 1) max-index inference)
      (if 
        subgoal-reordering
	(begin (compare-subgoals (vector-ref subgoals index))
               (select-subgoal subgoals (+ index 1) max-index inference))
        (vector-ref subgoals index)))))

(define (compare-subgoals subgoal)
 (if current-subgoal
     (if current-complexity 
	 (begin (trail-stop)
;	        (set! new-complexity (termlist-v-complexity 
;		                                 (subgoal-terms subgoal)))
	        (set! new-complexity (subgoal-instantiation subgoal))
		(set! new-vars (unbind-bindings))
		(if (subgoal-conditions subgoal)
		    (begin (set! current-subgoal subgoal)
		           (set! current-vars new-vars)
                           (set! current-complexity new-complexity))))
         (begin (trail-stop)
;		(set! current-complexity (termlist-v-complexity 
;		                           (subgoal-terms current-subgoal)))
	        (set! current-complexity 
                      (subgoal-instantiation current-subgoal))
                (set! current-vars (unbind-bindings))
	        (trail-stop)
;	        (set! new-complexity (termlist-v-complexity 
;		                           (subgoal-terms subgoal)))
	        (set! new-complexity (subgoal-instantiation subgoal))
		(set! new-vars (unbind-bindings))
		(if (subgoal-conditions subgoal)
		    (begin (set! current-subgoal subgoal)
		           (set! current-vars new-vars)
                           (set! current-complexity new-complexity)))))
    (set! current-subgoal subgoal)))

(define (reorder-sides sg left right)
 (trail-stop)
 (set! left (term-instantiation (ref (car (subgoal-terms sg)))))
 (unbind-bindings)
 (trail-stop)
 (set! right (term-instantiation (ref (cadr (subgoal-terms sg)))))
 (unbind-bindings)
 (if (or (< left right)
	 (and ff (= left right) (> right 0)
	      (> (functor-frequency (car (ref (car (subgoal-terms sg)))))
		 (functor-frequency (car (ref (cadr (subgoal-terms sg))))))))
     (begin (set! left (car (subgoal-terms sg)))
	    (set! right (cadr (subgoal-terms sg)))
	    (set-car! (subgoal-terms sg) right)
	    (set-car! (cdr (subgoal-terms sg)) left))))

(define (unbind-bindings)
;; makes free all variables back to the previous trailstop, outputs number
;; of modified bindings 
 (if (trail-content)
     (begin (set-car! (caar trail) '()) 
            (trail-previous) (+ 1 (unbind-bindings)))
     (begin (trail-previous) 0)))

(define (ref-int int-term)
;; if int-term is a <bound_variable>, its binding is returned, 
;; otherwise int-term is returned
 (if (or (integer? int-term) (free-variable int-term) 
	 (vector? (car int-term)))
     int-term
     (ref-int (car int-term))))

(define (ref-ext ext-term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
 (if (or (not ext-term) (non-variable ext-term) (free-variable ext-term))
     ext-term
     (ref-ext (car ext-term))))

;;(define (ref-occ occ-ground-term)
;;;; an occ-ground-term is a bound variable, a term or a nonnegative integer
;;;; if term is a <bound_variable>, its binding is returned, 
;;;; otherwise occ-ground-term is returned
;; (if (or (non-variable occ-ground-term) (free-variable ext-term))
;;     ext-term
;;     (ref-ext1 (car ext-term))))

(define (termlist-var-occ terms)
; counts number of variables in term by setting the variable to
; its current frequency and incrementing this value;
; uses the trail: therefore, before calling main function, say (trail-stop)
 (if (pair? terms)
     (begin (term-var-occ (car terms))
;            (newline) (display "1 :") (set! l1 terms) (repl)
            (termlist-var-occ (cdr terms)))))

(define (term-var-occ term)
; instantiates free variables to 0 and variables with integer bindings n
; to n+1; auxiliary to procedure termlist-var-occ 
 (if (free-variable term) 
     (bind term 0)
     (if (integer? (car term))
         (set-car! term (+ (car term) 1))
         (if (vector? (car term))
             (termlist-var-occ (cdr term))
             (term-var-occ (car term))))))

(define (term-instantiation1 term)
; computes instantiation value of an extended term t as follows:
;           |  ( occ(t) - 1 ) / occ(t)    if t is a variable (occurrence)
; t-i(t) =  {  1                          if t is a constant 
;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
;
 (cond ((integer? (car term))
        (quotient (* 1000 (car term)) (+ (car term) 1)))
       ((constant term) 1000)
       ((vector? (car term))
        (+ (if fff (quotient 500 (functor-frequency (car term))) 500)
	   (quotient (termlist-instantiation1 (cdr term)) 
                              (* 2 (functor-arity (car term))))))
       (else (term-instantiation1 (car term)))))

(define (termlist-instantiation1 termlist)
 (if (null? (cdr termlist))
     (term-instantiation1 (car termlist))
     (+ (term-instantiation1 (car termlist))
        (termlist-instantiation1 (cdr termlist)))))

(define (termlist-v-complexity terms)
; computes number of symbols in terms - number of distinct variables
; uses the trail: therefore, before calling main function, say (trail-stop)
;                            and afterwards say: (unbind-bindings))
 (if (null? terms)
     0
     (+ (term-v-complexity (ref-ext (car terms))) 
        (termlist-v-complexity (cdr terms)))))

(define (term-v-complexity term)
; computes number of symbols in term - number of distinct variables
; uses the trail: therefore, before calling main function, say (trail-stop)
;                            and afterwards say: (unbind-bindings)
 (cond ((or (not term) (constant term)) 1)
       ((variable term) (bind term #f) 0)
       (else (+ 1 (termlist-v-complexity (cdr term))))))

(define (term-instantiation term)
; computes instantiation value of a term as follows:
;           |  0                          if t is a variable
; t-i(t) =  {  1                          if t is a constant 
;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
;
 (cond ((not term) var2) 
       ((variable term) (bind term #f) 0)
       ((constant term) 1000)
       (else (+ (if fff (quotient 500
			 (max 1 (quotient (functor-frequency (car term))
				   (* (max 1 (quotient nr-literals
						       nr-functors)) fff))))
		    500)
		(quotient (termlist-instantiation (cdr term)) 
                              (* 2 (functor-arity (car term))))))))

(define (termlist-instantiation termlist)
 (if (null? (cdr termlist))
     (term-instantiation (ref-ext (car termlist)))
     (+ (term-instantiation (ref-ext (car termlist)))
        (termlist-instantiation (cdr termlist)))))

(define (compute-weighted-depth depth free-depth100 entry-inf inf) 
; first line must remain, otherwise segmentation fault 
; (probably because of compiler optimization)
    (quotient (+ 50 free-depth100
                 (quotient (* 100 (- (* depth 100) free-depth100))
                           (+ 100 (* (- inf entry-inf) wd3))))
              100))

;;;*************************************************************************
;; ------------------------
;; COMPUTE AND CHECK BOUNDS
;; ------------------------

(define (free-signature-weights sigs)
 (if (pair? sigs)
     (begin (set-signature-weight (car sigs) 0)
            (free-signature-weights (cdr sigs)))))
 
(define (update-pred-numbers preds)
 (if (pair? preds)
     (begin (set-predsym-number (car preds) 
                         (+ (predsym-number (car preds)) nr-functors))
            (update-pred-numbers (cdr preds)))))
 
(define (sig-bound sigs)
;; the function updates free-sig and the signature vector sig-vector: 
;; free-sig is decremented with the sum of the signature weights of the 
;; new symbols in sigs and the respective positions in the sig-vector are
;; incremented with the signature weights; the new positions in the vector 
;; are remembered in the sig-trail;
;; sig-bound returns #f if the new value of free-sig is < 0
 (if (pair? sigs)
     (if (not (vector-ref sig-vector (signature-number (car sigs))))
         (begin (set! free-sig (- free-sig (signature-weight (car sigs))))
           (if (and (>= free-sig 0) (sig-bound (cdr sigs)))
               (begin (vector-set! sig-vector (signature-number (car sigs))
                                              (signature-weight (car sigs)))
                 (vector-set! sig-trail sig-trail-position 
                                        (signature-number (car sigs)))
                 (set! sig-trail-position (+ sig-trail-position 1)))
               #f))
         (sig-bound (cdr sigs)))
     #t))

(define (untrail-sig old-sig-trail-position)
 (if (> sig-trail-position old-sig-trail-position)
     (begin 
       (set! sig-trail-position (- sig-trail-position 1))
       (vector-set! sig-vector (vector-ref sig-trail sig-trail-position) #f)
       (untrail-sig old-sig-trail-position))))

(define (term-variables termlist)
; extracts set of variables occurring in termlist
; nondestructive, but internally misuses the trail
 (trail-stop)
 (termlist-var-occ termlist)
 (unbind-collect-vars '()))

(define (subgoals-termlists-appended subgoals index)
 (if (< (vector-length subgoals) index)
     (append (subgoal-terms (vector-ref subgoals index))
	     (subgoals-termlists-appended subgoals (+ index 1)))
     '()))

(define (set-subgoal-foldup-predecessors subgoals length predecessor)
 (if (>= length 0)
     (begin (set-subgoal-foldup-predecessor
	         (vector-ref subgoals length) predecessor)
            (set-subgoal-foldup-predecessors
	         subgoals (- length 1) predecessor))))

(define (update-time-stamp sg inference)
 (v-trail-next)
 (vector-set! (car v-trail) 0 sg)
 (vector-set! (car v-trail) 1 9)
 (vector-set! (car v-trail) 2 (subgoal-time-stamp sg))
; (newline) (display "update-time-stamp of subgoal ")
; (display-subgoal sg) (display " to inference ") (print inference)
 (set-subgoal-time-stamp sg inference))

(define (foldup-subgoal subgoal predecessor inference) 
; DESCRIPTION: whenever the subgoals of a tableau clause are completely solved,
; the subgoal L dominating the clause copy (copy-path copy) is complemented and
; inserted at its c-point. The c-point is immediate above the highest subgoal K
; above L used in a reduction step during the solution of L.
; The c-point is computed as follows.
; Whenever a subgoal is used in a reduction step, its subgoal-time-stamp is
; destructively set to the number of the current inference step (inference).
; In order to redo this after backtracking, the change is documented on the
; value trail.
; The folding up operation itself is implemented as follows.
; Every subgoal vector has a cell subgoal-foldup-predecessor (SFP). Initially,
; SFP(K) is its dominating subgoal. Folding up is done destructively by setting
; SFP(L) := SFP(K), SFP(K) := LSFP(K), and complementing the sign of L.
; In order to redo this after backtracking, the changes are documented on the
; value trail.
; For handling the case of a subgoal which can be folded up to the root, the
; special subgoal foldup-root was introduced, which is the SFP of the subgoals
; in the top clause. The SFP of foldup-root is #f.
;
 (if foldup
     (let ((c-point (find-c-point predecessor inference)))
       (v-trail-next) ; trailing:
       (vector-set! (car v-trail) 0 c-point)
       (vector-set! (car v-trail) 1 8)
       (vector-set! (car v-trail) 2 (subgoal-foldup-predecessor c-point))
       (v-trail-next)
       (vector-set! (car v-trail) 0 subgoal)
       (vector-set! (car v-trail) 1 0)
       (vector-set! (car v-trail) 2 (subgoal-sign subgoal))
       (v-trail-next)
       (vector-set! (car v-trail) 0 subgoal)
       (vector-set! (car v-trail) 1 8)
       (vector-set! (car v-trail) 2 (subgoal-foldup-predecessor subgoal))
       (if (eq? (subgoal-sign subgoal) '-) ; changing sign of subgoal:
	   (set-subgoal-sign subgoal '+)
	   (set-subgoal-sign subgoal '-))
       (set-subgoal-foldup-predecessor subgoal
				       (subgoal-foldup-predecessor c-point))
       (set-subgoal-foldup-predecessor c-point subgoal)
       (if (and index-foldup (eq? c-point foldup-root))
	   (begin (set! lemmatype 1)
; 		  (newline) (display "Root foldup at inference: ")
; 		  (display successful-inferences) (display " of ")
; 		  (display-subgoal subgoal) (display " (")
; 		  (display (subgoal-clause-number subgoal)) (display ".")
; 		  (display (subgoal-number subgoal)) (display ")")
; 		  (display " on depth ")
; 		  (display (- depth (subgoal-depth subgoal)))
; 		  (display " ")
; 		  (if proof-output (display proof-output))
		  (in-lemma-index-subgoal subgoal))))
     (if (c-point-at-root? subgoal inference) 
            ; foldup is off, but in case the solved subgoal is context free,
	    ; it can be inserted into the discrimination tree of unit lemmas
	 (begin (set! lemmatype 1)
	   (comp-in-lemma-index-subgoal subgoal))) ; subgoal sign unchanged!
     ))

(define (find-c-point subgoal inference)
  (if subgoal
      (if (>= (subgoal-time-stamp subgoal) inference) 
	  subgoal
	  (find-c-point (subgoal-foldup-predecessor subgoal) inference))
      foldup-root))

(define (c-point-at-root? subgoal inference)
 (if subgoal
     (if (< (subgoal-time-stamp subgoal) inference) 
	 (c-point-at-root? (subgoal-predecessor subgoal) inference)
	 #f)))

(define (insert-lemma sg tag)
 (let (i-comp t-comp)
;   (display "Complexity: ")
   (trail-stop) (set! i-comp (subgoal-instantiation sg)) (unbind-bindings) 
   (trail-stop) (set! t-comp (termlist-v-complexity (subgoal-terms sg)))
   (unbind-bindings)
;   (display (* i-comp t-comp)) (display " INF: ")
;   (display glob-inf) (display " ") (newline)
;   (display i-comp) (display " ") (display t-comp) (display " ")
;   (if tag (display-subgoal sg) (display-comp-subgoal sg))
;   (newline)
;   (print "Lemmas before: ") (display-ulemmas lemmas 1) (newline)
   (set! i-comp (* i-comp t-comp))
   (if (= lemmatype 0) (set! i-comp (quotient (* i-comp 100) conjecture-value)))
;   (display "BEFORE insert-sg-lemmas: ") (print (list i-comp t-comp))
   (set! i-comp (quotient i-comp glob-inf))
   (insert-sg-lemmas sg i-comp t-comp lemmas #f tag)
;   (print "Lemmas after: ") (display-ulemmas lemmas 1) (newline)
;   (display "------------------------------------------") (newline)
   ))

(define (insert-sg-lemmas sg comp t-comp loc-lemmas pred tag)
  (if (null? loc-lemmas) ; sg is best lemma: sg inserted at end of lemmas
      (if (> max-lemmas 0) ; enough space
	  (begin (set! max-lemmas (- max-lemmas 1))
	    (if (null? lemmas) ; sg is first lemma
		(begin (set! lemmas (list (make-vector 6)))
		  (overwrite-lemma sg comp t-comp tag (car lemmas)))
	        (begin (set-cdr! pred (list (make-vector 6)))
		  (overwrite-lemma sg comp t-comp tag (cadr pred)))))
	  ; not enough space: (car lemmas) to be overwritten
	  (begin (set-cdr! pred lemmas) (set! lemmas (cdr lemmas))
	    (set-cdr! (cdr pred) '())
	    (overwrite-lemma sg comp t-comp tag (cadr pred))))
      (if (better-lemma sg comp (car loc-lemmas))
	  (insert-sg-lemmas sg comp t-comp (cdr loc-lemmas) loc-lemmas tag)
	  ; sg is not best lemma: sg possibly inserted before loc-lemmas
	  (if (> max-lemmas 0) ; enough space
	      (begin (set! max-lemmas (- max-lemmas 1))
	       (if pred ; sg is not worst lemma
		 (begin (set-cdr! pred (cons (make-vector 6) loc-lemmas))
		   (overwrite-lemma sg comp t-comp tag (cadr pred)))
		 (begin (set! lemmas (cons (make-vector 6) loc-lemmas))
	           (overwrite-lemma sg comp t-comp tag (car lemmas)))))
	      ; not enough space:
	      (if pred ; sg is not worst lemma: (car lemmas) to be overwritten
		  (begin (set-cdr! pred lemmas) (set! lemmas (cdr lemmas))
		    (set-cdr! (cdr pred) loc-lemmas)
		    (overwrite-lemma sg comp t-comp tag (cadr pred)))
		  )))))

(define (better-lemma sg comp cell) (< comp (vector-ref cell 0)))

(define (overwrite-lemma sg comp t-comp tag cell)
  (vector-set! cell 0 comp)
  (vector-set! cell 1 t-comp)
  (vector-set! cell 2 (if tag (subgoal-sign sg)
			  (if (eq? (subgoal-sign sg) '-) '+ '-)))
  (vector-set! cell 3 lemmatype)
  (vector-set! cell 4 (subgoal-pred sg))
  (set! var-assoc '())
  (vector-set! cell 5 (ref-copy-termlist (subgoal-terms sg)))
  )
	  
; (define (steal-last-pair nelist pred)
; ; nelist is a nonempty list
;   (if (null? (cdr nelist))
;       (let ((last nelist))
; 	(set-cdr! pred '())
; 	last)
;       (steal-last-pair (cdr nelist) nelist)))

(define (display-ulemmas lemmas nr)
  (if (and (pair? lemmas) (> printlemmas 0))
      (begin
	(display-ulemma (car lemmas) nr)
;	(set! printlemmas (- printlemmas 1))
	(display-ulemmas (cdr lemmas) (+ nr 1)))))

(define (display-ulemma lemma nr)
  (display "input_clause(l") (display nr) (display ",")
  (if (= 1 (vector-ref lemma 3))
      (display "axiom,[ ")
      (display "conjecture,[ "))
  (if (eq? '+ (vector-ref lemma 2))
      (display "++") (display "--"))
  (if (eq? (predsym-symbol (vector-ref lemma 4)) 'equal)
      (display "equal")
      (display (predsym-symbol (vector-ref lemma 4))))
  (display "(")
  (display-ref-list (vector-ref lemma 5)) (display ") ]).")
  (newline)  
  )

(define (insert-new-clauses lemmas nrlemmas)
 (if (and (pair? lemmas) (> nrlemmas 0))
     (begin (insert-new-clause (car lemmas))
;	    (print "ULEMMA INSERTED: ") (display-ulemma (car lemmas) 0)
	    (insert-new-clauses (cdr lemmas) (- nrlemmas 1)))))

(define (insert-new-clause lemma)
 (let ((link (make-link lemma)))
;   (print "C1")
;   (display-clause (car link))
;   (print "C2")
   (add-unit-links (formula-clauses formula) (list (car link)))
;   (if (eq? (vector-ref lemma 2) '+)
;       (set-predsym-poslinks (vector-ref lemma 4)
; 	   (cons link (predsym-poslinks (vector-ref lemma 4))))
;       (set-predsym-neglinks (vector-ref lemma 4)
; 	   (cons link (predsym-neglinks (vector-ref lemma 4)))))
))
(define (make-link lemma)
  (let ((c-c (make-vector 12 '())))
    (set-clause-generic c-c (vector (make-subgoal lemma) #f))
    (set-clause-copies c-c
	 (list (make-clause-copy (clause-generic c-c) 1 '() c-c '())))
    (set-clause-length c-c 1)
    (set-clause-complexity c-c 1)
    (set-clause-all-copies c-c (clause-copies c-c))
    (set-clause-number c-c (+ 100 nr-lemmas))
    (set-clause-type c-c (vector-ref lemma 3))
; (print "B1")
;     (display-clause c-c)
; (print "B2")
    (cons c-c 0)))

(define (make-subgoal lemma)
  (let ((sg (make-vector 8 #f)))
    (set-subgoal-sign sg (vector-ref lemma 2))
    (set-subgoal-pred sg (vector-ref lemma 4))
    (set-subgoal-terms sg (vector-ref lemma 5))
    (set-subgoal-links sg '())
    (set-subgoal-number sg 1)
    sg))




