(module change-counters
	(include "semprop.sch")
	(import (globals "globals.scm")
		(display-clauses semprop "semprop.scm")
		(display-clause semprop "semprop.scm")
		(display-atom-names semprop "semprop.scm")
		)
	(export (decrease-counters clauses depth clauses-pred stop)
		(disactivate clauses atom depth clauses-pred)
		(increase-counters clauses depth clauses-pred stop)
		(reactivate clauses atom depth clauses-pred)
;		(check-open-clauses clauses)
		))

(define (disactivate clauses atom depth clauses-pred)
 (if (pair? clauses) 
     (begin
      (if (eq? #t (c-activating-atom (car clauses))) ; deleted lemma
	  (begin (set-cdr! clauses-pred (cdr clauses)) (set! clauses clauses-pred))
	  (if (not (c-activating-atom (car clauses))) ; clause is active
              (begin (set-c-activating-atom (car clauses) atom)
	        (if (input-clause? (car clauses))
		    (begin (set! nr-open-clauses (- nr-open-clauses 1))
; 		      (if #f ; (= nr-open-clauses (check-open-clauses clause-list)))
; 			  (begin (newline) (display "inference: ") (display inference)
; 			     (display ", branching literal: ")
; 			     (if (= 0 (a-sign atom)) (display "-"))
; 			     (display-atom-names (list atom))
; 			     (display ", clause: ")
; 			     (display-clause (car clauses))
; 			     (display ", soll/ist: ")
; 			     (display nr-open-clauses) (display "/")
; 			     (display (check-open-clauses clause-list))))
		      (if (> dynamic-depth depth)
			  (begin (set! weight (vector-ref clauses-weight-vector
						    (c-length (car clauses))))
			     (decrease-freq-pos-act weight
						    (c-pos-literals (car clauses)))
			     (decrease-freq-neg-act weight
						    (c-neg-literals (car clauses)))))
		      (if purity
			  (begin ; (set! pi atom) (set! rv (car clauses))
				 (update-pos-occ (c-pos-literals (car clauses)) -1)
				 (update-neg-occ (c-neg-literals (car clauses)) -1)))
		      )))))
      (disactivate (cdr clauses) atom depth clauses))))

(define (decrease-freq-pos-act weight-diff atoms)
 (if (pair? atoms)
     (begin
       (if (not (a-sign (car atoms)))
	   (set-a-pos-frequency (car atoms) 
				(- (a-pos-frequency (car atoms)) weight-diff)))
       (decrease-freq-pos-act weight-diff (cdr atoms)))))

(define (decrease-freq-neg-act weight-diff atoms)
 (if (pair? atoms)
     (begin
       (if (not (a-sign (car atoms)))
	   (set-a-neg-frequency (car atoms) 
				(- (a-neg-frequency (car atoms)) weight-diff)))
       (decrease-freq-neg-act weight-diff (cdr atoms)))))

(define (reactivate clauses atom depth clauses-pred)
 (if (pair? clauses) 
     (begin
       (if (eq? #t (c-activating-atom (car clauses))) ; deleted lemma
	   (begin (set-cdr! clauses-pred (cdr clauses)) (set! clauses clauses-pred))
	   (if (eq? (c-activating-atom (car clauses)) atom)
	       (begin (set-c-activating-atom (car clauses) #f) ; cl. reactivated
		 (if (input-clause? (car clauses))
		     (begin (set! nr-open-clauses (+ nr-open-clauses 1))
		       (if (> dynamic-depth depth)
			   (begin (set! weight (vector-ref clauses-weight-vector
							   (c-length (car clauses))))
				  (increase-freq-pos-act weight
						    (c-pos-literals (car clauses)))
				  (increase-freq-neg-act weight
						    (c-neg-literals (car clauses)))))
		       (if purity
			   (begin (update-pos-occ (c-pos-literals (car clauses)) 1)
				  (update-neg-occ (c-neg-literals (car clauses)) 1)))
		       )))))
       (reactivate (cdr clauses) atom depth clauses))))

(define (update-pos-occ atoms nr)
 (if (pair? atoms)
     (begin 
       (if (not (a-sign (car atoms)))
           (begin (set-a-pos-occ (car atoms) (+ nr (a-pos-occ (car atoms))))
		  (if (= 0 (a-pos-occ (car atoms)))
		      (begin (set! purity-index (+ 1 purity-index))
			     (vector-set! purity-vector purity-index (car atoms))))))
       (update-pos-occ (cdr atoms) nr))))
 
(define (update-neg-occ atoms nr)
 (if (pair? atoms)
     (begin 
       (if (not (a-sign (car atoms)))
	   (begin (set-a-neg-occ (car atoms) (+ nr (a-neg-occ (car atoms))))
		  (if (= 0 (a-neg-occ (car atoms)))
		      (begin (set! purity-index (+ 1 purity-index))
			     (vector-set! purity-vector purity-index (car atoms))))))
       (update-neg-occ (cdr atoms) nr))))
 
(define (increase-freq-pos-act weight-diff atoms)
 (if (pair? atoms)
     (begin 
       (if (not (a-sign (car atoms)))
	   (set-a-pos-frequency (car atoms) 
				(+ (a-pos-frequency (car atoms)) weight-diff)))
       (increase-freq-pos-act weight-diff (cdr atoms)))))

(define (increase-freq-neg-act weight-diff atoms)
 (if (pair? atoms)
     (begin 
       (if (not (a-sign (car atoms)))
	   (set-a-neg-frequency (car atoms) 
				(+ (a-neg-frequency (car atoms)) weight-diff)))
       (increase-freq-neg-act weight-diff (cdr atoms)))))

(define (decrease-counters clauses depth clauses-pred stop)
;; decreases counters with considering stop clause;
;; if a new unit is generated, it is put in the unit-vector; 
;; if stop is #t and a unit clause is found, stop-clause is set
 (if (pair? clauses) 
     (cond ((eq? #t (c-activating-atom (car clauses))) ; deleted lemma
	    (set-cdr! clauses-pred (cdr clauses)) 
	    (decrease-counters (cdr clauses) depth clauses-pred stop))
           ((c-activating-atom (car clauses)) ;; clause is not active
            (decrease-counters (cdr clauses) depth clauses stop))
           ((and stop (= 1 (c-length (car clauses)))) ; unit clause falsifies
            (set! stop-clause (car clauses))) ; complement assignment
           ((= 2 (c-length (car clauses))) ;; a new unit clause is produced
            (set-c-length (car clauses) 1) (set! unit-index (+ unit-index 1))
            (vector-set! unit-vector unit-index (car clauses)) 
            (decrease-counters (cdr clauses) depth clauses stop))
           (else (if (and (> dynamic-depth depth) (input-clause? (car clauses))
			  (begin (set! weight (vector-ref clauses-weight-diff-vector
				 		          (c-length (car clauses))))
				 (> weight 0)))
		     (begin (increase-freq-pos weight (c-pos-literals (car clauses)))
 		            (increase-freq-neg weight (c-neg-literals (car clauses)))))
		 (set-c-length (car clauses) (- (c-length (car clauses)) 1))
                 (decrease-counters (cdr clauses) depth clauses stop)))))

(define (increase-counters clauses depth clauses-pred stop)
;; increases counters until stop clause is reached
 (if (pair? clauses)
     (cond ((and stop (eq? stop-clause (car clauses))))
	   ((eq? #t (c-activating-atom (car clauses))) ; deleted lemma
	    (set-cdr! clauses-pred (cdr clauses)) 
	    (increase-counters (cdr clauses) depth clauses-pred stop))
	   ((c-activating-atom (car clauses)) ;; clause is not active
	    (increase-counters (cdr clauses) depth clauses stop))
	   (else (if (and (> dynamic-depth depth) (input-clause? (car clauses))
			  (> (c-length (car clauses)) 1)
			  (begin (set! weight (vector-ref clauses-weight-diff-vector
						    (+ 1 (c-length (car clauses)))))
			   (> weight 0)))
		     (begin (decrease-freq-pos weight (c-pos-literals (car clauses)))
			    (decrease-freq-neg weight (c-neg-literals (car clauses)))))
		 (set-c-length (car clauses) (+ 1 (c-length (car clauses))))
		 (increase-counters (cdr clauses) depth clauses stop)))))

(define (increase-freq-pos weight-diff atoms)
 (if (pair? atoms)
     (begin
       (if (not (a-sign (car atoms)))
	   (set-a-pos-frequency (car atoms) 
                (+ (a-pos-frequency (car atoms)) weight-diff))) 
       (increase-freq-pos weight-diff (cdr atoms)))))

(define (increase-freq-neg weight-diff atoms)
 (if (pair? atoms)
     (begin
       (if (not (a-sign (car atoms)))
	   (set-a-neg-frequency (car atoms) 
                (+ (a-neg-frequency (car atoms)) weight-diff)))
       (increase-freq-neg weight-diff (cdr atoms)))))

(define (decrease-freq-pos weight-diff atoms)
 (if (pair? atoms)
     (begin
       (if (not (a-sign (car atoms)))
	   (set-a-pos-frequency (car atoms) 
                (- (a-pos-frequency (car atoms)) weight-diff)))
       (decrease-freq-pos weight-diff (cdr atoms)))))

(define (decrease-freq-neg weight-diff atoms)
 (if (pair? atoms)
     (begin
       (if (not (a-sign (car atoms)))
	   (set-a-neg-frequency (car atoms) 
                (- (a-neg-frequency (car atoms)) weight-diff)))
       (decrease-freq-neg weight-diff (cdr atoms)))))

; (define (check-open-clauses clauses)
;  (if (pair? clauses)
;      (if (or (lemma? (car clauses)) (c-activating-atom (car clauses)))
; 	 (check-open-clauses (cdr clauses))
; 	 (+ 1 (check-open-clauses (cdr clauses))))
;      0))
	 