(module evaluate
	(include "semprop.sch")
	(import
	        (globals "globals.scm")
	        (decrease-counters change-counters "change-counters.scm")
	        (disactivate change-counters "change-counters.scm")
	        (increase-counters change-counters "change-counters.scm")
	        (reactivate change-counters "change-counters.scm")
		(atom->= semprop "semprop.scm")
		(display-clause semprop "semprop.scm")
		(display-clauses semprop "semprop.scm")
		(display-path-index semprop "semprop.scm")
		(display-atom-names semprop "semprop.scm")
		(model semprop "semprop.scm")
		(order-atoms semprop "semprop.scm")
		(delete-lemmas lemmas "lemmas.scm")
		(generate-lemma0 lemmas "lemmas.scm")
		)
	(export 
		(evaluate atom-index depth path-index b-depth)
		(merge-boolean-vectors vec1 vec2 index)
	))
 
(define (evaluate atom-index depth path-index b-depth)
(let ((atom #f) sign (clause #f) clauses true-clauses depth-vector-r
      (inference1 inference))
(cond ; NOTE: not every value of inference passes through here
      ;       and some values may pass through more than once!
; A PURE LITERAL OR A UNIT CLAUSE EXISTS:
((or (and first-pure (>= purity-index 0) (get-pure-literal)) ; a pure literal exists
     (if (next-active-unit) ; an active unit clause exists
	 (begin (set! clause (vector-ref unit-vector unit-index))
                (set! new-atom (get-literal (c-pos-literals clause)))
		(if new-atom ; negative unit clause found
		    (set! current-sign 1) 
		    (begin (set! new-atom (get-literal (c-neg-literals clause)))
			   (set! current-sign 0))) ; positive unit found
		(set! inference (+ 1 inference))
		(set! unit-index (- unit-index 1)) #t)
	 #f)
     (and (not first-pure) (>= purity-index 0) (get-pure-literal)))
 (set! depth (+ 1 depth))
 (set! atom new-atom) (set! sign current-sign)
 (if (> b-depth 0)
     (begin (set-a-last-depth atom depth)
       (set-a-last-depth-mod30exp atom (vector-ref vector-mod30exp depth))))
 (if (= sign 0) ;; negative valuation of atom to be checked
     (begin (set! clauses (a-pos-clauses atom)) (set-a-sign atom 0)
	    (set! true-clauses (a-neg-clauses atom)))
     (begin (set! clauses (a-neg-clauses atom)) (set-a-sign atom 1)
	    (set! true-clauses (a-pos-clauses atom))))
 (set! stop-clause #f)
 (if clause (decrease-counters clauses b-depth '(0) #t)) ; no purity
 (if stop-clause ; no purity
     (begin (set! inference (+ 1 inference))
	    (if proof (begin (display-path-index (- path-index 1))
			     (if (= 0 sign) (display " -") (display "  "))
			     (display (a-realname atom)) (display " *")
			     (display inference) (display "* ")
			     (display-clause stop-clause) (newline)))    
       (increase-counters clauses b-depth '(0) #t)
       (set-a-sign atom #f);; atom is made available
       (set! depth-vector-r (next-last-depth-vector))
       (vector-false-upto depth-vector-r (vector-ref vector30 depth)) 
       (rplac-atom-depth (c-pos-literals stop-clause) depth-vector-r)
       (rplac-atom-depth (c-neg-literals stop-clause) depth-vector-r)
       (if (context-lemma? stop-clause)
	   (merge-boolean-vectors depth-vector-r (c-rel-vector stop-clause)
				  (c-rel-depth30 stop-clause)))
       (if (lemma? stop-clause)
	   (begin (set-c-usages stop-clause (+ 1 (c-usages stop-clause)))
		  (set! nr-lemma-uses (+ nr-lemma-uses 1))))
       (rplac-atom-depth (c-pos-literals clause) depth-vector-r)
       (rplac-atom-depth (c-neg-literals clause) depth-vector-r)
       (if (context-lemma? clause)
	   (merge-boolean-vectors depth-vector-r (c-rel-vector clause)
				  (c-rel-depth30 clause)))
       (if (lemma? clause)
	   (begin (set-c-usages clause (+ 1 (c-usages clause)))
		  (set! nr-lemma-uses (+ nr-lemma-uses 1)))))
     (begin ; no stop clause
       (disactivate true-clauses atom b-depth '(0))
       ; NOTE: since purity is checked wrt input clauses only, it may not hold for lemmas,
       ; hence for soundness reasons lemmas containing ~atom have to be disactivated, too:
       (if (and lemmas-on (not clause)) (disactivate clauses atom b-depth '(0)))
       (vector-set! path-vector path-index atom)
       (if current-switch (set! current-atom atom))
       (set! depth-vector-r (evaluate atom-index depth (+ 1 path-index) b-depth))
       (if current-switch (set! current-atom atom))
       (reactivate true-clauses atom b-depth '(0)); clauses reactivated
       (if clause
	   (increase-counters clauses b-depth '(0) #f) ; clause lengths incremented
	   (if lemmas-on (reactivate clauses atom b-depth '(0))))
       (set-a-sign atom #f); NOTE: atom must be made available after react., incr. !
       (if lemmas-on (begin (if (> nr-kept-lemmas max-nr-kept-lemmas)
				(set! max-nr-kept-lemmas nr-kept-lemmas))
			    (delete-lemmas (a-lemmas-out atom))
			    (set-a-lemmas-out atom '())))
       (if (and depth-vector-r (not (atom-irrelevant? depth-vector-r atom)))
	   (if clause
	       (begin (rplac-atom-depth (c-pos-literals clause) depth-vector-r)
		      (rplac-atom-depth (c-neg-literals clause) depth-vector-r)
		      (if (context-lemma? clause)
			  (merge-boolean-vectors depth-vector-r (c-rel-vector clause)
						 (c-rel-depth30 clause)))
		      (if (lemma? clause)
			  (begin (set-c-usages clause (+ 1 (c-usages clause)))
				 (set! nr-lemma-uses (+ nr-lemma-uses 1))))))
	   (if (and proof clause) 
	       (begin (display "-----literal")
		      (if (= sign 0) (display " -") (display " "))
		      (display (a-realname atom))
		      (display " irrelevant, hence inference ")
		      (display inference1) (print " irrelevant"))))))
 depth-vector-r)
; MODEL FOUND:
((= nr-open-clauses 0) (set! inference (+ 1 inference)) (model))
; BELOW NO PURE LITERAL AND NO UNIT, BUT FORMULA IS STILL OPEN:
; BACKJUMPING 
((and (> inference-limit 0) (>= branchings backjump)
      (begin (set! inference-limit (+ inference-limit
                (quotient (* inference-limit-increase inference-limit) 100)))
	     (set! backjump (+ inference-limit backjump)) (display branchings)
	     (set! backatom #f)
	     (set! backatom (compute-back-atom (- path-index 1) #f #f))
	     (if backatom (begin (display ": ") (display-atom-names (list backatom))))
	     (newline)
	     backatom))
 #f)
; PROPER SEMANTIC BRANCHING
(else ; not all input clauses are true
 (if (and order-unit (= b-depth 0))
     (begin (order-atoms) (set! atom-index 0) (set! current-atom #f)))
 (set! atom #f)
 (set! atom-index (next-atom atom-index))
 (if (= b-depth 0) (set! first-branching depth))
; (if (= b-depth 0) (begin  (display "FIRST BRANCHING DEPTH: ") (print depth)))
 (set! b-depth (+ 1 b-depth)) (set! branchings (+ 1 branchings))
 (let (depth-vector-l branch (inference0 (+ inference 1)) inference1)
  (if (and (> dynamic-steps 0) (>= depth-nontab b-depth))
      (set! atom (next-dynamic (vector-ref atom-vector atom-index)
			       (next-atom (+ 1 atom-index)) dynamic-steps))
      (if (and current-atom (a-sign current-atom) tableau)
	  (begin
	   (if (= 0 (a-sign current-atom))
	       (set! current-atom
		     (best-connected current-atom
			   (* (a-pos-frequency current-atom) leaf-factor)
			   (- path-index 1) #f
			   (if (a-left-branch current-atom) tableau tableau-right)))
	       (set! current-atom
		     (best-connected current-atom
			   (* (a-neg-frequency current-atom) leaf-factor)
			   (- path-index 1) #f
			   (if (a-left-branch current-atom) tableau tableau-right))))
	   (if (= 0 (a-sign current-atom))
	       (set! atom (shortest-neighbor (a-pos-clauses current-atom) #f #f #f))
	       (set! atom (shortest-neighbor (a-neg-clauses current-atom) #f #f #f)))
	   (if (not atom)
	       (set! atom (best-connected-for-predecessor tableau-steps
							  (- path-index 1) #f)))
; 	   (if (or (not atom)
; 		   (and (> compare-selected 1)
; 		     (let (compare (factor (+ compare-selected clause-divisor)))
; 		       (set-a-pos-frequency atom (* factor (a-pos-frequency atom)))
; 		       (set-a-neg-frequency atom (* factor (a-neg-frequency atom)))
; 		       (set! compare (atom->= atom (vector-ref atom-vector atom-index)))
; 		       (set-a-pos-frequency atom (quotient (a-pos-frequency atom) factor))
; 		       (set-a-neg-frequency atom (quotient (a-neg-frequency atom) factor))
; 		       (not compare)))))
	   )))
  (if (not atom)
      (if (> dynamic-steps 0)
	  (set! atom (next-dynamic (vector-ref atom-vector atom-index)
				   (next-atom (+ 1 atom-index)) dynamic-steps))
	  (set! atom (vector-ref atom-vector atom-index))))
  (set! current-atom atom)
  (vector-set! path-vector path-index atom)
  (set! depth (+ depth 1))
  (set-a-last-depth atom depth)
  (set-a-last-depth-mod30exp atom (vector-ref vector-mod30exp depth))
  (select-sign atom branch)
  (if branch ; pos atom frequency is higher than neg atom frequency
      (begin (set! sign 1) (set-a-sign atom 1)
	     (set! clauses (a-neg-clauses atom))
             (set! true-clauses (a-pos-clauses atom)))
      (begin (set! sign 0) (set-a-sign atom 0)
	     (set! clauses (a-pos-clauses atom))
             (set! true-clauses (a-neg-clauses atom))))
; FIRST LEFT BRANCH CONSIDERED
(set! depth-vector-l
      (left-evaluate atom atom-index clauses true-clauses depth path-index b-depth #f))
  (if (or (and depth-vector-l (atom-irrelevant? depth-vector-l atom) 
               ; either closed subtree and literal not used on left branch
	       (begin (if proof
			  (begin (display-path-index (- path-index 1))
			    (if (= (a-sign atom) 1) (display " -") (display "  "))
			    (display (a-realname atom))
			    (display " * since literal")
			    (if (= (a-sign atom) 0) (display " -") (display " "))
			    (display (a-realname atom))
			    (print " irrelevant")))
		      #t))
	  (and (not depth-vector-l) (not allmodels)))
	       ; or no closed subtree and only first model to be output
      (begin (set-a-sign atom #f) ; atom is made available
	     (if lemmas-on
		 (begin (if (> nr-kept-lemmas max-nr-kept-lemmas)
			    (set! max-nr-kept-lemmas nr-kept-lemmas))
			(delete-lemmas (a-lemmas-out atom))
			(set-a-lemmas-out atom '())
			(set-a-unit-lemmas atom '()))) 
	     depth-vector-l)
; NOW RIGHT BRANCH CONSIDERED
      (begin (set! unit-index -1) ; marks vectors of units and pure literals as empty
	     (set! purity-index -1)
	(if (= (a-sign atom) 1)
	    (begin (set! sign 0) (set-a-sign atom 0) (set! clauses (a-neg-clauses atom))
		   (set! true-clauses (a-pos-clauses atom)))
	    (begin (set! sign 1) (set-a-sign atom 1) (set! clauses (a-pos-clauses atom))
		   (set! true-clauses (a-neg-clauses atom))))
	(disactivate clauses atom b-depth '(0))
	(decrease-counters true-clauses b-depth '(0) #f)
	(set! current-atom atom)
	(if lemmas-on (units2index (a-unit-lemmas atom)))
	(set! inference1 inference)
	(set! depth-vector-r (evaluate atom-index depth (+ 1 path-index) b-depth))
	(reactivate clauses atom b-depth '(0)) ; clauses backtracked
	(increase-counters true-clauses b-depth '(0) #f) 
	(set-a-sign atom #f) ; atom is made available
	(if lemmas-on
	    (begin (if (> nr-kept-lemmas max-nr-kept-lemmas)
		       (set! max-nr-kept-lemmas nr-kept-lemmas))
		   (delete-lemmas (a-lemmas-out atom))
		   (set-a-lemmas-out atom '()) (set-a-unit-lemmas atom '()))) 
	(if (and depth-vector-r depth-vector-l)
	    (begin (last-depth-vectors-previous)
	      (if (not (atom-irrelevant? depth-vector-r atom)) ; lit. used
		  (begin (merge-boolean-vectors depth-vector-l depth-vector-r
					        (vector-ref vector30 (- depth 1)))
		   (if (and lemmas-on (> max-lemma-length 0))
		       (begin
			(if proof
			  (begin (display "Branch atom before lemma generation is ")
				 (display-atom-names (list atom))
				 (if (a-left-branch atom) (display "L") (display "R"))
				 ))
			(set! unit-atoms '())
			(if (and lemma-adaptive (> inference 1000)
				 (= 0 (modulo branchings 100)))
			    (begin 
;			      (display "<") (display nr-lemma-uses)
;			      (display "/") (display nr-gen-lemmas) (display ">")
;			      (display max-lemma-length) (display " ")
			      (set! max-lemma-length (min 128 (max 8 (+ max-lemma-length
				(- 13 (quotient (* 10 (+ nr-lemma-uses 1))
						(+ nr-gen-lemmas 1)))))))))
			(generate-lemma0 (- path-index 1) depth-vector-l '() '()))))
		  (begin ; below literal not used
		    (set-car! (car last-depth-vectors) depth-vector-r)
		    (set-car! (cadr last-depth-vectors) depth-vector-l)
		    (set! depth-vector-l depth-vector-r)
		    (if proof ; literal irrelevant
			(begin (display "----- literal")
			       (if (= (a-sign atom) 0) (display " -") (display " "))
			       (display (a-realname atom))
			       (display " irrelevant, hence inference ")
			       (display inference0) (display "-")
			       (display inference1) (print " irrelevant")
			       ))))
	      depth-vector-l) ; below no closed subtree on left or right branch
	    #f)))))
)))

(define (left-evaluate atom atom-index clauses true-clauses depth path-index b-depth
		       depth-vector)
  (set! unit-index -1) (set! purity-index -1)
  (decrease-counters clauses b-depth '(0) #f)
  (disactivate true-clauses atom b-depth '(0))
  (set-a-left-branch atom #t)
  (set! depth-vector (evaluate atom-index depth (+ 1 path-index) b-depth))
  (set-a-left-branch atom #f)
  (reactivate true-clauses atom b-depth '(0)) ; backtrack clauses
  (increase-counters clauses b-depth '(0) #f) 
  (if (and (not depth-vector) (eq? atom backatom)) ; backatom reached
      (begin (if (= 0 (a-sign atom)) (set-a-sign atom 1) (set-a-sign atom 0))
	     (left-evaluate atom atom-index true-clauses clauses depth path-index
			    b-depth #f))
      depth-vector))

(define (next-last-depth-vector)
; globals: last-depth-vectors, nr-atoms
; last-depth-vectors is a list (p_0 ... p_n) of dotted pairs where
; *) the car of each p_i is a last-depth-vector and
; *) the cdr of each p_i is p_i-1 or #f if i = 0
 (if (null? (cdr last-depth-vectors))
     (set-cdr! last-depth-vectors
	       (list (cons (make-vector (+ nr-atoms30 1) 0)
			   last-depth-vectors))))
 (set! last-depth-vectors (cdr last-depth-vectors))
 (caar last-depth-vectors))

(define (last-depth-vectors-previous)
 (set! last-depth-vectors (cdar last-depth-vectors)))

(define (merge-boolean-vectors vec1 vec2 index)
 (if (>= index 0)
     (begin (vector-set! vec1 index (bit-or (vector-ref vec1 index)
					    (vector-ref vec2 index)))
	    (merge-boolean-vectors vec1 vec2 (- index 1)))))

(define (vector-false-upto vector index)
 (if (>= index 0)
     (begin (vector-set! vector index 0)
	    (vector-false-upto vector (- index 1)))))

(define (next-dynamic atom atom-index steps)
 (if (and (> steps 0) atom-index)
     (if (atom->= atom (vector-ref atom-vector atom-index))
	 (next-dynamic atom (next-atom (+ 1 atom-index)) (+ steps 1))
	 (next-dynamic (vector-ref atom-vector atom-index)
		       (next-atom (+ 1 atom-index)) (+ steps 1)))
     atom))

(define (shortest-neighbor clauses shortest best atom)
; selects best atom in all shortest non-true clauses
 (if (and (pair? clauses) (input-clause? (car clauses)))
     (cond ((c-activating-atom (car clauses))
	    (shortest-neighbor (cdr clauses) shortest best #f))
	   (shortest
	    (if (> (c-length (car clauses)) (c-length shortest))
		(shortest-neighbor (cdr clauses) shortest best #f)
		(begin
		 (set! atom (best-literal (best-literal #f (c-pos-literals (car clauses)))
					  (c-neg-literals (car clauses))))
		 (if (and (= (c-length (car clauses)) (c-length shortest))
			  (atom->= best atom))
		     (shortest-neighbor (cdr clauses) shortest best #f)
		     (shortest-neighbor (cdr clauses) (car clauses) atom #f)))))
	   (else (set! atom (best-literal (best-literal #f (c-pos-literals (car clauses)))
					  (c-neg-literals (car clauses))))
		 (shortest-neighbor (cdr clauses) (car clauses) atom #f)))
     best))

(define (best-connected atom freq path-index new-atom steps)
 (if (and (> steps 0) (>= path-index 0))
     (begin (set! new-atom (vector-ref path-vector path-index))
	    (if (= (a-sign new-atom) 0)
		(if (>= freq (a-pos-frequency new-atom))
		    (best-connected atom freq (- path-index 1) new-atom
				    (- steps 1))
		    (best-connected new-atom (a-pos-frequency new-atom)
				    (- path-index 1) #f (- steps 1)))
		(if (>= freq (a-neg-frequency new-atom))
		    (best-connected atom freq (- path-index 1) new-atom
				    (- steps 1))
		    (best-connected new-atom (a-neg-frequency new-atom)
				    (- path-index 1) #f (- steps 1)))))
     atom))
	      
(define (best-literal max literals)
 (if (pair? literals)
     (if (or (a-sign (car literals))
	     (and max (atom->= max (car literals))))
	 (best-literal max (cdr literals))
	 (best-literal (car literals) (cdr literals)))
     max))

(define (next-active-unit)
 (if (>= unit-index 0)
     (if (c-activating-atom (vector-ref unit-vector unit-index)) ; clause is true
         (begin (set! unit-index (- unit-index 1)) (next-active-unit))
	 #t)
     #f))

(define (units2index units)
 (if (pair? units)
     (begin (set! unit-index (+ unit-index 1))
	    (vector-set! unit-vector unit-index (car units))
	    (units2index (cdr units)))))

(define (rplac-atom-depth atoms depth-vector)
 (if (pair? atoms) 
     (begin (if (> (a-last-depth (car atoms)) 0)
		(vector-set! depth-vector (a-last-depth30 (car atoms))
		  (bit-or (a-last-depth-mod30exp (car atoms))
		      (vector-ref depth-vector (a-last-depth30 (car atoms))))))
	    (rplac-atom-depth (cdr atoms) depth-vector))))

(define (get-literal clause-half)
 (if (pair? clause-half)
     (if (a-sign (car clause-half))
         (get-literal (cdr clause-half))
         (car clause-half))
     #f))

(define (next-atom atom-index)
 (if (< atom-index nr-atoms) 
     (if (a-sign (vector-ref atom-vector atom-index))
         (next-atom (+ 1 atom-index))
         atom-index)
     #f))

(define (get-pure-literal)
; globals: new-atom, current-sign
; returns #t/#f depending on whether a pure literal exists or not
 (set! new-atom (vector-ref purity-vector purity-index))
 (set! purity-index (- purity-index 1))
 (if (a-sign new-atom)
     (if (>= purity-index 0) (get-pure-literal) #f)
     (begin (if (= 0 (a-pos-occ new-atom)) ; atom does not occur positively
		(begin ;(display-clauses (a-pos-clauses new-atom))
		       (set! current-sign 0))
		(begin ;(display-clauses (a-neg-clauses new-atom))
		       (set! current-sign 1)))
;	    (print inference)
	    #t)))

(define (best-connected-for-predecessor steps path-index atom)
 (if (and (> steps 0) (>= path-index 0)
	  (> (a-last-depth (vector-ref path-vector path-index)) 0))
     (begin
       (set! atom (vector-ref path-vector path-index))
       (if (= 0 (a-sign atom))
	   (set! atom (shortest-neighbor (a-pos-clauses atom) #f #f #f))
	   (set! atom (shortest-neighbor (a-neg-clauses atom) #f #f #f)))
       (if atom
	   atom
	   (best-connected-for-predecessor (- steps 1) (- path-index 1) #f)))
     #f))

(define (compute-back-atom path-index back-atom back-value)
 (if (>= path-index 0)
     (let ((back-atom-tmp (vector-ref path-vector path-index)) back-value-tmp)
       (if (a-left-branch back-atom-tmp)
	   (begin
	     (if (= 0 (a-sign back-atom-tmp))
		 (set! back-value-tmp
		       (+ (quotient (* 100 (a-pos-frequency back-atom-tmp))
				    (a-neg-frequency back-atom-tmp))
			  (a-last-depth back-atom-tmp)))
		 (set! back-value-tmp
		       (+ (quotient (* 100 (a-neg-frequency back-atom-tmp))
				    (a-pos-frequency back-atom-tmp))
			  (a-last-depth back-atom-tmp))))
	     (if (or (not back-atom) (> back-value-tmp back-value))
		 (begin (set! back-atom back-atom-tmp)
			(set! back-value back-value-tmp)))))
       (compute-back-atom (- path-index 1) back-atom back-value))
     back-atom))