(module lemmas
	(include "semprop.sch")
	(import
	       (globals "globals.scm")
	       (display-atom semprop "semprop.scm")
	       (display-clause semprop "semprop.scm")
	       (merge-boolean-vectors evaluate "evaluate.scm")
	       )
	(export
	       (generate-lemma0 path-index rel-vector plits nlits)
	       (delete-lemmas clauses)
	       ))

(define (generate-lemma0 path-index rel-vector plits nlits)
; lemma clause has length 0 and no relevance
; call with (generate-lemma path-index rel-vector '() '())
; globals: max-lemma-length, path-atom 
 (if (>= path-index first-branching)
     (begin (set! path-atom (vector-ref path-vector path-index))
       (if (atom-irrelevant? rel-vector path-atom) ; literal irrelevant
	   (generate-lemma0 (- path-index 1) rel-vector plits nlits)
	   (begin ; complement of literal is in lemma
	     (if (= 0 (a-sign path-atom)) 
		 (extend-plits plits path-atom)
		 (extend-nlits nlits path-atom))
	     (set! foldup-index -1)
;	     (set! pi path-index) (set! rv path-atom)
	     (generate-lemma1 (- path-index 1) rel-vector plits nlits
			      0 #f))))))

(define (generate-lemma1 path-index rel-vector plits nlits relevance
			 last-rel-index)
; lemma clause has length 1 
 (if (>= path-index first-branching)
     (begin (set! path-atom (vector-ref path-vector path-index))
       (if (atom-irrelevant? rel-vector path-atom) ; complement of literal
						   ; not in lemma
	   (begin (if (a-left-branch path-atom) ; lemma may remove search space
		      (begin (set! foldup-index (+ foldup-index 1))
			(set! last-rel-index path-index)
			(vector-set! foldup-stack foldup-index path-atom)
			(set! relevance (+ lemma-relevance relevance))))
		  (generate-lemma1 (- path-index 1) rel-vector
				   plits nlits relevance last-rel-index))
	   ; complement of literal is in lemma
	   (if (> max-lemma-length 1)
	       (begin (if (= 0 (a-sign path-atom)) 
			  (extend-plits plits path-atom)
			  (extend-nlits nlits path-atom))
		      (generate-lemma (- path-index 1) rel-vector plits
				      nlits 2 relevance last-rel-index))
	       ; below only unit lemma generated
	       (if (> relevance 0)
		   (begin (make-lemma rel-vector path-atom plits nlits #f)
		     (set! path-index (+ 1 path-index))
		     (set! path-atom (vector-ref path-vector path-index))
		     ; on backtracking path-atom, lemma is deleted:
		     (set-a-lemmas-out path-atom
				       (cons lemma (a-lemmas-out path-atom)))
		     (insert-lemma lemma))))))
     ; path-index < first-branching:
     (if (> relevance 0) ; lemma is made and inserted for successor of root:
	 (begin (make-lemma #f #f plits nlits #f)
		; this unit lemma is context free and hence no more deleted
		(insert-lemma lemma)))))

(define (generate-lemma path-index rel-vector plits nlits length relevance
			last-rel-index)
; lemma clause has length > 1
 (if (>= path-index first-branching)
     (begin (set! path-atom (vector-ref path-vector path-index))
       (if (atom-irrelevant? rel-vector path-atom) ; complement of literal
						   ; not in lemma
	   (begin (if (a-left-branch path-atom) ; lemma may remove search space
		      (begin (set! last-rel-index path-index)
			     (set! relevance (+ lemma-relevance relevance))))
	     (generate-lemma (- path-index 1) rel-vector plits nlits
			     length relevance last-rel-index))
	   (begin ; complement of literal is in lemma
	     (if (and (< length max-lemma-length) 
		      (or (= 0 lemma-lengthrel)
			  (< length (* relevance lemma-lengthrel))))
		 (begin ; lemma length is increased
		   (if (= 0 (a-sign path-atom)) ; atom is in lemma
		       (extend-plits plits path-atom)
		       (extend-nlits nlits path-atom))
		   (generate-lemma (- path-index 1) rel-vector plits nlits
				   (+ 1 length) relevance last-rel-index))
		 ; max lemma length is reached or other filter criterion:
		 (if (> relevance length) ; lemma made and inserted
					  ; for successor of path-atom
		     (begin (make-lemma rel-vector path-atom plits nlits
					last-rel-index)
		       (set! path-index (+ path-index 1))
		       (set! path-atom (vector-ref path-vector path-index))
		       ; on backtracking path-atom, lemma is deleted:
		       (set-a-lemmas-out path-atom 
					 (cons lemma (a-lemmas-out path-atom)))
		       (insert-lemma lemma)))))))
     ; path-index < first-branching:
     (if (> relevance length) ; lemma made and inserted for successor of root
	 (begin (make-lemma #f #f plits nlits #f)
		(set! root-lemmas (cons lemma root-lemmas))
		(set! nr-root-lemmas (+ nr-root-lemmas 1))
		; this non-unit lemma is context free and hence no more deleted
		(insert-lemma lemma)))))

(define (make-lemma rel-vector atom plits nlits last-rel-index)
; atom is lowest path atom at which lemma becomes invalid
 (set! nr-gen-lemmas (+ 1 nr-gen-lemmas))
 (set! nr-kept-lemmas (+ 1 nr-kept-lemmas))
 (set! lemma-diff 0)
 (set! lemma (make-vector 8))
 (set-c-activating-atom lemma #f)
 (set-c-length lemma 0)
 (set-c-usages lemma 0)
 (vector-set! lemma 7 inference)
;  (if (= inference 23768)
;      (begin (display "Valid up to depth: ") (display last-rel-index)
; 	    (display "atom at path-index 568 is: ")
; 	    (display (a-realname (vector-ref path-vector 568)))
; 	    (newline)))
; (if (and (eq? plits '()) (eq? nlits '()))
;		(begin (print "ERROR before delete-above-last-rel") (repl)))
 (if (and #f last-rel-index)
     (begin (set! path-atom (vector-ref path-vector last-rel-index))
            (set! plits (delete-above-last-rel (a-last-depth path-atom) plits))
            (set! nlits (delete-above-last-rel (a-last-depth path-atom) nlits))
	    (set! atom (vector-ref path-vector (- last-rel-index 1)))
;	    (if (and (eq? plits '()) (eq? nlits '()))
;		(begin (print "ERROR in delete-above-last-rel")
;		       (print inference) (repl)))
		))
 (set-c-pos-literals lemma plits)
 (set-c-neg-literals lemma nlits)
; (set-c-relevance lemma nlits)
 (insert-unit-lemmas foldup-index lemma)
 (if atom
     (begin
       (set-c-rel-vector lemma (make-vector (+ nr-atoms30 1) 0))
       (merge-boolean-vectors (c-rel-vector lemma) rel-vector
			      (a-last-depth30 atom))
       (subtract-from-rel-vector plits (c-rel-vector lemma))
       (subtract-from-rel-vector nlits (c-rel-vector lemma))
       (set-c-rel-depth30 lemma (a-last-depth30 atom)))
     (set-c-rel-vector lemma #f)) ; empty context, hence lemma will be treated
 (if proof                        ; as if input clause
     (begin (display " context lemma: ") (display-clause lemma)
	    (if atom
		(begin (display " up to excluding atom ")
		       (print (a-realname atom)))
		(print " up to root")))))

(define (subtract-from-rel-vector atoms rel-vector)
 (if (pair? atoms)
     (begin
       (vector-set! rel-vector (a-last-depth30 (car atoms))
	      (bit-and (bit-not (a-last-depth-mod30exp (car atoms)))
		       (vector-ref rel-vector (a-last-depth30 (car atoms)))))
       (subtract-from-rel-vector (cdr atoms) rel-vector))))

(define (delete-above-last-rel depth lits)
 (if (pair? lits)
     (if (< (a-last-depth (car lits)) depth)
	 (delete-above-last-rel depth (cdr lits))
	 lits)
     '()))

(define (insert-unit-lemmas index clause)
 (if (>= index 0)
     (begin (set-a-unit-lemmas (vector-ref foldup-stack index)
	      (cons clause
		    (a-unit-lemmas (vector-ref foldup-stack index))))
	    (insert-unit-lemmas (- index 1) clause))))
 
(define (delete-lemmas clauses)
 (if (pair? clauses)
     (begin (set-c-activating-atom (car clauses) #t)
	    (set-c-pos-literals (car clauses) '())
	    (set-c-neg-literals (car clauses) '())
	    (set! nr-kept-lemmas (- nr-kept-lemmas 1))
	    (set! lemma-usages (+ (c-usages (car clauses)) lemma-usages))
	    (if (> (c-usages (car clauses)) 0)
		(set! used-lemmas (+ 1 used-lemmas)))
	    (delete-lemmas (cdr clauses)))))

; (define (delete-pos-lemma atoms clause)
;  (if (pair? atoms)
;      (begin (once-cdr-remq! clause (a-pos-lemmas (car atoms))
; 			    (cdr (a-pos-lemmas (car atoms))))
; 	    (delete-pos-lemma (cdr atoms) clause))))

; (define (delete-neg-lemma atoms clause)
;  (if (pair? atoms)
;      (begin (once-cdr-remq! clause (a-neg-lemmas (car atoms))
; 			    (cdr (a-neg-lemmas (car atoms))))
; 	    (delete-neg-lemma (cdr atoms) clause)))) 

; (define (once-cdr-remq! element list cdr-list)
; ; destructively removes first occurrence of element from cdr of nonempty list
; ; #unspecified is returned
;  (if (pair? cdr-list)
;      (if (eq? element (car cdr-list))
; 	 (set-cdr! list (cdr cdr-list))
; 	 (once-cdr-remq! element cdr-list (cdr cdr-list)))))

(define (insert-lemma clause)
 (insert-pos-lemma (c-pos-literals clause) clause)
 (insert-neg-lemma (c-neg-literals clause) clause))

(define (insert-pos-lemma atoms clause)
 (if (pair? atoms)
     (begin
       (if (pair? (a-pos-lemmas (car atoms)))
	   (set-cdr! (a-pos-lemmas (car atoms))
		     (cons clause (cdr (a-pos-lemmas (car atoms)))))
	   (begin
	     (display-atom (car atoms))
	     (display "ERROR IN LEMMA INSERTION" ) (print inference) (exit 1)
	     ))
      (insert-pos-lemma (cdr atoms) clause))))

(define (insert-neg-lemma atoms clause)
 (if (pair? atoms)
     (begin
       (if (pair? (a-pos-lemmas (car atoms)))
	   (set-cdr! (a-neg-lemmas (car atoms))
		     (cons clause (cdr (a-neg-lemmas (car atoms)))))
	   (begin
	     (display-atom (car atoms))
	     (display "ERROR IN LEMMA INSERTION" ) (print inference) (exit 1)
	     ))
       (insert-neg-lemma (cdr atoms) clause))))
