;;;*************************************************************************
;; ---------------------
;; UNIFICATION PROCEDURE
;; ---------------------
(module unification
	(include "dctp.sch")
	(export (unbind-vars) (unbind-vars-with-count) (ref term) (unify-lists k l) (unify-terms s t)
		(unbind-collect-vars variables) (unbind-last) (unbind-vars-and-undo-var-counter) (occurs-list var termlist)
		(occurs-term var term) (top-level-unify-stop s t) (gr-unify-lists-gathering-local-bindings k l))
	(import (display "display.scm") (globals "globals.scm") (free-index-variable index "index.scm") (index-variable-binding index "index.scm")))

(define (ref term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
  (if (index-variable term) 
      (if (positive? term)
	  (if (free-index-variable term) 
	      term
	      (ref (index-variable-binding term)))
	  term)
      (if (or (non-variable term) (free-variable term) (domain-element term))
	  term
	  (ref (car term)))))

(define (unify-lists k l)
;; unifies termlists k and l.
;; If unification succeeds, #t is returned, else #f.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     #t
     (if (unify-terms (ref (car k)) (ref (car l)))
	 (unify-lists (cdr k) (cdr l))
	 #f)))

(define (unify-terms s t)          
;; unifies terms s and t. 
;; If unification succeeds, #t is returned, else #f.
;; Side effect: terms and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (eq? s t)
     #t
     (if (variable s)
         (if (variable t) 
             (bind s t)
	     (if (occurs-list s (cdr t))
                 (unbind-vars)
  	         (bind s t)))
         (if (variable t)
             (if (occurs-list t (cdr s))
	         (unbind-vars)
                 (bind t s))
             (if (eq? (car s) (car t))
	         (unify-lists (cdr s) (cdr t))
                 (unbind-vars))))))

(define (occurs-list var termlist)              
;; returns #t if var occurs in termlist, and #f otherwise
 (if (null? termlist)
     #f
     (if (occurs-term var (ref (car termlist)))
         #t
         (occurs-list var (cdr termlist)))))

(define (occurs-term var term)       
;; returns #t if var occurs in term, and #f otherwise
 (if (eq? var term) 
     #t
     (and (non-variable term) (occurs-list var (cdr term)))))


(define (top-level-unify-stop s t)
  (let ((ref-s (ref s)) (ref-t (ref t)))
    (if (and (non-variable ref-s) (non-variable ref-t)
	     (eq? (car ref-s) (car ref-t)))
	(begin
	  (trail-stop)
	  #t)
	(if (and (non-variable ref-s) (variable ref-t))
	    (begin
	      (trail-stop)
	      (bind ref-t ref-s)
	      #t)
	    (if (and (variable ref-s) (non-variable ref-t))
		(begin
		  (trail-stop)
		  (bind ref-s ref-t)
		  #t)
		#f)))))
  
(define (gr-unify-lists-gathering-local-bindings k l)
;; gr-unifies termlists k and l.
;; If unification succeeds, #t is returned, else #f.
;; Side effect: termlists and trail destructively modified.
;; Before every unification, one has to call (trail-stop) in order
;; to mark the trail position for the stored substitution
 (if (null? k)
     '()
     (let ((unify-subterm (gr-unify-terms (ref (car k)) (ref (car l)))))
       (if unify-subterm
	   (let ((unify-restterms (gr-unify-lists-gathering-local-bindings (cdr k) (cdr l))))
	     (if unify-restterms
		 (append unify-subterm unify-restterms)))
	   (begin
	     (display "subterm unification of ")
	     (display-ref (car k))
	     (display " and ")
	     (display-ref (car l))
	     (display " failed!!!!!")
	     (newline)
	     #f)))))

(define (gr-unify-terms s t)          
  (if (eq? s t)
      '()
      (if (local-variable s)
	  (if (local-variable t)
	      (begin (bind s t) (list (cons s t)))
	      (if (and (not (variable t)) (occurs-list s (cdr t)))
		  (begin (print "occurs-check of s in t failed") (unbind-vars))
		  (begin (bind s t) (list (cons s t)))))
	  (if (variable s)
	      (if (local-variable t)
		  (begin (bind t s) '())
		  (if (variable t)
		      '()
		      (begin (print "non-variable t with shared s failed") (unbind-vars))))
	      (if (local-variable t)
		  (if (occurs-list t (cdr s))
		      (begin (print "occurs-check of t in s failed") (unbind-vars))
		      (begin (bind t s) '()))
		  (if (variable t)
		      (begin (print "variable t with non-variable s failed") (unbind-vars))
		      (if (eq? (car s) (car t))
			  (gr-unify-lists-gathering-local-bindings (cdr s) (cdr t))
			  (begin (print "non-equal functors failed") (unbind-vars)))))))))
		     



;;;END *********************************************************************
;;;*************************************************************************
;; -----------------------------
;; TRAIL MODIFICATION PROCEDURES

(define (unbind-vars)
;; makes free all variables back to the previous trailstop
 (if (trail-content)
     (begin 
       ;;(display "unbinding ") 
       ;;(if (variable (caar trail)) (display "variable "))
       ;;(if (caaar trail)
	   ;;(display-ref (ref (caar trail)))
	   ;;(display (caaar trail)))
       (set-car! (caar trail) '()) 
       ;;(display " to ") 
       ;;(display-ref (caar trail))
       ;;(newline)
       (trail-previous) (unbind-vars))
     (begin 
       (trail-previous) 
       (if (not (pair? trail))
	   (error/location "unbind-vars" "argument not a pair" trail "unification.scm" 5607))
       #f)))

(define (unbind-vars-and-undo-var-counter)
;; makes free all variables back to the previous trailstop, decrementing var-counter for each freed variable
 (if (trail-content)
     (begin 
       (set-car! (caar trail) '()) 
       (set! var-counter (- var-counter 1)) 
       (trail-previous) 
       (unbind-vars-and-undo-var-counter))
     (begin (trail-previous) #f)))

(define (unbind-vars-with-count)
;; makes free all variables back to the previous trailstop, returning the number of freed vars
 (if (trail-content)
     (let ((binding (ref (caar trail))))
       (set-car! (caar trail) '()) (trail-previous) 
       (if (variable binding)
	   (unbind-vars-with-count)
	   (+ 1 (unbind-vars-with-count))))
     (begin (trail-previous) 0)))

(define (unbind-last)
;; unbinds the last bound variable on the trail
 (if (trail-content)
     (begin (set-car! (caar trail) '()) (trail-previous))
     (begin (trail-previous) #f)))

(define (unbind-collect-vars variables)
;; makes free and collects all variables back to the previous trailstop
 (if (trail-content)
     (begin (set-car! (caar trail) '()) 
	    (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (unbind-collect-vars variables))
     (begin (trail-previous) variables)))

(define (last-binding-variables)
 (let ((old-trail trail) variables)
      (set! variables (last-binding-vars '()))
      (set! trail old-trail)
      variables))

(define (last-binding-vars variables)
 (if (trail-content)
     (begin (set! variables (cons (caar trail) variables))
	    (trail-previous)
	    (last-binding-vars variables))
     variables))
