
/******************************************************************
*                                                                 *
* This file contains a function to test a clause against          *
* the current SFINDER model without altering it as well as one    *
* for updating the current model by adding a new clause to the    *
* theory.                                                         *
* Also contained in this file are functions for statistics        *
* determination.                                                  *
*                                                                 *
******************************************************************/


#include "SOFTIE_SCOTT.h"


/***********************************
* Returns true iff passed in clause
* is true in the current SFINDER model,
* returns false otherwise.
* Called from find_given_clause(...) in clause.c.
************************************/
boolean test_in_model(struct clause *cl) {

  dynstring tmpStr;              // SFINDER form of clause
  int rv;

  if (cl->pick_weight >= BUCKETS-1)
    return 0;

  dynstring_initialise(&tmpStr, "");
  convert_to_sfinder_clause(cl, 1, &tmpStr);  // get SFINDER input clause
  rv = found(&tmpStr, 0, 0);
  dynstring_free(&tmpStr);
  
  return rv;

} /**************** end function test_in_model(...) ****************/



/***********************************
* Updates SFINDER theory with passed in clause.
* Returns true iff passed in clause
* is true in the current SFINDER model,
* returns false otherwise.
* Called from find_given_clause(...) in clause.c.
************************************/
boolean update_model(struct clause *cl) {

  dynstring tmpStr;
  int old_model_clauses = model_clauses;
  int rc;

  if (cl->pick_weight >= BUCKETS-1)
    return 0;

  dynstring_initialise(&tmpStr, "");
  convert_to_sfinder_clause(cl, 1, &tmpStr);  // get SFINDER input clause
  CLOCK_START(FINDER_TIME);
  rc = found(&tmpStr, 1, 1);                  // do it (the test/update)
  CLOCK_STOP(FINDER_TIME);
  dynstring_free(&tmpStr);
  if (model_clauses > old_model_clauses)
    retest_list(Sos);
  return rc;

} /**************** end function update_model(...) ****************/
