/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
//  file Statistics.hpp
//  defines class Statistics
//


#ifndef __Statistics__
#define __Statistics__


// class ostream;

#include "VampireKernel.hpp"


using namespace std;


// ******************* class Statistics, definition *********************


/**
 * Implements statistics about Vampire runs.
 */
class Statistics {
 public:
  // constructor
  Statistics (const VampireKernel &);
  void output (ostream&) const;
  void toXML (ostream&) const;
  void tabulate (ostream&) const;
  void result (char res);

 private:
  // general
  const float _versionNumber;
  const char* _versionDescription;
  const VampireKernel::TerminationReason _terminationReason;
  const float _time;   // seconds
  const long _memory;  // bytes
  // number of various categories of clauses
  const long _generated;
  const long _retained;
  const long _used;
  const long _currentlyActive;
  const long _currentlyPassive;
  // generated inferences
  const long _resolutions;
  const long _forwardSuperpositions;
  const long _backwardSuperpositions;
  const long _equalityResolutions;
  const long _equalityFactorings;
  const long _preprocessed; 
  const long _reanimated; 
  bool anyGeneratingInference () const;
  // simplifying inferences
  const long _forwardSubsumptionResolutions;
  const long _forwardDemodulated;
  const long _backwardDemodulated; 
  const long _theoryInferences;
  bool anySimplifyingInference () const;
  // redundant
  const long _forwardSubsumed;
  const long _backwardSubsumed;
  const long _discardedPropositionalTautologies;
  const long _discardedEquationalTautologies;
  const long _murderedPassiveOrphans;
  const long _murderedActiveOrphans;
  bool anyRedundantClause () const;
  // splitting
  const long _splittingInferences;
  const long _differentSplitComponents;
  const float _averageLiteralsPerComponent;
  const long _forwardDemodulatedOnBranches;
  const long _backwardDemodulatedOnBranches;
  bool anySplitting () const;
  // discarded
  const long _discardedWithTooHeavyLiterals;
  const long _discardedTooDeep;
  const long _discardedTooHeavy;
  const long _discardedWithTooManyVariables;
  const long _newClausesDiscardedDueToLackOfMemory;
  const long _recycledDueToLackOfResources;
  const char* _result;
  bool anyDiscarded () const;

  // tabulation
  static void tabulate (ostream&, const char* field, const char* value);
  static void tabulate (ostream&, const char* field, int value);
  static void tabulate (ostream&, const char* field, long value);
  static void tabulate (ostream&, const char* field, float value);
  static void startTab (ostream&, const char* field);
  static void endTab (ostream&);
  // XML output
  static void toXML (ostream&, const char* field, const char* value);
  static void toXML (ostream&, const char* field, int value);
  static void toXML (ostream&, const char* field, long value);
  static void toXML (ostream&, const char* field, float value);
  static void startXMLElement (ostream&, const char* field);
  static void endXMLElement (ostream&);
  static void endEmptyXMLElement (ostream&);
  void outputTerminationReason (ostream& str) const;
}; // class Statistics


// ******************* class Statistics, implementation *********************

/**
 * Output statistics to a stream.
 * 
 * @param str the output stream.
 * @param stat the statistics.
 */
inline 
ostream& operator << (ostream& str, const Statistics& stat)
{
  stat.output (str);
  return str;
} // ostream& operator << (ostream& str, const Statistics& stat)


#endif // __Statistics__
