/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
/** This software is an adaptation of the theorem prover Vampire for
 * working with large knowledge bases in the KIF format, see 
 * http://www.prover.info for publications on Vampire.
 *
 * Copyright (C) Andrei Voronkov and Alexandre Riazanov
 *
 * @author Alexandre Riazanov <riazanov@cs.man.ac.uk>
 * @author Andrei Voronkov <voronkov@cs.man.ac.uk>, <andrei@voronkov.com>
 *
 * @date 06/06/2003
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//
//  file Substitution.cpp
//  implements classes Binding and Substitution
//  currently not a very fast implementation
//


#include "Substitution.hpp"


// 03/10/2002 Manchester
bool Substitution::bound (Var v, Term& t) const
{
  TRACER( "Substitution::bound" );

  Iterator<Binding> bs (*this);
  while (bs.more()) {
    Binding b (bs.next ());
    if (b.var() == v) {
      t = b.term();
      return true;
    }
  }

  // not found
  return false;
} // Substitution::get


// collect into VarList lhs of all bindings
// 03/08/2002 Torrevieja
// 05/09/2002 Trento, changed
// 03/10/2002 Manchester, changed to new data structures
void Substitution::domain (VarList& vs, Lst<Binding> bs) const
{
  if (bs.isEmpty()) {
    return;
  }

  VarList tl;
  domain (tl, bs.tail());

  vs = VarList (bs.head().var(), tl);
} // Substitution::domain


// collect into VarList rhs of all bindings
// 03/10/2002 Manchester, copied from domain ()
void Substitution::range (TermList& ts, Lst<Binding> bs) const
{
  if (bs.isEmpty()) {
    return;
  }

  TermList tl;
  range (tl, bs.tail());

  ts = TermList (bs.head().term(), tl);
} // Substitution::range


// undo the last binding for v
// 03/10/2002 Manchester
void Substitution::undoBinding (Var v)
{
  *this = undoBinding (v, *this);
} // Substitution::undoBinding


// undo the last binding for v
// 03/10/2002 Manchester
Lst<Binding> Substitution::undoBinding (Var v, Lst<Binding> bs)
{
  ASS ( bs.isNonEmpty() );

  if (bs.head().var() == v) {
    return bs.tail();
  }

  return Lst<Binding> (bs.head(), undoBinding(v,bs.tail()));
} // Substitution::undoBinding


// t is in the range of the substitution
// WARNING: may not work correctly if the substitution contains
// multiple bindings of the same variable
bool Substitution::inRange (Term t) const
{
  Iterator<Binding> bs (*this);
  while ( bs.more() ) {
    if ( t.equal(bs.next().term()) ) {
      return true;
    }
  }

  return false;
} // Substitution::inRange
