/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         AutoFunctorPrecedence.cpp
// Description:  Automatic assignment of precedences to the functors and headers of the signature.
// Created:      Apr 24, 2001, 17:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "AutoFunctorPrecedence.h"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_AUTO_FUNCTOR_PRECEDENCE
 #define DEBUG_NAMESPACE "AutoFunctorPrecedence"
#endif 
#include "debugMacros.h"
//=================================================  

long VK::AutoFunctorPrecedence::InitStatic::_count; // Implicitely initialised to 0L

#ifdef DEBUG_ALLOC_OBJ_TYPE
#include "ClassDesc.h"
using namespace BK;
ClassDesc 
VK::AutoFunctorPrecedence::ArityFuncSortTree::LeftRightIter::NodeList::_classDesc("AutoFunctorPrecedence::ArityFuncSortTree::LeftRightIter::NodeList",
									      VK::AutoFunctorPrecedence::ArityFuncSortTree::LeftRightIter::NodeList::minSize(),
									      VK::AutoFunctorPrecedence::ArityFuncSortTree::LeftRightIter::NodeList::maxSize());

ClassDesc 
VK::AutoFunctorPrecedence::ArityFuncSortTree::RightLeftIter::NodeList::_classDesc("AutoFunctorPrecedence::ArityFuncSortTree::RightLeftIter::NodeList",
									      VK::AutoFunctorPrecedence::ArityFuncSortTree::RightLeftIter::NodeList::minSize(),
									      VK::AutoFunctorPrecedence::ArityFuncSortTree::RightLeftIter::NodeList::maxSize());

ClassDesc 
VK::AutoFunctorPrecedence::ArityFuncSortTree::Node::_classDesc("AutoFunctorPrecedence::ArityFuncSortTree::Node",
							       VK::AutoFunctorPrecedence::ArityFuncSortTree::Node::minSize(),
							       VK::AutoFunctorPrecedence::ArityFuncSortTree::Node::maxSize());

 
#endif


namespace VK
{
AutoFunctorPrecedence::AutoFunctorPrecedence() 
  : _dependenceOnArity(0L),
    _inversePolarity(false),
    _kinkyHeaderPrecedence(false),
    _somePrecedenceSet(false),
    _minimalFixedPrecedence(0L),
    _maximalFixedPrecedence(0L),
    _minFunctorNumber(0UL),
    _maxFunctorNumber(0UL),
#ifdef DEBUG_NAMESPACE
    _precedence(0L,"AutoFunctorPrecedence::_precedence"), 
    _precedenceIsFixed(false,"AutoFunctorPrecedence::_precedenceIsFixed"),
    _arity(0UL,"AutoFunctorPrecedence::_arity"),
    _registeredFunctor(false,"AutoFunctorPrecedence::_registeredFunctor")
#else
  _precedence(0L,0), 
  _precedenceIsFixed(false,0),
  _arity(0UL,0),
  _registeredFunctor(false,0)
#endif
{
  CALL("constructor AutoFunctorPrecedence()");   
};


AutoFunctorPrecedence::~AutoFunctorPrecedence()
{
  CALL("destructor ~AutoFunctorPrecedence()");
};

bool AutoFunctorPrecedence::precedenceIsReserved(long precedence,ulong& funNum)
{
  CALL("precedenceIsReserved(long precedence,ulong& funNum)");
  if (!_somePrecedenceSet) return false;
  long normPrec = normalisedPrecedence(precedence);  
  if ((normPrec < _minimalFixedPrecedence) ||
      (normPrec > _maximalFixedPrecedence))
    {
      return false;
    };    

  for (ulong fn = 0; fn < _registeredFunctor.size(); fn++)
    if ((_registeredFunctor[fn]) && 
	(_precedenceIsFixed.sub(fn)))
      {
	if (_precedence[fn] == normPrec)
	  {
	    funNum = fn;
	    return true;
	  };
      };
  return false;
}; // bool AutoFunctorPrecedence::precedenceIsReserved(long precedence,ulong& funNum)


void AutoFunctorPrecedence::assignFixed(ulong funNum,long precedence) 
{ 
  CALL("assignFixed(ulong funNum,long precedence)");
  _registeredFunctor.sub(funNum) = true;
  long normPrec = normalisedPrecedence(precedence);  
  if (!_somePrecedenceSet) 
    { 
      _somePrecedenceSet = true;
      _minimalFixedPrecedence = normPrec;
      _maximalFixedPrecedence = normPrec;
    }
  else 
    if (normPrec > _maximalFixedPrecedence) { _maximalFixedPrecedence = normPrec; }
    else 
      if (normPrec < _minimalFixedPrecedence) { _minimalFixedPrecedence = normPrec; };
   
  _precedence.sub(funNum) = normPrec;    
  _precedenceIsFixed.sub(funNum) = true;   
};

void AutoFunctorPrecedence::assignMaximalPrecedence(ulong funNum)
{
  _registeredFunctor.sub(funNum) = true;
  _somePrecedenceSet = true;
  _precedence.sub(funNum) = maximalPrecedence();
  _precedenceIsFixed.sub(funNum) = true;    
};

void AutoFunctorPrecedence::assignArity(ulong funNum,ulong arity) 
{ 
  CALL("assignArity(ulong funNum,ulong arity)");

  _registeredFunctor.sub(funNum) = true; 
  _arity.sub(funNum) = arity;
};
  
void AutoFunctorPrecedence::assignAllPrecedences() 
{ 
  CALL("assignAllPrecedences()");
  
  _maximalFixedPrecedence = max(abs(_maximalFixedPrecedence),abs(_minimalFixedPrecedence));
  _minimalFixedPrecedence = -_maximalFixedPrecedence;

  if (_dependenceOnArity)
    { 
      ArityFuncSortTree sortTree;   

      for (ulong fn = 0; fn < _registeredFunctor.size(); fn++)
	if ((_registeredFunctor[fn]) && 
	    (!_precedenceIsFixed.sub(fn))) 
	  sortTree.insert(ArityFuncPair(_arity.sub(fn),fn));

      if (_dependenceOnArity > 0L)     
	{
	  // greater arity => greater precedence 
	  ArityFuncSortTree::RightLeftIter sortTreeIter(sortTree);      
	  while (sortTreeIter)
	    { 
	      assignSmallerPrecedence(sortTreeIter.value().el2); 
	      sortTreeIter.next();     
	    };
	}
      else // _dependenceOnArity < 0L 
	{
	  // greater arity => smaller precedence   
	  ArityFuncSortTree::LeftRightIter sortTreeIter(sortTree);
	  while (sortTreeIter)
	    { 
	      assignSmallerPrecedence(sortTreeIter.value().el2); 
	      sortTreeIter.next();     
	    };
	};
    }  
  else // !_dependenceOnArity
    {
      for (ulong fn = 0; fn < _registeredFunctor.size(); fn++)
	if ((_registeredFunctor[fn]) && 
	    (!_precedenceIsFixed.sub(fn)))
	  assignSmallerPrecedence(fn);
    };
}; // void AutoFunctorPrecedence::assignAllPrecedences() 

void AutoFunctorPrecedence::assignSmallerPrecedence(ulong funNum) 
{ 
  CALL("assignSmallerPrecedence(ulong funNum)");
  _minimalFixedPrecedence = previousNormalisedPrecedence(_minimalFixedPrecedence);
  _precedence.sub(funNum) = _minimalFixedPrecedence;
};


}; // namespace VK


//====================================================
