/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         BuiltInTheoryDispatcher.cpp
// Description:  Built-in theory abstraction.  
// Created:      Jul 06, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//==================================================================
#include <cstring>
#include "GlobAlloc.hpp"
#include "BuiltInTheoryDispatcher.hpp"
#include "BuiltInFloatingPointArithmetic.hpp"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BUILT_IN_THEORY_DISPATCHER
 #define DEBUG_NAMESPACE "BuiltInTheoryDispatcher"
#endif
#include "debugMacros.hpp"
//===================================================================

using namespace BK;
using namespace VK;

BuiltInTheoryDispatcher* BuiltInTheoryDispatcher::_current;

BuiltInTheoryDispatcher::BuiltInTheoryDispatcher(Clause::NewNumber* clauseNumberGenerator) :
  _clauseNumberGenerator(clauseNumberGenerator),
  _floatingPointArithmetic(0)
{
};

BuiltInTheoryDispatcher::BuiltInTheoryDispatcher() 
{
};


BuiltInTheoryDispatcher::~BuiltInTheoryDispatcher()
{
  CALL("destructor ~BuiltInTheoryDispatcher()");
  if (_floatingPointArithmetic) delete _floatingPointArithmetic;
};


void* BuiltInTheoryDispatcher::operator new(size_t)
{
  CALL("operator new(size_t)");
  return GlobAlloc::allocate(sizeof(BuiltInTheoryDispatcher));
}; // void* BuiltInTheoryDispatcher::operator new(size_t)

void BuiltInTheoryDispatcher::operator delete(void* obj)
{
  CALL("operator delete(void* obj)");
  GlobAlloc::deallocate(obj,sizeof(BuiltInTheoryDispatcher));
}; // void BuiltInTheoryDispatcher::operator delete(void* obj)

bool BuiltInTheoryDispatcher::interpret(ulong inputSymbolId,
					const char* theoryName,
					const char* nativeSymbol)
{
  CALL("interpret(..)");
  BuiltInTheory* th = theory(theoryName);
  if (!th) return false;
  if (!th->interpret(inputSymbolId,nativeSymbol)) return false;
  return true;
}; // bool BuiltInTheoryDispatcher::interpret(ulong inputSymbolId,onst char* theoryName,const char* nativeSymbol)



Clause* BuiltInTheoryDispatcher::simplify(TmpLiteral* lit,TmpLiteral* newLit)
{
  CALL("simplify(TmpLiteral* lit,TmpLiteral* newLit)");  
  Clause* result = 0;
  if (_floatingPointArithmetic)
    result = _floatingPointArithmetic->simplify(lit,newLit);
  if (result) return result;
  return 0;
}; // Clause* BuiltInTheoryDispatcher::simplify(TmpLiteral* lit,TmpLiteral* newLit)


      
BuiltInTheory* BuiltInTheoryDispatcher::theory(const char* theoryName)
{
  CALL("theory(const char* theoryName)");
  if (!strcmp(theoryName,"floating_point_arithmetic"))
    {
      if (!_floatingPointArithmetic) 
	{
	  _floatingPointArithmetic = 
	    new BuiltInFloatingPointArithmetic("floating_point_arithmetic",_clauseNumberGenerator);
	};
      return _floatingPointArithmetic;
    };
  return 0; // no theory with this name
}; // BuiltInTheory* BuiltInTheoryDispatcher::theory(const char* theoryName)



//==================================================================
