/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         Chaining.hpp
// Description:  Ordered chaining for transitive predicates.  
// Created:      Jun 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//==================================================================
#ifndef CHAINING_H
#define CHAINING_H
//==================================================================
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "premise_factoring.hpp"
#include "weight_formula.hpp"
#include "SimplificationOrdering.hpp"
#include "iu_machine.hpp"
namespace VK
{
  class TERM;
  class Clause;
  class NewClause;
  class SymbolInfo;
  class OR_INDEX;
  class RES_PAIR_LIST;
}; // namespace VK
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CHAINING
#  define DEBUG_NAMESPACE "Chaining"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{
  class Chaining
  {
  public:
    Chaining(NewClause* newClauseBuffer,OR_INDEX* unifIndex);
    ~Chaining();
    void init(NewClause* newClauseBuffer,OR_INDEX* unifIndex) ;
    void destroy();
    void setTransitivity(const SymbolInfo* table) 
    {
      _transitivity = table;
    };
    void setChainingThroughVariablesFlag(bool fl)
    {
      _chainingThroughVariables = fl;
    };
    void setMainPremiseSubsumption(bool fl)
    {
      _mainPremiseSubsumption = fl;
    };
    void setSidePremiseSubsumption(bool fl)
    {
      _sidePremiseSubsumption = fl;
    };


    void setMainSelectedLitDemodulation(bool fl)
    {
      _mainSelectedLitDemodulation = fl;
    };
    void setSideSelectedLitDemodulation(bool fl)
    {
      _sideSelectedLitDemodulation = fl;
    };


    ulong statBlockedChainingsThroughVariables() const
    {
      return _statBlockedChainingsThroughVariables;
    };
    
    ulong statSubsumedMainPremiseInstances() const
    {
      return _statSubsumedMainPremiseInstances;
    };
    ulong statSubsumedSidePremiseInstances() const
    {
      return _statSubsumedSidePremiseInstances;
    };

    ulong statDemodulatedMainSelectedLits() const
    {
      return _statDemodulatedMainSelectedLits;
    };
    ulong statDemodulatedSideSelectedLits() const
    {
      return _statDemodulatedSideSelectedLits;
    };
    

    bool loadQuery(Clause* cl);
    bool nextInference();
  private:
    enum OuterContinuation
    {
      LoadForwardPositive,
      NextForwardPositive,
      NextBackwardPositive,
      NextForwardNegative1,
      NextBackwardNegative1,
      NextForwardNegative2,
      NextBackwardNegative2
    }; // enum OuterContinuation

    enum InnerContinuation
    {
      FirstLit,
      FirstFactoring,
      NextFactoring
    }; // enum InnerContinuation
  private:
    
    bool nextForwardPositive();
    bool nextBackwardPositive();
    bool nextForwardNegative1();
    bool nextBackwardNegative1();
    bool nextForwardNegative2();
    bool nextBackwardNegative2();



    bool posChainingLitSelectedIn(Clause* cl);
    bool negChainingLitSelectedIn(Clause* cl);

    bool buildResolvent(PrefixSym* newLitArg1,const TERM& newLitArg2);
    
    bool buildResolvent(const TERM& newLitArg1,PrefixSym* newLitArg2);


    bool nextFactoring();

    bool subsumeMainPremise();
    bool collectMainPremiseInstance();
    bool subsumeSidePremise();
    bool collectSidePremiseInstance();


    bool demodulateSelectedLits();
    bool collectMainPremiseSubtermsAvailableForParamodulationInto();
    bool collectSidePremiseSubtermsAvailableForParamodulationInto();


  private:  
    const SymbolInfo* _transitivity;
    bool _chainingThroughVariables;
    Clause* _query;
    NewClause* _newClauseBuffer;
    OR_INDEX* _unifIndex;
    BK::Comparison _headTailComparison;
    OuterContinuation _outerContinuation;
    InnerContinuation _innerContinuation;
    OR_PREMISE_WITH_FACTORING _queryPremise;
    OR_PREMISE_WITH_FACTORING _storedPremise;
    ulong _queryVarIndex;
    ulong _storedVarIndex;
    WEIGHT_FORMULA _residualQuerySizeFormula;
    long _residualQuerySize;
    ulong _currentQueryLit;
    SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
    SimplificationOrdering::CompiledConstraintForUnifiers* _constraintOnQuery;
    SimplificationOrdering::ConstraintCheckOnUnifier* _storedConstraintCheck;
    SimplificationOrdering::StoredConstraint* _storedConstraint;
    TERM _pseudoHeader;
    IU_MACH _unificationMachine;
    const Clause* _currentTransitivityLaw;
    PrefixSym* _queryHead;
    PrefixSym* _queryTail;
    RES_PAIR_LIST* _currentCandidate;
    TERM _newLitHeader;
    ulong _statBlockedChainingsThroughVariables;
    bool _mainPremiseSubsumption;
    bool _sidePremiseSubsumption;
    ulong _statSubsumedMainPremiseInstances;
    ulong _statSubsumedSidePremiseInstances;
    bool _mainSelectedLitDemodulation;
    bool _sideSelectedLitDemodulation;    
    ulong _statDemodulatedMainSelectedLits;
    ulong _statDemodulatedSideSelectedLits;
    bool _mainSelectedLitDemodulationAttempted;
    bool _mainSelectedLitDemodulated;
  }; // class Chaining

}; // namespace VK

//=================================================================
/*
  Expalanatory scheme for notation.

  Forward positive chaining:
     Query: C \/ r(queryTail,head) : head > queryTail
     Stored: D \/ r(head,storedTail) : head > storedTail
     Conclusion: C \/ D \/ r(queryTail,storedTail)
     
  Backward positive chaining:
     Query: C \/ r(head,queryTail) : head > queryTail
     Stored: D \/ r(storedTail,head) : head > storedTail
     Conclusion: C \/ D \/ r(storedTail,queryTail)
     
  Forward negative chaining 1: 
     Query: C \/ r(head,queryTail) : head > queryTail
     Stored: D \/ ~r(head,storedTail) : head > storedTail
     Conclusion: C \/ D \/ ~r(queryTail,storedTail)

  Backward negative chaining 1:
    Query: C \/  ~r(head,queryTail) : head > queryTail
    Stored: D \/ r(head,storedTail) : head > storedTail
    Conclusion: C \/ D \/ ~r(storedTail,queryTail)
     
  Forward negative chaining 2: 
    Query: C \/ r(queryTail,head) : head > queryTail
    Stored: D \/ ~r(storedTail,head) : head >= storedTail
    Conclusion: C \/ D \/ ~r(storedTail,queryTail)
 
  Backward negative chaining 2: 
    Query: C \/ ~r(queryTail,head) : head >= queryTail
    Stored: D \/ r(storedTail,head) : head > storedTail
    Conclusion: C \/ D \/ ~r(queryTail,storedTail)


 */

//=================================================================
#endif
