/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         ClauseBackground.cpp
// Description:  Ancestors of a clause.
// Created:      Mar 3, 2000, 13:25
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk 
//==================================================================
#include "ClauseBackground.hpp"
//==================================================================
using namespace VK;
using namespace BK;

#ifdef DEBUG_ALLOC_OBJ_TYPE
 ClassDesc ClauseBackground::_classDesc("ClauseBackground",sizeof(ClauseBackground),sizeof(ClauseBackground));
 ClassDesc ClauseBackground::ClauseList::_classDesc("ClauseBackground::ClauseList",
                                                    ClauseBackground::ClauseList::minSize(),
                                                    ClauseBackground::ClauseList::maxSize());
#endif

//==================================================================
#ifdef DEBUG_CLAUSE_BACKGROUND
#undef DEBUG_NAMESPACE
#define DEBUG_NAMESPACE "ClauseBackground"
#endif
#include "debugMacros.hpp"
//==================================================================


      void ClauseBackground::copyTo(ClauseBackground& copy) const 
	{
	  CALL("ClauseBackground& copy");
	  if (isInput())
	    {
	      copy = *this;
	    }
	  else
	    {
	      ClauseList* orgClauses = _ancestors;
	      ClauseList* targClauses = 0;
	      ClauseList** targClausesAddr = &targClauses;      
	      while (orgClauses)
		{
		  *targClausesAddr = 
		    new ClauseList(orgClauses->hd(),static_cast<ClauseList*>(0));
		  targClausesAddr = &((*targClausesAddr)->tl());
		  orgClauses = orgClauses->tl();
		};
	      copy.init(_rules,targClauses,_inferenceDepth);
	    };
	}; // void ClauseBackground::copyTo(ClauseBackground& copy) const 




      ostream& ClauseBackground::outputRules(ostream& str,const BitWord& _rules) 
	{
	  if (_rules.bit(ClauseBackground::Input)) str << "in ";
	  if (_rules.bit(ClauseBackground::Prepro)) str << "pp ";
	  if (_rules.bit(ClauseBackground::Subgoal)) str << "sg ";
	  if (_rules.bit(ClauseBackground::BuiltInTheoryFact)) str << "bitf ";	
	  if (_rules.bit(ClauseBackground::Support)) str << "sprt ";  
	  if (_rules.bit(ClauseBackground::BinRes)) str << "br ";
	  if (_rules.bit(ClauseBackground::ForwSup)) str << "fs ";
	  if (_rules.bit(ClauseBackground::BackSup)) str << "bs ";
	  if (_rules.bit(ClauseBackground::EqRes)) str << "er ";
	  if (_rules.bit(ClauseBackground::EqFact)) str << "ef ";
	  if (_rules.bit(ClauseBackground::ForwDemod)) str << "fd ";
	  if (_rules.bit(ClauseBackground::BackDemod)) str << "bd ";
	  if (_rules.bit(ClauseBackground::EqResSimp)) str << "ers ";
	  if (_rules.bit(ClauseBackground::ForwSubsRes)) str << "fsr ";
	  if (_rules.bit(ClauseBackground::BackSubsRes)) str << "bsr ";
	  if (_rules.bit(ClauseBackground::Split)) str << "sp ";
	  if (_rules.bit(ClauseBackground::Reanimation)) str << "rea ";
	  if (_rules.bit(ClauseBackground::Name)) str << "nm ";
	  if (_rules.bit(ClauseBackground::SimplificationByBuiltInTheories)) str << "sbit ";
	  return str;
	}; //  ostream& ClauseBackground::outputRules(ostream& str,const BitWord& _rules) 


//========================================
#ifdef DEBUG_CLAUSE_BACKGROUND
#undef DEBUG_NAMESPACE
#define DEBUG_NAMESPACE "OpenClauseBackground"
#endif
#include "debugMacros.hpp"
//========================================

      void OpenClauseBackground::InverseOrder() 
	{
	  Clause** b = _ancestors.begin();
	  Clause** e = _ancestors.last();    
	  while (b < e) 
	    {
	      Clause* tmp = *b;
	      *b = *e;
	      *e = tmp;
	      b++;
	      e--;
	    };
	}; // void OpenClauseBackground::InverseOrder() 



      void OpenClauseBackground::writeNewBackgroundTo(ClauseBackground& b) 
	{
	  CALL("writeNewBackgroundTo(ClauseBackground& b)");
	  ClauseBackground::ClauseList* clauses = 0;

	  for (Clause** anc = _ancestors.begin(); anc < _ancestors.end(); anc++)
	    {
	      clauses = new ClauseBackground::ClauseList(*anc,clauses);
	      
	    };
     
	  b.init(_rules,clauses,_inferenceDepth);
	}; // void OpenClauseBackground::writeNewBackgroundTo(ClauseBackground& b) 




//==================================================================
