/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         ClauseNumIndex.cpp
// Description:  Searching clauses by numbers.
// Created:      Apr 08, 2000, 22:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#include "ClauseNumIndex.hpp"
#include "ClassDesc.hpp"
#include "Clause.hpp"
//===========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_NUM_INDEX
 #define DEBUG_NAMESPACE "ClauseNumIndex"
#endif
#include "debugMacros.hpp"
//============================================================================

using namespace BK;
using namespace VK;

#ifdef DEBUG_ALLOC_OBJ_TYPE
 ClassDesc ClauseNumIndex::List::_classDesc("ClauseNumIndex::List",
                                             ClauseNumIndex::List::minSize(),
                                             ClauseNumIndex::List::maxSize());

 ClassDesc ClauseNumIndex::List::Node::_classDesc("ClauseNumIndex::List::Node",
                                                   ClauseNumIndex::List::Node::minSize(),
                                                   ClauseNumIndex::List::Node::maxSize());
#endif

void ClauseNumIndex::insert(Clause* cl)
{
 CALL("insert(Clause* cl)");
 bool newNode; 
 List::Node* node = _list.insert(cl->Number(),newNode);
 ASSERT(newNode);
 node->value() = cl;
 ASSERT(!cl->flag(Clause::FlagInClauseNumIndex));
 cl->setFlag(Clause::FlagInClauseNumIndex);
}; // void ClauseNumIndex::insert(Clause* cl)

void ClauseNumIndex::remove(Clause* cl)
{
 CALL("remove(Clause* cl)");
 ASSERT(cl->flag(Clause::FlagInClauseNumIndex));
 Clause* removedClause;
 ALWAYS(_list.remove(cl->Number(),removedClause));
 ASSERT(removedClause == cl);
 ASSERT(cl->flag(Clause::FlagInClauseNumIndex));
 cl->clearFlag(Clause::FlagInClauseNumIndex);
}; // void ClauseNumIndex::remove(Clause* cl)

Clause* ClauseNumIndex::find(ulong num) 
{
 CALL("find(ulong num)");
 _surfer.reset(_list);
 const List::Node* node;
 if (_surfer.find(num,node)) return node->value();
 return 0;
}; // Clause* ClauseNumIndex::find(ulong num) 

//============================================================================
           
