/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         ClauseNumIndex.h
// Description:  Searching clauses by numbers.
// Created:      Apr 08, 2000, 22:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef CLAUSE_NUM_INDEX_H
//============================================================================
#define CLAUSE_NUM_INDEX_H
#include "jargon.h"
#include "VampireKernelDebugFlags.h"    
#include "SkipList.h"
#include "GlobAlloc.h"
#include "DestructionMode.h"
namespace VK
{
class Clause;
}; // namespace VK
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_NUM_INDEX
 #define DEBUG_NAMESPACE "ClauseNumIndex"
#endif
#include "debugMacros.h"
//============================================================================
namespace VK
{
  using namespace BK;
class ClauseNumIndex
{
 public:
  ClauseNumIndex() : _list(), _surfer() 
  {
   CALL("constructor ClauseNumIndex()");
  };
  ~ClauseNumIndex() 
    { 
      CALL("destructor ~ClauseNumIndex()"); 
      if (DestructionMode::isThorough()) reset(); 
    };
  void insert(Clause* cl);
  void remove(Clause* cl);
  const Clause* find(ulong num);
  void reset()
  {
   _list.reset();  
  };
  #ifndef NO_DEBUG
   bool clean() const { return !_list; };
  #endif
 private:
  typedef SkipList<GlobAlloc,const Clause*,ulong,2,28,ClauseNumIndex> List;
 private:
  List _list;
  List::Surfer _surfer;
}; // class ClauseNumIndex

}; // namespace VK

//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
