/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         ClauseNumIndex.hpp
// Description:  Searching clauses by numbers.
// Created:      Apr 08, 2000, 22:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef CLAUSE_NUM_INDEX_H
//============================================================================
#define CLAUSE_NUM_INDEX_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"    
#include "SkipList.hpp"
#include "GlobAlloc.hpp"
#include "DestructionMode.hpp"
namespace VK
{
class Clause;
}; // namespace VK
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_NUM_INDEX
 #define DEBUG_NAMESPACE "ClauseNumIndex"
#endif
#include "debugMacros.hpp"
//============================================================================
namespace VK
{
class ClauseNumIndex
{  
 public:
  // List cannot be made private since it has a static data member
  typedef BK::SkipList<BK::GlobAlloc,Clause*,ulong,2,28,ClauseNumIndex> List;
 public:
  ClauseNumIndex() 
  {
   CALL("constructor ClauseNumIndex()");
  };
  ~ClauseNumIndex() 
    { 
      CALL("destructor ~ClauseNumIndex()");  
    };
  void init()
  {
    CALL("init()");
    _list.init();
    _surfer.init();
  };

  void destroy()
  {
    CALL("destroy()");
    
    _surfer.destroy();
    _list.destroy();
  };
  void reset()
  {
   _list.reset();  
  };

  void insert(Clause* cl);
  void remove(Clause* cl);
  Clause* find(ulong num);
  #ifndef NO_DEBUG
   bool clean() const { return !_list; };
  #endif
 private:
  List _list;
  List::Surfer _surfer;
}; // class ClauseNumIndex

}; // namespace VK

//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
