/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         CustomLiteralSelection.cpp
// Description:  Flexible schemes for literal selection.
// Created:      Jun 03, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//============================================================================
#include "CustomLiteralSelection.hpp"
using namespace BK;
using namespace VK;
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::~CustomLiteralSelection()
{
};

void 
CustomLiteralSelection::select(TmpLitList& lits,
			       TmpLitList& selectedLits,
			       TmpLitList& nonselectedLits,
			       bool& allMaximalSelected,
			       const TmpLitList::Statistics& stat)
{
  CALL("select(..)");    
  ASSERT(!lits.containsAnswerLiterals());    
  ASSERT(!lits.containsSplittingBranches());
  DOP(ulong debugTotalNumOfLits = 
      lits.numOfAllLiterals() + 
      selectedLits.numOfAllLiterals() +
      nonselectedLits.numOfAllLiterals());

  ulong numOfLits = lits.numOfAllLiterals();
  ASSERT(numOfLits);
  if (numOfLits == 1UL)
    {
      selectedLits.push(lits.pop());
      allMaximalSelected = true;
      return;
    };
  allMaximalSelected = false;
  _select(lits,selectedLits,nonselectedLits,allMaximalSelected,stat);
  ASSERT(lits.empty());
  ASSERT(debugTotalNumOfLits == selectedLits.numOfAllLiterals() + nonselectedLits.numOfAllLiterals());
}; // void CustomLiteralSelection::select(..)


long CustomLiteralSelection::literalSize(TmpLiteral* lit)
{
  CALL("literalSize(TmpLiteral* lit)");
  return static_cast<long>(lit->size());
};

long CustomLiteralSelection::numOfDifferentVariables(TmpLiteral* lit)
{
  CALL("numOfDifferentVariables(TmpLiteral* lit)");
  return static_cast<long>(lit->numOfDifferentVariables());
};

bool CustomLiteralSelection::isNegativeEquality(TmpLiteral* lit)
{
  CALL("isNegativeEquality(TmpLiteral* lit)");
  return lit->isNegative() && lit->isEquality();
};


bool CustomLiteralSelection::isNotPositiveEquality(TmpLiteral* lit)
{
  CALL("isNotPositiveEquality(TmpLiteral* lit)");
  return (!lit->isEquality()) || lit->isNegative();
};




bool CustomLiteralSelection::oneLiteralSelected(TmpLiteral* firstSelected,
						TmpLiteral* firstNonselected)
{
  CALL("oneLiteralSelected(..)");
  ASSERT(firstSelected);
  return !firstSelected->next();
};

bool CustomLiteralSelection::noBigNegativeNonselected(TmpLiteral* firstSelected,
						      TmpLiteral* firstNonselected)
{
  CALL("noBigNegativeNonselected(..)");
  ulong maxSizeOfSelected = firstSelected->size();
  TmpLiteral* lit = firstSelected->next();
  while (lit)
    {
      ulong litSize = lit->size();
      if (litSize > maxSizeOfSelected) maxSizeOfSelected = litSize;
      lit = lit->next();
    };

  lit = firstNonselected;
  
  while (lit)
    {
      if (lit->isNegativeForSelection() &&
	  (lit->size() > maxSizeOfSelected))
	return false;
      lit = lit->next();
    };

  return true;
}; // bool CustomLiteralSelection::noBigNegativeNonselected(..)

bool CustomLiteralSelection::noNegativeEqualitiesNonselected(TmpLiteral* firstSelected,
							     TmpLiteral* firstNonselected)
{
  CALL("noNegativeEqualitiesNonselected(..)");
  while (firstNonselected)
    {
      if (firstNonselected->isEquality() &&
	  firstNonselected->isNegativeForSelection())
	return false;
      firstNonselected = firstNonselected->next();
    };
  return true;
}; // bool CustomLiteralSelection::noNegativeEqualitiesNonselected(..)

bool CustomLiteralSelection::noNegativeWithManyDiffVarsNonselected(TmpLiteral* firstSelected,
								   TmpLiteral* firstNonselected)
{
  CALL("noNegativeWithManyDiffVarsNonselected(..)");
  ulong maxNumOfDiffVars = firstSelected->numOfDifferentVariables();
  TmpLiteral* lit = firstSelected->next();
  while (lit)
    {
      ulong numOfDiffVars = lit->numOfDifferentVariables();
      if (numOfDiffVars > maxNumOfDiffVars) maxNumOfDiffVars = numOfDiffVars;
      lit = lit->next();
    };

  lit = firstNonselected; 
  
  while (lit)
    {
      if (lit->isNegativeForSelection() &&
	  (lit->numOfDifferentVariables() > maxNumOfDiffVars))
	return false;
      lit = lit->next();
    };

  return true;
}; // bool CustomLiteralSelection::noNegativeWithManyDiffVarsNonselected(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ComplexLiteralCharacteristic"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::ComplexLiteralCharacteristic::~ComplexLiteralCharacteristic()
{  
}; 


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::InverseVarDepthProfile"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::InverseVarDepthProfile::~InverseVarDepthProfile()
{  
}; 

void CustomLiteralSelection::InverseVarDepthProfile::computeFor(TmpLiteral* lit)
{
  CALL("computeFor(TmpLiteral* lit)");
  lit->collectVarDepthProfile(_data);
}; // void CustomLiteralSelection::InverseVarDepthProfile::computeFor(TmpLiteral* lit)

Comparison CustomLiteralSelection::InverseVarDepthProfile::compareWith(ComplexLiteralCharacteristic* c)
{
  CALL("compareWith(ComplexLiteralCharacteristic* c)");
  return 
    _data.compareByInverselyInducedOrder((static_cast<InverseVarDepthProfile*>(c))->_data);
}; // Comparison CustomLiteralSelection::InverseVarDepthProfile::compareWith(ComplexLiteralCharacteristic* c)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::Filter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::Filter::~Filter()
{  
}; 

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::MaxValFilter::~MaxValFilter()
{
};


void 
CustomLiteralSelection::MaxValFilter::selectOne(TmpLitList& lits,
						       TmpLitList& selectedLits,
						       TmpLitList& nonselectedLits)
{
  CALL("selectOne(..)");  
  TmpLiteral* cand = lits.pop();
  long candVal = _val(cand);
  TmpLiteral* nextLit;
  long nextVal;
  while (lits.nonempty())
    {
      nextLit = lits.pop();
      nextVal = _val(nextLit);
      if (nextVal > candVal)
	{
	  nonselectedLits.push(cand);
	  cand = nextLit;
	  candVal = nextVal;
	}
      else
	nonselectedLits.push(nextLit);
    };
  selectedLits.push(cand);
}; // void CustomLiteralSelection::MaxValFilter::selectOne(..)

void 
CustomLiteralSelection::MaxValFilter::selectAll(TmpLitList& lits,
						       TmpLitList& selectedLits,
						       TmpLitList& nonselectedLits)
{
  CALL("selectAll(..)");
  long candVal = _val(lits.last());
  long nextVal;
  for (TmpLiteral* lit = lits.last()->previous();
       lit;
       lit = lit->previous())
    {
      nextVal = _val(lit);
      if (nextVal != candVal)	
	if (nextVal > candVal)
	  {
	    candVal = nextVal;
	    do 
	      {
		nonselectedLits.push(lits.pop());
	      }
	    while (lits.last() != lit);
	  }
	else // nextVal < candVal
	  {
	    TmpLiteral* tmp = lit->next();
	    ASSERT(tmp);
	    lits.remove(lit);
	    nonselectedLits.push(lit);
	    lit = tmp;
	  };
    };

  // all remaining literals have maximal _val
  ASSERT(lits.nonempty());
  selectedLits.collectFrom(lits);
}; // void CustomLiteralSelection::MaxValFilter::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxComplexValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::MaxComplexValFilter::~MaxComplexValFilter()
{
};


void 
CustomLiteralSelection::MaxComplexValFilter::selectOne(TmpLitList& lits,
						       TmpLitList& selectedLits,
						       TmpLitList& nonselectedLits)
{
  CALL("selectOne(..)");  
  ASSERT(lits.nonempty());
  ComplexLiteralCharacteristic* candVal = _tmpVal1;
  ComplexLiteralCharacteristic* bufferVal = _tmpVal2;
 next_candidate:
  ASSERT(lits.nonempty());
  candVal->computeFor(lits.last());
  
  for (TmpLiteral* lit = lits.last()->previous(); 
       lit; 
       lit = lit->previous())
    {
      bufferVal->computeFor(lit);
      switch (candVal->compareWith(bufferVal))
	{
	case Less:
	  nonselectedLits.push(lits.pop());
	  goto next_candidate;

	case Equal:
	  // as below
	case Greater:
	  {
	    TmpLiteral* tmp = lit->next();
	    ASSERT(tmp);
	    lits.remove(lit);
	    nonselectedLits.push(lit);
	    lit = tmp;
	  };
	  break;
	  
	case Incomparable:
	  break;
	}; // switch (candVal->compareWith(bufferVal))
    }; // for 
  // lits.last() is OK
  selectedLits.push(lits.pop());
  while (lits.nonempty()) // only one has to be selected
    nonselectedLits.push(lits.pop());
}; // void CustomLiteralSelection::MaxComplexValFilter::selectOne(..)

void 
CustomLiteralSelection::MaxComplexValFilter::selectAll(TmpLitList& lits,
						       TmpLitList& selectedLits,
						       TmpLitList& nonselectedLits)
{
  CALL("selectAll(..)");
  ASSERT(lits.nonempty());
  ComplexLiteralCharacteristic* candVal = _tmpVal1;
  ComplexLiteralCharacteristic* bufferVal = _tmpVal2;
 next_candidate:
  ASSERT(lits.nonempty());
  if (lits.numOfAllLiterals() == 1UL)
    {
      selectedLits.push(lits.pop());
      return;
    };
  candVal->computeFor(lits.last());
  
  for (TmpLiteral* lit = lits.last()->previous(); 
       lit; 
       lit = lit->previous())
    {
      bufferVal->computeFor(lit);
      switch (candVal->compareWith(bufferVal))
	{
	case Less:
	  nonselectedLits.push(lits.pop());
	  goto next_candidate;

	case Equal:
	  break;

	case Greater:
	  {
	    TmpLiteral* tmp = lit->next();
	    ASSERT(tmp);
	    lits.remove(lit);
	    nonselectedLits.push(lit);
	    lit = tmp;
	  };
	  break;
	  
	case Incomparable:
	  break;
	}; // switch (candVal->compareWith(bufferVal))
    }; // for ..
  // lits.last() is OK
  selectedLits.push(lits.pop());
  if (lits.empty()) return;
  goto next_candidate;
}; // void CustomLiteralSelection::MaxComplexValFilter::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::AllPOrAllFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::AllPOrAllFilter::~AllPOrAllFilter()
{
};


void 
CustomLiteralSelection::AllPOrAllFilter::selectOne(TmpLitList& lits,
						       TmpLitList& selectedLits,
						       TmpLitList& nonselectedLits)
{
  CALL("selectOne(..)"); 
  TmpLiteral* lit = lits.last();
  while (lit && (!(_prop(lit))))
    {
      lit = lit->previous();      
    };
  
  if (lit)
    {
      lits.remove(lit);
      selectedLits.push(lit);
    }
  else // select any one
    {
      selectedLits.push(lits.pop());
    };
  nonselectedLits.collectFrom(lits);
}; // void CustomLiteralSelection::AllPOrAllFilter::selectOne(..)

void 
CustomLiteralSelection::AllPOrAllFilter::selectAll(TmpLitList& lits,
						       TmpLitList& selectedLits,
						       TmpLitList& nonselectedLits)
{
  CALL("selectAll(..)");
   
  TmpLiteral* lit = lits.last();
  while (lit && (!(_prop(lit))))
    {
      lit = lit->previous();      
    };
  
  if (lit) 
    {
      // select all with the property
      while (lits.last() != lit)
	nonselectedLits.push(lits.pop());      
      selectedLits.push(lits.pop());
      while (lits.nonempty())
	{
	  lit = lits.pop();
	  if (_prop(lit)) 
	    {
	      selectedLits.push(lit);
	    }
	  else
	    nonselectedLits.push(lit);
	};
    }
  else // select all
    {
      selectedLits.collectFrom(lits);
    };
}; // void CustomLiteralSelection::AllPOrAllFilter::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::CompositionFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::CompositionFilter::~CompositionFilter()
{
};


void 
CustomLiteralSelection::CompositionFilter::selectOne(TmpLitList& lits,
						       TmpLitList& selectedLits,
						       TmpLitList& nonselectedLits)
{
  CALL("selectOne(..)");  
  if (_filters.empty()) 
    {
      selectedLits.push(lits.pop());
      nonselectedLits.collectFrom(lits);
      return;
    };

  TmpLiteral* lastSelected = selectedLits.last();
  
  Filter** lastFilter = _filters.last();
  Filter** filt = _filters.begin();
  while (filt != lastFilter)
    {
      (*filt)->select(lits,selectedLits,nonselectedLits);
      while (selectedLits.last() != lastSelected)
	lits.push(selectedLits.pop());
      ++filt;
    };
  
  // apply the last filter
  (*lastFilter)->selectOne(lits,selectedLits,nonselectedLits);
}; // void CustomLiteralSelection::CompositionFilter::selectOne(..)

void 
CustomLiteralSelection::CompositionFilter::selectAll(TmpLitList& lits,
						     TmpLitList& selectedLits,
						     TmpLitList& nonselectedLits)
{
  CALL("selectAll(..)");
  if (_filters.empty()) { selectedLits.collectFrom(lits); return; };
  
  TmpLiteral* lastSelected = selectedLits.last();
  
  Filter** lastFilter = _filters.last();
  Filter** filt = _filters.begin();
  while (filt != lastFilter)
    {
      (*filt)->select(lits,selectedLits,nonselectedLits);
      while (selectedLits.last() != lastSelected)
	lits.push(selectedLits.pop());
      ++filt;
    };
  
  // apply the last filter
  (*lastFilter)->select(lits,selectedLits,nonselectedLits);
}; // void CustomLiteralSelection::CompositionFilter::selectAll(..)

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::OneOfEachPolarityFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::OneOfEachPolarityFilter::~OneOfEachPolarityFilter()
{
};


void 
CustomLiteralSelection::OneOfEachPolarityFilter::selectOne(TmpLitList& lits,
						       TmpLitList& selectedLits,
						       TmpLitList& nonselectedLits)
{
  CALL("selectOne(..)");
  selectedLits.push(lits.pop());
  nonselectedLits.collectFrom(lits);
}; // void CustomLiteralSelection::OneOfEachPolarityFilter::selectOne(..)

void 
CustomLiteralSelection::OneOfEachPolarityFilter::selectAll(TmpLitList& lits,
						     TmpLitList& selectedLits,
						     TmpLitList& nonselectedLits)
{
  CALL("selectAll(..)");
  bool posSelected = false;
  bool negSelected = false;
  while (lits.nonempty())
    {
      TmpLiteral* lit = lits.pop();
      if (lit->isPositiveForSelection())
	{
	  if (posSelected)
	    {
	      nonselectedLits.push(lit);
	    }
	  else
	    {
	      posSelected = true;
	      selectedLits.push(lit);
	    };
	}
      else
	{
	  if (negSelected)
	    {
	      nonselectedLits.push(lit);
	    }
	  else
	    {
	      negSelected = true;
	      selectedLits.push(lit);
	    };
	};
    };
}; // void CustomLiteralSelection::OneOfEachPolarityFilter::selectAll(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::SelectionAcceptability"
#endif
#include "debugMacros.hpp"
//============================================================================

CustomLiteralSelection::SelectionAcceptability::~SelectionAcceptability()
{  
}; 


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability"
#endif
#include "debugMacros.hpp"
//============================================================================


CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability::~ConjOfPrimitiveSelectionAcceptability()
{
}; 


bool CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability::accepts(TmpLiteral* firstSelected,
									    TmpLiteral* firstNonselected)
{
  CALL("accepts(..)");

  Primitive* endOfPrim = _primitives.end();
  for (Primitive* prim = _primitives.begin(); 
       prim != endOfPrim;
       ++prim)
    {
      if (!((*prim)(firstSelected,firstNonselected))) return false;
    };
  return true;
}; //  bool CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability::accepts()




//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "TotalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================


void 
TotalLiteralSelection::_select(TmpLitList& lits,
			       TmpLitList& selectedLits,
			       TmpLitList& nonselectedLits,
			       bool& allMaximalSelected,
			       const TmpLitList::Statistics& stat)
{
  CALL("_select(..)");
  selectedLits.collectFrom(lits);
  allMaximalSelected = true;
}; // void TotalLiteralSelection::_select(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaximalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

void 
MaximalLiteralSelection::_select(TmpLitList& lits,
			       TmpLitList& selectedLits,
			       TmpLitList& nonselectedLits,
			       bool& allMaximalSelected,
			       const TmpLitList::Statistics& stat)
{
  CALL("_select(..)");
  allMaximalSelected = true;
  selectMaximal(lits,selectedLits,nonselectedLits);
}; // void MaximalLiteralSelection::_select(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "HyperLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
HyperLiteralSelection::_select(TmpLitList& lits,
			       TmpLitList& selectedLits,
			       TmpLitList& nonselectedLits,
			       bool& allMaximalSelected,
			       const TmpLitList::Statistics& stat)
{
  CALL("_select(..)");
  if (stat.negative_nonanswer_lits)
    {
      allMaximalSelected = false;
      if (stat.positive_nonanswer_lits())
	{
	  // mixed polarity, unselect all positive
	  unselectPositive(lits,nonselectedLits,stat.positive_nonanswer_lits());
	};
      ASSERT(lits.nonempty());
      if (lits.numOfAllLiterals() == 1UL)
	{
	  selectedLits.push(lits.pop());
	}
      else
	_filterForNegative->select(lits,selectedLits,nonselectedLits);      
    }
  else // only positive literals
    {
      allMaximalSelected = true;
      selectMaximal(lits,selectedLits,nonselectedLits);
    };
}; // void HyperLiteralSelection::_select(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaxWithCorrectionLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
MaxWithCorrectionLiteralSelection::_select(TmpLitList& lits,
					   TmpLitList& selectedLits,
					   TmpLitList& nonselectedLits,
					   bool& allMaximalSelected,
					   const TmpLitList::Statistics& stat)
{
  CALL("_select(..)");  
  if (stat.negative_nonanswer_lits)
    { 
      TmpLiteral* lastSelected = selectedLits.last();
      TmpLiteral* lastNonselected = nonselectedLits.last();
      selectMaximal(lits,selectedLits,nonselectedLits);
      if (!_maxSelectionAcceptability->accepts(((lastSelected) ? lastSelected->next() : selectedLits.first()),
					       ((lastSelected) ? lastNonselected->next() : nonselectedLits.first())))
	{      
	  allMaximalSelected = false;
	  // undo selection for negative literals
	  TmpLiteral* lit = nonselectedLits.last();
	  while (lit != lastNonselected)
	    {
	      if (lit->isNegativeForSelection())
		{
		  TmpLiteral* tmp = lit->previous();
		  nonselectedLits.remove(lit);
		  lits.push(lit);
		  lit = tmp;
		}
	      else
		lit = lit->previous();
	    };
	  while (selectedLits.last() != lastSelected)
	    {
	      lit = selectedLits.pop();
	      if (lit->isNegativeForSelection())
		{
		  lits.push(lit);
		}
	      else
		nonselectedLits.push(lit);
	    };

	  ASSERT(lits.nonempty());
	  
	  // select appropriate negative 	
	  
	  _filterForNegative->select(lits,selectedLits,nonselectedLits);
	}
      else // maximal selection accepted
	{
	  allMaximalSelected = true;
	};
     
    }
  else // only positive literals
    {
      allMaximalSelected = true;
      selectMaximal(lits,selectedLits,nonselectedLits);
    };
}; // void MaxWithCorrectionLiteralSelection::_select(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "BestOnlyLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
BestOnlyLiteralSelection::_select(TmpLitList& lits,
				  TmpLitList& selectedLits,
				  TmpLitList& nonselectedLits,
				  bool& allMaximalSelected,
				  const TmpLitList::Statistics& stat)
{
  CALL("_select(..)");
  TmpLiteral* cand = 0;
  if (stat.negative_nonanswer_lits)
    {
      if (stat.positive_nonanswer_lits())
	{
	  TmpLiteral* lastSelected = selectedLits.last();
	  TmpLiteral* lastNonselected = nonselectedLits.last();
	  selectMaximal(lits,selectedLits,nonselectedLits);

	  // now try to beat the maximal selection
	  
	  // light-weight test first: is a negative literal selected?
	  TmpLiteral* lit;
	  for (lit = (lastSelected) ? lastSelected->next() : selectedLits.first(); 
	       lit; 
	       lit = lit->next())
	    if (lit->isNegativeForSelection()) goto undo_maximal_selection;

	  // all selected are positive
	  
	  for (cand = (lastNonselected) ? lastNonselected->next() : nonselectedLits.first();
	       cand;
	       cand = cand->next())	    
	    {
	      if (cand->isNegativeForSelection())		
		for (lit = (lastSelected) ? lastSelected->next() : selectedLits.first();
		     lit; 
		     lit = lit->next())
		  {
		    Comparison cmp = _litOrdering->compare(lit,cand);
		    if ((cmp == BK::Less) || (cmp == BK::Equal))
		      goto undo_maximal_selection;
		  };
	    };

	  // maximal selection is OK
	  allMaximalSelected = true;
	  return;

	undo_maximal_selection:
	  lit = (lastNonselected) ? lastNonselected->next() : nonselectedLits.first();
	  while (lit)
	    {
	      if (lit->isNegativeForSelection())
		{
		  TmpLiteral* tmp = lit;
		  lit = lit->next();
		  nonselectedLits.remove(tmp);
		  lits.push(tmp);
		}
	      else
		lit = lit->next();
	    };

	  while (selectedLits.last() != lastSelected)
	    {
	      lit = selectedLits.pop();
	      if (lit->isNegativeForSelection())
		{
		  lits.push(lit);
		}
	      else
		nonselectedLits.push(lit);
	    };

	}; // if (stat.positive_nonanswer_lits())
      // only negative literals in lits if we get here
      allMaximalSelected = false;
      selectOneMaximal(_litOrdering,lits,selectedLits,nonselectedLits);
    }
  else // positive fact
    {
      allMaximalSelected = true;
      selectMaximal(lits,selectedLits,nonselectedLits);
    };
  
}; // void BestOnlyLiteralSelection::_select(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "ArbitraryLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
ArbitraryLiteralSelection::_select(TmpLitList& lits,
			       TmpLitList& selectedLits,
			       TmpLitList& nonselectedLits,
			       bool& allMaximalSelected,
			       const TmpLitList::Statistics& stat)
{
  CALL("_select(..)");
  allMaximalSelected = false;
  _filter->select(lits,selectedLits,nonselectedLits);
}; // void ArbitraryLiteralSelection::_select(..)


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
CompleteBidirectionalLiteralSelection::_select(TmpLitList& lits,
			       TmpLitList& selectedLits,
			       TmpLitList& nonselectedLits,
			       bool& allMaximalSelected,
			       const TmpLitList::Statistics& stat)
{
  CALL("_select(..)");

  if (stat.negative_nonanswer_lits)
    {
      allMaximalSelected = false;
      if (stat.positive_nonanswer_lits())
	{
	  // rule (mixed polarity lit. list)
	  if (lits.numOfAllLiterals() == 2UL)
	    {
	      selectedLits.collectFrom(lits);
	      return;
	    };

	  // collect all positive literals in the buffer
	  ASSERT(_buffer.empty());
	  ASSERT(lits.numOfAllLiterals() > 2UL);
	  TmpLiteral* lit = lits.first();
	check_lit:
	  
	  if (lit->isPositiveForSelection())
	    {
	      TmpLiteral* tmp = lit->next();
	      lits.remove(lit);
	      _buffer.push(lit);
	      if (tmp)
		{
		  lit = tmp;
		  goto check_lit;
		};
	    }
	  else
	    {
	      lit = lit->next();
	      if (lit) goto check_lit;
	    };
	  
	  // all positive ones are in _buffer now
	  _filterForNegativeInRules->select(lits,selectedLits,nonselectedLits);
	  _filterForPositiveInRules->select(_buffer,selectedLits,nonselectedLits);
	}
      else // negative fact
	{
	  _filterForNegativeFacts->select(lits,selectedLits,nonselectedLits);
	};
    }
  else // positive fact
    {
      allMaximalSelected = true;
      selectMaximal(lits,selectedLits,nonselectedLits);
    };  
}; // void CompleteBidirectionalLiteralSelection::_select(..)



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "IncompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================



void 
IncompleteBidirectionalLiteralSelection::_select(TmpLitList& lits,
			       TmpLitList& selectedLits,
			       TmpLitList& nonselectedLits,
			       bool& allMaximalSelected,
			       const TmpLitList::Statistics& stat)
{
  CALL("_select(..)");
  allMaximalSelected = false;

  if (stat.negative_nonanswer_lits)
    {
      if (stat.positive_nonanswer_lits())
	{
	  // rule (mixed polarity lit. list)
	  if (lits.numOfAllLiterals() == 2UL)
	    {
	      selectedLits.collectFrom(lits);
	      return;
	    };

	  // collect all positive literals in the buffer
	  ASSERT(_buffer.empty());
	  ASSERT(lits.numOfAllLiterals() > 2UL);
	  TmpLiteral* lit = lits.first();
	check_lit:
	  
	  if (lit->isPositiveForSelection())
	    {
	      TmpLiteral* tmp = lit->next();
	      lits.remove(lit);
	      _buffer.push(lit);
	      if (tmp)
		{
		  lit = tmp;
		  goto check_lit;
		};
	    }
	  else
	    {
	      lit = lit->next();
	      if (lit) goto check_lit;
	    };
	  
	  // all positive ones are in _buffer now
	  _filterForNegativeInRules->select(lits,selectedLits,nonselectedLits);
	  _filterForPositiveInRules->select(_buffer,selectedLits,nonselectedLits);
	}
      else // negative fact
	{
	  _filterForNegativeFacts->select(lits,selectedLits,nonselectedLits);
	};
    }
  else // positive fact
    {
      _filterForPositiveFacts->select(lits,selectedLits,nonselectedLits);
    };  
}; // void IncompleteBidirectionalLiteralSelection::_select(..)



//============================================================================


