/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         CustomLiteralSelection.hpp
// Description:  Flexible schemes for literal selection.
// Created:      Jun 03, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//============================================================================
#ifndef CUSTOM_LITERAL_SELECTION_H
//============================================================================
#define CUSTOM_LITERAL_SELECTION_H 
#include "jargon.hpp"
#include "RuntimeError.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TmpLitList.hpp"
#include "StandardLiteralOrdering.hpp"
#include "ExpandingStack.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection
  {
  public:
    typedef long (*NumericLiteralCharacteristic)(TmpLiteral*);
    typedef bool (*LiteralProperty)(TmpLiteral*);


    class ComplexLiteralCharacteristic;
    class InverseVarDepthProfile;
    // More shallow occurences of variables
    // lead to bigger instances of InverseVarDepthProfile 
    // in the corresponding ordering.
    
    


    class Filter;
    class MaxValFilter;
    class MaxComplexValFilter;
    class AllPOrAllFilter;
    class CompositionFilter;
    class OneOfEachPolarityFilter;
    
    class SelectionAcceptability;
    class ConjOfPrimitiveSelectionAcceptability;

  public:
    CustomLiteralSelection() {};
    virtual ~CustomLiteralSelection();
    void setStandardLiteralOrdering(StandardLiteralOrdering* ord) { _standardLiteralOrdering = ord; };
    void select(TmpLitList& lits,
		TmpLitList& selectedLits,
		TmpLitList& nonselectedLits,
		bool& allMaximalSelected,
		const TmpLitList::Statistics& stat); 
   
    static long literalSize(TmpLiteral* lit);


    static long numOfDifferentVariables(TmpLiteral* lit);

    
    static bool isNegativeEquality(TmpLiteral* lit);

    
    static bool isNotPositiveEquality(TmpLiteral* lit);


    static bool oneLiteralSelected(TmpLiteral* firstSelected,
				   TmpLiteral* firstNonselected); 

    static bool noBigNegativeNonselected(TmpLiteral* firstSelected,
					 TmpLiteral* firstNonselected); 

    static bool noNegativeEqualitiesNonselected(TmpLiteral* firstSelected,
						TmpLiteral* firstNonselected); 

    static bool noNegativeWithManyDiffVarsNonselected(TmpLiteral* firstSelected,
						      TmpLiteral* firstNonselected);


  protected:
    void selectMaximal(TmpLitList& lits,
		       TmpLitList& selectedLits,
		       TmpLitList& nonselectedLits);

    void selectOneMaximal(LiteralOrdering* litOrd,
		       TmpLitList& lits,
		       TmpLitList& selectedLits,
		       TmpLitList& nonselectedLits);


    void unselectPositive(TmpLitList& lits,TmpLitList& nonselectedLits,ulong numOfPositive);

  protected:
    StandardLiteralOrdering* _standardLiteralOrdering;

  private:
    virtual
    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat) = 0;
    
  }; // class CustomLiteralSelection 
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ComplexLiteralCharacteristic"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{
  class CustomLiteralSelection::ComplexLiteralCharacteristic
  {
  public:
    ComplexLiteralCharacteristic() {};
    virtual ~ComplexLiteralCharacteristic();
    virtual void computeFor(TmpLiteral* lit) = 0;
    virtual BK::Comparison compareWith(ComplexLiteralCharacteristic* c) = 0;
  }; // class CustomLiteralSelection::ComplexLiteralCharacteristic
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::InverseVarDepthProfile"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{
  class CustomLiteralSelection::InverseVarDepthProfile : public ComplexLiteralCharacteristic
  {
  public:
    InverseVarDepthProfile() {};
    ~InverseVarDepthProfile();
    void computeFor(TmpLiteral* lit);
    BK::Comparison compareWith(ComplexLiteralCharacteristic* c);
  private:
    TmpLiteral::VarDepthProfile _data;
  }; // class CustomLiteralSelection::ComplexLiteralCharacteristic
}; // namespace VK





//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::Filter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::Filter
  {
  public:
    Filter() : _alwaysSelectOne(false) {};
    virtual ~Filter();
    void setSelectOne(bool fl) { _alwaysSelectOne = fl; };
    void select(TmpLitList& lits,
		TmpLitList& selectedLits,
		TmpLitList& nonselectedLits)
    {
      CALL("select(..)");
      ASSERT(lits.nonempty());
      DOP(ulong debugNumOfLits = lits.numOfAllLiterals());
      DOP(ulong debugNumOfSelectedLits = selectedLits.numOfAllLiterals());
      DOP(ulong debugNumOfNonselected = nonselectedLits.numOfAllLiterals());

      if (_alwaysSelectOne)
	{
	  selectOne(lits,selectedLits,nonselectedLits);
	  ASSERT(selectedLits.numOfAllLiterals() - debugNumOfSelectedLits == 1UL);
	}
      else
	{
	  selectAll(lits,selectedLits,nonselectedLits);
	  ASSERT(selectedLits.numOfAllLiterals() > debugNumOfSelectedLits);
	};
      ASSERT(lits.empty());
      ASSERT(selectedLits.numOfAllLiterals() + 
	     nonselectedLits.numOfAllLiterals() 
	     ==
	     debugNumOfLits + 
	     debugNumOfSelectedLits + 
	     debugNumOfNonselected);
    };
  private:
    virtual void selectOne(TmpLitList& lits,
			   TmpLitList& selectedLits,
			   TmpLitList& nonselectedLits) = 0;
    virtual void selectAll(TmpLitList& lits,
			   TmpLitList& selectedLits,
			   TmpLitList& nonselectedLits) = 0;
  private:
    bool _alwaysSelectOne;
    friend class CustomLiteralSelection::CompositionFilter;
  }; // class CustomLiteralSelection::Filter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::MaxValFilter : 
    public Filter
  {
  public:
    MaxValFilter(NumericLiteralCharacteristic val) : _val(val)
    {
    };
    ~MaxValFilter();
  private:    
    void selectOne(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
    void selectAll(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
  private:
    NumericLiteralCharacteristic _val;
  }; // class CustomLiteralSelection::MaxValFilter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::MaxComplexValFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::MaxComplexValFilter : 
    public Filter
  {
  public:
    MaxComplexValFilter(ComplexLiteralCharacteristic* tmpVal1,
			ComplexLiteralCharacteristic* tmpVal2) : 
      _tmpVal1(tmpVal1),
      _tmpVal2(tmpVal2)
    {
    };
    ~MaxComplexValFilter();
  private:    
    void selectOne(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
    void selectAll(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
  private:
    ComplexLiteralCharacteristic* _tmpVal1;
    ComplexLiteralCharacteristic* _tmpVal2;
  }; // class CustomLiteralSelection::MaxComplexValFilter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::AllPOrAllFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::AllPOrAllFilter : 
    public Filter
  {
  public:
    AllPOrAllFilter(LiteralProperty prop) : _prop(prop)
    {
    };
    ~AllPOrAllFilter();
  private:    
    void selectOne(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
    void selectAll(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
  private:
    LiteralProperty _prop;
  }; // class CustomLiteralSelection::AllPOrAllFilter
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::CompositionFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::CompositionFilter : 
    public Filter
  {
  public:
    CompositionFilter() 
    {
    };
    ~CompositionFilter();
    void resetFilters() { _filters.reset(); };
    void addFilter(Filter* f) { _filters.pushSafe(f); };
  private:    
    void selectOne(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
    void selectAll(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
  private:
    BK::ExpandingStack<BK::GlobAlloc,Filter*,10UL,CompositionFilter> _filters;
  }; // class CustomLiteralSelection::CompositionFilter
}; // namespace VK



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::OneOfEachPolarityFilter"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::OneOfEachPolarityFilter : 
    public Filter
  {
  public:
    OneOfEachPolarityFilter() 
    {
    };
    ~OneOfEachPolarityFilter();
  private:    
    void selectOne(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
    void selectAll(TmpLitList& lits,
		   TmpLitList& selectedLits,
		   TmpLitList& nonselectedLits);
  }; // class CustomLiteralSelection::OneOfEachPolarityFilter
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::SelectionAcceptability"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::SelectionAcceptability
  {
  public:
    SelectionAcceptability() {};
    virtual ~SelectionAcceptability();
    virtual bool accepts(TmpLiteral* firstSelected,
			 TmpLiteral* firstNonselected) = 0;
  }; // class CustomLiteralSelection::SelectionAcceptability
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability :
    public SelectionAcceptability
  {
  public:
    ConjOfPrimitiveSelectionAcceptability() 
    {
    };
    ~ConjOfPrimitiveSelectionAcceptability();
    bool accepts(TmpLiteral* firstSelected,
		 TmpLiteral* firstNonselected);
    void resetPrimitives() { _primitives.reset(); };
    void addPrimitive(bool (*fun)(TmpLiteral*,TmpLiteral*)) 
    {
      _primitives.pushSafe(fun);
    };
  private:
    typedef bool (*Primitive)(TmpLiteral*,TmpLiteral*);
  private:
    BK::ExpandingStack<BK::GlobAlloc,Primitive,10UL,ConjOfPrimitiveSelectionAcceptability> _primitives;
  }; // class CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "TotalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class TotalLiteralSelection : public CustomLiteralSelection
  {
  public:
    TotalLiteralSelection() {};
    ~TotalLiteralSelection() {};
  private:    
    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat);    
  }; // class TotalLiteralSelection
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaximalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class MaximalLiteralSelection : public CustomLiteralSelection
  {
  public:
    MaximalLiteralSelection() {};
    ~MaximalLiteralSelection() {};
  private:    

    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat);    
  }; // class MaximalLiteralSelection
}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "HyperLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class HyperLiteralSelection : public CustomLiteralSelection
  {
  public:
    HyperLiteralSelection() {};
    ~HyperLiteralSelection() {};
    void setFilterForNegative(CustomLiteralSelection::Filter* f)
    {
      _filterForNegative = f;
    };
  private:  
    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filterForNegative;    
  }; // class HyperLiteralSelection
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "MaxWithCorrectionLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class MaxWithCorrectionLiteralSelection : public CustomLiteralSelection
  {
  public:
    MaxWithCorrectionLiteralSelection() 
    {
      CALL("constructor MaxWithCorrectionLiteralSelection()");
      DOP(_filterForNegative = 0);
      DOP(_maxSelectionAcceptability = 0);
    };
    ~MaxWithCorrectionLiteralSelection() {};
    void setFilterForNegative(CustomLiteralSelection::Filter* f)
    {
      _filterForNegative = f;
    };
    void setMaxSelectionAcceptability(CustomLiteralSelection::SelectionAcceptability* test)
    {
      _maxSelectionAcceptability = test;
    };
  private:  
    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat);    
  private:
    CustomLiteralSelection::SelectionAcceptability* _maxSelectionAcceptability;
    CustomLiteralSelection::Filter* _filterForNegative;
  }; // class MaxWithCorrectionLiteralSelection
}; // namespace VK



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "BestOnlyLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class BestOnlyLiteralSelection : public CustomLiteralSelection
  {
  public:
    BestOnlyLiteralSelection() {};
    ~BestOnlyLiteralSelection() {};
    void setLiteralQualityOrdering(LiteralOrdering* o)
    {
      _litOrdering = o;
    };
  private:  
    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat); 
  private:
    LiteralOrdering* _litOrdering;
  }; // class BestOnlyLiteralSelection
}; // namespace VK



//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "ArbitraryLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class ArbitraryLiteralSelection : public CustomLiteralSelection
  {
  public:
    ArbitraryLiteralSelection() {};
    ~ArbitraryLiteralSelection() {};
    void setFilter(CustomLiteralSelection::Filter* f)
    {
      _filter = f;
    };
  private:  
    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filter;    
  }; // class ArbitraryLiteralSelection
}; // namespace VK


//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class CompleteBidirectionalLiteralSelection : public CustomLiteralSelection
  {
  public:
    CompleteBidirectionalLiteralSelection() {};
    ~CompleteBidirectionalLiteralSelection() {};
    void setFilterForPositiveInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForPositiveInRules = f;
    };
    void setFilterForNegativeInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeInRules = f;
    };
    void setFilterForNegativeFacts(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeFacts = f;
    };

  private:  
    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filterForPositiveInRules; 
    CustomLiteralSelection::Filter* _filterForNegativeInRules;    
    CustomLiteralSelection::Filter* _filterForNegativeFacts;    
    TmpLitList _buffer;
  }; // class CompleteBidirectionalLiteralSelection
}; // namespace VK





//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "IncompleteBidirectionalLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
  class IncompleteBidirectionalLiteralSelection : public CustomLiteralSelection
  {
  public:
    IncompleteBidirectionalLiteralSelection() {};
    ~IncompleteBidirectionalLiteralSelection() {};
    void setFilterForPositiveInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForPositiveInRules = f;
    };
    void setFilterForNegativeInRules(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeInRules = f;
    };
    void setFilterForPositiveFacts(CustomLiteralSelection::Filter* f)
    {
      _filterForPositiveFacts = f;
    };
    void setFilterForNegativeFacts(CustomLiteralSelection::Filter* f)
    {
      _filterForNegativeFacts = f;
    };

  private:  
    void 
    _select(TmpLitList& lits,
	    TmpLitList& selectedLits,
	    TmpLitList& nonselectedLits,
	    bool& allMaximalSelected,
	    const TmpLitList::Statistics& stat);    
  private:
    CustomLiteralSelection::Filter* _filterForPositiveInRules; 
    CustomLiteralSelection::Filter* _filterForNegativeInRules;    
    CustomLiteralSelection::Filter* _filterForPositiveFacts;      
    CustomLiteralSelection::Filter* _filterForNegativeFacts;    
    TmpLitList _buffer;
  }; // class IncompleteBidirectionalLiteralSelection
}; // namespace VK






//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CUSTOM_LITERAL_SELECTION
#  define DEBUG_NAMESPACE "CustomLiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{
  inline
  void 
  CustomLiteralSelection::selectMaximal(TmpLitList& lits,
					TmpLitList& selectedLits,
					TmpLitList& nonselectedLits)
  {
    CALL("selectMaximal(..)");

    TmpLiteral* el;
    if (_standardLiteralOrdering->isStratified())
      {
	// filter out literals from nonmaximal layers
	TmpLiteral* maxLayerRepresentative = lits.last();  
      
	for (el = lits.last()->previous();
	     el;
	     el = el->previous())
	  {
	    switch (_standardLiteralOrdering->compareLayers(el,maxLayerRepresentative))
	      {
	      case BK::Greater:
		maxLayerRepresentative = el;	      
		do 
		  { 
		    nonselectedLits.push(lits.pop()); 
		  } 
		while (lits.last() != el);
		break;

	      case BK::Less:
		{
		  TmpLiteral* tmp = el->next();
		  ASSERT(tmp);
		  lits.remove(el);
		  nonselectedLits.push(el);
		  el = tmp;
		};
		break;

	      case BK::Equal:
		break;
	      
#ifdef DEBUG_NAMESPACE
	      default: ICP("ICP0"); break;
#else
#  ifdef _SUPPRESS_WARNINGS_
	      default: 
		BK::RuntimeError::report("Wrong BK::Comparison value in CustomLiteralSelection::selectMaximal(..)");
		break;
#  endif
#endif	     
	      }; // switch ..
	  }; //  for ..    
      };

    // all remaining literals belong to the maximal layer
  
    TmpLiteral* cand; 

  next_candidate:
    cand = lits.pop();
    if (lits.empty()) { selectedLits.push(cand); return; };
    el = lits.last();
    while (el)
      {      
	switch (_standardLiteralOrdering->compare(el,cand))
	  {
	  case BK::Greater:
	    nonselectedLits.push(cand); 
	    goto next_candidate;

	  case BK::Less:
	    {
	      TmpLiteral* tmp = el->next();
	      lits.remove(el);
	      nonselectedLits.push(el);
	      if (tmp) 
		{
		  el = tmp->previous();
		}
	      else
		el = lits.last(); 
	    };
	    break;
 
	  case BK::Equal: // same as below

	  case BK::Incomparable: 
	    el = el->previous();
	    break;

#ifdef DEBUG_NAMESPACE
	  default: ICP("ICP1"); break;
#else
#  ifdef _SUPPRESS_WARNINGS_
	  default: 
	    BK::RuntimeError::report("Wrong BK::Comparison value in CustomLiteralSelection::selectMaximal(..)");
	    break;
#  endif
#endif	     
	  }; // switch ..
      }; // while (el)

    selectedLits.push(cand);
    if (lits.empty()) return;
    goto next_candidate;
  }; // void CustomLiteralSelection::selectMaximal(..)



  inline
  void 
  CustomLiteralSelection::selectOneMaximal(LiteralOrdering* litOrd,
					TmpLitList& lits,
					TmpLitList& selectedLits,
					TmpLitList& nonselectedLits)
  {
    CALL("selectOneMaximal(LiteralOrdering* litOrd,..)");

    TmpLiteral* el;
    if (litOrd->isStratified())
      {
	// filter out literals from nonmaximal layers
	TmpLiteral* maxLayerRepresentative = lits.last();  
      
	for (el = lits.last()->previous();
	     el;
	     el = el->previous())
	  {
	    switch (litOrd->compareLayers(el,maxLayerRepresentative))
	      {
	      case BK::Greater:
		maxLayerRepresentative = el;	      
		do 
		  { 
		    nonselectedLits.push(lits.pop()); 
		  } 
		while (lits.last() != el);
		break;

	      case BK::Less:
		{
		  TmpLiteral* tmp = el->next();
		  ASSERT(tmp);
		  lits.remove(el);
		  nonselectedLits.push(el);
		  el = tmp;
		};
		break;

	      case BK::Equal:
		break;
	      
#ifdef DEBUG_NAMESPACE
	      default: ICP("ICP0"); break;
#else
#  ifdef _SUPPRESS_WARNINGS_
	      default: 
		BK::RuntimeError::report("Wrong BK::Comparison value in CustomLiteralSelection::selectMaximal(..)");
		break;
#  endif
#endif	     
	      }; // switch ..
	  }; //  for ..    
      };

    // all remaining literals belong to the maximal layer
  
    TmpLiteral* cand; 

  next_candidate:
    cand = lits.pop();
    if (lits.empty()) { selectedLits.push(cand); return; };
    el = lits.last();
    while (el)
      {      
	switch (litOrd->compare(el,cand))
	  {
	  case BK::Greater:
	    nonselectedLits.push(cand); 
	    goto next_candidate;

	  case BK::Less: // same as below
	  case BK::Equal: // same as below
	  case BK::Incomparable:
	    {
	      TmpLiteral* tmp = el->next();
	      lits.remove(el);
	      nonselectedLits.push(el);
	      if (tmp) 
		{
		  el = tmp->previous();
		}
	      else
		el = lits.last(); 
	    };
	    break;
	   

#ifdef DEBUG_NAMESPACE
	  default: ICP("ICP1"); break;
#else
#  ifdef _SUPPRESS_WARNINGS_
	  default: 
	    BK::RuntimeError::report("Wrong BK::Comparison value in CustomLiteralSelection::selectMaximal(..)");
	    break;
#  endif
#endif	     
	  }; // switch ..
      }; // while (el)

    selectedLits.push(cand);
    if (lits.empty()) return;
    goto next_candidate;
  }; // void CustomLiteralSelection::selectOneMaximal(LiteralOrdering* litOrd,..)




  
  inline
  void CustomLiteralSelection::unselectPositive(TmpLitList& lits,TmpLitList& nonselectedLits,ulong numOfPositive)
  {
    CALL("unselectPositive(..)");    
    TmpLiteral* el = lits.first();
    while (numOfPositive)
      {
	ASSERT(el);
	if (el->isPositiveForSelection())
	  {
	    TmpLiteral* tmp = el->previous();
	    lits.remove(el);
	    nonselectedLits.push(el);
	    if (tmp) { el = tmp->next(); }
	    else
	      el = lits.first();
	    numOfPositive--;
	  }
	else
	  el = el->next();
      };
  }; // void CustomLiteralSelection::unselectPositive(..)


}; // namespace VK


//============================================================================
#endif
