/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         DefinedHeaderTable.hpp
// Description:  Set of defined headers for stratified resolution.
// Created:      Dec 24, 2001.
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef DEFINED_HEADER_TABLE_H
//============================================================================

#define DEFINED_HEADER_TABLE_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_DEFINED_HEADER_TABLE
 #define DEBUG_NAMESPACE "DefinedHeaderTable"
#endif
#include "debugMacros.hpp"
//============================================================================
namespace VK
{
class DefinedHeaderTable
{
 public:
  DefinedHeaderTable() :
#ifdef NO_DEBUG 
    _defined(false,0)
#else
    _defined(false,"DefinedHeaderTable::_defined")
#endif   
  {
   CALL("DefinedHeaderTable()");
   DOP(_defined.freeze());
  };

  ~DefinedHeaderTable() 
  {
    CALL("destructor ~DefinedHeaderTable()");    
  };

  void init()
  {
   CALL("init()");
    #ifdef NO_DEBUG 
     _defined.init(false,0);
    #else
     _defined.init(false,"DefinedHeaderTable::_defined");
    #endif
   DOP(_defined.freeze());
  };

  void destroy()
  {
    CALL("destroy()");
    _defined.destroy();
  };

  void registerDefined(ulong headerNum) 
  {
   CALL("registerDefined(ulong headerNum)");
   DOP(_defined.unfreeze());
   _defined.sub(headerNum) = true;
   DOP(_defined.freeze());
  };

  bool isDefined(ulong headerNum) const 
  {
   CALL("isDefined(ulong headerNum) const");
   if (headerNum >= _defined.size()) return false;
   return _defined[headerNum];
  };  

 private:
  BK::Array<BK::GlobAlloc,bool,128UL,DefinedHeaderTable> _defined;
}; // class DefinedHeaderTable

}; // namespace VK
//======================================================================
#endif
