/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         InferenceDispatcher.hpp
// Description:  Interface to several inference engines.
// Created:      Mar 5, 2000, 20:40
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef INFERENCE_DISPATCHER_H
//====================================================
#define INFERENCE_DISPATCHER_H
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "or_index.hpp"
#include "binary_resolution.hpp"
#include "forward_superposition.hpp"
#include "eq_factoring.hpp"
#include "backward_superposition.hpp"
#include "Chaining.hpp"
namespace VK
{
  class Clause;
  class NewClause;
  class SymbolInfo;
}; // namespace VK
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INFERENCE_DISPATCHER
 #define DEBUG_NAMESPACE "InferenceDispatcher"
#endif
#include "debugMacros.hpp"
//====================================================
namespace VK
{
class InferenceDispatcher
{
public:
  InferenceDispatcher(NewClause* freshClause);
  ~InferenceDispatcher();
  void init(NewClause* freshClause);
  void destroy();
  void reset();


  void setParamodulationFlag(bool fl) 
    {
      _paramodulationFlag = fl;
      _ordResIndex.SetParamodulation(fl); 
    }; // void setParamodulationFlag(bool fl) 

  void setParamodulateFromVariablesFlag(bool fl) 
  {
    _paramodulateFromVariables = fl;
    _ordResIndex.setParamodulateFromVariablesFlag(fl);
    _backwardSuperposition.setParamodulateFromVariablesFlag(fl);
  };

  void setSimBackSup(long degree)
    {
      _backwardSuperposition.setSimultaneous(degree);
    };
  
  void setChainingTable(const SymbolInfo* table)
  {
    _chainingTable = table;
    _chaining.setTransitivity(table);
    _ordResIndex.setChainingTable(table);
  };

  void setChainingThroughVariablesFlag(bool fl)
  {
    _chainingThroughVariables = fl;
    _chaining.setChainingThroughVariablesFlag(fl);
    _ordResIndex.setChainingThroughVariablesFlag(fl);
  };

  void setMainPremiseSubsumption(bool fl)
  {
    _binRes.setMainPremiseSubsumption(fl);
    _forwardSuperposition.setMainPremiseSubsumption(fl);
    _backwardSuperposition.setMainPremiseSubsumption(fl);
    _chaining.setMainPremiseSubsumption(fl);
  };

  void setSidePremiseSubsumption(bool fl)
  {
    _binRes.setSidePremiseSubsumption(fl);
    _forwardSuperposition.setSidePremiseSubsumption(fl);
    _backwardSuperposition.setSidePremiseSubsumption(fl);
    _chaining.setSidePremiseSubsumption(fl);
  };

  void setMainSelectedLitDemodulation(bool fl)
  {
    _binRes.setMainSelectedLitDemodulation(fl);
    _forwardSuperposition.setMainSelectedLitDemodulation(fl);
    _backwardSuperposition.setMainSelectedLitDemodulation(fl);
    _chaining.setMainSelectedLitDemodulation(fl);
  };

  void setSideSelectedLitDemodulation(bool fl)
  {
    _binRes.setSideSelectedLitDemodulation(fl);
    _forwardSuperposition.setSideSelectedLitDemodulation(fl);
    _backwardSuperposition.setSideSelectedLitDemodulation(fl);
    _chaining.setSideSelectedLitDemodulation(fl);
  };

  void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;
      _binRes.setNoInfBetweenRulesFlag(fl);
      _forwardSuperposition.setNoInfBetweenRulesFlag(fl);
      _backwardSuperposition.setNoInfBetweenRulesFlag(fl);
    };

  ulong statBlockedParamodulationsFromVariables() const
  {
    return 
      _ordResIndex.statBlockedParamodulationsFromVariables() +
      _backwardSuperposition.statBlockedParamodulationsFromVariables();
  };

  ulong statBlockedChainingsThroughVariables() const
  {
    return 
      _ordResIndex.statBlockedChainingsThroughVariables() +
      _chaining.statBlockedChainingsThroughVariables();
  };

  
  ulong statSubsumedMainPremiseInstances() const
  {
    return 
      _binRes.statSubsumedMainPremiseInstances() +
      _forwardSuperposition.statSubsumedMainPremiseInstances() +
      _backwardSuperposition.statSubsumedMainPremiseInstances() +
      _chaining.statSubsumedMainPremiseInstances();
  };

  ulong statSubsumedSidePremiseInstances() const
  {
    return 
      _binRes.statSubsumedSidePremiseInstances() + 
      _forwardSuperposition.statSubsumedSidePremiseInstances() +
      _backwardSuperposition.statSubsumedSidePremiseInstances() +
      _chaining.statSubsumedSidePremiseInstances();
  };

  ulong statDemodulatedMainSelectedLits() const
  {
    return 
      _binRes.statDemodulatedMainSelectedLits() +
      _forwardSuperposition.statDemodulatedMainSelectedLits() +
      _backwardSuperposition.statDemodulatedMainSelectedLits() +
      _chaining.statDemodulatedMainSelectedLits();
  };

  ulong statDemodulatedSideSelectedLits() const
  {
    return 
      _binRes.statDemodulatedSideSelectedLits() + 
      _forwardSuperposition.statDemodulatedSideSelectedLits() +
      _backwardSuperposition.statDemodulatedSideSelectedLits() +
      _chaining.statDemodulatedSideSelectedLits();
  };

  void newQuery(Clause* cl) 
    { 
      CALL("newQuery(Clause* cl)");
      ASSERT(!cl->isAnswer());
      _query = cl; 
      if (builtInTransitivityLaw(_query))
	{
	  _noMoreInferences = true;
	}
      else
	{
	  _noMoreInferences = false;
	  _continuation = FirstBinRes;
	}; 
    };
  bool nextInference() { return nextInf(); };
  void integrateDefinition(Clause* cl,const TERM& definedHeader);
  void removeDefinition(Clause* cl);
  void makeActive(Clause* cl);
  void remove(Clause* cl);
 public:
  bool binResFlag;
 private:
  enum Continuation 
  { 
    FirstBinRes, 
    NextBinRes, 
    FirstChaining,
    NextChaining,
    FirstForwSup, 
    NextForwSup,
    FirstEqFactor,
    NextEqFactor,
    FirstBackSup,
    NextBackSup 
  };
 private:
  bool nextInf();
  bool builtInTransitivityLaw(const Clause* cl)
    {
      CALL("builtInTransitivityLaw(const Clause* cl)");
      if (!_chainingTable) return false;
      ulong transPred;
      if (!cl->isTransitivityLaw(transPred)) return false;
      return _chainingTable->transitivityLaw(transPred);
    };
 private:
  bool _paramodulationFlag;
  bool _paramodulateFromVariables;
  bool _noInfBetweenRulesFlag;
  const SymbolInfo* _chainingTable;
  bool _chainingThroughVariables;
  bool _noMoreInferences;
  OR_INDEX _ordResIndex;
  BIN_ORD_RES _binRes;
  FORWARD_SUPERPOSITION _forwardSuperposition;
  EQ_FACTORING _equalityFactoring;
  BACKWARD_SUPERPOSITION _backwardSuperposition;
  Chaining _chaining;
  Clause* _query;
  NewClause* _freshClause;
  Continuation _continuation;
}; // class InferenceDispatcher


}; // namespace VK
//====================================================
#endif
