/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         InferenceTree.h
// Description:  Collecting inference trees.
// Created:      Feb 13, 2000, 18:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef INFERENCE_TREE_H
//=============================================================================
#define INFERENCE_TREE_H 
#include "jargon.h"
#include "VampireKernelDebugFlags.h"
#include "GSortTree.h"
#include "Clause.h"
#include "GlobAlloc.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INFERENCE_TREE
 #define DEBUG_NAMESPACE "InferenceTree"
#endif
#include "debugMacros.h"
//============================================================================
namespace VK
{
  using namespace BK;
class InferenceTree
{
 public:
  InferenceTree() 
  {
   CALL("constructor InferenceTree()");
  };
  InferenceTree(const Clause* root) : collected(), iter()
  {
   CALL("constructor InferenceTree(const Clause* root)");
   collect(root);
  };
  ~InferenceTree() { CALL("destructor ~InferenceTree()"); };
  void collect(const Clause* root);
  void resetIter() { iter.reset(collected); };
  const Clause* getNextClause() 
  {
   if (iter)
    {
     const Clause* res = iter.value().clause();
     iter.next();
     return res; 
    }; 
   return (const Clause*)0; 
  };
 private:
  class ClWrapper
  {
   public:
    ClWrapper(const Clause* c) : cl(c) {};
    ~ClWrapper() {};
    const Clause* clause() const { return cl; };
    bool operator==(const ClWrapper& cw) const
    {
     return cl->Number() == cw.clause()->Number();
    };
    bool operator<(const ClWrapper& cw) const
    {
     return cl->Number() < cw.clause()->Number();
    };
    bool operator>(const ClWrapper& cw) const
    {
     return cl->Number() > cw.clause()->Number();
    };
   private:
    ClWrapper() { ERROR("Default constructor called"); };
   private:  
    const Clause* cl; 
  };
  typedef GSortTree<GlobAlloc,ClWrapper,InferenceTree> ClBag;
 private:
  void collectInf(const Clause* root);
 private:
  ClBag collected;
  ClBag::LeftRightIter iter;    
}; //class InferenceTree 

}; // namespace VK

//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
