/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         Input.cpp
// Description:  Input signature, terms and clauses.
// Created:      Jun 04, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//======================================================================
#include "Input.h"
//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Symbol"
#endif
#include "debugMacros.h"
//======================================================================

using namespace VK;

using namespace BK;


#ifdef DEBUG_ALLOC_OBJ_TYPE
ClassDesc Input::TermList::_classDesc("Input::TermList",
				      Input::TermList::minSize(),
				      Input::TermList::maxSize());


ClassDesc Input::LiteralList::_classDesc("Input::LiteralList",
					 Input::LiteralList::minSize(),
					 Input::LiteralList::maxSize());

ClassDesc Input::ClauseList::_classDesc("Input::ClauseList",
					Input::ClauseList::minSize(),
					Input::ClauseList::maxSize());

#endif	
			      

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Signature"
#endif
#include "debugMacros.h"
//======================================================================

const Input::Symbol* Input::Signature::registerVariable(const char* name) 
{
  CALL("registerVariable(const char* name)");
  Symbol** pos = lookUp(name,&_symbols);
  if (!(*pos))
    {
      *pos = new Symbol(Symbol::Variable,name);
      (*pos)->setNext(_symbolList);
      _symbolList = *pos;
      (*pos)->variableTimeStamp() = _currentVariableTimeStamp;
      (*pos)->normalisedVariable() = nextNormalisedVariable();
      return (*pos)->normalisedVariable();
    };
  if ((*pos)->type() != Symbol::Variable)
    {
      if ((*pos)->type() == Symbol::Unknown) 
	{
	  (*pos)->setType(Symbol::Variable);
	}
      else
	return (Symbol*)0;
    };
  if ((*pos)->variableTimeStamp() == _currentVariableTimeStamp)
    {
      return (*pos)->normalisedVariable();
    };
  (*pos)->variableTimeStamp() = _currentVariableTimeStamp;
  (*pos)->normalisedVariable() = nextNormalisedVariable();
  return (*pos)->normalisedVariable();
}; // Symbol* Input::Signature::registerVariable(const char* name) 
  
const Input::Symbol* Input::Signature::registerFunction(const char* name,ulong arity)
{
  CALL("registerFunction(const char* name,ulong arity)");
  Symbol** pos = lookUp(name,&_symbols);
  if (!(*pos))
    {
      *pos = new Symbol(Symbol::Function,name,arity);
      (*pos)->setNext(_symbolList);
      _symbolList = *pos;
      return *pos;
    };
  
  if ((*pos)->type() != Symbol::Function)
    {
      if ((*pos)->type() == Symbol::Unknown)
	{
	  (*pos)->setType(Symbol::Function);
	}
      else
	return (Symbol*)0;	
    };
  
  if ((*pos)->arityAssigned())
    {
      if ((*pos)->arity() != arity) return (Symbol*)0;
    }
  else
    (*pos)->setArity(arity);

  return *pos;
}; // Symbol* Input::Signature::registerFunction(const char* name,ulong arity)
  
const Input::Symbol* Input::Signature::registerPredicate(const char* name,ulong arity)
{
  CALL("registerPredicate(const char* name,ulong arity)");
  Symbol** pos = lookUp(name,&_symbols);
  if (!(*pos))
    {
      *pos = new Symbol(Symbol::Predicate,name,arity);
      (*pos)->setNext(_symbolList);
      _symbolList = *pos;
      return *pos;
    };
  
  if ((*pos)->type() != Symbol::Predicate)
    {
      if ((*pos)->type() == Symbol::Unknown)
	{
	  (*pos)->setType(Symbol::Predicate);
	}
      else
	return (Symbol*)0;	
    };
  
  if ((*pos)->arityAssigned())
    {
      if ((*pos)->arity() != arity) return (Symbol*)0;
    }
  else
    (*pos)->setArity(arity);

  return *pos;
}; // Symbol* Input::Signature::registerPredicate(const char* name,ulong arity)

Input::Symbol* Input::Signature::registerSymbol(const char* name)
{
  CALL("registerSymbol(const char* name)");
  Symbol** pos = lookUp(name,&_symbols);
  if (!(*pos))
    {
      *pos = new Symbol(Symbol::Unknown,name);
      (*pos)->setNext(_symbolList);
      _symbolList = *pos;
    };
  return *pos;
}; // Input::Symbol* Input::Signature::registerSymbol(const char* name)

bool Input::Signature::registerPrecedence(const char* name,long prec)
{
  CALL("registerPrecedence(const char* name,long prec)");
  Symbol* symbol = registerSymbol(name);
  if (symbol->precedenceAssigned())
    return (symbol->precedence() == prec);
  
  Symbol** pos = lookUpPrecedence(prec,&_precedences);
  if (*pos) return false; // this precedence has been reserved for another symbol
  
  *pos = symbol;
  symbol->assignPrecedence(prec);
  return true;
}; // bool Input::Signature::registerPrecedence(const char* name,long prec)

bool Input::Signature::registerAnswerPredicate(const char* name)
{
  CALL("registerAnswerPredicate(const char* name)");
  Symbol* symbol = registerSymbol(name);
  if (symbol->type() != Symbol::Predicate)
    {
      if (symbol->type() == Symbol::Unknown)
	{
	  symbol->setType(Symbol::Predicate);
	}
      else
	return false;
    };
  symbol->markAsAnswer();
  return true;
}; // bool Input::Signature::registerAnswerPredicate(const char* name)


Input::Symbol* Input::Signature::nextNormalisedVariable()
{
  CALL("nextNormalisedVariable()");
  Symbol* res = *_nextNormalisedVariable;
  if (res)
    {
      _nextNormalisedVariable = &(res->greaterNames());
      return res;
    };
  res = new Symbol(Symbol::Variable,
		   normalisedVariableName(_nextNormalisedVariableNumber));
  res->normalisedVariableNumber() = _nextNormalisedVariableNumber;
  ++_nextNormalisedVariableNumber;
  *_nextNormalisedVariable = res;
  _nextNormalisedVariable = &(res->greaterNames());
  return res; 
}; // Symbol* Input::Signature::nextNormalisedVariable()
  
const char* Input::Signature::normalisedVariableName(ulong varNum)
{
  CALL("normalisedVariableName(ulong varNum)");
  static char memory[256];
  snprintf(memory,256,"X%lu",varNum);
  return memory;
}; // const char* Input::Signature::normalisedVariableName(ulong varNum)

void Input::Signature::destroyTree(Symbol* tree)
{
  CALL("destroyTree(Symbol* tree)");
  if (tree)
    {
      destroyTree(tree->smallerNames());
      destroyTree(tree->greaterNames());
      delete tree;
    };
}; // void Input::Signature::destroyTree(Symbol* tree)
  
Input::Symbol** Input::Signature::lookUp(const char* name,Symbol** tree)
{
  CALL("lookUp(const char* name,Symbol** tree)");
  Symbol* root = *tree;
  if (!root) return tree;
  int cmp = strcmp(name,root->printName());
  if (!cmp) return tree;
  if (cmp < 0) return lookUp(name,&(root->smallerNames()));
  return lookUp(name,&(root->greaterNames()));
}; // Symbol** Input::Signature::lookUp(const char* name,Symbol** tree)

Input::Symbol** Input::Signature::lookUpPrecedence(long precedence,Symbol** tree)
{
  CALL("lookUpPrecedence(long precedence,Symbol** tree)");  
  Symbol* root = *tree;
  if (!root) return tree;
  if (root->precedence() == precedence) return tree;
  if (root->precedence() > precedence)
    return lookUpPrecedence(precedence,&(root->smallerPrecedences()));
  return lookUpPrecedence(precedence,&(root->greaterPrecedences()));
}; // Input::Symbol** Input::Signature::lookUpPrecedence(long precedence,Symbol** tree)




//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Term"
#endif
#include "debugMacros.h"
//======================================================================

#ifndef NO_DEBUG_VIS
ostream& operator<<(ostream& str,const VK::Input::Term* t)
{
  CALL("operator<<(ostream& str,const VK::Input::Term* t)");
  str << t->top();
  if (t->arguments())
    {
      str << "(";
      for (const Input::TermList* arg = t->arguments(); arg; arg = arg->tl())
	{
	  str << arg->hd();
	  if (arg->tl()) str << ",";
	};
      str << ")";
    };
  return str;
};
#endif

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input::Clause"
#endif
#include "debugMacros.h"
//======================================================================


#ifndef NO_DEBUG_VIS

ostream& operator<<(ostream& str,const VK::Input::Clause* cl)
{
  CALL("operator<<(ostream& str,const VK::Input::Clause* cl)");
  str << cl->name() << ": ";
  for (const Input::LiteralList* lit = cl->literals(); lit; lit = lit->tl())
    {
      str << lit->hd();
      if (lit->tl()) str << ", ";
    };
  return str;
};

#endif


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INPUT
 #define DEBUG_NAMESPACE "Input"
#endif
#include "debugMacros.h"
//======================================================================

void Input::destroyTermList(TermList* lst)
{
  CALL("destroyTermList(TermList* lst)");
  if (lst)
    {
      destroyTermList(lst->tl());
      delete lst->hd();
      delete lst;
    };
}; // void Input::destroyTermList(TermList* lst)


void Input::destroyLiteralList(LiteralList* lst)
{
  CALL("destroyLiteralList(LiteralList* lst)");
  if (lst)
    {
      destroyLiteralList(lst->tl());
      delete lst->hd();
      delete lst;
    };
}; // void Input::destroyLiteralList(LiteralList* lst)


void Input::destroyClauseList(ClauseList* lst)
{
  CALL("destroyClauseList(ClauseList* lst)");
  if (lst)
    {
      destroyClauseList(lst->tl());
      delete lst->hd();
      delete lst;
    };
}; // void Input::destroyClauseList(ClauseList* lst)

#ifndef NO_DEBUG_VIS
ostream& operator<<(ostream& str,const Input::ClauseList* lst)
{
  while (lst)
    {
      str << lst->hd() << "\n";
      lst = lst->tl();
    };
  return str;
}; // ostream& operator<<(ostream& str,const Input::ClauseList* lst)

#endif

//======================================================================
