/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         InstanceRetrievalForDemodulation.cpp
// Description:  Indexing for matching in backward demodulation.
// Created:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//====================================================
#include "InstanceRetrievalForDemodulation.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_INSTANCE_RETRIEVAL_FOR_DEMODULATION
 #define DEBUG_NAMESPACE "InstanceRetrievalForDemodulation"
#endif
#include "debugMacros.h"
//====================================================


using namespace VK;
using namespace BK;
using namespace Gematogen;

#ifdef DEBUG_ALLOC_OBJ_TYPE

ClassDesc InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::_classDesc("InstanceRetrievalForDemodulation::Index::NonmaxPathRelation",
										  InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::minSize(),
										  InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::maxSize());

ClassDesc InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::_headerClassDesc("InstanceRetrievalForDemodulation::Index::NonmaxPathRelation(header)",
											InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::headerMinSize(),
											InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::headerMaxSize());



ClassDesc InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::Node::_classDesc("InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::Node",
											InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::Node::minSize(),
											InstanceRetrievalForDemodulation::Index::NonmaxPathRelation::Node::maxSize());





ClassDesc InstanceRetrievalForDemodulation::Index::MaxPathRelation::_classDesc("InstanceRetrievalForDemodulation::Index::MaxPathRelation",
									       InstanceRetrievalForDemodulation::Index::MaxPathRelation::minSize(),
									       InstanceRetrievalForDemodulation::Index::MaxPathRelation::maxSize());

ClassDesc InstanceRetrievalForDemodulation::Index::MaxPathRelation::_headerClassDesc("InstanceRetrievalForDemodulation::Index::MaxPathRelation(header)",
										     InstanceRetrievalForDemodulation::Index::MaxPathRelation::headerMinSize(),
										     InstanceRetrievalForDemodulation::Index::MaxPathRelation::headerMaxSize());



ClassDesc InstanceRetrievalForDemodulation::Index::MaxPathRelation::Node::_classDesc("InstanceRetrievalForDemodulation::Index::MaxPathRelation::Node",
										     InstanceRetrievalForDemodulation::Index::MaxPathRelation::Node::minSize(),
										     InstanceRetrievalForDemodulation::Index::MaxPathRelation::Node::maxSize());





ClassDesc InstanceRetrievalForDemodulation::Index::PathIndexForInstanceRetrieval::Node::_funcClassDesc(
												       "InstanceRetrievalForDemodulation::Index::PathIndexForInstanceRetrieval::Node(func)",
												       InstanceRetrievalForDemodulation::Index::PathIndexForInstanceRetrieval::Node::funcMinSize(),
												       InstanceRetrievalForDemodulation::Index::PathIndexForInstanceRetrieval::Node::funcMaxSize());

ClassDesc InstanceRetrievalForDemodulation::Index::PathIndexForInstanceRetrieval::Node::_constClassDesc(
													"InstanceRetrievalForDemodulation::Index::PathIndexForInstanceRetrieval::Node(const)",
													InstanceRetrievalForDemodulation::Index::PathIndexForInstanceRetrieval::Node::constMinSize(),
													InstanceRetrievalForDemodulation::Index::PathIndexForInstanceRetrieval::Node::constMaxSize());


#endif

InstanceRetrievalForDemodulation::InstanceRetrievalForDemodulation() 
{
};
InstanceRetrievalForDemodulation::~InstanceRetrievalForDemodulation() 
{
};

void InstanceRetrievalForDemodulation::destroy()
{
  CALL("destroy()");
  _index.destroyThoroughly();
};

void InstanceRetrievalForDemodulation::integrate(const TERM* complexTerm)
{
  CALL("integrate(const TERM* complexTerm)");
  _index.integrate(TERM((TERM*)complexTerm));
}; 
void InstanceRetrievalForDemodulation::remove(const TERM* complexTerm)
{
  CALL("remove(const TERM* complexTerm)");
  _index.remove(TERM((TERM*)complexTerm));
}; 

bool InstanceRetrievalForDemodulation::newQuery(const TERM* query)
{
  CALL("newQuery(const TERM* query)");
  _noMoreMatches = !_index.compileQuery(TERM((TERM*)query));
  DOP(_debugQuery = query);
  return !_noMoreMatches;
}; 
bool InstanceRetrievalForDemodulation::nextMatch(const TERM*& match)
{
  CALL("nextMatch(const TERM*& match)");
  if (_noMoreMatches) return false;
  TERM matchId;
  bool result = _index.nextMatch(matchId);
  if (result)
    {
      ASSERT(matchId.IsReference());
      match = matchId.First();
      ASSERT(isValidMatch(match));
      return true;
    };
  _noMoreMatches = true;
  return false;
}; 


#ifdef DEBUG_NAMESPACE
bool InstanceRetrievalForDemodulation::isValidMatch(const TERM* term) const
{
  static Substitution::Instance queryInst;
  static TERM::Iterator matchIter; 
  queryInst.reset(_debugQuery,substitution());
  matchIter.Reset(term);
  do
    {
      if (queryInst.symbol() != matchIter.symbolRef()) return false;
      matchIter.Next();
    }
  while (queryInst.next());
  return true;
}; // bool InstanceRetrievalForDemodulation::isValidMatch(const TERM* term) const

#endif

//====================================================
