/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         LiteralSelection.h
// Description:  Defines various literal selection functions
// Created:      Dec 12, 2001
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Note:         Was a part of NewClause.
// Revised:      Dec 15, 2001. Bug fix in setSplittingFlag.
//============================================================================
#ifndef LITERAL_SELECTION_H
//=============================================================================

#define LITERAL_SELECTION_H 
#include "jargon.h"
#include "RuntimeError.h"
#include "VampireKernelDebugFlags.h"
#include "TmpLitList.h"
class SymbolInfo;
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#define DEBUG_NAMESPACE "LiteralSelection"
#endif
#include "debugMacros.h"
//============================================================================

namespace VK
{
  using namespace BK;
  using namespace Gematogen;
class LiteralSelection
{
 public:
  LiteralSelection() 
    {
      init();
    };
  
  ~LiteralSelection() {};
  void init()
    { 
      _literalComparisonMode = 0L;
      _inversePolarity = false;
      _basicSelFun = function1; 
      _splittingIsEnabled = false;
      _splittingStyle = 0L;
    };
  void setFunNum(long num);
  void setLiteralComparisonMode(long mode) { _literalComparisonMode = mode; };

  void setSplittingFlag(bool fl) 
    { 
      _splittingIsEnabled = (_splittingIsEnabled || fl);
      //  _splittingIsEnabled can only be changed from false to true.
    };
  void setSplittingStyle(long splittingStyle) { _splittingStyle = splittingStyle; };


  inline 
    bool 
    isPositiveForSelection(const TERM& header) const;
  
  inline
    bool 
    isNegativeForSelection(const TERM& header) const;


    void 
    select(TmpLitList& lits,TmpLitList& selectedLits,
	   TmpLitList& nonselectedLits,bool& negSelectionUsed) const;

  bool 
    selectForElimination(TmpLitList& lits,TmpLitList& selectedLits,
			 TmpLitList& nonselectedLits,  
			 const SymbolInfo* eliminationTable) const;

 private:

  void 
    selectWithNonblockingSplitting(TmpLitList& lits,TmpLitList& selectedLits,
				   TmpLitList& nonselectedLits,bool& negSelectionUsed,
	                           const TmpLitList::Statistics& stat) const;

  void 
    selectWithBlockingSplitting(TmpLitList& lits,TmpLitList& selectedLits,
				TmpLitList& nonselectedLits,bool& negSelectionUsed,
	                        const TmpLitList::Statistics& stat) const;


  static 
    void 
    function1(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& negSelectionUsed,
	      const TmpLitList::Statistics& stat);

  static 
    void 
    function2(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& negSelectionUsed,
	      const TmpLitList::Statistics& stat);

  static 
    void 
    function3(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& negSelectionUsed,
	      const TmpLitList::Statistics& stat);

  static 
    void 
    function4(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& negSelectionUsed,
	      const TmpLitList::Statistics& stat); 

  static 
    void 
    function5(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& negSelectionUsed,
	      const TmpLitList::Statistics& stat); 

  static 
    void 
    function6(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& negSelectionUsed,
	      const TmpLitList::Statistics& stat);  
  
  static 
    void 
    function7(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& negSelectionUsed,
	      const TmpLitList::Statistics& stat);


  static 
    void 
    function1002(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& negSelectionUsed,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1003(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& negSelectionUsed,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1004(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& negSelectionUsed,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1005(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& negSelectionUsed,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1006(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& negSelectionUsed,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1007(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& negSelectionUsed,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1008(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& negSelectionUsed,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1009(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& negSelectionUsed,
		 const TmpLitList::Statistics& stat);  


  inline
    static
    void 
    selectInherentlySelected(TmpLitList& lits,ulong numOfSelected,TmpLitList& selectedLits,TmpLitList& nonselectedLits);


  inline
    bool 
    satisfactorySelection1(TmpLitList& selectedLits,TmpLitList& nonselectedLits) const;

  inline
    bool 
    satisfactorySelection2(TmpLitList& selectedLits,TmpLitList& nonselectedLits) const;
  
  inline
    bool
    satisfactorySelection3(TmpLitList& selectedLits,TmpLitList& nonselectedLits) const;
   

  inline
    void 
    selectNegativeWithMaximalWeight(TmpLitList& lits,TmpLitList& selectedLits,
				    TmpLitList& nonselectedLits) const;

  inline
    void 
    selectNegativeLeastProlific(TmpLitList& lits,TmpLitList& selectedLits,
				TmpLitList& nonselectedLits) const;  
  
  inline
    void 
    selectNegativeMaximallyNongroundWithMaximalWeight(TmpLitList& lits,
						      TmpLitList& selectedLits,
						      TmpLitList& nonselectedLits) const;

  inline 
    void 
    unselectAnswerLiterals(TmpLitList& lits,
			   TmpLitList& selectedLits,
			   TmpLitList& nonselectedLits,
                           ulong numOfAnswerLits) const;


 private:
  static const LiteralSelection* _currentObject; 
  // To communicate the current object to the static selection functions.
  // They are made static to be represented by pointers.   
  void (*_basicSelFun)(TmpLitList&,TmpLitList&,TmpLitList&,bool&,const TmpLitList::Statistics&);
  long _literalComparisonMode;
  bool _inversePolarity;
  bool _splittingIsEnabled;
  long _splittingStyle;
}; // class LiteralSelection

}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#define DEBUG_NAMESPACE "LiteralSelection"
#endif
#include "debugMacros.h"
//============================================================================


namespace VK
{
  using namespace BK;
  using namespace Gematogen;
inline 
void 
LiteralSelection::setFunNum(long num)
{
  _inversePolarity = (num < 0);
  switch (num) 
    {
    case 1L : 
      _basicSelFun = function1; 
      return;
    case 2L : 
      _basicSelFun = function2;     
      return;
    case 3L : 
      _basicSelFun = function3; 
      return;
    case 4L : 
      _basicSelFun = function4; 
      return;
    case 5L : 
      _basicSelFun = function5;
      return;
    case 6L : 
      _basicSelFun = function6; 
      return;
    case 7L : 
      _basicSelFun = function7; 
      return;


    case 1002L :
      _basicSelFun = function1002; 
      return;
    case 1003L :
      _basicSelFun = function1003; 
      return;
    case 1004L :
      _basicSelFun = function1004; 
      return;
    case 1005L :
      _basicSelFun = function1005; 
      return;
    case 1006L :
      _basicSelFun = function1006; 
      return;
    case 1007L :
      _basicSelFun = function1007; 
      return;
    case 1008L :
      _basicSelFun = function1008; 
      return;    
    case 1009L :
      _basicSelFun = function1009; 
      return;          
     

    case -2L : 
      _basicSelFun = function2;
      return;
    case -3L : 
      _basicSelFun = function3;
      return;
    case -4L : 
      _basicSelFun = function4;
      return;
    case -5L : 
      _basicSelFun = function5;
      return;
    case -6L : 
      _basicSelFun = function6;
      return;
    case -7L : 
      _basicSelFun = function7;
      return;

    case -1005L :
      _basicSelFun = function1005; 
      return;
    case -1009L :
      _basicSelFun = function1009; 
      return;

    default: RuntimeError::report("Wrong selection number.");
    };
}; // void LiteralSelection::setFunNum(long num)





inline 
bool 
LiteralSelection::isPositiveForSelection(const TERM& header) const
{
  if (header.IsEquality() || header.IsSplitting() || (!_inversePolarity)) return header.Positive();
  return header.Negative();
}; 


inline 
bool 
LiteralSelection::isNegativeForSelection(const TERM& header) const
{
  if (header.IsEquality() || header.IsSplitting()  || (!_inversePolarity)) return header.Negative();
  return header.Positive();
}; 


}; // namespace VK


//======================================================================
#endif
