/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         LiteralSelection.hpp
// Description:  Defines various literal selection functions
// Created:      Dec 12, 2001
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Note:         Was a part of NewClause.
// Revised:      Dec 15, 2001. Bug fix in setSplittingFlag.
//============================================================================
#ifndef LITERAL_SELECTION_H
//=============================================================================

#define LITERAL_SELECTION_H 
#include "jargon.hpp"
#include "RuntimeError.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TmpLitList.hpp"
#include "StandardLiteralOrdering.hpp"
#include "CustomLiteralSelection.hpp"
class SymbolInfo;
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#define DEBUG_NAMESPACE "LiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace VK
{
class LiteralSelection
{
 public:
  LiteralSelection() :
    _oneOfMaxSizeFilter(CustomLiteralSelection::literalSize),
    _allOfMaxSizeFilter(CustomLiteralSelection::literalSize),
    _allNegEqOrAllFilter(CustomLiteralSelection::isNegativeEquality),
    _allExceptPosEqOrAllFilter(CustomLiteralSelection::isNotPositiveEquality),
    _allWithMaxNumOfDiffVarsFilter(CustomLiteralSelection::numOfDifferentVariables),
    _allWithMostShallowVarsFilter(&_inverseVarDepthProfile1,&_inverseVarDepthProfile2),
    _complexFilter3(&_complexCharacteristic1_1,&_complexCharacteristic1_2),
    _complexFilter5(&_complexCharacteristic1_1,&_complexCharacteristic1_2)
    {      
      _literalComparisonMode = 0L;
      _standardLiteralOrdering = 0;
      _basicSelFun = function1; 
      _splittingIsEnabled = false;
      _splittingStyle = 0L;
      
      _oneOfEachPolarityFilter.setSelectOne(false);

      _oneOfMaxSizeFilter.setSelectOne(true);
      _allOfMaxSizeFilter.setSelectOne(false);

      _allNegEqOrAllFilter.setSelectOne(false);

      _allExceptPosEqOrAllFilter.setSelectOne(false);


      _oneBigNegEqOrOneBigFilter.setSelectOne(true);
      _oneBigNegEqOrOneBigFilter.resetFilters();
      _oneBigNegEqOrOneBigFilter.addFilter(&_allNegEqOrAllFilter);
      _oneBigNegEqOrOneBigFilter.addFilter(&_oneOfMaxSizeFilter);

      _allWithMaxNumOfDiffVarsFilter.setSelectOne(false);
      _allWithMostShallowVarsFilter.setSelectOne(false);

      _oneOfMaxSizeMaxNongroundFilter.setSelectOne(true);
      _oneOfMaxSizeMaxNongroundFilter.resetFilters();
      _oneOfMaxSizeMaxNongroundFilter.addFilter(&_allWithMaxNumOfDiffVarsFilter);
      _oneOfMaxSizeMaxNongroundFilter.addFilter(&_allWithMostShallowVarsFilter);
      _oneOfMaxSizeMaxNongroundFilter.addFilter(&_oneOfMaxSizeFilter);


      _complexFilter1.setSelectOne(true);
      _complexFilter1.resetFilters();
      _complexFilter1.addFilter(&_allExceptPosEqOrAllFilter);
      _complexFilter1.addFilter(&_allWithMaxNumOfDiffVarsFilter);
      _complexFilter1.addFilter(&_allWithMostShallowVarsFilter);
      _complexFilter1.addFilter(&_oneOfMaxSizeFilter);
      
      _complexFilter2.setSelectOne(true);
      _complexFilter2.resetFilters();
      _complexFilter2.addFilter(&_allExceptPosEqOrAllFilter);
      _complexFilter2.addFilter(&_allOfMaxSizeFilter);
      _complexFilter2.addFilter(&_allWithMaxNumOfDiffVarsFilter);
      _complexFilter2.addFilter(&_allWithMostShallowVarsFilter);

      _complexFilter3.setSelectOne(false);

      _complexFilter4.setSelectOne(true);
      _complexFilter4.resetFilters();
      _complexFilter4.addFilter(&_allExceptPosEqOrAllFilter);
      _complexFilter4.addFilter(&_complexFilter3);
      _complexFilter4.addFilter(&_oneOfMaxSizeFilter);

      _complexFilter5.setSelectOne(true);
      

      _maxSelectionAcceptability1.resetPrimitives();
      _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::oneLiteralSelected);
      _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::noBigNegativeNonselected);
      _maxSelectionAcceptability1.addPrimitive(CustomLiteralSelection::noNegativeEqualitiesNonselected);

      _maxSelectionAcceptability2.resetPrimitives();
      _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::oneLiteralSelected);
      _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::noNegativeWithManyDiffVarsNonselected);
      _maxSelectionAcceptability2.addPrimitive(CustomLiteralSelection::noNegativeEqualitiesNonselected);
    };
  
  ~LiteralSelection() {};
  void init()
    { 
     new (this) LiteralSelection();
    };
  void destroy()
  {
    delete this;
  };
  void setFunNum(long num);
  void setLiteralComparisonMode(long mode) { _literalComparisonMode = mode; };
  void setStandardLiteralOrdering(StandardLiteralOrdering* ord) 
  { 
    _standardLiteralOrdering = ord;
    _totalSelection.setStandardLiteralOrdering(ord);
    _maximalSelection.setStandardLiteralOrdering(ord);
    _hyperSelection.setStandardLiteralOrdering(ord);
    _maxWithCorrectionSelection.setStandardLiteralOrdering(ord);
    _arbitrarySelection.setStandardLiteralOrdering(ord);
    _completeBidirectSelection.setStandardLiteralOrdering(ord);
    _incompleteBidirectSelection.setStandardLiteralOrdering(ord);
    _bestOnlySelection.setStandardLiteralOrdering(ord);
  };

  void setSplittingFlag(bool fl) 
    { 
      _splittingIsEnabled = (_splittingIsEnabled || fl);
      //  _splittingIsEnabled can only be changed from false to true.
    };
  void setSplittingStyle(long splittingStyle) { _splittingStyle = splittingStyle; };


#ifdef VKERNEL_FOR_STEP_RESOLUTION
  void setStepResolutionLiteralSelection(bool fl)
  {
    _stepResolutionLiteralSelectionFlag = fl;    
  };
#endif



    void 
      select(TmpLitList& lits,TmpLitList& selectedLits,
	     TmpLitList& nonselectedLits,bool& allMaximalSelected) const;
    
    void 
      selectForSupport(TmpLitList& lits,
		       TmpLitList& selectedLits,
		       TmpLitList& nonselectedLits) const;
    
  bool 
    selectForElimination(TmpLitList& lits,TmpLitList& selectedLits,
			 TmpLitList& nonselectedLits,  
			 const SymbolInfo* eliminationTable) const;

private:

  // Literals of different polarities are incomparable.
  // In positive literals, the more different variables
  // and var. occurences, the bigger the characteristic value.
  // In negative literals the other way around.

  class ComplexLiteralCharacteristic1 : public CustomLiteralSelection::ComplexLiteralCharacteristic
  {
  public:
    ComplexLiteralCharacteristic1() {};
    ~ComplexLiteralCharacteristic1();
    void computeFor(TmpLiteral* lit);
    BK::Comparison compareWith(ComplexLiteralCharacteristic* c);
  private:
    bool _positive;
    ulong _numOfDifferentVars;
    ulong _numOfVarOccurences;
  }; // class ComplexLiteralCharacteristic1

  class ComplexLiteralOrdering1 : public LiteralOrdering
  {
  public:
    ComplexLiteralOrdering1() 
    {
      _isStratified = true;
    };
    ~ComplexLiteralOrdering1();
    void destroy();
    BK::Comparison compare(TmpLiteral* lit1,
			   TmpLiteral* lit2) const;
    BK::Comparison compareLayers(const TmpLiteral* lit1,
				 const TmpLiteral* lit2) const;
  }; // class ComplexLiteralOrdering1


private:

  void* operator new(size_t,LiteralSelection* ptr) 
  {
    return ptr;
  };

  void operator delete(void* obj)
  {
  };

  void 
    selectWithNonblockingSplitting(TmpLitList& lits,TmpLitList& selectedLits,
				   TmpLitList& nonselectedLits,bool& allMaximalSelected,
	                           const TmpLitList::Statistics& stat) const;

  void 
    selectWithBlockingSplitting(TmpLitList& lits,TmpLitList& selectedLits,
				TmpLitList& nonselectedLits,bool& allMaximalSelected,
	                        const TmpLitList::Statistics& stat) const;


  static 
    void 
    function1(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& allMaximalSelected,
	      const TmpLitList::Statistics& stat);

  static 
    void 
    function2(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& allMaximalSelected,
	      const TmpLitList::Statistics& stat);

  static 
    void 
    function3(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& allMaximalSelected,
	      const TmpLitList::Statistics& stat);

  static 
    void 
    function4(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& allMaximalSelected,
	      const TmpLitList::Statistics& stat); 

  static 
    void 
    function5(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& allMaximalSelected,
	      const TmpLitList::Statistics& stat); 

  static 
    void 
    function6(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& allMaximalSelected,
	      const TmpLitList::Statistics& stat);  
  
  static 
    void 
    function7(TmpLitList& lits,TmpLitList& selectedLits,
	      TmpLitList& nonselectedLits,bool& allMaximalSelected,
	      const TmpLitList::Statistics& stat);


  static 
    void 
    function1002(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& allMaximalSelected,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1003(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& allMaximalSelected,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1004(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& allMaximalSelected,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1005(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& allMaximalSelected,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1006(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& allMaximalSelected,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1007(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& allMaximalSelected,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1008(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& allMaximalSelected,
		 const TmpLitList::Statistics& stat);

  static 
    void 
    function1009(TmpLitList& lits,TmpLitList& selectedLits,
		 TmpLitList& nonselectedLits,bool& allMaximalSelected,
		 const TmpLitList::Statistics& stat);  


  static 
    void 
    customBasicSelFun(TmpLitList& lits,TmpLitList& selectedLits,
		      TmpLitList& nonselectedLits,bool& allMaximalSelected,
		      const TmpLitList::Statistics& stat);  




  inline
    static
    void 
    selectInherentlySelected(TmpLitList& lits,ulong numOfSelected,TmpLitList& selectedLits,TmpLitList& nonselectedLits);


  inline
    bool 
    satisfactorySelection1(TmpLitList& selectedLits,TmpLitList& nonselectedLits) const;

  inline
    bool 
    satisfactorySelection2(TmpLitList& selectedLits,TmpLitList& nonselectedLits) const;
  
  inline
    bool
    satisfactorySelection3(TmpLitList& selectedLits,TmpLitList& nonselectedLits) const;
   

  inline
    void 
    selectNegativeWithMaximalWeight(TmpLitList& lits,TmpLitList& selectedLits,
				    TmpLitList& nonselectedLits) const;

  inline
    void 
    selectNegativeLeastProlific(TmpLitList& lits,TmpLitList& selectedLits,
				TmpLitList& nonselectedLits) const;  
  
  inline
    void 
    selectNegativeMaximallyNongroundWithMaximalWeight(TmpLitList& lits,
						      TmpLitList& selectedLits,
						      TmpLitList& nonselectedLits) const;

  inline 
    void 
    unselectAnswerLiterals(TmpLitList& lits,
			   TmpLitList& selectedLits,
			   TmpLitList& nonselectedLits,
                           ulong numOfAnswerLits) const;

#ifdef VKERNEL_FOR_STEP_RESOLUTION
  static 
  inline
  bool 
  isNextStepLiteral(const TmpLiteral* lit);

  static
  inline
  bool 
  isCurrentStepLiteral(const TmpLiteral* lit);

  void 
  inline
  unselectCurrentStateLiterals(TmpLitList& lits,
			       TmpLitList& selectedLits,
			       TmpLitList& nonselectedLits) const;
#endif



 private:
  static const LiteralSelection* _currentObject; 
  // To communicate the current object to the static selection functions.
  // They are made static to be represented by pointers.   
  void (*_basicSelFun)(TmpLitList&,TmpLitList&,TmpLitList&,bool&,const TmpLitList::Statistics&);
  long _literalComparisonMode;
  StandardLiteralOrdering* _standardLiteralOrdering;
  bool _splittingIsEnabled;
  long _splittingStyle;

  static CustomLiteralSelection* _customSelection; 


  // components for custom selection schemes
  CustomLiteralSelection::OneOfEachPolarityFilter _oneOfEachPolarityFilter;
  CustomLiteralSelection::MaxValFilter _oneOfMaxSizeFilter;
  CustomLiteralSelection::MaxValFilter _allOfMaxSizeFilter;
  CustomLiteralSelection::AllPOrAllFilter _allNegEqOrAllFilter;
  CustomLiteralSelection::AllPOrAllFilter _allExceptPosEqOrAllFilter;
  CustomLiteralSelection::CompositionFilter _oneBigNegEqOrOneBigFilter;
  CustomLiteralSelection::MaxValFilter _allWithMaxNumOfDiffVarsFilter;
  
  CustomLiteralSelection::InverseVarDepthProfile _inverseVarDepthProfile1;
  CustomLiteralSelection::InverseVarDepthProfile _inverseVarDepthProfile2;

  CustomLiteralSelection::MaxComplexValFilter _allWithMostShallowVarsFilter;
  CustomLiteralSelection::CompositionFilter _oneOfMaxSizeMaxNongroundFilter;

  ComplexLiteralCharacteristic1 _complexCharacteristic1_1;
  ComplexLiteralCharacteristic1 _complexCharacteristic1_2;
  

  CustomLiteralSelection::CompositionFilter _complexFilter1;
  // If possible, no positive equalities are selected.
  // From avail. literals we select max. nonground.
  // From remaining literals we select one of max. size.

  CustomLiteralSelection::CompositionFilter _complexFilter2;
  // If possible, no positive equalities are selected.
  // From avail. literals we select those of max. size.
  // From remaining literals we select one max. nonground.

  CustomLiteralSelection::MaxComplexValFilter _complexFilter3;
  // Select all maximal literals w.r.t. ComplexLiteralCharacteristic1.

  CustomLiteralSelection::CompositionFilter _complexFilter4;
  // First, try to get rid of positive equalities if
  // literals of other kinds exist. 
  // Then use _complexFilter3, then finish with _oneOfMaxSizeFilter.

  CustomLiteralSelection::MaxComplexValFilter _complexFilter5;
  // Same as _complexFilter3, but always selects 1 literal. 

  CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability _maxSelectionAcceptability1;
  CustomLiteralSelection::ConjOfPrimitiveSelectionAcceptability _maxSelectionAcceptability2;

  ComplexLiteralOrdering1 _complexLiteralOrdering1;
  

  // custom selection schemes
  TotalLiteralSelection _totalSelection;
  MaximalLiteralSelection _maximalSelection;
  HyperLiteralSelection _hyperSelection;
  MaxWithCorrectionLiteralSelection _maxWithCorrectionSelection;
  ArbitraryLiteralSelection _arbitrarySelection;
  CompleteBidirectionalLiteralSelection _completeBidirectSelection;
  IncompleteBidirectionalLiteralSelection _incompleteBidirectSelection;
  BestOnlyLiteralSelection _bestOnlySelection;


#ifdef VKERNEL_FOR_STEP_RESOLUTION
  bool _stepResolutionLiteralSelectionFlag;
#endif
}; // class LiteralSelection

}; // namespace VK

//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SELECTION
#define DEBUG_NAMESPACE "LiteralSelection"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{
inline 
void 
LiteralSelection::setFunNum(long num)
{
  TmpLiteral::setInversePolarityForSelection(num < 0);
  switch (num) 
    {
    case 0L : // every literal is selected  
      _basicSelFun = customBasicSelFun;
      _customSelection = &_totalSelection;
      return;
    case 1L : 
      _basicSelFun = function1; 
      return;
    case 2L : 
      _basicSelFun = function2;     
      return;
    case 3L : 
      _basicSelFun = function3; 
      return;
    case 4L : 
      _basicSelFun = function4; 
      return;
    case 5L : 
      _basicSelFun = function5;
      return;
    case 6L : 
      _basicSelFun = function6; 
      return;
    case 7L : 
      _basicSelFun = function7; 
      return;

    case 8L : // simulates 1L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maximalSelection;
      return;

    case 9L : // simulates 2L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeFilter);
      return;

    case 10L : // simulates 4L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneBigNegEqOrOneBigFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability1);
      return;

    case 11L : // simulates 6L      
      _basicSelFun = customBasicSelFun;
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundFilter);
      return;

    case 12L : // simulates 7L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability2);
      return;

    case 13L : // experimental
      _basicSelFun = customBasicSelFun;
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1);
      return;


    case 14L : // complete analogue of 1007L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_completeBidirectSelection;
      _completeBidirectSelection.setFilterForPositiveInRules(&_oneOfMaxSizeFilter);
      _completeBidirectSelection.setFilterForNegativeInRules(&_oneOfMaxSizeFilter);
      _completeBidirectSelection.setFilterForNegativeFacts(&_oneOfMaxSizeFilter);
      return;

    case 15L : // complete analogue of 1009L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_completeBidirectSelection;
      _completeBidirectSelection.setFilterForPositiveInRules(&_complexFilter5);
      _completeBidirectSelection.setFilterForNegativeInRules(&_complexFilter5);
      _completeBidirectSelection.setFilterForNegativeFacts(&_complexFilter5);
      return;






    case 1002L :
      _basicSelFun = function1002; 
      return;
    case 1003L :
      _basicSelFun = function1003; 
      return;
    case 1004L :
      _basicSelFun = function1004; 
      return;
    case 1005L :
      _basicSelFun = function1005; 
      return;
    case 1006L :
      _basicSelFun = function1006; 
      return;
    case 1007L :
      _basicSelFun = function1007; 
      return;
    case 1008L :
      _basicSelFun = function1008; 
      return;    
    case 1009L :
      _basicSelFun = function1009; 
      return;          



    case 1010L : // simulates 1002L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter1);
      return;

    case 1011L : // simulates 1003L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter2);
      return;

    case 1012L : // simulates 1005L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter4);
      return;

    
    case 1013L : // simulates 1007L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_incompleteBidirectSelection;
      _incompleteBidirectSelection.setFilterForPositiveInRules(&_oneOfMaxSizeFilter);
      _incompleteBidirectSelection.setFilterForNegativeInRules(&_oneOfMaxSizeFilter);
      _incompleteBidirectSelection.setFilterForPositiveFacts(&_oneOfMaxSizeFilter);
      _incompleteBidirectSelection.setFilterForNegativeFacts(&_oneOfMaxSizeFilter);
      return;

    case 1014L : // simulates 1009L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_incompleteBidirectSelection;
      _incompleteBidirectSelection.setFilterForPositiveInRules(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForNegativeInRules(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForPositiveFacts(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForNegativeFacts(&_complexFilter5);
      return;




     

    case -2L : 
      _basicSelFun = function2;
      return;
    case -3L : 
      _basicSelFun = function3;
      return;
    case -4L : 
      _basicSelFun = function4;
      return;
    case -5L : 
      _basicSelFun = function5;
      return;
    case -6L : 
      _basicSelFun = function6;
      return;
    case -7L : 
      _basicSelFun = function7;
      return;


    case -9L : // simulates -2   
      _basicSelFun = customBasicSelFun;
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeFilter);
      return;


    case -10L : // simulates -4       
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneBigNegEqOrOneBigFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability1);
      return;


    case -11L : // simulates -6L      
      _basicSelFun = customBasicSelFun;
      _customSelection = &_hyperSelection;
      _hyperSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundFilter);
      return;

    case -12L : // simulates -7L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_maxWithCorrectionSelection;
      _maxWithCorrectionSelection.setFilterForNegative(&_oneOfMaxSizeMaxNongroundFilter); 
      _maxWithCorrectionSelection.setMaxSelectionAcceptability(&_maxSelectionAcceptability2);
      return;

    case -13L : // experimental
      _basicSelFun = customBasicSelFun;
      _customSelection = &_bestOnlySelection;
      _bestOnlySelection.setLiteralQualityOrdering(&_complexLiteralOrdering1);
      return;

    case -14L : // complete analogue of 1007L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_completeBidirectSelection;
      _completeBidirectSelection.setFilterForPositiveInRules(&_oneOfMaxSizeFilter);
      _completeBidirectSelection.setFilterForNegativeInRules(&_oneOfMaxSizeFilter);
      _completeBidirectSelection.setFilterForNegativeFacts(&_oneOfMaxSizeFilter);
      return;

    case -15L : // complete analogue of -1009L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_completeBidirectSelection;
      _completeBidirectSelection.setFilterForPositiveInRules(&_complexFilter5);
      _completeBidirectSelection.setFilterForNegativeInRules(&_complexFilter5);
      _completeBidirectSelection.setFilterForNegativeFacts(&_complexFilter5);
      return;





    case -1005L :
      _basicSelFun = function1005; 
      return;
    case -1009L :
      _basicSelFun = function1009; 
      return;



    case -1012L : // simulates -1005L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_arbitrarySelection;
      _arbitrarySelection.setFilter(&_complexFilter4);
      return;

    case -1014L : // simulates -1009L
      _basicSelFun = customBasicSelFun;
      _customSelection = &_incompleteBidirectSelection;
      _incompleteBidirectSelection.setFilterForPositiveInRules(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForNegativeInRules(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForPositiveFacts(&_complexFilter5);
      _incompleteBidirectSelection.setFilterForNegativeFacts(&_complexFilter5);
      return;



    default: BK::RuntimeError::report("Wrong selection number.");
    };
}; // void LiteralSelection::setFunNum(long num)


}; // namespace VK


//======================================================================
#endif
