/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         LiteralSharing.hpp
// Description:  Implements literal sharing.
// Created:      Dec 13, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Note:         Was LITERAL_INDEX in clause_index.h.
// Revised:      Dec 25, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Freed from HellConst::MaxNumOfFunctors.
//============================================================================
#ifndef LITERAL_SHARING_H
//=============================================================================
#define LITERAL_SHARING_H 
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
#include "Term.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_LITERAL_SHARING
 #define DEBUG_NAMESPACE "LiteralSharing"
#endif
#include "debugMacros.hpp"
//============================================================================
namespace VK
{
class Flatterm;
class DTREE_NODE;
class SHARING_INDEX;
class LiteralSharing
{
 public:
  LiteralSharing(SHARING_INDEX* si);
  ~LiteralSharing();
  void init(SHARING_INDEX* si);
  void destroy(); 
  void reset();

  DTREE_NODE** Tree(const TERM& header)
  {
   CALL("Tree(const TERM& header)");
   DOP(_discTree.unfreeze());
   DTREE_NODE** res = &(_discTree.sub(header.HeaderNum()));
   DOP(_discTree.freeze());
   return res;
  };

  DTREE_NODE* Integrate(TERM* lit); // returns the leaf
 
  DTREE_NODE* Integrate(const Flatterm* lit); // returns the leaf 
  DTREE_NODE* IntegrateEq(const TERM& header,
                          const Flatterm* arg1,
                          const Flatterm* arg2); // returns the leaf 


  DTREE_NODE* FindLeaf(TERM* lit) // returns the leaf
  {
   // temporary version
   return Integrate(lit); 
  };

  void InfoForDeletion(TERM* lit,
                       DTREE_NODE**& branch_to_destroy,
                       DTREE_NODE*& leaf,
                       DTREE_NODE*& last_fork); // for nonpropositional literals


 private:
 
  static void DestroyLeafContent(DTREE_NODE* leaf);
  TERM* ShareSplittingBranchName(const TERM& prop);
  TERM* ShareProp(const TERM& prop);

 private:

  BK::Array<BK::GlobAlloc,TERM*,128UL,LiteralSharing> _propositional;  

  SHARING_INDEX* _sharingIndex;
  BK::Array<BK::GlobAlloc,DTREE_NODE*,128UL,LiteralSharing> _discTree;
}; // class LiteralSharing

}; // namespace VK
//============================================================================
#endif
