/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         Master.cpp
// Description:  Main loop.
// Created:      Mar 5, 2000, 20:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Master.h"
#include "Clause.h"
#include "ClauseBackground.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
#define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.h"
//====================================================


namespace VK
{
  using namespace BK;
  using namespace Gematogen;


void Master::endOfInput()
{
  CALL("endOfInput()");
  // destroy obsolete clauses
  ASSERT(_retainedNewClauses.empty());
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());

  disposeObsolete(); 

  ASSERT(_queueForDisposal.empty());

}; // void Master::endOfInput()

void Master::mainLoop0()
{
  CALL("mainLoop0()");
  timer.reset();
  timer.start();
  _timerIsOn = true;
  _externalTickFlag = false;
  _startMemory = (float)GlobAlloc::occupiedByObjects();  
  terminated = false;  
  proof = (Clause*)0;

  Clause* current;
 select_current: 

  PROGRESSCHECK("<select current> in Master::mainLoop0()",10.0);
  current = select0(); 
  if (!current) 
    { 
      terminate(EmptyPassive);
      timer.stop();
      _timerIsOn = false;
      return; 
    };
  if (_active.Total() >= _parameters.maxActive) 
    { 
      terminate(MaxActive); 
      timer.stop();
      _timerIsOn = false;
      return; 
    }; 

  if ((_parameters.orphan_murder)
      && (!current->IsVIP())
      && (current->IsOrphan()))
    {      
      _statPassiveOrphansMurdered++;
      disposeClause(current);  
      estimateResources();
      goto select_current;
    }; 
  
  makeActive0(current); 

  eliminateAllocationBufferDeficit0();  

  _inferenceDispatcherSemaphore = true;
  _infDispatcher.newQuery(current);
  while (_infDispatcher.nextInference())
    {
      PROGRESSCHECK("<next inference> in Master::mainLoop0()",3.0);
      processCollected0();
      if (proof) 
	{ 
	  terminate(Proof); 
	  _inferenceDispatcherSemaphore = false; 
	  timer.stop();
	  _timerIsOn = false;
	  return; 
	};
  
      if (_passive.Total() >= _parameters.maxPassive) 
	{ 
	  terminate(MaxPassive);  
	  _inferenceDispatcherSemaphore = false; 
	  timer.stop();
	  _timerIsOn = false;
	  return; 
	};
      if (terminated) 
	{ 
	  _inferenceDispatcherSemaphore = false; 
	  timer.stop();
	  _timerIsOn = false;
	  return; 
	};

      internalTick0();
 
      backSimplify();
      if (terminated) 
	{ 
	  _inferenceDispatcherSemaphore = false; 
	  timer.stop();
	  _timerIsOn = false;
	  return; 
	};
      integrateIntoBackDemIdexFromQueue();
    };
  // no more inferences with current
  _inferenceDispatcherSemaphore = false;
  
  internalTick0(); 
    

  // destroy obsolete clauses
  ASSERT(_retainedNewClauses.empty());
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());

  disposeObsolete(); 
  ASSERT(_queueForDisposal.empty());
  estimateResources();
  if ((_parameters.lrs == 2) || (_parameters.lrs == 3))
    {
      recycleUnreachablePassive0();
    };  
  goto select_current; 
}; // void Master::mainLoop0()

void Master::backSimplify()
{
  CALL("backSimplify()");
  ASSERT((_parameters.backward_demodulation > 0) || _queueForBackSimplification.empty());
  _backwardDemodulationSemaphore = true;
  Clause* currentSimplifier;
  Clause* backSimplified;
  bool usedSimplified = false;
  if (_parameters.backward_demodulation > 0)
    {
      while (_queueForBackSimplification.dequeue(currentSimplifier))
	{ 
	  ASSERT(currentSimplifier->flag(Clause::FlagInQueueForBackSimplification));
	  currentSimplifier->clearFlag(Clause::FlagInQueueForBackSimplification);
	  if (backDemodulationQuery(currentSimplifier))
	    { 
	      while (nextBackDemodulated())
		{         
		  _statBackSimplified++;
         
		  if (collectBackSimplified())
		    {   
		      //currentSimplifier->MarkAsSimplifier();
		      currentSimplifier->setFlag(Clause::FlagHasChildren);
		      backSimplified = _backDemodulation.simplified();
		      usedSimplified = usedSimplified 
			|| (backSimplified->mainSet() == Clause::MainSetActive) 
			|| (backSimplified->mainSet() == Clause::MainSetDefinitions); 

		      backSimplified->setFlag(Clause::FlagHasChildren);

		      enqueueForDisposal(backSimplified);        
		      processCollected0(); // not before enqueueForDisposal(backSimplified)!
		      if (proof) { terminate(Proof); _backwardDemodulationSemaphore = false; return; };
		      if (_passive.Total() >= _parameters.maxPassive) 
			{ 
			  terminate(MaxPassive); 
			  _backwardDemodulationSemaphore = false; 
			  return; 
			}; 
		    }
		  else
		    if (_freshClause.WasTautology())
		      {
			backSimplified = _backDemodulation.simplified();

			enqueueForDisposal(backSimplified);
		      };
		  internalTick0();
		};  
	    };
	};
      if (_parameters.orphan_murder && usedSimplified) killActiveOrphans0();
    };
  _backwardDemodulationSemaphore = false;
}; // void Master::backSimplify()


void Master::eliminateDefinitions()
{
  CALL("Master::eliminateDefinitions()");
  ASSERT(!_parameters.main_alg);
  timer.reset();
  timer.start();
  Clause* cl;
 select:
  _passive.ResetIterator();
 try_next:
  cl = _passive.IterNext();
  if (cl) 
    {
      if (cl->ContainsEliminated(_parameters.pred_elim_table))
	{
	  _passive.Remove(cl);
      
	  // perform resolutions
	  makeActive0(cl);
	  _inferenceDispatcherSemaphore = true;
	  _infDispatcher.newQuery(cl);     
	  while (_infDispatcher.nextInference())
	    {
	      _freshClause.UsedRule(ClauseBackground::Prepro);
	      processCollectedInput0();
	      if (proof) 
		{ 
		  terminate(Proof); 
		  _inferenceDispatcherSemaphore = false; 
		  goto collect_results; 
		};
	      backSimplify();
	      if (terminated) 
		{ 
		  _inferenceDispatcherSemaphore = false; 
		  goto collect_results; 
		};
	    };
	  _inferenceDispatcherSemaphore = false;
	  // destroy obsolete clauses
	  ASSERT(_retainedNewClauses.empty());
	  ASSERT(_queueForBackSimplification.empty());
	  ASSERT(_queueForIntegrationIntoBackwardDemodIndex.empty());
	  disposeObsolete(); 

	  // use next
	  goto select;     
	}
      else goto try_next;      
    };

  // all done, collect results
 collect_results: 
  ASSERT(_retainedNewClauses.empty());
  ASSERT(proof || (_queueForBackSimplification.empty() && _queueForIntegrationIntoBackwardDemodIndex.empty()));
  _queueForBackSimplification.reset();
  _queueForIntegrationIntoBackwardDemodIndex.reset();

  // all _active are obsolete now, make them _history
  Clause* act;
  while (_active.Nonempty())
    {
      act = _active.Top();  
      enqueueForDisposal(act);
    };
  disposeObsolete();    
  timer.stop();
  _timerIsOn = false;
}; // void Master::eliminateDefinitions()

void Master::inputClause0(Clause* cl)
{
  CALL("inputClause0(Clause* cl)");
  timer.reset();
  timer.start();
  _timerIsOn = true;
  terminated = false;  
  proof = (Clause*)0; 
  if (_freshClause.LoadClause(cl))
    {
      _freshClause.PushAncestor(cl); 
      _freshClause.UsedRule(ClauseBackground::Prepro); 
      _passive.MaxPriorityOn(); 
      processCollectedInput0(); 
      _passive.MaxPriorityOff(); 
      if (proof) 
	{  
	  terminate(Proof); 
	  ASSERT(_retainedNewClauses.empty());    
	  disposeObsolete();  
	  timer.stop();
	  _timerIsOn = false;
	  return; 
	}; 
      backSimplify(); 
      integrateIntoBackDemIdexFromQueue();
    };
  timer.stop();
  _timerIsOn = false;
}; // void Master::inputClause0(Clause* cl)

void Master::processCollected0()
{
  CALL("processCollected0()");

  if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';         
  _freshClause.ProcessCollected(_retainedNewClauses);


  Clause* newClause;  
  while (_retainedNewClauses.dequeue(newClause))
    { 

      ASSERT(newClause->mainSet() == Clause::MainSetNew);
      makePassive0(newClause);
      if (newClause->isAnswer()) 
	{
	  proof = newClause; 
	  return; 
	}; 

      if (newClause->weight() > _parameters.maxWeight)
	{        

	  enqueueForDisposal(newClause);
	  _statRecycled++;

	}
      else
	{
	  if (_parameters.backward_subsumption) 
	    if (backSubsumeBy(newClause))
	      if (_parameters.orphan_murder) 
		killActiveOrphans0();

	  if (_parameters.backward_demodulation > 0) 
	    {
	      ASSERT(!newClause->flag(Clause::FlagInQueueForBackSimplification));

	      _queueForBackSimplification.enqueue(newClause);

	      newClause->setFlag(Clause::FlagInQueueForBackSimplification);
	    };

	  integratePassive(newClause);

	  eliminateAllocationBufferDeficit0();
	};
    };
}; // void Master::processCollected0()

void Master::processCollectedInput0()
{
  CALL("processCollectedInput0()");

  ASSERT_IN(_backSubsumption._termTraversal.checkIntegrity(),"COP1");

  if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
  _freshClause.ProcessCollected(_retainedNewClauses);
  Clause* newClause;  
  while (_retainedNewClauses.dequeue(newClause))
    { 
      ASSERT(newClause->mainSet() == Clause::MainSetNew);

      makePassive0(newClause); 
      if (newClause->isAnswer()) 
	{
	  proof = newClause; 
	}
      else 
	{
	  if (_parameters.backward_subsumption) backSubsumeBy(newClause);
	  if (_parameters.backward_demodulation > 0) 
	    {
	      ASSERT(!newClause->flag(Clause::FlagInQueueForBackSimplification));
	      _queueForBackSimplification.enqueue(newClause);
	      newClause->setFlag(Clause::FlagInQueueForBackSimplification);
	    }; 
	  integratePassive(newClause);

	}; 
    };
  ASSERT(_retainedNewClauses.empty());
}; // void Master::processCollectedInput0()

Clause* Master::select0()
{
  CALL("select0()");
  Clause* selected;
 try_next_passive:
  selected = _passive.Next();
  if (!selected)
    {// _passive is empty, try _reservedPassive 
      _parameters.lrs = 0;
      _parameters.maxWeight = VampireKernelConst::MaxClauseSize;  
      _freshClause.SetMaxWeight(_parameters.maxWeight);  
      if ((_parameters.showWeightChange) 
	  && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
	_parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight());    
      loadReservedClauses();    
      selected = _passive.Next();
      if (selected) 
	{
	  ASSERT(selected->mainSet() == Clause::MainSetPassive);
	  selected->setMainSet(Clause::MainSetFree);
	};  
      return selected;
    };
  if (selected->MinResWeight() > _parameters.maxWeight)
    { 
      makeReserved(selected); 
      goto try_next_passive;         
    };
  if (selected)     
    {
      ASSERT(selected->mainSet() == Clause::MainSetPassive);
      selected->setMainSet(Clause::MainSetFree);
    }; 
  return selected; 
}; // Clause* Master::select0()  

void Master::makeActive0(Clause* cl)
{
  CALL("makeActive0(Clause* cl)");
  //cl->SetStore(CS_USED);  
  cl->setMainSet(Clause::MainSetActive);
  cl->setFlag(Clause::FlagHasChildren); // too memory inefficient, must be changed 
  _active.Insert(cl);
  _infDispatcher.makeActive(cl);
  if (_parameters.showActive) _parameters.showActive(cl); 
}; // void Master::makeActive0(Clause* cl)

void Master::makePassive0(Clause* cl)
{
  CALL("makePassive0(Clause* cl)");
  cl->SetNumber(_clauseNumberGenerator->generate());
  INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTRODUCE_CLAUSE_ID(cl);
  cl->setMainSet(Clause::MainSetPassive);
  _passive.Insert(cl);
  if (_parameters.useClauseNumIndex) _clauseNumIndex.insert(cl);
  if (_parameters.showPassive) _parameters.showPassive(cl);
}; // void Master::makePassive0(Clause* cl)

void Master::integratePassive(Clause* cl)
{
  CALL("integratePassive(Clause* cl)");
  // see also removeFromIndices(Clause* cl)

  if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
  if (_parameters.backward_subsumption) integrateIntoBackwardSubsumptionIndex(cl);
  if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
  if (_parameters.backward_demodulation > 0) 
    {
      if (_backwardDemodulationSemaphore)
	{
	  _queueForIntegrationIntoBackwardDemodIndex.enqueue(cl);
	  cl->setFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
	}
      else 
	{
	  cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
	  integrateIntoBackDemodulationIndex(cl);
	};
    };
}; // void Master::integratePassive(Clause* cl) 


void Master::integrateIntoBackDemIdexFromQueue()
{
  CALL("integrateIntoBackDemIdexFromQueue()");
  ASSERT(!_backwardDemodulationSemaphore);
  Clause* cl;
  while (_queueForIntegrationIntoBackwardDemodIndex.dequeue(cl))
    { 
      ASSERT(cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex));
      ASSERT(cl->mainSet() == Clause::MainSetPassive);
      cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
      integrateIntoBackDemodulationIndex(cl);
    };
}; // void Master::integrateIntoBackDemIdexFromQueue()

void Master::removeFromIndices(Clause* cl)
{
  CALL("removeFromIndices(Clause* cl)");
  // see also integratePassive(Clause* cl)
  if (cl->flag(Clause::FlagInBackDemodulationIndex)) removeFromBackDemodulationIndex(cl);
  if (cl->flag(Clause::FlagInForwardDemodulationIndex)) removeFromForwardDemodulationIndex(cl);
  if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) removeFromBackwardSubsumptionIndex(cl);
  if (cl->flag(Clause::FlagInForwardSubsumptionIndex)) removeFromForwardSubsumptionIndex(cl);
}; // void Master::removeFromIndices(Clause* cl)

void Master::makeReserved(Clause* cl)
{
  CALL("makeReserved(Clause* cl)");
  cl->setMainSet(Clause::MainSetReservedPassive);
  _reservedPassive.Insert(cl);
}; // void Master::makeReserved(Clause* cl)

void Master::enqueueForDisposal(Clause* cl)
{
  CALL("enqueueForDisposal(Clause* cl)"); 
  if (cl->mainSet() == Clause::MainSetTrash) return;
  switch (cl->mainSet()) 
    {
    case Clause::MainSetFree: break;
    case Clause::MainSetNew:
      ASSERT(_parameters.main_alg == 3);
      ASSERT(cl == proof);
      break;
    case Clause::MainSetPassive: _passive.Remove(cl); break;
    case Clause::MainSetReservedPassive: _reservedPassive.Remove(cl); break;    
    case Clause::MainSetActive: _active.Remove(cl); break;
#ifdef DEBUG_NAMESPACE
    default: 
      ICP("ICP1"); 
      return;
#else
    default: 
      RuntimeError::report("Wrong clause affiliation.");
#endif
    };

 
  if (cl->flag(Clause::FlagInQueueForBackSimplification)) 
    {
      _queueForBackSimplification.removeFirst(cl); 
      cl->clearFlag(Clause::FlagInQueueForBackSimplification);
    };
  if (cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex)) 
    {
      _queueForIntegrationIntoBackwardDemodIndex.removeFirst(cl); 
      cl->clearFlag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex);
    };

  bool canBeDestroyed = true;
  if (!_inferenceDispatcherSemaphore)
    {
      if (cl->flag(Clause::FlagInInferenceDispatcherAsActive)) _infDispatcher.remove(cl);
      if (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)) _infDispatcher.removeDefinition(cl);
    }
  else
    {
      if ((cl->flag(Clause::FlagInInferenceDispatcherAsActive)) 
	  || (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)))
	canBeDestroyed = false;
    };

  if (!_backwardDemodulationSemaphore)
    {
      if (cl->flag(Clause::FlagInBackDemodulationIndex)) removeFromBackDemodulationIndex(cl); 
    }
  else
    if (cl->flag(Clause::FlagInBackDemodulationIndex)) canBeDestroyed = false;

  if (cl->flag(Clause::FlagInForwardDemodulationIndex)) removeFromForwardDemodulationIndex(cl);

  if (!_backwardSubsumptionSemaphore)
    {
      if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) removeFromBackwardSubsumptionIndex(cl);
    }
  else if (cl->flag(Clause::FlagInBackwardSubsumptionIndex)) canBeDestroyed = false;

  if (cl->flag(Clause::FlagInForwardSubsumptionIndex)) removeFromForwardSubsumptionIndex(cl);
 
  if (canBeDestroyed)
    {
      if (cl->flag(Clause::FlagHasChildren))
	{

	  // must be in history
	  if (cl->flag(Clause::FlagInClauseIndex)) 
	    {
	      //_clauseAssembler.removeFromSharing(cl,false);  
	      cl->clearFlag(Clause::FlagInClauseIndex);
	    };
	  cl->setMainSet(Clause::MainSetHistory); 
	  _history.Insert(cl); 
	}
      else // can be completely destroyed
	{
      

	  bool shared = cl->flag(Clause::FlagLiteralsShared);
	  if (cl->flag(Clause::FlagInClauseIndex)) 
	    {
	      if (shared) _clauseAssembler.removeFromSharing(cl);
	      cl->clearFlag(Clause::FlagInClauseIndex);
	      cl->clearFlag(Clause::FlagLiteralsShared);
	    };
	  if (cl->flag(Clause::FlagInClauseNumIndex)) _clauseNumIndex.remove(cl);
	  if (shared) { cl->Destroy(); } else cl->DestroyUnshared();
	};

    }
  else // must wait in the queue
    {
      cl->setMainSet(Clause::MainSetTrash);
      _queueForDisposal.enqueue(cl); 
    };
}; // void Master::enqueueForDisposal(Clause* cl)

void Master::disposeObsolete()
{
  CALL("disposeObsolete()"); 
  Clause* disposedClause;
  while (_queueForDisposal.dequeue(disposedClause))
    { 
      ASSERT(disposedClause->mainSet() == Clause::MainSetTrash);
      disposedClause->setMainSet(Clause::MainSetFree);
      disposeClause(disposedClause);
    };
}; // void Master::disposeObsolete()

void Master::disposeClause(Clause* cl)
{
  CALL("disposeClause(Clause* cl)");
  ASSERT(cl->mainSet() == Clause::MainSetFree);
  ASSERT((!_inferenceDispatcherSemaphore) || ((!cl->flag(Clause::FlagInInferenceDispatcherAsActive)) && (!cl->flag(Clause::FlagInInferenceDispatcherAsDefinition))));
  if (cl->flag(Clause::FlagInInferenceDispatcherAsActive)) _infDispatcher.remove(cl);
  if (cl->flag(Clause::FlagInInferenceDispatcherAsDefinition)) _infDispatcher.removeDefinition(cl);
  removeFromIndices(cl);
  ASSERT((_parameters.main_alg == 2) || (_parameters.main_alg == 3) || (cl->flag(Clause::FlagInClauseIndex)));
  ASSERT((_parameters.main_alg == 2) || (_parameters.main_alg == 3) || (cl->flag(Clause::FlagLiteralsShared)));


  ASSERT(!cl->flag(Clause::FlagInQueueForBackSimplification));
  ASSERT(!cl->flag(Clause::FlagInQueueForIntegrationIntoBackwardDemodIndex));

  if (cl->flag(Clause::FlagHasChildren))
    {
      // must be in history
      if (cl->flag(Clause::FlagInClauseIndex)) 
	{
	  //_clauseAssembler.removeFromSharing(cl,false);  
	  cl->clearFlag(Clause::FlagInClauseIndex);
	};
      cl->setMainSet(Clause::MainSetHistory); 
      _history.Insert(cl); 

    }
  else // can be completely destroyed
    {
      bool shared = cl->flag(Clause::FlagLiteralsShared);
      if (cl->flag(Clause::FlagInClauseIndex)) 
	{
	  if (shared) _clauseAssembler.removeFromSharing(cl);
	  cl->clearFlag(Clause::FlagInClauseIndex);
	  cl->clearFlag(Clause::FlagLiteralsShared);
	};
      if (cl->flag(Clause::FlagInClauseNumIndex)) _clauseNumIndex.remove(cl);
      if (shared) { cl->Destroy(); } else cl->DestroyUnshared();

    };

}; // void Master::disposeClause(Clause* cl)


void Master::loadReservedClauses()
{
  CALL("loadReservedClauses()");
  Clause* cl;
  _reservedPassive.ResetIterator();
  while (!_reservedPassive.Empty())
    {
      cl = _reservedPassive.Pop();
      ASSERT(cl->mainSet() == Clause::MainSetReservedPassive); 
      cl->setMainSet(Clause::MainSetPassive);
      _passive.Insert(cl); 
    };
}; // void Master::loadReservedClauses()

bool Master::collectBackSimplified()
{
  CALL("collectBackSimplified()");
  return _backDemodulation.collectInference();  
}; // bool Master::collectBackSimplified()


bool Master::backSubsumeBy(Clause* cl)
{
  CALL("backSubsumeBy(Clause* cl)");
  _backwardSubsumptionSemaphore = true;
  Clause* subsumedClause;
  bool usedSubsumed = false;
  if (_backSubsumption.query(cl))
    {
      while (_backSubsumption.subsumeNext())
	{ 
	  _statBackSubsumed++;
	  cl->MakeVIP();
	  subsumedClause = (Clause*)_backSubsumption.subsumed();
	  ASSERT(subsumedClause != cl);
	  ASSERT(subsumedClause->Active()); 
	  usedSubsumed = (usedSubsumed 
			  || (subsumedClause->mainSet() == Clause::MainSetActive) 
			  || (subsumedClause->mainSet() == Clause::MainSetDefinitions));     
	  enqueueForDisposal(subsumedClause);     
	};
    };
  _backwardSubsumptionSemaphore = false;
  return (bool)usedSubsumed;
}; // bool Master::backSubsumeBy(Clause* cl)

void Master::estimateResources1()
{
  CALL("estimateResources1()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = timer.elapsedDeciseconds();

  bool noTimeCheck = used_time < _parameters.timeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.timeLimit - used_time)*moved_from_passive)/used_time);

  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable1(_parameters.reachable);
  _parameters.maxWeight = _passive.MaxAcceptedWeight();
  if (_parameters.maxWeight > _parameters.maxMaxWeight) 
    {
      _parameters.maxWeight = _parameters.maxMaxWeight;
    };
  _freshClause.SetMaxWeight(_parameters.maxWeight);
  if ((_parameters.showWeightChange) 
      && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
    _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight()); 

  if (_parameters.showUsagePace) _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources1()



void Master::estimateResources2()
{
  CALL("estimateResources2()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = timer.elapsedDeciseconds();

  bool noTimeCheck = used_time < _parameters.timeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.timeLimit - used_time)*moved_from_passive)/used_time);

  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable2(_parameters.reachable);
  _parameters.maxWeight = _passive.MaxAcceptedWeight();
  if (_parameters.maxWeight > _parameters.maxMaxWeight) 
    {
      _parameters.maxWeight = _parameters.maxMaxWeight;
    };
  _freshClause.SetMaxWeight(_parameters.maxWeight);
  if ((_parameters.showWeightChange) 
      && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
    _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight()); 

  if (_parameters.showUsagePace) _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources2()


void Master::estimateResources3()
{
  CALL("estimateResources3()");
  ASSERT(_parameters.lrs);
  float used_memory = usedMemory();
  float used_time = timer.elapsedDeciseconds();

  bool noTimeCheck = used_time < _parameters.timeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    LONG_MAX 
    : 
    (long)(((_parameters.timeLimit - used_time)*moved_from_passive)/used_time);

  long reachableByMemory = (noMemoryCheck)? 
    LONG_MAX
    :
    (long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory);
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  _passive.SetNumOfReachable3(_parameters.reachable);
  _parameters.maxWeight = _passive.MaxAcceptedWeight();
  if (_parameters.maxWeight > _parameters.maxMaxWeight) 
    {
      _parameters.maxWeight = _parameters.maxMaxWeight;
    };
  _freshClause.SetMaxWeight(_parameters.maxWeight);
  if ((_parameters.showWeightChange) 
      && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
    _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight()); 

  if (_parameters.showUsagePace) _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources3()




void Master::estimateResources4()
{
  CALL("estimateResources4()");
  ASSERT(_parameters.lrs == 4);
  float used_memory = usedMemory();
  float used_time = timer.elapsedDeciseconds();

  bool noTimeCheck = used_time < _parameters.timeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    (_halfLongMax) 
    : 
    ((long)(((_parameters.timeLimit - used_time)*moved_from_passive)/used_time));

  long reachableByMemory = (noMemoryCheck)? 
    (_halfLongMax)
    :
    ((long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory));
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  long currentlyUnreachable = _passive.Size() - _parameters.reachable;

  if (currentlyUnreachable <= 0)
    { 
      if ((currentlyUnreachable <= _parameters.unreachable) && (_parameters.maxWeight < ((long)VampireKernelConst::MaxClauseSize)))
	{ 
	  // weight limit increases 
	  _parameters.maxWeight++;
	  if (_parameters.maxWeight > _parameters.maxMaxWeight) 
	    {
	      _parameters.maxWeight = _parameters.maxMaxWeight;
	    };
	};
    }
  else // currentlyUnreachable > 0
    { 
      if (currentlyUnreachable >= _parameters.unreachable)
	{
	  // weight limit decreases 
	  if (_parameters.maxWeight > _passive.MaxPresentWeight()) 
	    {
	      _parameters.maxWeight = _passive.MaxPresentWeight(); 
	    }       
	  else 
	    if (_parameters.maxWeight > _passive.MinPresentWeight()) 
	      { 
		_parameters.maxWeight--;
	      };
	};
    };

  _parameters.unreachable = currentlyUnreachable;

  _freshClause.SetMaxWeight(_parameters.maxWeight);
  if ((_parameters.showWeightChange) 
      && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
    _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight()); 
  if (_parameters.showUsagePace) _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources4()



void Master::estimateResources5()
{
  CALL("estimateResources5()");
  ASSERT(_parameters.lrs == 5);
  float used_memory = usedMemory();
  float used_time = timer.elapsedDeciseconds();

  bool noTimeCheck = used_time < _parameters.timeLimit * _parameters.first_time_check;
  bool noMemoryCheck = used_memory < _parameters.memoryLimit * _parameters.first_mem_check;

  if (noTimeCheck && noMemoryCheck) return;

  long moved_from_passive = _passive.UsedAndRemoved();   
  long reachableByTime = (noTimeCheck)? 
    (_halfLongMax) 
    : 
    ((long)(((_parameters.timeLimit - used_time)*moved_from_passive)/used_time));

  long reachableByMemory = (noMemoryCheck)? 
    (_halfLongMax)
    :
    ((long)(((_parameters.memoryLimit - used_memory)*moved_from_passive)/used_memory));
  _parameters.reachable = (reachableByTime < reachableByMemory) ? reachableByTime : reachableByMemory;
  long currentlyUnreachable = _passive.Size() - _parameters.reachable;

  if (currentlyUnreachable <= 0)
    { 
      if ((currentlyUnreachable <= _parameters.unreachable) && (_parameters.maxWeight < ((long)VampireKernelConst::MaxClauseSize)))
	{ 
	  // weight limit increases 
	  _parameters.maxWeight++;
	  if (_parameters.maxWeight > _parameters.maxMaxWeight) 
	    {
	      _parameters.maxWeight = _parameters.maxMaxWeight;
	    };
	};
    }
  else // currentlyUnreachable > 0
    { 
      if (currentlyUnreachable >= _parameters.unreachable)
	{
	  // weight limit decreases 
	  if (_parameters.maxWeight > _passive.MaxPresentWeight()) 
	    {
	      _parameters.maxWeight = _passive.MaxPresentWeight(); 
	    }       
	  else 
	    if (_parameters.maxWeight >= _passive.MinPresentWeight())
	      { 
		_parameters.maxWeight--;
	      };
	};
    };

  _parameters.unreachable = currentlyUnreachable;

  _freshClause.SetMaxWeight(_parameters.maxWeight);
  if ((_parameters.showWeightChange) 
      && (_freshClause.PreviousMaxWeight() != _freshClause.MaxWeight()))
    _parameters.showWeightChange(_freshClause.PreviousMaxWeight(),_freshClause.MaxWeight()); 
  if (_parameters.showUsagePace) _parameters.showUsagePace(moved_from_passive/used_time);
 
}; // void Master::estimateResources5()


bool Master::completeStrategyUsed()
{
  return !(_parameters.incompleteAPriori
	   || statTooHeavy() 
	   || statTooHeavyLiterals() 
	   || statTooManyVariables()
	   || statTooLong()
	   || statTooDeep()
	   || statRecycled()
	   || statRefusedAllocation()
	   || (_parameters.selection >= 1000)
	   || (_parameters.selection <= -1000));      
}; // bool Master::completeStrategyUsed()


void Master::recycleUnreachablePassive0()
{
  CALL("recycleUnreachablePassive0()");
  Clause* cl;
 recycle_next:
  if (_passive.UnreachableKept() <= (_passive.CurrentlyKept()/5)) 
    return; // nothing to recycle
  cl = _passive.TheMostUnreachableClause();    
  if (!cl) return;  
  ASSERT(cl->mainSet() == Clause::MainSetPassive);
  _passive.Cancel(cl);
  cl->setMainSet(Clause::MainSetFree);
  enqueueForDisposal(cl);
  _statRecycled++;
  goto recycle_next;
}; // void Master::recycleUnreachablePassive0()


void Master::killActiveOrphans0()
{
  CALL("killActiveOrphans0()");
  _active.ResetIterator();
  for (Clause* cl = _active.Next(); cl; cl = _active.Next())
    if ((cl->Active())
	&& (!cl->IsVIP())
	&& (cl->IsOrphan()))
      {
	enqueueForDisposal(cl);
	_statActiveOrphansMurdered++;
      };
}; // void Master::killActiveOrphans0()

void Master::destroyPassive0() 
{ 
  CALL("destroyPassive0()");
  _passive.DestroyAllClauses();
 
  _passive.ResetIterator();
  for (Clause* cl = _passive.IterNext(); cl; cl = _passive.IterNext())
    enqueueForDisposal(cl);
}; // void Master::destroyPassive0() 

void Master::destroy()
{
  CALL("destroy()");

  // sets of persistent clauses
  switch (_parameters.main_alg)
    {
    case 0: destroyPassive0(); break;
    case 3: destroyPassive3(); break;
    default: ICP("ICP0"); break; 
    }; 

  while (_reservedPassive.Nonempty())
    enqueueForDisposal(_reservedPassive.Pop());
  _reservedPassive.Purge(); // to make it clean()

  while (_active.Nonempty())
    {
      enqueueForDisposal(_active.Pop());
    };


  _active.Purge(); // to make it clean()


  // queues
  Clause* retainedNewClause;

  while (_retainedNewClauses.dequeue(retainedNewClause)) 
    { 

      ASSERT(retainedNewClause->mainSet() == Clause::MainSetNew);
      retainedNewClause->setMainSet(Clause::MainSetFree);
      disposeClause(retainedNewClause);  
    };
  if (proof &&  (_parameters.main_alg == 3))
    { 
      enqueueForDisposal(proof); 
      proof = (Clause*)0; 
    };

  // queue for disposal 
  // not after _history.DestroyAllClauses()! 
  disposeObsolete();
  _history.DestroyAllClauses();  // not before disposeObsolete()!

  _queueForBackSimplification.reset();
  _queueForIntegrationIntoBackwardDemodIndex.reset();

  // indices and inference engines   


  _infDispatcher.destroy();
  _forwardSubsumption.Destroy();
  _backSubsumption.destroyIndex();
  _forwardDemodulation.Destroy();
  _backDemodulation.destroy();
  _sharingIndex.Destroy(); 
  _clauseAssembler.destroy();    
  _clauseNumIndex.reset();
}; // void Master::destroy()

void Master::reset() 
{
  CALL("reset()");
  _unifier.reset();
  Unifier::setCurrent(&_unifier);
  destroy();
  options.reset(); 
  resetStatistics();
  _backDemodulation.resetStat(); 
  ASSERT(clean());    
}; //  void Master::reset() 


}; // namespace VK

//====================================================
