/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         Master3.cpp
// Description:  Main loop implementing main_alg 3
// Created:      Apr 29, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      May 20, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Too strong an assertion removed in processCollectedInput3().
//====================================================
#include "Master.h"
#include "GlobalStopFlag.h"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_MASTER
 #define DEBUG_NAMESPACE "Master"
#endif                  
#include "debugMacros.h"
//====================================================

namespace VK
{
  using namespace BK;
  using namespace Gematogen;
void Master::inputClause3(Clause* cl)
{
 CALL("inputClause3(Clause* cl)");
 timer.reset();
 timer.start();
 _timerIsOn = true;
 terminated = false;  
 proof = (Clause*)0; 

 if (_freshClause.LoadClause(cl))
  {
   _freshClause.PushAncestor(cl);    
   _freshClause.UsedRule(ClauseBackground::Prepro);
   _pagedPassive.maximalPriorityOn();
   processCollectedInput3();
   _pagedPassive.maximalPriorityOff();
   if (proof) 
    { 
     terminate(Proof);
     ASSERT(_retainedNewClauses.empty());    
     disposeObsolete(); 
     timer.stop();
     _timerIsOn = false; 
     return; 
    };
  };

}; // void Master::inputClause3(Clause* cl)

void Master::processCollectedInput3()
{
  CALL("processCollectedInput3()");
  if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';

  _freshClause.ProcessCollected2(_retainedNewClauses);
  Clause* newClause;  


  while (_retainedNewClauses.dequeue(newClause))
    { 
      ASSERT(newClause->mainSet() == Clause::MainSetNew);
      newClause->SetNumber(_clauseNumberGenerator->generate());

      INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_INTRODUCE_CLAUSE_ID(newClause);  

      if (_parameters.showPassive) _parameters.showPassive(newClause);    
 
      if (newClause->isAnswer()) 
	{
	  proof = newClause;
	}
      else 
	{   
	  newClause->setMainSet(Clause::MainSetFree);  
	  if (!insertIntoPassive3(newClause)) ++_statClausesRefusedAllocation;
	  disposeClause(newClause);      
	};
    };
  ASSERT(_retainedNewClauses.empty());
}; // void Master::processCollectedInput3()


bool Master::insertIntoPassive3(Clause* cl)
{
 CALL("Master::insertIntoPassive3(Clause* cl)");
 ASSERT(_parameters.main_alg == 3);
 ulong max_weight = _freshClause.MaxWeight();
 _freshClause.SetMaxWeight(VampireKernelConst::MaxClauseSize); 
 
 ALWAYS(_freshClause.LoadClause(cl));
 _freshClause.setClauseNumber(cl->Number());
      
 for (ClauseBackground::ClauseList* anc = cl->background().ancestors(); anc; anc = anc->tl())
  _freshClause.PushAncestor(anc->hd()); 

 _freshClause.InverseAncestorOrder();

 _freshClause.UsedRules(cl->background().properties());

 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';
  
 
 bool res = _freshClause.transferToPagedPassive(_pagedPassive,(bool)_parameters.showPassive);  
 
 _freshClause.SetMaxWeight(max_weight);

 return res;
}; // bool Master::insertIntoPassive3(Clause* cl)



void Master::mainLoop3()
{
 CALL("mainLoop3()");
 timer.reset();
 timer.start();
 _timerIsOn = true;
 _startMemory = (float)GlobAlloc::occupiedByObjects(); 
 terminated = false;  
 proof = (Clause*)0;

 endOfInput();
 if (terminated) 
   {
     timer.stop();
     _timerIsOn = false; 
     return;
   };
 
 PagedPassiveClauseStorage::StoredClause selectedClause;
 Clause* reanimated;
 select_current: 
  PROGRESSCHECK("<select current> in Master::mainLoop0()",10.0);
  if (!_pagedPassive.select(selectedClause)) 
    { 
      terminate(EmptyPassive); 
      timer.stop();
      _timerIsOn = false; 
      return; 
    };
  if (_active.Total() >= _parameters.maxActive) 
    { 
      terminate(MaxActive); 
      timer.stop();
      _timerIsOn = false; 
      return; 
    };


  if ((_parameters.orphan_murder)
      && (!selectedClause.isVIP())
      && (selectedClause.isOrphan()))
   {      
    _statPassiveOrphansMurdered++;
    _pagedPassive.popSelected();
    goto select_current;
   };


  reanimated = reanimate3(selectedClause);
  
  // _pagedPassive.popSelected() is done in reanimate3(selectedClause)

  if (!reanimated) { GlobalStopFlag::check(); goto select_current; };
  if (reanimated->isAnswer()) 
    { 
      proof = reanimated; 
      terminate(Proof);  
      timer.stop();
      _timerIsOn = false; 
      return; 
    };
  if (_parameters.backward_subsumption) 
   if (backSubsumeBy(reanimated)) 
    if (_parameters.orphan_murder) killActiveOrphans3();      
    
  disposeObsolete();
  

  backSimplifyBy3(reanimated);
  if (proof) 
    {
      makeActive3(reanimated); // to make it available for destruction
      terminate(Proof);  
      timer.stop();
      _timerIsOn = false; 
      return; 
    };    
  disposeObsolete();  
  makeActive3(reanimated);

  _inferenceDispatcherSemaphore = true;  
  _infDispatcher.newQuery(reanimated);
  while (_infDispatcher.nextInference())
   {
    PROGRESSCHECK("<next inference> in Master::mainLoop0()",3.0);
    processCollected3();
    if (proof) 
     { 
      terminate(Proof); 
      _inferenceDispatcherSemaphore = false;  
      timer.stop();
      _timerIsOn = false; 
      return; 
     };
  
    if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
     { 
       terminate(MaxPassive); 
       _inferenceDispatcherSemaphore = false; 
       timer.stop();
       _timerIsOn = false; 
       return; 
     };
    if (terminated) 
      { 
	_inferenceDispatcherSemaphore = false; 
	timer.stop();
	_timerIsOn = false; 
	return; 
      };

    if (terminated) 
      { 
	_inferenceDispatcherSemaphore = false; 
	timer.stop();
	_timerIsOn = false; 
	return; 
      };
    GlobalStopFlag::check();
   };
  _inferenceDispatcherSemaphore = false;
 

  GlobalStopFlag::check();

  // destroy obsolete clauses
  ASSERT(_retainedNewClauses.empty());
  ASSERT(_queueForBackSimplification.empty());
  ASSERT(_queueForDisposal.empty());
  GlobalStopFlag::check();
  goto select_current; 

};// void Master::mainLoop3()


void Master::backSimplifyBy3(Clause* simplifier)
{
 CALL("backSimplifyBy3(Clause* simplifier)");
 _backwardDemodulationSemaphore = true; 
 Clause* backSimplified; 
 bool somethingSimplified = false;
 if (backDemodulationQuery(simplifier))
  { 
   while (nextBackDemodulated())
    {         
     _statBackSimplified++;
         
     if (collectBackSimplified())
      { 
       backSimplified = _backDemodulation.simplified();
       ASSERT(backSimplified->mainSet() == Clause::MainSetActive);       
       enqueueForDisposal(backSimplified);        
       somethingSimplified = true;
       processCollected3(); // not before enqueueForDisposal3(backSimplified)!

       if (proof) 
        { 
         terminate(Proof); 
         _backwardDemodulationSemaphore = false; 
         return; 
        };
       if (_pagedPassive.totalNumberOfClauses() >= _parameters.maxPassive) 
        { 
         terminate(MaxPassive); 
         _backwardDemodulationSemaphore = false; 
         return; 
        }; 
      }
     else
      if (_freshClause.WasTautology())
       {
        backSimplified = _backDemodulation.simplified();
        enqueueForDisposal(backSimplified);
       };
     GlobalStopFlag::check();
    };  
  }; 
 if (_parameters.orphan_murder && somethingSimplified) killActiveOrphans3();
 _backwardDemodulationSemaphore = false;
}; // void Master::backSimplifyBy3(Clause* simplifier)

Clause* Master::reanimate3(PagedPassiveClauseStorage::StoredClause& cl)
{
 CALL("reanimate3(PagedPassiveClauseStorage::StoredClause& cl)");
 ASSERT(_parameters.main_alg == 3); 

 ulong max_weight = _freshClause.MaxWeight();
 _freshClause.SetMaxWeight(VampireKernelConst::MaxClauseSize); 
 
 _freshClause.LoadClause(cl.literals());
 
 cl.resetAncestors();
 Clause* ancestor;
 while (cl.nextAncestor(ancestor)) _freshClause.PushAncestor(ancestor);
 _freshClause.InverseAncestorOrder();

 _freshClause.UsedRules(cl.rules());
 _freshClause.UsedRule(ClauseBackground::Reanimation);
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';

 _pagedPassive.popSelected();

 Clause* res = _freshClause.Reanimate();
 if (res)
  {
   res->SetNumber(_clauseNumberGenerator->generate());
   res->setMainSet(Clause::MainSetFree); 
   if (_parameters.useClauseNumIndex) _clauseNumIndex.insert(res);
   if (_parameters.showPassive) _parameters.showPassive(res);
  }; 
 _freshClause.SetMaxWeight(max_weight);

 return res;
}; // Clause* Master::reanimate3(PagedPassiveClauseStorage::StoredClause& cl) 

void Master::processCollected3()
{
 CALL("processCollected3()");
 if (_parameters.showGen) _freshClause.output(cout << "% GEN " << statGenerated() << ' ') << '\n';         
 Clause* answerClause;
 ulong numOfNewClauses;
 ulong numOfRejectedNewClauses;
 _freshClause.ProcessCollected3(answerClause,
                                _pagedPassive,
                                numOfNewClauses,
                                numOfRejectedNewClauses,
                                _clauseNumberGenerator,
                                (bool)_parameters.showPassive);
 _statClausesRefusedAllocation += numOfRejectedNewClauses;
 if (answerClause)  
  {
   proof = answerClause; 
   return; 
  };
}; // void Master::processCollected3()



void Master::makeActive3(Clause* cl)
{
 CALL("makeActive3(Clause* cl)");
 cl->setMainSet(Clause::MainSetActive);
 cl->setFlag(Clause::FlagHasChildren); // too memory inefficient, must be changed 
 _active.Insert(cl);
 _infDispatcher.makeActive(cl);
 if (_parameters.forward_subsumption) integrateIntoForwardSubsumptionIndex(cl);
 if (_parameters.backward_subsumption) 
  { 
   integrateIntoBackwardSubsumptionIndex(cl);
  };
 if (_parameters.forward_demodulation > 0) integrateIntoForwardDemodulationIndex(cl);
 if (_parameters.backward_demodulation > 0) integrateIntoBackDemodulationIndex(cl);
 if (_parameters.showActive) _parameters.showActive(cl);
}; // void Master::makeActive3(Clause* cl)


void Master::destroyPassive3()
{
 CALL("destroyPassive3()");
 ASSERT(_passive.isEmpty());
 _pagedPassive.destroy();
 _pagedPassive.init();  
}; // void Master::destroyPassive3()


bool Master::releaseMemory3(ulong size)
{
 CALL("releaseMemory3(ulong size)");
 if (size > PagedPassiveClauseStorage::pageSize()) 
   return false;
 try_free_page:
  if (_pagedPassive.recyclePage()) return true;
  if (_pagedPassive.cancelWorstClause()) goto try_free_page; 
 return false;
}; // bool Master::releaseMemory3(ulong size) 


void Master::killActiveOrphans3()
{
 CALL("killActiveOrphans3()");
 _active.ResetIterator();
 for (Clause* cl = _active.Next(); cl; cl = _active.Next())
  if ((cl->Active())
      && (!cl->IsVIP())
      && (cl->IsOrphan()))
   {
    enqueueForDisposal(cl);
    _statActiveOrphansMurdered++;
   };
}; // void Master::killActiveOrphans3()

}; // namespace VK



//====================================================




