/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         NonuniformConstantTermWeighting.hpp
// Description:  A term weighting scheme which assumes
//               that symbols may have different weights.
//               Not argument sensitive.    
// Created:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//===================================================================
#ifndef NONUNIFORM_CONSTANT_TERM_WEIGHTING_H
#define NONUNIFORM_CONSTANT_TERM_WEIGHTING_H
//===================================================================
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TermWeighting.hpp"
#include "TermWeightType.hpp"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NONUNIFORM_CONSTANT_TERM_WEIGHTING
#define DEBUG_NAMESPACE "NonuniformConstantTermWeighting"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{
  class LocalUnification;

  class NonuniformConstantTermWeighting : public TermWeighting
    {
    public: 
      NonuniformConstantTermWeighting() 
	{
	  setScheme(NonuniformConstant);
	};    
      ~NonuniformConstantTermWeighting();
      void* operator new(size_t);
      void operator delete(void* obj);

      void collectWeight(WeightPolynomial& weight,const Flatterm* term);
      void collectWeight(WeightPolynomial& weight,TermWeightType coefficient,const Flatterm* term);
      TermWeightType computeMinimalInstanceWeight(const Flatterm* term,
							       bool& containsVariables);
      TermWeightType computeWeightIfGroundAndLessOrEqual(TermWeightType coefficient,
								      const Flatterm* term,
								      TermWeightType weightLimit);
      // ^ returns 0 if nonground and -1L if too heavy

      void collectWeight(WeightPolynomial& weight,const TERM& term);
      void collectWeight(WeightPolynomial& weight,TermWeightType coefficient,const TERM& term);
      void collectWeight(WeightPolynomial& weight,TermWeightType coefficient,const PrefixSym* term);
      void collectWeight(WeightPolynomial& weight,const PrefixSym* term);
      void collectWeight(WeightPolynomial& weight,const TERM& term,ulong variableIndex);

      void 
	collectWeight(WeightPolynomial& weight,
		      TermWeightType coefficient,
		      const TERM* complexTerm,
		      const InstRet::Substitution* subst);
      void 
        collectWeight(WeightPolynomial& weight,
		      const TERM* complexTerm,
		      const InstRet::Substitution* subst);

      void collectWeight(WeightPolynomial& weight,
			 const TERM& term,
			 const ForwardMatchingSubstitutionCache* subst);



      void
      collectWeight(WeightPolynomial& weight,
		    const TERM& term,
		    const LocalUnification* unifier);

      TermWeightType computeMinimalInstanceWeight(const PrefixSym* term,bool& containsVariables);
      TermWeightType computeWeightIfGroundAndLessOrEqual(TermWeightType coefficient,const PrefixSym* term,TermWeightType weightLimit);
      // returns 0 if nonground and -1L if too heavy
      TermWeightType computeWeightIfGroundAndLessOrEqual(const PrefixSym* term,TermWeightType weightLimit);
      // returns 0 if nonground and -1L if too heavy
      TermWeightType computeMinimalInstanceWeightIfLessOrEqual(TermWeightType coefficient,
									    const PrefixSym* term,
									    TermWeightType weightLimit,
									    bool& ground);
      // ^ returns -1 if too heavy

      TermWeightType computeMinimalInstanceWeightIfLessOrEqual(const PrefixSym* term,
									    TermWeightType weightLimit,
									    bool& ground);
      // ^ returns -1 if too heavy
  



      TermWeightType computeMinimalInstanceWeight(const TERM* term,bool& containsVariables);
      TermWeightType computeWeightIfGroundAndLessOrEqual(TermWeightType coefficient,
								      const TERM* term,
								      TermWeightType weightLimit);
      // returns 0 if nonground and -1L if too heavy

      TermWeightType computeWeightIfGroundAndLessOrEqual(const TERM* term,TermWeightType weightLimit);
      // returns 0 if nonground and -1L if too heavy

      TermWeightType computeMinimalInstanceWeightIfLessOrEqual(TermWeightType coefficient,
									    const TERM* term,
									    TermWeightType weightLimit,
									    bool& ground);
      // ^ returns -1 if too heavy

      TermWeightType computeMinimalInstanceWeightIfLessOrEqual(const TERM* term,
									    TermWeightType weightLimit,
									    bool& ground);
      // ^ returns -1 if too heavy


    private:
  
      InstRet::Substitution::Instance _instance;
      TERM::Iterator _iter;


    }; // class NonuniformConstantTermWeighting : public TermWeighting


}; // namespace VK

//===================================================================
#endif

