/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         OldBackwardSubsumption.h
// Description:  Implements back subsumption using discrimination trees.
// Created:      Jan 14, 2001, 15:10
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef OLD_BACKWARD_SUBSUMPTION_H
//=============================================================================
#define OLD_BACKWARD_SUBSUMPTION_H 
#include "jargon.h"
#include "VampireKernelDebugFlags.h"
#include "VampireKernelConst.h"
#include "OldBSIndex.h"
#include "OldBSMachine.h"
#include "ClauseNumIndex.h"
#include "Clause.h"
#include "Term.h"
#include "SubsumptionCheck.h" 
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OLD_BACKWARD_SUBSUMPTION
 #define DEBUG_NAMESPACE "OldBackwardSubsumption"
#endif
#include "debugMacros.h"
//============================================================================

namespace VK
{
  using namespace BK;
  using namespace Gematogen;
class OldBackwardSubsumption
{
 public:
  OldBackwardSubsumption(ClauseNumIndex* clauseNumIndex) 
   : _index(),
     _integrator(&_index),
     _removal(&_index),
     _retrieval(&_index),
     _machine(),        
     _termTraversal(),
     _clauseNumIndex(clauseNumIndex),
     _setMode(false),
     _optimisedMultisetMode(false)
     #ifdef DEBUG_NAMESPACE
      , _debugSubsumptionCheck()
     #endif 
  {
  };
  ~OldBackwardSubsumption() 
  {
   CALL("destructor ~OldBackwardSubsumption()");      
  };
  bool subsumptionAllowedInSetMode(const Clause* subsumingCl,const Clause* subsumedCl)
  {
   return subsumingCl->NumOfPosEq() <= subsumedCl->NumOfPosEq(); 
  };
  bool subsumptionAllowedInMultisetMode(const Clause* subsumingCl,const Clause* subsumedCl)
  {
   return true; 
  };
  void destroyIndex() { CALL("destroyIndex()"); _index.destroy(); };
  #ifndef NO_DEBUG
   bool clean() const { return _index.clean(); };
  #endif        
  void integrate(const Clause* cl);
  void remove(const Clause* cl);
  bool query(const Clause* cl);
  bool subsumeNext() 
  {
   if (_setMode) return subsumeNextSetMode();
   if (_optimisedMultisetMode && _queryClause->allMaximalSelected()) return subsumeNextSetMode();
   return subsumeNextMultisetMode();
  };
  bool subsumeNextSetMode();
  bool subsumeNextMultisetMode();
  const Clause* subsumed() const { return _recentlySubsumedCl; };
  void setSetMode(bool fl) { _setMode = fl; }; 
  void setOptimisedMultisetMode(bool fl) { _optimisedMultisetMode = fl; }; 
 private:
  void preproQuery(const Clause* cl);
  void preproQueryLit(const TERM* lit);
  void preproQuerySymLit(const TERM* lit);
  void preproQueryComplexTerm(const TERM* t);
  void preproQueryComplexTermArguments(const TERM* t);
  void preproQuerySymLitComplexTerm(const TERM* t);
  void preproQuerySymLitComplexTermArguments(const TERM* t);
 private:
  OldBSIndex _index;
  OldBSIndex::Integrator _integrator;
  OldBSIndex::Removal _removal;
  OldBSIndex::Retrieval _retrieval;
  OldBSMachine _machine;  
  TERM::Traversal _termTraversal; 
  ClauseNumIndex* _clauseNumIndex;
  const Clause* _recentlySubsumedCl;
  const Clause* _queryClause; 
  bool _setMode;
  bool _optimisedMultisetMode;
  #ifdef DEBUG_NAMESPACE
   SubsumptionCheck _debugSubsumptionCheck;
  #endif
}; // class OldBackwardSubsumption

inline void OldBackwardSubsumption::preproQuery(const Clause* cl)
{
 CALL("preproQuery(const Clause* cl)");
 _machine.resetQuery(cl); 

 for (LiteralList::Iterator iter(cl->LitList()); iter.notEnd(); iter.next()) preproQueryLit(iter.currentLiteral());
 _machine.endOfQuery();
}; // void OldBackwardSubsumption::preproQuery(const Clause* cl)


inline void OldBackwardSubsumption::preproQueryLit(const TERM* lit)
{
 CALL("preproQueryLit(const TERM* lit)");
}; // void OldBackwardSubsumption::preproQueryLit(const TERM* lit) 


inline void OldBackwardSubsumption::preproQueryComplexTermArguments(const TERM* t) 
{
 CALL("preproQueryComplexTermArguments(const TERM* t)");
 ASSERT(t->isComplex());
 ASSERT(t->arity());
}; // void OldBackwardSubsumption::preproQueryComplexTermArguments(const TERM* t)  


inline void OldBackwardSubsumption::preproQueryComplexTerm(const TERM* t)
{
 CALL("preproQueryComplexTerm(const TERM* t)");
 ASSERT(t->isComplex());
}; // void OldBackwardSubsumption::preproQueryComplexTerm(const TERM* t) 

inline void OldBackwardSubsumption::preproQuerySymLit(const TERM* lit)
{
 CALL("preproQuerySymLit(const TERM* lit)");
 ASSERT(lit->IsSymLitHeader());
}; // void OldBackwardSubsumption::preproQuerySymLit(const TERM* lit)

inline void OldBackwardSubsumption::preproQuerySymLitComplexTerm(const TERM* t)
{
 CALL("preproQuerySymLitComplexTerm(const TERM* t)");
 ASSERT(t->isComplex());
 
}; // void OldBackwardSubsumption::preproQuerySymLitComplexTerm(const TERM* t) 


inline void OldBackwardSubsumption::preproQuerySymLitComplexTermArguments(const TERM* t) 
{
 CALL("preproQuerySymLitComplexTermArguments(const TERM* t)");
 ASSERT(t->isComplex());
 ASSERT(t->arity());
}; // void OldBackwardSubsumption::preproQuerySymLitComplexTermArguments(const TERM* t)  

}; // namespace VK

 
//======================================================================
#undef DEBUG_NAMESPACE
//======================================================================
#endif
