/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         Parser.h
// Description:  TPTP parser front end.
// Created:      Jun 04, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//======================================================================
#ifndef PARSER_H
#define PARSER_H
//======================================================================
#ifndef NO_DEBUG_VIS
#include <iostream.h>
#endif
#include <string.h>
#include "jargon.h"
#include "VampireKernelDebugFlags.h"
#include "DestructionMode.h"
#include "GList.h"
#include "Malloc.h"
#include "Input.h"
#include "pr_abstract_syntax.h"
#include "pr_parser.h"
//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PARSER
 #define DEBUG_NAMESPACE "Parser"
#endif
#include "debugMacros.h"
//======================================================================

namespace VK
{
class Parser
{
 public:
  Parser() :
    _clauses((Input::ClauseList*)0)
    {
    };
  
  ~Parser() 
    {
      if (DestructionMode::isThorough())
	{
	  Input::destroyClauseList(_clauses);
	};
    };
  
  bool parseFile(const char* tptpDir,const char* fileName);
  const Input::ClauseList* clauses() const { return _clauses; };
  Input::Signature& signature() { return _signature; };

 private:
  enum ItemSort
  {
    TPTPIncludeDirective,
    TPTPInputClause,
    TPTPInputFormula,
    TPTPExtensionSymbolPrecedence,
    TPTPExtensionSymbolWeight,
    TPTPExtensionAnswerPredicateDeclaration,
    UnknownItemSort
  };

  enum ItemType
  {
    TPTPAxiom,
    TPTPConjecture,
    TPTPHypothesis,
    UnknownItemType
  };
  
 private:    
  bool processInputItem(PR_LITERAL* item,
			const char* tptpDir,
			const char* fileName);

  static ItemSort itemSort(PR_LITERAL* item);
  static ItemType itemType(PR_TERM* it);
  bool readClause(PR_LITERAL* item);
  bool readLiterals(PR_TERM_LIST* lst,Input::LiteralList*& result);
  bool readLiteral(PR_TERM* lit,Input::Literal*& result);
  bool readAtom(PR_TERM* at,Input::Term*& result);
  bool readTerms(PR_TERM_LIST* terms,Input::TermList*& result);
  bool readTerm(PR_TERM* term,Input::Term*& result);
  bool readSymbolPrecedence(PR_LITERAL* item);
  bool readSymbolWeight(PR_LITERAL* item);
  bool readAnswerPredicateDeclaration(PR_LITERAL* item);
 private:
  Input::Signature _signature;
  Input::ClauseList* _clauses;
}; // class Parser

}; // namespace VK


//======================================================================
#endif
