/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         Signature.cpp
// Description:  Various signature related operations.  
// Created:      Apr 20, 2001, 14:30
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#include "Signature.h"
#include "Tabulation.h"
//=================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_SIGNATURE
 #define DEBUG_NAMESPACE "Signature"
#endif 
#include "debugMacros.h"
//=================================================  


namespace VK
{
long Signature::InitStatic::_count; // Implicitely initialised to 0L
Signature* Signature::_current;


Signature::Signature() 
  : _inputVariableName(DOP("Signature::_inputVariableName")),
    _nextInputVariableNumber(0UL),
    _functorName(DOP("Signature::_functorName")),
    _isInputFunctor(DOP("Signature::_isInputFunctor")),
    _isPredicate(DOP("Signature::_isPredicate")),
    _isCommutative(DOP("Signature::_isCommutative")),
#ifdef NO_DEBUG
    _arity(ULONG_MAX,(int)0),
#else
    _arity(ULONG_MAX,"Signature::_arity"),
#endif
#ifdef NO_DEBUG
    _isAnswerPredicate(false,(int)0),
#else
    _isAnswerPredicate(false,"Signature::_isAnswerPredicate"),
#endif
    _nextFunctorNumber(0UL),
  
    _numberOfSkolemSymbols(0L),
    _maxlNumOfSkolemSymbols(LONG_MAX),

    _nextSkolemConstantSuffix(0UL),
    _nextSkolemFunctorSuffix(0UL),
    _nextSkolemPredicateSuffix(0UL),
#ifdef DEBUG_NAMESPACE
    _weightIsSet(false,"Signature::_weightIsSet"),
    _weightConstantPart((TermWeightType)1,"Signature::_weightConstantPart"),
    _positiveHeaderPrecedence(DefaultHeaderPrecedence,"Signature::_positiveHeaderPrecedence"),  
    _negativeHeaderPrecedence(DefaultHeaderPrecedence,"Signature::_negativeHeaderPrecedence"),
#else
    _weightIsSet(false,0),
    _weightConstantPart((TermWeightType)1,0),
    _positiveHeaderPrecedence(DefaultHeaderPrecedence,0),  
    _negativeHeaderPrecedence(DefaultHeaderPrecedence,0),
#endif
    _minimalHeaderPrecedence(DefaultHeaderPrecedence),
    _maximalHeaderPrecedence(DefaultHeaderPrecedence),
    _autoFunctorPrecedence(),
    _symbolWeightByArity(1L)
{
  CALL("constructor Signature()");

  ASSERT(PositivePolarity == 0UL);
  ASSERT(NegativePolarity == 1UL);
  ASSERT(DefaultHeaderPrecedence == 0L);
  ASSERT(MinimalPossibleHeaderPrecedence == (LONG_MIN/4)); 
  ASSERT(MaximalPossibleHeaderPrecedence == (LONG_MAX/4));

  _functorName.sub(VampireKernelConst::UnordEqNum) = "equal"; // equality predicate in the TPTP format 
  _isInputFunctor.sub(VampireKernelConst::UnordEqNum) = true;
  _isPredicate.sub(VampireKernelConst::UnordEqNum) = true;
  _isCommutative.sub(VampireKernelConst::UnordEqNum) = true;           
  _arity.sub(VampireKernelConst::UnordEqNum) = 2UL;
  setSymbolWeightConstantPart(VampireKernelConst::UnordEqNum,(TermWeightType)1);
  if (VampireKernelConst::UnordEqNum >= _nextFunctorNumber) _nextFunctorNumber = VampireKernelConst::UnordEqNum + 1;
  
  _functorName.sub(VampireKernelConst::OrdEqNum) = "Equal"; // can not be an input functor in the TPTP format 
  _isInputFunctor.sub(VampireKernelConst::OrdEqNum) = true;
  _isPredicate.sub(VampireKernelConst::OrdEqNum) = true; 
  _isCommutative.sub(VampireKernelConst::OrdEqNum) = true; 
  _arity.sub(VampireKernelConst::OrdEqNum) = 2UL; 
  setSymbolWeightConstantPart(VampireKernelConst::OrdEqNum,(TermWeightType)1);
  if (VampireKernelConst::OrdEqNum >= _nextFunctorNumber) _nextFunctorNumber = VampireKernelConst::OrdEqNum + 1; 

  _functorName.sub(VampireKernelConst::ReservedPred0) = "ReservedPred0"; // can not be an input functor in the TPTP format 
  _isPredicate.sub(VampireKernelConst::ReservedPred0) = true; 
  _arity.sub(VampireKernelConst::ReservedPred0) = 2UL; 
  setSymbolWeightConstantPart(VampireKernelConst::ReservedPred0,(TermWeightType)1);
  if (VampireKernelConst::ReservedPred0 >= _nextFunctorNumber) _nextFunctorNumber = VampireKernelConst::ReservedPred0 + 1; 

  _functorName.sub(VampireKernelConst::ReservedPred1) = "ReservedPred1"; // can not be an input functor in the TPTP format 
  _isPredicate.sub(VampireKernelConst::ReservedPred1) = true; 
  _arity.sub(VampireKernelConst::ReservedPred1) = 2UL; 
  setSymbolWeightConstantPart(VampireKernelConst::ReservedPred1,(TermWeightType)1);
  if (VampireKernelConst::ReservedPred1 >= _nextFunctorNumber) _nextFunctorNumber = VampireKernelConst::ReservedPred1 + 1; 

  _functorName.sub(VampireKernelConst::ReservedPred2) = "ReservedPred2"; // can not be an input functor in the TPTP format
  _isPredicate.sub(VampireKernelConst::ReservedPred2) = true;
  _arity.sub(VampireKernelConst::ReservedPred2) = 2UL; 
  setSymbolWeightConstantPart(VampireKernelConst::ReservedPred2,(TermWeightType)1);
  if (VampireKernelConst::ReservedPred2 >= _nextFunctorNumber) _nextFunctorNumber = VampireKernelConst::ReservedPred2 + 1; 


  // Fill the "holes" between the reserved functors
   
  for (ulong fn = 0; fn < _nextFunctorNumber; fn++) 
    if (!reservedFunctorNumber(fn)) _functorName.sub(fn) = "NoFunctor"; // can not be an input functor in the TPTP format  


  // Create canonical names for all variables
  for (ulong vn = 0; vn < VampireKernelConst::MaxNumOfVariables; vn++)
    {
      _canonicalVariableName[vn] = createCanonicalVariableName(vn);
    };

  _autoFunctorPrecedence.assignFixed(VampireKernelConst::UnordEqNum,0L);
  _autoFunctorPrecedence.assignFixed(VampireKernelConst::OrdEqNum,0L);
  ASSERT_IN(checkObj(),"END");
}; // Signature::Signature() 



void Signature::destroy()
{
  CALL("destroy()");

}; // void Signature::destroy()





void Signature::resetInputVariables() 
{ 
  CALL("resetInputVariables()");
  while (_nextInputVariableNumber)  
    {
      _nextInputVariableNumber--; 
      delete [] (_inputVariableName.sub(_nextInputVariableNumber));
    };
}; // void Signature::resetInputVariables() 

ulong Signature::inputVariableNumber(const char* varName)
{
  CALL("inputVariableNumber(const char* varName)");
  for (ulong vn = 0; vn < _nextInputVariableNumber; vn++)
    if (!(strcmp(varName,_inputVariableName.sub(vn)))) return vn;
  _inputVariableName.sub(_nextInputVariableNumber) = new char[strlen(varName) + 1];
  strcpy(_inputVariableName.sub(_nextInputVariableNumber),varName);
  _nextInputVariableNumber++;
  return _nextInputVariableNumber - 1; 
}; // ulong Signature::inputVariableNumber(const char* varName)

ulong Signature::inputFunctorNumber(const char* funName)
{
  CALL("inputFunctorNumber(const char* funName)");
  for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
    if (!(strcmp(funName,_functorName.sub(fn)))) return fn;
  _functorName.sub(_nextFunctorNumber) = new char[strlen(funName) + 1];
  strcpy(_functorName.sub(_nextFunctorNumber),funName);
  _isInputFunctor.sub(_nextFunctorNumber) = true;
  _isPredicate.sub(_nextFunctorNumber) = false;   
  _isCommutative.sub(_nextFunctorNumber) = false; 
  _nextFunctorNumber++;
  ASSERT(_nextFunctorNumber <= _functorName.size());
  ASSERT(_nextFunctorNumber <= _isInputFunctor.size());
  ASSERT(_nextFunctorNumber <= _isPredicate.size());
  ASSERT(_nextFunctorNumber <= _isCommutative.size());
  return _nextFunctorNumber - 1;  
}; // ulong Signature::inputFunctorNumber(const char* funName)


void Signature::setFunctorProperties(ulong funNum,ulong arity,bool isPredicate)
{
  CALL("setFunctorProperties(ulong funNum,ulong arity,bool isPredicate)");
  _arity.sub(funNum) = arity;
  _isPredicate.sub(funNum) = isPredicate;
}; // void Signature::setFunctorProperties(ulong funNum,ulong arity,bool isPredicate)

void Signature::makePredicate(ulong funNum) 
{
  CALL("makePredicate(ulong funNum)"); 
  _isPredicate.sub(funNum) = true;
};


void Signature::makeNonpredicate(ulong funNum)
{
  CALL("makeNonpredicate(ulong funNum)");
  _isPredicate.sub(funNum) = false;
};


void Signature::makeCommutative(ulong funNum)
{
  CALL("makeCommutative(ulong funNum)");
  _isCommutative.sub(funNum) = true;
}; 

void Signature::makeNoncommutative(ulong funNum)
{
  CALL("makeNoncommutative(ulong funNum)");
  _isCommutative.sub(funNum) = false;
};



void Signature::setCommutativity(const SymbolInfo* commutativityTable,
				 const SymbolInfo* symmetryTable)
{
  CALL("setCommutativity(const Clause* const *commutativityTable,const Clause* const *symmetryTable)");
  ASSERT(_nextFunctorNumber <= _isCommutative.size());
  for (ulong fn = 0; fn < _nextFunctorNumber; fn++)     
    _isCommutative[fn] = 
      (commutativityTable && commutativityTable->commutativityLaw(fn)) || 
      (symmetryTable && symmetryTable->symmetryLaw(fn));
};



bool Signature::createNewSkolemConstantNumber(ulong& res)
{
  CALL("createNewSkolemConstantNumber()");
  static char tmpString[256]; 
  if (_numberOfSkolemSymbols >= _maxlNumOfSkolemSymbols) 
    return false;
  strcpy(tmpString,VampireKernelConst::SkolemConstantNamePrefix);
  char* endOfPrefix = tmpString + strlen(VampireKernelConst::SkolemConstantNamePrefix);
  char* suffix;
 try_suffix:
  suffix = endOfPrefix; 
  print(_nextSkolemConstantSuffix,suffix);
  suffix[0] = '\0';
  if (isRegisteredFunctorName(tmpString)) 
    {
      // a function with this name already exists, try bigger suffix  
      _nextSkolemConstantSuffix++;
      goto try_suffix;
    };
  // this candidate is OK, register it
   
  _functorName.sub(_nextFunctorNumber) = new char[strlen(tmpString) + 1];
  strcpy(_functorName.sub(_nextFunctorNumber),tmpString);
  _isInputFunctor.sub(_nextFunctorNumber) = false;
  _isPredicate.sub(_nextFunctorNumber) = false;   
  _isCommutative.sub(_nextFunctorNumber) = false; 
  _arity.sub(_nextFunctorNumber) = 0UL;
  setSymbolWeightConstantPart(_nextFunctorNumber,autoWeight(0UL));
  res = _nextFunctorNumber;
  _nextFunctorNumber++;

  ASSERT(_nextFunctorNumber <= _functorName.size());
  ASSERT(_nextFunctorNumber <= _isInputFunctor.size());
  ASSERT(_nextFunctorNumber <= _isPredicate.size());
  ASSERT(_nextFunctorNumber <= _isCommutative.size());

  _numberOfSkolemSymbols++;
  return true; 
}; // bool Signature::createNewSkolemConstantNumber(ulong& res)
  
bool Signature::createNewSkolemFunctorNumber(ulong arity,ulong& res)
{
  CALL("createNewSkolemFunctorNumber(arity,ulong& res)");
  static char tmpString[256]; 
  if (_numberOfSkolemSymbols >= _maxlNumOfSkolemSymbols) 
    return false;
  strcpy(tmpString,VampireKernelConst::SkolemFunctorNamePrefix);
  char* endOfPrefix = tmpString + strlen(VampireKernelConst::SkolemFunctorNamePrefix);
  char* suffix;
 try_suffix:
  suffix = endOfPrefix; 
  print(_nextSkolemFunctorSuffix,suffix);
  suffix[0] = '\0';
  if (isRegisteredFunctorName(tmpString)) 
    {
      // a function with this name already exists, try bigger suffix  
      _nextSkolemFunctorSuffix++;
      goto try_suffix;
    };
  // this candidate is OK, register it
   
  _functorName.sub(_nextFunctorNumber) = new char[strlen(tmpString) + 1];
  strcpy(_functorName.sub(_nextFunctorNumber),tmpString);
  _isInputFunctor.sub(_nextFunctorNumber) = false;
  _isPredicate.sub(_nextFunctorNumber) = false;   
  _isCommutative.sub(_nextFunctorNumber) = false; 
  _arity.sub(_nextFunctorNumber) = arity;
  setSymbolWeightConstantPart(_nextFunctorNumber,autoWeight(arity));
  res = _nextFunctorNumber;
  _nextFunctorNumber++;

  ASSERT(_nextFunctorNumber <= _functorName.size());
  ASSERT(_nextFunctorNumber <= _isInputFunctor.size());
  ASSERT(_nextFunctorNumber <= _isPredicate.size());
  ASSERT(_nextFunctorNumber <= _isCommutative.size());

  _numberOfSkolemSymbols++; 
  return true;
}; // bool Signature::createNewSkolemFunctorNumber(ulong arity,ulong& res)

bool Signature::createNewSkolemPredicateNumber(ulong arity,ulong& res)
{
  CALL("createNewSkolemPredicateNumber(arity,ulong& res)");
  static char tmpString[256]; 
  if (_numberOfSkolemSymbols >= _maxlNumOfSkolemSymbols) 
    return false;
  strcpy(tmpString,VampireKernelConst::SkolemPredicateNamePrefix);
  char* endOfPrefix = tmpString + strlen(VampireKernelConst::SkolemPredicateNamePrefix);
  char* suffix;
 try_suffix:
  suffix = endOfPrefix; 
  print(_nextSkolemPredicateSuffix,suffix);
  suffix[0] = '\0';
  if (isRegisteredFunctorName(tmpString)) 
    {
      // a function with this name already exists, try bigger suffix  
      _nextSkolemPredicateSuffix++;
      goto try_suffix;
    }; 
  // this candidate is OK, register it
   
  _functorName.sub(_nextFunctorNumber) = new char[strlen(tmpString) + 1];
  strcpy(_functorName.sub(_nextFunctorNumber),tmpString);
  _isInputFunctor.sub(_nextFunctorNumber) = false;
  _isPredicate.sub(_nextFunctorNumber) = true;   
  _isCommutative.sub(_nextFunctorNumber) = false; 
  _arity.sub(_nextFunctorNumber) = arity;
  setSymbolWeightConstantPart(_nextFunctorNumber,autoWeight(arity));
  res = _nextFunctorNumber;
  _nextFunctorNumber++;

  ASSERT(_nextFunctorNumber <= _functorName.size());
  ASSERT(_nextFunctorNumber <= _isInputFunctor.size());
  ASSERT(_nextFunctorNumber <= _isPredicate.size());
  ASSERT(_nextFunctorNumber <= _isCommutative.size());

  _numberOfSkolemSymbols++;  
  return true;
}; // bool Signature::createNewSkolemPredicateNumber(ulong arity,ulong& res)


ostream& Signature::output(ostream& str)
{ 
  str << "%=================== Signature: =====================\n";

  ASSERT(_nextFunctorNumber <= _functorName.size());
  ASSERT(_nextFunctorNumber <= _isInputFunctor.size());
  ASSERT(_nextFunctorNumber <= _isPredicate.size());
  ASSERT(_nextFunctorNumber <= _isCommutative.size());

  for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
    {
      if ((!reservedFunctorNumber(fn)) || (fn == VampireKernelConst::UnordEqNum))
	  
	{
	  str << fn << ": " << _functorName[fn] << "/";
	  if (arityIsSet(fn)) 
	    { 
	      ASSERT(fn < _arity.size());
	      str << _arity[fn]; 
	    }
	  else 
	    str << "?";
	  if (_isInputFunctor[fn]) { str << " input  "; } else str << " skolem ";
	  if (_isPredicate[fn]) 
	    { 
	      str << "predicate ";
	      if (_isCommutative[fn]) { str << "symmetric      "; } else str << "nonsymmetric  "; 
	    } 
	  else
	    {
	      str << "function  ";
	      if (_isCommutative[fn]) str << "commutative    ";  
	    };
	  str << "\n";
	};

    };
  for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
    {
      if ((!reservedFunctorNumber(fn)) || 
	  (fn == VampireKernelConst::UnordEqNum))
	{
	  if (_isPredicate[fn]) 
	    { 
	      str << "header_precedence(++" << _functorName[fn] << ") = " << headerPrecedence(PositivePolarity,fn) << ",     "
		  << "header_precedence(--" << _functorName.sub(fn) << ") = " << headerPrecedence(NegativePolarity,fn) << "\n";
	    };
	  str << "symbol_precedence(" << _functorName.sub(fn) << ") = " << symbolPrecedence(fn) << "\n";
	  
	  str << "weight("  << _functorName[fn] << ") = "
	      << weightConstantPart(fn) << "\n";

	};
    };
  return str << "%============ End of signature. =================\n";
}; // ostream& Signature::output(ostream& str)




bool Signature::isRegisteredFunctorName(const char* str)
{
  CALL("isRegisteredFunctorName(const char* str)");
  for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
    if (!(strcmp(str,_functorName.sub(fn)))) return true;
  return false;  
}; // bool Signature::isRegisteredFunctorName(const char* str)

const char* Signature::createCanonicalVariableName(ulong varNum)
{
  CALL("createCanonicalVariableName(ulong varNum)");
  static char tmpString[256];
  strcpy(tmpString,VampireKernelConst::CanonicalVariableNamePrefix);
  char* cursor = tmpString + strlen(VampireKernelConst::CanonicalVariableNamePrefix);
  print(varNum,cursor);
  cursor[0] = '\0';
  char* res = new char[strlen(tmpString) + 1]; 
  strcpy(res,tmpString);
  return res;
}; // const char* Signature::createCanonicalVariableName(ulong varNum)



void Signature::endOfInput()
{
  CALL("endOfInput()");

  // Calculate precedences
  for (ulong fn = 0; fn < nextFunctorNumber(); fn++)
    {
      _autoFunctorPrecedence.assignArity(fn,arity(fn));
    }; 
  _autoFunctorPrecedence.assignAllPrecedences();
  long headerPrecedence;
  for (ulong fn = 0; fn < nextFunctorNumber(); fn++)
    { 
      if (_autoFunctorPrecedence.headerPrecedence(true,fn,headerPrecedence))
	{  
	  setHeaderPrecedence(1UL,fn,headerPrecedence); // 1UL <=> negative
	  if (headerPrecedence > _maximalHeaderPrecedence)
	    {
	      _maximalHeaderPrecedence = headerPrecedence;
	    }
	  else
	    if (headerPrecedence < _minimalHeaderPrecedence)
	      {
		_minimalHeaderPrecedence = headerPrecedence;
	      };

	  ALWAYS(_autoFunctorPrecedence.headerPrecedence(false,fn,headerPrecedence));
	  setHeaderPrecedence(0UL,fn,headerPrecedence); // 0UL <=> positive
	  if (headerPrecedence > _maximalHeaderPrecedence)
	    {
	      _maximalHeaderPrecedence = headerPrecedence;
	    }
	  else
	    if (headerPrecedence < _minimalHeaderPrecedence)
	      {
		_minimalHeaderPrecedence = headerPrecedence;
	      };
	};
    };

  // Assign weights
  for (ulong fn = 0; fn < nextFunctorNumber(); fn++)
    {
      if (!weightIsSet(fn))
	setSymbolWeightConstantPart(fn,autoWeight(arity(fn)));
    };
 
  ASSERT(_nextFunctorNumber <= _functorName.size());
  ASSERT(_nextFunctorNumber <= _isInputFunctor.size());
  ASSERT(_nextFunctorNumber <= _isPredicate.size());
  ASSERT(_nextFunctorNumber <= _isCommutative.size());

}; // void Signature::endOfInput()
  

void Signature::tabulate(Tabulation* tabulation)
{
  CALL("tabulate(Tabulation* tabulation)");
  ASSERT(_nextFunctorNumber <= _functorName.size());
  ASSERT(_nextFunctorNumber <= _isInputFunctor.size());
  ASSERT(_nextFunctorNumber <= _isPredicate.size());
  ASSERT(_nextFunctorNumber <= _isCommutative.size());

  if (!tabulation) return;
  for (ulong fn = 0; fn < _nextFunctorNumber; fn++)
    {
      if ((arityIsSet(fn)) && 
	  ((!reservedFunctorNumber(fn)) || (fn == VampireKernelConst::UnordEqNum)) &&
	  (!isSkolem(fn)))
	{
	  if (_isPredicate[fn]) 
	    { 
	      tabulation->signaturePredicate(isSkolem(fn),functorName(fn),arity(fn));
	      tabulation->headerPrecedence(true,functorName(fn),headerPrecedence(PositivePolarity,fn));
	      tabulation->headerPrecedence(false,functorName(fn),headerPrecedence(NegativePolarity,fn));	      
	    }
	  else
	    {
	      tabulation->signatureFunction(isSkolem(fn),functorName(fn),arity(fn));
	    };
	  tabulation->symbolPrecedence(functorName(fn),symbolPrecedence(fn));
	  tabulation->symbolWeight(functorName(fn),
				   weightConstantPart(fn));
	};
    };

}; // void Signature::tabulate(Tabulation* tabulation)

}; // namespace VK

//====================================================
