/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         StandardLiteralOrdering.hpp
// Description:  Extension of the simplification ordering on terms
//               to literals.  
// Created:      Nov 07, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//==================================================================
#include "StandardLiteralOrdering.hpp"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_LITERAL_ORDERING
 #define DEBUG_NAMESPACE "StandardLiteralOrdering"
#endif
#include "debugMacros.hpp"
//===================================================================

using namespace BK;
using namespace VK;


StandardLiteralOrdering::StandardLiteralOrdering() :
  _mode(0L)
{
};
   
StandardLiteralOrdering::~StandardLiteralOrdering()
{
};

void StandardLiteralOrdering::init()
{
  CALL("init()");
  new (this) StandardLiteralOrdering();
}; // void StandardLiteralOrdering::init()

void StandardLiteralOrdering::destroy()
{
  CALL("destroy()");
}; // void StandardLiteralOrdering::destroy()

Comparison StandardLiteralOrdering::compare(TmpLiteral* lit1,
					    TmpLiteral* lit2) const
{
  CALL("compare(TmpLiteral* lit1,TmpLiteral* lit2) const");
  Comparison cmp;
  switch (_mode)
    {
    case 0L:
      if (lit1->isEquality())
	{
	  if (lit2->isEquality())
	    {
	      return compareEq(lit1,lit2);
	    }
	  else
	    return Less;
	}
      else
	if (lit2->isEquality())
	  return Greater;
 
      // both are nonequalities

      // compare atoms with the simplification ordering
    compare_atoms:
      cmp =  
	SimplificationOrdering::current()->compareAtoms(lit1->flatterm(),
							lit1->weight(),
							lit2->flatterm(),
							lit2->weight()); 
      if (cmp == Equal)
	{
	  if (lit1->polarity() == lit2->polarity()) return Equal;
	  return (lit1->isPositive()) ? Less : Equal;
	};
      return cmp;

    case 1L:
      if (lit1->isEquality())
	{
	  if (lit2->isEquality())
	    {
	      if (lit1->isNegative())
		{
		  if (lit2->isNegative())
		    {
		      // both are negative
		      return compareEq(lit1,lit2);
		    }
		  else
		    return Greater;
		}
	      else
		if (lit2->isNegative())
		  {
		    return Less;
		  }
		else // both are positive
		  return compareEq(lit1,lit2);
	    }
	  else
	    return Less;
	}
      else
	if (lit2->isEquality())
	  return Greater;
 
      // both are nonequalities

      // compare headers first
      cmp = 
	Signature::current()->compareHeaderPrecedences(lit1->polarity(),
						       lit1->header().functor(),
						       lit2->polarity(),
						       lit2->header().functor());
      ASSERT(cmp != Incomparable);
      if (cmp == Equal)
	{
	  ASSERT(lit1->header() == lit2->header());

	  // compare atoms with the simplification ordering
	  return SimplificationOrdering::current()->compareAtoms(lit1->flatterm(),
								 lit1->weight(),
								 lit2->flatterm(),
								 lit2->weight()); 
	}
      else
	return cmp;	

    case 2L:
      if (lit1->isEquality())
	{
	  if (lit2->isEquality())
	    {
	      return compareEq(lit1,lit2);
	    }
	  else
	    return Less;
	}
      else
	if (lit2->isEquality())
	  return Greater;
 
      // both are nonequalities
      // compare predicates first
      if (lit1->header().functor() == lit2->header().functor())
	{
	  goto compare_atoms;
	}
      else
	return 
	  Signature::current()->compareHeaderPrecedences(lit1->polarity(),
							 lit1->header().functor(),
							 lit2->polarity(),
							 lit2->header().functor());
    case 3L:      
      if (lit1->isEquality())
	{
	  if (lit2->isEquality())
	    {
	      return compareEq(lit1,lit2);
	    }
	  else
	    {
	      return inverse(compareNeqWithEq(lit2,lit1));
	    };
	}
      else
	if (lit2->isEquality())
	  {
	    return compareNeqWithEq(lit1,lit2);
	  }
	else // neither is equality, compare atoms
	  { 
	    goto compare_atoms;
	  };


#ifdef DEBUG_NAMESPACE
    default:
      ICP("ICP0");
      return Incomparable;
#else
#  ifdef _SUPPRESS_WARNINGS_      
    default:
      RuntimeError::report("Wrong value of _mode in StandardLiteralOrdering::compare(..)");
      return Incomparable;
#  endif
#endif

    }; // switch (_mode)
}; // Comparison StandardLiteralOrdering::compare(TmpLiteral* lit1,TmpLiteral* lit2) const




Comparison StandardLiteralOrdering::compareLayers(const TmpLiteral* lit1,
						      const TmpLiteral* lit2) const
{
  CALL("compareLayers(const TmpLiteral* lit1,const TmpLiteral* lit2) const");
  if (_mode == 3L) return Equal;
 
  if (_mode == 0L)
    {      
      if (lit1->isEquality())
	{
	  if (lit2->isEquality())
	    {
	      return Equal;
	    }
	  else
	    return Less;
	}
      else
	if (lit2->isEquality())
	  return Greater;
 
      // both are nonequalities
      return Equal;
    };


  if (_mode == 1L)
    {      
      if (lit1->isEquality())
	{
	  if (lit2->isEquality())
	    {
	      if (lit1->isNegative())
		{
		  if (lit2->isNegative())
		    {
		      // both are negative
		      return Equal;
		    }
		  else
		    return Greater;
		}
	      else
		if (lit2->isNegative())
		  {
		    return Less;
		  }
		else // both are positive
		  return Equal;
	    }
	  else
	    return Less;
	}
      else
	if (lit2->isEquality())
	  return Greater;
 
      // both are nonequalities

      return 
	Signature::current()->compareHeaderPrecedences(lit1->polarity(),
						       lit1->header().functor(),
						       lit2->polarity(),
						       lit2->header().functor());
    };

  ASSERT(_mode == 2L);
    
  if (lit1->isEquality())
    {
      if (lit2->isEquality())
	{
	  return Equal;
	}
      else
	return Less;
    }
  else
    if (lit2->isEquality())
      return Greater;
 
  // both are nonequalities

  if (lit1->header().functor() == lit2->header().functor())
    {
      return Equal;
    }
  else
    return 
      Signature::current()->compareHeaderPrecedences(lit1->polarity(),
						     lit1->header().functor(),
						     lit2->polarity(),
						     lit2->header().functor());
}; // Comparison StandardLiteralOrdering::compareLayers(const TmpLiteral* lit1,const TmpLiteral* lit2) const






inline
Comparison StandardLiteralOrdering::compareEq(TmpLiteral* lit1,
				      TmpLiteral* lit2) const
{
  CALL("compareEq(TmpLiteral* lit1,TmpLiteral* lit2) const");
  ASSERT(lit1->isEquality());
  ASSERT(lit2->isEquality());
  
  // compare atoms first
  Comparison acmp;
  if (lit1->isOrderedEquality())
    {
      if (lit2->isOrderedEquality())
	{
	  // both are ordered
	  acmp = compareEqAtomsOO(lit1,lit2);
	}
      else
	{
	  acmp = compareEqAtomsOU(lit1,lit2);
	};
    }
  else
    if (lit2->isOrderedEquality())
      {
	acmp = inverse(compareEqAtomsOU(lit2,lit1));
      }
    else // both are unordered
      {
	acmp = compareEqAtomsUU(lit1,lit2);
      };
  
  if (acmp == Equal)
    {     
      if (lit1->polarity() == lit2->polarity()) return Equal;
      return (lit1->isPositive()) ? Less : Equal;
    };
  return acmp;
}; // Comparison StandardLiteralOrdering::compareEq(TmpLiteral* lit1,TmpLiteral* lit2) const




Comparison StandardLiteralOrdering::compareEqAtomsOO(TmpLiteral* lit1,
					     TmpLiteral* lit2) const

{
  CALL("compareEqAtomsOO(TmpLiteral* lit1,TmpLiteral* lit2) const");
  ASSERT(lit1->isOrderedEquality());
  ASSERT(lit2->isOrderedEquality());

  Comparison cmp11 = 
    SimplificationOrdering::current()->compare(lit1->arg1(),lit1->arg1Weight(),
					       lit2->arg1(),lit2->arg1Weight()); 
  if (cmp11 != Equal) return cmp11;
  return SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),
						    lit2->arg2(),lit2->arg2Weight()); 

}; // Comparison StandardLiteralOrdering::compareEqAtomsOO(TmpLiteral* lit1,TmpLiteral* lit2) const

Comparison StandardLiteralOrdering::compareEqAtomsOU(TmpLiteral* lit1,
					     TmpLiteral* lit2) const

{
  CALL("compareEqAtomsOU(TmpLiteral* lit1,TmpLiteral* lit2) const");
  ASSERT(lit1->isOrderedEquality());
  ASSERT(!lit2->isOrderedEquality());
  
  // lit1->arg1() > lit1->arg2()
  switch (SimplificationOrdering::current()->compare(lit1->arg1(),lit1->arg1Weight(),
						     lit2->arg1(),lit2->arg1Weight()))
    {
    case Less:
      // lit1->arg2() < lit1->arg1() < lit2->arg1()
      return Less;
	  
    case Equal: 
      // lit1->arg1() = lit2->arg1() * lit2->arg2(), compare lit1->arg2() with lit2->arg2()
      return SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),
							lit2->arg2(),lit2->arg2Weight());	
    case Greater: 
      // lit1->arg1() > lit2->arg1(), compare lit1->arg1() with lit2->arg2()
      return SimplificationOrdering::current()->compare(lit1->arg1(),lit1->arg1Weight(),
							lit2->arg2(),lit2->arg2Weight());
    case Incomparable:         
      // compare lit1->arg2() with lit2->arg1() and lit2->arg2()
      switch (SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),
							 lit2->arg1(),lit2->arg1Weight()))
	{
	case Less: 
	  if (SimplificationOrdering::current()->lessOrEqual(lit1->arg1(),lit1->arg1Weight(),
							     lit2->arg2(),lit2->arg2Weight()))
	    {
	      return Less;
	    }
	  else
	    return Incomparable;

	case Equal:
	  return SimplificationOrdering::current()->compare(lit1->arg1(),lit1->arg1Weight(),
							    lit2->arg2(),lit2->arg2Weight());

	case Greater:
	  // lit1->arg1() * lit2->arg1(), lit1->arg2() > lit2->arg1()
	  switch (SimplificationOrdering::current()->compare(lit1->arg1(),lit1->arg1Weight(),
							     lit2->arg2(),lit2->arg2Weight()))
	    {
	    case Less: return Less;
	    case Equal: return Greater;
	    case Greater: return Greater;
	    case Incomparable: 
	      if (SimplificationOrdering::current()->greaterOrEqual(lit1->arg2(),lit1->arg2Weight(),
								    lit2->arg2(),lit2->arg2Weight()))
		{
		  return Greater;
		}
	      else
		return Incomparable;
	    };

	case Incomparable:
	  // lit1->arg1() * lit2->arg1(), lit1->arg2() * lit2->arg1()
	  if (SimplificationOrdering::current()->less(lit1->arg1(),lit1->arg1Weight(),
						      lit2->arg2(),lit2->arg2Weight()))
	    {
	      return Less;
	    }
	  else
	    return Incomparable;
	};

    }; // switch

#ifdef DEBUG_NAMESPACE
  ICP("ICP0");
  return Incomparable;
#else
#  ifdef _SUPPRESS_WARNINGS_
  return Incomparable;
  
#  endif
#endif

}; // Comparison StandardLiteralOrdering::compareEqAtomsOU(TmpLiteral* lit1,TmpLiteral* lit2) const


Comparison StandardLiteralOrdering::compareEqAtomsUU(TmpLiteral* lit1,
					     TmpLiteral* lit2) const

{
  CALL("compareEqAtomsUU(TmpLiteral* lit1,TmpLiteral* lit2) const");
  ASSERT(!lit1->isOrderedEquality());
  ASSERT(!lit2->isOrderedEquality());

  Comparison cmp = SimplificationOrdering::current()->compare(lit1->arg1(),lit1->arg1Weight(),lit2->arg1(),lit2->arg1Weight());
  switch (cmp)
    {
    case Greater:
      // lit1->arg1() > lit2->arg1()
      cmp = SimplificationOrdering::current()->compare(lit1->arg1(),lit1->arg1Weight(),lit2->arg2(),lit2->arg2Weight()); 
      switch (cmp)
	{
	case Greater:
	  // lit1->arg1() > lit2->arg1(), lit1->arg1() > lit2->arg2()
	  return Greater;
        
	case Less:
	  // lit1->arg1() > lit2->arg1(), lit1->arg1() < lit2->arg2(). Not an error since v2.73
	  return SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg2(),lit2->arg2Weight()); 
       
	case Incomparable:
	  // lit1->arg1() > lit2->arg1(), lit1->arg1() * lit2->arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg2(),lit2->arg2Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1->arg1() > lit2->arg1(), lit1->arg1() * lit2->arg2(), lit1->arg2() > lit2->arg2()
	      return Greater;
            
	    case Less:
	      // lit1->arg1() > lit2->arg1(), lit1->arg1() * lit2->arg2(), lit1->arg2() < lit2->arg2()
	      return Incomparable;
           
	    case Incomparable:
	      // lit1->arg1() > lit2->arg1(), lit1->arg1() * lit2->arg2(), lit1->arg2() * lit2->arg2()
	      return Incomparable;
           
	    case Equal:
	      // lit1->arg1() > lit2->arg1(), lit1->arg1() * lit2->arg2(), lit1->arg2() = lit2->arg2()
	      return Greater;
	    };
        
	case Equal:
	  // lit1->arg1() > lit2->arg1(), lit1->arg1() = lit2->arg2() : error because lit2->arg1() * lit2->arg2()
#ifdef DEBUG_NAMESPACE 
	  ICP("DE20"); 
#endif
	  return Incomparable;
	};

    case Less:
      // lit1->arg1() < lit2->arg1()
      cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg1(),lit2->arg1Weight()); 
      switch (cmp)
	{
	case Greater:
	  // lit1->arg1() < lit2->arg1(), lit1->arg2() > lit2->arg1(). Not an error since v2.73
	  return SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg2(),lit2->arg2Weight()); 
        
	case Less:
	  // lit1->arg1() < lit2->arg1(), lit1->arg2() < lit2->arg1()
	  return Less; // was Greater in the CASC-16 version (obvious incompleteness)
        
	case Incomparable:
	  // lit1->arg1() < lit2->arg1(), lit1->arg2() * lit2->arg1()
	  cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg2(),lit2->arg2Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1->arg1() < lit2->arg1(), lit1->arg2() * lit2->arg1(), lit1->arg2() > lit2->arg2()
	      return Incomparable;
           
	    case Less:
	      // lit1->arg1() < lit2->arg1(), lit1->arg2() * lit2->arg1(), lit1->arg2() < lit2->arg2()
	      return Less;
           
	    case Incomparable:
	      // lit1->arg1() < lit2->arg1(), lit1->arg2() * lit2->arg1(), lit1->arg2() * lit2->arg2()
	      return Incomparable;
           
	    case Equal:
	      // lit1->arg1() < lit2->arg1(), lit1->arg2() * lit2->arg1(), lit1->arg2() = lit2->arg2()
	      return Less; 
	    }; 
        
	case Equal:
	  // lit1->arg1() < lit2->arg1(), lit1->arg2() = lit2->arg1() : error because lit1->arg1() * lit1->arg2()
#ifdef DEBUG_NAMESPACE 
	  ICP("DE40");
#endif
	  return Incomparable;
	};
      
    case Incomparable:
      // lit1->arg1() * lit2->arg1()
      cmp = SimplificationOrdering::current()->compare(lit1->arg1(),lit1->arg1Weight(),lit2->arg2(),lit2->arg2Weight()); 
      switch (cmp)
	{
	case Greater:
	  // lit1->arg1() * lit2->arg1(), lit1->arg1() > lit2->arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg1(),lit2->arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() > lit2->arg2(), lit1->arg2() > lit2->arg1()
	      return Greater; 
           
	    case Less:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() > lit2->arg2(), lit1->arg2() < lit2->arg1()
	      return Incomparable;
           
	    case Incomparable:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() > lit2->arg2(), lit1->arg2() * lit2->arg1()
	      return Incomparable;
           
	    case Equal:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() > lit2->arg2(), lit1->arg2() = lit2->arg1()
	      return Greater;
	    };
          
	case Less:
	  // lit1->arg1() * lit2->arg1(), lit1->arg1() < lit2->arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg1(),lit2->arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() < lit2->arg2(), lit1->arg2() > lit2->arg1()
	      cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg2(),lit2->arg2Weight());
	      ASSERT(cmp != Equal);
	      return cmp;
          
	    case Less:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() < lit2->arg2(), lit1->arg2() < lit2->arg1()
	      return Less;
            
	    case Incomparable:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() < lit2->arg2(), lit1->arg2() * lit2->arg1()
	      cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg2(),lit2->arg2Weight()); 
	      if (cmp == Less) return Less;
	      return Incomparable;
            
	    case Equal:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() < lit2->arg2(), lit1->arg2() = lit2->arg1()
	      return Less; 
	    }; 
         

	case Incomparable:
	  // lit1->arg1() * lit2->arg1(), lit1->arg1() * lit2->arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg1(),lit2->arg1Weight()); 
	  switch (cmp)
	    {
	    case Greater:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() * lit2->arg2(), lit1->arg2() > lit2->arg1()
	      cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg2(),lit2->arg2Weight()); 
	      if (cmp == Greater) return Greater;
	      ASSERT(cmp != Equal);
	      return Incomparable;
           
	    case Less:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() * lit2->arg2(), lit1->arg2() < lit2->arg1()
	      return Incomparable;
           
	    case Incomparable:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() * lit2->arg2(), lit1->arg2() * lit2->arg1()
	      return Incomparable;
           
	    case Equal:
	      // lit1->arg1() * lit2->arg1(), lit1->arg1() * lit2->arg2(), lit1->arg2() = lit2->arg1()
	      return Incomparable;
	    }; 
        
	case Equal:
	  // lit1->arg1() * lit2->arg1(), lit1->arg1() = lit2->arg2()
	  cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg1(),lit2->arg1Weight()); 
	  return cmp;
	};
      
    case Equal:
      // lit1->arg1() = lit2->arg1()
      cmp = SimplificationOrdering::current()->compare(lit1->arg2(),lit1->arg2Weight(),lit2->arg2(),lit2->arg2Weight()); 
      return cmp;
    };
#ifdef DEBUG_NAMESPACE 
  ICP("DE100");
  return Incomparable; 
#else
#ifdef _SUPPRESS_WARNINGS_
  return Incomparable; 
#endif  
#endif

}; // Comparison StandardLiteralOrdering::compareEqAtomsUU(TmpLiteral* lit1,TmpLiteral* lit2) const


Comparison StandardLiteralOrdering::compareNeqWithEq(TmpLiteral* neq,
						     TmpLiteral* eq) const
{
  CALL("compareNeqWithEq(TmpLiteral* neq,TmpLiteral* eq) const");
  ASSERT(!neq->isEquality());
  ASSERT(eq->isEquality());
  Comparison cmp1 = 
    SimplificationOrdering::current()->compare(neq->flatterm(),neq->weight(),
					       eq->arg1(),eq->arg1Weight()); 
    
  if (eq->isOrderedEquality()) return cmp1;
  
  if (cmp1 == Less) return Less;
  
  Comparison cmp2 = 
    SimplificationOrdering::current()->compare(neq->flatterm(),neq->weight(),
					       eq->arg2(),eq->arg2Weight()); 
  
  if (cmp2 == Incomparable) return Incomparable;
  if (cmp2 == Less) return Less;
  ASSERT(cmp2 == Greater);
  return cmp1;
}; // Comparison StandardLiteralOrdering::compareNeqWithEq(..)



//=================================================================
