/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         Tabulation.h
// Description:  
// Created:      Mar 21, 2000, 16:35
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//=======================================================
#ifndef TABULATION_H
//=======================================================
#define TABULATION_H
#include <iostream.h>
#include "jargon.h"
#include "GlobAlloc.h"
#include "VampireKernelDebugFlags.h"
#include "Term.h"
#include "Clause.h"
#include "TermWeightType.h"
//========================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TABULATION
 #define DEBUG_NAMESPACE "Tabulation"
#endif
#include "debugMacros.h"
//========================================================
namespace VK
{

class Tabulation   
{
 public:
  Tabulation() : _str((ostream*)0), _jobId("unknown_job_id") {};
  ~Tabulation() {};
  void assignStream(ostream* str) { _str = str; };
  void assignJob(const char* jobId) { _jobId = jobId; };
  void assignKernelSession(ulong id) { _kernelSessionId = id; };

  // basic operations  
  void flush() { _str->flush(); };
  void open_unit() 
    {
      CALL("open_unit()");
      *_str << "vout("; 
      quoted_atom(_jobId);
      *_str << ','; 
    };    

  void open_ks_unit() 
    {
      CALL("open_ks_unit()");
      *_str << "vkout("; 
      quoted_atom(_jobId);
      *_str << ',' << _kernelSessionId << ',';
    };    
  
  void close_unit() { *_str << ").\n"; flush(); };
  
  void open_name() { *_str << '['; }; 
  void close_name() { *_str << ']'; };

  void comma() { *_str << ','; };
  void lbr() { *_str << '['; };
  void rbr() { *_str << ']'; };
  void quoted_atom(const char* s) { *_str << '\'' << s << '\''; };  
  void string(const char* s) { *_str << '\"' << s << '\"'; };
  void number(int n) { *_str << n; }; 
  void number(unsigned n) { *_str << n; }; 
  void number(long n) { *_str << n; }; 
  void number(unsigned long n) { *_str << n; }; 
  void number(float n) { *_str << n; };
  void number(double n) { *_str << n; };
  void boolean(bool b) { *_str << b; };
  void day(int dd) { number(dd); };
  void month(int mm) { number(mm); };
  void year(int yy) { number(yy); };
  void date(int dd,int mm,int yy) 
    { 
      lbr(); day(dd); comma(); month(mm); comma(); year(yy); rbr(); 
    };
  void hour(int hh) { number(hh); }; 
  void minute(int mm) { number(mm); };
  void second(int ss) { number(ss); }; 
  void time(int hh,int mm,int ss) 
    {
      lbr(); hour(hh); comma(); minute(mm); comma(); second(ss); rbr(); 
    };
  void simple_name(const char* n) { open_name(); quoted_atom(n); close_name(); };
  void simple_name(const char* n1,const char* n2)
    { 
      open_name(); quoted_atom(n1); comma(); quoted_atom(n2); close_name(); 
    };
  void simple_name(const char* n1,const char* n2,const char* n3)
    { 
      open_name(); quoted_atom(n1); comma(); quoted_atom(n2); comma(); quoted_atom(n3); close_name(); 
    };

  //============ derived shorthands: =======================================

  void header(const char* testId,const char* problemName)
    { 
      *_str << "%========  " << testId << " : " << problemName << "  ==========\n";
   
    };
  void problemName(const char* pn) 
    { 
      open_unit(); simple_name("problem_name"); comma(); quoted_atom(pn); close_unit();
    
    };
  void testId(const char* tid)
    {
      open_unit(); simple_name("test_id"); comma(); quoted_atom(tid); close_unit();
   
    };
  void hostName(const char* hn)
    {
      open_unit(); simple_name("host_name"); comma(); quoted_atom(hn); close_unit();   
    };
  void jobDate(int dd,int mm,int yy)
    {
      open_unit(); simple_name("job_date"); comma(); date(dd,mm,yy); close_unit();
   
    };
  void jobTime(int hh,int mm,int ss)
    {
      open_unit(); simple_name("job_time"); comma(); time(hh,mm,ss); close_unit();
   
    };
  void version(float num,const char* desc) 
    {
      open_unit(); simple_name("version"); comma();
      lbr(); number(num); comma(); quoted_atom(desc); rbr(); close_unit(); 
   
    };
  void versionDate()
    {
      open_unit(); simple_name("version_date"); comma(); quoted_atom(__DATE__); close_unit();
   
    };
  void versionTime()
    {
      open_unit(); simple_name("version_time"); comma(); quoted_atom(__TIME__); close_unit();
   
    };
  void option(const char* optName,long optVal)
    {
      open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
    };
  void option(const char* optName,unsigned long optVal)
    {
      open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
    };
  void option(const char* optName,bool optVal)
    {
      open_unit(); simple_name("option",optName); comma(); boolean(optVal); close_unit();
      
    };
  void option(const char* optName,const char* optVal) 
    {
      open_unit(); simple_name("option",optName); comma(); quoted_atom(optVal); close_unit();
   
    }; 
  void option(const char* optName,float optVal)
    {
      open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
    }; 
  void option(const char* optName,double optVal)
    {
      open_unit(); simple_name("option",optName); comma(); number(optVal); close_unit();
   
    };
  
  void stat(const char* paramName,bool paramVal)
    {   
      open_unit(); simple_name("stat",paramName); comma(); boolean(paramVal); close_unit(); 
    };
  void stat(const char* paramName,int paramVal)
    {   
      open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit(); 
    };
  void stat(const char* paramName,unsigned paramVal)
    {
      open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
    };
  void stat(const char* paramName,long paramVal)
    {
      open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
    };
  void stat(const char* paramName,unsigned long paramVal)
    {
      open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
    };
  void stat(const char* paramName,float paramVal)
    {
      CALL("stat(const char* paramName,float paramVal)");
      open_unit(); 
      simple_name("stat",paramName); 
      comma(); 
      number(paramVal); 
      close_unit();
    };
  void stat(const char* paramName,double paramVal)
    {
      open_unit(); simple_name("stat",paramName); comma(); number(paramVal); close_unit();
    };
  void stat(const char* paramName,const char* paramVal)
    {
      open_unit(); simple_name("stat",paramName); comma(); quoted_atom(paramVal); close_unit();
    };
  void kernelSessionResult(char r)
    {
      open_unit();
      simple_name("result"); comma();
      if (r == '+') { quoted_atom("+"); }
      else 
	if (r == '-') { quoted_atom("-"); }
	else quoted_atom("0");       
      close_unit();
    };

  void result(char r)
    {
      open_unit();
      simple_name("result"); comma();
      if (r == '+') { quoted_atom("+"); }
      else 
	if (r == '-') { quoted_atom("-"); }
	else quoted_atom("0");       
      close_unit();
    };

  void terminationReason(const char* tr)
    {
      open_unit(); 
      simple_name("termination_reason"); 
      comma(); 
      quoted_atom(tr); 
      close_unit();
    };  


  void headerPrecedence(bool positive,const char* predicateName,long precedence)
    {
      open_unit(); 
      simple_name("header_precedence");  
      comma();
      lbr(); 
      if (positive) { *_str << "++"; } else *_str << "--";
      *_str << predicateName; 
      comma(); 
      number(precedence);     
      rbr();
      close_unit();
    };

  void symbolPrecedence(const char* symbolName,long precedence)
    {
      open_unit(); 
      simple_name("symbol_precedence");   
      comma();
      lbr(); 
      *_str << symbolName; 
      comma(); 
      number(precedence);     
      rbr();
      close_unit();
    };

  void symbolWeight(const char* symbolName,
		    long weightConstPart)
    {
      open_unit(); 
      simple_name("symbol_weight");  
      comma(); 
      lbr();
      *_str << symbolName;  
      comma(); 
      number(weightConstPart);
      rbr();
      close_unit();      
    };


  void signatureFunction(bool skolem,const char* symbolName,ulong arity)
    {
      open_unit(); 
      simple_name("signature", ((skolem)? "skolem" : "input"),"function"); 
      comma(); 
      lbr(); 
      *_str << symbolName; 
      comma(); 
      number(arity);     
      rbr();
      close_unit();
    };

  void signaturePredicate(bool skolem,const char* symbolName,ulong arity)
    {
      open_unit(); 
      simple_name("signature", ((skolem)? "skolem" : "input"),"predicate"); 
      comma(); 
      lbr(); 
      *_str << symbolName; 
      comma(); 
      number(arity);     
      rbr();
      close_unit();
    };


  
  void answerPredicate(const char* predicateName)
    {
      open_unit(); 
      simple_name("answer_predicate"); 
      comma(); 
      lbr(); 
      *_str << predicateName;
      rbr();
      close_unit();
    };
  

  void proofClause(const Clause* cl) 
    {
      *_str << "vproof("; 
      quoted_atom(_jobId); 
      comma(); 
      //number(_kernelSessionId);
      //comma(); 
      cl->outputAsPrologTerm(*_str); 
      *_str << ").\n"; 
      flush(); 
    };  

  void errorMessage(const char* errorQualifier,const char* msg)
    {
      *_str << "verror("; 
      quoted_atom(_jobId);
      comma(); 
      lbr(); 
      quoted_atom(errorQualifier);
      rbr();
      comma();
      string(msg); 
      *_str << ").\n"; 
      flush(); 
    };

  
 private:
  ostream* _str;
  const char* _jobId;  
  ulong _kernelSessionId;
}; // class Tabulation

}; // namespace VK

//=======================================================
#endif
