/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         TmpLitList.cpp
// Description:  Lists of temporary (buffer) literals.
// Created:      Dec 17, 2001 
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
// Note:         Reimplementation of TMP_LIT_LIST (tmp_lit_list.h)
//============================================================================
#include "TmpLitList.h"
#include "Signature.h"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TMP_LIT_LIST
 #define DEBUG_NAMESPACE "TmpLitList"
#endif
#include "debugMacros.h"
//============================================================================


namespace VK
{


bool TmpLitList::substPossible(ulong var,unsigned long substDepth,long substSize) 
{
  CALL("substPossible(ulong var,unsigned long substDepth,long substSize) const");
  ASSERT(substDepth);
  for (TMP_LITERAL* el = first(); el; el = el->next())
    if (!el->instantiationPossible(var,substDepth,substSize))
      return false;
  return true;
}; // bool TmpLitList::substPossible(ulong var,unsigned long substDepth,long substSize) 
  

long TmpLitList::computeWeightBase() const 
{
  CALL("computeWeightBase() const");
  long res = 0L;
  for (const TMP_LITERAL* el = first(); el; el = el->next())
    {
      res += el->weightBase(); 
    };
  return res;
}; // long computeWeightBase() const 

void TmpLitList::computeSizeAndWeightBase(long& sz,long& weightBase) const
{
  CALL("computeSizeAndAndWeightBaseNumOfPosEq(long& sz,long& weightBase) const");
  sz = 0L;
  weightBase = 0L; 
  for (const TMP_LITERAL* el = first(); el; el = el->next())
    {
      sz += el->size();
      weightBase += el->weightBase();
    };
}; // void TmpLitList::computeSizeAndAndWeightBase(long& sz,long& weightBase) const


void TmpLitList::computeSizeAndWeightBaseAndNumOfPosEq(long& sz,long& weightBase,long& numOfPosEq) const
{
  CALL("computeSizeAndWeightBaseAndNumOfPosEq(long& sz,long& weightBase,long& numOfPosEq) const");
  sz = 0L;
  weightBase = 0L; 
  numOfPosEq = 0L;
  for (const TMP_LITERAL* el = first(); el; el = el->next())
    {
      sz += el->size();
      weightBase += el->weightBase();
      if (el->isEquality() && el->isPositive())
	numOfPosEq++;     
    };

}; // void TmpLitList::computeSizeAndWeightBaseAndNumOfPosEq(long& sz,long& weightBase,long& numOfPosEq) const



long TmpLitList::minlPosEqLHSSize() const
{
  CALL("minlPosEqLHSSize() const");

  for (const TMP_LITERAL* el = first(); el; el = el->next())
    {
      if (el->isEquality() && el->isPositive()) 
	{
	  long res = el->minPosEqLHSSize();   
	  for (el = el->next(); el; el = el->next())
	    {  
	      if (el->isEquality() && el->isPositive()) 
		{
		  long pemlw = el->minPosEqLHSSize();
		  if (pemlw < res) res = pemlw;
		};
	    };
	  return res;   
	};
    };
  return 0L;
}; // long TmpLitList::minlPosEqLHSSize() const




void TmpLitList::collectStatistics(Statistics& stat,bool inversePolarityForSelection) const
{
  CALL("collectStatistics(Statistics& stat,bool inversePolarityForSelection) const"); 
  for (const TMP_LITERAL* el = first(); el; el = el->next())
    {
      if (el->isAnswerLiteral()) 
	{
	  ASSERT(!el->isInherentlySelected());
	  ASSERT(!el->isEquality());      
	  stat.answer_literals++;
	}
      else // non-answer literal
	{
	  if (el->isInherentlySelected()) stat.inherently_selected++;
	  if (el->isEquality()) { stat.equations++; }
	  else
	    {
	      if (stat.noneq_nonanswer_lits)
		{
		  //if (GlobalHeaderWeighting.Compare(el->header(),stat.max_noneq_header) == Greater)
		  if (Signature::current()->compareHeaderPrecedences(el->header().polarity(),
							    el->header().functor(),
							    stat.max_noneq_header.polarity(),
							    stat.max_noneq_header.functor()) == Greater)
		 
		    {
		      stat.max_noneq_header = el->header();
		    };
		}
	      else stat.max_noneq_header = el->header(); 
	      stat.noneq_nonanswer_lits++;
	    };
	  if (isNegativeForSelection(el->header(),inversePolarityForSelection)) stat.negative_nonanswer_lits++;
	};
    };
      
}; // void TmpLitList::collectStatistics(Statistics& stat,bool inversePolarityForSelection) const

  

void TmpLitList::collectStatisticsWithSplitting(Statistics& stat,bool inversePolarityForSelection) const
{
  CALL("collectStatisticsWithSplitting(Statistics& stat,bool inversePolarityForSelection) const"); 
  for (const TMP_LITERAL* el = first(); el; el = el->next())
    {
      if (el->isAnswerLiteral()) 
	{
	  ASSERT(!el->isInherentlySelected());
	  ASSERT(!el->isEquality());    
	  stat.answer_literals++;
	}
      else
	{
	  if (el->isInherentlySelected()) stat.inherently_selected++;
	  if (el->isEquality()) { stat.equations++; }
	  else
	    {
	      if (stat.noneq_nonanswer_lits)
		{
		  //if (GlobalHeaderWeighting.Compare(el->header(),stat.max_noneq_header) == Greater)
		  if (Signature::current()->compareHeaderPrecedences(el->header().polarity(),
							    el->header().functor(),
							    stat.max_noneq_header.polarity(),
							    stat.max_noneq_header.functor()) == Greater)
		    {
		      stat.max_noneq_header = el->header();
		    };
		}
	      else stat.max_noneq_header = el->header(); 
	      stat.noneq_nonanswer_lits++;
	      if (el->header().IsSplitting())
		{
		  if (isNegativeForSelection(el->header(),inversePolarityForSelection))
		    {
		      stat.negative_splitting_branches++;
		    }
		  else stat.positive_splitting_branches++;
		};
	    };
	  if (isNegativeForSelection(el->header(),inversePolarityForSelection)) stat.negative_nonanswer_lits++;
	};
    };
}; // void TmpLitList::collectStatisticsWithSplitting(Statistics& stat,bool inversePolarityForSelection) const
  
void TmpLitList::renameVariables(INT_VAR_RENAMING& varRenaming)
{
  CALL("renameVariables(INT_VAR_RENAMING& varRenaming)");
  for (TMP_LITERAL* el = last(); el; el = el->previous()) 
    // ^ order is significant
    { 
      el->renameVariables(varRenaming);
    };
}; // void TmpLitList::renameVariables(INT_VAR_RENAMING& varRenaming)



bool TmpLitList::weed(bool& eqTautology)
{
  CALL("weed(bool& eqTautology)");
  ASSERT(checkIntegrity());

  TmpLitList remainingLits;
  remainingLits.collectFrom(*this);
  ASSERT(empty());
  while (remainingLits.nonempty())
    {
      TMP_LITERAL* lit = remainingLits.pop();
      TERM header = lit->header();
      if (header.IsEquality())
	{
	  if (lit->argsAreEqual())
	    {
	      if (header.Positive()) 
		{
		  push(lit);
		  collectFrom(remainingLits);
		  eqTautology = true; 
		  return false; 
		}
	      else
		lit->recycle();
	    }
	  else // different arguments
	    {
	      for (TMP_LITERAL* el = first(); el; el = el->next())
		if (el->isEquality())
		  {
		    if (lit->equalEqLit(*el))
		      {
			lit->recycle();
			goto next_lit;
		      }
		    else
		      {
			

			if (lit->complementaryEqLit(*el))
			  {
			    // propositional tautology
			    push(lit);
			    collectFrom(remainingLits);
			    eqTautology = false; 
			    return false; 
			  };
		      };
		  };
	      push(lit);
	      lit->orderEqArgs(); 
	    };
	}
      else // nonequality literal
	{
	  for (TMP_LITERAL* el = first(); el; el = el->next())
	    if (lit->equal(*el))
	      {
		lit->recycle();
		goto next_lit;
	      }
	    else
	      if (lit->isComplementary(*el))
		{
		  // propositional tautology
		  push(lit);
		  collectFrom(remainingLits);
		  eqTautology = false; 
		  return false; 
		};
	  push(lit);
	};
    next_lit: {}; // nothing to do here, just the end of the loop body
    };
  
  ASSERT(checkIntegrity());
  return true;  
}; // bool TmpLitList::weed(bool& eqTautology)


bool TmpLitList::weed(const Clause* const * commutativityTable,bool& eqTautology)
{
  CALL("weed(const Clause* const * commutativityTable,bool& eqTautology)");
  TmpLitList remainingLits;
  remainingLits.collectFrom(*this);
  ASSERT(empty());

  while (remainingLits.nonempty())
    {
      TMP_LITERAL* lit = remainingLits.pop();
      TERM header = lit->header();
      if (header.IsEquality())
	{
	  if (lit->argsAreEqual(commutativityTable))
	    {
	      if (header.Positive()) 
		{
		  push(lit);
		  collectFrom(remainingLits);
		  eqTautology = true; 
		  return false; 
		}
	      else
		lit->recycle();
	    }
	  else // different arguments
	    {
	      for (TMP_LITERAL* el = first(); el; el = el->next())
		if (el->isEquality())
		  {
		    if (lit->equalEqLit(*el,commutativityTable))
		      {
			lit->recycle();
			goto next_lit;
		      }
		    else
		      if (lit->complementaryEqLit(*el,commutativityTable))
			{
			  // propositional tautology
			  push(lit);
			  collectFrom(remainingLits);
			  eqTautology = false; 
			  return false; 
			};
		  };
	      push(lit);
	      lit->orderEqArgs(); 
	    };
	}
      else // nonequality literal
	{
	  for (TMP_LITERAL* el = first(); el; el = el->next())
	    if (lit->equal(*el,commutativityTable))
	      {
		lit->recycle();
		goto next_lit;
	      }
	    else
	      if (lit->isComplementary(*el,commutativityTable))
		{
		  // propositional tautology
		  push(lit);
		  collectFrom(remainingLits);
		  eqTautology = false; 
		  return false; 
		};
	  push(lit);
	};
    next_lit: {}; // nothing to do here, just the end of the loop body
    };
  
  ASSERT(checkIntegrity());
  return true;  

}; // bool TmpLitList::weed(const Clause* const * commutativityTable,bool& eqTautology)


void TmpLitList::order(TmpLitList& result,const VAR_WEIGHTING& varWeighting)
{
  CALL("order(TmpLitList& result,const VAR_WEIGHTING& varWeighting)");
  ASSERT(nonempty());
  if (numOfAllLiterals() == 1UL) { result.push(pop()); return; }; 
  if (numOfAllLiterals() == 2UL)
    { 
      TMP_LITERAL* fst = pop();
      TMP_LITERAL* snd = pop();
      if (fst->compareWithVarWeighting(*snd,varWeighting) == Less) 
	{
	  result.push(fst); 
	  result.push(snd); 
	  return;
	};
      result.push(snd); 
      result.push(fst);
      ASSERT(result.checkIntegrity());
      return; 
    };
  // numOfAllLiterals() > 2UL, use the quicksort algorithm
  TMP_LITERAL* pivot = pop();  
  TmpLitList smaller;
  TmpLitList greater; 
  do
    {
      TMP_LITERAL* lit = pop();
      if (lit->compareWithVarWeighting(*pivot,varWeighting) == Less) 
	{ 
	  smaller.push(lit); 
	} 
      else 
	greater.push(lit); 
    }
  while (nonempty()); 
  if (smaller.nonempty()) smaller.order(result,varWeighting);
  result.push(pivot);
  if (greater.nonempty()) greater.order(result,varWeighting);
  ASSERT(result.checkIntegrity());
}; // void TmpLitList::order(TmpLitList& result,const VAR_WEIGHTING& varWeighting)



ostream& TmpLitList::output(ostream& str) const 
{
  CALL("output(ostream& str) const");
  for (const TMP_LITERAL* el = first(); el; el = el->next())
    {
      str << el;
      if (el->next()) str << " % "; 
    };
  return str;
}; //  ostream& output(ostream& str) const 

}; // namespace VK

//============================================================================
