/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
// File:         VampireKernelConst.h
// Description:  Global constants for VampireKernel.
// Created:      Feb 12, 2000, 17:50
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef HELL_CONST_H
//====================================================
#define HELL_CONST_H 
#include <limits.h>
#include <stddef.h>
#include "jargon.h"
//====================================================
class VampireKernelConst
{
 public:
  // interface 
  static const char* const SkolemConstantNamePrefix = "c__";
  static const char* const SkolemFunctorNamePrefix = "f__";
  static const char* const SkolemPredicateNamePrefix = "p__";
  static const char* const CanonicalVariableNamePrefix = "X";
  // basic parameters
  static const ulong MaxlNumOfLits = 65535; // can not be greater than 2^16 - 1 NEW!
  static const ulong MaxSymbolArity = 1023; // can not be greater than 2^12 - 1, see term.h
#ifdef LARGE_TERMS
  static const ulong MaxNumOfVariables = 128;
  static const ulong MaxTermDepth = 48;
  static const ulong MaxTermSize = 256;  
  static const ulong MaxClauseSize = 512;
#else 
#ifdef HUGE_TERMS
  static const ulong MaxNumOfVariables = 256;
  static const ulong MaxTermDepth = 1024;
  static const ulong MaxTermSize = 1024; 
  static const ulong MaxClauseSize = 2048;
#else
  static const ulong MaxNumOfVariables = 64;
  static const ulong MaxTermDepth = 24;  
  static const ulong MaxTermSize = 128;   
  static const ulong MaxClauseSize = 256;
#endif
#endif
 
  static const ulong MaxlFunctorNum = 65535; 
 
  static const long MinMaxWeight = 1L;

  static const ulong MaxNumOfStandardIndices = 2; // for 2 premises in a rule

 // hard values: don't change this unless you know what you are doing!
  static const ulong UnordEqNum = 0;
  static const ulong OrdEqNum = 1;
  static const ulong FirstNoneqFun = 2; 
  static const ulong ReservedPred0 = 2;
  static const ulong ReservedPred1 = 3;
  static const ulong ReservedPred2 = 4;
 
  // derived values
  static const ulong MaxlHeaderNum = MaxlFunctorNum*2 + 1; 

  static const ulong MaxNumOfIndices = MaxNumOfStandardIndices + 0; 

  // checking integrity
  static bool checkIntegrity() 
    { 
      return (UnordEqNum == 0)
	&& (OrdEqNum == 1)
	&& (FirstNoneqFun == 2);
 
    };
  // functions
  static bool isEquality(const ulong& num) { return (num < FirstNoneqFun); }; 
  static bool isOrderedEq(const ulong& num) { return (bool)num; }; 
  static bool isUnorderedEq(const ulong& num) { return !num; }; 
  static bool isReservedPred(const ulong& num) { return (num > 1) && (num < 5); };
}; // namespace VampireKernelConst
//====================================================
#endif



