/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         backward_superposition.hpp
// Description:  Classes dealing with backward superposition.
// Created:      Mar 12, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Apr 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::StoredConstraint 
//               replaced SharedKBConstr
// Revised:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::CompiledConstraintForUnifiers
//               replaced ORD_CONSTR
// Revised:      Jul 15, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Major refurbishment.
//====================================================
#ifndef BACKWARD_SUPERPOSITION_H
//==================================================================
#define BACKWARD_SUPERPOSITION_H
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"   
#include "DestructionMode.hpp"
#include "para_machine.hpp"
#include "NewClause.hpp"
#include "SimplificationOrdering.hpp"
#include "weight_formula.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "BACKWARD_SUPERPOSITION"
#endif
#include "debugMacros.hpp"
//================================================= 

namespace VK
{
class BACKWARD_SUPERPOSITION
{
 public:
  BACKWARD_SUPERPOSITION(NewClause* res,OR_INDEX* unifIndex); 
  ~BACKWARD_SUPERPOSITION(); 
  void init(NewClause* res,OR_INDEX* unifIndex);
  void destroy();

  void setSimultaneous(long degree) { _simultaneousness = degree; };
  void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;     
    };
  void setParamodulateFromVariablesFlag(bool fl) 
  {
    _paramodulateFromVariables = fl;
  };
   
  ulong statBlockedParamodulationsFromVariables() const
  {
    return _statBlockedParamodulationsFromVariables;
  };
  
  void setMainPremiseSubsumption(bool fl)
  {
    _mainPremiseSubsumption = fl;
  };

  void setSidePremiseSubsumption(bool fl)
  {
    _sidePremiseSubsumption = fl;
  };

  
  void setMainSelectedLitDemodulation(bool fl)
  {
    _mainSelectedLitDemodulation = fl;
  };
  void setSideSelectedLitDemodulation(bool fl)
  {
    _sideSelectedLitDemodulation = fl;
  };
  
  ulong statSubsumedMainPremiseInstances() const
  {
    return _statSubsumedMainPremiseInstances;
  };

  ulong statSubsumedSidePremiseInstances() const
  {
    return _statSubsumedSidePremiseInstances;
  };
  
  ulong statDemodulatedMainSelectedLits() const
  {
    return _statDemodulatedMainSelectedLits;
  };
  ulong statDemodulatedSideSelectedLits() const
  {
    return _statDemodulatedSideSelectedLits;
  };
    

  bool SetQuery(Clause* cl); 
  bool CollectResult();
  bool NextInference(); 
  
private:
  void compileMainPremiseResidualSizeFormula();
  void recompileMainPremiseSelectedLitResidualSizeFormula();
  void loadMainPremise(Clause* cl);
  const TERM* findNextMainSelectedPosEq();
  void computeMainPremiseResidualWeight()
    {
      _mainPremiseResidualSize =
	_mainPremiseResidualSizeFormula.Compute() + 
	_mainPremiseSelectedLitResidualSizeFormula.Compute();
    }; 
  bool collectMainPremiseLiterals();
  void LoadEq(const TERM* eq_lit,ulong eq_ind);
  bool Paramodulate();
  bool collectSidePremiseLiterals();

  void LoadLHS(PrefixSym* t);
            
  bool NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr,
		     Clause*& candCl,
		     BackParaCandidates::LitRedexPair*& candLitRedexPair); 

  bool subsumeMainPremise(); 
  bool collectMainPremiseInstance();
  bool subsumeSidePremise(); 
  bool collectSidePremiseInstance();

  
  bool demodulateSelectedLits();
  bool collectMainPremiseSubtermsAvailableForParamodulationInto();
  bool collectSidePremiseSubtermsAvailableForParamodulationInto();


private:
  // main parameters
  NewClause* _resolvent;
  OR_INDEX* _unifIndex; 
  long _simultaneousness;
  bool _noInfBetweenRulesFlag;
  bool _paramodulateFromVariables;
  bool _mainPremiseSubsumption;
  bool _sidePremiseSubsumption;
  bool _mainSelectedLitDemodulation;
  bool _sideSelectedLitDemodulation;

  // statistics   
  ulong _statSubsumedMainPremiseInstances;
  ulong _statSubsumedSidePremiseInstances;
  ulong _statDemodulatedMainSelectedLits;
  ulong _statDemodulatedSideSelectedLits; 
  ulong _statBlockedParamodulationsFromVariables;

  // main premise data
  ulong _mainPremiseVarIndex;
  Clause* _mainPremise;

  LiteralList* _allMainPremiseLiterals;
  LiteralList::Element* _firstNonselectedMainLit;
  LiteralList::Iterator _mainSelectedLitIter;
  LiteralList::Element* _currentSelectedLitListElement;

  WEIGHT_FORMULA _mainPremiseResidualSizeFormula;
  WEIGHT_FORMULA _mainPremiseSelectedLitResidualSizeFormula;
  long _mainPremiseResidualSize;

  // current selected literal in the main premise
  const TERM* _currentMainPosEqLit;
  PrefixSym _currentMainPosEqArg1Mem[VampireKernelConst::MaxTermSize];
  PrefixSym _currentMainPosEqArg2Mem[VampireKernelConst::MaxTermSize];
  PrefixSym* _currentMainPosEqRHS;
  bool _currentMainPosEqIsUnordered;
  SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
  SimplificationOrdering::CompiledConstraintForUnifiers* _currentCompiledConstraint;

  // side premise data
  Clause* _sidePremise;
  ulong _sidePremiseVarIndex;
  SimplificationOrdering::ConstraintCheckOnUnifier* _storedConstraintCheck;



  // retrieval
  PARA_MACH _retrievalOfUnifiable;
  BackParaCandidates::Iterator _candidates;
  BackParaCandidates::LitRedexPair* _candLitRedexPair;


  // control
  bool _newRetrievalAttempt;
  bool _isFromVariable;
  bool _swapArguments;
  bool _noRulesAsSidePremises;
  bool _mainSelectedLitDemodulationAttempted;
  bool _mainSelectedLitDemodulated;
}; // class BACKWARD_SUPERPOSITION

}; // namespace VK
//================================================
#undef DEBUG_NAMESPACE
//==============================================================
#endif 







 
