/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
#include "binary_resolution.hpp"
#include "Clause.hpp"
#include "NewClause.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BINARY_RESOLUTION
 #define DEBUG_NAMESPACE "BIN_ORD_RES"
#endif
#include "debugMacros.hpp"
//=================================================

VK::BIN_ORD_RES::BIN_ORD_RES(NewClause* resolvent,OR_INDEX* resIndex) 
  : premise1(resolvent), 
  premise2(resolvent),
  index1(0),
  index2(1),
  ORMachine(resIndex),
  _resolvent(resolvent),
  _noInfBetweenRulesFlag(false),
  _statSubsumedMainPremiseInstances(0UL),
  _statSubsumedSidePremiseInstances(0UL),
  _statDemodulatedMainSelectedLits(0UL),
  _statDemodulatedSideSelectedLits(0UL)
{
  weight_form.SetIndex(index1);
};

void VK::BIN_ORD_RES::init(NewClause* resolvent,OR_INDEX* resIndex) 
{
  CALL("init(NewClause* resolvent,OR_INDEX* resIndex)");     
  premise1.init(resolvent);
  premise2.init(resolvent);
  index1 = 0;
  index2 = 1;
  ORMachine.init(resIndex);
  weight_form.init();
  _resolvent = resolvent;
  _noInfBetweenRulesFlag = false;
  _statSubsumedMainPremiseInstances = 0UL;
  _statSubsumedSidePremiseInstances = 0UL;
  _statDemodulatedMainSelectedLits = 0UL;
  _statDemodulatedSideSelectedLits = 0UL;
  weight_form.SetIndex(index1);
}; // void VK::BIN_ORD_RES::init(NewClause* resolvent,OR_INDEX* resIndex) 

void VK::BIN_ORD_RES::destroy()
{
  CALL("destroy()");
  weight_form.destroy();
  ORMachine.destroy();
  premise2.destroy();
  premise1.destroy();
}; // void VK::BIN_ORD_RES::destroy()


void VK::BIN_ORD_RES::Load(Clause* c)
 {
  // reset everithing here
  ORMachine.SetIndex(index2);
  premise1.Load(index1,c);
  CollectWeightForm(c);
  _sideClauseCanNotBeRule = _noInfBetweenRulesFlag && c->IsRule();
  currentLit = 0;
  _continuation = ContFirstLit;
 }; // inline void VK::BIN_ORD_RES::Load(Clause* c)   

bool VK::BIN_ORD_RES::NextInference()
 {
  CALL("NextInference()");
  switch (_continuation)
   {
    case ContFirstLit: goto next_lit;
    case ContFirstFactor: goto first_factor;
    case ContNextFactor: goto next_factor; 
   };
  
  ICP("ICP1");

  next_lit:
   if (currentLit > premise1.LastMax()) { return false; };
 
   ASSERT(premise1.nthMaxLit(currentLit));
   ORMachine.CompileLiteral(premise1.nthMaxLit(currentLit));
   ORMachine.Inversepolarity();
   premise1.Cut(currentLit);
  
   currentLit++;
  
   Unifier::current()->UnbindAll();
   RES_PAIR_LIST* candidate;
  
   if (ORMachine.FindFirstLeaf())  
    {
     process_leaf: 
      _residualMainPremiseSize = weight_form.Compute();
      if (_residualMainPremiseSize > _resolvent->MaxWeight())
	{
	  _resolvent->AnotherTooHeavyClause();
	  goto try_next_leaf;
	};

      ORMachine.ResetCandidatesFromLeaf();
      if (demodulateMainSelectedLits()) 
	{
	  while ((candidate  = ORMachine.NextCandidate()))
	    candidate->TheClause()->protectParamodulationsInto();
	  goto try_next_leaf;
	};
      if (subsumeMainPremise()) goto try_next_leaf;

      try_next_cand:
       candidate  = ORMachine.NextCandidate();
       if (!candidate) goto try_next_leaf;       
       if (_sideClauseCanNotBeRule && candidate->TheClause()->IsRule()) 
        {
         goto try_next_cand;
        };
       if (_residualMainPremiseSize + 
	   candidate->TheClause()->MinResWeight() >
	   _resolvent->MaxWeight())
        {
         _resolvent->AnotherTooHeavyClause();
         goto try_next_cand;
        };   
            
       premise2.LoadOptimized(index2,candidate);

       if (demodulateSideSelectedLits(!_mainSelectedLitDemodulation))
	 goto try_next_cand;
       if (subsumeSidePremise()) goto try_next_cand;

       // first resolvent without factoring
       
       if (!MakeNewResolvent()) 
        {
         goto first_factor;
        };
       _continuation = ContFirstFactor;
       return true;
       
       first_factor:
        ResetFactoringState();
       
        next_factor:
         if (NextFactor())
	   {
	     if (demodulateMainSelectedLits()) 
	       {
		 premise2.TheClause()->protectParamodulationsInto();
		 goto next_factor;
	       }; 
	     if (demodulateSideSelectedLits(!_mainSelectedLitDemodulation)) 
	       goto next_factor; 
	     if (subsumeMainPremise()) goto next_factor;  
	     if (subsumeSidePremise()) goto next_factor; 
	     if (!MakeNewResolvent()) goto next_factor;            
	     _continuation = ContNextFactor;
	     return true;          
	   };
  
         // no_more_factors:
         premise2.Unload();
         goto try_next_cand;
     
         // no more candidates in this leaf, try to find another one
         try_next_leaf:
          if (ORMachine.FindNextLeaf()) { goto process_leaf; };
     };
   
   // no more resolvents for this literal
   premise1.UnmarkCut();
   goto next_lit;
  
 }; // bool VK::BIN_ORD_RES::NextInference()


 
bool VK::BIN_ORD_RES::NextFactor()
 {
  CALL("NextFactor()");
  if (premise2.WasFactored())
   {
    if (premise2.GetNextFactor())
     {
      return true;
     }
    else 
     {
      premise2.ResetFactoring();
      if (premise1.WasFactored())
       {
        return premise1.GetNextFactor();
       }
      else
       {
        return premise1.GetFirstFactor(premise1.CutLit()+1);
       };
     };
   }
  else 
   {
    if (premise2.GetFirstFactor(0))
     {
      return true;
     }
    else
     {
      if (premise1.WasFactored())
       {
        return premise1.GetNextFactor();
       }
      else
       {
        return premise1.GetFirstFactor(premise1.CutLit()+1);
       };
     };
   };
 }; // bool VK::BIN_ORD_RES::NextFactor()
 

bool VK::BIN_ORD_RES::MakeNewResolvent()
 {
  CALL("MakeNewResolvent()");
  if (!CollectResult()) return false;

  // resolvent collected in Resolvent
  _resolvent->UsedRule(ClauseBackground::BinRes);
  _resolvent->PushAncestor(FirstPremise());
  _resolvent->PushAncestor(SecondPremise());
  
  return true;
  //_resolvent->ProcessCollected(); 
 }; // bool VK::BIN_ORD_RES::MakeNewResolvent() 


inline bool VK::BIN_ORD_RES::CollectResult() 
    { 
     CALL("CollectResult()");
     if ((!premise1.TheClause()->Active())
         || (!premise2.TheClause()->Active()))
      return false;
     _resolvent->openGenerated(ClauseBackground::BinRes);
     return premise1.CollectNonmarkedLiterals() && premise2.CollectNonmarkedLiterals();
    }; // inline bool VK::BIN_ORD_RES::CollectResult() 


inline void VK::BIN_ORD_RES::ResetFactoringState()
 {
  premise1.ResetFactoring();
  premise2.ResetFactoring();
 }; // void VK::BIN_ORD_RES::ResetFactoringState() 
 
inline void VK::BIN_ORD_RES::CollectWeightForm(const Clause* c)
 {
  weight_form.Reset();
  LiteralList* lits = c->LitList();
  for (LiteralList::Iterator iter(lits,lits->nth(lits->numOfSelectedLits()));
       iter.notEnd();
       iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      weight_form.Compile(iter.currentElement()->literal()); 
 }; // inline void CollectWeightForm() 


bool VK::BIN_ORD_RES::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (premise1.TheClause()->hasSubsumedPremiseInstance()) return false;
  premise1.TheClause()->blockForSubsumption();
  ulong maxWeight = _resolvent->MaxWeight();
  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _resolvent->openGenerated(ClauseBackground::Unspecified);
  if (premise1.collectAllLiterals())
    {
      _resolvent->endOfClause();
      subsumingClause = _resolvent->FSubsume();
      ASSERT(subsumingClause != premise1.TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_resolvent->WasTautology()) goto discarded;
      goto return_false;
    };

 return_false:
  premise1.TheClause()->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++_statSubsumedMainPremiseInstances;
 discarded:
  premise1.TheClause()->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return true;
}; // bool VK::BIN_ORD_RES::subsumeMainPremise()


bool VK::BIN_ORD_RES::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  if (premise2.TheClause()->hasSubsumedPremiseInstance()) return false;
  premise2.TheClause()->blockForSubsumption();
  ulong maxWeight = _resolvent->MaxWeight();
  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _resolvent->openGenerated(ClauseBackground::Unspecified);
  if (premise2.collectAllLiterals())
    {
      _resolvent->endOfClause();
      subsumingClause = _resolvent->FSubsume();
      ASSERT(subsumingClause != premise2.TheClause());
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_resolvent->WasTautology()) goto discarded;
      goto return_false;
    };

 return_false:
  premise2.TheClause()->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++_statSubsumedSidePremiseInstances;
 discarded:
  premise2.TheClause()->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return true;
}; // bool VK::BIN_ORD_RES::subsumeSidePremise()


bool VK::BIN_ORD_RES::demodulateMainSelectedLits()
{
  CALL("demodulateMainSelectedLits()");
  Clause* simplifyingClause;
  if (!_mainSelectedLitDemodulation) return false;
  premise1.TheClause()->blockForDemodulationFrom();
  ulong maxWeight = _resolvent->MaxWeight();
  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _resolvent->openGenerated(ClauseBackground::Unspecified);
  if (premise1.collectSubtermsAvailableForParamodulationInto(true))
    {
      _resolvent->endOfClause();
      simplifyingClause = _resolvent->canBeDemodulated();
      ASSERT(simplifyingClause != premise1.TheClause());
      if (simplifyingClause) goto demodulated;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_resolvent->WasTautology()) goto discarded;
      goto return_false;
    };
    
 return_false:
  premise1.TheClause()->unblockForDemodulationFrom();
  _resolvent->SetMaxWeight(maxWeight);
  return false;
  
 demodulated:
  premise1.TheClause()->protectParamodulationsInto();
  ++_statDemodulatedMainSelectedLits;
 discarded:
  premise1.TheClause()->unblockForDemodulationFrom();
  _resolvent->SetMaxWeight(maxWeight);
  return true;
}; // bool BIN_ORD_RES::demodulateMainSelectedLits()



bool VK::BIN_ORD_RES::demodulateSideSelectedLits(bool includeMarked)
{
  CALL("demodulateSideSelectedLits(bool includeMarked)");
  Clause* simplifyingClause;
  if (!_sideSelectedLitDemodulation) return false;
  premise2.TheClause()->blockForDemodulationFrom();
  ulong maxWeight = _resolvent->MaxWeight();
  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _resolvent->openGenerated(ClauseBackground::Unspecified);
  if (premise2.collectSubtermsAvailableForParamodulationInto(includeMarked))
    {
      _resolvent->endOfClause();
      simplifyingClause = _resolvent->canBeDemodulated();
      ASSERT(simplifyingClause != premise2.TheClause());
      if (simplifyingClause) goto demodulated;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_resolvent->WasTautology()) goto discarded;
      goto return_false;
    };
    
 return_false:
  premise2.TheClause()->unblockForDemodulationFrom();
  _resolvent->SetMaxWeight(maxWeight);
  return false;
  
 demodulated:
  premise1.TheClause()->protectParamodulationsInto();
  premise2.TheClause()->protectParamodulationsInto();
  ++_statDemodulatedSideSelectedLits;
 discarded:
  premise2.TheClause()->unblockForDemodulationFrom();
  _resolvent->SetMaxWeight(maxWeight);
  return true;
}; // bool BIN_ORD_RES::demodulateSideSelectedLits(bool includeMarked)

//=============================================================
 
 
 
 
 
 
 
 
