/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
#ifndef BINARY_RESOLUTION_H
//======================================================
#define BINARY_RESOLUTION_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "premise_factoring.hpp"
#include "or_machine.hpp"
#include "weight_formula.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BINARY_RESOLUTION
 #define DEBUG_NAMESPACE "BIN_ORD_RES"
#endif
#include "debugMacros.hpp"
//=================================================

namespace VK
{
class Clause;
class NewClause;

class BIN_ORD_RES
 {
  public:
   BIN_ORD_RES(NewClause* resolvent,OR_INDEX* resIndex);
   ~BIN_ORD_RES() { CALL("destructor ~BIN_ORD_RES()"); };
   void init(NewClause* resolvent,OR_INDEX* resIndex);
   void destroy();
   void setMainPremiseSubsumption(bool fl)
   {
     _mainPremiseSubsumption = fl;
   };
   void setSidePremiseSubsumption(bool fl)
   {
     _sidePremiseSubsumption = fl;
   };
   void setMainSelectedLitDemodulation(bool fl)
   {
     _mainSelectedLitDemodulation = fl;
   };
   void setSideSelectedLitDemodulation(bool fl)
   {
     _sideSelectedLitDemodulation = fl;
   };
   void setNoInfBetweenRulesFlag(bool fl)
   {
    _noInfBetweenRulesFlag = fl;
   };
   void Load(Clause* c);
   bool NextInference();  

   ulong statSubsumedMainPremiseInstances() const
   {
     return _statSubsumedMainPremiseInstances;
    };
   ulong statSubsumedSidePremiseInstances() const
   {
     return _statSubsumedSidePremiseInstances;
    };
   ulong statDemodulatedMainSelectedLits() const
   {
     return _statDemodulatedMainSelectedLits;
    };
   ulong statDemodulatedSideSelectedLits() const
   {
     return _statDemodulatedSideSelectedLits;
    };
   #ifndef NO_DEBUG 
    bool clean() const { return true; };
   #endif

  private:
   enum Continuation { ContFirstLit, ContFirstFactor, ContNextFactor }; 

  private: 
   void ResetFactoringState();
   bool NextFactor();
   bool MakeNewResolvent();
   void CollectWeightForm(const Clause* c);
   Clause* FirstPremise() const { return premise1.TheClause(); };
   Clause* SecondPremise() const { return premise2.TheClause(); };
   inline bool CollectResult();
   bool subsumeMainPremise();
   bool subsumeSidePremise();
   bool demodulateMainSelectedLits();
   bool demodulateSideSelectedLits(bool includeMarked);


  private:
   OR_PREMISE_WITH_FACTORING premise1,premise2;
   ulong index1;
   ulong index2;
   ulong currentLit;
   OR_MACH ORMachine;
   WEIGHT_FORMULA weight_form;
   long _residualMainPremiseSize;
   NewClause* _resolvent;
   Continuation _continuation;
   bool _noInfBetweenRulesFlag;
   bool _sideClauseCanNotBeRule;
   bool _mainPremiseSubsumption;
   bool _sidePremiseSubsumption;
   ulong _statSubsumedMainPremiseInstances;
   ulong _statSubsumedSidePremiseInstances;
   bool _mainSelectedLitDemodulation;
   bool _sideSelectedLitDemodulation;
   ulong _statDemodulatedMainSelectedLits;
   ulong _statDemodulatedSideSelectedLits;
 }; // class BIN_ORD_RES 
  
 
}; // namespace VK


//================================================
#undef DEBUG_NAMESPACE 
//======================================================
#endif
