/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

#ifndef FORWARD_SUBSUMPTION_H
//===========================================================
#define FORWARD_SUBSUMPTION_H
#ifndef NO_DEBUG_VIS
#include <iostream.h>
#endif
#include "jargon.h"
#include "VampireKernelDebugFlags.h"
#include "VampireKernelConst.h"
#include "fs_code.h"
#include "Stack.h"
#include "ExpandingStack.h"
#include "ObjectPool.h"
#include "Array.h"
#include "GlobAlloc.h"
#include "Clause.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION
#define DEBUG_NAMESPACE "FS_QUERY"
#endif 
#include "debugMacros.h"
//=================================================

namespace VK
{
class FS_COMMAND;
class FS_QUERY
{
 public:
  class LIT
    {    
    public:  
      LIT() {};
      ~LIT() {};
      void* operator new(size_t)
	{
	  CALL("LIT::operator new(size_t)");
	  return GlobAlloc::allocate(sizeof(LIT));
	};
      void operator delete(void* obj)
	{
	  CALL("LIT::operator delete(void* obj)");
	  GlobAlloc::deallocate(obj,sizeof(LIT));
	};
      const Flatterm* Args() const { return arg1; };
      void SetArgs(const Flatterm* a) { arg1 = a; };      
      const Flatterm* Arg1() const { return arg1; }; // for equations
      const Flatterm* Arg2() const { return arg2; }; // for equations
      void SetArgs(const Flatterm* a1,const Flatterm* a2)
	{ arg1 = a1; arg2 = a2; };
      LIT* Next() const { return next; };
      void SetNext(LIT* nl) { next = nl; };
      void Set(const Flatterm* a,LIT* nl) { arg1 = a; captured = 0L; next = nl; };
      void Set(const Flatterm* a1,const Flatterm* a2,LIT* nl) 
	{ arg1 = a1; arg2 = a2; captured = 0L; next = nl; };
      void Capture() { captured++; };
      void Release() { captured--; };
      long Captured() { return captured; };
    public:
#ifndef NO_DEBUG_VIS
      ostream& output(ostream& str) const { return (Args() - 1)->outputTerm(str); };
#endif
    private:
      const Flatterm* arg1;
      const Flatterm* arg2;
      long captured;
      LIT* next;
    }; // class LIT

 public:
  FS_QUERY() 
    : 

#ifdef NO_DEBUG 
    _litsWithHeader((LIT*)0,(int)0),
#else
    _litsWithHeader((LIT*)0,"CN_QUERY::_litsWithHeader"),
#endif

    _posOrdEqHeaderNum(TERM::positiveOrdEqHeader().HeaderNum()),
    _negOrdEqHeaderNum(TERM::negativeOrdEqHeader().HeaderNum()),

    _presentHeaders(DOP("FS_QUERY::_presentHeaders")), 
    collect_signature(true) // default
    {
      CALL("constructor FS_QUERY()");
      DOP(_litsWithHeader.freeze());
      signature = (FunSet*)0; 
    };

  ~FS_QUERY() { CALL("destructor ~FS_QUERY()"); };
  void ResetQuery()
    {
      CALL("ResetQuery()");

      _litPool.reset();
      while (_presentHeaders.nonempty())
	{
	  _litsWithHeader[_presentHeaders.pop()] = (LIT*)0;
	}; 

      last_pos_unord_eq = (LIT*)0;
      last_neg_unord_eq = (LIT*)0;
      num_of_pos_eq = 0;
      if (collect_signature) 
	{
	  if (signature) delete signature;
	  signature = new FunSet();
	};
    };
  void SetCollectSignature(bool fl) 
    { 
      collect_signature = fl; 
    };
  void LinkEquations()
    {
      CALL("LinkEquations()");
      if (last_pos_unord_eq) 
	last_pos_unord_eq->SetNext(FirstLit(_posOrdEqHeaderNum));
      if (last_neg_unord_eq) 
	last_neg_unord_eq->SetNext(FirstLit(_negOrdEqHeaderNum));
    }; 
  void EndOfQuery()
    {
      LinkEquations(); 
    }; 
  void PushLit(const Flatterm* lit)
    {
      CALL("PushLit(const Flatterm* lit)");
      ASSERT(!lit->Symbol().IsEquality());

      DOP(_litsWithHeader.unfreeze());
      LIT** lit_lst_ptr = &(_litsWithHeader.sub(lit->Symbol().HeaderNum()));
      DOP(_litsWithHeader.freeze());


      LIT* currLit = _litPool.reserveObject();
      if (*lit_lst_ptr)
	{
	  currLit->Set(lit + 1,*lit_lst_ptr);
	}
      else // no literals with this header yet
	{
	  _presentHeaders.pushSafe(lit->Symbol().HeaderNum());

	  currLit->Set(lit + 1,(LIT*)0);
	};
      *lit_lst_ptr = currLit;
      if (collect_signature) CollectFunctors(lit);
    };   
    
  void PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
    {
      CALL("PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)");
      ASSERT(header.IsUnorderedEq());
      if (header.Positive()) num_of_pos_eq++;


      DOP(_litsWithHeader.unfreeze());
      LIT** lit_lst_ptr = &(_litsWithHeader.sub(header.HeaderNum()));
      DOP(_litsWithHeader.freeze());


      LIT* currLit = _litPool.reserveObject();
      if (*lit_lst_ptr)
	{
	  currLit->Set(arg1,arg2,*lit_lst_ptr);
	}
      else // no literals with this header yet
	{
	  _presentHeaders.pushSafe(header.HeaderNum()); 


	  currLit->Set(arg1,arg2,(LIT*)0);
 
	  if (header.Negative()) 
	    { 
	      last_neg_unord_eq = currLit;
	    }
	  else // positive
	    {
	      last_pos_unord_eq = currLit;
	    };
	};

      *lit_lst_ptr = currLit;
      if (collect_signature) 
	{
	  AddFunctor(header.HeaderNum());
	  CollectFunctors(arg1); 
	  CollectFunctors(arg2);
	};
    };
    
  void PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
    {
      CALL("PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)");
      ASSERT(header.IsEquality() && header.IsOrderedEq());  
      if (header.Positive()) num_of_pos_eq++;

      DOP(_litsWithHeader.unfreeze());
      LIT** lit_lst_ptr = &(_litsWithHeader.sub(header.HeaderNum()));
      DOP(_litsWithHeader.freeze());



      LIT* currLit = _litPool.reserveObject();
      if (*lit_lst_ptr)
	{
	  currLit->Set(arg1,arg2,*lit_lst_ptr);
	}
      else // no literals with this header yet
	{

	  _presentHeaders.pushSafe(header.HeaderNum());

	  currLit->Set(arg1,arg2,(LIT*)0);
	};
      *lit_lst_ptr = currLit;
      if (collect_signature) 
	{
	  header.MakeUnordered();
	  AddFunctor(header.HeaderNum());
	  CollectFunctors(arg1); 
	  CollectFunctors(arg2);
	};
    }; 
    
  LIT* FirstLit(ulong header_num)  
    {
      CALL("FirstLit(ulong header_num)");
      if (header_num >= _litsWithHeader.size()) return (LIT*)0;
      return _litsWithHeader[header_num];
    }; 

  bool SubsumptionAllowedInSetMode(Clause* subsuming_cl)
    {
      // This doesn't guarantee completeness
      return subsuming_cl->NumOfPosEq() <= num_of_pos_eq;
    }; 

  bool SubsumptionAllowedInMultisetMode(Clause* subsuming_cl)
    {
      return true;
    };  
  bool SubsumptionAllowedInOptimisedMultisetMode(Clause* subsuming_cl,long numOfFactorings)
    {
      return (!numOfFactorings) 
	|| (subsuming_cl->allMaximalSelected());
    };  
  const FunSet* Signature() const { return signature; }; 

 private:

  void AddFunctor(unsigned long fun_num)
    {
      if (!(signature->write((size_t)fun_num)))
	{
	  FunSet* new_sig = signature->add((size_t)fun_num);
	  delete signature;
	  signature = new_sig; 
	};
    };
  void CollectFunctors(const Flatterm* t)
    {
      CALL("CollectFunctors(const Flatterm* t)");
      ASSERT(collect_signature);
      Flatterm::Iterator iter(t);
      TERM sym;
      while (iter.NotEndOfTerm())
	{
	  sym = iter.CurrPos()->symbolRef();
	  if (sym.isComplex()) AddFunctor(sym.HeaderNum());       
	  iter.NextSym();
	};
    };
    
 public:
#ifndef NO_DEBUG_VIS
  ostream& output(ostream& str) const;
#endif 

 private: // structure  
  ObjectPool<GlobAlloc,LIT> _litPool;

  Array<GlobAlloc,LIT*,128UL,FS_QUERY> _litsWithHeader;
  const ulong _posOrdEqHeaderNum;
  const ulong _negOrdEqHeaderNum;
  
  ExpandingStack<GlobAlloc,ulong,32UL,FS_QUERY> _presentHeaders;

  LIT* last_pos_unord_eq;
  LIT* last_neg_unord_eq;
  long num_of_pos_eq;
  bool collect_signature;
  FunSet* signature;
}; // class FS_QUERY
 
#ifdef FS_STAT
class FS_STATISTICS
{
 public:
  FS_STATISTICS() 
    : totInstr(0UL),
    compInstr(0UL) 
    {   
      CALL("constructor FS_STATISTICS()");
    };
  ~FS_STATISTICS() {};
  void anotherCall() { calls++; };
  void anotherInstr() { totInstr++; };
  void anotherCompInstr() { compInstr++; };
  void anotherCompSucc() { compSucc++; };
  unsigned long statCalls() const { return calls; };
  unsigned long statTotInstr() const { return totInstr; };
  unsigned long statNonCompInstr() const { return totInstr - compInstr; };
  unsigned long statCompInstr() const { return compInstr; };
  unsigned long statCompSucc() const { return compSucc; };
  unsigned long statCompFail() const { return compInstr - compSucc; };  
 private: 
  unsigned long calls;
  unsigned long totInstr;
  unsigned long compInstr;
  unsigned long compSucc;
};
#endif
}; // namespace VK

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUBSUMPTION
#define DEBUG_NAMESPACE "ForwardSubsumption"
#endif 
#include "debugMacros.h"
//=================================================
namespace VK
{
class ForwardSubsumption : public FS_CODE, public FS_QUERY 

#ifdef FS_STAT
, public FS_STATISTICS
#endif

{
 public:
  ForwardSubsumption() 
    : FS_CODE(subst), 
    setMode(false), // default
    backtrack_lit(DOP("ForwardSubsumption::backtrack_lit")),
    backtrack_instr(DOP("ForwardSubsumption::backtrack_instr")),
    backtrack_cursor(DOP("ForwardSubsumption::backtrack_cursor"))
    {
      CALL("constructor ForwardSubsumption()");
      DOP(backtrack_lit.freeze());
    };
  ~ForwardSubsumption() {};
  void Destroy() { FS_CODE::Destroy(); };
#ifndef NO_DEBUG
  bool clean() const { return FS_CODE::clean(); };
#endif
  Clause* Subsume() 
    {
      CALL("Subsume()");
  
      ulong maxLitNum = FS_CODE::maxlNumOfLits();
    
      DOP(backtrack_lit.unfreeze()); 
      backtrack_lit.expand(maxLitNum);
      DOP(backtrack_lit.freeze());

      if (setMode) return SubsumeInSetMode();
      if (optimisedMultisetMode) return SubsumeInOptimisedMultisetMode();
      return SubsumeInMultisetMode();
    };
  void SetUseSigFilters(bool fl) 
    { 
      FS_QUERY::SetCollectSignature(fl);
      FS_CODE::SetUseSigFilters(fl);
    };
  void SetSetMode(bool fl) 
    { 
      setMode = fl; 
    };
  void SetOptimisedMultisetMode(bool fl)
    {
      optimisedMultisetMode = fl;
    }; 
#ifndef NO_DEBUG_VIS 
  ostream& outputQuery(ostream& str) { return FS_QUERY::output(str); };
#endif
 
 private:
  Clause* SubsumeInMultisetMode();
  Clause* SubsumeInOptimisedMultisetMode();
  Clause* SubsumeInSetMode();

 private:
  const Flatterm* subst[VampireKernelConst::MaxClauseSize];
  bool setMode;
  bool optimisedMultisetMode;

  ExpandingStack<GlobAlloc,LIT*,32UL,ForwardSubsumption> backtrack_lit;

  Stack<FS_COMMAND*,MAX_FS_CODE_DEPTH> backtrack_instr;
  Stack<const Flatterm*,MAX_FS_CODE_DEPTH> backtrack_cursor;
}; // class ForwardSubsumption  : public FS_CODE 
}; // namespace VK
//================================================
#undef DEBUG_NAMESPACE
//================================================
#endif
