/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         forward_superposition.cpp
// Description:  Representation of stored terms.
// Created:      Feb 14, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 04, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//               The array FPARA_PREMISE::_inherentlySelected
//               has been made extendable.
// Revised:      Apr 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::StoredConstraint 
//               replaced SharedKBConstr
// Revised:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::CompiledConstraintForUnifiers
//               replaced ORD_CONSTR
// Revised:      Jul 05, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Major refurbishment.
//====================================================
#include "forward_superposition.hpp"
using namespace BK;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "FORWARD_SUPERPOSITION"
#endif
#include "debugMacros.hpp"
//=================================================



VK::FORWARD_SUPERPOSITION::FORWARD_SUPERPOSITION(NewClause* res,OR_INDEX* unifIndex) : 
  // main parameters 
  _resolvent(res),
  _unifIndex(unifIndex),
  _noInfBetweenRulesFlag(false),
  
  // statistics
  _statSubsumedMainPremiseInstances(0UL),
  _statSubsumedSidePremiseInstances(0UL),
  _statDemodulatedMainSelectedLits(0UL),
  _statDemodulatedSideSelectedLits(0UL),

  // main premise data
  _mainPremiseVarIndex(0UL), 
  _mainSelectedLiterals(DOP("FORWARD_SUPERPOSITION::_mainSelectedLiterals")),
  _inherentlySelectedMainLit(DOP("FORWARD_SUPERPOSITION::_inherentlySelectedMainLit")),

  // current selected literal in the main premise
  _nonvarSubtermPositions("_nonvarSubtermPositions"),
  _compiledConstraint(0),
  
  // side premise data
  _sidePremiseVarIndex(1UL),
  _storedConstraintCheck(0), 
    
  // retrieval
  _pseudoHeader((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity)
{
  CALL("constructor FORWARD_SUPERPOSITION(NewClause* res,OR_INDEX* unifIndex)");
  
  _mainPremiseResidualSizeFormula.SetIndex(_mainPremiseVarIndex);
  _retrievalOfUnifiable.SetIndex(_sidePremiseVarIndex); 
}; 

VK::FORWARD_SUPERPOSITION::~FORWARD_SUPERPOSITION() 
{ 
  CALL("destructor ~FORWARD_SUPERPOSITION()"); 
  if (DestructionMode::isThorough())
    {
      if (_storedConstraintCheck)
	SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
      _storedConstraintCheck = 0;
      if (_compiledConstraint)
	{
	  SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
	};      
    };
};

void VK::FORWARD_SUPERPOSITION::init(NewClause* res,OR_INDEX* unifIndex) 
{
  CALL("init(NewClause* res,OR_INDEX* unifIndex)");
  
  // main parameters 
  _resolvent = res;
  _unifIndex = unifIndex;
  _noInfBetweenRulesFlag = false;

  // statistics
  _statSubsumedMainPremiseInstances = 0UL;  
  _statSubsumedSidePremiseInstances = 0UL;
  _statDemodulatedMainSelectedLits = 0UL; 
  _statDemodulatedSideSelectedLits = 0UL; 
  // main premise data
  _mainPremiseVarIndex = 0UL;
  _mainSelectedLiterals.init(DOP("FORWARD_SUPERPOSITION::_mainSelectedLiterals"));
  _inherentlySelectedMainLit.init(DOP("FORWARD_SUPERPOSITION::_inherentlySelectedMainLit"));
  _mainPremiseResidualSizeFormula.init();

  // current selected literal in the main premise
  _nonvarSubtermPositions.init("_nonvarSubtermPositions");
  _compiledConstraint = 0;

  // side premise data
  _sidePremiseVarIndex = 1UL; 
  _storedConstraintCheck = 0;
    

  // retrieval
  _retrievalOfUnifiable.init();
  _pseudoHeader.init((TERM::Arity)1,(TERM::Functor)0,TERM::PositivePolarity);
  
  
  DOP(_mainSelectedLiterals.freeze());
  DOP(_inherentlySelectedMainLit.freeze());

  _mainPremiseResidualSizeFormula.SetIndex(_mainPremiseVarIndex);
  _retrievalOfUnifiable.SetIndex(_sidePremiseVarIndex); 
}; // void VK::FORWARD_SUPERPOSITION::init(NewClause* res,OR_INDEX* unifIndex) 

void VK::FORWARD_SUPERPOSITION::destroy()
{
  CALL("destroy()");
  if (DestructionMode::isThorough())
    {
      if (_storedConstraintCheck)
	SimplificationOrdering::current()->releaseConstraintCheckOnUnifier(_storedConstraintCheck);
      _storedConstraintCheck = 0;
      if (_compiledConstraint)
	{
	  SimplificationOrdering::current()->releaseCompiledConstraintForUnifiers(_compiledConstraint);
	};      
    };


  // retrieval
  _pseudoHeader.destroy();
  _retrievalOfUnifiable.destroy();

  // current selected literal in the main premise  
  _nonvarSubtermPositions.destroy();

  // main premise data
  _mainPremiseResidualSizeFormula.destroy();
  _inherentlySelectedMainLit.destroy();
  _mainSelectedLiterals.destroy();
}; // void VK::FORWARD_SUPERPOSITION::destroy()



bool VK::FORWARD_SUPERPOSITION::Load(Clause* cl)
{
  CALL("Load(Clause* cl)");
  
  DOP(_mainSelectedLiterals.unfreeze());
  _mainSelectedLiterals.expand(cl->numOfAllLiterals());
  DOP(_mainSelectedLiterals.freeze());   


  DOP(_inherentlySelectedMainLit.unfreeze());
  _inherentlySelectedMainLit.expand(cl->numOfAllLiterals());
  DOP(_inherentlySelectedMainLit.freeze());   


  _equalityResolutionApplicable = false;
  _mainPremise = cl;
  _allMainPremiseLiterals = cl->LitList();
  _numOfSelectedMainLits = _allMainPremiseLiterals->numOfSelectedLits();
  _firstNonselectedMainLit = _allMainPremiseLiterals->nth(_numOfSelectedMainLits);

  LiteralList::Iterator iter(_allMainPremiseLiterals);  
  _mainSelectedLiterals.reset();
  _freeMainSelectedLitMemory = _mainSelectedLiteralMemory;
  for (ulong ln = 0; ln < _numOfSelectedMainLits; ln++)
    {
      ASSERT(iter.notEnd());
      _mainSelectedLiterals.push(_freeMainSelectedLitMemory);
      StandardTermToPrefix(_mainPremiseVarIndex,iter.currentLiteral(),_freeMainSelectedLitMemory);
      ASSERT(_freeMainSelectedLitMemory - _mainSelectedLiteralMemory <= (long)VampireKernelConst::MaxClauseSize);
      _inherentlySelectedMainLit[ln] = iter.currentElement()->isInherentlySelected();
      if ((iter.currentLiteral()->Negative())
	  && (iter.currentLiteral()->IsEquality())
	  && (iter.currentLiteral()->IsUnorderedEq()))
	_equalityResolutionApplicable = true;
       
      iter.next();   };
     
  compileMainPremiseResidualSizeFormula();
     
  resetIterationOverMainLits();



  _noRulesAsSidePremises = _noInfBetweenRulesFlag && cl->IsRule();
  PrefixSym* lit;
  bool inherentlySelected;
 try_next_lit:

  lit = nextSelectedMainLit(inherentlySelected);

  if (!lit) return false;
  if (LoadLiteral(lit,inherentlySelected)) { _continuation = ContFSup; return true; };
  goto try_next_lit; 
}; // bool VK::FORWARD_SUPERPOSITION::Load(Clause* cl)
 


bool VK::FORWARD_SUPERPOSITION::NextInference()
{ 
  CALL("NextInference()");
  switch (_continuation)
    {
    case ContFSup:
      while (NextFSupInference())
	if (CollectResult()) return true;
      // no more superpositions, start equality resolution 
      if (!_equalityResolutionApplicable) return false;
      resetIterationOverMainLits();
      _continuation = ContEqRes;
      // no break here, proceed with the next case
    case ContEqRes:
      while (NextEqRes())
	{
	  if (CollectEqResResult())
	    {
	      UndoEqRes();         
	      return true;
	    };
	  UndoEqRes(); 
	}; 
      // no more inferences
      return false;
    };
  ICP("ICP1");
  return false;
}; // bool VK::FORWARD_SUPERPOSITION::NextInference()




void VK::FORWARD_SUPERPOSITION::compileMainPremiseResidualSizeFormula()
{
  CALL("compileMainPremiseResidualSizeFormula()");
  _mainPremiseResidualSizeFormula.Reset();

  for(LiteralList::Iterator iter(_allMainPremiseLiterals,
				 _firstNonselectedMainLit);
      iter.notEnd();
      iter.next())
    if (!iter.currentLiteral()->isAnswerLiteral())
      _mainPremiseResidualSizeFormula.Compile(iter.currentLiteral());


     
  resetIterationOverMainLits();
  for (PrefixSym* m_lit = nextSelectedMainLitMemory(); 
       m_lit; 
       m_lit = nextSelectedMainLitMemory())
    {
      TERM header = m_lit->Head().Func();
      if (header.IsEquality() && header.IsOrderedEq()) _mainPremiseResidualSizeFormula.Compile(m_lit->Arg2()); 
    };
      
}; // void VK::FORWARD_SUPERPOSITION::compileMainPremiseResidualSizeFormula()



bool VK::FORWARD_SUPERPOSITION::NextFSupInference()
{
  if (Paramodulate()) return true;  
  PrefixSym* lit;
  bool inherentlySelected;
 try_next_lit:

  lit = nextSelectedMainLit(inherentlySelected);


  if (!lit) return false;
  if ((LoadLiteral(lit,inherentlySelected)) && (Paramodulate())) 
    return true;
  goto try_next_lit; 
}; // bool VK::FORWARD_SUPERPOSITION::NextFSupInference()

bool VK::FORWARD_SUPERPOSITION::NextEqRes()
{
  CALL("NextEqRes()");
  PrefixSym* lit; 
 try_next_lit:

  lit = nextSelectedMainLitMemory();


  if (!lit) return false;
      

  TERM header =  lit->Head().Func();
  if (header.Negative() && header.IsEquality() && header.IsUnorderedEq())
    {
      PrefixSym* arg1 = lit+1;
      PrefixSym* arg2 = arg1->Bracket();
      if (UNIFICATION::Unify(arg1,arg2)) return true;
    };
  goto try_next_lit; 
}; // bool VK::FORWARD_SUPERPOSITION::NextEqRes()



bool VK::FORWARD_SUPERPOSITION::LoadLiteral(PrefixSym* lit,bool inhSel)
{
  CALL("LoadLiteral(PrefixSym* lit,bool inhSel)");

  if (!lit->Head().Func().arity()) return false; // propositional
  // here lit must be a plain nonpropositional literal with all brackets set
  _currentMainLiteral = lit;
  _nonvarSubtermPositions.reset();
  if (lit->Head().Func().IsEquality())
    {
      _eqArg1 = lit + 1;
      _eqArg2 = _eqArg1->Bracket();

      if (lit->Head().Func().IsOrderedEq())
	{
	  _currentCompiledConstraint = 0;

	  PrefixSym* end_of_arg = _eqArg1->Bracket();
	  for (PrefixSym* pos = lit + 1; pos < end_of_arg; pos++)
	    if (pos->Head().IsFunctor()) 
	      {
		_nonvarSubtermPositions.push(pos);
	      };

      
	  _swapArguments = false;

	}
      else // unordered equality
	{ 
	  if (_eqArg1->Head().isVariable() || _eqArg2->Head().isVariable())
	    {   
	      _currentCompiledConstraint = 0;


	      PrefixSym* end_of_lit = lit->Bracket();
	      for (PrefixSym* pos = _eqArg1; pos < end_of_lit; pos++)
		if (pos->Head().IsFunctor()) _nonvarSubtermPositions.push(pos);


	      _swapArguments = false;

	    }
	  else // both arguments are complex 
	    {

	      if (!_compiledConstraint)
		_compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
	      ASSERT(_compiledConstraint);
	      _compiledConstraint->load(_eqArg2,_eqArg1);
	      if (_compiledConstraint->rhsVarInstMayOrder())
		{
		  _compiledConstraint->compileGreaterOrEqual();
		  _currentCompiledConstraint = _compiledConstraint;
		}
	      else
		{
		  _currentCompiledConstraint = 0;
		};


	      PrefixSym* end_of_arg1 = _eqArg1->Bracket();
	      for (PrefixSym* pos = _eqArg1; pos < end_of_arg1; pos++)
		if (pos->Head().IsFunctor()) _nonvarSubtermPositions.push(pos);

	      //eq_second_arg = _eqArg2;

	      _swapArguments = true;
        
	    };
	};
    }
  else // noneq. literal
    {
      _currentCompiledConstraint = 0;

      PrefixSym* end_of_lit = lit->Bracket();
      for (PrefixSym* pos = lit + 1; pos < end_of_lit; pos++)
	if (pos->Head().IsFunctor()) _nonvarSubtermPositions.push(pos);

      _swapArguments = false;

    };
 
  if (_nonvarSubtermPositions.empty()) return false;

  LoadTermToRewrite(_nonvarSubtermPositions.pop());
  _currentMainLitIsInherentlySelected = inhSel;
  return true; 
}; // bool VK::FORWARD_SUPERPOSITION::LoadLiteral(PrefixSym* lit,bool inhSel)


bool VK::FORWARD_SUPERPOSITION::LoadRestOfLit()
{
  CALL("LoadRestOfLit()");

  if (!_swapArguments) return false;
  _swapArguments = false;

  if (!_compiledConstraint)
    _compiledConstraint = SimplificationOrdering::current()->reserveCompiledConstraintForUnifiers();
  ASSERT(_compiledConstraint);
  _compiledConstraint->load(_eqArg1,_eqArg2);
  if (_compiledConstraint->rhsVarInstMayOrder())
    {
      _compiledConstraint->compileGreaterOrEqual();
      _currentCompiledConstraint = _compiledConstraint;
    }
  else
    {
      _currentCompiledConstraint = 0;
    };




  PrefixSym* end_of_arg2 = _eqArg2->Bracket();
  for (PrefixSym* pos = _eqArg2; pos < end_of_arg2; pos++)
    if (pos->Head().IsFunctor()) _nonvarSubtermPositions.push(pos);
    
  
  return (!_nonvarSubtermPositions.empty());
}; // bool VK::FORWARD_SUPERPOSITION::LoadRestOfLit()


bool VK::FORWARD_SUPERPOSITION::Paramodulate() 
{
  CALL("Paramodulate()");
 try_another_candidate:
  _currentCandidate = NextCandidate(_currentCompiledConstraint);
  if (_currentCandidate) 
    {
      _sidePremise = _currentCandidate->TheClause();
      if (_noRulesAsSidePremises && _sidePremise->IsRule()) 
	goto try_another_candidate;
      if (demodulateSelectedLits())
	goto try_another_candidate;

      if (subsumeSidePremise())
	goto try_another_candidate;
      return true; 
    }
  else // try another position
    {
    try_another_position:
      if (_nonvarSubtermPositions.empty()) 
	{
	  if (!LoadRestOfLit()) return false;
	};
      ASSERT(!_nonvarSubtermPositions.empty());
    
      PrefixSym* term_to_rewrite = _nonvarSubtermPositions.pop();
      LoadTermToRewrite(term_to_rewrite);
      _currentCandidate = NextCandidate(_currentCompiledConstraint);
      if (_currentCandidate) 
	{
	  _sidePremise = _currentCandidate->TheClause();
	  if (_noRulesAsSidePremises && _sidePremise->IsRule()) 
	    goto try_another_candidate;
	  if (demodulateSelectedLits())
	    goto try_another_candidate;
	  if (subsumeSidePremise())
	    goto try_another_candidate;
	  return true;
	};
      goto try_another_position;
    };
}; // bool VK::FORWARD_SUPERPOSITION::Paramodulate() 



VK::RES_PAIR_LIST* 
VK::FORWARD_SUPERPOSITION::NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr) 
{
  CALL("NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr)");
  RES_PAIR_LIST* result;

  switch (_paramodulationMode)
    {
    case Init:
      if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->FParaTree(_currentParamodulationTargetTerm->Head().Func()))))
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();

	  _mainSelectedLitDemodulationAttempted = false;
	  _paramodulationMode = ParaFromComplex;

	  if (constr && constr->holdsOnCurrentUnifier()) goto try_next_leaf;
	  if (subsumeMainPremise()) goto try_next_leaf;

	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
	  ASSERT_IN(result,"A10");
	  return result;
	};
      goto try_para_from_var; 
     
    case ParaFromComplex:
      result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
      if (result) return result;
    try_next_leaf:
      if (_retrievalOfUnifiable.FindNextLeaf())
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();

	  _mainSelectedLitDemodulationAttempted = false;
	  if (constr && constr->holdsOnCurrentUnifier()) goto try_next_leaf;  
	  if (subsumeMainPremise()) goto try_next_leaf;

	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
	  ASSERT_IN(result,"A20");
	  return result;
	};
      goto try_para_from_var;
        
    case ParaFromVar:
      result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
      if (result) return result;
      if (_retrievalOfUnifiable.FindNextLeaf())
	{
	  _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();
	  _mainSelectedLitDemodulationAttempted = false;
	  _retrievalOfUnifiable.ResetCandidatesFromLeaf();
	  result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
	  ASSERT_IN(result,"A30"); 
	  return result;
	};
      return 0;
    };
     
 try_para_from_var:
  _paramodulationMode = ParaFromVar;
  _retrievalOfUnifiable.CompilePseudoLiteral(_pseudoHeader,_currentParamodulationTargetTerm);
      
  if (_retrievalOfUnifiable.FindFirstLeaf(*(_unifIndex->FParaFromVarTree())))
    {
      _mainPremiseResidualSize = _mainPremiseResidualSizeFormula.Compute();
      _mainSelectedLitDemodulationAttempted = false;
      _retrievalOfUnifiable.ResetCandidatesFromLeaf();
      result = _retrievalOfUnifiable.NextCandidateFromTheLeaf();
      ASSERT_IN(result,"A20");
      return result;
    };
  return 0; 
}; // RES_PAIR_LIST* VK::FORWARD_SUPERPOSITION::NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr)  



bool VK::FORWARD_SUPERPOSITION::CollectResult()
{
  CALL("CollectResult()");

  if (!_mainPremise->Active()) return false;
  
  if (!_sidePremise->Active()) return false;
 

  const SimplificationOrdering::StoredConstraint* constr = 
    _currentCandidate->orderingConstraint();
  if (constr)
    {
      if (!_storedConstraintCheck)
	_storedConstraintCheck = SimplificationOrdering::current()->reserveConstraintCheckOnUnifier();
      ASSERT(_storedConstraintCheck);
      if (_storedConstraintCheck->holds(constr,_sidePremiseVarIndex))
	return false;
    };


  if (_mainPremiseResidualSize
      + _sidePremise->MinResWeight() 
      > _resolvent->MaxWeight())
    {
      _resolvent->AnotherTooHeavyClause();
      return false;
    }; 
      
  _resolvent->openGenerated(ClauseBackground::ForwSup);
  if (collectMainPremiseLiterals() && collectRewrittenSidePremise())
    {
      //_resolvent->SetBackground(ClauseBackground::RESOL);
      _resolvent->UsedRule(ClauseBackground::ForwSup);
      _resolvent->PushAncestor(_mainPremise);
      _resolvent->PushAncestor(_sidePremise);
      return true;
    }
  else return false; 
}; // bool VK::FORWARD_SUPERPOSITION::CollectResult()   
 
bool VK::FORWARD_SUPERPOSITION::CollectEqResResult()
{
  CALL("CollectEqResResult()");
  if (!_mainPremise->Active()) return false;
  _resolvent->openGenerated(ClauseBackground::EqRes);
  if (collectMainPremiseLiterals())
    {
      //_resolvent->SetBackground(ClauseBackground::RESOL);
      _resolvent->UsedRule(ClauseBackground::EqRes);
      _resolvent->PushAncestor(_mainPremise);
      return true;
    }
  else return false; 
}; // bool VK::FORWARD_SUPERPOSITION::CollectEqResResult() 



bool VK::FORWARD_SUPERPOSITION::collectRewrittenSidePremise()
{
  CALL("collectRewrittenSidePremise()");
  return 
    _resolvent->PushRewrittenLit(_currentMainLiteral,
				 _currentParamodulationTargetTerm,
				 _currentCandidate->ReplacingTerm(),
				 _sidePremiseVarIndex,
				 _currentMainLitIsInherentlySelected) &&
    CollectRestOfTheRewritingClause();
}; // bool VK::FORWARD_SUPERPOSITION::collectRewrittenSidePremise()


bool VK::FORWARD_SUPERPOSITION::CollectRestOfTheRewritingClause()
{
  CALL("CollectRestOfTheRewritingClause()");
  unsigned ln = 0;
  unsigned used_lit_num = _currentCandidate->LitNum();
  for (LiteralList::Iterator iter(_sidePremise->LitList());
       iter.notEnd();
       iter.next())
    {
      if (ln != used_lit_num) 
	if (!_resolvent->PushStandardLit(_sidePremiseVarIndex,iter.currentLiteral(),iter.currentElement()->isInherentlySelected())) return false;
      ln++;
    }; 
  return true;
}; // bool VK::FORWARD_SUPERPOSITION::CollectRestOfTheRewritingClause()


bool VK::FORWARD_SUPERPOSITION::collectMainPremiseLiterals()
{
  CALL("collectMainPremiseLiterals()");
  // all the selected literals except the rewritten one
  for (ulong ln = _numOfSelectedMainLits; ln; ln--)
    if ((ln - 1) != _nextMainSelectedLitNum)
      if (!_resolvent->PushPrefixLit(_mainSelectedLiterals.nth(ln-1),_inherentlySelectedMainLit[ln-1])) return false;     
     
  // all nonselected
  for(LiteralList::Iterator iter(_allMainPremiseLiterals,
				 _firstNonselectedMainLit);
      iter.notEnd();
      iter.next())
    {
      ASSERT(!(iter.currentElement()->isInherentlySelected()));
      if (!_resolvent->PushStandardLit(_mainPremiseVarIndex,iter.currentLiteral(),false)) return false; 
    }; 
  return true; 
}; // bool VK::FORWARD_SUPERPOSITION::collectMainPremiseLiterals()

bool VK::FORWARD_SUPERPOSITION::subsumeMainPremise()
{
  CALL("subsumeMainPremise()");
  Clause* subsumingClause;
  if (!_mainPremiseSubsumption) return false;
  if (_mainPremise->hasSubsumedPremiseInstance()) return false;
  _mainPremise->blockForSubsumption();
  ulong maxWeight = _resolvent->MaxWeight();
  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _resolvent->openGenerated(ClauseBackground::Unspecified);
  if (collectMainPremiseInstance())   
    {
      _resolvent->endOfClause();
      subsumingClause = _resolvent->FSubsume();
      ASSERT(subsumingClause != _mainPremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_resolvent->WasTautology()) goto discarded;
      goto return_false;
    };
  
 return_false:
  _mainPremise->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++_statSubsumedMainPremiseInstances;
 discarded:
  _mainPremise->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return true;
}; // bool VK::FORWARD_SUPERPOSITION::subsumeMainPremise()


bool VK::FORWARD_SUPERPOSITION::collectMainPremiseInstance()
{
  CALL("collectMainPremiseInstance()");
  for(LiteralList::Iterator iter(_mainPremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!(_resolvent->PushStandardLit(_mainPremiseVarIndex,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) 
      return false; 
  return true;
}; // bool VK::FORWARD_SUPERPOSITION::collectMainPremiseInstance()



bool VK::FORWARD_SUPERPOSITION::subsumeSidePremise()
{
  CALL("subsumeSidePremise()");
  
  Clause* subsumingClause;
  if (!_sidePremiseSubsumption) return false;
  ASSERT(_sidePremise == _currentCandidate->TheClause());
  if (_sidePremise->hasSubsumedPremiseInstance()) return false;
  _sidePremise->blockForSubsumption();
  ulong maxWeight = _resolvent->MaxWeight();
  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
  _resolvent->openGenerated(ClauseBackground::Unspecified);
  if (collectSidePremiseInstance())   
    {
      _resolvent->endOfClause();
      subsumingClause = _resolvent->FSubsume();
      ASSERT(subsumingClause != _sidePremise);
      if (subsumingClause) goto subsumed;
      goto return_false;
    }
  else // analyse the reason
    {
      if (_resolvent->WasTautology()) goto discarded;
      goto return_false;
    };
  
 return_false:
  _sidePremise->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return false;

 subsumed:
  subsumingClause->registerPremiseInstanceSubsumption();
  ++_statSubsumedSidePremiseInstances;
 discarded:
  _sidePremise->unblockForSubsumption();
  _resolvent->SetMaxWeight(maxWeight);
  return true;
}; // bool VK::FORWARD_SUPERPOSITION::subsumeSidePremise()


bool VK::FORWARD_SUPERPOSITION::collectSidePremiseInstance()
{
  CALL("collectSidePremiseInstance()");  
  for(LiteralList::Iterator iter(_sidePremise->LitList());
      iter.notEnd();
      iter.next())     
    if (!(_resolvent->PushStandardLit(_sidePremiseVarIndex,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) 
      return false; 
  return true;  
}; // bool VK::FORWARD_SUPERPOSITION::collectSidePremiseInstance()


bool VK::FORWARD_SUPERPOSITION::demodulateSelectedLits()
{
  CALL("demodulateSelectedLits()");
  if (_mainPremise->paramodulationsIntoProtected() &&
      _sidePremise->Unit())
    return false; // paramodulations into this clause are protected

  Clause* simplifyingClause;
  if (_mainSelectedLitDemodulation)
    {   
      if (_mainSelectedLitDemodulationAttempted)
	{
	  if (_mainSelectedLitDemodulated)
	    {
	      ++_statDemodulatedMainSelectedLits;
	      return true;
	    };
	  // proceed with the side premise
	}
      else
	{
	  _mainSelectedLitDemodulationAttempted = true;
	  _mainPremise->blockForDemodulationFrom();
	  _sidePremise->blockForDemodulationFrom();
	  ulong maxWeight = _resolvent->MaxWeight();
	  _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
	  _resolvent->openGenerated(ClauseBackground::Unspecified);
	  if (collectMainPremiseSubtermsAvailableForParamodulationInto())
	    {	   
	      _resolvent->endOfClause();   	
      
	      simplifyingClause = _resolvent->canBeDemodulated();
	      ASSERT(simplifyingClause != _mainPremise);
	      ASSERT(simplifyingClause != _sidePremise);
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      _resolvent->SetMaxWeight(maxWeight);
	      
	      if (simplifyingClause)
		{	
		  _mainPremise->protectParamodulationsInto();
		  _sidePremise->protectParamodulationsInto();
		  ++_statDemodulatedMainSelectedLits;
		  _mainSelectedLitDemodulated = true;
		  return true;
		}
	      else
		_mainSelectedLitDemodulated = false;
	      // proceed with the side premise
	    }
	  else // analyse the result
	    {
	      _mainPremise->unblockForDemodulationFrom();
	      _sidePremise->unblockForDemodulationFrom();
	      _resolvent->SetMaxWeight(maxWeight);
	      return _resolvent->WasTautology();
	    };
	};
    };

  if (_sideSelectedLitDemodulation)
    {
      _mainPremise->blockForDemodulationFrom();
      _sidePremise->blockForDemodulationFrom();
      ulong maxWeight = _resolvent->MaxWeight();
      _resolvent->SetMaxWeight(VampireKernelConst::MaxClauseSize); 
      _resolvent->openGenerated(ClauseBackground::Unspecified);
      if (collectSidePremiseSubtermsAvailableForParamodulationInto())
	{	   
	  _resolvent->endOfClause();   	
      
	  simplifyingClause = _resolvent->canBeDemodulated();
	  ASSERT(simplifyingClause != _mainPremise);
	  ASSERT(simplifyingClause != _sidePremise);
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  _resolvent->SetMaxWeight(maxWeight);
	      
	  if (simplifyingClause)
	    {	
	      _mainPremise->protectParamodulationsInto();
	      _sidePremise->protectParamodulationsInto();
	      ++_statDemodulatedSideSelectedLits;
	      return true;
	    }
	  else
	    return false;
	}
      else // analyse the result
	{
	  _mainPremise->unblockForDemodulationFrom();
	  _sidePremise->unblockForDemodulationFrom();
	  _resolvent->SetMaxWeight(maxWeight);
	  return _resolvent->WasTautology();
	};
    }
  else
    return false;
}; // bool VK::FORWARD_SUPERPOSITION::demodulateSelectedLits()




bool 
VK::FORWARD_SUPERPOSITION::collectMainPremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectMainPremiseSubtermsAvailableForParamodulationInto()");  
  bool atLeastOneLiteralCollected = false;
  for(LiteralList::Iterator iter(_allMainPremiseLiterals);
      iter.currentElement() != _firstNonselectedMainLit;
      iter.next())
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_resolvent->pushPseudoLit(_mainPremiseVarIndex,
					     TERM::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_resolvent->PushStandardLit(_mainPremiseVarIndex,lit,false))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_resolvent->PushStandardLit(_mainPremiseVarIndex,lit,false))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
    };
  return atLeastOneLiteralCollected;
}; // bool VK::FORWARD_SUPERPOSITION::collectMainPremiseSubtermsAvailableForParamodulationInto()



bool 
VK::FORWARD_SUPERPOSITION::collectSidePremiseSubtermsAvailableForParamodulationInto()
{
  CALL("collectSidePremiseSubtermsAvailableForParamodulationInto()"); 
  ulong numOfSelectedLits = _sidePremise->LitList()->numOfSelectedLits();
  bool atLeastOneLiteralCollected = false;
  ulong numOfExaminedLits = 0UL;
  LiteralList::Iterator iter(_sidePremise->LitList());
  while (numOfExaminedLits < numOfSelectedLits)
    {
      TERM* lit = iter.currentLiteral();
      if (lit->IsEquality())
	{
	  if (lit->IsOrderedEq())
	    {
	      // only the first argument
	      if (!_resolvent->pushPseudoLit(_sidePremiseVarIndex,
					     TERM::dummyPositiveHeader1(),
					     *(lit->Args())))
		return false;
	    }
	  else
	    if (!_resolvent->PushStandardLit(_sidePremiseVarIndex,lit,false))
	      return false;
	  atLeastOneLiteralCollected = true;
	}
      else
	if (lit->arity())
	  {
	    if (!_resolvent->PushStandardLit(_sidePremiseVarIndex,lit,false))
	      return false;
	    atLeastOneLiteralCollected = true;
	  };
      iter.next();
      ++numOfExaminedLits;
    };
  return atLeastOneLiteralCollected;
}; // bool VK::FORWARD_SUPERPOSITION::collectSidePremiseSubtermsAvailableForParamodulationInto()


//=======================================================================
