/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         forward_superposition.hpp
// Description:  Representation of stored terms.
// Created:      Feb 14, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 04, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//               The array FPARA_PREMISE::_inherentlySelected
//               has been made extendable.
// Revised:      Dec 22, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               In FPARA_PREMISE it was implicitely assumed that 
//               <num of lits>*HellConst::MaxTermSize <= HellConst::MaxClauseSize
//               which is now wrong since HellConst::MaxClauseSize is
//               not defined via HellConst::MaxNumOfLiterals.
// Revised:      Apr 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::StoredConstraint 
//               replaced SharedKBConstr
// Revised:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::CompiledConstraintForUnifiers
//               replaced ORD_CONSTR
// Revised:      Jul 05, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Major refurbishment.
// Revised:      Jul 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bugfix in nextSelectedMainLitMemory()
//====================================================
#ifndef FORWARD_SUPERPOSITION_H
//============================================================
#define FORWARD_SUPERPOSITION_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Stack.hpp"
#include "para_machine.hpp"
#include "NewClause.hpp"
#include "weight_formula.hpp"
#include "Array.hpp"
#include "ExpandingStack.hpp"
#include "GlobAlloc.hpp"
#include "SimplificationOrdering.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "FORWARD_SUPERPOSITION"
#endif
#include "debugMacros.hpp"
//=================================================

namespace VK
{
  class FORWARD_SUPERPOSITION
  {
  public:
    FORWARD_SUPERPOSITION(NewClause* res,OR_INDEX* unifIndex);
    ~FORWARD_SUPERPOSITION();
    void init(NewClause* res,OR_INDEX* unifIndex);
    void destroy();

    void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;   
    };
  
    void setMainPremiseSubsumption(bool fl)
    {
      _mainPremiseSubsumption = fl;
    };
    void setSidePremiseSubsumption(bool fl)
    {
      _sidePremiseSubsumption = fl;
    };
    
    void setMainSelectedLitDemodulation(bool fl)
    {
      _mainSelectedLitDemodulation = fl;
    };
    void setSideSelectedLitDemodulation(bool fl)
    {
      _sideSelectedLitDemodulation = fl;
    };

  
    ulong statSubsumedMainPremiseInstances() const
    {
      return _statSubsumedMainPremiseInstances;
    };
    ulong statSubsumedSidePremiseInstances() const
    {
      return _statSubsumedSidePremiseInstances;
    };

    ulong statDemodulatedMainSelectedLits() const
    {
      return _statDemodulatedMainSelectedLits;
    };
    ulong statDemodulatedSideSelectedLits() const
    {
      return _statDemodulatedSideSelectedLits;
    };
    

    bool Load(Clause* cl);
    bool NextInference();

  private: 
    enum Continuation { ContFSup, ContEqRes };
    enum ParamodulationMode { Init, ParaFromComplex, ParaFromVar };

  private:
    void compileMainPremiseResidualSizeFormula();
    bool NextFSupInference();
    bool NextEqRes();
    void UndoEqRes()
    {
      Unifier::current()->UnbindLastSection();
    }; 
    void resetIterationOverMainLits() 
    { 
      _nextMainSelectedLitNum = _numOfSelectedMainLits; 
    };
    PrefixSym* nextSelectedMainLit(bool& inherentlySelected)
    {
      if (_nextMainSelectedLitNum)
	{
	  _nextMainSelectedLitNum--;
	  inherentlySelected = _inherentlySelectedMainLit[_nextMainSelectedLitNum];
	  return _mainSelectedLiterals.nth(_nextMainSelectedLitNum);
	};
      return 0;
    }; 
    PrefixSym* nextSelectedMainLitMemory()
    {
      if (_nextMainSelectedLitNum)
	{
	  _nextMainSelectedLitNum--;
	  return _mainSelectedLiterals.nth(_nextMainSelectedLitNum);
	};
      return 0;
    }; 
    bool LoadLiteral(PrefixSym* lit,bool inhSel);
    bool LoadRestOfLit();
    void LoadTermToRewrite(PrefixSym* t)
    {
      CALL("LoadTermToRewrite(PrefixSym* t)");
      // t must be a plain complex term with all brackets set
      _currentParamodulationTargetTerm = t;
      _retrievalOfUnifiable.CompileComplexTerm(_currentParamodulationTargetTerm);
      _paramodulationMode = Init;
    };
    bool Paramodulate();
    RES_PAIR_LIST* NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr);  
    bool CollectResult();
    bool CollectEqResResult();
    bool collectRewrittenSidePremise();
    bool CollectRestOfTheRewritingClause(); 
    bool collectMainPremiseLiterals();


    bool subsumeMainPremise();
    bool collectMainPremiseInstance();
    bool subsumeSidePremise();
    bool collectSidePremiseInstance();
    
    bool demodulateSelectedLits();
    bool collectMainPremiseSubtermsAvailableForParamodulationInto();
    bool collectSidePremiseSubtermsAvailableForParamodulationInto();

  private:
    // main parameters    
    NewClause* _resolvent;
    OR_INDEX* _unifIndex;
    bool _mainPremiseSubsumption;
    bool _sidePremiseSubsumption;
    bool _mainSelectedLitDemodulation;
    bool _sideSelectedLitDemodulation;
    bool _noInfBetweenRulesFlag; 

    // statistics
    ulong _statSubsumedMainPremiseInstances;
    ulong _statSubsumedSidePremiseInstances;
    ulong _statDemodulatedMainSelectedLits;
    ulong _statDemodulatedSideSelectedLits;

    // main premise data
    Clause* _mainPremise;
    ulong _mainPremiseVarIndex;
    LiteralList* _allMainPremiseLiterals;
    LiteralList::Element* _firstNonselectedMainLit;
    ulong _numOfSelectedMainLits;
    bool _equalityResolutionApplicable;
    PrefixSym _mainSelectedLiteralMemory[VampireKernelConst::MaxClauseSize];
    PrefixSym* _freeMainSelectedLitMemory;
    BK::ExpandingStack<BK::GlobAlloc,PrefixSym*,32UL,FORWARD_SUPERPOSITION> _mainSelectedLiterals;
    BK::Array<BK::GlobAlloc,bool,32UL,FORWARD_SUPERPOSITION> _inherentlySelectedMainLit;    
    WEIGHT_FORMULA _mainPremiseResidualSizeFormula;
    long _mainPremiseResidualSize;
    bool _mainSelectedLitDemodulationAttempted;
    bool _mainSelectedLitDemodulated;

    // current selected literal in the main premise
    ulong _nextMainSelectedLitNum;
    PrefixSym* _currentMainLiteral;
    bool _currentMainLitIsInherentlySelected;
    BK::Stack<PrefixSym*,VampireKernelConst::MaxTermSize> _nonvarSubtermPositions;
    SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
    SimplificationOrdering::CompiledConstraintForUnifiers* _currentCompiledConstraint;   
    PrefixSym* _eqArg1;
    PrefixSym* _eqArg2;
    bool _swapArguments;

    // side premise data
    Clause* _sidePremise;    
    ulong _sidePremiseVarIndex;
    SimplificationOrdering::ConstraintCheckOnUnifier* _storedConstraintCheck;

    // retrieval
    PARA_MACH _retrievalOfUnifiable;
    PrefixSym* _currentParamodulationTargetTerm;
    TERM _pseudoHeader;
    RES_PAIR_LIST* _currentCandidate; 
    bool _noRulesAsSidePremises;

    // control
    ParamodulationMode _paramodulationMode;
    Continuation _continuation;

  }; // class FORWARD_SUPERPOSITION
}; // namespace VK

//=================================================================
#endif
