/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

#include "GlobalStopFlag.h"
#include "iu_machine.h"
#include "unification.h"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "IU_MACH"
#endif
#include "debugMacros.h"
//=================================================

namespace VK
{
inline bool IU_MACH::CompleteSearch()
{
  CALL("CompleteSearch()"); 
 
  IU_COMMAND* com = command;
  bool same_var;
  bool free_var;
  void* bind;
  
  
 check_tag:
  
  ASSERT((com > Code()) && (com < end_of_code));
  ASSERT((com->Tag() == IU_COMMAND::END) || (!(proc.CurrNode()->IsLeaf())));
  
   
  switch (com->Tag())
    {
    case IU_COMMAND::START : 
      ICP("R0");
      return false;
    case IU_COMMAND::VAR :
      bind = com->Symbol().var()->Unref(free_var);
      if (free_var)
	{
	  if (proc.BindWithOccCheck((Unifier::Variable*)bind,same_var))
	    {
              _backtrackPoints.push(com);

              if (same_var) 
               { 
		_backtrackActions.push(ALTER_BINDING_WITH_OC_SAME_VAR); 
               }
	      else
		_backtrackActions.push(ALTER_BINDING_WITH_OC); 

	      //PushBacktrackPoint(com);
              //if (same_var) { PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR); }
              //else { PushBacktrackAction(ALTER_BINDING_WITH_OC); };

	      com++;
          
	      goto check_tag;  
	    }
	  else { goto backtrack; };
	}
      else
	{
       
	  ASSERT(PREFIX::CorrectComplexTerm((PREFIX_SYM*)bind));
       
	  if (proc.ExtractUnifiable((PREFIX_SYM*)bind))
	    {
              _backtrackPoints.push(com);
              _backtrackActions.push(ALTER_UNIFIABLE);

	      //PushBacktrackPoint(com);
	      //PushBacktrackAction(ALTER_UNIFIABLE);

	      com++;
          
	      goto check_tag;
	    }
	  else 
	    {
	      goto backtrack; 
	    };
	};
      
    case IU_COMMAND::VAR_FIRST_OCC :
      proc.Bind(com->Symbol().var());

      _backtrackPoints.push(com);
      _backtrackActions.push(ALTER_BINDING);

      //PushBacktrackPoint(com);
      //PushBacktrackAction(ALTER_BINDING);

      com++;
      
      goto check_tag;  
   
    case IU_COMMAND::FUNC :
    func_check_node:
    switch (proc.CurrNode()->Tag())
      {
      case OR_INDEX_TREE_NODE::FUNC : 
	if (com->Symbol() == proc.CurrNode()->Symbol())
	  {

            _backtrackPoints.push(com);
            _backtrackActions.push(TRY_ALTERNATIVE_NODE);

            //PushBacktrackPoint(com);
            //PushBacktrackAction(TRY_ALTERNATIVE_NODE);

            com++;
            proc.Below();
            
            goto check_tag;
	  }
	else
	  {
	  func_try_alternative:
	    proc.Alternative();
	    if (proc.CurrNode()) { goto func_check_node;}
	    else { goto backtrack; };
	  };
      case OR_INDEX_TREE_NODE::CONST : goto func_try_alternative;
      case OR_INDEX_TREE_NODE::VAR :
	bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
	if (free_var)
	  {
            if (OCCUR_CHECK::Occurs((Unifier::Variable*)bind,com->Term()))
	      {
		goto func_try_alternative;
	      }
            else // assignment possible
	      {
		Unifier::current()->BindAlone((Unifier::Variable*)bind,com->Term());

                _backtrackPoints.push(com);
                _backtrackActions.push(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

		//PushBacktrackPoint(com);
		//PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

		com = com->NextCommand();
		proc.Below();
              
		goto check_tag;
	      };
	  }
	else // complex term, unification needed
	  {
            if (UNIFICATION::UnifyComplex((PREFIX_SYM*)bind,com->Term()))
	      {

                _backtrackPoints.push(com);
                _backtrackActions.push(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);

		//PushBacktrackPoint(com);
		//PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);


		com = com->NextCommand();
		proc.Below();
              
		goto check_tag;
	      }
            else { goto func_try_alternative; };
	  };
      case OR_INDEX_TREE_NODE::VAR_FIRST_OCC :
	Unifier::current()->BindAlone(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());

        _backtrackPoints.push(com);
        _backtrackActions.push(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

	//PushBacktrackPoint(com);
	//PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);


	com = com->NextCommand();
	proc.Below();
          
	goto check_tag;
      case OR_INDEX_TREE_NODE::LEAF :
	ICP("R1");
	return false;
          
#ifdef DEBUG_NAMESPACE
      default : 
	ICP("A1");
	return false;    
#endif
      };
     
    case IU_COMMAND::FUNC_PLAIN :
    func_plain_check_node:
    switch (proc.CurrNode()->Tag())
      {
      case OR_INDEX_TREE_NODE::FUNC : 
	if (com->Symbol() == proc.CurrNode()->Symbol())
	  {

            _backtrackPoints.push(com);
            _backtrackActions.push(TRY_ALTERNATIVE_NODE);

            //PushBacktrackPoint(com);
            //PushBacktrackAction(TRY_ALTERNATIVE_NODE);


            com++;
            proc.Below();
            
            goto check_tag;
	  }
	else
	  {
	  func_plain_try_alternative:
	    proc.Alternative();
	    if (proc.CurrNode()) { goto func_plain_check_node;}
	    else { goto backtrack; };
	  };
      case OR_INDEX_TREE_NODE::CONST : goto func_plain_try_alternative;
      case OR_INDEX_TREE_NODE::VAR :
	bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
	if (free_var)
	  {
            Unifier::current()->BindAlone((Unifier::Variable*)bind,com->Term());

            _backtrackPoints.push(com);
            _backtrackActions.push(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

            //PushBacktrackPoint(com);
            //PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

            com = com->NextCommand();
            proc.Below();
            
            goto check_tag;
	  }
	else // complex term
	  {
            if (UNIFICATION::UnifyPlain((PREFIX_SYM*)bind,com->Term()))
	      {

                _backtrackPoints.push(com);
                _backtrackActions.push(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);

		//PushBacktrackPoint(com);
		//PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);

		com = com->NextCommand();
		proc.Below();
		goto check_tag;
	      }
            else { goto func_plain_try_alternative; }; 
	  };
      case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
	Unifier::current()->BindAlone(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());

        _backtrackPoints.push(com);
        _backtrackActions.push(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

	//PushBacktrackPoint(com);
	//PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

	com = com->NextCommand();
	proc.Below();
          
	goto check_tag;
      case OR_INDEX_TREE_NODE::LEAF :
	ICP("S1");
	return false;
#ifdef DEBUG_NAMESPACE
      default : 
	ICP("L1");
	return false;    
#endif
      }; 
      
    /************/
      
    case IU_COMMAND::FUNC_GROUND :
    func_ground_check_node:
    switch (proc.CurrNode()->Tag())
      {
      case OR_INDEX_TREE_NODE::FUNC : 
	if (com->Symbol() == proc.CurrNode()->Symbol())
	  {

            _backtrackPoints.push(com);
            _backtrackActions.push(TRY_ALTERNATIVE_NODE);

            //PushBacktrackPoint(com);
            //PushBacktrackAction(TRY_ALTERNATIVE_NODE);

            com++;
            proc.Below();
            goto check_tag;
	  }
	else
	  {
	  func_ground_try_alternative:
	    proc.Alternative();
	    if (proc.CurrNode()) { goto func_ground_check_node;}
	    else { goto backtrack; };
	  };
      case OR_INDEX_TREE_NODE::CONST : goto func_ground_try_alternative;
      case OR_INDEX_TREE_NODE::VAR :
	bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
	if (free_var)
	  {
            Unifier::current()->BindAlone((Unifier::Variable*)bind,com->Term());

            _backtrackPoints.push(com);
            _backtrackActions.push(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

            //PushBacktrackPoint(com);
            //PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);


            com = com->NextCommand();
            proc.Below();
            goto check_tag;
	  }
	else // complex term
	  {
            if (UNIFICATION::MatchGroundPlain((PREFIX_SYM*)bind,com->Term()))
	      {

                _backtrackPoints.push(com);
                _backtrackActions.push(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);

		//PushBacktrackPoint(com);
		//PushBacktrackAction(UNBIND_SECTION_TRY_ALTERNATIVE_NODE);

		com = com->NextCommand();
		proc.Below();
		goto check_tag;
	      }
            else { goto func_ground_try_alternative; }; 
	  };
      case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
	Unifier::current()->BindAlone(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());

        _backtrackPoints.push(com);
        _backtrackActions.push(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

	//PushBacktrackPoint(com);
	//PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);


	com = com->NextCommand();
	proc.Below();
	goto check_tag;
      case OR_INDEX_TREE_NODE::LEAF :
	ICP("R2");
	return false;
#ifdef DEBUG_NAMESPACE
      default : 
	ICP("A2");
	return false; 
#endif
      };
     
    case IU_COMMAND::CONST :
    const_check_node:
    switch (proc.CurrNode()->Tag())
      {
      case OR_INDEX_TREE_NODE::FUNC :
      const_try_alternative:
      proc.Alternative();
      if (proc.CurrNode()) { goto const_check_node;}
      else { goto backtrack; };
      case OR_INDEX_TREE_NODE::CONST : 
	if (com->Symbol() == proc.CurrNode()->Symbol())
	  {

            _backtrackPoints.push(com);
            _backtrackActions.push(TRY_ALTERNATIVE_NODE);

            //PushBacktrackPoint(com);
            //PushBacktrackAction(TRY_ALTERNATIVE_NODE);

            com++;
            proc.Below();
            goto check_tag;
	  }
	else { goto const_try_alternative; };
      case OR_INDEX_TREE_NODE::VAR : 
	bind = (proc.Shift(proc.CurrNode()->Symbol().var()))->Unref(free_var);
	if (free_var)
	  {
            Unifier::current()->BindAlone((Unifier::Variable*)bind,com->Term());


            _backtrackPoints.push(com);
            _backtrackActions.push(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

            //PushBacktrackPoint(com);
            //PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

            com++;
            proc.Below();
            goto check_tag;
	  }
	else // complex term
	  {
            if (com->Symbol() == ((PREFIX_SYM*)bind)->Head())
	      {


                _backtrackPoints.push(com);
                _backtrackActions.push(TRY_ALTERNATIVE_NODE);

		//PushBacktrackPoint(com);
		//PushBacktrackAction(TRY_ALTERNATIVE_NODE);


		com++;
		proc.Below();
		goto check_tag;
	      }
            else { goto const_try_alternative; };
	  };
      case OR_INDEX_TREE_NODE::VAR_FIRST_OCC : 
	Unifier::current()->BindAlone(proc.Shift(proc.CurrNode()->Symbol().var()),com->Term());

        _backtrackPoints.push(com);
        _backtrackActions.push(UNBIND_LAST_TRY_ALTERNATIVE_NODE);

	//PushBacktrackPoint(com);
	//PushBacktrackAction(UNBIND_LAST_TRY_ALTERNATIVE_NODE);


	com++;
	proc.Below();
	goto check_tag;
      case OR_INDEX_TREE_NODE::LEAF :
	ICP("R3"); 
	return false;
#ifdef DEBUG_NAMESPACE
      default : 
	ICP("A3"); 
	return false;
#endif
      };
    case IU_COMMAND::END :
      ASSERT(proc.CurrNode()->IsLeaf());      
      return true;
#ifdef DEBUG_NAMESPACE
    default :
      ICP("A3"); 
      return false; 
#endif  
    };
 backtrack:
  command = com;
  if (Backtrack()) { com = command; goto check_tag; }
  else { return false; };
}; // bool IU_MACH::CompleteSearch() 


bool IU_MACH::FindFirstLeaf(OR_INDEX_TREE_NODE* tree)
{
  command = Code() + 1;
  if (tree)
    {
      _backtrackPoints.push(Code());
      _backtrackActions.push(FAILURE);
      //PushBacktrackPoint(Code());
      //PushBacktrackAction(FAILURE);

      proc.Start(tree);
      return CompleteSearch();
    }
  else 
    {
      return false; 
    };
}; // bool IU_MACH::FindFirstLeaf()

bool IU_MACH::FindNextLeaf()
{
  return ((Backtrack()) && (CompleteSearch()));
}; // inline bool IU_MACH::FindNextLeaf()


inline bool IU_MACH::Backtrack()
{
  CALL("Backtrack()");
 
  IU_COMMAND* com;
  bool same_var;
  Unifier::Variable* var;
  
  GlobalStopFlag::check();
  
 rollback:
  com = _backtrackPoints.pop();
   
  BACKTRACK_ACTION act = _backtrackActions.pop(); 
   
  // switch (PopBacktrackAction())
  switch (act) 
    {
    case FAILURE : return false;
    case ALTER_BINDING_WITH_OC : 
     
      ASSERT(com->Symbol().var()->Binding());
      if (proc.AlterBindingWithOccCheck(Unifier::current()->LastBound(),same_var))
	{

          _backtrackPoints.push(com);

	  //PushBacktrackPoint(com);
	  if (same_var)
	    {
	      _backtrackActions.push(ALTER_BINDING_WITH_OC_SAME_VAR);
	      //PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR);
	    }
	  else
	    {
	      ASSERT(com->Symbol().var()->Binding());
              _backtrackActions.push(ALTER_BINDING_WITH_OC);
	      //PushBacktrackAction(ALTER_BINDING_WITH_OC);
	    };
	  command = com + 1;
	  return true;
	}
      else 
	{
	  goto rollback; 
	};

    case ALTER_BINDING_WITH_OC_SAME_VAR :
      
      var = (Unifier::Variable*)(com->Symbol().var()->Unref());
      ASSERT(Unifier::current()->isVariable((void*)var));
      if (proc.AlterBindingWithOccCheck(var,same_var))
	{
          _backtrackPoints.push(com);
	  //PushBacktrackPoint(com);
	  if (same_var)
	    {
              _backtrackActions.push(ALTER_BINDING_WITH_OC_SAME_VAR);
	      //PushBacktrackAction(ALTER_BINDING_WITH_OC_SAME_VAR);
	    }
	  else
	    {
              _backtrackActions.push(ALTER_BINDING_WITH_OC);
	      //PushBacktrackAction(ALTER_BINDING_WITH_OC);
	    };
	  command = com + 1;
	  return true;
	}
      else 
	{
	  goto rollback; 
	};
      
    case ALTER_BINDING : 
     
      ASSERT(com->Tag() == IU_COMMAND::VAR_FIRST_OCC);
      ASSERT(com->Symbol().var()->Binding());
      
      if (proc.AlterBinding())
	{
	  ASSERT(com->Symbol().var()->Binding());

          _backtrackPoints.push(com);
          _backtrackActions.push(ALTER_BINDING);
	  //PushBacktrackPoint(com);
	  //PushBacktrackAction(ALTER_BINDING);


	  command = com + 1;
	  return true;
	}
      else 
	{
	  goto rollback; 
	};
      
    case ALTER_UNIFIABLE :
     
      if (proc.AlterUnifiable())
	{
          _backtrackPoints.push(com);
          _backtrackActions.push(ALTER_UNIFIABLE);
	  //PushBacktrackPoint(com);
	  //PushBacktrackAction(ALTER_UNIFIABLE);
	  command = com + 1;
	  return true;
	}
      else 
	{
	  goto rollback; 
	};
      
    case TRY_ALTERNATIVE_NODE :
      
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
	{
	  goto rollback; 
	};
      
    case UNBIND_LAST_TRY_ALTERNATIVE_NODE :
     
      Unifier::current()->UnbindLast();
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
	{
	  goto rollback; 
	};
      
    case UNBIND_SECTION_TRY_ALTERNATIVE_NODE :
     
      Unifier::current()->UnbindLastSection();
      proc.RestoreCurrent();
      proc.Alternative();
      if (proc.CurrNode()) { command = com; return true; }
      else 
	{
	  goto rollback; 
	};
    };   
  ICP("R1");
  return false;
}; // bool IU_MACH::Backtrack()


#ifndef NO_DEBUG_VIS
ostream& IU_MACH::output(ostream& str) const
{
  str << "CODE:\n";
  return outputCode(str) << "INDEX = " << GetIndex() << '\n';
}; // ostream& IU_MACH::output(ostream& str) const
 
ostream& IU_MACH::outputCode(ostream& str) const
{
  const IU_COMMAND* com;
  for (com = Code(); com->Tag() != IU_COMMAND::END; com++) com->output(str,Code()) << '\n';
  return com->output(str,Code()) << '\n';
}; // ostream& IU_MACH::outputCode(ostream& str) const 
#endif


}; // namespace VK


//==============================================================
