/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

#include "or_premise.h"

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_PREMISE
 #define DEBUG_NAMESPACE "OR_PREMISE"
#endif
#include "debugMacros.h"
//=================================================

namespace VK
{
OR_PREMISE::OR_PREMISE() 
  : _resolvent((NewClause*)0), 
  _termMemory(DOP("OR_PREMISE::_termMemory")),
  _maxLiteral(DOP("OR_PREMISE::_maxLiteral")),
  _litIsInherentlySelected(DOP("OR_PREMISE::_litIsInherentlySelected")),
  _litIsMarked(DOP("OR_PREMISE::_litIsMarked"))
{
  CALL("constructor OR_PREMISE()");
  DOP(_termMemory.freeze());
  DOP(_maxLiteral.freeze());
  DOP(_litIsInherentlySelected.freeze());
  DOP(_litIsMarked.freeze());
}; // OR_PREMISE::OR_PREMISE() 

void OR_PREMISE::init() 
{ 
  _resolvent = (NewClause*)0; 
  _termMemory.init(DOP("OR_PREMISE::_termMemory")); 
  DOP(_termMemory.freeze());
  _maxLiteral.init(DOP("OR_PREMISE::_maxLiteral"));
  DOP(_maxLiteral.freeze());
  _litIsInherentlySelected.init(DOP("OR_PREMISE::_litIsInherentlySelected"));
  DOP(_litIsInherentlySelected.freeze());

  _litIsMarked.init(DOP("OR_PREMISE::_litIsMarked")); 
  DOP(_litIsMarked.freeze());

}; // void OR_PREMISE::init() 

OR_PREMISE::OR_PREMISE(NewClause* res) 
  : _resolvent(res), 
  _termMemory(DOP("OR_PREMISE::_termMemory")),
  _maxLiteral(DOP("OR_PREMISE::_maxLiteral")),
  _litIsInherentlySelected(DOP("OR_PREMISE::_litIsInherentlySelected")),
  _litIsMarked(DOP("OR_PREMISE::_litIsMarked"))
{
  CALL("constructor OR_PREMISE(NewClause* res)"); 
  DOP(_termMemory.freeze());
  DOP(_maxLiteral.freeze());
  DOP(_litIsInherentlySelected.freeze());
  DOP(_litIsMarked.freeze());
}; // OR_PREMISE::OR_PREMISE(NewClause* res) 


void OR_PREMISE::init(NewClause* res) 
{ 
  _resolvent = res; 
  _termMemory.init(DOP("OR_PREMISE::_termMemory")); 
  DOP(_termMemory.freeze());
  _maxLiteral.init(DOP("OR_PREMISE::_maxLiteral"));
  DOP(_maxLiteral.freeze());
  _litIsInherentlySelected.init(DOP("OR_PREMISE::_litIsInherentlySelected"));
  DOP(_litIsInherentlySelected.freeze());

  _litIsMarked.init(DOP("OR_PREMISE::_litIsMarked")); 
  DOP(_litIsMarked.freeze());

}; // void OR_PREMISE::init(NewClause* res) 


void OR_PREMISE::Load(ulong ind,const Clause* c)
{
  CALL("Load(ulong ind,const Clause* c)");
  ASSERT(!c->isAnswer());
  DOP(_termMemory.unfreeze());
  _termMemory.expand(c->size()); 
  DOP(_termMemory.freeze());

  _freeTermMemory = _termMemory.memory();
  //free_memory = memory;

  DOP(_maxLiteral.unfreeze());
  _maxLiteral.expand(c->numOfAllLiterals());
  DOP(_maxLiteral.freeze()); 

  DOP(_litIsInherentlySelected.unfreeze());
  _litIsInherentlySelected.expand(c->numOfAllLiterals());
  DOP(_litIsInherentlySelected.freeze());

  DOP(_litIsMarked.unfreeze());
  _litIsMarked.expand(c->numOfAllLiterals());
  DOP(_litIsMarked.freeze());

  clause = c;
  index = ind;

  _allLiterals = clause->LitList(); 
  _firstNonselectedLit = _allLiterals->nth(_allLiterals->numOfSelectedLits());
  last_max = _allLiterals->numOfSelectedLits() - 1;
  ulong max_lit = 0; 
  for(LiteralList::Iterator iter(_allLiterals);iter.currentElement() != _firstNonselectedLit;iter.next())
    {
      _maxLiteral[max_lit] = _freeTermMemory;
      _litIsMarked[max_lit] = false;
      StandardTermToPrefix(ind,iter.currentLiteral(),_freeTermMemory); 
      _litIsInherentlySelected[max_lit] = iter.currentElement()->isInherentlySelected();
      max_lit++;
    };   
}; // void OR_PREMISE::Load(ulong ind,const Clause* c)

void OR_PREMISE::LoadOptimized(ulong ind,const Clause* c)
{
  CALL("LoadOptimized(ulong ind,const Clause* c)");
  // if there is only one maximal literal we do not transform it
  // into prefix representation
  
  ASSERT(!c->isAnswer());
  DOP(_termMemory.unfreeze());
  _termMemory.expand(c->size()); 
  DOP(_termMemory.freeze());
   
  _freeTermMemory = _termMemory.memory();
  //free_memory = memory;

  DOP(_maxLiteral.unfreeze());
  _maxLiteral.expand(c->numOfAllLiterals());
  DOP(_maxLiteral.freeze()); 

  DOP(_litIsInherentlySelected.unfreeze());
  _litIsInherentlySelected.expand(c->numOfAllLiterals());
  DOP(_litIsInherentlySelected.freeze());

  DOP(_litIsMarked.unfreeze());
  _litIsMarked.expand(c->numOfAllLiterals());
  DOP(_litIsMarked.freeze());

  clause = c;
  index = ind;

  _allLiterals = clause->LitList(); 
  _firstNonselectedLit = _allLiterals->nth(_allLiterals->numOfSelectedLits());
  last_max = _allLiterals->numOfSelectedLits() - 1;

  if (last_max)
    {
      ulong max_lit = 0; 
      for(LiteralList::Iterator iter(_allLiterals);iter.currentElement() != _firstNonselectedLit;iter.next())
	{
	  _maxLiteral[max_lit] = _freeTermMemory;
	  _litIsMarked[max_lit] = false;
	  StandardTermToPrefix(ind,iter.currentLiteral(),_freeTermMemory); 

	  _litIsInherentlySelected[max_lit] = iter.currentElement()->isInherentlySelected();
	  max_lit++;
	};   
    }
  else
    {
      _maxLiteral[0] = (PREFIX_SYM*)0; // just in case
      _litIsMarked[0] = false; // just in case
    };
}; // void OR_PREMISE::LoadOptimized(ulong ind,const Clause* c) 


#ifndef NO_DEBUG_VIS  
ostream& OR_PREMISE::output(ostream& str) const
{
  str <<  "<<<<<< PREMISE >>>>>>>\n"
      << " Clause : " << clause << "\n Index = " << index
      << "\n Maximal literals: \n";
  for (ulong ml = 0; ml <= last_max; ml++)
    {
      str << "  [" << _litIsMarked[ml] << "] ";
      PREFIX::outputTerm(str,_maxLiteral[ml]);
      str << '\n';
    };
  return str; 
}; // ostream& OR_PREMISE::output(ostream& str) const
#endif
}; // namespace VK


//=================================================
