/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
#ifndef OR_PREMISE_H
//=================================================================
#define OR_PREMISE_H
#ifndef NO_DEBUG_VIS 
#  include <iostream>
#endif
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "Clause.hpp"
#include "prefix.hpp" 
#include "or_index.hpp"
#include "NewClause.hpp"
#include "Array.hpp"
#include "GlobAlloc.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_OR_PREMISE
 #define DEBUG_NAMESPACE "OR_PREMISE"
#endif
#include "debugMacros.hpp"
//=================================================
namespace VK
{

class OR_PREMISE
 {
  public: // methods
   OR_PREMISE(); 
   OR_PREMISE(NewClause* res); 
   ~OR_PREMISE();
   void init();
   void init(NewClause* res); 
   void destroy();


   void setResolvent(NewClause* res) { _resolvent = res; };

   PrefixSym* nthMaxLit(ulong n) 
     {
       CALL("nthMaxLit(ulong n)");
       ASSERT(n <= last_max);
       return _maxLiteral[n]; 
     };
   Clause* TheClause() const { return clause; };

   ulong LastMax() { return last_max; };
   ulong Index() { return index; };
   ulong CutLit() { return cut_lit; };

   void Load(ulong ind,Clause* c);
   void LoadOptimized(ulong ind,Clause* c);
   //            if there is only one maximal literal we do not transform it
   //            into prefix representation
  
   void Unload() 
   {
   };

   void Cut(ulong lit) 
    {
     CALL("Cut(ulong lit)");
     cut_lit = lit;
     _litIsMarked[cut_lit] = true;
    };
   void Load(ulong ind,RES_PAIR_LIST* res_pair)
    {
     CALL("Load(ulong ind,RES_PAIR_LIST* res_pair)");
     Load(ind,res_pair->TheClause());
     Cut((ulong)(res_pair->LitNum()));
    };
   void LoadOptimized(ulong ind,RES_PAIR_LIST* res_pair)
    {
     CALL("LoadOptimized(ulong ind,RES_PAIR_LIST* res_pair)");
     LoadOptimized(ind,res_pair->TheClause());
     Cut((ulong)(res_pair->LitNum()));
    };
   void Cut(TERM* lit) { Cut(clause->NumberOfLiteral(lit)); };
   void UnmarkCut() 
   { 
    CALL("UnmarkCut()");
    _litIsMarked[cut_lit] = false; 
   };
   bool CollectNonmarkedLiterals();
   bool collectAllLiterals(); 
   bool collectSubtermsAvailableForParamodulationInto(bool collectMarked);
   bool NonmarkedExist() { return nonmarkedExist; };
  public: // output for debugging
   #ifndef NO_DEBUG_VIS  
    ostream& output(ostream& str) const;
   #endif

  protected: // structure
   NewClause* _resolvent;
   Clause* clause;
   ulong index;
   
   BK::Array<BK::GlobAlloc,PrefixSym,256UL,OR_PREMISE> _termMemory;
   PrefixSym* _freeTermMemory;
   
   BK::Array<BK::GlobAlloc,PrefixSym*,32UL,OR_PREMISE> _maxLiteral;

   BK::Array<BK::GlobAlloc,bool,32UL,OR_PREMISE> _litIsInherentlySelected;

   ulong last_max;

   BK::Array<BK::GlobAlloc,bool,32UL,OR_PREMISE> _litIsMarked;
   
   LiteralList* _allLiterals;
   LiteralList::Element* _firstNonselectedLit;

   ulong cut_lit;
   bool nonmarkedExist;

 }; // class OR_PREMISE
 
//******************* Definitions: *************************************

inline bool OR_PREMISE::CollectNonmarkedLiterals()
 {
  CALL("CollectNonmarkedLiterals()");
  ulong currMax = 0;

  nonmarkedExist = (_firstNonselectedLit != 0);   

  while (currMax <= last_max)
   {   
    if (!(_litIsMarked[currMax])) 
     {
      nonmarkedExist = true;
      if (!(_resolvent->PushPrefixLit(_maxLiteral[currMax],_litIsInherentlySelected[currMax])))
       {
        return false; 
       };
     };
    currMax++;
   };
  
   
  for(LiteralList::Iterator iter(_allLiterals,_firstNonselectedLit);iter.notEnd();iter.next())     
   if (!(_resolvent->PushStandardLit(index,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) return false;
  
  return true;
 }; // inline bool OR_PREMISE::CollectNonmarkedLiterals()



inline bool OR_PREMISE::collectAllLiterals()
 {
  CALL("collectAllLiterals()");

  // works even if LoadOptimised was used

  for(LiteralList::Iterator iter(_allLiterals);iter.notEnd();iter.next())     
   if (!(_resolvent->PushStandardLit(index,iter.currentLiteral(),iter.currentElement()->isInherentlySelected()))) return false;
  
  return true;
 }; // inline bool OR_PREMISE::collectAllLiterals()




inline 
bool OR_PREMISE::collectSubtermsAvailableForParamodulationInto(bool collectMarked)
{
  CALL("collectSubtermsAvailableForParamodulationInto(bool collectMarked)");
  bool atLeastOneLiteralCollected = false;
  ulong litNum = 0UL;
  bool markedCollected = !collectMarked;
  for(LiteralList::Iterator iter(_allLiterals);
      iter.currentElement() != _firstNonselectedLit;
      iter.next())
    {
      if ((!_litIsMarked[litNum]) || (!markedCollected))
	{
	  markedCollected = _litIsMarked[litNum];
	  TERM* lit = iter.currentLiteral();
	  if (lit->IsEquality())
	    {
	      if (lit->IsOrderedEq())
		{
		  // only the first argument
		  if (!_resolvent->pushPseudoLit(index,
						 TERM::dummyPositiveHeader1(),
						 *(lit->Args())))
		    return false;
		}
	      else
		if (!_resolvent->PushStandardLit(index,lit,false))
		  return false;
	      atLeastOneLiteralCollected = true;
	    }
	  else
	    if (lit->arity())
	      {
		if (!_resolvent->PushStandardLit(index,lit,false))
		  return false;
		atLeastOneLiteralCollected = true;
	      };
	};
      ++litNum;
    };
  return atLeastOneLiteralCollected;
}; // bool OR_PREMISE::collectSubtermsAvailableForParamodulationInto(bool collectMarked)

}; // namespace VK


//=================================================================
#endif
