/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002, 2003

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/
//
// File:         tmp_literal.cpp
// Description:  Temporary (buffer) literals.
// Created:      Dec 21, 2001 
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#include "tmp_literal.hpp"

using namespace BK;
using namespace VK;
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEW_CLAUSE
 #define DEBUG_NAMESPACE "TmpLiteral"
#endif 
#include "debugMacros.hpp"
//=================================================

TmpLiteral::Domain* TmpLiteral::_currentDomain; // implicitely initialised to 0



ostream& TmpLiteral::output(ostream& str) const
{
  if (_inherentlySelected) str << "[[InhSel]] ";
  if (isEquality())
    { 
      str << '(';
      arg1()->outputTerm(str) << ')' << header() << '(';
      arg2()->outputTerm(str);
      str << ')';
    }
  else 
    flatterm()->outputTerm(str);
  return str; 
}; // ostream& TmpLiteral::output(ostream& str) const


void* TmpLiteral::operator new(size_t)
{
  CALL("operator new(size_t)");
  void* res = static_cast<void*>(GlobAlloc::allocate(sizeof(TmpLiteral)));
  return res;
}; // void* TmpLiteral::operator new(size_t)

void TmpLiteral::operator delete(void* obj)
{
  CALL("operator delete(void*)");
  GlobAlloc::deallocate(obj,sizeof(TmpLiteral));
}; // void TmpLiteral::operator delete(void*)

void TmpLiteral::operator delete(void* obj,size_t)
{
  CALL("operator delete(void*,size_t)");
  GlobAlloc::deallocate(obj,sizeof(TmpLiteral));
}; // void TmpLiteral::operator delete(void*)



bool TmpLiteral::instantiateVar(TmpLiteral& new_lit,ulong var,const Flatterm* subst_term)
{
  CALL("instantiateVar(TmpLiteral& new_lit,ulong var,const Flatterm* subst_term)");
  TERM new_header = header();
  ASSERT(new_header.arity());
  if(new_header.IsEquality()) new_header.MakeUnordered();
  new_lit.reset();
  new_lit.pushNonpropHeader(new_header);
  Flatterm::Iterator iter(flatterm());
  Flatterm::ConstIterator subst_iter;
  iter.NextSym(); // skipping the header
  while (iter.NotEndOfTerm())
    {
      TERM sym = iter.CurrPos()->Symbol();  
      if (sym.isVariable() && sym.var() == var) 
	{
	  subst_iter.Reset(subst_term);             
	  do
	    {
	      if (!new_lit.pushSym((subst_iter.CurrPos())->symbolRef())) return false;             
	      subst_iter.NextSym();
	    }
	  while (subst_iter.NotEndOfTerm());
	}
      else if (!new_lit.pushSym(sym)) return false;
      iter.NextSym();
    };
  new_lit.endOfLiteral();
  return true;
}; // bool TmpLiteral::instantiateVar(TmpLiteral& new_lit,ulong var,const Flatterm* subst_term)


//============================================================================
