/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
//  file atom.h 
//  defines class Atom
//

#ifndef __atom__
#define __atom__


class Substitution;


#include "typedefs.h"
#include "int.h"
#include "term.h"
#include "signature.h"
#include "memory.h"
#include "clist.h"


// ****************** class Atom, definition *************************


class Atom 
#   if debug_memory
    : public Memory <CID_ATOM>
#   endif
{
 public:
  typedef CList <Atom*,CID_ATOM_LIST> List;

  // constructors
  Atom ( Signature::Pred* p, Term::List* args ) :
    _functor(p),
    _args (args),
    _count (0)
    { if (args->isNonEmpty()) 
        args->inc (); } 

  Atom ( Term* l, Term* r ); // for equality

  // reference counting
  void inc () { _count++; }
  void dec () 
    { ASSERT(debug_atom,_count > 0); 
      _count--; 
      if (_count == 0) 
        destroy ();
    }

  // query the structure
  Signature::Pred* functor () const { return _functor; } 
  Term::List* args () const { return _args; } 
  int arity () const { return _functor->arity (); } 

  bool isEquality () const 
    { return functor()->isEquality (); }
  // deallocation and copying
  void destroy ();

  // statistics
  bool occurs (Var v) const 
    { return Term::occurs (args(), v); }


/*
  Atom* copy () const;

  int countOccurrences ( const FSymbol* ) const;
  int countOccurrences ( const Var ) const;
  Var maxVar ();
  int weight () const;
  bool occurs ( const FSymbol* ) const;

  void rename ( VarSubst& );
*/

  // miscellaneous
  bool equal ( Atom* ) const;
  bool hasVarsNotIn (const VarListList*) const; 
  bool isFlat () const // atom has the form P (x1,...,xn) where xi are variables
    { return Term::varsOnly (args()); }
  Term* isDefinition ( Term*& rhs ) const; // the result is lhs of the definition
  Compare compare (const Atom* l) const; // comparison, used for normalisationy
  void normalize ();
  Atom* apply ( const Substitution& ) const;  

/*
  Atom* apply ( Subst&, VarSubst&, int& lastVar ); // copies the argument
  Atom* replaceRelationByFunction ( PSymbol* pred, FSymbol* fun, int argNo, int* vs );
  void replace ( Var, Term* );  // actually reuses the memory allocated for the term
  bool isLinearAndFlat () const; // atom has the form P (x1,...,xn) where xi are pairwise different
  void introduceInverse ( int argNo,FSymbol* inv );
  void splitLargeArity ( const PSymbol* p, FSymbol::List* newFs );
  void splitLargeArity ( const FSymbol* f, FSymbol::List* newFs );
  Atom* negativeEqualitySplit ( int thresholdWeight );
*/
  // functions for checking equality axioms
  bool isTautology () const;
  bool swap (const Atom* a) const; // true if this is t1 = t2 and a is t2 = t1
  static bool transitivity (const Atom* a1, const Atom* a2, const Atom* a3);
  static bool functionMonotonicity (const Atom* a1, const Atom* a2);
  static bool predicateMonotonicity (const Atom* a1, const Atom* a2, const Atom* a3);

 private: 
  // structure
  Signature::Pred* _functor;
  Term::List* _args;
  long _count;

  inline void functor ( Signature::Pred* f ) { _functor = f; } 
  inline void args ( Term::List* ts ) 
    { 
      ts->inc ();
      _args->dec ();
      _args = ts; 
    } 
}; // class Atom


ostream& operator << ( ostream& str, const Atom* t );


#endif // __atom__
