#!/usr/bin/perl

use File::Copy;
use File::Basename;

# Script for controlling the interaction of SETHEO and E
# Working principle:
# SETHEO and E are run in an oszillating fashion, After each
# individual run, unit lemmas are given to the next sytem.
# also, the run times are successively increased

#
# some general configurations, adapt to your system
#
$BINDIR = $ENV{BINDIR};
if ( $BINDIR eq "" ) {
    $BINDIR = ".";
}
print "BINDIR $BINDIR\n";
$WORKDIR = $ENV{WORKDIR};
if ( $WORKDIR eq "" ) {
    $WORKDIR = "/tmp";
}
print "WORKDIR $WORKDIR\n";

# Adapt paths etc:
$DCTP = "$BINDIR/dctp -silent -tptp-lemmas -inference-limit ";
$E = "$BINDIR/eprover --tptp-format --tptp-out -xAuto -tAuto --memory-limit=384 --print-saturated=eiEI --filter-saturated=ucF -s --error-on-empty --eqn-no-infix -C ";
$STEXPOSU = "$BINDIR/stexposu -tptpout ";

# Program variables:
$MAX_NR_LEMMAS = 10;
$RESULT = 0;
$RUN = 0;
$SLIMIT = 2000;
$DLIMIT = 100;
$ELIMIT = 250;
$FILE = shift(@ARGV) or die "no problem specified\n";
$REMAINING = shift(@ARGV) or $REMAINING = 300;
$TOTAL_TIME = $REMAINING;
$FILECOPY = basename($FILE,"") . "-pid" . $$;
$ENRICHED_INPUT = "$WORKDIR" . "/" . "$FILECOPY" . "+lemmas";
$START_TIME = (times)[2];
$ELAPSED_LAST = $START_TIME;

#  $INPUT = "$WORKDIR/$FILECOPY";
#  copy($FILE,$INPUT); 
#  copy($INPUT,$ENRICHED_INPUT); 
copy($FILE,$ENRICHED_INPUT); 

# Main loop:
while( $RESULT == 0 ) {
# First Setheo run
    $RUN = $RUN + 1;
		      print "Run $RUN: setheo.... "; 
    $OUTFILE = "$WORKDIR" . "/" . "run" . "$RUN" . "-" . "$FILECOPY";
    $LIMIT = int $REMAINING;
    $SLIMIT = $SLIMIT * ( 3 / 2 );
#    print "MAX_NR_LEMMAS = $MAX_NR_LEMMAS ";
    $SETHEO = "$BINDIR/scheme-setheo  -wdr -wd1 90 -wd2 100 -wd3 100 -foldup -lemmas $MAX_NR_LEMMAS 500 -input-lemmas 10 -uselemmas 5 50 -sgindex -indexfoldup -cons -mate -dynsgreord 0 -elimit ";
    $MAX_NR_LEMMAS = int($MAX_NR_LEMMAS * 15 /10);
    $RETCODE = 0xffff & system "$BINDIR/limit_cpu $LIMIT $SETHEO $SLIMIT $ENRICHED_INPUT >$OUTFILE 2>&1";

    $ELAPSED = (times)[2] - $START_TIME;
    $REMAINING = $TOTAL_TIME - $ELAPSED;
    $USED_TIME = $ELAPSED - $ELAPSED_LAST; $ELAPSED_LAST = $ELAPSED; 

    if ( system("grep -c ' PROOF' $OUTFILE >/dev/null")  == 0 ) {
	$RESULT = 1;
    }
    elsif ( system("grep -c ' SATISFIABLE' $OUTFILE >/dev/null") == 0 ) {
	$RESULT = 2;
    }
    elsif ( system("grep -c ' EMPTY' $OUTFILE >/dev/null") == 0 ) {
	$RESULT = 3;
    }
    if ( $RESULT != 0 || $REMAINING < 1 || $RUN > 100 ) {
	last;
    }
    elsif ( $RETCODE != 0 ) {
	if ( $REMAINING >= 1 ) {
	    $RESULT = 3;
	}
	last;
    }
    $NR_LEMMAS = 0;
    open(LEMMAFILE,">>$ENRICHED_INPUT");
    open(OUT_LEMMAS,"<$OUTFILE");
    while(<OUT_LEMMAS>) {
	if ( /input_clause.*\./ ) {
	    s/,hypothesis,/,lemma,/;
	    s/,axiom,/,lemma,/;
	    s/,conjecture,/,lemma,/;
	    print LEMMAFILE $_;
	    $NR_LEMMAS = $NR_LEMMAS + 1;
	}
    }
       print "generated $NR_LEMMAS lemmas";
       print ", time: $USED_TIME\n";
    close OUT_LEMMAS;
    close LEMMAFILE;
    $SAVE_ENRICHED = "$WORKDIR" . "/" . "lemmas" . "$RUN" . "-" . "$FILECOPY";
    copy($ENRICHED_INPUT,$SAVE_ENRICHED);
    # then E run
    $RUN = $RUN + 1;
    print "Run $RUN: eprover... ";
    $OUTFILE = "$WORKDIR" . "/run" . "$RUN" . "-" . "$FILECOPY";
    # print "$OUTFILE\n"; 
    $ELIMIT = int($ELIMIT * (39 / 25));
    $LIMIT = int $REMAINING;
    $RETCODE = 0xffff & system "$BINDIR/limit_cpu $LIMIT $E $ELIMIT $ENRICHED_INPUT > $OUTFILE 2>&1";
    $ELAPSED = (times)[2] - $START_TIME;
    $REMAINING = $TOTAL_TIME - $ELAPSED;
    $USED_TIME = $ELAPSED - $ELAPSED_LAST; $ELAPSED_LAST = $ELAPSED;

    if ( system ("grep -c 'Proof found' $OUTFILE >/dev/null") == 0 ) {
	$RESULT = 1;
    }
    elsif ( system("grep -c 'No proof found' $OUTFILE >/dev/null") == 0 ) {
	$RESULT = 2;
    }
    elsif ( system("grep -c 'Input file contains no clauses' $OUTFILE >/dev/null") == 0 ) {
	$RESULT = 3;
    }
    if ( $RESULT != 0 || $REMAINING < 1 || $RUN > 100 ) {
	last;
    }
    elsif ( $RETCODE != 0 ) {
	if ( $REMAINING >= 1 ) {
	    $RESULT = 3;
	}
	last;
    }
    $NR_LEMMAS = 0;
    open(OUT_LEMMAS,"<$OUTFILE");
    open(NEW_LEMMAS,">>$ENRICHED_INPUT");
    while(<OUT_LEMMAS>) {
	if ( /input_clause.*\./ ){
	    s/,unknown,/,lemma,/;
	    s/,hypothesis,/,lemma,/;
	    s/,axiom,/,lemma,/;
	    s/,conjecture,/,lemma,/;
	    print NEW_LEMMAS $_;
	    $NR_LEMMAS = $NR_LEMMAS + 1;
	}
    }
    print "generated $NR_LEMMAS lemmas";
    print ", time: $USED_TIME\n";
    close OUT_LEMMAS;
    close NEW_LEMMAS;
}

# Formula status
if ( $RESULT == 0 ) {
    $RESULT = unknown;
}
elsif ( $RESULT == 1 ) {
    $RESULT = unsatisfiable;
}
elsif ( $RESULT == 2 ) {
    $RESULT = satisfiable;
}
else {
    $RESULT = ERROR;
}

print "Formula $FILECOPY status: $RESULT\n";
#unlink <$WORKDIR/*$$*>;

exit 0;
