/*-------------------------------------------------------------*/
/*      const.h            constants used by THEO and HERBY
/*
/*
/*---------------------------------------------------------------*/
 
/*----------------------------------------------*/
/* A theorem can have at most 1024 base clauses */
/* and at most 30 literals in any base clause   */
/* or any inferred clause.                      */
/* NOTE: data structure limit the number of     */
/* base clauses to 1024 (elim_inferences array!)*/
/*----------------------------------------------*/
 
#define MAX_BASE_CLAUSES   1024           /* max # base clauses     */
#define MAX_LIT_IN_CLAUSE  30             /* max # literals in cls  */

/*----------------------------------------------*/
/* MAX_VARS (in theorem) must be a power of 2   */
/*----------------------------------------------*/
 
#define MAX_VARS           64     /* max # distinct vars in theorem */
#define MAX_VARS_M1        63     /* MAX_VARS_M1 = MAX_VARS - 1     */
#define MAX_VARS_X2       128     /* MAX_VARS_X2 = MAX_VARS * 2     */

/*------------------------------------------------------------------*/
/* MAX_PF = max number of predicates,functions permitted in theorem */
/* Of course, each pred or fxn could appear many times.             */
/* PF_SPACE_SIZE = size of the array "pfspace".                     */
/*------------------------------------------------------------------*/

#define MAX_PF             5000 /* MAX_PF=MAX_PRED_IDS+MAX_FXNS_IDS */ 
#define PF_SPACE_SIZE      6*MAX_PF /* space for 100 6 char pred/fxn*/ 
 
#define MAX_PRED           10000  /* max # of predicate appearances */
#define MAX_CHAR           32000  /* MAX_CHAR = MAX_PRED * 4        */
#define PRED_PTR_SIZE      3*MAX_PRED /* size of pred_ptr_list      */
 
#define MAX_PRED_IDS       800     /* max number of different preds */
#define MAX_FXNS_IDS       800     /* max number of different fxns  */

/*------------------------------------------------------------------*/
/* MAX_INPUT_THM_SIZE is the maximum number of characters that      */
/* can appear in the input clauses.                                 */
/* MAX_INPUT_THM_SIZE = size of array "kode".                       */
/*                                                                  */
/* MAX_COMPILED_THM_SIZE is the maximum number of long words that   */
/* that the compiled theorem can occupy.                            */
/* MAX_COMPILED_THM_SIZE = size of array "compiled_code_buffer.     */
/*------------------------------------------------------------------*/
 
#define MAX_INPUT_THM_SIZE     80000  
#define MAX_COMPILED_THM_SIZE  80000  
 
/*------------------------------------------------------------------*/
/* ELIM_ARRAY_SIZE is the maximum number of resolvents that can     */
/* be generated during the eliminate phase (Phase 0)                */
/* ELIM_ARRAY_SIZE = size of array "elim_inferences".               */
/*------------------------------------------------------------------*/
 
#define ELIM_ARRAY_SIZE        100000            
 
/*------------------------------------------------------------------*/
/* MAX_ITERATION =        maximum number of iterations to carry out */
/* MAX_PROOF_LENGTH =      maximum number of inferences in a proof. */
/* PROOF_ARRAY_SIZE must be twice the value of MAX_PROOF_LENGTH.    */
/* DEPTH_PTR_SPACE = 4*MAX_PROOF_LENGTH = size of array "depth".    */
/* PROOF_ARRAY_SIZE = size of the array "cl" (used in pass2.c)      */
/*------------------------------------------------------------------*/
 
#define MAX_ITERATIONS        30 

#define MAX_PROOF_LENGTH       800           /* size of depth array */
#define PROOF_ARRAY_SIZE      1600              /* size of cl_array */
#define PROOF_ARRAY_SIZE_M1   1599              /* size of cl_array */

#define DEPTH_PTR_SPACE        4*MAX_PROOF_LENGTH
 
/*------------------------------------------------------------------*/
/* CSPACE_CLAUSES =         maximum number of clauses on clist      */
/* CSPACE_SIZE    =          maximum size of clauses in cspace      */
/* USPACE_CLAUSES =          maximum number of clauses on ulist     */
/* HSPACE_CLAUSES =         maximum number of clauses on hb_list    */
/* HSPACE_SIZE    =           maximum size of clauses on hb_list    */
/*------------------------------------------------------------------*/
 
#define CSPACE_CLAUSES     8000
#define CSPACE_SIZE      800000

#define MSPACE_CLAUSES      100
#define MSPACE_SIZE       80000
 
#define PLY_MASK    0x00ff          /* mask off ply; used in hash.c */
 
/*------------------------------------------------------------------*/
/* constants used by pass2.c to decode inferences in array  "cl".   */
/*------------------------------------------------------------------*/
 
#define PARENT_MASK  0x00000fff                   /*   parent mask  */
#define INDEX_MASK   0x0000000f            /* index of parents mask */
#define CIDMASK      0x07ff0000                       /* clause id  */
#define CIDMASK2     0x000007ff                /* shifted clause id */

/*------------------------------------------------------------------*/
/* message tags-- used by OCTOPUS                                   */
/*------------------------------------------------------------------*/

#define CCS     3       /* message for "compiled code size */
#define CCB     4       /* message for compiled code buffr */
#define YN      5       /* message for yes/no to continue  */
#define YNR     6       /* message for yes/no reply        */
#define CCSR    7       /* message for reply on buf size   */
#define CNAME   8       /* message for name of computer    */
#define MPROOF  9       /* message master found proof      */
#define SPROOF  10      /* message slave found proof       */

