/*-------------------------------------------------------------*/
/*      elim.c
/*
/*
/*      functions:
/*          bres_ply_1()            .................... #1
/*          res_ply_1()             .................... #2
/*          uncomple_lit()          .................... #3
/*          base_subsume()          .................... #4
/*          successor_subsume()     .................... #5
/*          single_resolvent()      .................... #6
/*          unit_resolvent()        .................... #7
/*          elim_link_lits()        .................... #8
/*          elim_long_cls()         .................... #9
/*          det_min_link_lit()      .................... #10
/*          add_link_resolvents()   .................... #11
/*          delete_link_clauses()   .................... #12
/*          simple_subsume()        .................... #13
/*          delete_clause()         .................... #14
/*          add_clause()            .................... #15
/*          get_good_clauses()      .................... #15a
/*          bres_replace()          .................... #16
/*          bres_link()             .................... #17
/*          eliminate()              ....................#18
/*-------------------------------------------------------------
/*      function linkage
/*
/*      eliminate()
/*          res_ply_1()
/*              bres_ply_1()
/*          uncomple_lit()
/*              delete_clause()
/*          base_subsume()
/*              simple_subsume()
/*                  delete_clause()
/*          successor_subsume()
/*              add_clause()
/*                  bres_replace()
                delete_clause() 
/*          single_resolvent()
/*              add_clause()
/*                  bres_replace()
                delete_clause() 
/*          elim_link_lits()
/*              det_min_link_lit()
/*              add_link_resolvents()
/*                  bres_link()
/*              delete_link_clauses()
/*          elim_long_cls()
/*--------------------------------------------------------------*/
 
#include <stdio.h>
#include <strings.h>
 
#include "clause.h"
#include "const.h"
 
/*------------------------------------------------------------*/
/* these masks are to allow a resolution to be compacted into */
/* one long word                                              */
/*------------------------------------------------------------*/
 
#define CLPARMASK       0xffc00000  /* mask for bits 23 - 32  */
#define CRPARMASK       0x003ff000  /* mask for bits 13 - 22  */
#define CNLITMASK       0x00000f00  /* mask for bits  9 - 12  */ 
#define CINDEX1         0x000000f0  /* mask field for intx1   */
#define CINDEX2         0x0000000f  /* mask field for intx2   */
#define HT_HIT_MASK     0xfffff0ff  /* mask field for c_nlit  */
#define P1023           0x000003ff  /* 1023 mask              */

/*------------------------------------------------------------*/
/* externally defined functions                               */
/*------------------------------------------------------------*/
	
extern void print_clause();
extern void print_clause2();
extern void print0();                           /* tdump.c    */
extern int time_now();                          /* theo.c     */
extern void clear_var_substitution_tble();      /* infer.c    */
extern void standardize();                      /* infer.c    */
extern int unify();                             /* unify.c    */
extern int install_clause();                    /* search.c   */
extern clause *generate_clause();               /* gen.c      */
extern int test_equality();                     /* gen.c      */
extern int add_clause_to_ht();                  /* hash.c     */
extern int calc_lit_hc();
extern void calc_clause_hashcode();
extern int calc_alt_hashcodes();
extern int copy_clause4();
extern int copy_clause5();
extern void calc_alt_hash();

/*------------------------------------------------------------*/
/* externally defined constants                               */
/*------------------------------------------------------------*/
 
extern char *OutPtr;

extern int elim_inferences[]; /* where cls_code-ed cl's are   */
extern clause *clist[];  /* search.c:  pointers to clauses    */
extern clause *ulist[500]; /* search.c:  pointers to clauses  */
extern dnode depth[]; /* search.c:  pointers to levels of cls */
extern int skolem[MAX_FXNS_IDS];
extern int good_clauses[2000];

extern int base, org_base, nextc, nextu, ply, iter_depth, nc;
extern int clauses_alive; /* count of the # of ACTIVE clauses */
extern int equal_pred;    /* = 1 if equal pred is in theorem  */
extern int o_elimprint;  /* theo.c: o_elimprint=1 => prt info */
extern int o_maxlit, o_maxterm, o_maxvar;
extern int o_learn_var;
extern int o_debug;
extern int o_merge, o_nc, o_hash, o_extend;
extern int o_batch;
extern int o_learn;
extern int number_preds;    /* theo.c:  # of preds in theorem */
extern int number_constants, number_const_nc;
extern int in_elim;
extern int simple_sol;                        /* theo.c       */
extern int eliminate_time;
extern int elim_starting_time;
extern int equal_clist_index;
extern int alt_hx, alt_hy, rows_less_nfxns, number_const_hash, var_array[64];


/*------------------------------------------------------------*/
/* internally defined functions                               */
/*------------------------------------------------------------*/

void bres_ply_1();
int res_ply_1();
int uncomple_lit();
int base_subsume();
int successor_subsume();
int single_resolvent();
int unit_resolvent();
int elim_link_lits();
void elim_long_cls();
int det_min_link_lit();
int add_link_resolvents();
int delete_link_clauses();
int simple_subsume();
void delete_clause();
int add_clause();
int bres_replace();
int bres_link();
void add_skolem_clauses();
void get_good_clauses();
void add_ulist_clauses();
void eliminate();
int calc_subsumed_hcodes();
int check_if_subsumed();

/*------------------------------------------------------------*/
/* internally defined constants                               */
/*------------------------------------------------------------*/
 
int subsumes_parent[50];  /* encoding of subsuming cls's info */

int subsumes_parent_id[50];
int sp;
int infoptr;
int units, nc_units;
int count_pass, inf_try, unify_try, unify_count, inf_count;
int clause_code;
int o_maxlit_temp,o_maxterm_temp,o_maxvar_temp,o_merge_temp;
int reinstate[1000],reinstate_loop[1000];

/*------------------------------------------------------------------*/ 
/*      bres_ply_1(c1,c2,intx1,intx2)         .................. #1
/*
/*      generate a single resolvent: on the intx1 lit of c1 and the
/*      intx2 lit of c2.  Extract information from the clause and
/*      store it in clause_code.  The format for clause_code is:
/*
/*      bits:  32 - 23    22 - 13   12 - 9    8 - 5     4 - 1
/*             l_par      r_par     c_nlit    intx1     intx2
/*
/*      called by:      res_ply_1()                        elim.c
/*
/*      calls:          standardize()                      infer.c
/*                      clear_var_substitution_tble()      infer.c
/*                      unify()                            unify.c
/*                      generate_clause()                  gen.c
/*                      simple_subsume()                   elim.c
/*                      print_clause()                     tdump.c
/*
/*      returns:        nil
/*------------------------------------------------------------------*/ 
 
void bres_ply_1(c1,c2,int1x,int2x)
clause *c1, *c2;
int int1x, int2x;
{
clause *c;
literal *l1, *l2;
int id1, id2;
int ix1, ix2, i, j;

clause_code = 0;
id1 = c1->c_bits; id2 = c2->c_bits;     
if(!((id1 & id2) & ACTIVE)) {return;}
standardize(c1,c2);                /* standardize vars apart */
 
ix1 = int1x;
ix2 = int2x;
 
l1 = c1->c_lit;
while(l1->l_id & ELIMLIT) l1 = l1->l_next;
for( j = 1 ; j < int1x; j++) {
   l1 = l1->l_next;
   while(l1->l_id & ELIMLIT) l1 = l1->l_next;}

l2 = c2->c_lit;
while(l2->l_id & ELIMLIT) l2 = l2->l_next;
for( j = 1 ; j < int2x; j++) {
   l2 = l2->l_next;
   while(l2->l_id & ELIMLIT) l2 = l2->l_next;}

id1 = l1->l_id & IDMASK;               /* pred id's  */
id2 = l2->l_id & IDMASK;
	
/*----------------------------------------------------------*/
/* if IDs are the same then we are in a predicate group.    */
/* we attempt unification if literals are complementary.    */
/*----------------------------------------------------------*/

inf_try++;
if(id1 == id2) {
   if(((l1->l_id ^ l2->l_id) & NEGMASK) == NEGMASK) {
	
/*------------------------------------------------------*/
/* if id's are the same and complementary, try unify    */
/*------------------------------------------------------*/

      clear_var_substitution_tble(c1->c_bits,c1->c_nv);
      clear_var_substitution_tble(c2->c_bits,c2->c_nv);
      unify_try++;
      if(unify(&l1->l_id,&l2->l_id)) { 
	 unify_count++;
	 if(c = generate_clause(c1,c2,l1,l2,ix1,ix2)) {

/*----------------------------------------------------------------*/
/* pack into clause_code a description of the                     */
/* binary resolution as follows:                                  */
/*  bits 23-32     bits 13-22    bits 9-12   bits 5-8   bits 1-4  */
/*  c_lpar->c_id   c_rpar->c_id  c->c_nlit   ix2        ix1       */
/*----------------------------------------------------------------*/

	         clause_code  = c->c_lpar->c_id * 4194304;
	         clause_code += c->c_rpar->c_id * 4096;
	         clause_code += c->c_nlit * 256;
		 if(c->c_nlit > 15) return;
	         clause_code += ix1 * 16 + ix2;
		 inf_count++;
 
           if(c->c_nlit == 1) {
              units++;
              if(c->c_lpar->c_id > org_base - nc + 1) {
                 if(c->c_lpar->c_id <= org_base + 1) {
                    nc_units++;}}}

/*--------------------------------------------------------------*/
/* if clause simple-subsumes either its left or right parent,   */
/* then set subsumes_parent to that encoded clause              */
/*--------------------------------------------------------------*/

         if(simple_subsume(c,c1) == 2) {    /* record clause */
            if(sp <49) sp++;
            subsumes_parent[sp] = clause_code;/*subsumes parent*/
            subsumes_parent_id[sp] = c1->c_id;}

         if(simple_subsume(c,c2) == 2) {
            if(sp < 49) sp++; 
            subsumes_parent[sp] = clause_code;
         subsumes_parent_id[sp] = c2->c_id;}}}}}

if(o_elimprint > 2 && clause_code) {
   print0("\n");
   print_clause(c);
   if(o_elimprint > 3) {getchar();}}
return;
}
 
 
/*------------------------------------------------------------------*/
/*      res_ply_1( )                    .............. #2
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 0------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      Generate all first-level resolvents one at a time.  They are
/*      placed in "clause_code" and then added to elim_inferences 
/*      list.  At most ELIM_ARRAY_SIZE resolvents can be formed.
/*
/*      called by:          eliminate()           elim.c
/*
/*      calls:              bres_ply_1()          elim.c
/*
/*      returns:            the # of resolvents of the current
/*                          base clauses
/*------------------------------------------------------------------*/
 
int res_ply_1()
{
dnode *d;
int i, j;
int int1, int2;

depth[0].d_tc = base;   /* pters to beginning, end of base clauses  */
depth[0].d_bc = 0;
 
d = &depth[1];
d->d_tc = nextc;
d->d_bc = nextc;
sp = -1;
 
units = nc_units = 0;
inf_count = inf_try = unify_try = unify_count = 0;
infoptr = 0;
ply = 1;
 
if(o_elimprint == 2) {
   print0("\n");}
 
/*----------------------------------------------------------*/
/* Loop through base clauses generating resolvents          */
/*----------------------------------------------------------*/

for(j = base; j >= 0; j--) {
   if((clist[j]->c_bits & ACTIVE) == 0) {continue;}
   if(j > 0) {
      for(i = j - 1; i >= 0; i--) {
         if((clist[i]->c_bits & ACTIVE) == 0) {continue;}
         for(int1 = 1 ; int1 <= clist[j]->c_nlit ; int1++) {
	    for(int2 = 1 ; int2 <= clist[i]->c_nlit ; int2++) {
	       if(infoptr == ELIM_ARRAY_SIZE) {
	          print0("\nToo many clauses. Skip Elim.\n");
	             return(-1);}
	       bres_ply_1(clist[j],clist[i],int1,int2);
	       if(!(clause_code & CNLITMASK) && clause_code) {
	          return(-1);}              /* return on null clause */
	       if(clause_code) {
	          elim_inferences[infoptr++] = clause_code;}}}}}}
if(count_pass++ == 0) {
//   printf(" INF %d/%d\n",inf_try,inf_count);
   OutPtr += sprintf(OutPtr," INF %d/%d\n",inf_try,inf_count);
   /*for(i = 1; i < 81; i++) {
      printf("\n%2d ",i);
      for(j = 1; j < 81; j++) {
	 if(i == j && (i % 5) == 0) printf(".");
	 else if(i == j) printf(" ");
	 else {printf("%d",resolve_pair_array[j][i]);}}}*/}


return(infoptr);                /*      return to eliminate      */
}
 
/*------------------------------------------------------------------*/
/*      uncomple_lit(numb_res)   ............................... #3 
/* 
/*------------------------------------------------------------------*/
/*------------------------------Step 1------------------------------*/
/*------------------------------------------------------------------*/
/* 
/*      Search for literal which does not appear complemented in 
/*      another clause.  Delete clause.
/* 
/*      called by:        eliminate()             elim.c 
/* 
/*      calls:            delete_clause()         elim.c
/*                        print_clause()          tdump.c
/* 
/*      returns:          1 if there is an uncomplemented lit in
/*                        some clause 0 otherwise 
/*------------------------------------------------------------------*/
 
int uncomple_lit(numb_res)
int numb_res;
{
int l, lit_max, lit_count;
int deleted;
int i, j;
		
/*--------------------------------------------------------------*/
/* delete all clauses containing uncomplemented literals        */
/*--------------------------------------------------------------*/
			
deleted = 0;
for(i = base; i >= 0; i--) {
   if(clist[i]->c_bits & ACTIVE) {
      lit_max = clist[i]->c_nlit;       
      for(l = 1;l <= lit_max;l++) { /* search for lit "l" in cl "i" */
         lit_count = 0;
         for(j = 0; j < numb_res ;  j++) {   
            if(((elim_inferences[j]&CLPARMASK)>>22)==clist[i]->c_id) { 
               if(((elim_inferences[j] & CINDEX1) >> 4)== l) { 
	          lit_count++; break;}}                
        if(((elim_inferences[j] & CRPARMASK)>>12)==clist[i]->c_id){ 
	   if((elim_inferences[j] & CINDEX2)  ==  l) {       
               lit_count++; break;}}}
     if(!lit_count)  {   /* lit "l" in clause "i" not found */  
        if(o_elimprint) {      
           printf("\n\nUncomp_litrl(%d):  delete ",l);        
           OutPtr += sprintf(OutPtr,"\n\nUncomp_litrl(%d):  delete ",l);
           print_clause(clist[i]);}
        delete_clause(i); 
	reinstate[i] = 1;
	reinstate_loop[i] = count_pass;
        deleted = 1;
          break;}}}}
			
return(deleted);
}
 
/*------------------------------------------------------------------*/
/*      base_subsume()                  ..................... #4
/*
/*      if base clause j is subsumed by some other base clause,
/*      eliminate cls j
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 2------------------------------*/
/*------------------------------------------------------------------*/
/*      called by:      eliminate()         elim.c
/*
/*      calls:          simple_subsume()    elim.c
/*                      delete_clause()     elim.c
/*                      print_clause()      tdump.c
/*
/*      returns:        1 if some clause is subsumed
/*                      0 otherwise
/*------------------------------------------------------------------*/
 
int base_subsume()
{
int i, j;
int deleted;

deleted = 0;	
for(i = 0; i <= base ; i++) {
   if(!(clist[i]->c_bits & ACTIVE)) {continue;}
   for(j = 0 ; j <= base ; j++) {
      if(!(clist[j]->c_bits & ACTIVE)) {continue;}
      if(i != j) {
	  if(simple_subsume(clist[i],clist[j])) {
	     if(o_elimprint) {
	         print0("\n\nBase_subsume: delete ");
	         print_clause(clist[j]);
	         print0("\n        Subsumed by: ");
	         print_clause(clist[i]);}
	      delete_clause(j);
	      reinstate[j] = 2;
	      reinstate_loop[j] = count_pass;
	      deleted = 1;}}}}

return(deleted);
}
 
/*------------------------------------------------------------------*/
/*      successor_subsume(c)       ........................ #5
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 3------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      if some clause with id sub_clause_id is subsumed by its child
/*      subsumes_parent, replace clause.
/*
/*      called by:          eliminate()        elim.c
/*
/*      calls:              add_clause()       elim.c
/*                          print_clause()     tdump.c
/*                          delete_clause()    elim.c
/*
/*      returns:            1 if a successor is subsumed; 0 otherwise
/*------------------------------------------------------------------*/
 
int successor_subsume()
{
int i;
int ssp;
clause *c;
int added;

ssp = 0;

if(sp == -1) {
   return(0);}
for(i = 0 ; i <= sp ; i++) { 
   c = clist[subsumes_parent_id[i] - 1];
   if(!(c->c_bits & ACTIVE)) {continue;}
   ssp = 1;
   if(o_elimprint) {
      print0("\n\nSuccessr_Sub: replace ");
      print_clause(clist[subsumes_parent_id[i] - 1]);}
   added = add_clause(subsumes_parent_id[i] - 1,subsumes_parent[i]);
   delete_clause(subsumes_parent_id[i] - 1);
   reinstate[subsumes_parent_id[i] - 1] = 3;
   reinstate_loop[subsumes_parent_id[i] - 1] = count_pass;}
return(ssp);
}

/*------------------------------------------------------------------*/
/*      single_resolvent(numb_res)    ............................ #6
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 4------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      If literal in clause has only one resolvent, replace
/*
/*      called by:  eliminate()                 elim.c
/*
/*      calls:      add_clause()                elim.c
/*                  delete_clause()             elim.c
/*                  print_clause()              tdump.c
/*
/*      returns:    1 if clauses replaced ; 0 otherwise
/*----------------------------------------------------------------*/
 
int single_resolvent(numb_res)
int numb_res;
{
literal *lit;
clause *c;
int i, j, k, l, num, lit_max, min_lit;
int i1, repl_base[MAX_BASE_CLAUSES], repl_resol[MAX_BASE_CLAUSES];
int done, lit_count;
int added;
		
min_lit = 1000;
num = 0;
added = 0;
	
for(i = 0; i < MAX_BASE_CLAUSES; i++) {
   repl_resol[i] = 0;}
	
/*----------------------------------------------------------*/
/* replace all single resolvent clauses with its resolvent  */
/*----------------------------------------------------------*/

for(i = base; i >= 0 && min_lit != 1; i--) {         
   if(!(clist[i]->c_bits & ACTIVE)) {continue;}
   lit_max = clist[i]->c_nlit;
   for(l = 1; l <= lit_max; l++) {/*for each ACTIVE base clause */
      lit_count = 0;
      for(j = 0; j < numb_res; j++ ) {
	     if(((elim_inferences[j] & CLPARMASK) >> 22) == 
                                         clist[i]->c_id) {
	        if(((elim_inferences[j] & CINDEX1) >> 4) == l) {
	           lit_count++;     
	           k = j;   
	           if(lit_count > 1) break;
	           lit = clist[i]->c_lit;   
	           for(i1 = 1; i1 < l ; i1++) {
		          lit = lit->l_next;}}}
	     if(((elim_inferences[j]&CRPARMASK)>>12)==clist[i]->c_id) {
	        if((elim_inferences[j] & CINDEX2) == l) {
	           lit_count++;
	           k = j;
	           if(lit_count > 1) break;
	           lit = clist[i]->c_lit;
	           for(i1 = 1; i1 < l ; i1++) {
		          lit = lit->l_next;}}}}
			
/*----------------------------------------------------------*/
/* cannout use NULL clause to replace another clause        */
/* only can replace a longer clause                         */
/*----------------------------------------------------------*/

      if(lit_count == 1 && 
          ((elim_inferences[k] & CNLITMASK) >> 8)) {
	       done = 1; 

/*----------------------------------------------------------*/
/* DON'T LOOP (See s18.dbabhp.thm                           */
/* The next several lines were modified Dec 13, 1994        */
/*----------------------------------------------------------*/

     /*    if(nextc - 1 < org_base) {
            done = 0;}
*/
	if(nextc > org_base + 1) {
	   if((((elim_inferences[k] & CLPARMASK) >> 22) == nextc) ||
             (((elim_inferences[k] & CRPARMASK) >> 12) == nextc)) {
             if(clist[nextc]->c_rpar->c_id == nextc ||
                       clist[nextc]->c_lpar->c_id == nextc) {


              if((clist[(elim_inferences[k] & CLPARMASK) >> 22]->c_id
           +clist[(elim_inferences[k] & CRPARMASK) >> 12]->c_id - 1)
                        == (clist[nextc - 1]->c_lpar->c_id +
                        clist[nextc - 1]->c_rpar->c_id)) {
                    done = 0;}}}}

 	  if(done == 1) {
	         if(elim_inferences[k] & CNLITMASK) {/*if not in HT:*/
	             repl_base[num] = i;
	             repl_resol[num++] = k;
	             if(((elim_inferences[k]&CNLITMASK)>>8)<min_lit) {
		             min_lit = 
                       ((elim_inferences[k] & CNLITMASK) >> 8);}}}}}} 


if(!done) return(0);

for(j = 0; j < num; j++) {                            
   i = repl_base[j];     
   k = repl_resol[j];    
   if(!reinstate[i]) {
      if(o_elimprint) {
	        print0("\n\nSingle_resol: replace ");
	        print_clause(clist[i]);
	        print0("\n                 with ");}
	
/*-------------------------------------------*/
/* add clause with condensed info in         */
/* elim_inferences[k].                       */
/*-------------------------------------------*/

      added = added + add_clause(i,elim_inferences[k]);
      delete_clause(i);
      reinstate[i] = 4;
      if(clist[i]->c_nv == 0) reinstate[i] = 5;
      reinstate_loop[i] = count_pass;
      return(1);}} 

if(added) return(1);
return(0);
}
 
/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/
/*      unit_resolvent()    ........................ #7
/*
/*------------------------------------------------------------------*/
/*------------------------------Step x -----------------------------*/
/*------------------------------------------------------------------*/
/*
/*      add unit clauses
/*
/*      called by:  eliminate()                 elim.c
/*
/*      calls:      add_clause()                elim.c
/*                  print_clause()              tdump.c
/*
/*      returns:    1 if clauses added ; 0 otherwise
/*----------------------------------------------------------------*/
 
int unit_resolvent()
{
literal *lit;
clause *c;
int i, j, k, num;
int repl_base[MAX_BASE_CLAUSES], repl_resol[MAX_BASE_CLAUSES];
int added;
clause *lp, *rp;
int lp_int, rp_int;

num = added = 0;
	
for(i = 0; i < MAX_BASE_CLAUSES; i++) {repl_resol[i] = 0;}
	
/*----------------------------------------------------------*/
/* replace all single resolvent clauses with its resolvent  */
/*----------------------------------------------------------*/
	
for(j = 0; j < infoptr; j++ ) {
   if(num >= MAX_BASE_CLAUSES) {break;}
   if(((elim_inferences[j] & CNLITMASK) >> 8) == 1) {
      i = (elim_inferences[j] & CLPARMASK) >> 22; 
      repl_base[num] = i;
      repl_resol[num++] = j;}}

for(j = 0; j < num; j++) { 
   i = repl_base[j];     
   k = repl_resol[j];
   c = clist[i - 1]; 
   lp_int = (((elim_inferences[k]&CLPARMASK)>>22)&P1023) - 1;
   rp_int = ((elim_inferences[k]&CRPARMASK)>>12) - 1;
   lp = clist[lp_int];
   rp = clist[rp_int];
   if(lp == rp) {continue;}
   if((lp->c_bits & EQCLS)  || (rp->c_bits & EQCLS)) {continue;} 
      if(o_elimprint) {
         print0("\n\nUnit_resol: delete     ");
         print_clause(clist[i - 1]);
         print0("\n");} 

/*-------------------------------------------*/
/* add clause with condensed info in         */
/* elim_inferences[k].                       */
/*-------------------------------------------*/ 

      if(base < org_base*1.5 && base < org_base + 20) {
         added = added + add_clause(i,elim_inferences[k]);
      reinstate[i - 1] = 6;
      reinstate_loop[i - 1] = count_pass;}}

if(added) return(1);
return(0);
}
 
/*------------------------------------------------------------------*/
/*      elim_link_lits()              .................... #8
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 5------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      determine the simplest link literal and then add and
/*      delete clauses
/*
/*      called by:  eliminate()                elim.c
/*
/*      calls:      det_min_link_lit()         elim.c
/*                  add_link_resolvents()      elim.c
/*                  delete_link_clauses()      elim.c
/*
/*      returns:    1 if link lit is found and clauses changed
/*                  0 otherwise
/*------------------------------------------------------------------*/
 
int elim_link_lits()
{
int base0, min_lit, number_resolvents, deleted_link_cls;
	
if((min_lit = det_min_link_lit()) == -1) return(0);

base0 = base;
number_resolvents = add_link_resolvents(min_lit);
base = base + number_resolvents;
clauses_alive = clauses_alive + number_resolvents;
deleted_link_cls = delete_link_clauses(min_lit,base0); 
if(!deleted_link_cls) {
   return(0);}	
return(1);
}
 
/*------------------------------------------------------------------*/
/*      elim_long_cls(maxlitrl)          .................... #9
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 6------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      sets the ACTIVE flag to "off" on all "long" clauses that have
/*      been added by the elim. procedure.  Clauses are considered
/*      to be long if they have more than maxlit literals.
/*
/*      called by:     eliminate()          elim.c
/*
/*      calls:          nobody
/*
/*      returns:        Nil
/*------------------------------------------------------------------*/
 
void elim_long_cls(maxlitrl)
int maxlitrl;
{
int i;
 
for(i = org_base + 1; i <= base; i++) {
   if(clist[i]->c_bits & ACTIVE) {
      if(clist[i]->c_nlit > maxlitrl) {
	     clauses_alive--;
	     clist[i]->c_bits &= ~ACTIVE;
	     reinstate[i] = 7;
	     reinstate_loop[i] = count_pass;}}}
return;
}
 
/*------------------------------------------------------------------*/
/*      det_min_link_lit()                 .................... #10
/*
/*      determine the literal that is to be called the minimum link
/*      literal
/*
/*      called by:      elim_link_lits()            elim.c
/*
/*      calls:          Nobody
/*
/*      returns:        the number of the min_link_literal
/*------------------------------------------------------------------*/
 
int det_min_link_lit()
{
literal *litrl;
int lit,j;
int min_link_lit;       /* returned by subroutine as pred to elim   */
int min_link_lit_cls;   /* number of cls in which min_link_lit is   */
int lit_total_count;    /* number of appearances of lit under study */
int lit_clause_count;   /* number of times lit appears in a clause  */
int  dead_clause;
int  pred_args;
	
lit_clause_count = 0;
min_link_lit = -1;
min_link_lit_cls = 1000;
pred_args = -1;
	
for(lit = 1; lit <= number_preds; lit++) {
   dead_clause = 0;
   lit_total_count = 0;
   for(j = 0; j <= base; j++) {
      lit_clause_count = 0;
      if(clist[j]->c_bits & ACTIVE) {
	 lit_clause_count = 0;
	 for(litrl=clist[j]->c_lit;litrl!=0;litrl=litrl->l_next) {
	          if(((litrl->l_id&IDMASK)>>ARGSHIFT)==lit) {
		  if(!(litrl->l_id & ARGMASK)) {
		     pred_args = 0;}
	             lit_clause_count++;
	             lit_total_count++;}}}
      if(lit_clause_count == 1 && reinstate[j]) {dead_clause++;}
      if(lit_clause_count > 1) {
         break;}}
   if(lit_clause_count <= 1 && !dead_clause) {
      if(lit_total_count < min_link_lit_cls && lit_total_count>0) {
	        min_link_lit = lit;
	        min_link_lit_cls = lit_total_count;
		if(!pred_args) return(min_link_lit);}}}

return(min_link_lit);
}
 
/*------------------------------------------------------------------*/
/*      add_link_resolvents(min_lit)      .................... #11
/*
/*      add clauses that are the resolvents of link literals
/*
/*      called by:      elim_link_lits()            elim.c
/*
/*      calls:          bres_link()                 elim.c
/*                      add_clause_to_ht()          hash.c
/*                      print_clause()              tdump.c
/*
/*      returns:        the number of new resolvents added
/*------------------------------------------------------------------*/
 
int add_link_resolvents(min_lit)
int min_lit;
{
dnode *d;
int i, j, cr_bits, cl_bits;
int num_resol, resol;      /*   number of resolvents           */
clause *c;
	
depth[0].d_tc = base;       /* initial generating clause       */
depth[0].d_bc = 0;
d = &depth[1];
	
ply = 1;
d->d_tc = nextc;
d->d_bc = nextc;
	
num_resol = 0;
	
if(o_elimprint) {
   print0(
	 "\n\nAdd clauses produced by eliminating link literal:\n");}
	
for(j = base; j >= 0; j--) {
   if(reinstate[j] != 0) {continue;}
   if(j > 0) {
      for(i = j - 1; i >= 0; i--) {
       if(reinstate[i] > 0) {continue;}
	     resol = bres_link(clist[i],clist[j],min_lit);
	     num_resol = num_resol + resol;
	     if(resol) {
	        clist[nextc-1]->c_id = nextc;
	        clist[nextc-1]->c_bits |= MERGE | BASE;
	        cr_bits = clist[nextc - 1]->c_rpar->c_bits;
	        cl_bits = clist[nextc - 1]->c_lpar->c_bits;
	        if((cr_bits & NC) || (cl_bits & NC)) {
	           clist[nextc - 1]->c_bits |= NC;} 
          c = clist[nextc - 1];
	        add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
	        if(o_elimprint) {
	           print0("                      ");
	           print_clause(clist[nextc - 1]);}}}}}
return(num_resol);           /*   return to elim_link_lits()  */
 
}
 
/*------------------------------------------------------------------*/
/*      delete_link_clauses(min_lit,base0)     ............... #12
/*
/*      delete clauses containing the link literal
/*
/*      called by:          elim_link_lits()        elim.c
/*
/*      calls:              print_clause()          tdump.c
/*
/*      returns:            nil
/*------------------------------------------------------------------*/
 
int delete_link_clauses(min_lit,base0)
int min_lit, base0;
{
literal *litrl;
int j, del_cls;
	
if(o_elimprint) {
   print0("\nDelete clauses with link literal:\n");}
	
del_cls = 0;
for(j = 0; j <= base0; j++) {
   if(clist[j]->c_bits & ACTIVE) {
      for(litrl=clist[j]->c_lit; litrl!=0; litrl=litrl->l_next) {
	     if(((litrl->l_id & IDMASK) >> ARGSHIFT) == min_lit) {
                if(!reinstate[j]) {
                   del_cls = 1;}
	        if(litrl->l_id & ARGMASK) {
                   reinstate[j] = 8;
		   reinstate_loop[j] = count_pass;}
	        clist[j]->c_bits &= ~ACTIVE;
	        if(o_elimprint) {
	           print0("                      ");
	           print_clause(clist[j]);}
	        break;}}}}
 
if(o_elimprint) {
   print0("\n");}
return(del_cls);
}
 
/*------------------------------------------------------------------*/
/*      simple_subsume(c1,c2)           .................... #13
/*
/*      simple_subsume examines the hash codes of the literals.
/*      Returns "1" if c1 = c2.  Returns 2 if c1 s-subsumes c2.
/*      That is, if every literal in c1 appears in exactly the
/*      same way in c2.
/*
/*      called by:  bres_ply_1()                   elim.c
/*                  base_subsume()                 elim.c
/*
/*      calls:      Nobody
/*
/*      returns:    1 if c1 = c2, 2 if c1 s-subsumes c2; 0 otherwise
/*------------------------------------------------------------------*/
 
int simple_subsume(c1,c2)
clause *c1, *c2;
{
literal *l1, *l2;
int found;
	
if(c1->c_nlit > c2->c_nlit) {
   return(0);}
		
for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) {
   found = 0;
   for( l2 = c2->c_lit; l2 != 0; l2 = l2->l_next) {
      if(l1->l_X == l2->l_X && l1->l_Y == l2->l_Y) {
	     found = 1;
	     break;}}
   if(found == 0) {
      return(0);}}

if(c1->c_nlit == c2->c_nlit) {
   return(1);}
return(2);
}                       
	
/*------------------------------------------------------------------*/
/*      delete_clause(i)                 .................... #14
/*
/*      delete the clause on clist whose index is i
/*
/*      called by:              uncomple_lit()          elim.c
/*                              base_subsume()          elim.c
/*
/*      calls:                  nobody
/*
/*      returns:                Nil
/*------------------------------------------------------------------*/
 
void delete_clause(i)
int i;
{       
clist[i]->c_bits &= ~ACTIVE;    
clauses_alive--;
if(o_elimprint > 1) getchar();
return;             /*  return to uncomple_lit    */
}
 
/*------------------------------------------------------------------*/
/*      add_clause(i,k)                 .................... #15
/*
/*      replace clause i with clause k
/*
/*      called by:          successor_subsume()         elim.c
/*                          single_resolvent()          elim.c
/*                          unit_resolvent()            elim.c
/*
/*      calls:              bres_replace()              elim.c
/*                          add_clause_to_ht()          hash.c
/*                          print_clause()              tdump.c
/*                          test_equality()             gen.c
/*
/*      returns:            nil
/*------------------------------------------------------------------*/
 
int add_clause(i,k)
int i, k;
{
clause *l_p, *r_p, *c;
int l_p_lit, r_p_lit;
int cl_bits, cr_bits;
int resolvent;

l_p      = clist[(((k & CLPARMASK) >> 22) & P1023) - 1]; 
l_p_lit  =  (k & CINDEX1)>>4;
r_p      = clist[((k & CRPARMASK) >> 12) - 1];
r_p_lit  = k & CINDEX2;
cr_bits  = clist[((k & CRPARMASK) >> 12) - 1]->c_bits;
cl_bits  = clist[(((k & CLPARMASK) >> 22) & P1023) - 1]->c_bits;

if(l_p == r_p) {return(0);}

resolvent = bres_replace(l_p,r_p,l_p_lit,r_p_lit);

c = clist[nextc - 1];

if(resolvent == 1) {
   base++;
   clauses_alive++;
   clist[nextc - 1]->c_id        = nextc;
   clist[nextc - 1]->c_lpar      = ( i != k ) ? l_p : r_p;
   clist[nextc - 1]->c_rpar      = ( i != k ) ? r_p : l_p;
   clist[nextc - 1]->c_indices   = ( i != k ) ? l_p_lit : r_p_lit;
   clist[nextc - 1]->c_indices   = 
                  clist[nextc - 1]->c_indices << LEFT_INDEX_SHIFT;
   clist[nextc - 1]->c_indices  += ( i != k ) ? r_p_lit : l_p_lit;
   clist[nextc - 1]->c_bits     |= MERGE | BASE; 

   if(((cr_bits & NC) != 0) || ((cl_bits & NC) != 0)) {
      clist[nextc - 1]->c_bits |= NC;} 

   c = clist[nextc - 1];
   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);

   if((c->c_lit->l_id & (~VARMASK)) == equal_pred) {
   if(test_equality(c->c_lit) == 1) {
      c->c_bits |= ACTIVE;
      if(o_elimprint) print0("..NO! SIMPLY DELETE CLAUSE\n");}}
   if(o_elimprint) {
      print_clause(clist[nextc - 1]);
         if(o_elimprint > 1) {getchar();}}}
 
return(resolvent);            /*  return to single resolvent  */
}
 
/*------------------------------------------------------------------*/
/*      bres_replace(c1,c2,l1_int,l2_int)    ................. #16
/*
/*      resolve the l1_int lit of c1 with the l2_int literal of c2
/*
/*      called by:      add_clause()                       elim.c
/*
/*      calls:          standardize()                      infer.c
/*                      clear_var_substitution_tble()      unify.c
/*                      unify()                            unify.c
/*                      generate_clause()                  gen.c
/*                      install_clause()                   search.c
/*                      print_clause()                     tdump.c
/*
/*      returns:        1 if clause is installed; 0 otherwise
/*------------------------------------------------------------------*/
 
int bres_replace(c1,c2,l1_int,l2_int)
clause *c1, *c2;
int l1_int, l2_int;
{
clause *c;
literal *l1, *l2, *ql, *pl;
int ix1, ix2, i, j, new;
	
ix1 = l1_int;  ix2 = l2_int;
	
standardize(c1,c2);              /* standardize variables apart  */

l1 = c1->c_lit;   
while(l1->l_id & ELIMLIT) {
   l1 = l1->l_next;
   ix1++;}
for ( j = 1 ; j < l1_int; j++) { 
   while(l1->l_id & ELIMLIT) {
      l1 = l1->l_next;
      ix1++;}
   l1 = l1->l_next;}

l2 = c2->c_lit;   
while(l2->l_id & ELIMLIT) {
   l2 = l2->l_next;
   ix2++;} 
for ( j = 1 ; j < l2_int; j++) {
   while(l2->l_id & ELIMLIT) {
      l2 = l2->l_next;
      ix2++;}
   l2 = l2->l_next;}

/*------------------------------------------------------*/
/* attempt unification if literals are complementary    */
/*------------------------------------------------------*/

clear_var_substitution_tble(c1->c_bits,c1->c_nv);
clear_var_substitution_tble(c2->c_bits,c2->c_nv);

in_elim = 1;                     /* necessary to elim tautologys */
if(unify(&l1->l_id,&l2->l_id)) {
   if(c = generate_clause(c1,c2,l1,l2,ix1,ix2)) {
      in_elim = 0;
      
/*------------------------------------------------------------------*/
/*      see if clause is already on clist by checking hashcodes
/*------------------------------------------------------------------*/
 
      new = 0;
      for(i = 0; i <= base; i++) {
         if(c->c_X == clist[i]->c_X && c->c_Y == clist[i]->c_Y) {
            new = 1;break;}}
      if(new == 1) {
	     if(o_elimprint) {
	        print_clause(c);
                print0("\n..... Wait!...\n ");
	        print0("Clause is in the clause hash table.\n");}
	     return(-1);}
     if(install_clause(c)) {
	     if(c->c_nlit == 0) simple_sol = 1;
	     return(1);} 
      return(0);}}
in_elim = 0;      /* added jan 8,2004 to match octopus elim */
return(0);
}
 
/*------------------------------------------------------------------*/ 
/*      bres_link(c1,c2,min_lit)           ................. #17
/*
/*      generate all resolvents of c1 and c2 and put them on clist
/*
/*      called by:      add_link_resolvents()              elim.c
/*
/*      calls:          standardize()                      bres.c
/*                      clear_var_substitution_tble()      infer.c
/*                      unify()                            unify.c
/*                      generate_clause()                  gen.c
/*                      install_clause()                   search.c
/*
/*      returns:        nr, the number of resolvents
/*------------------------------------------------------------------*/ 
 
int bres_link(c1,c2,min_lit)
clause *c1, *c2;
int min_lit;
{
clause *c;
literal *l1, *l2;
int ix1, ix2;  
int nr, id1, id2, cid1, cid2;
 
/*----------------------------------------------*/
/* check to see if both c1 and c2 are ACTIVE    */
/*----------------------------------------------*/
 
cid1 = c1->c_bits; 
cid2 = c2->c_bits;
if(!((cid1 & cid2) & ACTIVE)) {
   return(0);}
	
/*--------------------------*/
/* standardize vars apart   */
/*--------------------------*/
 
standardize(c1,c2);                  
	
/*------------------------------------------*/
/* set counter of number of resolvents to 0 */
/*------------------------------------------*/
 
nr = 0;
					     
ix1 = 1;
 
/*----------------------*/
/* scan down literals   */
/*----------------------*/
 
for(l1 = c1->c_lit; l1 != 0; l1 = l1->l_next) { 
   ix2 = 1;
   for(l2 = c2->c_lit; l2 != 0; l2 = l2->l_next) {
      id1 = l1->l_id & IDMASK;             /* predicate IDs  */
      id2 = l2->l_id & IDMASK;
 
/*----------------------------------------------------------*/
/* if IDs are the same then we are in a predicate group.    */
/* we attempt unification if literals are complementary.    */
/*----------------------------------------------------------*/
 
      if(id1 == id2 && (id1 >> ARGSHIFT) == min_lit) {
	     if(((l1->l_id ^ l2->l_id)&NEGMASK) == NEGMASK) {
	        clear_var_substitution_tble(c1->c_bits,c1->c_nv);
	        clear_var_substitution_tble(c2->c_bits,c2->c_nv);
	        in_elim = 1;
	        if(unify(&l1->l_id,&l2->l_id)) {
	           if(c = generate_clause(c1,c2,l1,l2,ix1,ix2)){
		          in_elim = 0;
		          if(install_clause(c)) {
		             c->c_bits |= BASE | MERGE;   
		             if(!c->c_nlit) {
		                simple_sol = 1;}
		             nr++;}}}
	        in_elim = 0;}}
  
      ix2++;}
   ix1++;}      
return (nr);
}

/*------------------------------------------------------------*/
/*      get_good_clauses()
/*
/*------------------------------------------------------------*/

void get_good_clauses()
{
clause *c;
int i;

if(o_learn == 0) return;

for(i = 0;i <= org_base + 1; i++) {
    if(good_clauses[i] > 0) {
       clist[i]->c_bits = clist[i]->c_bits | NC;}}
}

/*------------------------------------------------------------------*/
/* add_ulist_clauses() ..................................... #18
/*------------------------------------------------------------------*/

void add_ulist_clauses()
{
int i;
clause *c;
literal *l;

for(i = 0; i < nextu; i++) {
   c = ulist[i];
   if(c->c_bits & ACTIVE == 0) {continue;}
   if(c->c_nv >= o_maxvar_temp) {
      o_learn_var = o_maxvar_temp =  o_maxvar_temp + 3;}
   if(copy_clause4(c)) {
      base++;
      c = clist[nextc - 1];
      if(c->c_nlit > 2) c->c_bits &=~NC;
      for(l = c->c_lit; l != 0; l = l->l_next) {
         calc_lit_hc(l,c);}
      calc_clause_hashcode(c);
      add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
      if(c->c_nlit == 1) {
         normalize_literal(c->c_lit,c);
         add_clause_to_ht(c->c_lit->l_X,c->c_lit->l_Y,1);
         if(c->c_nv) {
            calc_alt_hashcodes(c->c_lit,c);}}}
   clauses_alive++;}
}

/*------------------------------------------------------------------*/
/* add_skolem_clauses() .......................................#19
/*
/* add clauses that result from the assumption that the skolem
/* constants are one of the constants mentioned in the other 
/* (base) clauses.
/*------------------------------------------------------------------*/

void add_skolem_clauses()
{
int i, j, *p;
clause *c;
literal *lit, *lit_new;

/*------------------------------------------*/
/* if equality clause Equal(x,x) not there  */
/* return!                                  */
/*------------------------------------------*/

return;
if(equal_clist_index <= 0) return;
/*if(number_constants == number_const_nc) return;*/

for(i = 1; i <= number_const_nc; i++) {
   if(skolem[i] == 0) {continue;}

/*-----------------------------------------*/
/* for each skolem constant, add a clause  */
/*-----------------------------------------*/

   copy_clause5(clist[equal_clist_index]);
   base++;

/*------------------------------------------*/
/* add literals and then replace variable x */
/* on one side of equality with skolem      */
/* constant and on the other with each      */
/* constant that is not a skolem constant   */
/*------------------------------------------*/

   c = clist[base];
   lit = c->c_lit;
   p = &lit->l_id;
   *++p = FCNMASK | (i << 5);
   *++p = 0;
   calc_lit_hc(lit,c);
   for(j = 1; j < number_constants - number_const_nc; j++) {
   printf("NC ");
      lit->l_next = (literal *) c->c_end; 
      lit_new = lit->l_next;
      lit_new->l_id = lit->l_id;
      lit_new->l_X = lit->l_X;
      lit_new->l_Y = lit->l_Y;
      lit_new->l_next = 0;
      lit = lit->l_next;
      p = &lit_new->l_id;
      *++p = FCNMASK | (i << 5);
      *++p = 0;
      c->c_nlit++;
      calc_lit_hc(lit_new,c);
      c->c_end = (clause *) ++p;}
   calc_clause_hashcode(c);   
   clist[base + 1] = c->c_end;
   printf("\n...");print_clause(c);getchar();
   clauses_alive++;}
return;
}

/*------------------------------------------------------------------*/
/*      eliminate()                 ..........................#20
/*
/*      Jan 26, 1991 (added to mac version,mn)
/*
/*      called by:      main()                  theo.c
/*
/*      calls:          time_now()              theo.c
/*                      res_ply_1()             elim.c
/*                      print_clause()          tdump.c
/*                      uncomple_lit()          elim.c
/*                      base_subsume()          elim.c
/*                      successor_subsume()     elim.c
/*                      single_resolvent()      elim.c
/*                      unit_resolvent()        elim.c
/*                      elim_link_lits()        elim.c
/*                      elim_long_cls()         elim.c
/*
/*      returns:        nil 
/*------------------------------------------------------------------*/
 
void eliminate()
{ 
int modify_cls;
int o_nc_temp,o_hash_temp,o_extend_temp;
int resol_count,cl_ptr,ii,i;
int input_char, input_char2;
int resolve_once;
clause *c;
	
/*----------------------------------------------------------*/
/* Allocate space for 6000 resolvents of the current base   */
/* clauses.  If more memory space is available to the user, */
/* the size of this array should be increased.              */
/* The array is not big enough to handle the "lions &       */
/* unicorn puzzle                                           */
/*----------------------------------------------------------*/

resolve_once = 0;
iter_depth = 1;
count_pass = 0;
 
elim_starting_time = time_now(); /* det starting time of elim proc */
	
o_nc_temp     = o_nc;        /* save search opts and restore     */
o_hash_temp   = o_hash;     /* when leaving subroutine          */
o_extend_temp = o_extend;
o_maxlit_temp = o_maxlit;
o_maxvar_temp  = o_maxvar;
o_maxterm_temp =o_maxterm;
o_merge_temp  = o_merge;
	
o_nc = o_hash = o_extend = o_merge = 0;
o_maxlit = 1000;
o_maxterm = 100;
o_maxvar = 30;
 
clauses_alive = base + 1;
modify_cls = 1;

for(ii = 0; ii <= 1000; ii++) {
   reinstate[ii] = reinstate_loop[ii] = 0;}

elim_long_cls(16);
get_good_clauses();
add_ulist_clauses(); 

/*----------------------------------------------------------*/
/* Top of main loop of Steps 0 - 5                          */
/* Loop back to this point until no clause is modified      */
/* or the number of clauses_alive is less than or equal     */
/* to one or a simple_sol has been found                    */
/*----------------------------------------------------------*/

while(modify_cls != 0 && clauses_alive > 1 && simple_sol == 0) {
   modify_cls = 0;
			
/*----------------------------------------------------------*/
/* Step 0                                                   */
/* Form a list of all binary resolvents of the base         */
/* clauses. The resolvents are placed on the list           */
/* elim_inferences.  Each resolvent takes one word.         */
/* the resolution is performed by bres_ply1                 */
/* on return, resol_count = the number of resolvents        */
/*----------------------------------------------------------*/
 
   resol_count = res_ply_1();            /* form all resol  */
   if(resol_count == -1)  {
         goto exit_elim;}   /* too many clses or NULL found */
   if(base > 900) {goto exit_elim;}

   if(o_elimprint) {
      printf("\nThere are %d base clauses;",clauses_alive);
      printf(" they have %d resolvents.",resol_count);
      OutPtr += sprintf(OutPtr,
	     "\nThere are %d base clauses;",clauses_alive);
      OutPtr += sprintf(OutPtr," they have %d resolvents.",resol_count);
      print0(
      "\n?(procede:'CR', list clauses:'c', list active clauses:'a')");
      if(o_elimprint == 2) {input_char = ' ';}
      else {input_char = getchar();}
      if(input_char == 'c' || input_char == 'a') {
	     for(ii = 0; ii <= base; ii++) {
		    if(input_char == 'c')  {
		       printf("\n");print_clause(clist[ii]);}
		    else if(clist[ii]->c_bits & ACTIVE) {
		       printf("\n");print_clause(clist[ii]);}}}}

/*----------------------------------------------------------*/
/* Step 1                                                   */
/* See if there is a literal that is unresolable            */
/* in some clause.  If so, eliminate the clause             */
/*----------------------------------------------------------*/
	
   if(resol_count != -1) {
      if(o_elimprint) {
         print0("u");}
      modify_cls = uncomple_lit(resol_count);}
 
/*----------------------------------------------------------*/
/* Step 2                                                   */
/* See if there is a base clause that s-subsumes another    */
/* If so, eliminate base clause that is subsumed            */
/*----------------------------------------------------------*/
 
   if(!modify_cls && resol_count != -1) {
      if(o_elimprint) {
         OutPtr += sprintf(OutPtr,"b%d ",base);
         printf("b%d ",base);}
      modify_cls += base_subsume();}
		
/*----------------------------------------------------------*/
/* Step 3                                                   */
/* Successor subsumption:       if the successor of some    */
/* base clause s-subsumes a base clause, replace base       */
/* clause with successor                                    */
/*----------------------------------------------------------*/

   if(!modify_cls) {
      if(o_elimprint) {
         OutPtr += sprintf(OutPtr,"s%d ",base);
         printf("s%d ",base);}
      modify_cls += successor_subsume();} 


/*----------------------------------------------------------*/
/* Step 4                                                   */
/* Single resolvent: if a literal in a clause resolves with */
/* only one other literal in one other clause, replace      */
/* the clause with its resolvent                            */
/*----------------------------------------------------------*/
 
   if(!modify_cls) {
      if(o_elimprint) {
         OutPtr += sprintf(OutPtr,"r%d ",base);
         printf("r%d ",base);}
      modify_cls += single_resolvent(resol_count);} 

/*----------------------------------------------------------*/
/* Step 5                                                   */
/* determine link literal in fewest clauses                 */
/* then replace clauses having link literal with their      */
/* resolvents                                               */
/*----------------------------------------------------------*/

   if(!modify_cls) {
      if(o_elimprint) {
         print0("l");}
      modify_cls += elim_link_lits();}

   if(!modify_cls && !resolve_once) {
      unit_resolvent();
      resolve_once++;modify_cls++;}}

/*----------------------------------------------------------*/
/* Step 6                                                   */
/* Eliminate long clauses (longer than o_maxlit_temp)       */
/*----------------------------------------------------------*/

if(o_elimprint) {
   print0("x");}
elim_long_cls(o_maxlit_temp);

/*----------------------------------------------------------*/
/* Exit eliminate: return variables to original values      */
/* then free space taken by elim_inferences                 */
/*----------------------------------------------------------*/

add_skolem_clauses();
 
exit_elim:              

for(i = 0; i < 9; i++) {
   for(ii = 0; ii <= base; ii++) {
      if(reinstate[ii] == i && i != 2) {
         clist[ii]->c_bits |= ACTIVE;}}}

o_nc        = o_nc_temp;
o_hash      = o_hash_temp;
o_extend    = o_extend_temp;
o_maxlit    = o_maxlit_temp;
o_maxterm   = o_maxterm_temp;
o_maxvar    = o_maxvar_temp;
o_merge     = o_merge_temp;
ply         = 0;
 
/*----------------------------------------------------------*/
/* User interface:                                          */
/* when e-option is on, user can now quit (q), print        */
/* the clauses (c), or print only the "ACTIVE" clauses, ie, */
/* the clauses that will be used in the search for a        */
/* contradiction, or proceed (p).                           */
/*----------------------------------------------------------*/
 
if(o_elimprint) {
   print0("\n\nExit eliminate:\n");
decide:
   printf("proceed:'CR'; quit:'q'; list clauses: 'c';");
   printf(" list active clauses: 'a'\n?");
get_character:
   if(o_elimprint == 2) {input_char = '\n';}
   else {input_char = getchar();}
   if(input_char != '\n') {
      while((input_char2 = getchar()) != '\n') { };}
   if(input_char == 'c' || input_char == 'a') { /* print clauses   */
      print0("\n");
      for(cl_ptr = 0; cl_ptr <= base; cl_ptr++ ) {
       if(input_char=='a'&&(clist[cl_ptr]->c_bits & ACTIVE) == 0) {
            continue;}
	       printf("\n");
	       print_clause(clist[cl_ptr]);}
      print0("\n");
      goto decide;}
 
   else if(input_char == 'q') simple_sol = 1; /* quit; no search    */
   else if(input_char == '\n') {
      print0("\n");}                          /* proceed to search  */
   else {
      printf("\n? ");
      goto decide;}}                        /* re-enter a character */
 
/*-------------------------------------------------------*/
/* print time taken by eliminate(), followed by          */
/* the original number of base clauses and the revised   */
/* number after eliminate()                              */
/*-------------------------------------------------------*/
 
eliminate_time = time_now() - elim_starting_time;
printf("   %d <BC: %d NC: %d AC: %d U: %d>\n",
   eliminate_time,org_base + 1,nc,clauses_alive,units);
OutPtr += sprintf(OutPtr,"   %d <BC: %d NC: %d AC: %d U: %d>\n",
      eliminate_time,org_base + 1,nc,clauses_alive,units);

return;
}
