#! /usr/bin/python

# finds TPTP problems (in tme format) which match certain criteria

import os
import re


TPTP_PATH = "/home/alex/TPTP/protein/"
SUFFIX = ".tme"



#
# filter functions
#
# return true for problems matching the criterion

# filter by problem status
def filter_status (problem, status):
    for line in problem:
        match = re.search ("^% Status   : (?P<value>.*)$", line)
        if match:
            if match.group ("value") == status:
                return 1
            
            else:
                return 0

def filter_satisfiable (problem):
    # % Status   : satisfiable
    filter_status (problem, "satisfiable")

def filter_unsatisfiable (problem):
    # % Status   : unsatisfiable
    filter_status (problem, "unsatisfiable")


# drop non-Horn problems
def filter_horn (problem):
    for line in problem:
        # % Syntax   : Number of clauses    :   20 (   0 non-Horn;   2 unit;  20 RR)
        match = re.search ("(?P<value>\d+) non-Horn", line)
        if match:
            if match.group ("value") == "0":
                return 1
            
            else:
                return 0

    return 0


# drop problems with equality
def filter_non_equality (problem):
    for line in problem:
        # %            Number of literals   :   38 (   0 equality)
        match = re.search ("(?P<value>\d+) equality", line)
        if match:
            if match.group ("value") == "0":
                return 1
            
            else:
                return 0

    return 0


# horn problems without equality
def filter_hne (problem):
    return filter_horn (problem) and filter_non_equality (problem)

# non-horn problems without equality
def filter_nne (problem):
    return (not (filter_horn (problem))) and filter_non_equality (problem)


# invalid specifications
def filter_non_FOF (problem):
    for line in problem:
        match = re.search ("%----FOF format not yet installed in PROTEIN", line)
        if match:
            return 0

    return 1
    
# top filter function
def filter (problem):
    return (
        filter_non_FOF (problem)
        and
        #filter_hne (problem)
        filter_nne (problem)
        )
    



# check every file if it's a wanted problem description
def check_problem_file (problems, dir_name, files_in_dir):
    for file_name in files_in_dir:
        file_path = os.path.join (dir_name, file_name)

        if ((file_name[-len(SUFFIX):] == SUFFIX)
            and
            (os.path.isfile (file_path))):
            try:
                file = open (file_path)
                problem = file.readlines ()
                file.close ()

                if filter (problem):
                    problems.append (file_name)

            except IOError, (error_nr, error_string):
                print ("Couldn't open " + file_path + ": " + error_string)



# get all wanted problems and print them line by line
problems = []
os.path.walk (TPTP_PATH, check_problem_file, problems)
problems.sort ()
for problem in problems:
    # remove the SUFFIX
    print problem[:-len(SUFFIX)]
