#include "FINDER.h"

#ifdef SOFTIE_SCOTT
#include "SOFTIE_SCOTT.h"
#endif

/*------------------------------------------------------------------------*/
void initialize()

     /*Called by init_finder().
      *The first function is called from main right at the start.
      *The name says it all.*/

{
  int i, j;
  functor f;
  kind k;

  init_sets();  

  /*---First zero all the array parts of array elements which are ------
   ----------to be null iff this element is undefined.-----------------*/

  for ( i = 0; i < LEX_MAX; i++ ) lexicon[i].word[0] = 0;
  for ( i = 0; i < LEX_MAX; i++ ) lexicon[i].iskind = false;
  for ( i = 0; i < ST_MAX; i++ ) sort[i].sname = 0;
  for ( i = 0; i < FN_MAX; i++ ) fn[i].fname = 0;
  for ( i = 0; i < WF_MAX; i++ ) form[i].sym = 0;
  for ( i = 0; i < MAXISET_ARRAY; i++ ) {
    for ( j = 0; j < CL_MAX; j++ ) {
      theory[i][j].prem = theory[i][j].conc = 0;
      theory[i][j].nextclause = theory[i][j].lastclause = 0;
    }
  }
  current_theory = 0;
  latest_clause.prem = 0;
  latest_clause.conc = 0;
  latest_clause.testlist = 0;    
  latest_clause.varlist = 0;
  latest_clause.nextclause = 0;
  latest_clause.lastclause = 0;
  latest_clause.softweight = 0;
  latest_clause.preprocessed = false;
  FN = 0;

  /*-------Set up the pre-defined types bool and int, and the ----------
   ------pre-defined relation and function symbols =, <, > and +.------*/

  k = newkind( "bool" );
  k->siz = k->sizmin = k->sizmax = 2;
  canon( k, "FALSE" );  canon( k, "TRUE" );
  /* convention wrt special functors is the same as in OTTER */

  f = newfunctor( "NOT", Func, 1 );
  f->val_type = f->arg_type[0] = k;
  f->scope = 1;	f->table = neg;

  f = newfunctor( "=", Func, 2 );
  f->val_type = k;	f->scope = 10;	f->table = id;
  /*  f = newfunctor( "equal", Func, 2 );
  f->val_type = k;	f->scope = 10;	f->table = id;*/
  f = newfunctor( "<", Func, 2 );
  f->val_type = k;	f->scope = 10;	f->table = less;
  f = newfunctor( ">", Func, 2 );
  f->val_type = k;	f->scope = 10;	f->table = greater;	
  k = newkind( "int" );
  k->sizmin = k->sizmax = SZ-1;
  f = newfunctor( "DIF", Func, 2 );
  f->val_type = k;	f->scope = 5;	f->table = difference;	
  f = newfunctor( "+", Func, 2 );   /* The dummy typeless addition */
  f = newfunctor( "-", Func, 2 );   /* The dummy typeless subtraction */
  f = newfunctor( "EST", Func, 1 );  /* The dummy typeless "exists" */
  f = newfunctor( "LEN", Func, 1 ); /* The dummy typeless length */

  /*----------Defualt values for parameters ----------------------------*/

  setting.max_models = 0;      /*solutions within a theory*/
  setting.max_backtracks = 0;
  setting.max_invalues = 0;
  setting.random = false;

#ifdef SOFTIE_SCOTT
  setting.pretest_limit = SHORT_CLAUSE_CUTOFF;  // synchronize short clause cutoff as set in SOFTIE_SCOTT.h
#else
  setting.pretest_limit = 3;  // the standard SFINDER pretest limit
#endif

  setting.batches = S_MAXIMAL;
  setting.extra_batches = 5;
  setting.maxref = MAX_REF_SIZE;
  setting.test_maxref = MAX_REF_SIZE;
  setting.maxtime = 0;
  setting.target_bound = MAX_PENALTY;
  setting.topsoff = true;
  setting.vars_over_wt = true;
  setting.report_breaks = true;
  setting.noisy = false;
  setting.sortcut = true;
  setting.timestamps = false;
  setting.Nosecs = false;
  setting.invert_change = false;
  setting.pm.verbosity = pfmodels | pbstats;
  setting.pm.pretty = true;
 
  /*----------Place the addresses of the test functions in the -----------
   --------------- structures to be passed to transref ------------------*/

  init_trin( &tr_parameters );
  tr_parameters.Test = test;
  tr_parameters.Onerefs = setposs;
  tr_parameters.Smallrefs = do_smallrefs;
  
  init_trin( &ttr_par );
  ttr_par.Test = model_t;
  ttr_par.Onerefs = single_t;
  ttr_par.Smallrefs = tuple_t;

  /*-----Next there are various integers, booleans and characters.-------*/

  neg[0] = 2; neg[1] = 1; neg[2] = 0;
  id[0] = less[0] = greater[0] = difference[0] = SZ;
  for ( i = 0; i < SZ; i++ ) 
    for ( j = 0; j < SZ; j++ ) {
      id[(SZ*i)+j+1] = (i == j);
      less[(SZ*i)+j+1] = (i < j);
      greater[(SZ*i)+j+1] = (i > j);
      difference[(SZ*i)+j+1] = (i < j? j-i: i-j);
    }
  zero = 0;
  
  for ( i = 0; i < V_LENGTH; i++ )  tr_parameters.forbidden[i] = false;

  for (i=0; i< MAXISET_ARRAY; i++) {
    for ( j = 0; j < ST_MAX; j++ ) {
      last_sizes[i][j] = 0;
      emptyset(last_sizmasks[i][j]);
    }
    for (j=0; j<V_LENGTH; j++) last_model[i][j] = 0;
    last_vlength[i] = 0;
    TH[i] = theory[i];
  }
  
  over_argmax = false;
  cmd_line_input = false;
  normal_input = true;
  single_model = false;
  cc_badguy = false;
  for ( i = 0; i < VARMAX; i++ ) varlex[i][0] = 0;
  dummy_symbol = (functor) malloc(sizeof(FUNCTOR));
  dummy_symbol->fname = 0;
  first_newval = true;
}
 /* initialize() */


/*------------------------------------------------------------------------*/
void canon(kind k, char new_name[])

     /*Called by initialise() and sort_spec().
      *This function adds a canonical name new_name to the enumeration
      *of the objects of sort k, setting sizmin and sizmax.*/

{
  name_list *lpp;
  functor f;

  f = newfunctor( new_name, Const, 0 );
  f->val_type = k;

    /*count the number of elements in the enumeration -- the hard way*/
  k->siz = 1;
  for (lpp = &(k->canonical); *lpp; lpp = &((*lpp)->nextname) )
    k->siz++;

  if (!(f->table)) f->table = (int*) malloc(sizeof(int));
  *(f->table) = k->siz-1;

  k->len[k->siz-1] = strlen(f->fname);

  *lpp = (name_list) malloc(sizeof(NAME_LIST));

  (*lpp)->name = f->fname;
  (*lpp)->nextname = 0;

  k->sizmin = k->sizmax = k->siz;
} /* canon() */


/*------------------------------------------------------------------------*/
int lex_pos(char word[])

     /*Called by newkind() and newfunctor() (in this file)
      *and various functions in parser.c and input.c.
      
      *Lex_pos scans the lexicon for the given "word", returning
      *its position in the lexicon. 
      *If it is new to the lexicon then returns -1.*/
     
{
  int i;
  
  for ( i = 0; lexicon[i].word[0]; i++ )
    if (!strcmp(word,lexicon[i].word)) return i;
  return -1; /* indicates that word is new */

} /* lex_pos */



/*------------------------------------------------------------------------*/
kind newkind( char new_name[] )

     /*Called by initialise() and get_sort().
      *If the sort identifier with name "new_name" already exists, 
      *return a pointer to it. Otherwise add it to the list, put its name
      *in the lexicon and initialize it to the default values.*/

{
  int i,k;
  functor f;
  dynstring string;

  dynstring_initialise(&string, "");
  
  if ( (i = lex_pos(new_name)) != -1){
    if (lexicon[i].iskind)
      return( sort+lexicon[i].offset );  /*return pointer to existing sort*/
    else
      inerror(new_name,"doubly defined as a sort and a functor");
  }
  /*find first undefined word*/
  for ( i=0; (i < LEX_MAX) && lexicon[i].word[0]; i++ );

  /*find first undefined sort*/
  for ( k=0; (k < ST_MAX) && sort[k].sname; k++ );

  lexicon[i].offset = k;
  lexicon[i].iskind = true;
  strcpy(lexicon[i].word,new_name);
  sort[k].sname = lexicon[i].word; /*sort points to its name in the lexicon*/
  sort[k].sizmin = 1;  
  sort[k].sizmax = SZ-1;
  sort[k].canonical = 0;
  sort[k].len[0] = 0;

  /* Define builtins for new sort */

  dynstring_strcpyc(&string,"+_"); dynstring_strcatc(&string,new_name);
  f = newfunctor( string.c_str, Func, 2 );
  f->total = false;
  f->arg_type[0] = &sort[k];
  f->arg_type[1] = Integer;
  f->val_type = &sort[k];
  f->table = sort[k].sum;

  dynstring_strcpyc(&string,"-_"); dynstring_strcatc(&string,new_name);
  f = newfunctor( string.c_str, Func, 2 );
  f->total = false;
  f->arg_type[0] = &sort[k];
  f->arg_type[1] = Integer;
  f->val_type = &sort[k];
  f->table = sort[k].minus;

  dynstring_strcpyc(&string,"EST_"); dynstring_strcatc(&string,new_name);
  f = newfunctor( string.c_str, Func, 1 );
  f->arg_type[0] = &sort[k];
  f->arg_type[1] = 0;
  f->val_type = Bool;
  f->table = sort[k].Ex;
 
  dynstring_strcpyc(&string,"LEN_"); dynstring_strcatc(&string,new_name);
  f = newfunctor( string.c_str, Func, 1 );
  f->arg_type[0] = &sort[k];
  f->arg_type[1] = 0;
  f->val_type = Integer;
  f->table = sort[k].len;

  for ( i = 0; i < SZ; i++ ) sort[k].len[i] = (i < 10? 1: 2);
  sort[k].piped = pipe_open;

  dynstring_free(&string);

  return &sort[k];
} /* newkind() */




/*------------------------------------------------------------------------*/
functor newfunctor( char new_name[], syntype category, int adicity )

     /*Called by initialize(), newsort(), and conon() (in this file)
      *and by input.c and parser.c.
      *If the functor identifier with name "new_name" already exists, 
      *in ther functor stack fn[] return a pointer to it. 
      *Otherwise add it to the stack, put its name in the lexicon and 
      *initialize it to the default values. */
{
  int i, j;
  functor f;
  
  if (adicity > ARGMAX) inerror(new_name,"given too many argument places");
  
  if ( (i = lex_pos(new_name)) != -1){
    if (!lexicon[i].iskind)
      return( fn+lexicon[i].offset );
    else
      inerror(new_name,"doubly defined as a sort and a functor");
  }
  /*find first undefined word   */
  for ( i = 0; (i < LEX_MAX) && lexicon[i].word[0]; i++ ); 

  /*find first undefined functor*/
  for ( j = 0; (j < FN_MAX) && fn[j].fname; j++ );  

  lexicon[i].offset = j;
  lexicon[i].iskind = false;
  strcpy(lexicon[i].word,new_name);
  fn[j].fname = lexicon[i].word;
  fn[j].category = category;
  fn[j].scope = adicity? adicity-1: 0;
  fn[j].adicity = adicity;

  if ( category == Var )
    inerror(new_name,"both a variable and a functor");
  /* for redefinitions of prexisting functors */
  else if ( new_name[0] == '=' ) fn[j].table = id;
  /*  else if ( !strcmp(new_name, "equal") ) fn[j].table = id; */
  else if ( new_name[0] == '<' ) fn[j].table = less;
  else if ( new_name[0] == '>' ) fn[j].table = greater;
  else if ( !strcmp(new_name,"DIF") ) fn[j].table = difference;
  else if (category==Const) fn[j].table = (int*) malloc(sizeof(int));
  else if ( !strchr("+-",new_name[0]) )
    fn[j].table = 0;
  for (i=0; i<ARGMAX; i++) fn[j].fats[i] = 1;
  if (strchr("=><",new_name[0]) || !strncmp(new_name,"DIF",3)) {
    fn[j].fats[0] = SZ;
    fn[j].fats[1] = 1;
  }

  fn[j].fcc = 0;
  for (i=0; i < MAXISET_ARRAY; i++) fn[j].old_fcc[i] = 0;
  fn[j].total = true;
  fn[j].injective = fn[j].leftin = fn[j].rightin = false;
  fn[j].surjective = fn[j].leftsur = fn[j].rightsur = false;
  fn[j].comm = false;
  fn[j].nop = false;
  fn[j].cut_it = false;
  fn[j].piped = pipe_open;
  for (i=0; i<ARGMAX; i++) fn[j].arg_type[i] = 0;
  fn[j].val_type = 0;
  fn[j].pmode = null;
  fn[j].nextfunct = 0;

  /*Append it to the testable functor stack FN[] */
  if ( category != Const && !strchr("+-<>=",new_name[0]) )
/*    if ( strcmp(new_name,"equal" ) ) */
    if ( strncmp(new_name,"EST",3) && strncmp(new_name,"LEN",3) &&
	 strncmp(new_name,"NOT",3) && strncmp(new_name,"DIF",3) ) {
      if ( !FN ) FN = &fn[j];
      else {
	for ( f = FN; f->nextfunct; f = f->nextfunct );
	f->nextfunct = &fn[j];
      }
    }
  
  return &fn[j];
} /* newfunctor() */

/*=============================== END =====================================*/
