/*------------------------------------------------------------------------*/
/* Main routines for obtaining finder input */
/*------------------------------------------------------------------------*/

#include "FINDER.h"
#include <stdlib.h>

static dynstring job_line; /* Line of input text		    */
static int  current_line = 0; /* Current line number		    */
static int  input_point = 0;  /*Current input point on the job_line */
static FILE *source;	      /* Current input stream               */


/*------------------------------------------------------------------------*/
char next_char()

     /*Called by get_a_char() and get_word().
      *Returns the next character from the job_line[]
      *Skipping over comments and if job_line[] is exhausted, reads
      *the next input line.
      */

{
  while (job_line.c_str[input_point] == '%' ||  /* skip comments    */
	 job_line.c_str[input_point] == '\0') { /* skip empty lines */

    if (source &&                       /* if reading from a file */
	!dynstring_fgets(&job_line,source)) /* refill the job_line */
      return '\0';                 /* patch for end_of_file */
    input_point = 0;               /* restart at the begining of a line */
    current_line++;                /* have moved to the next line */
  }	

  return job_line.c_str[input_point++]; /* next charater on the job_line */

} /* next_char() */



/*------------------------------------------------------------------------*/
char get_a_char()

     /*Called by get_word().
      *Returns next meaningful character from the job_line*/
     
{
  unsigned char c;

  while ( isspace(c = next_char()) ) ;
  return c;

} /* get_a_char() */



/*------------------------------------------------------------------------*/
void get_word( dynstring *word )
     /*Called by get_key_word().
      *most fcn's in this file and by fcn's in parser.c;
      
      *Gets next meaningful word from the job_line*/
{
  int i;

  dynstring_ensure_space(word, 1);

  word->c_str[0] = get_a_char();

  if ( strchr("=><+,;:.%)(}{",word->c_str[0]) )
    word->c_str[1] = '\0';
  else {
    for (i=1; !(isspace(word->c_str[i] = next_char())); i++ ) {
      dynstring_ensure_space(word, i + 1);
      if (i==1 && !strncmp(word->c_str,"->",2)) {
	i = 2;
	break;
      }
      else if ((i==1 && word->c_str[0]=='-') ||
	       strchr("=><+-,;:.%)(}{",word->c_str[i])) {
	input_point--;
	break;
      }
    }
    word->c_str[i] = '\0';
    word->length = strlen(word->c_str);
  }
} /* get_word() */



/*------------------------------------------------------------------------*/
void input_to_terminator(char *s)

     /* Append input characters to the job_line until a character from
      * s turns up. Incidentally turn newlines into spaces */
{
  char c;
  int i;

  for (i=0; job_line.c_str[i]; i++)
    if (isspace(job_line.c_str[i])) job_line.c_str[i] = ' ';
  do {
    c = fgetc(source);
    if (isspace(c)) c = ' ';
    dynstring_append_char(&job_line, c);
  }
  while (!strchr(s,c));
}




/*------------------------------------------------------------------------*/
statement_type get_key_word()

     /*Called by read_job().
      *Gets next key word from the job_line and determines its type.*/
     
{
  dynstring wd;
  statement_type result = nul;

  dynstring_initialise(&wd, "");
  
  get_word(&wd);
  if ( !strcmp(wd.c_str,".") || !strcmp(wd.c_str,"end") ) result = nul;
  else if ( !strcmp(wd.c_str,"sort") ) result = srt;
  else if ( !strcmp(wd.c_str,"function") ) result = fcn;
  else if ( !strcmp(wd.c_str,"constant") ) result = fcn0;
  else if ( !strcmp(wd.c_str,"clause") ) result = cls;
  else if ( !strcmp(wd.c_str,"setting") ) result = prm;
  else if ( !strcmp(wd.c_str,"include") ) result = inc;
  else inerror
           (wd.c_str,"found where key word (sort/function/clause/setting/end) expected");
  dynstring_free(&wd);
  return result;
} /* get_key_word() */



/*------------------------------------------------------------------------*/
void read_job( FILE *infil )
     
     /*Called by FINDER fcn main() and SCOTT fcn model_finder_file() .
      *Loops continuously over input file until get_key_word() returns null.
      *Integrates the object according to its type*/
{
  if ( cmd_line_input )
    dynstring_strcpy(&job_line,&command_line);
  else
    dynstring_clear(&job_line);
  input_point = 0;
  source = infil;
  for (;;)
    switch( get_key_word() ) {
    case nul:
      return;
    case srt:
      get_item(get_sort,0,0);
      break;
    case fcn:
      get_item(get_func,0,0);
      break;
    case fcn0:
      get_item(get_const,0,0);
      break;
    case cls:
      get_item(get_input_cl,0,0);
      break;
    case prm:
      get_item(get_set,0,0);
      break;
    case inc:
      get_incl_file();
      break;
    }
} /* read_job */



/*------------------------------------------------------------------------*/
void get_incl_file()

     /*Called by
      *Get_incl causes a file to be read as though it were
      *included at the current point. Note that it must end
      *with the word `end'. The file name is enclosed in 
      *(round) parentheses. Since the routine get_a_char() 
      *is used, the name may have white space such as line
      *breaks inserted in it: these will be ignored.
      */
     
     
{
  FILE *oldsource, *infil;
  dynstring include_filename;
  dynstring oldjobline;
  int oldinputpoint, i;

  dynstring_initialise(&oldjobline, "");
  dynstring_initialise(&include_filename, "");
  
  get_word( &include_filename );
  if ( include_filename.c_str[0] != '(' )
    inerror("Included filename must be in parentheses","");
  
  i = 0;
  dynstring_ensure_space(&include_filename, 0);
  while ( (include_filename.c_str[i] = get_a_char()) != ')' ) {
    i++;
    dynstring_ensure_space(&include_filename, i);
  }
  include_filename.c_str[i] = '\0';
  
  oldsource = source;
  dynstring_strcpy(&oldjobline, &job_line);
  oldinputpoint = input_point;
  if ( !(infil = fopen(include_filename.c_str,"r")) )
    inerror("Cannot open file",include_filename.c_str);
  read_job( infil );
  fclose(source);
  source = oldsource;
  dynstring_strcpy(&job_line, &oldjobline);
  input_point = oldinputpoint;

  dynstring_free(&oldjobline);
  dynstring_free(&include_filename);

} /* get_incl_file() */


/*------------------------------------------------------------------------*/
int get_int()

     /*Called by get_set(), get_item(), sort_spec(), func_spec(), 
      *and got_function().
      *Gets an expected unsigned integer charater from the input line.
      *Exits in error if one is not found.*/

{
  dynstring s;
  int rc, i;
  int sign = 1;

  dynstring_initialise(&s, "");
  
  do
    get_word(&s);
  while ( s.c_str[0] == ':' );
  rc = 0;
  for ( i = 0; s.c_str[i]; i++ ) {
    if ( i==0 && s.c_str[0]=='-' )
      sign = -1;
    else {
      if ( !isdigit(s.c_str[i]) )
	inerror(s.c_str,"found where integer expected");
      rc *= 10;
      rc += (s.c_str[i] - '0');
    }
  }
  /*
   *  if (rc > 63 || sign < 0)
   *    inerror("Integers must be in range 0...60",
   *	    "(system still under development)");
   */

  dynstring_free(&s);
  
  return (rc * sign);

} /*get_int()*/

/*------------------------------------------------------------------------*/
float get_float()

     /*Not used at present. Could be used to get priorities etc.
      *Gets an expected float from the input line.
      *Exits in error if one is not found.*/
{
  dynstring word;
  int digit;
  float sum, decimal;

  dynstring_initialise(&word, "");

  do
    get_word(&word);
  while ( word.c_str[0] == ':' );
  sum = 0;
  for ( digit = 0; word.c_str[digit]; digit++ ) {
    if ( word.c_str[digit] == '.')
      break;
    if ( !isdigit(word.c_str[digit]) )
      inerror(word.c_str,"found where integer expected");
    sum *= 10;
    sum += (word.c_str[digit] - '0');
  }

  do
    get_word(&word);
  while ( word.c_str[0] == '.' );
  decimal = 1;
  for (digit = 0; word.c_str[digit]; digit++ ) {
    if ( !isdigit(word.c_str[digit]) )
      inerror(word.c_str,"found where integer expected");
    decimal *= 10;
    sum += ((word.c_str[digit] - '0') / decimal);
  }

  return sum;

} /*get_float()*/


/*------------------------------------------------------------------------*/
void get_item( void (*item)(), kind k, functor f )

     /*Called by read_job(), sort_spec(), get_func(), get_cons(), 
      *and get_set().

      *Calls the appropriate function (*item)() to get either
      *one item of the given species (e.g. one clause) or a set of them.
      *A set, naturally, is specified by listing its members within a 
      *pair of braces.
      *
      *Some items like clauses have an optional integer parameter given
      *in parentheses after the key word, so get_item must search for 
      *this first.
      *
      *Functions called by get_item have all four of the parameters it 
      *issues, but generally don't use most of them.  Although the above
      *function read_job does not make use of the parameters k and f, 
      *functions (*item)() may do so when they themselves call get_item 
      *for their sub-items.
      *All calls to (*item)() are pased the first word of the item.*/

{
  dynstring s;
  int arg = 0;

  dynstring_initialise(&s, "");

  get_word(&s);

  if ( s.c_str[0] == '(' ) {  /* initial arg of item is an int in parentheses */
    arg = get_int();
    get_word(&s); /*get closing parentheses */
    if ( s.c_str[0] != ')' ) inerror("Missing right parenthesis","");
    get_word(&s); /*get string following integer argument */
  }

  if ( s.c_str[0] == '{' ) {  /* list of items in braces */
    get_word(&s);
    while ( s.c_str[0] != '}' ) {
      if ( s.c_str[0] != '.' ) (*item)(k,f,&s,arg);
      get_word(&s);  /* get next item */
    }
  }
  else if ( s.c_str[0] != '.' ) (*item)(k,f,&s,arg);   /* single item */

  dynstring_free(&s);
}



/*------------------------------------------------------------------------*/
void get_sort( kind k, functor f, dynstring *name, int x )

     /*Called by get_item().
      *Creates a new sort if "name" does'nt already exist.
      *Gets a (possibly) revised specification for that sort.*/
{	
  get_item( sort_spec, newkind(name->c_str), 0 );

} /* get_sort() */



/*------------------------------------------------------------------------*/
void sort_spec( kind k, functor f, dynstring *spec, int x )

     /* Called by get_item().
      *	Revises the specifications (if any) for the kind k.*/

{
  int canon_count;

  if ( !strcmp(spec->c_str,"cardinality") || !strcmp(spec->c_str,"card") ) {
    if ( k->canonical ) {
      inerror(k->sname,"an enumerated sort - can't set cardinality");
      return;
    }
    get_word(spec);
    switch (spec->c_str[0]) {
    case '=':
      k->sizmax = k->sizmin = get_int();
      return;
    case '<':
      k->sizmax = get_int()-1;
      if ( k->sizmax < k->sizmin )
	k->sizmin = k->sizmax;
      return;
    case '>':
      k->sizmin = get_int() + 1;
      if ( k->sizmax < k->sizmin )
	k->sizmax = k->sizmin;
      return;
    }
    inerror("","Statement of cardinality needs =, < or >");		
  }
  if (!strcmp(spec->c_str,"enum") ) {
    de_enumerate( k );
    for ( canon_count = 0; canon_count < SZ; canon_count++ ) {
      get_word(spec);
      if ( spec->c_str[0] == ':' )
	get_word(spec);
      canon( k, spec->c_str );
      get_word(spec);
      if ( spec->c_str[0] == '.' )
	return;
      if ( spec->c_str[0] != ',' )
	inerror(spec->c_str,"found where comma expected");
    }
  }
  inerror(spec->c_str,"found where key word \"cardinality\" or \"enum\" expected");

} /* sort_spec() */



/*------------------------------------------------------------------------*/
void de_enumerate( kind k )

     /*Called by sort_spec().
      *Removes the existing enumeration (if any) of kind k. */
{
  free_canon( k->canonical );
  k->canonical = 0;

} /* de_enumerate() */


/*------------------------------------------------------------------------*/
void free_canon( name_list lp )

     /*Called by de_emumerate().
      *Recusively frees pointers to the cannical names in the 
      *enumeration list. */
{
  if ( !lp ) return;
  free_canon( lp->nextname );
  free( lp->nextname );
  lp->nextname = 0;

} /* free_canon() */



/*------------------------------------------------------------------------*/
void get_func( kind k, functor f, dynstring *name, int x )
     
     /*Called by get_item(), and get_cons().
      *Scans for the argument and value types of the (new) functor. 
      *If the functor all ready exists then this is redefined.
      *Since newfuntor() is called and sets the print mode to null
      *(which it needs to do because it usually sets up variables, etc) 
      *so this function resets it to the default full.
      *Then calls func_spec() to scan any function specifications.
 
      *Parameters k, f and x are dummies for get_func.  The string s
      *holds the name of the function. */
     
{
  int i;
  int arity=0;
  boolean got_arg_types=false;
  boolean got_val_type=false;
  dynstring type_str, punctuation;
  dynstring fcn_types[ARGMAX+1];

  dynstring_initialise(&type_str, "");
  dynstring_initialise(&punctuation, "");
  
  for ( i = 0; i <= ARGMAX; i++ ) dynstring_initialise(&fcn_types[i], "");
  
  get_word(&type_str);
  
  if ( type_str.c_str[0] == ':' ) get_word(&type_str); /* skip ':' */

  /* get the types of the functor's arguments, and then its value type */
  do {
#ifdef SCOTT
    if (arity > ARGMAX){
      inerror("Adicity too great for function",name->c_str);
      over_argmax = true; 
      return;
    }
#else
    if (arity > ARGMAX) inerror("Adicity too great for function",name->c_str);
#endif
    dynstring_strcpy(&fcn_types[arity++],&type_str); /*record the next type*/
    if (got_arg_types) {
      got_val_type = true;
      arity--; /* the functions values is stored as the last type */
    }
    else {
      get_word(&punctuation);

      if (!strchr(".,-",punctuation.c_str[0]))
	inerror(type_str.c_str,"found where punctuation ( , -> or . ) expected");

      switch(punctuation.c_str[0]) {
      case '.':
	arity--;
	got_val_type = true;
	break;
      case ',': 
	if (got_arg_types) inerror(name->c_str,"can only have ONE value type");
	get_word(&type_str); /* get next argument type */
	break;
      case '-':
	if ( punctuation.c_str[1] != '>' )
	  inerror(punctuation.c_str,"found where \"->\" expected");
	got_arg_types = true;
	get_word(&type_str); /* get value type */
	break;
      }
    }
  }
  while (!got_val_type);

  f = newfunctor( name->c_str, Func, arity);
  f->val_type = set_up_type(&fcn_types[arity]);
  for (i=0; i<arity; i++)
    f->arg_type[i] = set_up_type(&fcn_types[i]);
  f->pmode = full;
  get_item( func_spec, 0, f );

  dynstring_free(&type_str);
  dynstring_free(&punctuation);
  
  for ( i = 0; i <= ARGMAX; i++ ) dynstring_free(&fcn_types[i]);
  
} /* get_func */



/*------------------------------------------------------------------------*/
void get_const( kind k, functor f, dynstring *name, int x )
     
     /*Called by get_item().
      *Calls get_func() for 0-adic function symbol. */

{
  dynstring const_type;

  dynstring_initialise(&const_type, "");

  f = newfunctor( name->c_str, Func, 0 );

  do	get_word(&const_type);
  while ( const_type.c_str[0] == ':' );

  f->val_type = set_up_type(&const_type);
  f->pmode = full;
  get_item( func_spec, 0, f );

  dynstring_free(&const_type);

} /* get_const() */




/*------------------------------------------------------------------------*/
kind set_up_type( dynstring *type )

     /*Called by get_func() and get_const().
      *Returns a pointer to the kind coresponding to the given "type".*/
{
  int i;
  
  for ( i = 0; sort[i].sname; i++ ) 
    if ( !strcmp(sort[i].sname,type->c_str) ) return( &sort[i] );
  inerror("Non-existent sort",type->c_str);
  return 0;
} /* set_up_type() */




/*------------------------------------------------------------------------*/
void func_spec( kind k, functor f, dynstring *key_word, int x )

     /* Called by get_func() and get_const().
      *Sets up the various properties of the function that can be 
      *specified in the section following its declaration. 
      *(The paramters x and f are just dummies). */
     
{
  dynstring specs;

  dynstring_initialise(&specs, "");
  
  if ( !strcmp(key_word->c_str,"total") ) f->total = true;
  else if ( !strcmp(key_word->c_str,"partial") ) f->total = false;
  else if ( !strcmp(key_word->c_str,"injective") ||
	    !strcmp(key_word->c_str,"all_different") ) f->injective = true;
  else if ( !strcmp(key_word->c_str,"left_injective") || 
	    !strcmp(key_word->c_str,"left_all_different") || 
	    !strcmp(key_word->c_str,"column_injective") || 
	    !strcmp(key_word->c_str,"column_all_different") )
    f->leftin = (f->adicity == 2);
  else if ( !strcmp(key_word->c_str,"right_injective") || 
	    !strcmp(key_word->c_str,"right_all_different") || 
	    !strcmp(key_word->c_str,"row_injective") || 
	    !strcmp(key_word->c_str,"row_all_different") )
    f->rightin = (f->adicity == 2);
  else if ( !strcmp(key_word->c_str,"surjective") ) f->surjective = true;
  else if ( !strcmp(key_word->c_str,"left_surjective") || 
	    !strcmp(key_word->c_str,"column_surjective") )
    f->leftsur = (f->adicity == 2);
  else if ( !strcmp(key_word->c_str,"right_surjective") || 
	    !strcmp(key_word->c_str,"row_surjective") )
    f->rightsur = (f->adicity == 2);
  else if ( !strcmp(key_word->c_str,"bijective") ||
	    !strcmp(key_word->c_str,"one_one") )
    f->surjective = f->injective = true;
  else if ( !strcmp(key_word->c_str,"left_bijective") || 
	    !strcmp(key_word->c_str,"left_one_one") || 
	    !strcmp(key_word->c_str,"column_bijective") || 
	    !strcmp(key_word->c_str,"column_one_one") )
    f->leftsur = f->leftin = (f->adicity == 2);
  else if ( !strcmp(key_word->c_str,"right_bijective") || 
	    !strcmp(key_word->c_str,"right_one_one") || 
	    !strcmp(key_word->c_str,"row_bijective") || 
	    !strcmp(key_word->c_str,"row_one_one") )
    f->rightsur = f->rightin = (f->adicity == 2);
  else if ( !strcmp(key_word->c_str,"commutative") )
    f->comm = ((f->adicity == 2) && (f->arg_type[0] == f->arg_type[1]));
  /*
  else if ( !strcmp(key_word->c_str,"cut") || key_word[0] == '!' ) f->cut_it = true;
  else if ( !strcmp(key_word->c_str,"no_cut") )
    f->cut_it = false;
  */
  else if ( !strcmp(key_word->c_str,"priority") ) f->nop = false;
  else if ( !strcmp(key_word->c_str,"no_priority") ) f->nop = true;
  else if ( !strcmp(key_word->c_str,"print") ) {
    get_word(&specs);
    if ( specs.c_str[0] == ':' ) get_word(&specs);
    if ( !strcmp(specs.c_str,"none") ) {
      f->pmode = null;
      /*
      f->cut_it = true;
      */
    }
    else if ( !strcmp(specs.c_str,"brief") ) f->pmode = brief;
    else if ( !strcmp(specs.c_str,"full") ) f->pmode = full;
    else inerror(specs.c_str,"found where print mode expected");
    return;
  }
  else if ( !strcmp(key_word->c_str,"scope") ) {
    f->scope = get_int();
    if ( f->scope > 10 ) f->scope = 10;
    return;
  }
  else inerror(key_word->c_str,"found where function specification expected");

  dynstring_free(&specs);

} /* func_spec() */


/*------------------------------------------------------------------------*/
boolean add_clause(dynstring *clause_string)

     /*Called by model() and find_theories().
      *Parse clause_string as tho' it was next on the input line. 
      *Returns TRUE iff clause is non-trivial*/

{
  boolean result = false;
  dynstring lead_word;

  dynstring_initialise(&lead_word, "");

  dynstring_strcpy(&job_line,clause_string);
  input_point = 0;

  get_word(&lead_word);
  if (get_cl(&lead_word,0)) {
    result = true;
  } else {
    inwarning("\nPassed clause is trivially satisfiable.","");
  }

  dynstring_free(&lead_word);
  
  return result;

}/* add_clause() */


void add_declaration(dynstring *declaration)
  /*----------------------------------------------------------------------*\
    Called by SCOTT fcn declare_otter_symbols().
    Parse "declaration" as tho' it was next on the input line. 
  \*----------------------------------------------------------------------*/
{
  dynstring_strcpy(&job_line,declaration);
  input_point = 0;
  source = NULL; // not reading from a file

  switch( get_key_word() ) {
  case nul:
    return;
  case srt:
    get_item(get_sort,0,0);
    break;
  case fcn:
    get_item(get_func,0,0);
    break;
  case fcn0:
    get_item(get_const,0,0);
    break;
  case cls:
    get_item(get_input_cl,0,0);
    break;
  case prm:
    get_item(get_set,0,0);
    break;
  case inc:
    get_incl_file();
    break;
  }

}/*--add_declaration()----------------------------------------------------*/


/*------------------------------------------------------------------------*/
void get_input_cl( kind k, functor f, dynstring *first_word, int softweight)

     /*Called by get_item(). 
      *Get the clause off the input line and, provided it's non-trivial,
      *add it to the current theory.*/
     
{
  if (first_word->c_str[0] != '.' &&
      !strchr(job_line.c_str+input_point,'.') && 
      !strchr(job_line.c_str+input_point,'}'))
    input_to_terminator(".}");

  if ( get_cl(first_word,softweight) ) {
    if (normal_input) add_to_theory();
  }
  else 
    inwarning("Input clause is trivially satisfiable.","");
}


/*------------------------------------------------------------------------*/
boolean get_cl( dynstring *lead_word, int softweight )

     /*Called by get_input_cl() and add_clause().
      *Parse the next clause on the input line into latest_clause.
      *Returns TRUE iff the clause is non-trivial.*/ 

{
  wff_stack *ins = 0;
  wff tempwff = 0 ;
  boolean parsed_premises;

  latest_clause.softweight = softweight;
  begin_parse();
  if (strcmp(lead_word->c_str,"->")  && strstr(job_line.c_str+input_point,"->"))
    parsed_premises = false;
  else parsed_premises = true;

  do {
    if (strcmp(lead_word->c_str,"->")) {
      tempwff = parse( lead_word );
      if (parsed_premises)
	ins = &(latest_clause.conc);
      else
	ins = &(latest_clause.prem);
    }

    if (!strcmp(lead_word->c_str,"->"))
      parsed_premises = true;
    else if (!strcmp(lead_word->c_str,",") && parsed_premises)
      inwarning("",
		"Consequents should be separated by semicolon, not comma");
    else if (!strcmp(lead_word->c_str,";") && !parsed_premises)
      inwarning("",
		"Antecedents should be separated by comma, not semicolon");
    else if (!strchr(",;.",lead_word->c_str[0]))
      inerror(lead_word->c_str,"found in clause where punctuation expected");

    if ( tempwff ) {
      while ( *ins ) ins = &((*ins)->nxt);
      *ins = (wff_stack) malloc(sizeof(WFF_STACK));
      (*ins)->w = tempwff;
      (*ins)->nxt = 0;
    }

    if (strcmp(lead_word->c_str,"."))
      get_word(lead_word); /*get next predicate name*/
  }
  while (strcmp(lead_word->c_str,"."));   /* all clauses terminated with "." */

  typecheck();
  end_parse();
  get_weight();

  return true;

} /* get_cl() */



/*------------------------------------------------------------------------*/
boolean isint( unsigned char *s )
     
     /*Called by one line in get_set(). This could be dodgy. */
     
{
  do if ( !isdigit(*s) ) return false;
  while ( *(++s) );
  return true;
}/* isint() */


/*------------------------------------------------------------------------*/
void get_set( kind k, functor f, dynstring *set, int x )
     
     /*Called by get_item().
      *Extracts setting information.*/
     
{
  if ( !strcmp(set->c_str,"solutions") ) setting.max_models = get_int();
  else if ( !strcmp(set->c_str,"backtracks") ) 
    setting.max_backtracks = get_int();
  else if ( !strcmp(set->c_str,"insertions") ) 
    setting.max_invalues = get_int();
  else if ( !strcmp(set->c_str,"stack") ) {
    get_word(set);
    if ( set->c_str[0] == ':' ) get_word(set);
    if ( !strcmp(set->c_str,"itsy") ) setting.batches = S_ITSY;
    else if ( !strcmp(set->c_str,"bitsy") ) setting.batches = S_BITSY;
    else if ( !strcmp(set->c_str,"tiny") ) setting.batches = S_TINY;
    else if ( !strcmp(set->c_str,"small") ) setting.batches = S_SMALL;
    else if ( !strcmp(set->c_str,"medium") ) setting.batches = S_MEDIUM;
    else if ( !strcmp(set->c_str,"large") ) setting.batches = S_LARGE;
    else if ( !strcmp(set->c_str,"LARGE") ) setting.batches = S_VERYLARGE;
    else if ( !strcmp(set->c_str,"huge") ) setting.batches = S_HUGE;
    else if ( !strcmp(set->c_str,"maximal") ) setting.batches = S_MAXIMAL;
    else if ( isint(set->c_str) ) sscanf(set->c_str,"%d",&(setting.batches));
    else inerror(set->c_str,"found where stack size expected");
    return;
  }
  else if ( !strcmp(set->c_str,"stack_margin") ) 
    setting.extra_batches = get_int();
  else if ( !strcmp(set->c_str,"verbosity") ) get_item( verb_spec, 0, 0 );
  else if ( !strcmp(set->c_str,"ref_length") ) setting.maxref = get_int();
  else if ( !strcmp(set->c_str,"pre_test") ) setting.pretest_limit = get_int();
  else if ( !strcmp(set->c_str,"test_ref_length") ) 
    setting.test_maxref = get_int();
  else if ( !strcmp(set->c_str,"time_limit") ) setting.maxtime = get_int();

  else if ( !strcmp(set->c_str,"max_bound") ) setting.target_bound = get_int();

  else if ( !strcmp(set->c_str,"primary") ) setting.Nosecs = true;
  else if ( !strcmp(set->c_str,"secondary") ) setting.Nosecs = false;
  
  else if ( !strcmp(set->c_str,"noisy") ) setting.noisy = true;
  else if ( !strcmp(set->c_str,"quiet") ) setting.noisy = false;
  
  else if ( !strcmp(set->c_str,"one_cardinality") ) setting.sortcut = true;
  else if ( !strcmp(set->c_str,"all_cardinalities") ) setting.sortcut = false;
  
  else if ( !strcmp(set->c_str,"time_stamp") ) setting.timestamps = true;
  else if ( !strcmp(set->c_str,"no_time_stamp") ) setting.timestamps = false;
  
  else if ( !strcmp(set->c_str,"report_breaks") ) setting.report_breaks = true;
  else if ( !strcmp(set->c_str,"silent_breaks") ) setting.report_breaks = false;
  
  else if ( !strcmp(set->c_str,"tops_off") ) setting.topsoff = true;
  else if ( !strcmp(set->c_str,"tops_on") ) setting.topsoff = false;
  
  else if ( !strcmp(set->c_str,"reverse_test") ) setting.invert_change = true;
  else if ( !strcmp(set->c_str,"obverse_test") ) setting.invert_change = false;
  
  else if ( !strcmp(set->c_str,"test_by_vars") ) setting.vars_over_wt = true;
  else if ( !strcmp(set->c_str,"test_by_weight") ) setting.vars_over_wt = false;

  else if ( !strcmp(set->c_str,"random") ) {
    setting.random = true;
    srand48((long) get_int());
  }
  else if ( !strcmp(set->c_str,"non_random") )
    setting.random = false;
  
  else if ( !strcmp(set->c_str,"pretty") ) setting.pm.pretty = true;
  else if ( !strcmp(set->c_str,"ugly") ) {
    setting.pm.pretty = false;
    setting.pm.verbosity = pugly;
  }
  else inerror("Unrecognosed setting option",set->c_str);

} /* get_set() */



/*------------------------------------------------------------------------*/
void verb_spec( kind k, functor f, dynstring *s, int x )

     /*Called by get_set() only.*/
{
  unsigned vb;
  
  vb = setting.pm.verbosity;
  if ( !strcmp(s->c_str,"models") ) vb = verb_spec_spec(vb,pmodels);
  else if ( !strcmp(s->c_str,"job") ) vb = verb_spec_spec(vb,pjob);
  else if ( !strcmp(s->c_str,"stats") ) vb = verb_spec_spec(vb,pstats);
  else {
    inerror("Unrecognised verbosity option",s->c_str);
    verb_spec_spec(0,0);
  }
  setting.pm.verbosity = vb;

} /*verb_spec()*/



/*------------------------------------------------------------------------*/
unsigned verb_spec_spec( unsigned vb, unsigned mask )

/*Called by verb_spec() only. */

{
  dynstring s;
  unsigned result = vb;

  dynstring_initialise(&s, "");

  get_word(&s);
  if ( s.c_str[0] == ':' ) get_word(&s);
  vb &= ~mask;
  if ( !strcmp(s.c_str,"none") ) result = vb;
  else if ( !strcmp(s.c_str,"brief") ) result = vb | (mask & (mask >> 1));
  else if ( !strcmp(s.c_str,"full") ) result = vb | (mask & (mask << 1));
  else inerror(s.c_str, "found where level (none, brief or full) expected");

  dynstring_free(&s);
  
  return vb;

} /* verb_spec_spec() */



/*------------------------------------------------------------------------*/
boolean got_partial()

/*Called by FINDER fcn main.c. -- not relvant to SCOTT*/

{
  boolean result = true;
  dynstring s;

  dynstring_initialise(&s, "");

  source = stdin;
  if ( !piping ) {
    result = false;
  } else {
    get_word(&s);
    if ( !strcmp(s.c_str,"end") ) {
      result = false;
    } else {
      if ( strcmp(s.c_str,"model") ) {
        fprintf(WARNING_STREAM,"\n\"%s\" found where \"model\" expected\n",s.c_str);
        result = false;
      } else {
        while ( got_function() );
      }
    }
  }

  return result;
}/* got_partial() */



/*------------------------------------------------------------------------*/
boolean got_function()

/* Called by got_partial(). -- not relvant to SCOTT*/

{
  dynstring s;
  int theArguments[ARGMAX];
  int i;
  functor f = NULL;
  kind k = NULL;
  int count;

  dynstring_initialise(&s, "");

  get_word(&s);
  if ( s.c_str[0] == '.' ) {
    dynstring_free(&s);
    return false;
  }
  if ( !strcmp(s.c_str,"size") ) {
    get_word(&s);
    if ( (i = lex_pos(s.c_str)) != -1) {
      if (lexicon[i].iskind )
	k = sort+lexicon[i].offset;
      else skipout("Incorrect sort name transfer",1);
    }
    k->siz = get_int();
    dynstring_free(&s);
    return true;
  }
  if ( (i = lex_pos(s.c_str)) != -1) {
    if (!lexicon[i].iskind ) 
      f = fn+lexicon[i].offset;
    else skipout("Incorrect function name transfer",1);
  }

  if (f->adicity) count = 2; else count = 1;
  for (i=0; i<f->adicity; i++) count *= (f->arg_type[i]->siz);
  if (f->table) free(f->table);
  f->table = (int*) malloc(sizeof(int) * count);

  if (f->adicity)
    f->fats[f->adicity-1] = 1;
  for (i=f->adicity-2; i>= 0; i--)
    f->fats[i] = f->fats[i+1] * (f->arg_type[i+1]->siz);

  if (!f->adicity) *(f->table) = get_int();
  else if (f->adicity == 1) {
    for ( i = 0; i < f->arg_type[0]->siz; i++ )
      *(f->table + i + 1) = get_int();
  }
  else multiread(f,0,theArguments);

  dynstring_free(&s);
  return true;

} /* got_function() */


/*---------------------------------------------------------------*/

/*
  To read back the values of a function ugly-printed by multiprint
*/

void multiread(functor f, int x, int theArguments[])
{
  for (theArguments[x] = 0;
       theArguments[x] < f->arg_type[x]->siz;
       theArguments[x]++) {
    if (x == f->adicity-1)
      *(f->table + getoffset(f,theArguments)) = get_int();
    else multiread(f,x+1,theArguments);
  }
}


/*------------------------------------------------------------------------*/
void inerror(char *s1, char *s2)

     /*Called by numerous functions in files:
      *FINDER.c, initial.c, input.c, parser.c, and prepare.c.
      
      *Print input error and exit with status set to current line number*/
{
  fprintf(WARNING_STREAM,
	  "\nInput error on line %d: %s\n\n%s %s\n\n",
	  current_line,
	  job_line.c_str,
	  s1,
	  s2);
  fflush(WARNING_STREAM);
  exit(current_line);
}


/*------------------------------------------------------------------------*/
void inwarning(char *s1, char *s2)

     /*Called by verb_spec(), verb_spec_spec(), get_set(), func_spec(), 
      *set_up_type(), sort_spec(), and get_key_word() (all in this file).
      
      *Print input warning but continue jobs.*/
{
  fprintf(WARNING_STREAM,
	  "\nWarning: possible error on line %d: %s\n%s %s\n\n\n", 
	  current_line, job_line.c_str,s1,s2);
  fflush(WARNING_STREAM);
}

/*================================ END ===================================*/
