/*
*			parser.c
*/


#include "FINDER.h"

static WFF newcl[NEWCL_MAX]; /*a parse wff stack for the new clause */
static boolean type_added;

static PSW parse_wffs[NEWCL_MAX]; /*parse() fcn wff stack*/
static int psw_next;    /* index of next (sub)formula to be parsed */
static int previous_end;

/*------------------------------------------------------------------------*/
void free_wff_list( wff_stack ptr )

/*Called by begin_parse().*/

{
    if ( !ptr ) return;
    free_wff_list( ptr->nxt );
    free( ptr );

} /* free_wff_list() */



/*------------------------------------------------------------------------*/
void begin_parse()

     /*Called by get_cl().
      *Initializes the temporary wff stack and the various lists of the 
      *current clause before parsing starts.*/

{
  int i;

  for ( i = 0; i < NEWCL_MAX; i++ ) {
    newcl[i].sym = 0;
    newcl[i].val = -1; /* dummy value */
  }
  free_wff_list(latest_clause.prem);
  latest_clause.prem = 0;
  free_wff_list(latest_clause.conc);
  latest_clause.conc = 0;
  free_wff_list(latest_clause.varlist);
  latest_clause.varlist = 0;
  free_wff_list(latest_clause.testlist);
  latest_clause.testlist = 0;
  psw_next = 0;

  /* get last name on the functor stack */
  for (i=0; fn[i].fname; i++ );
  previous_end = i;
} /* begin_parse() */


/*------------------------------------------------------------------------*/
wff parse( dynstring *symbol )
     
     /*Called by sparse() and get_cl().
      *Parsing routine which allow infix binary function symbols.

      *A general wff (gwf) consists of one or more simple wffs (swf) 
      *separated by binary functors.  
      *An swf is either a functor followed by the appropriate number 
      *of swf to form its arguments or a gwf in parentheses.

      *Hence we need two interacting parse functions, one for 
      *gwf (parse) and one for swf (sparse).  
      *In each case the symbol[] parameter holds the first symbol of 
      *the sequence to be parsed. After the parse it holds the symbol 
      *which follows the formula. 

      *Returns the fully parsed wff*/

{
  psw pp, qq;
  functor mf;
  int i;
  pp = &(parse_wffs[psw_next]);

  if (++psw_next >= NEWCL_MAX) inerror("Too many formulae parsed","");

  pp->leftsub = pp->rightsub = 0;
  pp->main = 0;
  pp->actual = sparse(symbol); /* get first swf */

  get_word(symbol);

    /* if next word is not a terminator,
       then it must be a binary operator */

  while ( !strchr(".,;)",symbol->c_str[0]) && strncmp("->",symbol->c_str,2) ) { 
    i = lex_pos(symbol->c_str);
    if ( i < 0 )
      inerror(symbol->c_str,"found where a function symbol was expected");
    mf = &(fn[lexicon[i].offset]);
    if ( mf->adicity != 2 )
      inerror(symbol->c_str,"found where a binary operator was expected");

    qq = (psw) malloc(sizeof(PSW));
    qq->leftsub = qq->rightsub = 0;
    qq->main = 0;

    get_word(symbol);
    qq->actual = sparse(symbol); /* get the second argument of the 
				    binary operator */

    pw_merge( &pp, qq, mf );     /* merge these structures into 
				    something more resonable */
    get_word(symbol);
  }
  return pf_transcribe(pp);

} /* parse() */




/*-----------------------------------------------------------------------*/
void pw_merge( psw *pp, psw qq, functor mf)

     /*Called by parse().
       
      *Recursive routine for converting infix to prefix forms.
      *Puts the next functor and subformula into the tree of 
      *partially parsed subformulas. 
      *Where the tree so far is, say
      *                     pp
      *	                   /  \
      *	                  L    R
      *		              / \
      *		            Rl   Rr
      *
      *the result of merging (...qq mf ) with it will be either
      *
      *		pp        or else       mf
      *	       /  \                    /  \
      *	      L    R                  pp   qq
      *		  / \                / \
      *		 Rl  mf             L   R
      *		    / \                / \
      *		   Rr  qq            Rl   Rr
      *
      *Note we pass pp by address since memory has already been 
      *allocated and this is what we want to return.*/

{
  psw rr;

    /*right diagram*/
  if ( !(*pp)->main || mf->scope >= (*pp)->main->scope ) {
    rr = (psw) malloc(sizeof(PSW));
    rr->leftsub = *pp;
    rr->rightsub = qq;
    rr->main = mf;
    rr->actual = 0;
    *pp = rr;
    return;
  }

    /*else left diagram*/
  pw_merge( &((*pp)->rightsub), qq, mf );
  return;

} /* pw_merge()*/



/*-------------------------------------------------------------------------*/
wff pf_transcribe( psw pp )

     /*Called by parse() and itself.
      *Pf_transcribe is part of getting the semi-parsed inside of a pair of
      *parentheses into fully transcribed form. It returns the fully parsed 
      *subformula whose partially parsed version is rooted at pp.*/

{
  wff subforms[2];

  if ( pp->actual ) return pp->actual;
  subforms[0] = pf_transcribe(pp->leftsub);
  subforms[1] = pf_transcribe(pp->rightsub);
  return locate( pp->main, subforms );

} /* pf_transcribe() */



/*-------------------------------------------------------------------------*/
wff sparse( dynstring *symbol )

     /*Called by parse() and itself.
      *Parse the cases other than that of a parenthesised subformula.
      *Returns the fully parsed simple wff. */

{    
  wff parsed_subform, args[ARGMAX];
  functor f = 0;
  int i, pos;

  /*if its in parentheses call parse() to deal with it*/
  if ( symbol->c_str[0] == '(' ) {
    get_word(symbol);
    parsed_subform = parse(symbol);
    if ( symbol->c_str[0] != ')' )
      inerror(symbol->c_str,"found where a right parenthesis was expected");
    return parsed_subform;
  }

  /*otherwise it's either a function, a variable or a number.*/
  for (i=0; i<ARGMAX; i++) args[i] = 0;
  pos = lex_pos(symbol->c_str);
  if ( pos >= 0 ) f = &fn[lexicon[pos].offset];   /* must be a function  */
  else {
    if ( isalpha(symbol->c_str[0]) )                     /* ok, its a varaiable */
      return locate( newvar(symbol->c_str,0), args );

    else if ( isdigit(symbol->c_str[0]) ) {              /* ok, its a number    */
      for ( i = 0; symbol->c_str[i]; i++ ) if ( !isdigit(symbol->c_str[i]) )
	inerror(symbol->c_str,"starts numerically but isn't a number");

      /* make a new constant for that number */
      f = newfunctor( symbol->c_str, Const, 0 );
      f->val_type = Integer;
      *(f->table) = symbol->c_str[0] - '0';
      if ( symbol->c_str[1] ) {
	*(f->table) *= 10; /* all numbers should be less than 100 */
	*(f->table) += (symbol->c_str[1] - '0');
      }
      return locate( f, args );
    }
    else inerror(symbol->c_str,"found where a function symbol was expected");
  }

  /*Now if its a  function ...*/
  if ( f->adicity > 0 ) {
    get_word(symbol);
    if ( symbol->c_str[0] != '(' )
      inerror(symbol->c_str,"found where a left parenthesis was expected");
    for (i=0; i<f->adicity; i++) {
      get_word(symbol);
      args[i] = parse(symbol);
      if (i == f->adicity-1) {
	if ( symbol->c_str[0] != ')' )
	  inerror(symbol->c_str,"found where a right parenthesis was expected");
      }
      else {
	if ( symbol->c_str[0] != ',' )
	  inerror(symbol->c_str,"found where a comma was expected");
      }
    }
  }

  return locate( f, args );

} /* sparse() */


/*-------------------------------------------------------------------------*/
wff locate( functor main_func, wff subform[ARGMAX] )

     /*Called by sparse() and pf_transcribe().
      *Returns a pointer to the (sub)formula with main functor main_func
      *and argument list subform[].
      *This is especialy important for finding pre-existing variables.

      *Note that the formula is located in the "parse wff stack" newcl[], 
      *not in form[]. This is where newcl[] gest built.
     
      *This is sometime called with a null 2nd argument. Would prefer to
      *do something more elegant for this case.*/

{
  int i, j;
  boolean subs_match;
  
  /* search for formula in the stack */
  for ( i = 0; newcl[i].sym; i++ ){
    if ( i >= NEWCL_MAX ) inerror("Clause is too long","");
    if ( newcl[i].sym->fname == main_func->fname ) {

      if (main_func->category == Var || main_func->category == Const ) 
	return &(newcl[i]); /*found it!!*/

      /*NOTE constants and variables have no args so the following
	code would be dangerous to execute */

      /*compare subformulae*/
      subs_match = true;
      for (j=0; j<main_func->adicity; j++)
	if (subform[j] != newcl[i].sb[j])
	  subs_match = false;
      if (subs_match) return &(newcl[i]); /* found it!! */
    }
  }

  /*Otherwise we have to add it to the stack -- the i-th element*/

  newcl[i].sym = main_func;
  newcl[i].matrix = main_func->table;
  newcl[i].wcc = -1;

  if ( main_func->category == Const ) newcl[i].val = *(main_func->table);

  /*what about variables!!!*/

  /* set any subformulae */
  for (j=0; j<ARGMAX; j++) {
    newcl[i].sb[j] = (j<main_func->adicity? subform[j]: 0);
    newcl[i].sbval[j] = (j<main_func->adicity? &(subform[j]->val): &zero);
  }

  return &(newcl[i]);

} /* locate() */


/*-------------------------------------------------------------------------*/
void typecheck()
     
     /*Called by get_cl().
      *Checking for agreement of types, especially for variables and for
      *typically ambiguous symbols like + and =, is a messy business.*/
     
{
  wff_stack wf;

  /* add any missing types */
  do {
    type_added = false;
    for ( wf = latest_clause.prem;wf; wf = wf->nxt ) type_needed(wf->w);
    for ( wf = latest_clause.conc;wf; wf = wf->nxt ) type_needed(wf->w);
  }
  while ( type_added ); /* dont like this logic!! */

  /*check for completness */
  for ( wf = latest_clause.prem;wf; wf = wf->nxt ) type_complete( wf->w );
  for ( wf = latest_clause.conc;wf; wf = wf->nxt ) type_complete( wf->w );

  /*check for consistency */
  for ( wf = latest_clause.prem;wf; wf = wf->nxt ) type_consistent( wf->w );
  for ( wf = latest_clause.conc;wf; wf = wf->nxt ) type_consistent( wf->w );

} /* typecheck() */



/*-------------------------------------------------------------------------*/
void type_needed( wff w )

     /*Called by typecheck() and itself.
      *Not every symbol is born with a type. This is the function for 
      *adding types to symbols without. There are various cases, for
      *sort-relative symbols like + and EST, and then for variables.
      *Types have to be deduced from the context.*/

{
  int i;

  if ( !w ) return;
  if ( !strcmp(w->sym->fname,"+") ) AT_plus( w );
  if ( !strcmp(w->sym->fname,"-") ) AT_plus( w );
  if ( !strcmp(w->sym->fname,"EST") ) AT_up( w, "EST_" );
  if ( !strcmp(w->sym->fname,"LEN") ) AT_up( w, "LEN_" );

  /*if argument types are known, check each argument
    --- what about outermost symbol? 
    --- does equality go thru' here? No, "=" has no argument types.*/
  for (i=0; i<w->sym->adicity; i++)
    if ( w->sym->arg_type[i] && !w->sb[i]->sym->val_type )
      add_type(w->sb[i],w->sym->arg_type[i]);

  /* types of predefined binary ops must match !!*/
  if (w->sym->adicity == 2 && 
      (strchr("=<>",w->sym->fname[0]) || !strcmp("DIF",w->sym->fname))) {
    if ( w->sb[0]->sym->val_type && !w->sb[1]->sym->val_type)
      add_type( w->sb[1], w->sb[0]->sym->val_type );
    if ( w->sb[1]->sym->val_type && !w->sb[0]->sym->val_type)
      add_type( w->sb[0], w->sb[1]->sym->val_type );
  }

  /*check any subformulae */
  for (i=0; i<w->sym->adicity; i++)
    type_needed(w->sb[i]);

}/* type_needed() */



/*-------------------------------------------------------------------------*/
void add_type( wff w, kind k )
     
     /*Called by type_needed() and AT_plus().
      *Gives w the  type k. 
      *Variables get their types this way -- same as kind. */

{
  dynstring string;
  dynstring_initialise(&string, "");
  
  if ( !strcmp(w->sym->fname,"EST") ) return;
  if ( !strcmp(w->sym->fname,"LEN") ) return;
  if ( w->sym->fname[0] == '+' ) {

    dynstring_sprintf(&string, "+_%s", k->sname);
    w->sym = fn+lexicon[lex_pos(string.c_str)].offset;
    w->matrix = w->sym->table;
  }
  else if ( w->sym->fname[0] == '-' ) {
    dynstring_sprintf(&string, "-_%s", k->sname);
    w->sym = fn+lexicon[lex_pos(string.c_str)].offset;
    w->matrix = w->sym->table;
  }
  else if ( w->sym->category == Var )
    w->sym = newvar(w->sym->fname,k); /* get variable types */
  else inerror("Missing type for",w->sym->fname);
  type_added = true;

  dynstring_free(&string);
} /* add_type() */



/*-------------------------------------------------------------------------*/
void AT_plus( wff w )

     /*Called by type_needed().
      *Ensure that in formula like (A + B) the + is of the same type 
      *as the A.*/
     
{
  if ( w->sb[0]->sym->val_type )
    add_type( w, w->sb[0]->sym->val_type );

} /* AT_plus() */



/*-------------------------------------------------------------------------*/
void AT_up( wff w, char *s )

     /*Called by type_needed().
      *Ensure that in formula LEN(A) or EST(A), the type of the predicate 
      *is determined by that of its argument.*/

{
  char string[LEXLENGTH];

  if ( w->sb[0]->sym->val_type ) {
    strcpy( string, s );
    strcat( string, w->sb[0]->sym->val_type->sname );
    w->sym = fn+lexicon[lex_pos(string)].offset;
    w->matrix = w->sym->table;
    type_added = true;
  }
} /* AT_up() */



/*-------------------------------------------------------------------------*/
functor newvar( char s[], kind k )
     
     /*Called by sparse() and add_type(). 
      *Returns a "fresh" variable.*/

{
  functor f;
  int i,j;

/*  for (j=previous_end; (j < FN_MAX) && fn[j].fname; j++ ){ */
  for (j=0; (j < FN_MAX) && fn[j].fname; j++ ){
    if ( !strcmp(s,fn[j].fname) &&  fn[j].val_type == k ) {
      return &(fn[j]);
    }
  }

  if (j >= FN_MAX) inerror("FN_MAX exceeded: too many functors!","");
  f = &(fn[j]); /* set to next open position */

  /*Find pre-existing variable name or the first open position*/
  for ( i = 0; (i < VARMAX) && varlex[i][0]; i++ ) 
    if ( !strcmp(s,varlex[i]) ) break;
  
  if (i >= VARMAX) inerror("VARMAX exceeded: too many variables!","");
  if ( !varlex[i][0] ) strcpy(varlex[i],s);

  /*copy it to the list of functors */
  f->fname = varlex[i];
  f->category = Var;
  f->scope = 0;
  f->adicity = 0;
  f->table = 0;
  f->fcc = 0;
  for (j=0; j<MAXISET_ARRAY; j++) fn->old_fcc[j] = 0;
  for (j=0; j<ARGMAX; j++) f->fats[j] = 0;
  f->total = true;
  f->injective = false;
  f->cut_it = false;
  for (j=0; j<ARGMAX; j++)
    f->arg_type[j] = 0;
  f->val_type = k;         /*same as the kind passed*/
  f->pmode = null;
  return f;
  
} /* newvar() */



/*-------------------------------------------------------------------------*/
void type_complete( wff w )

     /*Called by typecheck() and itself.
      *Checks that every symbol in the current formula has a
      *value type. Some (like `=') have no argument types.*/

{
  int i;
  
  if ( !w ) return;
  if ( !w->sym->val_type ) inerror("Type missing for",w->sym->fname);
  for (i=0; i<w->sym->adicity; i++) type_complete( w->sb[i] );

} /* type_complete( ) */


/*-------------------------------------------------------------------------*/
void type_consistent( wff w )

     /*Called by typecheck() and itself.
      *After everything has a type, there may be errors. This function 
      *goes recursively through a formula looking for type mismatches.*/

{
  int i;

  if (strchr("=<>",w->sym->fname[0]) || !strcmp("DIF",w->sym->fname)) {
    if ( w->sb[0]->sym->val_type != w->sb[1]->sym->val_type )
      inerror("Type mismatch: arguments of",w->sym->fname);
  }
  else
    for (i=0; i<w->sym->adicity; i++)
      if ( w->sym->arg_type[i] != w->sb[i]->sym->val_type )
	inerror("Type mismatch with argument of",w->sym->fname);
  
  for (i=0; i<w->sym->adicity; i++)
    type_consistent( w->sb[i] );

} /* type_consistent() */



/*-------------------------------------------------------------------------*/
boolean trivial_cl()

     /*Called by get_cl().
      *Return TRUE iff the functors in the new clause are all
      *trivially satisfied.*/

{
  int i;
  
  for (i=0 ; newcl[i].sym ;i++) {
    if ( newcl[i].sym->category == Var ) continue;
    if ( strchr("+=<>",newcl[i].sym->fname[0]) ) continue;
    if ( !strcmp("equal",newcl[i].sym->fname)) continue;
    return false;
  }
  return true;

}/* trivial_cl() */


/*-------------------------------------------------------------------------*/
void end_parse()

     /*Called by get_cl().
      *After the current clause has been parsed, this is called to set up
      *the various lists (varlist and testlist) and to set the pointers 
      *to subformula values correctly.*/

{
  wff_stack wptestlist, wpvarlist;	/* Ends of testlist and varlist */
  int i;

  wptestlist = wpvarlist = 0;
  /* for each symbol on the parse stack */
  for ( i = 0; newcl[i].sym; i++ ) {
    if ( newcl[i].sym->category == Var ) {     /* create a variable list  */
      if ( !wpvarlist ) {                      /* at first variable       */
	wpvarlist = (wff_stack) malloc(sizeof(WFF_STACK));
	latest_clause.varlist = wpvarlist;
      }
      else {                                   /* at subsequent variables */
	wpvarlist->nxt = (wff_stack) malloc(sizeof(WFF_STACK));
	wpvarlist = wpvarlist->nxt;
      }
      wpvarlist->w = newcl+i;                  /* reset end of varlist */
      wpvarlist->nxt = 0;
    }
    else if ( newcl[i].sym->category != Const ) { /* create a test list  */
      if ( !wptestlist ){                       /* at first subformula   */
	wptestlist = (wff_stack) malloc(sizeof(WFF_STACK));
	latest_clause.testlist = wptestlist;
      }
      else {                                    /* at subsequent subforms  */
	wptestlist->nxt = (wff_stack) malloc(sizeof(WFF_STACK));
	wptestlist = wptestlist->nxt;
      }
      wptestlist->w = newcl+i;               /* reset end of testlist */
      wptestlist->nxt = 0;
    }
    newcl[i].matrix = newcl[i].sym->table;
  }

  /* set the value pointers in each testlist, why not varlists too?!*/
  for ( wptestlist = latest_clause.prem; wptestlist; 
	wptestlist = wptestlist->nxt )
    set_valptrs( wptestlist->w );

  for ( wptestlist = latest_clause.conc; wptestlist; 
	wptestlist = wptestlist->nxt )
    set_valptrs( wptestlist->w );

  /* do something obscure with positive last clauses */
  if ( latest_clause.conc &&
       !(latest_clause.conc->nxt) &&
       *(latest_clause.conc->w->sym->fname) == '=' &&
       is_open(latest_clause.conc->w->sb[0]->sym) &&
       !occurs_inside(latest_clause.conc->w->sb[0],
		      latest_clause.conc->w->sb[1]) &&
       !occurs_in_list(latest_clause.conc->w->sb[0],
		       latest_clause.prem) )
    positive_last();

} /* end_parse() */



/*-------------------------------------------------------------------------*/
void set_valptrs( wff wf )

     /*Called by end_parse().
      *Sets the subformula value pointers.*/

{
  int i;

  if ( !wf ) return;
  if ( wf->val >= 0 ) return;

  /* foreach subformula */
  for (i=0; i<wf->sym->adicity; i++)
    set_valptrs( wf->sb[i] );

  for (i=0; i<ARGMAX; i++)
    wf->sbval[i] = (i<wf->sym->adicity? &(wf->sb[i]->val): &zero);

} /* set_valptrs() */



/*-------------------------------------------------------------------------*/
void positive_last()

     /*Called by end_parse().
      *Certain Horn clauses can be treated specially with one entry in
      *the list of refutation points counted as "positive". The positive
      *point must be moved to the end of the testlist so that any values
      *needed to evaluate it (which need not be in its subformulae) will
      *have been calculated before it is encountered.
      *
      *This is rather ugly code, but inefficiency here is of no great
      *significance since the testlist is always fairly short.
      */

{
  wff_stack lc, cls, wp;  /* latest clause conclusion, its left sub, 
			     a ptr */
  
  for ( lc = latest_clause.testlist; lc->nxt; lc = lc->nxt );
  for (cls = latest_clause.testlist;
       cls && cls->w != lc->w->sb[0];
       cls = cls->nxt);
  if ( !cls || cls->nxt == lc ) return;
  if ( cls == latest_clause.testlist )
    latest_clause.testlist = cls->nxt;
  else {
    for ( wp = latest_clause.testlist; wp->nxt != cls; wp = wp->nxt );
    wp->nxt = cls->nxt;
  }
  cls->nxt = lc;
  for ( wp = latest_clause.testlist; wp->nxt != lc; wp = wp->nxt );
  wp->nxt = cls;
  
} /* positive_last() */



/*-------------------------------------------------------------------------*/
void add_to_theory()
     
     /*Called by model() and get_cl().
      *Move current clause from  the temporary stack newcl[] to the 
      *permanent one form[], and add it to current_theory.*/

{
  wff_stack wptr;
  clause target;      /* The new clause as part of the theory */

  /* copy clause from newcl[] to form[] */
  for ( wptr = latest_clause.prem; wptr; wptr = wptr->nxt )
    wptr->w = copy_to_form(wptr->w);
  for ( wptr = latest_clause.conc; wptr; wptr = wptr->nxt )
    wptr->w = copy_to_form(wptr->w);
  for ( wptr = latest_clause.varlist; wptr; wptr = wptr->nxt )
    wptr->w = copy_to_form(wptr->w);
  for ( wptr = latest_clause.testlist; wptr; wptr = wptr->nxt )
    wptr->w = copy_to_form(wptr->w);
  
  /* add it to the end of the current theory */
  if ( size[current_theory] == CL_MAX )
    skipout("Too many clauses (whew!)",1);

  target = theory[current_theory] + size[current_theory];
  target->prem = latest_clause.prem;
  target->conc = latest_clause.conc;
  target->varlist =  latest_clause.varlist;
  target->testlist = latest_clause.testlist;
  target->weight = latest_clause.weight;
  target->varcount = latest_clause.varcount;
  target->softweight = latest_clause.softweight;
  target->lastclause = 0;

  if ( size[current_theory] > 0) { // theory exists, TH is defined
                                   /* new clause goes at front of theory */
    target->nextclause = TH[current_theory];
    TH[current_theory]->lastclause = target;
    TH[current_theory] = target;
    normalize();
  }
  else target->nextclause = 0;

  size[current_theory]++; /* updates the actual size of the current_theory */
  
  /* new clause is no longer the latest_clause */
  latest_clause.prem = 0;
  latest_clause.conc = 0;
  latest_clause.varlist = 0;
  latest_clause.testlist = 0;
}


/*-------------------------------------------------------------------------*/
wff copy_to_form( wff w )

     /*Called by add_to_theory().
      *Recursively copies the wff from the given structure to the form[],
      *identifying subformulae that already occur as it goes.*/

{
  int i,j;
  boolean subs_match;

  if ( !w ) return 0;

  /* for each argument of the main functor */
  for (i=0; i<w->sym->adicity; i++)
    w->sb[i] = copy_to_form( w->sb[i] );

  /* check to see if its already in the stack form[] */
  for ( i=0; form[i].sym; i++ ) {
    if (form[i].sym->fname == w->sym->fname
	&& form[i].sym->val_type == w->sym->val_type) {     
      if (w->sym->category == Var || w->sym->category == Const ) 
	return &(form[i]); /*found it!!*/

      /*compare subformulae*/
      subs_match = true;
      for (j=0; j<w->sym->adicity; j++)
	if (form[i].sb[j] != w->sb[j])
	  subs_match = false;
      if (subs_match) return &(form[i]);
    }
  }

  /* doesn't already exist so make a new one */
  if (i >= WF_MAX ) skipout("Too many formulas parsed",1);
  form[i].sym = w->sym;
  for (j=0; j<ARGMAX; j++) {
    form[i].sb[j] = (j<w->sym->adicity? w->sb[j]: 0);
    form[i].sbval[j] = (w->sb[j]? &(w->sb[j]->val): &zero);
  }
  form[i].matrix = w->sym->table;
  if ( w->sym->category == Const ) form[i].val = w->val;
  form[i].wcc = -1;
  return &(form[i]); /* this is for the recursive call on subformulae */
}

/*================================ END ===================================*/
