#include "FINDER.h"



int getoffset(functor f, int argument[])
{
  int i;
  int offset = 1;
  int comx, comy;

  if (!f->adicity) return 0;
  if (f->comm) {
    for (i=0; i<2; i++)
      if (argument[i] >= f->arg_type[i]->siz || argument[i] < 0)
	return 0;
    if (argument[0] < argument[1]) {
      comx = argument[1];
      comy = argument[0];
    }
    else {
      comx = argument[0];
      comy = argument[1];
    }
    return ((comx*(comx+1))/2) + comy +1;
  }
  for (i=0; i<f->adicity; i++) {
    if (f->arg_type[i] &&
	(argument[i] >= f->arg_type[i]->siz || argument[i] < 0))
      return 0;
    offset += (argument[i] * f->fats[i]);
  }
  return offset;
}



int vec_pos(functor f, int argument[])
{
  return f->fcc + getoffset(f,argument);
}




/*
*	This simply counts the non-constant symbols in the latest
*	clause (merging duplicates) and assigns the result to its 
*	"weight" field. It also counts the constituent variables 
*	for its "varcount".
*/


void get_weight()
{
  wff_stack wf;
  int wt;
  functor f;
  boolean piped_content = false;

  wt = 0;
  for ( wf = latest_clause.testlist; wf; wf = wf->nxt ) {
    f = wf->w->sym;
    if ( f->piped ) piped_content = true;	
    if ( is_open(f) && f->category != Const )
      wt++;
  }

  if ( !(latest_clause.weight = wt) && !piped_content )
    inwarning("Clause contains no function symbol",""); 

#ifndef SOFTIE_SCOTT
  if (latest_clause.weight > setting.pretest_limit &&
      latest_clause.softweight)
    inerror("Soft clause contains too many function symbols","");
#endif

  latest_clause.varcount = 0;
  for ( wf = latest_clause.varlist; wf; wf = wf->nxt )
    latest_clause.varcount++;
}



/*
*	We have just added a new clause to the start of the theory
*	list and set TH to point at it. So move it into the middle
*	of the list until it is in its correct position, then move
*	TH to point back at the start of the list. Note that this 
*	version of normalize assumes that the clauses are added one
*	by one, and that the pre-existing list is already normal.
*/

void normalize()
{
  while ( is_after(TH[current_theory], 
		   TH[current_theory]->nextclause) ) {
    TH[current_theory]->nextclause->lastclause = 
      TH[current_theory]->lastclause;

    if ( TH[current_theory]->lastclause )
      TH[current_theory]->lastclause->nextclause = 
	TH[current_theory]->nextclause;

    TH[current_theory]->lastclause = TH[current_theory]->nextclause;

    TH[current_theory]->nextclause = 
      TH[current_theory]->lastclause->nextclause;

    TH[current_theory]->lastclause->nextclause = TH[current_theory];

    if ( TH[current_theory]->nextclause )
      TH[current_theory]->nextclause->lastclause = TH[current_theory];
  }
  while ( TH[current_theory]->lastclause )
    TH[current_theory] = TH[current_theory]->lastclause;
}




/*
*  Clause c1 is after clause c2 if its weight and variable count 
*  say so. This is where the order of clauses is determined. Note
*  that the latest clause is not included.
*
*  A clause of pre-testable weight beats one of greater weight
*  irrespective of any other conditions.
*
*  If both are pre-testable the lighter wins, then the one with 
*  fewer variables, then the accession order.
*
*  If neither is pre-testable the lower weight wins, then the
*  lower variable count (or variables followed by weight if
*  vars_over_wt is set), then accession order.
*/

boolean is_after( clause c1, clause c2 )
{
  if ( !c1 || !c2 ) return false;

  if ( c2->weight < setting.pretest_limit ) {
    if ( c1->weight > c2->weight ) return true;
    if ( c1->weight < c2->weight ) return false;
    if ( c1->varcount > c2->varcount ) return true;
    if ( c1->varcount < c2->varcount ) return false;
    return ( c1 > c2 );
  }
  if ( c1->weight < setting.pretest_limit ) return false;
  
  if ( setting.vars_over_wt ) {
    if ( c1->varcount > c2->varcount ) return true;
    if ( c1->varcount < c2->varcount ) return false;
    if ( c1->weight > c2->weight ) return true;
    if ( c1->weight < c2->weight ) return false;
  }
  else {
    if ( c1->weight > c2->weight ) return true;
    if ( c1->weight < c2->weight ) return false;
    if ( c1->varcount > c2->varcount ) return true;
    if ( c1->varcount < c2->varcount ) return false;
  }
  return ( c1 > c2 );
}




boolean set_up_cc()
{
  set_sum();
  set_minus();
  set_E();
  if (!set_cc())
    return false;
  tr_parameters.vlength = 0;
  if (!set_up_for(FN))
    return false;
  if (tr_parameters.vlength >= V_LENGTH) {
    fprintf(WARNING_STREAM,"Errrrk!!\n");
    return false;
  }
  if (!set_table(FN))
    return false;

  set_matrix(TH[current_theory]);
  if (latest_clause.prem || latest_clause.conc)
    set_matrix(&latest_clause);

#ifdef DEBUG
  { functor f;
  for ( f = FN; f; f = f->nextfunct ) if ( !f->piped ) print_set( f );
  }
#endif

  if ( !(setting.pm.verbosity & pstats) )
    tr_parameters.verb = 0;
  else if ( setting.pm.verbosity & pbstats )
    tr_parameters.verb = 1;
  else if ( setting.pm.verbosity & pfstats )
    tr_parameters.verb = 2;
  tr_parameters.Nosecs = setting.Nosecs;
  tr_parameters.topsoff = setting.topsoff;
  tr_parameters.report_breaks = setting.report_breaks;
  tr_parameters.random = setting.random;
  tr_parameters.batches = setting.batches;
  tr_parameters.extra_batches = setting.extra_batches;
  tr_parameters.maxref = setting.maxref;
  tr_parameters.maxbak = setting.max_backtracks;
  tr_parameters.maxinvals = setting.max_invalues;
  tr_parameters.max_penalty = setting.target_bound;
  ttr_par.maxref = setting.test_maxref;
  return true;
}




/*
 *  This sets the "matrix" field of each formula to point at the table
 *  of values for the corresponding symbol.
 */

void set_matrix(clause c)
{
  if (!c) return;
  set_matrix(c->nextclause);
  set_my_matrix(c->prem);
  set_my_matrix(c->conc);
}

void set_matrices_flat(wff w); // prototype

void set_my_matrix(wff_stack ws)
{
  if (!ws) return;
  set_my_matrix(ws->nxt);
  set_matrices(ws->w);
}


void set_matrices(wff w)
{
  int i;
  
  if (!w) return;
  for (i=0; i<w->sym->adicity; i++)
    set_matrices( w->sb[i] );
  w->matrix = w->sym->table; 

} /* set_matrices() */



/*
 * The following function mallocs memory for the table of values for
 * each function. The built-in ones for each sort already have their
 * tables allocated at declaration, so they are skipped here.
 * The value returned is "true" (success) or "false" (the required
 * table was too big to fit in the memory available).
 */

boolean set_table( functor f )
{
  int count, i;

  if (!f) return true;
  if (!set_table(f->nextfunct)) return false;
  if (f->piped) return true;
  if (strchr("=><+-",f->fname[0])) return true;
  if (!strncmp("DIF",f->fname,3)) return true;
  if (!strncmp("EST",f->fname,3)) return true;
  if (!strncmp("LEN",f->fname,3)) return true;
  if (f->adicity) count = 2; else count = 1;
  for (i=0; i<f->adicity; i++) {
    if (MAXMATRIX/((f->arg_type[i]->siz)) < count)
      return false;
    else count *= (f->arg_type[i]->siz);
  }
  if (f->table) free(f->table);
  f->table = (int*) malloc(sizeof(int) * count);
  return true;
}




boolean set_up_for( functor f )
{
  int i, j;
  int argument[ARGMAX];

  if ( !f ) return true;
  if (!set_up_for( f->nextfunct )) return false;
  if ( f->piped ) return true;
  f->fcc = tr_parameters.vlength;
  if (!f->adicity) {
    tr_parameters.forbidden[tr_parameters.vlength++] = f->cut_it;
    if (tr_parameters.vlength >= V_LENGTH) return false;
  }
  else {
    for (i=0; i<f->adicity; i++) argument[i] = 0;
    do {
      j = vec_pos(f,argument);
      if (j >= V_LENGTH) return false;
      tr_parameters.forbidden[j] = f->cut_it;
      if (j >= tr_parameters.vlength) tr_parameters.vlength = j+1;
    }
    while (another_argument(f,argument));
  }
  return true;
}



void set_sum()
{
  int i, j, k;
  functor f;
  char s[LEXLENGTH];

  for ( k = 0; sort[k].sname; k++ ) {
    sort[k].sum[0] = sort[k].siz;
    for ( i = 0; i < SZ; i++ )
      for ( j = 0; j < SZ; j++ )
	sort[k].sum[(SZ*i)+j+1] = i+j < sort[k].siz? i+j: sort[k].siz;
    strcpy(s,"+_"); strcat(s,sort[k].sname);
    f = newfunctor(s,Func,2);
    if (f->val_type != sort+k) {
      for (i=0; lexicon[i].word[0]; i++) printf("%s\n",lexicon[i].word);
      inerror("Incorrectly defined addition function",s);
    }
    f->fats[0] = SZ;
    f->fats[1] = 1;
    emptyset(sort[k].sizmask);
    for ( i = 0; i < sort[k].siz; i++ )
      intoset(i,sort[k].sizmask);
  }
}



void set_minus()
{
  int i, j, k;
  functor f;
  char s[LEXLENGTH];

  for ( k = 0; sort[k].sname; k++ ) {
    sort[k].minus[0] = sort[k].siz;
    for ( i = 0; i < SZ; i++ )
      for ( j = 0; j < SZ; j++ )
	sort[k].minus[(SZ*i)+j+1] = i >= j? i-j: sort[k].siz;
    strcpy(s,"-_"); strcat(s,sort[k].sname);
    f = newfunctor(s,Func,2);
    if (f->val_type != sort+k) {
      for (i=0; lexicon[i].word[0]; i++) printf("%s\n",lexicon[i].word);
      inerror("Incorrectly defined subtraction function",s);
    }
    f->fats[0] = SZ;
    f->fats[1] = 1;
    emptyset(sort[k].sizmask);
    for ( i = 0; i < sort[k].siz; i++ )
      intoset(i,sort[k].sizmask);
  }
}



void set_E()
{
  int i, k;

  for ( k = 0; sort[k].sname; k++ )
    for ( i = -1; i < SZ-1; i++ )
      sort[k].Ex[i+1] = (i>=0 && i < sort[k].siz)? true: false;
}



boolean set_fats(functor f)
{
  int i;

  if (f->adicity)
    f->fats[f->adicity-1] = 1;
  for (i=f->adicity-2; i>= 0; i--) {
    f->fats[i] = f->fats[i+1] * (f->arg_type[i+1]->siz);
    if (f->fats[i] * (f->arg_type[i]->siz) >= V_LENGTH) return false;
  }
  return true;
}



boolean set_cc()
{
  functor f;

  for ( f = FN; f; f = f->nextfunct )
    if (!set_fats(f)) return false;
  return true;
}


/*
 *  This is extremely slow, of course, but as it's only used for
 *  debugging, speed is less important than obvious correctness.
 */

functor un_vec_pos(int x, int theArguments[])
{
  functor f;
  int i;

  for (f=FN; f; f = f->nextfunct) {
    for (i=0; i<f->adicity; i++) theArguments[i] = 0;
    do {
      if (vec_pos(f,theArguments) == x) return f;
    }
    while (another_argument(f,theArguments));
  }
  return 0;
}


void report_constraint(ref thisref)
{
  intlist lp;
  int args[ARGMAX];
  functor f;
  int i;

  for (lp=thisref->cellvals; lp; lp = lp->p) {
    f = un_vec_pos(lp->v,args);
    if (!lp->positive) printf("not ");
    printf("%s(",f->fname);
    for (i=0; i<f->adicity; i++) {
      if (i) printf(",%d",args[i]);
      else printf("%d",args[i]);
    }
    printf(") = %d\n", lp->w);
  }
  printf("\n");
}



void print_set( functor f )
{
  int i, j;
  int argument[ARGMAX];

  if (f->adicity > 2) {
    printf("\nAdicity of %s (fcc %d) greater than 2\n",f->fname,f->fcc);
    return;
  }
  printf("\nIndex for %s",f->fname);
  if ( !f->adicity ) {
    printf(" %d\n", vec_pos(f,argument));
    return;
  }
  puts("");
  if ( f->adicity == 1 ) {
    for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
      argument[0] = i;
      printf("%4d",vec_pos(f,argument));
    }
    argument[0] = i;
    printf("%5d\n",vec_pos(f,argument));
    return;
  }
  for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
    argument[0] = i;
    for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
      argument[1] = j;
      printf("%4d",vec_pos(f,argument));
    }
    argument[1] = j;
    printf("%5d",vec_pos(f,argument));
    puts("");
  }
  puts("");
  argument[0] = i;
  for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
    argument[1] = j;
    printf("%4d",vec_pos(f,argument));
  }
  argument[1] = j;
  printf("%5d",vec_pos(f,argument));
  puts("");
}



void set_max_penalty(int x)
{
  setting.target_bound = x;
}
