
#include "FINDER.h"


/*
*  Do_smallrefs is called after the search space has been prepared, to 
*  record refutations known to be of cardinality setting.pre_test or less.
*  Injective and surjective functions are also treated at this stage. 
*
*  The first bit of this function sets the matrix values of the nulladic 
*  functions which are fixed and which therefore don't count in refutations.
*/

int do_smallrefs( set info[], TRS *T )
{
  int twt, i;
  clause c;
  functor f;
  boolean do_one_small_clause();

  for ( f = FN; f; f = f->nextfunct ) if ( !f->adicity )
    if ( issingle(info[f->fcc]) ) {
      for ( i = 0; !isinset(i,info[f->fcc]); i++ ) ;
      *(f->table) = i;
    }

  do_injectives(T);
  if ( !do_surjectives(T) ) return FALSE;

  for ( twt = 0; twt <= setting.pretest_limit; twt++ ) {
    for ( c = TH[current_theory]; c; c = c->nextclause ) 
      if (!do_one_small_clause(c,twt,info,T))
	return FALSE;
    if ((latest_clause.prem || latest_clause.conc)
	&& !do_one_small_clause(&latest_clause,twt,info,T))
      return FALSE;
  }

  return TRUE;
}



boolean do_one_small_clause(clause c, int twt, set info[], TRS *T)
{
  int lwt;
  wff plit;

  if ( !c->preprocessed && !T->cut_flag ) {
    lwt = localweight(info,c,&plit,T);
    if (lwt <= twt || (plit && lwt == twt+1)) {
      c->preprocessed = true;
#ifdef DEBUG
      printf("Preprocessing ");
      print_one_clause(stdout,c);
      printf("\n");
#endif
      if ( !twt_test(info, c, lwt, plit, T) )
	return FALSE;
    }
  }
  return TRUE;
}


/*
*  The test routine for a clause whose local weight is lwt.
*  Return value is false iff a 0-refutation has been found.
*/

boolean twt_test( set info[], clause c, int lwt, wff plit, TRS *T )
{
  wff_stack wf, tf[PRE_MAX];
  int maximum[PRE_MAX];
  int i;
  functor f;
  set locally_supported[PRE_MAX];

  boolean vacuous();

  tf[lwt] = 0;
  maximum[lwt] = 0;
  if ( lwt )
    tf[0] = firstopen( info, c->testlist, lwt );
  for ( i = 1; i < lwt; i++ )
    tf[i] = firstopen( info, tf[i-1]->nxt, 1 );
  for ( i = 0; i < lwt; i++ )
    maximum[i] = tf[i]->w->sym->val_type->siz;
  correct_order( tf );

/*
*  First put in the values of any fixed cells
*/
  for ( wf = c->testlist; wf; wf = wf->nxt ) {
    f = wf->w->sym;
    if ( is_open(f) ) {
      if ( !f->adicity && issingle(info[f->fcc]) )
	for ( i = 0; !isinset(i,info[f->fcc]); i++ ) ;
      *(f->table) = i;
    }
  }

/*
*  Next initialize all the variables to the first (highest) value.
*/
  for ( wf = c->varlist; wf; wf = wf->nxt )
    wf->w->val = wf->w->sym->val_type->siz-1;
/*
*  Now repeat the test until no more variable assignments remain.
*/
  do {
    if (!vacuous(c,tf)) {
      for (i=0; i<PRE_MAX; i++)
	emptyset(locally_supported[i]);

      if ( !alive(c,
		  tf,
		  tf,
		  maximum,
		  c->testlist,
		  info,
		  locally_supported,
		  plit,
		  T)
	   )
	return false;

      if (0) {
	if (!part_ac(info,locally_supported,tf,maximum,T))
	  return false;
      }
    }
  }
  while ( !T->cut_flag && next_assignment(c) );
  return true;
}



boolean vacuous(clause c, wff_stack tf[])
{
  int i;
  wff_stack wf;
  int subvalues[ARGMAX];
  boolean novalue;

  if (1) return false;

  for (i=0; tf[i]; i++) tf[i]->w->val = -1;
  for (wf = c->testlist; wf; wf = wf->nxt) {
    novalue = false;
    for (i=0; i<wf->w->sym->adicity; i++) {
      subvalues[i] = *(wf->w->sbval[i]);
      if (subvalues[i] < 0) {
	novalue = true;
	break;
      }
    }
    if (novalue && wf->w->val != -1) wf->w->val = -2;
    else wf->w->val = *(wf->w->matrix + getoffset(wf->w->sym,subvalues));
  }
  return true;
}



/*
 * THIS MAY BE GOING WRONG WHEN TWO OPEN CELLS REFER TO THE SAME SPOT IN
 * THE "INFO" VECTOR. WELL, SOMETHING IS GOING WRONG! AH, IT'S WHEN THE
 * INFO REFERENCE OF ONE CELL IS A FUNCTION OF THE VALUE IN ANOTHER.
 *
 * HAVE TO LOOK AT ALL OF THIS AGAIN. PROBABLY DO SYMBOLIC CONSTRAINTS
 * AND AC PROPERLY AFTER ALL.
 */


boolean part_ac(set info[], set locally_supported[],
		wff_stack *tf, int maximum[], trs T)
{
  int args[ARGMAX];
  int oneref_cells[2];
  int oneref_vals[2];
  boolean oneref_valency[2];
  int i,j,k;

  oneref_cells[1] = -1; oneref_vals[1] = 0;
  oneref_valency[0] = oneref_valency[1] = false;
  for (j=0; tf[j]; j++) {
    for (i=0; i<tf[j]->w->sym->adicity; i++)
      args[i] = *(tf[j]->w->sbval[i]);
    k = vec_pos(tf[j]->w->sym,args);
    if (info[k] != locally_supported[j]) {
      for (i=0; i<=maximum[j]; i++)
	if (isinset(i,info[k]) && !isinset(i,locally_supported[j])) {
	  oneref_cells[0] = k;
	  oneref_vals[0] = i;
	  new_small_ref(oneref_cells,oneref_vals,oneref_valency,0,T);
	  if (T->novect) return false;
	}
      intersect(info[k],locally_supported[j]);
    }
  }
  return true;
}



boolean alive(clause c, wff_stack *tf, wff_stack *ltf, int *maximum,
	      wff_stack wf, set info[], set locsup[], wff plit, TRS *T)
{
  int i;
  int subvalues[ARGMAX];

  for ( ; wf != *ltf; wf = wf->nxt ) {
    for (i=0; i<wf->w->sym->adicity; i++)
      subvalues[i] = *(wf->w->sbval[i]);
    wf->w->val = *(wf->w->matrix + getoffset(wf->w->sym,subvalues));
  }

  if ( !*ltf ) return good_all( c, tf, locsup, plit, T );

  if ( wf->w == plit )
    return alive(c,
		 tf,
		 ltf+1,
		 maximum+1,
		 (*ltf)->nxt,
		 info,
		 locsup,
		 plit,
		 T);

  for ( (*ltf)->w->val = 0; (*ltf)->w->val <= *maximum; (*ltf)->w->val++ )
    if ( poss_val((*ltf)->w,info) )
      if (!alive(c,
		 tf,
		 ltf+1,
		 maximum+1,
		 (*ltf)->nxt,
		 info,
		 locsup,
		 plit,
		 T)
	  )
	return false;
  return true;
}




/*
*  If the clause fails, inform TRANSREF. Return false iff the result
* refutes the entire search space (this will cause exit from do_twos).  */

boolean good_all( clause c, wff_stack tf[], set locsup[], wff plit, TRS *T)
{
  boolean failure = true;
  wff_stack wf;
  wff W;
  int cells[PRE_MAX], vals[PRE_MAX];
  boolean valency[PRE_MAX];
  int i,j;
  int subvalues[ARGMAX];

  for ( wf = c->prem; wf; wf = wf->nxt )
    if ( !wf->w->val ) failure = false;
  if ( plit ) plit->val = c->conc->w->sb[1]->val;
  else for ( wf = c->conc; wf; wf = wf->nxt )
    if ( wf->w->val ) failure = false;

  if ( failure ) {
    for ( i = 0; tf[i]; i++ ) {
      W = tf[i]->w;
      for (j=0; j<W->sym->adicity; j++)
	subvalues[j] = *(W->sbval[j]);
      cells[i] = vec_pos(W->sym, subvalues);
      vals[i] = W->val;
      valency[i] = (W == plit);
    }
    cells[i] = -1;
    new_small_ref( cells, vals, valency, c->softweight, T );
    if ( T->novect ) return false;
  }
  else {                     /* Current values are supported */
    for (i=0; tf[i]; i++)
      intoset(tf[i]->w->val,locsup[i]);
  }
  return true;
}




/*
*  Poss_val returns true iff the value assigned to w is possible
*  according to the vector of possible values. Note that this 
*  calls function vec_pos.
*
*  If one of the input values is the dummy, then poss_val is true 
*  iff the value of w is the dummy of the appropriate type.
*/

boolean poss_val( wff w, set info[] )
{
  int i;
  int args[ARGMAX];

  for (i=0; i<w->sym->adicity; i++)
    if (*(w->sbval[i]) == w->sym->arg_type[i]->siz)
      return (w->val==w->sym->val_type->siz);
  for (i=0; i<w->sym->adicity; i++)
    args[i] = *(w->sbval[i]);
  return isinset(w->val,info[vec_pos(w->sym,args)]);
}



/*
*  For testing purposes, tfb is assumed to be in the posterity
*  of the other open cell tfa. They just might have got assigned
*  the wrong way round, so this function tests for correctness.
*/

void correct_order( wff_stack tf[] )
{
  wff_stack wf;
  int i;

  if ( !tf[0] || !tf[1] ) return;
  for ( i = 0; tf[i+1]; i++ ) {
    for ( wf = tf[i]->nxt; wf; wf = wf->nxt )
      if ( wf == tf[i+1] ) break;
    if ( !wf ) skipout("Bad pointer sequence in a clause",1);
  }
}




/*
*  Injectiveness is special to TRANSREF. It has to be informed of
*  the set of cells whose values cannot be identical, of the upper
*  bound on the set of values which cannot be assigned to more 
*  than one of the cells, and of course of the TRS structure to 
*  which all of this pertains. The upper bound on values for this 
*  purpose is to allow partial functions to be injective.
*
*  THIS MECHANISM WILL BE REPLACED BY A PROPER ALLDIFF ONE WHEN I
*  GET AROUND TO HACKING IT.
*/

void do_injectives( TRS *T )
{
  boolean dinfo[V_LENGTH];
  int i, j;
  functor f;
  int argument[ARGMAX];

  for ( f = FN; f; f = f->nextfunct ) if ( !f->piped ) {
    if ( f->injective ) {
      for ( i = 0; i < tr_parameters.vlength; i++ )
	dinfo[i] = false;
      set_dinfo(f,f->adicity-1,argument,dinfo);
      injective( dinfo, f->val_type->siz, T );
      f->leftin = f->rightin = false;
    }
    if ( f->rightin ) {
      for ( i = 0; i < tr_parameters.vlength; i++ )
	dinfo[i] = false;
      for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	argument[0] = i;
	for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
	  argument[1] = j;
	  dinfo[vec_pos(f,argument)] = true;
	}
	injective( dinfo, f->val_type->siz, T );
	for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
	  argument[1] = j;
	  dinfo[vec_pos(f,argument)] = false;
	}
      }
    }
    if ( f->leftin ) {
      for ( i = 0; i < tr_parameters.vlength; i++ )
	dinfo[i] = false;
      for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
	argument[1] = j;
	for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	  argument[0] = i;
	  dinfo[vec_pos(f,argument)] = true;
	}
	injective( dinfo, f->val_type->siz, T );
	for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	  argument[0] = i;
	  dinfo[vec_pos(f,argument)] = false;
	}
      }
    }
  }
}



void set_dinfo(functor f, int x, int argument[], boolean dinfo[])
{
  for (argument[x]=0; argument[x]<f->arg_type[x]->siz; argument[x]++) {
    if (x)
      set_dinfo(f,x-1,argument,dinfo);
    else
      dinfo[vec_pos(f,argument)] = true;
  }
}




/*
*  Surjective functions are also treated specially by TRANSREF.
*  Left and right surjectiveness are treated here, as well as 
*  total surjectiveness.
*/

boolean do_surjectives( TRS *T )
{
  boolean dinfo[V_LENGTH];
  set vset;
  int i, j, m;
  int valcount, argcount;
  functor f;
  int argument[ARGMAX];

  for ( f = FN; f; f = f->nextfunct ) if ( !f->piped ) {
    if ( f->rightsur ) {
      for ( i = 0; i < tr_parameters.vlength; i++ ) dinfo[i] = false;
      for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	argument[0] = i;
	for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
	  argument[1] = j;
	  dinfo[vec_pos(f,argument)] = true;
	}
	emptyset(vset);
	valcount = 0;
	for ( m = 0; m < f->val_type->siz; m++ ) {
	  intoset(m,vset);
	  valcount++;
	}
	if ( valcount > f->arg_type[1]->siz ) return false;
	surjective( dinfo, vset, T );
	for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
	  argument[1] = j;
	  dinfo[vec_pos(f,argument)] = false;
	}
      }
      f->surjective = false;
    }
    if ( f->leftsur ) {
      for ( i = 0; i < tr_parameters.vlength; i++ ) dinfo[i] = false;
      for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
	argument[1] = j;
	for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	  argument[0] = i;
	  dinfo[vec_pos(f,argument)] = true;
	}
	emptyset(vset);
	valcount = 0;
	for ( m = 0; m < f->val_type->siz; m++ ) {
	  intoset(m,vset);
	  valcount++;
	}
	if ( valcount > f->arg_type[0]->siz ) return false;
	surjective( dinfo, vset, T );
	for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	  argument[0] = i;
	  dinfo[vec_pos(f,argument)] = false;
	}
	f->surjective = false;
      }
    }
    if ( f->surjective ) {
      for ( i = 0; i < tr_parameters.vlength; i++ ) dinfo[i] = false;
      set_dinfo(f,f->adicity-1,argument,dinfo);
      argcount = 1;
      for (i=0; i<f->adicity; i++)
	argcount *= f->arg_type[i]->siz;
      emptyset(vset);
      valcount = 0;
      for ( i = 0; i < f->val_type->siz; i++ ) {
	intoset(i,vset);
	valcount++;
      }
      if ( valcount > argcount ) return false;
      surjective( dinfo, vset, T );
    }
  }
  return true;
}



/*
*  Local weight is weight minus any nulladic functions with only
*  one possible value. The parameter *plit is set to point to the
*  left subformula of the sole positive literal (if any) which must
*  have identity as its predicate and an open left subformula.
*/ 

int localweight( set info[], clause c, wff *plit, TRS *T )
{
  wff_stack wf;
  int counter = 0;
  functor f;
  boolean poscon;    /* Positive literal condition */

  poscon = (
	    c->conc &&
	    !c->conc->nxt &&
	    *(c->conc->w->sym->fname) == '=' &&
	    is_open(c->conc->w->sb[0]->sym) &&
	    !occurs_inside(c->conc->w->sb[0], c->conc->w->sb[1]) &&
	    !occurs_in_list(c->conc->w->sb[0], c->prem) );
  if ( poscon ) *plit = c->conc->w->sb[0];
  else *plit = 0;

  for ( wf = c->testlist; wf; wf = wf->nxt ) {
    f = wf->w->sym;
    if ( is_open(f) ) {
      if ( f->adicity || !issingle(info[f->fcc]))
	counter++;
    }
  }
  return counter;
}



/*
*  The first open spot in a list of wffs is the first one which
*  is part of a function that is_open, other than a nulladic one
*  with only one possible value.
*/

wff_stack firstopen( set info[], wff_stack wfl, int opens )
{
  functor f;

  if (!opens) printf("Firstopen called without opens!\n");
  f = wfl->w->sym;
  while ( !is_open(f) ||
	  (opens && !f->adicity && issingle(info[f->fcc]))) {
    wfl = wfl->nxt;
    f = wfl->w->sym;
  }
  return wfl;
}



/*
*  A function is_open if it is really part of the search space 
*  and not one of the constant functions such as identity or a
*  function piped in from another FINDER.
*/

boolean is_open( functor f )
{
  if ( !f ) return false;
 /*
  * if ( f->category != Func && f->category != Const ) return false;
  */
  if ( f->category != Func ) return false;
  if ( strchr("+-=<>",f->fname[0]) ) return false;
  if ( !strncmp(f->fname,"DIF",3) ) return false;
  if ( !strncmp(f->fname,"EST",3) ) return false;
  if ( !strncmp(f->fname,"NOT",3) ) return false;
  if ( !strncmp(f->fname,"LEN",3) ) return false;
  if ( f->piped ) return false;
  return true;
}


/*
*  Formula fmla occurs as a subformula of something in wfl.
*/

boolean occurs_in_list( wff fmla, wff_stack wfl )
{
  if ( !wfl ) return false;
  do
    if ( occurs_inside(fmla,wfl->w) ) return true;
  while ( (wfl = wfl->nxt) );
  return false;
}


/*
*  Formula little occurs as a subformula of formula big.
*/

boolean occurs_inside( wff little, wff big )
{
  int i;

  if ( !big ) return false;
  if ( little == big ) return true;
  for (i=0; i<big->sym->adicity; i++)
    if ( occurs_inside(little, big->sb[i]) ) return true;
  return false;
}
