
#include "FINDER.h"


/*
* Setposs is called from transref to trim the search space by removing
* impossible values from the vector of sets of possibilities. Note that
* according to TRANSREF this function should only subtract values, never
* add them.
*
* First it sets upper bounds according to the cardinalities of the 
* domains, then it looks to clauses of weight 1 to take out further
* impossible values. Clauses which are of local weight 1 but not of 
* absolute weight 1 will get treated a little later in do_twos.
*
* Clauses of weight 0 (which must include piped stuff) are dealt with
* at this point, and need never again be checked.
*/

int setposs( set info[], TRS *T )
{
  functor f;
  clause c;

  set_priority( FN, T );

  for ( f = FN; f; f = f->nextfunct ) if ( !f->piped )
    set_possval( info, f, T );

  for ( c = TH[current_theory]; c; c = c->nextclause )
    c->preprocessed = false;

  for ( c = TH[current_theory]; c; c = c->nextclause )
    if ( !c->weight ) {
      c->preprocessed = true;
      if ( !setposs_test(info,c,0,T) )
	return FALSE;
    }
    else if ( c->weight==1 ) {
      c->preprocessed = true;
      if ( !good_one(info,c,T) )
	return FALSE;
    }
  return TRUE;
}



/*
* The set of possible values is trimmed to exclude any which are bigger
* than the highest within the domain of the value sort. Special allowance
* has to be made for the dummy value in cases of dummy arguments.
*/

void set_possval( set info[], functor f, TRS *T )
{
  int i, sz;
  set mask;
  int argument[ARGMAX];
  boolean undefined;

  sz = f->total? f->val_type->siz: f->val_type->siz+1;
  emptyset(mask);
  for ( i = 0; i < sz; i++ )
    intoset(i,mask);
  for (i=0; i<f->adicity; i++)
    argument[i] = 0;
  do {
    undefined = false;
    for (i=0; i<f->adicity; i++)
      if (argument[i] == f->arg_type[i]->siz)
	undefined = true;
    if (undefined)
      singleton(f->val_type->siz,info[vec_pos(f,argument)]);
    else
      intersect(info[vec_pos(f,argument)],mask);
  }
  while (another_argument(f,argument));
}


boolean another_argument(functor f, int argument[])
{
  int i;

  for (i=0; i<f->adicity; i++) {
    if (argument[i] < f->arg_type[i]->siz) {
      argument[i]++;
      return true;
    }
    else argument[i] = 0;
  }
  return false;
}




int set_priority( functor f, TRS *T )
{
  int i;
  int x;
  int argument[ARGMAX];

  if ( !f ) return 0;
  x = set_priority( f->nextfunct, T );
  if ( x > 0 && f->nop ) x--;
  for (i=0; i<f->adicity; i++) argument[i] = 0;
  do
    co_priority(vec_pos(f,argument),x,T);
  while (another_argument(f,argument));
  return x+1;
}





boolean good_one( set info[], clause c, trs T )
{  
  wff_stack tf;
  int sz;

  for (tf = c->testlist;
       tf && !is_open(tf->w->sym);
       tf = tf->nxt)
    ;
  sz = tf->w->sym->val_type->siz+1;
  for ( tf->w->val = 0; tf->w->val < sz; tf->w->val++ )
    if ( !setposs_test(info, c, tf, T) )
      return false;
  return true;
}




boolean setposs_test( set info[], clause c, wff_stack tf, trs T )
{
  wff_stack wf, check;
  int vp;
  boolean failure;
  int i;
  int subvalues[ARGMAX];

  for ( wf = c->varlist; wf; wf = wf->nxt )
    wf->w->val = wf->w->sym->val_type->siz-1;
  do {
    for ( wf = c->testlist; wf; wf = wf->nxt )
      if ( wf != tf ) {
	for (i=0; i<wf->w->sym->adicity; i++)
	  subvalues[i] = *(wf->w->sbval[i]);
	wf->w->val = *(wf->w->matrix + getoffset(wf->w->sym,subvalues));
      }
    failure = true;
    for ( check = c->prem; check; check = check->nxt )
      if ( !check->w->val ) failure = false;
    for ( check = c->conc; check; check = check->nxt )
      if ( check->w->val ) failure = false;
    if ( failure ) {
      if ( !tf ) return false;
      for (i=0; i<tf->w->sym->adicity; i++)
	subvalues[i] = *(tf->w->sbval[i]);
      vp = vec_pos(tf->w->sym, subvalues);
      if (c->softweight)
	add_penalty(vp,tf->w->val,c->softweight,T);
      else {
	outofset(tf->w->val,info[vp]);
	if ( isemptyset(info[vp]) ) return false;
      }
    }
  }
  while ( next_assignment(c) );
  return true;
}
