/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Hauptkomponenten.h                                                                                         */
/* - ******************                                                                                         */
/* -                                                                                                            */
/* - Enthaelt die Operationen, die zum Programm-Start mit den in der KP-Menge enthaltenen Gleichungen der Spezi-*/
/* - fikation ausgefuehrt werden (I) sowie eine Prozedur, die alle KP's der Reihe nach behandelt und sie zu-    */
/* - sammenzufuehren sucht (II).                                                                                */
/* - Ferner finden sich hier die verschiedenen Vervollstaendigungs-Algorithmen (IV).                            */
/* - Bisher stehen hier auch noch die Prozeduren zur Interreduktion sowie zur Verwaltung der KP-Bildung nach    */
/* - Aufnahme von neuen Regeln bzw. Gleichungen (wird evtl. noch ausgelagert, je nach Umfang).                  */
/* - Darueberhinaus finden sich hier auch die Prozeduren, die als Schnittstellen zu den implementierten Ord-    */
/* - nungen fungieren. (ORD 1)                                                                                  */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef HAUPTKOMPONENTEN_H
#define HAUPTKOMPONENTEN_H

#include "general.h"
#include "InfoStrings.h"
#include "TermOperationen.h"
#include "XFilesFuerPMh.h"

/* -------------------------------------------------------------------------------*/
/* --------------- KPMenge Behandeln (I und II) ----------------------------------*/
/* -------------------------------------------------------------------------------*/

/* (I) */
/*-----*/

BOOLEAN HK_GewichtsloseKPsBehandeln(void);
/* Behandelt der Reihe nach die kritischen Paare mit Gewicht Null (also die Ausgangsgleichungen) in
   Abhaengigkeit von den verschiedenen (einfachen!) Modi wie folgt:
   
   reduction:        Richten links->rechts,  -> R.
   termination:      Richten nach Ordnung, richtbare -> R, unrichtbare -> TP-Puffer.
   confluence :      Richten links->rechts, -> R. KP-Bildung.
   convergence:      Richten nach Ordnung, richtbare -> R, unrichtbare -> TP-Puffer. KPBildung.
   
   Rueckgabe                  == TRUE   : Alle behandelten KP's konnten gerichtet werden.
                                 FALSE  : nicht alle behandelten KP's konnten gerichtet werden.

   Seiteneffekte                        : Nicht richtbare Ausgangsgleichungen werden ausgegeben und in TP-Puffer aufgenommen.

   Vor.: Alle KPs mit Gewicht 0 sind noch gueltig.
   In allen Modi bleibt die Gleichungsmenge leer.
*/

/* (II) */
/*------*/

BOOLEAN HK_KPsZusammenfuehrbar(void);
/* Behandelt alle kritischen Paare der Reihe nach (wie sie in der KP-Menge zu finden sind) wie folgt:
        Es wird versucht, linke und rechte Seite zusammenzufuehren. War dies moeglich,
        so wird fortgefahren, andernfalls wird das kritische Paar (nach der Reduktion!) ausgegeben
   Der Rueckgabewert gibt an, ob alle KP's zusammenfuehrbar waren. 
   Seiteneffekt : nicht richtbare KP's werden ausgegeben.*/

/* -------------------------------------------------------------------------------*/
/* --------------- Vervollstaendigung (IV) ---------------------------------------*/
/* -------------------------------------------------------------------------------*/
/* Verteiler: Je nach Variante (failing/unfailing) wird der 
   entsprechende Alg. aufgerufen. Dies geschieht auf dieser hohen Ebene, um Abfragen innerhalb der
   Algorithmen einzusparen.*/

typedef enum {HK_VervollAbgebrochen, HK_Vervollstaendigt, HK_ZielBewiesen, HK_StrategieWechseln,
              HK_Ueberholt, HK_Beendet}
  HK_ErgebnisT;
  
  
HK_ErgebnisT HK_Vervollstaendigung(StrategieT *Strategie);
/* Vervollstaendigung. Ziele werden je nach Modus erst am Ende oder waehrend des Laufes behandelt.
   Rueckgabe ist FALSE, wenn bei failing-Variante Vervollstaendigung abgebrochen wurde. */

extern BOOLEAN HK_fair;
#define HK_fair() HK_fair

/* -------------------------------------------------------------------------------*/
/* --------------- Ziel-Behandlung (Ia und III) ----------------------------------*/
/* -------------------------------------------------------------------------------*/

void HK_GleichungenUndZieleHolen(void);
/* Die Gleichungen aus der Spezifikation plattklopfen und mit Gewicht 0 in die KP-Menge stecken.
   Die Ziele aus der Spezifikation plattklopfen, dabei skolemisieren und in die Ziele-Liste stecken.*/

BOOLEAN HK_ZieleAufZusammenfuehrbarkeitTesten(void);
/* Alle Ziele aus der Zielmenge bzgl. des aktuellen Regel- und Gleichungs-Systems auf NF bringen. 
   Rueckgabe = TRUE gdw. alle Ziele waren zusammenfuehrbar.
*/

#endif /* HAUPTKOMPONENTEN_H */
