/* *************************************************************************
 *
 * KDHeap.h
 *
 * Array-basierte Heap als PQ.
 *
 * Bernd Loechner, 12.10.2001
 *
 * ************************************************************************* */

#ifndef KDHEAP_H
#define KDHEAP_H

#include "general.h"

#include "KDHeapElements.h"

typedef enum {
  WTI_Nothing, WTI_Reweight, WTI_Stop, WTI_Delete
} KDH_WalkthroughInteractionT;

typedef KDH_WalkthroughInteractionT (*KDH_WalkthroughFct)(ElementT *xp);

typedef BOOLEAN (*KDH_comparisonFct)(unsigned dim, ElementT x, ElementT y);

typedef struct {
  ElementT *xs;
  unsigned long e;
  /* e gibt die Zahl der allokierten Element an. D.h unabhaengig von der
   * Variante sind die Array-Elemente xs[0]...xs[e-1] allokiert.
   */
  unsigned long i;
  /* i gibt die Zahl der Elemente an. Fuer k > MAGIC_K liegen die Elemente
   * in xs[0]...xs[i-1], ansonsten in xs[1]...xs[i].
   */
  unsigned k; /* die Dimension des Heaps. */
  unsigned level_i; /* Invariante: level_i == level(i) */
  KDH_comparisonFct greater;
} KDHeapStruct, *KDHeapT;

KDHeapT KDH_mkHeap(unsigned long initial, unsigned k, KDH_comparisonFct grt);
void    KDH_delHeap(KDHeapT h);

BOOLEAN KDH_isEmpty(KDHeapT h);
unsigned long KDH_size(KDHeapT h);

void KDH_insert (KDHeapT h, ElementT x);
void KDH_deleteMin(KDHeapT h, ElementT *res, unsigned d);
void KDH_deleteElement(KDHeapT h, unsigned long index, ElementT *res);
/* index gibt die _Nummer_ des Elementes an, d.h. unabhaengig von der 
 * Variante muss gelten: 0 <= index < KDH_size(h)
 */
void KDH_Walkthrough(KDHeapT h, KDH_WalkthroughFct f);
/* Iteriert in unspezifizierter Weise ueber alle Elemente in h.
 * liefert f WTI_Delete wird das Element geloescht (ggf. an Pointern haengende
 *   Werte muessen schon von f entsorgt werden!)
 * liefert f WTI_Reweight muss das Element neu einsortiert werden.
 * liefert f WTI_Stop muss das Element neu einsortiert werden, der Durchlauf
     wird danach beendet
 * liefert f WTI_Nothing passiert von Seiten des KDHeaps nichts!
 */
#if 0
void  KDH_halveHeap(KDHeapT h);
#endif

#endif
