/* *************************************************************************
 * KPVerwaltung.h
 *
 * Modul zur Verwaltung der passiven Fakten.
 *
 * Bernd Loechner, 21.04.2001
 *
 * ************************************************************************* */
/**
 * ALLES, WAS IN DIESEM MODUL NICHT MIT KPV ANFAENGT, WIRD ALS MOMENTANE
 * "WART" BETRACHTET.
 **/

#ifndef KPVERWALTUNG_H
#define KPVERWALTUNG_H

#include "general.h"
#include "InfoStrings.h"
#include "RUndEVerwaltung.h"
#include "TermOperationen.h"
#include "TermpairRepresentation.h"

/* --------------------------------------------------------------------------*/
/* ----- Typedefs -----------------------------------------------------------*/
/* --------------------------------------------------------------------------*/

typedef struct{
  WeightT w1, w2;
} WeightEntryT;

/* Reihenfolge der Werte nicht aendern - wird in NewClassification benoetigt */
typedef enum{CPWithFather, CPWithMother, Initial, KilledRE, 
	       ChaseManhattanCP, TWreceived, CPHistory_Last
} CPHistoryT;

/* --------------------------------------------------------------------------*/
/* ----- PARENTS ------------------------------------------------------------*/
/* --------------------------------------------------------------------------*/

void KPV_NewParent(RegelOderGleichungsT re);

void KPV_FinishParent(RegelOderGleichungsT re);

void KPV_KillParent(RegelOderGleichungsT re);

/* --------------------------------------------------------------------------*/
/* ----- SELECTION ----------------------------------------------------------*/
/* --------------------------------------------------------------------------*/

BOOLEAN KPV_Select(SelectRecT *selRec);

void SUE_selectedCPRE(SelectRecT *selRec);

void KPV_PassiveFaktenWegwerfen(void);

void KPS_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

void KPS_InitAposteriori(InfoStringT DEF, InfoStringT USR);

/* --------------------------------------------------------------------------*/
/* ----- KPs in KP-Menge aufnehmen ------------------------------------------*/
/* --------------------------------------------------------------------------*/

void KPV_CPinitial(TermT lhs, TermT rhs);
void KPV_GebildetesKPBehandelnMitVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition,
				       RegelOderGleichungsT actualParent, TermT lhsReplica);
void KPV_GebildetesKPBehandelnMitMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition,
					RegelOderGleichungsT actualParent, TermT lhsReplica);
void KPV_IROpferBehandeln (RegelOderGleichungsT opfer);

void KPV_KonklusionenAbarbeiten(RegelOderGleichungsT r);
/* Meldet die entsprechende Regel oder Gleichung dem Datenbankexperten an und
   verarbeitet die gezogenen Schlussfolgerungen.
*/

/* --------------------------------------------------------------------------*/
/* ----- Interreduktion der KP-Menge ----------------------------------------*/
/* --------------------------------------------------------------------------*/

void KPV_KPMengeInterreduzieren(void);
/* Interreduktion der KP-Menge gemaess Aufruf-Einstellung.*/

void KPV_WahlfreiInterreduzieren(char *howStrong);
/* Fuehrt die KP-Mengen-Interreduktion zu einem beliebigen Zeitpunkt aus.
   Funktionalitaet(Maechtigkeit und dyn. Neubewerten)  wie beim Programm-Start
   vorgesehen.
   howStrong = [res]
*/

void KPV_CGMengeInterreduzieren(BOOLEAN Erzwingen, char *Staerke);
/* Erzwingen=FALSE: IR wird entsprechend durchgefuehrt, 
                    falls Stichpunkt oder Periode erreicht.
   Erzwingen=TRUE:  IR wird mit Staerke durchgefuehrt
*/

/* --------------------------------------------------------------------------*/
/* ---- WALKTHROUGH ---------------------------------------------------------*/
/* --------------------------------------------------------------------------*/

void KPV_Dump (void);

/* --------------------------------------------------------------------------*/
/* ---- CLEANING WOMAN ------------------------------------------------------*/
/* --------------------------------------------------------------------------*/

void KPV_CLEANUP(void);
void KPV_SOFTCLEANUP(void);

/* --------------------------------------------------------------------------*/
/* ----- Initialisierung/parameter ------------------------------------------*/
/* --------------------------------------------------------------------------*/

void KPV_SetzeWaisenmord (BOOLEAN morden);

/* todo ... */
void KPB_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

BOOLEAN KPB_InitAposteriori(InfoStringT DEF, InfoStringT USR);

void KPIR_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

BOOLEAN KPIR_InitAposteriori(InfoStringT DEF, InfoStringT USR);
/* Syntax des Info-Strings:
   period=<periode>[res]:[res]=<stichpunkt>.<stichpunkt>.<stichpunkt>:[res]=<stichpunkt>.<stichpunkt>...[:statistik]

   Beispiel:
   r=100:period=500res:e=150.200.300:statistik

   Default ist: ""

   Die den Zeitpunkt der KPMengen-IR bezeichnenden Zahlen beschreiben die 
   bis dahin insgesamt enstandenen Regeln und Gleichungen, sind also eindeutig.
   
   "statistik" gibt an, ob waehrend der KP-Mengen-IR Informationen gesammelt 
   werden sollen.
*/

void SUE_InitApriori(void);

BOOLEAN SUE_InitAposteriori(InfoStringT DEFsue,  InfoStringT USRsue,
                            InfoStringT DEFclas, InfoStringT USRclas,
                            InfoStringT DEFreclas, InfoStringT USRreclas,
                            InfoStringT DEFcgclas, InfoStringT USRcgclas,
                            InfoStringT DEFpq,   InfoStringT USRpq,
                            InfoStringT DEFtpr,  InfoStringT USRtpr,
                            InfoStringT DEFcg,    InfoStringT USRcg);

void SUE_ParamInfo(ParamInfoArtT what, int whichModule,
                   InfoStringT DEFsue,  InfoStringT USRsue,
                   InfoStringT DEFclas, InfoStringT USRclas,
                   InfoStringT DEFreclas, InfoStringT USRreclas,
                   InfoStringT DEFcgclas, InfoStringT USRcgclas,
                   InfoStringT DEFpq,   InfoStringT USRpq,
                   InfoStringT DEFtpr,  InfoStringT USRtpr,
		   InfoStringT DEFcg,    InfoStringT USRcg);
/* whichModule: 0 - SUE
                1 - CLAS
                2 - RECLAS
                3 - CGCLAS
                4 - PQ
                5 - TPR
		6 - CG
*/

unsigned long int KPH_LebendeKPsZaehlen(void);

#endif
