/*

03.03.94 Angelegt.
07.03.94 Um den Parameter KPHeuristik erweitert.
*/

#ifndef PARAMETER_H
#define PARAMETER_H


/* ------------------------------------------------*/
/* ------------- Includes -------------------------*/
/* ------------------------------------------------*/

#include "parser.h"
#include "general.h"


/* ------------------------------------------------*/
/* -------- Aufruf-Optionen -----------------------*/
/* ------------------------------------------------*/


/* Jede zu ermoeglichende Aufruf-Option wird in einer solchen Struktur eingerichtet.
   Dabei sind die einzelnen Teile weitgehend selbsterklaerend, bis auf 'ParameterArt':
   Ist ParameterArt == 0, so wird das folgende Argument der Aufrufzeile als Wert dieser
                          Option erwartet und als String nach eingestellterWert geschrieben ('frei');
   ist ParameterArt == 1, so wird lediglich das Vorkommen des entsprechenden Parameters festgestellt
                          (der als default eingetragene Wert wird invertiert) ('einstellig');
                          Moeglich ist hier, ueber AN/AUS bzw. an/aus unabhaengig vom Default-Wert den Schalter zu setzen.
   ist ParameterArt == 2, so wird das folgende Argument in der Werte-Liste gesucht und sein Index nach
                          eingestellterWert geschrieben ('zweistellig').
   Ein gewuenschter Default-Wert wird einfach als eingestellter Wert in der Initialisierung des Arrays (s.u.)
   angegeben, dies ist dann der Wert, der verwendet wird, wenn diese Option beim Programm-Aufruf nicht
   explizit angegeben wurde.
*/

typedef enum {
   Arbeitsmodi, 
   KPStrategien, 
   NormalformStrategien,
   Zielbehandlung,
   Interreduktion,
   TechnischeDetails,
   ZusatzOptionen
} ParameterGruppenT;

typedef struct AufrufOptionTS{
   const char *Kuerzel;                           /* z.B. '-m' */
   const char *Name;                              /* z.B. 'Modus' */
   const char *Bedeutung;                         /* z.B. 'Einstellen des internen Modus' */
   ParameterGruppenT Gruppe;                /* Thematische Zuordnung */
   BOOLEAN aktiviert;                       /* TRUE = hat schon eine Auswirkung (ist FALSE zB. bei interner Modus). 
                                               Wirkung von FALSE: Option wird weder bei der Optionen-Liste noch im Menue unterstuetzt,
                                                                  wird aber beim Aufruf ganz normal behandelt.*/
   int ParameterArt;                        /* z.B. 1 :es wird kein Wert erwartet bzw. nur AN/AUS */
   BOOLEAN NurVervollstaendigung;           /* TRUE bedeutet, dass die Option nur bei eingestelltem Modus completion o. proof */
                                            /* Beruecksichtigung findet */
   union{
      struct {                         /* Parameter-Art == 3 */
         char *DefaultInfoString;
         char *UserInfoString;              /* Ist NULL, wenn die Option nicht angegeben wurde */
      } InfoString;
      /* Bei InfoStrings ist das System-Verhalten wie folgt:
         der UserInfoString wird ausgewertet, fehlende Informationen 
         werden aus dem DefaultInfoString gezogen, wobei der UserInfoString natuerlich
         Prioritaet hat.
       */
      struct {			       /* Parameter-Art == 2 */
         int eingestellterWert;             /* z.B. 0 fuer Normal */
         int defaultWert;             /* z.B. 0 fuer Normal */
         int AnzahlVerschiedenerWerte;      /* z.B. 3 */
         char *WerteImKlartext[10];         /* z.B. 'Normal, KPSharing, MultipleNF'*/
      } zweistellig;
      struct {			       /* Parameter-Art == 1 */
         BOOLEAN eingestellterWert;         /* z.B. FALSE  */
         BOOLEAN defaultWert;         /* z.B. FALSE  */
         char *BedeutungF, *BedeutungT;     /* Bedeutung von TRUE == Bedeutung  */
      } einstellig;
      struct {			       /* Parameter-Art == 0 */
         BOOLEAN wurdeAngegeben;
         char *eingestellterWert;           /* z.B. 'braid4.prot' */
         char *defaultWert;                     /* z.B. 'braid4.prot' */
      } frei;
   } ParameterTyp;
} AufrufOptionT;

/* +++++++++++++ Daten aus der eingelesenen Spezifikation etc. +++++++*/

 
extern ordering        PA_Ordering;                                              /* Die von aussen eingestellte Ordnung */
char *PA_NFBildungIO(BOOLEAN r, BOOLEAN g, BOOLEAN s);

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Beim Programm-Aufruf eingestellbare Optionen ++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef enum{OptNrFailing, OptNrSUE, OptNrClassification,
	     OptNrReClassification, OptNrCGClassification, OptNrOrdnungen,
	     OptNrPriorityQueue, OptNrTermpairRepresentation, OptNrKapur,
             OptNrEinsStern, OptNrNusfu, OptNrKern, OptNrGolem, OptNrBack, 
             OptNrGZFB, OptNrFork, OptNrInterRedRechtsNachP, OptNrWaisenmord,
             OptNrKPgen, OptNrKPsel, OptNrKPInterreduktion,
	     OptNrCGInterreduktion,
             OptNrNFBildung, OptNrZielbehandlung,
             OptNrAusgabeLevel, OptNrPCLProt, OptNrZeitlimit, OptNrRealZeitlimit,
             OptNrMemlimit, OptNrAutomodus,
             OptNrSymbnorm, OptNrDatabase,
             OptNrDummy,
             OptNrLAST} /* Diese Option muss immer letzte sein */
AufrufOptionenTyp;

extern AufrufOptionT AufrufOptionen[OptNrLAST]; /* wird im .c-File initialisiert */

/*-------------------------------------------------------*/
/*------ Arbeitsmodi-------------------------------------*/
/*-------------------------------------------------------*/

extern mode PA_ExtModus;

extern BOOLEAN PA_InteraktionAn,
               PA_FailingCompletion,
               PA_EinsSternAn,
               PA_NusfuAn,
               PA_KernAn,
               PA_GolemAn,
               PA_BackAn,
               PA_Automodus,
               PA_ForkErlaubt;

/*-------------------------------------------------------*/
/*------ TechnischeDetails ------------------------------*/
/*-------------------------------------------------------*/

/* Zeitlimit */
#define PA_ZeitlimitGesetzt() (AufrufOptionen[OptNrZeitlimit].ParameterTyp.frei.wurdeAngegeben)
#define PA_Zeitlimit() (atoi(AufrufOptionen[OptNrZeitlimit].ParameterTyp.frei.eingestellterWert))

/* RealZeitlimit */
#define PA_RealZeitlimitGesetzt() (AufrufOptionen[OptNrRealZeitlimit].ParameterTyp.frei.wurdeAngegeben)
#define PA_RealZeitlimit() (atoi(AufrufOptionen[OptNrRealZeitlimit].ParameterTyp.frei.eingestellterWert))

/* Memlimit */
#define PA_MemlimitGesetzt() (AufrufOptionen[OptNrMemlimit].ParameterTyp.frei.wurdeAngegeben)
#define PA_Memlimit() (atoi(AufrufOptionen[OptNrMemlimit].ParameterTyp.frei.eingestellterWert))


/*-------------------------------------------------------*/
/*------ ZusatzOptionen ---------------------------------*/
/*-------------------------------------------------------*/
#define PA_InteraktionAn() (PA_InteraktionAn)
#define PA_InteraktionAusschalten() (PA_InteraktionAn = FALSE)

#define PA_Datenbank() (AufrufOptionen[OptNrDatabase].ParameterTyp.frei.wurdeAngegeben? \
                        AufrufOptionen[OptNrDatabase].ParameterTyp.frei.eingestellterWert:NULL)

/* ------------------------------------------------*/
/* -------- Einstellen der Aufruf-Optionen --------*/
/* ------------------------------------------------*/
/* Die Funktionen stehen in der Reihenfolge, in der sie aufzurufen sind. */

void PA_InitApriori(void);
/* Parameter besetzen */


BOOLEAN PA_Vorbehandeln(int argc, char *argv[]);
/* - Reaktion auf Hilfe-Anforderung
   - Einstellung der Pseudo-Parameter, soweit moeglich
   Rueckgabe ist FALSE, wenn das Vorbehandlen bereits einen Abbruchgrund lieferte.
   Die vernutzten Argumente werden auf NULL gesetzt.
*/

BOOLEAN PA_AufrufParameterEinstellen(int argc, char *argv[], mode Modus);
/* Einstellen der beim Aufruf mitgegebenen Parameter. argc und argv sind die Parameter von main,
   es wird also davon ausgegangen, dass argv[0] den Programmnamen enthaelt, und das argv[argc] der
   Name der zu bearbeitenden Spezifikation ist. Die Angaben zwischen argv[1] und argv[argc-1] werden
   ausgewertet. Dabei werden die in Parameter.c spezifizierten Aufruf-Optionen herangezogen.
   Inkorrekte Parameter sowie der Filename werden ueberlesen.
   Auf Anforderung von Hilfe sowie Menue wird reagiert.
   Rueckgabe ist, ob - ggf. mit Menue-Hilfe - eine korrekte Einstellung vorgenommen werden konnte.
   Vor.: Die Spezifikation befindet sich bereits im Speicher.
*/

BOOLEAN PA_InitAposterioriEarly (void);
/* Aufruf vor PM_SpezifikationAnalysieren() */
   
void PA_InitAposteriori(void);
/* Daten aus der eingelesenen Spezifikation auswerten.
   Anschliessend die betroffenen Module mit den eingelesenen Parametern
   aposteriori initialisieren.
*/

/* -----------------------------------------------------*/
/* -------- Eingestellte Modi auf IO_Strom ausgeben ----*/
/* -----------------------------------------------------*/

void PA_DruckeModi(void);
/* Gibt die eingestellten Modi auf den IO_Strom aus.
   Verwendung bei Ausgabe der Statistik.*/

void PA_DefaultsAusgeben(char *Filename);

#endif
