/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Puffer.h                                                                                                 */
/* - **********                                                                                                 */
/* -                                                                                                            */
/* - Realisierung von Puffern fuer Objekte, die kurzfristig zwischengespeichert werden muessen, da ihre weitere */
/* - Behandlung noch nicht vorgenommen werden kann. Der Aufwand an Platz und Zeit soll minimal sein.            */
/* -                                                                                                            */
/* - REPuffer : Hier koennen Regeln und Gleichungen zwischengelagert werden, die der Interreduktion zum Opfer   */
/* -            gefallen sind.                                                                                  */
/* -            ACHTUNG : Schreib- und Lesezugriffe sind nicht verschraenkbar (zuerst nur Schreiben, dann nur   */
/* -                      Lesen). Vor Neu-Beschreiben ist Puffer neu zu initialisieren.                         */
/* - KPPuffer : Hier werden waehrend der Unifikation die entstandenen Term-Paare abgelegt, bis die Unifikation  */
/* -            abgeschlossen ist und die Terme weiterbehandelt werden koennen.                                 */
/* -                                                                                                            */
/* -                                                                                                            */
/* - 16.06. Arnim. Angelegt.                                                                                    */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef PUFFER_H
#define PUFFER_H

#include "general.h"
#include "RUndEVerwaltung.h"	/* wg. RegelOderGleichungsT */


void PU_InitAposteriori(void);
/* In den Modi completion und proof werden beide Puffer eingerichtet, in den Modi convergence und confluence
   nur der KP-Puffer, in den uebrigen Modi keiner der beiden Puffer.*/

/* ------------------------------------------ Funktionen RE-Puffer ---------------------------------------------*/

typedef void (*REPufferAnwendungsprozT) (RegelOderGleichungsT);

void PU_REPufferNeu(void);
/* Bereitet Puffer auf neue Eingaben vor (Initialisierung).*/

void PU_REObjektEinfuegen(RegelOderGleichungsT RoderE);
/* Fuegt ein neues Objekt in den Puffer ein.*/

void PU_REPufferDurchlauf(REPufferAnwendungsprozT AP);
/* Wendet AP auf alle Eintraege des Puffers an und leert den Puffer. */

void PU_REPufferDurchlauf2(REPufferAnwendungsprozT AP);
/* Wendet AP auf alle Eintraege des Puffers an, ohne den Puffer zu leeren.*/

void PU_REMerkerSetzen(void);
/* Merker auf aktuelle Puffer-Position setzen, um spaeter alle folgenden Eintraege behandeln zu koennen.*/

void PU_REPufferDurchlaufAbMerker(REPufferAnwendungsprozT AP);
/* Wendet AP auf alle Eintraege des Puffers an, die nach dem Setzen des Merkers eingefuegt wurden. */

BOOLEAN PU_ObjektNochNichtInREPuffer(RegelOderGleichungsT RoderE);
/* Liefert TRUE, wenn das angegebene Objekt noch nicht im REPuffer gespeichert ist. Test auf ==.*/

/* ------------------------------------------ Funktionen TP-Puffer ---------------------------------------------*/

void PU_TPPufferNeu(void);
/* Bereitet Puffer auf neue Eingaben vor (Initialisierung).*/

void PU_TPEinfuegen(TermT links, TermT rechts, unsigned long int Nummer);
/* Fuegt ein Termpaar je nach Modus aus unterschiedlicher Quelle nebst Nummer und Elterninformationen
   in den Puffer ein.*/

void PU_TPAuslesen(TermT *links, TermT *rechts, unsigned long int *nr);
/* Liefert ein Termpaar, das im Puffer steht. Vor.: Puffer nicht leer! */

BOOLEAN PU_TPPufferNichtLeer(void);
/* Liefert TRUE, wenn der TP-Puffer noch nicht leer ist.*/

void PU_TPPufferLoeschen(void);
/* Loescht alle Terme, die sich noch im TP-Puffer befinden. */

#endif
