/*=================================================================================================================
  =                                                                                                               =
  =  RUndEVerwaltung.h                                                                                            =
  =                                                                                                               =
  =  Verwaltung der Regel- und der Gleichungsmenge                                                                =
  =                                                                                                               =
  =                                                                                                               =
  =  22.03.94  Thomas.                                                                                            =
  =  29.06.94  Arnim. Operationen III. eingebaut und teilweise bestehende Operationen geaendert:                  =
  =            RE_???EinfuegenMitRueckgabe eingebaut, RE_???Entfernen geaendert, RE_Loesche???menge geaendert     =
  =            RE_IstRegel, RE_DatenHolenObjektLoeschen eingebaut.                                                =
  =                                                                                                               =
  =                                                                                                               =
  =================================================================================================================*/

#ifndef RUNDEVERWALTUNG_H
#define RUNDEVERWALTUNG_H


#include "DSBaumKnoten.h"
#include "general.h"
#include "InfoStrings.h"
#include "TermOperationen.h"
#include "Termpaare.h"

/*=================================================================================================================*/
/*= I. Bedingter Testbetrieb und Testfunktion                                                                     =*/
/*=================================================================================================================*/

#define RE_TEST 0

/*       Testfunktionen   Konsistenztest     Konsistenztest
          im Kompilat?     nach Beweis?    nach Indexaenderung?
   0 --       NEIN            NEIN               NEIN
   1 --        JA              JA                NEIN
   2 --        JA              JA                 JA
*/

void RE_RUndETest(void);                                              /* Interne Konsistenzpruefung beider Mengen. */


/*=================================================================================================================*/
/*= II. Regelmenge                                                                                                =*/
/*=================================================================================================================*/

/* 0. Typen und Objekte ------------------------------------------------------------------------------------------ */

typedef GNTermpaarT RegelOderGleichungsT;

typedef GNTermpaarT RegelT;

extern DSBaumT RE_Regelbaum;

typedef void * TermpairT;


typedef struct {
  TermT lhs;
  TermT rhs;
  PCL_ObjektNummerT pclId;
  RegelOderGleichungsT actualParent;
  RegelOderGleichungsT otherParent;
  TermpairT tp;
  VergleichsT result;
  BOOLEAN wasFIFO;
} SelectRecT;


/* 1. Erzeugen von Regeln aus Termpaaren ------------------------------------------------------------------------- */

RegelT RE_ErzeugteRegel(SelectRecT selRec);

/* 2. Praedikate und nebenwirkungsfreie Funktionen --------------------------------------------------------------- */


#define /*BOOLEAN*/ RE_RegelmengeLeer(/*void*/)                                                                     \
  (!RE_Regelbaum.Wurzel)

#define /*unsigned int*/ RE_maxTiefeRegel(/*void*/)                                                                 \
  (RE_Regelbaum.maxTiefeLinks)




/* 4. Entfernungs- und Loeschoperationen ------------------------------------------------------------------------- */
                                                     /* Das Loeschen der beteiligten Terme muss explizit erfolgen! */

void RE_RegelEntfernen(RegelT Regel); /* aus Baum entfernen, weder wegwerfen noch als geloescht markieren! (->KPV) */

void RE_RegelLoeschen(RegelT Regel);                   /* Regel als Objekt loeschen sowie Meldung an KP-Verwaltung
                                                          Terme nicht loeschen*/
void RE_RegelVerfluechtigen(RegelT Regel);
void RE_RegelmengeLoeschen(void);


/* 5. Ausgabeoperation ------------------------------------------------------------------------------------------- */

void RE_AusgabeRegelmenge(int Grad);
/* Je nach Einstellung auch graphische Ausgabe:
   0 - Textuell, zum Einlesen als Spezifikation
   1 - Textuell
   2 - Baum
   3 - Baum mit Grafik
   4 - Textuell ausfuehrlich (hier wie 1)
   5 - Baumdump
   6 - Baumdump mit Grafik
   7 - wie 1, aber mit Ueberschrift "Completed system: Set of rules"
*/

/* 6. Schleifenkonstrukte ---------------------------------------------------------------------------------------- */

#define RE_forRegelnRobust(/*RegelT*/ RegelIndex, /*Block*/ Action)                                                 \
  BK_forRegelnRobust(RegelIndex, RE_Regelbaum, Action)

#define /*RegelT*/ RE_ErsteRegel(/*void*/)                                             /* Zaehlen sukzessive alle */\
  BK_ErstesTermpaar(RE_Regelbaum)
#define /*RegelT*/ RE_NaechsteRegel(/*void*/)                                               /* Elemente von R auf */\
  BK_NaechstesTermpaar()

#define /*RegelT*/ RE_ErsteRegelM(/*KantenT **/ iBlatt, /*GNTermpaarT **/iTermpaar)    /* dito mit expliziten In- */\
  BK_ErstesTermpaarM(RE_Regelbaum,iBlatt,iTermpaar)                                     /* dizes fuer Schachtelung */
#define /*RegelT*/ RE_NaechsteRegelM(/*KantenT **/ iBlatt, /*GNTermpaarT **/iTermpaar)                              \
  BK_NaechstesTermpaarM(iBlatt,iTermpaar)

  
/* 8. Umbenennungen ---------------------------------------------------------------------------------------------- */

void RE_RegelUmbenennen(RegelT Regel, signed int Delta);      /* Jede Variable x(i) wird ersetzt durch x(i+Delta). */

SymbolT RE_neuesteVariable(RegelT Regel);

#define /*void*/ RE_RegelExternalisieren(/*RegelT*/ Regel)                                                          \
  RE_RegelUmbenennen(Regel,SO_VarNummer(BO_NeuesteBaumvariable()))
/* Ist Regel ein Element des Regelbaumes, so nimmt die Funktion eine Umbenennung der Variablen vor, 
   so dass die resultierenden Variablen nicht im Baum vorkommen. 
   Kann auch fuer einzelne Richtungen von Gleichungen verwendet werden. */

#define /*void*/ RE_RegelReinternalisieren(/*RegelT*/ Regel)                                                        \
  RE_RegelUmbenennen(Regel,-SO_VarNummer(BO_NeuesteBaumvariable()))
/* Stellt nach dem Aufruf von RE_RegelExternalisieren den status quo ante wieder her, allerdings nur
   solange, wie keine Regel oder Gleichung mit neuer maximaler Variablenanzahl eingefuegt worden ist. */



/*=================================================================================================================*/
/*= III. Gleichungsmenge                                                                                          =*/
/*=================================================================================================================*/

/* 0. Typen und Objekte ------------------------------------------------------------------------------------------ */

extern DSBaumT RE_Gleichungsbaum;
typedef GNTermpaarT GleichungsT;


/* 1. Erzeugen von Gleichungen aus Termpaaren -------------------------------------------------------------------- */

GleichungsT RE_ErzeugteGleichung(SelectRecT selRec);

/* 2. Praedikate und nebenwirkungsfreie Funktionen --------------------------------------------------------------- */

#define /*BOOLEAN*/ RE_GleichungsmengeLeer(/*void*/)                                                                \
  (!RE_Gleichungsbaum.Wurzel)

#define /*BOOLEAN*/ RE_GleichungsmengeNichtLeer(/*void*/)                                                           \
  (RE_Gleichungsbaum.Wurzel)

#define /*unsigned int*/ RE_maxTiefeGleichung(/*void*/)                                                             \
  (RE_Gleichungsbaum.maxTiefeLinks)


/* 4. Entfernungs- und Loeschoperationen ------------------------------------------------------------------------- */
                                                     /* Das Loeschen der beteiligten Terme muss explizit erfolgen! */

void RE_GleichungEntfernen(GleichungsT Gleichung);
/* Nur eine Richtung der Gleichung aus dem Baum entfernen. 
   ACHTUNG: Gegenrichtung muss entfernt sein, bevor erneut mit Gleichungsbaum reduziert wird. */

void RE_GleichungLoeschen(GleichungsT Gleichung);  /* Gleichung als Objekt loeschen sowie Meldung an KP-Verwaltung.
       Subgleichungen loeschen & Meldung an NFBildung, falls es sich um ausgezeichnete Seite der Gleichung handelt.*/

void RE_GleichungVerfluechtigen(GleichungsT Gleichung);
void RE_GleichungsmengeLoeschen(void);


/* 5. Ausgabeoperationen ----------------------------------------------------------------------------------------- */

void RE_AusgabeGleichungsmenge(int Grad);
/* Je nach Einstellung auch graphische Ausgabe:
   0 - Textuell, zum Einlesen als Spezifikation
   1 - Textuell
   2 - Baum
   3 - Baum mit Grafik
   4 - Textuell ausfuehrlich
   5 - Baumdump
   6 - Baumdump mit Grafik
   7 - wie 1, aber mit Ueberschrift "Completed system: Set of equations"
*/


/* 6. Schleifenkonstrukte ---------------------------------------------------------------------------------------- */

#define RE_forGleichungenRobust(/*GleichungsT*/ GleichungsIndex, /*Block*/ Action)      \
   BK_forRegelnRobust(GleichungsIndex, RE_Gleichungsbaum, Action)

#define RE_forGMReferenzen(/*GNTermpaarT*/ RegelIndex, /*Boolexpr*/ Test, /*Block*/ Action)     \
   BK_ReferenzDurchlauf(RegelIndex, RE_Gleichungsbaum, Test, Action)

GleichungsT RE_ErsteGleichung(void);                                       /* Zaehlen sukzessive alle Elemente von */
GleichungsT RE_NaechsteGleichung(void);                                   /* E in der ausgezeichneten Richtung auf */
GleichungsT RE_ErsteGleichungM(KantenT *iBlatt, GNTermpaarT *iTermpaar);                /* dito mit expliziten In- */
GleichungsT RE_NaechsteGleichungM(KantenT *iBlatt, GNTermpaarT *iTermpaar);             /* dizes fuer Schachtelung */

RegelOderGleichungsT RE_ErstesFaktum(void);                     /* analog fuer Durchlauf von R und anschliessend E */
RegelOderGleichungsT RE_NaechstesFaktum(void);
RegelOderGleichungsT RE_ErstesFaktumM(KantenT *iBlatt, GNTermpaarT *iTermpaar, BOOLEAN *SchonInGleichungen);
RegelOderGleichungsT RE_NaechstesFaktumM(KantenT *iBlatt, GNTermpaarT *iTermpaar, BOOLEAN *SchonInGleichungen);



/*=================================================================================================================*/
/*= IV. Ergebnisausgabe R-, E- und KP-Menge                                                                       =*/
/*=================================================================================================================*/

void RE_SonstigeErgebnisse(BOOLEAN flag);
/* flag dient zur Detailsteuerung in den verschiedenen Modi:
   convergence: TRUE  = im TP-Puffer befinden sich divergente KPs
                FALSE = im TP-Puffer befinden sich unrichtbare Ausgangsgleichungen
*/

void RE_DruckeEndsystemFuerPCL(void);



/*=================================================================================================================*/
/*= V. Regeln und Gleichungen als Eltern von KPs                                                                  =*/
/*=================================================================================================================*/

/* Prozeduren, die zum Loesen, Weiterbehandeln und Loeschen von Regeln und Gleichungen waehrend der Interreduktion
   benoetigt werden, und zwar unabhaengig davon, ob es sich um Regeln oder Gleichungen handelt. Deshalb einheit-
   liche Sichtweise auf 'Objekte'.                                                                                 */

#define /*BOOLEAN*/ RE_IstRegel(/*RegelOderGleichungsT*/ Faktum) /* Liefert TRUE, wenn das angegebene Objekt eine */\
  TP_TermpaarIstRegel(Faktum)                                                                         /* Regel ist */

BOOLEAN RE_IstUngleichung(RegelOderGleichungsT Faktum);

void RE_UngleichungEntmanteln(UTermT* s, UTermT* t, RegelOderGleichungsT Faktum);

BOOLEAN RE_UngleichungEntmantelt(UTermT* s, UTermT* t, RegelOderGleichungsT Faktum);



/*=================================================================================================================*/
/*= VI. Initialisierungen                                                                                         =*/
/*=================================================================================================================*/

void RE_InitApriori(void);

/*=================================================================================================================*/
/*= XXX. Sonstiges                                                                                                =*/
/*=================================================================================================================*/

void RE_FaktumEinfuegen(RegelOderGleichungsT Faktum);

#endif
