/*=================================================================================================================
  =                                                                                                               =
  =  Termpaare.h                                                                                                  =
  =                                                                                                               =
  =  Realisierung von Termtupeln                                                                                  =
  =                                                                                                               =
  =                                                                                                               =
  =  22.03.95 Thomas.                                                                                             =
  =                                                                                                               =
  =================================================================================================================*/

#ifndef TERMPAARE_H
#define TERMPAARE_H


#include "general.h"
#include "SpeicherVerwaltung.h"
#include "TermOperationen.h"
#include <limits.h>  /* fuer LONG_MAX */

/*=================================================================================================================*/
/*= I. Typen                                                                                                      =*/
/*=================================================================================================================*/

typedef enum {weiss=0, blau=1, gelb=2, gruen=3} FarbT;                      /* Faerbung fuer Interreduktion der GM */

typedef struct GNTermpaarTS {
  TermT               linkeSeite,
                      rechteSeite,
                      r2;
  long int            Nummer;
  void               *SUEInfo; /* XXX Argl: wg ChaseManhatten !!! */
  BOOLEAN             lebtNoch;
  unsigned int        generation; /* Axiome: generation = 1; ansonsten: gen(Kind) = 1 + max(gen(Eltern)) */

  PCL_ObjektNummerT   pclNr, pclNr2;

  struct KnotenTS     *Blatt;
  BOOLEAN             RichtungAusgezeichnet, 
                      IstMonogleichung;
  FarbT               Faerbung;
  struct GNTermpaarTS *Gegenrichtung,
                      *Vorg,
                      *Nachf;
  BOOLEAN             DarfNichtReduzieren,
                      Grundzusammenfuehrbar;
  GZ_statusT          gzfbStatus;
  TermT               GescheitertL, GescheitertR;
#ifdef CCE
  TermT               CCE_lhs, CCE_rhs;
#endif
} *GNTermpaarT;



/*=================================================================================================================*/
/*= II. Speicherverwaltung                                                                                        =*/
/*=================================================================================================================*/

extern SV_SpeicherManagerT TP_FreispeicherManager;

#define /*void*/ TP_GNTermpaarHolen(/*L-Ausdruck*/ Index)                                                           \
  SV_Alloc(Index,TP_FreispeicherManager,GNTermpaarT)

#define /*void*/ TP_GNTermpaarLoeschen(/*GNTermpaarT*/ Termpaar)                                                    \
  SV_Dealloc(TP_FreispeicherManager,(Termpaar))

#define /*void*/ TP_LinksRechtsLoeschen(/*GNTermpaarT*/ Termpaar)                                                   \
  TO_TermeLoeschen(TP_LinkeSeite(Termpaar),TP_RechteSeite(Termpaar))



/*=================================================================================================================*/
/*= III. Lesender Zugriff                                                                                         =*/
/*=================================================================================================================*/

#define /*BOOLEAN*/ TP_RichtungAusgezeichnet(/*GNTermpaarT*/ Termpaar)                                              \
  ((Termpaar)->RichtungAusgezeichnet)

#define /*BOOLEAN*/ TP_RichtungUnausgezeichnet(/*GNTermpaarT*/ Termpaar)                                            \
  (!(Termpaar)->RichtungAusgezeichnet)

#define /*BOOLEAN*/ TP_IstMonogleichung(/*GNTermpaarT*/ Termpaar)                                                   \
  ((Termpaar)->IstMonogleichung)

#define /*BOOLEAN*/ TP_IstKeineMonogleichung(/*GNTermpaarT*/ Termpaar)                                              \
  (!(Termpaar)->IstMonogleichung)

#define /*FarbT*/ TP_IstWeiss(/*GNTermpaarT*/ Termpaar)                                                             \
  ((Termpaar)->Faerbung == weiss)

#define /*FarbT*/ TP_IstBlau(/*GNTermpaarT*/ Termpaar)                                                              \
  ((Termpaar)->Faerbung == blau)

#define /*FarbT*/ TP_IstGelb(/*GNTermpaarT*/ Termpaar)                                                              \
  ((Termpaar)->Faerbung == gelb)

#define /*FarbT*/ TP_IstGruen(/*GNTermpaarT*/ Termpaar)                                                             \
  ((Termpaar)->Faerbung == gruen)

#define /*GNTermpaarT*/ TP_Gegenrichtung(/*GNTermpaarT*/ Termpaar)                                                  \
  ((Termpaar)->Gegenrichtung)

#define /*TermT*/ TP_LinkeSeite(/*GNTermpaarT*/ Termpaar)                                                           \
  ((Termpaar)->linkeSeite)

#define /*TermT*/ TP_RechteSeite(/*GNTermpaarT*/ Termpaar)                                                          \
  ((Termpaar)->rechteSeite)

#define /*void**/ TP_SUEInfo(/*GNTermpaarT*/ Termpaar)                                                            \
  ((Termpaar)->SUEInfo)

#define /*unsigned*/ TP_ElternNr(/*GNTermpaarT*/ Termpaar)                                                          \
  ((Termpaar)->Nummer)

#define /*PCL_ObjektNummerT**/ TP_PclID(/*GNTermpaarT*/ Termpaar)                                                   \
  (&((Termpaar)->pclNr))

#define /*PCL_ObjektNummerT**/ TP_PclID2(/*GNTermpaarT*/ Termpaar)                                                   \
  (TP_FreieVariablen(Termpaar)?(&((Termpaar)->pclNr2)):(&((Termpaar)->pclNr)))

#define /*PCL_ObjektNummerT*/ TP_PclIDAs(/*GNTermpaarT*/ Termpaar)                                                   \
  ((Termpaar)->pclNr)

#define /*PCL_ObjektNummerT*/ TP_PclID2As(/*GNTermpaarT*/ Termpaar)                                                   \
  (TP_FreieVariablen(Termpaar)?((Termpaar)->pclNr2):((Termpaar)->pclNr))

#define /*GNTermpaarT*/ TP_Vorg(/*GNTermpaarT*/ Termpaar)                                                           \
  ((Termpaar)->Vorg)

#define /*GNTermpaarT*/ TP_Nachf(/*GNTermpaarT*/ Termpaar)                                                          \
  ((Termpaar)->Nachf)

#define /* unsgigned int*/ TP_GetGeneration(/*GNTermpaarT*/ Termpaar)                                               \
  ((Termpaar)->generation)

/*=================================================================================================================*/
/*= IV. Schreibender Zugriff                                                                                      =*/
/*=================================================================================================================*/

#define /*void*/ TP_GegenrichtungBesetzen(/*GNTermPaarT*/ Termpaar, /*GNTermPaarT*/ gegenrichtung)                  \
  (Termpaar)->Gegenrichtung = gegenrichtung

#define /*GNTermpaarT*/ TP_VorgSei(/*GNTermpaarT*/ Termpaar, /*GNTermpaarT*/ Vorgaenger)                            \
  ((Termpaar)->Vorg = Vorgaenger)

#define /*GNTermpaarT*/ TP_NachfSei(/*GNTermpaarT*/ Termpaar, /*GNTermpaarT*/ Nachfolger)                           \
  ((Termpaar)->Nachf = Nachfolger)

#define /*FarbT*/ TP_WeissMalen(/*GNTermpaarT*/ Termpaar)                                                           \
  ((Termpaar)->Faerbung = weiss)

#define /*FarbT*/ TP_BlauMalen(/*GNTermpaarT*/ Termpaar)                                                            \
  ((Termpaar)->Faerbung += blau)

#define /*FarbT*/ TP_GelbMalen(/*GNTermpaarT*/ Termpaar)                                                            \
  ((Termpaar)->Faerbung += gelb)


/*=================================================================================================================*/
/*= V. Abgeleitete Praedikate und Funktionen                                                                      =*/
/*=================================================================================================================*/

#define /*BOOLEAN*/ TP_TermpaarIstRegel(/*GNTermpaarT*/ Termpaar)                                                   \
  !((Termpaar)->Gegenrichtung)

#define /*BOOLEAN*/ TP_TermpaareGleich(/*GNTermpaarT*/ Termpaar1, /*GNTermpaarT*/ Termpaar2)                        \
  (TO_TermeGleich(TP_LinkeSeite(Termpaar1),TP_LinkeSeite(Termpaar2))                                                \
  && TO_TermeGleich(TP_RechteSeite(Termpaar1),TP_RechteSeite(Termpaar2)))

#define /*BOOLEAN*/ TP_FreieVariablen(/*GNTermpaarT*/ Termpaar)                                                     \
  ((Termpaar)->r2)

#define /*BOOLEAN*/ TP_RechteSeiteUnfrei(/*GNTermpaarT*/ Termpaar)                                                  \
  (TP_FreieVariablen(Termpaar) ? (Termpaar)->r2 : TP_RechteSeite(Termpaar))



/*=================================================================================================================*/
/*= VI. Initialisierung                                                                                           =*/
/*=================================================================================================================*/

void TP_InitApriori(void);

#endif /* Termpaare_H */
