/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - pcl.h                                                                                                      */
/* - *****                                                                                                      */
/* -                                                                                                            */
/* - Enthaelt alle fuer den PCL-Modus notwendigen Routinen, Zaehler etc..                                       */
/* -                                                                                                            */
/* - 30.11.94 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Diese Datei enthaelt die verschiedenen Ausgaberoutinen, die zum Erstellen eines PCL-Protokolls noetig sind.*/
/* - Wird beim Aufruf von WaldmeisterII bei der Option -pcl ein Dateiname angegeben, so wird diese Datei in     */
/* - PCL_InitAposteriori geoeffnet. Am Programmende wird diese Datei dann mit PCL_Beenden wieder geschlossen.   */
/* - Beim Aufruf der Ausgaberoutinen ist jeweils ein VAR-Parameter nr anzugeben. Dieser wird dann auf die als   */
/* - naechstes zu vergebende PCL-ID gesetzt, ist also nicht weiter anzupassen.                                  */

/* - Zu Beachten ist noch, dass Zuweisungen von PCL-IDs (zB von einer Datenstruktur in die naechste) nur ueber  */
/* - das zu diesem Zweck vorgesehene Makro ablaufen, da sonst im 'rauskompilierten Falle' der Compiler die      */
/* - Komponenten gar nicht kennt, dieses Makro aber ebenso durch 'nix tun' ersetzt wird und so keine Probleme   */
/* - bereiten duerfte.                                                                                          */

/* - Der Einbau der PCL-IDs in die einzelnen Datenstrukturen wurde wie folgt realisiert:                        */
/* - - es wurde eine Komponente PclNr eingefuegt                                                                */
/* - - diese Komponente kann mit bedingter Compilierung (COMP_PCL=0) wieder entfernt werden                     */
/* - - ferner wird eine Zugriffsfunktion angeboten, die die Adresse (!) der Komponente PclNr liefert, um sie    */
/* -   in pcl.c manipulieren zu koennen. Diese Zugriffsfunktion wird beim rauuscompilieren durch ein 0-Makro    */
/* -   ersetzt.                                                                                                 */
/* -                                                                                                            */
/* - Einige Erweiterungen fuer integrated PCL                                                                   */
/* - Bernd, 15.05.97                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef PCL_H
#define PCL_H

#include <limits.h>  /* fuer USHRT_MAX */

#include "TermOperationen.h"
#include "general.h"
#include "RUndEVerwaltung.h"
#include "InfoStrings.h"


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Typ-Definitionen ----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

typedef unsigned short StellenT;
#define MaxStelle USHRT_MAX
/* oder: */

/* Falls ein Ueberlauf von StellenT auftritt:
typedef unsigned long StellenT;
#define MaxStelle ULONG_MAX
*/


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Initialisierung/Abschluss des Protokolls ----------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PCL_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

void PCL_InitAposteriori(InfoStringT DEF, InfoStringT USR);

BOOLEAN PCL_EsWirdProtokolliert(void);

void PCL_Beenden(void);


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Die PCL-Ausgabe-Prozeduren /-Makros ---------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern struct MeldesatzS {

  /* --------------------------- Gleichung/Ziel aus der Spezifikation --------------------------------------------*/

  void (*GleichungAusDerSpezifikation) (TermT links, TermT rechts, PCL_ObjektNummerT *nrPtr);
  /* Die Gleichungen aus der Spezifikation werden beim Einfuegen in die KP-Menge mit Gewicht 0
     als 'initial' ins PCL-Protokoll geschrieben. Die Nummer wird mit einer neuen PCL-ID besetzt.
     <pcl-id> : tes-eqn : links = rechts : initial  */

  void (*ZielAusDerSpezifikation) (TermT links, TermT rechts, PCL_ObjektNummerT *nrPtr);
  /* Einlesen der Ziele aus der Spezifikation:
     Die Ziele werden bei der Aufnahme in die interne Datenstruktur als tes-goal gemeldet.
     Die Nummer wird mit einer neuen PCL-ID besetzt.
     <pcl-id> : tes-goal : links = rechts : hypothesis */

  /* --------------------------------- Reduktion -----------------------------------------------------------------*/

  void (*Reduktionsbeginn) (PCL_ObjektT objektArt, PCL_ObjektNummerT *nrPtr, PCL_SeitenT seite, TermT andereSeite);
  /* Zu Beginn der Normalformbildung wird das zu behandelnde Objekt angemeldet.
     Die nr wird bei jedem Reduktionsschrittauf die neue PCL-ID gesetzt. */

  void (*Reduktion) (TermT ergebnis, UTermT betroffenerTeilterm, 
                     PCL_ObjektNummerT *AngewandtesFaktIDPtr, PCL_SeitenT angewandteSeite);
  /* Jeder Reduktionsschritt wird mit betroffenem <fact>, angewandter <rule>/<eqn>, betroffener Seite und Stelle
     sowie angewandter Seite ins PPP geschrieben. Aus dem betroffenen Teilterm werden Tiefe sowie Stelle berechnet
     (spart deren Berechnung bei per Option abgeschaltetem PCL-Protokoll).
     <pcl-id> : <step-type> : links = rechts : tes-red( <pcl-expr>, <place>, <pcl-expr>, <side>)
     ^Nummer   ^Art des entst.  ^Ergebnis              ^Nr. des red. ^Pos.     ^Nr der  ^Richtung der Anwendung
                 Faktums          der Red.               Faktums       der Red. angewandten Regel/Gleichung */

  /* --------------------------- Unifikation ---------------------------------------------------------------------*/

  void (*Unifikation) (PCL_ObjektNummerT *nrPtr, TermT links, TermT rechts, RegelOderGleichungsT Vater, 
                       TermT betroffenerVaterTerm, UTermT betroffenerVaterTeilterm, RegelOderGleichungsT Mutter);
  /* Jede gelungene Unifikation wird als <cp-expr> ins PP geschrieben: beide Elternteile (<rule> und/oder <eqn>) 
     sowie genaue Beschreibung der Position. Die nr wird mit einer neuen PCL-ID besetzt. Dabei beschreibt Tiefe die
     Anzahl von Eintraegen im Stellen-Array (dh. Tiefe = 0 bedeutet Stelle lambda).
     <pcl-id> : tes-eqn : links=rechts : cp(Vater, <place>, Mutter, <side>) */

  void (*ReUnifikation) (PCL_ObjektNummerT *nrPtr, TermT links, TermT rechts, RegelOderGleichungsT Vater, 
                         int NummerBetroffeneVaterStelle, RegelOderGleichungsT Mutter);
  /* NummerBetroffeneVaterStelle ist die Nummer desjenigen Symbols in der linken Vater-Seite,
     das an der Stelle steht, wo die Ueberlappung stattfand.*/

  /* --------------------------- Aufnahme von KPs als Regeln oder Gleichungen ------------------------------------*/

  void (*KPZuRegel) (TermT links, TermT rechts, PCL_RichtungsRichtungT RichtungsRichtung, 
		     PCL_ObjektNummerT *RegelIDPtr, long int nr, BOOLEAN fifo);
  /* Die Orientierung von kritischen Paaren, also <eqn>'s, wird nur im Erfolgsfall ins PP gemeldet:
     <pcl-id> : tes-rule: links=rechts : orient(<pcl-id>,x|u)
     Die Pcl-ID der Regel enthaelt bereits diejenige des KPs und wird aktualisiert. */

  void (*KPZuGleichung) (TermT links, TermT rechts, PCL_ObjektNummerT *GleichungsIDPtr, long int nr, BOOLEAN fifo);
  void (*FreieVarInstanz) (TermT links, TermT rechts, PCL_ObjektNummerT *GleichungsIDPtr, BOOLEAN left);
  /* --------------------------- Ziele zusammenfuehren -----------------------------------------------------------*/

  void (*ZielBewiesen) (PCL_ObjektNummerT *bewiesenesZielIDPtr, TermT links, TermT rechts);
  /* Konnte ein Ziel bewiesen, i.e. zusammengefuehrt, werden, so wird dies mit 
     <pcl-id> : tes-final: links=rechts:<quote-expr> gemeldet,
     wobei die <quote-expr> gerade die vorhergehende PCL-ID des Ziels bezeichnet.
     *bewiesenesZiel enthaelt die PCL-ID des bewiesenen Zieles und wird dann auf die neue ID gesetzt. */

/* --------------------------- Abschluss der Vervollstaendigung: alle Fakten ausgeben --------------------------*/

  void (*VVAbschlussFaktDrucken) (PCL_ObjektNummerT *nrPtr, TermT links, TermT rechts, PCL_ObjektT art);
  /* Abschluss der Vervollstaendigung: Wurde ein (grund?, siehe PCL, S. 20) konvergentes Regelsystem erzielt, so
     sollen alle erzeugten Regeln und Gleichungen nochmal mit tes-final gemeldet werden.
     Dies soll hier geschehen. Mit jeder Regel und Gleichung ist diese Prozedur aufzurufen.
     Die neue PCL-ID wird an nr weitergegeben.*/

/* --------------------------------- Datenbank -----------------------------------------------------------------*/

  void (*DatenbankKonklusion) (RegelOderGleichungsT Konklusion);

} PCL_Proc;


#define PCL_GleichungAusDerSpezifikation PCL_Proc.GleichungAusDerSpezifikation
#define PCL_ZielAusDerSpezifikation PCL_Proc.ZielAusDerSpezifikation
#define PCL_Reduktionsbeginn PCL_Proc.Reduktionsbeginn
#define PCL_Reduktion PCL_Proc.Reduktion
#define PCL_Unifikation PCL_Proc.Unifikation
#define PCL_ReUnifikation PCL_Proc.ReUnifikation
#define PCL_KPZuRegel PCL_Proc.KPZuRegel
#define PCL_KPZuGleichung PCL_Proc.KPZuGleichung
#define PCL_FreieVarInstanz PCL_Proc.FreieVarInstanz
#define PCL_ZielBewiesen PCL_Proc.ZielBewiesen
#define PCL_VVAbschlussFaktDrucken PCL_Proc.VVAbschlussFaktDrucken
#define PCL_DatenbankKonklusion PCL_Proc.DatenbankKonklusion


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Chaining ------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern struct ChainsatzS {
  void (*StartChain)    (BOOLEAN);  /* Alle folgenden Schritte nur vorlaeufig (oder garnicht) protokollieren. */
  void (*SaveChain)     (void);     /* Vorlaeufig protokollierte Schritte speichern. */
  void (*DeleteChain)   (void);     /* Vorlaeufig protokollierte Schritte loeschen. */
} PCL_Chain;

#define PCL_StartChain    PCL_Chain.StartChain
#define PCL_SaveChain     PCL_Chain.SaveChain
#define PCL_DeleteChain   PCL_Chain.DeleteChain

void PCL_Off (void);
void PCL_On(void);

#endif  /* ifdef PCL_H */
