/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
//  File int_name_table.cpp
//
//  Implements the class IntNameTable for a table of names
//
//  made from NameTable, 13/05/2000, flight Manchester-Atlanta
//


#include <string.h>


#include "int_name_table.h"


const int IntNameTable::_noOfBuckets = 127;
IntNameTable* IntNameTable::vars = new IntNameTable;


IntNameTable::IntNameTable ()
  :
  _buckets ( new EntryList* [_noOfBuckets] ),
  _nextNumber ( 0 )
{
  if ( ! _buckets ) {
    NO_MEMORY;
  }

  // initialize all buckets to empty
  for ( int i = _noOfBuckets - 1; i >= 0; i-- ) 
    _buckets [i] = EntryList::empty ();
} // IntNameTable::IntNameTable


int IntNameTable::insert ( const char* str )
{
  int bucket = hash ( str );

  for ( EntryList* l = _buckets [ bucket ]; ! l->isEmpty (); l = l->tail() ) {
    if ( ! strcmp (str, l->head()->right()) ) { // found
      return l->head()->left();
    }
  }

  char* copy = new char [strlen (str) + 1];
  if (! copy) {
    NO_MEMORY;
  }

  strcpy ( copy, str );
  int result = _nextNumber;
  _nextNumber ++;

  Entry* newEntry = new Entry ( result, copy );

  EntryList::push ( newEntry, _buckets [bucket] );

  return result;
} // IntNameTable::insert


IntNameTable::~IntNameTable ()
{
  // destroy all buckets and strings in them
  for ( int i = _noOfBuckets - 1; i >= 0; i-- ) {
    // delete all strings
    for ( EntryList* lst = _buckets [i]; ! lst->isEmpty (); lst = lst->tail() ) {
      delete [] lst->head()->right();
      delete lst->head ();
    }

    
    // destroy the bucket itself
    _buckets [i] ->destroy ();
  }

  delete [] _buckets;
} // IntNameTable::~IntNameTable


// very simple 
int IntNameTable::hash ( const char* str )
{
  int result = 0;

  for ( const char* s = str ; *s != 0; s++ ) 
    result = ( result * 10 + *s ) % _noOfBuckets;
  
  return result;
} // IntNameTable::hash

