/* Take a stream of interpretations and remove the isomorphic ones.
 */

#include "../ladr/header.h"

#include "../ladr/clock.h"
#include "../ladr/ioutil.h"
#include "../ladr/interp.h"
#include "../ladr/commands.h"

#define PROGRAM_NAME    "isofilter"
#define PROGRAM_VERSION "2003-B"
#define PROGRAM_DATE    "Oct 2003"

static char Help_string[] =
"\nThis program reads a stream of interpretations (from stdin) and removes\n"
"isomorphic ones.  (There can be operator declarations before the\n"
"interpretations.)  An optional argument \"slow\" (mostly for debugging) says\n"
"to disable the optimizations during isomorphism checking.\n\n"
"Interpretations look like this:\n\n"
"interpretation(2, [\n"
"        function(A, [1]),\n"
"        function(e(_,_), [1,0,0,1]),\n"
"        relation(P(_), [0,1])]).\n\n";

static BOOL member_args(int argc, char **argv, char *str)
{
  int i;
  for (i = 1; i < argc; i++)
    if (str_ident(argv[i], str))
      return TRUE;
  return FALSE;
}  /* member_args */

int main(int argc, char **argv)
{
  Term t;
  Plist interps = NULL;
  int interps_read = 0;
  int interps_kept = 0;

  int perm[100], i;

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "-help")) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }

  declare_standard_parse_types();
  init_options();

  /* Note that we do not read any commands. */
  /* terms(interpretations) and end_of_list are optional. */

  t = read_term(stdin, stderr);
  if (is_term(t, "terms", 1))
    t = read_term(stdin, stderr);

  /* Build a permutation that swaps 0 and 1. */
  for (i = 0; i < 100; i++)
    perm[i] = i;
  perm[0] = 1; perm[1] = 0;

  while (t != NULL && !end_of_list_term(t)) {
    Interp a = compile_interp(t);
    Interp b = permute_interp(a, perm);
    fprint_interp_2(stdout, b);  /* We print the non-canonicalized interp. */
    zap_interp(a);
    zap_interp(b);
    t = read_term(stdin, stderr);
  }

  printf("%% isofilter: input=%d, kept=%d, ", interps_read, interps_kept);
  printf("checks=%lu, perms=%lu, ", iso_checks(), iso_perms());
  printf("%.2f seconds.\n", user_time() / 1000.);

  exit(0);
}  /* main */

