/* Take a stream of interpretations and remove the isomorphic ones.
 */

#include "../ladr/header.h"

#include "../ladr/clock.h"
#include "../ladr/ioutil.h"
#include "../ladr/interp.h"
#include "../ladr/commands.h"

#define PROGRAM_NAME    "isofilter"
#define PROGRAM_VERSION "2003-B"
#define PROGRAM_DATE    "Oct 2003"

static BOOL Slow = FALSE;  /* disable isomorphism optimizations */

static char Help_string[] =
"\nThis program reads a stream of interpretations in portable format\n"
"(from stdin) and takes a command-line argument saying how to print them:\n\n"
"    portable      : one line per operation\n"
"    portable2     : portable, with binary operations in a square\n"
"    tabular       : as nice tables\n"
"    tex           : formatted for LaTeX\n\n";

static BOOL member_args(int argc, char **argv, char *str)
{
  int i;
  for (i = 1; i < argc; i++)
    if (str_ident(argv[i], str))
      return TRUE;
  return FALSE;
}  /* member_args */

enum {PORTABLE, PORTABLE2, TABULAR, TEX};

int main(int argc, char **argv)
{
  Term t;
  int type;

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "-help")) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }
  else if (member_args(argc, argv, "portable"))
    type = PORTABLE;
  else if (member_args(argc, argv, "portable2"))
    type = PORTABLE2;
  else if (member_args(argc, argv, "tabular"))
    type = TABULAR;
  else if (member_args(argc, argv, "tex"))
    type = TEX;
  else {
    printf("%s", Help_string);
    exit(1);
  }

  declare_standard_parse_types();
  init_options();

  /* Note that we do not read any commands. */
  /* terms(interpretations) and end_of_list are optional. */

  t = read_term(stdin, stderr);
  if (is_term(t, "terms", 1))
    t = read_term(stdin, stderr);

  while (t != NULL && !end_of_list_term(t)) {
    Interp a = compile_interp(t);

    if (type == PORTABLE)
      fprint_interp(stdout, a);
    else if (type == PORTABLE2)
      fprint_interp_2(stdout, a);
    else if (type == TABULAR)
      fprint_interp_tabular(stdout, a);
    else if (type == TEX)
      fprint_interp_tex(stdout, a);

    zap_interp(a);
    zap_term(t);
    t = read_term(stdin, stderr);
  }
  
  exit(0);
}  /* main */

