#include "../ladr/header.h"

#include "../ladr/clock.h"
#include "../ladr/ioutil.h"
#include "../ladr/interp.h"
#include "../ladr/commands.h"

#define PROGRAM_NAME    "interpfilter"
#define PROGRAM_VERSION "2003-B"
#define PROGRAM_DATE    "Oct 2003"

static char Help_string[] =

"\nThis program takes a file of clauses (arg 1), a test to\n"
"apply (arg 2), and a stream of interpretations (from stdin).\n"
"The tests are 'models' and 'nonmodels'.  We print (to\n"
"stdout) the interpretations that either model or don't model\n"
"the set of clauses.\n\n"
"For example, if the set of claues consists of associativity \n"
"and commutativity (AC), we can remove AC models by using the\n"
"'nonmodels' test.\n\n";

static BOOL member_args(int argc, char **argv, char *str)
{
  int i;
  for (i = 1; i < argc; i++)
    if (str_ident(argv[i], str))
      return TRUE;
  return FALSE;
}  /* member_args */

/*************
 *
 *  find_false_clause(clauses, interp)
 *
 *  Return the first clause that is false in the given interpretation.
 *
 *************/

static Clause find_false_clause(Clist clauses, Interp interp)
{
  Clist_pos p = clauses->first;
  while (p != NULL) {
    if (!eval_clause(p->c, interp))
      return p->c;
    else
      p = p->next;
  }
  return NULL;
}  /* find_false_clause */

int main(int argc, char **argv)
{
  FILE *clause_fp;
  Term t;
  Interp interp;
  unsigned long int checked = 0;
  unsigned long int passed = 0;
  Clist clauses;
  BOOL models = TRUE;

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "help") ||
      argc < 3) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }

  if (str_ident(argv[2], "models"))
    models = TRUE;
  else if (str_ident(argv[2], "nonmodels"))
    models = FALSE;
  else
    fatal_error("interpfilter: operation should be 'models' or 'nonmodels'");

  clause_fp = fopen(argv[1], "r");
  if (clause_fp == NULL)
    fatal_error("interpfilter: clause file cannot be opened for reading");

  declare_standard_parse_types();
  init_options();

  t = read_commands(clause_fp, stderr, FALSE, KILL_UNKNOWN);
  
  if (!is_term(t, "clauses", 1))
    fatal_error("interpfilter: clauses(...) not found");

  clauses = read_clause_list(clause_fp, stderr, NULL, FALSE);

  fclose(clause_fp);

  /* Print each interpretation on stdin that satisfies the query. */

  t = read_term(stdin, stderr);
  while (t != NULL) {
    Clause false_clause;
    checked++;
    interp = compile_interp(t);
    false_clause = find_false_clause(clauses, interp);
    if ((models && false_clause == NULL) ||
	(!models && false_clause != NULL)) {
      passed++;
      fprint_interp(stdout, interp);
    }
    zap_interp(interp);
    zap_term(t);
    t = read_term(stdin, stderr);
  }

  printf("%% interpfilter %s %s: checked %lu, passed %lu, %.2f seconds.\n",
	 argv[1], argv[2], checked, passed,
	 user_time() / 1000.0);

  exit(passed > 0 ? 0 : 1);
}  /* main */

