#include "../ladr/header.h"

#include "../ladr/clock.h"
#include "../ladr/ioutil.h"
#include "../ladr/interp.h"
#include "../ladr/commands.h"
#include "../ladr/fastparse.h"

#define PROGRAM_NAME    "modtester"
#define PROGRAM_VERSION "2003-C"
#define PROGRAM_DATE    "Nov 2003"

static char Help_string[] =
"\nThis program takes a file of interpretations (arg 1) and a stream of\n"
"clauses (stdin).  For each clause, it tells interpretations in which\n"
"the clause is true.  An optional third argument 'fast' says to read\n"
"and write the clauses in fastparse form (e.g., =mxxx.).  The file of\n"
"interpretations can have operator declarations at the beginning.\n"
"For example,\n\n"
"    modtester interps fast < clauses.in\n\n";

static BOOL member_args(int argc, char **argv, char *str)
{
  int i;
  for (i = 1; i < argc; i++)
    if (str_ident(argv[i], str))
      return TRUE;
  return FALSE;
}  /* member_args */

int main(int argc, char **argv)
{
  FILE *interp_fp;
  BOOL fast;
  Clause c;
  Term t;
  Interp interp;
  Plist interps = NULL;
  int icount = 0;
  int *counters, i;

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "-help") ||
      argc < 2) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }

  interp_fp = fopen(argv[1], "r");
  if (interp_fp == NULL)
    fatal_error("interpretation file cannot be opened for reading");

  fast = (argc == 3 && str_ident(argv[2], "fast"));

  declare_standard_parse_types();
  init_options();

  t = read_commands(interp_fp, stderr, FALSE, KILL_UNKNOWN);

  /* terms(interpretations) and end_of_list are optional. */
  
  if (is_term(t, "terms", 1))
    t = read_term(interp_fp, stderr);  /* get first interpretation */

  while (t != NULL && !end_of_list_term(t)) {
    icount++;
    interp = compile_interp(t);
    interps = plist_append(interps, interp);
    zap_term(t);
    t = read_term(interp_fp, stderr);
  }

  fclose(interp_fp);

  counters = calloc(icount, sizeof(int *));

  /* Evaluate each clause on stdin. */

  if (fast) {
    /* Declare the symbols for fastparse. */
    fast_set_defaults();
  }

  c = clause_reader(fast);

  while (c != NULL) {

    Plist p = interps;
    Term t = clause_to_term(c);
    fwrite_term(stdout, t);
    zap_term(t);
    printf(".  %%");

    i = 0;
    while (p != NULL) {
      if (eval_clause(c, p->v)) {
	counters[i]++;
	printf(" %2d", i+1);
      }
      i++;
      p = p->next;
    }
    printf("\n");
    fflush(stdout);
    zap_clause(c);
    c = clause_reader(fast);
  }

  for (i = 0; i < icount; i++) {
    printf("%% interp %d models %d clauses.\n", i+1, counters[i]);
  }
  
  exit(0);
}  /* main */

