#include "../ladr/header.h"

#include "../ladr/ioutil.h"
#include "../ladr/fastparse.h"

#define PROGRAM_NAME    "renamer"
#define PROGRAM_VERSION "2003-C"
#define PROGRAM_DATE    "Nov 2003"

static char Help_string[] =
"\nThis program takes a stream of clauses (from stdin) and\n"
"renumbers the variables.  An optional argument 'fast' says\n"
"that the clauses will be read and written in fastparse form.\n\n";

static BOOL member_args(int argc, char **argv, char *str)
{
  int i;
  for (i = 1; i < argc; i++)
    if (str_ident(argv[i], str))
      return TRUE;
  return FALSE;
}  /* member_args */

int main(int argc, char **argv)
{

  BOOL fast_parse;
  Clause c;

  if (member_args(argc, argv, "help") ||
      member_args(argc, argv, "-help")) {
    printf("\n%s, version %s, %s\n",PROGRAM_NAME,PROGRAM_VERSION,PROGRAM_DATE);
    printf("%s", Help_string);
    exit(1);
  }

  fast_parse = member_args(argc, argv, "fast");

  if (fast_parse) {
    /* Declare the symbols for fastparse. */
    fast_set_defaults();
  }
  else {
    declare_standard_parse_types();
    set_parse_type("^", 450, INFIX);
    set_parse_type("v", 450, INFIX);
  }

  c = clause_reader(fast_parse);

  while (c != NULL) {

    renumber_variables(c, 100);  /* fatal if too many variables */
    clause_writer(c, fast_parse);
    fflush(stdout);

    zap_clause(c);
    c = clause_reader(fast_parse);
  }
  exit(0);
}  /* main */

