#ifndef TP_ATTRIB_H
#define TP_ATTRIB_H

#include "header.h"
#include "term.h"
#include "unify.h"

/* INTRODUCTION
This package is about lists of attributes.  Each attribute is
a pair (attribute-id, attribute-value).  Each attribute-id
is associated with a data type.
<P>
To use an attribute, you first have to call register_attribute(),
giving the name type of the attribute, and it returns an
ID to be used with the "set" and "get"  operations.
<P>
For example,
<PRE>
  int label_attr = register_attribute("label", STRING_ATTRIBUTE);
  ...
  Attribute a = set_string_attribute(NULL, label_attr, "clause_32");
  ...
  char *s = get_string_attribute(a, label_attr, 1);
</PRE>
*/

/* Public definitions */

typedef enum { INT_ATTRIBUTE,
	       STRING_ATTRIBUTE,
               TERM_ATTRIBUTE
             } Attribute_type;

typedef struct attribute * Attribute;

typedef union {int i; char *s; Term t;} Atype;

/* End of public definitions */

/* Public function prototypes from attrib.c */

Attribute get_attribute(void);

void free_attribute(Attribute p);

void fprint_attrib_mem(FILE *fp, int heading);

void p_attrib_mem();

int register_attribute(char *name, Attribute_type type);

void declare_term_attribute_inheritable(int id);

Attribute set_int_attribute(Attribute a, int id, int val);

int get_int_attribute(Attribute a, int id, int n);

BOOL exists_int_attribute(Attribute a, int id);

Attribute set_term_attribute(Attribute a, int id, Term val);

void replace_term_attribute(Attribute a, int id, Term val, int n);

void replace_int_attribute(Attribute a, int id, int val, int n);

Term get_term_attribute(Attribute a, int id, int n);

Attribute set_string_attribute(Attribute a, int id, char *val);

char *get_string_attribute(Attribute a, int id, int n);

void zap_attributes(Attribute a);

Term build_attr_term(Attribute a);

Term attributes_to_term(Attribute a, char *operator);

int attribute_name_to_id(char *name);

Attribute term_to_attributes(Term t, char *operator);

Attribute transfer_inheritable_attributes(Attribute parent_attr,
			   Context subst,
			   Attribute child_attr);

Attribute copy_attributes(Attribute a);

Attribute copy_inheritable_attributes(Attribute a);

void renumber_vars_attributes(Attribute attrs, int vmap[], int max_vars);

void set_vars_attributes(Attribute attrs, char *vnames[], int max_vars);

#endif  /* conditional compilation of whole file */
