#include "backdemod.h"

/* Private definitions and types */

/*************
 *
 *   index_term_back_demod()
 *
 *************/

static
void index_term_back_demod(Term t, Mindex idx, Indexop op)
{
  if (!VARIABLE(t)) {
    int i;
    mindex_update(idx, t, op);
    for (i = 0; i < ARITY(t); i++)
      index_term_back_demod(ARG(t,i), idx, op);
  }
}  /* index_term_back_demod */

/*************
 *
 *   index_clause_back_demod()
 *
 *************/

/* DOCUMENTATION
This routine indexes or unindexes an equation as a left-to-right
demodulator.
*/

/* PUBLIC */
void index_clause_back_demod(Clause c, Mindex idx, Indexop op)
{
  Literal lit;
  for (lit = c->literals; lit != NULL; lit = lit->next) {
    Term atom = lit->atom;
    int i;
    for (i = 0; i < ARITY(atom); i++) {
      index_term_back_demod(ARG(atom,i), idx, op);
    }
  }
}  /* index_clause_back_demod */

/*************
 *
 *   rewritable_term()
 *
 *************/

static
BOOL rewritable_term(Term alpha, Term t, Context subst)
{
  Trail tr = NULL;
  if (match(alpha, subst, t, &tr)) {
    undo_subst(tr);
    return TRUE;
  }
  else {
    int i;
    BOOL ok = FALSE;
    for (i = 0; i < t->arity && !ok; i++)
      ok = rewritable_term(alpha, ARG(t,i), subst);
    return ok;
  }
}  /* rewritable_term */

/*************
 *
 *   rewritable_clause()
 *
 *************/

/* DOCUMENTATION
This Boolean function checks if Clause c is can be rewritten
by Clause demod, which is assumed to be an oriented (left-to-right)
equation.
*/

/* PUBLIC */
BOOL rewritable_clause(Clause demod, Clause c)
{
  Term alpha = demod->literals->atom->args[0];
  Literal lit;
  BOOL ok = FALSE;
  Context subst = get_context();
  for (lit = c->literals; lit != NULL && !ok; lit = lit->next) {
    Term atom = lit->atom;
    int i;
    for (i = 0; i < atom->arity && !ok; i++)
      ok = rewritable_term(alpha, ARG(atom,i), subst);
  }
  free_context(subst);
  return ok;
}  /* rewritable_clause */

/*************
 *
 *   back_demod_linear()
 *
 *************/

/* DOCUMENTATION
This routine returns a Plist: the subset of Clist lst that can be
rewritten with Clause demod (which is assumed to be an oriented
equation).  The Plist which is returned is ordered by decreasing
clause ID.
*/

/* PUBLIC */
Plist back_demod_linear(Clause demod, Clist lst, Plist rewritables)
{
  Clist_pos p;

  for (p = lst->first; p != NULL; p = p->next) {
    Clause c = p->c;
    if (c != demod && rewritable_clause(demod, c)) {
      rewritables = insert_clause_into_plist(rewritables, c, FALSE);
    }
  }
  return(rewritables);
}  /* back_demod_linear */

/*************
 *
 *   back_demod_indexed()
 *
 *************/

/* DOCUMENTATION
This routine returns a Plist: the set clauses that have terms
that can be rewritten with Clause demod (which is assumed to be
an oriented  equation).  The Plist which is returned is ordered
by decreasing clause ID.
*/

/* PUBLIC */
Plist back_demod_indexed(Clause demod, Mindex idx)
{
  Plist rewritables = NULL;
  Context subst = get_context();
  Mindex_pos pos;
  Term t = mindex_retrieve_first(demod->literals->atom->args[0],
				 idx, INSTANCE, subst, NULL, FALSE, &pos);
  while (t != NULL) {
    Clause c = t->container;
    if (c != demod) {          /* in case demod is already in idx */
      rewritables = insert_clause_into_plist(rewritables, c, FALSE);
    }
    t = mindex_retrieve_next(pos);
  }
  free_context(subst);
  return rewritables;
}  /* back_demod_indexed */

