#include "bsub_conflict.h"

/* Private definitions and types */

static Lindex Unit_idx;
static Lindex Nonunit_idx;

/*************
 *
 *   init_bsub_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_bsub_index(Mindextype mtype,
		     Uniftype utype,
		     int fpa_depth)
{
  Unit_idx    = lindex_init(mtype, utype, fpa_depth, mtype, utype, fpa_depth);
  Nonunit_idx = lindex_init(mtype, utype, fpa_depth, mtype, utype, fpa_depth);
}  /* init_bsub_index */

/*************
 *
 *   bsub_destroy_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void bsub_destroy_index(void)
{
  lindex_destroy(Unit_idx);
  lindex_destroy(Nonunit_idx);
}  /* bsub_destroy_index */

/*************
 *
 *   bsub_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void bsub_index(Clause c, Indexop op)
{
  /* If c is a nonunit, all literals will be indexed, because
     a subsumer has to match only a subset of c. */
  BOOL unit = (number_of_literals(c) == 1);
  lindex_update(unit ? Unit_idx : Nonunit_idx, c, op);
}  /* bsub_index */

/*************
 *
 *   back_subsumption()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Plist back_subsumption(Clause c)
{
  Plist p1 = back_subsume(c, Unit_idx);
  Plist p2 = back_subsume(c, Nonunit_idx);
  Plist p3 = plist_cat(p1, p2);
  return p3;
}  /* back_subsumption */

/*************
 *
 *   unit_conflict()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Clause unit_conflict(Clause c)
{
  return unit_conflict_by_index(c, Unit_idx);
}  /* unit_conflict */

