#ifndef TP_CLASH_H
#define TP_CLASH_H

#include "mindex.h"
#include "clause.h"
#include "just.h"

/* INTRODUCTION
This package deals with clash structures, which are used for
binary resolution, hyperresolution, and UR-resolution.
The inference rule sets up a clash list (corresponding
to the nucleus), and then calls clash() to compute all of the
resolvents.
*/

/* Public definitions */

typedef struct clash  * Clash;

struct clash {
  BOOL       clashable;
  BOOL       clashed;
  Literal    nuc_lit;
  Context    nuc_subst;
  Literal    sat_lit;
  Context    sat_subst;
  Mindex     mate_index;
  Mindex_pos mate_pos;
  Clash      next;
};

/* End of public definitions */

/* Public function prototypes from clash.c */

void fprint_clash_mem(FILE *fp, BOOL heading);

void p_clash_mem();

Clash append_clash(Clash p);

void  zap_clash(Clash p);

Literal atom_to_literal(Term atom);

Literal apply_lit(Literal lit, Context c);

void clash(Clash c,
	   BOOL (*sat_test) (Literal),
	   Just_type rule,
	   void (*proc_proc) (Clause));

#endif  /* conditional compilation of whole file */
