#ifndef TP_CLAUSE_H
#define TP_CLAUSE_H

#include "glist.h"
#include "term.h"
#include "termflag.h"
#include "attrib.h"

/* INTRODUCTION
*/

/* Public definitions */

typedef struct literal * Literal;

struct literal {
  int      sign;
  Term     atom;
  Literal  next;
};

typedef struct clause * Clause;

struct clause {
  int              id;
  int              weight;
  Literal          literals;
  Attribute        attributes;
  struct just      *justification;
  struct clist_pos *containers;
};

/* End of public definitions */

/* Public function prototypes from clause.c */

void free_literal(Literal p);

Clause get_clause(void);

void fprint_clause_mem(FILE *fp, BOOL heading);

void p_clause_mem();

void zap_literals(Literal l);

void zap_clause(Clause c);

void fprint_clause(FILE *fp, Clause c);

void p_clause(Clause c);

Literal new_literal(int sign, Term atom);

void append_literal(Clause c, Literal lit);

int clause_length(Clause c);

Clause term_to_clause(Term t);

Term clause_to_term(Clause c);

Term clause_to_term_x(Clause c);

void clause_set_variables(Clause c, int max_vars);

void renumber_variables(Clause c, int max_vars);

void term_renumber_variables(Term t, int max_vars);

int positive_literals(Clause c);

int negative_literals(Clause c);

BOOL positive_clause(Clause c);

BOOL any_clause(Clause c);

BOOL negative_clause(Clause c);

BOOL mixed_clause(Clause c);

int number_of_literals(Clause c);

BOOL unit_clause(Clause c);

BOOL horn_clause(Clause c);

BOOL definite_clause(Clause c);

int greatest_variable_in_clause(Clause c);

Clause copy_clause(Clause c);

Clause copy_clause_with_flag(Clause c, int flag);

void upward_clause_links(Clause c);

int literal_number(Clause c, Literal l);

int atom_number(Clause c, Term atom);

Literal ith_literal(Clause c, int i);

BOOL true_term(Term t);

BOOL false_term(Term t);

BOOL tautology(Clause c);

void inherit_attributes(Clause parent1, Context subst1,
			Clause parent2, Context subst2,
			Clause child);

Ilist function_symbols_in_clause(Clause c, Ilist g);

Ilist relation_symbols_in_clause(Clause c, Ilist g);

Literal remove_null_literals(Literal l);

void new_atom_nodes(Clause c);

void new_atoms(Clause c);

Literal first_negative_literal(Clause c);

Plist constants_in_clause(Clause c);

#endif  /* conditional compilation of whole file */
