#ifndef TP_CLIST_H
#define TP_CLIST_H

#include "clause.h"

/* INTRODUCTION
This package handles Clists, which are doubly-linked lists of (pointers
to) clauses.  This is the "official" way of building lists of clauses.
(If you need a temporary, singly linked list, you can use Plist instead.)
<P>
An important property of Clists is that each clause knows what Clists
it is on.  In particular, each clause has a (singly linked) list of
containing Clists, constructed from the same nodes as the ordinary Clist
(see the definition of struct clist_pos).
*/

/* Public definitions */

typedef struct clist_pos * Clist_pos;
typedef struct clist * Clist;

struct clist {
  char       *name;
  Clist_pos  first, last;
  int        length;
};

struct clist_pos {
  Clist_pos  prev, next;  /* previous and next member of Clist */
  Clist_pos  nocc;        /* next member of containment list */
  Clist      list;        /* the head of the list */
  Clause     c;           /* pointer to the clause */
};

/* End of public definitions */

/* Public function prototypes from clist.c */

void fprint_clist_mem(FILE *fp, BOOL heading);

void p_clist_mem();

Clist clist_init(char *name);

void name_clist(Clist p, char *name);

void clist_free(Clist p);

void clist_append(Clause c, Clist l);

void clist_prepend(Clause c, Clist l);

void clist_insert_before(Clause c, Clist_pos pos);

void clist_insert_after(Clause c, Clist_pos pos);

void clist_remove(Clause c, Clist l);

void clist_remove_all_clauses(Clist l);

int clist_remove_all(Clause c);

int clist_member(Clause c, Clist l);

void fprint_clist(FILE *fp, Clist l);

void p_clist(Clist l);

void clist_zap(Clist l);

void clist_check(Clist l);

void clist_append_all(Clist l1, Clist l2);

BOOL clist_empty(Clist lst);

int clist_length(Clist l);

int max_wt_in_clist(Clist l);

#endif  /* conditional compilation of whole file */
