#include "connection.h"

/* Private definitions and types */

/*
 * memory management
 */

static link link_avail;
static unsigned link_gets, link_frees, link_avails;

static link2 link2_avail;
static unsigned link2_gets, link2_frees, link2_avails;

/* end of counters for each type */

/*************
 *
 *    link get_link()
 *
 *************/

/* PUBLIC */
link get_link(void)
{
  link p;

  link_gets++;
  if (link_avail == NULL)
    p = tp_alloc(sizeof(struct link));
  else {
    link_avails--;
    p = link_avail;
    link_avail = link_avail->next;
  }
  /* initialization */	
  	
  p->next = NULL;

  /* end of initialization */	
  return(p);
}  /* get_link */

/*************
 *
 *    free_link()
 *
 *************/

/* PUBLIC */
void free_link(link p)
{
  link_frees++;
  link_avails++;
  p->next = link_avail;
  link_avail = p;
}  /* free_link */

/*************
 *
 *    link2 get_link2()
 *
 *************/

/* PUBLIC */
link2 get_link2(void)
{
  link2 p;

  link2_gets++;
  if (link2_avail == NULL)
    p = tp_alloc(sizeof(struct link2));
  else {
    link2_avails--;
    p = link2_avail;
    link2_avail = link2_avail->next;
  }
  /* initialization */	
  	
  p->next = NULL;

  /* end of initialization */	
  return(p);
}  /* get_link2 */

/*************
 *
 *    free_link2()
 *
 *************/

/* PUBLIC */
void free_link2(link2 p)
{
  link2_frees++;
  link2_avails++;
  p->next = link2_avail;
  link2_avail = p;
}  /* free_link2 */

/* end of get and free routines for each type */

/*************
 *
 *   fprint_connection_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to FILE *fp) memory usage statistics for data types
associated with ??.
The Boolean argument heading tells whether to print a heading on the table.
*/

/* PUBLIC */
void fprint_connection_mem(FILE *fp, int heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct link);
  fprintf(fp, "link (%4d)      %11u%11u%11u%11u%9.1f K\n",
	  n, link_gets, link_frees,
	  link_gets - link_frees, link_avails,
	  (((link_gets - link_frees) + link_avails) * n) / 1024.);

  n = sizeof(struct link2);
  fprintf(fp, "link2 (%4d)      %11u%11u%11u%11u%9.1f K\n",
	  n, link2_gets, link2_frees,
	  link2_gets - link2_frees, link2_avails,
	  (((link2_gets - link2_frees) + link2_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_connection_mem */

/*************
 *
 *   p_connection_mem()
 *
 *************/

/* DOCUMENTATION
This routine prints (to stdout) memory usage statistics for data types
associated with ??.
*/

/* PUBLIC */
void p_connection_mem()
{
  fprint_connection_mem(stdout, 1);
}  /* p_connection_mem */

/*
 *  end of memory management
 */
