#include "control_mem.h"

/* Private definitions and types */

static int Disabled = 0;

/*************
 *
 *   clist_distribution()
 *
 *************/

static
int *clist_distribution(Clist_pos p, int n)
{
  int *a = calloc(sizeof(int), n);
  Clist_pos q;
  for (q = p; q; q = q->next) {
    int wt = q->c->weight;
    if (wt <= 0)
      a[0]++;
    else if (wt >= n)
      a[n-1]++;
    else
      a[wt]++;
  }
  return a;
}  /* clist_distribution */

/*************
 *
 *   p_dist()
 *
 *************/

static
void p_dist(int *a, int n)
{
  int i;
  for (i = 0; i < n; i++)
    printf("%d:  %d\n", i, a[i]);
}  /* p_dist */

/*************
 *
 *   disable_plist()
 *
 *************/

static
void disable_plist(Plist p, void (*disable_proc) (Clause))
{
  Plist q;
  for (q = p; q; q = q->next)
    (*disable_proc)(q->v);
  zap_plist(p);
}  /* disable_plist */
		  
/*************
 *
 *   disable_sos()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int disable_sos(Clist sos,
		int number_to_keep,
		int old_max_wt,
		int ratio,
		void (*disable_proc) (Clause))
{
  int initial_skip, i, n, keep, asize, new_max_wt;
  int *dist;
  Clist_pos p, q;
  Plist to_be_disabled = NULL;

  /* If ratio is 3, don't disable first 25% of clauses. */

  if (ratio <= 0)
    initial_skip = 0;  /* no ratio (-1) or level saturation (0) */
  else
    initial_skip = number_to_keep / (ratio + 1);

  for (p = sos->first, i = 0; p && i < initial_skip; p = p->next)
    i++;
  if (p == NULL)
    fatal_error("disable_sos, error 1");

  asize = (old_max_wt == INT_MAX ? max_wt_in_clist(sos)+1 : old_max_wt+1);
  dist = clist_distribution(p, asize);
  keep = number_to_keep - initial_skip; /* keep at most this many of sublist p */
  i = n = 0;
  while (n < keep && i < asize)
    n += dist[i++];
  if (n < keep)
    fatal_error("disable_sos, error 2");
  new_max_wt = i-1;
  
  /* Disable all clauses, starting with p, with wt > new_max_wt. */

  for (q = p; q; q = q->next) {
    if (q->c->weight > new_max_wt)
      to_be_disabled = plist_append(to_be_disabled, q->c);
  }

  n = plist_count(to_be_disabled);
  Disabled += n;

  disable_plist(to_be_disabled, disable_proc);

  printf("\nTarget=%d, disabled %d, new_sos_size=%d, new_max_wt=%d.\n",
	 number_to_keep, n, clist_length(sos), new_max_wt);
	 
    
  free(dist);
  return new_max_wt;
}  /* disable_sos */

/*************
 *
 *   control_sos_2()
 *
 *************/

/* DOCUMENTATION
<P>
*/

/* PUBLIC */
int control_sos_2(Clist sos,
		  int given,
		  int limit_millisec,
		  int current_max_wt,
		  int ratio,
		  void (*disable_proc) (Clause))
{
  int current_millisec = user_time();
  if (current_millisec <= 0 || given == 0)
    return current_max_wt;
  else {
    int can_be_given = given * (limit_millisec / (double) current_millisec) - given;
    if (clist_length(sos) <= can_be_given)
      return current_max_wt;
    else
      return disable_sos(sos,can_be_given,current_max_wt,ratio,disable_proc);
  }
}  /* control_sos_2 */
