#include "demod.h"

/* Private definitions and types */

static int Demod_attempts = 0;
static int Demod_rewrites = 0;

/*************
 *
 *    apply_demod()
 *
 *    Apply always succeeds and returns a pointer to the
 *    instantiated term.  If any subterms or terms in the context
 *    have the flag set, set the flag of the
 *    corresponding terms in the result.
 *    Otherwise, this is similar to ordinary apply.
 *    
 *************/

static
Term apply_demod(Term t, Context c, int flag)
{
  Term t2;
  DEREFERENCE(t, c)

  /* A NULL context is ok.  It happens when c is built by match. */
  /* If the context is NULL, then apply just copies the term.    */
    
  if (VARIABLE(t)) {
    if (c == NULL)
      t2 = get_variable_term(VARNUM(t));
    else
      t2 = get_variable_term(c->multiplier * MAX_VARS + VARNUM(t));
  }
  else {  /* constant or complex term */
    int i;
    t2 = get_rigid_term_like(t);
    if (term_flag(t, flag))
      term_flag_set(t2, flag);
    for (i = 0; i < t->arity; i++)
      t2->args[i] = apply(t->args[i], c);
  }

  if (term_flag(t, flag))
    term_flag_set(t2, flag);
  return t2;
}  /* apply_demod */

/*************
 *
 *   demod()
 *
 *   For non-AC terms.
 *
 *************/

#if 0
static
Term demod(Term t, Mindex demods, int flag, Ilist *just_head)
{
  if (term_flag(t, flag) || VARIABLE(t))
    ;  /* leave it alone */
  else {
    int i;
    Mindex_pos pos;
    Term contractum, alpha, atom;
    Clause demodulator;
    Context c;

    for (i = 0; i < t->arity; i++)
      t->args[i] = demod(t->args[i], demods, flag, just_head);
    c = get_context();
    Demod_attempts++;
    alpha = mindex_retrieve_first(t,demods,GENERALIZATION,NULL,c,FALSE,&pos);
				  
    if (alpha != NULL) {
      Demod_rewrites++;
      demodulator = alpha->container;
      atom = demodulator->literals->atom;
      contractum = apply_demod(atom->args[1], c, flag);
      mindex_retrieve_cancel(pos);  /* This also clears the context. */

      zap_term(t);
      *just_head = ilist_prepend(*just_head, demodulator->id);
      t = demod(contractum, demods, flag, just_head);
    }
    free_context(c);
  }
  term_flag_set(t, flag);  /* Mark as fully demodulated. */
  return(t);
}  /* demod */
#else
static
Term demod(Term t, Mindex demods, int flag, Ilist *just_head)
{
  if (term_flag(t, flag) || VARIABLE(t))
    ;  /* leave it alone */
  else {
    Mindex_pos pos;
    Term found;
    Context c;  /* allocate after we demoulate the subterms */
    int i;

    for (i = 0; i < t->arity; i++)
      t->args[i] = demod(t->args[i], demods, flag, just_head);

    c = get_context();
    Demod_attempts++;
    found = mindex_retrieve_first(t,demods,GENERALIZATION,NULL,c,FALSE,&pos);
				  
    while (found != NULL) {
      Clause demodulator = found->container;
      Term atom = demodulator->literals->atom;
      Term alpha = ARG(atom, 0);
      Term beta = ARG(atom, 1);
      Term other = (found == alpha ? beta : alpha);
      Term contractum = apply_demod(other, c, flag);
      BOOL ok;

      if (oriented_eq(atom))
	ok = TRUE;
      else
	ok = lrpo(t, contractum);
      
      if (ok) {
	Demod_rewrites++;
	zap_term(t);
	*just_head = ilist_prepend(*just_head, demodulator->id);
	t = demod(contractum, demods, flag, just_head);
	mindex_retrieve_cancel(pos);  /* This also clears the context. */
	found = NULL;
      }
      else {
	zap_term(contractum);
	found = mindex_retrieve_next(pos);
      }
    }
    free_context(c);
  }
  term_flag_set(t, flag);  /* Mark as fully demodulated. */
  return(t);
}  /* demod */
#endif

/*************
 *
 *    contract_bt
 *
 *************/

static
Term contract_bt(Term t, Mindex demods, int flag, Clause *demodulator_ptr)
{
  Mindex_pos pos;
  Term contractum, alpha;
  Clause demodulator = NULL;
  Context c = get_context();

  alpha = mindex_retrieve_first(t,demods,GENERALIZATION,NULL,c,TRUE,&pos);
				  
  if (alpha == NULL)
    contractum = NULL;
  else {
    Term atom;
    demodulator = alpha->container;
    atom = demodulator->literals->atom;
    contractum = apply_demod(atom->args[1], c, flag);
    if (c->partial_term) {
      /* Get copy, including marks that indicate normal terms. */
      Term partial = apply_demod(c->partial_term, NULL, flag);
      contractum = build_binary_term(SYMNUM(t), contractum, partial);
    }
    mindex_retrieve_cancel(pos);
  }
  free_context(c);
  *demodulator_ptr = demodulator;
  return(contractum);
}  /* contract_bt */

/*************
 *
 *    demod_bt
 *
 *************/

static
Term demod_bt(Term t, Mindex demods, int psn, int flag, Ilist *just_head)
{
  if (term_flag(t, flag) || VARIABLE(t))
    ;  /* leave it alone */
  else {
    int i;
    int sn = SYMNUM(t);
    for (i = 0; i < t->arity; i++)
      t->args[i] = demod_bt(t->args[i], demods, sn, flag, just_head);

    if (sn == psn && is_assoc_comm(sn))
      ;  /* t is part of an AC term, so leave it alone. */
    else {
      Term contractum;
      Clause demodulator;
      ac_canonical(t, flag);
      Demod_attempts++;
      contractum = contract_bt(t, demods, flag, &demodulator);
      if (contractum) {
	Demod_rewrites++;
	zap_term(t);
	ac_canonical(contractum, flag);
	*just_head = ilist_prepend(*just_head, demodulator->id);
	t = demod_bt(contractum, demods, psn, flag, just_head);
      }
    }
  }
  /* Mark as fully demodulated.  This also means ac_canonical. */
  term_flag_set(t, flag);
  return(t);
}  /* demod_bt */

/*************
 *
 *   demodulate()
 *
 *************/

/* DOCUMENTATION
This routine demodulates a term.  ID numbers of demodulators
are put on the front of just_head, so you'll probably want
to reverse the list before putting it into the clause justification.
*/

/* PUBLIC */
Term demodulate(Term t, Mindex demods, Ilist *just_head)
{
  int flag = claim_term_flag();
  Term t2;

  if (demods->unif_type == ORDINARY_UNIF)
    t2 = demod(t, demods, flag, just_head);
  else
    t2 = demod_bt(t, demods, -1, flag, just_head);
  term_flag_clear_recursively(t2, flag);
  release_term_flag(flag);
  return t2;
}  /* demodulate */

/*************
 *
 *   demod_attempts()
 *
 *************/

/* DOCUMENTATION
Return the number of rewrite attempts so far (for the whole process).
*/

/* PUBLIC */
int demod_attempts()
{
  return Demod_attempts;
}  /* demod_attempts */

/*************
 *
 *   demod_rewrites()
 *
 *************/

/* DOCUMENTATION
Return the number of successful rewrites so far (for the whole process).
*/

/* PUBLIC */
int demod_rewrites()
{
  return Demod_rewrites;
}  /* demod_rewrites */

