#include "demodulate.h"

/* Private definitions and types */

static Mindex Demod_idx;
static Mindex Back_demod_idx;

/*************
 *
 *   init_demodulator_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_demodulator_index(Mindextype mtype, Uniftype utype, int fpa_depth)
{
  Demod_idx = mindex_init(mtype, utype, fpa_depth);
}  /* init_demodulator_index */

/*************
 *
 *   init_back_demod_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_back_demod_index(Mindextype mtype, Uniftype utype, int fpa_depth)
{
  Back_demod_idx = mindex_init(mtype, utype, fpa_depth);
}  /* init_back_demod_index */

/*************
 *
 *   index_demodulator()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void index_demodulator(Clause c, Indexop operation)
{
  Term atom = c->literals->atom;
  Term alpha = ARG(atom, 0);
  Term beta  = ARG(atom, 1);

  /* Index the left argument. */
  if (variables_subset(beta, alpha)) {
    mindex_update(Demod_idx, alpha, operation);
  }

  /* Maybe index the right argument. */
  if (!oriented_eq(atom) &&
      !renamable_flip_eq(atom) &&
      variables_subset(alpha, beta)) {
    mindex_update(Demod_idx, beta, operation);
  }
}  /* index_demodulator */

/*************
 *
 *   back_demod_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void back_demod_index(Clause c, Indexop operation)
{
  index_clause_back_demod(c, Back_demod_idx, operation);
}  /* back_demod_index */

/*************
 *
 *   destroy_demodulation_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void destroy_demodulation_index(void)
{
  mindex_destroy(Demod_idx);
}  /* destroy_demodulation_index */

/*************
 *
 *   destroy_back_demod_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void destroy_back_demod_index(void)
{
  mindex_destroy(Back_demod_idx);
}  /* destroy_back_demod_index */

/*************
 *
 *   demodulate_clause()
 *
 *************/

/* DOCUMENTATION
Demodulate Clause c, using demodulators alreadly known to this package.
If any rewriting occurs, the justification is appended to
the clause's existing justification.
*/

/* PUBLIC */
void demodulate_clause(Clause c)
{
  demod_clause(c, Demod_idx);
}  /* demodulate_clause */

/*************
 *
 *   back_demodulatable()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Plist back_demodulatable(Clause demod)
{
  return back_demod_indexed(demod, Back_demod_idx);
}  /* back_demodulatable */

