#ifndef TP_FORMULA_H
#define TP_FORMULA_H

#include "listterm.h"

/* INTRODUCTION
*/

/* Public definitions */

/* formula types */

typedef enum {
  ATOM_FORM, AND_FORM, OR_FORM, NOT_FORM, IFF_FORM,
  IMP_FORM, IMPBY_FORM, ALL_FORM, EXISTS_FORM} Ftype;

typedef struct formula * Formula;
struct formula {
  Ftype   type;
  int     arity;
  char    *qvar;     /* quantified variable */
  Formula *kids;     /* for non-atoms */
  Term    atom;      /* for atoms */
};

/* formula preference */

typedef enum { CONJUNCTION,
	       DISJUNCTION
             } Fpref;

/* macros */

#define TRUE_FORMULA(f)  ((f)->type == AND_FORM && (f)->arity == 0)
#define FALSE_FORMULA(f) ((f)->type == OR_FORM  && (f)->arity == 0)

/* End of public definitions */

/* Public function prototypes from formula.c */

void free_formula(Formula p);

void fprint_formula_mem(FILE *fp, BOOL heading);

void p_formula_mem(void);

void zap_formula(Formula f);

Formula formula_flatten(Formula f);

Formula term_to_formula(Term t);

Term formula_to_term(Formula f);

void fprint_formula(FILE *fp, Formula f);

void p_formula(Formula c);

BOOL formula_ident(Formula f, Formula g);

Formula formula_copy(Formula f);

Formula nnf2(Formula f, Fpref pref);

Formula nnf(Formula f);

Formula formula_simplify(Formula f);

Formula cnf(Formula f);

Formula dnf(Formula f);

Formula skolemize(Formula f);

Formula unique_quantified_vars(Formula f);

Formula remove_universal_quantifiers(Formula f);

Formula clausify_prepare(Formula f);

int greatest_qvar(Formula f);

int greatest_symnum_in_formula(Formula f);

Formula eliminate_rebinding(Formula f);

#endif  /* conditional compilation of whole file */
