#include "forward_subsume.h"

/* Private definitions and types */

#define NUM_INDEXES 10  /* Must be >= 2 */

static Lindex Idx[NUM_INDEXES];

/*************
 *
 *   init_fsub_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_fsub_index(Mindextype mtype,
		     Uniftype utype,
		     int fpa_depth)
{
  int i;
  if (NUM_INDEXES < 2)
    fatal_error("init_fsub_index: NUM_INDEXES < 2");

  for (i = 0; i < NUM_INDEXES; i++)
    Idx[i] = lindex_init(mtype, utype, fpa_depth, mtype, utype, fpa_depth);
}  /* init_fsub_index */

/*************
 *
 *   fsub_destroy_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void fsub_destroy_index(void)
{
  int i;
  for (i = 0; i < NUM_INDEXES; i++)
    lindex_destroy(Idx[i]);
}  /* fsub_destroy_index */

/*************
 *
 *   fsub_index()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void fsub_index(Clause c, Indexop op)
{
  int n = number_of_literals(c);
  if (n >= NUM_INDEXES)
    n = NUM_INDEXES-1;
  /* Index only the first literal of c, because any subsumee
     will have to match that first literal.
   */
  lindex_update_first(Idx[n], c, op);
}  /* fsub_index */

/*************
 *
 *   forward_subsumption()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Clause forward_subsumption(Clause c)
{
  int nc = number_of_literals(c);  /* Don't let a longer clause subsume c. */
  int i;
  for (i = 1; i < NUM_INDEXES && i <= nc; i++) {
    Clause subsumer = forward_subsume(c, Idx[i]);
    if (subsumer != NULL)
      return subsumer;
  }
  return NULL;
}  /* forward_subsumption */

/*************
 *
 *   unit_deletion()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void unit_deletion(Clause c)
{
  unit_delete(c, Idx[1]);
}  /* unit_deletion */
