#ifndef TP_GLIST_H
#define TP_GLIST_H

#include "memory.h"

/* INTRODUCTION
This package handles Glist (generic) lists, which are simple,
singly linked lists.
The members of the lists can be integers or pointers.
In particular, each is a union of (int) and (void *).
There is no way to determine the types of the members of the
list, so you have to remember what kind of data are there.
<P>
Note that when assigning to and from a (void *), you don't need
a cast, but when dereferencing a (void *), you do need a cast.
<P>
There are two informal subtypes: "ilist" for lists of integers,
and "plist" for lists of pointers.  Several of the routines
(which have "ilist" or "plist" in the name) apply to these
suptypes.  The rest of the routines apply to any Glist.
*/

/* Public definitions */

typedef struct glist * Glist;

struct glist {
  Glist next;
  union {
    int i;
    void *v;
  } u;
};

typedef struct plist * Plist;

struct plist {
  void *v;
  Plist next;
};

typedef struct ilist * Ilist;

struct ilist {
  int i;
  Ilist next;
};

/* End of public definitions */

/* Public function prototypes from glist.c */

Ilist get_ilist(void);

void free_ilist(Ilist p);

Plist get_plist(void);

void free_plist(Plist p);

void fprint_glist_mem(FILE *fp, BOOL heading);

void p_glist_mem(void);

Ilist ilist_cat(Ilist p1, Ilist p2);

Ilist ilist_pop(Ilist p);

int ilist_count(Ilist p);

Ilist reverse_ilist(Ilist p);

Ilist copy_ilist(Ilist p);

void zap_ilist(Ilist p);

Ilist ilist_append(Ilist lst, int i);

Ilist ilist_prepend(Ilist lst, int i);

BOOL ilist_member(Ilist lst, int i);

Ilist ilist_subtract(Ilist p1, Ilist p2);

Ilist ilist_intersect(Ilist p1, Ilist p2);

Plist plist_cat(Plist p1, Plist p2);

Plist plist_pop(Plist p);

int plist_count(Plist p);

Plist reverse_plist(Plist p);

Plist copy_plist(Plist p);

void zap_plist(Plist p);

Plist plist_append(Plist lst, void *v);

Plist plist_prepend(Plist lst, void *v);

BOOL plist_member(Plist lst, void *v);

BOOL plist_subset(Plist a, Plist b);

void fprint_ilist(FILE *fp, Ilist p);

void p_ilist(Ilist p);

Ilist ilist_copy(Ilist p);

#endif  /* conditional compilation of whole file */
