#include "hints.h"

/* Private definitions and types */

static Lindex Hints_idx = NULL;     /* FPA index for hints */

/*************
 *
 *   init_hints()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_hints(void)
{
  Uniftype utype = (assoc_comm_symbols() || comm_symbols() ?
		    BACKTRACK_UNIF :
		    ORDINARY_UNIF);
  Hints_idx = lindex_init(FPA, utype, 100, FPA, utype, 100);
}  /* init_hints */

/*************
 *
 *   done_with_hints()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void done_with_hints(void)
{
  if (!lindex_empty(Hints_idx))
    printf("ERROR: Hints index not empty!\n");
  lindex_destroy(Hints_idx);
}  /* done_with_hints */

/*************
 *
 *   index_hint()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void index_hint(Clause c)
{
  lindex_update(Hints_idx, c, INSERT);
}  /* index_hint */

/*************
 *
 *   unindex_hint()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void unindex_hint(Clause c)
{
  lindex_update(Hints_idx, c, DELETE);
}  /* unindex_hint */

/*************
 *
 *   find_hint()
 *
 *************/

static
Clause find_hint(Clause c)
{
  return back_subsume_one(c, Hints_idx);
}  /* find_hint */

/*************
 *
 *   adjust_weight_with_hints()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void adjust_weight_with_hints(Clause c,
			      int bsub_add_wt_at,
			      int bsub_wt_at,
			      int label_at,
			      BOOL degrade)
{
  Clause hint = find_hint(c);

  if (hint != NULL) {

    /* This code allows both bsub_wt and bsub_add_wt to be used. */
    /* I don't think this will happen much in practice. */

    int bsub_wt = get_int_attribute(hint->attributes, bsub_wt_at, 1);
    int bsub_add_wt = get_int_attribute(hint->attributes, bsub_add_wt_at, 1);

    if (bsub_wt != INT_MAX)
      c->weight = bsub_wt;
    if (bsub_add_wt != INT_MAX)
      c->weight += bsub_add_wt;
    
    /* If the hint has label attributes, copy them to the clause. */
    
    {
      int i = 0;
      char *s = get_string_attribute(hint->attributes, label_at, ++i);
      while (s) {
	c->attributes = set_string_attribute(c->attributes, label_at, s);
	s = get_string_attribute(hint->attributes, label_at, ++i);
      }
    }

    /* Veroff's hint degradation strategy.  Cut in half the bsub_add_wt. */

    if (degrade && bsub_add_wt != INT_MAX) {
      int new = bsub_add_wt / 2;
      replace_int_attribute(hint->attributes, bsub_add_wt_at, new, 1);
      /*
      printf("degraded hint %d from %d to %d\n",hint->id,bsub_add_wt,new);
      */
    }
  }
}  /* adjust_weight_with_hints */
