#include "maximal.h"

/* Private definitions and types */

int Nonmaximal_flag     = -1;  /* termflag to mark nonmaximal literals */

/*************
 *
 *   lrpo_literals()
 *
 *************/

static
BOOL lrpo_literals(Literal l1, Literal l2)
{
  /* Is l1 > l2? */
  /* The negation sign is greatest in the symbol ordering. */
  if (!l1->sign && l2->sign)
    return TRUE;
  else if (l1->sign && !l2->sign)
    return FALSE;
  else {
    Term t1 = l1->atom;
    Term t2 = l2->atom;
    Ordertype p = sym_precedence(SYMNUM(t1), SYMNUM(t2));
    if (p == GREATER_THAN)
      return TRUE;
    else if (p == LESS_THAN)
      return FALSE;
    else
      return lrpo(t1, t2);
  }
}  /* lrpo_literals */

/*************
 *
 *   mark_non_maximal_literals()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void mark_non_maximal_literals(Clause c)
{
  Literal l1;

  /* return; // EVERYTHING IS MAXIMAL!!!! */

  if (Nonmaximal_flag == -1)
    Nonmaximal_flag = claim_term_flag();
  
  for (l1 = c->literals; l1; l1 = l1->next) {
    /* If there is a greater literal, mark l1 as nonmaximal. */
    Literal l2 = c->literals;
    BOOL ok = FALSE;
    while (l2 && !ok) {
      if (l1 != l2 && lrpo_literals(l2, l1))
	ok = TRUE;
      else
	l2 = l2->next;
    }
    if (ok) {
      /* Note: this is a property of the literal, but we mark the atom. */
      term_flag_set(l1->atom, Nonmaximal_flag);
    }
  }
}  /* mark_non_maximal_literals */

/*************
 *
 *   maximal_literal()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
BOOL maximal_literal(Literal lit)
{
  return !term_flag(lit->atom, Nonmaximal_flag);
}  /* maximal_literal */
