#include "nonport.h"

/* Private definitions and types */

#ifdef PRIMITIVE_ENVIRONMENT
/* This means that we don't have some UNIXy things */
#else
#  include <pwd.h>
#  include <unistd.h>
#endif

/*************
 *
 *   username()
 *
 *************/

/* DOCUMENTATION
Return the name of the user who started the current job.
*/

/* PUBLIC */
char *username(void)
{
#ifdef PRIMITIVE_ENVIRONMENT
  return("an unknown user");
#else
  struct passwd *p;
  p = getpwuid(getuid());
  return(p ? p->pw_name : "???");
#endif
}  /* username */

/*************
 *
 *   hostname()
 *
 *************/

/* DOCUMENTATION
Return the hostname of the computer on which the current job is running.
*/

/* PUBLIC */
char *hostname(void)
{
#ifdef PRIMITIVE_ENVIRONMENT
  return("an unknown computer");
#else
  static char host[64];
  if (gethostname(host, 64) != 0)
    strcpy(host, "???");
  return(host);
#endif
}  /* hostname */

/*************
 *
 *   my_process_id()
 *
 *************/

/* DOCUMENTATION
Return the process ID of the current process.
*/

/* PUBLIC */
int my_process_id(void)
{
#ifdef PRIMITIVE_ENVIRONMENT
  return 0;
#else
  return getpid();
#endif
}  /* my_process_id */
