#include "paramod.h"

/* Private definitions and types */

static int Basic_prunes;  /* statisic */

/*************
 *
 *   basic_paramodulation_prunes()
 *
 *************/

/* DOCUMENTATION
How many paramodulants were killed because they failed the "basic" test.
*/

/* PUBLIC */
int basic_paramodulation_prunes(void)
{
  return Basic_prunes;
}  /* basic_paramodulation_prunes */

/*************
 *
 *   basic_check()
 *
 *************/

static
BOOL basic_check(Term into_term)
{
  if (basic_paramod() && nonbasic_term(into_term)) {
    Basic_prunes++;
    return FALSE;
  }
  else
    return TRUE;
}  /* basic_check */

/*************
 *
 *   apply_lit_para()
 *
 *************/

static
Literal apply_lit_para(Literal lit, Context c)
{
  if (basic_paramod())
    return new_literal(lit->sign, apply_basic(lit->atom, c));
  else
    return new_literal(lit->sign, apply(lit->atom, c));
}  /* apply_lit_para */

/*************
 *
 *   apply_substitute_para()
 *
 *************/

static
Term apply_substitute_para(Term t, Term beta, Context from_subst,
			   Term into, Context into_subst)
{
  if (basic_paramod())
    return apply_basic_substitute(t, beta, from_subst, into, into_subst);
  else
    return apply_substitute(t, beta, from_subst, into, into_subst);
}  /* apply_substitute_para */

/*************
 *
 *   paramodulate()
 *
 *   Construct a paramodulant.
 *
 *************/

static
Clause paramodulate(Literal from_lit,
		    Term alpha, Context from_subst,
		    Term into,  Context into_subst)
{
  Clause from_clause = alpha->container;
  Clause into_clause = into->container;
  Clause p = get_clause();
  Term beta = NULL;
  Literal lit;
  for (lit = from_clause->literals; lit != NULL; lit = lit->next) {
    if (lit != from_lit)
      append_literal(p, apply_lit_para(lit, from_subst));
    else
      beta = (ARG(lit->atom,0)==alpha ? ARG(lit->atom,1) : ARG(lit->atom,0));
  }
  for (lit = into_clause->literals; lit != NULL; lit = lit->next) {
    append_literal(p, new_literal(lit->sign,
				  apply_substitute_para(lit->atom,
							beta, from_subst,
							into, into_subst)));
  }
  inherit_attributes(from_clause, from_subst, into_clause, into_subst, p);
  upward_clause_links(p);
  return p;
}  /* paramodulate */

/*************
 *
 *   para_right()
 *
 *   Should we paramoulate (from or into) the right side?
 *   If it is oriented, then NO;
 *   else if it is a renamable_flip unit, then NO;
 *   else YES.
 *
 *************/

static
BOOL para_right(Term atom)
{
  /* Assume atom is an eq_atom. */
  if (oriented_eq(atom))
    return FALSE;
  else if (renamable_flip_eq(atom) && unit_clause(atom->container))
    return FALSE;
  else
    return TRUE;
}  /* para_right */

/*************
 *
 *   ordering_check()
 *
 *************/

static
BOOL ordering_check(Term from_atom, int from_side, Context from_subst,
		    Term into_atom, int into_side, Context into_subst)
{
  BOOL ok = TRUE;
  return TRUE;
  /* This checks from and into equality atoms only.
     It does not check that literals are maximal in clauses. */
  
  if (!oriented_eq(from_atom)) {
    /* Check instance of from_atom; fail if the side we're using is smaller. */
    Term a1 = apply(ARG(from_atom, 0), from_subst);
    Term a2 = apply(ARG(from_atom, 1), from_subst);
    if ((from_side == 1 && lrpo(a2, a1)) ||
	(from_side == 2 && lrpo(a1, a2)))
      ok = FALSE;
    zap_term(a1);
    zap_term(a2);
  }
  if (ok && eq_atom(into_atom) && !oriented_eq(into_atom)) {
    /* Check instance of into_atom; fail if the side we're using is smaller. */
    Term a1 = apply(ARG(into_atom, 0), into_subst);
    Term a2 = apply(ARG(into_atom, 1), into_subst);
    if ((into_side == 1 && lrpo(a2, a1)) ||
	(into_side == 2 && lrpo(a1, a2)))
      ok = FALSE;
    zap_term(a1);
    zap_term(a2);
  }
  return ok;
}  /* ordering_check */

/*************
 *
 *   para_into()
 *
 *   Paramodulate from a given side of a given literal into a given
 *   term and/or its subterms.  The position vectors are for the
 *   justification only.  The parameter "where" should be
 *   PARA_ALL, PARA_TOP_ONLY, or PARA_ALL_EXCEPT_TOP.
 *
 *************/

static
void para_into(Literal from_lit, Term alpha, Ilist from_vec, int from_side,
	       Literal into_lit, Term into, Ilist into_vec, int into_side,
	       Para_loc where,
	       void (*proc_proc) (Clause))
{
  if (!VARIABLE(into) && basic_check(into)) {
    int i;
    if (COMPLEX(into) && (where == PARA_ALL || where == PARA_ALL_EXCEPT_TOP)) {
      Ilist g = ilist_prepend(into_vec, 0);
      for (i = 0; i < into->arity; i++) {
	g->i += 1;
	para_into(from_lit, alpha, from_vec, from_side,
		  into_lit, ARG(into,i), g, into_side,
		  PARA_ALL, proc_proc);
      }
      free_ilist(g);  /* Frees first node only; into_vec is unchanged. */
    }
    if (where == PARA_ALL || where == PARA_TOP_ONLY) {
      Context cf = get_context();
      Context ci = get_context();
      Trail tr = NULL;
      if (unify(alpha, cf, into, ci, &tr)) {
	if (ordering_check(from_lit->atom, from_side, cf,
			   into_lit->atom, into_side, ci)) {
	  Clause p = paramodulate(from_lit, alpha, cf, into, ci);
	  p->justification = para_just(PARA_JUST,
				       alpha->container, from_vec,
				       into->container, into_vec);
	  (*proc_proc)(p);
	}
	undo_subst(tr);
      }
      free_context(cf);
      free_context(ci);
    }
  }
}  /* para_into */

/*************
 *
 *   para_lits_into_noneq()
 *
 *************/

static
void para_lits_into_noneq(Literal flit, Term alpha,
			  Literal ilit,
			  void (*proc_proc) (Clause))
{
  /* Position vectors are constructed backward. */
  Ilist from_vec = ilist_prepend(ilist_prepend(NULL,0),0);
  Ilist into_vec = ilist_prepend(ilist_prepend(NULL,0),0);
  Term a = ilit->atom;
  int i;
  int from_side = (ARG(flit->atom,0) == alpha ? 1 : 2);

  from_vec->next->i = literal_number(alpha->container, flit);
  from_vec->i = from_side;  /* arg of flit */
  into_vec->next->i = literal_number(ilit->atom->container, ilit);
  /* The atom is not equality, so do nothing special. */
  for (i = 0; i < a->arity; i++) {
    into_vec->i += 1;  /* increment arg number */
    para_into(flit, alpha, from_vec, from_side,
	      ilit, ARG(a,i), into_vec, -1,
	      PARA_ALL, proc_proc);
  }
  zap_ilist(from_vec);
  zap_ilist(into_vec);
}  /* para_lits_into_noneq */

/*************
 *
 *   para_lits_into_eq()
 *
 *************/

static
void para_lits_into_eq(Literal flit, Term alpha,
			Literal ilit, Term into_arg,
			Para_loc where,
			void (*proc_proc) (Clause))
{
  int from_side = (ARG(flit->atom,0) == alpha ? 1 : 2);
  int into_side = (ARG(flit->atom,0) == alpha ? 1 : 2);
  
  /* Position vectors are constructed backward. */

  Ilist from_vec = ilist_prepend(ilist_prepend(NULL,0),0);
  Ilist into_vec = ilist_prepend(ilist_prepend(NULL,0),0);
  
  from_vec->next->i = literal_number(alpha->container, flit);
  from_vec->i = from_side;  /* arg of flit */
  into_vec->next->i = literal_number(into_arg->container, ilit);
  into_vec->i = into_side;  /* arg of ilit */

  para_into(flit, alpha, from_vec, from_side,
	    ilit, into_arg, into_vec, into_side,
	    where, proc_proc);

  zap_ilist(from_vec);
  zap_ilist(into_vec);
}  /* para_lits_into_eq */

/*************
 *
 *   para_from_into()
 *
 *************/

/* DOCUMENTATION
Paramodulate from one clause into another (non-backtrack unification version).
The "where" parameter tells where to paramodulate into
equality literals:  PARA_ALL, or PARA_ALL_EXCEPT_TOP.
"TOP" means the left or right argument of the equality.
<P>
For oriented equality atoms, we go from and into left sides only.
For nonoriented equality atoms, we go from and into both sides.
<P>
Without AC, the normal sequence for paramodulating C and D is
<UL>
<LI> C -> D, PARA_ALL;
<LI> D -> C, PARA_ALL_EXCEPT_TOP;
</UL>
*/

/* PUBLIC */
void para_from_into(Clause from, Clause into, Para_loc where,
		    void (*proc_proc) (Clause))
{
  Literal flit;
  for (flit = from->literals; flit != NULL; flit = flit->next) {
    if (flit->sign && eq_atom(flit->atom)) {
      Term f1 = ARG(flit->atom,0);
      Term f2 = ARG(flit->atom,1);
      Literal ilit;
      for (ilit = into->literals; ilit != NULL; ilit = ilit->next) {
	if (eq_atom(ilit->atom)) {
	  Term i1 = ARG(ilit->atom,0);
	  Term i2 = ARG(ilit->atom,1);

	  /* For now, don't go into vars because they don't have containment.
	     Check if we should go from/into right sides.
	  */

	  if (!VARIABLE(f1) && !VARIABLE(i1))
	    para_lits_into_eq(flit,f1,ilit,i1,where,proc_proc);

	  if (!VARIABLE(f2) && !VARIABLE(i1) && para_right(flit->atom))
	    para_lits_into_eq(flit,f2,ilit,i1,where,proc_proc);

	  if (!VARIABLE(f1) && !VARIABLE(i2) && para_right(ilit->atom))
	    para_lits_into_eq(flit,f1,ilit,i2,where,proc_proc);

	  if (!VARIABLE(f2) && !VARIABLE(i2) &&
	      para_right(flit->atom) && para_right(ilit->atom))
	    para_lits_into_eq(flit,f2,ilit,i2,where,proc_proc);
	}
	else {
	  if (!VARIABLE(f1))
	    para_lits_into_noneq(flit,f1,ilit,proc_proc);
	  if (!VARIABLE(f2) && para_right(flit->atom))
	    para_lits_into_noneq(flit,f2,ilit,proc_proc);
	}
      }
    }
  }
}  /* para_from_into */
