#include "weight.h"

/* Private definitions and types */

static int Number_of_weights;
static int *Weights;    /* This maps symbol numbers to weights. */
static int Not_symnum;  /* Cache the negation symbol: it's needed often. */
static int Variable_weight = 1;

/*************
 *
 *   init_weight_scheme()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
void init_weight_scheme(Plist p)
{
  Plist q;
  int i;
  Number_of_weights = greatest_symnum() + 1;

  Weights = malloc(Number_of_weights * sizeof(int));  /* Allocate an array */
  for (i = 0; i < Number_of_weights; i++)
    Weights[i] = 1;  /* default weight of each symbol is 1 */

  for (q = p; q; q = q->next) {
    Term t = q->v;
    int w;
    if (!is_term(t, "weight", 2) || !term_to_int(ARG(t,1), &w))
      fatal_error("init_weight_scheme, bad weight template");
    else {
      Weights[SYMNUM(ARG(t,0))] = w;
    }
  }
  Not_symnum = str_to_sn(NOT_SYM, 1); /* Cache the symnum for neg literals. */
}  /* init_weight_scheme */

/*************
 *
 *   term_weight()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int term_weight(Term t)
{
  if (VARIABLE(t))
    return Variable_weight;
  else {
    int w = Weights[SYMNUM(t)];
    int i;
    for (i = 0; i < ARITY(t); i++)
      w += term_weight(ARG(t,i));
    return w;
  }
}  /* term_weight */

/*************
 *
 *   clause_weight()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
int clause_weight(Clause c)
{
  Literal l;
  int w = 0;
  for (l = c->literals; l; l = l->next) {
    if (l->sign == FALSE)
      w += Weights[Not_symnum];
    w += term_weight(l->atom);
  }
  return w;
}  /* clause_weight */

/*************
 *
 *   set_weights_by_arity()
 *
 *************/

/* DOCUMENTATION
Set weights of all symbols of the given arity to the specified value.
*/

/* PUBLIC */
void set_weights_by_arity(int arity, int weight)
{
  int i;
  for (i = 0; i < Number_of_weights; i++) {
    /* It's OK if symbol i doesn't exist: sn_to_arity will return -1. */
    if (sn_to_arity(i) == arity)
      Weights[i] = weight;
  }
}  /* set_weights_by_arity */

/*************
 *
 *   set_variable_weight()
 *
 *************/

/* DOCUMENTATION
Set the weight of all variables to the specified value.
*/

/* PUBLIC */
void set_variable_weight(int weight)
{
  Variable_weight = weight;
}  /* set_variable_weight */

