#include "mace4.h"
#include <unistd.h>

/*************
 *
 *   member_args()
 *
 *************/

BOOL member_args(int argc, char **argv, char *str)
{
  int i;
  for (i = 1; i < argc; i++)
    if (str_ident(argv[i], str))
      return TRUE;
  return FALSE;
}  /* member_args */

/*************
 *
 *   command_line_flag()
 *
 *************/

static
void command_line_flag(int id, char *optarg)
{
  if (!optarg || str_ident(optarg, "1")) {
    set_flag(id);
    printf("%% From the command line: set(%s).\n", flag_id_to_str(id));
  }
  else if (str_ident(optarg, "0")) {
    clear_flag(id);
    printf("%% From the command line: clear(%s).\n", flag_id_to_str(id));
  }
  else
    fatal_error("Command-line flag must have 0 or 1 as the value");
}  /* command_line_flag */

/*************
 *
 *   command_line_parm()
 *
 *************/

static
void command_line_parm(int id, char *optarg)
{
  int n = atoi(optarg);
  assign_parm(id, n);
  printf("%% From the command line: assign(%s, %d).\n", parm_id_to_str(id),n);
}  /* command_line_parm */

/*************
 *
 *   usage_message()
 *
 *************/

void usage_message(FILE *fp)
{
  fprintf(fp,

  "%s %s (%s) -- Search for finite models.\n"
  "\n"
  "Input commands, clauses, and formulas are taken from standard input.\n"
  "\n"
  "Command-line options override any settings in the input file.\n"
  "To set or clear a flag, you must give 1 or 0 as the value.\n"
  "\n"
  "Basic Options\n"
  "\n"
  "  -n n : (lower case) parm domain_size (%d).\n"
  "  -N n : (upper case) parm iterate_up_to (%d).\n"
  "  -i n : (lower case) parm increment (%d).\n"
  "  -p n : (lower case) flag print_models (%s).\n"
  "  -P n : (upper case) flag print_models_portable (%s).\n"
  "  -m n : parm max_models (%d).\n"
  "  -t n : parm max_seconds (%d).\n"
  "  -b n : parm max_megs (%d).\n"
  "  -V n : (upper case) flag prolog_style_variables (%s).\n"
  "  -v n : (lower case) flag verbose (%s).\n"
  "\n"
  "Advanced Options\n"
  "\n"
  "  -L n : flag lnh (%s).\n"
  "  -O n : parm selection_order (%d).\n"
  "  -M n : parm Selection_measure (%d).\n"
  "  -G n : flag negprop (%s).\n"
  "  -H n : flag neg_assign (%s).\n"
  "  -I n : flag neg_assign_near (%s).\n"
  "  -J n : flag neg_elim (%s).\n"
  "  -K n : flag neg_elim_near (%s).\n"
  "  -T n : flag trace (%s).\n"
  "  -R n : flag integer_ring (%s).\n"
  "  -q n : flag iterate_primes (%s).\n"
  "\n"
  "Special Flags (not corresponding to set/clear/assign commands)\n"
  "\n"
  "  -c   : Ignore unrecognized set/clear/assign commands in the input\n"
  "         file.  This is useful for running MACE4 on an input file\n"
  "         designed for another program such as a theorem prover.\n"
  "\n"

	  , PROGRAM_NAME, PROGRAM_VERSION, PROGRAM_DATE,
	  parm(DOMAIN_SIZE),
	  parm(ITERATE_UP_TO),
	  parm(INCREMENT),
	  flag(PRINT_MODELS) ? "set" : "clear",
	  flag(PRINT_MODELS_PORTABLE) ? "set" : "clear",
	  parm(MAX_MODELS),
	  parm(MAX_SECONDS),
	  parm(MAX_MEGS),
	  flag(PROLOG_STYLE_VARIABLES) ? "set" : "clear",
	  flag(VERBOSE) ? "set" : "clear",
	  flag(LNH) ? "set" : "clear",
	  parm(SELECTION_ORDER),
	  parm(SELECTION_MEASURE),
	  flag(NEGPROP) ? "set" : "clear",
	  flag(NEG_ASSIGN) ? "set" : "clear",
	  flag(NEG_ASSIGN_NEAR) ? "set" : "clear",
	  flag(NEG_ELIM) ? "set" : "clear",
	  flag(NEG_ELIM_NEAR) ? "set" : "clear",
	  flag(TRACE) ? "set" : "clear",
	  flag(INTEGER_RING) ? "set" : "clear",
	  flag(ITERATE_PRIMES) ? "set" : "clear"
	  );
}  /* usage_message */

/*************
 *
 *   process_command_line_args()
 *
 *************/

void process_command_line_args(int argc, char **argv)
{
  extern char *optarg;
  int c;

  /* No colons:  no argument.
     One colon:  argument required.
     Two colons: argument optional.  (GNU extension!  Don't use it!)
  */
  
  while ((c = getopt(argc, argv,
	     "n:N:m:t:b:O:M:p:P:v:V:L:G:H:I:J:K:T:R:i:q:c")) != EOF) {
    switch (c) {
    case 'n':
      command_line_parm(DOMAIN_SIZE, optarg);
      break;
    case 'N':
      command_line_parm(ITERATE_UP_TO, optarg);
      break;
    case 'm':
      command_line_parm(MAX_MODELS, optarg);
      break;
    case 't':
      command_line_parm(MAX_SECONDS, optarg);
      break;
    case 'b':
      command_line_parm(MAX_MEGS, optarg);
      break;
    case 'O':
      command_line_parm(SELECTION_ORDER, optarg);
      break;
    case 'M':
      command_line_parm(SELECTION_MEASURE, optarg);
      break;

    case 'p':
      command_line_flag(PRINT_MODELS, optarg);
      break;
    case 'P':
      command_line_flag(PRINT_MODELS_PORTABLE, optarg);
      break;
    case 'V':
      command_line_flag(PROLOG_STYLE_VARIABLES, optarg);
      break;
    case 'v':
      command_line_flag(VERBOSE, optarg);
      break;
    case 'L':
      command_line_flag(LNH, optarg);
      break;
    case 'G':
      command_line_flag(NEGPROP, optarg);
      break;
    case 'H':
      command_line_flag(NEG_ASSIGN, optarg);
      break;
    case 'I':
      command_line_flag(NEG_ASSIGN_NEAR, optarg);
      break;
    case 'J':
      command_line_flag(NEG_ELIM, optarg);
      break;
    case 'K':
      command_line_flag(NEG_ELIM_NEAR, optarg);
      break;
    case 'T':
      command_line_flag(TRACE, optarg);
      break;
    case 'R':
      command_line_flag(INTEGER_RING, optarg);
      break;
    case 'i':
      command_line_parm(INCREMENT, optarg);
      break;
    case 'q':
      command_line_flag(ITERATE_PRIMES, optarg);
      break;
    case 'c':
      /* do nothing---this is handled */
      break;

    case '?':
    default:
      usage_message(stderr);
      fatal_error("unrecognized command-line option or missing value (flags take 0 or 1)");
    }
  }
}  /* process_command_line_args */

