#include "estack.h"

/* Estack */

struct estack {
  void **p;
  void *v;
  Estack next;
};

static Estack estack_avail;
static unsigned estack_gets, estack_frees, estack_avails;

/*************
 *
 *    Estack get_estack()
 *
 *************/

static
Estack get_estack(void)
{
  Estack p;

  estack_gets++;
  if (estack_avail == NULL)
    p = tp_alloc(sizeof(struct estack));
  else {
    estack_avails--;
    p = estack_avail;
    estack_avail = estack_avail->next;
  }

  /* initialization */	
#if 0
  p->p = NULL;
  p->v = NULL;
  p->next = NULL;
#endif

  return(p);
}  /* get_estack */

/*************
 *
 *    free_estack()
 *
 *************/

#if 0
static
void free_estack(Estack p)
{
  estack_frees++;
  estack_avails++;
  p->next = estack_avail;
  estack_avail = p;
}  /* free_estack */
#endif

/*************
 *
 *   fprint_estack_mem()
 *
 *************/

void fprint_estack_mem(FILE *fp, int heading)
{
  int n;
  if (heading)
    fprintf(fp, "  type (bytes each)        gets      frees     in use      avail      bytes\n");

  n = sizeof(struct estack);
  fprintf(fp, "estack (%4d)       %11u%11u%11u%11u%9.1f K\n",
	  n, estack_gets, estack_frees,
	  estack_gets - estack_frees, estack_avails,
	  (((estack_gets - estack_frees) + estack_avails) * n) / 1024.);

  /* end of printing for each type */
  
}  /* fprint_estack_mem */

/*************
 *
 *   p_estack_mem()
 *
 *************/

void p_estack_mem()
{
  fprint_estack_mem(stdout, 1);
}  /* p_estack_mem */

/*************
 *
 *   update_and_push()
 *
 *************/

Estack update_and_push(void **p, void *new, Estack stack)
{
  Estack s = get_estack();  /* allocate a stack entry */

  s->p = p;                 /* record the assignment */
  s->v = *p;
  *p = new;                 /* make the assignment */
  s->next = stack;
  return s;                 /* return the updated stack */
}  /* update_and_push */

/*************
 *
 *   restore_from_stack()
 *
 *************/

void restore_from_stack(Estack stack)
{
#if 0
  /* This recursive version is prettier, but it can take 20% longer overall. */
  if (stack != NULL) {
    *(stack->p) = stack->v;  /* restore */
    restore_from_stack(stack->next);
    free_estack(stack);
  }
#else
  /* nonrecursive version */
  if (stack != NULL) {
    Estack curr, prev;
    int n = 0;
    for (curr = stack; curr != NULL; curr = curr->next) {
      *(curr->p) = curr->v;  /* restore */
      prev = curr;
      n++;
    }
    prev->next = estack_avail;
    estack_avail = stack;
    estack_frees += n;
    estack_avails += n;
  }
#endif  
}  /* restore_from_stack */

/*************
 *
 *   zap_estack()
 *
 *************/

void zap_estack(Estack s)
{
#if 0
  /* This recursive version can cause a system (unix) stack overflow,
     because these estacks can get VERY big. */
  if (s != NULL) {
    zap_estack(s->next);
    free_estack(s);
  }
#else
  /* nonrecursive version */
  if (s != NULL) {
    Estack t;
    int n = 1;
    for (t = s; t->next != NULL; t = t->next)
      n++;
    t->next = estack_avail;
    estack_avail = s;
    estack_frees += n;
    estack_avails += n;
  }
#endif  
}  /* zap_estack */
