/*********************************************** OPS includes */

#define TERM_VP_FIELD

#include "../ladr/header.h"
#include "../ladr/commands.h" /* includes most of what we need */
#include "../ladr/clausify.h"
#include "../ladr/clock.h"
#include "../ladr/nonport.h"

/*********************************************** local includes */

#include "estack.h"
#include "mstate.h"
#include "syms.h"
#include "ground.h"
#include "propagate.h"

/*********************************************** macro definitions */

#define MAX(a,b)  ((a) > (b) ? a : b)
#define MIN(a,b)  ((a) < (b) ? a : b)

#define PROGRAM_NAME     "MACE4"
#define PROGRAM_VERSION  "2004-D"
#define PROGRAM_DATE     "May 26, 2004"

/* Parmaters */

enum {
  DOMAIN_SIZE,
  MAX_MODELS,
  MAX_SECONDS,
  ITERATE_UP_TO,
  SELECTION_ORDER,
  SELECTION_MEASURE,
  MAX_MEGS,
  /* added after 2003-B */
  INCREMENT
};

/* Flags */

enum {
  PRINT_MODELS,
  PRINT_MODELS_PORTABLE,
  LNH,
  TRACE,
  NEGPROP,
  NEG_ASSIGN,
  NEG_ASSIGN_NEAR,
  NEG_ELIM,
  NEG_ELIM_NEAR,
  PROLOG_STYLE_VARIABLES,
  VERBOSE,
  /* added after 2003-B */
  INTEGER_RING,
  ITERATE_PRIMES
};

/* Exit codes (fatal_error exit is 1, memory limit is fatal_error) */

#define MAX_MODELS_EXIT    0
#define EXHAUSTED_EXIT     2
#define ALL_MODELS_EXIT    3
#define MAX_SEC_YES_EXIT   4
#define MAX_SEC_NO_EXIT    5

/*********************************************** structure declarations */

/* A cell is member of a function or relation table.  For example,
   if we have a group operation and the domain is size 6, then
   the term f(2,3) represents one cell of the group multiplication
   table.  We will dynamically allocate a (1-dimensional) array
   of cells (to hold all tables) at the beginning of the run after
   we know how many we'll need.  The array will be indexed by IDs
   that are consructed from the terms representing the cells.
   (Each symbol has a "base" value, and the cell ID is an offset
   from the base.)  We have, in effect, multidimensional arrays
   within our main array of cells, we calculate the IDs from
   the symbol base and the term arguments.  Anyway, given a term
   like f(2,3), in which all of the arguments are domain members,
   we can quickly get to the corresponding cell.
*/

struct cell {
  int id;
  Term eterm;          /* the term representation, e.g., f(2,3) */
  Term value;          /* current value of cell (domain element or NULL) */
  Term occurrences;    /* current occurrences of the term */
  Term *possible;      /* current set of possible values */
  int max_index;       /* maximum index for this cell */
  Symbol_data symbol;  /* data on the function or relation symbol */
};

/* These macros take you to the correct cell in the Cells table.
   That is, given the base for the symbol and the indexes, get the ID.
*/

#define X0(b)         (b)
#define X1(b,i)       (b) + (i)
#define X2(b,i,j)     (b) + (i)*Domain_size + (j)
#define X3(b,i,j,k)   (b) + (i)*Domain_size*Domain_size + (j)*Domain_size + (k)
#define X4(b,i,j,k,l) (b) + (i)*Domain_size*Domain_size*Domain_size + (j)*Domain_size*Domain_size + (k)*Domain_size + (l)

/*********************************************** function prototypes */

/* from mace4.c */

void mace4_exit(int exit_code);

/* from util.c */

void random_permutation(int *a, int n);
int int_power(int n, int exp);
BOOL prime(int n);

/* from print.c */

void print_model_portable(FILE *fp, BOOL print_head);
void p_model(BOOL print_head);
void p_matom(Term atom);
void p_mclause(Mclause c);
void p_eterms(void);
void p_stats(void);
void p_mem(void);
void reset_current_stats(void);

/* from select.c */

int select_cell(int max_constrained);

/* from negpropindex.c */

void init_negprop_index(void);
void free_negprop_index(void);
void p_negprop_index(void);
void insert_negprop_index(Term atom);
void negprop_n_index(Term atom);
void insert_negprop_eq(Term atom, Term alpha, int val, Mstate state);
void insert_negprop_noneq(Term atom, Mstate state);
BOOL nterm(Term t, int *ppos, int *pid);
Term negprop_find_near(int sign, int sn, int val, Term query, int pos);

/* from negprop.c */

void propagate_negative(int type, int id, Term alpha, Term beta, int pos,
			Mstate state);

/* from ordercells.c */

void order_cells(void);

/* from commandline.c */

BOOL member_args(int argc, char **argv, char *str);
void usage_message(FILE *fp);
void process_command_line_args(int argc, char **argv);
