#include "mace4.h"

extern Symbol_data Symbols;

extern int Number_of_cells;
extern struct cell *Cells;

extern Term *Domain;
extern int Domain_size;

extern int Negation_flag;
extern int Eq_sn;

/* Statistics for entire run */

extern unsigned All_models;

/* Statistics for curent domain size */

extern int Mace4_clock;

extern unsigned Current_models;
extern unsigned Selections;
extern unsigned Assignments;
extern unsigned Propagations;
extern unsigned Cross_offs;
extern unsigned Rewrite_terms;
extern unsigned Rewrite_bools;
extern unsigned Indexes;
extern unsigned Ground_clauses_seen;
extern unsigned Ground_clauses_kept;
extern unsigned Rules_from_neg;

extern unsigned Neg_elim_attempts;
extern unsigned Neg_elim;
extern unsigned Neg_elim_agone;
extern unsigned Neg_elim_egone;

extern unsigned Neg_assign_attempts;
extern unsigned Neg_assign;
extern unsigned Neg_assign_agone;
extern unsigned Neg_assign_egone;

extern unsigned Neg_near_assign_attempts;
extern unsigned Neg_near_assign;
extern unsigned Neg_near_assign_agone;
extern unsigned Neg_near_assign_egone;

extern unsigned Neg_near_elim_attempts;
extern unsigned Neg_near_elim;
extern unsigned Neg_near_elim_agone;
extern unsigned Neg_near_elim_egone;

/*************
 *
 *   f[01234]_val()
 *
 *************/

static int id2val(int id)
{
  return (Cells[id].value == NULL ? -1 : VARNUM(Cells[id].value));
}  /* id2val */

static
int f0_val(int base)
{
  int id = X0(base);
  return id2val(id);
}  /* f0_val */

static
int f1_val(int base, int i)
{
  int id = X1(base,i);
  return id2val(id);
}  /* f1_val */

static
int f2_val(int base, int i, int j)
{
  int id = X2(base,i,j);
  return id2val(id);
}  /* f2_val */

/*************
 *
 *   p_model()
 *
 *************/

void p_model(BOOL print_head)
{
  Symbol_data p;
  int n = Domain_size;

  if (print_head)
    printf("\n-------- Model %d at %.2f seconds --------\n",
	   All_models,
	   user_time() / 1000.0);
  
  for (p = Symbols; p != NULL; p = p->next) {
    char *name = sn_to_str(p->sn);
    if (p->attribute != EQUALITY_SYMBOL) {
      /* This prints both relations and functions. */
      if (p->arity == 0) {
	int v = f0_val(p->base);
	if (v < 0)
	  printf("\n %s : -\n", name);
	else
	  printf("\n %s : %d\n", name, v);
      }
      else if (p->arity == 1) {
	char *s1 = "%2d";
	char *s2 = "--";
	char *s3 = " -";
	int i;
	for (i = 0; i < n; i++) {
	  printf("\n %s :\n", name);
	  printf("       ");
	  for (i = 0; i < n; i++)
	    printf(s1, i);
	  printf("\n    ---");
	  for (i = 0; i < n; i++)
	    printf(s2);
	  printf("\n       ");
	  for (i = 0; i < n; i++) {
	    int v = f1_val(p->base, i);
	    if (v < 0)
	      printf(s3);
	    else
	      printf(s1, v);
	  }
	  printf("\n");	  
	}
      }
      else if (p->arity == 2) {
	char *s1 = "%2d";
	char *s2 = "--";
	char *s3 = " -";
	int i, j;
	printf("\n %s :\n", name);
	printf("      |");
	for (i = 0; i < n; i++)
	  printf(s1, i);
	printf("\n    --+");
	for (i = 0; i < n; i++)
	  printf(s2);
	printf("\n");

	for (i = 0; i < n; i++) {
	  printf("%5d |", i);
	  for (j = 0; j < n; j++) {
	    int v = f2_val(p->base, i, j);
	    if (v < 0)
	      printf(s3);
	    else
	      printf(s1, v);
	  }
	  printf("\n");
	}
      }
      else {
	int n = int_power(Domain_size, p->arity);
	int i;
	printf("Sorry, no pretty printing yet for arity %d.\n", p->arity);
	for (i = 0; i < n; i++) {
	  int id = p->base + i;
	  fwrite_term(stdout, Cells[id].eterm);
	  if (Cells[id].value == NULL)
	    printf(" = -.\n");
	  else
	    printf(" = %d.\n", VARNUM(Cells[id].value));
	}
      }
    }
  }
  if (print_head)
    printf("\n-------- end of model --------\n");

}  /* p_model */

/*************
 *
 *   print_model_portable()
 *
 *************/

void print_model_portable(FILE *fp, BOOL print_head)
{
  int syms_printed;
  Symbol_data s;

  if (print_head)
    fprintf(fp, "\n-------- Model %d at %.2f seconds --------\n",
	    All_models,
	    user_time() / 1000.0);
  
  fprintf(fp, "interpretation( %d, [\n", Domain_size);

  syms_printed = 0;

  for (s = Symbols; s != NULL; s = s->next) {
    if (s->attribute != EQUALITY_SYMBOL) {
      int i, n;
      if (syms_printed > 0)
	fprintf(fp, ",\n");

      fprintf(fp, "\n        %s(%s%s",
	      s->type == FUNCTION ? "function" : "relation",
	      sn_to_str(s->sn),
	      s->arity == 0 ? "" : "(_");
      for (i = 1; i < s->arity; i++)
	fprintf(fp, ",_");
      fprintf(fp,"%s, [", s->arity == 0 ? "" : ")");
      n = int_power(Domain_size, s->arity);
      for (i = 0; i < n; i++) {
	int id = s->base + i;
	if (Cells[id].value == NULL)
	  fprintf(fp, "-");
	else
	  fprintf(fp, "%d", VARNUM(Cells[id].value));
	if (i < n-1)
	  fprintf(fp, ",%s", (i+1) % Domain_size == 0 ? "\n\t\t\t  " : "");
	else
	  fprintf(fp, "])");
      }
      syms_printed++;
    }
  }

  fprintf(fp, "\n]).\n");

  if (print_head)
    fprintf(fp, "\n-------- end of model --------\n");
}  /* print_model_portable */

/*************
 *
 *   p_matom()
 *
 *************/

void p_matom(Term atom)
{
  if (atom == NULL)
    printf("(NULL)");
  else if (!NEGATED(atom))
    fwrite_term(stdout, atom);
  else if (EQ_TERM(atom)){
    fwrite_term(stdout, ARG(atom,0));
    printf(" != ");
    fwrite_term(stdout, ARG(atom,1));
  }
  else {
    printf("~(");
    fwrite_term(stdout, atom);
    printf(")");
  }
  printf(".\n");
}  /* p_matom */

/*************
 *
 *   p_mclause()
 *
 *************/

void p_mclause(Mclause c)
{
  int i;
  printf("numlits=%d, active=%d, subsumed=%d: ",
	 c->numlits, c->active, c->subsumed);
  for (i = 0; i < c->numlits; i++) {
    Term atom = LIT(c,i);
    if (!NEGATED(atom))
      fwrite_term(stdout, atom);
    else {
      printf("~(");
      fwrite_term(stdout, atom);
      printf(")");
    }
    if (i < c->numlits-1)
      printf(" | ");
    else
      printf(".\n");
  }
}  /* p_mclause */

/*************
 *
 *   p_eterms()
 *
 *************/

static
int eterms_count(Term t)
{
  return (t == NULL ? 0 : 1 + eterms_count(t->vp));
}

void p_eterms(void)
{
  int i, j;
  printf("\n------- Cells --------\n");
  for (i = 0; i < Number_of_cells; i++) {
    int n = eterms_count(Cells[i].occurrences);
    if (n > 0) {
      fwrite_term(stdout, Cells[i].occurrences);
      printf(": %d occ, id=%d, val=", n, i);
      if (Cells[i].value == NULL)
	printf("NULL");
      else
	fwrite_term(stdout, Cells[i].value);
      printf(", pvals=");
      for (j = 0; j < id_to_domain_size(i); j++) {
	if (Cells[i].possible[j] == NULL)
	  printf(" -");
	else
	  printf("%2d", j);
      }
      printf("\n");
    }
  }
}  /* p_eterms */

/*************
 *
 *   p_stats()
 *
 *************/

void p_stats(void)
{
  printf("\n--------- statistics for domain size %d ---------\n\n",Domain_size);

  printf("Current CPU time: %.2f seconds ", clock_value(Mace4_clock) / 1000.0);
  printf("(total CPU time: %.2f seconds).\n",user_time() / 1000.0);
  printf("Ground clauses: seen=%u, kept=%u.\n",
	 Ground_clauses_seen, Ground_clauses_kept);
  printf("Selections=%u, assignments=%u, propagations=%u, current_models=%u.\n",
	 Selections, Assignments, Propagations, Current_models);
  printf("Rewrite_terms=%u, Rewrite_bools=%u, indexes=%u.\n",
	 Rewrite_terms, Rewrite_bools, Indexes);
  printf("Rules from negative clauses: %u, Cross_offs=%d.\n",
	 Rules_from_neg, Cross_offs);
  printf("Negative propagation:\n");
  printf("                 successes   attempts      agone      egone\n");
  printf("Neg_elim        %10u %10u %10u %10u\n",
	 Neg_elim, Neg_elim_attempts, Neg_elim_agone, Neg_elim_egone);
  printf("Neg_assign      %10u %10u %10u %10u\n",
	 Neg_assign, Neg_assign_attempts, Neg_assign_agone, Neg_assign_egone);
  printf("Neg_near_elim   %10u %10u %10u %10u\n",
	 Neg_near_elim, Neg_near_elim_attempts, Neg_near_elim_agone, Neg_near_elim_egone);
  printf("Neg_near_assign %10u %10u %10u %10u\n",
	 Neg_near_assign, Neg_near_assign_attempts, Neg_near_assign_agone, Neg_near_assign_egone);

}  /* p_stats */

/*************
 *
 *   p_mem()
 *
 *************/

void p_mem(void)
{
  printf("\n------------- memory usage (for entire run) -------------------\n");

  printf("\nTotal malloced: %d megabytes\n", megs_malloced());

  fprint_strbuf_mem(stdout, 1);
  fprint_parse_mem(stdout, 0);
  fprint_glist_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_mclause_mem(stdout, 0);
  fprint_mstate_mem(stdout, 0);
  fprint_estack_mem(stdout, 0);
}  /* p_mem */

/*************
 *
 *   reset_current_stats()
 *
 *************/

void reset_current_stats(void)
{
  Current_models = 0;
  Selections = 0;
  Assignments = 0;
  Propagations = 0;
  Cross_offs = 0;
  Rewrite_terms = 0;
  Rewrite_bools = 0;
  Indexes = 0;
  Ground_clauses_seen = 0;
  Ground_clauses_kept = 0;
  Rules_from_neg = 0;

  Neg_elim_attempts = 0;
  Neg_elim = 0;
  Neg_elim_agone = 0;
  Neg_elim_egone = 0;

  Neg_assign_attempts = 0;
  Neg_assign = 0;
  Neg_assign_agone = 0;
  Neg_assign_egone = 0;

  Neg_near_assign_attempts = 0;
  Neg_near_assign = 0;
  Neg_near_assign_agone = 0;
  Neg_near_assign_egone = 0;

  Neg_near_elim_attempts = 0;
  Neg_near_elim = 0;
  Neg_near_elim_agone = 0;
  Neg_near_elim_egone = 0;
}  /* reset_current_stats */

