/*----------------------------------------------------------*/
/* modify.c (Version for Theo)                              */     
/*                                                          */
/* Added December 2002                                      */
/*                                                          */
/* Controls the modification of clauses                     */
/*     determine_potential_changes().............#1         */
/*     modify_clause()...........................#2         */
/*     form_weakened_clause()....................#3a        */
/*     mark_questionable().......................#3         */
/*     print_change_array()......................#4         */
/*     reconsider_proof()........................#5         */
/*     reinstate_base_clause()...................#6         */
/*     bres_instantiated_proof().................#7         */
/*     bfac_instantiated_proof().................#7a        */
/*     bsub_instantiated_proof().................#7b        */
/*     filter_ulist()............................#8         */
/*     skip_equal_all_vars().....................#9         */
/*     skip_equal_first_last()...................#10        */
/*                                                          */
/*----------------------------------------------------------*/

/*----------------------------------------------------------*/
/* Calling of functions                                     */
/*     modify_clause()...........................#2         */
/*        determine_potential_changes()..........#1         */
/*           skip_equal_all_vars()...............#9         */
/*           skip_equal_first_last()............#10         */
/*        form_weakened_clause()................#3a         */
/*        print_change_array()...................#4         */
/*     mark_questionable().......................#3         */
/*     reconsider_proof()........................#5         */
/*        reinstate_base_clause()................#6         */
/*        bres_instantiated_proof()..............#7         */
/*        bfac_instantiated_proof()..............#7a        */
/*        bsub_instantiated_proof()..............#7b        */
/*        filter_ulist().........................#8         */
/*                                                          */
/*----------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "const.h"
#include "clause.h"

/*-----------------------------------------------------------*/
/* externally defined functions                              */
/*-----------------------------------------------------------*/

extern void print_clause();
extern int grab_number();
extern void standardize();
extern void clear_var_substitution_tble();
extern int unify();
extern clause *generate_iclause();
extern void calc_clause_hashcode();
extern int time_now();
extern void clean_var_names();
extern void replace_vars();
extern void copy_clause3();
extern void normalize_literal();
		   
/*-----------------------------------------------------------*/
/* externally defined constants                              */
/*-----------------------------------------------------------*/

extern char *OutPtr;
extern int ply, nextu;
extern int base, org_base, nc;
extern int batch_tried;
extern int o_maxvar;
extern int o_debug;
extern int o_learn, weakened_ply, weakened_clause,o_weakened_index;
extern char *weak_proof_begins_here, *weak_proof_ends_here;
extern int satisfiable;
extern int equal_pred;
extern clause *clist[];
extern clause *ulist[];
extern int u_space;
extern char input_buffer[];
extern int cl[PROOF_ARRAY_SIZE][14];
extern int equal_xyyzxz;
extern dnode depth[DEPTH_PTR_SPACE];
extern int cade_format;
extern int proof_found;
extern int spawned;

/*-----------------------------------------------------------*/
/* internally defined functions                              */
/*-----------------------------------------------------------*/

void determine_potential_changes();
void modify_clause();
void mark_questionable();
void print_change_array();
void reconsider_proof();
void form_weakened_clause();
void reinstate_base_clause();
int bres_instantiated_proof();
int bfac_instantiated_proof();
int bsub_instantiated_proof();
void filter_ulist();
int skip_equal_all_vars();
int skip_equal_first_last();

/*-----------------------------------------------------------*/
/* internally defined constants                              */
/*-----------------------------------------------------------*/

char clause_number[5];
char literal_number[5];
char char_number[5];
char old_char[80];
char new_char[5];
int replaced_char;
int deleted_terms[200];
int questionable[2000];
int total_questionable;
int excellent_weakened_proof;

/*------------------------------------------------------------*/
/* change_array:                                              */
/*         column 0:  batch_tried                             */
/*         column 1:  clause number                           */
/*         column 2:  literal number                          */
/*         column 3:  position in literal                     */
/*         column 4:  substitute (variable)                   */
/*         cols 5-9:  characters substituted for              */
/*------------------------------------------------------------*/

int change_array[50][10];
int change_level;

/*------------------------------------------------------------*/
/* potential_change_array:                                    */
/*          column 0: clause                                  */
/*          column 1: literal                                 */
/*          column 2: position                                */
/*          column 3: char                                    */
/*          column 4: #var in clause                          */
/*          column 5: substituted char.                       */
/*------------------------------------------------------------*/

int p_change_array[2000][7];
int p_change_number;
int randy;

/*------------------------------------------------------------*/
/* determine_potential_changes() ..................#1         */
/*                                                            */
/* called by:                                                 */
/*                                                            */
/* calls:   skip_equal_all_vars()                             */
/*          skip_equal_first_last()                           */
/*          tdump.c/print_clause()                            */
/*                                                            */
/* determines a list of potential changes that can be made to */
/* the base clauses to weaken them.  At most 2000 changes can */
/* be listed on potential_change_array[][].                   */
/*------------------------------------------------------------*/

void determine_potential_changes()
{
int i, j, jj, k, kk, a, *p, v_cnt, position, litl_index;
int jjtemp, var_cnt[64], base_print[2000];
literal *lit;
clause *c;
int second_half;

/*--------------------------------------------------*/
/* Heuristic function to weight weakeneings:        */
/*     1: Constant in set-of-support                */
/*     2: function in set of support                */
/*     3: literal in set-of-support                 */
/*     4: Constant in axioms                        */
/*     5: Function in axioms                        */
/*     6: Literal in axioms                         */
/*--------------------------------------------------*/

/*--------------------------------------------------*/
/* List positions where a constant can be replaced  */
/*--------------------------------------------------*/

for(i = 0; i < 2000; i++) {
   for(j = 0; j < 7; j++) {
      p_change_array[i][j] = 0;}}

k = -1;
for(i = base; i >= 0; i--) {

/*--------------------------------------*/
/* for base clauses                     */
/*--------------------------------------*/

   for(j = 0; j < 64; j++) {var_cnt[j] = 0;}
   c = clist[i];
   if(c->c_bits & EQCLS) {continue;}
   if(c->c_nlit == 1 && skip_equal_all_vars(c)) {continue;}
   v_cnt = 0;
   kk = k;
   litl_index = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      p = &lit->l_id;
      litl_index++;position = second_half = 0;
      for(a = *p++ & ARGMASK; a > 0; a--) {
	 position++;
	 if(a == 1 && second_half == 0) {
	    second_half = position;}
         if(*p & FCNMASK) {
            if((*p & ARGMASK) == 0) {
               if(!skip_equal_first_last(c,p)) {

/*----------------------------------------*/
/* the replacement of any constant by a   */
/* variable is a potential change.        */
/* for clauses in the set-of-support      */
/*----------------------------------------*/

                  if(k < 1999) {
		     k++;
	             p_change_array[k][0] = c->c_id;
	             p_change_array[k][1] = litl_index;
	             p_change_array[k][2] = position;
                     p_change_array[k][3] = *p;
		     if(c->c_id >= base - nc + 1) {
                        p_change_array[k][6] = 100000 - position;}
                     else p_change_array[k][6] = 400000 - position;}}}
	    else {

/*---------------------------------------*/
/* replace function with variable        */
/* but..... first ELIMINATE weakenings   */
/* for unit clauses                      */
/* where pred is ~equal(f....,g....)     */
/* and weakening replaces f with var     */
/* or g with a var, as in these cases    */
/* weakening won't help much!!.          */
/*---------------------------------------*/

               if(lit->l_id == (equal_pred | FCNMASK) && c->c_nlit == 1) { 
	           if(position == 1 || position == second_half) { 
		       goto skip_saving;}}

	       if(k < 1999) {
                  k++;
	          p_change_array[k][0] = c->c_id;
	          p_change_array[k][1] = litl_index;
	          p_change_array[k][2] = position;
	          p_change_array[k][3] = *p;
		  if(c->c_id >= base - nc + 1) {
                     p_change_array[k][6] = 200000 - position;}
                  else p_change_array[k][6] = 500000 - position;}}
skip_saving:
            a += *p++ & ARGMASK;}
         else {
	    if(var_cnt[*p + 1] == 0) {
	       var_cnt[*p + 1] = *p + 1;
	       v_cnt++;}
            else {
	       if(k < 1999) {
	          k++;
		  p_change_array[k][0] = c->c_id;
		  p_change_array[k][1] = litl_index;
		  p_change_array[k][2] = position;
		  if(c->c_id >= base - nc) {
		     p_change_array[k][6] = 600000 - position;}
                  else {
		     if(litl_index == c->c_nlit) {
		        p_change_array[k][6] = 600000 - position;}
                     else {
		        p_change_array[k][6] = 600000 - position;}}}}
	   
            p++;}}}

   for(j = kk + 1; j <= k; j++) {
      p_change_array[j][4] = v_cnt;}}

/*-------------------------------------------------*/
/* list literals that can be deleted from clauses  */
/* with more than 2 literals.                      */
/*-------------------------------------------------*/

for(i = base; i >= 0; i--) {
   c = clist[i];
   if(c->c_bits & EQCLS) {continue;}
   litl_index = 0;
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      litl_index++;
      if((litl_index == c->c_nlit && litl_index > 1) || (c->c_nlit > 1 && (i >= base - nc))) {
         if(k < 1999) {
            k++;
            p_change_array[k][0] = c->c_id;
            p_change_array[k][1] = litl_index;
            p_change_array[k][2] = -1;
            p_change_array[k][3] = -1;
            p_change_array[k][4] = c->c_nv;
            if(c->c_id >= base - nc + 1) {
               p_change_array[k][6] = 300000 - c->c_id * 100 - litl_index;}
            else p_change_array[k][6] = 600000 - c->c_id * 100 - litl_index;}}}}

if(p_change_array[k][0] == 0 && k >= 0) k--;

/*-----------------------------------------*/
/* sort based on heuristic score           */
/*-----------------------------------------*/

if( k != -1) {
   for(i = 0; i < k; i++) {
      for(j = i + 1; j <= k; j++) {
         if(p_change_array[i][6] > p_change_array[j][6]) {
            for(jj = 0; jj < 7; jj++) {
               jjtemp = p_change_array[i][jj];
               p_change_array[i][jj] = p_change_array[j][jj];
               p_change_array[j][jj] = jjtemp;}}}}}

/*-------------------------------------------------*/
/* print potential changes                         */
/*-------------------------------------------------*/

if(cade_format == 0) {
   if(k == -1) {
      printf("\n\nPOTENTIAL CHANGE ARRAY: no modifications possible\n");
      p_change_number = 0;
      return;}
   else {
      printf("\n\nPOTENTIAL CHANGE ARRAY: %d modifications possible\n",k+1);}
      printf(" # Clause  Literal  Position  Contents  Variables Heuristic\n");}

for(i = 0; i < 2000; i++) base_print[i] = 0;

if(cade_format == 0) {
   for(i = 0;  i <= k; i++) {
      base_print[p_change_array[i][0]]++;
      printf("%2d %5d    %3d       %2d      %8x     %2d    %1d\n",
	 i,
	 p_change_array[i][0],
	 p_change_array[i][1],
	 p_change_array[i][2],
	 p_change_array[i][3],
	 p_change_array[i][4],
	 p_change_array[i][6]);}
   print0("\n\n");}

/*--------------------------------------------------*/
/* print clauses that can be modified               */
/*--------------------------------------------------*/

p_change_number = k + 1;
return;
}

/*------------------------------------------------------------*/
/* modify_clause() .................................#2        */
/*                                                            */
/* modifies a clause as desired                               */
/*                                                            */
/* called from: main()             theo.c                     */
/*                                                            */
/* calls:       determine_potential_changes()                 */
/*              print_clause()                                */
/*              grab_number()                                 */
/*              clean_var_names()                             */
/*                                                            */
/*------------------------------------------------------------*/

void modify_clause()
{
int f, i, j, k, kk, l, m, n, lit_end, nv;
int dt;                            /* index for deleted_terms */
char ichar;
int a, aa, *p, *pp, *ppp, position;
clause *c;
literal *lit;

/*---------------------------------------------*/
/* determine clause to modify                  */
/*---------------------------------------------*/

if(o_learn == 0) return;
if(weakened_ply == 1) {return;}

determine_potential_changes();

/*---------------------------------------------*/
/* if no weakenings possible, return           */
/*---------------------------------------------*/

if(p_change_number == 0) {
   weakened_ply = 1;
   return;}

/*---------------------------------------------*/
/* pick a weakening                            */
/*---------------------------------------------*/

if(o_learn == 1) {
 randy = time_now() % p_change_number; 
/*   randy = 0; */
   if(o_weakened_index != -1) {randy = o_weakened_index;}
   goto complete_decision;}

/*---------------------------------------------*/
/* if o_learn = 2, then                        */
/* the user decides how to weaken theorem      */
/*---------------------------------------------*/

printf("\nEnter weakened index number:");
i = 0;
while((ichar = getchar()) != '\n') {input_buffer[i++] = ichar;}
input_buffer[i] = '\n';
randy = grab_number(0);

/*----------------------------------------------*/
/* m is the integfer value of the new variable  */
/* determine pointer to literal to modify       */
/* and then modify literal: replace l with m    */
/*----------------------------------------------*/

complete_decision:
for(i = 0; i < p_change_number; i++) {
   for(k = 0; k < 200; k++) {deleted_terms[k] = 0;}
   j = p_change_array[i][0];
   c = clist[j - 1];
   form_weakened_clause(i);

/*-----------------------------------------------*/
/* mark clause j as "questionable"               */
/*-----------------------------------------------*/

   c->c_nv = p_change_array[i][4];
   change_array[change_level][0] = batch_tried;
   change_array[change_level][1] = j;
   change_array[change_level][2] = p_change_array[i][1];
   change_array[change_level][3] = p_change_array[i][2];
   change_array[change_level][4] = p_change_array[i][4];
   change_array[change_level][5] = replaced_char;
   reinstate_base_clause();}

final_change:
form_weakened_clause(randy);

j = p_change_array[randy][0];
questionable[j] = 1;
c = clist[j - 1];
c->c_bits |= NC;
c->c_nv = p_change_array[randy][4] + 1;
printf("\n%d ",randy);
print_clause(c);
printf("\n");
change_array[change_level][0] = batch_tried;
change_array[change_level][1] = j;
change_array[change_level][2] = p_change_array[randy][1];
change_array[change_level][3] = p_change_array[randy][2];
change_array[change_level][4] = p_change_array[randy][4];
change_array[change_level][5] = replaced_char;
clean_var_names(c);
//print_change_array();
return;
}

/*------------------------------------------------------------*/
/* form_weakened_clause() ........................#2a         */
/*                                                            */
/* mark clauses that were derived from modified clauses       */
/* and their descendants                                      */
/*                                                            */
/* called by:                                                 */
/*                                                            */
/* calls:     Nobody                                          */
/*                                                            */
/*------------------------------------------------------------*/

void form_weakened_clause(randy)
int randy;
{
int j, k, kk, l, f, m, nv, n, dt, aa, lit_end;
int position, a, *p, *pp, *ppp;
clause *c;
literal *lit;

j = p_change_array[randy][0];
k = p_change_array[randy][1];
l = p_change_array[randy][2];
f = p_change_array[randy][3];
m = p_change_array[randy][4];
nv = p_change_array[randy][4];
c = clist[j - 1];

/*----------------------------------------------*/
/* now carry out the weakening of the clause    */
/* first determine the literal to weaken.       */
/*----------------------------------------------*/

kk = k;
for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   if(--kk == 0) break;}

if(l != 0 && l != -1) {
   p = &lit->l_id;
   position = 1;
   if(nv > o_maxvar - 3) {o_maxvar++;}
   for(a = *p++ & ARGMASK; a > 0; a--) {
      if(*p & FCNMASK) {
         if(position++ == l) {
	    if((*p & ARGMASK) == 0) {

/*------------------------------------------*/
/* replace constant with a variable!        */
/*------------------------------------------*/

	       replaced_char = *p;*p = nv++;break;}
            else {

/*------------------------------------------*/
/* replace function with a variable!        */
/*------------------------------------------*/

               pp = p;
	       dt = 1;

/*------------------------------------*/
/* find end of function               */
/*------------------------------------*/

	       for(aa = *pp++ & ARGMASK; aa > 0; aa--) {
                  if(*pp & FCNMASK) {aa += *pp++ & ARGMASK;}
                  else {pp++;}}

/*------------------------------------*/
/* p points to fxn to be replaced     */
/* pp points to first term after fxn  */
/*------------------------------------*/

               if(lit->l_next != 0) lit_end = (int) lit->l_next;
	       else lit_end = (int) c->c_end;

	       for(ppp = p; (int) ppp < lit_end; ppp++) {
		  deleted_terms[dt++] = *ppp;}

               deleted_terms[0] = dt - 1;

/*-------------------------------------*/
/* find end of literal                 */
/*-------------------------------------*/

                replaced_char = *p;
                *p = nv++;
                p++;
                while((int) pp < (int) lit_end) {
                   *p = *pp;
                   p++;pp++;}
                break;}}
          a += *p++ & ARGMASK;}
       else {
          if(position++ == l) {
             replaced_char = *p;
             n = *p;*p = nv++;break;}
          p++;}}}

/*-------------------------------------*/
/* delete literal                      */
/*-------------------------------------*/

else if(l == -1) {
   replaced_char = 0;
   c->c_nlit--;
   lit->l_id = lit->l_id | ELIMLIT;}

/*-----------------------------------------------*/
/* mark clause j as "questionable"               */
/*-----------------------------------------------*/

return;
}

/*------------------------------------------------------------*/
/* mark_questionable() ............................#3         */
/*                                                            */
/* mark clauses that were derived from modified clauses       */
/* and their descendants                                      */
/*                                                            */
/* called by:                                                 */
/*                                                            */
/* calls:     Nobody                                          */
/*                                                            */
/*------------------------------------------------------------*/

void mark_questionable()
{
int i, lp, rp;

if(o_learn == 0) return;
if(weakened_ply == 1) {return;}

total_questionable = 0;

lp = rp = 0;
for(i = org_base; i <= base; i++) {
   if(clist[i]->c_lpar != 0) {
      lp = clist[i]->c_lpar->c_id;}
   if(clist[i]->c_rpar != 0) {
      rp = clist[i]->c_rpar->c_id;}
   if(questionable[lp] == 1) {questionable[i+1] = 1;total_questionable++;}
   if(questionable[rp] == 1) {questionable[i+1] = 1;total_questionable++;}}

for(i = 0;cl[i][0] != 0 ;i++) {
   lp = ((cl[i][0] >> 4) & PARENT_MASK);
   rp = ((cl[i][1] >> 4) & PARENT_MASK);
   if(questionable[lp] == 1) {questionable[base + 2 + i] = 1;total_questionable++;}
   if(questionable[rp] == 1) {questionable[base + 2 + i] = 1;total_questionable++;}}

return;
}

/*-----------------------------------------------------------*/
/* print_change_array() ........................#4           */
/*                                                           */
/* called by:                                                */
/*                                                           */
/* calls:      Nobody                                        */
/*                                                           */
/*-----------------------------------------------------------*/

void print_change_array()
{
int i;

printf("\nLEVEL  CLAUSE  LITERAL POSITION     NEW      OLD HEURISTIC\n");

printf("%3d   %5d %7d %8x  %8x    %8x\n",
      change_array[1][0],change_array[1][1],change_array[1][2],
      change_array[1][3],change_array[1][4],change_array[1][5]);
printf("\n");
}

/*------------------------------------------------------------*/
/* reconsider_proof() ............................#5          */
/*                                                            */
/* called by:                                                 */
/*                                                            */
/* calls:      reinstate_base_clause()                        */
/*             bres_instantiated_proof()                      */
/*             bfac_instantiated_proof()                      */
/*             bsub_instantiated_proof()                      */
/*             filter_ulist()                                 */
/*                                                            */
/* examine proof to see what can be used for next proof       */
/* (1) substitute constants, etc., back into original clauses.*/
/* (2) try proof with revised clauses.                        */
/* (3) add good clauses to ulist!                             */
/*------------------------------------------------------------*/

void reconsider_proof(proof_found)
int proof_found;
{
int i, j, jj, k, ui, clpar, crpar, clindex, crindex;
int c1, c2;                      /* index of clauses to resolve  */
int c1_index, c2_index;          /* index of literals to resolve */
int nextu0;

/*----------------------------------------*/
/* Undo changes to base clauses           */
/*----------------------------------------*/

if(o_learn == 0) return;
if(weakened_ply == 0) return;

nextu0 = nextu;

reinstate_base_clause();

if(satisfiable == 1) {
   printf("No proof sought; clauses are satisfiable!\n");
   fflush(stdout);
   nextu = 0;
   ulist[nextu] = (clause *) u_space;
   return;}

if(proof_found != 1) {
   printf("Proof not found; dont add clauses to ulist\n");
   fflush(stdout);
   nextu = 0;
   ulist[nextu] = (clause *) u_space;
   return;}

/*----------------------------------------*/
/* try proof with revised clauses         */
/* examine first Phase 0 clauses          */
/*----------------------------------------*/

for(i = org_base + 1; i <= base; i++) {
   if(!(clist[i]->c_bits & LIST)) {continue;}
   if(clist[i]->c_lpar == 0) {continue;}
   if(questionable[clist[i]->c_id] == 0) {continue;}
   clpar = clist[i]->c_lpar->c_id;
   crpar = clist[i]->c_rpar->c_id;
   clindex = clist[i]->c_indices >> 6;
   crindex = (clist[i]->c_indices & 0x0000003F);
   if(p_change_array[randy][0] == clpar && p_change_array[randy][1] <= clindex
      && p_change_array[randy][2] == -1) clindex++; 
   if(p_change_array[randy][0] == crpar && p_change_array[randy][1] <= crindex
      && p_change_array[randy][2] == -1) crindex++;
   if(bres_instantiated_proof(clpar,crpar,clindex,crindex)) {
      ulist[nextu - 1]->c_id = clist[i]->c_id;}}

/*----------------------------------------*/
/* Now consider phase 1,2 clauses         */
/*----------------------------------------*/

k = 0;
for(i = base + 1; i <= base + ply; i++) {
next_clause:
   if(excellent_weakened_proof == 1) {
      for(jj = 0; jj < nextu; jj++) {
         ulist[jj]->c_bits |= BASE;
         ulist[jj]->c_lpar = clist[i]->c_lpar;
         ulist[jj]->c_rpar = clist[i]->c_rpar;

         ulist[jj]->c_indices = clist[i]->c_indices;}
      return;}

   k++;
   if(questionable[i+1] == 0) {continue;}

/*--------------------------*/
/* substitute string??      */
/*--------------------------*/

   c1       = (cl[k - 1][0] >> 4) & PARENT_MASK;
   c1_index = (cl[k - 1][0] & INDEX_MASK);
   c2       = (cl[k - 1][1] >> 4) & PARENT_MASK;
   c2_index = (cl[k - 1][1] & INDEX_MASK);

   if(p_change_array[randy][0] == c1 && p_change_array[randy][1] <= c1_index
                                  && p_change_array[randy][2] == -1) c1_index++; 
   if(p_change_array[randy][0] == c2 && p_change_array[randy][1] <= c2_index
                                  && p_change_array[randy][2] == -1) c2_index++;

   if(cl[k - 1][11] > 0 && clist[c1 - 1]->c_nlit > 1) {
      goto det_clause;}

   if(c2 == 0 && c2_index == 0) {
      goto det_next_clause;}

   if(c1_index == 0) {
/*-----------------------------------------------------*/
/* substitute string-- not programmed to handle        */
/*-----------------------------------------------------*/

      for(j = i + 1; j < ply + base; j++) {
         clpar = clist[j]->c_lpar->c_id;
         if(clpar <= base) {
            i = j;
            break;}}
      if(i != j) i = j + 1;
      if(i >= base + ply) {
         break;}}

det_clause:
   clpar = clist[i]->c_lpar->c_id;
   if(clist[i]->c_rpar != 0) {
       crpar = clist[i]->c_rpar->c_id;}
   else crpar = 0;
   clindex = clist[i]->c_indices >> 6;
   crindex = (clist[i]->c_indices & 0x0000003F);
   if(p_change_array[randy][0] == clpar && p_change_array[randy][1] <= clindex
         && p_change_array[randy][2] == -1) clindex++;
   if(p_change_array[randy][0] == crpar && p_change_array[randy][1] <= crindex
         && p_change_array[randy][2] == -1) crindex++; 

    if(clpar*crpar*clindex*crindex == 0) {
det_next_clause:

/*-------------------------------*/
/* try to instantiate            */
/*-------------------------------*/

       if(c2 == 0 & c2_index == 0) {
          if(bsub_instantiated_proof(clpar)) {
	     for(ui = 0; ui < nextu - 2; ui++) {
	        if(ulist[ui]->c_id == clpar) {
		   ulist[ui]->c_bits &= ~ACTIVE;}}
             continue;}}

/*-------------------------------*/
/* try to factor                 */
/*-------------------------------*/

       if(crpar == 0) {
          if(bfac_instantiated_proof(clpar,clindex,crindex)) {
             ulist[nextu - 1]->c_id = clist[i]->c_id;
             for(ui = 0; ui < nextu - 2; ui++) {
                if(ulist[ui]->c_id == clpar) {
                   ulist[ui]->c_bits &= ~ACTIVE;}}
             continue;}}

       for(j = i + 1; j < base + ply; j++) {
          clpar = clist[j]->c_lpar->c_id;
	  if(clpar <= base) {
             i = j;break;}}
       if(i != j) i++;
       if(i < ply + base) {continue;}
       else {break;}}

    if(bres_instantiated_proof(clpar,crpar,clindex,crindex)) {
       ulist[nextu - 1]->c_id = clist[i]->c_id;
       for(ui = 0; ui < nextu - 2; ui++) {
	  if(ulist[ui]->c_id == clpar) {
	     ulist[ui]->c_bits &= ~ACTIVE;}}
       /*print_clause(ulist[nextu - 1]);*/}
    else {
/*       printf("fails -- try a second time"); */

       if(p_change_array[randy][0] != clpar && p_change_array[randy][1] != -1) {clindex++;}
      
       else {clindex--;}
       if(clindex == 0) {
	  for(j = i + 1; j < ply + base; j++) {
	     clpar = clist[j]->c_lpar->c_id;
	     if(clpar <= base) {
		i = j;
		break;}}
          if(i != j) i++;
	  if(i < base + ply) {
	     continue;}}

       if(bres_instantiated_proof(clpar,crpar,clindex,crindex)) {
          ulist[nextu - 1]->c_id = clist[i]->c_id;
          for(ui = 0; ui < nextu - 2; ui++) {
	     if(ulist[ui]->c_id == clpar) {
	        ulist[ui]->c_bits &= ~ACTIVE;}}
	  /*print_clause(ulist[nextu - 1]);*/}}}

filter_ulist(nextu0);

return;
}

/*------------------------------------------------------------*/
/* reinstate_base_clause()  ............................#6    */
/*                                                            */
/* called by: reconsider_proof()                              */
/*                                                            */
/* calls:                                                     */
/*                                                            */
/* substitute constants, etc., back into original clause.     */
/*------------------------------------------------------------*/

void reinstate_base_clause()
{
int i, *p, fix_lit;
clause *c;
literal *lit;

/*------------------------------------------*/
/* Undo changes to base clauses             */
/* determine modified clause and literal    */
/*------------------------------------------*/

c = clist[change_array[change_level][1] -1];
c->c_nv = change_array[change_level][4] + 1;
fix_lit = change_array[change_level][2];

for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
   if(--fix_lit == 0) {break;}}

/*------------------------------------------*/
/* three types of modifications possible    */
/*------------------------------------------*/

if(change_array[change_level][3] == -1) {
   c->c_nlit++;
   lit->l_id = lit->l_id & ~ELIMLIT;}

if((change_array[change_level][5] & FCNMASK) &&
      (change_array[change_level][5] & ARGMASK) > 0)  {

/*-------------------------------------------*/
/* move characters down to make space        */
/*-------------------------------------------*/

   p = &lit->l_id;
   p = p + change_array[change_level][3];
   for(i = 1; i <= deleted_terms[0]; i++) {
      *p++ = deleted_terms[i];}}

else {
   p = &lit->l_id;
   p = p + change_array[change_level][3];
   *p = change_array[change_level][5];}

clean_var_names(c);

}

/*--------------------------------------------------------*/
/* bres_instantiated_proof() ...................#7        */
/*                                                        */
/* called by:                                             */
/*                                                        */
/* calls:                                                 */
/*                                                        */
/*--------------------------------------------------------*/

int bres_instantiated_proof(ci1,ci2,ix1,ix2)
int ci1, ci2, ix1, ix2;
{
literal *l1, *l2;
int i, p1, p2;
clause *c, *c1, *c2, *c3, *c4;

/*---------------------------------------*/
/* get clauses from either ulist or clist*/
/*---------------------------------------*/

c3 = 0;
c4 = 0;
p1 = p2 = 0;

if(ci1 > org_base + 1) {
   if(questionable[ci1] == 0) {
      c1 = clist[ci1 - 1];}
   else {
      c1 = 0;
      for(i = 0; i < nextu; i++) {
         if(ulist[i]->c_id == ci1) {
	    c1 = ulist[i];
	    c3 = clist[ci1 - 1];
	    if(c1->c_nlit != c3->c_nlit) {
	       p1 = 1;}
            break;}}}}
else {c1 = clist[ci1 - 1];}

if(ci2 > org_base + 1) {
   if(questionable[ci2] == 0) {
      c2 = clist[ci2 - 1];}
   else {
      c2 = 0;
      for(i = 0; i < nextu; i++) {
         if(ulist[i]->c_id == ci2) {
	    c2 = ulist[i];
	    c4 = clist[ci2 - 1];
	    if(c2->c_nlit != c4->c_nlit) {
	       p2 = 1;}
	    break;}}}}
else {c2 = clist[ci2 - 1];}

if(c1 == 0 || c2 == 0) return(0);

l1 = c1->c_lit;
for(i = 1; i < ix1; i++) {
   l1 = l1->l_next;
   if(l1 == 0) return(0);}
   l2 = c2->c_lit;
for(i = 1; i < ix2; i++) {
   l2 = l2->l_next;
   if(l2 == 0) return(0);}

standardize(c1,c2);             /* standardize variables apart  */
clear_var_substitution_tble(c1->c_bits,c1->c_nv);
clear_var_substitution_tble(c2->c_bits,c2->c_nv);

if(((l1->l_id ^ l2->l_id) & ~VARMASK) == NEGMASK) {
   if(unify(&l1->l_id,&l2->l_id)) {
      if(c = generate_iclause(c1,c2,l1,l2,ix1,ix2)) {
          ulist[nextu++] = c;
          ulist[nextu] = c->c_end;
	  if(c->c_nlit == 0) {
	     excellent_weakened_proof = 1;
	     questionable[c->c_id] = 0;}
          return(1);}}}

else if(p1 & ~p2) {
   if(l1->l_next != 0) {
      l1 = l1->l_next;
      ix1++;
      goto now_resolve;}}

else if(~p1 & p2) {
   if(l2->l_next != 0) {
      l2 = l2->l_next;
      ix2++;
      goto now_resolve;}}

else if(p1 & p2) {
   if(l1->l_next != 0) {
      l1 = l1->l_next;
      ix1++;
      if(l2->l_next != 0) {
         l2 = l2->l_next;
         ix2++;
         goto now_resolve;}}}

return(0);

now_resolve:
if(((l1->l_id ^ l2->l_id) & ~VARMASK) == NEGMASK) {
   if(unify(&l1->l_id,&l2->l_id)) {
      if(c = generate_iclause(c1,c2,l1,l2,ix1,ix2)) {
          ulist[nextu++] = c;
          ulist[nextu] = c->c_end;
	  if(c->c_nlit == 0) {
	     excellent_weakened_proof = 1;
	     questionable[c->c_id] = 0;}
          return(1);}}}

return(0);
}

/*--------------------------------------------------------*/
/* bfac_instantiated_proof() ...................#7a       */
/*                                                        */
/* called by:                                             */
/*                                                        */
/* calls:                                                 */
/*                                                        */
/*--------------------------------------------------------*/

int bfac_instantiated_proof(ci1,ix1,ix2)
int ci1, ix1, ix2;
{
literal *l1, *l2;
int i;
clause *c, *c1;

/*---------------------------------------*/
/* get clauses from either ulist or clist*/
/*---------------------------------------*/

if(ci1 > org_base + 1) {
   if(questionable[ci1] == 0) {
      c1 = clist[ci1 - 1];}
   else {
      c1 = 0;
      for(i = 0; i < nextu; i++) {
         if(ulist[i]->c_id == ci1) {
	    c1 = ulist[i];
            break;}}}}
else {c1 = clist[ci1 - 1];}

if(c1 == 0) return(0);

l1 = c1->c_lit;
for(i = 1; i < ix1; i++) {
   l1 = l1->l_next;
   if(l1 == 0) return(0);}

l2 = c1->c_lit;
for(i = 1; i < ix2; i++) {
   l2 = l2->l_next;
   if(l2 == 0) return(0);}

clear_var_substitution_tble(c1->c_bits,c1->c_nv);

if(l1->l_id == l2->l_id) {
   if(unify(&l1->l_id,&l2->l_id)) {
      if(c = generate_iclause(c1,0,l1,l2,ix1,ix2)) {
          ulist[nextu++] = c;
          ulist[nextu] = c->c_end;
          return(1);}}}
return(0);
}

/*--------------------------------------------------------*/
/* bsub_instantiated_proof() ...................#7b       */
/*                                                        */
/* called by:                                             */
/*                                                        */
/* calls:                                                 */
/*                                                        */
/*--------------------------------------------------------*/

int bsub_instantiated_proof(ci1)
int ci1;
{
literal *l1, *l2;
int i;
clause *c, *c1, *c_new;
int var1, con1, var2, con2, v1_count, v2_count, substitution_info;

/*---------------------------------------*/
/* get clauses from either ulist or clist*/
/*---------------------------------------*/

if(ci1 > org_base + 1) {
   if(questionable[ci1] == 0) {
      c1 = clist[ci1 - 1];}
   else {
      c1 = 0;
      for(i = 0; i < nextu; i++) {
         if(ulist[i]->c_id == ci1) {
            c1 = ulist[i];
            break;}}}}
else {c1 = clist[ci1 - 1];}

if(c1 == 0) return(0);

substitution_info = c1->c_change;
var1 = substitution_info & 0x000000FF;
con1 = (substitution_info >> 8) & 0x000000FF;
var2 = (substitution_info >> 16) & 0x000000FF;
con2 = (substitution_info >> 24) & 0x000000FF;

copy_clause_all(c1);
v1_count = 1;
if(var2 != 0) v2_count = 1;
replace_vars(ulist[nextu],substitution_info);

normalize_literal(ulist[nextu]->c_lit,ulist[nextu]);
nextu++;
ulist[nextu] = ulist[nextu - 1]->c_end;
return(1);
}

/*----------------------------------------------------------------*/
/* filter_ulist() .............................#8                 */
/*                                                                */
/* called by:                                                     */
/*                                                                */
/* calls:      Nobody                                             */
/*                                                                */
/*----------------------------------------------------------------*/

void filter_ulist(utop)
int utop;
{
clause *c;
literal *lit;
int id1, i, j, ii, k, m, mm, ulist_min, ulist_min_loc;
int b_main_line, e_main_line;
int print_axioms_once, print_nc_once;
int ulistA[500];

if(o_learn == 0) return;
if(weakened_clause == -1) {return;}

if(total_questionable == 0) excellent_weakened_proof = 1;

for(i = utop + 1; i < nextu; i++) {
   ulist[i]->c_bits |= ACTIVE;
   if(ulist[i]->c_id == ulist[i - 1]->c_id + 1) {
      if(ulist[i]->c_nlit == ulist[i - 1]->c_nlit - 1) {
         ulist[i - 1]->c_bits &= ~ACTIVE;}}}

/*---------------------------------------------*/
/* scan for and delete duplicates on ulist     */
/*---------------------------------------------*/

for(i = 0; i < nextu; i++) {
   id1 = ulist[i]->c_id;
   for(j = i + 1; j < nextu; j++) {
      if(id1 == ulist[j]->c_id) {
         ulist[j]->c_bits &= ~ACTIVE;}}}

for(i = ply - 2; i >= 0; i--) {if(cl[i][3] != 0) break;}
b_main_line = cl[i + 1][0] >> 16;
e_main_line   = cl[ply - 1][0] >> 16;

for(i = 0; i < nextu; i++) {
   if(ulist[i]->c_id>=b_main_line && ulist[i]->c_id<=e_main_line-1) {
      ulist[i]->c_bits |= ACTIVE;}}

for(i = 0; i < nextu; i++) {
   c = ulist[i];
   for(lit = c->c_lit; lit != 0; lit = lit->l_next) {
      calc_lit_hc(lit,c);}
   calc_clause_hashcode(c);}

if(excellent_weakened_proof == 1) {
   questionable[p_change_array[randy][0]] = 0;
   printf("\nProof Found (Weakened proof is sufficient)\n\n");
   OutPtr += sprintf(OutPtr,"\nProof Found (Weakened proof is sufficient)\n\n");
   printf("\nTHE PROOF\n\n");
   OutPtr += sprintf(OutPtr,"\nTHE PROOF\n\n");
   print_axioms_once = print_nc_once = 0;
   for(i = 0; i <= org_base; i++) {
      if(org_base + 1 - nc) {
         if(print_axioms_once++ == 0) print0("\nAxioms:\n");}
      else if(print_nc_once++ == 0) print0("\nNegated conclusion:\n");
      print_clause(clist[i]);}

   print0("\nPhase 0 clauses used in proof:\n"); 

   for(i = org_base + 1; i <= base; i++) {
      if(clist[i]->c_bits & LIST) {
         if(questionable[i + 1]) {
	    questionable[i + 1] = 0;
	    for(ii = 0; ii < nextu; ii++) {
	       if(ulist[ii]->c_id == i + 1) {
	          ulist[ii]->c_bits = clist[i]->c_bits;
	          print_clause(ulist[ii]);}}}
         else print_clause(clist[i]);}}

   print0("\nPhases 1 and 2 clauses used in proof:\n"); 
	 
   for(i = 1; i <= ply; i++) {
      if(i > 1 && clist[depth[i].d_tc]->c_lpar->c_id !=
      clist[depth[i].d_tc]->c_id - 1) {
         printf("\n");
	 OutPtr += sprintf(OutPtr,"\n");} 
      if(questionable[base + i + 1]) {
	 questionable[base + i + 1] = 0;
         for(ii = 0; ii < nextu; ii++) {
	    if(ulist[ii]->c_id == base + i +1) {
	       ulist[ii]->c_bits = clist[depth[i].d_tc]->c_bits;
	       print_clause(ulist[ii]);}}}
      else print_clause(clist[depth[i].d_tc]);}

   printf("\nEND OF PROOF\n");
   OutPtr += sprintf(OutPtr,"\nEND OF PROOF\n\n");}

else {
   printf("\nEND OF WEAKENED PROOF\n\n");
   OutPtr += sprintf(OutPtr,"\nEND OF WEAKENED PROOF\n\n");
   printf("\nList of useful clauses from weakened proof:\n\n");
   OutPtr += sprintf(OutPtr,"\nList of useful clauses from weakened proof:\n\n");

/*------------------------------------------*/
/* sort while printing                      */
/*------------------------------------------*/

   for(i = 0; i < 500; i++) {ulistA[i] = -1;}

   for(i = 0; i < nextu; i++) {
      ulist_min = 10000;
      ulist_min_loc = -1;
      for(k = 0; k < nextu; k++) {
         if(ulistA[k] == 1) {continue;} 
         if(ulist[k]->c_id < ulist_min) {
	    ulist_min_loc = k;
	    ulist_min = ulist[k]->c_id;}} 
      if(ulist_min == 10000) {break;} 
      k = ulist_min_loc;
      ulist[k]->c_bits |= BASE;
      ulist[k]->c_lpar = 0;
      ulist[k]->c_rpar = 0;
      ulist[k]->c_indices = 0;
      questionable[ulist[k]->c_id] = 0;
      print_clause(ulist[k]);
      for(m = 0; m < nextu; m++) {
         if(ulist[m]->c_id == ulist[k]->c_id) {
            ulistA[m] = 1;}}}

   weak_proof_ends_here = OutPtr;}

printf("\n");OutPtr += sprintf(OutPtr,"\n");

}

/*----------------------------------------------------------------*/
/* skip_equal_all_vars() ..........................#9             */
/*                                                                */
/* called by:                                                     */
/*                                                                */
/* calls:                                                         */
/*                                                                */
/* don't use a clause of form equal(x,y) where x,y the same       */
/* variable for learning.  The clause is of no help               */
/*                                                                */
/* return: 0 if clause is not of useless form                     */
/* return: 1 if clause is useless                                 */
/*----------------------------------------------------------------*/

int skip_equal_all_vars(c)
clause *c;
{
int *p;

if(c->c_nlit > 1) {return(0);}
if((c->c_lit->l_id & 0x8000FFFF) != (equal_pred | FCNMASK)) {return(0);}

/*----------------------------------*/
/* first test for equal(a,b)        */
/*----------------------------------*/

p = &c->c_lit->l_id;
p++;
if((*p & FCNMASK) && ((*p & ARGMASK) == 0)) {
   p++;
   if((*p & FCNMASK) && ((*p & ARGMASK) == 0)) {return(1);}}

/*-----------------------------------*/
/* now test for equal(xxxxx,constant)*/
/* (to be completed??)               */
/*-----------------------------------*/

return(0);
}

/*----------------------------------------------------------------*/
/* skip_equal_first_last() ...........................#10         */
/*                                                                */
/* called by:                                                     */
/*                                                                */
/* calls:                                                         */
/*                                                                */
/* don't use a clause of form equal(c,xxxxxy) or equal(xxxx,c)    */
/*                                                                */
/* return: 0 if clause is not of useless form                     */
/* return: 1 if clause is useless                                 */
/*----------------------------------------------------------------*/

int skip_equal_first_last(c,pp)
int *pp;
clause *c;
{
int *p, a;
int string_length, string_length2;
int *const_ptr, *const_ptr2;

if(c->c_nlit > 1) {return(0);}
if((c->c_lit->l_id & 0x8000FFFF) != (equal_pred | FCNMASK)) {return(0);}

/*----------------------------------*/
/* first test for equal(a,b)        */
/*----------------------------------*/

string_length = 0; string_length2 = 1;

p = &c->c_lit->l_id;

for(a = *p++ & ARGMASK; a > 1; a--) {
   if(string_length == 0) {const_ptr = p;}
   string_length++;
   if(*p & FCNMASK) {a += *p++ & ARGMASK;}
   else {p++;}}

const_ptr2 = p;

if((*p & FCNMASK) == 0) {p++;}
else {
   for(a = *p++ & ARGMASK; a > 0; a--) {
      string_length2++;
      if(*p & FCNMASK) {a += *p++ & ARGMASK;}
      else {p++;}}}

if(string_length * string_length2 == string_length ||
   string_length * string_length2 == string_length2) {

   if(string_length == 1 && string_length2 > 1 && pp == const_ptr &&
            (*pp & FCNMASK)) {return(1);}

   if(string_length2 == 1 && string_length > 1 && pp == const_ptr2 &&
            (*pp & FCNMASK)) {return(1);}}

/*-----------------------------------*/
/* now test for equal(xxxxx,constant)*/
/* (to be completed??)               */
/*-----------------------------------*/

return(0);
}

