{-# OPTIONS -fglasgow-exts #-}
------------------------------------------------------------------
-- Paradox, a Finite Domain Model Finder for First Order Logic  --
-- Copyright (c) 2003 Koen Claessen and Niklas Srensson        --
-- {koen,nik}@cs.chalmers.se                                    --
--                                                              --
-- This file is part of Paradox.                                --
--                                                              --
-- Paradox is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public License  --
-- as published by the Free Software Foundation; either version --
-- 2 of the License, or (at your option) any later version.     --
--                                                              --
-- Paradox is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty  --
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See  --
-- the GNU General Public License for more details.             --
--                                                              --
-- You should have received a copy of the GNU General Public    --
-- License along with Paradox; if not, write to the Free        --
-- Software Foundation, Inc., 59 Temple Place, Suite 330,       --
-- Boston, MA 02111-1307 USA.                                   --
------------------------------------------------------------------
module Main where

{-

This module defines the main function of the Paradox program.

-}

import Flags
import ParseProblem
import SolveProblem
import Output
import Sat ( identify )
import Data
import Set

import Utils
  ( thd3
  )

import IO
  ( try
  )

-------------------------------------------------------------------------
-- main

main :: IO ()
main =
  do sat <- identify
     putStrLn ("This is Paradox v. 1.1-casc (with " ++ sat ++ "), June 2004.")
     flags <- getFlags
     if null (files flags)
       then putStrLn "Warning: No file(s) given! Try --help."
       else sequence_ [ problem flags (sizes flags) file
                      | file <- files flags
                      ]

-------------------------------------------------------------------------
-- problem

problem :: Flags -> [Int] -> FilePath -> IO ()
problem flags ns file =    
  do putHeader "Reading"
     mecs <- try $ readProblemWithRoots ("" : map (++"/") (roots flags)) file
     case mecs of
       Left err ->
         do putStrLn (show err)
            return ()
       
       Right cs ->
         do solveProblem flags' ns (map thd3 cs)
        where
         flags' = flags{ dimacs = patch `fmap` dimacs flags }

         patch ('%':'f':s) = name file ++ patch s
         patch (c:s)       = c : patch s
         patch ""          = ""

         name = reverse
              . dropExtension
              . takeWhile (not . isSlash)
              . reverse

         dropExtension s = case drop 1 (dropWhile (/= '.') s) of
                             "" -> s
                             s' -> s'

         isSlash '/'  = True
         isSlash '\\' = True
         isSlash _    = False

         -- normally not used
         -- this is here for experimental reasons
         trans = set . map transLit . elements
          where
           transLit (Pos x) = Pos (transAtom x)
           transLit (Neg x) = Neg (transAtom x)

           transAtom (Pred _ (p ::- _) [b,v1,v2,v3,v4,h,s1,s2,s3,s4,e1,e2]) =
             (p % 1 ::- Unknown) `prd`
               [ b
               , (name' "v" ::- Unknown) `app`
                   [ (name' "v1" ::- Unknown) `app` [v1,v2]
                   , (name' "v2" ::- Unknown) `app` [v3,v4]
                   ]
               , h
               , (name' "s" ::- Unknown) `app`
                   [ (name' "s1" ::- Unknown) `app` [s1,s2]
                   , (name' "s2" ::- Unknown) `app` [s3,s4]
                   ]
               , (name' "e" ::- Unknown) `app`
                   [ e1
                   , e2
                   ]
               ]

           transAtom a = a

-------------------------------------------------------------------------
-- the end.
