/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
//  file problem.h 
//  defines class Problem
//


#ifndef __problem__
#define __problem__


#include "typedefs.h"
#include "unit.h"
#include "formula.h"


class FSymbol;
class Options;


// ****************** class Problem, definition *******************


class Problem {
 public:
  // constructors & destructors
  explicit Problem ( Unit::List* );
  ~Problem ();
  void preprocess (Options&);

  // query the structure
  Unit::Chain& units () { return _units; }

  // properties
  bool hasFormulas () const { return _hasFormulas; }

  // miscellaneous
  int removeUnusedDefinitions (); // the result is the number of removed definitions
  int removeEqualityAxioms ();
  int split ();
  void ennf ();
  int removeIff ();
  int miniscope ();
  void normalize ();   // normalize and sort clauses/formulas in the input
  void flatten ();
  void rectify ();
  void skolemize (); 
  void clausify ();

/*
  void printDefinitions ( ostream& str );
  int removeLinearFunctionDefinitions ( ostream& tab ); // the result is the number of removed symbols
  int removeAllFunctionDefinitions ( ostream& tab ); // the result is the number of removed symbols
  int removeAllRelationalFunctionDefinitions ( ostream& tab ); // the result is the number of removed symbols
  int replaceCancellationAxioms ( ostream& tab ); 
  int eliminateResolvable ( ostream& tab);
  int eliminatePure ( ostream& tab);
  int removeVarEquality ( ostream& tab );
  int removeTautologies ( ostream& tab );
  int cutTooLongClauses ( int maxLiterals );
  bool splitLargeArity ( int maxArity );
  int negativeEqualitySplit ( int thresholdWeight, ostream& tab );

  // statistics
  bool hasEquality () const;
  bool hasEqualityInGoal () const;
  bool isHorn () const;
  bool hasNonEquality () const;
  bool hasFunctionSymbols () const;
  bool isPropositional ()const;
  bool isUnit () const;
  int numberOfClauses () const;
  int numberOfLiterals () const;
  void statistics ( ostream& ) const; // appends statistics to the file
  int maxLitWeight ();
  int goalWeight ();  
*/
 private:

  // structure
  Unit::Chain _units;
  bool _hasFormulas;

  // auxiliary functions
  Formula* hasFreeVars () const;
  int findDefinitions ();
  void clausify (Formula::List* formulas, 
		 Literal::List* literals,
		 Unit* parent);
/*
  int countOccurrences ( const FSymbol* f ) const;
  bool occurs ( const FSymbol* f ) const;
  bool occurs ( const PSymbol* p ) const;
  int countOccurrences ( const PSymbol* p ) const;
  void applyFunDef ( const Term* lhs, const Term* rhs, const Unit* def, ostream& report ); // replace definition using lhs = rhs
  bool removeOneRelationalFunctionDefinition ( ostream& ); 
  void replaceRelationByFunction ( PSymbol* pred, FSymbol* fun, int argNo, int* vs, ostream& tab );
  void removeResolvable ( const PSymbol* pred, bool sign, ostream& report );
  void removePredicate ( const PSymbol* p, ostream& report ); // remove all occurrences of p
  int polarities ( const PSymbol* p ) const;
*/
};


ostream& operator << ( ostream&, Problem& );

 
#endif // __problem_
