/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
//  file pure.cpp
//  implements class Pure: auxiliary class for defining pure predicates
//
//  27/05/2001 Manchester
//


#include "pure.h"
#include "clause.h"


Pure::Pure ( int noOfSymbols )
  :
  _noOfSymbols (noOfSymbols),
  _remaining (noOfSymbols),
  _symbols ( new Symbol [noOfSymbols] )
{
  // number of symbols must be non-zero
  ASSERT( debug_pure, noOfSymbols );  
} // Pure::Pure


Pure::Symbol::Symbol ()
  :
  pos (false),
  neg (false)
{
} // Pure::Symbol::Symbol


Pure::~Pure ()
{
  delete [] _symbols;
} // Pure::Pure


bool Pure::getClause (Clause* c)
{
  Literal::List::Iterator ls ( c->literals() );
  while ( ls.more() ) {
    Literal* l = ls.next();
    Atom* atom = l->atom();
    if ( atom->isEquality() )
      continue;
    Symbol& s = _symbols[atom->functor()->number()];

    if ( s.pos && s.neg ) { // s is not a candidate anyhow
      continue;
    }

    if ( l->positive() ) {
      s.pos = true;
    }
    else {
      s.neg = true;
    }

    if ( s.pos && s.neg ) { // s is not a candidate any more
      _remaining --;
      if ( _remaining == 0 ) { // no remaining candidate symbols
        return false;
      }
    }
  }

  return true;
} // Pure::getClause


ostream& operator << ( ostream& str, const Pure& res )
{
  PSymbol** signature = PSymbol::signature ();
  for ( int i = res._noOfSymbols - 1; i >= 0; i-- ) {
    if ( ! res._symbols[i].pos || ! res._symbols[i].neg ) {
      const char* diagnosis;
      if ( res._symbols[i].pos ) {
        diagnosis = "pure positive";
      }
      else if ( res._symbols[i].neg ) { // negative occurrence only
        diagnosis = "pure negative";
      }
      else {
        diagnosis = "no occurrences";
      }
      str << "  " << signature[i] << " (" << diagnosis << ")\n";
    }
  }

  delete [] signature;

  return str;
} // operator << (ostream&, const Pure&)


Pure::Iterator::Iterator ( Pure& pure )
  :
  _signature ( PSymbol::signature() ),
  _symbols ( pure.symbols() ),
  _nextSymbol ( PSymbol::noOfSymbols() - 1 )
{
} // Pure::Iterator::Iterator
 

Pure::Iterator::~Iterator ()
{
  delete [] _signature;
} // Pure::Iterator::~Iterator
 

PSymbol* Pure::Iterator::next ()
{
  while ( _nextSymbol >= 0 ) {
    Symbol& s = _symbols [_nextSymbol];
    if ( s.pos == ! s.neg ) { // 
      return _signature [_nextSymbol--];
    }
    // no occurrences
    _nextSymbol--;
  }

  return 0;
} // Pure::Iterator::Iterator

