/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
//  file pure.h 
//  defines class Pure: auxiliary class for finding pure predicate symbols
//
//  27/05/2001 Manchester
//

#ifndef __pure__
#define __pure__


#include "debug.h"


class Clause;
class PSymbol;


// ***************** class Pure, definition ****************************


class ostream;

// A symbol P if pure if and only if either all of its occurrences are
// positive or all of them are negative
class Pure {
 public:

  class Symbol { // made public since used by Iterator
   public:
    bool pos; // has at least one positive occurrence
    bool neg; // has at least one neg occurrence

    Symbol ();
  };  // class Pure::Symbol

  // constructor
  Pure ( int noOfSymbols );
  ~Pure ();
  bool getClause (Clause*); // true if there are still candidate symbols
  int remaining () { return _remaining; }
  Symbol* symbols () { return _symbols; }

  class Iterator {
   public:
    Iterator ( Pure& );
    ~Iterator ();
    PSymbol* next (); // 0 if none
   private:
    PSymbol** _signature;
    Symbol* _symbols;
    int _nextSymbol;
  };  // class Pure::Iterator
 
 private:
  // structure
  int _noOfSymbols;
  int _remaining;   // number of remaining eligible symbols
  Symbol* _symbols;
 
 friend ostream& operator << ( ostream&, const Pure& );
}; // class Resolvable


// for debugging only
ostream& operator << ( ostream&, const Pure& );

// ***************** class Pure, implementation ****************************




#endif // __pure__
