/*
COPYRIGHT UNIVERSITY OF MANCHESTER, 2002

Contributors:
  Alexandre Riazanov
  Andrei Voronkov

LICENCE TERMS

Please note the following:

(1) this code was intended for CASC only, can only be used for
CASC-related experiments and cannot be used otherwise;

(2) Vampire is free for non-commercial use and a standard non-commercial
licence can be found on the Vampire Web page that is currently:

  http://www.cs.man.ac.uk/~riazanoa/Vampire/index.html

BY USING THIS CODE YOU AGREE TO THE TERMS IN THIS LICENCE.
 
(3) To get a licence for commercial use, please consult the Vampire Web
page.

(4) Disclaimer.

THIS MATERIAL IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. THE ENTIRE RISK AS TO THE QUALITY AND
PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE PROGRAM PROVE
DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
CORRECTION.

*/

//
//  file signature.h 
//  defines class Signature consisting of predicate and function symbols
//
//  28/04/2002, Manchester, made by merging PSymbol and FSymbol
//  01/05/2002, Manchester, implementation of equality changed, Iterator added
//

#ifndef __signature__
#define __signature__

#define NO_OF_BUCKETS 2047

// ***************** class Signature, definition ****************************


class ostream;


#include "int.h"
#include "list.h"


class Signature {
 
 public:

  class Symbol;
  typedef List <Symbol*,CID_PSYM_LIST> List;

  class Symbol 
#   if debug_memory
    : public Memory <CID_SYMBOL>
#   endif
  {
   public:
    // query structure
    int arity () const { return _arity; }
    const char* name () const { return _name;}
    const int number () const { return _number; }
    void arity ( int ar ) { _arity = ar; };
    void kernelNumber (unsigned long n) { _kernelNumber = n; }
    unsigned long kernelNumber () const { return _kernelNumber; }

   protected:
    Symbol ( const char* name, int arity, int& number ) :
      _name (name),
      _arity (arity),
      _number (number++)
      {}
    // structure
    const char* _name;
    int _arity;
    int _number;
    unsigned long _kernelNumber;
  }; // class Signature::Symbol

  class Fun :
    public Symbol
  {
   public:
    Fun ( const char* name, int arity,int& count ) :
      Symbol (name,arity,count)
      {}
    Compare compare (const Fun*) const; // comparison, needed to normalize
  };

  class Pred :
    public Symbol
  {
   public:
    Pred ( const char* name, int arity,int& count ) :
      Symbol (name,arity,count)
      {}
    bool isEquality () const
      { return number() == 0; }
    Compare compare (const Pred*) const; // comparison, needed to normalize
  };

  // iterator, iterate over all function or predicate symbols
  class Iterator {
   public:
    explicit Iterator (List** buckets);
    bool more ();
    Symbol* next ();
   private:
    List** _firstBucket;
    List** _currentBucket;
    List* _currentList;
  }; // Iterator

  // iterator over all function symbols
  class FunIterator 
    : public Iterator {
   public:
    explicit FunIterator (Signature* s) :
      Iterator (s->_fbuckets)
      {}
  };

  // iterator over all predicate symbols
  class PredIterator 
    : public Iterator {
   public:
    explicit PredIterator (Signature* s) :
      Iterator (s->_pbuckets)
      {}
  };

  Signature ();
  ~Signature ();

  Fun* createFun ( const char* name, int arity )
    { return static_cast<Fun*>(create (name, arity, _fbuckets, false)); }
  Pred* createPred ( const char* name, int arity )
    { return static_cast<Pred*>(create (name, arity, _pbuckets, true)); }

  // structure
  Pred* equality () { return _equality; }
  int noOfPreds () const { return _noOfPreds; }
  int noOfFuns () const { return _noOfFuns; }
  bool arityCheck () const; // true if every symbol has only one arity

  // skolemization
  Fun* newSkolemFunction (int arity);
  
 private:  
  List * _fbuckets [NO_OF_BUCKETS];
  List * _pbuckets [NO_OF_BUCKETS];
  int _noOfPreds;
  int _noOfFuns;
  int _lastSkolem;
 
  static int hash ( const char* );  // hashing function
  Symbol* create ( const char* str, int arity, List** buckets, bool isPred );
  Pred* _equality;

  // miscellaneous
  bool existsFun ( const char* name ) const;

 friend class SymCounter;
 friend class FunIterator;
 friend class PredIterator;
}; // class Signature


ostream& operator << ( ostream& str, const Signature::Symbol* p );


// the global signature
extern Signature* sig;

#endif // __signature__
