/*=================================================================================================================
  =                                                                                                               =
  =  DomaenenMemorium.c                                                                                           =
  =                                                                                                               =
  =  Kapselung mmap-Speicherzugriff                                                                               =
  =                                                                                                               =
  =  19.02.97 Thomas: aus StrahlMemorium.c ausgelagert.                                                           =
  =                                                                                                               =
  =================================================================================================================*/

/* 02.09.96: AJ: Signalbehandlung fuer SUNOS5 (=Solaris2) und Linux */

#ifdef LINUX
#define _GNU_SOURCE 1
#endif

#include "antiwarnings.h"
#include "Ausgaben.h"
#include "DomaenenMemorium.h"
#include <fcntl.h>
#include "FehlerBehandlung.h"
#include <signal.h>
#include "SpeicherVerwaltung.h"
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>
#include "WaldmeisterII.h"
#include "Statistiken.h"

#if defined (LINUX)
# if !defined __GLIBC__
#  include <asm/sigcontext.h>
#  include <signal.h>
#  define sigcontext sigcontext_struct
   typedef __sighandler_t sighandler_t;
# endif
#elif defined (SUNOS5)
#include <sys/siginfo.h>
#elif defined (SUNOS4)
#else
#error "Signalbehandlung nur fuer SunOS4, SunOS5 und Linux definiert"
#endif


/* Bemerkung zur Implementierung:
   - AlterSignalbehandler kann den Wert SIG_DFL haben, in diesem Fall
     ist AlterSignalbehandler kein gueltiger Funktionsaufruf
   
   LINUX:  Um an die Adresse, die den Speicherfehler versucht hat, zu kommen, wurden die Kernel-Strukturen untersucht
           und der zusaetzliche Parameter sigcontext_struct lokalisiert. Dies funktioniert mit Kernel 2.0.17, aber
           muss nicht mit aktuelleren Kernels funktionieren.  Linux unterstuetzt bei mmap auch sogenanntes anonymous
           Mapping. Dazu muss kein File geoeffnet sein, aber das Flag MAP_ANON (manchmal auch MAP_ANONYMOUS)
           uebergegeben werden.

   SUNOS5: Solaris definiert bei sigaction zwei moegliche Funktionen: sa_handler (POSIX konform) und sa_sigaction
           (Solaris-spezifisch). sa_sigaction liefert die Informationen, die wir benoetigen.  */



/*=================================================================================================================*/
/*= I. Aufteilung des virtuellen Adressraums                                                                      =*/
/*=================================================================================================================*/

/* Jeder Domaene muss bei der Einrichtung ein Adressraum zugewiesen werden. Aus dem Aufaehlungstyp wird eine Angabe
   in der Form (a,b) abgeleitet mit 0x00 <= a <= b <= 0xFF, die sich bezieht auf den Adressraum 
   [a*0x1000000,b*0x1000000+0x0FFFFFF]. 
   Beim Einrichten wird ueberprueft, ob a und b sinnvolle Werte haben und der beanspruchte Adressraum nicht schon
   belegt ist. Waechst eine Domaene ueber ihre Grenze hinaus, so wird ein Systemabbruch ausgeloest. Nach dem Frei-
   geben einer Domaene steht ihr Adressraum wieder zur Verfuegung. 

   Zur Zeit stehen zwei Aufteilungen zur Verfuegung. Wenn sich ab a=0x20 eine Seite mappen laesst, so wird der ge-
   samte virtuelle Adressraum bis b=0xef als verfuegbar angenommen, wobei unter LINUX der Bereich (0x30, 0x3f) aus-
   gespart wird. Falls sich aber keine Seite ab 0x20 mappen laesst, so werden nur die Bereiche (0x10, 0x1f) sowie
   (0xe0, 0xef) verwendet.

   Der Bereich (0x00, 0x0f) ist reserviert als Adressraum fuer Programmkode, statische und dynamische Variablen. */

#define SechzehnMByte ((unsigned long int)0x1000000)

typedef struct {
  DomaenenbereichsT Bereich;
  unsigned long int Start;
  unsigned long int Obergrenze;
} BereichstabellenT [DM_Das_wars_jetzt_aber];

static BereichstabellenT BereicheDick = {
#if defined (LINUX)
  {DM_iPCLMemo,    0x20, 0x3e},   /* knapp  1/2  GByte Adressraum */
  {DM_iPCLBuffer,  0x3f, 0x3f},   /* 16   MByte Adressraum */
  {DM_SV,          0x60, 0xbf}   /* 1 1/2  GByte Adressraum fuer DOMAENE SpeicherManager */
},
#else
  {DM_SV,          0x40,  0xbf},   /*  2   GByte Adressraum fuer DOMAENE SpeicherManager */
  {DM_iPCLMemo,    0xc0,  0xdf},   /* 1/2  GByte Adressraum */
  {DM_iPCLBuffer,  0xe0,  0xe0}    /* 16   MByte Adressraum */
},
#endif
                         BereicheDuenn = {
  {DM_SV,          0x10, 0x1f},   /* 256  MByte Adressraum fuer DOMAENE SpeicherManager */
  {DM_KPH,         0xe0, 0xe7},   /* 128  MByte Adressraum alternativ fuer DOMAENE KPHeap */
  {DM_iPCLMemo,    0xe8, 0xeb},   /*  64  MByte Adressraum */
  {DM_iPCLBuffer,  0xec, 0xec}    /*  16  MByte Adressraum */
},
*Bereiche;

#define for_Bereiche(/*unsigned int*/ i)                                                                            \
  for(i=0; i<DM_Das_wars_jetzt_aber; i++)

static void GrenzenHolen(DomaenenbereichsT Bereich, void **Start, void **Obergrenze)
{
  unsigned int i;
  for_Bereiche(i)
    if ((*Bereiche)[i].Bereich==Bereich) {
      *Start = (void*)((*Bereiche)[i].Start*SechzehnMByte);
      *Obergrenze = (void*)(((*Bereiche)[i].Obergrenze+1)*SechzehnMByte);
      return;                                                  /* umsteigen auf Angabe als rechtsoffenes Intervall */
    }
  our_error("Illegaler Domaenenbereich");
}


static BOOLEAN our_mmap(void*,void*);
static BOOLEAN our_munmap(void*,void*);

static void DomaenenbereicheAbstecken(void)
{
#if 0
  unsigned long int i;
  for(i=0; i<16*16; i++) {
    void *a = (void*)(i*SechzehnMByte),
         *b = (void*)((char *)a+DM_Seitengroesse*2);
    if (our_mmap(a,b)) {
      printf("OK [%p,%p[\n",a,b);
      our_munmap(a,b);
    }
    else
      printf("Fehlschlag [%p,%p[\n",a,b);
  }
  exit(0);
#else
  void *a = (void*)((char *)NULL+0x20*SechzehnMByte),
       *b = (void*)((char *)a+DM_Seitengroesse);
  Bereiche = our_mmap(a,b) ? (our_munmap(a,b), &BereicheDick) : &BereicheDuenn;
#endif
}



/*=================================================================================================================*/
/*= II. Initialisierung; Allgemeines                                                                              =*/
/*=================================================================================================================*/

size_t DM_Seitengroesse;
#define CharToLong(a,b,c,d) 256*(256*(256*a+b)+c)+d                            /* fuer Markierung unserer Domaenen */
#define WMIdentifikation    CharToLong('W','M','I','I')                 /* im Ggs. zu von anderen gemappten Seiten */

volatile DomaenenT Domaenenliste;

#ifndef MAP_ANON
  static int Orkus;
#endif
#if   defined (SUNOS4)
 typedef void (*WM_SignalHandlerT)(int Signal, int Code, struct sigcontext *Kontext, char *Adresse);
 static void Kollisionsbehandlung(int Signal, int Code, struct sigcontext *Kontext, char *Adresse);
#elif defined (SUNOS5)
  static void Kollisionsbehandlung(int Signal, siginfo_t *Kontext, void *Adresse);
#elif defined (LINUX)
  typedef void (*WM_SignalHandlerT) (int Signal, struct sigcontext Kontext);
  static void Kollisionsbehandlung(int Signal, struct sigcontext Kontext);
#endif
static struct sigaction AlterSignalbehandler;


void DM_InitApriori(void)
{
  if (WM_ErsterAnlauf()) {
    struct sigaction sigact;
    DM_Seitengroesse = getpagesize();
#if !defined (MAP_ANON)
    if ((Orkus = open("/dev/zero",O_RDWR))==-1)
      our_error("/dev/zero kann nicht geoeffnet werden.");
#endif
#if defined (SUNOS4)
    sigact.sa_handler = &Kollisionsbehandlung;
    sigact.sa_flags =  0;
#elif defined (SUNOS5)
    sigact.sa_handler = NULL;
    sigact.sa_sigaction  = &Kollisionsbehandlung;
    sigact.sa_flags =  SA_SIGINFO|SA_RESTART;
#elif defined (LINUX)
    sigact.sa_handler = (sighandler_t)&Kollisionsbehandlung;
    sigact.sa_flags =  SA_RESTART;
#endif                                   
    sigemptyset(&sigact.sa_mask);
    if (sigaction(SIGSEGV, &sigact, &AlterSignalbehandler))
      our_error("Signalbehandler fuer SIGSEGV kann nicht gesetzt werden.");
    Domaenenliste = NULL;
    DomaenenbereicheAbstecken();
  }
}


void DM_NieMehrDomaenen_SagNiemalsNie(void)
{
#ifndef MAP_ANON   
  if (close(Orkus))
    our_error("/dev/zero kann nicht geschlossen werden.");
#endif
}


/*=================================================================================================================*/
/*= III. Mapping, Unmapping, Seitenschutz                                                                         =*/
/*=================================================================================================================*/

static BOOLEAN our_mmap(void *Start, void *Ende)
{
  return mmap(Start,(char*)Ende-(char*)Start,PROT_READ|PROT_WRITE,MAP_PRIVATE|MAP_FIXED          
#ifdef MAP_ANON
                                                                                       |MAP_ANON,-1
#else
                                                                                               ,Orkus
#endif
                                                                                                     ,0)!=(caddr_t)-1;
}

void *mmap_shared(size_t size)
{
  void *res;
  if ( (res = mmap(NULL, size,PROT_READ|PROT_WRITE,MAP_SHARED          
#ifdef MAP_ANON
                                                             |MAP_ANON,-1
#else
                                                                      ,Orkus
#endif
                                                                             ,0))!=(caddr_t)-1 ){
    if (res == NULL){
      our_warning("doofes OS: mmap_shared == NULL");
    }
    return res;
  }
  else {
    return NULL;
  }
}
void DM_mmap(void *Start, void *Ende)              /* Seiten [Start,Ende) holen. Kein Test, ob Ausdehnung erlaubt. */
{                                                                           /* Start und Ende sind Seitenadressen. */
  /*printf("mmap[%p,%p[\n",Start,Ende);*/
  if (!our_mmap(Start,Ende))
    our_out_of_mem_error("mmap kann nicht ausgefuehrt werden.");
}


static BOOLEAN our_munmap(void *Start, void *Ende)
{ return !munmap(Start,(size_t)((char*)Ende-(char*)Start)); }
void DM_munmap(void *Start, void *Ende)                                           /* Seiten [Start,Ende) weggeben. */
{                                                                           /* Start und Ende sind Seitenadressen. */
  if (!our_munmap(Start,Ende))
    our_error("munmap fehlgeschlagen");
}


typedef enum { 
  Gesperrt = PROT_NONE,
  Ungesperrt = PROT_READ|PROT_WRITE
} SchutzT;

static BOOLEAN our_mprotect(void *Ende, SchutzT Schutz, BOOLEAN Abbruch)            /* Fuer die Seite vor Ende Schutzbits */
{                                                                               /* setzen. Ende ist Seitenadresse. */
  if (mprotect((char*)Ende-DM_Seitengroesse(),DM_Seitengroesse(),Schutz)==-1) {                      /* Fehlerfall */
    if (Abbruch)
      our_error("mprotect kann nicht ausgefuehrt werden.");
    return FALSE;
  }
  else
    return TRUE;
}



/*=================================================================================================================*/
/*= IV. Domaenen                                                                                                  =*/
/*=================================================================================================================*/

#define _ ptr2ul                                                                                   /* Lesbarkeit++ */

void DM_AmEndeGesperrtAllokieren(DomaenenT Domaene, unsigned long int Delta, void *Hakendatum)
             /* Seiten [Ende,Ende+Delta) holen; zuvor jedoch testen, ob erlaubte Ausdehnung ueberschritten wuerde; */
{                       /* Domaenenadresse, Identifikation und Hakendatum in letzte Seite schreiben; diese sperren */
  void *NeuesEnde = (char*)Domaene->Ende+Delta;
  if (_(NeuesEnde)>_(Domaene->Obergrenze)) {
    char text[256];
    /* snprintf waere hier evtl. besser */
    sprintf (text, "Domaene %s waechst ueber vorgesehenen Adressraum.", Domaene->Name);
    our_error(text);
  }
  DM_mmap(Domaene->Ende,NeuesEnde);
  *((DomaenenT*)NeuesEnde-1) = Domaene;
  *((unsigned long int*)NeuesEnde-2) = WMIdentifikation;
  *((unsigned long int**)NeuesEnde-3) = Hakendatum;
  our_mprotect(NeuesEnde,Gesperrt,TRUE);
  Domaene->Ende = NeuesEnde;
}


static void LastRessortSignalHandler(void)  /* Wenn kein Signalbehandler gesetzt ist, hilft nur eins: raus! Nichts */
{       /* mehr ausgeben, da es sonst zu einer Endlosschleife kommen kann (auf Linux). Prozedur wurde ausgelagert, */
#if defined SUNOS4 || defined SUNOS5                               /* um einen Breakpoint darauf setzen zu koennen */
  fputs("Unrecoverable Segmentation Fault\n",stderr);                              /* mal austesten, ob das klappt */
#endif
  abort ();                                                                        /* weniger Aufwand als exit(-1) */
}

#define TestAlterSignalbehandler(Signalbehandler)         /* pruefen, ob ueberhaupt eigener Behandler gesetzt war */\
  if ((void *)Signalbehandler==(void *)SIG_DFL) {  /* Diese Abfrage wird vom Apogee bemeckert mit Warnung invalid */\
    LastRessortSignalHandler(); /* type conversion. Es ist aber die einzige Moeglichkeit, den GCC ruhigzustellen. */\
  }

#if defined (SUNOS4)
#define SonstigerFehler() (Code!=SEGV_PROT)
#define AlteBehandlung()                                                                                            \
{ TestAlterSignalbehandler(AlterSignalbehandler.sa_handler);                                                        \
  (*AlterSignalbehandler.sa_handler)(Signal,Code,Kontext,Adresse); }
#elif defined (SUNOS5)
#define SonstigerFehler() (Info->si_code!=SEGV_ACCERR)
#define AlteBehandlung()                                                                                            \
{ if ((AlterSignalbehandler.sa_flags) && SA_SIGINFO) {                                                              \
    TestAlterSignalbehandler(AlterSignalbehandler.sa_sigaction);                                                    \
    (*(AlterSignalbehandler.sa_sigaction))(Signal, Info, Kontext);                                                  \
  }                                                                                                                 \
  else {                                                                                                            \
    TestAlterSignalbehandler(AlterSignalbehandler.sa_handler);                                                      \
    (*(AlterSignalbehandler.sa_handler))(Signal);                                                                   \
  } }
#elif defined (LINUX)
#define SonstigerFehler() (Signal!=SIGSEGV)
#define AlteBehandlung()                                                                                            \
{ TestAlterSignalbehandler (AlterSignalbehandler.sa_handler);                                                       \
  (*(WM_SignalHandlerT)AlterSignalbehandler.sa_handler)(Signal,Kontext); }
#endif

static void Kollisionsbehandlung(int Signal, 
#if   defined (SUNOS4)
                                             int Code, struct sigcontext *Kontext, char *Adresse)
#elif defined (SUNOS5)    
                                             siginfo_t *Info, void *Kontext)
#elif defined (LINUX)    
                                             struct sigcontext Kontext)
#endif
{
  if (SonstigerFehler())
    AlteBehandlung()
  else {
#if   defined (SUNOS5)
    unsigned long Adresse = (unsigned long) Info->si_addr;
#elif defined (LINUX)
    unsigned long Adresse = Kontext.cr2;
#endif
    char *Kollision = (char*)((unsigned long int)Adresse & ~(DM_Seitengroesse()-1)) + DM_Seitengroesse();
    DomaenenT Domaene;
    if (!our_mprotect(Kollision,Ungesperrt,FALSE) ||   /* Kollisionsseite zum Lesen freigeben, sofern das moeglich */
      *((unsigned long int*)Kollision-2)!=WMIdentifikation)                                /* ist, sonst Ausstieg. */
      AlteBehandlung();                                          /* Dito, falls Seite keine WM-Identifikation hat. */
    Domaene = *((DomaenenT*)Kollision-1);
    (*Domaene->Haken)(Kollision);                                                        /* Domaenenhaken aufrufen */
  }
}


DomaenenT DM_NeueDomaene(DomaenenartenT Domaenenart, DomaenenbereichsT Bereich, 
  DomaenenhakenT Haken, const char *Name)
{
  unsigned long int Start,
                    Obergrenze;
  DomaenenT Domaene = our_alloc(sizeof(struct DomaenenTS));
  volatile DomaenenT Index = Domaenenliste;
  Domaene->Name = Name;
  GrenzenHolen(Bereich, (void**)&Start, (void**)&Obergrenze);
  if (Start>Obergrenze || Obergrenze>0xFF*SechzehnMByte)
    our_error("Ungueltige Adressraumangabe in DM_NeueDomaene.");
  Domaene->Start = Domaene->Ende = Domaene->Frei = (void*)Start;
  Domaene->Obergrenze = (void*)Obergrenze;
  /*  while (Index) {
    if (_(Domaene->Start)<_(Index->Obergrenze) && _(Domaene->Obergrenze)>_(Index->Start))
      our_error("Der Adressraum einer neuer Domaene ueberschneidet sich mit dem einer alten.");
    Index = Index->Nachf;
    }*/
  if ((Domaene->Domaenenart = Domaenenart)!=Strahldomaene && Domaenenart!=Managerdomaene)
    our_error("Unbekannte Domaenenart.");
  Domaene->Haken = Haken;
  Domaene->Vorg = NULL;
  Domaene->Nachf = Domaenenliste;
  Domaenenliste = Domaene;
if (Domaenenliste) Domaenenliste->Vorg = Domaene;  /* AAAAAAAARRRRRRGGGGGHHHHH!!!! */
  if_sst(
    Domaene->AnzahlDomaenenausdehnungen = 
    Domaene->AnzahlGefroreneStrahlen = 
    Domaene->AnzahlGeloeschteStrahlen = 
    Domaene->AnzahlGeunmapptePages = 
    Domaene->AnzahlGemapptePages = 
    Domaene->AnzahlMaxGemapptePages = 0);
  return Domaene;
}


void DM_DomaeneLeeren(volatile DomaenenT Domaene)
{
  DM_munmap(Domaene->Start,Domaene->Ende);  /* Adressbereich freigeben. Nach munmap-Dokumentation (natuerliche In- */
  Domaene->Ende = Domaene->Frei =Domaene->Start;/* terpretation) muss nicht mehr der gesamte Bereich gemappt sein. */
  if_sst(
    Domaene->AnzahlDomaenenausdehnungen = 
    Domaene->AnzahlGefroreneStrahlen = 
    Domaene->AnzahlGeloeschteStrahlen = 
    Domaene->AnzahlGeunmapptePages =
    Domaene->AnzahlGemapptePages = 
    Domaene->AnzahlMaxGemapptePages = 0);
}


void DM_DomaeneLoeschen(volatile DomaenenT Domaene)
{
  DM_DomaeneLeeren(Domaene);                                                            /* Adressbereich freigeben */
  if (Domaene->Vorg)
    Domaene->Vorg->Nachf = Domaene->Nachf;
  else                                                                /* erstes Element der Domaenenliste entfernt */
    Domaenenliste = Domaene->Nachf;
  if (Domaene->Nachf)
    Domaene->Nachf->Vorg = Domaene->Vorg;
  our_dealloc(Domaene);
}

