/**********************************************************************
* Filename : ClasHeuristics.
*
* Kurzbeschreibung : Heuristiken zur Classifikation
*
* Autoren : Andreas Jaeger
*
* 10.03.98: Erstellung
*
*
*
**********************************************************************/

#include <stdlib.h>
#include <string.h>
#include "compiler-extra.h"

#include "general.h"
#include "Ausgaben.h"
#include "ClasFunctions.h"
#include "ClasHeuristics.h"
#include "Ordnungen.h"
#include "Unifikation1.h"

/* Classifikationsroutinen haben alle folgende Aufrufsyntax:
   wtPtr: WeightEntryT
   lhs: TermT
   rhs: TermT
   actualParent: RegelOderGleichungsT
   otherParent: RegelOderGleichungsT
   actualPosition: UTermT : Position in actualParent?
Wo?-> CG/CP
   type ist ClassifyProcT
(WeightEntryT *wtPtr, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent, UTermT actualPosition);
*/


WeightT
CH_AddWeight (TermT lhs, TermT rhs,
	      RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM,
	      RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
	      UTermT actualPosition MAYBE_UNUSEDPARAM,
	      BOOLEAN isCP)
{

  return CF_Phi (lhs, isCP) + CF_Phi (rhs, isCP);
  
}


WeightT
CH_OrdWeight (TermT lhs, TermT rhs,
	      RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM,
	      RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
	      UTermT actualPosition MAYBE_UNUSEDPARAM,
	      BOOLEAN isCP MAYBE_UNUSEDPARAM)
{

  return CF_Phi_KBO (lhs) + CF_Phi_KBO (rhs);
  
}

WeightT
CH_GtWeight (TermT lhs, TermT rhs,
	     RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM,
	     RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
	     UTermT actualPosition MAYBE_UNUSEDPARAM,
	     BOOLEAN isCP)
{
  WeightT res;
#if 0
  WeightT phi1 = CF_Phi (lhs, isCP),
          phi2 = CF_Phi (rhs, isCP);

  if ((phi1 <= 5) && (phi2 <= 5)){ /* Bevorzugen der 3er Permuatationen */
    return min (phi1,phi2);
  }
#endif
  
   switch (ORD_VergleicheTerme(lhs, rhs))
     {
     case Groesser:
       res = CF_Phi (lhs, isCP);
       break;
     case Kleiner:
       res = CF_Phi (rhs, isCP);
       break;
     default:
       res = CF_Phi (lhs, isCP) + CF_Phi (rhs, isCP);
       break;
     }
   return res;
}


WeightT
CH_MixWeight (TermT lhs, TermT rhs,
	      RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM,
	      RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
	      UTermT actualPosition MAYBE_UNUSEDPARAM,
	      BOOLEAN isCP)
{
  WeightT w_lhs, w_rhs;
  WeightT res;

  w_lhs = CF_Phi (lhs, isCP);
  w_rhs = CF_Phi (rhs, isCP);
  
  switch (ORD_VergleicheTerme(lhs, rhs))
    {
    case Groesser:
      res = w_lhs;
      break;
    case Kleiner:
      res = w_rhs;
      break;
    default:
      res = w_lhs + w_rhs;
      break;
    }
  return (w_lhs + w_rhs) * res + res + (w_lhs + w_rhs);
}

WeightT
CH_MixWeight2 (TermT lhs, TermT rhs,
	      RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM,
	      RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
	      UTermT actualPosition MAYBE_UNUSEDPARAM,
	      BOOLEAN isCP)
{
  WeightT w_lhs, w_rhs;
  WeightT res;

  w_lhs = CF_Phi (lhs, isCP);
  w_rhs = CF_Phi (rhs, isCP);
  
  switch (ORD_VergleicheTerme(lhs, rhs))
    {
    case Groesser:
      res = w_lhs;
      break;
    case Kleiner:
      res = w_rhs;
      break;
    default:
      res = w_lhs + w_rhs;
      break;
    }
  return res * 10 + (w_lhs + w_rhs);
}


WeightT
CH_MaxWeight (TermT lhs, TermT rhs,
	      RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM,
	      RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
	      UTermT actualPosition MAYBE_UNUSEDPARAM,
	      BOOLEAN isCP)
{
  WeightT w_lhs, w_rhs;
  
  w_lhs = CF_Phi (lhs, isCP);
  w_rhs = CF_Phi (rhs, isCP);

  return max (w_lhs, w_rhs);
}


/* prod(plus(length), plus(unif)) */
WeightT
CH_Unifikationsmass (TermT lhs, TermT rhs,
		     RegelOderGleichungsT actualParent MAYBE_UNUSEDPARAM,
		     RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
		     UTermT actualPosition MAYBE_UNUSEDPARAM,
		     BOOLEAN isCP)
{
  WeightT w_lhs, w_rhs;
  w_lhs = CF_Phi (lhs, isCP);
  w_rhs = CF_Phi (rhs, isCP);

  return (w_lhs + w_rhs) * U1_Unifikationsmass(lhs, rhs);
}


WeightT
CH_Fifo (TermT lhs MAYBE_UNUSEDPARAM, TermT rhs MAYBE_UNUSEDPARAM,
         RegelOderGleichungsT actualParent  MAYBE_UNUSEDPARAM,
         RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
         UTermT actualPosition MAYBE_UNUSEDPARAM,
         BOOLEAN isCP MAYBE_UNUSEDPARAM)
{
   static WeightT fifo_zaehler = 0;
   return fifo_zaehler++;
  
}


WeightT
CH_Golem (TermT lhs, TermT rhs,
          RegelOderGleichungsT actualParent  MAYBE_UNUSEDPARAM,
          RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
          UTermT actualPosition MAYBE_UNUSEDPARAM,
          BOOLEAN isCP)
{
  return CF_GolemGewicht (lhs, isCP) + CF_GolemGewicht (rhs, isCP);
}

WeightT CH_PolyWeight (TermT lhs, TermT rhs,
          RegelOderGleichungsT actualParent  MAYBE_UNUSEDPARAM,
          RegelOderGleichungsT otherParent MAYBE_UNUSEDPARAM,
          UTermT actualPosition MAYBE_UNUSEDPARAM,
          BOOLEAN isCP)
{
  WeightT res, lphi, rphi;
  
  lphi = CF_Phi (lhs, isCP);
  rphi = CF_Phi (rhs, isCP);

  if (min(lphi,rphi) < 30){
    switch (ORD_VergleicheTerme(lhs, rhs)){
    case Groesser:
      res = lphi;
      break;
    case Kleiner:
      res = rphi;
      break;
    default:
      res = lphi + rphi;
      break;
    }
  }
  else {
    res = min (lphi, rphi);
  }
  res *= CF_PolyZuGross(lhs,rhs) ? 10 : 1;
  return res;
}

