/**********************************************************************
* Filename : NormaleZiele.c
*
* Kurzbeschreibung : Normale Zielverwaltung
*
* Autoren : Andreas Jaeger
*
* 10.10.96: Erstellung aus Zielverwaltung.h heraus
* 22.02.97: Mehrfachaufrufclean. AJ.
*
* Die Funktionen in diesem Modul werden nie direkt aufgerufen, sondern
* nur von der Zielverwaltung!
*
**********************************************************************/

#include <stdlib.h>
#include <string.h>

#include "compiler-extra.h"
#include "antiwarnings.h"
#include "general.h"
#include "pcl.h"

#include "Ausgaben.h"
#include "InfoStrings.h"
#include "MatchOperationen.h"
#include "NFBildung.h"
#include "NormaleZiele.h"
#include "RUndEVerwaltung.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "Termpaare.h"
#include "Zielverwaltung.h"
#include "Zielverwaltung_private.h"
#include "ZielAusgaben.h"

/*********************/
/* Typdeklarationen  */
/*********************/

/*
  reduziert: wird gesetzt, wenn links oder rechts reduziert wurde.  Falls im
  Wiederholenmodus gearbeitet wird, kann somit festgestellt werden, ob links/rechts
  die originalen Ziele sind, oder reduzierte Elemente.
*/
typedef struct privateNormaleZieleTS {
   TermT               links;
   TermT               rechts;
   BOOLEAN             reduziert;
   PCL_ObjektNummerT   orig_pclID;
} privateNormaleZieleT, *privateNormaleZielePtrT;


void NZ_InitZielBeobachtungsProz (ZieleT  *ziel) {

  ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
}


void NZ_ZielEinfuegen (ZieleT * ziel, TermT links, TermT rechts) 
{
   ziel->privat.NZ = (privateNormaleZieleT *) our_alloc (sizeof(privateNormaleZieleT));
   ziel->privat.NZ->links = TO_Termkopie (links);
   ziel->privat.NZ->rechts = TO_Termkopie (rechts);
   ziel->privat.NZ->reduziert = FALSE;
   ziel->privat.NZ->orig_pclID = ziel->pclID;
}


void NZ_LoeschePrivateData (ZieleT *ziel) 
{
   TO_TermLoeschen (ziel->privat.NZ->links);
   TO_TermLoeschen (ziel->privat.NZ->rechts);
   our_dealloc (ziel->privat.NZ);
}

INLINE static void InitElemente (ZieleT *ziel)
{
   if (ziel->privat.NZ->reduziert && ziel->modul == Z_Repeat) {
      TO_TermLoeschen (ziel->privat.NZ->links);
      TO_TermLoeschen (ziel->privat.NZ->rechts);
      ziel->privat.NZ->links = TO_Termkopie (ziel->orginalLinks);
      ziel->privat.NZ->rechts = TO_Termkopie (ziel->orginalRechts);
      ziel->privat.NZ->reduziert = FALSE;
      ziel->pclID = ziel->privat.NZ->orig_pclID;
   }
}

void NZ_AusgabeZiel (ZieleT * ziel, Z_AusgabeArtT Z_AusgabeArt) 
{
  if (ziel->modul == Z_Paramodulation) {
    switch (Z_AusgabeArt) {
    case Z_AusgabeAlles:
    case Z_AusgabeInitial:
      IO_DruckeFlex ("%t ?= %t\n",
                     ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      IO_DruckeFlex ("        using narrowing to prove %t ?= %t\n", 
                     ziel->orginalLinks, ziel->orginalRechts);
      break;
    case Z_AusgabeKurz:
      IO_DruckeFlex ("true ?= false, current %t ?= %t\n", 
                     ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      IO_DruckeFlex ("        using narrowing to prove %t ?= %t\n", 
                     ziel->orginalLinks, ziel->orginalRechts);
      break;
    case Z_AusgabeSpez:
      IO_DruckeFlex("\t%t = %t\n", ziel->orginalLinks, ziel->orginalRechts);
      break;
    }
  } else {
    switch (Z_AusgabeArt) {
    case Z_AusgabeAlles:
    case Z_AusgabeInitial:
      IO_DruckeFlex(" current  %t ?= %t\n", ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      break;
    case Z_AusgabeKurz:
      IO_DruckeFlex("%t = %t", ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      break;
    case Z_AusgabeSpez:
      IO_DruckeFlex("\t%t = %t\n", ziel->orginalLinks, ziel->orginalRechts);
      break;
    }
  }
}

   
void NZ_ZielBeobachtung (ZieleT *ziel, Z_BeobachtungsDurchlaufProcT AP)  
{
   AP (ziel->privat.NZ->links, ziel->privat.NZ->rechts, ziel->nummer);
}


BOOLEAN NZ_TesteZielErledigt (ZieleT *ziel) 
{
   if (TO_TermeGleich(ziel->privat.NZ->links, ziel->privat.NZ->rechts)){
      PCL_ZielBewiesen(&ziel->pclID, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      if (IO_AusgabeLevel() >= 2) {
         fprintf(IO_Strom, "joined goal:    %8u  ", ziel->nummer);
         IO_DruckeFlex("%t ?= %t to %t\n", ziel->orginalLinks, ziel->orginalRechts,
                       ziel->privat.NZ->rechts);
      }
      else
         fprintf(IO_Strom, "joined goal:    %8u\n", ziel->nummer);
      ZA_ZielZusammengefuehrtDrucken(); /* ... und somit zusammengefuehrt.*/
      ziel->status = Z_Joined;
      return TRUE;
   }
   return FALSE;
}


BOOLEAN NZ_ZielErledigtMitNeuemFaktum (ZieleT *ziel, RegelOderGleichungsT Faktum)
{
   BOOLEAN reduziert;

   InitElemente (ziel);
   if (ziel->modul == Z_Repeat) {
     reduziert = NF_NormalformRE2(ziel->privat.NZ->links, ziel->privat.NZ->rechts, &ziel->pclID, Ziel);
   } else {
     PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
     reduziert = NF_NormalformstRE(Faktum, ziel->privat.NZ->links);
     PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
     reduziert |= NF_NormalformstRE(Faktum, ziel->privat.NZ->rechts);
   }
   if (reduziert){
      ziel->privat.NZ->reduziert = TRUE;
      if (ziel->beobachtProz)
         ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      if (ziel->modul != Z_Repeat)
	ZA_ZielReduziertDrucken (ziel->nummer, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      /* Ziel .. konnte reduziert werden: l =? r */
      if (NZ_TesteZielErledigt (ziel))
         return TRUE;
   }
   return FALSE;
}


BOOLEAN NZ_ZielFinalerDurchlauf (ZieleT * ziel) 
{
   BOOLEAN reduziert;

   InitElemente (ziel);
   reduziert = NF_NormalformRE2(ziel->privat.NZ->links, ziel->privat.NZ->rechts, &ziel->pclID, Ziel);
   if (reduziert){
      ziel->privat.NZ->reduziert = TRUE;
      if (ziel->beobachtProz)
         ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      if (ziel->modul != Z_Repeat)
	 ZA_ZielReduziertDrucken (ziel->nummer, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      /* Ziel .. konnte reduziert werden: l =? r */
      if (TO_TermeGleich(ziel->privat.NZ->links, ziel->privat.NZ->rechts)){
         PCL_ZielBewiesen(&ziel->pclID, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
         ZA_ZielZusammengefuehrtDrucken(); /* ... und somit zusammengefuehrt.*/
         ziel->status = Z_Joined;
         return TRUE;
      }
   }

   return FALSE;
}


/*=================================================================================================================
  =  Zielbehandlung                                                                                               =
  =================================================================================================================*/


static void NZ_ParseInfoStrings (InfoStringT DEF MAYBE_UNUSEDPARAM,
                                 InfoStringT USR MAYBE_UNUSEDPARAM) 
{
   ;
}


void NZ_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   switch (what) {
   case ParamEvaluate:
      NZ_ParseInfoStrings (DEF, USR);
      break;
   case ParamExplain:
      fprintf (IO_Strom, "Classical Goal Treatment (without MNF)\n");
      fprintf (IO_Strom, "**************************************\n");
      fprintf (IO_Strom, "To select: snf (keeps goals always reduced)\n");
      fprintf (IO_Strom, "To select: rnf (tries to reduce initial goal)\n");
      break;
   }
}


BOOLEAN NZ_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
   NZ_ParseInfoStrings (DEF, USR);
   return TRUE;
}

