/*=================================================================================================================
  =                                                                                                               =
  =  ChaseManhattan.c                                                                                             =
  =                                                                                                               =
  =                                                                                                               =
  =  23.10.97  Thomas.                                                                                            =
  =                                                                                                               =
  =================================================================================================================*/

#include "Ausgaben.h"
#include "ChaseManhattan.h"
#include "DSBaumOperationen.h"
#include "Hauptkomponenten.h"
#include "parser.h"
#include "pcl.h"
#include <string.h>
#include "RUndEVerwaltung.h"
#include "SpeicherVerwaltung.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "Termpaare.h"
#include "Unifikation2.h"


/*=================================================================================================================*/
/*= I. Typedeklarationen                                                                                          =*/
/*=================================================================================================================*/

#define CM_TEST 0

typedef enum {parallel, gekreuzt} VerschwurbelungsT;

typedef struct HoehererFaktenTS {
  TermT links,
        rechts,
        OriginalLinks,
        OriginalRechts;
  long int Nummer;
  RegelOderGleichungsT Faktum;
  VerschwurbelungsT Verschwurbelung;
  unsigned int AnzVorigerBindungen;
  KantenT i;
  GNTermpaarT j;
  BOOLEAN k;
  unsigned int NeuestesX;
  struct HoehererFaktenTS *Vorg,
                          *Nachf;
} *HoehererFaktenT;
#define /*KantenT **/ i(/*HoehererFaktenT*/ Praemisse)                                                              \
  (&((Praemisse)->i))
#define /*GNTermpaarT **/ j(/*HoehererFaktenT*/ Praemisse)                                                          \
  (&((Praemisse)->j))
#define /*BOOLEAN **/ k(/*HoehererFaktenT*/ Praemisse)                                                              \
  (&((Praemisse)->k))

static SV_SpeicherManagerT CM_FreispeicherManager;


typedef struct EinlagenTS {
  HoehererFaktenT Praemissen,
                  Konklusionen;
  BOOLEAN KonklusionenGezogen;
  struct EinlagenTS *Nachf;
} *EinlagenT;

static EinlagenT ChaseManhattan;

static eq_list dcdr(eq_list *index)
{
  return (*index = eq_rest(*index));
}

#define /*EinlagenT*/ TP_Einlage(/*GNTermpaarT*/ Termpaar)                                                          \
  ((EinlagenT)((Termpaar)->SUEInfo))
#define /*void*/ TP_EinlageSei(/*GNTermpaarT*/ Termpaar, /*EinlagenT*/ Einlage)                                     \
  ((Termpaar)->SUEInfo = (void *)(Einlage))

#define /*HoehererFaktenT*/ TP_Konklusionsmuster(/*GNTermpaarT*/ Termpaar)                                          \
  ((HoehererFaktenT)((Termpaar)->Blatt))
#define /*void*/ TP_KonklusionsmusterSei(/*GNTermpaarT*/ Termpaar, /*HoehererFaktenT*/ Konklusion)                  \
  ((Termpaar)->Blatt = (KantenT)(Konklusion))

#define /*HoehererFaktenT*/ TP_Praemisseninstanzen(/*GNTermpaarT*/ Termpaar)                                        \
  ((HoehererFaktenT)((Termpaar)->Gegenrichtung))
#define /*void*/ TP_PraemisseninstanzenSei(/*GNTermpaarT*/ Termpaar, /*HoehererFaktenT*/ Praemisseninstanzen)       \
  ((Termpaar)->Gegenrichtung = (GNTermpaarT)Praemisseninstanzen)



/*=================================================================================================================*/
/*= II. Eintragung der Individuenvariablenbezeichner                                                              =*/
/*=================================================================================================================*/

static SymbolT ersteVariable,
               letzteVariable;

typedef struct KonversionsTS {
  char *Bezeichner;
  struct KonversionsTS *Nachf;
} *KonversionsT;
static KonversionsT Konversion;

static void AusTermHolen(term Term)
{
  if variablep(Term) {
    KonversionsT Index = Konversion;
    while (Index)
      if (strcmp(Index->Bezeichner,Term->RealVariableName))
        Index = Index->Nachf;
      else
        return;
    (Index = struct_alloc(struct KonversionsTS))->Nachf = Konversion;
    (Konversion = Index)->Bezeichner = Term->RealVariableName;
    SO_NaechstesVarSymbD(letzteVariable);
  }
  else {
    signed int i;
    for (i=0; i<sig_arity(term_func(Term)); i++)
      AusTermHolen(term_arg(Term,i));
  }
}

static void AusAbschnittHolen(eq_list *Index)
{
  while (*Index && eq_first_lhs(*Index)) {
    term l = eq_first_lhs(*Index),
         r = eq_first_rhs(*Index);
    AusTermHolen(l);
    AusTermHolen(r);
    dcdr(Index);
  }
  if (*Index)
    dcdr(Index);
}


static void IndividuenvariablenEintragen(void)
{
  eq_list index_eq = get_equations(),
          index_cc = get_conclusions();
  ersteVariable = SO_VarDelta(BO_NeuesteBaumvariable(),CM_VariablenDelta);
  letzteVariable = SO_VorigesVarSymb(ersteVariable);
  Konversion = NULL;
  do {
    AusAbschnittHolen(&index_eq);
    AusAbschnittHolen(&index_cc);
  } while (index_eq);
}



/*=================================================================================================================*/
/*= III. Transformationen                                                                                         =*/
/*=================================================================================================================*/

static TermT VariablenTransformation(term Variable)
{
  SymbolT Symbol = letzteVariable;
  KonversionsT Index = Konversion;
  while (strcmp(Variable->RealVariableName,Index->Bezeichner)) {
    SO_VorigesVarSymbD(Symbol);
    Index = Index->Nachf;
  }
  return TO_NullstelligerTerm(Symbol);
}

static SymbolT SymbolTransformation(func_symb symb)
{
  SymbolT F;
  if (sig_is_fvar(symb)) {
    func_symb_list sig = get_sig();
    SO_forFktVarSymbole(F) {
      func_symb f;
      while ((!sig_is_fvar(f = sig_first(sig))))
        sig = sig_rest(sig);
      if (f==symb)
        return F;
      sig = sig_rest(sig);
    }
  }
  else
    SO_forFktSymbole(F)
      if (!strcmp(IO_FktSymbName(F),map_func(symb)) && SO_Stelligkeit(F)==(unsigned)sig_arity(symb))
        return F;
  return SO_ErstesVarSymb;
}


static TermT TermTransformation(term Term)
{
  if (variablep(Term))
    return VariablenTransformation(Term);
  else {                                                                                       /* fuer Konstanten: */
    SymbolT Symbol = SymbolTransformation(term_func(Term));
    if (SO_SymbGleich(Symbol,SO_ErstesVarSymb))
      return NULL;
    else {
      short int i;
      UTermT Neu=NULL, Alt=NULL, Ende=NULL;               /* alle Teilterme in umgekehrter Reihenfolge durchlaufen */
      for(i=sig_arity(term_func(Term))-1; i>=0; i--) {     /* Invariante: Alt zeigt auf zuletzt erzeugten Teilterm */
        if ((Neu = TermTransformation(term_arg(Term,i)))) {
          TO_NachfBesetzen(TO_TermEnde(Neu),Alt);                                              /* Nachf-Verkettung */
          if (!Ende)                   /* vermerkt Endzeiger des letzten Teilterms; wird Endzeiger des Gesamtterms */
            Ende = TO_TermEnde(Neu);
          Alt = Neu;
        }
        else {
          if (Ende)
            TO_TermzellenlisteLoeschen(Alt,Ende);
          return NULL;
        }
      }
      TO_TermzelleHolen(Neu);                                                   /* Termzelle fuer Topsymbol bilden */
      TO_NachfBesetzen(Neu,Alt);
      TO_EndeBesetzen(Neu,Ende ? Ende : Neu);         /* Ende-Zeiger mit Ende aus Schleife besetzen bzw. mit Term- */
      TO_SymbolBesetzen(Neu,Symbol);                                       /* zelle selbst, falls Konstantensymbol */
      return Neu;
    }
  }
}

static unsigned int Gleichungsnummer;
static union {term d; TermT w;} UnbekanntLinks,
                                UnbekanntRechts;
static SymbolT FreiePatternvariable;

static HoehererFaktenT AbschnittTransformation(eq_list *Index, unsigned int *Anzahl)
{
  HoehererFaktenT Anfang = NULL,
                  Ende = NULL;
  *Anzahl = 0;
  while (*Index && eq_first_lhs(*Index)) {
    term l = eq_first_lhs(*Index),
         r = eq_first_rhs(*Index);
    TermT links = TermTransformation(l),
          rechts = TermTransformation(r);
    (*Anzahl)++;
    if (links && rechts) {
      HoehererFaktenT Neu;
      SV_Alloc(Neu,CM_FreispeicherManager,HoehererFaktenT);
      Neu->links = links;
      Neu->Nummer = ++Gleichungsnummer;
      Neu->rechts = rechts;
      Neu->Nachf = NULL;
      Neu->Vorg = Ende;
      if (Ende)
        Ende->Nachf = Neu;
      Ende = Neu;
      if (!Anfang)
        Anfang = Neu;
    }
    else {
      if (!UnbekanntLinks.d) {
        UnbekanntLinks.d = l;
        UnbekanntRechts.d = r;
      }
      if (links)
        TO_TermLoeschen(links);
      if (rechts)
        TO_TermLoeschen(rechts);
    }
    dcdr(Index);
  }
  if (*Index)
    dcdr(Index);
  return Anfang;
}


static BOOLEAN PatternvariablenUngebunden(HoehererFaktenT Praemissen, HoehererFaktenT Konklusionen)
{
  do {
    UTermT i = Konklusionen->links;
    BOOLEAN iINlinks = true;
    do {
      SymbolT F = TO_TopSymbol(i);
      if (SO_SymbIstFVar(F)) {
        HoehererFaktenT j = Praemissen;
        for (; j; j = j->Nachf)
          if (TO_SymbolEnthalten(F,j->links) || TO_SymbolEnthalten(F,j->rechts))
            break;
        if (!j) {
          FreiePatternvariable = F;
          UnbekanntLinks.w = Konklusionen->links;
          UnbekanntRechts.w = Konklusionen->rechts;
          return TRUE;
        }
      }
    } while ((i = TO_Schwanz(i)) ? TRUE : iINlinks ? (iINlinks = FALSE, i = Konklusionen->rechts, TRUE) : FALSE);
  } while ((Konklusionen = Konklusionen->Nachf));
  return FALSE;
}

static void ListeLoeschen(HoehererFaktenT Liste)
{
  while (Liste) {
    HoehererFaktenT Nachf = Liste->Nachf;
    TO_TermeLoeschen(Liste->links,Liste->rechts);
    SV_Dealloc(CM_FreispeicherManager,Liste)
    Liste = Nachf;
  }
}

static RegelOderGleichungsT ErsteKonklusion,
                            LetzteKonklusion;
static unsigned int MaximaleAnzahlPraemissen;

static void BankEroeffnen(void)
{
  eq_list index_eq = get_equations(),
          index_cc = get_conclusions();
  EinlagenT Ende = NULL;
  MaximaleAnzahlPraemissen = 0;
  Gleichungsnummer = 0;
  do {
    unsigned int AnzahlPraemissen,
                 AnzahlKonklusionen;                                                      /* aus Symmetriegruenden */
    HoehererFaktenT Praemissen = ( UnbekanntLinks.d = NULL,
                                   FreiePatternvariable = SO_ErstesVarSymb,
                                   AbschnittTransformation(&index_eq,&AnzahlPraemissen)),
                    Konklusionen = AbschnittTransformation(&index_cc,&AnzahlKonklusionen);
    if (UnbekanntLinks.d || PatternvariablenUngebunden(Praemissen,Konklusionen)) {
      ListeLoeschen(Praemissen);
      ListeLoeschen(Konklusionen);
      IO_DruckeFlex("\n\n********************************************************************\n");
      IO_DruckeFlex("*** ChaseManhattan: ");
      if (SO_SymbGleich(FreiePatternvariable,SO_ErstesVarSymb)) {
        IO_DruckeFlex("Function constant not found in specification ***\n");
        IO_DruckeFlex("***                 ");
        print_pair(UnbekanntLinks.d," = ",UnbekanntRechts.d);
      }
      else {
        fprintf(IO_Strom,"Function variable %s free in conclusion %*s***\n",
          IO_FktSymbName(FreiePatternvariable),max(0,7-(signed)strlen(IO_FktSymbName(FreiePatternvariable))),""); 
        IO_DruckeFlex("***                 %t = %t\n",UnbekanntLinks.w,UnbekanntRechts.w);
      }
      IO_DruckeFlex("***                 Skipping tuple EQUATIONS-CONCLUSION          ***\n");
      IO_DruckeFlex("********************************************************************\n\n\n");
    }
    else {
      EinlagenT Neu = struct_alloc(struct EinlagenTS);
      Neu->Praemissen = Praemissen;
      Neu->Konklusionen = Konklusionen;
      Neu->KonklusionenGezogen = FALSE;
      Neu->Nachf = NULL;
      if (Ende)
        Ende->Nachf = Neu;
      Ende = Neu;
      if (!ChaseManhattan)
        ChaseManhattan = Neu;
      Maximieren(&MaximaleAnzahlPraemissen,AnzahlPraemissen);
    }
  } while (index_cc);
  ErsteKonklusion = LetzteKonklusion = NULL;
}



/*=================================================================================================================*/
/*= IV. Ausgabefunktion                                                                                           =*/
/*=================================================================================================================*/

static void ListeAusgeben(HoehererFaktenT Liste)
{
  while (Liste) {
    fprintf(IO_Strom,"(%*ld) ",LaengeNatZahl(Gleichungsnummer),Liste->Nummer);
    IO_DruckeTermPaar(Liste->links,"=",Liste->rechts);
    IO_LF();
    Liste = Liste->Nachf;
  }
}

void CM_BankAusgeben(void)
{
  EinlagenT Index = ChaseManhattan;
  IO_DruckeFlex("\n\nContents of data base:\n----------------------\n");
  while (Index) {
    IO_LF();
    ListeAusgeben(Index->Praemissen);
    IO_DruckeFlex("======================\n");
    ListeAusgeben(Index->Konklusionen);
    Index = Index->Nachf;
  }
}



/*=================================================================================================================*/
/*= V. Variablenmanagement                                                                                        =*/
/*=================================================================================================================*/

/* Variablenbereiche in ChaseManhattan sind
        x1            < ... < BO_NeuesteBaumvariable() < ... 
      < ersteVariable <= ... <= letzteVariable 
      < temporaere Umbenennungen aus Mengenunifikation,
   wenn die Praemissen und Konklusionen Variablen enthalten;
   ansonsten ist letzteVariable = SO_VorigesVarSymb(ersteVariable). - 
   Steigende Ausdehnungen sollten an SO, TO, U2 gemeldet werden, was im folgenden als Propagieren firmiert.
   Wie entwickeln sich die Ausdehnungen?
   
   1. Es muss einen sinnvollen Initialzustand geben. Dabei ist letzteVariable zu propagieren.
   2. Mit BO_NeuesteBaumvariable() verschiebt sich letzteVariable nach oben -> letzteVariable propagieren
   3. Wird bei der Mengenunifikation ein weiteres Faktum ergriffen, so erhaelt dieses Faktum durch Umbenennung
      schliesslich Variablen bis NeuestesX+NeuesteBaumvariable -> diesen Wert propagieren. */

static void ListeVarUmbenennen(HoehererFaktenT Liste, SymbolT Delta)
{
  while (Liste) {
    TO_VariablenUmbenennen(Liste->links,Delta);
    TO_VariablenUmbenennen(Liste->rechts,Delta);
    Liste = Liste->Nachf;
  }
}

static void VariablePropagieren(SymbolT Variable)
{
  SO_VariablenfeldAnpassen(Variable);
  TO_FeldVariablentermeAnpassen(Variable);
  U2_VariablenfelderAnpassen(Variable);
}

void CM_VariablenmanagementAnpassen(SymbolT VarAusRegel)
{
  if (ChaseManhattan && SO_VariableIstNichtNeuer(ersteVariable,VarAusRegel)) {
    SymbolT ersteVariableNeu = SO_VarDelta(VarAusRegel,CM_VariablenDelta);
    signed int Delta = SO_VarNummer(ersteVariableNeu)-SO_VarNummer(ersteVariable);
    EinlagenT Index = ChaseManhattan;
    while (Index) {
      ListeVarUmbenennen(Index->Praemissen,Delta);
      ListeVarUmbenennen(Index->Konklusionen,Delta);
      Index = Index->Nachf;
    }
    ersteVariable = ersteVariableNeu;
    letzteVariable = SO_NtesNaechstesVarSymb(letzteVariable,Delta);
    VariablePropagieren(letzteVariable);
  }
}



/*=================================================================================================================*/
/*= VI. Mengenunifikation                                                                                         =*/
/*=================================================================================================================*/

#define /*HoehererFaktenT*/ ErstePraemisse(/*EinlagenT*/ Einlage)                                                   \
  (Einlage->Praemissen==Anfangspraemisse ? Einlage->Praemissen->Nachf : Einlage->Praemissen)
#define /*HoehererFaktenT*/ NaechstePraemisse(/*HoehererFaktenT*/ Praemisse)                                        \
  ((Praemisse)->Nachf==Anfangspraemisse ? (Praemisse)->Nachf->Nachf : (Praemisse)->Nachf)
#define /*HoehererFaktenT*/ VorigePraemisse(/*HoehererFaktenT*/ Praemisse)                                          \
  ((Praemisse)->Vorg==Anfangspraemisse ? (Praemisse)->Vorg->Vorg : (Praemisse)->Vorg)


#define /*BOOLEAN*/ AufNaechstePraemisse(/*void*/)                                                                  \
  ( NaechstePraemisse(Praemisse) ?                                                                                  \
      ( Praemisse->NeuestesX = NeuestesX,                                                                           \
        NeuestesX = RE_neuesteVariable(Praemisse->Faktum),                                                          \
        Praemisse = NaechstePraemisse(Praemisse))                                                                   \
      : FALSE)


static void Hinbenennen(HoehererFaktenT Praemisse, HoehererFaktenT Anfangspraemisse, 
                                                                                SymbolT NeuestesX, EinlagenT Einlage)
{
  BOOLEAN KopierenNoetig = Praemisse->Faktum==Anfangspraemisse->Faktum;
  HoehererFaktenT Index = ErstePraemisse(Einlage);
  for (; Index!=Praemisse; Index = NaechstePraemisse(Index))
    KopierenNoetig = KopierenNoetig || Praemisse->Faktum==Index->Faktum;
  if (KopierenNoetig) {
    Praemisse->OriginalLinks = TP_LinkeSeite(Praemisse->Faktum);
    Praemisse->OriginalRechts = TP_RechteSeite(Praemisse->Faktum);
    TP_LinkeSeite(Praemisse->Faktum) = TO_TermkopieVarUmbenannt(TP_LinkeSeite(Praemisse->Faktum),NeuestesX);
    TP_RechteSeite(Praemisse->Faktum) = TO_TermkopieVarUmbenannt(TP_RechteSeite(Praemisse->Faktum),NeuestesX);
  }
  else {
    Praemisse->OriginalLinks = NULL;
    RE_RegelUmbenennen(Praemisse->Faktum,SO_VarNummer(NeuestesX));
  }
}

static void Herbenennen(HoehererFaktenT Praemisse, SymbolT NeuestesX)
{
  if (Praemisse->OriginalLinks) {
    TP_LinksRechtsLoeschen(Praemisse->Faktum);
    TP_LinkeSeite(Praemisse->Faktum) = Praemisse->OriginalLinks;
    TP_RechteSeite(Praemisse->Faktum) = Praemisse->OriginalRechts;
  }
  else
    RE_RegelUmbenennen(Praemisse->Faktum,-SO_VarNummer(NeuestesX));
}


#define /*void*/ AufErstesFaktum(/*void*/)                                                                          \
  ( Praemisse->Faktum = RE_ErstesFaktumM(i(Praemisse),j(Praemisse),k(Praemisse)),                                   \
    Hinbenennen(Praemisse,Anfangspraemisse,NeuestesX,Einlage),                                                      \
    VariablePropagieren(BO_NeuesteBaumvariable()+NeuestesX),                                                        \
    Praemisse->Verschwurbelung = parallel,                                                                          \
    Praemisse->AnzVorigerBindungen = U2_AnzBindungen()                                                              \
  )

static BOOLEAN AufNaechstesFaktum_(HoehererFaktenT *Praemisse, HoehererFaktenT Anfangspraemisse, 
                                                                               SymbolT *NeuestesX, EinlagenT Einlage)
{
  do {
    U2_BindungenAufhebenBis((*Praemisse)->AnzVorigerBindungen);
    if ((*Praemisse)->Verschwurbelung==parallel) {
      (*Praemisse)->Verschwurbelung = gekreuzt;
      return TRUE;
    }
    else {
      RegelOderGleichungsT NaechstesFaktum = RE_NaechstesFaktumM(i(*Praemisse),j(*Praemisse),k(*Praemisse));
      Herbenennen(*Praemisse,*NeuestesX);
      if (NaechstesFaktum) {
        (*Praemisse)->Verschwurbelung = parallel;
        (*Praemisse)->Faktum = NaechstesFaktum;
        Hinbenennen(*Praemisse,Anfangspraemisse,*NeuestesX,Einlage);
        return TRUE;
      }
    }
  } while ((*Praemisse = VorigePraemisse(*Praemisse)) ? (*NeuestesX = (*Praemisse)->NeuestesX, TRUE) : FALSE);
  return FALSE;
}
#define /*BOOLEAN*/ AufNaechstesFaktum(/*void*/)                                                                    \
  AufNaechstesFaktum_(&Praemisse,Anfangspraemisse,&NeuestesX,Einlage)

#define /*TermT*/ L                                                                                                 \
  Praemisse->links
#define /*TermT*/ R                                                                                                 \
  Praemisse->rechts
#define /*TermT*/ l                                                                                                 \
  (Praemisse->Verschwurbelung==parallel ? TP_LinkeSeite(Praemisse->Faktum) : TP_RechteSeite(Praemisse->Faktum))
#define /*TermT*/ r                                                                                                 \
  (Praemisse->Verschwurbelung==gekreuzt ? TP_LinkeSeite(Praemisse->Faktum) : TP_RechteSeite(Praemisse->Faktum))

#define /*HoehererFaktenT*/ ErsteAnfangspraemisse(/*RegelOderGleichungsT*/ Faktum, /*EinlagenT*/ Einlage)           \
  ( Einlage->Praemissen->Verschwurbelung = parallel,                                                                \
    Einlage->Praemissen->Faktum = Faktum,                                                                           \
    Einlage->Praemissen->NeuestesX = SO_VorErstemVarSymb,                                                           \
    /* Einlage->Praemissen->OriginalLinks = NULL,                                         logisch nicht notwendig */\
    Einlage->Praemissen )
    

#define /*BOOLEAN*/ AufNaechsteAnfangspraemisse(/*RegelOderGleichungsT*/ Faktum)                                    \
  ( Anfangspraemisse->Verschwurbelung==parallel ? (Anfangspraemisse->Verschwurbelung = gekreuzt, TRUE)              \
    :                                                                                                               \
    ( Anfangspraemisse = Anfangspraemisse->Nachf) ?                                                                 \
        ( Anfangspraemisse->Verschwurbelung = parallel,                                                             \
          Anfangspraemisse->Faktum = Faktum,                                                                        \
          Anfangspraemisse->NeuestesX = SO_VorErstemVarSymb,                                                        \
          /* Anfangspraemisse->OriginalLinks = NULL,                                      logisch nicht notwendig */\
          TRUE )                                                                                                    \
    : FALSE )

#define /*TermT*/ AnfangLinks                                                                                       \
  (Anfangspraemisse->Verschwurbelung==parallel ? Anfangspraemisse->links : Anfangspraemisse->rechts)
#define /*TermT*/ AnfangRechts                                                                                      \
  (Anfangspraemisse->Verschwurbelung ? Anfangspraemisse->links : Anfangspraemisse->rechts)

static void Erfolgsaktion(EinlagenT Einlage);

static void Mengenunifikation(RegelOderGleichungsT Faktum, EinlagenT Einlage)
{
  TermT links = TP_LinkeSeite(Faktum),
        rechts = TP_RechteSeite(Faktum);
  SymbolT NeuestesX = RE_neuesteVariable(Faktum)+letzteVariable;
  HoehererFaktenT Anfangspraemisse = ErsteAnfangspraemisse(Faktum,Einlage);
  do {
    U2_SubstAufIdentitaetSetzen();
    if (U2_DoppeltWeiterUnifiziert(AnfangLinks,links, AnfangRechts,rechts)) {
      HoehererFaktenT Praemisse = ErstePraemisse(Einlage);       /* IO_DruckeFlex("Anfangspraemisse gefunden!\n"); */
      if (Praemisse) {
        AufErstesFaktum();
        do {
        LosGehts:                                                            /* Invariante: Praemisse, Faktum <> 0 */
          if (U2_DoppeltWeiterUnifiziert(L,l, R,r)) {
            if (AufNaechstePraemisse()) {
              AufErstesFaktum();
              goto LosGehts;
            }
            else
              Erfolgsaktion(Einlage);
	  }
        } while (AufNaechstesFaktum());
      }
      else
        Erfolgsaktion(Einlage);
    }
  } while (AufNaechsteAnfangspraemisse(Faktum));
} 



/*=================================================================================================================*/
/*= VII. Aktivierung fuer jedes neue Faktum                                                                       =*/
/*=================================================================================================================*/

void CM_IntegriertesFaktumMelden(RegelOderGleichungsT Faktum)
{
  EinlagenT Index = ChaseManhattan;
  for (; Index; Index = Index->Nachf)
    if (Index->Praemissen)
      Mengenunifikation(Faktum,Index);
    else if (!Index->KonklusionenGezogen) {
      U2_SubstAufIdentitaetSetzen();
      Erfolgsaktion(Index);
    }
}


static void Erfolgsaktion(EinlagenT Einlage)
{
  HoehererFaktenT Index;
  Einlage->KonklusionenGezogen = TRUE;
  for (Index = Einlage->Konklusionen; Index; Index = TP_Nachf(Index)) {
    TermT links = U2_MGU(Index->links),
          rechts = U2_MGU(Index->rechts);
    if (TO_TermeGleich(links,rechts)) 
      TO_TermeLoeschen(links,rechts);
    else if (TO_TermEnthaeltFVar(links) || TO_TermEnthaeltFVar(rechts)) {
      IO_DruckeFlex("free database conclusion: %t = %t\n",links,rechts);
      IO_DruckeFlex("     (yet unsupported)\n");
      TO_TermeLoeschen(links,rechts);
    }
    else {
      RegelOderGleichungsT Neu;
      HoehererFaktenT Praemissenindex = Einlage->Praemissen,
                      Ende = NULL;
      IO_DruckeFlex("database conclusion:      %t = %t\n",links,rechts);
      TP_GNTermpaarHolen(Neu);
      TP_LinkeSeite(Neu) = links;
      TP_RechteSeite(Neu) = rechts;
      TP_NachfSei(Neu,NULL);
      TP_VorgSei(Neu,LetzteKonklusion);
      if (PCL_EsWirdProtokolliert()) {
        TP_EinlageSei(Neu,Einlage);
        TP_KonklusionsmusterSei(Neu,Index);
        TP_PraemisseninstanzenSei(Neu,NULL);
        while (Praemissenindex) {
          HoehererFaktenT neu;
          SV_Alloc(neu,CM_FreispeicherManager,HoehererFaktenT);
          neu->Faktum = Praemissenindex->Faktum;
          neu->Verschwurbelung = Praemissenindex->Verschwurbelung;
          neu->Nachf = NULL;
          if (Ende)
            Ende->Nachf = neu;
          else
            TP_PraemisseninstanzenSei(Neu,neu);
          Ende = neu;
          Praemissenindex = Praemissenindex->Nachf;
        }
      }
      if (LetzteKonklusion)
        TP_NachfSei(LetzteKonklusion,Neu);
      LetzteKonklusion = Neu;
      if (!ErsteKonklusion)
        ErsteKonklusion  = Neu;
    }
  }
#if CM_TEST
  Index = Einlage->Praemissen;
  for (; Index; Index = Index->Nachf)
    IO_DruckeFlex("%t=%t  @=  %t=%t\n",Index->links,Index->rechts,
      Index->Verschwurbelung==parallel ? TP_LinkeSeite(Index->Faktum) : TP_RechteSeite(Index->Faktum),
      Index->Verschwurbelung==gekreuzt ? TP_LinkeSeite(Index->Faktum) : TP_RechteSeite(Index->Faktum));
  U2_DruckeMGU();
#endif
}



/*=================================================================================================================*/
/*= VIII. Ergebnispublikation fuer PCL                                                                            =*/
/*=================================================================================================================*/

RegelOderGleichungsT CM_ErsteKonklusionsinstanz(void)
{ 
  return ErsteKonklusion; 
}

RegelOderGleichungsT CM_NaechsteKonklusionsinstanz(void)
{
  RegelOderGleichungsT Alt = ErsteKonklusion;
  if (!(ErsteKonklusion = TP_Nachf(ErsteKonklusion)))
    LetzteKonklusion = NULL;
  TP_GNTermpaarLoeschen(Alt);
  return ErsteKonklusion;
}

RegelOderGleichungsT CM_Konklusion(RegelOderGleichungsT Konklusionsinstanz)
{ 
  return (RegelOderGleichungsT)TP_Konklusionsmuster(Konklusionsinstanz);
}


static void LRTausch(HoehererFaktenT Instanz, HoehererFaktenT Praemisse)
{
  if (Instanz && Instanz->Verschwurbelung==gekreuzt) {
    TermT links = Praemisse->links;
    Praemisse->links = Praemisse->rechts;
    Praemisse->rechts = links;
  }
}

static HoehererFaktenT Praemissenindex,
                       Instanzenindex;

RegelOderGleichungsT CM_ErstePraemisse(RegelOderGleichungsT Konklusion)
{
  Praemissenindex = TP_Einlage(Konklusion)->Praemissen;
  Instanzenindex = TP_Praemisseninstanzen(Konklusion);
  LRTausch(Instanzenindex, Praemissenindex);
  return (RegelOderGleichungsT) Praemissenindex;
}

RegelOderGleichungsT CM_NaechstePraemisse(void)
{ 
  LRTausch(Instanzenindex, Praemissenindex);
  Praemissenindex = Praemissenindex->Nachf;
  Instanzenindex = Instanzenindex->Nachf;
  LRTausch(Instanzenindex, Praemissenindex);
  return (RegelOderGleichungsT) Praemissenindex;
}


RegelOderGleichungsT CM_ErstePraemisseninstanz(RegelOderGleichungsT Konklusion)
{
  Instanzenindex = TP_Praemisseninstanzen(Konklusion);
  return Instanzenindex ? Instanzenindex->Faktum : NULL;
}

RegelOderGleichungsT CM_NaechstePraemisseninstanz(void)
{ 
  HoehererFaktenT Alt = Instanzenindex;
  Instanzenindex = Instanzenindex->Nachf;
  SV_Dealloc(CM_FreispeicherManager,Alt);
  return Instanzenindex ? Instanzenindex->Faktum : NULL;
}



/*=================================================================================================================*/
/*= IX. Initialisierung und Aufraeumen                                                                            =*/
/*=================================================================================================================*/

void CM_InitApostposteriori(char *Dateiname)
{
  parse_spec(Dateiname);
  SO_FunktionsvariablenEintragen();
  IO_FunktionsvariablenEintragen();
  U2_FunktionsvariablenEintragen();
  IndividuenvariablenEintragen();
  VariablePropagieren(letzteVariable);
  SV_ManagerInitialisieren(&CM_FreispeicherManager,struct HoehererFaktenTS,
    CM_AnzPraemissenUndKonklusionenJeBlock,"Praemissen und Konklusionen");
  BankEroeffnen();
  CM_BankAusgeben();
  clear_spec();
}

unsigned int CM_MaximaleAnzahlPraemissen(void)
{
  return MaximaleAnzahlPraemissen;
}

void CM_DatenbankLeeren(void)
{
  while (ChaseManhattan) {
    ListeLoeschen(ChaseManhattan->Praemissen);
    ListeLoeschen(ChaseManhattan->Konklusionen);
    ChaseManhattan = ChaseManhattan->Nachf;
  }
  while (ErsteKonklusion) {
    LetzteKonklusion = TP_Nachf(ErsteKonklusion);
    TP_LinksRechtsLoeschen(ErsteKonklusion);
    TP_GNTermpaarLoeschen(ErsteKonklusion);
    ErsteKonklusion = LetzteKonklusion;
  }
}
