/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Hauptkomponenten.c                                                                                         */
/* - *******************                                                                                        */
/* -                                                                                                            */
/* - Enthaelt die Operationen, die zum Programm-Start mit den in der KP-Menge enthaltenen Gleichungen der Spezi-*/
/* - fikation ausgefuehrt werden (I) sowie eine Prozedur, die alle KP's der Reihe nach behandelt und sie zu-    */
/* - sammenzufuehren sucht (II).                                                                                */
/* - 29.06.94 Arnim. AlsRegel/GleichungAufnehmen fertiggemacht.                                                 */
/* - 10.94/1.11.94 Arnim. Prozeduren fuer die failing completion eingebaut.                                     */
/* -------------------------------------------------------------------------------------------------------------*/
#include "antiwarnings.h"
#include "compiler-extra.h"
#include "general.h"

#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <unistd.h>

#include "Ausgaben.h"
#include "DSBaumOperationen.h"
#include "Grundzusammenfuehrung.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "Interreduktion.h"
#include "Konsulat.h"
#include "KPVerwaltung.h"
#include "LispLists.h"
#include "NFBildung.h"
#include "Ordnungen.h"
#include "Parameter.h"
#include "PhilMarlow.h"
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "Signale.h"
#include "SpezNormierung.h"
#include "Statistiken.h"
#include "Subsumption.h"
#include "Termpaare.h"
#include "Unifikation1.h"
#include "WaldmeisterII.h"
#include "general.h"
#include "pcl.h"
#include "Zielverwaltung.h"



/* -------------------------------------------------------------------------------*/
/* - static Variablen ------------------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

BOOLEAN HK_fair;

static BOOLEAN Proofmodus;

static unsigned int SelectZaehler = 0;


/* -------------------------------------------------------------------------------*/
/* --------------- KPMenge Behandeln (I und II) ----------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Ausgaben  */
/*-----------*/

static void NeueRegelDrucken(RegelT regel)
{
   if (IO_level2()){
       fprintf(IO_Strom, "%snew rule:            %5ld  ", Forked ? (Vadder ? "0> " : "1> ") : "", regel->Nummer);
       IO_DruckeFaktum(TP_LinkeSeite(regel), TRUE, TP_RechteSeite(regel));
   }
   else if (IO_level3())
      fprintf(IO_Strom, "... and added as new rule %ld.\n\n", regel->Nummer);
   else if (IO_level4()){
      fprintf(IO_Strom, "... and added as new rule %ld:\n", regel->Nummer);
      IO_DruckeFaktum(TP_LinkeSeite(regel), TRUE, TP_RechteSeite(regel));
      fprintf(IO_Strom, "\n");
   }
#if !COMP_COMP
   if (IO_level2()||IO_level3()||IO_level4()){
     if (regel->Grundzusammenfuehrbar) IO_DruckeFlex("...grundzusammenfuehrbar\n");
     else if (0) IO_DruckeFlex("...NICHT grundzusammenfuehrbar\n");
   }
#endif
#ifdef CCE
   IO_CCElog(CCE_Add, regel->linkeSeite, regel->rechteSeite);
#endif
}

static void NeueGleichungDrucken(GleichungsT gl)
{
   if (IO_level2()){
      fprintf(IO_Strom, "%snew equation:        %5ld  ", Forked ? (Vadder ? "0> " : "1> ") : "", -gl->Nummer);
      IO_DruckeFaktum(TP_LinkeSeite(gl), FALSE, TP_RechteSeite(gl));
   }
   else if (IO_level3())
      fprintf(IO_Strom, "...and added to E as %s with number %ld.\n\n",TP_IstMonogleichung(gl)?"mono equation":"such", -gl->Nummer);
   else if (IO_level4()){
      fprintf(IO_Strom, "...and added to E as %s with number %ld:\n       ",TP_IstMonogleichung(gl)?"mono equation":"such", -gl->Nummer);
      IO_DruckeFaktum(TP_LinkeSeite(gl), FALSE, TP_RechteSeite(gl));
   }
#if !COMP_COMP
   if (IO_level2()||IO_level3()||IO_level4()){
     if (gl->Grundzusammenfuehrbar) IO_DruckeFlex("...grundzusammenfuehrbar\n");
     else if (0) IO_DruckeFlex("...NICHT grundzusammenfuehrbar\n");
   }
#endif
#ifdef CCE
   IO_CCElog(CCE_Add, gl->linkeSeite, gl->rechteSeite);
#endif
}



BOOLEAN HK_GewichtsloseKPsBehandeln(void)
{
  our_error("Only the modi proof and completion at the moment available, sorry!");
  return FALSE;
}

BOOLEAN HK_KPsZusammenfuehrbar(void)
{
  our_error("Only the modi proof and completion at the moment available, sorry!");
  return FALSE;
}


/* -------------------------------------------------------------------------------*/
/* --------------- Ziel-Behandlung (III) -----------------------------------------*/
/* -------------------------------------------------------------------------------*/

void HK_GleichungenUndZieleHolen(void)
/* Die Gleichungen aus der Spezifikation plattklopfen und mit Gewicht 0 in die KP-Menge stecken.
   Fuer jedes Symbol testen, ob es AC-Eigenschaft hat.
   Die Ziele aus der Spezifikation plattklopfen, dabei skolemisieren und in die Ziele-Liste stecken.*/
{
   ListenT Index;
#if !COMP_POWER
   if (PA_InteraktionAn)
     IO_DruckeUeberschrift("Fetching goals and equations");
#endif
   if (PA_ExtModus!=completion && PA_ExtModus!=proof) IO_AusgabenAus();

   for (Index=PM_Gleichungen; Index; Cdr(&Index)) {
     ListenT Eintrag = car(Index);
     TermT Flunder1 = TO_Termkopie(car(Eintrag)),
           Flunder2 = TO_Termkopie(cdr(Eintrag));
     KPV_CPinitial(Flunder1,Flunder2);
   }
   Z_ZieleAusSpezifikationHolen();

   if (PA_ExtModus!=completion && PA_ExtModus!=proof) IO_AusgabenEin();
}


/* (III) */
/*-------*/


BOOLEAN HK_ZieleAufZusammenfuehrbarkeitTesten(void)
/* Alle Ziele aus der Zielmenge bzgl. des aktuellen Regel- und Gleichungs-Systems abschliessend auf NF bringen. */
{
   if (Z_ZielmengeLeer())
     return TRUE;
   IO_DruckeUeberschrift("Treatment of Goals");
   return Z_FinalerZielDurchlauf();
}


/* -------------------------------------------------------------------------------*/
/* --------------- Vervollstaendigung (IV) ---------------------------------------*/
/* -------------------------------------------------------------------------------*/
/* Hauptschleife(n) der Vervollstaendigung.

   Die Routine Vervollstaendigen enthaelt die Hauptschleife des Beweisers.

   Die verschiedenen parametrisierten Auspraegungen sind durch Funktionsvariablen realisiert.
   Im failing-Modus wir im wesentlichen die Gleichungsaufnahme verhindert.

   Auffaecherung der Parametrisierung:
   +-----------+------------+------+------+-----+-----+-----+
   |  PrCo     |   FunF     | NF-R | NF-E | Sub | psR | psE |
   +-----------+------------+------+------+-----+-----+-----+
   |Proof      |  Unfailing |  _   |  _   | _   |  _  |  _  |
   |           |            |      |      |     |     |     |
   |Completion |  Failing   |  R   |  E   | S   |  P  |  E  |
   +-----------+------------+------+------+-----+-----+-----+
   Es gibt also 2^7 = 128 verschiedene Moeglichkeiten.

   Im Proofmodus ergibt sich ein geringfuegig veraendertes Verhalten gegenueber dem
   Vervollstaendigungsmodus.

*/

/* Regel/Gleichung aufnehmen */
/*****************************/


VergleichsT HK_VariantentestUndVergleich(TermT links, TermT rechts)
{
  BOOLEAN Wegwerfen = (PA_GolemAn && (Z_IstSkolemvarianteZuUngleichung(links,rechts) || Z_IstGrundungleichung(links,rechts)));
  VergleichsT Resultat = Wegwerfen ? Gleich : ORD_VergleicheTerme(links,rechts);
  HK_fair = HK_fair && (!Wegwerfen || TO_TermeGleich(links,rechts));
  return Resultat;
}

static StrategieT *Strategie;
BOOLEAN Wechsle;
BOOLEAN HK_TestStrategieWechsel (void)
{
  Wechsle = Strategie && ++SelectZaehler==Strategie->Schrittzahl;
  return Wechsle;
}


/* Neue Regel/Gleichung muss man nicht testen. */
/* Achtung: Variablennormierung muss dieselbe bleiben. */

static void RueckwaertsGrundzusammenfuehrbarkeit(RegelOderGleichungsT Neues)
{
  GNTermpaarT Faktum;
  if(!GZ_doGZFB())  return; 
  if (GZ_TEST) IO_DruckeFlex("                       Rueckwaertstests\n");
  RE_forRegelnRobust(Faktum, {
    if (Faktum != Neues) {
      Faktum->DarfNichtReduzieren = TRUE;
      if ((!GZ_ZSFB_BEHALTEN || !Faktum->Grundzusammenfuehrbar) && 
	  GZ_GrundzusammenfuehrbarRueckwaerts(Faktum,Neues)) {
	if (GZ_ZSFB_BEHALTEN){
	  KPV_KillParent(Faktum); /* XXXX */
	  Faktum->DarfNichtReduzieren = FALSE;
	  Faktum->Grundzusammenfuehrbar = TRUE;
	}
	else {
	  Faktum->DarfNichtReduzieren = FALSE;
	  RE_RegelVerfluechtigen(Faktum);
	}
	if(!COMP_COMP)
	  IO_DruckeFlex("Regel %d grundzusammenfuehrbar\n",TP_ElternNr(Faktum));
      }
      else
	Faktum->DarfNichtReduzieren = FALSE;
    }
  });
  RE_forGleichungenRobust(Faktum, if ((Faktum != Neues) && TP_RichtungAusgezeichnet(Faktum)) {
    Faktum->DarfNichtReduzieren = Faktum->Gegenrichtung->DarfNichtReduzieren = TRUE;
    if ((!GZ_ZSFB_BEHALTEN || !Faktum->Grundzusammenfuehrbar) && 
        GZ_GrundzusammenfuehrbarRueckwaerts(Faktum,Neues)) {
      if (GZ_ZSFB_BEHALTEN){
	KPV_KillParent(Faktum); /* XXXX */
	Faktum->DarfNichtReduzieren = Faktum->Gegenrichtung->DarfNichtReduzieren = FALSE;
        Faktum->Grundzusammenfuehrbar = TRUE;
        Faktum->Gegenrichtung->Grundzusammenfuehrbar = TRUE;
      }
      else{
	Faktum->DarfNichtReduzieren = Faktum->Gegenrichtung->DarfNichtReduzieren = FALSE;
        RE_GleichungVerfluechtigen(Faktum);
      }
      if (!COMP_COMP)
	IO_DruckeFlex("Gleichung %d grundzusammenfuehrbar\n",-TP_ElternNr(Faktum));
    }
    else
      Faktum->DarfNichtReduzieren = Faktum->Gegenrichtung->DarfNichtReduzieren = FALSE;
  });
}

static void ArbeitsAufnahme(RegelOderGleichungsT zuletztEingefuegt, BOOLEAN ProofModus)
/* Was in beiden Faellen (neue Regel/neue Gleichung) gemeinsam getan werden muss */
{
   IR_InterreduktionLinks(zuletztEingefuegt);
   RE_FaktumEinfuegen(zuletztEingefuegt);
   IR_InterreduktionRechts(zuletztEingefuegt);

   if (ProofModus){
      if (Z_ZieleErledigtMitNeuemFaktum(zuletztEingefuegt)){  /* Falls Ziele erledigt, keine KP-Bildung mehr! */
#if 1 /* anderseits core-dumps */
         IO_AusgabenAus();     /* Dafuer aber noch den IR-Puffer abarbeiten!!! */
         IR_PufferAuslesen();
         IO_AusgabenEin();
#endif
         return;
      }
   }
   KPV_KPMengeInterreduzieren();
   KPV_CGMengeInterreduzieren(FALSE, "");
   U1_KPsBildenZuFaktum(zuletztEingefuegt);
   IR_PufferAuslesen();
   KPV_KonklusionenAbarbeiten(zuletztEingefuegt);
   RueckwaertsGrundzusammenfuehrbarkeit(zuletztEingefuegt);
   /*    ST_SignalUsrStatistik("###");*/
}

/* Vervollstaendigungs-Hauptschleife */
/*************************************/

HK_ErgebnisT HK_Vervollstaendigung (StrategieT *strategie)
{
  RegelOderGleichungsT zuletztEingefuegt;
  SelectRecT selRec;

  HK_fair = TRUE;
  Proofmodus = (PA_ExtModus == proof);
  if (WM_ErsterAnlauf() && ((!Forked) || Vadder)) {
     if (Proofmodus)
        IO_DruckeUeberschrift("COMPLETION - PROOF");
     else
        IO_DruckeUeberschrift("COMPLETION");
  }

  Strategie = strategie;
  SelectZaehler = 0;
  Wechsle = FALSE;

  while ((!Proofmodus || !Z_ZieleBewiesen()) && KPV_Select(&selRec)){
    switch (selRec.result){
    case Unvergleichbar:
      if (PA_FailingCompletion){
	IO_DruckeFlex("Completion aborted since critical pair can't be oriented:\n %t <<##>> %t\n",
                       selRec.lhs, selRec.rhs);
	TO_TermeLoeschen(selRec.lhs, selRec.rhs); /* eigtl.     TPR_DestroyPar(...) */
	return HK_VervollAbgebrochen;
      }
      else {
	SUE_selectedCPRE(&selRec);
	zuletztEingefuegt = RE_ErzeugteGleichung(selRec);
	NeueGleichungDrucken(zuletztEingefuegt);
	ArbeitsAufnahme(zuletztEingefuegt, Proofmodus);
      }
      break;
    case Kleiner :
    case Groesser:
      /* kp als neue Regel rechts->links oder links->rechts aufnehmen */
      SUE_selectedCPRE(&selRec);
      zuletztEingefuegt = RE_ErzeugteRegel(selRec);
      NeueRegelDrucken(zuletztEingefuegt);
      ArbeitsAufnahme(zuletztEingefuegt, Proofmodus);
      break;
    case Gleich:
    default:
      our_error("violation of internal invariant (main loop)");
      break;
    }
    if (SI_InterruptflagGesetzt())       ST_IRMenue("main loop");
    KO_Konsultieren();
  
    if (Forked && (COM_getState() != STATE_UNSET)){
      return HK_Ueberholt;
    }
  }

  if (Wechsle) return HK_StrategieWechseln;
  if (Proofmodus && Z_ZieleBewiesen()) {
      return HK_ZielBewiesen;
  }
  if (!COMP_POWER)
    fprintf(IO_Strom, "\n%s.\n", HK_fair() ? "system completed" : "no more critical pairs");

  return HK_Vervollstaendigt;
}

