/*=================================================================================================================
  =                                                                                                               =
  =  Interreduktion.c                                                                                             =
  =                                                                                                               =
  =  Interreduktion der Regel- und Gleichungsmenge                                                                =
  =  Subsumption                                                                                                  =
  =                                                                                                               =
  =  24.06.94  Arnim.                                                                                             =
  =  14.07.94  Arnim. Subsumption, dh. Durchlauf + Abspeichern aller Subgleichungen.                              =
  =  15.07.94  Arnim. Subgleichungsmenge implementiert.                                                           =
  =                                                                                                               =
  =================================================================================================================*/
#include "Interreduktion.h"

#include "Ausgaben.h"
#include "compiler-extra.h"
#include "DSBaumOperationen.h"
#include "Grundzusammenfuehrung.h"
#include "KPVerwaltung.h"
#include "MatchOperationen.h"
#include "NFBildung.h"
#include "Parameter.h"
#include "pcl.h"
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "Statistiken.h"
#include "Subsumption.h"
#include "TermOperationen.h"
#include "Termpaare.h"


static BOOLEAN rechts_IR_nach_P = FALSE;

/* -------------------------------------------------------------------------------*/
/* --------------- Parametereinstellung ------------------------------------------*/
/* -------------------------------------------------------------------------------*/

void IR_SetzeRechtsBehandlung(BOOLEAN nachP)
{
  rechts_IR_nach_P = nachP;
}

/*=================================================================================================================
  =                                                                                                               =
  =  Ein Wort zu den waehrend der Interreduktion ausgeloesten Meldungen:                                          =
  =                                                                                                               =
  =  - Wird eine Regel links oder eine Gleichung reduziert, so wird dies sofort (nur) unter                       =
  =    Angabe der Nummer gemeldet. Wird eine Regel rechts reduziert, so wird zusaetzlich (falls Ausgabelevel      =
  =    entsprechend) die auf NF gebrachte rechte Seite mit ausgegeben (dies ist ja bei den linken Seiten der      =
  =    Regeln und den Gleichungen nicht sofort moeglich, da NF-Bildung noch nicht konsistent).                    =
  =  - Wird nun der Puffer ausgelesen, in dem die reduzierten Regeln oder Gleichungen zwischengespeichert         =
  =    sind, so wird die Entfernung der entsprechenden Regel/Gleichung aus der Regel/Gleichungsmenge              =
  =    ohne nochmalige Angabe der Ursache gemeldet. Die Terme werden dann ausgegeben, wenn die Bildung            =
  =    eines kritischen Paares aus der ehemaligen Regel/Gleichung gemeldet wird.                                  =
  =  - Wird eine Gleichung subsummiert, so wird sie ohne Federlesen aus der GM entfernt und incl. Termen ver-     =
  =    nichtet (natuerlich nach einer entsprechenden Meldung).                                                    =
  =================================================================================================================*/

/* ------------- Regeln ----------------------------------*/

static void DruckeMeldungRegelRechtsReduziert(RegelT r)
{
    if (IO_level2())
       fprintf(IO_Strom, "simplify rhs of rule:%5ld\n", TP_ElternNr(r));
    else if (IO_level3())
       fprintf(IO_Strom, "simplify rhs of rule:%5ld.\n", TP_ElternNr(r));
    else if (IO_level4()){
       fprintf(IO_Strom, "simplify rhs of rule:%5ld:\n       ", TP_ElternNr(r));
       IO_DruckeFaktum(TP_LinkeSeite(r), TRUE, TP_RechteSeite(r));
    }
#ifdef CCE
    IO_CCElog(CCE_Delete, r->CCE_lhs, r->CCE_rhs);
    TO_TermLoeschen(r->CCE_rhs);
    r->CCE_rhs = TO_Termkopie(r->rechteSeite);
#endif
}

static void DruckeMeldungRegelLinksReduziert(RegelT r)
{
    if (IO_level2())
       fprintf(IO_Strom, "remove rule:         %5ld\n", TP_ElternNr(r));  /* Eigentlich "Reduziere Regel" */
    else if ((IO_level3()) || (IO_level4()))
       fprintf(IO_Strom, "simplify lhs of rule %ld, removed from R.\n", TP_ElternNr(r));
#ifdef CCE
    IO_CCElog(CCE_Delete, r->CCE_lhs, r->CCE_rhs);
    TO_TermeLoeschen(r->CCE_lhs, r->CCE_rhs);
    r->CCE_lhs = NULL; r->CCE_rhs = NULL;
#endif
} 

static void DruckeMeldungRegelGeloescht(RegelT r)
{
    if ((IO_level4()) || (IO_level3()))
       fprintf(IO_Strom, "rule %ld removed from R.\n", TP_ElternNr(r));
} 

/* ------------- Gleichungen -----------------------------*/


static void DruckeMeldungGleichungReduziert(GleichungsT g)
{
    if (IO_level2())
       fprintf(IO_Strom, "remove equation:     %5ld\n", -TP_ElternNr(g));
    else if ((IO_level4()) || (IO_level3()))
       fprintf(IO_Strom, "equation no. %ld removed.\n", TP_ElternNr(g));
#ifdef CCE
    IO_CCElog(CCE_Delete, g->CCE_lhs, g->CCE_rhs);
    TO_TermeLoeschen(g->CCE_lhs, g->CCE_rhs);
    g->CCE_lhs = NULL; g->CCE_rhs = NULL;
#endif
} 
 
static void DruckeMeldungGleichungGeloescht(GleichungsT g)
{
    if ((IO_level4()) || (IO_level3()))
       fprintf(IO_Strom, "equation no. %ld removed.\n", TP_ElternNr(g));
} 


static void DruckeMeldungNeueGleichungSubsummiertAlte(GleichungsT g)
{
    if (IO_level2())
       fprintf(IO_Strom, "remove equation:     %5ld\n", -TP_ElternNr(g));
    else if (IO_level3())
       fprintf(IO_Strom, "... the new equation subsumes equation no. %ld.\n", TP_ElternNr(g));
    else if (IO_level4()){
       fprintf(IO_Strom, "... the new equation subsumes equation no. %ld:\n       ", TP_ElternNr(g));
       IO_DruckeFaktum(TP_LinkeSeite(g), FALSE, TP_RechteSeite(g));
    }
#ifdef CCE
    IO_CCElog(CCE_Delete, g->CCE_lhs, g->CCE_rhs);
    TO_TermeLoeschen(g->CCE_lhs, g->CCE_rhs);
    g->CCE_lhs = NULL; g->CCE_rhs = NULL;
#endif
} 

/* -------------------------------------------------------------------------------*/
/* --------------- Pufferausleseroutinen -----------------------------------------*/
/* -------------------------------------------------------------------------------*/

static void Anwendungsprozedur(RegelOderGleichungsT objekt)
{
   if(RE_IstRegel(objekt)){
      DruckeMeldungRegelGeloescht(objekt);
      KPV_IROpferBehandeln(objekt);
      RE_RegelLoeschen(objekt);        /* Loeschen der Regel als Objekt, -> Meldung an KPVerwaltung u. Schirm*/
   }
   else
      if (TP_RichtungAusgezeichnet(objekt)){
         if (TP_IstBlau(objekt) || TP_IstMonogleichung(objekt)){       /* Gegenrichtung ist nicht im Puffer */
            TP_LinksRechtsLoeschen(TP_Gegenrichtung(objekt));   /* sonst: Gegenrichtung wird noch/wurde schon */
            RE_GleichungLoeschen(TP_Gegenrichtung(objekt));            /* behandelt */
         }
         DruckeMeldungGleichungGeloescht(objekt);
	 KPV_IROpferBehandeln(objekt);
         RE_GleichungLoeschen(objekt);    /* Loeschen der Gleichung als Objekt, -> Meldung an KPVerwaltung und Schirm */
      }
      else{                     /* Richtung ist nicht ausgezeichnet */
         if (TP_IstBlau(objekt)){ /* ausgezeichnete Richtung nicht im Puffer! */
            RegelOderGleichungsT merker = TP_Gegenrichtung(objekt); /* ausgezeichnete Gegenrichtung merken */
            TP_LinksRechtsLoeschen(objekt);
            RE_GleichungLoeschen(objekt);
            DruckeMeldungGleichungGeloescht(merker);
	    KPV_IROpferBehandeln(merker);
            RE_GleichungLoeschen(merker);    /* Loeschen der Gleichung als Objekt, -> Meldung an KPVerwaltung und Schirm */
         }
         else{                  /* ausgezeichnete Richtung selbst ist auch im Puffer */
            TP_LinksRechtsLoeschen(objekt);
            RE_GleichungLoeschen(objekt);
            return;
         }
      }
}

static void FinaleEntfernungsproz(RegelOderGleichungsT objekt)
/* Beim Puffer-Durchlauf zum Entfernen der jeweiligen Gegenrichtung der bereits einseitig entfernten Gleichungen.*/
{
   if (TP_RichtungAusgezeichnet(objekt)){
      DruckeMeldungGleichungReduziert(objekt);
      IncAnzGleichungenInterRed();
      if (TP_IstBlau(objekt)){
         if (TP_IstKeineMonogleichung(objekt))
            RE_GleichungEntfernen(TP_Gegenrichtung(objekt));
      }
      else if (TP_IstGelb(objekt) || TP_IstWeiss(objekt))
         our_error("\nGM-IR: weisses/gelbes Objekt im Puffer!");
   }
   else{
      if (TP_IstBlau(objekt)){
         RE_GleichungEntfernen(TP_Gegenrichtung(objekt));
         DruckeMeldungGleichungReduziert(TP_Gegenrichtung(objekt));
         IncAnzGleichungenInterRed();
      }
      if (TP_IstGelb(objekt))
         our_error("\nGM-IR: gelbes Objekt im Puffer!");
      if (TP_IstWeiss(objekt))
         our_error("\nGM-IR: weisses Objekt im Puffer!");
   }
}


static void FinaleKillprozSubsumption(RegelOderGleichungsT objekt)
/* Entfernen der Gegenrichtungen der gepufferten Gleichungen aus der Gleichungsmenge und
   physikalisches Loeschen der Objekte. Anwendung nach Subsumption.*/
{
  RE_GleichungEntfernen(TP_Gegenrichtung(objekt));
  TP_LinksRechtsLoeschen(TP_Gegenrichtung(objekt));
  TP_LinksRechtsLoeschen(objekt);
  RE_GleichungLoeschen(TP_Gegenrichtung(objekt));
  RE_GleichungLoeschen(objekt);
}

/* -------------------------------------------------------------------------------*/
/* --------------- Subsumption auf der Gleichungsmenge ---------------------------*/
/* -------------------------------------------------------------------------------*/

static void GMSubsummierenMitGleichung(GleichungsT NeueGleichung)
/* Sucht alle von der angegebenen Gleichung subsummierten Gleichungen aus GM, entfernt und loescht sie.
   Seiteneffekt: Meldung, dass Gleichung xyz von neuer Gleichung subsummiert wird.
                 Entfernen der subsummierten Gleichung(en) aus der Gleichungsmenge.
   Ferner entsprechende Statistik anpassen.*/
{
   GleichungsT BehGleichung;

   if (RE_GleichungsmengeNichtLeer()){
      RE_forGleichungenRobust(BehGleichung,{ 
         if (TP_RichtungAusgezeichnet (BehGleichung) && 
             SS_TermpaarSubsummiertTermpaar(TP_LinkeSeite(NeueGleichung), TP_RechteSeite(NeueGleichung), 
                                            TP_LinkeSeite(BehGleichung), TP_RechteSeite(BehGleichung))){
            DruckeMeldungNeueGleichungSubsummiertAlte(BehGleichung);
            DruckeMeldungGleichungGeloescht(BehGleichung);
            IncAnzErfSubsumptionenAlt();
            RE_GleichungEntfernen(BehGleichung); /* aber nur diese eine Seite entfernen! */
            PU_REObjektEinfuegen(BehGleichung);            /* In REPuffer stecken.*/
         }
      });
   }
   PU_REPufferDurchlauf(FinaleKillprozSubsumption);
   PU_REPufferNeu();
}

/* -------------------------------------------------------------------------------*/
/* --------------- Interreduktion der Gleichungsmenge ----------------------------*/
/* -------------------------------------------------------------------------------*/

static void GMInterred(RegelOderGleichungsT Objekt)
/* Entfernen aller Gleichungen aus der Gleichungsmenge (-> Puffer), die von dem angegebenen Objekt
   reduziert werden. (Das Objekt ist selbst noch nicht enthalten!) */
{
   GleichungsT Gleichung;
   PU_REMerkerSetzen();
   RE_forGMReferenzen(Gleichung, NF_ObjektAnwendbar(Objekt, TP_LinkeSeite(Gleichung)), {
     RE_GleichungEntfernen(Gleichung);
     TP_BlauMalen(Gleichung);
     TP_GelbMalen(TP_Gegenrichtung(Gleichung));
     PU_REObjektEinfuegen(Gleichung);
   });
   PU_REPufferDurchlaufAbMerker(FinaleEntfernungsproz); /* Gegenseite ggf. aus E entfernen, Red. ausfuehren. */
}


/* -------------------------------------------------------------------------------*/
/* --------------- Interreduktion der Regelmenge ---------------------------------*/
/* -------------------------------------------------------------------------------*/

static void RMLinksInterred(RegelOderGleichungsT Objekt)
/* Entfernen aller Regeln aus der Regelmenge (-> Puffer), die von dem angegebenen Objekt
   links reduziert werden. (Das Objekt ist selbst noch nicht enthalten!) */
{
   RegelT BehRegel;
   RE_forRegelnRobust(BehRegel,{            /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/ 
      if (NF_ObjektAnwendbar(Objekt, TP_LinkeSeite(BehRegel)) ||
          (rechts_IR_nach_P && NF_ObjektAnwendbar(Objekt, TP_RechteSeite(BehRegel)))){
         DruckeMeldungRegelLinksReduziert(BehRegel);
         IncAnzRegWgLinksRedWeg();
         RE_RegelEntfernen(BehRegel);                             /* Entfernen (nicht Loeschen) aus RM */
         PU_REObjektEinfuegen(BehRegel);                                        /* In REPuffer stecken.*/
      }
   });
}


static void RMRechtsInterred(RegelOderGleichungsT Objekt)
/* Reduktion aller rechten Seiten von Regeln, die mit dem angegebenen Objekt reduziert werden koennen, und
   Abschlus der NF-Bildung. Die rechte Seite der Regel selbst muss nicht betrachtet werden (dort ist wegen der
   Ordnung keine Reduktion moeglich!!). */

{
  RegelT BehRegel;
  if (!rechts_IR_nach_P) {
    RE_forRegelnRobust(BehRegel,{   /* Im Regelbaum gibt es jeweils nur einen Eintrag je Blatt.*/
      if(Objekt != BehRegel) {  /*Regel mit sich selbst? Niemals...*/
        PCL_Reduktionsbeginn(Regel, TP_PclID(BehRegel), rechteSeite, TP_LinkeSeite(BehRegel));
        if(NF_NormalformstRE(Objekt, TP_RechteSeite(BehRegel))){
          DruckeMeldungRegelRechtsReduziert(BehRegel);
          ON_GRUNDZUSAMMENFUEHRUNG(BehRegel->gzfbStatus = GZ_interred;)
            IncAnzRegRechtsRed();
        }
      }
    });
  }
}

/* -------------------------------------------------------------------------------*/
/* --------------- Schnittstelle nach aussen -------------------------------------*/
/* -------------------------------------------------------------------------------*/

void IR_InterreduktionLinks(RegelOderGleichungsT neuesObjekt)
{
  if ((IO_level3()) || (IO_level4()))
    fprintf(IO_Strom, ">>interreduction<<\n\n");
  PU_REPufferNeu();
  if (!TP_TermpaarIstRegel(neuesObjekt)){
    GMSubsummierenMitGleichung(neuesObjekt);
  }
  GMInterred(neuesObjekt);
  /* Muss fuer Gleichungen nach GMSubsummierenMitGleichung kommen, sonst Core-Dump! Warum?    */
  /* wg. der Puffer-Behandlung bei Subsumption! dort wird auch PU_REPufferNeu aufgerufen. BL. */
  RMLinksInterred(neuesObjekt);
}

void IR_InterreduktionRechts(RegelOderGleichungsT neuesObjekt)
{
  RMRechtsInterred(neuesObjekt);
  if ((IO_level4()) || (IO_level3()))
    fprintf(IO_Strom, "\n>>interreduction finished<<\n\n");
}

void IR_PufferAuslesen(void)
/* Auslesen des waehrend der Interreduktion gefuellten Regel- und Gleichungs-Puffers.
   Behandeln der darin befindlichen Objekte gemaess Modus. */
{
   PU_REPufferDurchlauf(Anwendungsprozedur);
}

/* -------------------------------------------------------------------------------*/
/* --------------- Test auf Killer-Eigenschaft -----------------------------------*/
/* -------------------------------------------------------------------------------*/

static BOOLEAN TermpaarAnwendbar (TermT lhs, TermT rhs)
{
  RegelOderGleichungsT Index;

  RE_forRegelnRobust(Index, {
    if (NF_TermpaarAnwendbar(lhs, rhs, TP_LinkeSeite(Index)))
      return TRUE;
  });
  RE_forGleichungenRobust(Index, {
    if (NF_TermpaarAnwendbar(lhs, rhs, TP_LinkeSeite(Index)))  /* reicht, da jede Seite indiziert ist */
      return TRUE;
  });
  return FALSE;
}

BOOLEAN IR_TPRReduziertRE(TermT Links, TermT Rechts)
{
  unsigned int dummy;
  /* wg. Anpassung der Variablenbaenke wird BO_TermpaarNormieren und nicht
     BO_TermpaarNormierenAlsKP verwendet. */
  BO_TermpaarNormieren (Links, Rechts, &dummy);
  return TermpaarAnwendbar (Links, Rechts);
}

BOOLEAN IR_TPEReduziertRE(TermT Links, TermT Rechts)
{
  unsigned int dummy;
  /* wg. Anpassung der Variablenbaenke wird BO_TermpaarNormieren und nicht
     BO_TermpaarNormierenAlsKP verwendet. */
  BO_TermpaarNormieren (Links, Rechts, &dummy);
  if (TermpaarAnwendbar (Links, Rechts))
    return TRUE;
  BO_TermpaarNormieren (Rechts, Links, &dummy);
  return TermpaarAnwendbar (Rechts, Links);
}

