/* **************************************************************************
 * KPVerwaltung.c
 *
 * Modul zur Verwaltung der passiven Fakten.
 *
 * Bernd Loechner, 21.04.2001
 *
 * ************************************************************************** */


#include "KPVerwaltung.h"

#include <string.h>

#include "Achtzigstel.h"
#include "Ausgaben.h"
#include "ChaseManhattan.h"
#include "compiler-extra.h"
#include "DSBaumOperationen.h"
#include "general.h"
#include "Grundzusammenfuehrung.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "KDHeap.h"
#include "NewClassification.h"
#include "NFBildung.h"
#include "Parameter.h"
#include "ParseInterleave.h"
#include "pcl.h"
#include "PhilMarlow.h"
#include "RUndEVerwaltung.h"
#include "Statistiken.h"
#include "Stichpunkte.h"
#include "Subsumption.h"
#include "TermOperationen.h"
#include "TermpairRepresentation.h"
#include "Unifikation1.h"
#include "Zielverwaltung.h"

#define WRITE_TRACES 0
#define ANDREI 0


/* --------------------------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

/* ------------------------------------------------------------------------*/
/* -------------------- globale Variablen ---------------------------------*/
/* ------------------------------------------------------------------------*/

/* Parameter: */
/* KPS Parameter: */
static BOOLEAN Permsub;
static BOOLEAN Normalform_R;
static BOOLEAN Normalform_E;
static BOOLEAN Subsumption;

/* fuer Interleaving von CP und CG */
static unsigned long int modulo;
static unsigned long int threshold;

/* KPBildung: Staerke von -kg */
static BOOLEAN dokgR, dokgE, dokgS, dokgP; 

/* fuer Interleaving der Queues */
static unsigned int moduloCP;
static unsigned int thresholdCP;

static BOOLEAN Waisenmord = FALSE;

/* fuer Interleaving der CG-Queues */
static unsigned int moduloCG;
static unsigned int thresholdCG;

/* Interreduktion */
static BOOLEAN doR_CGIR, doE_CGIR;
static StichpunktT *StichpunktListe;
static unsigned long int AnzahlStichpunkte;
static unsigned long int Period;
static int PeriodStrength;
static StichpunktT *StichpunktListe_CGIR;
static unsigned long int AnzahlStichpunkte_CGIR;
static unsigned long int Periode_CGIR;
static int PeriodenStaerke_CGIR;


static KDHeapT newCPHeap;
static KDHeapT newCGHeap;

static const char *strengthArrayCG[4]={"o", "r", "e", "re"};
static const char *strengthArray[16]={"o", "r", "e", "re", "s", "rs", "es", "res", "w",
				      "rw", "ew", "rew", "sw", "rsw", "esw", "resw"};



/***********************************************************************
  Ausgaben 
***********************************************************************/

static void KPinitialDrucken(TermT lhs, TermT rhs)
{
   if (IO_level3())
      fprintf(IO_Strom, "initial equation %lu turned into ues %lu.\n", GetKPV_AnzKPAusSpezifikation(), GetKPV_AnzKPGesamt());
   else if (IO_level4()){
      fprintf(IO_Strom, "initial equation %lu turned into ues %lu.\n", GetKPV_AnzKPAusSpezifikation(), GetKPV_AnzKPGesamt());
      IO_DruckeFlex("       %t # %t\n", lhs, rhs);
   }
   else if (IO_level2()){
      if (!IO_GUeberschrift){
	 fprintf(IO_Strom, "\nInitial equations:\n" );
	 fprintf(IO_Strom, "------------------\n\n" );
	 IO_GUeberschrift = TRUE;
      }
      fprintf(IO_Strom, "(%4ld)  ", GetKPV_AnzKPAusSpezifikation());
      IO_DruckeFlex("%t = %t\n", lhs, rhs);
   }
}

static void KPZusammengefuehrtDrucken (TermT ergebnis)
{
  if (IO_level3()) fprintf(IO_Strom, "... joined.\n");
  else if(IO_level4()) IO_DruckeFlex("... joined into %t.\n", ergebnis);
}

static void KPSubsummiertDrucken(void)
{
   if ((IO_level3()) || (IO_level4()))
      fprintf(IO_Strom, "... subsumed.\n");
}

static void KPPermSubsummiertDrucken (void)
{
   if ((IO_level3()) || (IO_level4()))
      fprintf(IO_Strom, "... perm-subsumed.\n");
}

static void    KPGZSFDrucken  (TermT lhs, TermT rhs)
{
  if ((IO_level2()) || (IO_level3()) || (IO_level4())){
     if (GZ_ZSFB_BEHALTEN){
       fprintf(IO_Strom, "... grundzusammengefuehrt.\n");
     }
     else {
       IO_DruckeFlex(" Termpaar %t = %t grundzusammengefuehrt.\n",lhs, rhs);
     } 
  }
}

static void KPwirdBehandeltDrucken(RegelOderGleichungsT Vater, RegelOderGleichungsT Mutter, TermT links, TermT rechts)
{
  if (IO_level1())
    fprintf(IO_Strom, ".");
  else if (IO_level3()||IO_level4()){
    /* Das folgende ist furchtbar, steht aber eh auf der Abschussliste. 30.08.01, BL. */
    if (AufrufOptionen[OptNrWaisenmord].ParameterTyp.einstellig.eingestellterWert){
      fprintf(IO_Strom, "\n************************************\nue (%ld, %ld) touched.\n", 
              (Vater == NULL) ? 0 :TP_ElternNr(Vater), (Mutter == NULL) ? 0 :TP_ElternNr(Mutter));
    }
    else {
      fprintf(IO_Strom, "\n************************************\nue (?, ?) touched.\n");
    }
  }
  if (IO_level4()){
    IO_DruckeFlex("       %t # %t\n",links, rechts);
  }
}

static void DruckeMeldungKPgebildet(unsigned long Nummer, long VaterNr, long MutterNr, TermT links, TermT rechts)
{
  if (IO_level3())
    fprintf(IO_Strom, "critical pair %lu built with parents %ld and %ld.\n", Nummer, VaterNr, MutterNr);
  else if (IO_level4()){
    fprintf(IO_Strom, "critical pair %lu built with parents %ld and %ld:\n", Nummer, VaterNr, MutterNr);
    BO_TermpaarNormierenAlsKP(links, rechts);                                       /* !!! Seiteneffekt !!! */
    IO_DruckeFlex("       %t # %t\n",links, rechts);
  }
} 

static void DruckeMeldungKPmitDBgebildet(unsigned long Nummer, TermT links, TermT rechts)
{
  if (IO_level3())
    fprintf(IO_Strom, "critical pair %lu generated from database.\n", Nummer);
  else if (IO_level4()){
    fprintf(IO_Strom, "critical pair %lu generated from database:\n", Nummer);
    BO_TermpaarNormierenAlsKP(links, rechts);     /* !!! Seiteneffekt !!! */
    IO_DruckeFlex("       %t # %t\n",links, rechts);
  }
}

static void DruckeMeldungZuKPgemacht(unsigned long Nummer, TermT links, TermT rechts)
{
  if (IO_level3())
    fprintf(IO_Strom, "... made ues %lu.\n",Nummer);
  else if (IO_level4())
    IO_DruckeFlex("... made ues %lu: %t # %t\n",Nummer, links, rechts);
}

static void DruckeMeldungAlsKPaufgenommen(TermT lhs, TermT rhs, WeightEntryT wt)
{
  if (IO_level4()){
    BO_TermpaarNormierenAlsKP(lhs, rhs);     /* !!! Seiteneffekt !!! */
    IO_DruckeFlex("... normalized to:\n       %t # %t\n... added to SUE: %d, %d.\n\n", lhs, rhs, wt.w1, wt.w2);
#if ANDREI
    IO_DruckeFlex("#Andrei: %t = %t", lhs, rhs);
    BO_TermpaarNormierenAlsKP(rhs, lhs);     /* !!! Seiteneffekt !!! */
    IO_DruckeFlex("#%t = %t\n", rhs, lhs);
#endif
  }
  else if (IO_level3())
    fprintf(IO_Strom, "... added to SUE.\n\n");
} 


static KDH_WalkthroughInteractionT UEOutput_TW(ElementT *xp)
{
  fprintf(IO_Strom, "CP(%6ld, %6ld), Weight: ", TP_ElternNr(xp->aP), TP_ElternNr(xp->oP));
  fprintf(IO_Strom, "%11ld.%11ld.", xp->w1, xp->w2);
  fprintf(IO_Strom, " : ");
  TPR_PrintTermpair(xp->tp, xp->aP, xp->oP, " ?= ");
  fprintf(IO_Strom, "\n");
  return WTI_Nothing;
}

void KPV_Dump (void)
{
  KDH_Walkthrough(newCPHeap,UEOutput_TW);
  KDH_Walkthrough(newCGHeap,UEOutput_TW); /* XXX: evtl. CG schreiben? */
}

static BOOLEAN ParseCGCP (const char *str, int *cg, int *cp)
{
   char *kopie, *kopieDummy, *token;

   kopie = (char *) our_alloc (sizeof (char) * strlen (str));
   strcpy (kopie, str);
   kopieDummy = kopie;
   
  token = strtok (kopie, ":");
  while (token != NULL) {
    if (strncasecmp (token, "cgcp=", strlen ("cgcp=")) == 0) {
      token = strtok (token+strlen("cgcp="), ".");
      *cg = atoi (token);
      token = strtok (NULL, ".:");
      if (token)
	*cp = atoi (token);
      else
	*cp = 1;
      our_dealloc (kopieDummy);
      if (*cg == 0 && *cp == 0)
        our_error ("At least one of the cg and cp parameters for cgcp must be non-zero!");
      return TRUE;
    }
    token = strtok (NULL, ":");
  }

  our_dealloc (kopieDummy);

  return FALSE;
}

static void Auswertung (unsigned long int *modulo, unsigned long int *threshold,
			InfoStringT DEFsue, InfoStringT USRsue)

{
  int cg, cp;
  *modulo = 2;
  *threshold = 1;
  if (ParseCGCP (USRsue, &cg, &cp) || ParseCGCP (DEFsue, &cp, &cg)) {
    *modulo = cg + cp;
    *threshold = cg;
  } else {
    cp = cg = 1;
  }
  if (!COMP_POWER) {
    fprintf(IO_Strom, "\n* SET OF UNSELECTED EQUATIONS:  Narrowing interval CG= %d, CP= %d\n", cg, cp);
  }
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------------ PARENTS -----------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
#if WRITE_TRACES
static CounterT kids, live_kids;
#endif

void KPV_NewParent(RegelOderGleichungsT re MAYBE_UNUSEDPARAM)
{
#if WRITE_TRACES
   kids = KPV_AnzKPGesamt;
   live_kids = KPV_AnzKPInserted;
#endif
}

void KPV_FinishParent(RegelOderGleichungsT re MAYBE_UNUSEDPARAM)
{
#if WRITE_TRACES
  if (re != NULL){
    printf("##P %5ld %5ld %5ld\n", re->Nummer, KPV_AnzKPGesamt-kids, KPV_AnzKPInserted-live_kids);
  }
#endif
}

void KPV_KillParent(RegelOderGleichungsT re)
{
#if WRITE_TRACES
  printf("##K %ld\n", re->Nummer);
#endif
  re->lebtNoch = FALSE;
  if (TP_Gegenrichtung(re) != NULL)
    TP_Gegenrichtung(re)->lebtNoch = FALSE;
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ GENERATION --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
/* In den folgenden Funktionen, die jeweils bei Entstehung eines KPs aufgerufen werden,
   wird bei bereits existierenden Termpaaren direkt der StartChain-Befehl erteilt. Bei neu erzeugten
   muss zun"achst ihre Bildung mittels TPR_NewFlatsGenerated() angezeigt werden, wo dann ggf.
   die entsprechende Chain angelegt wird.
*/

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Insert ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


static void insertElement(KDHeapT h, TermpairT tp, WeightEntryT wt,
			  RegelOderGleichungsT actPar, RegelOderGleichungsT otherPar)
{
  ElementT x;
  x.tp = tp;
  x.oP = otherPar;
  x.aP = actPar;
  x.w1 = wt.w1;
  x.w2 = wt.w2;

#if WRITE_TRACES
  printf("##A %ld,%ld,%ld,%ld\n",TP_ElternNr(x.aP), TP_ElternNr(x.oP), x.w1, x.w2);
#endif
  KDH_insert(h, x);
}

static void recentCPinsert(TermT recCPlhs, TermT recCPrhs, 
                    RegelOderGleichungsT ActualParent, RegelOderGleichungsT recCPOtherParent, 
                    UTermT recCPactualPosition, PCL_ObjektNummerT recCPpclID, 
                    CPHistoryT recCPGenerationHistory, TermT ReplicationOfLHS)
{
   TermpairT tp;
   WeightEntryT wt;
   TermT lhs,rhs;
   BOOLEAN IsCritGoal = Z_UngleichungEntmantelt(&lhs,&rhs,recCPlhs,recCPrhs);
   if (IsCritGoal){
     C_CG_Classify (&wt, lhs, rhs, ActualParent, recCPOtherParent);
   }
   else {
     C_Classify (&wt, recCPlhs, recCPrhs, ActualParent, recCPOtherParent, 
		 recCPactualPosition, recCPGenerationHistory);
   }

   DruckeMeldungAlsKPaufgenommen(recCPlhs, recCPrhs,wt);
   tp = TPR_Flat2TP(recCPlhs, recCPrhs, 
                    ActualParent, recCPOtherParent, 
                    recCPactualPosition, recCPpclID, 
                    recCPGenerationHistory == CPWithFather, ReplicationOfLHS);
   
   if (IsCritGoal){
     insertElement(newCGHeap,tp,wt,ActualParent,recCPOtherParent);
     KPV_IncCritGoalsGesamt();
     KPV_IncCritGoalsActual();
     KPV_MaximizeCritGoals();
   }
   else {
     insertElement(newCPHeap,tp,wt,ActualParent,recCPOtherParent);
     IncAnzKPEingefuegt ();
   }
   KPV_IncAnzKPInserted();
}

void KPV_CPinitial(TermT lhs, TermT rhs)
{
   PCL_ObjektNummerT  recCPpclID;
   KPV_IncAnzKPGesamt(); KPV_IncAnzKPAusSpezifikation();
   KPinitialDrucken(lhs, rhs);
   PCL_GleichungAusDerSpezifikation(lhs, rhs, &recCPpclID);
   PCL_StartChain(TRUE);

   recentCPinsert(lhs, rhs, 
                  NULL, NULL,
                  NULL, recCPpclID, 
                  Initial, NULL);
}

static BOOLEAN lohntSichBehandlung(TermT lhs, TermT rhs)
{
  return (TO_Termlaenge(lhs)+TO_Termlaenge(rhs) < 50);
}
static BOOLEAN KPBehandelt(TermT lhs, TermT rhs, PCL_ObjektNummerT *pclID)
{
  if(lohntSichBehandlung(lhs,rhs)){
    NF_Normalform2(dokgR, dokgE, lhs, rhs, pclID, KP);
    if (TO_TermeGleich(lhs, rhs)){
      KPZusammengefuehrtDrucken (lhs);
      KPV_IncAnzKPdirektZusammengefuehrt();
    }
    else if (dokgP && GZ_ACVerzichtbar(lhs, rhs)){
      KPPermSubsummiertDrucken ();
      KPV_IncAnzKPdirektPermSubsummiert();
    }
    else if (dokgS && SS_TermpaarSubsummiertVonGM(lhs, rhs)){
      KPSubsummiertDrucken ();
      KPV_IncAnzKPdirektSubsummiert();
    }
    else if (FALSE && GZ_Grundzusammenfuehrbar(lhs, rhs)){ /* nur zum Testen XXX */
      KPPermSubsummiertDrucken ();
      KPV_IncAnzKPdirektPermSubsummiert();
    }
    else {
      return FALSE;
    }
    PCL_DeleteChain();
    TO_TermeLoeschen(lhs, rhs);
    return TRUE;
  }
  return FALSE;
}

void KPV_GebildetesKPBehandelnMitVater(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition,
				       RegelOderGleichungsT actualParent, TermT lhsReplica)
{
  PCL_ObjektNummerT pclID;
  KPV_IncAnzKPGesamt(); 
  DruckeMeldungKPgebildet(GetKPV_AnzKPGesamt(), TP_ElternNr(actualParent), TP_ElternNr(father), lhs, rhs);
  TPR_NewFlatsGenerated();
  PCL_Unifikation(&pclID, lhs, rhs,
		  father, TP_LinkeSeite(father), actualPosition, actualParent);
  if (!KPBehandelt(lhs, rhs, &pclID)){
    recentCPinsert(lhs, rhs, 
                   actualParent, father, 
                   actualPosition, pclID, 
                   CPWithFather, lhsReplica);
  }
  else {
#if WRITE_TRACES
    printf("##R %ld,%ld\n", actualParent->Nummer, father->Nummer);
#endif
  }
}

void KPV_GebildetesKPBehandelnMitMutter(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition,
					RegelOderGleichungsT actualParent, TermT lhsReplica)

{
  PCL_ObjektNummerT pclID;
  KPV_IncAnzKPGesamt(); 	
  DruckeMeldungKPgebildet(GetKPV_AnzKPGesamt(), TP_ElternNr(actualParent), TP_ElternNr(mother), lhs, rhs);
  TPR_NewFlatsGenerated();
  PCL_Unifikation(&pclID, lhs, rhs,
		  actualParent, lhsReplica, actualPosition, mother);
  if (!KPBehandelt(lhs, rhs, &pclID)){
    recentCPinsert(lhs, rhs, 
                   actualParent, mother, 
                   actualPosition, pclID, 
                   CPWithMother, lhsReplica);
  }
  else {
#if WRITE_TRACES
    printf("##R %ld,%ld\n", actualParent->Nummer, mother->Nummer);
#endif
  }
}

void KPV_IROpferBehandeln(RegelOderGleichungsT opfer)
{
  TermT ll, rr;
  ll = TO_Termkopie(TP_LinkeSeite(opfer));
  rr = TO_Termkopie(TP_RechteSeite(opfer));KPV_IncAnzKPGesamt(); 
  KPV_IncAnzKPIROpfer();
  DruckeMeldungZuKPgemacht(GetKPV_AnzKPGesamt(), ll, rr);
  PCL_StartChain(TRUE);
  if (!KPBehandelt(ll, rr, &(TP_PclIDAs(opfer)))){
    recentCPinsert(ll, rr, 
                   opfer, NULL, 
                   NULL, TP_PclIDAs(opfer), 
                   KilledRE, NULL);
  }
}

void KPV_KonklusionenAbarbeiten(RegelOderGleichungsT r)
/* Meldet die entsprechende Regel oder Gleichung dem Datenbankexperten an und
   verarbeitet die gezogenen Schlussfolgerungen.
*/
{
  RegelOderGleichungsT Konklusion;
  
  CM_IntegriertesFaktumMelden(r);
  for (Konklusion=CM_ErsteKonklusionsinstanz(); Konklusion; Konklusion = CM_NaechsteKonklusionsinstanz()){
    KPV_IncAnzKPGesamt(); 
    KPV_IncAnzKPDatenbank();
    DruckeMeldungKPmitDBgebildet(GetKPV_AnzKPGesamt(), TP_LinkeSeite(Konklusion), TP_RechteSeite(Konklusion));
    PCL_StartChain(TRUE);
    PCL_DatenbankKonklusion(Konklusion);

    recentCPinsert(TP_LinkeSeite(Konklusion), TP_RechteSeite(Konklusion), 
		   NULL, NULL, 
		   NULL, TP_PclIDAs(Konklusion), 
		   ChaseManhattanCP, NULL);
    /* XXX: Wo werden die Konklusionen geloescht? */
  }
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ SELECTION ---------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static BOOLEAN selectNonOrphan (KDHeapT h, unsigned int d, ElementT *xp, BOOLEAN *wasFIFO)
{
  while (!KDH_isEmpty(h)){
    KDH_deleteMin(h,xp,d);
    IncAnzKPEntfernt();
    if (!Waisenmord || (xp->aP == NULL) || (xp->oP == NULL) || (xp->aP->lebtNoch && xp->oP->lebtNoch)){
#if WRITE_TRACES
      printf("##D %ld,%ld,%ld,%ld\n",
	     (xp->aP == NULL) ? 0 : TP_ElternNr(xp->aP), (xp->oP == NULL) ? 0 : TP_ElternNr(xp->oP), xp->w1, xp->w2);
#endif
      *wasFIFO = d;
      return TRUE;
    }
#if WRITE_TRACES
    printf("##O %ld,%ld,%ld,%ld\n",
	   (xp->aP == NULL) ? 0 : TP_ElternNr(xp->aP), (xp->oP == NULL) ? 0 : TP_ElternNr(xp->oP), xp->w1, xp->w2);
#endif
    IncAnzKPSpaeterWaisenmord();
    TPR_DestroyTP(xp->tp, xp->aP, xp->oP);
  }
  return FALSE;
}

static BOOLEAN Select(KDHeapT h, unsigned int d, SelectRecT *selRec)
{
  ElementT x;
  if (selectNonOrphan(h,d,&x,&(selRec->wasFIFO))){
    selRec->tp           = x.tp;
    selRec->actualParent = x.aP;
    selRec->otherParent  = x.oP;
    TPR_TP2ParFinal(selRec->tp, &(selRec->lhs), &(selRec->rhs), 
		    selRec->actualParent, selRec->otherParent, &(selRec->pclId));

    KPwirdBehandeltDrucken(selRec->actualParent, selRec->otherParent, selRec->lhs, selRec->rhs);
    KPV_IncAnzBehandelterKPsVervollst();
    return TRUE;
  }
  return FALSE;
}

static BOOLEAN CGbevorzugen(void)
{
  return GetKPV_AnzAktivierterFakten() % modulo < threshold;
}

/* BOOLEAN lastFIFO = 0;*/

static unsigned int CPdimension(void)
{
  return GetKPV_AnzAktivierterRE() % moduloCP < thresholdCP;
#if 0
  if (0&&lastFIFO){
    lastFIFO = 0;
    if (GetKPV_AnzAktivierterREFifo() % 5 == 0){
      moduloCP++;
    }
  }
  /*if (GetKPV_AnzAktivierterRE() % moduloCP < thresholdCP){*/
  /*  if (GetKPV_AnzAktivierterRE()+1 > GetKPV_AnzAktivierterREFifo()*moduloCP){*/
    /*    printf ("1 %d: akt = %ld mod = %ld tresh = %ld fifo = %ld\n",
	    (GetKPV_AnzAktivierterRE()+1 > GetKPV_AnzAktivierterREFifo()*moduloCP) ? 1 : 0,
	    GetKPV_AnzAktivierterRE(), moduloCP, thresholdCP,GetKPV_AnzAktivierterREFifo());*/
    return 1;
  }
  else {
    /*    printf ("0 %d: akt = %ld mod = %ld tresh = %ld fifo = %ld\n",
	    (0.9*GetKPV_AnzAktivierterRE()+1 > GetKPV_AnzAktivierterREFifo()*moduloCP) ? 1 : 0,
	    GetKPV_AnzAktivierterRE(), moduloCP, thresholdCP,GetKPV_AnzAktivierterREFifo());*/
    return 0;
  }
#endif
}
static unsigned int CGdimension(void)
{
  return (GetKPV_AnzAktivierterG() % moduloCG < thresholdCG) ? 1 : 0;
}
static BOOLEAN WasSelektiert(SelectRecT *selRec)
{
  if (PM_Existenzziele()) {
    if (CGbevorzugen()){
      if (Select(newCGHeap, CGdimension(), selRec)) return TRUE;
      if (Select(newCPHeap, CPdimension(), selRec)) return TRUE;
    }
    else {     
      if (Select(newCPHeap, CPdimension(), selRec)) return TRUE;
      if (Select(newCGHeap, CGdimension(), selRec)) return TRUE;
    }
  }
  else{
    if (Select(newCPHeap, CPdimension(), selRec))   return TRUE;
  }
  return FALSE;
} 

void SUE_selectedCPRE(SelectRecT *selRec)
{
    TPR_Par2Flat(selRec->tp, selRec->lhs, selRec->rhs, selRec->actualParent, selRec->otherParent);
    KPV_IncAktivierterFakten();
    if (Z_IstGemantelteUngleichung(selRec->lhs, selRec->rhs)) {
      KPV_IncAktivierterG();
      if (selRec->wasFIFO)
        KPV_IncAktivierterGFifo();
      else 
        KPV_IncAktivierterGHeu();
    } else {
      KPV_IncAktivierterRE();
      if (selRec->wasFIFO){
	/* lastFIFO++; */
        KPV_IncAktivierterREFifo();
      }
      else
        KPV_IncAktivierterREHeu();
    }
    KPV_IncAnzKPZuRE();
}

VergleichsT HK_VariantentestUndVergleich(TermT links, TermT rechts); /* XXX */
BOOLEAN HK_TestStrategieWechsel (void); /* XXXXXXXXXXXXXXXXX */

BOOLEAN KPV_Select(SelectRecT *selRec)
{
  while (!HK_TestStrategieWechsel() && WasSelektiert(selRec)){
    NF_Normalform2(Normalform_R, Normalform_E, selRec->lhs, selRec->rhs, &(selRec->pclId), Gleichung);
    selRec->result = HK_VariantentestUndVergleich(selRec->lhs, selRec->rhs);
    if (selRec->result == Gleich){
      KPZusammengefuehrtDrucken (selRec->lhs);
      KPV_IncAnzKPspaeterZusammengefuehrt();
    }
    else if (Permsub && GZ_ACVerzichtbar(selRec->lhs, selRec->rhs)){
      KPPermSubsummiertDrucken();
      KPV_IncAnzKPspaeterPermSubsummiert();
    }
    else if ((selRec->result == Unvergleichbar) && Subsumption && SS_TermpaarSubsummiertVonGM(selRec->lhs, selRec->rhs)){
      KPSubsummiertDrucken();
      KPV_IncAnzKPspaeterSubsummiert();
    }
    else if (!GZ_ZSFB_BEHALTEN && GZ_doGZFB() && GZ_Grundzusammenfuehrbar(selRec->lhs,selRec->rhs)){
      KPGZSFDrucken(selRec->lhs, selRec->rhs);
      KPV_IncAnzKPspaeterPermSubsummiert(); /* XXX */
    }
    else {
      return TRUE;
    }
    TPR_DestroyPar(selRec->tp, selRec->lhs, selRec->rhs, selRec->actualParent, selRec->otherParent);
  }
  return FALSE;
}

/* Dieses Makro soll die Auswertung der Parameter uebernehmen. Die auszufuehrenden
   Operationen werden als Bloecke uebergeben.*/

#define KPS_Auswertung(PERMSUB, NFR, NFE, SUB)                                        \
{                                                                                               \
  PERMSUB=FALSE;NFR=NFE=SUB=FALSE;                                                            \
                                                                                                \
  if (!COMP_POWER)                                                                              \
    fprintf(IO_Strom, "\n* TREATMENT OF CPS AFTER SELECTION:  ");                               \
                                                                                                \
  if (IS_FindOption(DEF, USR, 1, "p") == 0) PERMSUB  = TRUE;                                    \
  if (IS_FindOption(DEF, USR, 1, "s") == 0) SUB = TRUE;                                        \
  if (IS_FindOption(DEF, USR, 1, "r") == 0) NFR = TRUE;                                        \
  if (IS_FindOption(DEF, USR, 1, "e") == 0) NFE = TRUE;                                        \
  if (!COMP_POWER) {                                                                            \
    fprintf(IO_Strom, " %s%s", NFR?"R":"-", NFE?"E":"-");                                     \
    fprintf(IO_Strom, "%s", SUB?"S":"-");                                                       \
    fprintf(IO_Strom, "%s", PERMSUB?"P":"-");                                                   \
    fprintf(IO_Strom,"\n");                                                                     \
   }                                                                                            \
}

void KPS_ParamInfo (ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  if (what == ParamEvaluate){
    BOOLEAN PSUBInfo;
    BOOLEAN Rinfo, Einfo, Sinfo;
    KPS_Auswertung(PSUBInfo, Rinfo, Einfo, Sinfo);
  }
  else{
    IO_DruckeKurzUeberschrift("Treatment of CPs after Selection", '-');
    fprintf(IO_Strom, "1. [r]\n");
    fprintf(IO_Strom, "2. [e]\n");
    fprintf(IO_Strom, "3. [s]\n");
    fprintf(IO_Strom, "4. [p]\n");
    fprintf(IO_Strom, "\n");
    fprintf(IO_Strom, "Explanation:\n");
    fprintf(IO_Strom, "\n");
    fprintf(IO_Strom, "1-3.  These letters cause Waldmeister to normalize newly generated\n");
    fprintf(IO_Strom, "      critical pairs with _r_ules and/or _e_quations, with or without\n");
    fprintf(IO_Strom, "      additional _s_ubsumption\n");
    fprintf(IO_Strom, "4.    Suppress redundant AC provable rules and equations.\n\n");
  }
}

void KPS_InitAposteriori(InfoStringT DEF, InfoStringT USR)
/* Einrichtung der Funktionspointer fuer die Ordnungsaufrufe.*/
{
   KPS_Auswertung(Permsub,Normalform_R,Normalform_E,Subsumption);
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

/* Programmende */
static KDH_WalkthroughInteractionT abraeumFct(ElementT *xp)
{
  TPR_DestroyTP(xp->tp, xp->aP, xp->oP);
  return WTI_Nothing;
}

void KPV_PassiveFaktenWegwerfen(void)
{ 
#if 0
  KDH_Walkthrough(newCPHeap, abraeumFct);
  /*KDH_Walkthrough(newCGHeap, abraeumFct);*/
#else
  SelectRecT selRec;
  while (WasSelektiert(&selRec))    
    TPR_DestroyPar(selRec.tp, selRec.lhs, selRec.rhs, selRec.actualParent, selRec.otherParent);
#endif
}

void KPV_CLEANUP(void)
{
  KDH_Walkthrough(newCPHeap, abraeumFct);
  KDH_delHeap(newCPHeap);
  newCPHeap = NULL;

  KDH_Walkthrough(newCGHeap, abraeumFct);
  /* XXX todo: 
  KDH_delHeap(newCGHeap);
  newCGHeap = NULL;
  */
}

void KPV_SOFTCLEANUP(void)
{
#if WRITE_TRACES
  printf("##C\n");
#endif
  KDH_delHeap(newCPHeap);
  newCPHeap = NULL;
  KDH_delHeap(newCGHeap);
  newCGHeap = NULL;
}

/* ################################################################################################################################### */
/* ################################################################################################################################### */

static KDH_WalkthroughInteractionT DurchlaufFct(ElementT *xp)
{
  TermT lhs, rhs;
  PCL_ObjektNummerT pclID;
  AZ_DruckeMeldungItem();
  TPR_TP2ParIntermed (xp->tp, &lhs, &rhs, xp->aP, xp->oP, &pclID);
  /* XXX
   * Dumfug: Entweder lhs oder rhs sind ``false'', das andere ist eq(s,t)!
   */
  if ((doR_CGIR || doE_CGIR) && NF_Normalform2 (doR_CGIR, doE_CGIR, lhs, rhs, &pclID, KP)){
    UTermT s,t;
    if (Z_UngleichungEntmantelt(&s,&t, lhs,rhs)) {                 /* gdw weder SUCCESS noch wegzuwerfendes CG */
      WeightEntryT wt;                /* Wir haben jetzt: eq (s,t) = false und brauchen s,t zur Classifikation */
      C_CG_Classify (&wt, s, t, xp->aP, xp->oP);
      if (xp->w1 != wt.w1) {
	xp->w1 = wt.w1;                                                            /* w2 wird nicht veraendert */
      }
      xp->tp = TPR_Par2TP (xp->tp, lhs, rhs,xp->aP, xp->oP,TRUE, pclID);
      KPV_IncIRReduziert();
      return WTI_Reweight;
    }
    else if (TO_TermeGleich (lhs, rhs)) {                               /* false = false entstanden; wegwerfen */
      TPR_DestroyPar (xp->tp, lhs, rhs, xp->aP, xp->oP);
      KPV_IncIRGleich();
      KPV_DecCritGoalsActual();
      return WTI_Delete;
    }
    else {                                                                 /* true = false entstanden; SUCCESS */
      xp->w1 = minimalWeight();                                            /* entsprechendes CG sofort bringen */
      xp->tp = TPR_Par2TP (xp->tp, lhs, rhs,xp->aP, xp->oP,TRUE, pclID);
      KPV_IncIRReduziert();
      return WTI_Stop;                          /* Schleife direkt verlassen, weitere Arbeit ist ueberfluessig */
    }
  }
  else {                                                       /* fokussiertes kritisches Ziel war irreduzibel */
    xp->tp = TPR_Par2TP(xp->tp, lhs, rhs,xp->aP, xp->oP,FALSE, pclID);
    return WTI_Nothing;
  }
}

static void Durchlauf(unsigned int strength)
{
  doR_CGIR = strength & 1;
  doE_CGIR = strength & 2;
  
  AZ_DruckeMeldungStart("critical goals", strengthArrayCG[strength], KDH_size(newCGHeap));
  KDH_Walkthrough(newCGHeap, DurchlaufFct);
  AZ_DruckeMeldungEnde();
}


void KPV_CGMengeInterreduzieren(BOOLEAN Erzwingen, char *Staerke)
{
  unsigned int strength;
  if (Erzwingen) {
    strength = 3;
    while (strcmp(Staerke,strengthArrayCG[strength]) && strength)
      strength--;
    Durchlauf(strength);
  }
  else if (SP_IstStichpunkt(StichpunktListe_CGIR, AnzahlStichpunkte_CGIR,
                            RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt(), 
                            Periode_CGIR, PeriodenStaerke_CGIR,
                            &strength))
    Durchlauf(strength);
}

     


/* ################################################################################################################################### */
/* ################################################################################################################################### */

void KPB_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   if (what == ParamEvaluate){
      BOOLEAN dokgR, dokgE, dokgS, dokgP;
      
      dokgR = (IS_FindOption(DEF, USR, 1, "r") == 0);
      dokgE = (IS_FindOption(DEF, USR, 1, "e") == 0);
      dokgS = (IS_FindOption(DEF, USR, 1, "s") == 0);
      dokgP = (IS_FindOption(DEF, USR, 1, "p") == 0);

      fprintf(IO_Strom, "\n* TREATMENT OF CPS ON GENERATION:    ");

      if (dokgR || dokgE || dokgS || dokgP){
	if (dokgR) fprintf(IO_Strom, "R");
	if (dokgE) fprintf(IO_Strom, "E");
	if (dokgS) fprintf(IO_Strom, "S");
	if (dokgP) fprintf(IO_Strom, "P");
	fprintf(IO_Strom, "\n");
      }
      else 
	fprintf(IO_Strom, "-\n");
   }
   else{
      fprintf(IO_Strom, "Treatment of CPs after Generation\n");
      fprintf(IO_Strom, "*********************************\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. [r]\n");
      fprintf(IO_Strom, "2. [e]\n");
      fprintf(IO_Strom, "3. [s]\n");
      fprintf(IO_Strom, "4. [p]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "Explanation:\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1-3. These letters cause Waldmeister to normalize newly generated\n");
      fprintf(IO_Strom, "     critical pairs with _r_ules and/or _e_quations, with or without\n");
      fprintf(IO_Strom, "     additional _s_ubsumption\n");
      fprintf(IO_Strom, "4.   This option tells Waldmeister to discard AC-redundant equations.\n");
   }
}

BOOLEAN KPB_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
  dokgR = (IS_FindOption(DEF, USR, 1, "r") == 0);
  dokgE = (IS_FindOption(DEF, USR, 1, "e") == 0);
  dokgS = (IS_FindOption(DEF, USR, 1, "s") == 0);
  dokgP = (IS_FindOption(DEF, USR, 1, "p") == 0);

#if !COMP_POWER
  fprintf(IO_Strom, "\n* TREATMENT OF CPS ON GENERATION:    ");
#endif


#if !COMP_POWER
  if (dokgR || dokgE || dokgS || dokgP){
    if (dokgR) fprintf(IO_Strom, "R");
    if (dokgE) fprintf(IO_Strom, "E");
    if (dokgS) fprintf(IO_Strom, "S");
    if (dokgP) fprintf(IO_Strom, "P");
    fprintf(IO_Strom, "\n");
  }
  else 
    fprintf(IO_Strom, "-\n");
#endif

  return TRUE;
}

/* -------------------------------------------------------------------------------*/
/* --------------- Strings in Strength umwandeln ---------------------------------*/
/* -------------------------------------------------------------------------------*/

static int String2Strength(char *str)
{
   int val = 0;

   if (strchr(str, 'r')) val |= 1;
   if (strchr(str, 'e')) val |= 2;
   if (strchr(str, 's')) val |= 4;
   if (strchr(str, 'w')) val |= 8;

   return val;
}
      
/* -------------------------------------------------------------------------------*/
/* --------------- Anwendungsprozeduren ------------------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Anwendungsprozeduren, die dem KP-Mengendurchlauf uebergeben werden, um die KP-Mengen-Interreduktion
   durchzufuehren.
   Vor Aufruf jeder Anwendungsprozedur kann davon ausgegangen werden, dass das entsprechende KP als aktuelles (bzgl. IR) angemeldet
   und physikalisch noch intakt ist. Es sind also sowohl Zugriffe auf die Eintrage des Objekts kp als auch ueber die
   KPO_...AktKP()-Funktionen statthaft.*/

static BOOLEAN doR, doE, doS, doW;

static KDH_WalkthroughInteractionT AP_generic(ElementT *xp)
/*
(TermT lhs, TermT rhs, 
 RegelOderGleichungsT Father MAYBE_UNUSEDPARAM, RegelOderGleichungsT Mother MAYBE_UNUSEDPARAM, 
PCL_ObjektNummerT * pclID MAYBE_UNUSEDPARAM)
*/
{
  BOOLEAN reduziert = FALSE;
  TermT lhs, rhs;
  PCL_ObjektNummerT pclID;

  IncAnzKPIRBetrachtet();
  AZ_DruckeMeldungItem(); 

  if (Waisenmord && !(xp->aP->lebtNoch && xp->oP->lebtNoch)){
    IncAnzKPIRWaisenmord();
    IncAnzKPEntfernt();
    TPR_DestroyTP(xp->tp, xp->aP, xp->oP);
    return WTI_Delete;
  }

  IncAnzKPIRBehandelt();
  TPR_TP2ParIntermed(xp->tp, &lhs, &rhs, xp->aP, xp->oP, &pclID);

  if ((doR || doE) && NF_Normalform2(doR, doE, lhs, rhs, &pclID, KP)){
    reduziert = TRUE;
    IncAnzKPIRReduziert();
    if (TO_TermeGleich(lhs, rhs)) {
      IncAnzKPIRZusammengefuehrt();
      IncAnzKPEntfernt();
      IncAnzKPIRGeloescht();
      TPR_DestroyPar (xp->tp, lhs, rhs,xp->aP, xp->oP);
      return WTI_Delete;
    }
  }

  if (doS && SS_TermpaarSubsummiertVonGM(lhs, rhs)) {
    IncAnzKPIRSubsummiert();
    IncAnzKPEntfernt();
    IncAnzKPIRGeloescht();
    TPR_DestroyPar (xp->tp, lhs, rhs,xp->aP, xp->oP);
    return WTI_Delete;
  }

  if (doW || reduziert) {
    WeightEntryT wt;
               
    IncAnzKPIRVeraendert();

    wt.w1 = xp->w1;
    wt.w2 = xp->w2;
    C_ReClassify(&wt, lhs, rhs, xp->aP, xp->oP);
    xp->tp = TPR_Par2TP (xp->tp, lhs, rhs,xp->aP, xp->oP, TRUE, pclID);
    xp->w1 = wt.w1;
    xp->w2 = wt.w2;

    return WTI_Reweight;
  }
  else {
    xp->tp = TPR_Par2TP (xp->tp, lhs, rhs,xp->aP, xp->oP, TRUE, pclID); /* XXX verm. FALSE statt TRUE */
    return WTI_Nothing;
  }
}

/* -------------------------------------------------------------------------------*/
/* --------------- KP-Mengen-Durchlaeufe -----------------------------------------*/
/* -------------------------------------------------------------------------------*/

static void do_KPIR(int strength)
{
   AZ_DruckeMeldungStart("critical pairs", strengthArray[strength], KDH_size(newCPHeap));
   doR = strength & 1;
   doE = strength & 2;
   doS = strength & 4;
   doW = strength & 8;
   KDH_Walkthrough(newCPHeap, AP_generic);
   AZ_DruckeMeldungEnde();
}   


void KPV_KPMengeInterreduzieren(void)
/* Interreduktion der KP-Menge gemaess Aufruf-Einstellung.*/
{
   int strength;
   /* Bedingungsreihenfolge relevant: Stichpunkte werden immer geprueft. Damit
      wird kein Stichpukt durch paralleles periodisches Interreduzieren uebersprungen!
      */
   if (SP_IstStichpunkt(StichpunktListe, AnzahlStichpunkte,
			RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt(),
			Period, PeriodStrength, &strength))
      do_KPIR(strength);
}


void KPV_WahlfreiInterreduzieren(char *howStrong)
/* Fuehrt die KP-Mengen-Interreduktion zu einem beliebigen Zeitpunkt aus.
   Funktionalitaet (Maechtigkeit und dyn. Neubewerten)  wie beim Programm-Start
   vorgesehen. */
{
   int strength = String2Strength(howStrong);

   do_KPIR(strength);
}


/* -------------------------------------------------------------------------------*/
/* --------------- Initialisieren ------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

 
  
void KPIR_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   if (what == ParamEvaluate){
     StichpunktT *localStichpunktListe = StichpunktListe; /* retten */
     long unsigned int Period=0;
     int PeriodStrength=0;
     unsigned long int AnzahlStichpunkte;
     BOOLEAN StatistikAn;

     StatistikAn = (IS_ValueOfOption(DEF, USR, "statistics") 
		    && (strcmp(IS_ValueOfOption(DEF, USR, "statistics"), "on") == 0));
     
     SP_StichpunktlisteGenerieren(DEF,USR, &Period, &PeriodStrength,
				  &StichpunktListe, &AnzahlStichpunkte,16,
				  strengthArray);
      
     fprintf(IO_Strom, "\n* INTERMEDIATE CP-SET-REDUCTION\n\n");
     SP_Ausgabe (Period, PeriodStrength, StichpunktListe, AnzahlStichpunkte,
		 strengthArray, StatistikAn);

     our_dealloc(StichpunktListe);
     StichpunktListe = localStichpunktListe;
   }
   else{
      IO_DruckeKurzUeberschrift("INTERMEDIATE CP-SET-REDUCTION", '-');
      fprintf(IO_Strom, "1. [period=<n><strength>]\n");
      fprintf(IO_Strom, "2. [<strength>=<n1>.<n2>.<n3>...]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "Explanation:\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. Tells Waldmeister to periodically normalize the members of SUE every <n> new rules/equations\n");
      fprintf(IO_Strom, "    with <strength>. One may declare only one period.\n");
      fprintf(IO_Strom, "2. Tells Waldmeister to normalize all members of SUE after a certain number of new rules/equations\n");
      fprintf(IO_Strom, "    (<n1>, <n2>, <n3>) with <strength>.\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "<strength> is a string within which only the following letters will be recognized:\n");
      fprintf(IO_Strom, "r - rules           : use rules for normalization\n");
      fprintf(IO_Strom, "e - equations       : use equations for normalization\n");
      fprintf(IO_Strom, "s - subsumption     : subsume\n");
      fprintf(IO_Strom, "w - weight          : reweight every touched ue\n");
      fprintf(IO_Strom, "o - orphan-murder   : remove all the orphans.\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "If normalizing the cps intermediately, reweighting of the simplified ones\n");
      fprintf(IO_Strom, "will be done automatically. Those joined or subsumed will be removed from\n");
      fprintf(IO_Strom, "the set.\n");
      fprintf(IO_Strom, "It is not necessary to state an 'o' when intermediate reductions are invoked:\n");
      fprintf(IO_Strom, "full orphan murder will be executed on these ocasions automatically.\n");
   }
}
   

BOOLEAN KPIR_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
  
  SP_StichpunktlisteGenerieren(DEF,USR, &Period, &PeriodStrength,
			       &StichpunktListe, &AnzahlStichpunkte,16,
			       strengthArray);
#if !COMP_POWER  
  fprintf(IO_Strom, "\n* INTERMEDIATE CP-SET-REDUCTION\n\n");
  SP_Ausgabe (Period, PeriodStrength, StichpunktListe, AnzahlStichpunkte,
	      strengthArray, KPIR_STATISTIK);
#endif
  
  return TRUE;
}

void SUE_ParamInfo(ParamInfoArtT what, int whichModule,
                   InfoStringT DEFsue,    InfoStringT USRsue,
                   InfoStringT DEFclas,   InfoStringT USRclas,
                   InfoStringT DEFreclas, InfoStringT USRreclas,
                   InfoStringT DEFcgclas, InfoStringT USRcgclas,
                   InfoStringT DEFpq MAYBE_UNUSEDPARAM,     InfoStringT USRpq MAYBE_UNUSEDPARAM,
                   InfoStringT DEFtpr,    InfoStringT USRtpr,
		   InfoStringT DEFcg,    InfoStringT USRcg)
{
   if ((whichModule==0) || (whichModule==99)){                /* SUE */
     if (what == ParamEvaluate){
       unsigned long int infoMod, infoTh;
       Auswertung(&infoMod, &infoTh, DEFsue, USRsue);
     }
     else{
       IO_DruckeKurzUeberschrift("Set of unselected equations", '-');
       fprintf(IO_Strom, "[cgcp=<n1>.<n2>]\n");
       fprintf(IO_Strom, "Intervall for paramodulation with existentially quantified goals.\n");
     }
   }

   if ((whichModule==1) || (whichModule==2) ||(whichModule==3))
     C_ParamInfo(what, whichModule, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);
   if (whichModule == 99) {   
      C_ParamInfo(what, 1, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);
      C_ParamInfo(what, 2, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);
      C_ParamInfo(what, 3, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);
   }
   
   if ((whichModule==5) || (whichModule==99)) TPR_ParamInfo(what, DEFtpr, USRtpr);

   if ((whichModule==6) || (whichModule==99)) TPR_ParamInfo(what, DEFcg, USRcg);
}
      

static BOOLEAN CG_InitAposteriori (InfoStringT DEF, InfoStringT USR)
{
  int fifo, heuristic;

  if (PI_ParseInterleave (USR, &fifo, &heuristic)
      || PI_ParseInterleave (DEF, &fifo, &heuristic)) {
    moduloCG = fifo + heuristic;
    thresholdCG = fifo;

#if !COMP_POWER
    fprintf (IO_Strom, "\n* Interleaving of Critical Goal Queues\n");
    fprintf (IO_Strom, "  Fifo: %d, Heuristic: %d\n", fifo, heuristic);
#endif
  } else {
    moduloCG = 1;
    thresholdCG = 0;
  }
  

  SP_StichpunktlisteGenerieren(DEF,USR,&Periode_CGIR,&PeriodenStaerke_CGIR,
                               &StichpunktListe_CGIR, &AnzahlStichpunkte_CGIR,
                               4, strengthArrayCG);
#if !COMP_POWER
  fprintf(IO_Strom, "\n* INTERMEDIATE CG-SET-REDUCTION\n\n");
  SP_Ausgabe (Periode_CGIR, PeriodenStaerke_CGIR,
              StichpunktListe_CGIR, AnzahlStichpunkte_CGIR,
              strengthArrayCG, TRUE);
#endif

  SetKPV_AnzCritGoalsActual(0); /* Tatsaechliche Groesse */

  newCGHeap = KDH_mkHeap(2000, (thresholdCG) ? 2 : 1, NULL);
  return TRUE;
}

static BOOLEAN KPH_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
  int fifo, heuristic;

#if !COMP_POWER
   fprintf(IO_Strom, "Regular Heap. statistics = %s. ", KPH_STATISTIK ? "on":"off");
#endif
   if (PI_ParseInterleave (USR, &fifo, &heuristic)
       || PI_ParseInterleave (DEF, &fifo, &heuristic)) {
     moduloCP = fifo + heuristic;
     thresholdCP = fifo;
#if !COMP_POWER
     fprintf (IO_Strom, "  Interleaving of CP Queues\n");
     fprintf (IO_Strom, "    Fifo: %d, Heuristic: %d\n", fifo, heuristic);
#endif
   } else {
     moduloCP = 1;
     thresholdCP = 0;
   }

   newCPHeap = KDH_mkHeap(10000, (thresholdCP) ? 2 : 1, NULL);
   return TRUE;
}

BOOLEAN SUE_InitAposteriori(InfoStringT DEFsue,    InfoStringT USRsue,
                            InfoStringT DEFclas,   InfoStringT USRclas,
                            InfoStringT DEFreclas, InfoStringT USRreclas,
                            InfoStringT DEFcgclas, InfoStringT USRcgclas,
                            InfoStringT DEFpq,     InfoStringT USRpq,
                            InfoStringT DEFtpr,    InfoStringT USRtpr,
                            InfoStringT DEFcg,     InfoStringT USRcg)
{
   if (!TPR_InitAposteriori(DEFtpr, USRtpr))
      return FALSE;

   if (!C_InitAposteriori(DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas))
      return FALSE;

   if (!KPH_InitAposteriori(DEFpq, USRpq))
      return FALSE;

   if (!CG_InitAposteriori(DEFcg, USRcg))
      return FALSE;

   Auswertung(&modulo, &threshold, DEFsue, USRsue);

   return TRUE;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Waisenbehandlung ------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void KPV_SetzeWaisenmord (BOOLEAN morden)
{
  Waisenmord = morden;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Restekiste ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

unsigned long int KPH_LebendeKPsZaehlen(void)
{
   /* XXX todo: KDH */
  return LONG_MAX;
}


