/**********************************************************************
* Filename : MNF_Print.c
*
* Kurzbeschreibung : MultipleNormalFormen: Ausgabe der Mengen/Elemente
*
* Autoren : Andreas Jaeger
*
* 18.11.96: Erstellung
*
*
*
* Benoetigt folgende Module:
*  - Ausgaben: Ausgeben von Termen
*  - Stack:
*
* Benutzte Globale Variablen aus anderen Modulen:
*   zur Zeit keine
*
**********************************************************************/

#include "antiwarnings.h"
#include "general.h"
#include "compiler-extra.h"
#include "MNF_Print.h"
#include <limits.h>
#include "Hash.h"
#include "MNF_intern.h"
#include "MNF.h"
#include "NFBildung.h"
#include "Ausgaben.h"
#include "Stack.h"
#include "pcl.h"


INLINE static MNF_ColTermPtrT
MNF_Get_ID (MNF_MengeT *menge, PCL_ObjektNummerT id) {
   MNF_ColTermPtrT       colTermPtr = NULL;
   HashEntryPtrT         TraversePtr;

   HA_InitTraverseN(&menge->HashTabelle, &TraversePtr);
   while (!HA_EndTraverseN(&TraversePtr)) {
      HA_NextTraverseN(&TraversePtr, (void **)&colTermPtr);
      if (colTermPtr->pcl_info.pclID == id) {
         return colTermPtr;
      }
   }
   return NULL;
}

static void BuildStack (MNF_MengeT *menge, StackPtrT stack,
                        MNF_ColTermPtrT lastElem)
{
   MNF_ColTermPtrT       acctTermPtr;
   PCL_ObjektNummerT     parentID;

   acctTermPtr = lastElem;
   parentID = acctTermPtr->pcl_info.parentID;
   ST_New (stack);
   ST_Push (stack, acctTermPtr);
   
   while (parentID != 0) {
      acctTermPtr = MNF_Get_ID (menge, parentID);
      ST_Push (stack, acctTermPtr);
      parentID = acctTermPtr->pcl_info.parentID;
   }
}

static void Output_OneStep (PCL_ObjektNummerT *pclIDPtr, PCL_SeitenT Seite,
                            MNF_ColTermPtrT colTerm, TermT oldTerm,
                            TermT TermAndereSeite)
{
   TermT      copyTerm;
   UTermT     Index;
   UTermT     redPos;
   
   if (oldTerm != NULL) {
      /* Wurde Term mit NF_Bildung reduziert?-> Spezialbehandlung */
      if (colTerm->pcl_info.RENummer == 0) {
         copyTerm = TO_Termkopie (oldTerm);      
         PCL_Reduktionsbeginn(Ziel, pclIDPtr, Seite, TermAndereSeite);
         (void)NF_NormalformRE (copyTerm);
         TO_TermLoeschen (copyTerm);
      } else {
         PCL_Reduktionsbeginn(Ziel, pclIDPtr, Seite, TermAndereSeite);
         /* pcl_info.redPosition ist eine Stelle in oldTerm, wir suchen aber die analoge
            Stelle in colTerm->term fuer die PCL-Ausgabe */
         redPos = colTerm->term;
         TO_doStellen(oldTerm, Index, 
                      {
                         if (Index == colTerm->pcl_info.redPosition)
                            break;
                         else
                            redPos = TO_Schwanz (redPos);
                      });
         PCL_Reduktion (colTerm->term, redPos,
                        &(colTerm->pcl_info.REpclID), colTerm->pcl_info.RESeite);
      }
   }
}

/*
  Ausgabe des Zielpaares mit allen Vorfahren fuer PCL
  */
void MNF_PrintJoinedPCL (MNF_MengeT *menge, PCL_ObjektNummerT *pclIDPtr) {
   MNF_ColTermPtrT       leftPtr, rightPtr;
   PCL_ObjektNummerT     acctLeftID, acctRightID;
   StackT                stackLeft, stackRight;
   TermT                 leftTerm, rightTerm;
   
   if (!menge->joinable) {
      return;
   }

   if (MNF_tcolour_red(*menge->joinedCTerm1)) {
      rightPtr = menge->joinedCTerm1;
      leftPtr = menge->joinedCTerm2;
      /* Wenn der linke Pointer NULL ist, dann muessen wir das initiale Element nehmen */
      if (leftPtr == NULL)
         leftPtr = menge->initialCTermLeft;
   }
   else {
      rightPtr = menge->joinedCTerm2;
      leftPtr = menge->joinedCTerm1;
      /* Wenn der rechte Pointer NULL ist, dann muessen wir das initiale Element nehmen */
      if (rightPtr == NULL)
         rightPtr = menge->initialCTermRight;
   }

   BuildStack (menge, &stackLeft, leftPtr);
   BuildStack (menge, &stackRight, rightPtr);

   /* mal erst die Anfangsziele nehmen */
   leftPtr  = ST_Pop (&stackLeft);
   rightPtr = ST_Pop (&stackRight);

   /* leftTerm/rightTerm speichern immer den Vorgaenger, dieser wird
      vom PCL-Protokoll benoetigt.
      */

   leftTerm = leftPtr->term;
   rightTerm = rightPtr->term;
   /* nach left, right die jeweils erste Ableitung,
      sofern es eine gibt
    */
   if (ST_Empty (&stackLeft)) {
      leftPtr = NULL;
      acctLeftID = LONG_MAX;
   } else {
      leftPtr = ST_Pop (&stackLeft);
      acctLeftID  = leftPtr->pcl_info.pclID;
   }
   if (ST_Empty (&stackRight)) {
      rightPtr = NULL;
      acctRightID = LONG_MAX;
   } else {
      rightPtr = ST_Pop (&stackRight);
      acctRightID = rightPtr->pcl_info.pclID;
   }


   while (leftPtr || rightPtr) {
      if (leftPtr && acctLeftID <= acctRightID) {
         /* Element von links ausgeben */
         Output_OneStep (pclIDPtr,linkeSeite, leftPtr, leftTerm,
                         rightTerm);
         leftTerm = leftPtr->term;
         if (ST_Empty (&stackLeft)) {
            leftPtr = NULL;
            acctLeftID = LONG_MAX;
         } else {
            leftPtr = ST_Pop (&stackLeft);
            acctLeftID  = leftPtr->pcl_info.pclID;
         }
      } else if (rightPtr && acctRightID <= acctLeftID) {
         /* Element von rechts ausgeben */
         Output_OneStep (pclIDPtr,rechteSeite,rightPtr, rightTerm,
                         leftTerm);
         rightTerm = rightPtr->term;
         if (ST_Empty (&stackRight)) {
            rightPtr = NULL;
            acctRightID = LONG_MAX;
         } else {
            rightPtr = ST_Pop (&stackRight);
            acctRightID = rightPtr->pcl_info.pclID;
         }
      }
   }
  
   PCL_ZielBewiesen (pclIDPtr, leftTerm, rightTerm);
   ST_Delete (&stackLeft);
   ST_Delete (&stackRight);
}


/* Ausgeben aller neuen Elemente */
void MNF_PrintMengeNew (MNF_MengeT *menge, const char *leftStr, const char *rightStr,
                        const char *title, TermT t1, TermT t2)
{
   MNF_ColTermPtrT       colTermPtr = NULL;
   BOOLEAN               first = TRUE;
   unsigned int          countPrintedR = 0;
   unsigned int          countPrintedG = 0;
   int                   AusgabeLevel = IO_AusgabeLevel();
   const unsigned int    maxPrintElems = 20;
   HashEntryPtrT         TraversePtr;

   if ((AusgabeLevel < 1) || (menge->joinable))
      return;
   HA_InitTraverseN(&menge->HashTabelle, &TraversePtr);
   while (!HA_EndTraverseN (&TraversePtr)) {
      HA_NextTraverseN (&TraversePtr, (void **)&colTermPtr);
      if (MNF_tnew (*colTermPtr)) {
         if (first) {
            fprintf (IO_Strom, title);
            if (AusgabeLevel > 1)
               IO_DruckeFlex(": %t ?= %t\n", t1, t2);
            else
               fprintf (IO_Strom, "\n");
            first = FALSE;
         }
         MNF_setold (*colTermPtr);
         if (AusgabeLevel > 2) {
            if (MNF_tcolour_green (*colTermPtr) && (countPrintedG < maxPrintElems)) {
               countPrintedG++;
               fprintf (IO_Strom, leftStr);
               IO_DruckeFlex("%t\n", colTermPtr->term);
            }
            else if (MNF_tcolour_red (*colTermPtr) && (countPrintedR < maxPrintElems)) {
               countPrintedR++;
               fprintf (IO_Strom, rightStr);
               IO_DruckeFlex("%t\n", colTermPtr->term);
            }
         }
      }
   }
   /* Wenn modifiziert wurde, wird noch die Summe der Elemente ausgegeben */
   if (!first) {
      fprintf (IO_Strom, "no. of elems in goal set: %u\n", menge->size);
      (void)fflush (IO_Strom);
   }
}


/* Ausgeben aller Elemente */
void MNF_PrintMengeComplete (MNF_MengeT *menge, const char *leftStr, const char *rightStr)
{
   MNF_ColTermPtrT       colTermPtr = NULL;
   HashEntryPtrT         TraversePtr;

   MNF_PrintMengeStatLong (menge);
   if (menge->size == 0)
      return;

   /* erst die linke Seite */
   HA_InitTraverseN (&menge->HashTabelle, &TraversePtr);
   while (!HA_EndTraverseN (&TraversePtr)) {
      HA_NextTraverseN (&TraversePtr, (void **)&colTermPtr);
      if (MNF_tcolour_green (*colTermPtr)) {
         fprintf (IO_Strom, leftStr);
         IO_DruckeFlex("%t\n", colTermPtr->term);
      }
   }
   HA_InitTraverseN (&menge->HashTabelle, &TraversePtr);
   while (!HA_EndTraverseN (&TraversePtr)) {
      HA_NextTraverseN (&TraversePtr, (void **)&colTermPtr);
      if (MNF_tcolour_red (*colTermPtr)) {
         fprintf (IO_Strom, rightStr);
         IO_DruckeFlex("%t\n", colTermPtr->term);
      }
   }
}

/* Ausgabe einer kurzen Statistik zur Nachfolgermenge */
void MNF_PrintMengeStatShort (MNF_MengeT *menge) 
{
   fprintf (IO_Strom, "\tno. of elems in set: %u\n", menge->size);
}


/* Ausgabe einer langen Statistik zur Nachfolgermenge */
void MNF_PrintMengeStatLong (MNF_MengeT *menge) 
{
   fprintf (IO_Strom, "\tno. of elems in set: %u\n", menge->size);
   fprintf (IO_Strom, "\tuse analyse: %s\n", menge->useAnalyse ? "yes" : "no");
   fprintf (IO_Strom, "\ttry all elements: %s\n", menge->MNF_AddTryAllElements ? "yes" : "no");
   fprintf (IO_Strom, "\thash size: %d\n", menge->HashTabelle.size);

}


/*
  Ausgabe eines CTerms mit allen Informationen
  */
void MNF_PrintColTerm (MNF_ColTermT *colTermPtr) 
{

   if (colTermPtr->pcl_info.parentID == 0)
      fprintf(IO_Strom, "Term no. %lu (initial", colTermPtr->pcl_info.pclID);
   else {
      fprintf(IO_Strom, "Term no %lu (from: %4lu with %s %5ld order: ",
              colTermPtr->pcl_info.pclID,
              colTermPtr->pcl_info.parentID,
              colTermPtr->pcl_info.RENummer > 0 ? "rule" : "equation",
              colTermPtr->pcl_info.RENummer);
      switch (colTermPtr->flags.vgl) {
      case Kleiner:
         fprintf (IO_Strom, " less");
         break;
      case Groesser:
         fprintf (IO_Strom, " greater");
         break;
      case Gleich:
         fprintf (IO_Strom, " equal");
         break;
      case Unvergleichbar:
         fprintf (IO_Strom, " uncomparable");
         break;
      default:
         our_error ("Unbekannter Vergleichswert");
      }
   }         
   IO_DruckeFlex("): %t\n", colTermPtr->term);
   if (colTermPtr->pcl_info.redPosition != NULL)
      IO_DruckeFlex("Position: %t\n", colTermPtr->pcl_info.redPosition);
#if MNF_TEST
   fprintf(IO_Strom, "  in=%4lu, out=%4lu, outUp=%4lu, status=%4lu, GlS=%3lu, GesS=%3lu\n",
           colTermPtr->in, colTermPtr->out, colTermPtr->outUp, colTermPtr->status,
           colTermPtr->anzGlSchritte, colTermPtr->anzGesSchritte);
#endif   
}

/*
  Ausgabe der kompletten Nachfolgermenge mit allen Informationen,
  wird beim Debugging benoetigt.
  */
void MNF_PrintMenge (MNF_MengeT *menge) {
   MNF_ColTermPtrT       colTermPtr = NULL;
   HashEntryPtrT         TraversePtr;

   fprintf (IO_Strom, "\nComplete MNF-Set\n");
   HA_InitTraverseN (&menge->HashTabelle, &TraversePtr);
   while (!HA_EndTraverseN (&TraversePtr)) {
      HA_NextTraverseN (&TraversePtr, (void **)&colTermPtr);
      MNF_PrintColTerm (colTermPtr);
   }   
}


/*
  Ausgabe des Zielpaares mit allen Vorfahren
  */
static void MNF_PrintJoinedInternal (MNF_MengeT *menge)
{
   MNF_ColTermPtrT       acctTermPtr;
   PCL_ObjektNummerT     parentID;
   
   if (!menge->joinable) {
      fprintf (IO_Strom, "Elements are not joinable -> no Output from MNF_PrintJoinedInternal\n");
      return;
   }

   fprintf (IO_Strom, "\nJoined/Subsumed term:\n");
   acctTermPtr = menge->joinedCTerm1;
   parentID = acctTermPtr->pcl_info.parentID;
   MNF_PrintColTerm (acctTermPtr);

   while (parentID != 0) {
      acctTermPtr = MNF_Get_ID (menge, parentID);
      MNF_PrintColTerm (acctTermPtr);
      parentID = acctTermPtr->pcl_info.parentID;
   }
   
   acctTermPtr = menge->joinedCTerm2;
   if (acctTermPtr == NULL) {
      fprintf (IO_Strom, "Joined with other side of unmodified goal.\n");
      return;
   }
   fprintf (IO_Strom, "\nJoined/Subsumed term - other side:\n");

   parentID = acctTermPtr->pcl_info.parentID;
   MNF_PrintColTerm (acctTermPtr);

   
   while (parentID != 0) {
      acctTermPtr = MNF_Get_ID (menge, parentID);
      MNF_PrintColTerm (acctTermPtr);
      parentID = acctTermPtr->pcl_info.parentID;
   }
}

void MNF_PrintJoinedComplete (MNF_MengeT *menge) {

   MNF_PrintJoinedInternal (menge);
}


void MNF_PrintJoinedShort (MNF_MengeT *menge) {
   MNF_ColTermPtrT       leftPtr, rightPtr;

   if (!menge->joinable) {
      fprintf (IO_Strom, "Elements are not joinable -> no Output from MNF_PrintJoinedShort\n");
      return;
   }

   /* joinedCTerm1 existiert immer, joinedCTerm2 nicht unbedingt */
   if (menge->joinedCTerm2 == NULL) {
      if (MNF_tcolour_red (*menge->joinedCTerm1)) {
         rightPtr = menge->joinedCTerm1;
         leftPtr = menge->initialCTermLeft;
      } else {
         leftPtr = menge->joinedCTerm1;
         rightPtr = menge->initialCTermRight;
      }
   } else {
      if (MNF_tcolour_red (*menge->joinedCTerm1)) {
         rightPtr = menge->joinedCTerm1;
         leftPtr = menge->joinedCTerm2;
      } else {
         rightPtr = menge->joinedCTerm2;
         leftPtr = menge->joinedCTerm1;
      }
   }
      
   fprintf (IO_Strom, "\nJoined/Subsumed term pair:\n");
   IO_DruckeFlex(" %t = %t\n", leftPtr->term, rightPtr->term);
}
