/**********************************************************************
* Filename : Ordnungen.c
*
* Kurzbeschreibung : Ordnungen
*
* Autoren : Andreas Jaeger, Arnim Buch
*
* 14.12.96: Erstellung
* 25.02.97: Vergleichs- und Groessertest-Funktionen aus Hauptkomponenten
*           hierher verlagert. Arnim.
*
*
* Benoetigt folgende Module:
*
* Benutzte Globale Variablen aus anderen Modulen:
*
**********************************************************************/

#include "Ausgaben.h"
#include "compiler-extra.h"
#include <ctype.h>
#include "general.h"
#include "parser.h"
#include "FehlerBehandlung.h"
#include "InfoStrings.h"
#include "KBO.h"
#include "LPO.h"
#include "LPOVortests.h"
#include "Ordnungen.h"
#include "PhilMarlow.h"
#include "Praezedenzgenerator.h"
#include <string.h>
#include "Statistiken.h"
#include "SymbolOperationen.h"


BOOLEAN ORD_OrdnungTotal = FALSE;

/* -------------------------------------------------------------------------------*/
/* --------------- Term-Vergleich (ORD-1) ----------------------------------------*/
/* -------------------------------------------------------------------------------*/

typedef VergleichsT (*ORDVergleichsT)  (UTermT, UTermT);
typedef BOOLEAN     (*ORDTestT)  (UTermT, UTermT);

static ORDVergleichsT ORD_Vergleichsfkt;
static ORDTestT       ORD_UnifTestfkt, ORD_RedTestfkt;



/* -------------------------------------------------------------------------------*/
/* --------------- Spezialordnungen ----------------------------------------------*/
/* -------------------------------------------------------------------------------*/

#define /*BOOLEAN*/ Vergleich2Test(s,t, eq, gt) \
  (eq(s,t) ? Gleich : gt(s,t) ? Groesser : gt(t,s) ? Kleiner : Unvergleichbar)
  /* zu selten / speziell fuer Plazierung in general.h */

static BOOLEAN LeererTest(UTermT s MAYBE_UNUSEDPARAM, UTermT t MAYBE_UNUSEDPARAM)
{ return FALSE; }

static VergleichsT LeererVergleich(UTermT s, UTermT t)
{ return TO_TermeGleich(s,t) ? Gleich : Unvergleichbar; }

/* Fassung fuer allquantifizierte Ziele.. */
static BOOLEAN Col3Test(UTermT s, UTermT t)
{ 
  if (TO_IstKombinatorTupel(s,t))
    return TRUE;
  else {
    SymbolT c = TO_TopSymbol(s);
    if (SO_SymbIstKonst(c))
      if ( (  !PM_Existenzziele()
	    || (SO_SymbUngleich(c,SO_TRUE) && SO_SymbUngleich(c,SO_FALSE)))
           && TO_TermIstGrund(t))
	return TRUE;
    return FALSE;
  }
}

static VergleichsT Col3Vergleich(UTermT s, UTermT t)
{ return Vergleich2Test(s,t, TO_TermeGleich, Col3Test); }


static BOOLEAN Col4Test(UTermT s_, UTermT t_)
{ 
  if (PM_Existenzziele()) {
    SymbolT s = TO_TopSymbol(s_),
            t = TO_TopSymbol(t_);
    return 
      (s==SO_EQ && (t==SO_TRUE || t==SO_FALSE)) ||
      (s==SO_TRUE && t==SO_FALSE);
  }
  else
    return FALSE;
}

static VergleichsT Col4Vergleich(UTermT s, UTermT t)
{ return Vergleich2Test(s,t, TO_TermeGleich, Col4Test); }

/* ORD-1 */
/*-------*/

VergleichsT ORD_VergleicheTerme(UTermT term1, UTermT term2)
/* Vergleicht die beiden angegebenen Terme nach der eingestellten Ordnung. */
{
  VergleichsT result;
  /* ganz normaler Ordnungsaufruf ... */
  result = ((*ORD_Vergleichsfkt) (term1, term2));
  return result;
}

/* ORD-1a*/
/*-------*/
BOOLEAN ORD_TermGroesserRed(UTermT term1, UTermT term2)
/* Testet, ob term1 groesser als term2 ist (bzgl. der eingestellten Ordnung). */
{
  BOOLEAN res = ((*ORD_RedTestfkt) (term1, term2));
  return  res; 
}

BOOLEAN ORD_TermGroesserUnif(UTermT term1, UTermT term2)
/* Testet, ob term1 groesser als term2 ist (bzgl. der eingestellten Ordnung). */
{
  BOOLEAN res = ((*ORD_UnifTestfkt) (term1, term2));
  return res;
}

void ORD_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  if (what == ParamEvaluate){
    ordering ord = PA_Ordering;
    char *newprec;
    char *newperm;  
    fprintf(IO_Strom, "\n* ORDERING: "); 
    if (IS_FindOption (DEF, USR, 1, "kbo") == 0)  
      ord  = ordering_kbo; 
    else if (IS_FindOption (DEF, USR, 1, "lpo") == 0)                  
      ord  = ordering_lpo; 
    else if (IS_FindOption (DEF, USR, 1, "empty") == 0)                  
      ord  = ordering_empty; 
    else if (IS_FindOption (DEF, USR, 1, "col3") == 0)                  
      ord  = ordering_col3; 
    else if (IS_FindOption (DEF, USR, 1, "col4") == 0)                  
      ord  = ordering_col4; 
    switch (ord ) { 
    case ordering_kbo:   
      fprintf(IO_Strom, "KBO"); 
      break; 
    case ordering_lpo:   
      fprintf(IO_Strom, "LPO"); 
      break; 
    case ordering_empty:   
      fprintf(IO_Strom, "empty"); 
      break; 
    case ordering_col3:   
      fprintf(IO_Strom, "col3"); 
      break; 
    case ordering_col4:   
      fprintf(IO_Strom, "col4"); 
      break; 
    default:  
      ST_ErrorStatistik("Unknown Ordering",FALSE);
      break; 
    } 
    newprec =IS_ValueOfOption (DEF,USR,"prec"); 
    newperm=IS_ValueOfOption (DEF,USR,"perm"); 
    if (newprec) { 
      fprintf(IO_Strom, ", (new) precedence %s",newprec );
      if (!(SO_PraezedenzKorrekt(newprec)) )
        fprintf(IO_Strom, " invalid, leaving unchanged");
    } 
    else if (newperm) { 
      char *punkt; 
      int permno; 
      int permdegree= SO_AnzahlFunktionen ; 
      punkt=strchr(newperm,'.'); 
      if (punkt){ 
        *punkt='\0'; 
        permdegree = atoi(punkt+1); 
      } 
      permno = atoi(newperm); 
      if ((newprec =PG_Permutator(permno, permdegree)))
        fprintf(IO_Strom, ", permutation %u is %s with degree %d", permno, newprec , permdegree);  
      else
        fprintf(IO_Strom, ", permutation %u with degree %d invalid or precedence not total", permdegree, permno);  
    } 
    fprintf(IO_Strom, "\n            current precedence is "); 
    if (SO_PraezedenzTotal ) { 
      unsigned int i; 
      for(i=0; i< SO_AnzahlFunktionen ; i++) { 
        IO_DruckeSymbol(SO_TotaleKette[i]); 
        if (i<SO_AnzahlFunktionen-1) 
          fprintf(IO_Strom, " > "); 
      } 
    } 
    else 
      fprintf(IO_Strom, "not total"); 
    fprintf(IO_Strom, "\n"); 
  }
  else
    fprintf(IO_Strom, "-ord [kbo|lpo|empty]:prec=a.b.c..d.e.a  redefine ordering and precedence\n"
            "KBO only with weights=1; precedence above would be a>b>c;d>e>a\n");
}
  
/*=================================================================================================================*/
/*= N+1. Weiter im Text                                                                                           =*/
/*=================================================================================================================*/

static void GgfGewichteAendern(InfoStringT DEF, InfoStringT USR);

BOOLEAN ORD_InitAposteriori (InfoStringT DEF, InfoStringT USR) 
{
  ordering oldOrd = PA_Ordering;
  char *newprec;
  char *newperm;  

  fprintf(IO_Strom, "\n* ORDERING: ");
  if (IS_FindOption (DEF, USR, 1, "kbo") == 0)  
    PA_Ordering  = ordering_kbo;
  else if (IS_FindOption (DEF, USR, 1, "lpo") == 0)
    PA_Ordering  = ordering_lpo;
  else if (IS_FindOption (DEF, USR, 1, "empty") == 0)
    PA_Ordering  = ordering_empty;
  else if (IS_FindOption (DEF, USR, 1, "col3") == 0)
    PA_Ordering  = ordering_col3;
  else if (IS_FindOption (DEF, USR, 1, "col4") == 0)
    PA_Ordering  = ordering_col4;

  ORD_OrdnungTotal = FALSE;

  switch (PA_Ordering) { 
  case ordering_kbo:
    ORD_Vergleichsfkt = KB_KBOVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = KB_KBOGroesser;
    ORD_OrdnungTotal = SO_PraezedenzTotal;
    if (oldOrd != PA_Ordering)
      fprintf(IO_Strom, "changed to KBO");
    break;
  case ordering_lpo:   
    if (oldOrd != PA_Ordering) 
      fprintf(IO_Strom, "changed to LPO");
    ORD_Vergleichsfkt = LP_LPO;
    ORD_RedTestfkt =  ORD_UnifTestfkt = LP_LPOGroesser;
    LV_InitAposteriori();
    ORD_OrdnungTotal = SO_PraezedenzTotal;
    break;
  case ordering_empty:
    if (oldOrd != PA_Ordering) 
      fprintf(IO_Strom, "changed to empty");
    ORD_Vergleichsfkt = LeererVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = LeererTest;
    break;
  case ordering_col3:
    if (oldOrd != PA_Ordering) 
      fprintf(IO_Strom, "changed to col3");
    ORD_Vergleichsfkt = Col3Vergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = Col3Test;
    TO_KombinatorapparatAktivieren();
    break;
  case ordering_col4:
    if (oldOrd != PA_Ordering) 
      fprintf(IO_Strom, "changed to col4");
    ORD_Vergleichsfkt = Col4Vergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = Col4Test;
    TO_KombinatorapparatAktivieren();
    break;
  default:
    ST_ErrorStatistik("Unknown Ordering",FALSE);
    break;
  } 

  newprec =IS_ValueOfOption (DEF,USR,"prec");
  newperm=IS_ValueOfOption (DEF,USR,"perm");
  if (newprec) {
    fprintf(IO_Strom, ", (new) precedence %s", newprec );
    if (!SO_PraezedenzErneuert(newprec))  
      our_error("Invalid Precedence");
  } 
  else if (newperm) { 
    char *punkt;
    int permno;
    int permdegree= SO_AnzahlFunktionen ;
    punkt=strchr(newperm,'.');
    if (punkt) { 
      *punkt='\0';
      permdegree = atoi(punkt+1);
    } 
    permno = atoi(newperm);
    if ((newprec =PG_Permutator(permno, permdegree))) { 
      fprintf(IO_Strom, ", permutation %u is %s with degree %d", permno, newprec , permdegree);
      if(!SO_PraezedenzErneuert(newprec)) 
        ST_ErrorStatistik("Permutator generated invalid precedence\n",FALSE);
    } 
    else { 
      fprintf(IO_Strom, ", permutation %u with degree %d invalid or precedence not total", permdegree, permno);
      ST_ErrorStatistik("Invalid permutation", FALSE);
    } 
  }  

  GgfGewichteAendern(DEF,USR);
  if (PA_Ordering==ordering_kbo)
    KB_InitAposteriori();    

  fprintf(IO_Strom, "\n            current precedence is ");
  if (SO_PraezedenzTotal) { 
    unsigned int i;
    for(i=0; i< SO_AnzahlFunktionen ; i++) { 
      IO_DruckeSymbol(SO_TotaleKette[i]);
      if (i<SO_AnzahlFunktionen-1) 
        fprintf(IO_Strom, " > ");
    } 
  } 
  else 
    fprintf(IO_Strom, "not total"); 
  fprintf(IO_Strom, "\n"); 

  return TRUE;
}

void ORD_OrdnungUmstellen(ordering Ordnung)
{
  ORD_OrdnungTotal = FALSE;
  switch (Ordnung){
  case ordering_lpo:
    ORD_Vergleichsfkt = LP_LPO;
    ORD_RedTestfkt = ORD_UnifTestfkt = LP_LPOGroesser;
    LV_InitAposteriori();
    ORD_OrdnungTotal = SO_PraezedenzTotal;
    break;
  case ordering_kbo:
    ORD_Vergleichsfkt = KB_KBOVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = KB_KBOGroesser;
    KB_InitAposteriori();   /* Phi auf Gueltigkeit testen.*/
    ORD_OrdnungTotal = SO_PraezedenzTotal;
    break;
  case ordering_empty:
    ORD_Vergleichsfkt = LeererVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = LeererTest;
    break;
  case ordering_col3:
    ORD_Vergleichsfkt = Col3Vergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = Col3Test;
    break;
  case ordering_col4:
    ORD_Vergleichsfkt = Col4Vergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = Col4Test;
    break;
  default:
    ST_ErrorStatistik("Invalid Ordering in ORD_OrdnungUmstellen", FALSE);
    break;
  }
  PA_Ordering = Ordnung;
}


static void GgfGewichteAendern(InfoStringT DEF, InfoStringT USR)
{
  BOOLEAN DefaultAngegeben = IS_ValueOfOption(DEF, USR, "DEF")!=NULL,
          Geaendert = DefaultAngegeben;
  long int Default = DefaultAngegeben ? atoi(IS_ValueOfOption(DEF, USR, "DEF")) : 1,
           *Gewichte;
  SymbolT i;
  Gewichte = array_alloc(SO_AnzahlFunktionen,long int);
  SO_forFktSymbole(i) {
    char *wert = IS_ValueOfOption(DEF, USR, IO_FktSymbName(i));
    if (wert && (isdigit((int)*wert) || *wert=='-')) {
      Gewichte[i] = atoi(wert);
      Geaendert = TRUE;
    }
    else
      Gewichte[i] = -1;
  }
  if (Geaendert)
    SO_PhiUmstellen(Gewichte,DefaultAngegeben,Default);
  our_dealloc(Gewichte);
}
