/* -------------------------------------------------------*/
/* ------------- Includes --------------------------------*/
/* -------------------------------------------------------*/

#include "antiwarnings.h"
#if defined (SUNOS5)
#include <fcntl.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/fault.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#endif
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/utsname.h>

#include "compiler-extra.h"
#include "Ausgaben.h"
#include "DomaenenMemorium.h"
#include "DSBaumAusgaben.h"
#include "DSBaumKnoten.h"
#include "DSBaumOperationen.h"                                                   /* wegen Ausdehnung Variablenbank */
#include "FehlerBehandlung.h"
#include "general.h"
#include "Interreduktion.h"
#include "KPVerwaltung.h"
#include "MatchOperationen.h"                                                    /* wegen Ausdehnung Variablenbank */
#include "MNF.h"
#include "NewClassification.h"
#include "NFBildung.h"                                                      /* wegen Ausdehnung des Reduktionspfades */
#include "Parameter.h"
#include "parser.h"
#include "pcl.h"
#include "RUndEVerwaltung.h"
#include "Signale.h"
#include "Statistiken.h"
#include "StrahlMemorium.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "TermIteration.h"
#include "TermVerfolgung.h"
#include "Unifikation1.h"                                                 /* wegen Ausdehnung Variablenbank, Stack */
#include "Unifikation2.h"
#include "ZeitMessung.h"
#include "Zielverwaltung.h"
#include "version.h"
#include "WaldmeisterII.h"

extern char **environ;



StatisticsT Stat;

extern unsigned long int BK_AnzEinzelnAllokierterNachfolgerArrays;

extern volatile DomaenenT Domaenenliste;


extern BOOLEAN MNF_Initialised;

extern unsigned long int AnzRegelnAktuell;
extern unsigned long int AnzGleichungenAktuell;
extern unsigned long int AnzMonogleichungen;
extern unsigned long int AnzAktMonogleichungen;


extern SV_SpeicherManagerT *ManagerListe;

static void DruckeSpeicherManagerStatistik(int, const char*, unsigned long, unsigned long, unsigned long,
  unsigned long, unsigned long, unsigned long) MAYBE_UNUSEDFUNC;
static void DruckeSpeicherManagerStatistik(int Grad, const char *Name, unsigned long Blockgroesse, 
  unsigned long ElementGroesse, unsigned long Anforderungen, unsigned long Freigaben,
  unsigned long InsgesamtAngelegteBloecke, unsigned long MaximalGleichzeitigBelegteElemente )
/* Grad: 0 - Ausgabe nur, wenn Manager benutzt wurde und Diskrepanzen
         1 - Ausgabe, wenn Manager benutzt wurde, unabhaengig von Diskrepanzen
*/
{
  BOOLEAN Ausgeben = FALSE;
  
  switch (Grad){
  case 0:
    Ausgeben = ((Anforderungen+Freigaben>0) && (Anforderungen != Freigaben));
    break;
  case 1:
    Ausgeben = (Anforderungen+Freigaben>0);
    break;
  default:
    our_error("Falscher Ausgabe-Grad in IO_DruckeSpeicherManagerStatistik\n");
    break;
  }
  if (Ausgeben){
    IO_DruckeKurzUeberschrift(Name, '+');
    IO_DruckeStatistikZeile("Block size", Blockgroesse);
    IO_DruckeStatistikZeile("Element size", ElementGroesse);
    IO_DruckeStatistikZeile("Total number of allocations",Anforderungen );
    IO_DruckeStatistikZeile("Total number of deallocations",Freigaben );
    IO_DruckeStatistikZeile("Total number of allocated blocks",InsgesamtAngelegteBloecke );
    IO_DruckeStatistikZeile("Max. number of elements used at the same time",MaximalGleichzeitigBelegteElemente );
  }
}


static void DruckeNestedSpeicherManagerStatistik(const char*, unsigned long, unsigned long, unsigned long,
  BOOLEAN, unsigned long, unsigned long) MAYBE_UNUSEDFUNC;
static void DruckeNestedSpeicherManagerStatistik(const char *Name, unsigned long Blockgroesse,
  unsigned long ElementGroesse, unsigned long Anforderungen, BOOLEAN Fehlerfall, 
  unsigned long AnzNochBelegterElemente, unsigned long InsgesamtAngelegteBloecke )
{
  char VerlaengerterName[SV_maxLaengeNameNestedManager];
  strcpy(VerlaengerterName,Name);
  strcat(VerlaengerterName," (Nested-Manager)");
  IO_DruckeKurzUeberschrift(VerlaengerterName, '+');
  IO_DruckeStatistikZeile("Block size", Blockgroesse);
  IO_DruckeStatistikZeile("Element size", ElementGroesse);
  IO_DruckeStatistikZeile("Total number of allocations",Anforderungen );
  if (Fehlerfall)
    IO_DruckeStatistikZeileString("Remaining number of allocated elements","nicht ermittelbar");
  else
    IO_DruckeStatistikZeile("Remaining number of allocated elements",AnzNochBelegterElemente );
  IO_DruckeStatistikZeile("Total number of allocated blocks",InsgesamtAngelegteBloecke );
  IO_DruckeStatistikZeile("Max. number of elements usable at the same time",Blockgroesse*InsgesamtAngelegteBloecke);
}


void SV_Statistik(BOOLEAN Fehlerfall MAYBE_UNUSEDPARAM, BOOLEAN NurDiskrepanzen MAYBE_UNUSEDPARAM, InfoStringT IS)   
{                                                                /* Gibt die Statistik der Speicherverwaltung aus. */
  
  if (IS_IsIn("complete", IS) || IS_IsIn("SV", IS)){
#if DM_STATISTIK
    BOOLEAN UeberschriftGedruckt = FALSE;
    unsigned long GesamtAnforderungen = 0,
                  GesamtFreigaben = 0,
                  GesamtAngelegteBloecke = 0,
                  GesamtBelegterSpeicher = 0;
                  SV_SpeicherManagerT *runny = ManagerListe;
      
    if (!NurDiskrepanzen) {
      UeberschriftGedruckt = TRUE;
      IO_DruckeUeberschrift("Memory statistics");
    }
      
    while (runny) {
      SV_SpeicherManagerT mbm = *runny;
      if (mbm.Nested) {
        if (Fehlerfall) {
          if (!UeberschriftGedruckt) {
            IO_DruckeKurzUeberschrift("Discrepancies in memory allocation", '*');
            UeberschriftGedruckt = TRUE;
          }
          DruckeNestedSpeicherManagerStatistik(mbm.Name, mbm.block_size, mbm.element_size, 
                                                   mbm.Anforderungen, TRUE, 0, mbm.InsgesamtAngelegteBloecke);
        } else {
          unsigned long int AnzNochBelegterElemente = SV_AnzNochBelegterElemente(mbm);
          GesamtFreigaben += mbm.Anforderungen - AnzNochBelegterElemente;
          /* Hier kann nur festgestellt werden, ob noch nicht freigegebene Elemente in den Bloecken stehen */
          if (AnzNochBelegterElemente || FALSE)
            DruckeNestedSpeicherManagerStatistik(mbm.Name, mbm.block_size, mbm.element_size, mbm.Anforderungen,
                                                 FALSE, AnzNochBelegterElemente, mbm.InsgesamtAngelegteBloecke);
        }
      }
      else {
        int AusgabeGrad = NurDiskrepanzen?0:1;
        GesamtFreigaben += mbm.Freigaben;
        if (!UeberschriftGedruckt && mbm.Anforderungen != mbm.Freigaben) {
          IO_DruckeKurzUeberschrift("Discrepancies in memory allocation", '*');
          UeberschriftGedruckt = TRUE;
        }
        DruckeSpeicherManagerStatistik(AusgabeGrad,mbm.Name, mbm.block_size, mbm.element_size, mbm.Anforderungen,
                                       mbm.Freigaben, mbm.InsgesamtAngelegteBloecke, mbm.MaximalGleichzeitigBelegteElemente);
      }
      GesamtAnforderungen += mbm.Anforderungen;
      GesamtAngelegteBloecke += mbm.InsgesamtAngelegteBloecke;
      GesamtBelegterSpeicher += mbm.InsgesamtAngelegteBloecke * mbm.element_size * mbm.block_size;
      runny = runny->Nachf;
    }
    if (UeberschriftGedruckt) {
      IO_DruckeZeile('+');
      IO_DruckeStatistikZeile("Total number of allocations",GesamtAnforderungen );
      IO_DruckeStatistikZeile("Total number of deallocations",GesamtFreigaben );
      IO_DruckeStatistikZeile("Total number of allocated blocks",GesamtAngelegteBloecke );
      IO_DruckeStatistikZeile("Thereby allocated memory in bytes", GesamtBelegterSpeicher);
    }
#elif !COMP_POWER
    fprintf(IO_Strom, "\nMemory statistics not compiled into object module.\n");
    fprintf(IO_Strom, "If desired, define constant DM_STATISTIK as 1\n");
    fprintf(IO_Strom, "in header file DomaenenMemorium.h.\n\n");
#endif
  }
}



enum Zeitsorten {User,System,Wall}; /* XXX */
extern struct timeval Start[3];
extern struct timeval Dauer[3];
static void DruckeZeitSolo(FILE *fp, struct timeval time) /* Neufassung. BL. */
{ 
  long seconds, milliseconds;

  milliseconds = (time.tv_usec + 500) / 1000;
  seconds      = time.tv_sec + milliseconds / 1000;
  milliseconds = milliseconds % 1000;

  fprintf(fp, "%ld.%03ld", seconds, milliseconds);
}


static void ZM_Statistik(InfoStringT IS)
{
   if (!(IS_IsIn("ZM", IS) || IS_IsIn("all", IS)))
     return;

   if (IS_IsIn("standard", IS)){
      IO_DruckeUeberschrift("TIMES (user- and system-time in seconds)");
      DruckeZeitSolo(IO_Strom, Dauer[User]);
      fprintf(IO_Strom, "s ut ");
      DruckeZeitSolo(IO_Strom, Dauer[System]);
      fprintf(IO_Strom, "s st (");
      DruckeZeitSolo(IO_Strom, Dauer[Wall]);
      fprintf(IO_Strom, " wct)");
      fprintf(IO_Strom, "   total time (without initialisation)\n");
   }
   else if (IS_IsIn("latex", IS)){
      fprintf(IO_Strom, " ");
      DruckeZeitSolo(IO_Strom, Dauer[User]);
      fprintf(IO_Strom, "s ut ");
      DruckeZeitSolo(IO_Strom, Dauer[System]);
      fprintf(IO_Strom, "s st (");
      DruckeZeitSolo(IO_Strom, Dauer[Wall]);
      fprintf(IO_Strom, " wct)");
   }
}



/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Interne Funktionen ++++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

typedef struct SPTS{
   char strength[6];
   unsigned long int sp;
} StichpunktTyp;

static StichpunktTyp *SPArray;

static int SPArrayPos=0, SPArrayLen=0;

static void StichpunktMerken(unsigned long int sp, char *strength)
/* Stichpunkte der wahlfreien KP-Mengen-Interreduktion merken fuer spaetere Ausgabe. */
{
  if (SPArray == NULL){
     SPArrayLen = 10;              /* Stichpunkt-Array initialisieren. */
     SPArray = our_alloc(SPArrayLen*sizeof(StichpunktTyp));
     SPArrayPos = 0;
  }
   if (SPArrayPos == SPArrayLen-1){
      SPArray = our_realloc(SPArray, (SPArrayLen+10) * sizeof(StichpunktTyp), "IO:Stichpunkt-Array");
      SPArrayLen +=10;
   }
   strncat(SPArray[SPArrayPos].strength, strength, 5);
   SPArray[SPArrayPos++].sp = sp;
   
}


/*
  Schreiben der gesamten Regel/Gleichungsmenge als Spezifikation auf Platte
  */

static void SpezifikationSchreiben (void)
{
   char            AusgabeDatei[INPUT_STRLEN];
   mode            Mode = get_mode();
   SymbolT         symbol;
   

   fprintf(IO_Strom, "\nName of file to write specification to: ");
   fgets(AusgabeDatei, INPUT_STRLEN, stdin);
   chop(AusgabeDatei);

   IO_AusgabeStromUmlenken(AusgabeDatei);

   fprintf (IO_Strom, "NAME\t\t%s\n\n", get_spec_name());
   fprintf (IO_Strom, "MODE\t\t");
   switch (Mode) {
   case completion:
      fprintf(IO_Strom,"COMPLETION");
      break;
   case reduction:
      fprintf(IO_Strom,"REDUCTION");
      break;
   case confluence:
      fprintf(IO_Strom,"CONFLUENCE");
      break;
   case proof:
      fprintf(IO_Strom,"PROOF");
      break;
   case termination:
      fprintf(IO_Strom,"TERMINATION");
      break;
   case convergence:
      fprintf(IO_Strom,"CONVERGENCE");
      break;
   default:
      our_error("unexpected label in SpezifikationSchreiben");
      break;
   }
   fprintf (IO_Strom, "\n\n");
   
   fprintf (IO_Strom, "SORTS\t");
   printSorts (IO_Strom);
   fprintf (IO_Strom, "\nSIGNATURE\n");
   printSig (IO_Strom);
   fprintf (IO_Strom, "ORDERING\t");
   switch (PA_Ordering) {
   case ordering_kbo: fprintf (IO_Strom, "KBO\n");break;
   case ordering_lpo: fprintf (IO_Strom, "LPO\n");break;
   default:
      our_error("unexpected label in SpezifikationSchreiben");
      break;
   }
   
   printOrdering (IO_Strom);

   fprintf (IO_Strom, "VARIABLES\n");
   fprintf (IO_Strom, "\t");
   IO_DruckeSymbol (-1);
   for (symbol = -2; symbol >= BO_NeuesteBaumvariable(); symbol--) {
      fprintf (IO_Strom, ", ");
      IO_DruckeSymbol (symbol);
   }
   fprintf (IO_Strom, " : ");
   /* wir geben allen Variablen die gleiche Sorte, nicht schoen,
      aber anscheinend einzige Moeglichkeit, da Variablen keine
      Sorte haben - hier geht Information verloren */
   printSort (IO_Strom, 1);
   fprintf (IO_Strom, "\n\n");
   

   fprintf (IO_Strom, "EQUATIONS\n");
   if (!RE_RegelmengeLeer())
      RE_AusgabeRegelmenge (0);
   if (!RE_GleichungsmengeLeer())
      RE_AusgabeGleichungsmenge(0);
   
   fprintf (IO_Strom, "CONCLUSION\n");
   Z_AusgabeZielmenge (Z_AusgabeSpez);
   IO_AusgabeStromZuruecksetzen();
}



/* -------------------------------------------------------*/
/* -------------- Initialisierung ------------------------*/
/* -------------------------------------------------------*/

void ST_InitAposteriori(void)
{
  SPArrayPos = 0;
}

/* -------------------------------------------------------*/
/* -------------- Statistiken der einzelnen Module -------*/
/* -------------------------------------------------------*/


static void ST_Statistik_standard(void)
{
  int tiefe1, tiefe2, i;
  char str[100];
  IO_DruckeKurzUeberschrift("Unselected Equations",'+');
  
  IO_DruckeStatistikZeile("Total number of ues",Stat.KPV_AnzKPGesamt);
  tiefe1 = IO_Einruecktiefe(0, Maximum(8, Stat.KPV_AnzKPAusSpezifikation, Stat.KPV_AnzKPIROpfer,
                                       Stat.KPV_AnzKPdirektZusammengefuehrt, Stat.KPV_AnzKPdirektSubsummiert, Stat.KPV_AnzKPInserted,
                                       Stat.KPV_AnzKPWeggeworfen, Stat.KPV_AnzKPDatenbank,
                                       Stat.KPV_AnzKPdirektPermSubsummiert));
  IO_DruckeStatistikZeileEingerueckt("of them from Specification",Stat.KPV_AnzKPAusSpezifikation, tiefe1);
  IO_DruckeStatistikZeileEingerueckt("of them from Killed RE", Stat.KPV_AnzKPIROpfer, tiefe1);
  IO_DruckeStatistikZeileEingerueckt("of them by   ChaseManhattan", Stat.KPV_AnzKPDatenbank, tiefe1);
  IO_DruckeStatistikZeileEingerueckt("of them from Generation", 
                                     Stat.KPV_AnzKPGesamt-Stat.KPV_AnzKPAusSpezifikation-Stat.KPV_AnzKPIROpfer-Stat.KPV_AnzKPDatenbank, tiefe1);
  IO_DruckeStatistikLeerZeile();
  
  IO_DruckeStatistikZeileEingerueckt("of them joined immediately",Stat.KPV_AnzKPdirektZusammengefuehrt, tiefe1);
  IO_DruckeStatistikZeileEingerueckt("of them subsumed immediately", Stat.KPV_AnzKPdirektSubsummiert, tiefe1);
  IO_DruckeStatistikZeileEingerueckt("of them perm-subsumed immediately", Stat.KPV_AnzKPdirektPermSubsummiert, tiefe1);
  IO_DruckeStatistikZeileEingerueckt("of them thrown away", Stat.KPV_AnzKPWeggeworfen, tiefe1);
  IO_DruckeStatistikZeileEingerueckt("of them inserted into PQ", Stat.KPV_AnzKPInserted, tiefe1);
  IO_DruckeStatistikLeerZeile();
  IO_DruckeStatistikZeileEingerueckt("ues selected",Stat.KPV_AnzKPBehandeltInVervollst, tiefe1);
  
  tiefe2 = IO_Einruecktiefe(tiefe1, Maximum(5, Stat.KPV_AnzKPspaeterZusammengefuehrt, Stat.KPV_AnzKPZuRE,
                                            Stat.KPV_AnzKPZurueckgestellt, Stat.KPV_AnzKPspaeterSubsummiert, 
                                            Stat.KPV_AnzKPspaeterPermSubsummiert));
  
  IO_DruckeStatistikZeileEingerueckt("of them not orientable, postponed", Stat.KPV_AnzKPZurueckgestellt, tiefe1);
  IO_DruckeStatistikZeileEingerueckt("of them joined", Stat.KPV_AnzKPspaeterZusammengefuehrt, tiefe2);
  IO_DruckeStatistikZeileEingerueckt("of them subsumed (forward subsumption)", Stat.KPV_AnzKPspaeterSubsummiert, tiefe2);
  IO_DruckeStatistikZeileEingerueckt("of them perm-subsumed (forward subsumption)", Stat.KPV_AnzKPspaeterPermSubsummiert, tiefe2);
  IO_DruckeStatistikZeileEingerueckt("of them made RE", Stat.KPV_AnzKPZuRE, tiefe2);
#if KPH_STATISTIK
  {
    int tiefe1, tiefe2, tiefe3;
    int AnzGueltigeKPsInMenge = KPH_LebendeKPsZaehlen();

    IO_DruckeKurzUeberschrift("Priority-Queue: Regular heap", '+');

    IO_DruckeStatistikZeile("ues inserted into PQ", Stat.AnzKPEingefuegt);
    IO_DruckeStatistikZeile("ues removed from PQ", Stat.AnzKPGeloescht);
    IO_DruckeStatistikZeileEingerueckt("of them because of early orphan murder", Stat.AnzKPFrueherWaisenmord, 
                                       IO_Einruecktiefe(0, Maximum(2,Stat. AnzKPFrueherWaisenmord, Stat.AnzKPSpaeterWaisenmord)));
    IO_DruckeStatistikZeileEingerueckt("of them because of late orphan murder", Stat.AnzKPSpaeterWaisenmord, 
                                       IO_Einruecktiefe(0, Maximum(2, Stat.AnzKPFrueherWaisenmord, Stat.AnzKPSpaeterWaisenmord)));
    IO_DruckeStatistikLeerZeile();
    IO_DruckeStatistikZeile("ues currently in PQ", Stat.AnzKPEingefuegt - Stat.AnzKPGeloescht);
    IO_DruckeStatistikZeileEingerueckt("of them still valid", AnzGueltigeKPsInMenge, IO_Einruecktiefe(0, AnzGueltigeKPsInMenge));
    IO_DruckeStatistikLeerZeile();
    IO_DruckeStatistikZeile("Max. No. of ues in PQ",Stat. AnzKPBlaeh);
    
    IO_DruckeStatistikLeerZeile();
    IO_DruckeStatistikZeile("ues touched during IRP", Stat.AnzKPIRBetrachtet);
    tiefe1 = IO_Einruecktiefe(0, Maximum(2, Stat.AnzKPIRWaisenmord, Stat.AnzKPIRBetrachtet-Stat.AnzKPIRWaisenmord));
    tiefe2 = IO_Einruecktiefe(tiefe1, Maximum(2, Stat.AnzKPIRGeloescht, Stat.AnzKPIRVeraendert));
    tiefe3 = IO_Einruecktiefe(tiefe2, Maximum(2, Stat.AnzKPIRVorgezogen, Stat.AnzKPIRZurueckgeschoben));
    
    IO_DruckeStatistikZeileEingerueckt("of them removed as orphans", Stat.AnzKPIRWaisenmord, tiefe1);
    IO_DruckeStatistikZeileEingerueckt("of them reprocessed", Stat.AnzKPIRBetrachtet-Stat.AnzKPIRWaisenmord, tiefe1);
    IO_DruckeStatistikZeileEingerueckt("of them removed", Stat.AnzKPIRGeloescht, tiefe2);
    IO_DruckeStatistikZeileEingerueckt("of them changed", Stat.AnzKPIRVeraendert, tiefe2);
    IO_DruckeStatistikZeileEingerueckt("of them prefered", Stat.AnzKPIRVorgezogen, tiefe3);
    IO_DruckeStatistikZeileEingerueckt("of them defered", Stat.AnzKPIRZurueckgeschoben, tiefe3);
  }
#endif

  IO_DruckeKurzUeberschrift("Classification", '+');
  if (ClasCriteria) {
    for (i = 0; ClasCriteria[i].criterion != Crit_LastMarker; i++){
      strcpy (str, get_criterion (ClasCriteria[i].criterion));
      strcat (str, " during CLASSIFY");
      IO_DruckeStatistikZeile(str, ClasCriteria[i].fullfilled);
    }
  }
  if (ReClasCriteria) {
    for (i = 0; ReClasCriteria[i].criterion != Crit_LastMarker; i++){
      strcpy (str, get_criterion (ReClasCriteria[i].criterion));
      strcat (str, " during RECLASSIFY");
      IO_DruckeStatistikZeile(str, ReClasCriteria[i].fullfilled);
    }
  }

  IO_DruckeKurzUeberschrift("Critical Goals", '+');
  IO_DruckeStatistikZeile("total number of critical goals", Stat.KPV_AnzCritGoalsGesamt);
  IO_DruckeStatistikZeile("current number of critical goals", Stat.KPV_AnzCritGoalsActual);
  IO_DruckeStatistikZeile("max number of critical goals", Stat.KPV_AnzMaxCritGoals);
  IO_DruckeStatistikZeile("joined during interreduction", Stat.KPV_AnzIRGleich);
  IO_DruckeStatistikZeile("reduced during interreduction", Stat.KPV_AnzIRReduziert);
  IO_DruckeStatistikZeile("killed orphans during interreduction", Stat.KPV_AnzIRWMord);
  
  IO_DruckeKurzUeberschrift("Interreduktion", '+');
  IO_DruckeStatistikZeile("Regeln fielen der Interreduktion zum Opfer", Stat.AnzRegWgLinksRedWeg);
  IO_DruckeStatistikZeile("Regeln rechts reduziert", Stat.AnzRegRechtsRed);
  IO_DruckeStatistikZeile("Gleichungen fielen der Interreduktion zum Opfer", Stat.AnzGleichungenInterRed);
  IO_DruckeStatistikZeile("Gleichungen fielen der Backward-Subsumption zum Opfer", Stat.AnzErfSubsumptionenAlt);
#if KPIR_STATISTIK
  if (StatistikAn){
    int tiefe1 = IO_Platzverbrauch(Maximum(2,Stat. AnzKPIRReduziert, Stat.AnzKPIRSubsummiert));
     
     IO_DruckeKurzUeberschrift("Intermediate Reprocessing (IRP)", '+');
     IO_DruckeStatistikZeile("CPs touched (total)", Stat.AnzKPIRBehandelt);
     IO_DruckeStatistikZeileEingerueckt("of them subsumed", Stat.AnzKPIRSubsummiert, tiefe1);
     IO_DruckeStatistikZeileEingerueckt("of them simplified", Stat.AnzKPIRReduziert, tiefe1);
     IO_DruckeStatistikZeileEingerueckt("of them joined", Stat.AnzKPIRZusammengefuehrt, tiefe1+IO_Platzverbrauch(Stat.AnzKPIRZusammengefuehrt));
     IO_DruckeStatistikZeile("CPs touched only for reweighting", Stat.AnzKPnurW);
   }
#endif
   {
     volatile DomaenenT Domaene = Domaenenliste;
     while (Domaene) {
       IO_DruckeKurzUeberschrift(Domaene->Name, '+');
       IO_DruckeStatistikZeile(Domaene->Domaenenart==Strahldomaene ? "Mapped memory in bytes (beams)"
                               : "Mapped memory in bytes (managers)",(unsigned long int)((char*)Domaene->Ende-(char*)Domaene->Start));
       if_sst(
              IO_DruckeStatistikZeile("Domain expansions",Domaene->AnzahlDomaenenausdehnungen);
              IO_DruckeStatistikLeerZeile();
              IO_DruckeStatistikZeile("Mapped pages", Domaene->AnzahlGemapptePages);
              IO_DruckeStatistikZeile("Unmapped pages", Domaene->AnzahlGeunmapptePages);
              IO_DruckeStatistikZeile("Max. mapped pages", Domaene->AnzahlMaxGemapptePages);
              IO_DruckeStatistikLeerZeile();
              IO_DruckeStatistikZeile("Mapped memory", Domaene->AnzahlGemapptePages * DM_Seitengroesse());
              IO_DruckeStatistikZeile("Unmapped memory", Domaene->AnzahlGeunmapptePages * DM_Seitengroesse());
              IO_DruckeStatistikZeile("Max. mapped memory", Domaene->AnzahlMaxGemapptePages * DM_Seitengroesse());
              switch (Domaene->Domaenenart) {
              case Strahldomaene:
                IO_DruckeStatistikZeile("Frozen beams", Domaene->AnzahlGefroreneStrahlen);
                IO_DruckeStatistikZeile("Removed beams", Domaene->AnzahlGeloeschteStrahlen);
                break;
              case Managerdomaene:
              default:
                break;
              }
              );
       Domaene = Domaene->Nachf;
     }
   }
   {
     int tiefe1;
     /******** Regelmenge ****************************** */
     if (RE_AnzRegelnGesamt() > 0){
       unsigned int KnotenImBaum, GesparteKnoten;
       BO_KnotenZaehlen(&KnotenImBaum, &GesparteKnoten, RE_Regelbaum);
       IO_DruckeKurzUeberschrift("Set of Rules",'+');
       tiefe1 = IO_Einruecktiefe(0, Stat.AnzRegelnAktuell);
       IO_DruckeStatistikZeile("Total Number of Rules",Stat.RE_AnzRegelnGesamt);
       IO_DruckeStatistikZeileEingerueckt("of them still in R", Stat.AnzRegelnAktuell, tiefe1);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Nodes in R-tree", KnotenImBaum);
       IO_DruckeStatistikZeile("Nodes saved", GesparteKnoten);
       IO_DruckeStatistikZeileAnteil("Percentage of saved nodes (compared with SDT)", 
                                     GesparteKnoten, GesparteKnoten+KnotenImBaum);
     }
     /******** Gleichungsmenge ************************* */
     if (RE_AnzGleichungenGesamt() > 0){
       unsigned int KnotenImBaum, GesparteKnoten;
       BO_KnotenZaehlen(&KnotenImBaum, &GesparteKnoten, RE_Gleichungsbaum);
       IO_DruckeKurzUeberschrift("Set of Equations",'+');
       IO_DruckeStatistikZeile("Total Number of Equations",Stat.RE_AnzGleichungenGesamt);
       tiefe1 = IO_Einruecktiefe(0, Maximum(2, Stat.AnzGleichungenAktuell, Stat.AnzMonogleichungen));
       IO_DruckeStatistikZeileEingerueckt("of them still in E", Stat.AnzGleichungenAktuell, tiefe1);
       IO_DruckeStatistikZeileEingerueckt("of them Mono-Equations", Stat.AnzMonogleichungen, tiefe1);
       IO_DruckeStatistikZeileEingerueckt("of them still in E", Stat.AnzAktMonogleichungen, tiefe1+IO_Platzverbrauch(Stat.AnzAktMonogleichungen)+2);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Nodes in E-tree", KnotenImBaum);
       IO_DruckeStatistikZeile("Nodes saved", GesparteKnoten);
       IO_DruckeStatistikZeileAnteil("Percentage of saved nodes (compared with SDT)", 
                                     GesparteKnoten, GesparteKnoten+KnotenImBaum);
     }
   }
#if U1_STATISTIK
   {
     int tiefe1;
     if(WerteDa(2,Stat.AnzUnifVersRegelnToplevel, Stat.AnzUnifVersGleichgnToplevel)){
       IO_DruckeKurzUeberschrift("Unifikation", '+');
       tiefe1 = IO_Einruecktiefe(0, Maximum(6, Stat.AnzUnifErfRegelnToplevel, Stat.AnzUnifErfRegelnTeilterme,
                                            Stat.AnzUnifErfRegelnEigeneTterme, Stat.AnzUnifErfGleichgnToplevel,
                                            Stat.AnzUnifErfGleichgnTeilterme, Stat.AnzUnifErfGleichgnEigeneTterme));
       IO_DruckeStatistikZeile("Kapur-Opfer", Stat.AnzKPOpferKapur);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Unterdrueckte Nicht-Eins-Stern-Paare", Stat.AnzNonEinsSternKPs);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Unterdrueckte Nusfu-Paare", Stat.AnzNonNusfuKPs);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Unifikationsversuche mit der Regelmenge Toplevel",Stat.AnzUnifVersRegelnToplevel);
       IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",Stat.AnzUnifErfRegelnToplevel, tiefe1);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Unifikationsversuche mit Teiltermen der Regelmenge",Stat.AnzUnifVersRegelnTeilterme);
       IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",Stat.AnzUnifErfRegelnTeilterme, tiefe1);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Unifikationsversuche mit eigenen Teiltermen (RM)",Stat.AnzUnifVersRegelnEigeneTterme);
       IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",Stat.AnzUnifErfRegelnEigeneTterme, tiefe1);
       if (RE_AnzGleichungenGesamt > 0){
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Unifikationsversuche mit der Gleichungsmenge Toplevel",Stat.AnzUnifVersGleichgnToplevel);
         IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich"Stat.,AnzUnifErfGleichgnToplevel, tiefe1);
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Unifikationsversuche mit Teiltermen der Gleichungsmenge",Stat.AnzUnifVersGleichgnTeilterme);
         IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",Stat.AnzUnifErfGleichgnTeilterme, tiefe1);
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Unifikationsversuche mit eigenen Teiltermen (GM)",Stat.AnzUnifVersGleichgnEigeneTterme);
         IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",Stat.AnzUnifErfGleichgnEigeneTterme, tiefe1);
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Ueberlappungen erzeugt", Stat.AnzUnifUeberlappungenGebildet);
         IO_DruckeStatistikZeile("Wegen >-Test fehlgeschlagene KP-Bildungen Regel-Gleichung", Stat.AnzUnifMisserfolgeOrdnungRG);
         IO_DruckeStatistikZeile("Wegen >-Test fehlgeschlagene KP-Bildungen Gleichung-Gleichung",Stat. AnzUnifMisserfolgeOrdnungGG);
         IO_DruckeStatistikZeile("Dabei ueberfluessigerweise gebildete kritische Terme", Stat.AnzUnifUeberflKritTerme);
       }
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Insgesamt erzeugte KPs", 
                               Stat.AnzUnifErfRegelnToplevel+Stat.AnzUnifErfRegelnTeilterme+
                               Stat.AnzUnifErfRegelnEigeneTterme+Stat.AnzUnifErfGleichgnToplevel+
                               Stat.AnzUnifErfGleichgnTeilterme+Stat.AnzUnifErfGleichgnEigeneTterme);
       tiefe1 = IO_Einruecktiefe(0, Maximum(3, Stat.AnzKPRR, Stat.AnzKPGG,Stat.AnzKPRG));
       IO_DruckeStatistikZeileEingerueckt("davon enstanden zwischen 2 Regeln", Stat.AnzKPRR, tiefe1);
       IO_DruckeStatistikZeileEingerueckt("davon enstanden zwischen 2 Gleichungen", Stat.AnzKPGG, tiefe1);
       IO_DruckeStatistikZeileEingerueckt("davon enstanden zwischen Regel und Gleichung", Stat.AnzKPRG, tiefe1);
     }
   }
#endif
   if(WerteDa(2,Stat.AnzUnifVers,Stat.AnzUnifErf)) {
     int Tiefe1 = IO_Einruecktiefe(0, Maximum(1, Stat.AnzUnifErf));
     IO_DruckeKurzUeberschrift("Pattern-Unifikation", '+');
     IO_DruckeStatistikZeile("Unifikationsversuche",Stat.AnzUnifVers);
     IO_DruckeStatistikZeileEingerueckt("dabei erfolgreich",Stat.AnzUnifErf, Tiefe1);
   }

#if NFB_STATISTIK
   {
     if (WerteDa(2, Stat.AnzRedVersMitRegeln, Stat.AnzRedVersMitGleichgn)){
       IO_DruckeKurzUeberschrift("rewriting",'+');
       IO_DruckeStatistikZeile("match queries R",Stat.AnzRedVersMitRegeln);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Reductions with R",Stat.AnzRedErfRegeln);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("match queries E",Stat.AnzRedVersMitGleichgn);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("Reductions with E",Stat.AnzRedErfGleichgn);
       IO_DruckeStatistikLeerZeile();
       IO_DruckeStatistikZeile("total match queries", Stat.AnzRedVersMitRegeln + Stat.AnzRedVersMitGleichgn);
       IO_DruckeStatistikZeile("total reductions", Stat.AnzRedErfRegeln + Stat.AnzRedErfGleichgn);
     }
   }
#else
   IO_DruckeStatistikZeile("rewriting-statistics OFF", 0);
#endif

   if(WerteDa(1,Stat.AnzRedMisserfolgeOrdnung)) {
     IO_DruckeKurzUeberschrift("MatchOperationen", '+');
     IO_DruckeStatistikZeile("An >-Test gescheiterte Reduktionsversuche mit GM", Stat.AnzRedMisserfolgeOrdnung);
   }
   if (Stat.AnzEingefRegeln>0){
     IO_DruckeKurzUeberschrift("Operationen auf Regellisten (insgesamt, dh. R & E)", '+');
     IO_DruckeKurzUeberschrift("Operationen auf digitalen Suchbaeumen (insgesamt, d.h. R & E)",'+');
     IO_DruckeStatistikZeile("Einfuege-Operationen",Stat.AnzEingefRegeln);
     IO_DruckeStatistikZeile("Loesch-Operationen",Stat.AnzEntfernterRegeln);
     IO_DruckeStatistikZeile("Aufspaltungen von Termsuffici aus Blaettern",Stat.AnzSchrumpfPfadExp);
     IO_DruckeStatistikZeile("Zusammenziehungen von linearen Unterbaeumen in Blaetter",Stat.AnzPfadschrumpfungen);
   }
   IO_DruckeKurzUeberschrift("Treatment of Goals", '+');
   IO_DruckeStatistikZeile("Goals proven by rewriting", Stat.AnzErfZusfuehrungenZiele);
   if (Z_ZielModul == Z_MNF_Ziele)
     IO_DruckeStatistikZeile("Maximal size of elems in goal set",  Stat.MNFZ_MaxCountMNFSet);
#if MNF_STATISTIK   
   if (MNF_Initialised) {
     IO_DruckeKurzUeberschrift("MultipleNormalForms",'+');
     
     IO_DruckeStatistikZeile("Total Termpairs", Stat.MNF_AnzCalls);
     IO_DruckeStatistikZeile("Joinable Termpairs", Stat.MNF_AnzTermGleichOpp);
     IO_DruckeStatistikZeile("Immediatly joinable Termpairs", Stat.MNF_AnzDirektZus);
     IO_DruckeStatistikZeile("Reductions", Stat.MNF_AnzReduktionen);
     IO_DruckeStatistikZeile("Match queries with equations", Stat.MNF_AnzMatchVersucheE);
     IO_DruckeStatistikZeile("Match queries with rules", Stat.MNF_AnzMatchVersucheR);
     IO_DruckeStatistikZeile("Maximal Size of MNF-Set", Stat.MNF_MaxNachfolger);
     IO_DruckeStatistikZeile("Stops: MNF-set > maxSize", Stat.MNF_AbbruchMaxSize);
     IO_DruckeStatistikZeile("Total steps up with E", Stat.MNF_AnzEQUp);
     IO_DruckeStatistikZeile("Maximal steps up with E", Stat.MNF_MaxEQUp);
     IO_DruckeStatistikZeile("Total steps up with R", Stat.MNF_AnzAntiR);
   }
#endif
}

static void ST_Statistik_short(void)
{
  IO_DruckeKurzUeberschrift("Unselected Equations",'+');
  /*  IO_DruckeStatistikZeile("Generated", Stat.KPV_AnzKPGesamt);*/
  IO_DruckeStatistikZeile("Generated", Stat.KPV_AnzKPGesamt-Stat.KPV_AnzKPAusSpezifikation-Stat.KPV_AnzKPIROpfer-Stat.KPV_AnzKPDatenbank);
  IO_DruckeStatistikZeile("Selected", Stat.KPV_AnzKPBehandeltInVervollst);
  IO_DruckeStatistikZeile("Perm-Subsumed", Stat.KPV_AnzKPspaeterPermSubsummiert);
  IO_DruckeKurzUeberschrift("Sets of Rules and Equations",'+');
  IO_DruckeStatistikZeile("Total Number of Rules",Stat.RE_AnzRegelnGesamt);
  IO_DruckeStatistikZeile("Total Number of Equations",Stat.RE_AnzGleichungenGesamt);
#if NFB_STATISTIK
  IO_DruckeKurzUeberschrift("rewriting", '+');
  IO_DruckeStatistikZeile("Match Queries", Stat.AnzRedVersMitRegeln + Stat.AnzRedVersMitGleichgn);
  IO_DruckeStatistikZeile("Reductions", Stat.AnzRedErfRegeln + Stat.AnzRedErfGleichgn);
#endif
  IO_DruckeStatistikZeile("Forw. Matches R", Stat.VorwaertsMatchR);
  IO_DruckeStatistikZeile("Forw. Matches E", Stat.VorwaertsMatchE);
  IO_DruckeStatistikZeile("Forw. Matches S", Stat.VorwaertsMatchS);
  
  if (Z_ZielModul == Z_MNF_Ziele){
    IO_DruckeKurzUeberschrift("Treatment of Goals", '.');
    IO_DruckeStatistikZeile("Maximal size of elems in goal set",  Stat.MNFZ_MaxCountMNFSet);
  }
#if MNF_STATISTIK   
  if (MNF_Initialised) {
    IO_DruckeKurzUeberschrift("MultipleNormalForms",'+');
    IO_DruckeStatistikZeile("MNF Reductions", Stat.MNF_AnzReduktionen);
    IO_DruckeStatistikZeile("Total steps up (R+E)", Stat.MNF_AnzEQUp+Stat.MNF_AnzAntiR);
  }
#endif
}

static void ST_Statistik_slatex_head(void)
{
  fprintf(IO_Strom, "CPs gen    & selected   & perm-subs  & ");
  fprintf(IO_Strom, "Rules      & Equations  & ");
#if NFB_STATISTIK
  fprintf(IO_Strom, "Matches    & Reductions & ");
#endif
  fprintf(IO_Strom, "F.Match. R & F.Match. E & F.Match. S & ");
  
  if (Z_ZielModul == Z_MNF_Ziele){
    fprintf(IO_Strom, "G:MaxSize &");
  }
#if MNF_STATISTIK   
  if (MNF_Initialised) {
    fprintf(IO_Strom, "MNF: Red & Tot. Up & ");
  }
#endif
}

/* -------------------------------------------------------*/
/* -------------- Prozess-Statistik ----------------------*/
/* -------------------------------------------------------*/
#if 0
  d e f u n c t ;;;
#include "DomaenenMemorium.h"
extern  DomaenenT Domaene;
#endif

static void ST_PlatzverbrauchAusgeben(BOOLEAN Kopf)
     /* Ausgabe des Platzverbrauchs im Latex-Tabellenformat */
{
#if 0
  d e f u n c t ;;;
  if (Kopf) fprintf(IO_Strom, "mem Heap & ");
  else      IO_DruckeStatistikZeile("mem Heap", 
                   (Domaene == NULL) ? 0 : Domaene->AnzahlGemapptePages * getpagesize());
#endif
#if defined (SUNOS4) /* || defined (LINUX) */
  if (Kopf) fprintf(IO_Strom, "max.res.mem & ");
  else{
    struct rusage rusageBuf;
    
    if (getrusage(RUSAGE_SELF, &rusageBuf) == -1) fprintf(IO_Strom, " <error> &");
    else IO_DruckeStatistikZeile("bla",rusageBuf.ru_maxrss * getpagesize());
  }
#elif defined (LINUX)

  if (Kopf) {
    fprintf(IO_Strom, "vsize & rss.mem &");
  } else {
    FILE *statFd;
    unsigned long vsize, rss, pagesize;

    statFd = fopen("/proc/self/statm", "r");
    if (statFd != NULL) {
      fscanf(statFd, "%ld %ld", &vsize, &rss); /* XXX nochmals ueberpruefen ! */
      fclose(statFd);
      pagesize = sysconf(_SC_PAGESIZE);
      vsize = vsize * pagesize;
      rss = rss * pagesize;
      if (0){
	fprintf(IO_Strom, " %lu & %lu &", vsize, rss);
      }
      else {
	vsize = vsize / 1024;
	rss = rss / 1024;
	if (vsize > 10240){
	  fprintf(IO_Strom, " %.1fM & %.1fM &", vsize/1024.0, rss/1024.0);
	}
	else {
	  fprintf(IO_Strom, " %luk & %luk &", vsize, rss);
	}
      }
    } else {
      fprintf(IO_Strom, " <error> & <error> &");
    }
  }  

#elif defined (SUNOS5)
  if (Kopf) fprintf(IO_Strom, "proc.size& res.mem &");
  else{
    static    int             fd;
    static    BOOLEAN         status;
    prusage_t       prusage;
    prpsinfo_t      prpsinfo;
    
    if (status == 0){
      if ((fd = open("/proc/self",O_RDONLY)) == -1){
        status = 2;
      }
      else {
        status = 1;
      }
    }
    if ((status == 2) ||
        (ioctl(fd, PIOCUSAGE, &prusage) == -1) ||
        (ioctl(fd, PIOCPSINFO, &prpsinfo) == -1))
      fprintf(IO_Strom, " <error> & <error> &");
    else
      fprintf(IO_Strom, " %ld & %ld &", prpsinfo.pr_size*sysconf(_SC_PAGESIZE),
              prpsinfo.pr_rssize *sysconf(_SC_PAGESIZE));
  }
#else
#error "ST_PlatzverbrauchAusgeben muss angepasst werden!"      
#endif
}

void ST_ProzessStatistik(void)
/* Ausgabe von Prozess-Daten (aus getrusage)  */
{

#if defined (SUNOS4) || defined (LINUX)
   struct rusage rusageBuf;
   
   if (getrusage(RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("IO_ProzessStatistik: getrusage fehlgeschlagen");
   
   IO_DruckeUeberschrift("Prozess-Statistik");
   IO_DruckeStatistikZeile("maximum resident set size/pages", (unsigned long int)rusageBuf.ru_maxrss);
   IO_DruckeStatistikZeile("system pagesize (may differ from HW-pagesize)",(unsigned long int) getpagesize());
   IO_DruckeStatistikZeile("==> maximum resident amount of memory (bytes)", rusageBuf.ru_maxrss * getpagesize());
   IO_DruckeStatistikZeile("memory in use by the process while running (integral value: pages * clock ticks)", (unsigned long int) rusageBuf.ru_idrss);
   IO_DruckeStatistikZeile("page faults serviced without any physical I/O activity", (unsigned long int) rusageBuf.ru_minflt);
   IO_DruckeStatistikZeile("page faults serviced with physical I/O activity", (unsigned long int) rusageBuf.ru_majflt);
   IO_DruckeStatistikZeile("times process was swapped out of main memory", (unsigned long int) rusageBuf.ru_nswap);
   IO_DruckeStatistikZeile("file system inputs (read)", (unsigned long int) rusageBuf.ru_inblock);
   IO_DruckeStatistikZeile("file system outputs (write)", (unsigned long int) rusageBuf.ru_oublock);
   IO_DruckeStatistikZeile("messages sent over sockets", (unsigned long int) rusageBuf.ru_msgsnd);
   IO_DruckeStatistikZeile("messages received from sockets", (unsigned long int) rusageBuf.ru_msgrcv);
   IO_DruckeStatistikZeile("signals delivered", (unsigned long int) rusageBuf.ru_nsignals);
   IO_DruckeStatistikZeile("context switches due to voluntarily giving up the processor before time slice completed", (unsigned long int) rusageBuf.ru_nvcsw);
   IO_DruckeStatistikZeile("context switches due to a higher priority process or exceeded time slice", (unsigned long int) rusageBuf.ru_nivcsw);
#elif defined (SUNOS5)
   int             fd;
   char            proc[40];
   prusage_t       prusage;
   prpsinfo_t      prpsinfo;
   
   sprintf(proc,"/proc/%d", (int)getpid());
   if ((fd = open(proc,O_RDONLY)) == -1) {
      our_assertion_failed ("IO_ProzessStatistik: open /proc fehlgeschlagen");
      }
   if (ioctl(fd, PIOCUSAGE, &prusage) == -1) {
      our_assertion_failed ("IO_ProzessStatistik: open prusage-ioctl fehlgeschlagen");
   }
   if (ioctl(fd, PIOCPSINFO, &prpsinfo) == -1) {
      our_assertion_failed ("IO_ProzessStatistik: open prpsinfo-ioctl fehlgeschlagen");
   }
   IO_DruckeUeberschrift("Prozess-Statistik");
   IO_DruckeStatistikZeile("system pagesize (may differ from HW-pagesize)",(unsigned long int) sysconf(_SC_PAGESIZE));
   IO_DruckeStatistikZeile("size of process /pages", (unsigned long int)prpsinfo.pr_size);
   IO_DruckeStatistikZeile("==> process size (bytes)", prpsinfo.pr_size*sysconf(_SC_PAGESIZE) );
   IO_DruckeStatistikZeile("resident set size/pages", (unsigned long int)prpsinfo.pr_rssize);
   IO_DruckeStatistikZeile("==> resident amount of memory (bytes)", prpsinfo.pr_rssize *sysconf(_SC_PAGESIZE) );
   IO_DruckeStatistikZeile("page faults serviced without any physical I/O activity", (unsigned long int) prusage.pr_minf);
   IO_DruckeStatistikZeile("page faults serviced with physical I/O activity", (unsigned long int) prusage.pr_majf);
   IO_DruckeStatistikZeile("times process was swapped out of main memory", (unsigned long int) prusage.pr_nswap);
   IO_DruckeStatistikZeile("file system inputs (read)", (unsigned long int) prusage.pr_inblk);
   IO_DruckeStatistikZeile("file system outputs (write)", (unsigned long int) prusage.pr_oublk);
   IO_DruckeStatistikZeile("messages sent over sockets", (unsigned long int) prusage.pr_msnd);
   IO_DruckeStatistikZeile("messages received from sockets", (unsigned long int) prusage.pr_mrcv);
   IO_DruckeStatistikZeile("signals delivered", (unsigned long int) prusage.pr_sigs);
   IO_DruckeStatistikZeile("context switches due to voluntarily giving up the processor before time slice completed", (unsigned long int) prusage.pr_vctx);
   IO_DruckeStatistikZeile("context switches due to a higher priority process or exceeded time slice", (unsigned long int) prusage.pr_ictx);
#else
#error "ST_ProzessStatistik muss angepasst werden!"      
#endif
}

typedef enum {Finished, Timeout, Error, Interrupt, NoMemory, GoalProved} ReasonT;


static void STEXZeileAusgeben(ReasonT Situation)
{
  fprintf(IO_Strom, "\n");
  IO_StatistikModusAufLatexSetzen();
  fprintf(IO_Strom, " %-10s & ", "problem");
  ST_Statistik_slatex_head();
  ST_PlatzverbrauchAusgeben(TRUE);
  fprintf(IO_Strom, " time & Result\n ");
  IO_DruckeBeispiel();
  ST_Statistik_short();
  ST_PlatzverbrauchAusgeben(FALSE);
  ZM_Statistik("ZM:latex");
  switch (Situation){
  case Finished:
    if (Z_ZielGesamtStatus == Z_AlleWiederlegt || Z_ZielGesamtStatus == Z_EinigeWiederlegt)
      fprintf(IO_Strom, " & R ");
    else
      fprintf(IO_Strom, " & S ");
    break;
  case Timeout:
    fprintf(IO_Strom, " & T ");
    break;
  case Error:
    fprintf(IO_Strom, " & E ");
    break;
  case NoMemory:
    fprintf(IO_Strom, " & M ");
    break;
  case Interrupt:
    fprintf(IO_Strom, " & I ");
    break;
  case GoalProved:
    fprintf(IO_Strom, " & G ");
    break;
  default:
    break;
  }
  fprintf(IO_Strom, "\n");
  IO_StatistikModusAufNormalSetzen();
}

/* -------------------------------------------------------*/
/* -------------- Menue ----------------------------------*/
/* -------------------------------------------------------*/


static void Menue(ReasonT Situation, const char* Ueberschrift, const char *Bemerkung, BOOLEAN Resultat)
/* Ausgabe von Statistiken und Ergebnissen, auch als Interrupt-Menue. */
{
   BOOLEAN MenueWiederholen = TRUE;
   char    s[INPUT_STRLEN];
   char    AusgabeDatei[INPUT_STRLEN];
   char    eing[INPUT_STRLEN];
   BOOLEAN AusgabeDateiAngegeben = FALSE;
   BOOLEAN DruckeMenue = TRUE;
   
   if (Situation ==Interrupt){ 
      ZM_StoppGesamtZeit();
      SI_InterruptflagLoeschen();  /* Zuruecksetzen des Flags */
   }
   do{
     if (DruckeMenue) {
       
      IO_DruckeUeberschrift(Ueberschrift);
      IO_DruckeBeispiel();
      if (Bemerkung) fprintf(IO_Strom, "\n***> %s", Bemerkung);
      fprintf(IO_Strom, "\n\n");

      fprintf(IO_Strom, "* How to proceed\n");
      fprintf(IO_Strom, "       q - quit\n");
      if (Situation != Error && Situation != NoMemory)
        fprintf(IO_Strom, "       c - continue\n"); /* nach Fehler gibt's kein continue */
      if (Situation == Interrupt) fprintf(IO_Strom, "      ol - output-level (0,...,4)\n");

      fprintf(IO_Strom, "\n* Statistics\n");
      fprintf(IO_Strom, "       s - statistics on run\n");
      fprintf(IO_Strom, "      sh - short statistics on run\n");
      fprintf(IO_Strom, "       p - statistics on process\n");
      fprintf(IO_Strom, "    stex - short statistics in LaTeX-format\n");

      fprintf(IO_Strom, "\n* Input-Data & Environment\n");
      fprintf(IO_Strom, "     par - parameterization\n");
      fprintf(IO_Strom, "   setup - Waldmeister set-up\n");
      fprintf(IO_Strom, "    view - specification\n");
      fprintf(IO_Strom, "    conf - configuration\n");


      if ((Situation == Finished) || (Situation == Interrupt)){ /* Bei Timeout und Error keine Konsistenz-Garantie */
        fprintf(IO_Strom, "\n* State of Computation\n");
        fprintf(IO_Strom, "       r - current set of rules\n");
        fprintf(IO_Strom, "       e - current set of equations\n");
        fprintf(IO_Strom, "       g - current set of goals\n");
        fprintf(IO_Strom, "     sue - current set of ues --- beware of data overflow\n");
        if ( (Situation == Finished) &&((PA_ExtModus == convergence) || (PA_ExtModus == termination) || (PA_ExtModus == confluence)))
          fprintf(IO_Strom, "      or - other results (Sonstige Ergebnisse)\n");
        fprintf(IO_Strom, "    spec - save current system as new specification\n");
      }

      fprintf(IO_Strom, "\n* Redirect outputs from this menue\n");
      fprintf(IO_Strom, "     sfo - following outputs into file (start file output)\n");
      fprintf(IO_Strom, "     ffo - finish file output (finish file output)\n");


      if ((Situation ==Interrupt) || (SPArrayPos>0))
	fprintf(IO_Strom, "\n* Action\n");
      if (Situation ==Interrupt)
	fprintf(IO_Strom, "     irp - intermediate reprocessing of SUE --- order coffee first\n");
      if (Situation ==Interrupt)
	fprintf(IO_Strom, "   cgirp - intermediate reprocessing of critical goals \n");
      if (SPArrayPos>0)
	fprintf(IO_Strom, "      ip - IRP-points\n");
        
      fprintf(IO_Strom, "\n* Check consistency\n");
      if (Situation != Timeout){ /* Bei Timeout keine Konsistenz-Garantie */
	fprintf(IO_Strom, "      ic - check WDTs for consistency\n");
#if TV_TEST
	fprintf(IO_Strom, "      tc - check termcells for consistency\n");
#endif
	fprintf(IO_Strom, "      ms - memory statistics\n");
      }
      else
	fprintf(IO_Strom, "         - No consistency-checking because of Timeout\n");
#if COMP_TW        
      fprintf(IO_Strom, "     ref - output referee results\n");
#endif
     }
     if (!DruckeMenue)
       fprintf(IO_Strom, "\n('h' for topic-list) ");
     DruckeMenue = FALSE;
     fprintf(IO_Strom, ">>> ");
     fgets(s, INPUT_STRLEN, stdin);
     chop (s);
      
      fprintf(IO_Strom, "\n");

      if (strcmp(s, "q") == 0)
        exit(0);
      else if (strcmp(s, "h") == 0){
        DruckeMenue = TRUE;
      }
      else if (strcmp(s, "s") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         IO_DruckeUeberschrift("statistics");
         ST_Statistik_standard();
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "sh") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         IO_DruckeUeberschrift("short statistics");
         ST_Statistik_short();
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else IO_WarteAufReturnMitText("RETURN");
      }
      else if ((strcmp(s, "or") == 0) && (Situation == Finished) && 
               ((PA_ExtModus == convergence) || (PA_ExtModus == termination) || 
                (PA_ExtModus == confluence))){
        if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
        RE_SonstigeErgebnisse(Resultat);
        if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
        else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "sfo") == 0){
         fprintf(IO_Strom, "\nName of file: ");
         fgets(AusgabeDatei, INPUT_STRLEN, stdin);
         chop(AusgabeDatei);
         AusgabeDateiAngegeben = TRUE;
      }
      else if (strcmp(s, "ffo") == 0){
         AusgabeDateiAngegeben = FALSE;
      }
      else if (strcmp(s, "par") == 0){
         PA_DefaultsAusgeben(AusgabeDateiAngegeben?AusgabeDatei:"stdout");
         IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "p") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         ST_ProzessStatistik();
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else  IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "setup") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         IO_DruckeUeberschrift("Waldmeister set-up");
         PA_DruckeModi();
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "view") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         IO_DruckeUeberschrift("specification");
         IO_DateiAusgabe(IO_Spezifikationsname(), TRUE);
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
      }
      else if (strcmp(s, "r") == 0){
         if (RE_RegelmengeLeer()){
           fprintf(IO_Strom, "\n\nset of rules is empty.\n\n");
           IO_WarteAufReturnMitText("RETURN");
         }
         else{
            int wert = 1;
#if BA_AUSGABEN
            do{
               fprintf(IO_Strom, "output R:\n");
               fprintf(IO_Strom, " 1 - textual\n");
               fprintf(IO_Strom, " 2 - text-tree\n");
               fprintf(IO_Strom, " 3 - graphics-tree (emacs)\n");
	       fprintf(IO_Strom, " 4 - textual tree-dump\n");
	       fprintf(IO_Strom, " 5 - graphical tree-dump\n");
               fprintf(IO_Strom, "\n>>> ");
               fgets(eing, INPUT_STRLEN, stdin);
               wert = atoi(eing);
            }while((wert<1)||(wert>5));
#endif
            if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
            RE_AusgabeRegelmenge(wert);
            if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
            else IO_WarteAufReturnMitText("RETURN");
         }
      }
      else if (strcmp(s, "e") == 0){
         if (RE_GleichungsmengeLeer()){
           fprintf(IO_Strom, "\n\nset of equations is empty\n\n");
           IO_WarteAufReturnMitText("RETURN");
         }
         else{
#if BA_AUSGABEN
            int wert = 1;
            do{
               fprintf(IO_Strom, "output E:\n");
               fprintf(IO_Strom, " 1 - textual\n");
               fprintf(IO_Strom, " 2 - text-tree\n");
               fprintf(IO_Strom, " 3 - graphics-tree (emacs)\n");
	       fprintf(IO_Strom, " 4 - textual tree-dump\n");
	       fprintf(IO_Strom, " 5 - graphical tree-dump\n");
               fprintf(IO_Strom, "\n>>> ");
               fgets(eing, INPUT_STRLEN, stdin);
               wert = atoi(eing);
            }while((wert<1)||(wert>5));
            if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
            RE_AusgabeGleichungsmenge(wert);
            if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
            else IO_WarteAufReturnMitText("RETURN");
#endif
         }
      }
      else if (strcmp(s, "g") == 0){
            int wert;
            do{
               fprintf(IO_Strom, "output goals:\n");
               fprintf(IO_Strom, "1 - short\n");
               fprintf(IO_Strom, "2 - long\n\n");
               fprintf(IO_Strom, ">>> ");
               fgets(eing, INPUT_STRLEN, stdin);
               wert = atoi (eing);
            } while((wert<1) &&(wert>2));
            if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
            switch (wert) {
            case 1:
               Z_AusgabeZielmenge(Z_AusgabeInitial);
               break;
            case 2:
            default:
               Z_AusgabeZielmenge(Z_AusgabeAlles);
            }
            if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
            else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "ol") == 0){
         int wert;
         do{
            fprintf(IO_Strom, "current output level: %d\n", IO_AusgabeLevel());
            fprintf(IO_Strom, "new level >>> ");
            fgets(eing, INPUT_STRLEN, stdin);
            wert = atoi(eing);
            fprintf(IO_Strom, "\n");
         }while((wert<0) || (wert > 4));
         IO_AusgabeLevelBesetzen(wert);
      }
      else if (strcmp(s, "c") == 0)
        MenueWiederholen = FALSE;
      else if (strcmp(s, "sue") == 0){
         if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
         KPV_Dump();
         if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
         else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "ms") == 0){
	int wert;
	do{
	  fprintf(IO_Strom, "memory statistics\n");
	  fprintf(IO_Strom, "1 - discrepancies only\n");
	  fprintf(IO_Strom, "2 - complete\n\n");
	  fprintf(IO_Strom, ">>> ");
	  fgets(eing, INPUT_STRLEN, stdin);
	  wert = atoi(eing);
	}while((wert<1) &&(wert>2));
	if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
	SV_Statistik((Situation == Error || Situation == NoMemory), (wert==1), "complete");
	if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
	else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "i") == 0){
	if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
	ST_InterneStatistik();
	if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
	else IO_WarteAufReturnMitText("RETURN");
      }
      else if (TV_TEST && strcmp(s, "tc") == 0){
	if ((Situation == Timeout) || (Situation == Error) || Situation == NoMemory)   /* Fehler-Situation */
	  fprintf(IO_Strom, "\nerror was encountered ==> test not possible.\n");
	else{
	  if (Situation == Interrupt){
	    TV_TermzellenVerfolger(3);
	  }
	  else
	    TV_TermzellenVerfolger(3);
	}
	IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "ic") == 0){
	if ((Situation == Timeout) || (Situation == Error) || Situation == NoMemory)   /* Fehler-Situation */
	  fprintf(IO_Strom, "\nerror was encountered ==> test not possible.\n");
	else
	  RE_RUndETest();
	IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "irp") == 0){
	printf("specify >>> ");
	fgets(eing, INPUT_STRLEN, stdin);
	chop(eing);
	if ((SPArrayPos>0)&&(SPArray[SPArrayPos-1].sp == RE_AnzRegelnGesamt() + RE_AnzGleichungenGesamt()))
	  fprintf(IO_Strom, "SUE already reprocessed!\n");
	else{
	  ZM_StartGesamtZeit(); /*Zeitmessung soll konsistent bleiben (sonst: T(NFBildung) > T(Gesamt) */
	  KPV_WahlfreiInterreduzieren(eing);
	  ZM_StoppGesamtZeit();
	  fprintf(IO_Strom, "\nMoment for reproduction of run is %ld.\n", RE_AnzRegelnGesamt() + RE_AnzGleichungenGesamt());
	  StichpunktMerken(RE_AnzRegelnGesamt() + RE_AnzGleichungenGesamt(), eing);
	}
	IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "cgirp") == 0) {
	printf("specify >>> ");
	fgets(eing, INPUT_STRLEN, stdin);
	chop(eing);
	ZM_StartGesamtZeit(); /*Zeitmessung soll konsistent bleiben (sonst: T(NFBildung) > T(Gesamt) */
	KPV_CGMengeInterreduzieren(TRUE, eing);
	ZM_StoppGesamtZeit();
	fprintf(IO_Strom, "\nMoment for reproduction of run is %ld.\n", RE_AnzRegelnGesamt() + RE_AnzGleichungenGesamt());
	IO_WarteAufReturnMitText("RETURN");
      }
      else if ((SPArrayPos>0) && (strcmp(s, "ip") == 0)){
	int i;
	if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
	fprintf(IO_Strom, "IRP-Stichpunkte are: ");
	for (i=0; i<SPArrayPos; i++){
	  fprintf(IO_Strom, "%ld %s", SPArray[i].sp, SPArray[i].strength);
	  if (i<SPArrayPos-1)
	    fprintf(IO_Strom, ":");
	}
	fprintf(IO_Strom, "\n\n");
	if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
	else IO_WarteAufReturnMitText("RETURN");
      }
      else if (strcmp(s, "conf") == 0){
	if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
	ST_DruckeKonfiguration();
          if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
          else IO_WarteAufReturnMitText("RETURN");
      }
      else if ((strcmp(s, "stex") == 0)){
	if (AusgabeDateiAngegeben) IO_AusgabeStromUmlenken(AusgabeDatei);
	STEXZeileAusgeben(Situation);
	if (AusgabeDateiAngegeben) IO_AusgabeStromZuruecksetzen();
      }
      else if (strcmp (s, "spec") == 0) {
           SpezifikationSchreiben();
      }
   }while (MenueWiederholen);
   
   if (Situation == Interrupt) ZM_StartGesamtZeit(); /* Interrupt-Modus => Zeit wieder anwerfen */

}

/* -------------------------------------------------------*/
/* -------------- Statistik ------------------------------*/
/* -------------------------------------------------------*/

static void AutoStatistik(ReasonT Situation, const char *Bemerkung);


/* -------------------------------------------------------*/
/* -------------- Statistik ------------------------------*/
/* -------------------------------------------------------*/


void ST_SuccessStatistik(const char *Bemerkung, BOOLEAN Resultat)
/* Bei eingeschalteter Interaktion: Interaktives Ergebnis-Menue.
   Sonst Ausgabe diverser Daten gemaess Aufruf.
*/
{
   if (PA_InteraktionAn())
      Menue(Finished, "RESULTS-STATISTICS", Bemerkung, Resultat);
   
   else
      AutoStatistik(Finished, Bemerkung);
}

#if COMP_POWER
void ST_PowerStatistik(const char *Bemerkung)
{
  DauerT Laufzeit = ZM_Laufzeit();
#if defined (SUNOS5)
  int fd;
  char proc[40];
  prusage_t prusage;
  prpsinfo_t prpsinfo;
  unsigned long int Prozessgroesse;
  sprintf(proc,"/proc/%d", (int)getpid());
  if ((fd = open(proc,O_RDONLY))==-1)
    our_assertion_failed ("IO_ProzessStatistik: open /proc fehlgeschlagen");
  if (ioctl(fd, PIOCUSAGE, &prusage)==-1)
    our_assertion_failed ("IO_ProzessStatistik: open prusage-ioctl fehlgeschlagen");
  if (ioctl(fd, PIOCPSINFO, &prpsinfo)==-1)
    our_assertion_failed ("IO_ProzessStatistik: open prpsinfo-ioctl fehlgeschlagen");
  Prozessgroesse = prpsinfo.pr_size*sysconf(_SC_PAGESIZE);
  Prozessgroesse = 1000*((Prozessgroesse+500)/1000);
#endif

  fprintf (IO_Strom, "\nWaldmeister states: %s.\n", Bemerkung);

  fprintf(IO_Strom,"\nStatistics:\n-----------\n");
  fprintf(IO_Strom,"Rules         %11ld\n", RE_AnzRegelnGesamt());
  fprintf(IO_Strom,"Equations     %11ld\n", RE_AnzGleichungenGesamt());
  fprintf(IO_Strom,"Critical pairs%11ld\n", GetKPV_AnzKPGesamt());
#if defined (SUNOS5)
  fprintf(IO_Strom,"Process size  %11ld\n", Prozessgroesse);
#endif
  fprintf(IO_Strom,"Search time [s]%9ld.%1ld\n",Laufzeit.Sekunden,Laufzeit.Zehntelsekunden);
}
#endif

void ST_ErrorStatistik(const char *Bemerkung, BOOLEAN Resultat)
{
   if (PA_InteraktionAn()){
      IO_RingBell();
      Menue(Error, "ERROR-MENUE", Bemerkung, Resultat);
   }
   else{
      AutoStatistik(Error, Bemerkung);
   }

   if (Forked && Vadder) {
     kill( son_pid, SIGVTALRM );
     waitpid( son_pid, NULL, 0 );
   }

}

void ST_NoMemory (const char *Bemerkung, BOOLEAN Resultat)
{
   if (PA_InteraktionAn()) {
      IO_RingBell();
      Menue(NoMemory, "ERROR-MENUE", Bemerkung, Resultat);
   }
   else {
      AutoStatistik(NoMemory, Bemerkung);
   }

   if (Forked && Vadder) {
     kill( son_pid, SIGVTALRM );
     waitpid( son_pid, NULL, 0 );
   }
}

void ST_TimeoutStatistik(const char *Bemerkung, BOOLEAN Resultat)
{
   if (PA_InteraktionAn())
      Menue(Timeout, "TIMEOUT-MENUE", Bemerkung, Resultat);
   
   else {
      AutoStatistik(Timeout, Bemerkung);
   }

   if (Forked && Vadder) {
     kill( son_pid, SIGVTALRM );
     waitpid( son_pid, NULL, 0 );
   }

}

static void
ZwischenStatistik (ReasonT Situation, const char *Bemerkung)
{
   ZM_StoppGesamtZeit();
   SI_Usr1InterruptflagLoeschen ();
   SI_InterruptflagLoeschen ();
   AutoStatistik(Situation, Bemerkung);
   fflush (NULL);  /* Alle Output-puffer leeren */
   ZM_StartGesamtZeit();   
}

void ST_SignalUsrStatistik(const char *Bemerkung)
{
  ZwischenStatistik (Interrupt, Bemerkung);
}

void ST_GoalProvedStatistik(const char *Bemerkung)
{
  ZwischenStatistik (GoalProved, Bemerkung);
}



/* -------------------------------------------------------*/
/* -------------- Interne Statistik ----------------------*/
/* -------------------------------------------------------*/
extern unsigned int GroesseSymbolInfo;
unsigned int U1_AusdehnungKopierstapel (void);
unsigned int U1_AusdehnungKopierstapel2 (void);
unsigned int U2_AusdehnungKopierstapel (void);
extern unsigned long DomainGran;
extern int xUDimension;
extern int xUDimension2;

void ST_InterneStatistik(void)
{
   IO_DruckeUeberschrift("Interne Statistik");

   IO_DruckeStatistikZeile("Max. Expansion of Leaf-depth-array in R-tree", RE_Regelbaum.TiefeBlaetterDimension);
   IO_DruckeStatistikZeile("Max. Expansion of Leaf-depth-array in E-tree", RE_Gleichungsbaum.TiefeBlaetterDimension);
   IO_DruckeStatistikZeile("Maximale Ausdehnung der Unifikationsvariablenbank", -xUDimension);
   IO_DruckeStatistikZeile("Maximale Ausdehnung des Unifikationsstacks", -xUDimension);
   IO_DruckeStatistikZeile("Maximale Ausdehnung der drei U1_Endestapel", -xUDimension);
   IO_DruckeStatistikZeile("Maximale Ausdehnung des U1_Kopierstapels", U1_AusdehnungKopierstapel());
   IO_DruckeStatistikZeile("Maximale Ausdehnung des U1_Kopierstapels2", U1_AusdehnungKopierstapel2());
   IO_DruckeStatistikZeile("Maximale Ausdehnung der Unifikationsvariablenbank", -xUDimension);
   IO_DruckeStatistikZeile("Konstante Ausdehnung der Bank fuer Patternvariablen", SO_AnzahlFunktionsvariablen());
   IO_DruckeStatistikZeile("Maximale Ausdehnung des Unifikationsstacks", -xUDimension);
   IO_DruckeStatistikZeile("Maximale Ausdehnung der drei U2_Endestapel", -xUDimension);
   IO_DruckeStatistikZeile("Maximale Ausdehnung des U2_Kopierstapels", U2_AusdehnungKopierstapel());
   IO_DruckeStatistikZeile("Maximale Ausdehnung der Reduktionsvariablenbank",MO_AusdehnungVariablenbank());
   IO_DruckeStatistikZeile("Maximale Ausdehnung des MO_Kopierstapels", MO_AusdehnungKopierstapel());
   IO_DruckeStatistikZeile("Groesste aufgetretene Tiefe einer linken Seite", BO_MaximaleTiefe);
   IO_DruckeStatistikZeile("Maximale Ausdehnung der BO_Umbenennungsvariablenbank",BO_AusdehnungVariablenbank());
   IO_DruckeStatistikZeile("Maximale Ausdehnung des BO_Sprungstapels",BO_AusdehnungSprungstapel());
   IO_DruckeStatistikZeile("Maximale Ausdehnung des BO_Kopierstapels", BO_AusdehnungKopierstapel());
   IO_DruckeStatistikZeile("Max. Expansion of SO_SymbolInfo", GroesseSymbolInfo);
   IO_DruckeStatistikZeile("BK-successor-arrays allocated via malloc", BK_AnzEinzelnAllokierterNachfolgerArrays);
#if MNF_STATISTIK
   if (MNF_Initialised) {
     IO_DruckeStatistikZeile("Calls of HA_AddData", Stat.MNF_AnzCallsHAAddData);
     IO_DruckeStatistikZeile("Count MNF: same Terms", Stat.MNF_AnzTermGleich);
     IO_DruckeStatistikZeile("Count MNF: same Terms, same colour", Stat.MNF_AnzTermGleichCol);
     IO_DruckeStatistikZeile("Count MNF: same Terms, opp. colour", Stat.MNF_AnzTermGleichOpp);
     if (Stat.MNF_AnzCalls > 0)
       IO_DruckeStatistikZeile("Average Size of MNF-Set", Stat.MNF_SumAnzNachfolger/Stat.MNF_AnzCalls);
   }
#endif
#if HA_STATISTIK
   IO_DruckeStatistikZeile("Hash Collisions", Stat.HA_AnzBucketGleich);
   IO_DruckeStatistikZeile("keys1 equal (Hash)", Stat.HA_AnzKey1Gleich);
   IO_DruckeStatistikZeile("keys2 equal (Hash)", Stat.HA_AnzKey2Gleich);
   IO_DruckeStatistikZeile("Maximal Size of Hash", Stat.HA_MaxHashSize);
   IO_DruckeStatistikZeile("Maximal no. of elems in a bucket", Stat.HA_MaxElemsInBucket);
#endif
   
   IO_DruckeStatistikZeile("Maximale Ausdehnung des TO_Kopierstapels", TO_AusdehnungKopierstapel());
   IO_DruckeStatistikZeile("Maximale Ausdehnung des IO_Kopierstapels", IO_AusdehnungAusgabestapel());
}

/* -------------------------------------------------------*/
/* -------------- Konfiguration --------------------------*/
/* -------------------------------------------------------*/

void ST_DruckeKonfiguration(void)
/* Ausgabe der durch die defines in diesem Modul bestimmten Konfiguration. */
{
  struct utsname  utsbuf;

  IO_DruckeUeberschrift("Konfiguration");

  fprintf(IO_Strom, "%s", configuration_banner);
  
  IO_DruckeZeile('+');
  uname(&utsbuf);
  fprintf(IO_Strom, "Run executed on %s (%s, %s %s)\n", 
          utsbuf.nodename, utsbuf.machine, 
          utsbuf.sysname, utsbuf.release);
  IO_DruckeZeile('+');
  
#if IO_AUSGABEN
  fprintf(IO_Strom, "Ausgaben waehrend des Laufes AN, ");
#else
  fprintf(IO_Strom, "Ausgaben waehrend des Laufes AUS, ");
#endif
#if COMP_OPTIMIERUNG
  fprintf(IO_Strom, "Optimized Compilation.\n");
#endif
#if !COMP_OPTIMIERUNG
  fprintf(IO_Strom, "Non-Optimized Compilation.\n");
#endif
}

/* -------------------------------------------------------*/
/* -------------- Ergebnis-Menue -------------------------*/
/* -------------------------------------------------------*/

void ST_IRMenue(const char *where)
/* Ausgabe von Statistiken und Ergebnissen, auch als Interrupt-Menue. */
{
   char str[200] = "Run interrupted by user in ";
   
   if (SI_Usr1InterruptflagGesetzt ()) {
      ST_SignalUsrStatistik ("Received signal SIGUSR1");
   }
   else {
      strcat (str, where);
      Menue(Interrupt, "INTERRUPT-MENUE", str, FALSE);
   }
}

static void AutoStatistik(ReasonT Situation, const char *Bemerkung)
{
   STEXZeileAusgeben(Situation);
   
#if COMP_POWER
   switch (Situation){
   case Timeout:
     ST_PowerStatistik ("Run out of time");
     break;
   case NoMemory:
     ST_PowerStatistik ("Run out of memory");
     break;
   case Error:
     fprintf(IO_Strom, "Error: %s\n", Bemerkung);
     break;
   case Finished: 
   default: 
     break;
   }
#else
   switch (Situation){
   case Finished: 
     if (Bemerkung != NULL){
       fprintf(IO_Strom, "\n\nWaldmeister states: %s.\n", Bemerkung);
     }
     break;
   case Timeout: 
     fprintf(IO_Strom, "\n\n TIMEOUT: %s\n", Bemerkung);
     break;
   case Error:
   case NoMemory:
     fprintf(IO_Strom, "\n\n ERROR: %s \n", Bemerkung);
     break;
   default: 
     break;
   }
#endif
}


