/*
01.03.94 Angelegt.
02.03.94 Weitergemacht. Optionen eingebaut. Arnim.
04.03.94 Aufruf mit symbolischen Dateinamen eingebaut. Arnim.
07.03.94 KPHeuristik als Aufrufoption eingebaut.
16.06.98 Komplett umgebaut. AJ.
*/

/* ------------------------------------------------*/
/* ------------- Includes -------------------------*/
/* ------------------------------------------------*/
#include "antiwarnings.h"
#include "compiler-extra.h"


#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "Ausgaben.h"
#include "ChaseManhattan.h"
#include "DomaenenMemorium.h"
#include "DSBaumOperationen.h"
#include "DSBaumTest.h"
#include "FehlerBehandlung.h"
#include "FileMenue.h"
#include "Hauptkomponenten.h"
#include "KPVerwaltung.h"
#include "MatchOperationen.h"
#include "MNF.h"
#include "Parameter.h"
#include "PhilMarlow.h"
#include "Puffer.h"
#include "Signale.h"
#include "Statistiken.h"
#include "Unifikation1.h"
#include "Unifikation2.h"
#include "VariablenMengen.h"
#include "WaldmeisterII.h"
#include "ZeitMessung.h"
#include "Zielverwaltung.h"
#include "general.h"
#include "pcl.h"


/* ------------------------------------------------*/
/* -------- globale Variablen ---------------------*/
/* ------------------------------------------------*/

unsigned int WM_AnlaufNr;
BOOLEAN LaufBeendet = FALSE;
pid_t my_pid;
pid_t father_pid;
pid_t son_pid;
BOOLEAN Vadder = TRUE;
BOOLEAN Forked = FALSE;
BOOLEAN communicationPossible = FALSE;
BOOLEAN ueberholt = FALSE;

/* ------------------------------------------------*/
/* -------- lokale Prozeduren ---------------------*/
/* ------------------------------------------------*/

/* zum Suchen in den angegebenen Beispielverzeichnissen.*/

static const char  *LaufErgebnis;         /* Ergebnis des Laufes im Klartext */
static BOOLEAN Resultat;
static BOOLEAN RunProof = FALSE;         /* wird auf TRUE gesetzt, wenn Beweis konstruiert werden soll */

static void SpezifikationEinlesen(char *filename)
/* Nur Spezifikation einlesen, damit extModus bekannt wird. */
{
#if !COMP_POWER
   fprintf(IO_Strom, "\nparsing %s", filename);
#endif
   parse_spec(filename);
#if !COMP_POWER
   fprintf(IO_Strom, "... done\n");
#endif
}
   
static void AllesAprioriInitialisieren(void)
/* Alles, was VOR Einlesen der Spezifikation passieren mu\3 oder kann. Die Aufrufparameter sind aber noch nicht eingestellt! */
{

   IO_InitApriori();  /* IO_Strom initialisieren */
   PA_InitApriori();  /* Parameter definieren */
   DM_InitApriori();  /* Signalbehandlung umbiegen fuer Domaenenverwaltung */
   SV_InitApriori();  /* Falls SV-Debug-Betrieb: Aufruf malloc_debug(2) */
   BK_InitApriori();  /* Speichermanager fuer innere und Blattknoten und Baumverwaltungseintraege */
   BO_InitApriori();  /* Variablenbank; Speichermanager fuer gerichtet normierte Termpaare */
   RE_InitApriori();  /* Variablenumbenennungsbank dynamisch erzeugen; NaechsteBaumvariable auf x1 setzen */
   MO_InitApriori();  /* Variablenbank fuer Matching dynamisch erzeugen */
   VM_InitApriori();  /* Speichermanager fuer Variablenmengen */
   U1_InitApriori();  /* Bindungsstack, Substitutions-Variablenbank, TermIteration */
   U2_InitApriori();  /* Bindungsstack, Substitutions-Variablenbank, TermIteration */
   TP_InitApriori();  /* Termpaar-Manager initialisieren. */
   BT_InitApriori();  /* Stapel in DSBaumTest initialisieren. */
   LL_InitApriori();  /* Listenzellenmanager einrichten */
}

static void SpezifikationGelesen(void)
/* Eingelesene Strukturen freigeben. */
{
  /*clear_spec();*/
}

static HK_ErgebnisT ModusCompletionProof (StrategieT *Strategie)
{
  HK_ErgebnisT Ergebnis;

  Ergebnis = HK_Vervollstaendigung(Strategie);

  switch (Ergebnis) {
  case HK_VervollAbgebrochen:   /* Failing completion */
    LaufErgebnis = "Failing completion aborted";
    break;
  case HK_Vervollstaendigt:
    LaufErgebnis = HK_fair() ? "System completed" : "Unfair completion run out of critical pairs";
    if (HK_fair()) {
      COM_markSuccess();
    }
    break;
  case HK_ZielBewiesen:
    LaufErgebnis = Z_AnzahlZiele()==1 ? "Goal proved" : 
      Z_AnzahlBewiesenerZiele()==Z_AnzahlZiele() ? "All goals proved" : "One goal proved";
    COM_markSuccess();
    break;
  case HK_Ueberholt:
    ueberholt = TRUE;
  case HK_StrategieWechseln:
    break;
  case HK_Beendet:              /* wird nur in anderen Modi benutzt */
    our_error ("HK_Beendet in Completion/Proof encountered");
    break;
  }
  
  if (PA_ExtModus==completion) {
     RE_AusgabeRegelmenge(7);
     if (!PA_FailingCompletion && RE_GleichungsmengeNichtLeer())
        RE_AusgabeGleichungsmenge (7);
     HK_ZieleAufZusammenfuehrbarkeitTesten();
  }
  RunProof = 
    (PA_ExtModus==completion && (Z_MindEinZielBewiesen() || HK_fair())) ||
    (PA_ExtModus==proof && Z_MindEinZielBewiesen());
  /* Note: The proof construction tool cannot cope with refutations. */
  if ((Ergebnis != HK_StrategieWechseln) && !ueberholt)
    Z_AusgabeZieleEnde(HK_fair() && (PA_ExtModus!=proof || !Z_ZieleBewiesen())); /* System konvergent? */

  return Ergebnis;
}


static HK_ErgebnisT ModusProof (StrategieT *Strategie)
{
   if (!SO_PraezedenzTotal) {
      our_error ("Precedence is not total");
   }
   else if (Z_AnzahlZiele()==0) {
     our_error ("Proof mode without goals");
   }
   
   /* alles ok */
   return ModusCompletionProof (Strategie);
}


static HK_ErgebnisT ModusConfluence (void)
{
   /* Ausgangsgleichunugen von links nach rechts richten,*/
   /* in R aufnehmen und dabei alle KPs bilden. */
   HK_GewichtsloseKPsBehandeln();
   
   if (HK_KPsZusammenfuehrbar()) { /* Test auf Zusammenfuehrbarkeit der KPs */
      if (HK_ZieleAufZusammenfuehrbarkeitTesten()) { /* .. falls erfolgreich: Ziele behandeln */
         LaufErgebnis = "System confluent, goals joined";         
         RunProof = TRUE;
      }
      else
         LaufErgebnis = "System confluent, goals not joined"; 
   }
   else
      LaufErgebnis = "System confluent, goals not joined";
   return HK_Beendet;
}


static HK_ErgebnisT ModusConvergence (void)
{
   /* Ausgangsgleichungen nach Ordnung richten,*/
   /* in R aufnehmen und dabei alle KPs bilden. */
   /* .. Wenn alle Gleichungen richtbar waren: */
   if ((Resultat = HK_GewichtsloseKPsBehandeln())) {
      if (HK_KPsZusammenfuehrbar()) { /*    .. Test auf Zusammenfuehrbarkeit der KPs */
         if (HK_ZieleAufZusammenfuehrbarkeitTesten())
            LaufErgebnis = "System convergent, goals joined";
         else
            LaufErgebnis = "System convergent, goals not joined";
      }
      else
         LaufErgebnis = "System noetherian but not convergent, goals not joined";
   }
   else
      LaufErgebnis = "System is not noetherian, goals not joined";   

   return HK_Beendet;
}


static HK_ErgebnisT ModusReduction (void)
{
   /* Ausgangsgleichungen von links nach rechts richten, ohne */
   /* Interreduktion und KP-Bildung in die Regelmenge stecken */
   HK_GewichtsloseKPsBehandeln();
   if (HK_ZieleAufZusammenfuehrbarkeitTesten())
      LaufErgebnis = "All goals joinable";
   else
      LaufErgebnis = "Not all goals joinable";

   return HK_Beendet;
}


static HK_ErgebnisT ModusTermination (void)
{
   /* Ausgangsgleichungen nach Ordnung richten */
   /* und KP-Bildung in die Regelmenge stecken */
   if ((Resultat = HK_GewichtsloseKPsBehandeln()),TRUE) {
      if (HK_ZieleAufZusammenfuehrbarkeitTesten()) /* ... Ziele behandeln. */
         LaufErgebnis = "System terminates, all goals joinable";
      else
         LaufErgebnis = "System terminates, not all goals joinable";
   }
   else
      LaufErgebnis = "System not orientable with specified ordering, goals not considered";
   return HK_Beendet;
}


static HK_ErgebnisT RunModus (StrategieT *Strategie)
{
   switch (PA_ExtModus) {
   case proof:
      return ModusProof (Strategie);
      break;
   case completion:
      return ModusCompletionProof (Strategie);
      break;
   case confluence:
      return ModusConfluence ();
      break;
   case convergence:
      return ModusConvergence ();
      break;
   case reduction:
      return ModusReduction ();
      break;
   case termination:
      return ModusTermination ();
      break;
   default: break;
   }
   return HK_Beendet;
}


static void InitDatenbank (void)
{
   if (PA_Datenbank()) {
      char *db = FM_DatenbankErkannt(PA_Datenbank());
      if (db)
         CM_InitApostposteriori(db);
      else
         our_error ("Data base not found");
   }
}

static void AbfrageConfigurationOK (void)
{
   if (PA_InteraktionAn()){
      char eing[INPUT_STRLEN];
      do {
         printf ("\nConfiguration ok (y/n)?\n>>> ");
         fgets(eing, INPUT_STRLEN, stdin);
       } while ((eing[0] != 'j') && (eing[0] != 'n') && (eing[0] != 'y') && (eing[0]!= 'c'));
      if (eing[0] == 'n') {
         printf("\n\n... try it again, Sam.\n\n");
         exit(0);
      }
   }
}


static void LeseSpezifikation (int argc, char *argv[])
{
   char *Spezifikation;

   Spezifikation = FM_SpezifikationErkannt(argc, argv);
  
   if (Spezifikation == NULL){
     IO_AusgabeStromAn ();
     fprintf(IO_Strom, "\n\nTerminated: no specification!\n\n");
     exit(1);
   }

   /*********************************
    Gewaehlte Spezifikation einlesen
   ********************************/

   IO_SpezifikationsnamenSetzen (Spezifikation);

   SpezifikationEinlesen (Spezifikation);
}

static void DoSoftCleanUp (void)
{
         
  KPV_SOFTCLEANUP();
  SV_AlleManagerLeeren();
}

/* optstr != "" */
static void ArgumenteUmbauen (char *opt_str, int *argc, char **argv[])
{
   int altargc = *argc;
   int i = 0;
   int newargc;
   char **newargv;

   BuildArgv (opt_str, (*argv)[0], &newargc, &newargv);
   
   *argc += newargc-1;
   *argv = our_realloc (*argv, sizeof(char *) * *argc, "ArgumenteUmbauen");
   /* Argumente nach hinten schieben */
   for (i = 1; i < altargc; i++)
      (*argv)[*argc-i] = (*argv)[altargc-i];

   for (i = 1; i < newargc; i++)
     (*argv)[i] = newargv[i];

   our_dealloc (newargv);
}

/* ------------------------------------------------*/
/* ---------- DAS HAUPTPROGRAMM -------------------*/
/* ------------------------------------------------*/

int    
#if COMP_POWER
    WM_main
#else
       main
#endif
            (int argc, char *argv[])
{
  HK_ErgebnisT Ergebnis;
  int WM_argc;
  char **WM_argv,**WM_argv_copy;

  /* Wir benutzen eigene Variablen um argc, argv beliebig manipulieren zu koennen */
  WM_argc = argc;
  WM_argv = our_alloc (sizeof (char *) * WM_argc);
  WM_argv_copy = our_alloc (sizeof (char *) * WM_argc);
  memcpy (WM_argv, argv, sizeof (char *) * WM_argc);
  

  /************************
    Vorab-Initialisierung
   *************************/

  WM_AnlaufNr = 1;
  AllesAprioriInitialisieren(); /* Alle Module initialisieren. */
  
  /************************************************************
    Aufruf abarbeiten: Spezifikation und Parameter einstellen
   *************************************************************/
  
  PA_Vorbehandeln(WM_argc, WM_argv);                                /* interaction, protocol, automodus etc. */
  LeseSpezifikation(WM_argc, WM_argv);
  
  if (PA_Automodus)
    memcpy (WM_argv_copy, WM_argv, sizeof (char *) * WM_argc);
  
  PA_InitAposterioriEarly ();
  PM_SpezifikationAnalysieren();

  COM_init();
  

  /************************************************************
    Schleife fuer StrategieWechsel
   ************************************************************/
  do {
    StrategieT *Strategie = NULL;    
    if (PA_Automodus) {
      Strategie = PM_Strategie(WM_AnlaufNr);
      if ((WM_AnlaufNr == 1) && (Strategie->Schrittzahl != 0)){ /* mehr als eine Strategie */
	if ((!PA_ForkErlaubt) || (!communicationPossible)){
	  Forked = FALSE;
	}
	else { /* Also probieren wir es... */
	  pid_t pid =  fork();
	  my_pid = getpid();
	  if (pid == -1){ /* fork hat nicht geklappt */
	    our_warning("fork failed!");
	    /* weiter wie bisher */
	    Forked = FALSE;
	  }
	  else if (pid == 0){/* Sohn */
	    ZM_StartGesamtZeit();                                               /* Gesamtzeit starten */

 	    Strategie->Schrittzahl = 0;

	    nice(15);
	    Vadder = FALSE;
	    Forked = TRUE;
	    son_pid = my_pid;
	    father_pid = getppid();
	  }
	  else { /* Vadder */
	    ZM_StartGesamtZeit();                                               /* Gesamtzeit starten */

	    WM_AnlaufNr++;
	    Strategie = PM_Strategie(WM_AnlaufNr);

	    Vadder = TRUE;
	    Forked = TRUE;
	    son_pid = pid;
	    father_pid = my_pid;
	  }
	}
      }

      if (!Forked && WM_ErsterAnlauf()) {
	ZM_StartGesamtZeit();                                               /* Gesamtzeit starten */
      }

      if (!WM_ErsterAnlauf()) {
         if (!PA_InteraktionAn())
            IO_AusgabeStromAus ();
         WM_argc = argc;
         memcpy (WM_argv, WM_argv_copy, sizeof (char *) * WM_argc);
      }
      ArgumenteUmbauen (Strategie->Optionen, &WM_argc, &WM_argv);
      if (!WM_ErsterAnlauf())
        AllesAprioriInitialisieren();
    }
    
    if (!PA_AufrufParameterEinstellen(WM_argc, WM_argv, get_mode())){
      IO_AusgabeStromAn ();
      fprintf(IO_Strom, "\nProgram aborted because of inconsistent parameter setting!\n\n");
      exit(1);
    }
    
    PA_InitAposteriori();                                  /* System gemaess Parameter-Einstellung initialisieren. */

#if !COMP_POWER
    IO_DruckeZeile('+');
#endif
    if (WM_ErsterAnlauf())
      AbfrageConfigurationOK ();

    if (!PA_InteraktionAn()) {             /* Bei !PA_InteraktionAn wird Ausgabestrom auf /dev/null initialisiert, */
      IO_AusgabeStromAn ();                                                          /* um Ausgaben zu unterbinden */
      if (WM_ErsterAnlauf() && ((!Forked) || Vadder)) 
         IO_DruckeKopf();                                                                /* Programm-Kopf ausgeben */
    }

    SI_SignalsSetzen (); 
    HK_GleichungenUndZieleHolen();                        /* Gleichungen und Ziele in Systemstrukturen uebernehmen */
    SpezifikationGelesen();
    if (WM_ErsterAnlauf() && ((!Forked) || Vadder))
      PM_StrategienDrucken();
    InitDatenbank ();
    
    /*************************************************************************
     * Eigentlicher Programm-Lauf nach Art der Spezifikationsdatei
     *************************************************************************/
        
    Ergebnis = RunModus (Strategie);
    if (Ergebnis == HK_StrategieWechseln) {
      DoSoftCleanUp ();
    }
    ++WM_AnlaufNr;
  }  while (Ergebnis == HK_StrategieWechseln);
  
  
  LaufBeendet = TRUE;

  ZM_StoppGesamtZeit();  
  SI_SignalsDeaktivieren ();
  IO_RingBell();                  /* Benutzer wecken */
  
  /**************************************************
    Ausgabe der Ergebnisse
    ***************************************************/
  if (Forked && Vadder){
    waitpid(son_pid, NULL, 0);
#if !COMP_COMP
    IO_DruckeFlex("Sohn eingesammelt\n");
#endif
  }
  
  if (Forked){
    if (!ueberholt){
      if (Vadder){ /* Sohn erlegen */
	if (!USE_COM || (COM_getState() == STATE_FATHER_READY)) {
	  RE_DruckeEndsystemFuerPCL(); /* Fuer PCL-Modus ggf. Abschlussfakten drucken */
	  PCL_Beenden();               /* PCL-Protokoll abschliessen */
	  /*	  kill(son_pid, SIGUSR2);*/
	}
      }
      else { /* Vadder in Rente schicken */
	
	if (!USE_COM || (COM_getState() == STATE_SON_READY)) {
	  RE_DruckeEndsystemFuerPCL(); /* Fuer PCL-Modus ggf. Abschlussfakten drucken */
	  PCL_Beenden();               /* PCL-Protokoll abschliessen */
	  /* kill(father_pid, SIGUSR2);*/
	}
      }
    }
    else {
      /* leider verloren ... */
    }
  } 
  else {  
    RE_DruckeEndsystemFuerPCL(); /* Fuer PCL-Modus ggf. Abschlussfakten drucken */
    PCL_Beenden();               /* PCL-Protokoll abschliessen */
  }

  /**************************************************
    Statistiken
    ***************************************************/
  
#if COMP_POWER
  ST_PowerStatistik(LaufErgebnis);
#else
  ST_SuccessStatistik(LaufErgebnis, Resultat);
#endif
  
  /**************************************************
    Speicher aufraeumen
    ***************************************************/

#if !COMP_COMP
  COM_free();  
  IO_AusgabenAus();                                                          /* verbliebene Strukturen loeschen */
  RE_RUndETest();                                     /* falls eincompiliert, Mengenkonsistenztest durchfuehren */
  
  if (PA_InteraktionAn() || !COMP_OPTIMIERUNG) {
    Z_Cleanup();    /* Ziele loeschen */
    MNF_Cleanup();
    KPV_CLEANUP();         /* muss vor dem Loeschen der R- bzw. E-Menge passieren */
    
    RE_GleichungsmengeLoeschen();
    RE_RegelmengeLoeschen();   /* E- und R-Menge loeschen */
    
    PU_TPPufferLoeschen();
    TO_FeldVariablentermeLeeren(); /* ... */
    CM_DatenbankLeeren();
    PM_ArchivLeeren();
  }
  else {
    PM_ArchivLeeren();
    DoSoftCleanUp ();
  }
  
  IO_AusgabenEin();
  
  SO_CleanUp();
  
  if (PA_InteraktionAn() || !COMP_OPTIMIERUNG)
    SV_Statistik(FALSE, TRUE, "complete");      /* FALSE: kein System-Fehler aufgetreten */
  
  IO_CleanUp(); /* falls Ausgaben umgelenkt waren, die Protokoll-Datei schliessen */
  
  /**************************************************
    Ausgaben beenden
    ***************************************************/
  
  
  DM_NieMehrDomaenen_SagNiemalsNie();
#endif

  /* Programm-Ende */
#if COMP_POWER
  /* return = 0 bedeutet alles ok */
  return (RunProof == FALSE);
#else
  return (0);
#endif
}

/* ------------------------------------------------*/
/* -------VADDER / SOHN Communication--------------*/
/* ------------------------------------------------*/

static int lock_reg(int fd, int cmd, int type, off_t offset, int whence, off_t len)
{
  struct flock	lock;

  lock.l_type = type;		/* F_RDLCK, F_WRLCK, F_UNLCK */
  lock.l_start = offset;	/* byte offset, relative to l_whence */
  lock.l_whence = whence;	/* SEEK_SET, SEEK_CUR, SEEK_END */
  lock.l_len = len;		/* #bytes (0 means to EOF) */

  return( fcntl(fd, cmd, &lock) );	/* -1 upon error */
}

#define	writew_lock(fd, offset, whence, len) lock_reg(fd, F_SETLKW, F_WRLCK, offset, whence, len)
#define	un_lock(fd, offset, whence, len)     lock_reg(fd, F_SETLK, F_UNLCK, offset, whence, len)

struct syncData {
  short state;
  int fd;
} syncData;

static struct syncData *sharedState; 

void COM_init(void) {
#if USE_COM
  char *pathname;
  sharedState = mmap_shared(sizeof(syncData));
  if (sharedState == NULL){
    communicationPossible = FALSE;
    return;
  }
  sharedState->state = STATE_UNSET;

  pathname = tempnam("/tmp", "wm");
  if (pathname == NULL){
    communicationPossible = FALSE;
    return;
  }
    
  sharedState->fd = open(pathname, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  if (sharedState->fd == -1){
    communicationPossible = FALSE;
    return;
  }

  if (unlink(pathname) == -1){
    communicationPossible = FALSE;
    return;
  }

  if (writew_lock(sharedState->fd, 0, SEEK_SET, 0) == -1){
    communicationPossible = FALSE;
    return;
  }

  if (un_lock(sharedState->fd, 0, SEEK_SET, 0) == -1){
    communicationPossible = FALSE;
    return;
  }

  communicationPossible = TRUE;
#else
  communicationPossible = FALSE;
#endif
}

void COM_free() {
#if USE_COM
/*if (communicationPossible){
    un_lock(sharedState->fd, 0, SEEK_SET, 0);
  }*/
#endif
}

#define NO_OF_TRIES 3
void COM_setState( short state ) {
#if USE_COM
  int i, res = 0;

  for(i=0; i < NO_OF_TRIES; i++){
#if !COMP_COMP
    printf("%sTry to get lock(%d)\n", Vadder ? "0>" : "1>", i);
#endif
    res = writew_lock(sharedState->fd, 0, SEEK_SET, 0);
    /* es kann z.B. ein signal einschlagen, waehrend wir warten ;-(( */
    if (res != -1) break;
  } 

  if (res == -1){
    our_warning("problems with acquiring lock");
    return;
  }

#if !COMP_COMP
  printf("%s>got lock(%d) --> \n", Vadder ? "0>" : "1>", i, state);
#endif

  if (sharedState->state == STATE_UNSET) {
    sharedState->state = state;
  } 
  else {
    our_warning( "illegal state change" );
  }

  for(i=0; i < NO_OF_TRIES; i++){
#if !COMP_COMP
    printf("%sTry to release lock(%d)\n", Vadder ? "0>" : "1>", i);
#endif
    res = un_lock(sharedState->fd, 0, SEEK_SET, 0);
    if (res != -1) break;
  } 

  if (res == -1){
    our_warning("problems with releasing lock");
  }

#endif
}

short COM_getState() {
#if USE_COM
  return sharedState->state;
#else
  return STATE_UNDEF;
#endif
}

/* speichere, dass ich einen Beweis gefunden habe */
void COM_markSuccess() { 
#if !COMP_COMP
  printf("%d success\n", Vadder);
#endif
  if (Forked) {
    if (Vadder) {
      COM_setState(STATE_FATHER_READY);
    } else {
      COM_setState(STATE_SON_READY);
    }
  }
}
