/*
Dieses Modul enthaelt die Ausgabe-Funktionen sowie alle 'Bezeichner' der eingelesenen Spezifikation.

28.02.94 Angelegt.
02.03.94 DruckeKopf und DruckeStatistik eingebaut.

*/
/*
  ToDo:
  Anpassung an englische Gepflogenheiten: Decimalpunkt, Tausenderkomma?
*/
/* -------------------------------------------------------*/
/* ------------- Includes --------------------------------*/
/* -------------------------------------------------------*/
#include "antiwarnings.h"
#include <signal.h> 
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#if defined (LINUX) || defined (SUNOS5)
#include <sys/wait.h>
#endif

#include "Ausgaben.h"
#include "FehlerBehandlung.h"
#include "general.h"
#include "Grundzusammenfuehrung.h"
#include "Parameter.h"
#include "parser.h"
#include "PhilMarlow.h"
#include "SpezNormierung.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "TermIteration.h"
#include "version.h"
#include "WaldmeisterII.h"
#include "Zielverwaltung.h"

TI_IterationDeklarieren(IO)

unsigned long int IO_AusdehnungAusgabestapel(void)
{
   return IO_AusdehnungKopierstapel();
}

/* ------------------------------------------------------------------------*/
/* -------------------- globale Variablen (siehe Ausgaben.h) --------------*/
/* ------------------------------------------------------------------------*/

unsigned int IO_level;
BOOLEAN IO_GUeberschrift;
FILE *IO_Strom;

/* -------------------------------------------------------*/
/* ------------- Globals ---------------------------------*/
/* -------------------------------------------------------*/

static char *Spezifikationsname = NULL;
static char **IO_SymbolNamen;
static char IO_SpezName[MaximaleBezeichnerLaenge];
static char IO_Modus[MaximaleBezeichnerLaenge];
static int Ausgabelevel;

static BOOLEAN LatexAusgabeAktiviert = FALSE;

#define IO_Ueberschriftenbreite 80
#define IO_KOMBINATORAUSGABE 0


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ AusgabeStrom steuern ++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

static FILE *AlterStrom = NULL;

void IO_AusgabeStromUmlenken(const char *Filename)
{
#if !COMP_COMP
  fprintf(stderr,"%s IO_AusgabeStromUmlenken: %s \n",
	  Forked ? (Vadder ? "Vadder" : "Sohny") : "ich",
	  Filename);
#endif
  if (0 && (AlterStrom != NULL))
      our_fatal_error("Ausgabestrom zweimal umgelenkt!");
   else{
      AlterStrom = IO_Strom;
      if (!(IO_Strom = fopen(Filename, "a"))){
         IO_Strom = AlterStrom;
         AlterStrom = NULL;
         fprintf(IO_Strom, "\nAusgabe konnte nicht umgelenkt werden!\n\n");
      }
   }
}

void IO_AusgabeStromZuruecksetzen(void)
{
#if !COMP_COMP
  fprintf(stderr,"%s IO_AusgabeStromZuruecksetzen:\n",
	  Forked ? (Vadder ? "Vadder" : "Sohny") : "ich");
#endif
   if (AlterStrom == NULL)
      our_fatal_error("Ausgabestrom nicht umgelenkt -> kein Zuruecksetzen moeglich!\n");
   else{
      fclose(IO_Strom);
      IO_Strom = AlterStrom;
      AlterStrom = NULL;
   }
}

static BOOLEAN AusgabeIstAus = FALSE;

void IO_AusgabeStromAus (void)
{
  AusgabeIstAus = TRUE;
  IO_AusgabeStromUmlenken ("/dev/null");
}

void IO_AusgabeStromAn (void)
{
  if (AusgabeIstAus) {
    AusgabeIstAus = FALSE;
    IO_AusgabeStromZuruecksetzen ();
  }
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Initialisierung +++++++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void IO_InitApriori(void)
/* Vor der ersten Ausgabe auf IO_Strom aufzurufen!!!! */
{
  if (WM_ErsterAnlauf())
   IO_Strom = stdout;
  /* Line buffering benutzen */
#if !COMP_WM
  if (setvbuf (IO_Strom, NULL, _IOLBF, 0) != 0)
    perror ("setvbuf");
#endif
}

BOOLEAN IO_SymbolErkannt(char *str, SymbolT *Symbol)
/* Liefert TRUE, wenn str dem Ausgabe-String eines Symbols aus der Spezifikation oder aus der Paramodulation entspricht. */
{
   SymbolT i;

   SO_forFktSymbole(i)
      if (strcmp(str, IO_SymbolNamen[i]) == 0){
         *Symbol = i;
         return TRUE;
      }
   return FALSE;
}


static void ParamodSymbolnamenEtablieren(SymbolT ParamodSymbol, const char *Name)
{
  unsigned int n = 0;
  IO_SymbolNamen[ParamodSymbol] = our_alloc(MaximaleBezeichnerLaenge*sizeof(char *));
  while (TRUE) {
    SymbolT i;
    BOOLEAN ok = TRUE;
    if (n)
      sprintf(IO_SymbolNamen[ParamodSymbol],"%s%d",Name,n);
    else
      sprintf(IO_SymbolNamen[ParamodSymbol],"%s",Name);
    SO_forSpezSymbole(i)
      if (strcmp(IO_SymbolNamen[i],IO_SymbolNamen[ParamodSymbol])==0) 
        ok = FALSE;
    if (ok)
      return;
    n++;
  }
}

void IO_CleanUp(void)
{
  SymbolT i;

  SO_forFktSymbole(i)
    our_dealloc(IO_SymbolNamen[i]);
  our_dealloc(IO_SymbolNamen);

  if (IO_Strom != stdout){
     fclose(IO_Strom);
     IO_Strom = stdout;
  }
}

void IO_ParamInfo(ParamInfoArtT what, int level, char* protfile)
{
  if (what == ParamEvaluate){
    fprintf(IO_Strom, "\n* OUTPUT:  Level %d to %s%s\n", 
            level, 
            (protfile||(strcmp(protfile,"stdout")!=0))?"file ":"",
            (protfile||(strcmp(protfile,"stdout")!=0))?protfile:"stdout");
  }
  else{
    fprintf(IO_Strom, "-a (0..5) -prot <protfile>\n");
  }
}

void IO_InitAposteriori(int level, char *Protokolldatei)
/* Daten aus der Spezifikation holen. */
{
   Ausgabelevel = level;
   IO_level     = level;
   IO_GUeberschrift = FALSE;
   if (strcmp(Protokolldatei, "stdout") != 0) {
     fprintf(IO_Strom, "\nWaldmeister is running, output goes to %s.\n", Protokolldatei);
     IO_AusgabeStromUmlenken (Protokolldatei);
     PA_InteraktionAusschalten();
  }
#if IO_KOMBINATORAUSGABE
   TO_KombinatorapparatAktivieren();
#endif
}  


void IO_InitAposterioriEarly (void)
/* Daten aus der Spezifikation holen. */
{
   func_symb_list p;
   SymbolT i;
   
   IO_SymbolNamen = array_alloc(SO_AnzahlFunktionen,char*);
   SO_forSpezSymbole(i)
     IO_SymbolNamen[i] = string_alloc(MaximaleBezeichnerLaenge);
   for (p=get_sig(); p; p = sig_rest(p)) {
     func_symb f = sig_first(p);
     strcpy(IO_SymbolNamen[SN_WMvonDISCOUNT(f)],map_func(f));
   }
   if (/*PA_KernAn*/ TRUE) {  /* XXX 29.06.98 */
     strcpy(IO_SymbolNamen[SO_const1],"const1");
     strcpy(IO_SymbolNamen[SO_const2],"const2");
   }
   if (PM_Existenzziele()) {
     ParamodSymbolnamenEtablieren(SO_EQ   ,"eq");
     ParamodSymbolnamenEtablieren(SO_TRUE ,"true");
     ParamodSymbolnamenEtablieren(SO_FALSE,"false");
   }
   sprintf(IO_SpezName,get_spec_name());
   sprintf(IO_Modus,mode_names[PA_ExtModus]);
   /* default-Werte setzen */
   Ausgabelevel = 2;
   IO_level     = 2;
   IO_GUeberschrift = FALSE;
   LatexAusgabeAktiviert = FALSE;
}  


void IO_FunktionsvariablenEintragen(void)
{
  SymbolT i;
  func_symb_list sig = get_sig();
  array_realloc(IO_SymbolNamen,SO_AnzahlFunktionen+SO_AnzahlFunktionsvariablen(),char*);
  SO_forFktVarSymbole(i) {
    func_symb f;
    while ((!sig_is_fvar(f = sig_first(sig))))
      sig = sig_rest(sig);
    sig = sig_rest(sig);
    IO_SymbolNamen[i] = string_alloc(MaximaleBezeichnerLaenge);
    strcpy(IO_SymbolNamen[i],map_func(f));
  }
}


int IO_AusgabeLevel(void)
{
   return Ausgabelevel;
}

void IO_AusgabeLevelBesetzen(int level)
{
  IO_level     = level;
  Ausgabelevel = level;
}

void IO_AusgabenAus(void)
/* Schaltet unabhaengig vom aktuellen Ausgabelevel die Ausgaben aus.*/
{
  IO_level = 0;
}

void IO_AusgabenEin(void)
/* Schaltet die mit IO_AusgabenAus zuvor ausgeschalteten Ausgaben gemaess Einstellung wieder ein.*/
{
  IO_level = Ausgabelevel;
}

void IO_RingBell(void)
{
  if (PA_InteraktionAn())
    fprintf(IO_Strom, "\a");
}

void IO_SpezifikationsnamenSetzen(char *Filename)
/* Setzt den Namen der bearbeiteten Spezifikation (incl. Pfad) ein.*/
{
   Spezifikationsname = strdup(Filename);
}

char *IO_Spezifikationsname(void)
/* Gibt den Namen (und ggf. Pfad) der zu verwendeten Spez.datei zurueck.*/
{
   return Spezifikationsname;
}

/* -------------------------------------------------------*/
/* -------------- Kopf und Spezifikation -----------------*/
/* -------------------------------------------------------*/

void IO_DruckeKopf(void)
{
#if !COMP_POWER
  if (PA_InteraktionAn)
    fprintf(IO_Strom, "\n\n");
#endif
  fprintf(IO_Strom, "********************************************************************************\n");
  fprintf(IO_Strom, "*                             W A L D M E I S T E R           \\|  \\ /      \\|/ *\n");
  fprintf(IO_Strom, "*                                                              |/  |    \\/  |  *\n");
  fprintf(IO_Strom, "*              (C) 1994-2002  A. Buch and Th. Hillenbrand,      \\ /      \\ /   *\n");
  fprintf(IO_Strom, "*                             A. Jaeger and B. Loechner          |        |    *\n");
  fprintf(IO_Strom, "*                             <waldmeister@informatik.uni-kl.de>          |    *\n");
  fprintf(IO_Strom, "********************************************************************************\n");
#if !COMP_POWER
  fprintf(IO_Strom, "\n");
#endif
}

void IO_DruckeBeispiel(void)
{
  if (Spezifikationsname != NULL){
   if (LatexAusgabeAktiviert){
      char *Dateiname_ = strrchr(Spezifikationsname,'/'),
           *Dateiname = Dateiname_ ? Dateiname_+1 : Spezifikationsname;
      fprintf(IO_Strom, "%-10s & ", Dateiname);
   }
   else{
      if (IO_SpezName && IO_SpezName [0] != '\0') {
         IO_DruckeKopfZeileText(TRUE, "problem           ", IO_SpezName);
         IO_LF();
         IO_DruckeKopfZeileText(TRUE, "... from file     ", Spezifikationsname);
         IO_LF();
         IO_DruckeKopfZeileText(TRUE, "Computation Mode", IO_Modus);
      }
   }
  }
}  

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Elementare Druck-Funktionen +++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void IO_DruckeGanzzahl(int n)
{
  fprintf(IO_Strom, "%d",n);
}

void IO_DruckeAdresse(void *Adresse)
{
  fprintf(IO_Strom, "%p",Adresse);
}

static void DruckeStellenfolgeStrom(FILE *Strom, PCL_SeitenT seite, UTermT Term, UTermT Teilterm);

void IO_DruckeFlex(const char *fmt, ...)
/* Neufassung vom 1.4.95, Arnim. */
{
   va_list ap;
   int ival;
   double dval;
   unsigned long uval;
   long int lval;
   TermT tval,tval2;
   PCL_SeitenT seite;

   char *ArgPosition;
   
   va_start(ap, fmt);
   do{
     ArgPosition = strchr(fmt, '%');
     if (ArgPosition != NULL){
       fwrite(fmt, sizeof(char), ArgPosition-fmt, IO_Strom); /* (ArgPosition-fmt) Objekte aus fmt auf Ausgabe-Strom schreiben */
       switch (*++ArgPosition){
       case 'b':
         ival = va_arg(ap, int);
         fprintf(IO_Strom, "%s", ival ? "True" : "False");
	 break;
       case 'd':
         ival = va_arg(ap, int);
         fprintf(IO_Strom, "%d", ival);
         break;
       case 'f':
         dval = va_arg(ap, double);
         fprintf(IO_Strom, "%f", dval);
         break;
       case 's':
         fprintf(IO_Strom, "%s", va_arg(ap, char *));
         break;
       case 'l':
         lval = va_arg(ap, long);
         fprintf(IO_Strom, "%ld",lval);
         break;
       case 'u':
         uval = va_arg(ap, unsigned long);
         fprintf(IO_Strom, "%lu", uval);
         break;
       case 'v':
         lval = va_arg(ap, long);
	 switch (lval){
	 case Gleich:         fprintf(IO_Strom, "Gleich");         break;
	 case Unvergleichbar: fprintf(IO_Strom, "Unvergleichbar"); break;
	 case Groesser:       fprintf(IO_Strom, "Groesser");       break;
	 case Kleiner:        fprintf(IO_Strom, "Kleiner");        break;
	 default:             fprintf(IO_Strom, "??VGL??");        break;
	 }
         break;
       case 'S':
	 {
	   SymbolT sym = va_arg(ap,SymbolT);
	   IO_DruckeSymbol(sym);
	 }
	 break;
       case 't':
         tval = va_arg(ap, TermT);
	 if (tval != NULL){
	   IO_DruckeTerm(tval);
	 }
	 else {
	   fprintf(IO_Strom, "<nullterm>");
	 }
         break;
       case 'p':
         seite = va_arg(ap, PCL_SeitenT);
         tval = va_arg(ap, TermT);
	 tval2 = va_arg(ap, TermT);
         DruckeStellenfolgeStrom(IO_Strom,seite,tval,tval2);
         break;
       case 'q':
         seite = va_arg(ap, PCL_SeitenT);
	 putc(seite == linkeSeite ? 'L' : 'R', IO_Strom);
         break;
       default:
         putc(*ArgPosition, IO_Strom);
         break;
       }
       fmt = ArgPosition + 1;
     }
     else
       fprintf(IO_Strom, fmt);
   }while(ArgPosition);
   va_end(ap);
   return;
}

void IO_DruckeFlexStrom(FILE *Strom, const char *fmt, ...)
{
   va_list ap;
   int ival;
   double dval;
   unsigned long uval;
   long int lval;
   TermT tval,tval2;
   PCL_SeitenT seite;
   
   char *ArgPosition;
   
   va_start(ap, fmt);
   do{
     ArgPosition = strchr(fmt, '%');
     if (ArgPosition != NULL){
       fwrite(fmt, sizeof(char), ArgPosition-fmt, Strom); /* (ArgPosition-fmt) Objekte aus fmt auf Ausgabe-Strom schreiben */
       switch (*++ArgPosition){
       case 'b':
         ival = va_arg(ap, int);
         fprintf(Strom, "%s", ival ? "True" : "False");
	 break;
       case 'd':
         ival = va_arg(ap, int);
         fprintf(Strom, "%d", ival);
         break;
       case 'f':
         dval = va_arg(ap, double);
         fprintf(Strom, "%f", dval);
         break;
       case 's':
         fprintf(Strom, "%s", va_arg(ap, char *));
         break;
       case 'l':
         lval = va_arg(ap, long);
         fprintf(Strom, "%ld",lval);
         break;
       case 'u':
         uval = va_arg(ap, unsigned long);
         fprintf(Strom, "%lu", uval);
         break;
       case 'v':
         lval = va_arg(ap, long);
	 switch (lval){
	 case Gleich:         fprintf(Strom, "Gleich");         break;
	 case Unvergleichbar: fprintf(Strom, "Unvergleichbar"); break;
	 case Groesser:       fprintf(Strom, "Groesser");       break;
	 case Kleiner:        fprintf(Strom, "Kleiner");        break;
	 default:             fprintf(Strom, "??VGL??");        break;
	 }
         break;
       case 'S':
	 {
	   SymbolT sym = va_arg(ap,SymbolT);
	   IO_DruckeSymbolStrom(Strom,sym);
	 }
	 break;
       case 't':
         tval = va_arg(ap, TermT);
	 if (tval != NULL){
	   IO_DruckeTermStrom(Strom, tval);
	 }
	 else {
	   fprintf(Strom, "<nullterm>");
	 }
         break;
       case 'p':
         seite = va_arg(ap, PCL_SeitenT);
         tval = va_arg(ap, TermT);
	 tval2 = va_arg(ap, TermT);
         DruckeStellenfolgeStrom(Strom,seite,tval,tval2);
         break;
       case 'q':
         seite = va_arg(ap, PCL_SeitenT);
	 putc(seite == linkeSeite ? 'L' : 'R', Strom);
         break;
       default:
         putc(*ArgPosition, Strom);
         break;
       }
       fmt = ArgPosition + 1;
     }
     else
       fprintf(Strom, fmt);
   }while(ArgPosition);
   va_end(ap);
   return;
}


/* Diese beiden Funktionen geben eine Zeile des Programm-Kopfes aus.
   Dabei wird die Ausrichtung automatisch vorgenommen.
   Die Ausgabe erfolgt nur, wenn Bedingung erfuellt ist.*/

void IO_DruckeKopfZeileText(BOOLEAN Bedingung, const char *Titel, const char *Text)
{
  if(Bedingung){
    fprintf(IO_Strom, "   %-40s: ", Titel);
    fprintf(IO_Strom, "%s\n", Text);
  }
}

void IO_DruckeKopfZeileInfoStrings(BOOLEAN Bedingung, const char *Titel, InfoStringT is1, InfoStringT is2)
{
  if(Bedingung){
     fprintf(IO_Strom, "   %-40s: ", Titel);
     if ((strlen(is1)>0) && (strlen(is2)>0))
        fprintf(IO_Strom, "%s:%s \n", is1, is2);
     else if (strlen(is1)>0)
        fprintf(IO_Strom, "%s \n", is1);
     else if (strlen(is2)>0)
        fprintf(IO_Strom, "%s \n", is2);
     else fprintf(IO_Strom, "\n");
  }
}


void IO_DruckeKopfZeileZahl(BOOLEAN Bedingung, const char *Titel, int Zahl)
{
  if(Bedingung){
    fprintf(IO_Strom, "   %-40s: ", Titel);
    fprintf(IO_Strom, "%d\n", Zahl);
  }
}



void IO_DruckeZeile (char zeichen)
/* Druckt eine ganze Zeile mit dem Zeichen zeichen. */
{
  short i;

  for (i=0; i<IO_Ueberschriftenbreite; i++){
    putc (zeichen, IO_Strom);
  }
  IO_LF();
}

void IO_DruckeUeberschrift(const char *text)
/* Ueberschrift drucken */
{
  short i, rahmen;

  IO_LF();
  IO_DruckeZeile('*');
  rahmen = ((IO_Ueberschriftenbreite - strlen(text) - 2) / 2);
  for (i=0; i<rahmen; i++){
    fprintf(IO_Strom, "*");
  }
  fprintf(IO_Strom, " ");
  fprintf(IO_Strom, text);
  fprintf(IO_Strom, " ");
  for (i=rahmen+strlen(text)+2; i<IO_Ueberschriftenbreite; i++){
    fprintf(IO_Strom, "*");
  }
  IO_LF();
  IO_DruckeZeile('*');
  IO_LF();
}  

void IO_WarteAufReturnMitText(const char *text)
/* Gibt text aus und Wartet auf Druck der Return-Taste.*/
{
  char bla[INPUT_STRLEN];

  fprintf(IO_Strom, "\n>> %s <<\n", text);
  rewind(stdin);
  fgets(bla, INPUT_STRLEN, stdin);
}


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Druck-Funktionen fuer System-Strukturen +*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void IO_DruckeVergleich(VergleichsT Vergleich)
{
  switch (Vergleich) {
  case Kleiner:        fprintf(IO_Strom, " < "); break;
  case Gleich:         fprintf(IO_Strom, " = "); break;
  case Groesser:       fprintf(IO_Strom, " > "); break;
  case Unvergleichbar: fprintf(IO_Strom, " # "); break;
  }
}


void IO_DruckeSymbol(SymbolT symbol)
{
#if GZ_GRUNDZUSAMMENFUEHRUNG
  if (SO_IstExtrakonstante(symbol))
    fprintf(IO_Strom, "c%d",symbol-SO_Extrakonstante(1)+1); 
  else
#endif
  if (SO_SymbIstFkt(symbol))
    fprintf(IO_Strom, IO_SymbolNamen[symbol]);
  else
    fprintf(IO_Strom, "x%d",SO_VarNummer(symbol));
}

char *IO_FktSymbName(SymbolT FktSymb)
/* Zeiger auf Zeichenketten-Darstellung des Druckenamens des Funktionssymbols */
{
  return IO_SymbolNamen[FktSymb];
}

#define Symb(t)   IO_DruckeSymbol(TO_TopSymbol(t))   
#define Oft(n)    fprintf(IO_Strom, "^%d",n); 
#define Auf()     putc( '(', IO_Strom)
#define Komma()   putc( ',', IO_Strom)
#define Spc()     putc( ' ', IO_Strom)
#define Zu()      putc( ')', IO_Strom)


#if IO_KOMBINATORAUSGABE

#define IstKTerm(t)  (SO_SymbGleich(TO_TopSymbol(t),TO_ApplyOperator))

static void DruckeTerm(UTermT t, BOOLEAN Klammern)
{
  unsigned int n = 1;
  UTermT ti;
  switch (TO_AnzahlTeilterme(t)) {
  case 0:
    Symb(t);
    break;
  case 1:
    Symb(t);
    for (ti=TO_Schwanz(t); TO_TopSymboleGleich(t,ti); n++, ti=TO_Schwanz(ti));
    if (n>=StartzahlVerkettungskuerzung)
      Oft(n); 
    Auf();
    DruckeTerm(n>=StartzahlVerkettungskuerzung ? ti : TO_Schwanz(t),TRUE);
    Zu();
    break;
  case 2:
    if (IstKTerm(t)) {
      UTermT t1 = TO_ErsterTeilterm(t),
        t2 = TO_NaechsterTeilterm(t1);
      if (Klammern) Auf();
      DruckeTerm(t1,FALSE);
      Spc();
      DruckeTerm(t2,TRUE);
      if (Klammern) Zu();
      break;
    }
  default:
    Symb(t);
    Auf();
    DruckeTerm(ti = TO_ErsterTeilterm(t),TRUE);
    while (TO_PhysischUngleich(ti = TO_NaechsterTeilterm(ti),TO_NULLTerm(t))) {
      Komma();
      DruckeTerm(ti,TRUE);
    }
    Zu();
  }
}

void IO_DruckeTerm(UTermT t)
{
  DruckeTerm(t,FALSE);
}

#else


void IO_DruckeTerm(UTermT t)
{
  if (TO_Nullstellig(t))
    Symb(t);
  else
    do
      switch (TO_AnzahlTeilterme(t)) {
      case 0: {
        Symb(t);
        while (!(--TI_TopZaehler)) {
          Zu();
          if (TI_PopUndLeer())
            return;
        }
        Komma();
        t = TO_Schwanz(t);
        break;
      }
      case 1: {
        SymbolT Symbol = TO_TopSymbol(t);
        UTermT Teilterm = TO_ErsterTeilterm(t);
        unsigned int n = 1;
        while (SO_SymbGleich(Symbol,TO_TopSymbol(Teilterm))) {
          n++;
          Teilterm = TO_ErsterTeilterm(Teilterm);
        }
        if (n>=StartzahlVerkettungskuerzung) {
          Symb(t);
          Oft(n);
          Auf();
          TI_PushOT(t);
        }
        else
          while (n--) {
            Symb(t);
            Auf();
            TI_PushOT(t);
          }
        t = Teilterm;
        break;
      }
      default: {
        Symb(t);
        Auf();
        TI_PushOT(t);
        t = TO_Schwanz(t);
        break;
      }
      }
    while (TRUE);
}
#endif

void IO_DruckeSymbolStrom(FILE *Strom, SymbolT symbol)
/* Ausgabe eines Symbols auf dem angegebenen Ausgabestrom.*/
{
  if (SO_SymbIstFkt(symbol))
    fprintf(Strom,IO_SymbolNamen[symbol]);
  else
    fprintf(Strom,"x%d",SO_VarNummer(symbol));
}


void IO_DruckeTermStrom(FILE *Strom, UTermT Term)
/* Druckt den Term Term auf den Ausgabestrom Strom aus. Dabei werden Folgen gleicher, einstelliger
   Funktionssymbole nicht gekuerzt und Konstanten mit nachfolgendem Klammernpaar ausgegeben.*/
{
  if (TO_Nullstellig(Term)){
    IO_DruckeSymbolStrom(Strom, TO_TopSymbol(Term));
    if (TO_TermIstNichtVar(Term)) /* bei Konstanten Klammernpaar ausgeben */
      fprintf(Strom, "()");
  }
  else
    do
      switch (TO_AnzahlTeilterme(Term)) {
      case 0: {
        IO_DruckeSymbolStrom(Strom, TO_TopSymbol(Term));
        if (TO_TermIstNichtVar(Term)) /* bei Konstanten Klammernpaar ausgeben */
          fprintf(Strom, "()");
        while (!(--TI_TopZaehler)) {
          fprintf(Strom, ")");
          if (TI_PopUndLeer())
            return;
        }
        fprintf(Strom, ",");
        Term = TO_Schwanz(Term);
        break;
      }
      case 1: {
        IO_DruckeSymbolStrom(Strom, TO_TopSymbol(Term));
        fprintf(Strom, "(");
        TI_PushOT(Term);
        Term = TO_ErsterTeilterm(Term);
        break;
      }
      default: {
        IO_DruckeSymbolStrom(Strom, TO_TopSymbol(Term));
        fprintf(Strom, "(");
        TI_PushOT(Term);
        Term = TO_Schwanz(Term);
        break;
      }
      }
    while (TRUE);
}


#define IO_DruckeSymbolString(buf, symb)     \
  (SO_SymbIstFkt(symb) ? our_sprintf((buf,IO_SymbolNamen[symb])) : our_sprintf((buf, "x%d",SO_VarNummer(symb))))


unsigned int IO_DruckeTermString(char* str, UTermT Term)
/* Druckt den Term Term auf den Ausgabestring str aus. Dabei werden Folgen gleicher, einstelliger
   Funktionssymbole nicht gekuerzt und Konstanten mit nachfolgendem Klammernpaar ausgegeben.
   In str muss genuegend Platz vorhanden sein. Gibt die Zahl ausgegebener Zeichen zurueck.*/
{
  char *str_org = str;

  if (TO_Nullstellig(Term)){
    str += IO_DruckeSymbolString(str, TO_TopSymbol(Term));
    if (TO_TermIstNichtVar(Term)) /* bei Konstanten Klammernpaar ausgeben */
      str += our_sprintf((str, "()"));
  }
  else
    do
      switch (TO_AnzahlTeilterme(Term)) {
      case 0:
        str += IO_DruckeSymbolString(str, TO_TopSymbol(Term));
        if (TO_TermIstNichtVar(Term)) /* bei Konstanten Klammernpaar ausgeben */
          str += our_sprintf((str, "()"));
        while (!(--TI_TopZaehler)) {
          str += our_sprintf((str, ")"));
          if (TI_PopUndLeer())
            return str - str_org;
        }
        str += our_sprintf((str, ","));
        Term = TO_Schwanz(Term);
        break;
      case 1:
        str += IO_DruckeSymbolString(str, TO_TopSymbol(Term));
        str += our_sprintf((str, "("));
        TI_PushOT(Term);
        Term = TO_ErsterTeilterm(Term);
        break;
      default:
        str += IO_DruckeSymbolString(str, TO_TopSymbol(Term));
        str += our_sprintf((str, "("));
        TI_PushOT(Term);
        Term = TO_Schwanz(Term);
        break;
      }
    while (TRUE);
  return str - str_org;
}

unsigned int IO_TermAusgabelaenge(UTermT Term)
/* gibt die Anzahl der Zeichen aus, die bei der Ausgabe von Term mittels IO_DruckeTermString ausgegeben wuerden */
{
  UTermT NULLTerm = TO_NULLTerm(Term);
  unsigned int Laenge = 0;
  do {
    SymbolT Symbol = TO_TopSymbol(Term),
            Stelligkeit = SO_Stelligkeit(Symbol);
    Laenge += SO_SymbIstVar(Symbol) ? 
      1+LaengeNatZahl(SO_VarNummer(Symbol))
      : 
      strlen(IO_SymbolNamen[Symbol])+2+(Stelligkeit>=2 ? Stelligkeit-1 : 0);
  } while (TO_PhysischUngleich(NULLTerm,Term = TO_Schwanz(Term)));
  return Laenge;
}

void IO_DruckeTermPaar(UTermT term1,const char *trenner,UTermT term2)
{
   IO_DruckeFlex("%t %s %t", term1, trenner, term2);
}

void IO_DruckeFaktum(TermT l, BOOLEAN orientierbar, TermT r)
{
  UTermT s,t;
  if (Z_UngleichungEntmantelt(&s,&t, l,r))
    IO_DruckeTermPaar(s,"?",t);
  else
    IO_DruckeTermPaar(l, orientierbar ? "->" : "=", r);
  IO_DruckeFlex("\n");
}


static void DruckeFolgenelement(FILE *Strom, unsigned int i)
{
  fprintf(Strom, "%d",TO_Stelligkeit(TI_Kopierstapel[i].Teilterm)+1-TI_Kopierstapel[i].Zaehler);
}

static void DruckeStellenfolgeStrom(FILE *Strom, PCL_SeitenT seite, UTermT Term, UTermT Teilterm)
{
  putc (seite == linkeSeite ? 'L' : 'R', Strom);
  if (TO_PhysischUngleich(Term,Teilterm)) {
    unsigned int i;
    TI_Push(Term);
    while (TO_PhysischUngleich(Term = TO_Schwanz(Term),Teilterm))
      if (TO_Nullstellig(Term))
        while (!(--TI_TopZaehler))
          TI_Pop();
      else
        TI_Push(Term);
    for (i=1; i<=TI_KSIndex; i++) {
      putc('.',Strom);
      DruckeFolgenelement(Strom,i);
    }
    TI_KSLeeren();
  }
  /* Gilt TO_PhysischGleich(Term,Teilterm), was fuer Term nullstellig stets der Fall
     sein muss, so handelt es sich um die Wurzelstelle, wofuer die leere Stellenfolge 
     ausgegeben wird. */
}

/* -------------------------------------------------------*/
/* -------------------------------------------------------*/
/* -------------- Hilfe-Text -----------------------------*/
/* -------------------------------------------------------*/
/* -------------------------------------------------------*/
extern char **environ;

static void DateiAusgabeOhnePager(const char *name)
{
   FILE *test;
   if (!(test = fopen(name, "r"))){
      fprintf(IO_Strom, "File '%s' not found.", name);
      fclose(test);
   }
   else{
      int c;
      
      while( (c = getc(test)) != EOF)
         putc(c, IO_Strom);
   }
}

void IO_DateiAusgabe(const char* name, BOOLEAN pager)
/* Ausgabe eines Hilfe-Textes (System-Beschreibung).*/
{
  if (pager){
     int PID, status, ww;
     
     PID = fork();                 /* fork erforderlich, weil execve sonst den WM-Prozess terminiert! */

     if (PID==0){
	const char *argv[3];
	
        
        argv[0] = getenv("PAGER");
        argv[1] = name;
        argv[2] = NULL;
        
        if (argv[0] == NULL)
           argv[0] = "more";
        
        if (execve(argv[0],  (char *const *) argv,environ) == -1){
           printf("ENVIRONMENT-VARIABLE PAGER not specified and 'more' not available.\n");
           perror(NULL);
           DateiAusgabeOhnePager(name);
           IO_WarteAufReturnMitText("RETURN");
           exit(-1);            /* Kindprozess terminieren */
        }
     }
     /* Jetzt Vaterprozess */
     ww=wait(&status);
  }
  else
     DateiAusgabeOhnePager(name);
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* ++++++++++++++ Statistik ++++++++++++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

static int Stelligkeit(unsigned long int zahl)
{
   int stellenzahl = 0;
   while (zahl > 0){
      zahl = zahl / 10;
      stellenzahl++;
   }
   return stellenzahl;
}

int IO_Platzverbrauch(unsigned long int zahl)
{
   int st = Stelligkeit(zahl);

   return st + ((st-1)/3);
}

int IO_Einruecktiefe(int bisherigeTiefe, unsigned long int maxEintrag)
/* Berechnet die Einruecktiefe, die fuer die folgenden Eintraege noetig ist.
   Dabei ist maxEintrag der groesste der folgenden Eintraege.
*/
{
   return (bisherigeTiefe + IO_Platzverbrauch(maxEintrag) +2);
}

static void DZFR(unsigned long int zahl, int *stellen, int stellenzahl)
{
  if (zahl == 0){
    *stellen = 0;
    return;
  }
  else if ((zahl / 10) == 0){
    fprintf(IO_Strom, "%lu", zahl);
    *stellen = 1;
  }
  else{
    DZFR(zahl/10, stellen, stellenzahl);
    /*
    if ( (((stellenzahl - *stellen)/3)*3 == stellenzahl-*stellen) && ((stellenzahl-*stellen) > 0))
    */
    if (((stellenzahl - *stellen) % 3 == 0) && ((stellenzahl-*stellen) > 0))
      fprintf(IO_Strom, ".");
    /*
      fprintf(IO_Strom, "%lu", zahl - (zahl/10)*10);
      */
    fprintf(IO_Strom, "%lu", zahl % 10);
    (*stellen)++;
  }
}
      

static void DruckeZahlFormatiert(unsigned long int zahl, int platz)
/* Gibt die Zahl zahl aus, setzt nach jeder dritten Ziffer (von hinten) einen Punkt,
   und fuellt die Ausgabe auf insgesamt platz Zeichen von links mit Leerzeichen auf.*/
{
  int stellenzahl = Stelligkeit(zahl), Leerzeichen, i;
  Leerzeichen = platz - IO_Platzverbrauch(zahl);
  for (i = 0; i < Leerzeichen; i++)
    fprintf(IO_Strom, " ");
  i = 0;
  DZFR(zahl,&i, stellenzahl);
}

static void DruckeStringFormatiert(const char *str, int platz)
/* Gibt die Zeichenkette str aus, nachdem die Ausgabe auf insgesamt platz Zeichen von links 
   mit Leerzeichen aufgefuellt wurde. */
{
  unsigned int i = platz-strlen(str);
  while (i--)
    fprintf(IO_Strom," ");
  fprintf(IO_Strom,"%s",str);
}


const int RechterRandDatum = 14;


/* -------------------------------------------------------*/
/* ------------- Statistik-Ausgaben ----------------------*/
/* -------------------------------------------------------*/

void IO_StatistikModusAufLatexSetzen(void)
{
   LatexAusgabeAktiviert = TRUE;
}

void IO_StatistikModusAufNormalSetzen(void)
{
   LatexAusgabeAktiviert = FALSE;
}

void IO_DruckeKurzUeberschrift(const char *text, char zeichen)
{
   if (!LatexAusgabeAktiviert){
      short i, rahmen;
      
      IO_LF();
      rahmen = 5; /*((IO_Ueberschriftenbreite - strlen(text) - 2) / 2);*/
      for (i=0; i<rahmen; i++){
         putc (zeichen, IO_Strom);
      }
      putc (' ', IO_Strom);
      fprintf(IO_Strom, text);
      putc (' ', IO_Strom);
      for (i=rahmen+strlen(text)+2; i<IO_Ueberschriftenbreite; i++){
         putc (zeichen, IO_Strom);
      }
      IO_LF();
      IO_LF();
   }
}  


   
void IO_DruckeStatistikZeile (const char *str,unsigned long Datum)
/* Gibt den String und dahinter das Datum aus, falls Datum nicht null ist.*/
{
   if (LatexAusgabeAktiviert){
     fprintf(IO_Strom, "%10lu & ", Datum);
     /*
      if (Datum == 0)
         fprintf(IO_Strom, " 0 & ");
      else{
         int stellenzahl = Stelligkeit(Datum), i;
         DZFR(Datum, &i, stellenzahl);
         fprintf(IO_Strom, " & ");
      }
     */
   }
   else if (Datum){
      DruckeZahlFormatiert(Datum, RechterRandDatum);
      fprintf(IO_Strom, " %s\n", str);
  }
}

void IO_DruckeStatistikZeileString(const char *str, const char *Datum)
/* Gibt den String und dahinter das Datum aus, falls Datum nicht null ist.*/
{
   if (LatexAusgabeAktiviert)
      fprintf(IO_Strom, " %s & ", Datum);
   else if (Datum){
      DruckeStringFormatiert(Datum, RechterRandDatum);
      fprintf(IO_Strom, " %s\n", str);
  }
}

void IO_DruckeStatistikZeileEingerueckt(const char *str, unsigned long Datum, int Einruecktiefe)
/* Druckt die Statistik-Zeile um Einruecktiefe 'Stellen' eingerueckt aus.
   Einruecktiefe gibt die Position des rechten Zeichens des Datums an.
*/
{
   if (LatexAusgabeAktiviert){
      if (Datum == 0)
         fprintf(IO_Strom, " 0 & ");
      else{
         int stellenzahl = Stelligkeit(Datum), i;
         DZFR(Datum, &i, stellenzahl);
         fprintf(IO_Strom, " & ");
      }
   }
   else if (Datum){
      DruckeZahlFormatiert(Datum, RechterRandDatum + Einruecktiefe);
      fprintf(IO_Strom, " %s\n", str);
   }
      
}

void IO_DruckeStatistikLeerZeile(void)
{
   if (!LatexAusgabeAktiviert)  fprintf(IO_Strom, "\n");
}

void IO_DruckeStatistikZeileAnteil(const char *str,unsigned long int Zaehler, unsigned long int Nenner)
/* Gibt den String und dahinter das Datum aus, falls der Nenner nicht null ist.*/
{
  if (LatexAusgabeAktiviert){
    if ((Zaehler == 0) || (Nenner == 0))
      fprintf(IO_Strom, " 0 & ");
    else{
      double Quotient = 100.0*(double)Zaehler/(double)Nenner;
      fprintf (IO_Strom, "%3.1f & \n", Quotient);
    }
  }
  else if (Nenner) {
    double Quotient = 100.0*(double)Zaehler/(double)Nenner;
    unsigned int Vorkomma = (Quotient = ((double)(unsigned int)(10.0*Quotient+0.5))/10.0),
                 Nachkomma = (unsigned int)(10.0*(Quotient-(double)Vorkomma)+0.5),
                 i = RechterRandDatum-LaengeNatZahl(Vorkomma)-LaengeNatZahl(Nachkomma)-1;
    while (i--)
      fprintf(IO_Strom," ");
    fprintf(IO_Strom,"%d,%d %s\n",Vorkomma,Nachkomma,str);
  }
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++++++++++ Sonstiges +++++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

unsigned int IO_SymbAusgabelaenge(SymbolT Symbol)
{
  return SO_SymbIstFkt(Symbol) ? strlen(IO_SymbolNamen[Symbol]) : (1+LaengeNatZahl(SO_VarNummer(Symbol)));
}

#ifdef CCE

static CCE_loggingT CCE_State = CCE_Undef;
static FILE *CCEFile;
static BOOLEAN do_CCE = FALSE;
static char *CCEFilename;
static char CCEInitialized = FALSE;

void IO_CCEinit(char *filename)
{
  CCEFilename = filename;
  do_CCE = TRUE;
}
static void CCE_OpenFile()
{
  SymbolT symbol;

  if (!do_CCE) return;
  
  if (!(CCEFile = fopen(CCEFilename, "w"))){
    our_error("Cannot open file for logging CCE!");
  }  
  CCEInitialized = TRUE;

  fprintf (CCEFile, "NAME\t%s\n\n", get_spec_name());
  fprintf (CCEFile, "MODE\tCCE\n\n");
  fprintf (CCEFile, "SORTS\t");
  printSorts (CCEFile);

  fprintf (CCEFile, "\nSIGNATURE\n");
  printSig (CCEFile);

  fprintf (CCEFile, "ORDERING\t");
  switch (PA_Ordering) {
  case ordering_kbo: fprintf (CCEFile, "KBO\n");break;
  case ordering_lpo: fprintf (CCEFile, "LPO\n");break;
  default:
    our_error("unexpected label in IO_CCEinit");
    break;
  }
  printOrdering (CCEFile);

  fprintf (CCEFile, "VARIABLES\n");
  fprintf (CCEFile, "\tx1");
  for (symbol = 2; symbol <= 10; symbol++) {
    fprintf (CCEFile, ", x%d", symbol);
  }
  fprintf (CCEFile, " : ");
  /* wir geben allen Variablen die gleiche Sorte, nicht schoen,
     aber anscheinend einzige Moeglichkeit, da Variablen keine
     Sorte haben - hier geht Information verloren */
  printSort (CCEFile, 1);
  fprintf (CCEFile, "\n\n");
}

void IO_CCElog(CCE_loggingT tp, TermT lhs, TermT rhs)
{
  if (!do_CCE) return;
  if (!CCEInitialized)
    CCE_OpenFile();
  if (CCE_State == tp){
    IO_DruckeFlexStrom(CCEFile, "\t%t = %t\n", lhs, rhs);
  }
  else {
    CCE_State = tp;
    switch (tp){
    case CCE_Add:    IO_DruckeFlexStrom(CCEFile, "\nADD\t%t = %t\n",    lhs, rhs); break;
    case CCE_Delete: IO_DruckeFlexStrom(CCEFile, "\nDELETE\t%t = %t\n", lhs, rhs); break;
    case CCE_Test:   IO_DruckeFlexStrom(CCEFile, "\nTEST\t%t = %t\n",   lhs, rhs); break;
    default:         our_error("unexpected value for CCE_loggingT in IO_CCElog!");
    }
  }
}
#endif
