(* $Id: config.ml,v 1.30.2.1 2004/06/15 21:11:54 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)

exception NO_SOLUTION



(*** console argument types ****)

type flag = {
  flag_name: string;
  mutable flag_value: bool;
}

type int_argument = {
  int_name: string;
  mutable int_value: int;
}

type string_argument = {
  string_name: string;
  mutable string_value: string;
}



(*** data ***)

(* bool command line arguments *)
let use_resolve = {
  flag_name = "-ur";
  flag_value = false;
}

let use_subsume = {
  flag_name = "-us";
  flag_value = false;
}

let use_compact = {
  flag_name = "-uc";
  flag_value = false;
}

let use_productivity = {
  flag_name = "-up";
  flag_value = false;
}

let use_plus_v = {
  flag_name = "-uv";
  flag_value = false;
}

let use_chronological_backtracking = {
  flag_name = "-ubj";
  flag_value = false;
}

let use_horn_detection = {
  flag_name = "-uh";
  flag_value = false;
}

let use_term_indexing = {
  flag_name = "-utx";
  flag_value = true;
}

let show_options = {
  flag_name = "-vo";
  flag_value = false;
}

let show_statistic = {
  flag_name = "-vst";
  flag_value = false;
}

let show_model = {
  flag_name = "-vm";
  flag_value = false;
}

let show_derivation = {
  flag_name = "-vd";
  flag_value = false;
}

let show_context_unifier = {
  flag_name = "-vcu";
  flag_value = false;
}

let show_derivation_context = {
  flag_name = "-vdx";
  flag_value = false;
}

let show_computed_assert_candidates = {
  flag_name = "-vca";
  flag_value = false;
}

let show_computed_split_candidates = {
  flag_name = "-vcs";
  flag_value = false;
}




(* int command line arguments *)
let initial_depth_bound = {
  int_name = "-b";
  int_value = 2;
}

let verbosity = {
  int_name = "-v";
  int_value = 1;
}


(* string command line arguments *)
let print_model_to_file = {
  string_name = "-pmf";
  string_value = "";
}

(* string command line arguments *)
let derivation_to_dot_file_name = {
  string_name = "-pdf";
  string_value = "";
}



let problem_file_name =
  ref ""

(* should a special treatment for horn clauses be used?
   only evaluated if horn_detection is true. *)
let horn =
  ref false

(* current depth bound *)
let current_depth_bound =
  ref 2

(* previous depth bound *)
let prev_depth_bound =
  ref (-1)


(* lowest depth of candidates dropped because of the depth bound *)
let lowest_depth_cut =
  ref None





(*** access functions ***)



(* verbosity level sets the show options *)
let verbosity_flags = [
  show_options; show_statistic;
  show_model;
  show_derivation; show_context_unifier; show_derivation_context; 
  show_computed_assert_candidates; show_computed_split_candidates
]

let verbosity_level_0 = []

let verbosity_level_1 = [
  show_statistic
]

let verbosity_level_2 = [
  show_statistic; show_model
]

let verbosity_level_3 = [
  show_statistic; show_model; show_derivation
]

let verbosity_level_4 = [
  show_statistic; show_model; show_derivation; show_context_unifier
]

let verbosity_levels = 
  [| verbosity_level_0; verbosity_level_1; verbosity_level_2; verbosity_level_3; verbosity_level_4 |]

let flags_for_verbosity_level (level: int) : flag list =
  if level >= Array.length verbosity_levels then
    verbosity_levels.(Array.length verbosity_levels - 1)

  else if level < 0 then
    verbosity_levels.(0)

  else
    verbosity_levels.(level)


let evaluate_verbosity () : unit =
  List.iter
    (fun flag ->
       flag.flag_value <- false;
    )
    verbosity_flags;

  List.iter
    (fun flag ->
       flag.flag_value <- true;
    )
    (flags_for_verbosity_level verbosity.int_value)



let set_verbosity (value: int) : unit =
  verbosity.int_value <- value;

  evaluate_verbosity ()
  

let get_use_resolve () : bool =
   use_resolve.flag_value

let set_use_resolve (value: bool) : unit =
   use_resolve.flag_value <- value

let get_use_subsume () : bool =
   use_subsume.flag_value

let set_use_subsume (value: bool) : unit =
   use_subsume.flag_value <- value

let get_use_compact () : bool =
   use_compact.flag_value

let set_use_compact (value: bool) : unit =
   use_compact.flag_value <- value

let get_use_productivity () : bool =
   use_productivity.flag_value

let set_use_productivity (value: bool) : unit =
   use_productivity.flag_value <- value

let get_use_plus_v () : bool =
   use_plus_v.flag_value

let set_use_plus_v (value: bool) : unit =
   use_plus_v.flag_value <- value

let get_use_chronological_backtracking () : bool =
   use_chronological_backtracking.flag_value

let set_use_chronological_backtracking (value: bool) : unit =
   use_productivity.flag_value <- value

let set_use_chronological_backtracking (value: bool) : unit =
   use_chronological_backtracking.flag_value <- value

let get_use_horn_detection () : bool =
   use_horn_detection.flag_value

let set_use_horn_detection (value: bool) : unit =
   use_horn_detection.flag_value <- value

let get_use_term_indexing () : bool =
   use_term_indexing.flag_value

let set_use_term_indexing (value: bool) : unit =
   use_term_indexing.flag_value <- value

let get_show_options () : bool =
   show_options.flag_value

let get_show_statistic () : bool =
   show_statistic.flag_value

let get_show_model () : bool =
   show_model.flag_value

let get_print_model_to_file () : string =
  print_model_to_file.string_value

let set_print_model_to_file (file_name: string) : unit =
  print_model_to_file.string_value <- file_name

let get_show_derivation () : bool =
   show_derivation.flag_value

let set_show_derivation (value: bool) : unit =
   show_derivation.flag_value <- value

let get_derivation_to_dot_file_name () : string =
  derivation_to_dot_file_name.string_value

let set_derivation_to_dot_file_name (file_name: string) : unit =
  derivation_to_dot_file_name.string_value <- file_name

let get_derivation_to_dot_file () : bool =
  get_use_chronological_backtracking ()
  &&
  derivation_to_dot_file_name.string_value <> ""

let get_show_context_unifier () : bool =
   show_context_unifier.flag_value

let set_show_context_unifier (value: bool) : unit =
   show_context_unifier.flag_value <- value

let get_show_derivation_context () : bool =
   show_derivation_context.flag_value

let set_show_derivation_context (value: bool) : unit =
   show_derivation_context.flag_value <- value

let get_show_computed_assert_candidates () : bool =
   show_computed_assert_candidates.flag_value

let set_show_computed_assert_candidates (value: bool) : unit =
   show_computed_assert_candidates.flag_value <- value

let get_show_computed_split_candidates () : bool =
   show_computed_split_candidates.flag_value

let set_show_computed_split_candidates (value: bool) : unit =
   show_computed_split_candidates.flag_value <- value


let get_problem_file_name () : string =
  !problem_file_name

let set_problem_file_name (name: string) : unit =
  problem_file_name := name





let get_horn () : bool =
  !horn

let set_horn () : unit =
  horn := true;
  use_chronological_backtracking.flag_value <- true












(*** representation ***)



let flag_to_string (flag: flag) : string =
  flag.flag_name ^ ": " ^ string_of_bool (flag.flag_value) ^ "\n"

let int_argument_to_string (argument: int_argument) : string =
  argument.int_name ^ ": " ^ string_of_int (argument.int_value) ^ "\n"

let string_argument_to_string (argument: string_argument) : string =
  if argument.string_value != "" then
    argument.string_name ^ ": " ^ argument.string_value ^ "\n"
  else
    ""

let to_string () : string =
  (* to keep the order we can not just dump the Hashtable *)
  flag_to_string use_resolve
  ^ flag_to_string use_subsume
  ^ flag_to_string use_compact
  ^ flag_to_string use_productivity
  ^ flag_to_string use_plus_v
  ^ flag_to_string use_chronological_backtracking
  ^ flag_to_string use_horn_detection
  ^ flag_to_string use_term_indexing
  ^ int_argument_to_string initial_depth_bound

  ^ flag_to_string show_statistic
  ^ flag_to_string show_model
  ^ string_argument_to_string print_model_to_file
  ^ flag_to_string show_derivation
  ^ string_argument_to_string derivation_to_dot_file_name
  ^ flag_to_string show_context_unifier
  ^ flag_to_string show_derivation_context
  ^ flag_to_string show_computed_assert_candidates
  ^ flag_to_string show_computed_split_candidates
  ^ flag_to_string show_options

  ^ int_argument_to_string verbosity






(*** evaluation of the command line argumens ***)

let usage_msg =
  "Usage: darwin [OPTIONS] PROBLEM_FILE[.tme]\n"
  ^ "Example: darwin -v 3 PUZ37-3.tme\n"
  ^ "Options:"



let eval_problem_file_name (argument: string) : unit =
  problem_file_name := argument
  

let create_spec_entry_flag (flag: flag) (help: string) =
  (
    flag.flag_name,
    Arg.Bool (fun value -> flag.flag_value <- value),
    "[" ^ string_of_bool (flag.flag_value) ^ "]: " ^ help
  )

let create_spec_entry_int (argument: int_argument) (help: string) =
  (
    argument.int_name,
    Arg.Int (fun value -> argument.int_value <- value),
    "[" ^ string_of_int (argument.int_value) ^ "]: " ^ help
  )

let create_spec_entry_string (argument: string_argument) (help: string) =
  (
    argument.string_name,
    Arg.String (fun value -> argument.string_value <- value),
    "[" ^ argument.string_value ^ "]: " ^ help
  )

let create_spec_entry_verbosity () =

  let level_to_string level description =
    "      " ^ string_of_int level
    ^ " -> " ^ description
    ^ " (" ^ String.concat ", " (List.map (fun flag -> flag.flag_name) (flags_for_verbosity_level level))
    ^ ")" ^ "\n"
  in
  (
    verbosity.int_name,
    Arg.Int set_verbosity,
    ("[" ^ string_of_int (verbosity.int_value) ^ "]: " ^ "\n"
     ^ level_to_string 0 "show only the derivation result"
     ^ level_to_string 1 "... and show the derivation statistic "
     ^ level_to_string 2 "... and show the found model"
     ^ level_to_string 3 "... and show the derivation steps"
     ^ level_to_string 4 "... and show the context unifiers underlying the derivation steps"
    )
  )


let create_spec_list () = [
  create_spec_entry_flag use_resolve "use the resolve inference rule";
  create_spec_entry_flag use_subsume "use the subsume inference rule";
  create_spec_entry_flag use_compact "use the compact inference rule";
  create_spec_entry_flag use_productivity "split only on productive remainders";
  create_spec_entry_flag use_plus_v "use +v instead of -v as the default interpretation";
  create_spec_entry_flag use_chronological_backtracking "use backjumping instead of dynamic backtracking (always used for Horn)";
  create_spec_entry_flag use_horn_detection "use an optimization on Horn problems (otherwise ignored)";
  create_spec_entry_flag use_term_indexing "use term indexing for context checks";
  create_spec_entry_int initial_depth_bound "the initial term depth bound";
  create_spec_entry_flag show_statistic "print the derivation statistic after the derivation";
  create_spec_entry_flag show_model "print the model found for a satisfiable problem";
  create_spec_entry_string print_model_to_file "print the model found for a satisfiable problem to a file";
  create_spec_entry_flag show_derivation "print the derivation";
  create_spec_entry_string derivation_to_dot_file_name "print the derivation/proof in the graphviz dot format to a file (backjumping only).";
  create_spec_entry_flag show_context_unifier "print the context unifier underlying each derivation step";
  create_spec_entry_flag show_derivation_context "print the context at each derivation step";
  create_spec_entry_flag show_computed_assert_candidates "print each computed assert candidate";
  create_spec_entry_flag show_computed_split_candidates "print each computed split candidate (remainder)";
  create_spec_entry_flag show_options "print the options settings";
  create_spec_entry_verbosity ();
]









(*** depth bound *)

let set_initial_depth_bound (value:int) : unit =
  initial_depth_bound.int_value <- value

let get_current_depth_bound () : int =
  !current_depth_bound

let set_current_depth_bound (depth: int) : unit =
  current_depth_bound := depth

let get_prev_depth_bound () : int =
  !prev_depth_bound

let set_prev_depth_bound (depth: int) : unit =
  prev_depth_bound := depth


let check_depth_bound (depth: int) : bool =
  if depth > !current_depth_bound then begin
    begin
      match !lowest_depth_cut with
	| None ->
	    lowest_depth_cut := Some depth
	      
	| Some depth_cut ->
	    if depth_cut > depth then
	      lowest_depth_cut := Some depth
    end;
    false
  end

  else
    true


let update_depth_bound () : bool =
  match !lowest_depth_cut with
    | None ->
	false

    | Some depth_cut ->
	set_prev_depth_bound (get_current_depth_bound ());
	set_current_depth_bound depth_cut;
	lowest_depth_cut := None;
	true

let reset_depth_bound () : unit =
  current_depth_bound := initial_depth_bound.int_value;
  prev_depth_bound := -1
  



(*** initialization ***)

let set_defaults () : unit =
  use_resolve.flag_value <- true;
  use_subsume.flag_value <- true;
  use_compact.flag_value <- true;
  use_productivity.flag_value <- true;
  use_plus_v.flag_value <- false;
  use_chronological_backtracking.flag_value <- false;
  use_horn_detection.flag_value <- true;
  use_term_indexing.flag_value <- true;
  
  initial_depth_bound.int_value <- 2;
  verbosity.int_value <- 1;
  evaluate_verbosity ();

  horn := false;
  current_depth_bound := 2;
  prev_depth_bound := -1;
  lowest_depth_cut := None




let init () =
  set_defaults ();

  let spec_list =
    create_spec_list ()
  in
    Arg.parse spec_list eval_problem_file_name usage_msg;
    current_depth_bound := initial_depth_bound.int_value


let print_usage_message () : unit =
  Arg.usage (create_spec_list ()) usage_msg



let debug =
  false






(* startup initialization *)
let _ =
  init ()
