(* $Id: context_unifier_check.mli,v 1.3 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** basic types for and operations on context unifiers

  contains also the types for the search for context unifiers,
  but this is actually done in {!Context_unifier_space}.
*)


(** {6 Types} *)

type var = Var.var
type literal = Term.literal
type clause = Term.clause
type subst = Subst.subst
type state = State.state
type explanation = State.explanation
type packed_subst = Unification.packed_subst
type context_unifier_space = Context_unifier.context_unifier_space
type context_partners = Context_unifier.context_partners
type raw_context_unifier = Context_unifier.raw_context_unifier



(** {6 Functions} *)


(** {2 Checks} *)
  
(** raises {!Context_unifier.CLOSED} on an empty remainder. *)
val check_close: subst -> context_unifier_space -> context_partners -> unit


(** returns the [raw_context_unifier] representation of the assert literal,
  if [subst] is a valid 'assert context unifier'
  and obeys the current depth bound (see {!Config.check_depth_bound}).
  I.e. exactly one context_partner in [context_partners]
  is {!Context_unifier.assert_partner}
  and considered to be the gap in the context unifier.
  Then the assert rule criteria of the model evolution calculus must apply
  to the remaining context unifier (empty remainder)
  and the assert literal (parameter free) *)
val check_assert: subst -> context_unifier_space -> context_partners -> raw_context_unifier option

(** returns the [raw_context_unifier] representation
  for the context unifier [subst] between
  the input clause [context_unifier_space]
  and the context literals [context_partners],
  if the depth of the shallowest remainder literal
  obeys the current depth bound (see {!Config.check_depth_bound}).
  Does not check if the remainder is empty,
  this must be done separately with {!Context_unifier_check.check_close} *)
val check_split: subst -> context_unifier_space -> context_partners -> raw_context_unifier option
