(* $Id: context_unifier_space.mli,v 1.14 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** computes context unifiers for a clause

  Stores for a clause all current partial context unifiers against the context.

  Registers all literals at {!Problem_literals} and
  extends all partial context unifiers
  whenever a new context literal is added.
  
  Completed context unifiers are checked if valid
  for {e Close}, {e Assert}, and {e Split}.

  Calls the callback functions of {!Context_unifier.context_unifier_space}
  for new candidates.
*)

(** {6 Types} *)

type literal = Term.literal
type clause = Term.clause
type context_unifier_space = Context_unifier.context_unifier_space
type raw_context_unifier = Context_unifier.raw_context_unifier



(** {6 Functions} *)


(** {2 Creation} *)

(** [create: clause process_assert_candidate process_split_candidate]
  - creates a {!Context_unifier_space.context_unifier_space} for a {!Term.clause}
    using [process_assert_candidate] resp. [process_split_candidate]
    as callback functions when assert resp. split candidates are computed
  - registers the literals at {!Problem_literals}
  - and returns a literal to assert if this is a unit clause
*)
val create: clause ->
  (raw_context_unifier -> unit) ->
  (raw_context_unifier -> unit) ->
  context_unifier_space * literal option



(** {2 Backtracking} *)

(** dependend on {!Problem_literals.backtrack} *)
val backtrack: context_unifier_space -> unit
