(* $Id: derivation_to_dot.mli,v 1.2 2004/06/13 18:22:21 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** derivation visualization

  prints the proof in the {{:http://www.graphviz.org/cgi-bin/man?dot} graphviz dot format}
  to a file.

  inference steps are represented by different node colors:
  - Assert -> blue
  - Unit Split (Split on a unit remainder) -> yellow
  - Split -> green
  - Close -> red

  only works for chronological backtracking
*)


(** {6 Types} *)

type state = State.state
type literal = Term.literal



(** {6 Functions} *)


(** {2 Initialization/Finalization} *)

(** [init file_name]
  starts a new derivation visualization to the file named [file_name] *)
val init: string -> unit

(** finishes a proof *)
val finish: unit -> unit




(** {2 Compact Visualization} *)

(** builds a graph where the nodes contain the literal added to the context. *)

val print_assert: literal ->  unit

val print_unit_split: literal -> unit

(** must be called in the state created by the left split *)
val print_split_left: literal -> unit

(** must be called in the state created by the right split *)
val print_split_right: literal -> unit

val print_close: unit -> unit




(** {2 Detailed Visualization} *)

(** builds a graph where
  - the nodes contain the context unifier underlying the Split/Assert
  - the edges are annotated with the literal added to the context
*)

val print_assert_detailed: literal -> string -> unit

val print_unit_split_detailed: literal -> string -> unit

(** must be called in the state created by the left split *)
val print_split_left_detailed: literal -> string -> unit

(** must be called in the state created by the right split *)
val print_split_right_detailed: literal -> unit

val print_close_detailed: string -> unit
