(* $Id: selection.mli,v 1.12 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** selection of a split or assert literal

  manages the computation of candidates for {e Assert} and {e Split}
  with {!Context_unifier} and {!Context_unifier_space}
  after each change of the context. *)


(** {6 Types} *)

type clause = Term.clause
type explanation = State.explanation
type selected_assert = Selection_types.selected_assert
type selected_split = Selection_types.selected_split


(** the {e Selection} data structure *)
type selection



(** type of selected candidate *)
type selected =
  | Assert of selected_assert
      (** assertion of an assert literal *)
  | Assert_split of selected_assert
      (** assertion of a unit remainder *)
  | Split of selected_split
      (** split on a remainder literal *)


(** no further selection possible - satisfiability detected *)
exception SATISFIABLE



(** {6 Functions} *)


(** [create input_clauses] creates a new [selection]
  over the input clauses. *)
val create: clause list -> selection


(** to be called when a new term is added to the context.
  Updates the set of assert and split candidates.
  May raise {!Context_unifier.CLOSED}. *)
val add: selection -> Context.element -> unit


(** selects a literal to add to the context.
  Raises {!Selection.SATISFIABLE} if no valid candidate literal exists. *)
val select: selection -> selected



(** removes candidates based on invalid states. *)
val backtrack: selection -> explanation -> unit
