(* $Id: selection_lookahead.mli,v 1.3.2.2 2004/06/14 05:28:15 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** contradiction lookahead for assert candidates

  all assert (and unit split) candidates added are kept in a set.
  other assert (and unit split) candidates can be checked
  for contradiction with the lookahead set.

  I.e. if there are two contradictory candidates
  then one is chosen, added to the context,
  and thus {e Close} is triggered as soon as possible.

  Otherwise - depending on the selection heuristic and bad luck --
  those contradictory candidates might not be chosen at all
  in a reasonable amount of time *)


(** {6 Types} *)

type literal = Term.literal
type raw_context_unifier = Context_unifier.raw_context_unifier


(** {6 Functions} *)

(** clears the lookahead set *)
val init: unit -> unit


(** [add candidate_literal candidate_context_unifier]
  adds an mandatory candidate with its context unifier to the lookahead set *)
val add: literal -> raw_context_unifier -> unit

(** [remove candidate_literal candidate_context_unifier]
  removes an assert candidate from the lookahead set.
  only valid candidates should be removed this way,
  invalid candidates should be handled by {!Selection_lookahead.backtrack} *)
val remove: literal -> raw_context_unifier -> unit

(** [check candidate_literal candidate_context_unifier]
  checks if the candidate is contradictory with the lookahead set *)
val check: literal -> raw_context_unifier -> bool


(** removes invalid candidates from the lookahead set *)
val backtrack: unit -> unit
