(* $Id: selection_split.mli,v 1.10 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** selection of a split literal

  stores all split candidates (remainders) and selects literals to split on. *)


(** {6 Types} *)

type literal = Term.literal
type explanation = State.explanation
type raw_context_unifier = Context_unifier.raw_context_unifier
type selected_assert = Selection_types.selected_assert
type selected_split = Selection_types.selected_split

type candidates

type selected =
  | Assert of selected_assert
      (** split literal from a unit remainder ->
	  can be treated like an assert literal *)
  | Split of selected_split
      (** ordinary split literal *)



(** {6 Functions} *)

(** creates an empty candidate repository *)
val create: unit -> candidates

(** adds a split candidate to the repository *)
val add: candidates -> raw_context_unifier -> unit

(** selects (priority based) a split literal from the repository.
  Ignores already used candidates
  and candidates
  - subsumed by
  - contradictory with
  - not productive with
  the context.

  Returns None if no split candidate exists. *)
val select: candidates -> selected option

(** removes the split candidates based on invalid states.
  readds previously used but now from the context removed candidates *)
val backtrack: candidates -> unit


(** the remainder of the context unifier as a clause. *)
val remainder_literals_to_string: raw_context_unifier -> string
