(* $Id: statistic.ml,v 1.5 2004/06/13 18:22:22 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


type counter = Counter.counter

type attribute = {
  name: string;
  mutable value: counter;
}


(*** data ***)
let close = {
  name = "close";
  value = Counter.create ();
}

let assert_ = {
  name = "close";
  value = Counter.create ();
}

let split = {
  name = "split";
  value = Counter.create ();
}

let resolve = {
  name = "resolve";
  value = Counter.create ();
}

let subsume = {
  name = "subsume";
  value = Counter.create ();
}

let compact = {
  name = "compact";
  value = Counter.create ();
}

let filtered_by_productivity = {
  name = "filtered_by_productivity";
  value = Counter.create ();
}

let computed_assert_candidates = { 
  name = "computed_assert_candidates";
  value = Counter.create ();
}

let computed_split_candidates = { 
  name = "computed_split_candidates";
  value = Counter.create ();
}

let max_context_size = {
  name = "max_context_size";
  value = Counter.create ();
}

let debug = {
  name = "debug";
  value = Counter.create ();
}


let get (attribute: attribute) : int =
  Counter.value attribute.value

let inc (attribute: attribute) : unit =
  Counter.inc attribute.value

let set (attribute: attribute) (value: int) : unit =
  Counter.set attribute.value value





let inc_close () : unit =
  inc close

let get_close () : int =
  get close


let inc_assert () : unit =
  inc assert_

let get_assert () : int =
  get assert_


let inc_split () : unit =
  inc split

let get_split () : int =
  get split


let inc_resolve () : unit =
  inc resolve

let get_resolve () : int =
  get resolve


let inc_subsume () : unit =
  inc subsume

let get_subsume () : int =
  get subsume


let inc_compact () : unit =
  inc compact

let get_compact () : int =
  get compact


let inc_filtered_by_productivity () : unit =
  inc filtered_by_productivity

let get_filtered_by_productivity () : int =
  get filtered_by_productivity


let inc_computed_assert_candidates () : unit =
  inc computed_assert_candidates

let get_computed_assert_candidates () : int =
  get computed_assert_candidates


let inc_computed_split_candidates () : unit =
  inc computed_split_candidates

let get_computed_split_candidates () : int =
  get computed_split_candidates


let set_max_context_size (value: int) : unit =
  set max_context_size value

let get_max_context_size () : int =
  get max_context_size


let inc_debug () : unit =
  inc debug

let get_debug () : int =
  get debug


(*** print ***)
let to_string () : string =
  "Close: " ^ string_of_int (get close) ^ "\n"
  ^ "Assert: " ^ string_of_int (get assert_) ^ "\n"
  ^ "Split: " ^ string_of_int (get split) ^ "\n"
  ^ "Resolve: " ^ string_of_int (get resolve) ^ "\n"
  ^ "Subsume: " ^ string_of_int (get subsume) ^ "\n"
  ^ "Compact: " ^ string_of_int (get compact) ^ "\n"
  ^ "Productivity filtered: " ^ string_of_int (get filtered_by_productivity) ^ "\n"
  ^ "Computed Assert Candidates: " ^ string_of_int (get computed_assert_candidates) ^ "\n"
  ^ "Computed Split Candidates: " ^ string_of_int (get computed_split_candidates) ^ "\n"
  ^ "Maximum Context Size: " ^ string_of_int (get max_context_size) ^ "\n"
  ^ "Depth Bound: " ^ string_of_int (Config.get_current_depth_bound ()) ^ "\n"
  ^ "Horn optimization: " ^ string_of_bool (Config.get_horn ()) ^ "\n"
  ^ "Debug: " ^ string_of_int (get debug) ^ "\n"



(*** initialization ***)

let restart () =
  set close 0;
  set assert_ 0;
  set split 0;
  set resolve 0;
  set subsume 0;
  set compact 0;
  set filtered_by_productivity 0;
  set computed_assert_candidates 0;
  set computed_split_candidates 0;
  set max_context_size 0;
  set debug 0
