(* $Id: substitution_tree.mli,v 1.6 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** term indexing with a substitution tree

  A substitution tree stores only substitutions.
  Indexing is achieved by building a tree
  whose nodes are generalizations over similar terms
  and whose edges are substitutions instantiating towards the terms.

  E.g.:
  
  [f(a, g(a, x), g(y, y))] and [f(b, g(b, b), g(x, y))]

  can be abstracted to the node [f(x, g(x, y), g(z, u))]
  with the edges [x -> a; u -> z]
  and [x -> b; y -> b].

  (See P. Graf, {{:http://citeseer.ist.psu.edu/graf94substitution.html}Substitution Tree Indexing})

  Term variants may be inserted several times with different data
  with no loss of information.

  Currently, only terms with the same top symbol may be inserted.
  So adding f(x) and g(x) is not allowed.
  A substitution tree is an index exactly for one predicate/function symbol.
*)

(** functor to create a substitution tree. *)
module Make (DataType: Term_indexing.DataType):
  Term_indexing.Index with type data = DataType.t
