(* $Id: symbol.mli,v 1.6 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** constant and function symbols

  A symbol consists of
  - its unique name (eg {e p}, {e f} or {e a})
  - its arity (0 for a constant)
  - and its unique id provided by this module.

  All symbols are irrevocably registered in a global symbol table.
  A symbol may be replaced by a new symbol with the same name
  but a different arity.
  (Thus running several problems in interactive mode is possible.) *)


(** {6 Types} *)


(** a constant or function symbol *)
type symbol



(** {6 Functions} *)



(** {2 Creation} *)

(** [Symbol.create_symbol name arity] registers a symbol and returns it.
  Replaces an old entry if the symbol is already registered
  with a different arity. *)
val create_symbol: string -> int -> symbol


(** creates a unique fresh constant.
  Can e.g. be used for skolemization. *)
val create_constant: unit -> symbol


(** {2 Decomposition} *)

val id_of_symbol: symbol -> int

val name_of_symbol: symbol -> string

val arity_of_symbol: symbol -> int


(** {2 Comparison} *)

val equal: symbol -> symbol -> bool



(** {2 String Representation} *)

(** the name of the symbol *)
val to_string: symbol -> string
