(* $Id: term_indexing.ml,v 1.5 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


module type DataType = 
sig
  type t

  val is_equal: t -> t -> bool
  val to_string: t -> string
end



module type Index =
  sig
    type t

    type data
    type term = Term.term
    type subst = Subst.subst
      
    val create: unit -> t
    val clear: t -> unit
      
    val add: t -> term -> data -> unit
      
    val remove: t -> term -> data option -> bool 
     
    val iter: (term -> data -> unit) -> t -> unit
      
    val find_all_variants: t -> term -> data list
    val find_all_generalizations: p_preserving:bool -> t -> term -> data list
    val find_all_unifiable: p_preserving:bool -> t -> term -> data list
    val find_all_instances: p_preserving:bool -> t -> term -> data list

    val find_generalization: p_preserving:bool -> t -> term -> data option
    val find_unifiable: p_preserving:bool -> t -> term -> data option
    val find_instance: p_preserving:bool -> t -> term -> data option
    val find_non_productive: t -> term -> term -> data option
    val find_non_msg: t -> term -> term -> data option

    val to_string: t -> string
  end
