(* $Id: term_quality.mli,v 1.9 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** qualities for term comparisons *)


(** {6 Types} *)


type literal = Term.literal
type subst = Subst.subst

(** {6 Functions} *)


(** {2 Pureness } *)

(** does the term contain parameters? *)
val is_parametric: literal -> bool

(** compare_pureness is_first_literal_parametric is_second_literal_parametric
  prefers universal to parametric literals *)
val cmp_pureness: bool -> bool -> int



(** {2 Depth } *)

(** the deepest branch if the term is seen as a tree.

  Examples:
  - a -> 0
  - f(a, b) -> 1
  - f(a, g(b)) -> 2 *)
val depth_of_literal : literal -> int

(** [depth_of_literal_subst subst literal offset]
  is like {!Term_quality.depth_of_literal},
  but variables in [literal] are replaced by their bindings
  in [subst] *)
val depth_of_literal_subst : subst -> literal -> int -> int





(** {2 Weight } *)

(** the number of symbol occurences in the literal.

  Examples:
  - a -> 1
  - f(a, a) -> 3
  - f(x, g(b)) -> 3 *)
val weight_of_literal: literal -> int
