(* $Id: tools.mli,v 1.5 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(** general functions *)

(** {2 List} *)

(** [list_remove_first predicate list]
   returns [list] without the first element meeting the [predicate].
   returns the original list if no element meets the condition *)
val list_remove_first: ('a -> bool) -> 'a list -> 'a list

(** [lists_unordered_equal equal list1 list2].
  are [list1] and [list2] isomorph with respect to [equal]
  if the order of the elements is not taken into account? *)
val lists_unordered_equal: ('a -> 'b -> bool) -> 'a list -> 'b list -> bool

(** [lists_merge equal list1 list2]
  adds all elements which are not already in [list1] from [list2] to [list1]. *)
val lists_merge: ('a -> 'a -> bool) -> 'a list -> 'a list -> 'a list

(** [lists_shared equal list1 list2].
  do the [list1] and [list2] have at least one element in common? *)
val do_lists_intersect: ('a -> 'a -> bool) -> 'a list -> 'a list -> bool

(** [lists_shared equal lists]
   finds all elements which are in more than one list *)
val lists_shared: ('a -> 'a -> bool) -> 'a list list -> 'a list


(** [mapping_extend mapping key_equal value_equal key value].
  [mapping] is considered to be a bijective mapping from [key]s to [value]s.
  [mapping_extend] tries to extend [mapping] by the mapping [key] -> [value].
  If [key] -> [value] is already part of [mapping], [mapping] remains unchanged.
  @raise Exit on failure (i.e. when key or map are already in [mapping]). *)
val mapping_extend: ('a * 'b) list -> ('a -> 'a -> bool) -> ('b -> 'b -> bool) -> 'a -> 'b -> ('a * 'b) list



(** {2 Array} *)

(** [array_fold2 f init array1 array2]
  folds on two arrays.
  @raise Invalid_argument "Tools.array_fold2"
  if [array1] and [array2] are of different length *)
val array_fold2: ('a -> 'b -> 'c -> 'a) -> 'a -> 'b array -> 'c array -> 'a

(** like [List.exists] for an array *)
val array_exists: ('a -> bool) -> 'a array -> bool
