(* $Id: var.ml,v 1.4 2004/06/13 18:22:23 alexf Exp $ *)

(*
This file is part of the first order theorem prover Darwin
Copyright (C) 2004  The University of Iowa
                    Universitaet Koblenz-Landau 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*)


(* indicator variables introduce new preformance/memory overhead.
   using bools for universal/indicator is
   approx. equal in terms of performance and worse in terms of memory *)
type var =
  | Universal of int
  | Parametric of int
  | Universal_Indicator of int
  | Parametric_Indicator of int



let create_universal (id: int) : var =
  Universal id

let create_parametric (id: int) : var =
  Parametric id

let create_universal_indicator (id: int) : var =
  Universal_Indicator id

let create_parametric_indicator (id: int) : var =
  Parametric_Indicator id

let clone_renumbered (var: var) (id: int) : var = 
  match var with
    | Universal _ ->
	create_universal id

    | Parametric _ ->
	create_parametric id

    | Universal_Indicator _ ->
	create_universal_indicator id

    | Parametric_Indicator _ ->
	create_parametric_indicator id


let clone_as_indicator (var: var) : var = 
  match var with
    | Universal id ->
	create_universal_indicator id

    | Parametric id ->
	create_parametric_indicator id

    | Universal_Indicator _
    | Parametric_Indicator _ ->
	var

let clone_as_non_indicator (var: var) : var = 
  match var with
    | Universal _
    | Parametric _ ->
	var

    | Universal_Indicator id ->
	create_universal id

    | Parametric_Indicator id ->
	create_parametric id


let is_universal (var: var) : bool =
  match var with
    | Universal _
    | Universal_Indicator _ ->
	true

    | Parametric _
    | Parametric_Indicator _ ->
	false
      
let is_parametric (var: var) : bool =
  not (is_universal var)

let is_indicator (var: var) : bool =
  match var with
    | Universal _
    | Parametric _ ->
	false

    | Parametric_Indicator _
    | Universal_Indicator _ ->
	true


let equal (var1: var) (var2: var) : bool =
  match var1, var2 with
    | Universal id1, Universal id2
    | Parametric id1, Parametric id2
    | Universal_Indicator id1, Universal_Indicator id2
    | Parametric_Indicator id1, Parametric_Indicator id2
	when id1 = id2 ->
	true

    | _ ->
	false


let id_of_var (var: var) : int = 
  match var with
    | Universal id
    | Parametric id
    | Parametric_Indicator id
    | Universal_Indicator id ->
	id


let to_string (var: var) : string = 
  match var with
    | Universal id ->
	"_" ^ string_of_int id

    | Parametric id ->
	"=" ^ string_of_int id

    | Universal_Indicator id ->
	"*_" ^ string_of_int id

    | Parametric_Indicator id ->
	"*=" ^ string_of_int id
