/*------------------------------------------------------------------*/
/*  tptpi.c
/*
/*
/*  A routine that converts tptp-formated theorems to
/*  theo-formated theorems.  Note that all the functions
/*  in this file only call other functions in this file
/*  as shown below.  CompileTptp is called from main().
/*
/*  functions:
/*      ReadTptp()  ................. #1
/*      DetermineRating()............ #1a
/*      ConvertTptp() ............... #2
/*      InsertChar()................. #3
/*      DeleteChar()................. #4
/*      CheckClauseType()............ #5
/*      CompareStrings()............. #6
/*      SetPtrToNextLine()........... #7
/*      AssignCostToClause()......... #8
/*      DumpClause()................. #9
/*      OutString()..................#10
/*      FinalSort()..................#11
/*      CompileTptp()................#12
/*
/*  relation of functions:
/*      CompileTptp()
/*         ReadTptp()
/*         ConvertTptp()
/*            InsertChar()
/*            DeleteChar()
/*            SetPtrToNextLine()
/*            CheckClauseType()
/*               CompareStrings()
/*         AssignCostToClause()
/*         FinalSort()
/*            DumpClause()
/*            OutString()
/*
/*------------------------------------------------------------------*/


#include <stdio.h>
#include <strings.h>
#include <ctype.h>

#include "const.h"

/*-----------------------------------*/
/* externally defined functions      */
/*-----------------------------------*/

extern void print0();

/*-----------------------------------*/
/* externally defined constants      */
/*-----------------------------------*/

extern char theorem_name[];
extern int slash_char;
extern FILE *infile;
extern char *OutPtr;
extern int satisfiable;
extern int unknown, tough;
extern int o_batch;

/*-----------------------------------*/
/* internally defined functions      */
/*-----------------------------------*/

void ReadTptp();
void DetermineRating();
void ConvertTptp();
void InsertChar();
void DeleteChar();
void CheckClauseType();
int  CompareStrings();
void SetPtrToNextLine();
void AssignCostToClause();
void DumpClause();
void OutString();
void FinalSort();
void CompileTptp();

/*----------------------------------*/
/* internally defined constants     */
/*----------------------------------*/

FILE *includeFile;

char tptpCodeBuffer[80000];
char *tptpCodePtr;
char *tptpCodeEndPtr;
char *tptpCodePtrTemp;
char *tptpCodePtrTemp2;
char Rating[40];
int RatingPtr;
char Sat[40];
int SatPtr;
int theorem_rating;
                             
char tptpIncludeCodeBuffer[80000];  /* handling of include directive */
char *currentIncludeCodeBuffer;
char *tptpCodePtrSave;                      /* saved position in the */
char *tptpCodeEndPtrSave;                   /* including file */
int  inInclude;                     /* to prevent recursive includes */
                                    /*-------------------------------*/
char includeName[128];
char includeAltName[128] = {'.','.','/','.','.','/'};

char theoCodeBuffer[80000];
char *theoCodePtr;

int ClauseCount;
int CharCount;

int ClauseType[MAX_BASE_CLAUSES];
int LitCount[MAX_BASE_CLAUSES];
int ClauseLength[MAX_BASE_CLAUSES];
int ClauseHead[MAX_BASE_CLAUSES];

/*-----------------------------------------------------------------*/
/*      ReadTptp()                   .......................#1
/*
/*  called by:   CompileTptp()       tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*-----------------------------------------------------------------*/

void ReadTptp()
{
int InChar, i;
char *p1, *p2, *p3;

inInclude = 0;
CharCount = 0;
ClauseCount = 0;

for(i = 0; i < MAX_BASE_CLAUSES ; i++) {
   ClauseType[i] = LitCount[i] = ClauseLength[i] = ClauseHead[i] = 0;}

p1 = tptpCodeBuffer;
p2 = tptpIncludeCodeBuffer;
p3 = theoCodeBuffer;

for(i = 0; i < 80000; i++) {
   *p1 = *p2 = *p3 = 0;
   p1++; p2++; p3++;}

tptpCodePtr = tptpCodeBuffer;

/*------------------------------------------*/
/* read theorem                             */
/*------------------------------------------*/

for(;;) {
   *tptpCodePtr = getc(infile);
   CharCount++;
   if(*tptpCodePtr == -1) {break;}
   tptpCodePtr++;}
tptpCodeEndPtr = tptpCodePtr;

currentIncludeCodeBuffer = tptpIncludeCodeBuffer;

return;
}

/*-------------------------------------------------------------------*/
/*  DetermineRating()
/*
/* Called from: CompileClauses()
/*
/*  Searches from rating of theorem and store for later printing
/*
/*-------------------------------------------------------------------*/

void DetermineRating()
{
int i, n, s;

for(i = 0; i < CharCount; i++) {
   if(tptpCodeBuffer[i] == 'R' &&
      tptpCodeBuffer[i+1] == 'a' &&
	 tptpCodeBuffer[i+2] == 't' &&
	 tptpCodeBuffer[i+3] == 'i' &&
	 tptpCodeBuffer[i+4] == 'n' &&
	 tptpCodeBuffer[i+5] == 'g') {
      RatingPtr = 0;
      while(tptpCodeBuffer[i] != '\n' && 
			     tptpCodeBuffer[i] != ',') {
	 Rating[RatingPtr++] = tptpCodeBuffer[i++];}
      if(tptpCodeBuffer[i] != '\n') Rating[RatingPtr] = ' ';
      break;}}


n = 0;
for(s = 0; s < 40 ; s++) {
   if(Rating[s] != '.') {continue;}
   n = n * 10 + (Rating[s - 1] - '0');
   n = n * 10 + (Rating[s + 1] - '0');
   n = n * 10 + (Rating[s + 2] - '0');
   break;}
		 
theorem_rating = n;

for(i = 0; i < CharCount; i++) {
   if(tptpCodeBuffer[i] == 'S' &&
      tptpCodeBuffer[i+1] == 't' &&
	 tptpCodeBuffer[i+2] == 'a' &&
	 tptpCodeBuffer[i+3] == 't' &&
	 tptpCodeBuffer[i+4] == 'u' &&
	 tptpCodeBuffer[i+5] == 's') {
      SatPtr = 0;
      while(tptpCodeBuffer[i] != '\n' && 
			     tptpCodeBuffer[i] != ',') {
	 Sat[SatPtr++] = tptpCodeBuffer[i++];}
      if(tptpCodeBuffer[i] != '\n') Sat[SatPtr] = '\n';
      break;}}

for(i = 0; i <= RatingPtr; i++) {
   OutPtr += sprintf(OutPtr,"%c",Rating[i]);
   printf("%c",Rating[i]);}
print0("   ");
for(i = 0; i <= SatPtr; i++) {
   OutPtr += sprintf(OutPtr,"%c",Sat[i]);
   printf("%c",Sat[i]);}
print0("\n");

for(i = 0; i < 30; i++) {
   if(Rating[i] == ' ' && Rating[i+1] == '1' && 
	    Rating[i + 2] == '.' && Rating[i + 3] == '0'){
      tough = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == 'o' && Sat[i + 1] == 'p' && Sat[i + 2] == 'e') {
      unknown = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == 'u' && Sat[i + 1] == 'n' && Sat[i + 2] == 'k') {
      unknown = 1;break;}}

for(i = 0; i < 30; i++) {
   if(Sat[i] == ' ' && Sat[i + 1] == 's' && Sat[i + 2] == 'a') {
      satisfiable = 1;break;}}

}

/*-------------------------------------------------------------------*/
/*  ConvertTptp()        ........................... #2
/*
/*  called by:  CompileTptp()      tptpi.c
/*
/*  calls:      InsertChar()       tptpi.c
/*              DeleteChar()       tptpi.c
/*              CheckClauseType()  tptpi.c
/*              SetPtrToNextLine() tptpi.c
/*
/*  returns:    Nil
/*
/*-------------------------------------------------------------------*/

void ConvertTptp()
{
char *j, *jj, *jjj, *k, InChar;
char char1, char2;
int bracket_count;
int i, ii;
char *CommaPtr, *CRPtr;
int CRfound;

tptpCodePtr = tptpCodeBuffer;

STATE1:
if(*tptpCodePtr == 10) {    /* skip over carriage returns */
   SetPtrToNextLine();
   goto STATE1;}

if(*tptpCodePtr == ' ') {    /* skip over carriage returns */
   DeleteChar(1);
   goto STATE1;}

if(*tptpCodePtr == '%') {   /* skip over comments */
   *tptpCodePtr = ';';
   SetPtrToNextLine();
   goto STATE1;}

if(*tptpCodePtr == 'i') { 
   if(tptpCodePtr[2] == 'c') { 

	      /*-------------------------------*/
              /* handling of include directive */
	      /*-------------------------------*/

      for( ;*tptpCodePtr != '\''; tptpCodePtr++); 
      tptpCodePtr++;

      for(i = 0; *tptpCodePtr!='\''; i++) {
	 includeAltName[i + 6] = includeName[i] = *tptpCodePtr++;}
      includeAltName[i + 6] = includeName[i] = 0;

      for(; *tptpCodePtr != '.'; tptpCodePtr++);
      tptpCodePtr++;
      for(ii = slash_char + 1; ii <= slash_char + 6 + i; ii++) {
	 theorem_name[ii] = includeAltName[ii - slash_char - 1];}

      if(inInclude) { 
         printf("Can't do recursive include..."); }
      else { 
         if((includeFile = fopen(includeName,"r")) == NULL) {
	    if((includeFile = 
		  fopen(theorem_name,"r")) == NULL) {
               printf("Can't include required file");}}

         inInclude = 1;
         tptpCodePtrSave = tptpCodePtr;
         tptpCodeEndPtrSave = tptpCodeEndPtr;
         tptpCodePtr = currentIncludeCodeBuffer;

         for(;;) {
             InChar = getc(includeFile);/* read the included file */
             *tptpCodePtr = InChar;
             if(InChar == -1) {break;}
             tptpCodePtr++;} 
          fclose(includeFile);
          tptpCodeEndPtr = tptpCodePtr;
          tptpCodePtr = currentIncludeCodeBuffer;
          goto STATE1;}}
   else {
      ClauseHead[ClauseCount] = (int) tptpCodePtr;
      LitCount[ClauseCount++] = 1;
      CheckClauseType();
      *tptpCodePtr++ = ';';
      DeleteChar(12);
      SetPtrToNextLine();
      goto STATE2;}}

if(*tptpCodePtr == -1) {         /*        -1 = EOF               */
   if(inInclude) {               /*-------------------------------*/
      inInclude = 0;
      currentIncludeCodeBuffer = tptpCodeEndPtr + 1;
      tptpCodePtr = tptpCodePtrSave;   /* restore position in the */
      tptpCodeEndPtr = tptpCodeEndPtrSave;      /* including file */
      goto STATE1;}
   return;}                      /*-------------------------------*/

STATE2:

if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) {
   DeleteChar(1);
   goto STATE2;} 
   /*
   if(*tptpCodePtr == '.' || *tptpCodePtr == ',') {
      tptpCodePtr++;
      while(*tptpCodePtr == ' ') {
	 DeleteChar(1);}
      goto STATE1;}
   else {
      tptpCodePtr++;
      goto STATE2;}} */

/*if(*tptpCodePtr == ' ' || *tptpCodePtr == 10) {
   DeleteChar(1);
   goto STATE1;}*/ 

if(*tptpCodePtr == '%') {
   while(*tptpCodePtr != 10) {
      DeleteChar(1);}
   goto STATE2;}

if(*tptpCodePtr == '.') {
   DeleteChar(1);
   tptpCodePtr++;
   ClauseLength[ClauseCount - 1] =
      (int) tptpCodePtr - (int) ClauseHead[ClauseCount - 1];
   goto STATE1;}

if(*tptpCodePtr == '[') {
   DeleteChar(1);
   goto STATE2;}

if(*tptpCodePtr ==  '+') {  /* 43 = + */
   tptpCodePtr++;
   if(*tptpCodePtr ==  '+' ) {
      tptpCodePtr--;
      DeleteChar(2);
      goto STATE2;}
   tptpCodePtr--;
   goto STATE2;}

if(*tptpCodePtr == '-') {
   tptpCodePtr++;
   if(*tptpCodePtr == '-') {
      tptpCodePtr--;
      DeleteChar(1);
      *tptpCodePtr++ = '~';
      goto STATE2;}
   tptpCodePtr--;
   goto STATE2;}

if(*tptpCodePtr == ',') {
   CRfound = 0;
   CommaPtr = tptpCodePtr;
   tptpCodePtr++;
   while(*tptpCodePtr == ' ' || *tptpCodePtr == 9) {
      DeleteChar(1);}
   if(*tptpCodePtr == '+' || *tptpCodePtr == '-') {
      *CommaPtr = -2;
      goto next_literal;}
   if(*tptpCodePtr != 10) {
      goto STATE2;}
   CRPtr = tptpCodePtr;
   CRfound = 1;
   tptpCodePtr++;
   while(*tptpCodePtr == 32) {
      DeleteChar(1);}
next_literal: 
   tptpCodePtrTemp = tptpCodePtr;
   tptpCodePtrTemp++;

   if(*tptpCodePtr == '+' && *tptpCodePtrTemp == '+') {
      LitCount[ClauseCount - 1]++;
      tptpCodePtr--;
      *tptpCodePtr++ = ' ';
      *tptpCodePtr++ = '|';
      *tptpCodePtr++ = ' ';
      if(CRfound != 0) {*CRPtr = -2;}
      *CommaPtr = -2;
      goto STATE2;}
   if(*tptpCodePtr == '-' && *tptpCodePtrTemp == '-') {
      LitCount[ClauseCount - 1]++;
      tptpCodePtr--;
      *tptpCodePtr++ = ' ';
      *tptpCodePtr++ = '|';
      *tptpCodePtr++ = '~';
      if(CRfound != 0) {*CRPtr = -2;}
      *CommaPtr = -2;
      goto STATE2;}
   *CRPtr = -2;
   goto STATE2;}

if(isalnum(*tptpCodePtr)) {
   j = tptpCodePtr;
   j--;
   k = tptpCodePtr;
   while(*tptpCodePtr != ')' && *tptpCodePtr != ',' &&
         *tptpCodePtr != ' ' && *tptpCodePtr != '(' &&
	 *tptpCodePtr != ']') {
      tptpCodePtr++;}
   if(*tptpCodePtr == ' ') {
      *tptpCodePtr = '(';
      bracket_count = 0;
      tptpCodePtr++;
      for(jj = tptpCodePtr ; ; jj++) {
	 if(*jj == '(') {bracket_count++;}
	 if(bracket_count == 0 && (*jj == ',' || *jj == ')')) {
	    jjj = tptpCodePtr;
	    tptpCodePtr = jj;
	    InsertChar(')'); 
	    tptpCodePtr = --jjj;
	    break;}
	 if(*jj == ')') {
	    bracket_count--;}}}

   if(*j == '+' || *j == '-' || isupper(*k) || *tptpCodePtr ==
           '(' || *tptpCodePtr == ' ') { 
      goto STATE2;}

   --tptpCodePtr;
   char1 = *tptpCodePtr;
   --tptpCodePtr;
   char2 = *tptpCodePtr;
   tptpCodePtr++;tptpCodePtr++;
   if(char2 == '(' || char2 == ',') {
      if(islower((unsigned char) char1) && char1 > 'o') {
	 InsertChar('C');
	 tptpCodePtr++;}}

   InsertChar('^');
   tptpCodePtr++;
   goto STATE2;}

if(*tptpCodePtr == ']') { 
   DeleteChar(2);
   goto STATE2;}
else {
   tptpCodePtr++;
   goto STATE2;}
}

/*------------------------------------------------------------------*/
/*  InsertChar()           ...................... #3
/*
/*  called by:  ConvertTptp()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*------------------------------------------------------------------*/

void InsertChar(s)
char s;
{
char *tptpCodeEndPtr1;
char *tptpCodeEndPtr2;

tptpCodeEndPtr++;
tptpCodeEndPtr1 = tptpCodeEndPtr;
tptpCodeEndPtr2 = tptpCodeEndPtr1;
tptpCodeEndPtr2--;

while(tptpCodeEndPtr2 >= tptpCodePtr) {
   *tptpCodeEndPtr1 = *tptpCodeEndPtr2;
   tptpCodeEndPtr2--;
   tptpCodeEndPtr1--;}

*tptpCodePtr = s;
return;
}

/*------------------------------------------------------------------*/
/*  DeleteChar()          ..................  #4
/*
/*  called by:   ConvertTptp()     tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*------------------------------------------------------------------*/

void DeleteChar(d)
int d;
{
int i;

for(i = d; i > 0; i--) {
   *tptpCodePtr++ = -2;}
return;
}

/*------------------------------------------------------------------*/
/*  CheckClauseType()        ................  #5
/*
/*  called by:    ConvertTptp()     tptpi.c
/*
/*  calls:        CompareStrings()  tptpi.c
/*
/*  returns:      Nil
/*
/*------------------------------------------------------------------*/

void CheckClauseType()
{
char *ccp;

ccp = tptpCodePtr;
while(*ccp != 10) {
   if(CompareStrings(ccp,"conjecture",10)) {
      ClauseType[ClauseCount - 1] = 1; return;}
   if(CompareStrings(ccp,"hypothesis",10)) {
      ClauseType[ClauseCount - 1] = 2; return;}
   ccp++;}
ClauseType[ClauseCount - 1] = 3;
return;
}

/*------------------------------------------------------------------*/
/*  CompareStrings(s,t,n)       .................  #6
/*
/*  called by:  CheckClauseType()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    1 if strings are same, 0 otherwise
/*
/*------------------------------------------------------------------*/

int CompareStrings(s,t,n)
char s[],t[];
int n;
{
int i;

i = 0;
while(s[i] == t[i]) {
   i++;
   if(i == n) {return(1);}}
return(0);
}

/*------------------------------------------------------------------*/
/*  SetPtrToNextLine()      ....................  #7
/*
/*  called by:   ConvertTptp()      tptpi.c
/*
/*  calls:       Nobody
/*
/*  returns:     Nil
/*
/*------------------------------------------------------------------*/

void SetPtrToNextLine()
{
while(*tptpCodePtr != 10) { 
   tptpCodePtr++;}
tptpCodePtr++;
return;
}

/*------------------------------------------------------------------*/
/*  AssignCostToClause()            .......................  #8
/*
/*  called by:  CompileTptp()     tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*  assigns a cost to clauses so that they are first sorted with
/*  the negated conclusion at the end, then by the number of
/*  literals, then by the length of the clause in characters.
/*------------------------------------------------------------------*/

void AssignCostToClause()
{
int i;

for(i = 0; i < ClauseCount; i++) {
    LitCount[i] =
       100000*ClauseType[i] + 1000*LitCount[i] + ClauseLength[i];}
return;
}

/*------------------------------------------------------------------*/
/*  DumpClause()         .....................  #9
/*
/*  called by:  FinalSort()        tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*------------------------------------------------------------------*/

void DumpClause(Imax)
int Imax;
{
int CRctr;

CRctr = 0;
tptpCodePtr = (char *) ClauseHead[Imax];

for( ; ; ) {
   if(*tptpCodePtr == -2) {
      tptpCodePtr++;                     /* skip deleted chars */
      continue;}
   *theoCodePtr = *tptpCodePtr;
   if(*theoCodePtr == 10 && CRctr++ == 1) {
      break;}
   if(*theoCodePtr == '^') {        /* replace "^" with "()" */
      *theoCodePtr++ = '(';
      *theoCodePtr = ')';}
   if(*theoCodePtr == '|') {
      theoCodePtr++;             /* delete OR symbol   */
      *theoCodePtr = ' ';}
   tptpCodePtr++;
   theoCodePtr++;}
theoCodePtr++;
*theoCodePtr++ = 10;

return;
}

/*------------------------------------------------------------------*/
/*  OutString()       ....................  #10
/*
/*  called by:  FinalSort()        tptpi.c
/*
/*  calls:      Nobody
/*
/*  returns:    Nil
/*
/*  Place a string "s" of length "Length" in the theoCodeBuffer 
/*------------------------------------------------------------------*/

void OutString(s,Length)
char s[];
int Length;
{
int i;

for(i = 0; i < Length ; i++) {
   *theoCodePtr++ = s[i];}
*theoCodePtr++ = 10;
return;
}

/*------------------------------------------------------------------*/
/*  FinalSort()          ..................... #11
/*
/*  called by:  CompileTptp()         tptpi.c
/*
/*  calls:      AssignCostToClause()  tptpi.c
/*
/*  returns:     Nil
/*        
/*------------------------------------------------------------------*/

void FinalSort()
{
int IndexMax, LitMax, i, NegConclusion;
int ii;
char *OutputCodePtr;

theoCodePtr = theoCodeBuffer;
NegConclusion = 0;
for( ; ;   ) {
   IndexMax = 0;
   LitMax = 0;
   for(i = 0; i < ClauseCount; i++) {
      if(LitCount[i] > LitMax) {
         LitMax = LitCount[i];
         IndexMax = i;}}
   if(LitMax == 0) {goto finaldump;}
   LitCount[IndexMax] = 0;
   if(NegConclusion == 0 && LitMax < 300000) {
      NegConclusion++;
      OutString("negated_conclusion",18);}
   DumpClause(IndexMax);}

finaldump:
*theoCodePtr = -1;


return;
}
 
/*------------------------------------------------------------------*/ 
/*      compile_tptp()         .......................#12
/*
/*      called by:  main()                  theo.c
/*
/*      calls:      ReadTptp()              tptpi.c
/*                  ConvertTptp()           tptpi.c
/*                  AssignCostToClause()    tptpi.c
/*                  FinalSort()             tptpi.c
/*
/*      returns:    Nil
/*------------------------------------------------------------------*/
 
void CompileTptp()
{
int i;
char *z;

ReadTptp(); 
DetermineRating();
ConvertTptp();
AssignCostToClause();
FinalSort();
/*
for(z = tptpCodeBuffer ; z < tptpCodePtr ; z++) {
   printf(" %c ",(char ) *z);}

for(i=0; i < RatingPtr; i++) {
   printf("%c",Rating[i]);
   OutPtr += sprintf(OutPtr,"%c",Rating[i]);}
print0("   ");

for(i=0; i < SatPtr; i++) {
   printf("%c",Sat[i]);
   OutPtr += sprintf(OutPtr,"%c",Sat[i]);}
print0("\n");
*/
return;
}

