
/*
 * File Assertion.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#if DEBUG_PREPRO

#include "Assertion.hpp"
#include "Formula.hpp"


bool Assertion::isInENNF (const Formula& f)
{
  switch (f.connective()) {
  case ATOM:
    return true;

  case AND:
  case OR:
    {
      Iterator<Formula> fs(f.args());
      while (fs.hasNext()) {
	if (! isInENNF(fs.next())) {
	  return false;
	}
      }
    }
    return true;

  case IMP:
    return false;

  case IFF:
  case XOR:
    return isInENNF(f.left()) && isInENNF(f.right());

  case NOT:
    return f.uarg().connective() == ATOM;

  case FORALL:
  case EXISTS:
    return isInENNF(f.qarg());

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
}


bool Assertion::isInNNF (const Formula& f)
{
  switch (f.connective()) {
  case ATOM:
    return true;

  case AND:
  case OR:
    {
      Iterator<Formula> fs(f.args());
      while (fs.hasNext()) {
	if (! isInNNF(fs.next())) {
	  return false;
	}
      }
    }
    return true;

  case IMP:
  case IFF:
  case XOR:
    return false;

  case NOT:
    return f.uarg().connective() == ATOM;

  case FORALL:
  case EXISTS:
    return isInNNF(f.qarg());

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 


bool Assertion::isFlat (const Formula& f)
{
  switch (f.connective()) {
  case ATOM:
    return true;

  case AND:
  case OR:
    {
      Iterator<Formula> fs(f.args());
      while (fs.hasNext()) {
	Formula sub(fs.next());
	if (sub.connective() == f.connective() || ! isFlat(sub)) {
	  return false;
	}
      }
    }
    return true;

  case IMP:
  case IFF:
  case XOR:
    return isFlat(f.left()) && isFlat(f.right());

  case NOT:
    return isFlat(f.uarg());

  case FORALL:
  case EXISTS:
    {
      Formula sub(f.qarg());
      return f.connective() != sub.connective() && isFlat(sub);
    }

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 


#endif 
