
/*
 * File Atom.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __Atom__
#define __Atom__

class AtomList;
class PredicateSymbol;

#include "Term.hpp"
#include "Symbol.hpp"





class Atom {
 private:
  class Data;

 public:
  
  Atom () : _data (0) {}

  Atom (const PredicateSymbol& p, const TermList& args); 
  Atom (const PredicateSymbol& p);                       
  Atom (const Term& l, const Term& r);             
  Atom (const Atom& a);
  ~Atom ();
  void operator = (const Atom& rhs);

  const PredicateSymbol& functor () const;
  const TermList& args () const;
  bool operator == (const Atom& rhs) const;
  bool operator != (const Atom& rhs) const;
  bool isEquality () const;

  
  void* operator new (size_t);

  
  XML::Element* toXML () const;
  bool occurs (const PredicateSymbol&) const;
  bool occurs (Var v) const;
  bool equals (const Atom& t) const;
  bool hasVarsNotIn (const VarListList&) const;

  
  string toString () const;
 private:
  
  Data* _data;
}; 


class AtomList 
: public Lst<Atom>
{
public:
  
  AtomList () : Lst<Atom> () {}

  
  AtomList::AtomList (const AtomList& ts)
    : Lst<Atom> (ts) {}

  explicit AtomList (const Atom& a); 
  
  AtomList (const Atom &hd, const AtomList& tl)
    : Lst<Atom> (hd,tl) {}
  
  explicit AtomList (LstData<Atom>* d)
    : Lst<Atom> (d) {}

  
  const AtomList& tail () const
    { return static_cast<const AtomList&>(Lst<Atom>::tail()); }
}; 


class Atom::Data 
{
 public:
  Data ();
  ~Data () { ASS (_counter == 0); }

  
  Data (const PredicateSymbol& f, const TermList& args)
    : _counter (1), _functor (f), _args (args) {} 
  
  explicit Data (const PredicateSymbol& f)
    : _counter (1), _functor (f) {}
  Data (const Term& l, const Term& r);

  void ref () { ASS (this); _counter++; }
  void deref ();

  
  const PredicateSymbol& functor () const { return _functor; }
  
  const TermList& args () const { return _args; }

  
  CLASS_NAME("Atom");
  USE_ALLOCATOR(Data);
 protected:
  
  int _counter;
  const PredicateSymbol& _functor;
  TermList _args;
}; 





inline
Atom::Atom (const Atom& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 

inline
Atom::~Atom () 
{
  if (_data) {
    _data->deref ();
  }
} 

inline
bool Atom::operator == (const Atom& rhs) const
{ 
  ASS (_data && rhs._data);

  return _data == rhs._data; 
} 


inline
bool Atom::operator != (const Atom& rhs) const
{ 
  ASS (_data && rhs._data);

  return _data != rhs._data; 
} 



inline
const PredicateSymbol& Atom::functor () const
{
  return _data->functor ();
} 


inline
const TermList& Atom::args () const
{
  return _data->args ();
} 




inline
bool Atom::hasVarsNotIn (const VarListList& vs) const
{
  return args().hasVarsNotIn (vs);
} 



inline
bool Atom::isEquality () const
{ 
  return functor().isEquality (); 
} 


#endif 
