
/*
 * File AtomIterator.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "AtomIterator.hpp"
#include "Tracer.hpp"
#include "Formula.hpp"

class AtomIterator::Element 
{
public:
  Element (const FormulaList& list, Element* rest) 
    : _isList(true), _list(const_cast<FormulaList*>(&list)), _rest(rest) {}
  Element (const Formula& f, Element* rest) 
    : _isList(false), _formula(const_cast<Formula*>(&f)), _rest(rest) {}
  bool _isList;
  union{ 
    FormulaList* _list;
    Formula* _formula;
  };
  Element* _rest;
};



void AtomIterator::popReserve ()
{
  TRACER("AtomIterator::popReserve()");
  ASS(_reserve);

  Element* next = _reserve->_rest;
  delete _reserve;
  _reserve = next;
} 




AtomIterator::AtomIterator (const Formula& t)
  : _current(const_cast<Formula*>(&t)),
    _reserve(0)
{
} 



AtomIterator::AtomIterator (const FormulaList& ts)
  : _current(0),
    _reserve(new Element(ts,0))
{
} 



bool AtomIterator::hasNext ()
{
  TRACER("AtomIterator::hasNext");

  if (_current) {
    return true;
  }

  
  while (_reserve) {
    if (_reserve->_isList) {
      FormulaList* first = _reserve->_list;
      if (first->isEmpty()) {
	popReserve();
      }
      else { 
	_current = const_cast<Formula*>(&(first->head()));
	_reserve->_list = const_cast<FormulaList*>(&(first->tail()));
	return true;
      }
    }
    else { 
      _current = _reserve->_formula;
      popReserve();
      return true;
    }
  }
  
  return false;
} 



const Atom& AtomIterator::next ()
{
  for(;;) {
    switch (_current->connective()) {
    case ATOM:
      {
	const Atom& a = _current->atom();
	_current = 0;
	return a;
      }
      
    case AND:
    case OR:
      _reserve = new Element(_current->args().tail(),_reserve);
      _current = const_cast<Formula*>(&_current->args().head());
      break;
      
    case IMP:
    case IFF:
    case XOR:
      _reserve = new Element(_current->right(),_reserve);
      _current = const_cast<Formula*>(&_current->left());
      break;
    
    case NOT:
      _current = const_cast<Formula*>(&_current->uarg());
      break;

    case FORALL:
    case EXISTS:
      _current = const_cast<Formula*>(&_current->qarg());
      break;

#if DEBUG_PREPRO
    default:
      ASS(false);
#endif
    }
  }
} 



AtomIterator::~AtomIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
