//
// File:         CountedReference.hpp
// Description:  Template for counted references. 
// Created:      Oct 10, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk 
// Status:       Sketch.
//============================================================================
#ifndef COUNTED_REFERENCE_H
#define COUNTED_REFERENCE_H
//============================================================================
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COUNTED_REFERENCE
#  define DEBUG_NAMESPACE "CountedReference<T>"
#endif
#include "debugMacros.hpp"
//============================================================================

namespace BK 
{

  // Requirements:
  //   void T::incReferenceCounter()
  //   void T::decReferenceCounter()

  template <class T>
  class CountedReference
  {
  public:
    CountedReference() : _content(0) {};
    CountedReference(T* ptr) : _content(ptr)
    {
      if (_content)
	_content->incReferenceCounter(); 
    };
    CountedReference(const CountedReference& ref) : _content(ref._content)
    {
      if (_content)
	_content->incReferenceCounter();
    };
    ~CountedReference()
    {
      if (_content)
	_content->decReferenceCounter();
    };

    CountedReference& operator=(const CountedReference& ref)
    {
      if (_content) 
	_content->decReferenceCounter();
      _content = ref->_content;
      if (_content)
	_content->incReferenceCounter();
    };
  
    bool operator==(const CountedReference& ref) const
    {
      return _content == ref._content;
    };

    bool operator==(const T* ptr) const
    {
      return _content == ptr;
    };

    void clear()
    {
      if (_content)
	{
	  _content->decReferenceCounter();
	  _content = 0;
	};
    };

    T* content() const { return _content; };
    const T* constContent() const { return _content; };

    T& operator*() const
    {
      return *_content; // even if _content == 0
    };

    T* operator->() const
    {
      return _content; // even if _content == 0
    };

  private:
    T* _content;
  }; // class CountedReference<class T>

}; // namespace BK


//============================================================================
#endif
