//
// File:         Defragmentation.hpp
// Description:  Memory defragmentation implemented with hashing
//               beginnings and ends of memory pieces.
// Created:      Apr 20, 2004, Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//==========================================================================
#ifndef DEFRAGMENTATION_H
#define DEFRAGMENTATION_H
//==========================================================================
#include "jargon.hpp"
#include "BlodKorvDebugFlags.hpp"
using namespace std;
//==========================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_DEFRAGMENTATION
#  define DEBUG_NAMESPACE "Defragmentation"
#endif
#include "debugMacros.hpp"
//==========================================================================

namespace BK 
{
  class Defragmentation
  {
  public:
    Defragmentation();
    ~Defragmentation();
    void reset();
    bool merge(void** piece,
	       ulong size);
    bool integrate(void** piece,
                   ulong size); // in words
    void resetResultCollection();
    void** collectNextPiece(ulong& size);
    // ^ piece of memory obtained externally, i.e. not
    // by merging some integrated pieces 
  private: 
    static void** freeCellValue() { return reinterpret_cast<void**>(0); };
    static void** deletedCellValue() { return reinterpret_cast<void**>(1); };
    static float criticalLoadFactor() { return 0.75; };
    float currentLoadFactor() const 
    { 
      return static_cast<float>(_currentLoad) / _currentHashTableSize;
    };
    ulong hashFunction(void** ptr)
    {
      CALL("hashFunction(void** ptr)");
      ASSERT(_currentHashTableSize);
      return
	(reinterpret_cast<ulong>(ptr) >> 2) % _currentHashTableSize;
    };
    
    void*** findInBeginnings(void** ptr);
    // ^ ptr is assumed to be in _hashTableForBeginnings
    void*** findInEnds(void** keyPtr,void** valPtr);
    // ^ valPtr is assumed to be in _hashTableForEnds with the index keyPtr
   
    void*** tryToFindInBeginnings(void** ptr);
    // ^ returns 0 if ptr is not found in _hashTableForBeginnings
    
    void*** tryToFindInEnds(void** ptr);
    // ^ returns 0 if no piece ending in ptr is registered in _hashTableForEnds
    
    void*** locationFor(void** ptr,void*** hashTable);
    // ^ ptr is assumed not to be in hashTable
    bool rehash();
  private:
    ulong _eligibleHashSizes[10]; 
    static ulong lastEligibleHashSizeNumber() 
    {
      /******** DF;
      return 4UL; 
      DF; *******/

      return 9UL;
    };
    ulong _nextEligibleHashSizeNumber;
    ulong _currentHashTableSize;
    ulong _currentLoad;
    void*** _hashTableForBeginnings;
    void*** _hashTableForEnds;

    ulong _nextResult;
  }; // class Defragmentation
}; // namespace BK

//==========================================================================
#endif
