//
// File:         jargon.hpp
// Description:  Frequently used definitions.
// Created:      May 5, 2000, 20:20
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//============================================================================
#ifndef JARGON_H
#define JARGON_H
#include <new>
using namespace std;

//============================================================================
// Target operational environment and compiler

// Supported compilers

#define BK_COMPILER_GCC_2_95_GENERIC 0
#define BK_COMPILER_GCC_3_GENERIC 1
#define BK_COMPILER_MSVCPP_GENERIC 2

// Supported compiler families

#define BK_COMPILER_FAMILY_GCC_2_95 0
#define BK_COMPILER_FAMILY_GCC_3 1
#define BK_COMPILER_FAMILY_MSVCPP 2

// Supported operational environments

#define BK_TARGET_OPERATIONAL_ENVIRONMENT_UNIX_GENERIC 0
#define BK_TARGET_OPERATIONAL_ENVIRONMENT_MSWIN32_GENERIC 1

// Supported operational environment types

#define BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_UNIX 0
#define BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_MSWIN32 1

// Identify the compiler family and version

#undef BK_COMPILER
#undef BK_COMPILER_FAMILY 


#if defined __GNUG__
#  if ((__GNUC__ == 2) && (__GNUC_MINOR__ == 95))
#    define BK_COMPILER BK_COMPILER_GCC_2_95_GENERIC
#    define BK_COMPILER_FAMILY BK_COMPILER_FAMILY_GCC_2_95
#  elif (__GNUC__ == 3)
#    define BK_COMPILER BK_COMPILER_GCC_3_GENERIC
#    define BK_COMPILER_FAMILY BK_COMPILER_FAMILY_GCC_3
#  else
#    error "Currently unsupported GCC version. Contact the BlodKorv library maintainers."
#  endif



#elif (defined _MSC_VER) 
#  if (_MSC_VER >= 1200)
#    define BK_COMPILER BK_COMPILER_MSVCPP_GENERIC
#    define BK_COMPILER_FAMILY BK_COMPILER_FAMILY_MSVCPP
#  else
#    error "Currently unsupported version of Microsoft C/C++ compiler. Contact the BlodKorv library maintainers."
#  endif

#else
#  error "Currently unsupported compiler family. Contact the BlodKorv library maintainers."
#endif


// Identify operational environment

#undef BK_TARGET_OPERATIONAL_ENVIRONMENT
#undef BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE

#if ((defined unix) || (defined __unix__) || (defined linux) || (defined __linux__))
#  define BK_TARGET_OPERATIONAL_ENVIRONMENT BK_TARGET_OPERATIONAL_ENVIRONMENT_UNIX_GENERIC
#  define BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_UNIX
#elif (defined _WIN32)
#  define BK_TARGET_OPERATIONAL_ENVIRONMENT BK_TARGET_OPERATIONAL_ENVIRONMENT_MSWIN32_GENERIC
#  define BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE BK_TARGET_OPERATIONAL_ENVIRONMENT_TYPE_MSWIN32

#else
#  error "Currently unsupported target operational environment. Contact the BlodKorv library maintainers."

#endif


//============================================================================


typedef unsigned long ulong;

namespace BK 
{

  class DummyArgument
  {
  public:
    DummyArgument() {};
    void touch() const {};
  };

  static DummyArgument dummyArgument;

  inline
  void* dummyVoidPtr() { return 0; }; 

  template <class C>
  inline
  static 
  void init(C& rawMemory)
  {
    ::new (static_cast<void*>(&rawMemory)) C();
  };

  template <class C,class A1>
  inline
  static 
  void init(C& rawMemory,const A1& arg1)
  {
    ::new (static_cast<void*>(&rawMemory)) C(arg1);
  };



  template <class C,class A1,class A2>
  inline
  static 
  void init(C& rawMemory,const A1& arg1,const A2& arg2)
  {
    ::new (static_cast<void*>(&rawMemory)) C(arg1,arg2);
  };

  

  template <class C,class A1,class A2,class A3>
  inline
  static 
  void init(C& rawMemory,const A1& arg1,const A2& arg2,const A3& arg3)
  {
    ::new (static_cast<void*>(&rawMemory)) C(arg1,arg2,arg3);
  };

  


  template <class C>
  inline
  static 
  void destroy(C& obj)
  {
    obj.~C();
  };
  

  template <class C>
  class DefaultRelocation
  {
  public:
    static void move(C& source,void* target)
    {
      ::new (target) C(source);
      source.~C();
    };
  };

  template <class C>
  class FastRelocation
  {
  public:
    static void move(C& source,void* target)
    {
      ::new (target) C(source);
    };
  };
  
  template <class C>
  class CustomRelocation
  {
  public:
    static void move(C& source,void* target)
    {
      source.relocateTo(target);
    };
  };
  
  
  template <class C>
  inline
  static
  void hibernate(C& obj)
  {
    // "do nothing" is the default semantics of hybernation 
  };

  template <class C>
  class AutoContextSwitch
  {
  public:
    typedef typename C::Context Context;
  public:
    AutoContextSwitch(Context* c) :
      _previous(Context::getCurrent())
    {
      Context::setCurrent(c);
    };
    ~AutoContextSwitch()
    {
      Context::setCurrent(_previous);
    };
  private:
    AutoContextSwitch() {};
  private:
    Context* _previous;
  }; // class AutoContextSwitch<class C>
  


}; // namespace BK




//============================================================================
#endif
