
/*
 * File Debugger.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Debugger.hpp"

#if DEBUG_PREPRO


bool Debugger::allocated (void* memory, 
			  int words,
			  const string& className)
{
  
  

  ASS(!_allocatedPieces.find(memory));
  _allocatedPieces.insert(memory,Info(2*words,className));

  return true;
} 


bool Debugger::released (void* memory, 
			 int words,
			 const string& className)
{
  
  
  Info found(-1,"");
  ASS (_allocatedPieces.find(memory,found));
  ASS (found.size % 2 == 0);
  ASS (!(found.size > 2*words));
  ASS (!(found.size < 2*words));
  ASS (found.className == className);

  _allocatedPieces.replace(memory,Info(2*words+1,""));
  return true;
} 


bool Debugger::reallocated (void* memory, 
			    int words, 
			    const string& className)
{
  
  

  Info found(-1,"");
  
  if (! _allocatedPieces.find(memory,found)) {
    cout << "Memory address " 
	 << memory 
	 << " of size "
	 << words
	 << " requested for allocation by " 
	 << className
	 << " has not been previously allocated\n";
    ASSERTION_VIOLATION;
  }
  
  ASS (found.size % 2 != 0);
  
  ASS (found.size == 2*words+1);

  _allocatedPieces.replace(memory,Info(2*words,className));
  return true;
} 

void Debugger::reportLeaks (ostream& str)
{
  
  int words = 0;
  str << "Memory leaks:" << "\n";
  for (int i = _allocatedPieces._capacity-1; i >= 0; i--) {
    Memory::Entry& entry(_allocatedPieces._entries[i]);
    if (entry.isOccupied()) {
      int size = entry.value().size;
      if (size % 2 == 0) {
	str << entry.key() << ": " << entry.value().className << "\n";
	words += size / 2;
      }
    }
  }
  str << "Total: " << words << " words\n";
} 

#endif
