
/*
 * File Grishin.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Grishin.hpp"
#include "Sequent.hpp"
#include "SignedFormula.hpp"
#include "Substitution.hpp"

bool Grishin::proveOr (SequentList&, Substitution&, Sequent*,
		       bool sign, const FormulaList&)
{
  return false;
}

bool Grishin::proveAtom (SequentList&, Substitution&, Sequent*,
			 bool sign, const Atom&)
{
  return false;
}


class Grishin::SequentList
  : public List<Sequent*>
{
public:
  ~SequentList() { delete _head; }
  static SequentList* empty ()
  { return static_cast<SequentList*>(List<Sequent*>::empty()); }
  SequentList* tail () const
  { return static_cast<SequentList*>(List<Sequent*>::tail()); }
  SequentList (Sequent* head, SequentList* tail) :
    List<Sequent*>(head,tail) {}
  SequentList (Sequent* head) :
    List<Sequent*>(head) {}
}; 


Grishin::Grishin ()
  : _initial(new Sequent())
{
  _skolem.setSkolemPrefix("$");
} 


void Grishin::add (const Formula& f)
{
  _initial = new Sequent(_initial,true,f);
} 


bool Grishin::findDerivation ()
{
  SequentList lst(_initial);
  Substitution sbst;
  prove(lst,sbst);
} 


bool Grishin::prove (SequentList& lst, Substitution& sbst)
{
  if (lst.isEmpty()) {
    return true;
  }

  Sequent* sequent = lst.head();
  SignedFormula sf;
  Sequent* subsequent = sequent->popFormula(sf);
  if (! subsequent) { 
    return false;
  }

  Formula f (sf.formula());
  bool sign = sf.sign();
  SequentList& tail = *lst.tail();
  switch (sf.type()) {
  case ATOM:
    return proveAtom(tail,sbst,subsequent,sign,f.atom());

  case AND:
    {
      
      SequentList newLst(new Sequent(subsequent,sign,f.args()),&tail);
      return prove(newLst,sbst);
    }

  case OR:
    {
      
      return proveOr(tail,sbst,subsequent,sign,f.args());
    }

  case IMP:
    {
      
      if (sign) {
	SequentList newLst1(new Sequent(subsequent,true,f.right()),&tail);
	SequentList newLst2(new Sequent(subsequent,false,f.left()),&newLst1);
	return prove(newLst2,sbst);
      }
      
      SequentList newLst(new Sequent(new Sequent(subsequent,false,f.right()),
				     true,f.left()),
			 &tail);
      return prove(newLst,sbst);
    }

  case IFF:
  case XOR:
  case NOT:
  case FORALL:
  case EXISTS:
    return true;
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 
