
/*
 * File InferenceInfo.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __InferenceInfo__
#define __InferenceInfo__

namespace XML {
  class Element;
};

#include <string>
#include "Position.hpp"
#include "Term.hpp"
#include "InferenceRule.hpp"
#include "VampireKernel.hpp"


class InferenceInfo
{
public:
  InferenceInfo (InferenceRule rule, 
		 const Position& position, 
		 InferenceInfo* next);
  InferenceInfo (InferenceRule rule, 
		 InferenceInfo* next);
  ~InferenceInfo () {}

  CLASS_NAME("InferenceInfo");
  USE_ALLOCATOR(InferenceInfo);

  
  InferenceInfo* next () const { return _next; }
  
  InferenceRule rule () const { return _rule; }
  
  const Position& position () const { return _position; }
  void destroy ();
  void destroyList ();
  XML::Element* toXML () const;
  string toString() const;
private:
  
  InferenceRule _rule;
  
  Position _position;
  
  InferenceInfo* _next;
}; 



class InferenceInfoTermList
  : public InferenceInfo
{
public:
  
  const TermList& terms () const { return _terms; }
  InferenceInfoTermList (InferenceRule rule,
			 const Position& position, 
			 InferenceInfo* next,
			 const TermList& terms);

  CLASS_NAME("InferenceInfoTermList");
  USE_ALLOCATOR(InferenceInfoTermList);
private:
  TermList _terms;
}; 


enum KernelRule {
  KR_IsInput = 1,
  KR_Preprocessing = 2,
  KR_BinaryResolution = 4,
  KR_ForwardTransitiveChaining = 8,
  KR_BackwardTransitiveChaining = 16,
  KR_ForwardSuperposition = 32,
  KR_BackwardSuperposition = 64,
  KR_EqualityResolution = 128,
  KR_EqualityFactoring = 256,
  KR_ForwardDemodulation = 512,
  KR_BackwardDemodulation = 1024,
  KR_SimplificationByEqualityResolution = 2048,
  KR_ForwardSubsumptionResolution = 4096,
  KR_BackwardSubsumptionResolution = 8192,
  KR_PropositionalSplitting = 16384, 
  KR_NegativeEqualitySplitting = 32768, 
  KR_IsReanimatedPassive = 65536,
  KR_IsPartOfNameIntroduction = 131072,
  KR_SimplificationByBuiltInTheories = 262144,
  KR_IsNameIntro = 524288,
  KR_Folding = 1048576,
  KR_Clausification = 2097152,
  KR_IsBuiltInTheoryFact = 4194304
}; 


class KernelInferenceInfo
  : public InferenceInfo
{
public:
  explicit KernelInferenceInfo (VampireKernel::Clause::BackgroundFlagIterator&);

  CLASS_NAME("KernelInferenceInfo");
  USE_ALLOCATOR(KernelInferenceInfo);
  std::string rulesAsString () const;
private:
  unsigned _inferenceCodes;
  static std::string toString (KernelRule);
}; 


#endif 
