
/*
 * File Kernel.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __Kernel__
#define __Kernel__

#include "RunResult.hpp"
#include "VampireKernel.hpp"

namespace XML {
  class DeepElement;
};

class Problem;
class Options;
class Term;
class Timer;
class TermList;
class RenamingToNew;
class Signature;
class Clause;
class Literal;
class LiteralList;


class Kernel
{
public:
  int passClauses (const Problem& problem, const Options& opts);
  void passTerm (const Term& t, 
		 RenamingToNew& renaming);
  static void passOptions (VampireKernel::Options& kernelOptions,
			   const Options& vampireOptions);
  void passSignature (Signature& sig);
  void getClause (const VampireKernel::Clause*, Clause& clause) const;
  void getTerm (const VampireKernel::Subterm*, Term& term) const;
  
  const VampireKernel::Clause* refutation () const 
    { return _kernel.refutation(); }
  
  VampireKernel& kernel () { return _kernel; }
  RunResult run (long timeLimitInDeciseconds,
		 VampireKernel::Options& kernelOptions, 
		 Problem& problem,
		 const Options& options,
		 Timer& timer,
		 XML::DeepElement* log);

  RunResult rerun ();
private:
  void getLiteralList (const VampireKernel::Literal*, 
		       LiteralList& literals) const;
  void getLiteral (const VampireKernel::Literal*, Literal& literal) const;
  void getTermList (const VampireKernel::Subterm*, TermList& terms) const;

  
  VampireKernel _kernel;
};

#endif
