
/*
 * File Literal.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "Literal.hpp"
#include "XML.hpp"
#include "Signature.hpp"


LiteralList:: LiteralList (const Literal& hd, const LiteralList& tl)
  : Lst<Literal> (hd,tl) 
{
}

void Literal::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;

  if (_counter == 0) {
    delete this;
  }
} 



void Literal::operator = (const Literal& t)
{
  if (t._data) {
    t._data->ref ();
  }

  if (_data) {
    _data->deref ();
  }

  _data = t._data;
} 



XML::Element* Literal::toXML() const
{
  XML::DeepElement* lit = new XML::DeepElement("literal");
  lit->addAttribute(new XML::TextAttribute("sign",sign() ? "+" : "-"));
  lit->addChild(atom().toXML());

  return lit;
} 



string Literal::toString () const
{
  if (isPositive()) {
    return atom().toString();
  }
  Atom a (atom());
  if (a.isEquality()) {
    TermList ts (a.args());
    return ts.head().toString() + "!=" + ts.second().toString();
  }
  return string("~") + a.toString();
} 



bool Literal::equals (const Literal& l) const
{
  return sign() == l.sign() &&
         atom().equals(l.atom()); 
} 



string LiteralList::toString (string separator) const
{
  if (isEmpty()) {
    return "";
  }
  string result = head().toString();
  Iterator<Literal> lst(tail());
  while (lst.hasNext()) {
    result += separator + lst.next().toString();
  }
  return result;
} 



