
/*
 * File Literal.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#ifndef __Literal__
#define __Literal__


class SubstitutionWithBacktracking;


#include "Atom.hpp"





class Literal 
{
 private:
  class Data;

 public:
  
  Literal () : _data (0) {}
  Literal (bool sign, const Atom& atom);
  Literal (const Literal& t);
  ~Literal ();
  void operator= (const Literal& rhs);

  
  bool operator == (const Literal& rhs) const { return _data == rhs._data; }
  bool operator != (const Literal& rhs) const { return _data != rhs._data; }
  bool isPositive () const;
  bool isNegative () const;
  bool sign () const;
  const Atom& atom () const;
  
  bool isNull () const { return _data == 0; }

  
  XML::Element* toXML () const;
  bool equals (const Literal&) const;

  
  string toString () const;
 private:
  
  Data* _data;
}; 


class Literal::Data 
{
 public:
  Data ();
  
  Data (bool sign, const Atom& atom)  
    : _counter (1), _sign (sign), _atom (atom) {}
  
  bool isPositive () const { return _sign; }
  
  bool isNegative () const  { return ! _sign; }
  
  bool sign () const  { return _sign; }
  
  const Atom& atom () const  { return _atom; }

  void ref () { ASS (this); _counter++; }
  void deref ();

  
  CLASS_NAME("Literal");
  USE_ALLOCATOR(Data);
 protected:
  
  int _counter;
  bool _sign;
  Atom _atom;
}; 


class LiteralList 
: public Lst<Literal>
{
 public:
  
  LiteralList () : Lst<Literal> () {}
  
  LiteralList (const LiteralList& ts) : Lst<Literal> (ts) {}
  
  explicit LiteralList (const Literal& l) : Lst<Literal> (l) {}
  LiteralList (const Literal& hd, const LiteralList& tl);
  
  explicit LiteralList (LstData<Literal>* d) : Lst<Literal> (d) {}

  
  const LiteralList& tail () const
    { return static_cast<const LiteralList&>(Lst<Literal>::tail()); }

  string toString (string separator) const;
}; 





inline
Literal::~Literal () 
{
  if (_data) {
    _data->deref ();
  }
} 


inline
Literal::Literal (const Literal& t)
  :
  _data (t._data)
{
  if (_data) {
    _data->ref ();
  }
} 


inline
Literal::Literal (bool sign, const Atom& atom)
  :
  _data (new Data(sign,atom))
{
} 


inline
bool Literal::isPositive () const
{
  return _data->sign();
} 


inline
bool Literal::isNegative () const
{
  return ! _data->sign();
} 


inline
bool Literal::sign () const
{
  return _data->sign();
} 


inline
const Atom& Literal::atom () const
{
  return _data->atom();
} 


#endif 
