
/*
 * File Problem.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include <string>
#include <climits>
#include <cstdlib>


#include "Tracer.hpp"
#include "Problem.hpp"
#include "XML.hpp"


Problem::Problem (const string& name, const UnitList& lst)
  : 
  _hasFormulas (false),
  _name (name)
{
  
  Iterator<Unit> us (lst);

  while ( us.hasNext() ) {
    Unit u (us.next());
    switch ( u.unitType() ) {
      case FORMULA:
        _hasFormulas = true;
        break;

      default:
        break;
    }
    _units.addLast (u);
  }
} 


Problem::~Problem ()
{
  
  while (! _units.isEmpty() ) {
    _units.delFirst ();
  }
} 









void Problem::removeAllConjectures ()
{
  TRACER ( "Problem::removeAllConjectures" );

  UnitChain::DelIterator us (_units);
  while ( us.hasNext() ) {
    switch (us.next().inputType()) {
    case CONJECTURE:
      us.del ();
      break;

    default:
      break;
    }
  }

  _originalAnswerVariables.makeEmpty();
  Literal empty;
  _answerLiteral = empty; 
} 



void Problem::addUnit (Unit u)
{
  TRACER ( "Problem::addUnit" );

  _units.addLast (u);
} 



XML::Element* Problem::toXML () const
{
  XML::DeepElement* problem = new XML::DeepElement("problem");

  UnitChain::Iterator us (_units);
  while (us.hasNext()) {
    problem->addChild(us.next().toXML());
  }

  return problem;
} 



string Problem::toString () const
{
  string problem("");

  UnitChain::Iterator us (_units);
  while (us.hasNext()) {
    problem += us.next().toString() + '\n';
  }

  return problem;
} 
