
/*
 * File Problem.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */
#ifndef __Problem__
#define __Problem__


#include <cstring>

#include "Unit.hpp"

namespace XML {
  class Element;
};

class FSymbol;
class Options;






class Problem 
{
 public:
  
  Problem (const string& name, const UnitList&);
  ~Problem ();

  
  const UnitChain& units () const { return _units; }
  
  UnitChain& giveUnits () { return _units; }
  
  bool hasFormulas () const { return _hasFormulas; }
  
  string name () const {return _name;};

  
  int split ();
  void addUnit (Unit u); 
  void removeAllConjectures (); 
  const VarList& originalAnswerVariables () const
    { return _originalAnswerVariables; }
  
  void setOriginalAnswerVariables (const VarList& vs)
    { _originalAnswerVariables = vs; }
  const Literal& answerLiteral () const 
    { return _answerLiteral; }
  
  void setAnswerLiteral (const Literal& lit) { _answerLiteral = lit; }
  XML::Element* toXML () const;
  string toString() const;

 private:
  
  void* operator new (size_t);

  
  UnitChain _units;
  bool _hasFormulas;
  VarList _originalAnswerVariables;
  Literal _answerLiteral;
  string _name;
};


#endif 
