
/*
 * File Property.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */



#ifndef __Property__
#define __Property__


#include <string>

class Problem;
class Unit;
class Clause;
class Literal;
class Atom;
class Term;
class Formula;
class VarCounter;



enum Category {
  NEQ,  
  HEQ,
  PEQ,
  HNE,
  NNE,
  FEQ,
  FNE, 
  EPR,
  UEQ
};


enum Prop {
  
  PR_HAS_X_EQUALS_Y = 1u,
  
  PR_HAS_FUNCTION_DEFINITIONS = 2u,
  
  PR_HAS_SUBSET = 4u,
  
  PR_HAS_EXTENSIONALITY = 8u
}; 


class Property 
{
 public:
  
  explicit Property (const Problem&);

  void scanProblem ();

  
  Category category() const { return _category;}
  string categoryString() const;

  string toString () const;

   
  int clauses () const { return _goalClauses + _axiomClauses; }
   
  int formulas () const { return _goalFormulas + _axiomFormulas; }
   
  int unitClauses () const { return _unitGoals + _unitAxioms; }
   
  int hornClauses () const { return _hornGoals + _hornAxioms; }

  
  bool hasProp (Prop p) const { return _props & p; }
  
  void addProp (Prop p) { _props |= p; }
  
  unsigned int props () const { return _props; }

 private:
  static bool hasXEqualsY (const Clause&);
  static bool isXEqualsY (const Atom&);
  static bool hasXEqualsY (const Formula&, VarCounter&, int polarity);

  
  void scan (const Unit&);
  void scan (const Formula&);
  void scan (const Clause&, bool isAxiom);
  void scan (const Atom&, bool& isGround);
  void scan (const Term&, bool& isGround);

  char axiomTypes () const;
  char goalTypes () const;
  char equalityContent () const;
  char nonGroundUnitContent () const;
  char groundPositiveContent () const;
  char goalsAreGround () const;
  char setClauseSize () const;
  char setLiteralSize () const;
  char setTermSize () const;
  char maxFunArity () const;
  char maxPredArity () const;

  
  const Problem& _problem;
  int _goalClauses;
  int _axiomClauses;

  int _equalityAtoms;
  int _atoms;

  int _goalFormulas;
  int _axiomFormulas;
  int _subformulas;

  int _terms;
  int _unitGoals;
  int _unitAxioms;
  int _hornGoals;
  int _hornAxioms;
  int _equationalClauses;
  int _pureEquationalClauses;
  int _groundUnitAxioms;
  int _positiveAxioms;
  int _groundPositiveAxioms;
  int _groundGoals;
  int _maxFunArity;
  int _maxPredArity;

  
  int _variablesInThisClause;
  
  int _totalNumberOfVariables;
  
  int _maxVariablesInClause;

  
  unsigned _props;

  
  Category _category;
}; 


#endif 
