
/*
 * File Query.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __query__
#define __query__


#include <string>


#include "Formula.hpp"


class Query {
 public:
  class Attribute;
  enum QueryType {
    ASSERTION,
    GOAL,
    TERMINATE
  };

  
  Query (QueryType qtype, const Formula& formula, Attribute* attributes);
  Query () : _type (TERMINATE) {}

  
  QueryType type () const { return _type; }
  const Formula& formula () const { return _formula; }
  int timeLimitInSeconds () const { return _timeLimit; }
  int bindingsLimit () const { return _bindingsLimit; }

  
  CLASS_NAME("Query");
  USE_ALLOCATOR(Query);
 private:
  
  QueryType _type; 
  Formula _formula;
  Attribute* _attributes;
  
  int _timeLimit;
  int _depthLimit;
  int _bindingsLimit;
}; 


class Query::Attribute {
 public:
  
  Attribute (const string name, const string value) :
    _name (name),
    _value (value)
    {}

  
  const string name () const { return _name; }
  const string value () const { return _value; }
  Attribute* next () const { return _next; }
  void next (Attribute* att) { _next = att; }

  
  USE_ALLOCATOR(Attribute);
 private:
  const string _name;
  const string _value;
  Attribute* _next;
}; 


#endif 
