
/*
 * File SignedFormula.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "SignedFormula.hpp"
#include "XML.hpp"


SignedFormulaList:: SignedFormulaList (const SignedFormula& hd, 
				       const SignedFormulaList& tl)
  : Lst<SignedFormula> (hd,tl) 
{
}

void SignedFormula::Data::deref () 
{ 
  ASS (this);
  ASS (_counter > 0);
  _counter--;

  if (_counter == 0) {
    delete this;
  }
} 



void SignedFormula::operator = (const SignedFormula& sf)
{
  if (sf._data) {
    sf._data->ref ();
  }

  if (_data) {
    _data->deref ();
  }

  _data = sf._data;
} 



XML::Element* SignedFormula::toXML() const
{
  XML::DeepElement* lit = new XML::DeepElement("signed_formula");
  lit->addAttribute(new XML::TextAttribute("sign",sign() ? "+" : "-"));
  lit->addChild(formula().toXML());

  return lit;
} 



string SignedFormula::toString () const
{
  return (string)(sign() ? "+" : "-") + formula().toString();
} 



Connective SignedFormula::type () const
{
  if (sign()) {
    return formula().connective();
  }
  
  switch (formula().connective()) {
  case ATOM:
    return ATOM;
  case AND:
    return OR;
  case OR:
    return AND;
  case IMP:
    return IMP;
  case IFF:
    return XOR;
  case XOR:
    return IFF;
  case NOT:
    return NOT;
  case FORALL:
    return EXISTS;
  case EXISTS:
    return FORALL;
#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }
} 
