
/*
 * File SubformulaIterator.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "SubformulaIterator.hpp"
#include "Tracer.hpp"
#include "Formula.hpp"


SubformulaIterator::SubformulaIterator (const Formula& t)
  : _current(const_cast<Formula*>(&t)),
    _reserve(0)
{
} 



SubformulaIterator::SubformulaIterator (const FormulaList& ts)
  : _current(0),
    _reserve(new Element(ts,0))
{
} 



bool SubformulaIterator::hasNext ()
{
  TRACER("SubformulaIterator::hasNext");

  if (_current) {
    return true;
  }

  
  while (_reserve) {
    if (_reserve->_isList) {
      FormulaList* first = _reserve->_list;
      if (first->isEmpty()) {
	Element* rest = _reserve->_rest;
	delete _reserve;
	_reserve = rest;
      }
      else { 
	_current = const_cast<Formula*>(&(first->head()));
	_reserve->_list = const_cast<FormulaList*>(&(first->tail()));
	return true;
      }
    }
    else { 
      _current = _reserve->_formula;
      Element* rest = _reserve->_rest;
      delete _reserve;
      _reserve = rest;
    }
  }
  
  return false;
} 



const Formula& SubformulaIterator::next ()
{
  Formula& result = *_current;

  switch (result.connective()) {
  case ATOM:
    _current = 0;
    break;

  case AND:
  case OR:
    _reserve = new Element(result.args(),_reserve);
    _current = 0;
    break;
    
  case IMP:
  case IFF:
  case XOR:
    _current = const_cast<Formula*>(&result.left());
    _reserve = new Element(result.right(),_reserve);
    break;

  case NOT:
    _current = const_cast<Formula*>(&result.uarg());
    break;

  case FORALL:
  case EXISTS:
    _current = const_cast<Formula*>(&result.qarg());
    break;

#if DEBUG_PREPRO
  default:
    ASS(false);
#endif
  }

  return result;
} 



SubformulaIterator::~SubformulaIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
