
/*
 * File SubtermIterator.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "SubtermIterator.hpp"
#include "Tracer.hpp"
#include "Literal.hpp"


SubtermIterator::SubtermIterator (const Term& t)
  : _current(const_cast<Term*>(&t)),
    _reserve(0)
{
} 



SubtermIterator::SubtermIterator (const TermList& ts)
  : _current(0),
    _reserve(new Element(ts,0))
{
} 



SubtermIterator::SubtermIterator (const Atom& a)
  : _current(0),
    _reserve(new Element(a.args(),0))
{
} 



SubtermIterator::SubtermIterator (const Literal& l)
  : _current(0),
    _reserve(new Element(l.atom().args(),0))
{
} 



bool SubtermIterator::hasNext ()
{
  TRACER("SubtermIterator::hasNext");

  if (_current) {
    return true;
  }

  
  while (_reserve) {
    TermList* first = _reserve->_list;
    if (first->isEmpty()) {
      Element* rest = _reserve->_rest;
      delete _reserve;
      _reserve = rest;
    }
    else { 
      _current = const_cast<Term*>(&(first->head()));
      _reserve->_list = const_cast<TermList*>(&(first->tail()));
      return true;
    }
  }
  
  return false;
} 



const Term& SubtermIterator::next ()
{
  Term* result = _current;
  if (_current->tag() == COMPOUND) {
    _reserve = new Element(result->args(),_reserve);
  }
  _current = 0;
  return *result;
} 



SubtermIterator::~SubtermIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
