
/*
 * File SymCounter.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __sym_counter__
#define __sym_counter__


#include "Symbol.hpp"

class Problem;
class Signature;
class FunctionSymbol;
class PredicateSymbol;
class Unit;
class Formula;
class Clause;
class Term;
class Atom;
class Literal;

using namespace std;






class SymCounter 
{ 
 public:

  class Fun {
    FunctionSymbol* _fun;
    int _occ;
   public:
    Fun () : _occ(0) {}
    
    void fun (const FunctionSymbol& f) 
    { 
      _fun = &(const_cast<FunctionSymbol&>(f)); 
    }
    int occ () const { return _occ; }
    const FunctionSymbol& fun () { return *_fun; }
    void add (int add) { _occ += add; }
  };

  class Pred {
    PredicateSymbol* _pred;
    int _pocc;  
    int _nocc;  
    int _docc;  
   public:
    Pred () 
      : _pocc (0), 
        _nocc (0),
        _docc (0) 
        {}
    void pred (const PredicateSymbol& p) 
    { 
      _pred = &(const_cast<PredicateSymbol&>(p));
    }
    int pocc () const { return _pocc; }
    int nocc () const { return _nocc; }
    int docc () const { return _docc; }
    const PredicateSymbol& pred () { return *_pred; }
    void add (int polarity, int add);
  };

  explicit SymCounter (Signature&);
  ~SymCounter ();

  
  void count (const Problem&, int);
  void count (const Unit&, int);
  void count (const Clause&, int);
  void count (const Formula&, int polarity, int add);
  void count (const Literal&, int);
  void count (const Atom& a, int polarity, int add);
  void count (const Term& t, int add);

  
  Pred* get (const PredicateSymbol& p) { return _preds + p.number(); }
  Fun* get (const FunctionSymbol& f) { return _funs + f.number(); }

 private:
  
  
  int _noOfPreds;
  
  int _noOfFuns;
  
  Pred* _preds;
  
  Fun* _funs;
}; 


#endif 
