
/*
 * File Symbol.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __Symbol__
#define __Symbol__


#include "List.hpp"


class Symbol 
{
public: 
  
  enum Property {
    
    FUNCTION = 1,
    
    EQUALITY_PROXY = 2,
    
    SKOLEM = 4
  };
public:
  
  int arity () const { return _arity; }
  
  const string& name () const { return _name;}
  
  int number () const { return _number; }
  
  void setNumber (int number) { _number = number; }
  
  bool operator == (const Symbol& symbol) const 
  { return this == &symbol; }
  
  bool operator != (const Symbol& symbol) const 
  { return this != &symbol; }
  
  unsigned long kernelNumber () const { return _kernelNumber; }
  
  void setKernelNumber (unsigned long n) { _kernelNumber = n; }
  
  int weight () const { return _weight; }
  
  int precedence () const { return _precedence; }
  
  int negativePrecedence () const { return _negativePrecedence; }
  
  void setWeight (int weight) { _weight = weight; }
  
  void setPrecedence (int precedence) { _precedence = precedence; }
  
  void setNegativePrecedence (int precedence) { _negativePrecedence = precedence; }
  
  void addProperty (Property p) { _properties = _properties | p; }
  
  bool hasProperty (Property p) const { return (_properties & p) != 0; }
  
  bool isFunction () const { return hasProperty(FUNCTION); }
  
  bool isEqualityProxy () const { return hasProperty(EQUALITY_PROXY); }
  
  bool isSkolemFunction () const { return hasProperty(SKOLEM); }

  XML::Element* toXML() const;

  CLASS_NAME("Symbol");
  USE_ALLOCATOR(Symbol);
protected:
  Symbol (const string& name, int arity, int& number);
  
  Symbol (const Symbol&);

  
  string _name;
  
  int _arity;
  
  int _number;
  
  unsigned long _kernelNumber;
  
  int _weight;
  
  int _precedence;
  
  int _negativePrecedence;
  
  int _properties;
}; 



class SymbolList :
  public List <Symbol*> 
{
public:
  ~SymbolList() { delete _head; }
  static SymbolList* empty ()
  { return static_cast<SymbolList*>(List<Symbol*>::empty()); }
  SymbolList* tail () const
  { return static_cast<SymbolList*>(List<Symbol*>::tail()); }
  SymbolList (Symbol* head, SymbolList* tail) :
    List<Symbol*>(head,tail) {}
};


class FunctionSymbol :
  public Symbol
{
public:
  FunctionSymbol (const string& name, int arity, int& count);
  
  USE_ALLOCATOR(FunctionSymbol);
}; 


class PredicateSymbol :
  public Symbol
{
public:
  PredicateSymbol (const string& name, int arity, int& count) 
    : Symbol (name,arity,count) {}

  
  bool isEquality () const { return _number == 0; }

  
  USE_ALLOCATOR(PredicateSymbol);
}; 


ostream& operator << (ostream& str, const Symbol* p);

#endif
