
/*
 * File TermFunIterator.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "TermFunIterator.hpp"
#include "Tracer.hpp"
#include "Literal.hpp"


TermFunIterator::TermFunIterator (const Term& t)
  : _current(const_cast<Term*>(&t)),
    _reserve(0)
{
} 



TermFunIterator::TermFunIterator (const TermList& ts)
  : _current(0),
    _reserve(new Element(ts,0))
{
} 



TermFunIterator::TermFunIterator (const Atom& a)
  : _current(0),
    _reserve(new Element(a.args(),0))
{
} 



TermFunIterator::TermFunIterator (const Literal& l)
  : _current(0),
    _reserve(new Element(l.atom().args(),0))
{
} 



bool TermFunIterator::hasNext ()
{
  TRACER("TermFunIterator::hasNext");

  for (;;) { 
    if (! _current) { 
      
      while (_reserve) {
	TermList* first = _reserve->_list;
	if (first->isEmpty()) {
	  Element* rest = _reserve->_rest;
	  delete _reserve;
	  _reserve = rest;
	}
	else { 
	  _current = const_cast<Term*>(&(first->head()));
	  _reserve->_list = const_cast<TermList*>(&(first->tail()));
	  goto current_exists;
	}
      }
      
      return false;
    }

    
  current_exists:
    switch (_current->tag()) {
    case VAR:
    case NUMERIC:
      _current = 0;
      break;
    case COMPOUND:
      return true;
    }
  }
} 



const FunctionSymbol& TermFunIterator::next ()
{
  const FunctionSymbol& result = _current->functor();
  _reserve = new Element(_current->args(), _reserve);
  _current = 0;
  return result;
} 



TermFunIterator::~TermFunIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 
