
/*
 * File TheoryFinder.hpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#ifndef __TheoryFinder__
#define __TheoryFinder__

class Atom;
class Clause;
class Formula;
class Unit;
class Symbol;
class Problem;
class Signature;


enum Axiom {
  
  AX_A,
  
  AX_C,
  
  AX_EXTENSIONALITY,
  
  AX_SUBSET,
  
  AX_SUBSET1
};


class TheoryFinder {
public:
  class Record;

  TheoryFinder (const Problem&);
  ~TheoryFinder ();
  int search ();
  
  const Record* records () { return _records; }
private:
  class MatcherState;
  class Matcher;
  class AtomMatcher;
  class LiteralListMatcher;
  class ClauseMatcher;
  class FormulaMatcher;
  class QuantifiedFormulaMatcher;
  class FormulaPairMatcher;
  class FormulaListMatcher;

  bool matchAll (const Unit& unit);
  bool matchAll (const Clause&);
  bool matchAll (const Formula&);
  bool matchC (const Clause&);
  bool matchC (const Formula&);
  bool matchC (const Atom&);
  bool matchExtensionality (const Formula&);
  bool matchExtensionality (const Clause&);
  bool matchSubset (const Formula&);
  bool matchSubset1 (const Clause&);
  void stripForall (Formula& f);

  
  enum Code {
    CODE_NEWVAR = 0,
    CODE_OLDVAR = 1,
    CODE_NEWSYM = 2,
    CODE_OLDSYM = 3,
    CODE_AND = 4,
    CODE_OR = 5,
    CODE_IMP = 6,
    CODE_NOT = 7,
    CODE_IFF = 8,
    CODE_XOR = 9,
    CODE_FORALL = 10,
    CODE_EXISTS = 11,
    CODE_EQL = 12
  };

  
  const Problem& _problem;
  
  Record* _records;
  
  Unit* _currentUnit;
  
  static char _commutativityAtom[];
};


class TheoryFinder::Record 
{
public:
  Record (Axiom kind,Unit& u,const Symbol* s1,Record* rec);
  Record (Axiom kind,Unit& u,const Symbol* s1,const Symbol* s2,Record* rec);
  Record (Axiom kind,Unit& u,
	  const Symbol* s1,
	  const Symbol* s2,
	  const Symbol* s3,
	  Record* rec);

  
  Axiom kind;
  
  Unit unit;
  
  const Symbol* symbol1;
  
  const Symbol* symbol2;
  
  const Symbol* symbol3;
  
  Record* next;

  
  CLASS_NAME("TheoryFinder::Record");
  USE_ALLOCATOR(Record);
}; 


#endif
