
/*
 * File Tracer.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include <climits>
#include "Tracer.hpp"
#include "XML.hpp"

#if DEBUG_TRACE_ALL

#include <iostream>


#define CP_LAST_POINT 0

#define NUMBER_OF_POINTS_TO_OUTPUT 200

MyTracer::MyTracer (const string& fun) 
  :
  _fun (fun),
  _previous (_current)
{

  _current = this;
  controlPoint(_fun,CP_ENTRY);
} 


MyTracer::~MyTracer () 
{

  _current = _previous;
  controlPoint(_fun,CP_EXIT);
} 



void MyTracer::controlPoint (const string& description, ControlPointKind cpk)
{








  _lastPointKind = cpk;
  _passedControlPoints++;
  _lastControlPoint = description;
  if (_passedControlPoints <= CP_LAST_POINT &&
      _passedControlPoints > CP_LAST_POINT - NUMBER_OF_POINTS_TO_OUTPUT) {
    outputLastControlPoint(cout);
  }
} 



void MyTracer::controlPoint (const string& description)
{
  controlPoint (description,CP_MID);
} 



void MyTracer::outputLastControlPoint (ostream& str)
{
  str << _passedControlPoints 
      << ": " 
      << (_lastPointKind == CP_ENTRY ? '+' : 
          _lastPointKind == CP_EXIT ? '-' : ' ')
      << _lastControlPoint 
      << "\n";
} 



void MyTracer::printStack (ostream& str)
{
  int depth = 0;

  str << "Control points passed: " << _passedControlPoints << "\n"
      << "last control point:\n";
  outputLastControlPoint(str);
  _current->printStack (str, depth);
} 



XML::Element* MyTracer::stackToXML ()
{
  XML::DeepElement* xml = new XML::DeepElement("tracer_stack");
  for (MyTracer* s = _current; s; s = s->_previous) {
    xml->addChild(new XML::TextElement("function",s->_fun));
  }
  return xml;
} 



void MyTracer::printStack (ostream& str, int& depth)
{
  if (! this) { 
    return;
  }
  _previous->printStack (str, depth);
  for (int i = depth; i > 0; i--) {
    str << ' ';
  }
  str << _fun << "\n";
  depth ++;
} 


#endif 
