//
// File:         BackwardSubsumption.hpp
// Description:  Interface for several BS related modules.
// Created:      Feb 26, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 03, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Slightly messed the code up in order to compile it by VC++ 6.0.
//============================================================================
#ifndef BACKWARD_SUBSUMPTION_H
//=============================================================================
#define BACKWARD_SUBSUMPTION_H 
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelInterceptMacros.hpp"
#include "BSIndex.hpp"
#include "BSMachine.hpp"
#include "ClauseNumIndex.hpp"
#include "Clause.hpp"
#include "Term.hpp"
#ifndef NO_DEBUG
#  include "SubsumptionCheck.hpp" 
#endif
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BACKWARD_SUBSUMPTION
#define DEBUG_NAMESPACE "BackwardSubsumption"
#endif
#include "debugMacros.hpp"
//============================================================================


namespace VK
{

class BackwardSubsumption
{
 public:
  typedef BSIndex::LTPair LTPair;  
  
  typedef BSIndex::LAPair LAPair;
  typedef BSIndex::LATTriple LATTriple;
  typedef BSIndex::LList LList;
  typedef BSIndex::LTList LTList;
  typedef BSIndex::LAList LAList;
  typedef BSIndex::LATList LATList;
  typedef BSIndex::CSkipList CSkipList;
  typedef BSIndex::CLSkipList CLSkipList;
  typedef BSIndex::CLTSkipList CLTSkipList;
  typedef BSIndex::CLASkipList CLASkipList;
  typedef BSIndex::CLATSkipList CLATSkipList;
  
 public:
  // VC++ 6.0 breaks down when compiling any
  // constructor or destructor definition for BackwardSubsumption.

  /**************
  BackwardSubsumption(ClauseNumIndex* clauseNumIndex) : 
    _integrator(&_index),
    _removal(&_index),
    _retrieval(&_index),
    _termTraversal(DOP("BackwardSubsumption::_termTraversal")),  
    _clauseNumIndex(clauseNumIndex),
    _setMode(false)
  {
    CALL("constructor BackwardSubsumption(ClauseNumIndex* clauseNumIndex)");
  }; // BackwardSubsumption(ClauseNumIndex* clauseNumIndex)

  ~BackwardSubsumption()
  {
    CALL("destructor ~BackwardSubsumption()");
    unlock();
  };
  ***************/

  void init(ClauseNumIndex* clauseNumIndex)
  {
    CALL("init(ClauseNumIndex* clauseNumIndex)");
    BK::init(_index);
    _integrator.init(&_index);
    _removal.init(&_index);
    _retrieval.init(&_index);
    BK::init(_machine);
#ifdef DEBUG_NAMESPACE
    BK::init(_termTraversal,"BackwardSubsumption::_termTraversal");
#else
    BK::init(_termTraversal);
#endif

    _clauseNumIndex = clauseNumIndex;
    _setMode = false;
    unlock();
#ifdef DEBUG_NAMESPACE
    BK::init(_debugSubsumptionCheck);
#endif
    
  };


  void destroy()
  {
    CALL("destroy()");
#ifdef DEBUG_NAMESPACE
    BK::destroy(_debugSubsumptionCheck);
#endif
    BK::destroy(_termTraversal);
    BK::destroy(_machine);
    _retrieval.destroy();
    _removal.destroy();
    _integrator.destroy();
    BK::destroy(_index);
    BK_CORRUPT(*this);
  };

  void reset()
  {
    CALL("reset()");
    _index.reset();
  }; // void reset()
  
  bool isLocked() const { return _isLocked; };
  
  bool subsumptionAllowedInSetMode(const Clause* subsumingCl,const Clause* subsumedCl)
    {
      return subsumingCl->NumOfPosEq() <= subsumedCl->NumOfPosEq(); 
    };
  bool subsumptionAllowedInMultisetMode(const Clause* subsumingCl,const Clause* subsumedCl)
    {
      return true; 
    };     
  void integrate(Clause* cl);
  void remove(Clause* cl);
  bool query(Clause* cl);
  void abandonCurrentQuery();
  bool subsumeNext() 
    {
      CALL("subsumeNext()");
      ASSERT(isLocked());
      bool res;
      if (_setMode) 
	{
	  res = subsumeNextSetMode();
	}
      else 
         res = subsumeNextMultisetMode();
  
#ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
      if (!res) { INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING_RETRIEVAL_TOTAL; };
#endif 

      return res;
    };
  Clause* subsumed() const 
  { 
    CALL("subsumed() const");
    ASSERT(isLocked());
    return _recentlySubsumedCl; 
  };
  void setSetMode(bool fl) { _setMode = fl; }; 

 private:
  void lock() { _isLocked = true; };
  void unlock() { _isLocked = false; };
  bool subsumeNextSetMode();
  bool subsumeNextMultisetMode();
  bool preproQuery(const Clause* cl);
  bool preproQueryLit(const TERM* lit);
  bool preproQuerySymLit(const TERM* lit);
  bool preproQueryComplexTerm(const TERM* t);
  bool preproQueryComplexTermArguments(const TERM* t);
  bool preproQuerySymLitComplexTerm(const TERM* t);
  bool preproQuerySymLitComplexTermArguments(const TERM* t);
 private:
  BSIndex _index;
  BSIndex::Integrator _integrator;
  BSIndex::Removal _removal;
  BSIndex::Retrieval _retrieval;
  BSMachine _machine; 
  TERM::Traversal _termTraversal;
  ClauseNumIndex* _clauseNumIndex;
  Clause* _recentlySubsumedCl;
  Clause* _queryClause; 
  bool _setMode;
  bool _isLocked;
#ifdef DEBUG_NAMESPACE
  SubsumptionCheck _debugSubsumptionCheck;
#endif
}; // class BackwardSubsumption

}; // namespace VK

//======================================================================
#endif
