// Revised:     May 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//              Representation of indexed objects changed: ulong* ---> ulong
// Revised:     Jun 20, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//              Bug fix: el->linkEqArgs() was missing
//              in CN_QUERY::Load(TmpLitList& component,ulong numOfVars)
//===========================================================
#include "ComponentName.hpp"
#include "Signature.hpp"
#include "GlobalStopFlag.hpp"
#include "TempLitListExtras.hpp"
#include "TemporaryLiteral.inl"
#include "ClauseAssembler.hpp"
#include "Clause.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
#define DEBUG_NAMESPACE "ComponentName"
#endif 
#include "debugMacros.hpp"
//=================================================

using namespace BK;
using namespace Gem;
using namespace VK;



ComponentName::ComponentName(ClauseAssembler* nameIntroAssembler,
			     BK::Receiver<Clause*>* nameIntroReceiver) : 
  _nameIntroAssembler(nameIntroAssembler),
  _nameIntroReceiver(nameIntroReceiver),
  variantIndex(subst), 
  backtrack_lit(DOP("backtrack_lit")),
  backtrack_instr(DOP("backtrack_instr")),
  backtrack_cursor(DOP("backtrack_cursor"))
{
  CALL("constructor ComponentName()");
  TemporaryLiteral::Extras::freshLiteral(dummyTmpLiteral);
  variantIndex.SetUseSigFilters(false);
  query.SetCollectSignature(false); 
  dummyComponent.push(dummyTmpLiteral);
  DOP(backtrack_lit.freeze());
};

ComponentName::~ComponentName() 
{ 
  CALL("destructor ~ComponentName()");
};


bool ComponentName::findName(BasicTempLitList& component,Clause*& nameIntroClause)
{
  CALL("findName(BasicTempLitList& component,Clause*& nameIntroClause)");

  if (!query.Load(component,
		  TempLitListExtras::countDifferentVariables(component)))
    return false;
  
  void* indexedObject;
  bool res = findVariant(indexedObject);

  nameIntroClause = static_cast<Clause*>(indexedObject);

  ASSERT(!res || nameIntroClause);

  return res;
}; // bool ComponentName::findName(BasicTempLitList& component,Clause*& nameIntroClause)




Clause* ComponentName::registerNameIntro(const BasicTempLitList& component,
					 const TemporaryLiteral& definedLit)
{
  CALL("registerNameIntro(const BasicTempLitList& component,..)");
  Clause* res = 
    _nameIntroAssembler->assembleNameIntro(definedLit,component);
  _nameIntroReceiver->receive(res);
  variantIndex.Integrate(component,
			 static_cast<void*>(res));
  return res;
}; // Clause* ComponentName::registerNameIntro(const BasicTempLitList& component,..)




Clause* ComponentName::registerNameIntro(const Flatterm* term,
					 const TemporaryLiteral& definedLit,
					 const BasicTempLitList& definitionBody)
{
  CALL("registerNameIntro(const Flatterm* term,..)");  
  Clause* res = 
    _nameIntroAssembler->assembleNameIntro(definedLit,definitionBody);
  _nameIntroReceiver->receive(res);
  makeDummyLit(term);
  variantIndex.Integrate(dummyComponent,
			 reinterpret_cast<void*>(res));
  return res;
}; // Clause* ComponentName::registerNameIntro(const Flatterm* term,..)


inline bool ComponentName::findVariant(void*& indexedObject)
{ 
  CALL("ComponentName::findVariant(void*& indexedObject)");

  FS_COMMAND* curr_instr = variantIndex.Tree();
  if (!curr_instr) return false;
 
  query.EndOfQuery();
  const Flatterm* cursor = 0;
  TERM curr_sym;
  TERM curr_sym2;
  CN_QUERY::LIT* curr_lit;
  CN_QUERY::LIT* next_lit;

  DOP(backtrack_lit.unfreeze());
  if (!backtrack_lit.tryToExpand(variantIndex.maxlNumOfLits()))
    return false;
  DOP(backtrack_lit.freeze());

  backtrack_lit.reset();
  backtrack_instr.reset();
  backtrack_cursor.reset();
  //-----------------------------------------------------------

    
 check_tag:
  switch (curr_instr->Tag())
    {
    case FS_COMMAND::LIT_HEADER :
      curr_lit = query.FirstLit(curr_instr->HeaderNum());
    check_lit:
      if (curr_lit)
        {
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_lit;
	    }
	  else
	    { 
	      curr_lit->Capture();
             
	      backtrack_lit.push(curr_lit); // to be released on backtracking
	      backtrack_instr.push(curr_instr);

	      cursor = curr_lit->Args();
	      curr_instr = curr_instr->Next();
	      goto check_tag;
	    };
        }
      else // no literal with this header
        goto try_fork;
     
    case FS_COMMAND::ORD_EQ :
      curr_lit = query.FirstLit(curr_instr->HeaderNum());
    check_ord_eq_lit:
      if (curr_lit)
        { 
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_ord_eq_lit;
	    }
	  else
	    { 
	      curr_lit->Capture();

	      backtrack_lit.push(curr_lit); // to be released on backtracking
	      backtrack_instr.push(curr_instr);

	      cursor = curr_lit->Arg1();
	      curr_instr = curr_instr->Next();
	      goto check_tag;
	    };
        }
      else // no ordered equations with this polarity
        goto try_fork;
       
       
    case FS_COMMAND::UNORD_EQ : 
      curr_lit = query.FirstLit(curr_instr->HeaderNum());
    check_unord_eq_lit:
      if (curr_lit)
        {
	  if (curr_lit->Captured())
	    {
	      curr_lit = curr_lit->Next();  
	      goto check_unord_eq_lit;
	    }
	  else
	    {
	      curr_lit->Capture(); 
           
	      backtrack_lit.push(curr_lit);

	      curr_instr->SwapOn();
        
	      backtrack_instr.push(curr_instr);
        
	      cursor = curr_lit->Arg1();
        
	      curr_instr = curr_instr->Next();
        
	      goto check_tag;
	    };
        }
      else // no equations with this polarity (neither ordered nor unordered)
        goto try_fork; 
       
    case FS_COMMAND::PROP_LIT :
      if (query.FirstLit(curr_instr->HeaderNum())) // the query contains such literal
	{
	  if (curr_instr->Fork()) 
	    {           
	      backtrack_instr.push(curr_instr);
	    }; 
	  curr_instr = curr_instr->Next(); 
	  goto check_tag;
	}
      else // no such literal in the query, try fork
	goto try_fork; 
       
       
    case FS_COMMAND::SECOND_EQ_ARG :
     
      if (cursor->IsBackJump()) 
	{
	  cursor = cursor->after();
	}; 
      
      curr_instr = curr_instr->Next(); 
      goto check_tag; 
       
    case FS_COMMAND::FUNC :
      curr_sym = cursor->Symbol();
      if (curr_sym.isVariable()) goto backtrack;
      
    compare_func:
      if (curr_sym == curr_instr->Func()) 
        {
	  cursor++;
	  curr_instr = curr_instr->Next();
	  goto check_tag; 
        };
      if (curr_sym < curr_instr->Func()) goto backtrack;
      curr_instr = curr_instr->Fork();
      if (!curr_instr) goto backtrack;
       
      if (curr_instr->IsFunc()) 
	{
          goto compare_func;
	};
                
      goto check_tag;
         

    case FS_COMMAND::VAR :
      
      if (!cursor->isVariable()) goto try_fork;
      *(curr_instr->Var()) = cursor;
      cursor = cursor->after();
      curr_instr = curr_instr->Next();
      goto check_tag;

    case FS_COMMAND::COMPARE_NF :
      ASSERT((*(curr_instr->Var1()))->isVariable());
      ASSERT((*(curr_instr->Var2()))->isVariable());
      if ((*(curr_instr->Var1()))->symbolRef() == 
          (*(curr_instr->Var2()))->symbolRef())
	{
	  curr_instr = curr_instr->Next(); 
	  goto check_tag;     
	}
      else goto backtrack;
 
      
    case FS_COMMAND::COMPARE :
      ASSERT((*(curr_instr->Var1()))->isVariable());
      ASSERT((*(curr_instr->Var2()))->isVariable());
      if ((*(curr_instr->Var1()))->symbolRef() == 
          (*(curr_instr->Var2()))->symbolRef())
	{
	  if (curr_instr->Fork()) 
	    {

	      backtrack_instr.push(curr_instr);
	      backtrack_cursor.push(cursor);
	    };
	  curr_instr = curr_instr->Next(); 
	  goto check_tag; 
	}
      else goto try_fork;

    case FS_COMMAND::SIG_FILTER :
      if (curr_instr->Functors()->subset(*(query.Signature())))
	{
	  curr_instr = curr_instr->Next();       
	  goto check_tag;
	};
      goto backtrack;

    case FS_COMMAND::SUCCESS : 
      if ((curr_instr->NumOfDiffVars() == query.NumOfDiffVars())
	  && (curr_instr->NumOfLits() == query.NumOfLits()))
	{
	  indexedObject = curr_instr->IndexedObject();
	  return true;
	}
      else goto try_fork; 
      
#ifdef DEBUG_NAMESPACE  
    default: ICP("ICP1"); return false; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: return false; 
#endif
#endif  
    }; 
   
 try_fork:
  curr_instr = curr_instr->Fork();
  if (curr_instr) goto check_tag;
  
  // no fork, try to backtrack
 backtrack: 
  if (backtrack_instr.empty())
    //if (next_backtrack_instr == backtrack_instr) 
    {
      return false;
    }; 
   
  curr_instr = backtrack_instr.pop();
   
  switch (curr_instr->Tag())
    {
    case FS_COMMAND::LIT_HEADER :
 
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();
 
      ASSERT(curr_lit); 
      ASSERT(curr_lit->Captured());
      curr_lit->Release();
      curr_lit = curr_lit->Next();       
      goto check_lit;
     
    case FS_COMMAND::ORD_EQ :
 
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();

      ASSERT(curr_lit); 
      ASSERT(curr_lit->Captured());
      curr_lit->Release();
      curr_lit = curr_lit->Next();       
      goto check_ord_eq_lit;      

    case FS_COMMAND::UNORD_EQ :
 
      GlobalStopFlag::check();
      curr_lit = backtrack_lit.pop();

      ASSERT(curr_lit);
      ASSERT(curr_lit->Captured());
      if (curr_instr->SwapFlag()) 
	{        
	  // curr_lit remains captured, so there is no curr_lit->Release() here 
	  next_lit = curr_lit->Next();

	  backtrack_lit.push(curr_lit);

	  curr_instr->SwapOff();

	  backtrack_instr.push(curr_instr);
         
	  cursor = curr_lit->Arg2();
	  curr_instr = curr_instr->Next();
	  goto check_tag;
	}
      else
	{       
	  curr_lit->Release();        
	  curr_lit = curr_lit->Next(); 
	  goto check_unord_eq_lit;
	}; 
       
    case FS_COMMAND::PROP_LIT :
      curr_instr = curr_instr->Fork();
      goto check_tag; 
     
    case FS_COMMAND::COMPARE :
      curr_instr = curr_instr->Fork();

      cursor = backtrack_cursor.pop();

      goto check_tag;
    
#ifdef DEBUG_NAMESPACE  
    default: ICP("ICP2"); return false; 
#else
#ifdef _SUPPRESS_WARNINGS_
    default: return false; 
#endif
#endif 
    };
  ICP("ICP3");
    return false;


}; // bool ComponentName::findVariant(void*& indexedObject)


//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
#define DEBUG_NAMESPACE "CN_QUERY"
#endif 
#include "debugMacros.hpp"
//=================================================

CN_QUERY::CN_QUERY() : 
  _litsWithHeader(static_cast<LIT*>(0),"CN_QUERY::_litsWithHeader"),
  _posOrdEqHeaderNum(TERM::positiveOrdEqHeader().HeaderNum()),
  _negOrdEqHeaderNum(TERM::negativeOrdEqHeader().HeaderNum()),
  _presentHeaders(DOP("CN_QUERY::_presentHeaders")),
  collect_signature(true) // default
{
  CALL("constructor CN_QUERY()");
  DOP(_litsWithHeader.freeze());
  signature = 0; 
};


void CN_QUERY::ResetQuery()
{
  CALL("ResetQuery()");
  _litCache.reset();
  while (_presentHeaders.nonempty())
    {
      _litsWithHeader[_presentHeaders.pop()] = 0;
    };

  last_pos_unord_eq = 0;
  last_neg_unord_eq = 0;
  if (collect_signature) 
    {
      if (signature) delete signature;
      signature = new FunSet();
    };
  numOfLits = 0;
}; // void CN_QUERY::ResetQuery()

inline void CN_QUERY::LinkEquations()
{
  CALL("LinkEquations()") 
    if (last_pos_unord_eq) 
      last_pos_unord_eq->SetNext(FirstLit(_posOrdEqHeaderNum));
  if (last_neg_unord_eq) 
    last_neg_unord_eq->SetNext(FirstLit(_negOrdEqHeaderNum));
}; 

void CN_QUERY::EndOfQuery()
{
  LinkEquations();
}; 


bool CN_QUERY::PushLit(const Flatterm* lit)
{
  CALL("PushLit(const Flatterm* lit)");
  ASSERT(!lit->Symbol().IsEquality());

  DOP(_litsWithHeader.unfreeze());
  LIT** lit_lst_ptr = &(_litsWithHeader.sub(lit->Symbol().HeaderNum()));
  DOP(_litsWithHeader.freeze());


  LIT* currLit = _litCache.tryToReserveObject();
  if (!currLit) return false;

  if (*lit_lst_ptr)
    {
      currLit->Set(lit + 1,*lit_lst_ptr);
    }
  else // no literals with this header yet
    {
      _presentHeaders.pushSafe(lit->Symbol().HeaderNum());
      currLit->Set(lit + 1,static_cast<LIT*>(0));
    };
  *lit_lst_ptr = currLit; 
  currLit->Release(); 
  if (collect_signature) CollectFunctors(lit);
  return true;
}; // bool CN_QUERY::PushLit(const Flatterm* lit)
  
  
    
bool CN_QUERY::PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
{
  CALL("PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)");
  ASSERT(header.IsUnorderedEq());
     

  DOP(_litsWithHeader.unfreeze());
  LIT** lit_lst_ptr = &(_litsWithHeader.sub(header.HeaderNum()));
  DOP(_litsWithHeader.freeze());


  LIT* currLit = _litCache.tryToReserveObject();
  
  if (!currLit) return false;

  if (*lit_lst_ptr)
    {
      currLit->Set(arg1,arg2,*lit_lst_ptr);
    }
  else // no literals with this header yet
    {

      _presentHeaders.pushSafe(header.HeaderNum());
      currLit->Set(arg1,arg2,static_cast<LIT*>(0));
      if (header.Negative()) { last_neg_unord_eq = currLit; }
      else // positive
	last_pos_unord_eq = currLit;
    };
  *lit_lst_ptr = currLit;
  currLit->Release();
  if (collect_signature) 
    {
      AddFunctor(header.HeaderNum());
      CollectFunctors(arg1); 
      CollectFunctors(arg2);
    };
  return true;
}; // bool CN_QUERY::PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
    
bool CN_QUERY::PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)
{
  CALL("PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)");
  ASSERT(header.IsEquality() && header.IsOrderedEq());

  DOP(_litsWithHeader.unfreeze());
  LIT** lit_lst_ptr = &(_litsWithHeader.sub(header.HeaderNum()));
  DOP(_litsWithHeader.freeze());


  LIT* currLit = _litCache.tryToReserveObject();
  if (!currLit) return false;
  if (*lit_lst_ptr)
    {
      currLit->Set(arg1,arg2,*lit_lst_ptr);
    }
  else // no literals with this header yet
    {

      _presentHeaders.pushSafe(header.HeaderNum());
      currLit->Set(arg1,arg2,static_cast<LIT*>(0));
    };
  *lit_lst_ptr = currLit;
  currLit->Release(); 
  if (collect_signature) 
    {
      header.MakeUnordered();
      AddFunctor(header.HeaderNum());
      CollectFunctors(arg1); 
      CollectFunctors(arg2);
    };
  return true;
}; // bool CN_QUERY::PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2)


bool CN_QUERY::Load(BasicTempLitList& component,ulong numOfVars)
{
  CALL("Load(BasicTempLitList& component,ulong numOfVars)");
  ResetQuery(); 
  numOfDiffVars = numOfVars;
  numOfLits = component.length();
  for (BasicTempLitList::Element* el = component.first();
       el;
       el = el->next())
    {
      if (el->literal()->isEquality())
	{	      
	  el->literal()->linkEqArgs();
	  if (el->literal()->isOrderedEquality())   
	    {
	      if (!PushOrdEq(el->literal()->header(),
			     el->literal()->arg1(),
			     el->literal()->arg2()))
		return false;
	    }      
	  else 
	    {
	      if (!PushUnordEq(el->literal()->header(),
			       el->literal()->arg1(),
			       el->literal()->arg2()))
		return false;
	    };
	} 
      else 
	{
	  if (!PushLit(el->literal()->flatterm()))
	    return false;
	};
    };
  return true;
}; // bool CN_QUERY::Load(BasicTempLitList& component,ulong numOfVars)


//============================================================================

