// Revised:     May 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//              Representation of indexed objects changed: ulong* ---> ulong
//===========================================================
#ifndef COMPONENT_NAME_H
//===========================================================
#define COMPONENT_NAME_H
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "Receiver.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "fs_code.hpp"
#include "fs_command.hpp"
#include "ExpandingStack.hpp"
#include "ObjectCache.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "GlobAlloc.hpp"
#include "Array.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
#define DEBUG_NAMESPACE "CN_QUERY"
#endif 
#include "debugMacros.hpp"
//=================================================
namespace VK
{
  class Clause;
  class ClauseAssembler;

class CN_QUERY
{
 public:
  class LIT
    {    
    public:  
      LIT() {};
      ~LIT() {};
      void* operator new(size_t)
	{
	  CALL("LIT::operator new(size_t)");
	  return BK::GlobAlloc::allocate(sizeof(LIT));
	};
      void operator delete(void* obj)
	{
	  CALL("LIT::operator delete(void* obj)");
	  BK::GlobAlloc::deallocate(obj,sizeof(LIT));
	};
      const Flatterm* Args() const { return arg1; };
      void SetArgs(const Flatterm* a) { arg1 = a; };      
      const Flatterm* Arg1() const { return arg1; }; // for equations
      const Flatterm* Arg2() const { return arg2; }; // for equations
      void SetArgs(const Flatterm* a1,const Flatterm* a2)
	{ arg1 = a1; arg2 = a2; };
      LIT* Next() const { return next; };
      void SetNext(LIT* nl) { next = nl; };
      void Set(const Flatterm* a,LIT* nl) { arg1 = a; next = nl; };
      void Set(const Flatterm* a1,const Flatterm* a2,LIT* nl) 
	{ arg1 = a1; arg2 = a2; next = nl; };
      void Capture() { captured = true; };
      void Release() { captured = false; };
      bool Captured() const { return captured; };
    public:
      ostream& output(ostream& str) const { return (Args() - 1)->outputTerm(str); };
    private:
      const Flatterm* arg1;
      const Flatterm* arg2;
      bool captured;
      LIT* next;
    }; // class LIT

 public:
  CN_QUERY();

  ~CN_QUERY() { CALL("destructor ~CN_QUERY()"); };

  void ResetQuery();
  void SetCollectSignature(bool fl) 
    { 
      collect_signature = fl; 
    };
  void EndOfQuery();
  bool PushLit(const Flatterm* lit);
  bool PushUnordEq(TERM header,const Flatterm* arg1,const Flatterm* arg2);
  bool PushOrdEq(TERM header,const Flatterm* arg1,const Flatterm* arg2);
    
  LIT* FirstLit(ulong header_num)  
    {
      CALL("FirstLit(ulong header_num)");
      if (header_num >= _litsWithHeader.size()) return 0;
      return _litsWithHeader[header_num];
    }; 
   

  const FunSet* Signature() const { return signature; }; 

  ulong NumOfDiffVars() const { return numOfDiffVars; };
  ulong NumOfLits() const { return numOfLits; }
  bool Load(BasicTempLitList& component,ulong numOfVars);

 private:
  void LinkEquations();
  void AddFunctor(ulong fun_num)
    {
      if (!(signature->write((size_t)fun_num)))
	{
	  FunSet* new_sig = signature->add((size_t)fun_num);
	  delete signature;
	  signature = new_sig; 
	};
    };
  void CollectFunctors(const Flatterm* t)
    {
      CALL("CollectFunctors(const Flatterm* t)");
      ASSERT(collect_signature);
      Flatterm::ConstIterator iter(t);
      TERM sym;
      while (iter.NotEndOfTerm())
	{
	  sym = iter.CurrPos()->symbolRef();
	  if (sym.isComplex()) AddFunctor(sym.HeaderNum());       
	  iter.NextSym();
	};
    };
    
 public: 
  ostream& output(ostream& str) const
    {
      for (ulong p = 0; p < _presentHeaders.size(); p++)
	for (const LIT* lit = _litsWithHeader[_presentHeaders.nth(p)]; lit; lit = lit->Next())
	  lit->output(str) << '\n';
      return str;
    };

 
 private: // structure 
  BK::ObjectCache<BK::GlobAlloc,64UL,LIT> _litCache;

  BK::Array<BK::GlobAlloc,LIT*,BK::DefaultRelocation<LIT*>,512UL> _litsWithHeader;
  ulong _posOrdEqHeaderNum;
  ulong _negOrdEqHeaderNum;

  BK::ExpandingStack<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,32UL> _presentHeaders;

  LIT* last_pos_unord_eq;
  LIT* last_neg_unord_eq;
  bool collect_signature;
  FunSet* signature;
  ulong numOfDiffVars;
  ulong numOfLits;
}; // class CN_QUERY
 
}; // namespace VK

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_COMPONENT_NAME
#define DEBUG_NAMESPACE "ComponentName"
#endif 
#include "debugMacros.hpp"
//=================================================
namespace VK
{
class ComponentName 
{
 public:
  ComponentName(ClauseAssembler* nameIntroAssembler,
		BK::Receiver<Clause*>* nameIntroReceiver);
  ~ComponentName();
  

  bool findName(BasicTempLitList& component,
		Clause*& nameIntroClause);
  

  Clause* registerNameIntro(const BasicTempLitList& component,
			    const TemporaryLiteral& definedLit);

  bool findName(const Flatterm* term,Clause*& nameIntroClause)
    {
      CALL("findName(const Flatterm* term,Clause*& nameIntroClause)");
      makeDummyLit(term);
      return findName(dummyComponent,nameIntroClause);
    };
  
  Clause* registerNameIntro(const Flatterm* term,
			    const TemporaryLiteral& definedLit,
			    const BasicTempLitList& definitionBody);

  /*
  void integrate(const Flatterm* term,ulong name)
    {
      CALL("integrate(const Flatterm* term,ulong name)");
      makeDummyLit(term);
      variantIndex.Integrate(dummyComponent,
			     reinterpret_cast<void*>(name));
    };
  */

#ifndef NO_DEBUG_VIS 
  ostream& outputQuery(ostream& str) { return query.output(str); };
#endif

 private: 
   
  bool findVariant(void*& indexedObject);  
 
  void makeDummyLit(const Flatterm* term)
    {
      CALL("makeDummyLit(const Flatterm* term)");
      LiteralProperties prop; // default value
      dummyTmpLiteral->pushNonpropHeader(TERM::dummyPositiveHeader1(),
					 prop);
      dummyTmpLiteral->pushTerm(term); // very inefficient
      dummyTmpLiteral->endOfLiteral();
    };

 private:
  ClauseAssembler* _nameIntroAssembler;
  BK::Receiver<Clause*>* _nameIntroReceiver;
  const Flatterm* subst[VampireKernelConst::MaxClauseSize]; 
  FS_CODE variantIndex;
  CN_QUERY query;
  TemporaryLiteral::Extras::Handle dummyTmpLiteral;
  BasicTempLitList dummyComponent;

  BK::ExpandingStack<BK::GlobAlloc,CN_QUERY::LIT*,BK::FastRelocation<CN_QUERY::LIT*>,32UL> backtrack_lit;

  BK::Stack<FS_COMMAND*,MAX_FS_CODE_DEPTH> backtrack_instr;
  BK::Stack<const Flatterm*,MAX_FS_CODE_DEPTH> backtrack_cursor; 
}; // class ComponentName 
}; // namespace VK
//================================================
#undef DEBUG_NAMESPACE
//===============================================================
#endif
