//
// File:         FTUnification.hpp
// Description:  Unification for flatterms.
// Created:      Jun 6, 2000, 15:00
// Author:       Alexandre Riazanov
// mail:         riazanov@cs.man.ac.uk
//====================================================
#ifndef FT_UNIFICATION_H
//====================================================
#define FT_UNIFICATION_H
#ifndef NO_DEBUG_VIS
 #include <iostream>
#endif
#include "jargon.hpp"
#include "Stack.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp" 
#include "Flatterm.hpp" 
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FT_UNIFICATION
 #define DEBUG_NAMESPACE "FTUnification"
#endif 
#include "debugMacros.hpp"
//================================================= 

namespace VK
{
class FTUnification
{
 public:
  FTUnification() 
   : _bound("_bound"),
     _borderVar("_borderVar"),
     _borderTerm("_borderTerm")
  { 
    CALL("constructor FTUnification()");
   for (ulong v = 0; v < 2*VampireKernelConst::MaxNumOfVariables; v++) _subst[v] = 0;
  };
  
  /********
  void init() 
  {
    CALL("init()");
    _bound.init();
    _borderVar.init();
    _borderTerm.init();
    for (ulong v = 0; v < 2*VampireKernelConst::MaxNumOfVariables; v++) 
      _subst[v] = 0;
  };
  ***********/

  ~FTUnification() 
  {
    CALL("destructor ~FTUnification()");
  };

  bool unify(const Flatterm* t1,const Flatterm* t2); 
  
  const Flatterm* subst(ulong var) const { return _subst[var]; };

  #ifndef NO_DEBUG
   bool equalWithSubst(const Flatterm* t1,const Flatterm* t2);
  #endif

 private: 
  void reset();
  bool occurs(ulong var,const Flatterm* term);
  void bind(ulong var,const Flatterm* term);
 private:
  const Flatterm* _subst[2*VampireKernelConst::MaxNumOfVariables];
  BK::Stack<ulong,2*VampireKernelConst::MaxNumOfVariables> _bound;
  BK::Stack<ulong,VampireKernelConst::MaxTermSize> _borderVar;
  BK::Stack<const Flatterm*,VampireKernelConst::MaxTermSize> _borderTerm;
}; // class FTUnification

}; // namespace VK



//================================================
#undef DEBUG_NAMESPACE
//================================================= 
#endif
