//
// File:         Flatterm.hpp
// Description:  Flatterms.
// Created:      Feb 13, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Fob 03, 2004
//               Minor bug in equal(..,const SymbolInfo* commutativityTable,..)
//               was fixed.
//====================================================
#ifndef FLATTERM_H
//=================================================================
#define FLATTERM_H
#include <iostream>
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Term.hpp"
#include "MultisetOfVariables.hpp"
#include "variables.hpp"
#include "Stack.hpp"
#include "Comparison.hpp"
#include "Receiver.hpp"
#include "SymbolInfo.hpp"
namespace VK
{
  class Clause;
};

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm"
#endif
#include "debugMacros.hpp"
//=================================================

namespace VK
{
  class Flatterm    
  { 
  public: 
    class Iterator;
    class ConstIterator;
    class Open;
    class OpenWithMemory;
  public:
    Flatterm() : 
      _isSubstitutionPosition(false),
      _isInputNonBasicPosition(false)
    {
    };
    ~Flatterm() {};
    TERM Symbol() const { return _symbol; };
    TERM& symbolRef() { return _symbol; };
    const TERM& symbolRef() const { return _symbol; };
    bool isVariable() const { return _symbol.isVariable(); };
    bool isComplex() const 
    {
      CALL("isComplex() const");
      return _symbol.isComplex(); 
    };
    ulong Var() const 
    {
      CALL("Var() const");
      ASSERT(isVariable());
      return _symbol.var(); 
    };
    ulong functor() const 
    { 
      CALL("functor() const");
      ASSERT(isComplex());
      return _symbol.functor(); 
    };
    ulong arity() const { return _symbol.arity(); };
    bool isSubstitutionPosition() const { return _isSubstitutionPosition; };
    bool isInputNonBasicPosition() const { return _isInputNonBasicPosition; };
    bool isNonBasicPosition() const 
    { 
      return isSubstitutionPosition() || isInputNonBasicPosition();
    };

    void SetSymbol(const TERM& sym) { _symbol = sym; };
    void SetSymbol(const TERM& sym,bool substPos) 
    { 
      _symbol = sym; 
      setSubstitutionPositionFlag(substPos); 
    };
    void SetVar(ulong var) { _symbol.MakeVar(var); _after = this + 1; };
    void SetVar(ulong var,bool substPos) 
    { 
      _symbol.MakeVar(var); 
      _after = this + 1; 
      setSubstitutionPositionFlag(substPos); 
    };
    void SetVar(const TERM& var) { _symbol = var; _after = this + 1; };
    void SetVar(const TERM& var,bool substPos) 
    { 
      _symbol = var; 
      _after = this + 1; 
      setSubstitutionPositionFlag(substPos); 
    };
    void SetConst(const TERM& csym) { _symbol = csym; _after = this + 1; }; 
    void SetConst(const TERM& csym,bool substPos) 
    { 
      _symbol = csym; 
      _after = this + 1;
      setSubstitutionPositionFlag(substPos); 
    }; 
    void setSubstitutionPositionFlag(bool substPos)
    {
      _isSubstitutionPosition = substPos;
    };
    void setInputNonBasicPositionFlag(bool inpNonBasPos)
    {
      _isInputNonBasicPosition = inpNonBasPos;
    };

    const Flatterm* after() const { return _after; };
    Flatterm* after() { return _after; };
    const Flatterm* next() const { return (this + 1); }; 
    Flatterm* next() { return (this + 1); };
    const Flatterm* previous() const { return (this - 1); }; 
    Flatterm* previous() { return (this - 1); };
    Flatterm* arg1() { return next(); };
    const Flatterm* arg1() const { return next(); };
    Flatterm* arg2() { return arg1()->after(); };
    const Flatterm* arg2() const { return arg1()->after(); };
    const Flatterm* lastSymbol() const { return after() - 1; };
    Flatterm* lastSymbol() { return after() - 1; };
    ulong subtermPosition(const Flatterm* subterm) const
    {
      return subterm - this;
    };
    const Flatterm* subtermIn(ulong position) const 
    {
      return this + position;
    };
    bool containsPosition(ulong position) const
    {
      return (this + position) <= _after; 
    };
    bool containsSubterm(const Flatterm* subterm) const 
    {
      return (subterm >= this) && (subterm < _after);
    };
    bool properlyContainsSubterm(const Flatterm* subterm) const 
    {
      return (subterm > this) && (subterm < _after);
    };

    void SetNextTerm(Flatterm* nt) { _after = nt; };
    long size() const { return (_after - this); };
    ulong Depth() const;
    bool Nonground() const;
    bool containsSymbol(const TERM& sym) const;
    bool equal(const Flatterm* t) const;
    bool equal(const Flatterm* t,
	       const SymbolInfo* commutativityTable,
	       bool& commutativityUsed) const;
    void commutativityLawsForEqual(const Flatterm* t,
				   const SymbolInfo* commutativityTable,
				   BK::Receiver<const Clause*>* laws) const;
    bool EqualArguments(const Flatterm* t) const;
    bool EqualArguments(const Flatterm* t,
			const SymbolInfo* commutativityTable,
			bool& commutativityUsed) const;
    void commutativityLawsForEqualArguments(const Flatterm* t,
					    const SymbolInfo* commutativityTable,
					    BK::Receiver<const Clause*>* laws) const;
    bool Complementary(const Flatterm* t) const;
    bool Complementary(const Flatterm* t,
		       const SymbolInfo* commutativityTable,
		       bool& commutativityUsed) const;
    void commutativityLawsForComplementary(const Flatterm* t,
					   const SymbolInfo* commutativityTable,
					   BK::Receiver<const Clause*>* laws) const;
    BK::Comparison CompareLex(const Flatterm* t) const;

    void renameVariables(INT_VAR_RENAMING& var_renaming);
    bool IsBackJump() const { return _after < this; };
    void SetBackJump(Flatterm* jmp) 
    {
      CALL("SetBackJump(Flatterm* jmp)");
      ASSERT(jmp);
      ASSERT(jmp < this);
      _after = jmp; 
    };
    void BinSwapArgs();  

    float resistanceToUnification(ulong depthCoeff,
				  float constCoeff,
				  float varCoeff) const;

    ulong markInputNonBasicPositions(); 
    // ^ returns the number of marked non-variable positions
    void unmarkAllSubstitutionPositions();
    
    void unmarkSubstitutionPositionsAbove(Flatterm* subterm);

#ifndef NO_DEBUG
    bool CheckIntegrity() const { return (after() > this); };
#endif
    ostream& output(ostream& str) const { return str << _symbol; };
    ostream& outputTerm(ostream& str) const;

  private:
    void ShiftNextTerm(const long& shift) { _after += shift; };
    static void CopyWithShift(const Flatterm* org,Flatterm* dest,const long& shift);
      
  private:
    Flatterm* _after;
    TERM _symbol;
    bool _isSubstitutionPosition;
    bool _isInputNonBasicPosition;
  }; // class Flatterm
}; // namespace VK

namespace std
{
  inline ostream& operator<<(ostream& str,const VK::Flatterm* t) 
  {    
    return t->outputTerm(str);
  };
};

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm::Open"
#endif
#include "debugMacros.hpp"
//=================================================
namespace VK
{
  class Flatterm::Open
  {
  public:
    Open() 
      : open_parentheses("open_parentheses"),
	bracket_counter(),
	depth_counter()
    {};
    Open(Flatterm* mem) 
      : open_parentheses("open_parentheses"),
	bracket_counter()
    { 
      Reset(mem);
    };
    ~Open() {};
    void Reset(Flatterm* mem) 
    {
      CALL("Reset(Flatterm* mem)");
      DOP(debug_fst_pos = mem);
      curr_pos = mem; 
      open_parentheses.reset();
      bracket_counter.Reset();
      depth_counter.Reset(1UL);
    };

    ulong Depth() const { return depth_counter.Depth(); };
    void PushVar(ulong var)
    {
      CALL("PushVar(ulong var)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    }; 
    
    void PushVar(ulong var,bool substPos)
    {
      CALL("PushVar(ulong var,bool substPos)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var,substPos);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    };

    void PushVar(const TERM& var)
    {
      CALL("PushVar(const TERM& var)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    };
    void PushVar(const TERM& var,bool substPos)
    {
      CALL("PushVar(const TERM& var,bool substPos)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var,substPos);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    };

    void PushVar(const TERM& var,ulong varOffset)
    {
      CALL("PushVar(const TERM& var,ulong varOffset)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var.var() + varOffset);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    }; 

    void PushVar(const TERM& var,ulong varOffset,bool substPos)
    {
      CALL("PushVar(const TERM& var,ulong varOffset,bool substPos)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetVar(var.var() + varOffset,substPos);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    }; 

    void PushConst(const TERM& csym)
    {
      CALL("PushConst(const TERM& csym)");     
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetConst(csym);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    }; 
    void PushConst(const TERM& csym,bool substPos)
    {
      CALL("PushConst(const TERM& csym,bool substPos)");     
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetConst(csym,substPos);
      curr_pos++;
      CloseNParentheses(bracket_counter.Close());
    }; 
    void PushNonconstFun(const TERM& fsym)
    {
      CALL("PushNonconstFun(const TERM& fsym)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetSymbol(fsym);
      // fsym.arity() > 0 here
      bracket_counter.Open(fsym.arity());
      open_parentheses.push(curr_pos);
      depth_counter.Down();
      curr_pos++;
    }; 
    void PushNonconstFun(const TERM& fsym,bool substPos)
    {
      CALL("PushNonconstFun(const TERM& fsym,bool substPos)");
      ASSERT(curr_pos - debug_fst_pos < (long)VampireKernelConst::MaxTermSize);
      curr_pos->SetSymbol(fsym,substPos);
      // fsym.arity() > 0 here
      bracket_counter.Open(fsym.arity());
      open_parentheses.push(curr_pos);
      depth_counter.Down();
      curr_pos++;
    }; 
    void Push(const TERM& sym) 
    {
      if (sym.isVariable()) { PushVar(sym); }
      else if (sym.arity()) { PushNonconstFun(sym); }
      else PushConst(sym);
    };
    void Push(const TERM& sym,bool substPos) 
    {
      if (sym.isVariable()) { PushVar(sym,substPos); }
      else if (sym.arity()) { PushNonconstFun(sym,substPos); }
      else PushConst(sym,substPos);
    };
    void Push(const TERM& sym,ulong varOffset)
    {
      if (sym.isVariable()) { PushVar(sym,varOffset); }
      else if (sym.arity()) { PushNonconstFun(sym); }
      else PushConst(sym);
    };
    void Push(const TERM& sym,ulong varOffset,bool substPos)
    {
      if (sym.isVariable()) { PushVar(sym,varOffset,substPos); }
      else if (sym.arity()) { PushNonconstFun(sym,substPos); }
      else PushConst(sym,substPos);
    };
    Flatterm* CurrPos() const { return curr_pos; };
   
    void PushStandardComplexTerm(const TERM* t,ulong varOffset)
    { 
      CALL("PushStandardComplexTerm(const TERM* t,ulong varOffset)");
      ASSERT(t->isComplex());
      TERM::Iterator iter(t);  
      do
	{    
	  Push(iter.symbolRef(),varOffset);
	}
      while (iter.Next()); 
    };

  protected:
    void copyFrom(const Open& orig,long shift)
    {
      CALL("copyFrom(const Open& orig,long shift)");
      curr_pos = orig.curr_pos + shift;
      open_parentheses.reset();
      for (Flatterm* const * p = orig.open_parentheses.begin();
	   p != orig.open_parentheses.end();
	   ++p)
	{
	  open_parentheses.push((*p) + shift);
	};
      bracket_counter = orig.bracket_counter;
      depth_counter = orig.depth_counter;
      DOP(debug_fst_pos = orig.debug_fst_pos + shift);
    };

  private: 
    void CloseNParentheses(unsigned n)
    {
      depth_counter.Up(n);
      while (n)
	{
	  open_parentheses.pop()->SetNextTerm(curr_pos);
	  n--;
	};
    }; 
  private:
    Flatterm* curr_pos;
    BK::Stack<Flatterm*,VampireKernelConst::MaxTermSize> open_parentheses;
    BracketCounter<VampireKernelConst::MaxTermDepth> bracket_counter;
    DepthCounter depth_counter;
    DOP(Flatterm* debug_fst_pos);
  }; // class Flatterm::Open

}; // namespace VK
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm::OpenWithMemory"
#endif
#include "debugMacros.hpp"
//=================================================
namespace VK
{
  class Flatterm::OpenWithMemory : public Flatterm::Open
  {
  public:
    OpenWithMemory() : Open(memory) {};
    OpenWithMemory& operator=(const OpenWithMemory& orig);
    void Reset() { Open::Reset(memory); };
    const Flatterm* Term() const { return memory; }; 
    Flatterm* Term() { return memory; }; 
    long currentSize() const { return CurrPos() - memory; };
  private:
    Flatterm memory[VampireKernelConst::MaxTermSize+2];
  }; // class Flatterm::OpenWithMemory : public Flatterm::Open

}; // namespace VK
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm::Iterator"
#endif
#include "debugMacros.hpp"
//=================================================

namespace VK
{
  class Flatterm::Iterator
  {
  private:
    Flatterm* curr_pos;
    Flatterm* end_of_term;
  public:
    Iterator() {};
    Iterator(Flatterm* t) { Reset(t); };
    void Reset(Flatterm* t)
    {
      curr_pos = t;
      end_of_term = t->after();
    };
    Flatterm* CurrPos() const { return curr_pos; };
    bool EndOfTerm() const { return (curr_pos == end_of_term); };
    bool NotEndOfTerm() const { return (curr_pos < end_of_term); };
    void NextSym() { curr_pos++; };
    void after() { curr_pos = curr_pos->after(); }; 
  }; // class Flatterm::Iterator

}; // namespace VK

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm::ConstIterator"
#endif
#include "debugMacros.hpp"
//=================================================

namespace VK
{

  class Flatterm::ConstIterator
  {
  private:
    const Flatterm* curr_pos;
    const Flatterm* end_of_term;
  public:
    ConstIterator() {};
    ConstIterator(const Flatterm* t) { Reset(t); };
    void Reset(const Flatterm* t)
    {
      curr_pos = t;
      end_of_term = t->after();
    };
    const Flatterm* CurrPos() const { return curr_pos; };
    bool EndOfTerm() const { return (curr_pos == end_of_term); };
    bool NotEndOfTerm() const { return (curr_pos < end_of_term); };
    void NextSym() { curr_pos++; };
    void after() { curr_pos = curr_pos->after(); }; 
  }; // class Flatterm::ConstIterator

}; // namespace VK

//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FLATTERM
#define DEBUG_NAMESPACE "Flatterm"
#endif
#include "debugMacros.hpp"
//=================================================
namespace VK
{

  inline bool Flatterm::equal(const Flatterm* t) const
  {
    CALL("equal(const Flatterm* t) const");
    // can be applied to variables
    if (symbolRef() != t->symbolRef()) return false; 
    register long n = size();
    if (n != t->size()) return false;
    n--;
    while (n)
      {
	if ((this[n].symbolRef()) != (t[n].symbolRef())) return false;   
	n--;  
      };
    return true;
  }; // bool Flatterm::equal(const Flatterm* t) const


  inline bool Flatterm::equal(const Flatterm* t,
			      const SymbolInfo* commutativityTable,
			      bool& commutativityUsed) const
  {
    CALL("equal(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const");

    if (size() != t->size()) return false;

    ConstIterator iter1(this);

    ConstIterator iter2(t);
    const Flatterm* arg11;
    const Flatterm* arg21;
    const Flatterm* arg12;
    const Flatterm* arg22;
      
    commutativityUsed = false; 
      
    do
      {
	if (iter1.CurrPos()->symbolRef() != iter2.CurrPos()->symbolRef()) 
	  return false;

	if ((iter1.CurrPos()->isComplex()) && 
	    commutativityTable->commutativityLaw(iter1.CurrPos()->functor()))
	  {    
	    // recursion     
	    arg11 = iter1.CurrPos() + 1;
	    arg21 = iter2.CurrPos() + 1;
	    bool commUsed;
	    if (arg11->equal(arg21,commutativityTable,commUsed))
	      {      
		COP("equal 1st args");
		commutativityUsed = commutativityUsed || commUsed;
		iter1.NextSym();
		iter1.after();
		iter2.NextSym();
		iter2.after();      
	      }
	    else // try to swap arguments
	      {
		arg12 = arg11->after();
		arg22 = arg21->after();
		if (arg12->equal(arg21,commutativityTable,commUsed)
		    && arg22->equal(arg11,commutativityTable,commUsed))
		  {
		    commutativityUsed = true;
		    iter1.after();         
		    iter2.after();
		  }
		else return false;
	      };
	  }
	else
	  {
	    iter1.NextSym();
	    iter2.NextSym();
	  };
      }
    while (iter1.NotEndOfTerm());
    return true;
  }; // bool Flatterm::equal(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const

  inline 
  void 
  Flatterm::commutativityLawsForEqual(const Flatterm* t,
				      const SymbolInfo* commutativityTable,
				      BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForEqual(..) const");

    ConstIterator iter1(this);
    ConstIterator iter2(t);
    const Flatterm* arg11;
    const Flatterm* arg21;
    const Flatterm* arg12;
    const Flatterm* arg22;      
    const Clause* commutativityLaw;
    do
      {
	ASSERT(iter1.CurrPos()->symbolRef() == iter2.CurrPos()->symbolRef());
	if ((iter1.CurrPos()->isComplex()) && 
	    (commutativityLaw = commutativityTable->commutativityLaw(iter1.CurrPos()->functor())))
	  {    
	    // recursion     
	    arg11 = iter1.CurrPos() + 1;
	    arg21 = iter2.CurrPos() + 1;
	    bool commUsed;
	    if (arg11->equal(arg21,commutativityTable,commUsed))
	      {      
		COP("equal 1st args");
		if (commUsed)
		  arg11->commutativityLawsForEqual(arg21,commutativityTable,laws);
		iter1.NextSym();
		iter1.after();
		iter2.NextSym();
		iter2.after();      
	      }
	    else // swap arguments
	      {
		arg12 = arg11->after();
		arg22 = arg21->after();
		arg12->commutativityLawsForEqual(arg21,commutativityTable,laws);
		arg22->commutativityLawsForEqual(arg11,commutativityTable,laws);
		laws->receive(commutativityLaw);
		iter1.after();         
		iter2.after();
	      };
	  }
	else
	  {
	    iter1.NextSym();
	    iter2.NextSym();
	  };
      }
    while (iter1.NotEndOfTerm());
  }; // void Flatterm::commutativityLawsForEqual(const Flatterm* t,const SymbolInfo* commutativityTable,BK::Receiver<const Clause*>* laws) const

  inline bool Flatterm::EqualArguments(const Flatterm* t) const
  {
    CALL("EqualArguments(const Flatterm* t) const");
    const Flatterm* sym1 = this + 1;
    const Flatterm* sym2 = t + 1;
    while (sym1 != _after)
      {
	if (sym1->Symbol() != sym2->Symbol()) return false;
	sym1++;
	sym2++;
      };
    return true; 
  }; // bool Flatterm::EqualArguments(const Flatterm* t) const

  inline bool Flatterm::EqualArguments(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const
  {
    CALL("EqualArguments(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const");
    commutativityUsed = false;
    bool commUsed;
    ConstIterator iter1(this);
    ConstIterator iter2(t);
    iter1.NextSym();
    iter2.NextSym();
    while (iter1.NotEndOfTerm())
      {
	if (!iter1.CurrPos()->equal(iter2.CurrPos(),commutativityTable,commUsed)) 
	  return false;
	commutativityUsed = commutativityUsed || commUsed;
	iter1.after();
	iter2.after();
      };
    return true;
  }; // bool Flatterm::EqualArguments(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const



  inline
  void 
  Flatterm::commutativityLawsForEqualArguments(const Flatterm* t,
					       const SymbolInfo* commutativityTable,
					       BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForEqualArguments(..) const");
    
    ConstIterator iter1(this);
    ConstIterator iter2(t);
    iter1.NextSym();
    iter2.NextSym();
    while (iter1.NotEndOfTerm())
      {
	iter1.CurrPos()->commutativityLawsForEqual(iter2.CurrPos(),
						   commutativityTable,
						   laws);
	iter1.after();
	iter2.after();
      };
  }; // void Flatterm::commutativityLawsForEqualArguments(const Flatterm* t,const SymbolInfo* commutativityTable,BK::Receiver<const Clause*>* laws) const


  inline bool Flatterm::Complementary(const Flatterm* t) const
  {
    return (Symbol().ComplementaryHeader(t->Symbol()))
      && (EqualArguments(t));
  }; // bool Flatterm::Complementary(const Flatterm* t) const

  inline bool Flatterm::Complementary(const Flatterm* t,
				      const SymbolInfo* commutativityTable,
				      bool& commutativityUsed) const
  {
    return (Symbol().ComplementaryHeader(t->Symbol()))
      && (EqualArguments(t,commutativityTable,commutativityUsed));
  }; // bool Flatterm::Complementary(const Flatterm* t,const SymbolInfo* commutativityTable,bool& commutativityUsed) const

  inline 
  void 
  Flatterm::commutativityLawsForComplementary(const Flatterm* t,
					      const SymbolInfo* commutativityTable,
					      BK::Receiver<const Clause*>* laws) const
  {
    CALL("commutativityLawsForComplementary(..) const");
    ASSERT(Symbol().ComplementaryHeader(t->Symbol()));
    commutativityLawsForEqualArguments(t,commutativityTable,laws);
  }; // void Flatterm::commutativityLawsForComplementary(const Flatterm* t,const SymbolInfo* commutativityTable,BK::Receiver<const Clause*>* laws) const
  

  inline float Flatterm::resistanceToUnification(ulong depthCoeff,
						 float constCoeff,
						 float varCoeff) const
  {
    CALL("resistanceToUnification(ulong depthCoeff,float constCoeff,float varCoeff) const");
    if (isVariable()) return varCoeff;
    if (!arity()) return constCoeff;
    float res = 0.0;
    const Flatterm* end = after();
    for (const Flatterm* arg = next(); arg != end; arg = arg->after())
      res += arg->resistanceToUnification(depthCoeff,constCoeff,varCoeff);
    return 1.0 + (res/depthCoeff);
  }; // float Flatterm::resistanceToUnification(..) const

}; // namespace VK
//=============================================================
#endif
