//
// File:         NegativeEqualitySplitting.hpp
// Description:  Core module for negative equality splitting.
// Created:      Nov 07, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef NEGATIVE_EQUALITY_SPLITTING_H
#define NEGATIVE_EQUALITY_SPLITTING_H
//===================================================================
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
namespace VK
{
  class TemporaryClause;
  class TemporaryClauseCollector;
  class ComponentName;
}; // namespace VK
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_NEGATIVE_EQUALITY_SPLITTING
#  define DEBUG_NAMESPACE "NegativeEqualitySplitting"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class NegativeEqualitySplitting
  {
  public:
    typedef TemporaryLiteral::Extras::Handle LitHandle;
  public:
    NegativeEqualitySplitting(ComponentName* componentNaming);
    ~NegativeEqualitySplitting();
    void reset();
    void setMinimalEligibleArgSize(ulong n)
    {
      _minimalEligibleArgSize = n;
    };
    void setMaximalEligibleVarCount(ulong n)
    {
      _maximalEligibleVarCount = n;
      if (_maximalEligibleVarCount >= VampireKernelConst::MaxNumOfVariables)
	{
	  _maximalEligibleVarCount = VampireKernelConst::MaxNumOfVariables - 1;
	};
    };
    bool
    splitOffAppropriateNegEqArguments(TemporaryClause& cl,
				      TemporaryClauseCollector* newClauseCollector,
				      ulong& numOfSplitOffArgs,
				      ulong& numOfNewNames);
  private:
    NegativeEqualitySplitting() {};
    bool splitOffAppropriateArgument(LitHandle& lit,
				     TemporaryClauseCollector* newClauseCollector,
				     bool& newNameWasIntroduced,
				     Clause*& nameIntroClause);
    static ulong variableOutside(const WeightPolynomial::Monomials& vars);
  private:
    long _minimalEligibleArgSize;
    long _maximalEligibleVarCount;
    ComponentName* _componentNaming;
    Flatterm _newVariable[1];   
    LitHandle _definedLiteral;
    LitHandle _definitionBodyLiteral;
    BasicTempLitList _definitionBody;
  }; // class NegativeEqualitySplitting
}; // namespace VK



//===================================================================

#endif
