//
// File:         StandardLiteralOrdering.hpp
// Description:  Extension of the simplification ordering on terms
//               to literals.  
// Created:      Nov 07, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//==================================================================
#ifndef STANDARD_LITERAL_ORDERING_H
#define STANDARD_LITERAL_ORDERING_H
//==================================================================
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "Comparison.hpp"
#include "SimplificationOrdering.hpp"
#include "TemporaryLiteral.hpp"
#include "Signature.hpp"
#include "LiteralOrdering.hpp"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_STANDARD_LITERAL_ORDERING
 #define DEBUG_NAMESPACE "StandardLiteralOrdering"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{
  class StandardLiteralOrdering : public LiteralOrdering
  {
  public:
    StandardLiteralOrdering();    
    ~StandardLiteralOrdering();
    void setMode(long m) 
    { 
      _mode = m; 
      _isStratified = (_mode != 0L) && (_mode != 3L);
    };
    BK::Comparison compare(TemporaryLiteral& lit1,
			   TemporaryLiteral& lit2) const;
    BK::Comparison compareLayers(const TemporaryLiteral& lit1,
				 const TemporaryLiteral& lit2) const;
  private:

    BK::Comparison compareEq(TemporaryLiteral& lit1,
			 TemporaryLiteral& lit2) const;
    BK::Comparison compareEqAtomsOO(TemporaryLiteral& lit1,
				TemporaryLiteral& lit2) const;
    BK::Comparison compareEqAtomsOU(TemporaryLiteral& lit1,
				TemporaryLiteral& lit2) const;
    BK::Comparison compareEqAtomsUU(TemporaryLiteral& lit1,
				TemporaryLiteral& lit2) const;
    BK::Comparison compareNeqWithEq(TemporaryLiteral& neq,
				    TemporaryLiteral& eq) const;

    
  private:
    long _mode;
  }; // class StandardLiteralOrdering
}; // namespace VK


//=================================================================
#endif
