//
// File:         TemporaryClause.hpp
// Description:  Structure for temporary clauses.
// Created:      Nov 07, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Status:       sketch
//=================================================================
#ifndef TEMPORARY_CLAUSE_H
#define TEMPORARY_CLAUSE_H
//===================================================================
#include <iostream>
#include "jargon.hpp"
#include "DestructionMode.hpp"
#include "GlobAlloc.hpp"
#include "ObjectPool.hpp"
#include "ObjectPoolWithMemory.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "TemporaryLiteral.hpp"
#include "BasicTempLitList.hpp"
#include "TempLitListExtras.hpp"
#include "ClauseBackground.hpp"
//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause"
#endif
#include "debugMacros.hpp"
//===================================================================

namespace VK
{  
  class TemporaryClause
  {
  public:
    typedef BK::ObjectPool<BasicTempLitList>::EntryHandle LitListHandle;
    class Context;
    class Extras;
  public:
    TemporaryClause();
    ~TemporaryClause();
    void hibernate() { _literals.clear(); };
    void reset();
    const LitListHandle& literals() const { return _literals; };
    LitListHandle& literals() { return _literals; };
    const OpenClauseBackground& background() const { return _background; };
    OpenClauseBackground& background() { return _background; };
    bool persistentClauseNumberIsSet() const { return _persistentClauseNumberIsSet; };
    void setPersistenClauseNumber(ulong n) 
    {
      _persistentClauseNumberIsSet = true;
      _persistentClauseNumber = n;
    };

    void resetSimplificationFlags();
    
    bool inputNonBasicPositionsMarked() const { return _inputNonBasicPositionsMarked; };
    void markInputNonBasicPositions();
    bool weeded() const { return _weeded; };
    bool weededModuloCommutativity() const 
    { 
      return _weededModuloCommutativity; 
    };
    bool testedForSubsumption() const { return _testedForSubsumption; };
    bool simplifiedBySubsumptionResolution() const { return _simplifiedBySubsumptionResolution; };
    bool normalisedByCommutativity() const { return _normalisedByCommutativity; };
    bool testedForReductionOfNonBasicByCommutativity() const
    {
      return _testedForReductionOfNonBasicByCommutativity;
    };
    bool normalisedByDemodulation() const { return _normalisedByDemodulation; };
    bool testedForReductionOfNonBasic() const { return _testedForReductionOfNonBasic; };
    bool simplifiedByBuiltInTheories() const { return _simplifiedByBuiltInTheories; };
    bool simplifiedByDestructiveEqualityResolution() const { return _simplifiedByDestructiveEqualityResolution; };
    bool lockedForPropositionalSplitting() const { return _lockedForPropositionalSplitting; };
    bool lockedForNegativeEqualitySplitting() const { return _lockedForNegativeEqualitySplitting; };

    

    void markAsWeeded() { _weeded = true; };
    void markAsWeededModuloCommutativity() 
    {
      _weeded = true;
      _weededModuloCommutativity = true; 
    };
    void markAsTestedForSubsumption() { _testedForSubsumption = true; };
    void markAsSimplifiedBySubsumptionResolution() { _simplifiedBySubsumptionResolution = true; };
    void markAsNormalisedByCommutativity() { _normalisedByCommutativity = true; };
    void markAsTestedForReductionOfNonBasicByCommutativity()
    {
      _testedForReductionOfNonBasicByCommutativity = true;
    };
    void markAsNormalisedByDemodulation() {  _normalisedByDemodulation = true; };
    void markAsTestedForReductionOfNonBasic() { _testedForReductionOfNonBasic = true; };
    void markAsSimplifiedByBuiltInTheories() { _simplifiedByBuiltInTheories = true; };
    void markAsSimplifiedByDestructiveEqualityResolution() { _simplifiedByDestructiveEqualityResolution = true; };
    void markAsLockedForPropositionalSplitting() { _lockedForPropositionalSplitting = true; };
    void markAsLockedForNegativeEqualitySplitting() { _lockedForNegativeEqualitySplitting = true; };

    void registerIntraLiteralModification()
    {
      CALL("registerIntraLiteralModification()");   
      _inputNonBasicPositionsMarked = false;
      _weeded = false;
      _weededModuloCommutativity = false;
      _testedForSubsumption = false;
      _simplifiedBySubsumptionResolution = false;
      _normalisedByCommutativity = false;
      _normalisedByDemodulation = false;
      _simplifiedByBuiltInTheories = false;
      _simplifiedByDestructiveEqualityResolution = false;
      _lockedForPropositionalSplitting = false;
      _lockedForNegativeEqualitySplitting = false;
    };

    bool isAnswer() const
    {
      CALL("isAnswer() const");
      return 
	TempLitListExtras::containsOnlyAnswerLiterals(*_literals);
    };
    
    bool isNonSubsumable() const
    {
      CALL("isNonSubsumable() const");
      return _background.isNonSubsumable();
    };

    ostream& outputAnnounced(ostream& str) const;
    ostream& outputComplete(ostream& str) const;
    
  private:
    TemporaryClause& operator=(const TemporaryClause&) 
    { 
      return *this;
    };

  private:
    ulong _universalId;
    LitListHandle _literals;
    OpenClauseBackground _background;
    bool _persistentClauseNumberIsSet;
    ulong _persistentClauseNumber;
    
    bool _inputNonBasicPositionsMarked;
    bool _weeded;
    bool _weededModuloCommutativity;
    bool _testedForSubsumption;
    bool _simplifiedBySubsumptionResolution;
    bool _normalisedByCommutativity;
    bool _testedForReductionOfNonBasicByCommutativity;
    bool _normalisedByDemodulation;
    bool _testedForReductionOfNonBasic;
    bool _simplifiedByBuiltInTheories;
    bool _simplifiedByDestructiveEqualityResolution;
    bool _lockedForPropositionalSplitting;
    bool _lockedForNegativeEqualitySplitting;
    
  }; // class TemporaryClause
}; // namespace VK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "<global>"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace BK
{
  // specialisation of the hibernation template for TemporaryClause
  template <>
  inline
  static
  void hibernate(VK::TemporaryClause& obj)
  {
    CALL("hibernate(VK::TemporaryClause& obj)");
    obj.hibernate();
  };

}; // namespace BK



//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  inline
  void TemporaryClause::resetSimplificationFlags()
  {
    CALL("resetSimplificationFlags()");
    _inputNonBasicPositionsMarked = false;
    _weeded = false;
    _weededModuloCommutativity = false;
    _testedForSubsumption = false;
    _simplifiedBySubsumptionResolution = false;
    _normalisedByCommutativity = false;
    _testedForReductionOfNonBasicByCommutativity = false;
    _normalisedByDemodulation = false;
    _testedForReductionOfNonBasic = false;
    _simplifiedByBuiltInTheories = false;
    _simplifiedByDestructiveEqualityResolution = false;
    _lockedForPropositionalSplitting = false;
    _lockedForNegativeEqualitySplitting = false;
  }; // bool TemporaryClause::resetSimplificationFlags()

  
  inline
  void TemporaryClause::markInputNonBasicPositions()
  {
    CALL("markInputNonBasicPositions()");
    if (!_inputNonBasicPositionsMarked)
      {
	for (BasicTempLitList::Element* el = _literals->first();
	     el;
	     el = el->next())
	  el->literal()->markInputNonBasicPositions();
	_inputNonBasicPositionsMarked = true;
      };
  }; // void TemporaryClause::markInputNonBasicPositions()

}; // namespace BK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause::Context"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  class TemporaryClause::Context
  {
  public:
    class AutoActivator
    {
    public:
      AutoActivator() : _previous(Context::getCurrent())
      {
	Context::setCurrent(static_cast<Context*>(0));
      };
      AutoActivator(Context* c) : _previous(Context::getCurrent())
      {
	Context::setCurrent(c);
      };
      ~AutoActivator()
      {
	Context::setCurrent(_previous);
      };
    private:
      Context* _previous;
    }; // class AutoActivator

  public:
    Context() : _previousUniversalId(0UL) {};
    ~Context() {};
    static Context* getCurrent() { return _current; };
    static void setCurrent(Context* c) { _current = c; };
    ulong generateUniversalId() 
    {
      ++_previousUniversalId;
      return _previousUniversalId;
    };
    BK::ObjectPool<TemporaryClause>* defaultPool()
    {
      return static_cast<BK::ObjectPool<TemporaryClause>*>(&_defaultPool);
    };
  private:
    ulong _previousUniversalId;
    BK::ObjectPoolWithMemory<BK::GlobAlloc,TemporaryClause> _defaultPool;
    static Context* _current;
  }; // class TemporaryClause::Context
    

}; // namespace VK

//===================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_TEMPORARY_CLAUSE
#  define DEBUG_NAMESPACE "TemporaryClause::Extras"
#endif
#include "debugMacros.hpp"
//===================================================================


namespace VK
{
  class TemporaryClause::Extras
  {    
  public:
    typedef BK::ObjectPool<TemporaryClause> Pool;
    typedef BK::ObjectPool<TemporaryClause>::EntryHandle Handle;
  public:
    static void freshClause(Handle& h,Pool* pool)
    {
      CALL("freshClause(Handle& h,Pool* pool)");
      ASSERT(pool);
      pool->reserveObject(h);
      h->reset();
    };
    static void freshClause(Handle& h)
    {
      CALL("freshClause(Handle& h)");
      ASSERT(Context::getCurrent());
      freshClause(h,Context::getCurrent()->defaultPool());
    };
  }; // class TemporaryClause::Extras
}; // namespace VK


//===================================================================

#endif
