//
// File:         VampireKernel.cpp
// Description:  Implementation of Vampire kernel. 
// Created:      Jun 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 03, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Had to mess the code up to please VC++ 6.0
//======================================================================
#include <climits>
#include <cstdio>
#include <cmath>
#include "GlobAlloc.hpp"
#include "GlobalClock.hpp"
#include "RuntimeError.hpp"
#include "GlobalStopFlag.hpp"
#include "Receiver.hpp"
#include "VampireKernel.hpp"
#include "Signature.hpp"
#include "Clause.hpp"
#include "ClauseQueue.hpp"
#include "ProblemProfile.hpp"
#include "Tabulation.hpp"
#include "Master.hpp"
#include "ClauseNumIndex.hpp"
#include "TermWeighting.hpp"
#include "UniformTermWeighting.hpp"
#include "NonuniformConstantTermWeighting.hpp"
#include "NonrecursiveKBOrdering.hpp"
#include "StandardKBOrdering.hpp"
#include "InferenceTree.hpp"
#include "WeightPolynomial.hpp"
#include "BuiltInTheoryDispatcher.hpp"
#include "clause_set.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
#include "sharing.hpp"
#include "LiteralSharing.hpp"
#include "ClauseAssembler.hpp"
#include "LiteralProperties.inl"
#include "ComponentName.hpp"
using namespace BK;
using namespace VK;
//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#  define DEBUG_NAMESPACE "VampireKernel::_Implementation"
#endif    
#include "debugMacros.hpp"
//======================================================================

class VampireKernel::_Implementation
{
public:

  //
  //   VC++ 6.0 (debug mode) breaks down if I try to define a constructor
  //   or destructor for _Implementation

  /***********

	      _Implementation(const VampireKernel::Options& opt);
	      ~_Implementation();

  ******************/

  void init(const VampireKernel::Options& opt);
  void destroy();

  void openSession()
  {
    _currentSession = this;
  };

  static void closeSession()
  {
    _currentSession = 0;
  };
  
 
  void* operator new(size_t)
  {
    return GlobAlloc::allocate(sizeof(_Implementation));
  };
   
  void operator delete(void* obj)
  {
    GlobAlloc::deallocate(obj,sizeof(_Implementation));
  };

  static _Implementation* allocate()
  { 
    return 
      static_cast<_Implementation*>(GlobAlloc::allocate(sizeof(_Implementation)));
  };

  static void deallocate(_Implementation* obj)
  {
    GlobAlloc::deallocate(static_cast<void*>(obj),sizeof(_Implementation));
  };

  void beginInputSignature();

  void endInputSignature();
  void registerInputFunction(const char* name,
			     ulong arity,
			     ulong& id);
  void registerInputPredicate(const char* name,
			      ulong arity,
			      bool isAnswer,
			      ulong& id);
  void setSymbolWeightConstPart(ulong symId,
				const VampireKernel::TermWeightType& weight);
  void setSymbolPrecedence(ulong symId,long prec);
  void setHeaderPrecedence(bool positive,ulong symId,long prec); 
  void interpret(bool predicate,
		 const char* inputSymbol,
		 ulong arity,
		 const char* theory,
		 const char* nativeSymbol);
  void beginInputClauses();
  void endInputClauses();
  void beginInputClause(ulong clauseNumber,void* origin,const ClauseAnnotation& annotation);
  void endInputClause();

  void beginInputLiteralDefinition(ulong clauseNumber,
                                   void* origin,
				   bool goal,
				   bool support);
  
  void endInputLiteralDefinition();

  void beginInputLiteralDefinitionHead(bool positive,ulong predicateId);
  void endInputLiteralDefinitionHead();

  void beginInputLiteralDefinitionTail();
  void endInputLiteralDefinitionTail();

  void beginInputLiteralDefinitionTailComponent();
  void endInputLiteralDefinitionTailComponent();

  void beginLiteral(bool positive,
		    ulong predicateId,
		    const VampireKernel::LiteralAnnotation& annotation);
  void endLiteral();
  void receiveFunction(ulong symId);
  void receiveNumericConstant(const double& value);
  void receiveVariable(ulong varNum);
  void saturate(ulong nextVacantClauseId,long timeLimit);
  void preprocessAndRunMainLoop();
  void lookForAnotherAnswer();
  void suspend();
  void resume(); 
  void reportInterrupt(int signalNumber);
  void reportErrorSignal(int signalNumber);
  void reportError();
  VampireKernel::TerminationReason terminationReason() const;
  const VampireKernel::Clause* refutation() const;
  VampireKernel::ClauseSetIterator* activeClauses();
  void disposeOf(VampireKernel::ClauseSetIterator* iterator);
  bool isNumericConstantId(ulong symbolId) const;
  bool isInputSymbol(ulong symbolId) const; 
  bool isAnswerPredicateId(ulong symbolId) const;
  ulong symbolArity(ulong symbolId) const;
  const char* symbolPrintName(ulong symbolId) const; 
  const double& numericConstantValue(ulong symbolId) const; 

private:
  class ClauseSetIterator : public VampireKernel::ClauseSetIterator
  {
  public:
    ClauseSetIterator(const CLAUSE_SET& set) :
      _iter(set)
    {
    };
    ~ClauseSetIterator();
    void* operator new(size_t)
    {
      return GlobAlloc::allocate(sizeof(ClauseSetIterator));
    };
    void operator delete(void* obj)
    {
      GlobAlloc::deallocate(obj,sizeof(ClauseSetIterator));
    };
    bool isEnd() const;
    void next();
    const VampireKernel::Clause* clause() const;
    
  private:
    CLAUSE_SET::ConstIter _iter;
  }; // class ClauseSetIterator


private:
  static void handleAlarm(int);
  static void globalStopFlagHandler();
  void preprocessInputQueryClauses();
  void preprocessQueryBySimplification(ulong iterationNumber);
  void preprocessQueryByDefElimination(ulong iterationNumber);
  void prepareMasterForSimplification();
  void prepareMasterForDefElimination();
  void loadInputQueryClausesIntoMaster();
  void loadInputSupportClausesIntoMaster();
  void extractPreprocessedQueryClauses();
  void prepareMasterForSaturation(long timeLimit, // deciseconds
				  float memoryLimit); // bytes
  void reportTermination(); 
  void reportTimeLimitExpired();
  static void setHardMemoryLimit(float lim); // bytes
  static void setAllocationBufferSize(long percentage);
  static bool releaseMemory(ulong size);
  static bool releaseALotOfMemory();
  static void handleAllocationFailure(ulong size); // bytes
  void reportAllocationFailure(ulong size); // bytes
  static void outputClause(const VK::Clause* cl);
  static void outputWeightChange(long oldWL,long newWL);


private:

  enum Continuation
  {
    Continuation_Init,
    Continuation_Prepro_Simplification,
    Continuation_Prepro_DefElimination_LoadInput,
    Continuation_Prepro_DefElimination_DefElimination,
    Continuation_LoadInputQuery,
    Continuation_LoadInputSupport,
    Continuation_MainLoop
  };

  class TermWeightingWrapper
  {
  public:
    TermWeightingWrapper(const VampireKernel::Options::TermWeightingScheme& sc)
    {
      CALL("TermWeightingWrapper(const VampireKernel::Options::TermWeightingScheme& sc)");      
      switch (sc)
	{
	case VampireKernel::Options::UniformTermWeighting:
	  _uniformTermWeighting = new UniformTermWeighting();
	  TermWeighting::setCurrent(static_cast<TermWeighting*>(_uniformTermWeighting));
	  _nonuniformConstantTermWeighting = 0;
	  break;
	case VampireKernel::Options::NonuniformConstantTermWeighting:
	  _nonuniformConstantTermWeighting = new NonuniformConstantTermWeighting();
	  TermWeighting::setCurrent(static_cast<TermWeighting*>(_nonuniformConstantTermWeighting));
	  _uniformTermWeighting = 0;
	  break;
	};
    };
    ~TermWeightingWrapper()
    {
      if (DestructionMode::isThorough())
	{
	  if (_uniformTermWeighting) delete _uniformTermWeighting;
	  if (_nonuniformConstantTermWeighting) delete _nonuniformConstantTermWeighting;
	};
    };
  

  private:
    UniformTermWeighting* _uniformTermWeighting;
    NonuniformConstantTermWeighting* _nonuniformConstantTermWeighting;
  }; // class TermWeightingWrapper

  class SimplificationOrderingWrapper
  {
  public:
    SimplificationOrderingWrapper(const VampireKernel::Options::SimplificationOrdering& so)
    {
      switch (so)
	{ 
	case VampireKernel::Options::NonrecursiveKBOrdering:
	  _nonrecursiveKBOrdering = new NonrecursiveKBOrdering();
	  SimplificationOrdering::setCurrent(static_cast<SimplificationOrdering*>(_nonrecursiveKBOrdering));
	  _standardKBOrdering = 0;
	  break;
	case VampireKernel::Options::StandardKBOrdering:
	  _standardKBOrdering = new StandardKBOrdering();
	  SimplificationOrdering::setCurrent(static_cast<SimplificationOrdering*>(_standardKBOrdering));
	  _nonrecursiveKBOrdering = 0;
	  break;	
	};
    };
    ~SimplificationOrderingWrapper() 
    { 
      if (DestructionMode::isThorough())
	{
	  if (_nonrecursiveKBOrdering) delete _nonrecursiveKBOrdering;
	  if (_standardKBOrdering) delete _standardKBOrdering;
	};
    };
  private:
    NonrecursiveKBOrdering* _nonrecursiveKBOrdering;
    StandardKBOrdering* _standardKBOrdering;
  }; // class SimplificationOrderingWrapper


  class NameIntroReceiver : public BK::Receiver<VK::Clause*>
  {
  public:
    NameIntroReceiver(_Implementation* host) : _hostKernelSession(host) {};
    void receive(VK::Clause* const & cl);
  private:
    _Implementation* _hostKernelSession;
  }; // class NameIntroReceiver : public BK::Receiver<VK::Clause*>


private:
  void outputInputSupportClauses();
  void outputInputQueryClauses();
  const VampireKernel::Statistics& statistics();
  void updateExportedStatistics();
  void outputPreproStat(ostream* str,bool full);
  void outputCurrentLoopStat(ostream* str,bool full);
  void outputCumulativeStat(ostream* str,bool full);
  void autoStatOutput(ostream* str);

  void tabulateStatistics();
  const char* flagValue(bool flag);
  void tabulateOptions();
  void tabulate();
  void switchBuiltinEqualityOff();
  void switchChainingForTransitiveOff();
  void analysePreprocessedInput();
  void analyseOptions();
  void collectPreprocessingIterationStatistics();
  static bool isRelevantBackgroundFlag(const VK::ClauseBackground::Rule& fl);

  static 
  VampireKernel::Clause::BackgroundFlag
  convertBackgroundFlag(const VK::ClauseBackground::Rule& fl);

  static 
  VK::ClauseBackground::Rule
  convertBackgroundFlag(const VampireKernel::Clause::BackgroundFlag& fl);



private:  
  static VampireKernel::_Implementation* _currentSession;

  WeightPolynomial::Sharing _weightPolynomialSharing;

  TermWeightingWrapper _termWeightingWrapper;
  SimplificationOrderingWrapper _simplificationOrderingWrapper;

  VampireKernel* _wrapper;
  VampireKernel::Options _options;
  GlobalClock::Timer _timer;
  long _saturationTimeLimit;
  bool _saturationRunning;
  bool _mainLoopRunning;
  Signature _signature;

  VK::Statistics _inputStatistics;

  ulong _currentInputClauseNumber;
  void* _currentInputClauseOrigin;

  ClauseQueue _inputSupportClauses;
  ClauseQueue _inputQueryClauses;

  ulong _numberOfSubmittedInputClauses;
  bool _inputClauseRejected;
  ulong _numberOfRejectedInputClauses;
  bool _incompleteProcedure;
  ProblemProfile _problemProfile;
  Tabulation* _tabulation;
  VK::Clause::NewNumber _clauseNumberGenerator;

  BuiltInTheoryDispatcher* _builtInTheoryDispatcher;
  

  VK::Clause::Context _clauseContext;
  VK::Clause::Context::AutoActivator _clauseContextActivator;
  
  TemporaryLiteral::Context _temporaryLiteralContext;
  TemporaryLiteral::Context::AutoActivator _temporaryLiteralContextActivator;
  
  TempLitListExtras::Context _temporaryLitListContext;
  TempLitListExtras::Context::AutoActivator _temporaryLitListContextActivator;
  
  TemporaryClause::Context _temporaryClauseContext;
  TemporaryClause::Context::AutoActivator _temporaryClauseContextActivator;
  
  BasicTempLitList::Context _basicTempLitListContext;
  BasicTempLitList::Context::AutoActivator _basicTempLitListContextActivator;
  

  ComponentName _componentNamingService;
  SHARING_INDEX _staticSharingIndex;
  LiteralSharing _staticLiteralSharing;
  ClauseAssembler _staticClauseAssembler;

  Master _master;
  VK::TemporaryClauseCollector _inputClauseCollector;
  ClauseQueue _preprocessingHistory;
  CLAUSE_SET _nameIntroClauses;
  NameIntroReceiver _nameIntroReceiver;
  ClauseNumIndex _numberToClauseMapping;
 
  bool _solvedByPreprocessing;
  VampireKernel::TerminationReason _terminationReason;
  InferenceTree _proofTree;  
  bool _symbolWeightConstantPartInputIgnored;
  bool _unaryFunctionWithZeroWeightExists;
  ulong _unaryFunctionWithZeroWeight;
  Continuation _continuation;
  ulong _preproIterationNumber;
  long _preproEliminationCycles;

  float _preproLastTimeSec;
  long _preproLastOccupiedByObjectsBytes;
  long _preproLastOccupiedByChunksBytes;
  long _preproLastTotalBufferDeficitBytes;
  long _preproLastActiveBufferDeficitBytes;
  long _preproLastSuspendedBufferDeficitBytes;
  VK::Statistics _preproStatistics;
  VK::Statistics _cumulativeStatistics;
  VampireKernel::Statistics _exportedCumulativeStatistics;

  bool _lookingForMoreAnswers;
  long _numberOfSupportClauses;
  bool _equalityInSupport;
  bool _clientEnforcedLiteralSelection;
  bool _clientRestrictedEquality;
  VK::Clause* _currentInputQueryClause;
  VK::Clause* _currentInputSupportClause;
  friend class VampireKernel;
  friend class VampireKernel::Clause;
  friend class VampireKernel::Clause::BackgroundFlagIterator;
  friend class NameIntroReceiver;
}; // class VampireKernel::_Implementation

VampireKernel::_Implementation* 
VampireKernel::_Implementation::_currentSession;

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::_Implementation"
#endif    
#include "debugMacros.hpp"
//======================================================================


/***********
VampireKernel::_Implementation::_Implementation(const VampireKernel::Options& opt) :    
  _termWeightingWrapper(opt.term_weighting_scheme),
  _simplificationOrderingWrapper(opt.simplification_ordering),
  _options(opt),
  _saturationRunning(false),
  _mainLoopRunning(false),
  _incompleteProcedure(false),
  _tabulation(0),
  _builtInTheoryDispatcher(0),
  _clauseContextActivator(&_clauseContext),
  _temporaryLiteralContextActivator(&_temporaryLiteralContext),
  _temporaryLitListContextActivator(&_temporaryLitListContext),
  _temporaryClauseContextActivator(&_temporaryClauseContext),
  _basicTempLitListContextActivator(&_basicTempLitListContext),
  _componentNamingService(&_staticClauseAssembler,&_nameIntroReceiver),
  _staticLiteralSharing(&_staticSharingIndex),
  _staticClauseAssembler(&_staticLiteralSharing,static_cast<LiteralSelection*>(0)),
  _master(&_componentNamingService),
  _nameIntroClauses(VK::Clause::MainSetNameIntro,"_nameIntroClauses"),
  _nameIntroReceiver(this),
  _terminationReason(VampireKernel::UnknownTerminationReason),
  _continuation(Continuation_Init),
  _lookingForMoreAnswers(false),
  _numberOfSupportClauses(0L),
  _clientEnforcedLiteralSelection(false),
  _clientRestrictedEquality(false)
{
  CALL("constructor _Implementation(const VampireKernel::Options& opt)");
    
  Signature::setCurrent(&_signature);
  _signature.setMaxlNumOfSkolemSymbols(_options.max_skolem);
  if (_options.tabulation)
    {

      _tabulation = new Tabulation();
      _tabulation->assignStream(_options.tabulation);
      _tabulation->assignJob(_options.job_id);
      static ulong sessionId = 0UL;
      _tabulation->assignKernelSession(sessionId);
      ++sessionId;
    }
  else
    _tabulation = 0;
    
  _master.setClauseNumberGenerator(&_clauseNumberGenerator);
    
  WeightPolynomial::setSharing(&_weightPolynomialSharing);
    
  _inputClauseCollector.setShowAnnouncedClauses(0);
  _inputClauseCollector.setShowCompleteClauses(0);
  _inputClauseCollector.setPreprocessingInProgress(false);
  _inputClauseCollector.setStatistics(&_inputStatistics);
  _inputClauseCollector.setTautologyCheck(false);
  _inputClauseCollector.setLiteralFactoring(false);
}; // _Implementation(const VampireKernel::Options& opt)
  

VampireKernel::_Implementation::~_Implementation()
{
  CALL("destructor ~_Implementation()");
  if (DestructionMode::isThorough())
    {
      VK::Clause* clause;
      while (_inputSupportClauses.dequeue(clause))
        {
	  ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	  _staticClauseAssembler.removeFromSharing(clause);
	  clause->Destroy();
	};
      while (_inputQueryClauses.dequeue(clause))
	{
	  ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	  _staticClauseAssembler.removeFromSharing(clause);
	  clause->Destroy();
	};
      while (_preprocessingHistory.dequeue(clause))
	{
	  ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	  _staticClauseAssembler.removeFromSharing(clause);
	  clause->Destroy();
	};
      
      while (_nameIntroClauses.Nonempty())
	{
	  clause = _nameIntroClauses.Pop();
	  _staticClauseAssembler.removeFromSharing(clause);
	  clause->Destroy();
	};
 
      if (_tabulation) delete _tabulation;
      if (_builtInTheoryDispatcher) delete _builtInTheoryDispatcher;
    };
}; // ~_Implementation()

  ******************/

void 
VampireKernel::_Implementation::init(const VampireKernel::Options& opt) 
{
  CALL("init(const VampireKernel::Options& opt)");
 
  BK::init(_weightPolynomialSharing);

  BK::init(_termWeightingWrapper,opt.term_weighting_scheme);
  BK::init(_simplificationOrderingWrapper,opt.simplification_ordering);

  _options = opt;

  BK::init(_timer);

  _saturationRunning = false;
  _mainLoopRunning = false;

  BK::init(_signature);
  BK::init(_inputStatistics);

  BK::init(_inputSupportClauses);
  BK::init(_inputQueryClauses);

  _incompleteProcedure = false;
    
  BK::init(_problemProfile);

  _tabulation = 0;

  BK::init(_clauseNumberGenerator);

  _builtInTheoryDispatcher = 0;

  BK::init(_clauseContext);
  BK::init(_clauseContextActivator,&_clauseContext);

  BK::init(_temporaryLiteralContext);
  BK::init(_temporaryLiteralContextActivator,&_temporaryLiteralContext); 

  BK::init(_temporaryLitListContext);
  BK::init(_temporaryLitListContextActivator,&_temporaryLitListContext); 
 
  BK::init(_temporaryClauseContext);
  BK::init(_temporaryClauseContextActivator,&_temporaryClauseContext); 

  BK::init(_basicTempLitListContext);
  BK::init(_basicTempLitListContextActivator,&_basicTempLitListContext);

  BK::init(_componentNamingService,&_staticClauseAssembler,&_nameIntroReceiver);
  BK::init(_staticSharingIndex);
  BK::init(_staticLiteralSharing,&_staticSharingIndex);
  BK::init(_staticClauseAssembler,&_staticLiteralSharing,static_cast<LiteralSelection*>(0));
  _master.init(&_componentNamingService);
  BK::init(_inputClauseCollector);
  BK::init(_preprocessingHistory);
  BK::init(_nameIntroClauses,VK::Clause::MainSetNameIntro,"_nameIntroClauses");
  BK::init(_nameIntroReceiver,this);
  BK::init(_numberToClauseMapping);
    
  _terminationReason = VampireKernel::UnknownTerminationReason;

  BK::init(_proofTree);

  _continuation = Continuation_Init;

  BK::init(_preproStatistics);
  BK::init(_cumulativeStatistics);
  BK::init(_exportedCumulativeStatistics);

  _lookingForMoreAnswers = false;
  _numberOfSupportClauses = 0L;
  _clientEnforcedLiteralSelection = false;
  _clientRestrictedEquality = false;

  Signature::setCurrent(&_signature);
  _signature.setMaxlNumOfSkolemSymbols(_options.max_skolem);
  if (_options.tabulation)
    {
      _tabulation = new Tabulation();
      _tabulation->assignStream(_options.tabulation);
      _tabulation->assignJob(_options.job_id);
      static ulong sessionId = 0UL;
      _tabulation->assignKernelSession(sessionId);
      ++sessionId;
    }
  else
    _tabulation = 0;
    
  _master.setClauseNumberGenerator(&_clauseNumberGenerator);
    
  WeightPolynomial::setSharing(&_weightPolynomialSharing);

  _inputClauseCollector.setShowAnnouncedClauses(0);
  _inputClauseCollector.setShowCompleteClauses(0);
  _inputClauseCollector.setPreprocessingInProgress(false);
  _inputClauseCollector.setStatistics(&_inputStatistics);
  _inputClauseCollector.setTautologyCheck(false);
  _inputClauseCollector.setLiteralFactoring(false);
}; // void VampireKernel::_Implementation::init(const VampireKernel::Options& opt)



void  
VampireKernel::_Implementation::destroy() 
{ 
  CALL("destroy()");
   
  if (DestructionMode::isThorough())
    {
      VK::Clause* clause;
	while (_inputSupportClauses.dequeue(clause))
	  {
	    ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	    _staticClauseAssembler.removeFromSharing(clause);
	    clause->Destroy();
           };
	while (_inputQueryClauses.dequeue(clause))
	  {
	    ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	    _staticClauseAssembler.removeFromSharing(clause);
	    clause->Destroy();
           };
	while (_preprocessingHistory.dequeue(clause))
	  {
	    ASSERT(clause->flag(VK::Clause::FlagLiteralsShared));
	    _staticClauseAssembler.removeFromSharing(clause);
	    clause->Destroy();
           };
	while (_nameIntroClauses.Nonempty())
	  {
	    clause = _nameIntroClauses.Pop();
	    _staticClauseAssembler.removeFromSharing(clause);
	    clause->Destroy();
	  };
      if (_tabulation) delete _tabulation;
      if (_builtInTheoryDispatcher) delete _builtInTheoryDispatcher;
    };
  
  BK::destroy(_exportedCumulativeStatistics);
  BK::destroy(_cumulativeStatistics);
  BK::destroy(_preproStatistics);
  BK::destroy(_proofTree);
  BK::destroy(_numberToClauseMapping);
  BK::destroy(_preprocessingHistory);
  BK::destroy(_nameIntroReceiver);
  BK::destroy(_nameIntroClauses);
  BK::destroy(_inputClauseCollector);
  _master.destroy();
  BK::destroy(_staticClauseAssembler);
  BK::destroy(_staticLiteralSharing);
  BK::destroy(_staticSharingIndex);
  BK::destroy(_componentNamingService);

  BK::destroy(_basicTempLitListContextActivator);
  BK::destroy(_basicTempLitListContext);

  BK::destroy(_temporaryClauseContextActivator); 
  BK::destroy(_temporaryClauseContext);

  BK::destroy(_temporaryLitListContextActivator); 
  BK::destroy(_temporaryLitListContext);

  BK::destroy(_temporaryLiteralContextActivator); 
  BK::destroy(_temporaryLiteralContext);

  BK::destroy(_clauseContextActivator);
  BK::destroy(_clauseContext);

  BK::destroy(_clauseNumberGenerator);
  BK::destroy(_problemProfile);
  BK::destroy(_inputQueryClauses);
  BK::destroy(_inputSupportClauses);
  BK::destroy(_inputStatistics);
  BK::destroy(_signature);
  BK::destroy(_timer);	
  BK::destroy(_simplificationOrderingWrapper);
  BK::destroy(_termWeightingWrapper);
  BK::destroy(_weightPolynomialSharing);
}; // void VampireKernel::_Implementation::destroy() 





void  
VampireKernel::_Implementation::beginInputSignature() 
{
  _symbolWeightConstantPartInputIgnored = false;
  _unaryFunctionWithZeroWeightExists = false;
};

void  
VampireKernel::_Implementation::endInputSignature() 
{
  if (_symbolWeightConstantPartInputIgnored)
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: input of some symbol weight constant parts ignored\n"
			     << "%          because of --term_weighting_scheme " 
			     << _options.term_weighting_scheme << "\n";
    };

  if (_options.selection < 0L)
    {
      _signature.adjustHeaderPrecedenceForPositiveSelection();
    };
  if (_options.header_precedence_kinky)
    {
      _signature.makeHeaderPrecedenceKinky();
    };   
  _signature.setDependenceOfPrecedenceOnArity(_options.symbol_precedence_by_arity);
  _signature.setDependenceOfWeightOnArity(_options.symbol_weight_by_arity);
  _signature.endOfInput();
};
  
void  
VampireKernel::_Implementation::registerInputFunction(const char* name,
						      ulong arity,
						      ulong& id)
{
  id = _signature.inputSymbolNumber(false,name,arity);
};


inline
void  
VampireKernel::_Implementation::registerInputPredicate(const char* name,
						       ulong arity,
						       bool isAnswer,
						       ulong& id)
{
  id = _signature.inputSymbolNumber(true,name,arity);
  if (isAnswer) _signature.makeAnswerPredicate(id);
};
  
inline
void  
VampireKernel::_Implementation::setSymbolWeightConstPart(ulong symId,
							 const VampireKernel::TermWeightType& weight)
{
  if (_options.term_weighting_scheme == 
      VampireKernel::Options::UniformTermWeighting) 
    {
      _symbolWeightConstantPartInputIgnored = true;
      return;
    };
    
  VampireKernel::TermWeightType w = weight;
  if (w == (VampireKernel::TermWeightType)0)
    {
      ulong arity = _signature.arity(symId);
      if (arity == 0UL)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: zero weight of the constant/propositional variable ";
	      _signature.outputSymbol(*_options.std_output,symId) << " is reset to 1.\n";
	    };
	  w = (VampireKernel::TermWeightType)1;
	}
      else
	if ((arity == 1UL) &&
	    (!_signature.isPredicate(symId)))
	      
	  {
	    if (_unaryFunctionWithZeroWeightExists)
	      {
		if (_options.std_output)
		  {
		    *_options.std_output << "% WARNING: zero weight of the unary function ";
		    _signature.outputSymbol(*_options.std_output,symId) 
		      << " is reset to 1.\n"
		      << "           Zero weight was previously assigned to ";
		    _signature.outputSymbol(*_options.std_output,_unaryFunctionWithZeroWeight) << "\n";
		  };
		w = (VampireKernel::TermWeightType)1;
	      }
	    else
	      {
		_unaryFunctionWithZeroWeightExists = true;
		_unaryFunctionWithZeroWeight = symId;
		_signature.assignMaximalPossibleSymbolPrecedence(symId);
	      };	      
	  };
    };
  _signature.setSymbolWeightConstantPart(symId,w);
};

inline
void  
VampireKernel::_Implementation::setSymbolPrecedence(ulong symId,long prec)
{
  CALL("setSymbolPrecedence(ulong symId,long prec)");
  ulong previousSym;
  if (prec == 0L)
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of ";
	  _signature.outputSymbol(*_options.std_output,symId) 
	    << " ignored: bad value 0\n";
	};
      return;
    };
    
  if (_signature.symbolPrecedenceIsReserved(prec,previousSym))
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of ";
	  _signature.outputSymbol(*_options.std_output,symId) 
	    << " ignored: precedence value has been reserved for ";
	  _signature.outputSymbol(*_options.std_output,previousSym) << "\n";
	};
      return;
    };
  if (prec >= Signature::maximalPossibleSymbolPrecedence())
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of ";
	  _signature.outputSymbol(*_options.std_output,symId) << " ignored: too high.\n";
	};
      return;
    };
  if (prec <= Signature::minimalPossibleSymbolPrecedence())
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of ";
	  _signature.outputSymbol(*_options.std_output,symId) << " ignored: too low.\n";
	};
      return;
    };
  if ((_unaryFunctionWithZeroWeightExists) && 
      (symId == _unaryFunctionWithZeroWeight))
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of ";
	  _signature.outputSymbol(*_options.std_output,symId) 
	    << " ignored: maximal precedence assigned since the function has zero weight.\n";
	};
      return;
    };
    
  _signature.setInputSymbolPrecedence(symId,prec);
};


 
inline
void  
VampireKernel::_Implementation::setHeaderPrecedence(bool positive,ulong symId,long prec)
{
  CALL("setHeaderPrecedence(bool positive,ulong symId,long prec)");
  ulong previousSym;
  Signature::Polarity previousPol;
  if (prec == -1L ||
      prec == 0L ||
      prec == 1L)
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of header ";
	  if (positive) 
	    {
	      *_options.std_output << "++";
	    }
	  else
	    *_options.std_output << "--";
	  _signature.outputSymbol(*_options.std_output,symId) 
	    << " ignored: bad value " << prec << "\n";
	};
      return;
    };

      

  if (_signature.headerPrecedenceIsReserved(prec,previousPol,previousSym))
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of header ";
	  if (positive) 
	    {
	      *_options.std_output << "++";
	    }
	  else
	    *_options.std_output << "--";
	  _signature.outputSymbol(*_options.std_output,symId) 
	    << " ignored: precedence value has been reserved for ";
	  if (previousPol == Signature::PositivePolarity)
	    {
	      *_options.std_output << "++";
	    }
	  else
	    *_options.std_output << "--";
	  _signature.outputSymbol(*_options.std_output,previousSym) << "\n";
	};
      return;
    };
  if (prec >= Signature::maximalPossibleHeaderPrecedence())
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of header ";
	  if (positive) 
	    {
	      *_options.std_output << "++";
	    }
	  else
	    *_options.std_output << "--";
	  _signature.outputSymbol(*_options.std_output,symId) << " ignored: too high.\n";
	};
      return;
    };
  if (prec <= Signature::minimalPossibleHeaderPrecedence())
    {
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: precedence of header ";
	  if (positive) 
	    {
	      *_options.std_output << "++";
	    }
	  else
	    *_options.std_output << "--";
	  _signature.outputSymbol(*_options.std_output,symId) << " ignored: too low.\n";
	};
      return;
    };
  
  Signature::Polarity polarity = 
    (positive)? Signature::PositivePolarity : Signature::NegativePolarity;
    
  _signature.setInputHeaderPrecedence(polarity,symId,prec);

}; // void VampireKernel::_Implementation::setHeaderPrecedence(..)


inline
void  
VampireKernel::_Implementation::interpret(bool predicate,
					  const char* inputSymbol,
					  ulong arity,
					  const char* theory,
					  const char* nativeSymbol)
{
  if (!_builtInTheoryDispatcher)
    {
      _builtInTheoryDispatcher = new BuiltInTheoryDispatcher(&_clauseNumberGenerator);
      BuiltInTheoryDispatcher::setCurrent(_builtInTheoryDispatcher);
    };
  ulong symId = _signature.inputSymbolNumber(predicate,inputSymbol,arity);
  if (!_builtInTheoryDispatcher->interpret(symId,theory,nativeSymbol))
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: can not interpret " << inputSymbol << "/" << arity 
			     << " as " << theory << "/" << nativeSymbol << "\n";
    };
};

void  
VampireKernel::_Implementation::beginInputClauses()
{
  CALL("beginInputClauses()");
  ASSERT(_inputSupportClauses.empty());
  ASSERT(_inputQueryClauses.empty());
  _numberOfSubmittedInputClauses = 0UL;
  _numberOfRejectedInputClauses = 0UL;
  _numberOfSupportClauses = 0L;
  _clientEnforcedLiteralSelection = false;
  _clientRestrictedEquality = 
    (_options.all_pos_eq_mode == LiteralAnnotation::ResolutionOnly);
  _equalityInSupport = false;
};

void  
VampireKernel::_Implementation::endInputClauses()
{
  CALL("endInputClauses()");

  if (_numberOfSupportClauses && _options.std_output)
    {
      *_options.std_output << "% WARNING: " << _numberOfSupportClauses 
			   << " support clauses in the input.\n"
			   << "%          Preprocessing by definition elimination is blocked.\n"
			   << "%          Incompleteness is not assumed!\n";
      _options.elim_def = 0;
    };
  if (_clientEnforcedLiteralSelection)
    {	
      *_options.std_output << "% WARNING: some input clauses have preselected literals.\n" 
			   << "%          Incompleteness is not assumed!\n";
    };
  if (_clientRestrictedEquality)
    {	
      *_options.std_output << "% WARNING: some (possibly all) equality literals have restricted semantics.\n"
			   << "%          Incompleteness is not assumed!\n";
    };
};
  
void  
VampireKernel::_Implementation::beginInputClause(ulong clauseNumber,
						 void* origin,
						 const ClauseAnnotation& annotation)
{
  CALL("beginInputClause(ulong clauseNumber,void* origin,const ClauseAnnotation& annotation)");
  ++_numberOfSubmittedInputClauses;
  _inputClauseRejected = false;
  
  _inputClauseCollector.openNewClause(ClauseBackground::Input);
  _inputClauseCollector.announceNewClause();
  if (annotation.isGoal()) _inputClauseCollector.markAsSubgoal();
  if (annotation.isSupport()) 
    {
      _inputClauseCollector.markAsSupport();
      ++_numberOfSupportClauses;        
    };
  if (annotation.isNonSubsuming()) _inputClauseCollector.markAsNonSubsuming();
  if (annotation.isNonSubsumable()) _inputClauseCollector.markAsNonSubsumable();
  
  _currentInputClauseNumber = clauseNumber;
  _currentInputClauseOrigin = origin;
};

void  
VampireKernel::_Implementation::endInputClause()
{
  CALL("endInputClause()");
  ObjectPool<TemporaryClause>::EntryHandle collectedClause;
  VK::Clause* clause;
  _inputClauseRejected = 
    _inputClauseRejected ||
    (!_inputClauseCollector.finaliseClause(collectedClause)) ||
    (!((clause = _staticClauseAssembler.assembleSharedWithoutSelection(*collectedClause))));
  if (_inputClauseRejected)
    {
      ++_numberOfRejectedInputClauses;
    }
  else
    {
      clause->SetNumber(_currentInputClauseNumber);
      clause->setInputClauseOrigin(_currentInputClauseOrigin);
      
      if (clause->isSupportClause())
	{
	  if (_options.chaining_for_transitive &&
	      clause->isTransitivityLaw())
	    {
	      
	      // becomes a query clause
	      clause->markAsQuery();
	      clause->setMainSet(VK::Clause::MainSetInputQuery);
	      _inputQueryClauses.enqueue(clause);
	    }
	  else
	    {
	      if ((!_equalityInSupport) &&
		  clause->containsEquality())
		_equalityInSupport = true;
	      
	      clause->setMainSet(VK::Clause::MainSetInputSupport);
	      _inputSupportClauses.enqueue(clause);
	    };
	}
      else
	{
	  clause->setMainSet(VK::Clause::MainSetInputQuery);
	  _inputQueryClauses.enqueue(clause);
	};
      
      if (clause->Empty())
	{
	  if (_options.std_output)
	    *_options.std_output << "% WARNING: empty input clause.\n";
	};
    };
}; // void VampireKernel::_Implementation::endInputClause()
  


void 
VampireKernel::_Implementation::beginInputLiteralDefinition(ulong clauseNumber,
							    void* origin,
							    bool goal,
							    bool support)
{
  CALL("beginInputLiteralDefinition(..)");
  // temporary
}; // void VampireKernel::_Implementation::beginInputLiteralDefinition(..)
  
void VampireKernel::_Implementation::endInputLiteralDefinition()
{
  CALL("endInputLiteralDefinition()");
  // temporary
}; // void VampireKernel::_Implementation::endInputLiteralDefinition()

void 
VampireKernel::_Implementation::beginInputLiteralDefinitionHead(bool positive,
								ulong predicateId)
{
  CALL("beginInputLiteralDefinitionHead(..)");
  // temporary
}; // void VampireKernel::_Implementation::beginInputLiteralDefinitionHead(..)

void VampireKernel::_Implementation::endInputLiteralDefinitionHead()
{
  CALL("endInputLiteralDefinitionHead()");
  // temporary
}; // void VampireKernel::_Implementation::endInputLiteralDefinitionHead()

void VampireKernel::_Implementation::beginInputLiteralDefinitionTail()
{
  CALL("beginInputLiteralDefinitionTail()");
  // temporary
}; // void VampireKernel::_Implementation::beginInputLiteralDefinitionTail()

void VampireKernel::_Implementation::endInputLiteralDefinitionTail()
{
  CALL("endInputLiteralDefinitionTail()");
  // temporary
}; // void VampireKernel::_Implementation::endInputLiteralDefinitionTail()

void VampireKernel::_Implementation::beginInputLiteralDefinitionTailComponent()
{
  CALL("beginInputLiteralDefinitionTailComponent()");
  // temporary
}; // void VampireKernel::_Implementation::beginInputLiteralDefinitionTailComponent()

void VampireKernel::_Implementation::endInputLiteralDefinitionTailComponent()
{
  CALL("endInputLiteralDefinitionTailComponent()");
  // temporary
}; // void VampireKernel::_Implementation::endInputLiteralDefinitionTailComponent()



void VampireKernel::_Implementation::beginLiteral(bool positive,
						  ulong predicateId,
						  const VampireKernel::LiteralAnnotation& annotation)
{
  CALL("beginLiteral(..,const VampireKernel::LiteralAnnotation& annotation)");

  if (_inputClauseRejected) return;
  VK::LiteralProperties prop(annotation);
  TERM::Polarity polarity = 
    (positive)? Signature::PositivePolarity : Signature::NegativePolarity;
  TERM header((TERM::Arity)_signature.arity(predicateId),
	      (TERM::Functor)predicateId,
	      polarity);

  if (_inputClauseCollector.openNewLiteral(header,prop))
    {
      if (annotation.deploymentMode() != LiteralAnnotation::DefaultDeploymentMode ||
	  annotation.selectabilityCategory() != LiteralAnnotation::DefaultSelectability)
	_clientEnforcedLiteralSelection = true;
      if (annotation.equalityMode() == LiteralAnnotation::ResolutionOnly)
	_clientRestrictedEquality = true;
    }
  else
    _inputClauseRejected = true;
    

}; // void VampireKernel::_Implementation::beginLiteral(..,const VampireKernel::LiteralAnnotation& annotation)

void  
VampireKernel::_Implementation::endLiteral()
{
  if (_inputClauseRejected) return;
  if (!_inputClauseCollector.closeLiteral())
    _inputClauseRejected = true;
};
  
inline
void  
VampireKernel::_Implementation::receiveFunction(ulong symId)
{
  CALL("receiveFunction(ulong symId)");
  if (_inputClauseRejected) return;
  TERM symbol((TERM::Arity)_signature.arity(symId),
	      (TERM::Functor)symId,
	      Signature::PositivePolarity);
  if (!_inputClauseCollector.pushFunc(symbol))
    _inputClauseRejected = true;
}; // void receiveFunction(ulong symId)

inline
void  
VampireKernel::_Implementation::receiveNumericConstant(const double& value)
{
  CALL("receiveNumericConstant(const double& value)");
  ulong id = _signature.numericConstantNumber(value);
  receiveFunction(id);
}; // void receiveNumericConstant(const double& value


inline
void  
VampireKernel::_Implementation::receiveVariable(ulong varNum)
{
  CALL("receiveVariable(ulong varNum)");
  if (_inputClauseRejected) return;
  if (!_inputClauseCollector.pushVar(varNum))
    _inputClauseRejected = true;
};

void  
VampireKernel::_Implementation::saturate(ulong nextVacantClauseId,long timeLimit)
{
  CALL("saturate(ulong nextVacantClauseId,long timeLimit)");

  GlobalStopFlag::reset();
  _saturationRunning = true;
  _saturationTimeLimit = timeLimit;
  if (_saturationTimeLimit < 0L) _saturationTimeLimit = 0L;
  _mainLoopRunning = false;
  _timer.reset();
  _timer.start();
  GlobalStopFlag::HandlerActivator 
    globalStopFlagHandlerActivator(globalStopFlagHandler);
  GlobalClock::AlarmHandler alarmHandler(handleAlarm);

  _clauseNumberGenerator.reset(nextVacantClauseId);
  if (_options.std_output)
    *_options.std_output << "% New Vampire Kernel session, time limit: " 
			 << _saturationTimeLimit/10 << "." 
			 << _saturationTimeLimit%10 << " sec.\n";
  if (_numberOfRejectedInputClauses)
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: " << _numberOfRejectedInputClauses 
			     << " input clauses rejected.\n";
      _incompleteProcedure = true;
    };
    
  if (_options.show_opt) 
    {	
      if (_options.std_output)
	{
	  *_options.std_output << "%================== Options: =======================\n";
	  *_options.std_output<< "% time limit = " << ((float)_saturationTimeLimit)/10 << "\n";
	  _options.output(*_options.std_output);
	  *_options.std_output << "%========== End of options: ======================\n";
	};
    };

  _clauseContext.setSelectedPosEqPenaltyCoefficient(_options.selected_pos_eq_penalty_coefficient);
  
  _clauseContext.setNongoalPenaltyCoefficient(_options.nongoal_penalty_coefficient);
  _clauseContext.setShowLiteralProperties(_options.show_lit_prop);

  if (_options.show_input && _options.std_output)
    {
      *_options.std_output << "%================== Input clauses (support + query): ===================\n";
      outputInputSupportClauses();
      outputInputQueryClauses();
      *_options.std_output << "%============== End of input clauses: ================\n";
    };
  if (_options.show_profile && _options.std_output)
    {
      _problemProfile.load(_inputSupportClauses);
      _problemProfile.load(_inputQueryClauses);
      _problemProfile.summarise();
      *_options.std_output << "%========= Profile of input (support + query):\n"; 
      _problemProfile.output(*_options.std_output);
      *_options.std_output << "%========= End of profile. ========\n";
      *_options.std_output << _signature;
    };


  GlobalStopFlag::check();
  preprocessAndRunMainLoop();
}; // void VampireKernel::_Implementation::saturate(ulong nextVacantClauseId,long timeLimit)

void  
VampireKernel::_Implementation::preprocessAndRunMainLoop()
{
  CALL("preprocessAndRunMainLoop()");    
  long remainingTime;
  float remainingMemory;

  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_Prepro_Simplification:
      // same as below
    case Continuation_Prepro_DefElimination_LoadInput:
      // same as below
    case Continuation_Prepro_DefElimination_DefElimination:
      goto preprocess;
    case Continuation_LoadInputQuery:
      if (_options.std_output)
	*_options.std_output << "% Resuming loading preprocessed clauses...\n";
      goto load_input_query;
    case Continuation_LoadInputSupport:
      if (_options.std_output)
	*_options.std_output << "% Resuming loading preprocessed clauses...\n";
      goto load_input_support;
    case Continuation_MainLoop:
      if (_options.std_output)
	*_options.std_output << "% Resuming main loop...\n";
      goto main_loop;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::preprocessAndRunMainLoop()");
    };

  
  _preproLastTimeSec = _timer.elapsedSeconds();
  _preproLastOccupiedByObjectsBytes = 0L;
  _preproLastOccupiedByChunksBytes = 0L;
  _preproLastTotalBufferDeficitBytes = 0L;
  _preproLastActiveBufferDeficitBytes = 0L;
  _preproLastSuspendedBufferDeficitBytes = 0L;
  _preproStatistics.reset();

 preprocess:
  preprocessInputQueryClauses();

  if (_master.terminated())
    {
      _solvedByPreprocessing = true;
      reportTermination();
      _timer.stop();
      _saturationRunning = false; 
      return;
    };
  _continuation = Continuation_Init;

  _solvedByPreprocessing = false;

    
  _signature.setCommutativity(_problemProfile.commutativityTable(),
			      _problemProfile.symmetryTable());
  _signature.makeCommutative(VampireKernelConst::UnordEqNum);
  _signature.makeCommutative(VampireKernelConst::OrdEqNum);

#ifdef INTERCEPT_FORWARD_MATCHING_INDEXING
  FMILog << _signature;
#endif
#ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
  BSILog.unblock();
  BSILog << "MODE_DESCRIPTION\n";
  if (options.bs_set_mode)
    {
      BSILog << "  'set mode'\n";
    }
  else
    {
      BSILog << "  'multiset mode'\n";
    };
  BSILog << "END_OF_MODE_DESCRIPTION\n";
  BSILog.block();
#endif 

#ifdef INTERCEPT_BACKWARD_MATCHING_INDEXING
  BMILog.unblock();
  // Mode modifiers:
  BMILog << "$\n"; // end of mode description
  BMILog << _signature;
  INTERCEPT_BACKWARD_MATCHING_INDEXING_COMMENT(BMILog << "           // end of signature declaration");
  BMILog.block();
#endif


#ifdef INTERCEPT_INDEXED_SEARCH_FOR_GENERALISATIONS
  SFGILog.unblock();
  // Mode modifiers:
  SFGILog << "$\n"; // end of mode description
  SFGILog << _signature;
  INTERCEPT_INDEXED_SEARCH_FOR_GENERALISATIONS_COMMENT(SFGILog << "           // end of signature declaration");
  SFGILog.block();
#endif


#ifdef INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING
  BSIULog.unblock();
  // Mode modifiers:
  BSIULog << "$\n"; // end of mode description
  BSIULog << _signature;
  INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING_COMMENT(BSIULog << "           // end of signature declaration");
  BSIULog.block();
#endif


  if ((!_problemProfile.numOfEqLiterals()) &&
      (!_equalityInSupport))
    {
      switchBuiltinEqualityOff();
    };
    
  if (!_problemProfile.numOfTransitive())
    {
      switchChainingForTransitiveOff();
    };
    

  GlobalStopFlag::check();

  remainingTime = _saturationTimeLimit - _timer.elapsedDeciseconds();
  remainingMemory = 
    (((float)_options.memory_limit)*1024) - GlobAlloc::occupiedByObjects();
  if (remainingMemory < 0.0) remainingMemory = 0.0;

  analysePreprocessedInput();
  
  if (_options.input_non_basicness && 
      _options.show_profile && 
      _options.std_output)
    {
      if (_signature.nonBasicArgumentExists())
	{
	  *_options.std_output << "% Functions/headers with input-non-basic arguments:\n";
	  _signature.outputNonBasicArguments(*_options.std_output);
	}
      else
	{
	  *_options.std_output << "% No non-basic arguments in the preprocessed input.\n";
	};
    };
  
  analyseOptions();

  prepareMasterForSaturation(remainingTime,remainingMemory);

 load_input_support:
  loadInputSupportClausesIntoMaster();

  if (_master.terminated())
    {
      reportTermination();
      _timer.stop();
      _saturationRunning = false; 
      _continuation = Continuation_LoadInputSupport;
      return;
    };
  _continuation = Continuation_Init;

 load_input_query:

  loadInputQueryClausesIntoMaster();
   
  if (_master.terminated())
    {
      reportTermination();
      _timer.stop();
      _saturationRunning = false; 
      _continuation = Continuation_LoadInputQuery;
      return;
    };
  _continuation = Continuation_Init;

  GlobalStopFlag::check();

 main_loop:
  _mainLoopRunning = true;
  _master.mainLoop();
  _mainLoopRunning = false;
  reportTermination();
  _timer.stop();
  _saturationRunning = false; 
  _continuation = Continuation_MainLoop;
}; // void VampireKernel::_Implementation::preprocessAndRunMainLoop()

void  
VampireKernel::_Implementation::lookForAnotherAnswer()
{
  CALL("lookForAnotherAnswer()");
  ASSERT(refutation());
  ASSERT(!_timer.isTicking());
  ASSERT(!_saturationRunning);
  GlobalStopFlag::reset();
  _lookingForMoreAnswers = true;
  _timer.start();
  _saturationRunning = true;
  GlobalStopFlag::HandlerActivator 
    globalStopFlagHandlerActivator(globalStopFlagHandler);
  GlobalClock::AlarmHandler alarmHandler(handleAlarm);

  switch (_continuation)
    {
    case Continuation_Prepro_Simplification:
      // same as below
    case Continuation_Prepro_DefElimination_LoadInput:
      // same as below
    case Continuation_Prepro_DefElimination_DefElimination:
      // same as below
    case Continuation_LoadInputQuery:
      // same as below	
    case Continuation_LoadInputSupport:
      // same as below	
    case Continuation_MainLoop:
      if (_options.std_output)
	*_options.std_output << "% Resuming saturation in attempt to find another answer...\n";
      preprocessAndRunMainLoop();
      break;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::lookForAnotherAnswer()");
    };
}; // void VampireKernel::_Implementation::lookForAnotherAnswer()


void  
VampireKernel::_Implementation::suspend()
{
  if (_saturationRunning)
    {
      _master.suspend();
      _timer.stop();
    };
        
  if (_options.std_output)
    {
      *_options.std_output << "% Kernel session suspended.\n";   
      autoStatOutput(_options.std_output);
    };
};

void  
VampireKernel::_Implementation::resume()
{
  if (_options.std_output)
    *_options.std_output << "% Kernel session resumed.\n";
  if (_saturationRunning)
    {
      _timer.start();
      _master.resume();
    };
}; 


void  
VampireKernel::_Implementation::reportInterrupt(int signalNumber)
{
  CALL("reportInterrupt(int signalNumber)");
  _terminationReason = VampireKernel::AbortedOnExternalInterrupt;
    
  if (_options.std_output)
    {
      autoStatOutput(_options.std_output);
      *_options.std_output << "% Proof not found. Aborted on external signal " 
			   << signalNumber << ".\n";
      _options.std_output->flush();
    };
}; // void VampireKernel::_Implementation::reportInterrupt(int signalNumber)

void  
VampireKernel::_Implementation::reportErrorSignal(int signalNumber)
{
  CALL("reportErrorSignal(int signalNumber)");
  _terminationReason = VampireKernel::AbortedOnErrorSignal;
  if (_options.std_output)
    *_options.std_output << "% Proof not found. Kernel session aborted on error signal " 
			 << signalNumber << ".\n";
}; // void VampireKernel::_Implementation::reportErrorSignal(int signalNumber)

  
void  
VampireKernel::_Implementation::reportError()
{
  CALL("reportError()");
  _terminationReason = VampireKernel::AbortedOnError;
    
  if (_options.std_output)
    *_options.std_output << "% Proof not found. Kernel session aborted due to an error.\n";
}; // void VampireKernel::_Implementation::reportError(const char* errorQualifier,const char* errorMessage)

VampireKernel::TerminationReason  
VampireKernel::_Implementation::terminationReason() const
{
  return _terminationReason;
};

const VampireKernel::Clause*  
VampireKernel::_Implementation::refutation() const
{
  return static_cast<const VampireKernel::Clause*>(static_cast<const void*>(_master.proof()));
};

VampireKernel::ClauseSetIterator*  
VampireKernel::_Implementation::activeClauses()
{
  CALL("activeClauses()");
  return 
    static_cast<VampireKernel::ClauseSetIterator*>(new ClauseSetIterator(_master.activeClauses()));
};

void  
VampireKernel::_Implementation::disposeOf(VampireKernel::ClauseSetIterator* iter)
{
  CALL("disposeOf(VampireKernel::ClauseSetIterator* iter)");
  delete static_cast<ClauseSetIterator*>(iter);
};


bool  
VampireKernel::_Implementation::isNumericConstantId(ulong symbolId) const
{
  return _signature.isNumericConstant(symbolId);
};
bool  
VampireKernel::_Implementation::isInputSymbol(ulong symbolId) const
{
  return !_signature.isSkolem(symbolId);
}; 
bool  
VampireKernel::_Implementation::isAnswerPredicateId(ulong symbolId) const
{
  return _signature.isAnswerPredicate(symbolId);
};
ulong  
VampireKernel::_Implementation::symbolArity(ulong symbolId) const
{
  return _signature.arity(symbolId);
};
const char*  
VampireKernel::_Implementation::symbolPrintName(ulong symbolId) const
{
  CALL("symbolPrintName(ulong symbolId) const");
  ASSERT(!isNumericConstantId(symbolId));
  return _signature.symbolName(symbolId);
}; 
const double&  
VampireKernel::_Implementation::numericConstantValue(ulong symbolId) const
{
  CALL("numericConstantValue(ulong symbolId) const");
  ASSERT(isNumericConstantId(symbolId));
  return _signature.numericConstantValue(symbolId);
}; 




void  
VampireKernel::_Implementation::handleAlarm(int)
{
  if (_currentSession && _currentSession->_saturationRunning)
    {	
      if (_currentSession->_timer.elapsedDeciseconds() > _currentSession->_saturationTimeLimit)
	{
	  GlobalStopFlag::raise();
	}
      else
	{
	  if (_currentSession->_mainLoopRunning)
	    {
	      _currentSession->_master.externalTick();
	    };
	};
    };
};

void  
VampireKernel::_Implementation::globalStopFlagHandler()
{
  CALL("globalStopFlagHandler()");
  if (_currentSession) 
    { 
      _currentSession->reportTimeLimitExpired();
      DestructionMode::makeFast();
      throw VampireKernel::Emergency(VampireKernel::Emergency::TimeLimit);
    };
}; // void VampireKernel::_Implementation::globalStopFlagHandler()

void  
VampireKernel::_Implementation::preprocessInputQueryClauses()
{
  CALL("preprocessInputQueryClauses()"); 

  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_Prepro_Simplification:
      if (_options.std_output)
	*_options.std_output << "% Resuming preprocessing by saturation "
			     << _preproIterationNumber << "\n";
      goto simplify;
    case Continuation_Prepro_DefElimination_LoadInput:
      // same as below
    case Continuation_Prepro_DefElimination_DefElimination:
      if (_options.std_output)
	*_options.std_output << "% Resuming preprocessing by def. elimination "
			     << _preproIterationNumber << "\n";
      goto eliminate_definitions;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::preprocessInputQueryClauses()");
    };



  _preproIterationNumber = 0UL;
  _preproEliminationCycles = 0L;

 simplify:
  preprocessQueryBySimplification(_preproIterationNumber);
  collectPreprocessingIterationStatistics();
  _problemProfile.reset();


  _problemProfile.load(_inputQueryClauses);
  _problemProfile.summarise();
  if (_options.show_profile && _options.std_output)
    {
      *_options.std_output << "%========= Profile of preprocessed query (" 
			   << _preproIterationNumber << "):\n"; 
      _problemProfile.output(*_options.std_output);
      *_options.std_output << "%========= End of profile for preprocessed query (" 
			   << _preproIterationNumber << ") ========\n";
    }; 
  if (_master.terminated()) 
    {
      _continuation = Continuation_Prepro_Simplification;
      return;
    };
  _continuation = Continuation_Init;

  if (_problemProfile.numOfEliminated() && 
      (_preproEliminationCycles < _options.elim_def))
    { 
    next_elimination_cycle:
      _preproIterationNumber++;
      _preproEliminationCycles++;

    eliminate_definitions:
      preprocessQueryByDefElimination(_preproIterationNumber);
      collectPreprocessingIterationStatistics();
      _problemProfile.reset(); 

      _problemProfile.load(_inputQueryClauses);
      _problemProfile.summarise();   
      if (_options.show_profile && _options.std_output)
	{
	  *_options.std_output << "%========= Profile of preprocessed query (" 
			       << _preproIterationNumber << "):\n"; 
	  _problemProfile.output(*_options.std_output);
	  *_options.std_output << "%========= End of profile for preprocessed query (" 
			       << _preproIterationNumber << ") ========\n";
	}; 
      if (_master.terminated()) 
	{
	  return;
	};
      
      _continuation = Continuation_Init;
      if (_problemProfile.numOfEliminated() && 
	  (_preproEliminationCycles < _options.elim_def))
	goto next_elimination_cycle;
      goto simplify;
    };
  // end of story
  if (_options.std_output)
    *_options.std_output << "%======== Preprocessing has been done in " << _preproIterationNumber + 1 << " stages.\n";
    
}; // void VampireKernel::_Implementation::preprocessInputQueryClauses()

void  
VampireKernel::_Implementation::preprocessQueryBySimplification(ulong iterationNumber)
{
  CALL("preprocessQueryBySimplification(ulong iterationNumber)");    
  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_Prepro_Simplification:
      goto load_input;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::preprocessQueryBySimplification(ulong iterationNumber)");
    };
    
  if (_options.std_output)
    *_options.std_output << "%=============== Preprocessing by simplification " 
			 << iterationNumber << " ============\n";
  prepareMasterForSimplification();
    
 load_input:
  loadInputQueryClausesIntoMaster();
  if (_master.terminated()) return;
  _continuation = Continuation_Init;
  extractPreprocessedQueryClauses();
  if (_options.show_prepro && _options.std_output) 
    {
	
      *_options.std_output << "%=============== Preprocessing  by simplification " 
			   << iterationNumber << " statistics: ============\n";
      if (_options.auto_stat_full) 
	outputCurrentLoopStat(_options.std_output,true);
      if (_options.auto_stat_summary) 
	outputCurrentLoopStat(_options.std_output,false);
      if ((!_options.auto_stat_full) && (!_options.auto_stat_summary))
	{
	  *_options.std_output << "%==== Automatic statistics output is suppressed.\n";
	};
      *_options.std_output << "%=============== Query clauses preprocessed by simplification " 
			   << iterationNumber << " ================\n";
      outputInputQueryClauses();
      *_options.std_output << "%============== End of clauses preprocessed by " 
			   << iterationNumber << "  ==========\n";
    };
}; // void VampireKernel::_Implementation::preprocessQueryBySimplification(ulong iterationNumber)

void  
VampireKernel::_Implementation::preprocessQueryByDefElimination(ulong iterationNumber)
{
  CALL("preprocessQueryByDefElimination(ulong iterationNumber)");    

  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_Prepro_DefElimination_LoadInput:
      goto load_input;
    case Continuation_Prepro_DefElimination_DefElimination:
      ASSERT_IN(_inputQueryClauses.empty(),"CONT WITH ELIM");
      goto eliminate_definitions;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::preprocessQueryByDefElimination(ulong iterationNumber)");
    };


  if (_options.std_output) 
    *_options.std_output << "%=============== Preprocessing by def. elimination " 
			 << iterationNumber << " ============\n";
  prepareMasterForDefElimination();

 load_input:

  loadInputQueryClausesIntoMaster();
  if (_master.terminated()) 
    {
      _continuation = Continuation_Prepro_DefElimination_LoadInput;
      return;
    };  
  _continuation = Continuation_Init;

 eliminate_definitions:
  ASSERT_IN(_inputQueryClauses.empty(),"ELIMINATE");
  
  _master.eliminateDefinitions();
  if (_master.terminated())
    {
      ASSERT_IN(_inputQueryClauses.empty(),"PROVED BY DEF ELIM");
      _continuation = Continuation_Prepro_DefElimination_DefElimination;

      return;
    };
  _continuation = Continuation_Init;
  extractPreprocessedQueryClauses(); 

  if (_options.show_prepro && _options.std_output) 
    {
      *_options.std_output << "%=============== Preprocessing  by def. elimination " 
			   << iterationNumber << " statistics: ============\n";
      if (_options.auto_stat_full) 
	outputCurrentLoopStat(_options.std_output,true);
      if (_options.auto_stat_summary) 
	outputCurrentLoopStat(_options.std_output,false);
      if ((!_options.auto_stat_full) && (!_options.auto_stat_summary))
	{
	  *_options.std_output << "%==== Automatic statistics output is suppressed.\n";
	};
      *_options.std_output << "%=============== Query clauses preprocessed by def. elimination " 
			   << iterationNumber << " ================\n";
      outputInputQueryClauses();
      *_options.std_output << "%============== End of clauses preprocessed by " 
			   << iterationNumber << " ==========\n";
    };

}; // void VampireKernel::_Implementation::preprocessQueryByDefElimination(ulong iterationNumber)

void  
VampireKernel::_Implementation::prepareMasterForSimplification()
{
  CALL("prepareMasterForSimplification()");
  _master.reset();
  _master.options.main_alg = 0;
  _master.options.binary_resolution = false; 
  _master.options.no_inferences_between_rules = false;
  _master.options.splitting = _options.static_splitting;
  _master.options.splitting_style = _options.splitting_style;
  _master.options.small_basic_comp_in_hyp_splitting = _options.small_basic_comp_in_hyp_splitting;
  _master.options.neg_eq_splitting = _options.static_neg_eq_splitting;  
  _master.options.neg_eq_spl_var_count_limit = _options.stat_neg_eq_spl_var_count_limit;  
  _master.options.paramodulation = false;
  _master.options.paramodulate_from_variables = _options.paramodulate_from_variables;
  _master.options.sim_back_sup = 0;
  _master.options.forw_sup_no_dup_red = false;
  _master.options.back_sup_no_dup_red = false;
  _master.options.prohibit_non_basic_superposition = false;
  _master.options.prohibit_non_basic_fd = false;
  _master.options.prohibit_non_basic_bd = false;
  _master.options.reduce_non_basic = false;
  _master.options.input_non_basicness = false;
  _master.options.run_time_non_basicness = false;

  _master.options.selection = _options.selection;
  _master.options.literal_comparison_mode = _options.literal_comparison_mode;
  _master.options.inherited_negative_selection = _options.inherited_negative_selection;

  // strategy switches        
  _master.options.handle_equality = true;
  _master.options.forward_subsumption = true;
  _master.options.fs_set_mode = true; // not to worry about completeness here 
  _master.options.fs_use_sig_filters = false;
  _master.options.simplify_by_fs = _options.simplify_by_fs;
  _master.options.main_premise_subsumption = 0;
  _master.options.side_premise_subsumption = 0;
  _master.options.main_selected_lit_demodulation = 0;
  _master.options.side_selected_lit_demodulation = 0;
  _master.options.forward_demodulation = 0; 
  _master.options.fd_on_splitting_branches = false;
  if (_options.normalize_by_commutativity)
    {
      _master.options.normalize_by_commutativity = _problemProfile.commutativityTable();
    }
  else 
    _master.options.normalize_by_commutativity = 0;
  if (_options.eq_taut_mod_C)
    {
      _master.options.eq_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.eq_taut_mod_C = 0;
  if (_options.prop_taut_mod_C)
    {
      _master.options.prop_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.prop_taut_mod_C = 0;
  if (_options.trivial_eq_res_mod_C)
    {
      _master.options.trivial_eq_res_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.trivial_eq_res_mod_C = 0;
  if (_options.identical_lit_factoring_mod_C)
    {
      _master.options.identical_lit_factoring_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.identical_lit_factoring_mod_C = 0;

  _master.options.chaining_for_transitive = 0;
  _master.options.chaining_through_variables = _options.chaining_through_variables;
    
  _master.options.backward_demodulation = 0;
  _master.options.bd_on_splitting_branches = false;
  _master.options.simplify_by_eq_res = _options.simplify_by_eq_res;
  _master.options.backward_subsumption = true;
  _master.options.bs_set_mode = true; // not to worry about completeness here 
  _master.options.orphan_murder = 0;
  _master.options.lrs = 0;
  _master.options.pred_elim_table = 0;

  _master.options.orient_pos_equalities_in_support = true;
  _master.options.orient_neg_equalities_in_support = true;

  switch (_options.all_pos_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      _master.options.all_pos_eq_mode = LiteralProperties::UseEqForResolution;
      break;
    case LiteralAnnotation::ResolutionOnly:
      _master.options.all_pos_eq_mode = LiteralProperties::UseEqForResolutionOnly;
      break;
    default:
      _master.options.all_pos_eq_mode = LiteralProperties::DefaultEqualityMode;
    };


  switch (_options.all_neg_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      _master.options.all_neg_eq_mode = LiteralProperties::UseEqForResolution;
      break;
    default:
      _master.options.all_neg_eq_mode = LiteralProperties::DefaultEqualityMode;
    };



  // numeric options
  _master.options.age_weight_ratio_numerator = 1;
  _master.options.age_weight_ratio_denominator = 0;
  _master.options.best_good_weight_ratio_numerator = 1;
  _master.options.best_good_weight_ratio_denominator = 0;
  _master.options.timeLimit = LONG_MAX/10;
  _master.options.memoryLimit = (float)LONG_MAX;
  _master.options.allocation_buffer_price = (float)_options.allocation_buffer_price;
  _master.options.maxActive = LONG_MAX;
  _master.options.maxPassive = LONG_MAX;
  _master.options.maxWeight = LONG_MAX;
  _master.options.lrs_softness = 1.0;
  _master.options.first_time_check = 1.0; 
  _master.options.first_mem_check = 1.0;
  // interface options
  _master.options.outputStreamForGenerated = (_options.show_gen) ? _options.std_output : 0;
  _master.options.showPassive = (_options.show_passive) ? outputClause : 0;
  _master.options.showActive = (_options.show_active) ? outputClause : 0;
  _master.options.showLitProp = _options.show_lit_prop;
  _master.options.outputStreamForPassiveProfiles = _options.std_output;
  _master.options.showPassiveProfileInterval = _options.show_passive_profile_interval;
  _master.options.showWeightChange = (_options.show_weight_limit_changes) ? outputWeightChange : 0;

#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _master.options.step_resolution_literal_selection = _options.step_resolution_literal_selection;
#endif

  _master.setStrategy();
}; // void VampireKernel::_Implementation::prepareMasterForSimplification()

void  
VampireKernel::_Implementation::prepareMasterForDefElimination()
{
  CALL("prepareMasterForDefElimination()");
  _master.reset();

  _master.options.main_alg = 0;
  _master.options.binary_resolution = true; // for elimination of definitions
  _master.options.no_inferences_between_rules = false;
  _master.options.splitting = false;   
  _master.options.splitting_style = 1L;
  _master.options.small_basic_comp_in_hyp_splitting = false;
  _master.options.neg_eq_splitting = 0L; // off 
  _master.options.neg_eq_spl_var_count_limit = 0L;   
  _master.options.paramodulation = false;
  _master.options.paramodulate_from_variables = _options.paramodulate_from_variables;
  _master.options.sim_back_sup = 0;
  _master.options.forw_sup_no_dup_red = false;
  _master.options.back_sup_no_dup_red = false;
 
  _master.options.prohibit_non_basic_superposition = false;
  _master.options.prohibit_non_basic_fd = false;
  _master.options.prohibit_non_basic_bd = false;
  _master.options.reduce_non_basic = false;
  _master.options.input_non_basicness = false;
  _master.options.run_time_non_basicness = false;

  _master.options.selection = _options.selection;
  _master.options.literal_comparison_mode = _options.literal_comparison_mode;
  _master.options.inherited_negative_selection = _options.inherited_negative_selection;

  // strategy switches        
  _master.options.handle_equality = true;
  _master.options.forward_subsumption = true;
  _master.options.fs_set_mode = true; // not to worry about completeness here 
  _master.options.fs_use_sig_filters = false;
  _master.options.simplify_by_fs = _options.simplify_by_fs;
  _master.options.main_premise_subsumption = 0;
  _master.options.side_premise_subsumption = 0;
  _master.options.main_selected_lit_demodulation = 0;
  _master.options.side_selected_lit_demodulation = 0;
  _master.options.forward_demodulation = 0;
  _master.options.fd_on_splitting_branches = false;
  if (_options.normalize_by_commutativity)
    {
      _master.options.normalize_by_commutativity = _problemProfile.commutativityTable();
    }
  else 
    _master.options.normalize_by_commutativity = 0;
  if (_options.eq_taut_mod_C)
    {
      _master.options.eq_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.eq_taut_mod_C = 0;
  if (_options.prop_taut_mod_C)
    {
      _master.options.prop_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.prop_taut_mod_C = 0;
  if (_options.trivial_eq_res_mod_C)
    {
      _master.options.trivial_eq_res_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.trivial_eq_res_mod_C = 0;
  if (_options.identical_lit_factoring_mod_C)
    {
      _master.options.identical_lit_factoring_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.identical_lit_factoring_mod_C = 0;

  _master.options.chaining_for_transitive = 0;
  _master.options.chaining_through_variables = _options.chaining_through_variables;
  _master.options.backward_demodulation = 0;
  _master.options.bd_on_splitting_branches = false;
  _master.options.simplify_by_eq_res = _options.simplify_by_eq_res;
  _master.options.backward_subsumption = true;
  _master.options.bs_set_mode = true; // not to worry about completeness here 
  _master.options.orphan_murder = 0;
  _master.options.lrs = 0;
  _master.options.pred_elim_table = _problemProfile.eliminationTable();

  _master.options.orient_pos_equalities_in_support = true;
  _master.options.orient_neg_equalities_in_support = true;

  switch (_options.all_pos_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      _master.options.all_pos_eq_mode = LiteralProperties::UseEqForResolution;
      break;
    case LiteralAnnotation::ResolutionOnly:
      _master.options.all_pos_eq_mode = LiteralProperties::UseEqForResolutionOnly;
      break;
    default:
      _master.options.all_pos_eq_mode = LiteralProperties::DefaultEqualityMode;
    };


  switch (_options.all_neg_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      _master.options.all_neg_eq_mode = LiteralProperties::UseEqForResolution;
      break;
    default:
      _master.options.all_neg_eq_mode = LiteralProperties::DefaultEqualityMode;
    };



  // numeric options
  _master.options.age_weight_ratio_numerator = 1;
  _master.options.age_weight_ratio_denominator = 0;
  _master.options.best_good_weight_ratio_numerator = 1;
  _master.options.best_good_weight_ratio_denominator = 0;
  _master.options.timeLimit = LONG_MAX/10;
  _master.options.memoryLimit = (float)LONG_MAX;
  _master.options.allocation_buffer_price = (float)_options.allocation_buffer_price;
  _master.options.maxActive = LONG_MAX;
  _master.options.maxPassive = LONG_MAX;
  _master.options.maxWeight = LONG_MAX;
  _master.options.lrs_softness = 1.0; 
  _master.options.first_time_check = 1.0;
  _master.options.first_mem_check = 1.0;
  // interface options
  _master.options.outputStreamForGenerated = (_options.show_gen) ? _options.std_output : 0;
  _master.options.showPassive = (_options.show_passive) ? outputClause : 0;
  _master.options.showActive = (_options.show_active) ? outputClause : 0;
  _master.options.showLitProp = _options.show_lit_prop;
  _master.options.outputStreamForPassiveProfiles = _options.std_output;
  _master.options.showPassiveProfileInterval = _options.show_passive_profile_interval;
  _master.options.showWeightChange = (_options.show_weight_limit_changes) ? outputWeightChange : 0;

#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _master.options.step_resolution_literal_selection = _options.step_resolution_literal_selection;
#endif

  _master.setStrategy();
}; // void VampireKernel::_Implementation::prepareMasterForDefElimination()

void  
VampireKernel::_Implementation::loadInputQueryClausesIntoMaster()
{
  CALL("loadInputQueryClausesIntoMaster()");

  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_Prepro_Simplification:
      // same as below
    case Continuation_Prepro_DefElimination_LoadInput:
      // same as below
    case Continuation_LoadInputQuery:
      goto load_clause;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::loadInputQueryClausesIntoMaster()");
    };

 next_clause:
  if (_inputQueryClauses.dequeue(_currentInputQueryClause))
    {
      _currentInputQueryClause->setMainSet(VK::Clause::MainSetPreprocessingHistory);
      _preprocessingHistory.enqueue(_currentInputQueryClause);
    load_clause:
      _master.inputClause(_currentInputQueryClause);
      if (_master.terminated()) 
	{
	  return;
	};
      _continuation = Continuation_Init;	
      goto next_clause;
    };
    
  _master.endOfInput();
  ASSERT(_inputQueryClauses.empty());
}; // void VampireKernel::_Implementation::loadInputQueryClausesIntoMaster()



void  
VampireKernel::_Implementation::loadInputSupportClausesIntoMaster()
{
  CALL("loadInputSupportClausesIntoMaster()");

  switch (_continuation)
    {
    case Continuation_Init:
      break;
    case Continuation_LoadInputSupport:
      goto load_clause;
    default:
      ICP("ICP0");
      RuntimeError::report("Bad _continuation in VampireKernel::_Implementation::loadInputSupportClausesIntoMaster()");
    };

 next_clause:
  if (_inputSupportClauses.dequeue(_currentInputSupportClause))
    {
      _currentInputSupportClause->setMainSet(VK::Clause::MainSetPreprocessingHistory);
      _preprocessingHistory.enqueue(_currentInputSupportClause);
    load_clause:
      _master.inputClause(_currentInputSupportClause);
      if (_master.terminated()) 
	{
	  return;
	};
      _continuation = Continuation_Init;
      goto next_clause;
    };
  _master.endOfInput();
  ASSERT(_inputSupportClauses.empty());
}; // void VampireKernel::_Implementation::loadInputSupportClausesIntoMaster()



void  
VampireKernel::_Implementation::extractPreprocessedQueryClauses()
{
  CALL("extractPreprocessedQueryClauses()");

  ASSERT(_inputQueryClauses.empty());
  _master.passiveResetIter();
  _master.historyResetIter();
  VK::Clause* cl;
  for (cl = _master.nextPassive(); cl; cl = _master.nextPassive())
    {
      ASSERT(!cl->isBuiltInTheoryFact());

      VK::Clause* copy = 
	_staticClauseAssembler.assembleSharedCopy(cl);
      _numberToClauseMapping.insert(copy);
      copy->setMainSet(VK::Clause::MainSetInputQuery);
      _inputQueryClauses.enqueue(copy);
    };
  for (cl = _master.nextHistory(); cl; cl = _master.nextHistory())
    {
      VK::Clause* copy = 
	_staticClauseAssembler.assembleSharedCopy(cl);
      _numberToClauseMapping.insert(copy);
      copy->setMainSet(VK::Clause::MainSetPreprocessingHistory);
      _preprocessingHistory.enqueue(copy); 
    };

  const ClauseQueue::Entry* qe;
  for (qe = _inputQueryClauses.begin(); 
       qe; 
       qe = qe->next())
    {
      cl = qe->value();
      if (!cl->isInputClause() &&
	  !cl->isBuiltInTheoryFact())
	{
	  for (ClauseBackground::ClauseList* ancestors = 
		 cl->background().ancestors(); 
	       ancestors; 
	       ancestors = ancestors->tl())
	    {
	      VK::Clause* ancClause = ancestors->hd();
	      VK::Clause* ancCopy = 
		_numberToClauseMapping.find(ancClause->Number());
	      if (ancCopy)  { ancestors->hd() = ancCopy; }
	      else 
		{
		  ASSERT(ancClause->isInputClause() || 
			 ancClause->isBuiltInTheoryFact() ||
			 ancClause->isNameIntro());
		}; 
	    };  
	};

    };

  for (qe = _preprocessingHistory.begin(); 
       qe; 
       qe = qe->next())
    {
      cl = qe->value(); 
      if (!cl->isInputClause() &&
	  !cl->isBuiltInTheoryFact())
	{
	  for (ClauseBackground::ClauseList* ancestors = 
		 cl->background().ancestors(); 
	       ancestors; 
	       ancestors = ancestors->tl())
	    {
	      VK::Clause* ancClause = ancestors->hd();
	      VK::Clause* ancCopy = _numberToClauseMapping.find(ancClause->Number());
	      if (ancCopy)  { ancestors->hd() = ancCopy; }
	      else 
		{
		  ASSERT(ancClause->isInputClause() || 
			 ancClause->isBuiltInTheoryFact() ||
			 ancClause->isNameIntro());
		}; 
	    };
	};
    };
}; // void VampireKernel::_Implementation::extractPreprocessedQueryClauses()

void  
VampireKernel::_Implementation::prepareMasterForSaturation(long timeLimit, // deciseconds
							   float memoryLimit) // bytes
{
  _master.reset();
  _master.options.incompleteAPriori = _incompleteProcedure;
  _master.options.main_alg = _options.main_alg;
  _master.options.binary_resolution = true;
  _master.options.no_inferences_between_rules = _options.no_inferences_between_rules;
  _master.options.splitting = _options.dynamic_splitting;  
  _master.options.splitting_style = _options.splitting_style;
  _master.options.small_basic_comp_in_hyp_splitting = _options.small_basic_comp_in_hyp_splitting;
  _master.options.neg_eq_splitting = _options.dynamic_neg_eq_splitting; 
  _master.options.neg_eq_spl_var_count_limit = _options.dyn_neg_eq_spl_var_count_limit; 
  _master.options.paramodulation = _options.paramodulation;
  _master.options.paramodulate_from_variables = _options.paramodulate_from_variables;
  _master.options.sim_back_sup = _options.sim_back_sup;
  _master.options.forw_sup_no_dup_red = _options.forw_sup_no_dup_red;
  _master.options.back_sup_no_dup_red = _options.back_sup_no_dup_red;


  _master.options.prohibit_non_basic_superposition = 
    _options.prohibit_non_basic_superposition;

  _master.options.prohibit_non_basic_fd = 
    _options.prohibit_non_basic_fd;
    
  _master.options.prohibit_non_basic_bd = 
    _options.prohibit_non_basic_bd;
    
  _master.options.reduce_non_basic = 
    _options.reduce_non_basic;

  _master.options.input_non_basicness = 
    _options.input_non_basicness;

  _master.options.run_time_non_basicness = 
    _options.run_time_non_basicness;

  _master.options.selection = _options.selection;
  _master.options.literal_comparison_mode = _options.literal_comparison_mode;
  _master.options.inherited_negative_selection = _options.inherited_negative_selection;

  // strategy switches
  _master.options.handle_equality = _options.paramodulation;
  _master.options.forward_subsumption = _options.forward_subsumption;
  _master.options.fs_set_mode = _options.fs_set_mode;
  _master.options.fs_use_sig_filters = _options.fs_use_sig_filters;
  _master.options.simplify_by_fs = _options.simplify_by_fs;
  _master.options.main_premise_subsumption = _options.main_premise_subsumption;
  _master.options.side_premise_subsumption = _options.side_premise_subsumption;
  _master.options.main_selected_lit_demodulation = _options.main_selected_lit_demodulation;
  _master.options.side_selected_lit_demodulation = _options.side_selected_lit_demodulation;
  _master.options.forward_demodulation = _options.forward_demodulation;
  _master.options.fd_on_splitting_branches = _options.fd_on_splitting_branches;
  if (_options.normalize_by_commutativity)
    {
      _master.options.normalize_by_commutativity = _problemProfile.commutativityTable();
    }
  else
    _master.options.normalize_by_commutativity = 0;
  if (_options.eq_taut_mod_C)
    {
      _master.options.eq_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.eq_taut_mod_C = 0;
  if (_options.prop_taut_mod_C)
    {
      _master.options.prop_taut_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.prop_taut_mod_C = 0;
  if (_options.trivial_eq_res_mod_C)
    {
      _master.options.trivial_eq_res_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.trivial_eq_res_mod_C = 0;
  if (_options.identical_lit_factoring_mod_C)
    {
      _master.options.identical_lit_factoring_mod_C = _problemProfile.commutativityTable();
    }
  else 
    _master.options.identical_lit_factoring_mod_C = 0;


  if (_options.chaining_for_transitive)
    {
      _master.options.chaining_for_transitive = _problemProfile.transitivityTable();
    }
  else
    _master.options.chaining_for_transitive = 0;
  _master.options.chaining_through_variables = _options.chaining_through_variables;

  _master.options.backward_demodulation = _options.backward_demodulation;
  _master.options.bd_on_splitting_branches = _options.bd_on_splitting_branches;
  _master.options.simplify_by_eq_res = _options.simplify_by_eq_res;
  _master.options.backward_subsumption = _options.backward_subsumption;
  _master.options.bs_set_mode = _options.bs_set_mode;
  _master.options.orphan_murder = _options.orphan_murder;
  _master.options.lrs = _options.lrs;

  _master.options.pred_elim_table = 0;

  _master.options.orient_pos_equalities_in_support = _options.orient_pos_equalities_in_support;
  _master.options.orient_neg_equalities_in_support = _options.orient_neg_equalities_in_support;

  switch (_options.all_pos_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      _master.options.all_pos_eq_mode = LiteralProperties::UseEqForResolution;
      break;
    case LiteralAnnotation::ResolutionOnly:
      _master.options.all_pos_eq_mode = LiteralProperties::UseEqForResolutionOnly;
      break;
    default:
      _master.options.all_pos_eq_mode = LiteralProperties::DefaultEqualityMode;
    };


  switch (_options.all_neg_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      _master.options.all_neg_eq_mode = LiteralProperties::UseEqForResolution;
      break;
    default:
      _master.options.all_neg_eq_mode = LiteralProperties::DefaultEqualityMode;
    };


  // numeric options
  _master.options.age_weight_ratio_numerator = _options.age_weight_ratio_numerator;
  _master.options.age_weight_ratio_denominator = _options.age_weight_ratio_denominator;
  _master.options.best_good_weight_ratio_numerator = _options.best_good_weight_ratio_numerator;
  _master.options.best_good_weight_ratio_denominator = _options.best_good_weight_ratio_denominator;
  _master.options.timeLimit = timeLimit;
  _master.options.memoryLimit = memoryLimit;
  _master.options.allocation_buffer_price = (float)_options.allocation_buffer_price;
  _master.options.maxActive = _options.max_active;
  _master.options.maxPassive = _options.max_passive;
  _master.options.maxWeight = _options.max_weight;
  _master.options.maxInferenceDepth = _options.max_inference_depth;
  _master.options.lrs_softness = _options.lrs_softness;
  _master.options.first_time_check = ((float)_options.first_time_check)/100;
  _master.options.first_mem_check = ((float)_options.first_mem_check)/100;
  // interface options 
  _master.options.outputStreamForGenerated = (_options.show_gen) ? _options.std_output : 0;
  _master.options.showPassive = (_options.show_passive) ? outputClause : 0;
  _master.options.showActive = (_options.show_active) ? outputClause : 0;
  _master.options.showLitProp = _options.show_lit_prop;
  _master.options.outputStreamForPassiveProfiles = _options.std_output;
  _master.options.showPassiveProfileInterval = _options.show_passive_profile_interval;
  _master.options.showWeightChange = (_options.show_weight_limit_changes) ? outputWeightChange : 0;


#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _master.options.step_resolution_literal_selection = _options.step_resolution_literal_selection;
#endif


  _master.setStrategy();
#ifdef INTERCEPT_BACKWARD_MATCHING_INDEXING
  BMILog.unblock();
#endif

#ifdef INTERCEPT_INDEXED_SEARCH_FOR_GENERALISATIONS
  SFGILog.unblock();
#endif

#ifdef INTERCEPT_BACKWARD_SUPERPOSITION_INDEXING
  BSIULog.unblock();
#endif

#ifdef INTERCEPT_BACKWARD_SUBSUMPTION_INDEXING
  BSILog.unblock();
#endif
}; // void VampireKernel::_Implementation::prepareMasterForSaturation(long timeLimit,float memoryLimit)


void  
VampireKernel::_Implementation::reportTermination()
{
  CALL("reportTermination()");
    
  autoStatOutput(_options.std_output);
  switch (_master.terminationReason())
    {
    case Master::UnknownTR:
      ICP("ICP0");
      if (_options.std_output)
	*_options.std_output << "% Proof not found. Internal error: unknown termination type.\n"; 
      _terminationReason = VampireKernel::UnknownTerminationReason;
      break;  
    case Master::Proof: 
      if (_options.std_output)
	*_options.std_output << "% Proof found. Thanks to Tanya!\n"; 
      _terminationReason = VampireKernel::RefutationFound;
      if (_options.proof && _options.std_output)
	{
	  _proofTree.collect(_master.proof());    
	  *_options.std_output << "%================== Proof: ======================\n"; 
	  _proofTree.resetIter();   
	  for (const VK::Clause* cl = _proofTree.getNextClause(); 
	       cl; 
	       cl = _proofTree.getNextClause())
	    *_options.std_output << "% " << cl << '\n';
	    
	  *_options.std_output << "%==============  End of proof. ==================\n"; 
	}; 
      break; 
    case Master::EmptyPassive: 
      if (_master.completeStrategyUsed() && 
	  (!_lookingForMoreAnswers))
	{
	  if (_options.std_output)
	    *_options.std_output << "% Unprovable.\n";  
	  _terminationReason = VampireKernel::Satisfiable;
	}
      else 
	{ 
	  if (_options.std_output)
	    *_options.std_output << "% Proof not found. Empty passive.\n"; 
	  _terminationReason = VampireKernel::EmptyPassive;
	};
      break;

    case Master::MaxPassive: 
      if (_options.std_output)
	*_options.std_output << "% Proof not found. Max. num. of passive reached. \n";
      _terminationReason = VampireKernel::MaxPassive;
      break; 
    case Master::MaxActive: 
      if (_options.std_output)
	*_options.std_output << "% Proof not found. Max. num. of active reached. \n"; 
      _terminationReason = VampireKernel::MaxActive;
      break; 
    }; 
  if (_options.std_output) _options.std_output->flush();
}; // void VampireKernel::_Implementation::reportTermination()  

void  
VampireKernel::_Implementation::reportTimeLimitExpired()
{
  CALL("reportTimeLimitExpired()");  
  _terminationReason = VampireKernel::TimeLimitExpired;
  if (_saturationRunning)
    {
      autoStatOutput(_options.std_output);
    };
  if (_options.std_output)
    {
      *_options.std_output << "% Proof not found. Time limit expired.\n";
      _options.std_output->flush();
    };
}; // void VampireKernel::_Implementation::reportTimeLimitExpired()


void  
VampireKernel::_Implementation::setHardMemoryLimit(float lim) // bytes
{
  CALL("setHardMemoryLimit(float lim)");
  if (lim < 0.0) lim = 0.0;
  long truncated;
  if (lim > (LONG_MAX/2))
    {
      truncated = LONG_MAX/2;
    }
  else
    truncated = (long)lim;
  GlobAlloc::setHardMemoryLimit(truncated);
  GlobAlloc::basicAllocationFailureHandling().setMemoryReleaseRequest(releaseMemory);
  GlobAlloc::basicAllocationFailureHandling().setMassiveMemoryReleaseRequest(releaseALotOfMemory);
  GlobAlloc::basicAllocationFailureHandling().setUnrecoverableAllocationFailureHandler(handleAllocationFailure);
}; // VampireKernel::_Implementation::setHardMemoryLimit(float lim

void  
VampireKernel::_Implementation::setAllocationBufferSize(long percentage)
{
  GlobAlloc::setAllocationBufferSize(percentage);
}; 

bool  
VampireKernel::_Implementation::releaseMemory(ulong size) // bytes
{
  CALL("releaseMemory(ulong size)");
  if (_currentSession && (_currentSession->_mainLoopRunning))
    {
      return _currentSession->_master.releaseMemory(size);
    }
  else
    return false;
}; // bool VampireKernel::_Implementation::releaseMemory(ulong size)
  
bool  
VampireKernel::_Implementation::releaseALotOfMemory()
{
  CALL("releaseALotOfMemory()");
  if (_currentSession && (_currentSession->_mainLoopRunning))
    {
      return _currentSession->_master.releaseALotOfMemory();
    }
  else
    return false;
}; // bool VampireKernel::_Implementation::releaseALotOfMemory()
  
void  
VampireKernel::_Implementation::handleAllocationFailure(ulong size) // bytes
{
  CALL("handleAllocationFailure(ulong size)");
  if (_currentSession) 
    {
      _currentSession->reportAllocationFailure(size);
      DestructionMode::makeFast();
      throw VampireKernel::Emergency(VampireKernel::Emergency::AllocationFailure);
    }
  else
    {
      cout << "% Allocation failure occured when no kernel session is running.\n";
      DestructionMode::makeFast();
      throw VampireKernel::Emergency(VampireKernel::Emergency::AllocationFailure);
    };
}; // void VampireKernel::_Implementation::handleAllocationFailure(ulong size)

void  
VampireKernel::_Implementation::reportAllocationFailure(ulong size)
{
  CALL("reportAllocationFailure(ulong size)");
  _terminationReason = VampireKernel::AbortedOnAllocationFailure;
  if (_options.std_output)
    {
      *_options.std_output << "% ------------- ALLOCATION FAILURE: " 
			   << size << " bytes -----------------\n";
      autoStatOutput(_options.std_output);
      *_options.std_output << "% Proof not found. Allocation failure.\n";
      _options.std_output->flush();
    };
}; // void VampireKernel::_Implementation::reportAllocationFailure(ulong size)

void  
VampireKernel::_Implementation::outputClause(const VK::Clause* cl)
{
  if (_currentSession && (_currentSession->_options.std_output))
    *(_currentSession->_options.std_output) << "% " << cl << "\n";
};
  
void  
VampireKernel::_Implementation::outputWeightChange(long oldWL,long newWL)
{
  if (_currentSession && (_currentSession->_options.std_output))
    *(_currentSession->_options.std_output) << "% WEIGHT LIMIT CHANGED: " << oldWL << " ------> " << newWL << "\n";
};


void  
VampireKernel::_Implementation::outputInputSupportClauses()
{
  if (!_options.std_output) return;
  for (const ClauseQueue::Entry* cl = _inputSupportClauses.begin();
       cl;
       cl = cl->next())
    {
      *_options.std_output << "% " << cl->value() << "\n";
    };
};

void  
VampireKernel::_Implementation::outputInputQueryClauses()
{
  if (!_options.std_output) return;
  for (const ClauseQueue::Entry* cl = _inputQueryClauses.begin();
       cl;
       cl = cl->next())
    {
      *_options.std_output << "% " << cl->value() << "\n";
    };
};
  
const VampireKernel::Statistics&
VampireKernel::_Implementation::statistics()
{
  updateExportedStatistics();
  return _exportedCumulativeStatistics;
};

void VampireKernel::_Implementation::updateExportedStatistics()
{
  _cumulativeStatistics = _preproStatistics;
  _cumulativeStatistics.absorb(_master.statistics());
  
  _exportedCumulativeStatistics.time = _timer.elapsedSeconds();
  _exportedCumulativeStatistics.memory = GlobAlloc::occupiedByObjects();

  VampireKernel::Statistics::DeductionInferences& dedInf = 
    _exportedCumulativeStatistics.deductionInferences;
  VampireKernel::Statistics::SimplifyingInferences& simInf = 
    _exportedCumulativeStatistics.simplifyingInferences;
  VampireKernel::Statistics::PropositionalSplitting& propSpl =
    _exportedCumulativeStatistics.propositionalSplittingInferences;
  VampireKernel::Statistics::NegativeEqualitySplitting& negEqSpl = 
    _exportedCumulativeStatistics.negativeEqualitySplittingInferences;
  VampireKernel::Statistics::GeneratedClauses& genCl = 
    _exportedCumulativeStatistics.generatedClauses;
  VampireKernel::Statistics::RetainedClauses& retCl = 
    _exportedCumulativeStatistics.retainedClauses;

  
  dedInf.totalAccomplished = 
    _cumulativeStatistics.inferences.deduction.totalAccomplished();
  dedInf.resolutionAccomplished = 
    _cumulativeStatistics.inferences.deduction.resolution.accomplished;
  dedInf.forwardSuperposition = 
    _cumulativeStatistics.inferences.deduction.forwardSuperposition.accomplished;
  dedInf.backwardSuperposition = 
    _cumulativeStatistics.inferences.deduction.backwardSuperposition.accomplished;
  dedInf.equalityFactoring = 
    _cumulativeStatistics.inferences.deduction.equalityFactoring.accomplished;
  dedInf.equalityResolution = 
    _cumulativeStatistics.inferences.deduction.equalityResolution.accomplished;
  dedInf.forwardChaining = 
    _cumulativeStatistics.inferences.deduction.forwardChaining.accomplished;
  dedInf.backwardChaining = 
    _cumulativeStatistics.inferences.deduction.backwardChaining.accomplished;
  dedInf.totalRedundantDiscarded = 
    _cumulativeStatistics.inferences.deduction.totalRedundantDiscarded();
  dedInf.totalRedundantBranchesDiscarded = 
    _cumulativeStatistics.inferences.deduction.totalRedundantBranchesDiscarded();
  dedInf.totalNonredundantDiscarded = 
    _cumulativeStatistics.inferences.deduction.totalNonredundantDiscarded();
  dedInf.totalNonredundantBranchesDiscarded = 
    _cumulativeStatistics.inferences.deduction.totalNonredundantBranchesDiscarded();



  simInf.totalForward = 
    _cumulativeStatistics.inferences.simplifying.totalForward();
  simInf.propositionalTautologyDeletion = 
    _cumulativeStatistics.inferences.simplifying.propositionalTautologyDeletion;
  simInf.equationalTautologyDeletion = 
    _cumulativeStatistics.inferences.simplifying.equationalTautologyDeletion;
  simInf.forwardSubsumption = 
    _cumulativeStatistics.inferences.simplifying.forwardSubsumption;
  simInf.forwardSubsumptionResolution = 
    _cumulativeStatistics.inferences.simplifying.forwardSubsumptionResolution;
  simInf.forwardDemodulation = 
    _cumulativeStatistics.inferences.simplifying.forwardDemodulation;
  simInf.onBranchForwardDemodulation = 
    _cumulativeStatistics.inferences.simplifying.onBranchForwardDemodulation;
  simInf.destructiveEqualityResolution = 
    _cumulativeStatistics.inferences.simplifying.destructiveEqualityResolution;
  simInf.builtInTheory = 
    _cumulativeStatistics.inferences.simplifying.builtInTheory;


  simInf.totalBackward = 
    _cumulativeStatistics.inferences.simplifying.totalBackward();
  simInf.backwardSubsumption = 
    _cumulativeStatistics.inferences.simplifying.backwardSubsumption;
  simInf.backwardDemodulation = 
    _cumulativeStatistics.inferences.simplifying.backwardDemodulation;
  simInf.onBranchBackwardDemodulation = 
    _cumulativeStatistics.inferences.simplifying.onBranchBackwardDemodulation;


  propSpl.accomplished = 
    _cumulativeStatistics.inferences.propositionalSplitting.accomplished;
    
  if (_cumulativeStatistics.inferences.propositionalSplitting.accomplished)
    {
      propSpl.averageNumOfComponents = 
	_cumulativeStatistics.inferences.propositionalSplitting.components /
	_cumulativeStatistics.inferences.propositionalSplitting.accomplished;
    }
  else
    propSpl.averageNumOfComponents = 0.0;

  propSpl.introducedComponentNames = 
    _cumulativeStatistics.inferences.propositionalSplitting.introducedComponentNames;


    
  negEqSpl.accomplished = 
    _cumulativeStatistics.inferences.negativeEqualitySplitting.accomplished;
  if (_cumulativeStatistics.inferences.negativeEqualitySplitting.accomplished)
    {
      negEqSpl.averageNumOfArgumentsSplitOff = 
	_cumulativeStatistics.inferences.negativeEqualitySplitting.argumentsSplitOff /
	_cumulativeStatistics.inferences.negativeEqualitySplitting.accomplished;
    }
  else
    negEqSpl.averageNumOfArgumentsSplitOff = 0.0;
  negEqSpl.introducedNames = 
    _cumulativeStatistics.inferences.negativeEqualitySplitting.introducedNames;
    

  genCl.total = 
    _cumulativeStatistics.clauses.generated.total();
  genCl.byDeductionInferences = 
    _cumulativeStatistics.clauses.generated.totalGeneratedByDeduction();
  genCl.bySimplificationInferences = 
    _cumulativeStatistics.clauses.generated.totalGeneratedBySimplification();
  genCl.byPropositionalSplitting = 
    _cumulativeStatistics.clauses.generated.byPropositionalSplitting;
  genCl.byNegativeEqualitySplitting = 
    _cumulativeStatistics.clauses.generated.byNegativeEqualitySplitting;
  genCl.byOther = 
    _cumulativeStatistics.clauses.generated.byOther;
    
  genCl.totalRedundantDiscarded = 
    _cumulativeStatistics.clauses.generated.totalRedundant();
  genCl.discardedPropositionalTautologies = 
    _cumulativeStatistics.clauses.generated.discardedPropositionalTautologies;
  genCl.discardedEquationalTautologies = 
    _cumulativeStatistics.clauses.generated.discardedEquationalTautologies;
  genCl.discardedSubsumed = 
    _cumulativeStatistics.clauses.generated.subsumed;
    
  genCl.totalNonredundantDiscarded = 
    _cumulativeStatistics.clauses.generated.totalNonredundantDiscarded();
    
  genCl.totalSimplified = 
    _cumulativeStatistics.clauses.generated.totalSimplified();
  genCl.transformedByPropositionalSplitting = 
    _cumulativeStatistics.clauses.generated.transformedByPropositionalSplitting;
  genCl.transformedByNegativeEqualitySplitting = 
    _cumulativeStatistics.clauses.generated.transformedByNegativeEqualitySplitting;

  
  retCl.total = 
    _cumulativeStatistics.clauses.retained.total;
  retCl.selected = 
    _cumulativeStatistics.clauses.retained.selected;
  retCl.totalSubsumed = 
    _cumulativeStatistics.clauses.retained.passiveSubsumed +
    _cumulativeStatistics.clauses.retained.activeSubsumed;
  retCl.totalSimplified = 
    _cumulativeStatistics.clauses.retained.totalSimplified();
  retCl.totalSimplifiedByDemodulation = 
    _cumulativeStatistics.clauses.retained.passiveSimplifiedByDemodulation +
    _cumulativeStatistics.clauses.retained.activeSimplifiedByDemodulation;
  retCl.totalSimplifiedByOnBranchDemodulation = 
    _cumulativeStatistics.clauses.retained.passiveSimplifiedByOnBranchDemodulation +
    _cumulativeStatistics.clauses.retained.activeSimplifiedByOnBranchDemodulation;
  retCl.totalOrphansMurdered = 
    _cumulativeStatistics.clauses.retained.passiveOrphansMurdered +
    _cumulativeStatistics.clauses.retained.activeOrphansMurdered; 
  retCl.totalRecycled = 
    _cumulativeStatistics.clauses.retained.passiveRecycled +
    _cumulativeStatistics.clauses.retained.activeRecycled;
  retCl.currentlyPassive = _master.numOfCurrentlyPassive();
  retCl.currentlyActive = _master.numOfCurrentlyActive();
}; // void VampireKernel::_Implementation::updateExportedStatistics()

void 
VampireKernel::_Implementation::outputPreproStat(ostream* str,bool full)
{
  if (!str) return;
  *str << "%========================================================================\n";
  if (full)
    {
      *str << "%===== Full statistics for accomplished preprocessing passes: ============\n";
    }
  else
    {
      *str << "%===== Statistics summary for accomplished preprocessing passes: =========\n";
    };
  *str << "% time " << _preproLastTimeSec << "\n";
  *str << "% memory " << (_preproLastOccupiedByObjectsBytes/1024) 
       << "Kb (" << (_preproLastOccupiedByChunksBytes/1024) << "Kb in chunks)\n"
       << "%     buffer deficit " << (_preproLastTotalBufferDeficitBytes/1024) << "Kb ("
       << (_preproLastActiveBufferDeficitBytes/1024) << "Kb act. + "
       << (_preproLastSuspendedBufferDeficitBytes/1024) << "Kb susp.)\n";
  if (full)
    {
      _preproStatistics.outputDetailed(*str);
    }
  else
    {
      _preproStatistics.outputSummary(*str);
    };
  if (full)
    {
      *str << "%===== End of full statistics for accomplished prepro. passes. ===========\n";
    }
  else
    {
      *str << "%======= End of stat. summary for accomplished prepro. passes. ===========\n";
    };
}; // void VampireKernel::_Implementation::outputPreproStat(ostream* str,bool full)

void 
VampireKernel::_Implementation::outputCurrentLoopStat(ostream* str,bool full)
{
  if (!str) return;
  *str << "%========================================================================\n";
  if (full)
    {
      *str << "%============== Full statistics for current loop: =======================\n";
    }
  else
    {
      *str << "%=========== Statistics summary for current loop: =======================\n";
    };

  *str << "% time " << _master.timer().elapsedSeconds() << "\n";  
  
  ulong defragCalls;
  float defragAverageInputSize;
  ulong defragNumOfInputPieces;
  float defragAverageOutputSize;
  ulong defragNumOfOutputPieces;
  GlobAlloc::getDefragmentationStatistics(defragCalls,
					  defragAverageInputSize,
					  defragNumOfInputPieces,
					  defragAverageOutputSize,
					  defragNumOfOutputPieces);
  *str << "% memory " << (GlobAlloc::occupiedByObjects()/1024) 
       << "Kb (" << (GlobAlloc::occupiedByChunks()/1024) << "Kb in chunks)\n"
       << "%     buffer deficit " << (GlobAlloc::totalBufferDeficit()/1024) << "Kb ("
       << (GlobAlloc::activeBufferDeficit()/1024) << "Kb act. + "
       << (GlobAlloc::suspendedBufferDeficit()/1024) << "Kb susp.)\n"
       << "%     defragmentation: " << defragCalls << " calls";
  if (defragCalls)
    *str << ", " << defragNumOfInputPieces << "*" << floor(defragAverageInputSize)
	 << " bytes ---> " 
	 << defragNumOfOutputPieces << "*" << floor(defragAverageOutputSize) << " bytes";
  *str << "\n";
  if (full)
    {
      _master.statistics().outputDetailed(*str);
    }
  else
    {
      _master.statistics().outputSummary(*str);
    };

  *str << "%== Currently passive " << _master.numOfCurrentlyPassive()
       << " (" << _master.numOfReachablePassive() << " reachable)\n";
  *str << "%== Currently active " << _master.numOfCurrentlyActive() << "\n"; 

  if (full)
    {
      *str << "%============ End of full statistics for current loop. ==================\n";
    }
  else
    {
      *str << "%========== End of statistics summary for current loop. =================\n";
    };
}; // void VampireKernel::_Implementation::outputCurrentLoopStat(ostream* str,bool full)

void 
VampireKernel::_Implementation::outputCumulativeStat(ostream* str,bool full)
{
  if (!str) return;
  _cumulativeStatistics = _preproStatistics;
  _cumulativeStatistics.absorb(_master.statistics());
  *str << "%========================================================================\n";
  if (full)
    {
      *str << "%===== Full cumulative statistics (prepro. + current loop): =============\n";
    }
  else
    {
      *str << "%===== Summary of cumulative statistics (prepro. + curr. loop): =========\n";
    };
  
  *str << "% time " <<  _timer.elapsedSeconds()
       << " (" << _master.timer().elapsedSeconds() << " current loop)\n";
  
  ulong defragCalls;
  float defragAverageInputSize;
  ulong defragNumOfInputPieces;
  float defragAverageOutputSize;
  ulong defragNumOfOutputPieces;
  GlobAlloc::getDefragmentationStatistics(defragCalls,
					  defragAverageInputSize,
					  defragNumOfInputPieces,
					  defragAverageOutputSize,
					  defragNumOfOutputPieces);
  *str << "% memory " << (GlobAlloc::occupiedByObjects()/1024) 
       << "Kb (" << (GlobAlloc::occupiedByChunks()/1024) << "Kb in chunks)\n"
       << "%     buffer deficit " << (GlobAlloc::totalBufferDeficit()/1024) << "Kb ("
       << (GlobAlloc::activeBufferDeficit()/1024) << "Kb act. + "
       << (GlobAlloc::suspendedBufferDeficit()/1024) << "Kb susp.)\n"
       << "%     defragmentation: " << defragCalls << " calls";
  if (defragCalls)
    *str << ", " << defragNumOfInputPieces << "*" << floor(defragAverageInputSize)
	 << " bytes ---> " 
	 << defragNumOfOutputPieces << "*" << floor(defragAverageOutputSize) << " bytes";
  *str << "\n";
  if (full)
    {
      _cumulativeStatistics.outputDetailed(*str);
    }
  else
    {
      _cumulativeStatistics.outputSummary(*str);
    };

  *str << "%== Currently passive " << _master.numOfCurrentlyPassive()
       << " (" << _master.numOfReachablePassive() << " reachable)\n";
  *str << "%== Currently active " << _master.numOfCurrentlyActive() << "\n"; 

  if (full)
    {
      *str << "%================ End of full cumulative statistics. ====================\n";
    }
  else
    {
      *str << "%=============== End of cumulative statistics summary. ==================\n";
    };
}; // void VampireKernel::_Implementation::outputCumulativeStat(ostream* str,bool full)

  
void 
VampireKernel::_Implementation::autoStatOutput(ostream* str)
{
  if (!str) return;
  if (_options.auto_stat_prepro)
    {
      if (_options.auto_stat_full) outputPreproStat(str,true);
      if (_options.auto_stat_summary) outputPreproStat(str,false);
    };
  if (_options.auto_stat_curr_loop)
    {
      if (_options.auto_stat_full) outputCurrentLoopStat(str,true);
      if (_options.auto_stat_summary) outputCurrentLoopStat(str,false);
    };
  if (_options.auto_stat_cumulative)
    {
      if (_options.auto_stat_full) outputCumulativeStat(str,true);
      if (_options.auto_stat_summary) outputCumulativeStat(str,false);
    };
}; // VampireKernel::_Implementation::autoStatOutput(ostream* str)


void  
VampireKernel::_Implementation::tabulateStatistics()
{
  CALL("tabulateStatistics()");

  updateExportedStatistics();

  _tabulation->stat("time",_exportedCumulativeStatistics.time);
  _tabulation->stat("memory",_exportedCumulativeStatistics.memory);

  _tabulation->stat("ded_inf_total_accomplished",_exportedCumulativeStatistics.deductionInferences.totalAccomplished);
  _tabulation->stat("ded_inf_total_redundant_discarded",_exportedCumulativeStatistics.deductionInferences.totalRedundantDiscarded);
  _tabulation->stat("ded_inf_total_redundant_branches_discarded",_exportedCumulativeStatistics.deductionInferences.totalRedundantBranchesDiscarded);
  _tabulation->stat("ded_inf_nonredundant_discarded",_exportedCumulativeStatistics.deductionInferences.totalNonredundantDiscarded);
  _tabulation->stat("ded_inf_nonredundant_branches_discarded",_exportedCumulativeStatistics.deductionInferences.totalNonredundantBranchesDiscarded);



  _tabulation->stat("sim_inf_prop_tautology_deletion",_exportedCumulativeStatistics.simplifyingInferences.propositionalTautologyDeletion);
  _tabulation->stat("sim_inf_eq_tautology_deletion",_exportedCumulativeStatistics.simplifyingInferences.equationalTautologyDeletion);
  _tabulation->stat("sim_inf_forward_subsumption",_exportedCumulativeStatistics.simplifyingInferences.forwardSubsumption);
  _tabulation->stat("sim_inf_forward_subs_resolution",_exportedCumulativeStatistics.simplifyingInferences.forwardSubsumptionResolution);
  _tabulation->stat("sim_inf_forward_demodulation",
		    _exportedCumulativeStatistics.simplifyingInferences.forwardDemodulation + 
		    _exportedCumulativeStatistics.simplifyingInferences.onBranchForwardDemodulation);
  _tabulation->stat("sim_inf_destructive_eq_res",_exportedCumulativeStatistics.simplifyingInferences.destructiveEqualityResolution);
  _tabulation->stat("sim_inf_built_in_theory",_exportedCumulativeStatistics.simplifyingInferences.builtInTheory);
  _tabulation->stat("sim_inf_backward_subsumption",_exportedCumulativeStatistics.simplifyingInferences.backwardSubsumption);
  _tabulation->stat("sim_inf_backward_demodulation",
		    _exportedCumulativeStatistics.simplifyingInferences.backwardDemodulation + 
		    _exportedCumulativeStatistics.simplifyingInferences.onBranchBackwardDemodulation);



  _tabulation->stat("prop_spl_accomplished",_exportedCumulativeStatistics.propositionalSplittingInferences.accomplished);
  _tabulation->stat("prop_spl_avg_num_of_components",_exportedCumulativeStatistics.propositionalSplittingInferences.averageNumOfComponents);
  _tabulation->stat("prop_spl_introduced_comp_names",_exportedCumulativeStatistics.propositionalSplittingInferences.introducedComponentNames);

  _tabulation->stat("neg_eq_spl_accomplished",_exportedCumulativeStatistics.negativeEqualitySplittingInferences.accomplished);
  _tabulation->stat("neg_eq_spl_avg_num_of_args_split_off",_exportedCumulativeStatistics.negativeEqualitySplittingInferences.averageNumOfArgumentsSplitOff);
  _tabulation->stat("neg_eq_spl_introduced_names",_exportedCumulativeStatistics.negativeEqualitySplittingInferences.introducedNames);


  _tabulation->stat("gen_cl_total",_exportedCumulativeStatistics.generatedClauses.total);
  _tabulation->stat("gen_cl_total_redundant_discarded",_exportedCumulativeStatistics.generatedClauses.totalRedundantDiscarded);
  _tabulation->stat("gen_cl_total_nonredundant_discarded",_exportedCumulativeStatistics.generatedClauses.totalNonredundantDiscarded);



  _tabulation->stat("ret_cl_total",_exportedCumulativeStatistics.retainedClauses.total);
  _tabulation->stat("ret_cl_selected",_exportedCumulativeStatistics.retainedClauses.selected);
  _tabulation->stat("ret_cl_total_subsumed",_exportedCumulativeStatistics.retainedClauses.totalSubsumed);
  _tabulation->stat("ret_cl_total_simplified",_exportedCumulativeStatistics.retainedClauses.totalSimplified);
  _tabulation->stat("ret_cl_total_orphans_murdered",_exportedCumulativeStatistics.retainedClauses.totalOrphansMurdered);
  _tabulation->stat("ret_cl_total_recycled",_exportedCumulativeStatistics.retainedClauses.totalRecycled);
  _tabulation->stat("ret_cl_currently_passive",_exportedCumulativeStatistics.retainedClauses.currentlyPassive);
  _tabulation->stat("ret_cl_currently_active",_exportedCumulativeStatistics.retainedClauses.currentlyActive);

}; // void VampireKernel::_Implementation::tabulateStatistics()



inline  
const char*  
VampireKernel::_Implementation::flagValue(bool flag)
{
  if (flag) return "on";
  return "off";
};

void  
VampireKernel::_Implementation::tabulateOptions()
{
  CALL("tabulateOptions()");
  if (!_tabulation) return;

  // main algorithm
  _tabulation->option("main_alg",_options.main_alg);

  // inference rules
  _tabulation->option("no_inferences_between_rules",flagValue(_options.no_inferences_between_rules));
  _tabulation->option("static_splitting",flagValue(_options.static_splitting));
  _tabulation->option("dynamic_splitting",flagValue(_options.dynamic_splitting));
  _tabulation->option("splitting_style",_options.splitting_style);
  _tabulation->option("small_basic_comp_in_hyp_splitting",_options.small_basic_comp_in_hyp_splitting);
  _tabulation->option("static_neg_eq_splitting",_options.static_neg_eq_splitting);
  _tabulation->option("stat_neg_eq_spl_var_count_limit",_options.stat_neg_eq_spl_var_count_limit);
  _tabulation->option("dynamic_neg_eq_splitting",_options.dynamic_neg_eq_splitting);
  _tabulation->option("dyn_neg_eq_spl_var_count_limit",_options.dyn_neg_eq_spl_var_count_limit);
  _tabulation->option("paramodulation",flagValue(_options.paramodulation));
  _tabulation->option("paramodulate_from_variables",flagValue(_options.paramodulate_from_variables));
  _tabulation->option("sim_back_sup",_options.sim_back_sup);
  _tabulation->option("forw_sup_no_dup_red",flagValue(_options.forw_sup_no_dup_red));
  _tabulation->option("back_sup_no_dup_red",flagValue(_options.back_sup_no_dup_red));
  _tabulation->option("prohibit_non_basic_superposition",flagValue(_options.prohibit_non_basic_superposition));
  _tabulation->option("prohibit_non_basic_fd",flagValue(_options.prohibit_non_basic_fd));
  _tabulation->option("prohibit_non_basic_bd",flagValue(_options.prohibit_non_basic_bd));
  _tabulation->option("reduce_non_basic",flagValue(_options.reduce_non_basic));
  _tabulation->option("input_non_basicness",flagValue(_options.input_non_basicness));
  _tabulation->option("run_time_non_basicness",flagValue(_options.run_time_non_basicness));
  _tabulation->option("selection",_options.selection);
  _tabulation->option("literal_comparison_mode",_options.literal_comparison_mode);
  _tabulation->option("inherited_negative_selection",flagValue(_options.inherited_negative_selection));
     
  // strategy 
  _tabulation->option("forward_subsumption",flagValue(_options.forward_subsumption));
  _tabulation->option("fs_set_mode",flagValue(_options.fs_set_mode));
  _tabulation->option("fs_use_sig_filters",flagValue(_options.fs_use_sig_filters));
  _tabulation->option("simplify_by_fs",_options.simplify_by_fs);
  _tabulation->option("main_premise_subsumption",_options.main_premise_subsumption);
  _tabulation->option("side_premise_subsumption",_options.side_premise_subsumption);
  _tabulation->option("main_selected_lit_demodulation",_options.main_selected_lit_demodulation);
  _tabulation->option("side_selected_lit_demodulation",_options.side_selected_lit_demodulation);
  _tabulation->option("forward_demodulation",_options.forward_demodulation);
  _tabulation->option("fd_on_splitting_branches",flagValue(_options.fd_on_splitting_branches));
  _tabulation->option("normalize_by_commutativity",flagValue(_options.normalize_by_commutativity));

  _tabulation->option("eq_taut_mod_C",flagValue(_options.eq_taut_mod_C));
  _tabulation->option("prop_taut_mod_C",flagValue(_options.prop_taut_mod_C));
  _tabulation->option("trivial_eq_res_mod_C",flagValue(_options.trivial_eq_res_mod_C));
  _tabulation->option("identical_lit_factoring_mod_C",flagValue(_options.identical_lit_factoring_mod_C));

  _tabulation->option("chaining_for_transitive",flagValue(_options.chaining_for_transitive));
  _tabulation->option("chaining_through_variables",flagValue(_options.chaining_through_variables));
  _tabulation->option("backward_demodulation",_options.backward_demodulation);
  _tabulation->option("bd_on_splitting_branches",flagValue(_options.bd_on_splitting_branches));
  _tabulation->option("simplify_by_eq_res",flagValue(_options.simplify_by_eq_res));
  _tabulation->option("backward_subsumption",_options.backward_subsumption);
  _tabulation->option("bs_set_mode",flagValue(_options.bs_set_mode));
  _tabulation->option("orphan_murder",flagValue(_options.orphan_murder != 0L));
  _tabulation->option("lrs",_options.lrs);
  _tabulation->option("elim_def",_options.elim_def);
    
  switch (_options.simplification_ordering)
    {
    case VampireKernel::Options::NonrecursiveKBOrdering:
      _tabulation->option("simplification_ordering","nonrecursive_KB");
      break;
    case VampireKernel::Options::StandardKBOrdering:
      _tabulation->option("simplification_ordering","standard_KB");
      break;
    default: 
      RuntimeError::report("Wrong value of --simplification_ordering.");
    };

  _tabulation->option("symbol_precedence_by_arity",_options.symbol_precedence_by_arity);
  _tabulation->option("header_precedence_kinky",flagValue(_options.header_precedence_kinky));
  switch (_options.term_weighting_scheme)
    {
    case VampireKernel::Options::UniformTermWeighting:
      _tabulation->option("term_weighting_scheme","uniform");
      break;
    case VampireKernel::Options::NonuniformConstantTermWeighting: 
      _tabulation->option("term_weighting_scheme","nonuniform_constant");
      break;
    default:
      RuntimeError::report("Wrong value of --term_weighting_scheme.");
    };
  _tabulation->option("symbol_weight_by_arity",_options.symbol_weight_by_arity);
  _tabulation->option("orient_pos_equalities_in_support",flagValue(_options.orient_pos_equalities_in_support));
  _tabulation->option("orient_neg_equalities_in_support",flagValue(_options.orient_neg_equalities_in_support));



  switch (_options.all_pos_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      _tabulation->option("all_pos_eq_mode","add_res");
      break;
    case LiteralAnnotation::ResolutionOnly:
      _tabulation->option("all_pos_eq_mode","res_only");
      break;
    default:
      _tabulation->option("all_pos_eq_mode","default");      
    };


  switch (_options.all_neg_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      _tabulation->option("all_neg_eq_mode","add_res");
      break;
    default:
      _tabulation->option("all_neg_eq_mode","default");      
    };


  // numeric options 
  _tabulation->option("time_limit",((float)_saturationTimeLimit)/10);
  _tabulation->option("memory_limit",_options.memory_limit);
  _tabulation->option("allocation_buffer_size",_options.allocation_buffer_size); 
  _tabulation->option("allocation_buffer_price",_options.allocation_buffer_price); 
  _tabulation->option("max_skolem",_options.max_skolem);
  _tabulation->option("max_active",_options.max_active);
  _tabulation->option("max_passive",_options.max_passive);
  _tabulation->option("max_weight",_options.max_weight);
  _tabulation->option("max_inference_depth",_options.max_inference_depth);
  _tabulation->option("age_weight_ratio_numerator",_options.age_weight_ratio_numerator);
  _tabulation->option("age_weight_ratio_denominator",_options.age_weight_ratio_denominator);
  _tabulation->option("best_good_weight_ratio_numerator",_options.best_good_weight_ratio_numerator);
  _tabulation->option("best_good_weight_ratio_denominator",_options.best_good_weight_ratio_denominator);
  _tabulation->option("selected_pos_eq_penalty_coefficient",_options.selected_pos_eq_penalty_coefficient);    
  _tabulation->option("nongoal_penalty_coefficient",_options.nongoal_penalty_coefficient);
  _tabulation->option("lrs_softness",_options.lrs_softness);   
  _tabulation->option("first_time_check",_options.first_time_check);   
  _tabulation->option("first_mem_check",_options.first_mem_check);   

  // interface options
  _tabulation->option("proof",flagValue(_options.proof));

#ifdef VKERNEL_FOR_STEP_RESOLUTION
  _tabulation->option("step_resolution_literal_selection",flagValue(_options.step_resolution_literal_selection));
#endif
  
}; // void VampireKernel::_Implementation::tabulateOptions()


void  
VampireKernel::_Implementation::tabulate()
{
  CALL("tabulate()");
  if (!_tabulation) return;

  tabulateOptions();
  _signature.tabulate(_tabulation);
  tabulateStatistics();
    
  if ((_terminationReason == VampireKernel::RefutationFound) && 
      _options.proof)
    {
      // _proofTree is supposed to be collected here,
      // see reportTermination()
      _proofTree.resetIter(); 
      for (const VK::Clause* cl = _proofTree.getNextClause(); 
	   cl; 
	   cl = _proofTree.getNextClause())
	{
	  _tabulation->proofClause(cl);
	};
    }; 
}; // void VampireKernel::_Implementation::tabulate()


void  
VampireKernel::_Implementation::switchBuiltinEqualityOff()
{
  CALL("switchBuiltinEqualityOff()");
  if  (_options.paramodulation 
       || _options.run_time_non_basicness
       || _options.input_non_basicness
       || _options.run_time_non_basicness
       || _options.inherited_negative_selection
       || _options.forward_demodulation
       || _options.fd_on_splitting_branches
       || _options.backward_demodulation
       || _options.bd_on_splitting_branches
       || _options.simplify_by_eq_res
       || _options.main_selected_lit_demodulation
       || _options.side_selected_lit_demodulation)
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: No equality. All equality related features are switched off!\n";   
      _options.paramodulation = false;
      _options.prohibit_non_basic_superposition = false;
      _options.prohibit_non_basic_fd = false;
      _options.prohibit_non_basic_bd = false;
      _options.reduce_non_basic = false;
      _options.input_non_basicness = false;
      _options.run_time_non_basicness = false;
      _options.inherited_negative_selection = false;
      _options.forward_demodulation = false;
      _options.fd_on_splitting_branches = false;
      _options.backward_demodulation = false;
      _options.bd_on_splitting_branches = false;
      _options.simplify_by_eq_res = false;
      _options.main_selected_lit_demodulation = 0;
      _options.side_selected_lit_demodulation = 0;
    };
}; // void VampireKernel::_Implementation::switchBuiltinEqualityOff()

void  
VampireKernel::_Implementation::switchChainingForTransitiveOff()
{
  CALL("switchChainingForTransitiveOff()");
  if (_options.chaining_for_transitive)
    {
      if (_options.std_output)
	*_options.std_output << "% WARNING: No transitive predicates. Chaining is switched off!\n";  
      _options.chaining_for_transitive = false;
    };

}; // void VampireKernel::_Implementation::switchChainingForTransitiveOff()

void  
VampireKernel::_Implementation::analysePreprocessedInput()
{
  CALL("analysePreprocessedInput()");
  if (_options.input_non_basicness)
    {
      _signature.resetAllArgumentsAsNonBasic();
	  const ClauseQueue::Entry* entry;
      for (entry = _inputSupportClauses.begin();
	   entry;
	   entry = entry->next())
	{
	  for (const LiteralList::Element* el = entry->value()->LitList()->firstElem();
	       el;
	       el = el->next())
	    {
	      el->literal()->markBasicArgumentsInSignature();
	    };
	};
      for (entry = _inputQueryClauses.begin();
	   entry;
	   entry = entry->next())
	{
	  for (const LiteralList::Element* el = entry->value()->LitList()->firstElem();
	       el;
	       el = el->next())
	    {
	      el->literal()->markBasicArgumentsInSignature();
	    };
	};
      _signature.finaliseArgumentBasicness();
    };

}; // void VampireKernel::_Implementation::analysePreprocessedInput()

void  
VampireKernel::_Implementation::analyseOptions()
{
  CALL("analyseOptions()");
  switch (_options.selection)
    { 
    case 1006L:
    case 1007L:
    case 1008L:
    case 1009L:
    case -1009L:
      _options.inherited_negative_selection = false;
      if (_options.std_output)
	*_options.std_output << "% WARNING: bidirectional literal selection "
			     << _options.selection 
			     << " switches inherited_negative_selection off!\n";
    };

    
  if (_options.header_precedence_kinky)
    {
      _incompleteProcedure = true;
      // may be refined in the future 
    };
    
  if (_options.fs_set_mode || _options.bs_set_mode)
    {	
      _incompleteProcedure = true;
    };
    
  if (_options.input_non_basicness)
    {
      if (!_signature.nonBasicArgumentExists())
	{
	  if (_options.std_output)
	    *_options.std_output << "% WARNING: no non-basic arguments in the input.\n"
				 << "% input_non_basicness is switched off.\n";
	  _options.input_non_basicness = false;
	};
    };

  if (!_options.input_non_basicness)
    {
      if (_options.prohibit_non_basic_superposition || 
	  _options.prohibit_non_basic_bd)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: input_non_basicness is off.\n";
	      if (_options.prohibit_non_basic_superposition)
		*_options.std_output << "% prohibit_non_basic_superposition is switched off.\n";
	      if (_options.prohibit_non_basic_bd)
		*_options.std_output << "% prohibit_non_basic_bd is switched off.\n";
	    };
	  _options.prohibit_non_basic_superposition = false;
	  _options.prohibit_non_basic_bd = false;
	};
    };
  

  if (!_options.input_non_basicness && !_options.run_time_non_basicness)
    {
      if (_options.prohibit_non_basic_fd ||
	  _options.reduce_non_basic)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: no non-basic subterms can be identified.\n";
	      if (_options.prohibit_non_basic_fd)
		*_options.std_output << "% prohibit_non_basic_fd is switched off.\n";
	      if (_options.reduce_non_basic)
		*_options.std_output << "% reduce_non_basic is switched off.\n";
	    };
	  _options.prohibit_non_basic_fd = false;
	  _options.reduce_non_basic = false;
	};
    };

  if (_options.input_non_basicness)
    {
      if (!_options.prohibit_non_basic_superposition &&
	  !_options.prohibit_non_basic_fd &&
	  !_options.prohibit_non_basic_bd &&
	  !_options.reduce_non_basic)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: input_non_basicness is switched off since it cannot be used.\n";
	    };
	  _options.input_non_basicness = false;
	};
    };

  if (_options.run_time_non_basicness)
    {
      
      if (!_options.prohibit_non_basic_fd &&
	  !_options.reduce_non_basic)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: run_time_non_basicness is switched off since it cannot be used.\n";
	    };
	  _options.run_time_non_basicness = false;
	};
    };
  
  if (_options.reduce_non_basic)
    {
      if (!_options.prohibit_non_basic_fd || 
	  !_options.prohibit_non_basic_bd)
	{
	  if (_options.std_output)
	    {
	      *_options.std_output << "% WARNING: non-basic demodulation makes no sense.\n"
				   << "% since reduce_non_basic is on.\n";
	      if (!_options.prohibit_non_basic_fd)
		*_options.std_output << "% prohibit_non_basic_fd is switched on.\n";
	      if (!_options.prohibit_non_basic_bd)
		*_options.std_output << "% prohibit_non_basic_bd is switched on.\n";
	    };
	  _options.prohibit_non_basic_fd = true;
	  _options.prohibit_non_basic_bd = true;
	};
    };


  if (_options.prohibit_non_basic_superposition &&
      (!_options.prohibit_non_basic_fd ||
       !_options.prohibit_non_basic_bd))
    {	
      
      if (_options.std_output)
	{
	  *_options.std_output << "% WARNING: non-basic demodulation without non-basic superposition\n"
			       << "% causes incompleteness.\n";
	};
      _incompleteProcedure = true;
    };

  if (_options.forw_sup_no_dup_red ||
      _options.back_sup_no_dup_red)
    {
      _incompleteProcedure = true;
    };

}; // void VampireKernel::_Implementation::analyseOptions()


void  
VampireKernel::_Implementation::collectPreprocessingIterationStatistics()
{   
  _preproLastTimeSec = _timer.elapsedSeconds();
  _preproLastOccupiedByObjectsBytes = GlobAlloc::occupiedByObjects();
  _preproLastOccupiedByChunksBytes = GlobAlloc::occupiedByChunks();
  _preproLastTotalBufferDeficitBytes = GlobAlloc::totalBufferDeficit();
  _preproLastActiveBufferDeficitBytes = GlobAlloc::activeBufferDeficit();
  _preproLastSuspendedBufferDeficitBytes = GlobAlloc::suspendedBufferDeficit();
  _preproStatistics.absorb(_master.statistics());
  _master.resetStatistics();
  if (_preproStatistics.completenessViolated())
    _incompleteProcedure = true;
}; // void VampireKernel::_Implementation::collectPreprocessingIterationStatistics()



bool  
VampireKernel::_Implementation::isRelevantBackgroundFlag(const VK::ClauseBackground::Rule& fl)
{
  switch (fl)
    {
    case VK::ClauseBackground::Input:
    case VK::ClauseBackground::Prepro:
    case VK::ClauseBackground::BinRes:
    case VK::ClauseBackground::ForwTransChain:
    case VK::ClauseBackground::BackTransChain:
    case VK::ClauseBackground::ForwSup: 
    case VK::ClauseBackground::BackSup:
    case VK::ClauseBackground::EqRes:
    case VK::ClauseBackground::EqFact:
    case VK::ClauseBackground::ForwDemod:
    case VK::ClauseBackground::BackDemod: 
    case VK::ClauseBackground::EqResSimp:
    case VK::ClauseBackground::ForwSubsRes:
    case VK::ClauseBackground::BackSubsRes:  
    case VK::ClauseBackground::Split: 
    case VK::ClauseBackground::NegEqSplit:
    case VK::ClauseBackground::Reanimation:
    case VK::ClauseBackground::Name:
    case VK::ClauseBackground::SimplificationByBuiltInTheories:
    case VK::ClauseBackground::NameIntro:
    case VK::ClauseBackground::Folding:
    case VK::ClauseBackground::Clausification:
    case VK::ClauseBackground::BuiltInTheoryFact:
      return true;
    default:
      return false;
    };
}; // bool VampireKernel::_Implementation::isRelevantBackgroundFlag(const VK::ClauseBackground::Rule& fl)

   
VampireKernel::Clause::BackgroundFlag
VampireKernel::_Implementation:: convertBackgroundFlag(const VK::ClauseBackground::Rule& fl)
{
  switch (fl)
    {
	
    case VK::ClauseBackground::Input: 
      return VampireKernel::Clause::IsInputClause;
    case VK::ClauseBackground::Prepro: 
      return VampireKernel::Clause::IsObtainedDuringPreprocessing;
    case VK::ClauseBackground::BinRes: 
      return VampireKernel::Clause::BinaryResolutionUsed;
    case VK::ClauseBackground::ForwTransChain: 
      return VampireKernel::Clause::ForwardTransitiveChainingUsed;
    case VK::ClauseBackground::BackTransChain: 
      return VampireKernel::Clause::BackwardTransitiveChainingUsed;
    case VK::ClauseBackground::ForwSup: 
      return VampireKernel::Clause::ForwardSuperpositionUsed; 
    case VK::ClauseBackground::BackSup: 
      return VampireKernel::Clause::BackwardSuperpositionUsed;
    case VK::ClauseBackground::EqRes: 
      return VampireKernel::Clause::EqualityResolutionUsed;
    case VK::ClauseBackground::EqFact: 
      return VampireKernel::Clause::EqualityFactoringUsed;
    case VK::ClauseBackground::ForwDemod: 
      return VampireKernel::Clause::ForwardDemodulationUsed;
    case VK::ClauseBackground::BackDemod: 
      return VampireKernel::Clause::BackwardDemodulationUsed; 
    case VK::ClauseBackground::EqResSimp: 
      return VampireKernel::Clause::SimplificationByEqualityResolutionUsed;
    case VK::ClauseBackground::ForwSubsRes: 
      return VampireKernel::Clause::ForwardSubsumptionResolutionUsed;
    case VK::ClauseBackground::BackSubsRes: 
      return VampireKernel::Clause::BackwardSubsumptionResolutionUsed;  
    case VK::ClauseBackground::Split: 
      return VampireKernel::Clause::PropositionalSplittingUsed; 
    case VK::ClauseBackground::NegEqSplit: 
      return VampireKernel::Clause::NegativeEqualitySplittingUsed;
    case VK::ClauseBackground::Reanimation: 
      return VampireKernel::Clause::IsReanimatedPassive;
    case VK::ClauseBackground::Name: 
      return VampireKernel::Clause::IsPartOfNameIntroduction;
    case VK::ClauseBackground::SimplificationByBuiltInTheories: 
      return VampireKernel::Clause::SimplificationByBuiltInTheoriesUsed;
     case VK::ClauseBackground::NameIntro:
       return VampireKernel::Clause::IsNameIntro;
     case VK::ClauseBackground::Folding:
       return VampireKernel::Clause::FoldingUsed;
     case VK::ClauseBackground::Clausification:
       return VampireKernel::Clause::ClausificationUsed;
    case VK::ClauseBackground::BuiltInTheoryFact: 
      return VampireKernel::Clause::IsBuiltInTheoryFact;
	
    default:
      return VampireKernel::Clause::IsInputClause; // to avoid warnings
    };

}; // VampireKernel::Clause::BackgroundFlag VampireKernel::_Implementation::convertBackgroundFlag(const VK::ClauseBackground::Rule& fl)

   
VK::ClauseBackground::Rule
VampireKernel::_Implementation::convertBackgroundFlag(const VampireKernel::Clause::BackgroundFlag& fl)
{
  CALL("convertBackgroundFlag(const VampireKernel::Clause::BackgroundFlag& fl)");
  switch (fl)
    {
    case VampireKernel::Clause::IsInputClause:
      return VK::ClauseBackground::Input;
    case VampireKernel::Clause::IsObtainedDuringPreprocessing:
      return VK::ClauseBackground::Prepro;
    case VampireKernel::Clause::BinaryResolutionUsed:
      return VK::ClauseBackground::BinRes;
    case VampireKernel::Clause::ForwardTransitiveChainingUsed:
      return VK::ClauseBackground::ForwTransChain;
    case VampireKernel::Clause::BackwardTransitiveChainingUsed:
      return VK::ClauseBackground::BackTransChain;
    case VampireKernel::Clause::ForwardSuperpositionUsed:
      return VK::ClauseBackground::ForwSup;
    case VampireKernel::Clause::BackwardSuperpositionUsed:
      return VK::ClauseBackground::BackSup;
    case VampireKernel::Clause::EqualityResolutionUsed:
      return VK::ClauseBackground::EqRes;
    case VampireKernel::Clause::EqualityFactoringUsed:
      return VK::ClauseBackground::EqFact;
    case VampireKernel::Clause::ForwardDemodulationUsed:
      return VK::ClauseBackground::ForwDemod;
    case VampireKernel::Clause::BackwardDemodulationUsed:
      return VK::ClauseBackground::BackDemod;
    case VampireKernel::Clause::SimplificationByEqualityResolutionUsed:
      return VK::ClauseBackground::EqResSimp;
    case VampireKernel::Clause::ForwardSubsumptionResolutionUsed:
      return VK::ClauseBackground::ForwSubsRes;
    case VampireKernel::Clause::BackwardSubsumptionResolutionUsed:
      return VK::ClauseBackground::BackSubsRes;
    case VampireKernel::Clause::PropositionalSplittingUsed:
      return VK::ClauseBackground::Split; 
    case VampireKernel::Clause::NegativeEqualitySplittingUsed:
      return VK::ClauseBackground::NegEqSplit; 
    case VampireKernel::Clause::IsReanimatedPassive:
      return VK::ClauseBackground::Reanimation;
    case VampireKernel::Clause::IsPartOfNameIntroduction:
      return VK::ClauseBackground::Name;
    case VampireKernel::Clause::SimplificationByBuiltInTheoriesUsed:
      return VK::ClauseBackground::SimplificationByBuiltInTheories;
    case VampireKernel::Clause::IsNameIntro:
      return VK::ClauseBackground::NameIntro;
    case VampireKernel::Clause::FoldingUsed:
      return VK::ClauseBackground::Folding;
    case VampireKernel::Clause::ClausificationUsed:
      return VK::ClauseBackground::Clausification;
    case VampireKernel::Clause::IsBuiltInTheoryFact:
      return VK::ClauseBackground::BuiltInTheoryFact;
    default:
      return VK::ClauseBackground::Input; // to avoid warnings
    };
    
}; // VK::ClauseBackground::Rule VampireKernel::_Implementation::convertBackgroundFlag(const VampireKernel::Clause::BackgroundFlag& fl)



//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::_Implementation::ClauseSetIterator"
#endif    
#include "debugMacros.hpp"
//======================================================================

VampireKernel::_Implementation::ClauseSetIterator::~ClauseSetIterator()
{
  CALL("destructor ~ClauseSetIterator()");
};


bool VampireKernel::_Implementation::ClauseSetIterator::isEnd() const
{
  return _iter.isEnd();
}; 

void VampireKernel::_Implementation::ClauseSetIterator::next()
{
  _iter.next();
};

const VampireKernel::Clause* VampireKernel::_Implementation::ClauseSetIterator::clause() const
{
  return 
    static_cast<const VampireKernel::Clause*>(static_cast<const void*>(_iter.clause()));
};
    
//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::_Implementation::NameIntroReceiver"
#endif    
#include "debugMacros.hpp"
//======================================================================

void VampireKernel::_Implementation::NameIntroReceiver::receive(VK::Clause* const & cl)
{
  CALL("receive(VK::Clause* const & cl)");
  cl->SetNumber(_hostKernelSession->_clauseNumberGenerator.generate());
  _hostKernelSession->_nameIntroClauses.Insert(cl);
  if (_hostKernelSession->_options.std_output &&
      _hostKernelSession->_options.show_name_intro)
    *(_hostKernelSession->_options.std_output) << "% " << cl << "\n";
}; // void VampireKernel::_Implementation::NameIntroReceiver::receive(VK::Clause* const & cl)


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel"
#endif    
#include "debugMacros.hpp"
//======================================================================

VampireKernel::VampireKernel() 
{
  CALL("constructor VampireKernel()");
  _implementation = 0; 
};
VampireKernel::~VampireKernel() 
{
  CALL("destructor ~VampireKernel()");
  if (DestructionMode::isThorough())
    {
      if (_implementation)
	{
	  _Implementation::closeSession();

	  
	  _implementation->destroy();
	  _Implementation::deallocate(_implementation);	
	  
	  //delete _implementation;
	};
    };
};


void VampireKernel::openSession(const Options& opt) 
{
  CALL("openSession(const Options& opt)");
  ASSERT(!GlobAlloc::occupiedByObjects());
  GlobalStopFlag::reset();
  _Implementation::setHardMemoryLimit(((float)opt.memory_limit)*1024);
  _Implementation::setAllocationBufferSize(opt.allocation_buffer_size);
  //_implementation = new _Implementation(opt);
 
  _implementation = _Implementation::allocate();
  _implementation->init(opt);
  
  _implementation->openSession();
  _implementation->_wrapper = this;
}; // void VampireKernel::openSession(const Options& opt) 

void VampireKernel::closeSession()
{
  _Implementation::closeSession();
  _implementation = 0;
}; // void VampireKernel::closeSession()


float VampireKernel::versionNumber()
{
  return 7.00f;
};

string VampireKernel::versionDescription() 
{
  return string("Civatateo (CASC J2)");
};

void VampireKernel::beginInputSignature() 
{
  _implementation->beginInputSignature();
};  
void VampireKernel::endInputSignature() 
{
  _implementation->endInputSignature();
};


void VampireKernel::registerInputFunction(string name,
					  ulong arity,
					  ulong& id)
{
  _implementation->registerInputFunction(name.c_str(),arity,id);
}; 
  
  

void VampireKernel::registerInputPredicate(string name,
					   ulong arity,
					   bool isAnswer,
					   ulong& id)
{
  _implementation->registerInputPredicate(name.c_str(),arity,isAnswer,id);
};  

void VampireKernel::setSymbolWeightConstPart(ulong symId,const VampireKernel::TermWeightType& weight) 
{
  _implementation->setSymbolWeightConstPart(symId,weight);
};


void VampireKernel::setSymbolPrecedence(ulong symId,long prec) 
{
  _implementation->setSymbolPrecedence(symId,prec);
}; 

void VampireKernel::setHeaderPrecedence(bool positive,ulong symId,long prec)
{
  _implementation->setHeaderPrecedence(positive,symId,prec);
}; 


void VampireKernel::interpretFunction(string inputSymbol,
				      ulong arity,
				      string theory,
				      string nativeSymbol)
{
  _implementation->interpret(false,inputSymbol.c_str(),arity,
			     theory.c_str(),nativeSymbol.c_str());
};

void VampireKernel::interpretPredicate(string inputSymbol,
				       ulong arity,
				       string theory,
				       string nativeSymbol)
{
  _implementation->interpret(true,inputSymbol.c_str(),arity,
			     theory.c_str(),nativeSymbol.c_str());
};


  
void VampireKernel::beginInputClauses() 
{
  _implementation->beginInputClauses();
};

void VampireKernel::endInputClauses() 
{
  _implementation->endInputClauses();
};
  
void VampireKernel::beginInputClause(ulong clauseNumber,
				     void* origin,
				     bool goal,
				     bool support) 
{
  ClauseAnnotation annotation;
  if (goal) annotation.markAsGoal();
  if (support) annotation.markAsSupport();
  _implementation->beginInputClause(clauseNumber,origin,annotation); 
};


void VampireKernel::beginInputClause(ulong clauseNumber,
				     void* origin,
				     const ClauseAnnotation& annotation)
{
  _implementation->beginInputClause(clauseNumber,origin,annotation); 
};

void VampireKernel::endInputClause() 
{
  _implementation->endInputClause();
};
  

void VampireKernel::beginInputLiteralDefinition(ulong clauseNumber,
						void* origin,
						bool goal,
						bool support)
{
  _implementation->beginInputLiteralDefinition(clauseNumber,
					       origin,
					       goal,
					       support);
}; 
  
void VampireKernel::endInputLiteralDefinition()
{
  _implementation->endInputLiteralDefinition();
}; 

void VampireKernel::beginInputLiteralDefinitionHead(bool positive,ulong predicateId)
{
  _implementation->beginInputLiteralDefinitionHead(positive,predicateId);
}; 
void VampireKernel::endInputLiteralDefinitionHead()
{
  _implementation->endInputLiteralDefinitionHead();
}; 

void VampireKernel::beginInputLiteralDefinitionTail()
{
  _implementation->beginInputLiteralDefinitionTail();
}; 
void VampireKernel::endInputLiteralDefinitionTail()
{
  _implementation->endInputLiteralDefinitionTail();
}; 

void VampireKernel::beginInputLiteralDefinitionTailComponent()
{
  _implementation->beginInputLiteralDefinitionTailComponent();
}; 
void VampireKernel::endInputLiteralDefinitionTailComponent()
{
  _implementation->endInputLiteralDefinitionTailComponent();
}; 


void VampireKernel::beginLiteral(bool positive,ulong predicateId) 
{
  LiteralAnnotation annotation;
  _implementation->beginLiteral(positive,predicateId,annotation);
};

void VampireKernel::beginLiteral(bool positive,ulong predicateId,bool preselected) 
{
  LiteralAnnotation annotation;
  if (preselected)
    {
      LiteralAnnotation::SelectabilityInheritanceLimit inhLim;
      inhLim.makeInfinite();
      annotation.makeHighSelectability(static_cast<long>(LiteralAnnotation::MaximalSelectabilityPriority),
				       inhLim,
				       LiteralAnnotation::DefaultDeploymentMode,
				       LiteralAnnotation::DefaultEqualityMode);
    };
  _implementation->beginLiteral(positive,predicateId,annotation);
};

void VampireKernel::beginLiteral(bool positive,
				 ulong predicateId,
				 const LiteralAnnotation& annotation)
{
  _implementation->beginLiteral(positive,predicateId,annotation);
};

void VampireKernel::endLiteral() 
{
  _implementation->endLiteral();
};
  
void VampireKernel::receiveFunction(ulong symId) 
{
  _implementation->receiveFunction(symId);
};

void VampireKernel::receiveNumericConstant(double value)
{
  _implementation->receiveNumericConstant(value);
};

void VampireKernel::receiveVariable(ulong varNum) 
{
  _implementation->receiveVariable(varNum);
};
    
void VampireKernel::saturate(ulong nextVacantClauseId,long timeLimit) 
{
  _implementation->saturate(nextVacantClauseId,timeLimit);
};


void VampireKernel::lookForAnotherAnswer()
{
  _implementation->lookForAnotherAnswer();
};


VampireKernel* VampireKernel::currentSession()
{
  if (_Implementation::_currentSession)
    {
      return _Implementation::_currentSession->_wrapper;
    }
  else
    return 0;
};


void VampireKernel::suspend()
{
  _implementation->suspend();
};

void VampireKernel::resume()
{
  _implementation->resume();
}; 

void VampireKernel::tabulate()
{
  CALL("tabulate()");
  _implementation->tabulate();
};


void VampireKernel::reportInterrupt(int signalNumber)
{
  if (_implementation)
    {
      _implementation->reportInterrupt(signalNumber);
    };
};
  
void VampireKernel::reportErrorSignal(int signalNumber)
{
  if (_implementation)
    {
      _implementation->reportErrorSignal(signalNumber);
    };
};

void VampireKernel::reportError()
{
  if (_implementation)
    {
      _implementation->reportError();
    };
};

// Communicating search results

VampireKernel::TerminationReason VampireKernel::terminationReason() const
{
  return _implementation->terminationReason();
};
const VampireKernel::Clause* VampireKernel::refutation() const 
{
  return _implementation->refutation();
};

VampireKernel::ClauseSetIterator* VampireKernel::activeClauses()
{
  return _implementation->activeClauses();
};

void VampireKernel::disposeOf(VampireKernel::ClauseSetIterator* itr)
{
  _implementation->disposeOf(itr);
};

bool VampireKernel::isNumericConstantId(ulong symbolId) const
{
  return _implementation->isNumericConstantId(symbolId);
}; 
bool VampireKernel::isInputSymbol(ulong symbolId) const
{
  return _implementation->isInputSymbol(symbolId);
}; 

bool VampireKernel::isAnswerPredicateId(ulong symbolId) const
{
  return _implementation->isAnswerPredicateId(symbolId);
};

ulong VampireKernel::symbolArity(ulong symbolId) const
{
  return _implementation->symbolArity(symbolId);
};

string VampireKernel::symbolPrintName(ulong symbolId) const
{
  return string(_implementation->symbolPrintName(symbolId));
}; 
double VampireKernel::numericConstantValue(ulong symbolId) const
{
  return _implementation->numericConstantValue(symbolId);
}; 



// Getting the session statistics
 
const VampireKernel::Statistics& VampireKernel::statistics()
{
  return _implementation->statistics();
};
  
void VampireKernel::outputPreproStat(ostream* str,bool full)
{
  if (!str) return;
  _implementation->outputPreproStat(str,full);
}; // void VampireKernel::outputPreproStat(ostream* str,bool full)

void VampireKernel::outputCurrentLoopStat(ostream* str,bool full)
{
  if (!str) return;
  _implementation->outputCurrentLoopStat(str,full);
}; // void VampireKernel::outputCurrentLoopStat(ostream* str,bool full)

void VampireKernel::outputCumulativeStat(ostream* str,bool full)
{
  if (!str) return;
  _implementation->outputCumulativeStat(str,full);
}; // void VampireKernel::outputCumulativeStat(ostream* str,bool full)


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Options"
#endif    
#include "debugMacros.hpp"
//======================================================================

void VampireKernel::Options::setDefaultValues()
{
  // session attributes
  job_id = "default_job_id";
  
  // main algorithm
    
  main_alg = 0L;

  // inference rules
  no_inferences_between_rules = false;
  static_splitting = true;
  dynamic_splitting = false;
  splitting_style = 1L;
  small_basic_comp_in_hyp_splitting = false;
  static_neg_eq_splitting = 3L;
  stat_neg_eq_spl_var_count_limit = 0L;
  dynamic_neg_eq_splitting = 0L;
  dyn_neg_eq_spl_var_count_limit = 0L;
  paramodulation = true;
  paramodulate_from_variables = true;
  sim_back_sup = 0;
  forw_sup_no_dup_red = false;
  back_sup_no_dup_red = false;
  prohibit_non_basic_superposition = false;
  prohibit_non_basic_fd = true;
  prohibit_non_basic_bd = true;
  reduce_non_basic = false;
  input_non_basicness = false;
  run_time_non_basicness = false;
  selection = 4L;
  literal_comparison_mode = 0;
  inherited_negative_selection = false;
     
  // strategy 
  forward_subsumption = true;
  fs_set_mode = false;
  fs_use_sig_filters = false;
  simplify_by_fs = 1L;
  main_premise_subsumption = 0L;
  side_premise_subsumption = 0L;
  main_selected_lit_demodulation = 0L;
  side_selected_lit_demodulation = 0L;
  forward_demodulation = 4L;
  fd_on_splitting_branches = true;
  normalize_by_commutativity = true;
  eq_taut_mod_C = false;
  prop_taut_mod_C = false;
  trivial_eq_res_mod_C = false;
  identical_lit_factoring_mod_C = false;
  chaining_for_transitive = false;
  chaining_through_variables = true;
  backward_demodulation = 4L;
  bd_on_splitting_branches = true;
  simplify_by_eq_res = true;
  backward_subsumption = true;
  bs_set_mode = false;
  orphan_murder = false;
  lrs = 1L;
  elim_def = 5L;
  simplification_ordering = NonrecursiveKBOrdering;
  symbol_precedence_by_arity = 1L;
  header_precedence_kinky = false;
  term_weighting_scheme = UniformTermWeighting;
  symbol_weight_by_arity = 0L;
  orient_pos_equalities_in_support = true;
  orient_neg_equalities_in_support = true;
  all_pos_eq_mode = LiteralAnnotation::DefaultEqualityMode;
  all_neg_eq_mode = LiteralAnnotation::DefaultEqualityMode;

  // numeric options 
  memory_limit = LONG_MAX; // kilobytes
  allocation_buffer_size = 10L;
  allocation_buffer_price = 2.0;
  max_skolem = 1024; 
  max_active = LONG_MAX;
  max_passive = LONG_MAX;
  max_weight = LONG_MAX;
  max_inference_depth = LONG_MAX;
  age_weight_ratio_numerator = 1L;
  age_weight_ratio_denominator = 5L;
  best_good_weight_ratio_numerator = 1;
  best_good_weight_ratio_denominator = 0;
  selected_pos_eq_penalty_coefficient = 1.0;  
  nongoal_penalty_coefficient = 1.0;
  lrs_softness = 1.0;
  first_time_check = 5L;   
  first_mem_check = 200L;   
    
  // interface options    
  show_opt = false;
  show_input = false;
  show_prepro = false;
  show_gen = false;
  show_passive = false;
  show_active = false;
  show_name_intro = false;
  show_lit_prop = false; 
  show_passive_profile_interval = LONG_MAX;
  proof = true;
  auto_stat_prepro = false;
  auto_stat_curr_loop = false;
  auto_stat_cumulative = true;
  auto_stat_full = false;
  auto_stat_summary = true;
  show_profile = false;
  show_weight_limit_changes = false;  
  tabulation = 0;
  std_output = &cout;

#ifdef VKERNEL_FOR_STEP_RESOLUTION
  step_resolution_literal_selection = false;
#endif

}; // void VampireKernel::Options::setDefaultValues()


ostream& VampireKernel::Options::output(ostream& str)
{
  str << "% job_id = " << job_id << '\n';

  str << "% main_alg = " << main_alg << '\n';
  
  str << "% no_inferences_between_rules = " << no_inferences_between_rules << '\n';
  str << "% static_splitting = " << static_splitting << '\n';
  str << "% dynamic_splitting = " << dynamic_splitting << '\n';
  str << "% splitting_style = " << splitting_style << '\n'; 
  str << "% small_basic_comp_in_hyp_splitting = " << small_basic_comp_in_hyp_splitting << '\n'; 
  str << "% static_neg_eq_splitting = " << static_neg_eq_splitting << '\n';
  str << "% stat_neg_eq_spl_var_count_limit = " << stat_neg_eq_spl_var_count_limit << '\n';
  str << "% dynamic_neg_eq_splitting = " << dynamic_neg_eq_splitting << '\n';
  str << "% dyn_neg_eq_spl_var_count_limit = " << dyn_neg_eq_spl_var_count_limit << '\n';
  str << "% paramodulation = " << paramodulation << '\n'; 
  str << "% paramodulate_from_variables = " << paramodulate_from_variables << '\n'; 
  str << "% sim_back_sup = " << sim_back_sup << '\n';  
  str << "% forw_sup_no_dup_red = " << forw_sup_no_dup_red << '\n';   
  str << "% back_sup_no_dup_red = " << back_sup_no_dup_red << '\n';   
  str << "%  = prohibit_non_basic_superposition " << prohibit_non_basic_superposition << '\n'; 
  str << "%  = prohibit_non_basic_fd" << prohibit_non_basic_fd << '\n'; 
  str << "%  = prohibit_non_basic_bd" << prohibit_non_basic_bd << '\n'; 
  str << "% reduce_non_basic = " << reduce_non_basic << '\n';  
  str << "% input_non_basicness = " << input_non_basicness << '\n'; 
  str << "% run_time_non_basicness = " << run_time_non_basicness << '\n';   
  str << "% selection = " << selection << '\n';
  str << "% literal_comparison_mode = " << literal_comparison_mode << '\n';
  str << "% inherited_negative_selection = " << inherited_negative_selection << '\n'; 
     
  str << "% forward_subsumption = " << forward_subsumption << '\n'; 
  str << "% fs_set_mode = " << fs_set_mode << '\n'; 
  str << "% fs_use_sig_filters = " << fs_use_sig_filters << '\n';
  str << "% simplify_by_fs = " << simplify_by_fs << '\n';
  str << "% main_premise_subsumption = " << main_premise_subsumption << '\n';
  str << "% side_premise_subsumption = " << side_premise_subsumption << '\n';
  str << "% main_selected_lit_demodulation = " << main_selected_lit_demodulation << '\n';
  str << "% side_selected_lit_demodulation = " << side_selected_lit_demodulation << '\n';
  str << "% forward_demodulation = " << forward_demodulation << '\n';
  str << "% fd_on_splitting_branches = " << fd_on_splitting_branches << '\n';
  str << "% normalize_by_commutativity = " << normalize_by_commutativity << '\n';
  str << "% eq_taut_mod_C = " << eq_taut_mod_C << '\n';
  str << "% prop_taut_mod_C = " << prop_taut_mod_C << '\n';
  str << "% trivial_eq_res_mod_C = " << trivial_eq_res_mod_C << '\n';
  str << "% identical_lit_factoring_mod_C = " << identical_lit_factoring_mod_C << '\n';
  str << "% chaining_for_transitive = " << chaining_for_transitive << '\n';
  str << "% chaining_through_variables = " << chaining_through_variables << '\n';
  str << "% backward_demodulation = " << backward_demodulation << '\n';
  str << "% bd_on_splitting_branches = " << bd_on_splitting_branches << '\n';
  str << "% simplify_by_eq_res = " << simplify_by_eq_res << '\n'; 
  str << "% backward_subsumption = " << backward_subsumption << '\n'; 
  str << "% bs_set_mode = " << bs_set_mode << "\n";
  str << "% orphan_murder = " << orphan_murder << '\n'; 
  str << "% lrs = " << lrs << '\n';
  str << "% elim_def = " << elim_def << '\n';
  str << "% simplification_ordering = " << simplification_ordering << '\n';
  str << "% symbol_precedence_by_arity = " << symbol_precedence_by_arity << '\n';
  str << "% header_precedence_kinky = " << header_precedence_kinky << '\n';
  str << "% term_weighting_scheme = " << term_weighting_scheme << '\n';
  str << "% symbol_weight_by_arity = " << symbol_weight_by_arity <<  '\n';
  str << "% orient_pos_equalities_in_support = " << orient_pos_equalities_in_support <<  '\n';
  str << "% orient_neg_equalities_in_support = " << orient_neg_equalities_in_support <<  '\n';

  switch (all_pos_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      str << "% all_pos_eq_mode = add_res\n";
      break;
    case LiteralAnnotation::ResolutionOnly:
      str << "% all_pos_eq_mode = res_only\n";
      break;
    default:      
      str << "% all_pos_eq_mode = default\n";
    };

  switch (all_neg_eq_mode)
    {
    case LiteralAnnotation::AddResolution:
      str << "% all_neg_eq_mode = add_res\n";
      break;
    default:      
      str << "% all_neg_eq_mode = default\n";
    };

  str << "% memory_limit = " << memory_limit << '\n';
  str << "% allocation_buffer_size = " << allocation_buffer_size << '\n';
  str << "% allocation_buffer_price = " << allocation_buffer_price << '\n';
  str << "% max_skolem = " << max_skolem << '\n';
  str << "% max_active = " << max_active << '\n'; 
  str << "% max_passive = " << max_passive << '\n';  
  str << "% max_weight = " << max_weight << '\n';    
  str << "% max_inference_depth = " << max_inference_depth << '\n';  
  str << "% age_weight_ratio_numerator = " << age_weight_ratio_numerator << '\n';   
  str << "% age_weight_ratio_denominator = " << age_weight_ratio_denominator << '\n'; 
  str << "% best_good_weight_ratio_numerator = " << best_good_weight_ratio_numerator << '\n';
  str << "% best_good_weight_ratio_denominator = " << best_good_weight_ratio_denominator << '\n'; 
  str << "% selected_pos_eq_penalty_coefficient = " 
      << selected_pos_eq_penalty_coefficient << '\n';  
  str << "% nongoal_penalty_coefficient = "
      << nongoal_penalty_coefficient << '\n';
  str << "% lrs_softness = " << lrs_softness << '\n';  
  str << "% first_time_check = " << first_time_check << '\n';   
  str << "% first_mem_check = " << first_mem_check << '\n';  
    
  str << "% tabulation = " << (tabulation != 0) << '\n';
  str.flush();
  return str;
}; // ostream& VampireKernel::Options::output(ostream& str)

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Clause"
#endif    
#include "debugMacros.hpp"
//======================================================================

ulong VampireKernel::Clause::number() const
{
  CALL("number() const");
  return (static_cast<const VK::Clause*>(static_cast<const void*>(this)))->Number();
};

bool VampireKernel::Clause::isLiteralDefinition(const Literal*& definedLiteral,
					       const Literal*& bodyLiterals)
{
  CALL("isLiteralDefinition(const Literal*& definedLiteral,const Literal*& bodyLiterals)");
  if ((static_cast<const VK::Clause*>
       (static_cast<const void*>(this)))->isNameIntro())
    {
      definedLiteral = static_cast<const VampireKernel::Literal*>
	(static_cast<const void*>
	 ((static_cast<const VK::Clause*>
	   (static_cast<const void*>(this)))->LitList()->firstElem()));
      bodyLiterals = definedLiteral->next();
      return true;
    }
  else
    return false;
};


const VampireKernel::Literal* VampireKernel::Clause::literals() const
{
  CALL("literals() const");
  ASSERT(!isBuiltInTheoryFact());
  ASSERT(!(static_cast<const VK::Clause*>
	   (static_cast<const void*>(this)))->isNameIntro());
  return 
    static_cast<const VampireKernel::Literal*>
    (static_cast<const void*>
     ((static_cast<const VK::Clause*>
       (static_cast<const void*>(this)))->LitList()->firstElem()));
};


ulong VampireKernel::Clause::numberOfSelectedLiterals() const
{
  CALL("numberOfSelectedLiterals() const");
  ASSERT(!(static_cast<const VK::Clause*>
	   (static_cast<const void*>(this)))->isNameIntro());
  return 
    (static_cast<const VK::Clause*>
     (static_cast<const void*>(this)))->LitList()->numOfSelectedLits();
};

bool VampireKernel::Clause::isInputClause() const
{
  CALL("isInputClause() const");  
  ASSERT(!isBuiltInTheoryFact());
  return (static_cast<const VK::Clause*>
	  (static_cast<const void*>(this)))->isInputClause();
};

bool VampireKernel::Clause::isBuiltInTheoryFact() const
{
  CALL("isBuiltInTheoryFact() const"); 
  return (static_cast<const VK::Clause*>
	  (static_cast<const void*>(this)))->isBuiltInTheoryFact();
};

bool VampireKernel::Clause::isNameIntroduction() const
{
  CALL("isNameIntroduction() const");
  return (static_cast<const VK::Clause*>
	  (static_cast<const void*>(this)))->isNameIntro();
};

void* VampireKernel::Clause::inputClauseOrigin() const
{
  CALL("inputClauseOrigin() const");
  ASSERT(!isBuiltInTheoryFact());
  ASSERT(isInputClause());
  return (static_cast<const VK::Clause*>
	  (static_cast<const void*>(this)))->inputClauseOrigin();
};

const VampireKernel::Clause::Ancestor* VampireKernel::Clause::firstAncestor() const
{
  CALL("firstAncestor() const");  
  ASSERT(!isBuiltInTheoryFact());
  if (isInputClause()) return 0;
  return 
    static_cast<const VampireKernel::Clause::Ancestor*>
    (static_cast<const void*>
     ((((static_cast<const VK::Clause*>
	 (static_cast<const void*>(this)))->background().ancestors()))));
}; 

VampireKernel::Clause::BackgroundFlagIterator 
VampireKernel::Clause::backgroundFlagIterator() const
{
  CALL("backgroundFlagIterator() const");
  BackgroundFlagIterator res;
  res._base = static_cast<const void*>(this);
  ulong fl = 0UL;
  while ((fl < 32UL) && 
	 (!(VampireKernel::_Implementation::isRelevantBackgroundFlag((VK::ClauseBackground::Rule)fl) &&
	    (static_cast<const VK::Clause*>
	     (static_cast<const void*>(this)))->background().properties().bit(fl))))
    {
      ++fl;
    };
    
  res._offset = fl;
  return res;
}; // VampireKernel::Clause::BackgroundFlagIterator VampireKernel::Clause::backgroundFlagIterator() const

bool VampireKernel::Clause::backgroundFlagIsSet(BackgroundFlag fl) const
{
  CALL("backgroundFlagIsSet(BackgroundFlag fl) const");
  ulong nativeFlagNum = 
    (ulong)VampireKernel::_Implementation::convertBackgroundFlag(fl);
  return 
    (static_cast<const VK::Clause*>
     (static_cast<const void*>(this)))->background().properties().bit((ulong)nativeFlagNum);
}; // bool VampireKernel::Clause::backgroundFlagIsSet(BackgroundFlag fl) const


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Clause::Ancestor"
#endif    
#include "debugMacros.hpp"
//======================================================================

const VampireKernel::Clause* VampireKernel::Clause::Ancestor::clause() const
{
  CALL("clause() const");
  return 
    static_cast<const VampireKernel::Clause*>
    (static_cast<const void*>
     ((static_cast<const VK::ClauseBackground::ClauseList*>
       (static_cast<const void*>(this)))->hd()));
}; 

const VampireKernel::Clause::Ancestor* VampireKernel::Clause::Ancestor::next() const
{
  CALL("next() const");  
  return 
    static_cast<const VampireKernel::Clause::Ancestor*>
    (static_cast<const void*>
     ((static_cast<const VK::ClauseBackground::ClauseList*>
       (static_cast<const void*>(this)))->tl()));
};




//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Clause::BackgroundFlagIterator"
#endif    
#include "debugMacros.hpp"
//======================================================================


VampireKernel::Clause::BackgroundFlagIterator::BackgroundFlagIterator()
{
  _base = 0;
};

VampireKernel::Clause::BackgroundFlagIterator::~BackgroundFlagIterator()
{
};

bool VampireKernel::Clause::BackgroundFlagIterator::getCurrentFlag(BackgroundFlag& flag) const
{
  CALL("getCurrentFlag(BackgroundFlag& flag) const");
  if (_base && (_offset < 32UL)) 
    {
      flag = 
	VampireKernel::_Implementation::convertBackgroundFlag((const VK::ClauseBackground::Rule&)_offset);
      return true;
    }
  else
    return false;
};

void VampireKernel::Clause::BackgroundFlagIterator::goToNextFlag()
{
  CALL("goToNextFlag()");
  ++_offset;
  while ((_offset < 32UL) && 
	 (!(VampireKernel::_Implementation::isRelevantBackgroundFlag((VK::ClauseBackground::Rule)_offset) &&
	    (static_cast<const VK::Clause*>(_base))->background().properties().bit(_offset))))
      
    {
      ++_offset;
    };
};

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Literal"
#endif    
#include "debugMacros.hpp"
//======================================================================


const VampireKernel::Literal* 
VampireKernel::Literal::next() const
{
  CALL("next() const");
  return
    static_cast<const VampireKernel::Literal*>
    (static_cast<const void*>    
     ((static_cast<const VK::LiteralList::Element*>
       (static_cast<const void*>(this)))->next()));
}; 

bool VampireKernel::Literal::isPositive() const 
{
  CALL("isPositive() const");
  return 
    (static_cast<const VK::LiteralList::Element*>(static_cast<const void*>(this)))->positive();
};


ulong VampireKernel::Literal::predicate() const
{
  CALL("predicate() const");
  ulong predId = 
    (static_cast<const VK::LiteralList::Element*>
     (static_cast<const void*>(this)))->literal()->functor();
  if (predId == VampireKernelConst::OrdEqNum) predId = VampireKernelConst::UnordEqNum;
  return predId;
};

const VampireKernel::Subterm* VampireKernel::Literal::firstArgument() const
{
  CALL("firstArgument() const");
  const TERM* args = 
    ((static_cast<const VK::LiteralList::Element*>
      (static_cast<const void*>(this)))->literal()->Args());
  if (args->IsNull()) return 0;
  return 
    static_cast<const VampireKernel::Subterm*>(static_cast<const void*>(args));
};


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::Subterm"
#endif    
#include "debugMacros.hpp"
//======================================================================

const VampireKernel::Subterm* VampireKernel::Subterm::next() const
{
  CALL("next() const");
  const TERM* nxt = (static_cast<const TERM*>(static_cast<const void*>(this)))->Next();
  if (nxt->IsNull()) return 0;
  return 
    static_cast<const VampireKernel::Subterm*>(static_cast<const void*>(nxt));
};

bool VampireKernel::Subterm::isVariable() const
{
  CALL("isVariable() const");
  return (static_cast<const TERM*>(static_cast<const void*>(this)))->isVariable();
};

ulong VampireKernel::Subterm::variableNumber() const
{
  CALL("variableNumber() const");
  ASSERT(isVariable());
  return (static_cast<const TERM*>(static_cast<const void*>(this)))->var();
};

ulong VampireKernel::Subterm::topFunction() const
{
  CALL("topFunction() const");
  ASSERT(!isVariable());
  return (static_cast<const TERM*>(static_cast<const void*>(this)))->First()->functor();
};

const VampireKernel::Subterm* VampireKernel::Subterm::firstArgument() const
{
  CALL("firstArgument() const");
  ASSERT(!isVariable());
  const TERM* args = (static_cast<const TERM*>(static_cast<const void*>(this)))->First()->Args();  
  if (args->IsNull()) return 0;
  return 
    static_cast<const VampireKernel::Subterm*>(static_cast<const void*>(args));
};


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::ClauseSetIterator"
#endif    
#include "debugMacros.hpp"
//======================================================================

VampireKernel::ClauseSetIterator::~ClauseSetIterator()
{
  CALL("destructor ~ClauseSetIterator()");
};


//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::LiteralAnnotation"
#endif    
#include "debugMacros.hpp"
//======================================================================

VampireKernel::LiteralAnnotation::LiteralAnnotation() 
{
  makeDefault();
};

void VampireKernel::LiteralAnnotation::makeDefault()
{
  _selectabilityCategory = DefaultSelectability;
  _selectabilityPriority = 0L;
  _selectabilityInheritanceLimit.makeInfinite();
  _deploymentMode = DefaultDeploymentMode;
  _equalityMode = DefaultEqualityMode;
};

void 
VampireKernel::LiteralAnnotation::makeDefaultSelectability(long selectabilityPriority,
							   const DeploymentMode& deploymentMode,
							   const EqualityMode& equalityMode)
{
  _selectabilityCategory = DefaultSelectability;
  _selectabilityPriority = selectabilityPriority;
  if (_selectabilityPriority > MaximalSelectabilityPriority)
    {
      _selectabilityPriority = MaximalSelectabilityPriority;
    }
  else if (_selectabilityPriority < MinimalSelectabilityPriority)
    {
      _selectabilityPriority = MinimalSelectabilityPriority;
    };
  _selectabilityInheritanceLimit.makeInfinite();
  _deploymentMode = deploymentMode;
  _equalityMode = equalityMode;
};

void 
VampireKernel::LiteralAnnotation::makeHighSelectability(long selectabilityPriority,
							const SelectabilityInheritanceLimit& selectabilityInheritanceLimit,
							const DeploymentMode& deploymentMode,
							const EqualityMode& equalityMode)
{
  _selectabilityCategory = HighSelectability;
  _selectabilityPriority = selectabilityPriority;
  if (_selectabilityPriority > MaximalSelectabilityPriority)
    {
      _selectabilityPriority = MaximalSelectabilityPriority;
    }
  else if (_selectabilityPriority < MinimalSelectabilityPriority)
    {
      _selectabilityPriority = MinimalSelectabilityPriority;
    };
  _selectabilityInheritanceLimit = selectabilityInheritanceLimit;
  _deploymentMode = deploymentMode;
  _equalityMode = equalityMode;
};

void 
VampireKernel::LiteralAnnotation::makeLowSelectability(long selectabilityPriority,
						       const SelectabilityInheritanceLimit& selectabilityInheritanceLimit,
						       const DeploymentMode& deploymentMode,
						       const EqualityMode& equalityMode)
{
  _selectabilityCategory = LowSelectability;
  _selectabilityPriority = selectabilityPriority;
  if (_selectabilityPriority > MaximalSelectabilityPriority)
    {
      _selectabilityPriority = MaximalSelectabilityPriority;
    }
  else if (_selectabilityPriority < MinimalSelectabilityPriority)
    {
      _selectabilityPriority = MinimalSelectabilityPriority;
    };
  _selectabilityInheritanceLimit = selectabilityInheritanceLimit;
  _deploymentMode = deploymentMode;
  _equalityMode = equalityMode;
};

const VampireKernel::LiteralAnnotation::SelectabilityCategory& 
VampireKernel::LiteralAnnotation::selectabilityCategory() const
{
  return _selectabilityCategory;
}; 

long VampireKernel::LiteralAnnotation::selectabilityPriority() const
{
  return _selectabilityPriority;
};

const VampireKernel::LiteralAnnotation::SelectabilityInheritanceLimit& 
VampireKernel::LiteralAnnotation::selectabilityInheritanceLimit() const
{
  return _selectabilityInheritanceLimit;
};

const VampireKernel::LiteralAnnotation::DeploymentMode& 
VampireKernel::LiteralAnnotation::deploymentMode() const
{
  return _deploymentMode;
};

const VampireKernel::LiteralAnnotation::EqualityMode& 
VampireKernel::LiteralAnnotation::equalityMode() const
{
  return _equalityMode;
};

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#define DEBUG_NAMESPACE "VampireKernel::LiteralAnnotation::SelectabilityInheritanceLimit"
#endif    
#include "debugMacros.hpp"
//======================================================================


VampireKernel::LiteralAnnotation::SelectabilityInheritanceLimit::SelectabilityInheritanceLimit() :
  _finite(false) 
{
};

void VampireKernel::LiteralAnnotation::SelectabilityInheritanceLimit::makeFinite(ulong lim) 
{
  _finite = true;
  _value = lim;
  if (_value > MaximalFinite) _value = MaximalFinite;
};

void VampireKernel::LiteralAnnotation::SelectabilityInheritanceLimit::makeInfinite()
{
  _finite = false;
};

bool VampireKernel::LiteralAnnotation::SelectabilityInheritanceLimit::isFinite(ulong& lim) const
{
  lim = _value;
  return _finite;
};

//======================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_VAMPIRE_KERNEL
#  define DEBUG_NAMESPACE "VampireKernel::ClauseAnnotation"
#endif    
#include "debugMacros.hpp"
//======================================================================

VampireKernel::ClauseAnnotation::ClauseAnnotation() 
{
  makeDefault();
};

void VampireKernel::ClauseAnnotation::makeDefault()
{
  _isGoal = false;
  _isSupport = false;
  _isNonSubsuming = false;
  _isNonSubsumable = false;
};

void VampireKernel::ClauseAnnotation::make(bool isGoal,bool isSupport,bool isNonSubsuming,bool isNonSubsumable)
{
  _isGoal = isGoal;
  _isSupport = isSupport;
  _isNonSubsuming = isNonSubsuming;
  _isNonSubsumable = isNonSubsumable;
}; 

bool VampireKernel::ClauseAnnotation::isGoal() const
{
  return _isGoal;
};

bool VampireKernel::ClauseAnnotation::isSupport() const
{
  return _isSupport;
};

bool VampireKernel::ClauseAnnotation::isNonSubsuming() const
{
  return _isNonSubsuming;
};

bool VampireKernel::ClauseAnnotation::isNonSubsumable() const
{
  return _isNonSubsumable;
};
  
void VampireKernel::ClauseAnnotation::markAsGoal()
{
  _isGoal = true;
};

void VampireKernel::ClauseAnnotation::markAsSupport()
{
  _isSupport = true;
};

void VampireKernel::ClauseAnnotation::markAsNonSubsuming()
{
  _isNonSubsuming = true;
};

void VampireKernel::ClauseAnnotation::markAsNonSubsumable()
{
  _isNonSubsumable = true;
};

//======================================================================







