#ifndef BINARY_RESOLUTION_H
//======================================================
#define BINARY_RESOLUTION_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "premise_factoring.hpp"
#include "or_machine.hpp"
#include "weight_formula.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
namespace VK
{
  class Clause;
  class ForwardProcessing;
  class Statistics;
}; // namespace VK
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_BINARY_RESOLUTION
#define DEBUG_NAMESPACE "BIN_ORD_RES"
#endif
#include "debugMacros.hpp"
//=================================================

namespace VK
{
  class BIN_ORD_RES
  {
  public:
    BIN_ORD_RES(OR_INDEX* resIndex);
    ~BIN_ORD_RES() { CALL("destructor ~BIN_ORD_RES()"); };
    void setStatistics(Statistics* stat)
    {
      _dummyClauseCollector.setStatistics(stat);
      _statistics = stat;
    };

    void setClauseCollector(TemporaryClauseCollector* cc) 
    {
      _clauseCollector = cc;
    };
  
    void setForwardProcessing(ForwardProcessing* fp)
    {
      _forwardProcessing = fp;
    };

    void setMainPremiseSubsumption(bool fl)
    {
      _mainPremiseSubsumption = fl;
    };
    void setSidePremiseSubsumption(bool fl)
    {
      _sidePremiseSubsumption = fl;
    };
    void setMainSelectedLitDemodulation(bool fl)
    {
      _mainSelectedLitDemodulation = fl;
    };
    void setSideSelectedLitDemodulation(bool fl)
    {
      _sideSelectedLitDemodulation = fl;
    };
    void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;
    };

    void setClauseSizeLimit(long lim)
    {
      _clauseSizeLimit = lim;
    };

    bool Load(Clause* c);
    bool NextInference();  

#ifndef NO_DEBUG 
    bool clean() const { return true; };
#endif

  private:
    enum Continuation { ContFirstLit, ContFirstFactor, ContNextFactor }; 

  private: 
    void ResetFactoringState();
    bool NextFactor();
    void CollectWeightForm(const Clause* c);
    Clause* FirstPremise() const { return premise1.TheClause(); };
    Clause* SecondPremise() const { return premise2.TheClause(); };
    inline bool CollectResult();
    bool subsumeMainPremise();
    bool subsumeSidePremise();
    bool demodulateMainSelectedLits();
    bool demodulateSideSelectedLits(bool includeMarked);
    static bool eligible(const Clause* cl); 

  private:
    OPEN_BR_DECL(1);
    Statistics* _statistics;
    OR_PREMISE_WITH_FACTORING premise1,premise2;
    ulong index1;
    ulong index2;
    ulong currentLit;
    bool _rotationPossible;
    OR_MACH ORMachine;
    WEIGHT_FORMULA weight_form;
    long _residualMainPremiseSize;
    long _clauseSizeLimit;
    TemporaryClauseCollector* _clauseCollector;
    ForwardProcessing* _forwardProcessing;
    TemporaryClause::Extras::Handle _dummyClause;
    TemporaryClauseCollector _dummyClauseCollector;
    Continuation _continuation;
    bool _noInfBetweenRulesFlag;
    bool _sideClauseCanNotBeRule;
    bool _mainPremiseSubsumption;
    bool _sidePremiseSubsumption;
    bool _mainSelectedLitDemodulation;
    bool _sideSelectedLitDemodulation;
    CLOSE_BR_DECL(1);
  }; // class BIN_ORD_RES 
  
 
}; // namespace VK


//======================================================
#endif
