#include "clause_set.hpp"
#include "Clause.inl"
using namespace BK;
using namespace VK;
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_SET
#define DEBUG_NAMESPACE "CLAUSE_SET"
#endif
#include "debugMacros.hpp"
//============================================================================
CLAUSE_SET::CLAUSE_SET(const Clause::MainSet& tag) :
  _tag(tag),
  numOfClauses(0L), 
  removed(0L),
  _mainIteratorIsEngaged(false)
#ifdef DEBUG_NAMESPACE
  , name("?")
#endif
{
  
};

CLAUSE_SET::CLAUSE_SET(const Clause::MainSet& tag,const char* nm) : 
  _tag(tag),
  numOfClauses(0L), 
  removed(0L),
  _mainIteratorIsEngaged(false)
#ifdef DEBUG_NAMESPACE
  , name(nm)
#endif
{
};

CLAUSE_SET::~CLAUSE_SET() 
{
};

#ifndef NO_DEBUG
bool CLAUSE_SET::clean() const
{
  if (clauses.Next() != &clauses) 
    {
      return false;
    };  
  if (clauses.Previous() != &clauses) 
    {
      return false;
    }; 
  if (numOfClauses)
    {
      return false;
    };
  if (removed) 
    {
      return false;
    };
  return true; 
};
#endif


void CLAUSE_SET::Insert(Clause* c) 
{ 
  c->setMainSet(_tag);
  numOfClauses++; 
  Clause::InsertAfter(c,&clauses); 
};

void CLAUSE_SET::Remove(Clause* c) 
{
  CALLM(name,"Remove(Clause* c)"); 
  numOfClauses--; 
  ASSERT(numOfClauses >= 0L);
  removed++;
  // take care about the main iterator
  if (c == current)
    {
      current = current->Next();
      if (current == (&clauses)) current = 0; 
    };
  Clause::Remove(c); 
  c->setMainSet(Clause::MainSetFree);
}; // void CLAUSE_SET::Remove(Clause* c) 

void CLAUSE_SET::receive(Clause* const & cl)
{ 
  CALLM(name,"receive(Clause* const & cl)");
  Insert(cl);
}; // void CLAUSE_SET::receive(Clause* const & cl)


Clause* CLAUSE_SET::Next() 
{
  CALLM(name,"Next()");
  Clause* res = current;
  if (current)
    {
      current = current->Next();
      if (current == (&clauses)) current = 0; 
    }
  else
    _mainIteratorIsEngaged = false;
  return res;
}; // Clause* CLAUSE_SET::Next() 

Clause* CLAUSE_SET::Top() 
{
  CALLM(name,"Top()");
  if (Empty()) return 0;     
  return clauses.Next();
};

Clause* CLAUSE_SET::Pop()
{
  CALLM(name,"Pop()");
  ASSERT(Nonempty());
  Clause* c = clauses.Next();
  Remove(c);
  return c;
}; // Clause* CLAUSE_SET::Pop()


void CLAUSE_SET::Purge()
{
  CALL("Purge()");
  while (!Empty())
    {
      Pop();
    };
  ASSERT(numOfClauses == 0UL);
  removed = 0L;
}; // void CLAUSE_SET::Purge()


void CLAUSE_SET::DestroyAllClauses()
{
  CALLM(name,"DestroyAllClauses()");
  while (!Empty())
    {
      Pop()->Destroy();
    };
  removed = 0L;
}; // void CLAUSE_SET::DestroyAllClauses()


#ifndef NO_DEBUG
bool CLAUSE_SET::contains(const Clause* cl) const
{
  for (const Clause* c = clauses.Next();
       c != &clauses;
       c = c->Next())
    if (c == cl) return true;
  return false;
};
#endif

   
#ifndef NO_DEBUG_VIS
ostream& CLAUSE_SET::outputClauses(ostream& str)
{
  resetMainIterator();
  for (Clause* c = Next(); c; c = Next()) str << c << '\n';
  return str;
};
#endif

//============================================================================


