#ifndef CLAUSE_SET_H
//===================================================================
#define CLAUSE_SET_H
#ifndef NO_DEBUG_VIS
#include <iostream>
#endif
#include "jargon.hpp"
#include "Receiver.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "Clause.hpp"
//============================================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_CLAUSE_SET
#define DEBUG_NAMESPACE "CLAUSE_SET"
#endif
#include "debugMacros.hpp"
//============================================================================
namespace VK
{

  class CLAUSE_SET : public BK::Receiver< Clause* >
  {
  public:
    class ConstIter
    {
    public:
      ConstIter(const CLAUSE_SET& set)
      {
	CALL("ConstIter::constructor ConstIter(const CLAUSE_SET& set)");
	_current = (set.Empty()) ? 0 : set.clauses.Next();
	_end = &(set.clauses);
      };
      bool isEnd() const { return _current == 0; };
      void next()
      {
	CALL("ConstIter::next()");
	if (_current)
	  {
	    _current = _current->Next();
	    if (_current == _end) _current = 0;
	  };
      };
      const Clause* clause() const { return _current; };
    private:
      const Clause* _current;
      const Clause* _end;
    }; // class ConstIter

    class Iter
    {
    public:
      Iter(CLAUSE_SET& set)
      {
	CALL("Iter::constructor Iter(CLAUSE_SET& set)");
	_current = (set.Empty()) ? 0 : set.clauses.Next();
	_end = &(set.clauses);
      };
      bool isEnd() const { return _current == 0; };
      void next()
      {
	CALL("Iter::next()");
	if (_current)
	  {
	    _current = _current->Next();
	    if (_current == _end) _current = 0;
	  };
      };
      Clause* clause() const { return _current; };
    private:
      Clause* _current;
      Clause* _end;
    }; // class ConstIter

  public:
    CLAUSE_SET(const Clause::MainSet& tag);
    CLAUSE_SET(const Clause::MainSet& tag,const char* nm);
    ~CLAUSE_SET();

    bool Empty() const { return (clauses.Next() == (&clauses)); };
    bool Nonempty() const { return (clauses.Next() != (&clauses)); }; 
#ifndef NO_DEBUG
    bool clean() const;
#endif

    void Insert(Clause* c);
    void Remove(Clause* c);
    void receive(Clause* const & cl);

    long Size() const { return numOfClauses; };
    long Removed() const { return removed; };
    long Total() const { return (Size() + Removed()); };

    bool mainIteratorIsEngaged() const 
    {
      return _mainIteratorIsEngaged;
    };

    void resetMainIterator() 
    {
      CALL("resetMainIterator()");
      ASSERT(!_mainIteratorIsEngaged);
      _mainIteratorIsEngaged = true;
      current = Empty() ? 0 : clauses.Next(); 
    };
    void releaseMainIterator()
    {
      _mainIteratorIsEngaged = false;
    };
    Clause* Next();

    Clause* Top();

    Clause* Pop();
   
    void Purge();
    
    void DestroyAllClauses();

#ifndef NO_DEBUG
    bool contains(const Clause* cl) const;
#endif

   
#ifndef NO_DEBUG_VIS
    ostream& outputClauses(ostream& str);
#endif

  private: 
    Clause::MainSet _tag;
    Clause clauses;
    Clause* current;
    long numOfClauses;
    long removed;
    bool _mainIteratorIsEngaged;
#ifdef DEBUG_NAMESPACE
    const char* name;
#endif
    friend class ConstIter;
    friend class Iter;
  }; // class CLAUSE_SET
}; // namespace VK

//===================================================================
#endif








 
