//
// File:         forward_superposition.hpp
// Description:  Representation of stored terms.
// Created:      Feb 14, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 04, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//               The array FPARA_PREMISE::_inherentlySelected
//               has been made extendable.
// Revised:      Dec 22, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               In FPARA_PREMISE it was implicitely assumed that 
//               <num of lits>*HellConst::MaxTermSize <= HellConst::MaxClauseSize
//               which is now wrong since HellConst::MaxClauseSize is
//               not defined via HellConst::MaxNumOfLiterals.
// Revised:      Apr 28, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::StoredConstraint 
//               replaced SharedKBConstr
// Revised:      May 09, 2002, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               SimplificationOrdering::CompiledConstraintForUnifiers
//               replaced ORD_CONSTR
// Revised:      Jul 05, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Major refurbishment.
// Revised:      Jul 21, 2003, Alexandre Riazanov, riazanov@cs.man.ac.uk
//               Bugfix in nextSelectedMainLitMemory()
//====================================================
#ifndef FORWARD_SUPERPOSITION_H
//============================================================
#define FORWARD_SUPERPOSITION_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "VampireKernelConst.hpp"
#include "Stack.hpp"
#include "para_machine.hpp"
#include "weight_formula.hpp"
#include "Array.hpp"
#include "ExpandingStack.hpp"
#include "GlobAlloc.hpp"
#include "SimplificationOrdering.hpp"
#include "TemporaryClause.hpp"
#include "TemporaryClauseCollector.hpp"
namespace VK
{
  class ForwardProcessing;
  class Statistics;
}; // namespace VK
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_FORWARD_SUPERPOSITION
#define DEBUG_NAMESPACE "FORWARD_SUPERPOSITION"
#endif
#include "debugMacros.hpp"
//=================================================

namespace VK
{

  class FORWARD_SUPERPOSITION
  {
  public:
    FORWARD_SUPERPOSITION(OR_INDEX* unifIndex);
    ~FORWARD_SUPERPOSITION();

    void setStatistics(Statistics* stat)
    {
      _dummyClauseCollector.setStatistics(stat);
      _statistics = stat;
    };

    void setClauseCollector(TemporaryClauseCollector* cc) 
    {
      _clauseCollector = cc;
    };
  
    void setForwardProcessing(ForwardProcessing* fp)
    {
      _forwardProcessing = fp;
    };

    void setNoDuplicateRedexes(bool fl)
    {
      _noDuplicateRedexesFlag = fl;
    };

    void setNoInfBetweenRulesFlag(bool fl)
    {
      _noInfBetweenRulesFlag = fl;   
    };
  
    void setUseInputNonBasicness(bool fl)
    {
      _useInputNonBasicness = fl;
    };

    void setMainPremiseSubsumption(bool fl)
    {
      _mainPremiseSubsumption = fl;
    };
    void setSidePremiseSubsumption(bool fl)
    {
      _sidePremiseSubsumption = fl;
    };
    
    void setMainSelectedLitDemodulation(bool fl)
    {
      _mainSelectedLitDemodulation = fl;
    };
    void setSideSelectedLitDemodulation(bool fl)
    {
      _sideSelectedLitDemodulation = fl;
    };
    
    void setOrientEqualitiesInSupport(bool orientPos,bool orientNeg)
    {
      _orientPosEqualitiesInSupport = orientPos;
      _orientNegEqualitiesInSupport = orientNeg;
    };
  

    void setClauseSizeLimit(long lim)
    {
      _clauseSizeLimit = lim;
    };



    bool Load(Clause* cl);
    bool NextInference();

  private: 
    enum Continuation { ContFSup, ContEqRes };
    enum ParamodulationMode { Init, ParaFromComplex, ParaFromVar };

  private:
    void compileMainPremiseResidualSizeFormula();
    bool NextFSupInference();
    bool NextEqRes();
    void UndoEqRes()
    {
      Unifier::current()->UnbindLastSection();
    }; 
    void resetIterationOverMainLits() 
    { 
      _nextMainSelectedLitNum = _numOfSelectedMainLits; 
    };
    PrefixSym* nextSelectedMainLit(LiteralProperties& prop)
    {
      if (_nextMainSelectedLitNum)
	{
	  _nextMainSelectedLitNum--;
	  prop = _mainLitProperties[_nextMainSelectedLitNum];
	  return _mainSelectedLiterals.nth(_nextMainSelectedLitNum);
	};
      return 0;
    }; 
    PrefixSym* nextSelectedMainLitMemory()
    {
      if (_nextMainSelectedLitNum)
	{
	  _nextMainSelectedLitNum--;
	  return _mainSelectedLiterals.nth(_nextMainSelectedLitNum);
	};
      return 0;
    }; 
    bool LoadLiteral(PrefixSym* lit,const LiteralProperties& prop);
    bool LoadRestOfLit();
    void loadEligibleSubtermPositions(PrefixSym* term);
				      

    bool LoadTermToRewrite(PrefixSym* t)
    {
      CALL("LoadTermToRewrite(PrefixSym* t)");
      // t must be a plain complex term with all brackets set
      _currentParamodulationTargetTerm = t;
      if (!_retrievalOfUnifiable.CompileComplexTerm(_currentParamodulationTargetTerm))
	return false;
      _paramodulationMode = Init;
      return true;
    };
    bool Paramodulate();
    RES_PAIR_LIST* NextCandidate(SimplificationOrdering::CompiledConstraintForUnifiers* constr);  
    bool CollectResult();
    bool CollectEqResResult();
    bool collectSidePremiseLiterals(const LiteralProperties::SLModifier& slmod); 
    bool collectMainPremiseLiterals(const LiteralProperties::SLModifier& slmod);
    bool collectRewrittenLiteral(const LiteralProperties::SLModifier& slmod);

    bool subsumeMainPremise();
    bool collectMainPremiseInstance();
    bool subsumeSidePremise();
    bool collectSidePremiseInstance();
    
    bool demodulateSelectedLits();
    bool collectMainPremiseSubtermsAvailableForParamodulationInto();
    bool collectSidePremiseSubtermsAvailableForParamodulationInto();

  private:
    OPEN_BR_DECL(1);
    // main parameters
    Statistics* _statistics;
    TemporaryClauseCollector* _clauseCollector;
    ForwardProcessing* _forwardProcessing;
    OR_INDEX* _unifIndex;
    bool _mainPremiseSubsumption;
    bool _sidePremiseSubsumption;
    bool _mainSelectedLitDemodulation;
    bool _sideSelectedLitDemodulation;
    bool _noDuplicateRedexesFlag;
    bool _noInfBetweenRulesFlag;
    bool _useInputNonBasicness;
    bool _orientPosEqualitiesInSupport;
    bool _orientNegEqualitiesInSupport;
    bool _orientPosEqualitiesInMainPremise;
    bool _orientNegEqualitiesInMainPremise;    
    long _clauseSizeLimit;


    // main premise data
    Clause* _mainPremise;
    ulong _mainPremiseVarIndex;
    LiteralList* _allMainPremiseLiterals;
    LiteralList::Element* _firstNonselectedMainLit;
    ulong _numOfSelectedMainLits;
    bool _equalityResolutionApplicable;
    PrefixSym _mainSelectedLiteralMemory[VampireKernelConst::MaxClauseSize];
    PrefixSym* _freeMainSelectedLitMemory;
    BK::ExpandingStack<BK::GlobAlloc,PrefixSym*,BK::FastRelocation<PrefixSym*>,32UL> _mainSelectedLiterals; 
    BK::Array<BK::GlobAlloc,LiteralProperties,BK::FastRelocation<LiteralProperties>,32UL> _mainLitProperties;    
    WEIGHT_FORMULA _mainPremiseResidualSizeFormula;
    long _mainPremiseResidualSize;
    bool _mainSelectedLitDemodulationAttempted;
    bool _mainSelectedLitDemodulated;

    // current selected literal in the main premise
    ulong _nextMainSelectedLitNum;
    PrefixSym* _currentMainLiteral;
    LiteralProperties _currentMainLitProperties;
    BK::Stack<PrefixSym*,VampireKernelConst::MaxTermSize> _eligibleSubtermPositions;
    SimplificationOrdering::CompiledConstraintForUnifiers* _compiledConstraint;
    SimplificationOrdering::CompiledConstraintForUnifiers* _currentCompiledConstraint;   
    PrefixSym* _eqArg1;
    PrefixSym* _eqArg2;
    bool _swapArguments;

    // side premise data
    Clause* _sidePremise;    
    ulong _sidePremiseVarIndex;
    SimplificationOrdering::ConstraintCheckOnUnifier* _storedConstraintCheck;

    // retrieval
    PARA_MACH _retrievalOfUnifiable;
    PrefixSym* _currentParamodulationTargetTerm;
    TERM _pseudoHeader;
    RES_PAIR_LIST* _currentCandidate; 
    bool _noRulesAsSidePremises;

    // control
    ParamodulationMode _paramodulationMode;
    Continuation _continuation;

    // misc
    TemporaryClause::Extras::Handle _dummyClause;
    TemporaryClauseCollector _dummyClauseCollector;
    CLOSE_BR_DECL(1);
  }; // class FORWARD_SUPERPOSITION
}; // namespace VK

//=================================================================
#endif
