#include "kept.hpp"
//====================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_KEPT
 #define DEBUG_NAMESPACE "KEPT"
#endif
#include "debugMacros.hpp"
//====================================================



using namespace VK;
Clause* KEPT::Next() 
{
 if (!(EmptyCat(0)))
  {
   Clause* res = _category[0].Next();
   Clause::Remove(res);
   StatUsed(0); 
   return res;
  };
  
 if (_ageWeightRatioDenominator > 0)
  {
   if (EmptyCat(_minNonemptyCatNum))
    {
     return 0; 
    };

   if ((_totalSelectedByAgeOrWeight % 
	(_ageWeightRatioNumerator + _ageWeightRatioDenominator)) <
       _ageWeightRatioDenominator)
    // pick by weight
    {
      // determing the interval of eligible good weights
      long concession = 0UL;
      long tmp = _totalSelectedByWeight;
      while ((tmp % (_bestGoodWeightNumerator + _bestGoodWeightDenominator) >=
	      _bestGoodWeightNumerator) &&
	     _minNonemptyCatNum + concession + 1 < NUMBER_OF_KEPT_CATEGORIES)
	{
	  ++concession;
	  tmp = tmp / (_bestGoodWeightNumerator + _bestGoodWeightDenominator);
	};

      long cat = _minNonemptyCatNum + concession;
      while (EmptyCat(cat))
	{
	  ASSERT(cat > _minNonemptyCatNum);
	  --cat;
	};
      ASSERT(!EmptyCat(cat));
      Clause* res = _category[cat].Next();
      Clause::Remove(res);
      StatUsed(cat); 
      ++_totalSelectedByAgeOrWeight;
      ++_totalSelectedByWeight;
      return res;
    }
   else // pick by age
    {
     unsigned long oldestNum = _category[_minNonemptyCatNum].Next()->Number();
     long oldestCat = _minNonemptyCatNum;
     for (long cat = _minNonemptyCatNum + 1; cat<=_maxNonemptyCatNum; cat++)
      {
       if ((!(EmptyCat(cat))) && (_category[cat].Next()->Number() < oldestNum))
        {
         oldestNum = _category[cat].Next()->Number();
         oldestCat = cat;
        };
      };
     ++_totalSelectedByAgeOrWeight;
     Clause* res = _category[oldestCat].Next();
     Clause::Remove(res);
     StatUsed(oldestCat);
     return res;
    };
  }
 else // _ageWeightRatioDenominator <= 0
  {
   if (EmptyCat(0)) { return 0; };
   Clause* res = _category[0].Next();
   Clause::Remove(res);
   StatUsed(0);
   return res;
  };
}; //  Clause* KEPT::Next() 



Clause* KEPT::theWorstUnblockedClauseAfter(Clause* cl)
{
  CALL("theWorstUnblockedClauseAfter(Clause* cl)");
  long cat = Category(cl);
  Clause* cand = cl;


 previous:
  cand = cand->Previous();


  if (cand == &(_category[cat]))
    {
    try_smaller_cat:
      if (cat <= _minNonemptyCatNum) return 0;
      --cat;
      if (EmptyCat(cat)) goto try_smaller_cat;


      cand = _category[cat].Previous();


      ASSERT(cand != &(_category[cat]));
    };
  if (!cand->isProtected()) 
    {
      return cand;
    };
  goto previous;
}; // Clause* KEPT::theWorstUnblockedClauseAfter(Clause* cl)
  








ostream& KEPT::outputProfile(ostream& str) const
{
  CALL("outputProfile(ostream& str) const");
  str << "%============ Profile of Passive (weight:stored:oldest):";
  ulong columns = 0UL;
  for (long cat = _minNonemptyCatNum; cat <= _maxNonemptyCatNum; cat++)
    {
      if (!EmptyCat(cat))
	{
	  if (columns == 0UL) 
	    {
	      str << "\n% ";
	      columns = 5UL;
	    };
	  str << cat << ":" << currently_kept[cat] 
	      << ":" << _category[cat].Next()->Number() << "   ";
	  --columns;
	};
    };

  return
    str << "\n%======================================================\n";
}; // ostream& KEPT::outputProfile(ostream& str) const


void KEPT::DestroyAllClauses()
{
 CALL("DestroyAllClauses()");
 for (long cat = _minNonemptyCatNum; cat <= _maxNonemptyCatNum; cat++)
  {
   while (!(EmptyCat(cat)))
    {
     Clause* c = _category[cat].Next();
     Clause::Remove(c);
     c->Destroy();
    };
  };
 Reset();  
}; // void KEPT::DestroyAllClauses()


#ifndef NO_DEBUG
 bool KEPT::clean() const
 {
  if (_maxPriorityFlag 
      || num_of_currently_kept
      || num_of_used
      || num_of_removed) return false;
  if (numOfReachable != LONG_MAX) return false;
  for (long cat = 0; cat < NUMBER_OF_KEPT_CATEGORIES; cat++)
   {
    if (currently_kept[cat] || used[cat] || removed[cat]) return false;
    if (_category[cat].Next() != (_category + cat)) return false;
    if (_category[cat].Previous() != (_category + cat)) return false;  
   };         
  
 if (_minNonemptyCatNum || 
     _maxNonemptyCatNum || 
     _totalSelectedByAgeOrWeight ||
     _totalSelectedByWeight) 
   return false;
  if (max_accepted_cat != NUMBER_OF_KEPT_CATEGORIES-1) return false;
   return true;
 }; // bool KEPT::clean() const

 inline long KEPT::CountCurrentlyKept(long cat)
  {
   Clause* cl = _category[cat].Next();
   long res = 0L;
   while (cl != &(_category[cat]))
    {
     res++;
     cl = cl->Next();
    };
   return res;
  }; 
   
 bool KEPT::CheckStatistics()
  {
   long sum_used = 0L;
   long sum_removed = 0L;
   long sum_currently_kept = 0L;
   for (long c = 0; c < NUMBER_OF_KEPT_CATEGORIES; c++)
    {
     sum_used += used[c];
     sum_removed += removed[c];
     sum_currently_kept += currently_kept[c];
        
     long real_curr_kept = CountCurrentlyKept(c);
     if (currently_kept[c] != real_curr_kept)
      {
       DOP(DMSG("KEPT::CheckStatistics() > ") << "currently_kept[" << c << "] CORRUPT!\n");
       DOP(DMSG("KEPT::CheckStatistics() > stat = ") << currently_kept[c] << ", real = " << real_curr_kept << "\n");
       return false;
      };
    };  
   if (sum_used != num_of_used) { return false; };
   if (sum_removed != num_of_removed) { return false; };
   if (sum_currently_kept != num_of_currently_kept) { return false; };
   if (_minNonemptyCatNum > _maxNonemptyCatNum) { return false; };
   if (_minNonemptyCatNum < _maxNonemptyCatNum) 
    {
     if (EmptyCat(_minNonemptyCatNum)) 
      {
       DOP(DMSG("KEPT::CheckStatistics() > ") 
            << "_minNonemptyCatNum(" << _minNonemptyCatNum << ") < _maxNonemptyCatNum("   
            << _maxNonemptyCatNum << ") but EMPTY!\n");
       return false;
      };
     if (EmptyCat(_maxNonemptyCatNum))
      {
       DOP(DMSG("KEPT::CheckStatistics() > ") 
           << "_maxNonemptyCatNum(" << _maxNonemptyCatNum << ") > _minNonemptyCatNum(" << _minNonemptyCatNum << ") but EMPTY!\n");
       return false;
      };
    };
   return true; 
  }; // bool KEPT::CheckStatistics() 
#endif 




//====================================================
