//
// File:         premise_factoring.hpp
// Description:  OR premises with factoring.
// Created:      Feb 15, 2000, Alexandre Riazanov, riazanov@cs.man.ac.uk
// Revised:      Dec 10, 2001 , Alexandre Riazanov, riazanov@cs.man.ac.uk
//               The arrays OR_PREMISE::{max_literal,mark} replaced by 
//               extending arrays OR_PREMISE::{_maxLiteral,_litIsMarked}
// Revised:      Dec 11, 2001, Alexandre Riazanov, riazanov@cs.man.ac.uk 
//               The stack OR_PREMISE_WITH_FACTORING::factored replaced by 
//               extending stack OR_PREMISE_WITH_FACTORING::_factoredLits
//====================================================
#ifndef PREMISE_FACTORING_H
//======================================================================
#define PREMISE_FACTORING_H
#include "jargon.hpp"
#include "VampireKernelDebugFlags.hpp"
#include "or_premise.hpp"
#include "ExpandingStack.hpp"
#include "GlobAlloc.hpp"
//================================================
#undef DEBUG_NAMESPACE
#ifdef DEBUG_PREMISE_FACTORING
 #define DEBUG_NAMESPACE "OR_PREMISE_WITH_FACTORING"
#endif
#include "debugMacros.hpp"
//=================================================
namespace VK
{

class OR_PREMISE_WITH_FACTORING : public OR_PREMISE
 {
  public:
   OR_PREMISE_WITH_FACTORING();
   ~OR_PREMISE_WITH_FACTORING();

   
   bool Load(ulong ind,Clause* c)
   {
    CALL("Load(ulong ind,Clause* c)");
    if (!OR_PREMISE::load(ind,c)) return false;

    DOP(_factoredLits.unfreeze());
    if (!_factoredLits.tryToExpand(c->numOfAllLiterals()))
      return false;
    DOP(_factoredLits.freeze());

    DOP(_rotationRequired.unfreeze());
    if (!_rotationRequired.tryToExpand(c->numOfAllLiterals()))
      return false;
    DOP(_rotationRequired.freeze());
    return true;
   };
   bool LoadOptimized(ulong ind,Clause* c)
   {
    CALL("LoadOptimized(ulong ind,Clause* c)");
    if (!OR_PREMISE::loadOptimized(ind,c)) return false;

    DOP(_factoredLits.unfreeze());
    if (!_factoredLits.tryToExpand(c->numOfAllLiterals()))
      return false;    
    DOP(_factoredLits.freeze());

    DOP(_rotationRequired.unfreeze());
    if (!_rotationRequired.tryToExpand(c->numOfAllLiterals()))
      return false;    
    DOP(_rotationRequired.freeze());
    return true;
   };
   bool Load(ulong ind,RES_PAIR_LIST* res_pair)
   {
    CALL("Load(ulong ind,RES_PAIR_LIST* res_pair)");
    if (!OR_PREMISE::load(ind,res_pair)) return false;

    DOP(_factoredLits.unfreeze());
    if (!_factoredLits.tryToExpand(res_pair->TheClause()->numOfAllLiterals()))
      return false;
    DOP(_factoredLits.freeze());

    DOP(_rotationRequired.unfreeze());
    if (!_rotationRequired.tryToExpand(res_pair->TheClause()->numOfAllLiterals()))
      return false;
    DOP(_rotationRequired.freeze()); 
    return true;   
   };

   bool LoadOptimized(ulong ind,RES_PAIR_LIST* res_pair)
   {
    CALL("LoadOptimized(ulong ind,RES_PAIR_LIST* res_pair)");
    if (!OR_PREMISE::loadOptimized(ind,res_pair)) return false;

    DOP(_factoredLits.unfreeze());
    if (!_factoredLits.tryToExpand(res_pair->TheClause()->numOfAllLiterals()))
      return false;
    DOP(_factoredLits.freeze());

    DOP(_rotationRequired.unfreeze());
    if (!_rotationRequired.tryToExpand(res_pair->TheClause()->numOfAllLiterals()))
      return false;
    DOP(_rotationRequired.freeze());  
    return true; 
   };

   void ResetFactoring()
    {
     CALL("ResetFactoring()");
     was_factored = false;
     if (FactoringPossible())
      {
       _factoredLits.reset();
       _rotationRequired.reset();
       _mainLiteral = OR_PREMISE::_selectedLiteral[cut_lit]; 
       _mainLiteralRotation = OR_PREMISE::_selectedLiteralRotation[cut_lit]; 
      };
    };
   bool WasFactored() { return was_factored; };
   bool FactoringPossible() { return lastSelected() != 0UL; };
   bool GetFirstFactor(ulong first_candidate);
   bool GetNextFactor();
  private: // aux. methods
   bool FindFirstFactor(ulong& candidate,bool& rotationRequired);
   void factoriseLiteralProperties();  
 private:
   BK::ExpandingStack<BK::GlobAlloc,ulong,BK::FastRelocation<ulong>,32UL> _factoredLits;
   BK::ExpandingStack<BK::GlobAlloc,bool,BK::FastRelocation<bool>,32UL> _rotationRequired;
   
   PrefixSym* _mainLiteral; // the literal itself in prefix representation
   PrefixSym* _mainLiteralRotation;
   
   ulong _lastFactoredLitNum;
   bool _rotationRequiredForLastFactored;
   bool was_factored;
 }; // class OR_PREMISE_WITH_FACTORING

}; // namespace VK

//======================================================================
#endif
