
/*
 * File VarCounter.cpp.
 * Copyright (c) 2004 Andrei Voronkov and Alexandre Riazanov.
 *
 * You are not allowed to use or modify this software or
 * any part thereof.
 * The only reason this file appears on the Web is that
 * it is required by the CASC rules.
 * All questions related to licensing and distribution of
 * Vampire should be addressed to Andrei Voronkov.
 */


#include "VarCounter.hpp"
#include "Int.hpp"
#include "Tracer.hpp"
#include "XML.hpp"



VarCounter::VarCounter ()
  :  _top(0),
     _counts(0)
{
}


VarCounter::~VarCounter ()
{
  TRACER("VarCounter::~VarCounter");

  delete [] _counts;
} 


void VarCounter::incCount (Var w)
{
  TRACER("VarCounter::incCount");

  int v = w.toInt();
  if (v >= _top) { 
    expandToFit(v);
  }

  _counts[v] ++;
} 



void VarCounter::decCount (Var w)
{
  TRACER("VarCounter::decCount");

  int v = w.toInt();
  if (v >= _top) { 
    expandToFit(v);
  }

  _counts[v] --;
} 



void VarCounter::expandToFit (int v)
{
  TRACER("VarCounter::expandToFit");

  
  int newTop = Int::max(_top+16,v+1);

  
  int* newCounts = new int[newTop];
  if (! newCounts) {
    ALLOCATION_FAILURE;
  }
  
  for (int i = 0;i < _top;i++) {
    newCounts[i] = _counts[i];
  }
  for (int k = _top;k < newTop;k++) {
    newCounts[k] = 0;
  }
  delete [] _counts;
  _counts = newCounts;
  _top = newTop;
} 



int VarCounter::getCount (Var w) const 
{ 
  TRACER("VarCounter::getCount");

  int v = w.toInt();
  return v < _top ? _counts[v] : 0; 
} 



void VarCounter::setCount (Var w, int c)
{
  TRACER("VarCounter::setCount");

  int v = w.toInt();
  if (v >= _top) { 
    expandToFit(v);
  }

  _counts[v] = c;
} 



XML::Element* VarCounter::toXML () const
{
  XML::DeepElement* vc = new XML::DeepElement("var_counter");
  vc->addChild(new XML::IntegerElement("top",_top));
  XML::DeepElement* counters = new XML::DeepElement("counters");
  vc->addChild(counters);
  for (int i = 0;i < _top;i++) {
    if (_counts[i] != 0) {
      XML::EmptyElement* counter = new XML::EmptyElement("counter");
      counters->addChild(counter);
      counter->addAttribute(new XML::IntegerAttribute("var",i));
      counter->addAttribute(new XML::IntegerAttribute("value",_counts[i]));
    }
  }
  return vc;
} 
