;======================================================================
;
;                          resio.scm
;
;                   Gandalf version c-2.6 r1
;
; Input-output and auxiliary funs of the resolution proof searcher.
;
;                  Copyright (c) Tanel Tammet
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 1, or (at your option)
; any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; The author can be reached at tammet@staff.ttu.ee, tanel@cs.ioc.ee or
; tammet@cs.chalmers.se
;
;
;=======================================================================


;===================================================================
;
;              creating the input clause set
;
;===================================================================

; make-input-clauses takes a raw NUMERIFIED clause set

; and builds an input clause list

;(define (make-input-clauses cllist)
;  (set! *kept-cl-nr* 0)
;  (set! cllist (numerify-clause-list cllist))
;  (fill-predsymb-bit-vector!)
;  (set! cllist (map make-input-clause cllist))
;  cllist)

; make-input-clause takes an undecorated numerified clause
; and returns a decorated input clause

(define (make-input-clause cl sosflag)

  ;(newline)(display "make-input-clause entered with: ")(newline)
  ;(print-pure-clause cl)(newline)


  (set! cl (remove-duplicates cl))
  (let ((litinfolst (make-input-litinfolst cl))
        (tmp '()) )

    ;(newline)(display "litinfolst made: ")(display litinfolst)(newline)

    (set-x-vars-selfref-touched! *max-nr-of-vars-in-clause*)
    (set! cl (map (lambda(x)(subst-deco-term x #t)) cl))

    (set! *used-demodulators* '())
    (set! tmp (sort-clause (map cons litinfolst cl) #t))
    (set! cl  (map cdr tmp))
    (set! litinfolst (map car tmp))

    ;(newline)(display "clause after sorting: ")(newline)
    ;(print-pure-clause cl)(newline)
    ;(newline)(display "litinfolst after sorting: ")(display litinfolst)
    ;(newline)

    (set! *clause-var-nr* 0)
    (if (pure-clause-may-cont-vars? cl)
        (begin
          (set! cl (copy-list cl))
          ;create a list of lists (lit litrepvars litvars othervars follvars):
          (make-varinfo-lst! cl (list '())) ; cl is destr replaced to be varinfo-lst
          (set! *new-var-names* '())
          (make-fulldeco-clause! cl) )) ; cl is destr replaced to be a clause

    ;(newline)(debug-print-pure-clause cl)(display *clause-var-nr*) (newline)

    ; cl is now the finished pure clause.
    ; *clause-var-nr* is now the max var nr in the final pure clause
    ; start making the info-part.
    (let ((strengthint 0)
          (len 0)
          (predsymbsint 0)
          (weight 0)
          (minw 0)
          (clinfo '()) )
      (do ((clpart cl (cdr clpart)))
          ((null? clpart))
        (let* ((lit (car clpart))
               (fun (car lit)) )
          (if (countable-pred? fun) ; answer-preds may be uncountable
              (begin
                (if (zero? strengthint)
                    (set! strengthint (logand *strength-mask* fun))
                    (set! strengthint (sum-fun-strengths fun strengthint)) )
                (if (not (zero? (logand (predsymb-bit fun) predsymbsint)))
                    (set! predsymbsint
                          (logior *cl-rep-predsymbs-mask* predsymbsint) ))
                (set! predsymbsint (logior (predsymb-bit fun) predsymbsint))
                (set! len (+ 1 len))
                (set! weight (+ weight (lit-weight lit))) ))))
      (if (> len *cl-maxlen*) (set! len *cl-maxlen*))
      (if (> *clause-var-nr* *cl-maxvar*)(set! *clause-var-nr* *cl-maxvar*))
      (set! strengthint
            (logior (logior (shiftleft len *cl-len-shift*)
                            (shiftleft *clause-var-nr* *cl-maxvar-shift*) )
                    strengthint))
      (set! *kept-cl-nr* (+ 1 *kept-cl-nr*))
      (if (or *order-strategy-flag*
              *name-order-strategy-flag*
              *exp-order-strategy-flag*
              *epr-order-strategy-flag*
              *firstlitpref-order-strategy-flag*
              *posweight-order-strategy-flag*
              *negweight-order-strategy-flag*
              *ordered-paramodulation-flag*)
          (set! litinfolst
                (map cddr (order-prohibit-litinfos
                            (map (lambda(x)(cons '() (cons '() x)))
                                 litinfolst)
                            cl))))

      ; - - - -


      (if (and *split-strategy-flag*
             (not (null? (cdr cl))) )
        (begin
           (set! weight strengthint)
           (set! tmp (map (lambda(x)(make-var-depth-info x)) cl))

           (if  (every? (lambda(el)
                         (and (not (null? el))
                             (every? (lambda(el2)
                                       (or (eq? el2 el)
                                           (every? (lambda(v)
                                                    (not (assq (car v) el2)) )
                                                   el)))
                                     tmp)))
                       tmp)
               (begin
                 (if (> *display-info-level* 70)
                     (begin
                       (newline)
                       (display "splittable clause: ")
                       (print-pure-clause cl)
                       (newline) ))
                 (do ((clpart cl (cdr clpart)))
                     ((null? clpart)
                      (set! weight minw) )
                   (cond ((zero? minw)
                          (set! minw (logand *strength-mask* (caar clpart))) )
                         ((> (logand *strength-mask* (caar clpart))
                             minw)
                          (set! minw (logand *strength-mask*
                                             (caar clpart) )))))))))


      (set! clinfo
            (list strengthint
                  predsymbsint
                  litinfolst
                  weight
                  *kept-cl-nr*
                  (list 'input *kept-cl-nr*)
                  sosflag))

      (set! tmp (cons clinfo cl))
      (if *dynamic-demodulation-flag*
         (attempt-to-make-a-demodulator tmp) )
      tmp)))


(define (remove-duplicates lst)
  (cond
   ((null? lst) '())
   ((member (car lst) (cdr lst))
      (remove-duplicates (cdr lst)) )
   (else
      (cons (car lst) (remove-duplicates (cdr lst))) )))

(define (make-input-litinfolst cl)
  (cond
   ((not *hyper-strategy-flag*)
      (make-input-litinfolist-aux cl) )
   ((positive-cl? cl)
      (make-input-litinfolist-aux cl) )
   ; nucleus
   (else
      (let ((nr 1)
            (res '()) )
        (do ((clpart cl (cdr clpart)))
            ((null? clpart) (reverse res))
          (if (neg-fun? (caar clpart))
              (begin (set! res (cons (list (put-litinfo-ordernr
                                            nr *litinf-noparainto-mask*))
                                     res))
                     (set! nr (+ 1 nr)) )
              (set! res (cons (make-pos-lit-litinfo (car clpart))
                              res))))))))


(define (make-input-litinfolist-aux cl)
  (map make-pos-lit-litinfo cl) )

(define (make-pos-lit-litinfo x)
  (cond ((answer-lit? x)
           (list *answer-litinf*) )
        ((and (eq? *eq-pred-nr* (pure-fun (car x)))
              (equal? (eq-lit-arg1 x) (eq-lit-arg2 x)) )
           ; eq reflexivity axiom
           (list (logior *litinf-noparainto-mask*
                         (logior *litinf-noparafromleft-mask*
                                 *litinf-noparafromright-mask*))))
        ((and (composite? (cadr x))
              (eq? *eq-fun-nr* (pure-fun (caadr x)))
              (equal? (eq-lit-arg1 (cadr x)) (eq-lit-arg2 (cadr x))) )
           ; eq-fun reflexivity axiom
           (list (logior *litinf-noparainto-mask*
                         (logior *litinf-noparafromleft-mask*
                                 *litinf-noparafromright-mask*))))
        (else
           ; everything allowed
           (list 0) )))

; make-varinfo-lst! takes a pure clause and replaces it
; destructively with a list
; (lit litrepvars litvars othervars follvars)
; or for the ground literals:
; (lit)
; it returns the lists of all lists of vars in clauses.
; the second arg is the list of lists of vars in previous clauses
;
; answer-literals are treated specially: to the rest of the clause
; it seems that the answer-literal is ground.

(define (make-varinfo-lst! cl prevars)
  (if (null? cl)
      '()
      (let* ((lit (car cl))
             (fun (car lit))
             (lcvars
               (if (fun-may-cont-vars? fun)
                   (collect-vars-composite! lit)
                   (cons '() '()) ))
             (follvars '()) )
        ;(if (answer-pred? fun)
        ;    (set! follvars
        ;          (make-varinfo-lst! (cdr cl) (cons '() prevars)) )
        ;    (set! follvars
        ;          (make-varinfo-lst! (cdr cl) (cons (cdr lcvars) prevars)) ))
        (set! follvars
                 (make-varinfo-lst! (cdr cl) (cons (cdr lcvars) prevars)) )
        (if (fun-may-cont-vars? fun)
            (set-car! cl
                      (list lit
                            (car lcvars)
                            (cdr lcvars)
                            (append prevars follvars)
                            follvars))
            (set-car! cl (list lit)) )

        ;(if (answer-pred? fun)
        ;    (cons '() follvars)
        ;    (cons (cdr lcvars) follvars) ))))
        (cons (cdr lcvars) follvars) )))


; make-fulldeco-clause! takes a list of elements in the form
; (lit litrepvars litvars othervars follvars)
; and replaces its elements destructively by the fully
; decorated literals.

(define (make-fulldeco-clause! reslst)

  ;(newline)(display "make-fulldeco-clause! entered with: ")(newline)
  ;(display reslst)(newline)
  ;(display *clause-var-nr*)(newline)
  ;(display *new-var-names*)(newline)(newline)
  ;(for-each (lambda(x)(debug-print-lit (car x))(newline)
  ;                      (display (cdr x))(newline) )
  ;            reslst)

  (if (not (null? reslst))
      (let* ((el (car reslst)))
        (if (term-may-cont-vars? (car el))
            (decorate-vars-lit!
              (car el) ;lit
              (cadr el) ;repeating vars in the given literal
              (caddr el) ;all vars in the given literal
              (caddr (cdr el)) ;list of var lists of the other literals
              (caddr (cddr el)) )) ;list of var lists of the foll literals
        (set-car! reslst (car el)) ; el has been destructively changed
        (make-fulldeco-clause! (cdr reslst)) )))

; fill-predsymb-bit-vector uses the *pred-num-table* created by
; numerify-literal during numerify-clause-list.
; It is an alist ((puresymbol . nr) ...) where nr-s start from
; 1 (this is assumed!)
; if the ans-pred and equality pred are present, they are always taken
; into the first 13 bit-representable predsymbs.
;
; It also sets the value for *ans-pred-nr* and *eq-pred-nr*.

(define *ans-pred-nr* 0)
(define *eq-pred-nr* 0)
(define *eq-fun-nr* 0)
(define *case-fun-nr* 0)
(define *if-fun-nr* 0)

(define *subst-fun-nr* 0)
(define *subst-prim-fun-nr* 0)
(define *lambda-fun-nr* 0)
(define *lambda-dummypar-nr* 0)

(define *int-outmost-symb-labelfun* '())
(define *int-outmost-label-fun-nr* 0)

(define (pos-eq-pred? fn)
  (eq? *eq-pred-nr* fn) )

(define (pos-eq-fun? fn)
  (eq? *eq-fun-nr* fn) )


(define (fill-predsymb-bit-vector!)
  (let ((a (assq *ans-symbol* *pred-num-table*))
        (e (assq *eq-symbol* *pred-num-table*))
        (ef (assq *eq-fun-symbol* *fun-num-table*))
        (cf (assq *case-fun-symbol* *fun-num-table*))
        (ifs (assq *if-fun-symbol* *fun-num-table*))
        (intlabel
          (if *int-logic-flag*
              (assq *int-outmost-symb-labelfun* *pred-num-table*)
              #f)))
    (set! *ans-pred-nr* 0)
    (set! *eq-pred-nr* 0)
    (set! *eq-fun-nr* 0)
    (set! *case-fun-nr* 0)
    (set! *if-fun-nr* 0)
    (set! *int-outmost-label-fun-nr* 0)
    (if e
        (begin
          (set! *pred-num-table*
                (cons e (remove-eq e *pred-num-table*)) )
          (set! *eq-pred-nr* (cdr e)) ))
    (if a
        (begin
          (set! *pred-num-table*
                (cons a (remove-eq a *pred-num-table*)) )
          (set! *ans-pred-nr* (cdr a)) ))
    (if ef (set! *eq-fun-nr* (cdr ef)))
    (if cf (set! *case-fun-nr* (cdr cf)))
    (if ifs (set! *if-fun-nr* (cdr ifs)))
    (if intlabel (set! *int-outmost-label-fun-nr* (cdr intlabel)))
    (do ((part *pred-num-table* (cdr part))
         (n 1 (if (< n *cl-predsymbs-pos-max*)
                  (shiftleft n 1)
                  *cl-pred-overflow-mask*)))
        ((null? part))
      (vector-set! *predsymb-bit-vector*
                       (shiftright (cdar part) *fnsymb-shift*)
                       n)
      (vector-set! *predsymb-bit-vector*
                       (shiftright (logior *sgn-mask* (cdar part))
                                   *fnsymb-shift*)
                       (if (eq? n *cl-pred-overflow-mask*)
                           *cl-pred-overflow-mask*
                           (shiftleft n *cl-predsymbs-neg-shift*) )))))


;===================================================================
;
;              strategies and strategical flags
;
;===================================================================


;;; NB! Currently *ordered-factor-flag* works OK only for input sos,
;;;     otherwise it is like always #t

(define (turn-on-demodulation)
  (set! *demodulate-flag* #t)
  (set! *dynamic-demodulation-flag* #t)
  ;(set! *new-demodulators-flag* #f)
  (set! *new-demodulators-flag* #t)
  (set! *demodulate-constants-flag* #t)
  (set! *ordered-paramodulation-flag* #t) )


(define (turn-off-demodulation)
  (set! *demodulate-flag* #f)
  (set! *dynamic-demodulation-flag* #f)
  (set! *new-demodulators-flag* #f)
  (set! *demodulate-constants-flag* #f) )
  ; (set! *ordered-paramodulation-flag* #f)


(define (init-binres-strat)
  (set! *mace-model-strat* #f)
  (set! *sem-model-strat* #f)
  (set! *complete-unsat-strategy* #f)
  (set! *complete-sat-strategy* #f)
  (set! *unit-strategy-flag* #f)
  (set! *double-strategy-flag* #f)
  (set! *bliksem-strategy-flag* #f)
  (set! *user-criterias-post-flag* #f)
  (set! *factors-stratflag* #t)
  (set! *ordered-factor-flag* #f)
  (set! *name-order-strategy-flag* #f)
  (set! *epr-order-strategy-flag* #f)
  (set! *exp-order-strategy-flag* #f)
  (set! *firstlitpref-order-strategy-flag* #f)
  (set! *remove-identical-lit-stratflag* #t)
  (set! *paramodulation-allowed-flag* #t)
  (set! *hyper-strategy-flag* #f)
  (set! *type-theory-stratflag* #f)
  (set! *para-pos-unit-strategy* #f)
  (set! *para-unit-strategy* #f)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #t)
  (set! *paramodulate-from-vars-stratflag* #f)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-hyper-strat-new)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f)
   (set! *complete-unsat-strategy* #f)
   (set! *complete-sat-strategy* #f)
   (init-posweight-ordered-binres-strat)
   ; (set! *factors-stratflag* #f)
   (turn-on-demodulation)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *ordered-paramodulation-flag* #t)
   (set! *order-equality-kb-strat* #t)
   (set! *order-equality-lex-strat* #f)
   (set! *order-equality-firstpref-strat* #f)
   (set! *lex-order-prefer-bigger-arity* #t)
   (make-symbs-lex-order-vector)
   (make-symbs-kb-order-vector) )

(define (init-hyper-strat)
  (set! *mace-model-strat* #f)
  (set! *sem-model-strat* #f)
  (set! *complete-unsat-strategy* #f)
  (set! *complete-sat-strategy* #f)
  (set! *unit-strategy-flag* #f)
  (set! *double-strategy-flag* #f)
  (set! *bliksem-strategy-flag* #f)
  (set! *user-criterias-post-flag* #f)
  (set! *factors-stratflag* #t)
  (set! *ordered-factor-flag* #f)
  (set! *name-order-strategy-flag* #f)
  (set! *epr-order-strategy-flag* #f)
  (set! *exp-order-strategy-flag* #f)
  (set! *demodulate-with-clauses-stratflag* #f)
  (set! *remove-identical-lit-stratflag* #t)
  (set! *paramodulation-allowed-flag* #t)
  (set! *hyper-strategy-flag* #t)
  (set! *type-theory-stratflag* #f)
  (set! *para-pos-unit-strategy* #t)
  (set! *para-unit-strategy* #f)
  (turn-on-demodulation)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #t)
  (set! *paramodulate-from-vars-stratflag* #f)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-synth-strat)
  (set! *user-criterias-post-flag* #f)
  (set! *factors-stratflag* #t)
  (set! *ordered-factor-flag* #f)
  (set! *paramodulation-allowed-flag* #t)
  (set! *remove-identical-lit-stratflag* #t)
  (set! *hyper-strategy-flag* #f)
  (set! *type-theory-stratflag* #f)
  (set! *para-pos-unit-strategy* #f)
  (set! *para-unit-strategy* #f)
  (set! *classical-synth-stratflag* #t)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #t)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-binary-hyper-sos-strat)
   (init-binres-strat)
   (set! *binary-hyper-sos-strategy* #t) )

(define (init-unit-strat)
  (init-binres-strat)
  (set! *unit-strategy-flag* #t) )

(define (init-double-strat)
  (init-binres-strat)
  (set! *double-strategy-flag* #t) )

(define (init-binres-bliksem-strat)
  (init-posweight-ordered-binres-strat)
  (set! *bliksem-strategy-flag* #t) )

(define (init-posweight-ordered-unit-strat)
  (init-unit-strat)
  (set! *posweight-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )

(define (init-negweight-ordered-unit-strat)
  (init-unit-strat)
  (set! *negweight-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )

(define (init-ordered-binres-strat)
  (init-binres-strat)
  (set! *order-strategy-flag* #t)
  (set! *negweight-order-strategy-flag* #f)
  (set! *posweight-order-strategy-flag* #f) ; this used to be #t!!!
  (set! *name-order-strategy-flag* #f)
  (set! *ordered-factor-flag* #f) )

(define (init-ordered-binres-fulleq-strat)
  (init-binres-strat)
  (set! *order-strategy-flag* #t)
  (set! *paramodulate-from-vars-stratflag* #t)
  (set! *name-order-strategy-flag* #f)
  (set! *ordered-factor-flag* #f) )

(define (init-ordered-binres-sos-strat)
  (init-binres-strat)
  (set! *order-sos-strategy-flag* #t)
  (set! *order-strategy-flag* #t)
  (set! *name-order-strategy-flag* #f)
  (set! *ordered-factor-flag* #f) )


(define (init-name-ordered-binres-strat)
  (init-binres-strat)
  (set! *order-strategy-flag* #f)
  (set! *name-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )

(define (init-name-ordered-binres-sos-strat)
  (init-binres-strat)
  (set! *order-sos-strategy-flag* #t)
  (set! *order-strategy-flag* #f)
  (set! *name-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )

;(define (init-ordered-binres-0-strat)
;  (init-binres-strat)
;  (set! *order-strategy-flag* #t)
;  (set! *name-order-strategy-flag* #f)
;  (set! *ordered-factor-flag* #f) )

;(define (init-name-ordered-binres-0-strat)
;  (init-binres-strat)
;  (set! *order-strategy-flag* #f)
;  (set! *name-order-strategy-flag* #t)
;  (set! *ordered-factor-flag* #f) )

(define (init-weight-ordered-binres-sos-strat)
  (init-binres-strat)
  (set! *order-sos-strategy-flag* #t)
  (set! *firstlitpref-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )

(define (init-weight-ordered-binres-strat)
  (init-binres-strat)
  (set! *order-sos-strategy-flag* #f)
  (set! *firstlitpref-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )


(define (init-posweight-ordered-binres-sos-strat)
  (init-binres-strat)
  (set! *order-sos-strategy-flag* #t)
  (set! *posweight-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )

(define (init-negweight-ordered-binres-sos-strat)
  (init-binres-strat)
  (set! *order-sos-strategy-flag* #t)
  (set! *negweight-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )

(define (init-posweight-ordered-binres-strat)
  (init-binres-strat)
  (clear-eq-order-strat)
  (set! *posweight-order-strategy-flag* #t)
  (turn-on-demodulation)
  ; (set! *unit-strategy-flag* #t)
  ;(set! *unit-equality-strat* #t)
  ;(set! *para-pos-unit-strategy* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #t)
  ;(set! *order-equality-firstpref-strat* #f)
  ;(set! *sos-preference-ratio* 2) )
  (set! *ordered-factor-flag* #t) )

(define (init-negweight-ordered-binres-strat)
  (init-binres-strat)
  (set! *negweight-order-strategy-flag* #t)
  (set! *ordered-factor-flag* #f) )


(define (init-hyper-wos-strat)
   (init-hyper-strat)
   (set! *hyper-wos-strategy-flag* #t) )

(define (init-ordered-hyper-strat)
  (init-hyper-strat)
  (set! *order-strategy-flag* #t) )

(define (init-typet-strat)
  (set! *user-criterias-post-flag* #f)
  (set! *factors-stratflag* #f)
  (set! *remove-identical-lit-stratflag* #f)
  (set! *paramodulation-allowed-flag* #t)
  (set! *hyper-strategy-flag* #t)
  (set! *type-theory-stratflag* #t)
  (set! *para-pos-unit-strategy* #t)
  (set! *para-unit-strategy* #f) )

(define (init-binary-unit-depth-backdemod)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *sos-preference-ratio* 0) )

(define (init-binary-unit-depth-backdemod-alt)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *sos-preference-ratio* 2) )

(define (init-binary-unit-depth-root-backdemod-alt)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  ; (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *root-para-strat* #t)
  (set! *order-equality-lex-strat* #t)
  (set! *order-equality-firstpref-strat* #f)
  (set! *sos-preference-ratio* 6) )

(define (init-binary-unit-depth)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *sos-preference-ratio* 0) )

(define (init-binary-unit-depth-sospref)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *sos-preference-ratio* 0) )


(define (init-binary-unit-depth-sospref-alt)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *sos-preference-ratio* 2) )


(define (init-binary-unit-lex)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #t)
  (set! *order-equality-firstpref-strat* #f) )

(define (init-binary-unit-lex-backdemod)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #t)
  (set! *order-equality-lex-strat* #t)
  (set! *order-equality-firstpref-strat* #f) )

(define (init-binary-unit-firstpref)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #t) )

(define (init-binary-unit-firstpref-backdemod)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-strategy-flag* #t)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #t) )

(define (init-ordered-hyper-uniteq-strat)
   (init-ordered-hyper-strat)
   (set! *para-unit-strategy* #t) )

(define (init-unit-uniteq-strat)
   (init-unit-strat)
   (set! *unit-strategy-flag* #t)
   (set! *para-unit-strategy* #t) )

(define (init-name-ordered-binres-uniteq-strat)
   (init-name-ordered-binres-strat)
   (set! *para-unit-strategy* #t) )

(define (init-ordered-binres-uniteq-strat)
   (init-ordered-binres-strat)
   (set! *para-unit-strategy* #t) )

(define (init-binres-uniteq-strat)
   (init-binres-strat)
   (set! *para-unit-strategy* #t) )

; special equality ordering strategies


; equality ordering selection principles:
;
; - default is knuth-bendix, bigger arity preferred
; - for posweight-order:
;   - binary-posweight-ordered-binres-kb-big (default) is knuth-bendix, bigger arity preferred
;   - binary-posweight-order-kb-small is knuth-bendix, smaller arity preferred
;   - binary-posweight-order-lex-big is lex, bigger arity preferred
;   - binary-posweight-order-lex-small is lex, smaller arity preferred
;   - binary-posweight-order-firstpref is first arg term depth based
; - special unit-equality limits:
;   - hyper-uniteq is knuth-bendix, bigger arity preferred, units only
;   - binary-unit-uniteq is knuth-bendix, bigger arity preferred, units only
;   - binary-uniteq is knuth-bendix, bigger arity preferred, units only

;(init-posweight-ordered-binres-kb-big-strat)
;(init-posweight-ordered-binres-kb-small-strat)
;(init-posweight-ordered-binres-lex-big-strat)
;(init-posweight-ordered-binres-lex-small-strat)
;(init-posweight-ordered-firstpref-strat)



(define (init-exp-order-strat)
 ; (newline)(display "kb big init")(newline)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #f)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #f)
  (set! *ordered-factor-flag* #t)
  ;(turn-off-demodulation)
  (set! *ordered-paramodulation-flag* #f)
  (set! *order-equality-kb-strat* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #t)
  (set! *epr-order-strategy-flag* #f)
  (set! *exp-order-strategy-flag* #t)
  (set! *complete-unsat-strategy* #f)
  (set! *complete-sat-strategy* #f)
  (set! *paramodulate-from-vars-stratflag* #f)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-epr-order-strat)
 ; (newline)(display "kb big init")(newline)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #f)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #f)
  (set! *ordered-factor-flag* #t)
  (turn-off-demodulation)
  (set! *ordered-paramodulation-flag* #f)
  (set! *order-equality-kb-strat* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #t)
  (set! *epr-order-strategy-flag* #t)
  (set! *complete-unsat-strategy* #t)
  (set! *complete-sat-strategy* #f)
  (set! *paramodulate-from-vars-stratflag* #t)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )



(define (init-posweight-ordered-binres-kb-big-strat)
 ; (newline)(display "kb big init")(newline)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #t)
  (set! *ordered-factor-flag* #t)
  (turn-on-demodulation)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #t)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-posweight-ordered-binres-kb-small-strat)
  ;(newline)(display "kb small init")(newline)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #t)
  (set! *ordered-factor-flag* #t)
  (turn-on-demodulation)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #f)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-posweight-ordered-binres-kb-small-fulleq-strat)
  ;(newline)(display "kb small init")(newline)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #t)
  (set! *ordered-factor-flag* #t)
  (turn-on-demodulation)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #f)
  (set! *paramodulate-from-vars-stratflag* #t)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-posweight-ordered-binres-lex-big-strat)
  ;(newline)(display "lex big init")(newline)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #f)
  (set! *ordered-factor-flag* #t)
  (turn-on-demodulation)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #f)
  (set! *order-equality-lex-strat* #t)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #t)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-posweight-ordered-binres-lex-small-strat)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #f)
  (set! *ordered-factor-flag* #t)
  (turn-on-demodulation)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #f)
  (set! *order-equality-lex-strat* #t)
  (set! *order-equality-firstpref-strat* #f)
  (set! *lex-order-prefer-bigger-arity* #f)
  (set! *paramodulate-from-vars-stratflag* #t)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )


(define (init-posweight-ordered-firstpref-strat)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-kb-strat* #f)
  (set! *ordered-factor-flag* #t)
  (turn-on-demodulation)
  (set! *ordered-paramodulation-flag* #t)
  (set! *order-equality-kb-strat* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #t)
  (set! *lex-order-prefer-bigger-arity* #t)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector) )

; satisfiability checking strategies

(define (init-sat-hyper-strat)
   (set! *mace-nofuns-model-strat* #f)
   (set! *simple-model-strat* #f)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f)
   (set! *random-model-strat* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (init-hyper-strat)
   (turn-off-demodulation)
   (set! *complete-unsat-strategy* #t)
   (set! *complete-sat-strategy* #t)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *ordered-paramodulation-flag* #t)
   (set! *order-equality-lex-strat* #f)
   (set! *order-equality-firstpref-strat* #f)
   (set! *para-pos-unit-strategy* #f)
   (set! *para-unit-strategy* #f)
   (set! *paramodulate-from-vars-stratflag* #t) )

(define (init-sat-posweight-ordered-binres-strat)
   (set! *mace-nofuns-model-strat* #f)
   (set! *simple-model-strat* #f)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f)
   (set! *random-model-strat* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (init-posweight-ordered-binres-kb-big-strat)
   (turn-off-demodulation)
   ;(set! *posweight-order-strategy-flag* #f)
   (set! *complete-unsat-strategy* #t)
   (set! *complete-sat-strategy* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *ordered-paramodulation-flag* #f)
   (set! *order-equality-lex-strat* #f)
   (set! *order-equality-firstpref-strat* #f)
   (set! *para-pos-unit-strategy* #f)
   (set! *para-unit-strategy* #f)
   (set! *paramodulate-from-vars-stratflag* #t) )

(define (init-sat-ordered-binres-strat)
   (set! *mace-nofuns-model-strat* #f)
   (set! *simple-model-strat* #f)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f)
   (set! *random-model-strat* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (init-ordered-binres-strat)
   (turn-off-demodulation)
   (set! *posweight-order-strategy-flag* #f)
   (set! *complete-unsat-strategy* #t)
   (set! *complete-sat-strategy* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *ordered-paramodulation-flag* #f)
   (set! *order-equality-lex-strat* #f)
   (set! *order-equality-firstpref-strat* #f)
   (set! *para-pos-unit-strategy* #f)
   (set! *para-unit-strategy* #f)
   (set! *paramodulate-from-vars-stratflag* #t) )

(define (init-sat-weight-ordered-binres-strat)
   (set! *mace-nofuns-model-strat* #f)
   (set! *simple-model-strat* #f)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f)
   (set! *random-model-strat* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (init-weight-ordered-binres-strat)
   (turn-off-demodulation)
   (set! *remove-all-tautologies-stratflag* #f)
   (set! *back-subsume-stratflag* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *ordered-paramodulation-flag* #t)
   (set! *order-equality-lex-strat* #f)
   (set! *order-equality-firstpref-strat* #f)
   (set! *para-pos-unit-strategy* #f)
   (set! *para-unit-strategy* #f) )

(define (init-prop-unsat-nofun-strat)
   (set! *complete-unsat-strategy* #t)
   (set! *complete-sat-strategy* #t)
   (set! *prop-unsat-strat* #t)
   (set! *mace-nofuns-model-strat* #f)
   (set! *random-model-strat* #f)
   (set! *simple-model-strat* #t)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f))

(define (init-simple-model-strat)
   (init-prop-sat-aux)
   (set! *complete-unsat-strategy* #f)
   (set! *complete-sat-strategy* #t)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *mace-nofuns-model-strat* #f)
   (set! *random-model-strat* #f)
   (set! *simple-model-strat* #t)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f) )

(define (init-mace-model-strat)
   (init-prop-sat-aux)
   (set! *complete-unsat-strategy* #f)
   (set! *complete-sat-strategy* #t)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *mace-nofuns-model-strat* #f)
   (set! *random-model-strat* #f)
   (set! *simple-model-strat* #f)
   (set! *mace-model-strat* #t)
   (set! *sem-model-strat* #f) )

(define  (init-sem-model-strat)
   (init-prop-sat-aux)
   (set! *complete-unsat-strategy* #f)
   (set! *complete-sat-strategy* #t)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *mace-nofuns-model-strat* #f)
   (set! *random-model-strat* #f)
   (set! *simple-model-strat* #f)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #t) )

(define (init-mace-nofuns-model-strat)
   (init-prop-sat-aux)
   (set! *complete-unsat-strategy* #t)
   (set! *complete-sat-strategy* #t)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *mace-nofuns-model-strat* #t)
   (set! *random-model-strat* #f)
   (set! *simple-model-strat* #f)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f) )

(define (init-random-model-strat)
   (init-prop-sat-aux)
   (set! *complete-unsat-strategy* #f)
   (set! *complete-sat-strategy* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *mace-nofuns-model-strat* #f)
   (set! *random-model-strat* #t)
   (set! *simple-model-strat* #f)
   (set! *mace-model-strat* #f)
   (set! *sem-model-strat* #f) )

(define (init-prop-sat-aux)
   (turn-off-demodulation)
   (set! *back-subsume-stratflag* #f)
   (set! *demodulate-with-clauses-stratflag* #f)
   (set! *ordered-paramodulation-flag* #f)
   (set! *order-equality-lex-strat* #f)
   (set! *order-equality-firstpref-strat* #f)
   (set! *para-pos-unit-strategy* #f)
   (set! *para-unit-strategy* #f) )


(define (clear-eq-order-strat)
  (set! *order-equality-kb-strat* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f) )

; - - - - - - -- - - - - - - strategical flags - - - - - - - - - -


(define *complete-unsat-strategy* #f)
(define *complete-sat-strategy* #f)
(define *sat-strategy-flag* #f) ; global flag: search for
                                ; contradiction or satisfiability?
(define *full-proof-strategy-flag* #f) ; global flag: show full proof?
(define *sos-exhausted-flag* #f) ; set to #t iff sos exhausted

(define *split-strategy-flag* #f)
(define *unit-equality-strat* #f)
(define *root-para-strat* #f)
(define *unit-para-flag* #f)
(define *order-equality-lex-strat* #f) ; use depth- or lex-ordering
(define *order-equality-firstpref-strat* #f) ; use first-term-depth ordering
(define *sos-strategy-flag* #t) ; locally used flag: determines
    ; whether sos strategy is used in a pass
(define *keep-all-old-clauses-flag* #f) ; if #t, all clauses from old
       ; passes are kept for history
(define *auto-mode-flag* #f) ; if #t, automatic strategy selection

(define *stop-res-real-time* 0) ; time to stop res. If = 0, then
                                ; no stopping
(define *splitting-allowed-flag* #t) ; sos splitting: on or off
(define *max-kept-clauses-nr* 80000) ; the maximal number of kept clauses
(define *max-size-sum-nr* 2000000) ; the maximal sum of kept sizes

(define *display-info-level* 50) ; gradation:
   ; 20 : derivation found/not
   ; 30 : history shown
   ; 35 : statistics shown,
   ; 38 : strategy changes shown
   ; 40 : input clauses in final form
   ; 50 : given clauses
   ; 60 : kept clauses
   ; 63 ; back subsumption
   ; 65 : cutoff literals and clauses
   ; 70 : raw clauses
   ; 80 : removal reasons
   ; 90 : usable clauses

(define *display-info-char-flag* #f) ; special chars are output to trace search
(define *display-debug-info-flag* #f) ; debug forms of clauses also shown

(define *remove-all-tautologies-stratflag* #t)

(define *prop-unsat-strat* #f) ; propositional unsatisfiability

(define *hyper-wos-strategy-flag* #f) ;
(define *binary-hyper-sos-strategy* #f) ; first binary sos, then hyper
(define *hyper-contradiction-strategy* #f) ; iff #t, then kept clauses are
  ; immediately checked for full unit clash contradiction
(define *unit-clash-flag* #f) ; #t only during hyper-contradiction check

(define *factors-stratflag* #f)
(define *back-subsume-stratflag* #t)
(define *demodulate-with-clauses-stratflag* #t) ; clauses used for
        ; demodulation, giving max one fully clause-demodulated version
        ; as an extra
(define *havealready-timecounter* 0) ; counts the times havealready called:
                                     ; used for avoiding double work
(define *remove-identical-lit-stratflag* #t)
(define *paramodulation-allowed-flag* #t)   ; for user control only
(define *paramodulation-strategy-flag* #t)  ; actually used during search
(define *paramodulate-from-vars-stratflag* #f)
(define *para-pos-unit-strategy* #f)
(define *para-unit-strategy* #f)
(define *unit-strategy-flag* #f)
(define *double-strategy-flag* #f)
(define *hyper-strategy-flag* #f)
(define *bliksem-strategy-flag* #f)
(define *order-strategy-flag* #f)
(define *order-sos-strategy-flag* #f)
(define *name-order-strategy-flag* #f)
(define *exp-order-strategy-flag* #f)
(define *epr-order-strategy-flag* #f)
(define *firstlitpref-order-strategy-flag* #f)
(define *poslitpref-order-strategy-flag* #f)
(define *posweight-order-strategy-flag* #f)
(define *negweight-order-strategy-flag* #f)
(define *ordered-factor-flag* #f)
(define *neg-hyper-strategy-flag* #f)
(define *type-theory-stratflag* #f)
(define *int-logic-flag* #f)
(define *int-use-binres-flag* #f)
(define *classical-synth-stratflag* #f)
(define *int-use-hyperres-flag* #f)

(define *discr-tree-subsume-stratflag* #t)

(define *mace-nofuns-model-strat* #f)
(define *simple-model-strat* #f)
(define *mace-model-strat* #f)
(define *sem-model-strat* #f)
(define *random-model-strat* #f)


(define *discr-depth-first* #t) ;;; a search-guiding flag for discr-tree
                                ;;; unit subsumption. May be #t or #f.

(define *ordered-paramodulation-flag* #t) ;;; indicates whether
                              ;;; positive equalities are ordered

(define *dynamic-demodulation-flag* #f) ;;; indicates whether
                              ;;; dynamic demodulation is attempted
                              ;;; (this requires also the *ordered-param...
                              ;;; to be #t
(define *demodulate-lit-flag* #f) ;;; indicates whether lit demodulation
                              ;;;should be
                              ;;; attempted on initial and derived clauses
(define *back-demodulate-flag* #f)
(define *demodulate-flag* #f) ;;; indicates whether term demodulation should be
                              ;;; attempted on initial and derived clauses
(define *demodulate-constants-flag* #f) ;;; demodulation applied
       ;;; to constants? If *demodulate-flag* = #f, it must be #f too
(define *new-demodulators-flag* #f) ;;; demodulation to be applied
       ;;; to unchanged terms (ie ground)?
       ;;; If *demodulate-flag* = #f, it must be #f too


(define *user-criterias-post-flag* #f) ; set by the prover itself
(define *user-criterias-pre-flag* #f) ; set by the prover itself

(define *clause-usenr-limits* '())
(define *clause-usedepth-limits* '())
(define *clause-usestruct-limits* '())
(define *clause-keepsize-limit* 0)
(define *clause-midsearch-keepsize-limit* 0)
(define *clause-finalkeepsize-limit* 0)
(define *clause-keepdepth-limit* 0)
(define *clause-uselength-limit* 0)
(define *clause-keeplength-limit* 0)
(define *order-strategy-sos-flag* #f)
(define *global-sos-preference-ratio* 5)
(define *special-insertion-place* '()) ; used in back-demodulation

(define (init-one-pass-strategical-flags)
  (set! *sos-preference-ratio* *global-sos-preference-ratio*)
  (set! *split-strategy-flag* #f)
  (set! *sos-strategy-flag* #t)
  (set! *binary-hyper-sos-strategy* #f)
  (set! *hyper-wos-strategy-flag* #f)
  (set! *remove-all-tautologies-stratflag* #t)
  (set! *hyper-contradiction-strategy* #f)
  (set! *unit-clash-flag* #f)
  (set! *factors-stratflag* #f)
  (set! *back-subsume-stratflag* #t)

  (set! *demodulate-with-clauses-stratflag* #t)
  (set! *bliksem-strategy-flag* #f)
  (set! *prop-unsat-strat* #f)

  (set! *remove-identical-lit-stratflag* #t)
  (set! *paramodulation-allowed-flag* #t)
  (set! *paramodulation-strategy-flag* #t)
  (set! *paramodulate-from-vars-stratflag* #f)
  (set! *para-pos-unit-strategy* #f)
  (set! *para-unit-strategy* #f)
  (set! *unit-strategy-flag* #f)
  (set! *double-strategy-flag* #f)
  (set! *hyper-strategy-flag* #f)
  (set! *bliksem-strategy-flag* #f)
  (set! *order-strategy-flag* #f)
  (set! *order-strategy-sos-flag* #f)
  ; (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *name-order-strategy-flag* #f)
  (set! *epr-order-strategy-flag* #f)
  (set! *firstlitpref-order-strategy-flag* #f)
  (set! *poslitpref-order-strategy-flag* #f)
  (set! *posweight-order-strategy-flag* #f)
  (set! *negweight-order-strategy-flag* #f)
  (set! *ordered-factor-flag* #f)
  (set! *neg-hyper-strategy-flag* #f)
  (set! *type-theory-stratflag* #f)
  (set! *int-logic-flag* #f)
  (set! *int-use-binres-flag* #f)
  (set! *classical-synth-stratflag* #f)
  (set! *int-use-hyperres-flag* #f)
  (set! *discr-tree-subsume-stratflag* #t)
  (set! *discr-depth-first* #t)
  (set! *ordered-paramodulation-flag* #t)
  ; (set! *dynamic-demodulation-flag* #f)
  ; (set! *demodulate-lit-flag* #f)
  ; (set! *demodulate-flag* #f)
  ; (set! *demodulate-constants-flag* #f)
  ; (set! *new-demodulators-flag* #f)

  (set! *unit-equality-strat* #f)
  (set! *root-para-strat* #f)
  (set! *para-pos-unit-strategy* #f)
  (set! *para-unit-strategy* #f)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)


  (set! *user-criterias-post-flag* #f)
  (set! *user-criterias-pre-flag* #f)
  (set! *clause-keepdepth-limit* 0)
  (set! *clause-keeplength-limit* 0)
  (set! *clause-uselength-limit* 0)
  (set! *clause-finalkeepsize-limit* 0)

  ;(set! *clause-keepsize-limit* 0)

  (set! *clause-midsearch-keepsize-limit* 0)
  (set! *clause-usenr-limits* '())
  (set! *special-insertion-place* '())
  (set! *clause-usedepth-limits* '())
  (set! *clause-usestruct-limits* '()) )


; - - - - - - - - - -- - - - strategic functions - - - - - - -  -




(define *answer-litinf* (logior *litinf-noresolve-mask*
                                (logior *litinf-noparainto-mask*
                                        *litinf-isanswer-mask*)))

(define (empty-cl? cl)
  (if *int-logic-flag*
      (or (null? cl)
          (and (null? (cdr cl))
                  (eq? *int-outmost-label-fun-nr* (pure-fun (caar cl))) ))
      (let ((resflag #t))
        (do ((clpart cl (cdr clpart)))
            ((or (null? clpart) (not resflag))  resflag)
          (if (not (answer-pred? (caar clpart))) (set! resflag #f)) ))))


;(define (unit-cl? cl)
;  (if *classical-synth-stratflag*
;      (null? (cdr cl))
;      (let ((count 0))
;        (do ((clpart cl (cdr clpart)))
;            ((or (null? clpart) (eq? count 2))  (not (eq? count 2)))
;          (if (not (answer-pred? (caar clpart)))
;              (if (eq? count 0) (set! count 1) (set! count 2)) )))))

(define (unit-cl? cl)
   (null? (cdr cl)) )

(define (double-cl? cl)
   (or (null? (cdr cl))
       (null? (cddr cl)) ))

(define (first-nonans-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((or (null? clpart)
                 (not (answer-pred? (caar clpart))) )
                (if (null? clpart)
                    #f
             (car clpart) ))))

(define (member-nonans-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((or (null? clpart)
           (not (answer-pred? (caar clpart))) )
         (if (null? clpart)
             #f
             clpart) )))

(define (member-ans-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((or (null? clpart)
                 (answer-pred? (caar clpart)) )
               (if (null? clpart)
                   #f
             clpart) )))

(define (remove-ans-literals cl)
  (let* ((res (cons '() '()))
         (resend res) )
    (do ((clpart cl (cdr clpart)))
        ((null? clpart)
           (cdr res) )
      (if (not (answer-pred? (caar clpart)))
          (begin
            (set-cdr! resend (cons (car clpart) '()))
            (set! resend (cdr resend)) )))))

(define (ground-fullcl?        fullcl)
  (not (zero? (logand (caar fullcl) *cl-ground-flag*))) )

(define (positive-cl? cl)
  (let ((resflag #t))
    (do ((clpart cl (cdr clpart)))
        ((or (null? clpart) (not resflag))  resflag)
      (if (neg-fun? (caar clpart)) (set! resflag #f)) )))

(define (negative-cl? cl)
  (let ((resflag #t))
    (do ((clpart cl (cdr clpart)))
        ((or (null? clpart) (not resflag))  resflag)
      (if (not (neg-fun? (caar clpart))) (set! resflag #f)) )))

(define (mixed-cl? cl)
  (let ((negflag #f)
        (posflag #f) )
    (do ((clpart cl (cdr clpart)))
        ((or (null? clpart)
             (and negflag posflag) )
          (and negflag posflag) )
      (if (neg-fun? (caar clpart))
          (set! negflag #t)
          (set! posflag #t) ))))

(define (has-single-hyper-premiss? cl)
  (let ((count 0))
    (do ((clpart cl (cdr clpart)))
        ((or (null? clpart) (eq? count 2))  (not (eq? count 2)))
      (if (if *neg-hyper-strategy-flag*
              (and (pos-fun? (caar clpart)) (not (answer-pred? (caar clpart))))
              (neg-fun? (caar clpart)) )
          (if (eq? count 0) (set! count 1) (set! count 2)) ))))

(define (first-hyper-premiss-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((if *neg-hyper-strategy-flag*
           (and (pos-fun? (caar clpart)) (not (answer-pred? (caar clpart))))
           (neg-fun? (caar clpart)) )
         (car clpart) )))

(define (first-neg-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((neg-fun? (caar clpart))
         (car clpart) )))

(define (contains-eqlit-cl? cl)
  (do ((clpart cl (cdr clpart)))
      ((or (null? clpart) (pos-eq-pred? (pure-fun (caar clpart))))
         (not (null? clpart)) )))

(define (contains-eqorneg-cl? cl)
  (do ((clpart cl (cdr clpart)))
      ((or (null? clpart)
           (neg-fun? (caar clpart))
           (pos-eq-pred? (pure-fun (caar clpart))) )
         (not (null? clpart)) )))


(define (commutative-fun? fun) ;;; !!!!!!!! TMP !!!!!!!!! tmp always #f
  #f)

(define (reflexive-fun? fun)
  #f)

(define (removable-tautology-pr? litinfcl-pr)
  (if (and *remove-all-tautologies-stratflag*
           (tautology-pure-cl? (cdr litinfcl-pr)) )
      (begin
        (cond ((>= *display-info-level* 80)
                 (begin (newline)
                    (display "the derived clause removed as a tautology") ))
              (*display-info-char-flag*
                   (display "T") ))
        #t)
      #f))


(define (clause-pass-user-limits-post? purecl)
  (and (or (zero? *clause-keepsize-limit*) (pass-keepsize-limit? purecl))
       (or (zero? *clause-keepdepth-limit*) (pass-keepdepth-limit? purecl))
       (or (zero? *clause-keeplength-limit*)(pass-keeplength-limit? purecl)) ))

(define (pass-keepdepth-limit? purecl)
  (every? (lambda(lit)
            (<= (shiftright (logand *term-depth-mask* (car lit))
                           *term-depth-shift*)
                *clause-keepdepth-limit*))
          purecl))


(define (pass-keepsize-limit? purecl)
  (let ((sum 0)
        (tmp 0))
    (do ((part purecl (cdr part)))
        ((or (null? part)
             (> sum *clause-keepsize-limit*) )
          (<= sum *clause-keepsize-limit*) )
      (set! tmp (shiftright (logand *term-size-mask* (caar part))
                            *term-size-shift*))
      (if (>= tmp *term-size-max*)
          (set! tmp (compute-real-termsize (car part))) )
      (set! sum (+ sum tmp)) )))


(define (compute-real-termsize term)
  ;(newline)(display "compute-real-termsize on: ")(display term)
  (let ((sum 0)
        (tmp 0))
    (if (non-composite? term)
        1
        (begin
          (set! tmp (shiftright (logand *term-size-mask* (car term))
                                *term-size-shift*))
          (if (< tmp *term-size-max*)
              tmp
              (do ((part term (cdr part)))
                  ((null? part) (+ 1 sum))
                (set! sum (+ sum (compute-real-termsize (car part)))) ))))))

(define (compute-real-clausesize purecl)
  (let ((res 0))
    (do ((part purecl (cdr part)))
        ((null? part) res)
      (set! res (+ res (compute-real-termsize (car part)))) )))

(define (pass-keeplength-limit? purecl)
  (<= (length purecl) *clause-keeplength-limit*) )


(define (clause-pass-user-crit-post? inflist purecl hist)
  (and (or (null? *clause-usenr-limits*) (hist-pass-usenr-limits? hist))
       (or (null? *clause-usedepth-limits*) (hist-pass-usedepth-limits? hist))
       (or (null? *clause-usestruct-limits*)
           (hist-pass-usestruct-limits? hist)) ))

(define (hist-pass-usenr-limits? hist)
  (let ((countlst (map (lambda(x)(list (car x) 0 (cadr x)))
                       *clause-usenr-limits*)))
    (hist-count-check countlst hist) ))

(define (hist-count-check countlst hist)
  (if (pair? hist)
      (let ((constr (car hist)))
        (cond
         ((eq? constr 'hyper)
          (and (hist-count-check countlst (caddr hist))
               (every? (lambda(x)(hist-count-check countlst x))
                       (cdddr hist) )))
         ((eq? constr 'literal)
          (hist-count-check countlst (cadr hist)) )
         ((eq? constr 'input)
          (set! constr (assoc-memq (cadr hist) countlst))
          (if constr
              (if (eq? (cadr constr) (caddr constr))
                  #f
                  (begin (set-car! (cdr constr) (+ 1 (cadr constr)))
                         #t))
              #t))
         (else #t) ))
      #t))


(define (hist-pass-usedepth-limits? hist)
   (let ((constr (car hist))
         (given '())
         (limit1 '())
         (limit2 '()) )
    (cond
     ((eq? constr 'para)
       (set! limit1 (caddr hist))
       (set! limit2 (cadddr hist))
       (cond
         ((and (eq? 'input (car limit1))
               (begin
                 (set! given (assq (cadr limit1) *clause-usedepth-limits*))
                 given)
               (greater-clause-occ-depth? (cadr limit1) (cdr given) limit2) )
           #f)
         ((and (eq? 'input (car limit2))
               (begin
                 (set! given (assq (cadr limit2) *clause-usedepth-limits*))
                 given)
               (greater-clause-occ-depth? (cadr limit1) (cdr given) limit1) )
          #f)
         (else #t) ))
     ((eq? constr 'hyper)
         (set! given (assq (caddr hist) *clause-usedepth-limits*))
         (if (not given)
             #t
             (not (greater-clause-occ-depth?
                    (caddr hist) (cdr given) )
                  ....))))))



; *clause-usestruct-limits*

(define (hist-pass-usestruct-limits? hist)
  (let ((constr (car hist))
        (res #f)
        (one-found #f)
        (tmp '()) )
    (cond
     ((eq? constr 'hyper)
        (set! tmp (cadr (caddr hist))) ; tmp := nr of the nucleon
        (do ((place (member-if (lambda(x)(eq? (car x) tmp))
                               *clause-usestruct-limits*)
                    (member-if (lambda(x)(eq? (car x) tmp))
                               (cdr place) )))
            ((or (null? place)(not place) res)
                (if res #t (not one-found)) )
          (set! one-found #t)
          (if (hist-pass-usestruct-limits-hyper? hist (car place))
              (set! res #t) )))
     (else #t) )))

(define (hist-pass-usestruct-limits-hyper? hist limitcase)
  (let ((nr (cadr (caddr hist)))
        (res #t)
        (tmp '()) )
    (do ((els (cdddr hist) (cdr els)))
        ((or (null? els) (not res)) res)
      (if (and (pair? (car els))
               (begin (set! tmp (get-hyper-hist-el-limitnr (car els)))
                      (not (some? (lambda(x)
                                   (or (eq? x tmp)
                                       (and (pair? x) (eq? (car x) tmp)) ))
                                 (cdr limitcase) ))))
          (set! res #f) ))))

(define (get-hyper-hist-el-limitnr hist)
  (cond
    ((eq? 'literal (car hist))
        (get-hyper-hist-el-limitnr (cadr hist)) )
    ((eq? 'input (car hist))
        (cadr hist) )
    ((or (eq? 'hyper (car hist)) (eq? 'fact (car hist)))
        (get-hyper-hist-el-limitnr (caddr hist)) )
    (else 0) ))




;-------------------------------------------------------------------
;
;                    init nonstatistical globals
;
;-------------------------------------------------------------------


(define (init-globals)
  (init-one-pass-globals)
  (set! *unit-equality-strat* #f)
  (set! *root-para-strat* #f)
  (set! *para-pos-unit-strategy* #f)
  (set! *para-unit-strategy* #f)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *order-equality-firstpref-strat* #f)
  (set! *use-real-time* #f)
  (set! *run-forever-flag* #f)
  (if *full-proof-strategy-flag*
      (set! *keep-all-old-clauses-flag* #t)
      (set! *keep-all-old-clauses-flag* #f) )
  (set! *all-back-subsumed-clauses* '())
  (set! *default-time-limit* 300)
  (if (not *args-timelimit*)
      (set! *timelimit* #f)
      (set! *timelimit* *args-timelimit*) )
  (set! *foundflag* #f)
  (set! *sat-found-flag* #f)
  (set! *given-checkpoint-keptnrs* '())
  (set! *checkpoint-keptnrs* '())

  (set! *sos-strategy-flag* #t)
  (set! *bliksem-strategy-flag* #f)
  ;(set! *ans-prohibited-funsymbs* '())
  ;(set! *ans-prohibited-constsymbs* '())
  ;(set! *ans-prohibited-predsymbs* '())

  (set! *ans-prohibited-funnrs* '())
  (set! *ans-prohibited-constnrs* '())
  (set! *ans-prohibited-prednrs* '())

  (set! *discr-tree-list* '())
  (set! *unit-alist-list* '())
  (set! *sos-clauses-list-list* '())
  (set! *usable-clauses-list-list* '())
  (set! *lex-symbol-orderlst* '())

  (set! *unit-alist* '()) ; alist of all unit clauses, with assoc-keys being
  ; the pure predicate symbs of units.

  ;(set! *clause-keepsize-limit* 0)

  (set! *clause-finalkeepsize-limit* 0)
  (set! *clause-keepdepth-limit* 0)
  (set! *clause-keeplength-limit* 0)
  (set! *clause-uselength-limit* 0)

  (set! *prolog-style-variables* #t)
  (set! *const-num-table* '())
  (set! *const-num* 0)
  (set! *min-const-num-sub1* 100000)
  (set! *fun-num-table* '())
  (set! *fun-num* 0)
  (set! *min-fun-num-sub1* 0)
  (set! *max-nr-of-funs* 255)
  (set! *pred-num-table* '())
  (set! *pred-num* 0)
  (set! *min-pred-num-sub1* 0)
  (set! *max-nr-of-preds* 127)
  (set! *var-num-table* '())
  (set! *var-num* 0)
  (set! *min-var-num-sub1* 0)
  (set! *max-nr-of-vars-in-clause* 255) )

(define *one-pass-kept-clauses-firstnr* 1) ; the first kept clausnr of a pass

(define (init-one-pass-globals)
  (set! *stop-fast-loop-flag* #f)
  (set! *hyper-contradiction-strategy* #f)
  (set! *clause-usenr-limits* '())
  (set! *clause-usedepth-limits* '())
  (set! *clause-usestruct-limits* '())
  (set! *one-pass-kept-clauses-firstnr* (+ 1 *kept-clauses-nr*))

  (set! *foundflag* #f)      ; set to #t iff the empty clause is found
  (set! *sat-found-flag* #f)
  (set! *answer-cl* '())     ; set to a pure answer-clause if empty is found
  (set! *answer-history* '()) ; set to history of the answer clause
  (set! *too-many-vars-flag* #f) ; set when such a clause is generated

  (set! *lex-symbol-orderlst* '())
  (set! *sos-clauses-list* '())
  (set! *sos-clauses-list-end* '())   ; last pair of *sos-clauses-list
  (set! *ans-input-sos* '()) ; sos before removing answer-literals
  (set! *usable-clauses-list* '())
  (set! *usable-clauses-list-end* '()) ; last pair of *usable-clauses-list*
  (set! *subsumption-list* (list (list 0)) )
  (set! *unit-alist* '()) ; alist of all unit clauses, with assoc-keys being
  ; the pure predicate symbs of units.
  (set! *discr-tree* (list '()) )
  (set! *nonunit-discr-tree* (list '()) )
  (set! *discr-tree-demod-terms* (list '()) ) ; a special discrimination tree
  ; for demodulating terms
  (set! *discr-tree-demod-atoms* (list '()))  ; a special discrimination tree
  ; for demodulating atoms

  (set! *sos-clauses-list-list* '())
  (set! *usable-clauses-list-list* '()) )

(define (init-one-pass-globals-combinedstrategy)
  (set! *stop-fast-loop-flag* #f)
  (set! *hyper-contradiction-strategy* #f)
  (set! *clause-usenr-limits* '())
  (set! *clause-usedepth-limits* '())
  (set! *clause-usestruct-limits* '())
  (set! *one-pass-kept-clauses-firstnr* (+ 1 *kept-clauses-nr*))

  (set! *foundflag* #f)      ; set to #t iff the empty clause is found
  (set! *sat-found-flag* #f)
  (set! *answer-cl* '())     ; set to a pure answer-clause if empty is found
  (set! *answer-history* '()) ; set to history of the answer clause
  (set! *too-many-vars-flag* #f) ; set when such a clause is generated

  (set! *lex-symbol-orderlst* '())
  (set! *sos-clauses-list* '())
  (set! *sos-clauses-list-end* '())   ; last pair of *sos-clauses-list
  (set! *ans-input-sos* '()) ; sos before removing answer-literals
  (set! *usable-clauses-list* '())
  (set! *usable-clauses-list-end* '()) ; last pair of *usable-clauses-list*
  (set! *subsumption-list* (list (list 0)) )
  ;;;(set! *unit-alist* '()) ; alist of all unit clauses, with assoc-keys being
  ; the pure predicate symbs of units.
  (set! *discr-tree* (list '()) )
  ;;;(set! *nonunit-discr-tree* (list '()) )
  ;;;(set! *discr-tree-demod-terms* (list '()) ) ; a special discrimination tree
  ; for demodulating terms
  ;;;(set! *discr-tree-demod-atoms* (list '()))  ; a special discrimination tree
  ; for demodulating atoms

  ;;;(set! *sos-clauses-list-list* '())
  ;;;(set! *usable-clauses-list-list* '()) )
  )

; - - - - - - - - - - - - - - - - - - - - -




;-------------------------------------------------------------------
;
;                        statistics
;
;-------------------------------------------------------------------

(define *derived-clauses-nr* 0)
(define *kept-clauses-nr* 0)
(define *kept-size-sum-nr* 0)  ; sum of clause sizes currently kept
(define *kept-mid-nucleus-nr* 0)
(define *kept-new-demodulators-nr* 0)

(define *forward-subsumed-unit-nr* 0)
(define *forward-subsumed-double-nr* 0)
(define *forward-subsumed-overdouble-nr* 0)

(define *forward-subsumed-nonunit-nr* 0)

(define *back-subsumed-nr* 0)

(define *simple-unit-cutoff-nr* 0)
(define *full-unit-cutoff-nr* 0)
(define *double-unit-cutoff-nr* 0)

(define *proof-start-real-time* 0)
(define *proof-start-run-time* 0)

(define *subs-tried-nr* 0)
(define *subs-length-fail-nr* 0)
(define *subs-strength-fail-nr* 0)
(define *subs-aux-strength-fail-nr* 0)
(define *subs-gclpreds-fail-nr* 0)
(define *subs-funlist-fail-nr* 0)
(define *subs-reg-subs-tried-nr* 0)
(define *subs-reg-subs-fail-nr* 0)

(define (init-stats)
  (set! *given-clause-nr* 0)

  (set! *derived-clauses-nr* 0)
  (set! *kept-clauses-nr* 0)
  (set! *kept-size-sum-nr* 0)
  (set! *last-usable-cl-nr* 0)
  (set! *kept-mid-nucleus-nr* 0)
  (set! *kept-new-demodulators-nr* 0)

  (set! *forward-subsumed-unit-nr* 0)
  (set! *forward-subsumed-nonunit-nr* 0)
  (set! *forward-subsumed-double-nr* 0)
  (set! *forward-subsumed-overdouble-nr* 0)

  (set! *back-subsumed-nr* 0)
  (set! *havealready-timecounter* 0)
  (set! *simple-unit-cutoff-nr* 0)
  (set! *full-unit-cutoff-nr* 0)
  (set! *double-unit-cutoff-nr* 0)

  (set! *proof-start-real-time* (get-internal-real-time))
  (set! *proof-start-run-time* (get-internal-run-time))

  (set! *subs-tried-nr* 0)
  (set! *subs-length-fail-nr* 0)
  (set! *subs-strength-fail-nr* 0)
  (set! *subs-aux-strength-fail-nr* 0)
  (set! *subs-gclpreds-fail-nr* 0)
  (set! *subs-funlist-fail-nr* 0)
  (set! *subs-reg-subs-tried-nr* 0)
  (set! *subs-reg-subs-fail-nr* 0) )

(define (show-stats)
  (show-main-stats)
  (show-subsumption-stats) )

(define (show-main-stats)
  (newline)
  (display "Global statistics over all passes: ")(newline)(newline)
  (display " given clauses:    ") (display *given-clause-nr*)(newline)
  (display " derived clauses:   ") (display *derived-clauses-nr*)(newline)
  (display " kept clauses:      ") (display *kept-clauses-nr*)(newline)
  (display " kept size sum:     ") (display *kept-size-sum-nr*)(newline)
  (display " kept mid-nuclei:   ") (display *kept-mid-nucleus-nr*)(newline)
  (display " kept new demods:   ") (display *kept-new-demodulators-nr*)
  (newline)

  (display " forw unit-subs:    ") (display *forward-subsumed-unit-nr*)
  (newline)
  (display " forw double-subs: ") (display *forward-subsumed-double-nr*)
  (newline)
  (display " forw overdouble-subs: ")
  (display *forward-subsumed-overdouble-nr*)
  (newline)
  (display " backward subs:     ") (display *back-subsumed-nr*)(newline)

  (display " fast unit cutoff:  ") (display *simple-unit-cutoff-nr*)(newline)
  (display " full unit cutoff:  ") (display *full-unit-cutoff-nr*)(newline)
  (display " dbl  unit cutoff:  ") (display *double-unit-cutoff-nr*)(newline)

  (display " real runtime    :  ")
  (display-time-seconds (- (get-internal-real-time) *proof-start-real-time*))
  (newline)
  (display " process. runtime:  ")
  (display-time-seconds (- (get-internal-run-time) *proof-start-run-time*))
  (newline) )

(define (show-subsumption-stats)
  (display "specific non-discr-tree subsumption statistics: ")(newline)
  (display " tried:           ")(display *subs-tried-nr*)(newline)
  (display " length fails:    ")(display *subs-length-fail-nr*)(newline)
  (display " strength fails:  ")(display *subs-strength-fail-nr*)(newline)
  (display " predlist fails:  ")(display *subs-gclpreds-fail-nr*)(newline)
  (display " aux str. fails:  ")(display *subs-aux-strength-fail-nr*)(newline)
  (display " by-lit fails:    ")(display *subs-funlist-fail-nr*)(newline)
  (display " full subs tried: ")(display *subs-reg-subs-tried-nr*)(newline)
  (display " full subs fail:  ")(display *subs-reg-subs-fail-nr*)(newline) )


(define (display-time-seconds units)
  (let* ((full (quotient units internal-time-units-per-second))
         (rest (remainder units internal-time-units-per-second)) )
         ; (rest100s (quotient (* rest 10) 6)) )
    (display full)(display #\.)(display rest) ))

(define (mkstring-time-seconds units)
  (let* ((full (quotient units internal-time-units-per-second))
         (rest (remainder units internal-time-units-per-second)) )
         ; (rest100s (quotient (* rest 10) 6)) )
    (string-append
      (number->string full)
      (string-append "." (number->string rest)) )))


;=====================================================================
;
;
;                          I/O algorithms
;
;
;====================================================================



;--------------------------------------------------------------------
;
;                   top-level procedures
;
;-------------------------------------------------------------------


; fsat is a top-level function taking a file as an input and starting
; the satisfiability search process


(define *run-id* "") ; run-id will be a an appended string of real time plus
                     ; problem file name: used to create unique temporary
                     ; files, normally in /tmp
(define *gandalf-program-arguments* '())
(define *intepreter-mode* #f)


; gandalf-remove-tempfiles is called directly by the modified
; scm interpreter in case it sees an alarm or xcpu signal

(define (gandalf-remove-tempfiles)
 (if (and *prop-conversion-file-name*
          (not (null? *prop-conversion-file-name*))
          (not (equal? *prop-conversion-file-name* "")) )
  (begin
    (system (string-append "rm " *prop-conversion-file-name*))
    (system (string-append "rm " *prop-conversion-temp-file-name*))
    (system (string-append "rm " *prop-conversion-header-file-name*))
    (system (string-append "rm " *prop-conversion-result-file-name*)) )))

(define *gandalf-pid* 0)


; gandalf-set-pid is called directly by the modified
; scm interpreter when it sets itself up

(define (gandalf-set-pid i)
  (set! *gandalf-pid* i) )


(define (fsat filename)
  (set! *sat-strategy-flag* #t)
  (fprove filename) )

(define (fprove filename)
  (set! *run-id*
       (string-append
               (number->string *gandalf-pid*)
               (string-append
                      (string-lastn (number->string (current-time)) 5)
                      (string-firstn (path-string-endpart filename) 10) )))
  (set! *gandalf-program-arguments* (program-arguments))
  (if (or (string=? "scmlit" (path-string-endpart (car *gandalf-program-arguments*)))
          (string=? "scmlit.exe" (path-string-endpart (car *gandalf-program-arguments*))) )
      (set! *intepreter-mode* #t)
      (set! *intepreter-mode* #f) )
  (init-globals)

  ; (set! *sat-strategy-flag* #f) ; satisfiability search
  (set! *sos-exhausted-flag* #f)
  (set! *auto-mode-flag* #t)  ; default in auto mode
  (if *sat-strategy-flag*
     (set! *display-info-level* 30)
     (set! *display-info-level* 30) )

  ;(newline)(display "starting to preprocess file")(newline)

  (preprocess-otter-file filename)

  ;(newline)(display "ended preprocessing file")(newline)

  (let* ((tmpname *otter-tmpfile*)
               ; (string-append
               ;            *otter-tmpfile-prefix*
               ;            filename *otter-tmpfile-suffix*))

         (rawlst (read-processed-otter-file tmpname))
         ;(rawlst (read-processed-otter-file "out.tmp"))

         ; (dummy (begin (pretty-print rawlst) #f))

         (clauselst (read-otter-list-clauses rawlst))
         (readylst '())
         (res #f) )

    (set! *ans-prohibited-constsymbs* '())
    (set! *ans-prohibited-predsymbs* '())
    (set! *ans-prohibited-funsymbs* '())

    ;(pretty-print clauselst)

    (set! readylst (sort-out-otter-list-clauses clauselst))

    ;(newline)
    ;(pretty-print readylst)
    ;(newline)

    ;(newline)(display "starting to curry")(newline)

    ;(if  *sat-strategy-flag*
    ;     (set! readylst (groundprocess-input-readylst readylst)) )

    ;(display "groundprocessed readylst: ")(newline)
    ;(pretty-print readylst)
    ;(newline)

    (set! readylst (curry-input-readylst readylst))

    ;(newline)(display "starting to prove")(newline)

    (set! res (prove readylst 99999999 50))

    (if res
        (begin
           (if #f ; (>= *display-info-level* 30)
               (begin (newline)
                      (if (not (null? *answer-cl*))
                          (begin (display "answer-clause: ")
                                 (print-pure-clause *answer-cl*) (newline) ))
                      (if (not *prop-unsat-strat*)
                          (show-history *answer-history*)) ))
           (begin
              (newline)
              (display "***GANDALF_FOUND_A_REFUTATION***")
              (newline) )
            #t)
        (if (>= *display-info-level* 30)
            (if (and *sat-strategy-flag* *sos-exhausted-flag* *complete-unsat-strategy*)
              (begin (newline)
                  (if (not *prop-clauses-generation-limited-flag*)
                      (begin (display "Clause set found to be **SATISFIABLE** (satisfiable).")
                             (newline)
                             (set! *sat-found-flag* #t) ))
                  #f)
              (if (eq? *answer-history* 'not-enough-time)
                (begin (newline)
                  (display "time limit exhausted: proof search terminated.")
                  (newline)
                  #f)
                (begin (newline)
                  (display "search space exhausted.")
                  (newline)
                  (if (and *sat-strategy-flag*
                           *sos-exhausted-flag*
                           *complete-unsat-strategy*)
                      (begin
                         (if (not *prop-clauses-generation-limited-flag*)
                             (display "clause set was found to be satisfiable (**SATISFIABLE**).") )
                         (newline) ))
                  #f)))
            #f))
    (if (>= *display-info-level* 35) (show-stats))

    ;(system
    ;  (string-append "echo \'"
    ;    (string-append filename
    ;       (string-append (if res " 1 " " 0 ")
    ;         (string-append
    ;           (string-append (number->string *max-var-stat-pre*) " ")
    ;          (string-append
    ;           (string-append (number->string *max-var-stat-post*) " ")
    ;
    ;          (string-append
    ;            (string-append
    ;               (mkstring-time-seconds
    ;                  (- (get-internal-run-time) *proof-start-run-time*))
    ;                "\' | cat >> out.txt"))))))))

    (if res #t #f) ))


(define (sprove filename)
   (set! *sat-strategy-flag* #t)
   (fprove filename) )

(define (path-string-endpart str)
  (let ((origstr str)
        (resstr "") )
    (do ((i (- (string-length str) 1) (- i 1)))
        ((or (< i 0)
             (char=? #\/ (string-ref str i)) )
           resstr)
      (set! resstr (string-append (string (string-ref str i)) resstr)) )))

(define (path-string-startpart str)
  (let ((endpart (path-string-endpart str)))
    (substring str 0 (- (string-length str) (string-length endpart))) ))

(define (string-firstn str n)
  (let ((l (string-length str)))
    (if (> n l)
        str
        (substring str 0 n) )))

(define (string-lastn str n)
  (let ((l (string-length str)))
    (if (> n l)
        str
        (substring str (- l n) l) )))

(define (use-input-settings raw-input)
  (let ((setlist (assq 'set raw-input))
        (clearlist (assq 'clear raw-input))
        (assignlist (assq 'assign raw-input)) )
    (if (not setlist) (set! setlist '()))
    (if (not clearlist) (set! clearlist '()))
    (if (not assignlist) (set! assignlist '()))
    (for-each
      (lambda(el)
        ;(display el)
        (cond
          ((eq? el 'binary_exporder_res)
            (set! *auto-mode-flag* #f)
            (init-exp-order-strat) )
          ((eq? el 'binary_eprorder_res)
            (set! *auto-mode-flag* #f)
            (init-epr-order-strat) )
          ((eq? el 'binary_posweight_kb_big)
            (set! *auto-mode-flag* #f)
            (init-posweight-ordered-binres-kb-big-strat) )
          ((eq? el 'binary_posweight_kb_small)
            (set! *auto-mode-flag* #f)
            (init-posweight-ordered-binres-kb-small-strat) )
          ((eq? el 'binary_posweight_lex_big)
            (set! *auto-mode-flag* #f)
            (init-posweight-ordered-binres-lex-big-strat) )
          ((eq? el 'binary_posweight_lex_small)
            (set! *auto-mode-flag* #f)
            (init-posweight-ordered-binres-kb-big-strat) )
          ((eq? el 'binary_posweight_firstpred)
            (set! *auto-mode-flag* #f)
            (init-posweight-ordered-firstpref-strat) )
         ((eq? el 'hyper_order_uniteq_res)
            (set! *auto-mode-flag* #f)
            (init-ordered-hyper-uniteq-strat) )
         ((eq? el 'binary_unit_uniteq_res)
            (set! *auto-mode-flag* #f)
            (init-unit-uniteq-strat) )
         ((eq? el 'binary_nameorder_uniteq_res)
            (set! *auto-mode-flag* #f)
            (init-name-ordered-binres-uniteq-strat) )
         ((eq? el 'binary_order_uniteq_res)
            (set! *auto-mode-flag* #f)
            (init-ordered-binres-uniteq-strat) )
         ((eq? el 'binary_order_sos_res)
            (set! *auto-mode-flag* #f)
            (init-ordered-binres-sos-strat) )
         ((eq? el 'binary_nameorder_sos_res)
            (set! *auto-mode-flag* #f)
            (init-name-ordered-binres-sos-strat) )
         ((eq? el 'binary_uniteq_res)
            (set! *auto-mode-flag* #f)
            (init-binres-uniteq-strat) )
         ((eq? el 'binary_posweightorder_sos_res)
            (set! *auto-mode-flag* #f)
            (init-posweight-ordered-binres-sos-strat) )
         ((eq? el 'binary_negweightorder_sos_res)
            (set! *auto-mode-flag* #f)
            (init-negweight-ordered-binres-sos-strat) )
         ((eq? el 'binary_order_fulleq_res)
            (set! *auto-mode-flag* #f)
            (init-ordered-binres-fulleq-strat) )
         ((eq? el 'auto)
            (set! *auto-mode-flag* #t) )
         ((eq? el 'hyper_res)
            (set! *auto-mode-flag* #f)
            (init-hyper-strat) )
         ((eq? el 'hyper_order_res)
            (set! *auto-mode-flag* #f)
            (init-ordered-hyper-strat) )
         ((eq? el 'binary_res)
            (set! *auto-mode-flag* #f)
            (init-binres-strat) )
         ((eq? el 'binary_order_res)
            (set! *auto-mode-flag* #f)
            (init-ordered-binres-strat) )
         ((eq? el 'binary_nameorder_res)
            (set! *auto-mode-flag* #f)
            (init-name-ordered-binres-strat) )
         ((eq? el 'binary_weightorder_sos_res)
            (set! *auto-mode-flag* #f)
            (init-weight-ordered-binres-sos-strat) )
         ((eq? el 'binary_weightorder_res)
            (set! *auto-mode-flag* #f)
            (init-weight-ordered-binres-strat) )
         ((eq? el 'binary_posweightorder_res)
            (set! *auto-mode-flag* #f)
            (init-posweight-ordered-binres-strat) )
         ((eq? el 'binary_negweightorder_res)
            (set! *auto-mode-flag* #f)
            (init-negweight-ordered-binres-strat) )
         ((eq? el 'binary_hypersos_res)
            (set! *auto-mode-flag* #f)
            (init-binary-hyper-sos-strat) )
         ((eq? el 'binary_unit_res)
            (set! *auto-mode-flag* #f)
            (init-unit-strat) )
         ((eq? el 'binary_double_res)
            (set! *auto-mode-flag* #f)
            (init-double-strat) )
         ((eq? el 'binary_posweightunit_res)
            (set! *auto-mode-flag* #f)
            (init-posweight-ordered-unit-strat) )
         ((eq? el 'binary_negweightunit_res)
            (set! *auto-mode-flag* #f)
            (init-negweight-ordered-unit-strat) )
         ((eq? el 'binary_synth_res)
            (set! *auto-mode-flag* #f)
            (init-synth-strat) )
         ((eq? el 'eq_order_lex)
            (clear-eq-order-strat)
              (set! *order-equality-lex-strat* #t) )
         ((eq? el 'eq_order_kb)
            (clear-eq-order-strat)
              (set! *order-equality-kb-strat* #t) )
         ((eq? el 'eq_order_firstpref)
            (clear-eq-order-strat)
              (set! *order-equality-firstpref-strat* #t) )
         ((eq? el 'eq_order_depth)
            (clear-eq-order-strat)
              (set! *order-equality-lex-strat* #f) )
         ((eq? el 'check_hypercontra)
            (set! *hyper-contradiction-strategy* #t) )
         ((eq? el 'demodulate)
            (turn-on-demodulation) )
         ((eq? el 'back_demodulate)
            (turn-on-demodulation)
            (set! *back-demodulate-flag* #t) )
         ((eq? el 'split)
           (set! *split-strategy-flag* #t) )
         ((eq? el 'wall_time)
            (set! *use-real-time* #t) )
         ((eq? el 'run_forever)
            (set! *run-forever-flag* #t) )
         ((eq? 'print_tracechar el)
            (set!  *display-info-char-flag* #t) )
         ((eq? 'print_debug el)
            (set! *display-debug-info-flag* #t) )
         ((eq? el 'otter_style_variables)
            (set!  *prolog-style-variables* #f) )
         ((eq? el 'prolog_style_variables)
            (set!  *prolog-style-variables* #t) )))

      setlist)
    ;(newline)
    ;(display "clearlist: ")(display clearlist) (newline)
    ;(display "setlist: ")(display setlist) (newline)
    ;(display "assignlist: ")(display assignlist) (newline)
    (for-each
      (lambda(el)
        (cond
         ;((not (pair? el)))
         ;((not (pair? (cdr el))))
         ((eq? el 'prolog_style_variables)
            (set! *prolog-style-variables* #f) )
         ((eq? el 'split)
           (set! *splitting-allowed-flag* #f) )
         ((eq? el 'paramodulate)
           (set! *paramodulation-allowed-flag* #f) )
         ((eq? el 'demodulate)
            (turn-off-demodulation) )))
      clearlist)
    (for-each
       (lambda(el)
         (cond
          ((not (pair? el)))
          ((not (pair? (cdr el))))
          ((eq? 'preference_ratio (car el))
             (set! *global-sos-preference-ratio* (cadr el)) )
          ((eq? 'checkpoints (car el))
             (set! *given-checkpoint-keptnrs* (cdadr el)) )
          ((eq? 'max_kept (car el))
             (set! *max-kept-clauses-nr* (cadr el)) )
          ((eq? 'max_size_sum (car el))
             (set! *max-size-sum-nr* (cadr el)) )
          ((eq? 'max_seconds (car el))
             ; (set! *total-time-given* #t)
             (if (not *timelimit*)
                 (set! *timelimit* (number->string (cadr el))) ))
          ((eq? 'print_level (car el))
             (set! *display-info-level* (cadr el)) )
          ((eq? 'max_literals (car el))
             (set! *clause-keeplength-limit* (cadr el)) )
          ((eq? 'max_weight (car el))
             (set! *clause-keepsize-limit* (cadr el)) )
          ((eq? 'max_depth (car el))
             (set! *clause-keepdepth-limit* (cadr el)) )))
       assignlist) ))



;--------------------------------------------------------------------
;
;        Reading and parsing otter-syntax clauses from file
;
;--------------------------------------------------------------------


;(define (mtest)
;  (preprocess-otter-file "lifsch.in")
;  (let* ((lst (read-processed-otter-file "gandalf.tmp")))
;    (pretty-print lst)
;    (let* ((res (read-otter-list-clauses lst)))
;      (newline)
;      (for-each (lambda(x)(display x)(newline)) res) )))


(define (sort-out-otter-list-clauses otterlst)
  (let* ((soslist '())
         (usablelist '())
         (sosformulalist '())
         (sosusablelist '())
         (axiomslist '())
         (demodlist '())
         (commandlist '())
         (setlist '())
         (clearlist '())
         (assignlist '())
         (cl '())
         (curlist '())
         (eol-reached #t)
         (usableformulalist '())
         (res '()) )
    (do ((part otterlst (cdr part)))
        ((null? part)
           ;(newline)
           ;(display otterlst)
           ;(newline)
           ;(display soslist)
           ;(newline)
           ;(display demodlist)
           ;(quit)

           (if (not (null? soslist))
               (set! res
                     (cons (cons 'sos (reverse (prop-conv soslist))) res)) )
           (if (not (null? usablelist))
               (set! res
                 (cons (cons 'usable (reverse (prop-conv usablelist))) res)) )
           (if (not (null? sosformulalist))
               (set! res
                     (cons (cons 'sosformula (reverse
                                              (prop-conv sosformulalist))) res)) )
           (if (not (null? usableformulalist))
               (set! res
                 (cons (cons 'usableformula (reverse
                                             (prop-conv usableformulalist))) res)) )
           (if (not (null? axiomslist))
               (set! res (cons (cons 'axioms (reverse axiomslist)) res)) )
           (if (not (null? demodlist))
               (set! res (cons (cons 'demodulators (reverse demodlist)) res)) )
           (if (not (null? setlist))
               (set! res (cons (cons 'set setlist) res)) )
           (if (not (null? clearlist))
               (set! res (cons (cons 'clear clearlist) res)) )
           (if (not (null? assignlist))
               (set! res (cons (cons 'assign assignlist) res)) )
           (if (not (null? commandlist))
               (set! res (cons (cons 'command commandlist) res)) )
           res)
      (set! cl (car part))
      ; (display cl)(newline)
      (cond
       ((not (pair? cl))
          (report-error "syntax error in input file" cl) )
       ((eq? (car cl) 'end_of_list)
          (set! eol-reached #t)
          (set! curlist #f) )
       ;((not (pair? (car cl)))
       ;          (report-error "syntax error in input file" cl) )
       ((and (pair? (car cl))(eq? 'assign (caar cl))(eq? 'prohibit_ans_pred (cadar cl)))
          (set! *ans-prohibited-predsymbs* (caddar cl)) )
       ((and (pair? (car cl))(eq? 'assign (caar cl))(eq? 'prohibit_ans_fun (cadar cl)))
          (set! *ans-prohibited-funsymbs* (caddar cl)) )
       ((and (pair? (car cl))(eq? 'assign (caar cl))(eq? 'prohibit_ans_const (cadar cl)))
          (set! *ans-prohibited-constsymbs* (caddar cl)) )
       ((and (null? (cdr cl))
             eol-reached
             (and (pair? (car cl)))
             (eq? 'list (caar cl))
             (eq? 'sos (cadar cl)) )
          (set! eol-reached #f)
          (set! curlist 'sos) )
       ((and (null? (cdr cl))
             eol-reached
             (and (pair? (car cl)))
             (eq? 'list (caar cl))
             (eq? 'usable (cadar cl)) )
          (set! eol-reached #f)
          (set! curlist 'usable) )
       ((and (null? (cdr cl))
             eol-reached
             (and (pair? (car cl)))
             (eq? 'formula_list (caar cl))
             (eq? 'sos (cadar cl)) )
          (set! eol-reached #f)
          (set! curlist 'sosformula) )
       ((and (null? (cdr cl))
             eol-reached
             (and (pair? (car cl)))
             (eq? 'formula_list (caar cl))
             (eq? 'usable (cadar cl)) )
          (set! eol-reached #f)
          (set! curlist 'usableformula) )
       ((and (null? (cdr cl))
             eol-reached
             (and (pair? (car cl)))
             (eq? 'list (caar cl))
             (eq? 'axioms (cadar cl)) )
          (set! eol-reached #f)
          (set! curlist 'axioms) )
       ((and (null? (cdr cl))
             eol-reached
             (and (pair? (car cl)))
             (eq? 'list (caar cl))
             (eq? 'demodulators (cadar cl)) )
          (set! eol-reached #f)
          (set! curlist 'demodulators) )
       ((eq? curlist 'sos)
          (set! soslist (cons cl soslist)) )
       ((eq? curlist 'usable)
          (set! usablelist (cons cl usablelist)) )
       ((eq? curlist 'sosformula)
          (set! sosformulalist (cons cl sosformulalist)) )
       ((eq? curlist 'usableformula)
          (set! usableformulalist (cons cl usableformulalist)) )
       ((eq? curlist 'axioms)
          (set! axiomslist (cons cl axiomslist)) )
       ((eq? curlist 'demodulators)
          (set! demodlist (cons cl demodlist)) )

       ((not (pair? (car cl)))
                 (report-error "syntax error in input file" cl) )
       ((eq? 'set (caar cl))
          (if (not (pair? (cdar cl)))
              (report-error "syntax error in input file" cl) )
          (set! setlist (cons (cadar cl) setlist)) )
       ((eq? 'clear (caar cl))
          (if (not (pair? (cdar cl)))
              (report-error "syntax error in input file" cl) )
          (set! clearlist (cons (cadar cl) clearlist)) )
       ((eq? 'assign (caar cl))
          (if (or (not (pair? (cdar cl)))
                  (not (pair? (cddar cl))) )
              (report-error "syntax error in input file" cl) )
          (set! assignlist (cons (cdar cl) assignlist)) )
       (else
          (set! commandlist (cons cl commandlist)) )))))

(define (prop-conv lst)
  (map (lambda(cl)
         (map (lambda(lit)
                (if (pair? lit)
                    lit
                    (list lit 0) ))
              cl))
       lst))

; read-otter-list-clauses makes a list of lisp-syntax clauses from
; the list of clauses otterlst read in by read-processed-otter-file

(define (read-otter-list-clauses otterlst)
  (let* ((part otterlst)
         (clause-rest '())
         (res '()) )
    (set! *parse-formula-context* #f)
    (do ()
        ((null? part) (reverse res))
      (set! clause-rest (parse-otter-clause part))
      (set! res (cons (car clause-rest) res))
      (set! part (cdr clause-rest)) )))


; parse-otter-clause parses a clause from an input list.
; it returns a pair (parsed-clause . rest)

(define *otter-or-symbol* (string->symbol "|"))

(define *parse-formula-context* #f)

(define (parse-otter-clause lst)
  (let ((part lst)
        (res '())
        (endflag #f)
        (tmp1 '()))
    (do ()
        (endflag
           (if *parse-formula-context*
               (cons (list (make-prefix-formula (reverse res))) part)
               (cons (reverse res) part)) )

      (if (and (pair? part)
               (eq? (car part) 'formula_list))
          (set! *parse-formula-context* #t))
      (if (and (pair? part)
                     (eq? (car part) 'end_of_list) )
          (set! *parse-formula-context* #f))

      ;(if (or (equal? res '(usable formula_list))
      ;              (equal? res '(sos formula_list)) )
      ;          (set! *parse-formula-context* #t) )

      (if (and (pair? part) (pair? (cdr part)) (pair? (cddr part))
               (or (eq? '= (cadr part))
                   (eq? '!= (cadr part)) ))
          (begin
            (if (eq? '!= (cadr part))
                (set! tmp1 (string->symbol
                  (string-append "-" (symbol->string *eq-symbol*)) ))
                (set! tmp1 *eq-symbol*) )
            (set! part (append (list tmp1
                                     (list (car part) '! (caddr part)) )
                               (cdddr part) )) ))

      (if *parse-formula-context*

                 (cond ((eq? *otter-dot-repl* (car part))
                  (set! endflag #t) )
                (else
                 (set! res (cons (car part) res)) ))

          ;(if (and (pair? (car part)) (memq '= (car part)))
          ;          (begin (set! tmp1 (prepare-read-eqlit (car part) #t))
          ;                 (set! part (append tmp1 (cdr part))) ))
          ; (if (and (pair? (car part)) (memq '!= (car part)))
          ;          (begin (set! tmp1 (prepare-read-eqlit (car part) #f))
          ;                 (set! part (append tmp1 (cdr part))) ))
          (cond ((eq? *otter-dot-repl* (car part))
                 (set! endflag #t) )
                ((eq? *otter-or-symbol* (car part)))
                ((or (symbol? (car part)) (number? (car part)))
                 (cond ((null? (cdr part))
                        (report-error "syntax error in input file") )
                       ((eq? *otter-dot-repl* (cadr part))
                        (set! res (cons (car part) res)) )
                       ((eq? *otter-or-symbol* (cadr part))
                        (if (null? (cddr part))
                            (report-error "syntax error in input file")
                            (set! res (cons (car part) res)) ))
                       ((not (pair? (cadr part)))
                        (report-error "syntax error in input file") )
                       (else
                        (let ((tmp (parse-otter-explist (cadr part))))
                          (set! res (cons (cons (car part) tmp) res))
                          (set! part (cdr part)) ))))
                ((pair? (car part))
                 (report-error "syntax error in input file") )))

      (if (not (null? part)) (set! part (cdr part))) )))


;-----------------------------------------------------------------------
;
;                       FORMULA CONVERSIONS TO CNF
;
;
;-----------------------------------------------------------------------



(define (make-prefix-formula form)
  form)

;  (let ((tmp '()))
;    (display form)
;    (cond
;     ; atoms
;     ((not (pair? form)) form)
;     ; literals in list
;     ((and (not (pair? (car form) ))
;           (pair? (cdr form))
;           (pair? (cadr form))
;           (null? (cddr form)) )
;        (set! tmp (make-prefix-formula-term-aux (cons (car form) (cadr form)) #f))
;        (pretty-print tmp)
;        (read-char)
;        (set! tmp (move-not-and-equiv tmp *and-symbol*))
;        (pretty-print tmp)
;        (read-char)
;        (set! tmp (skolemize-formula tmp))
;        (pretty-print tmp)
;        (read-char)
;             (set! tmp (flatten-prefix-formula tmp))
;        (pretty-print tmp)
;        (read-char)
;        tmp)
;     (else
;        (display "formula syntax error: ")
;        (display form) ))))


(define  (make-prefix-formula-term-aux term termflag)
  ;(newline)
  ;(display "term: ")
  ;(display term)
  ;(read-char)
  ; (newline)
  (cond
   ((not (pair? term))
      term)
   ((list? term)
      (if (memq *otter-comma-repl* term)
          (set! term (remove-eq *otter-comma-repl* term)) )
      (cond ((and
              (not termflag)
              (or (eq? (car term) *exists-symbol*)
                  (eq? (car term) *all-symbol*) ))
               (cond
                   ((or (null? (cdr term))
                        (null? (cddr term)) )
                      (formula-syntax-error "arguments to quantifier lacking" term) )
                   ((pair? (cadr term))
                      (formula-syntax-error "incorrect bound variable syntax" term) )
                   ((and (symbol? (caddr term))
                         (null? (cdddr term)) )
                      (caddr term) )
                   ((and (symbol? (caddr term))
                         (pair? (cadddr term))
                         (null? (cddddr term)) )
                      (list (car term)
                            (cadr term)
                            (make-prefix-formula-term-aux
                                                 (cons (caddr term) (caddr (cdr term)))
                                                 #t)))
                   ((or (eq? (caddr term) *exists-symbol*)
                        (eq? (caddr term) *all-symbol*) )
                      (list (car term)
                            (cadr term)
                            (make-prefix-formula-term-aux (cddr term) #f) ))
                   ((not (null? (cdddr term)))
                      (formula-syntax-error "too many arguments to quantifier" term) )
                   (else
                      (list (car term)
                            (cadr term)
                               (make-prefix-formula-term-aux (caddr term) #f) ))))
            ((and
              (not termflag)
              (eq? (car term) *not-symbol*) )

               (cond
                  ((null? (cdr term))
                     (formula-syntax-error "arguments to NOT lacking" term)  )
                  ((and (symbol? (cadr term))
                        (null? (cddr term)) )
                     (make-neg-symbol (cadr term)) )
                  ((and (not (pair? (cadr term)))
                        (null? (cddr term)) )
                      (formula-syntax-error "incorrect argument to NOT " term) )
                  ((and (symbol? (cadr term))
                        (pair? (caddr term))
                        (null? (cdddr term)) )
                     (make-prefix-formula-term-aux
                         (cons
                           (make-neg-symbol (cadr term))
                           (caddr term) )
                         #t))
                   ; ((not (null? (cddr term)))
                   ;   (formula-syntax-error "too many arguments to NOT" term) )
                  (else
                      (list (car term)
                              (make-prefix-formula-term-aux (cdr term) #f) ))))
            ((and
               (not termflag)
               (pair? (cdr term))
               (or (eq? (cadr term) *and-symbol*)
                     (eq? (cadr term) *or-symbol*)
                   (eq? (cadr term) *imp-symbol*)
                   (eq? (cadr term) *equivalent-symbol*) )
               (null? (cddr term)) )
              (formula-syntax-error "connective has too few arguments" term) )
            ((and
               (not termflag)
               (or (eq? (car term) *and-symbol*)
                     (eq? (car term) *or-symbol*)
                   (eq? (car term) *imp-symbol*)
                   (eq? (car term) *equivalent-symbol*) )
               (or (null? (cdr term))
                   (null? (cddr term)) ))
              (formula-syntax-error "connective has too few arguments" term) )
            ((and
              (not termflag)
              (pair? (cdr term))
              (pair? (cddr term))
              (or (eq? (cadr term) *and-symbol*)
                  (eq? (cadr term) *or-symbol*)
                  (eq? (cadr term) *imp-symbol*)
                  (eq? (cadr term) *equivalent-symbol*) ))
                (cons (cadr term)
                      (cons (make-prefix-formula-term-aux (car term) #f)
                            (list (make-prefix-formula-term-aux (cddr term) #f)) )))
            ((and
              (not termflag)
              (not (pair? (car term)))
              (not (or (eq? (car term) *and-symbol*)
                       (eq? (car term) *or-symbol*)
                       (eq? (car term) *imp-symbol*)
                       (eq? (car term) *equivalent-symbol*) ))
              (pair? (cdr term))
              (pair? (cddr term))
              (pair? (cdddr term))
              (pair? (cadr term))
              (or (eq? (caddr term) *and-symbol*)
                  (eq? (caddr term) *or-symbol*)
                  (eq? (caddr term) *imp-symbol*)
                  (eq? (caddr term) *equivalent-symbol*) ))
                (cons (caddr term)
                      (cons (cons (car term)
                                  (make-prefix-formula-term-aux (cadr term) #t) )
                            (list (make-prefix-formula-term-aux (cdddr term) #f)) )))
            ((and (not (pair? (car term)))
                  (not (null? (cdr term)))
                  (pair? (cadr term)) )
               (cons (cons (car term)  (make-prefix-formula-term-aux (cadr term) #t))
                     (make-prefix-formula-term-aux (cddr term) termflag) ))
            (else
               (cons (make-prefix-formula-term-aux (car term) termflag)
                     (make-prefix-formula-term-aux (cdr term) termflag)) )))
   (else
      (formula-syntax-error "period symbol prohibited" term) )))


(define (formula-syntax-error msg term)
   (newline)
   (display "Syntax error in input formula (")
   (display msg)(display "):")(newline)
   ;(print-raw-formula term)
   (pretty-print term)
   (newline)
   (abort) )

(define (print-raw-formula term)
   (print-list-as-term (remove-formula-dots term)) )

(define (remove-formula-dots term)
   (if (pair? term)
       (if (eq? (car term) *otter-comma-repl*)
           (remove-formula-dots (cdr term))
           (cons (remove-formula-dots (car term))
                 (remove-formula-dots (cdr term)) ))
       term))



(define (move-not-and-equiv term mcon)
   (let ((arg '()))
     (cond
        ((not (pair? term))
            term)
        ((eq? *not-symbol* (car term))
            (set!  arg (cadr term))
            (cond
               ((not (pair? arg))
                  (make-neg-symbol arg) )
               ((pair? (car arg))
                   (move-not-and-equiv (list *not-symbol*  (car arg) mcon)) )
               ((eq? *not-symbol* (car arg))
                   (move-not-and-equiv arg mcon) )
               ((eq? *or-symbol* (car arg))
                   (list *and-symbol*
                         (move-not-and-equiv (list *not-symbol* (cadr arg)) mcon)
                         (move-not-and-equiv (list *not-symbol* (caddr arg)) mcon) ))
               ((eq? *and-symbol* (car arg))
                   (list *or-symbol*
                         (move-not-and-equiv (list *not-symbol* (cadr arg)) mcon)
                         (move-not-and-equiv (list *not-symbol* (caddr arg)) mcon) ))
               ((eq? *imp-symbol* (car arg))
                    (list *and-symbol*
                      (move-not-and-equiv (cadr term) mcon)
                      (move-not-and-equiv (list *not-symbol* (caddr term)) mcon) ))
               ((eq? *equivalent-symbol* (car arg))
                    (if (eq? mcon *or-symbol*)
                         (list *or-symbol*
                         (list *and-symbol*
                               (move-not-and-equiv (list *not-symbol* (cadr arg) mcon))
                               (move-not-and-equiv (caddr arg) mcon) )
                         (list *and-symbol*
                               (move-not-and-equiv (cadr arg) mcon)
                               (move-not-and-equiv (list *not-symbol* (caddr arg)) mcon) ))
                        (list *and-symbol*
                         (list *or-symbol*
                               (move-not-and-equiv (cadr arg) mcon)
                               (move-not-and-equiv (caddr arg) mcon)
                         (list *or-symbol*
                               (move-not-and-equiv (list *not-symbol* (cadr arg)) mcon)
                               (move-not-and-equiv (list *not-symbol* (caddr arg)) mcon) )))))
               ((eq? *all-symbol* (car arg))
                   (list *exists-symbol*
                         (cadr arg)
                         (move-not-and-equiv (list *not-symbol* (caddr arg)) mcon) ))
               ((eq? *exists-symbol* (car arg))
                   (list *all-symbol*
                         (cadr arg)
                         (move-not-and-equiv (list *not-symbol* (caddr arg)) mcon) ))
               (else
                   (cons (make-neg-symbol (car arg)) (cdr arg)) )))

        ((eq? *imp-symbol* (car term))
           (list *or-symbol*
                 (move-not-and-equiv (list *not-symbol* (cadr term)) mcon)
                 (move-not-and-equiv (caddr term) mcon)))
        ((eq? *equivalent-symbol* (car term))
           (if (eq? mcon *or-symbol*)
               (list *or-symbol*
                 (list *and-symbol*
                       (move-not-and-equiv (cadr term) mcon)
                       (move-not-and-equiv (caddr term) mcon) )
                 (list *and-symbol*
                       (move-not-and-equiv (list *not-symbol* (cadr term)) mcon)
                       (move-not-and-equiv (list *not-symbol* (caddr term)) mcon) ))
               (list *and-symbol*
                 (list *or-symbol*
                       (move-not-and-equiv (list *not-symbol* (cadr term)) mcon)
                       (move-not-and-equiv (caddr term) mcon) )
                 (list *or-symbol*
                       (move-not-and-equiv (cadr term) mcon)
                       (move-not-and-equiv (list *not-symbol* (caddr term)) mcon) ))))
        (else
            (map (lambda(x)(move-not-and-equiv x mcon)) term) ))))



(define (flatten-prefix-formula term)
   (cond
      ((not (pair? term))
         term)
       ;  ((and (pair? (car term))
      ;      (not (pair? (cdr term))) )
      ;   (flatten-prefix-formula (car term)) )
      ((not (pair? (cdr term)))
            (flatten-prefix-formula (car term)) )
      ((or (eq? (car term) *and-symbol*)
           (eq? (car term) *or-symbol*)
           (eq? (car term) *imp-symbol*) )
          (cond
             ((and (pair? (cadr term))
                   (eq? (caadr term) (car term))
                   (pair? (caddr term))
                   (eq? (caaddr term) (car term)) )
                (cons (car term)
                      (append (cdr (flatten-prefix-formula (cadr term)))
                              (append
                                   (cdr (flatter-prefix-formula (caddr term))
                                   (map flatten-prefix-formula
                                        (cdddr term) ))))))
             ((and (pair? (cadr term))
                   (eq? (caadr term) (car term)) )
                (cons (car term)
                      (append (cdr (flatten-prefix-formula (cadr term)))
                              (map flatten-prefix-formula
                                   (cddr term) ))))
             ((and (pair? (caddr term))
                   (eq? (caaddr term) (car term)) )
                (cons (car term)
                      (cons (flatten-prefix-formula (cadr term))
                            (append (cdr (flatten-prefix-formula (caddr term)))
                                    (map flatten-prefix-formula
                                         (cdddr term) )))))
             (else
                (map (lambda(x)(flatten-prefix-formula x)) term) )))

      (else
         (map (lambda(x)(flatten-prefix-formula x)) term) )))


; type-alist format:
; ((int 0 (s int)))
;  (boolnum 0 1)
;  (smallnums 0 1 2 3 4)
;  (listnum nil (c num listnum))
; )

(define *type-alist* '())

(define *free-var-base-string* "x@")
(define *skolemfn-base-string* "$")
(define *type-pred-name* 'hastype)
(define *unfold-finite-types-flag* #t)
(define *last-skolemfn-nr* 0)
(define *last-free-var-nr* 0)
(define *const-bindings* '())


(define (skolemize-formula term)
  (set! *last-skolemfn-nr* 0)
  (set! *last-free-var-nr* 0)
  (skolemize-formula-aux term '() '()) )

(define (skolemize-formula-aux term free-vars varbindings)
   (let ((tmp1 '())
         (tmp2 '())
         (var '())
         (typename '()) )
     (cond
        ((symbol? term)
           (set! tmp1 (assq term varbindings))
           ; (set! tmp2 (assq term *const-bindings*))
           (cond (tmp1 (cdr tmp1))
                 ; (tmp2 (cdr tmp2))
                 (else term) ))
        ((not (pair? term))
            term)
        ((eq? *all-symbol* (car term))
            (cond
               ((symbol? (cadr term))
                  (set! tmp1 (make-freevar-formula))
                  (set! varbindings
                          (remove-if (lambda(x)(eq? (car x) (cadr term))) varbindings) )
                  (skolemize-formula-aux
                     (caddr term)
                     (cons tmp1 free-vars)
                     (cons (cons (cadr term) tmp1) varbindings) ))
               ((pair? (cadr term))
                       (set! var (cadr (cadr term)))
                  (set typename (caddr (cadr term)))
                  (set! tmp2 (assq typename *type-alist*))
                  (set! varbindings
                          (remove-if (lambda(x)(eq? (car x) var)) varbindings) )
                  (if (or (not *unfold-finite-types-flag*)
                          (not (finite-type-symbol-formula typename)) )
                      (begin
                          (set! tmp1 (make-freevar-formula))
                            (skolemize-formula-aux
                             (list *or-symbol*
                                   (list *not-symbol*
                                        (make-type-literal tmp1 typename) )
                                   (caddr term) )
                                (cons tmp1 free-vars)
                             (cons (cons var tmp1) varbindings) ))
                      (begin
                         (skolemize-formula-aux
                            (make-finite-subs-formula term var typename)
                            free-vars
                            varbindings))))))
        ((eq? *exists-symbol* (car term))
            (cond
               ((symbol? (cadr term))
                  (set! tmp1 (make-skolem-term-formula (cadr term) free-vars))
                  (set! varbindings
                          (remove-if (lambda(x)(eq? (car x) (cadr term))) varbindings) )
                  (skolemize-formula-aux
                     (caddr term)
                     free-vars
                     (cons (cons (cadr term) tmp1) varbindings) ))
               ((pair? (cadr term))
                       (set! var (cadr (cadr term)))
                  (set typename (caddr (cadr term)))
                  (set! tmp2 (assq typename *type-alist*))
                  (set! varbindings
                          (remove-if (lambda(x)(eq? (car x) var)) varbindings) )
                  (if (or (not *unfold-finite-types-flag*)
                          (not (finite-type-symbol-formula typename)) )
                      (begin
                          (set! tmp1 (make-skolem-term-formula var free-vars))
                          (skolemize-formula-aux
                             (list *and-symbol*
                                   (make-type-literal tmp1 typename)
                                   (caddr term) )
                                free-vars
                             (cons (cons var tmp1) varbindings) ))
                      (begin
                         (skolemize-formula-aux
                            (make-finite-subs-formula term var typename)
                            free-vars
                            varbindings))))))
        (else
           (map (lambda(x)(skolemize-formula-aux x free-vars varbindings))
                term)))))



(define  (make-skolem-term-formula varsymb free-vars)
  (let ((tmp '()))
    (set! *last-skolemfn-nr* (+ 1 *last-skolemfn-nr*))
    (set! tmp (string->symbol (string-append *skolemfn-base-string*
                                                     (number->string *last-skolemfn-nr*) )))
    (cons tmp free-vars) ))


(define (make-finite-subs-formula term var typename)
   (let* ((typecases '())
          (form-list '())
          (connective '()) )
     (set! typecases (assq typename *type-alist*))
     (set! typecases (cdr typecases))
     (set! form-list
                (map (lambda(x)(subs-into-formula (caddr term) var x)) typecases) )
     (if (eq? (car term) *exists-symbol*)
         (set! connective *or-symbol*)
         (set! connective *and-symbol*) )
     (split-list-by-connective-formula connective form-list) ))


(define (split-list-by-connective-formula conn lst)
  (cond
     ((null? lst) lst)
     ((null? (cdr lst)) (car lst))
     ((null? (cddr lst)) (cons conn lst))
     (else
             (list conn
                   (car lst)
                   (split-list-by-connective-formula conn (cdr lst)) ))))



(define (subs-into-formula term var value)
   (cond
      ((eq? term var) value)
      ((not (pair? term)) term)
      ((and (or (eq? (car term) *all-symbol*)
                      (eq? (car term) *exists-symbol*) )
                  (or (and (symbol? (cadr term))
                               (eq? (cadr term) var) )
                          (and (pair? (cadr term))
                               (eq? (cadr (cadr term)) var) )))
                term)
      (else
          (map (lambda(x)(subs-into-formula x var value)) term) )))


(define (make-freevar-formula)
  (set! *last-free-var-nr* (+ 1 *last-free-var-nr*))
  (string->symbol (string-append *free-var-base-string*
                                      (number->string *last-free-var-nr*) )))


(define (make-type-literal var name)
  (list *type-pred-name* var name) )

(define (finite-type-symbol-formula x)
     #f)


(define (make-neg-symbol pred)
  (let* ((predstr (symbol->string pred))
         (polarity (if (char=? #\- (string-ref predstr 0)) #f #t)) )  ; #t or #f
    (if (not polarity)
        (string->symbol (substring predstr 1 (string-length predstr)))
        (string->symbol (string-append "-" predstr)) )))


(define (clausify-formula-aux term)
  term)

;  (let ((part1 '())
;        (part2 '()) )
;    (cond
;      ((not (pair? term))
;         term)
;      ((and (pair? (car term)) (null? (cdr term)))
;         (clausify-formula-aux (car term)) )
;      ((eq? *or-symbol* (car term))
;         (set!
;         (cond
;          ((memq? (cdr term) *true-symbol*)
;             *true-symbol*)
;          ((every? (lambda(x)(or (not (pair? x)) (not (eq? *and-symbol* (car x)))))
;                   (cdr term) )
;             (remove *false-symbol* term) )
;          (else
;             (set! part1 (filter-if
;                                 (lambda(x)
;                                     (or (not (pair? x)) (not (eq? *and-symbol* (car x))) ))
;                                 term))  )))))))
;              (set! part2 (filter-if (lambda (x) (not


(define *true-symbol* 'true)
(define *false-symbol* 'false)

(define *and-symbol* '&)
(define *or-symbol* '|)
(define *imp-symbol* '=>)
(define *not-symbol* '-)
(define *equivalent-symbol* '<=>)
(define *exists-symbol* 'exists)
(define *all-symbol* 'all)

(define (ftest) (fprove "lifsch.in"))




;--------------------------------------------------------

(define (prepare-read-eqsymb eqlit posflag)
  (let* ((eqsymb (if posflag '= '!=))
         (neweqsymb (if posflag
                        *eq-symbol*
                        (string->symbol
                         (string-append "-" (symbol->string *eq-symbol*)) )))
         (tmp (replace-one-eq eqsymb '! eqlit)) )
    (list neweqsymb tmp) ))

(define (prepare-read-eqlit eqlit posflag)
  (let* ((eqsymb (if posflag '= '!=))
         (neweqsymb (if posflag
                        *eq-symbol*
                        (string->symbol
                         (string-append "-" (symbol->string *eq-symbol*)) )))
         (tmp (replace-one-eq eqsymb '! eqlit)) )
    (list neweqsymb tmp) ))

; parse-otter-exp parses an argument list of a functional term.
; it returns the input arglist in a parsed form.

(define (parse-otter-explist exp)
  (let ((part exp)
        (res '()) )
    (do ()
        ((null? part) (reverse res))
      (cond ((eq? *otter-comma-repl* (car part)))
            ((or (symbol? (car part)) (number? (car part)))
             (cond ((null? (cdr part))
                    (set! res (cons (car part) res)) )
                   ((eq? *otter-comma-repl* (cadr part))
                    (if (null? (cddr part))
                        (report-error "syntax error in input file")
                        (set! res (cons (car part) res)) ))
                   ((not (pair? (cadr part)))
                    (report-error "syntax error in input file") )
                   (else
                    (let ((tmp (parse-otter-explist (cadr part))))
                      (set! res (cons (cons (car part) tmp) res))
                      (set! part (cdr part)) ))))
            ((pair? (car part))
             (set! res (cons (parse-otter-explist (car part)) res)) ))
    (if (not (null? part)) (set! part (cdr part))) )))



; preprocess-typet-file reads a file char-by-char and outputs
; a file with the following changes:
;   . is separated by extra spaces and replaced by ...
;   % is replaced by ;
;   ',' is separated by extra spaces and replaced by !
;   | is separated by extra spaces
;   => is surrounded by extra spaces
;   upper-case chars are prefixed by ^


(define *otter-uppercase-prefix* #\^)
(define *otter-comma-repl* '!)
(define *otter-dot-repl* '...)

;(define *otter-tmpfile* "gandalf.tmp")

(define *otter-tmpfile-suffix* ".tmp")
(define *otter-tmpfile-prefix* "")
;(define *otter-tmpfile-prefix* "/users/cs/tammet/gandalf/")
;(define *otter-tmpfile-prefix* "/users/cs/tammet/gandalf/prb.tmp")

(define *otter-outputfile-suffix* ".res")

;---------------

(define *otter-tmpfile* " ")
(define *otter-inport* '())

(define (string-port-handler port)
  port)

;-------------

(define (preprocess-otter-file filename)
  (let ((iport (open-input-file filename)))
    (if iport
      (let ((ostring '()))
        (set! *otter-inport* iport)
        (set! ostring (call-with-output-string preprocess-otter-file-aux))
        (set! *otter-tmpfile* ostring) )
      (begin (report-error "could not open file: " filename)
             #f))))


(define (preprocess-otter-file-aux oport)
  (if oport
      (let ((iport *otter-inport*))
        (do ((char (read-char iport) (read-char iport)))
            ((eof-object? char)
              oport) ; scm5d5
          ;(display char)
          (cond
           ((eq? char #\,)
            (write-char #\space oport)
            (write *otter-comma-repl* oport)
            (write-char #\space oport) )
           ((eq? char #\%)
            (write-char #\; oport)  )
           ((eq? char #\.)
            (write-char #\space oport)
            (write *otter-dot-repl* oport)
            (write-char #\space oport) )
           ((eq? char #\|)
            (write-char #\space oport)
            (write-char #\| oport)
            (write-char #\space oport) )
           ;((and (eq? char #\!) (eq? (peek-char iport) #\=))
           ; (read-char iport)
           ; (write-char #\space oport)
           ; (write-char #\! oport)
           ; (write-char #\= oport)
           ; (write-char #\space oport) )
           ;((eq? char #\=)
           ; (write-char #\space oport)
           ; (write-char #\= oport)
           ; (if (eq? (peek-char iport) #\>)
           ;     (begin (write-char #\> oport)
           ;            (write-char #\space oport)
           ;            (read-char iport) )
           ;     (write-char #\space oport) ))
           ((eq? char *otter-uppercase-prefix*)
             (write-char *otter-uppercase-prefix* oport)
             (write-char *otter-uppercase-prefix* oport) )
           ((eq? char #\$)
             (write-char #\$ oport)
             (write-char #\$ oport) )
           ((and (char-alphabetic? char) (char-upper-case? char))
            (write-char *otter-uppercase-prefix* oport)
            (write-char char oport) )
           (else
            (write-char char oport) ))))
      (begin
        ; (close-port iport)
        (report-error "could not open string for temporary output")
        #f)))

; read-processed-otter-file reads a preprocessed otter-file
; and outputs a list of all the read objects (in the original order).
;

(define (read-processed-otter-file filename)

  ;(newline)(display "read-processed-otter-file called")(newline)

  (let ((iport (call-with-input-string filename string-port-handler))
        ;(iport (open-input-file filename))
        ;(iport (call-with-input-file filename string-port-handler))
        (lst '())
        (lst-end '()) )
    (set! lst (cons '() '()))
    (set! lst-end lst)
    (do ((obj (read iport) (read iport)))
        ((eof-object? obj)
          (close-input-port iport)
          (set! lst (cdr lst))

          ;(newline)(display "read-processed-otter-file terminated")(newline)

          lst)
      (set-cdr! lst-end (cons obj '()))
      (set! lst-end (cdr lst-end)) )))


; ---------------- reading clauses ----------------------------------

(define *ans-symbol* '$$ans)
(define *eq-symbol* 'equal) ; 'eq
(define *eq-fun-symbol* '$$eq)
(define *case-fun-symbol* '$$case)
(define *if-fun-symbol* '$$if)
(define *splitpred0-symbol* 'splitpred0)
(define *splitpred1-symbol* 'splitpred1)

(define *inhabited-symbol* '$$inh)
(define *subst-fun-symbol* '$$idsubst)
(define *subst-prim-fun-symbol* '$$idsubst-prim)
(define *id-refl-fun-symbol* '$$idrefl)
(define *lambda-fun-symbol* '$$lambda)
(define *lambda-dummypar-symbol* '$$*)
(define *typet-speclit*
  (list *ans-symbol*
        (list *lambda-fun-symbol*
              (list *subst-prim-fun-symbol* *lambda-dummypar-symbol*)
              (list *subst-fun-symbol* *lambda-dummypar-symbol*) )))
(define *classicsynth-speclit*
  (list *ans-symbol* (list *if-fun-symbol* 1 2 3)) )


(define *demodulator-list-symbol* 'demodulators)
(define *lit-demodulator-list-symbol* 'lit-demodulators)
(define *sos-list-symbol* 'sos)
(define *usable-list-symbol* 'usable)

(define *ans-prohibited-funsymbs* '())
(define *ans-prohibited-constsymbs* '())
(define *ans-prohibited-predsymbs* '())

(define *ans-prohibited-funnrs* '())
(define *ans-prohibited-constnrs* '())
(define *ans-prohibited-prednrs* '())



; ---------------- numerifying clauses and terms ---------------------

; *prolog-style-variables* determines the var/nonvar criteria
;   for input symbols

(define *prolog-style-variables* #t)

;;; In the following constant, function and predicate symbols
;;; of a clause set have each a different numeration
;;; (since function symbs cannot be constant symbs cannot be pred symbs
;;; due to their position in the clause or a term)
;;; vars are enumerated separately for each clause

; - - - - - - - - - - - - - - - - - - - - - -

; *const-num-table*: ((foo . N) (bar . M) ...)
;     for all constant symbs as first els of the pair

(define *const-num-table* '())

; *const-num* is initially set to a smallest possible function number - 1

(define *const-num* 0)
(define *curry-const-num* 0)
(define *min-const-num-sub1* 100000) ; the smallest fun num - 1
                                     ; can be made as small as 65535

; - - - - - - - - - - - - - - - - - - - - - -

; *fun-num-table*: ((foo . N) (bar . M) ...)
;     for all function symbs as first els of the pair

(define *fun-num-table* '())
(define *fun-num-arity-table* '()) ; ((N . 1)(M . 2)(R . 2) ...)

; *fun-num* is initially set to a smallest possible function number - 1

(define *fun-num* 0)
(define *min-fun-num-sub1* 0) ; the smallest fun num - 1
(define *max-nr-of-funs* 255) ; can be made bigger by 1 bit

; - - - - - - - - - - - - - - - - - - - - - -

; *pred-num-table*: ((foo . N) (bar . M) ...)
;     for all pure pred symbs (NO -) as first els of the pair

(define *pred-num-table* '())

; *pred-num* is initially set to a smallest possible function number - 1

(define *pred-num* 0)
(define *min-pred-num-sub1* 0) ; the smallest fun num - 1
(define *max-nr-of-preds* 127)

; - - - - - - - - - - - - - - - - - - - - - -

; *var-num* is for each new clause set to a smallest possible var number - 1

; *var-num-table*: ((xv . N) (u . M) ...)
;     for all var symbs IN THE SINGLE CLAUSE as first els of the pair

(define *var-num-table* '()) ; new for each single clause

(define *var-num* 0)
(define *min-var-num-sub1* 0)

; the following def is given at the beginning of the file
; (define *max-nr-of-vars-in-clause* 255) ; can be made three bits bigger

; - - - - - - - - - - - - - - - - - - - - - -

;;; in the following the clause is a standard list like:
;;; ((p 1 (f x) x) (-p x y (f (f y))))

(define (numerify-clause-list clauses)
  (init-numerify-clause-list)
  (map numerify-clause clauses) )



(define (init-numerify-clause-list)
  (set! *const-num-table* '())
  (set! *fun-num-table* '())
  (set! *fun-num-arity-table* '())
  (set! *pred-num-table* '())
  (set! *const-num* (- *minconst* 1))
  (set! *curry-const-num* (- *mincurryconst* 1))
  (set! *fun-num* *min-fun-num-sub1*)
  (set! *pred-num* *min-pred-num-sub1*)
  (set! *ans-prohibited-funnrs* '())
  (set! *ans-prohibited-prednrs* '())
  (set! *ans-prohibited-constnrs* '())
  (if *type-theory-stratflag* (numerify-literal *typet-speclit*))
  (if *classical-synth-stratflag*
      (begin (numerify-literal *classicsynth-speclit*)
             (set! *ans-prohibited-prednrs*
                   (map  (lambda(x) (car (numerify-literal (list x))))
                         *ans-prohibited-predsymbs*))
             (set! *ans-prohibited-funnrs*
                   (map  (lambda(x) (car (numerify-term (list x))))
                         *ans-prohibited-funsymbs*))
             (set! *ans-prohibited-constnrs*
                   (map numerify-term *ans-prohibited-constsymbs*) ))))

(define (numerify-clause clause)
  (set! *var-num-table* '())
  (set! *var-num* *min-var-num-sub1*)
  (map numerify-literal clause) )

(define (numerify-literal lit)
  (let* ((tmp 0)
         (pred (car lit))
         (predstr (symbol->string pred))
         (polarity (if (char=? #\- (string-ref predstr 0)) #f #t)) ; #t or #f
         (purepred (if polarity pred
           (string->symbol (substring predstr 1 (string-length predstr))) ))
         (place (assq purepred *pred-num-table*))
         (resnr
            (if place
                (polarize-pred-nr (cdr place) polarity)
                (begin (set! *pred-num* (+ 1 *pred-num*))
                       (if (> *pred-num* *max-nr-of-preds*)
               (report-error "Too many predicate symbs in the clause set" #t) )
                       (set! *pred-num-table*
                             (cons (cons purepred
                                     (shiftleft *pred-num* *fnsymb-shift*) )
                                   *pred-num-table*))
                       (polarize-pred-nr
                           (shiftleft *pred-num* *fnsymb-shift*)
                           polarity)))))
    (if (special-curry-purepred? purepred)
        (begin
          (set! tmp (numerify-curry-const (cadr lit)))
          (cons resnr (cons tmp (map numerify-term (cddr lit)))) )
        (cons resnr (map numerify-term (cdr lit))) ) ))


(define (polarize-pred-nr nr boolsgn)
  (if boolsgn
      nr
      (logior *sgn-mask* nr) ))

(define (reverse-polarize-pred-nr nr)
  (logxor *sgn-mask* nr) )

(define (numerify-term term)
  (cond
    ((pair? term)
       (let* ((fun (car term))
              (nr 0)
              (tmp 0)
              (place (assq fun *fun-num-table*)))
         (if place
             (set! nr (cdr place))
             (begin (set! *fun-num* (+ 1 *fun-num*))
                    (if (> *fun-num* *max-nr-of-funs*)
        (report-error "Too many function symbs in the clause set" #t) )
                    (set! *fun-num-table*
                          (cons (cons fun
                                      (shiftleft *fun-num* *fnsymb-shift*) )
                                *fun-num-table*))
                    (set! nr (shiftleft *fun-num* *fnsymb-shift*))
                    (set! *fun-num-arity-table*
                          (cons (cons nr (length (cdr term)))
                                 *fun-num-arity-table*)) ))

         (if (special-curry-fun? (car term))
             (begin
               (set! tmp (numerify-curry-const (cadr term)))
               (cons nr (cons tmp (map numerify-term (cddr term)))) )
             (cons nr (map numerify-term (cdr term))) ) ))
    ((var-symbol? term) ; term is a variable
       (let ((place (assq term *var-num-table*)))
         (if place
             (cdr place)
             (begin (set! *var-num* (+ 1 *var-num*))
                    (if (> *var-num* *max-nr-of-vars-in-clause*)
        (report-error "Too many vars in the encountered clause" #t) )
                    (set! *var-num-table*
                          (cons (cons term *var-num*)
                                *var-num-table*))
                    *var-num*))))
   (else ; term is a constant symb
       (let ((place (assq term *const-num-table*)))
         (if place
             (cdr place)
             (begin (set! *const-num* (+ 1 *const-num*))
                    (set! *const-num-table*
                          (cons (cons term *const-num*)
                                *const-num-table*))
                    *const-num*))))))


(define (numerify-curry-const term)
  (let ((place (assq term *const-num-table*)))
    (if place
        (cdr place)
        (begin (set! *curry-const-num* (+ 1 *curry-const-num*))
               (set! *const-num-table*
                     (cons (cons term *curry-const-num*)
                            *const-num-table*))
                *curry-const-num*))))


(define (special-curry-purepred? purepred)
  (and (not (null? *curried-pred-const-table*))
       (assq purepred *curried-appred-arities*) ))


(define (special-curry-fun? funsymb)
  (and (not (null? *curried-fun-const-table*))
       (assq funsymb *curried-apfun-arities*) ))

(define (curry-const? x)
  (and (const? x)
       (>= x *mincurryconst*) ))

(define (var-symbol? symb)
  (and (symbol? symb)
       (let ((char1 (string-ref (symbol->string symb) 0)))
         (if *prolog-style-variables*
             (eqv? char1 #\^)
             (member char1 '(#\x #\y #\z #\u #\v #\w)) ))))


; ------------------------------------------------------------------
;                          funs for printing
; -------------------------------------------------------------------


(define *list-output* #f) ; #t: (p (f x) 1)  #f: p(f(x) 1)

; *comma-output* has effect only if *list-output* is #f

(define *comma-output* #t) ; #t: p(f(x),1)    #f: p(f(x) 1)

; print-full-clause restores the original input order of
; negative literals in nucleons (for hyperresolution only!)
; These are moreover placed before positive literals.

(define (print-full-clause fullcl)
  (if #f ;(and *hyper-strategy-flag*
         ;  (if *neg-hyper-strategy-flag*
         ;      (not (negative-cl? (cdr fullcl)))
         ;      (not (positive-cl? (cdr fullcl))) ))
      (let ((new (map (lambda(x)'()) (cdr fullcl)))
            (nr 0)
            (tmp '())
            (otherlst '()) )
        (for-each (lambda(inf lit)
                    (set! nr (get-litinfo-ordernr (car inf)))
                    (if (not (eq? nr 0))
                        (set-car! (nth-cdr (- nr 1) new) lit)
                        (set! otherlst (cons lit otherlst)) ))
                  (caddar fullcl) ; litinfolist
                  (cdr fullcl) ) ; pure cl
        (if (not (null? otherlst))
            (set! new (map (lambda(lit)
                             (if (null? lit)
                                 (begin (set! tmp (car otherlst))
                                        (set! otherlst (cdr otherlst))
                                        tmp)
                                 lit))
                           new)))
        (print-pure-clause new) )
      (begin
          (print-pure-clause (cdr fullcl))
          (newline)
          (display "resobits: ")
          (display (map
                    (lambda(inf)
                      (zero? (logand *litinf-noresolve-mask* (car inf))) )
                   (caddr (car fullcl)) ))
          (newline)
          (display "factbits: ")
          (display (map
                    (lambda(inf)
                      (zero? (logand *litinf-nofactor-mask* (car inf))) )
                   (caddr (car fullcl)) ))
          (newline)
          (display "parabits: ")
          (display (map
                    (lambda(inf)
                     (list ; (car inf)
                           (litinf-left-para? (car inf))
                           (litinf-right-para? (car inf)) ))
                    (caddr (car fullcl)) ))) ))


(define (print-clause fullcl)
  (print-pure-clause (cdr fullcl)) )

(define (print-pure-clause pureclause)
  (cond
   ((null? pureclause)
      (if *list-output*
          (display '())
          (display "EMPTY CLAUSE.") ))
  (*list-output*
     (display #\()
     (do ((part pureclause (cdr part)))
         ((null? part))
       (print-lit (car part))
       (if (not (null? (cdr part))) (display #\space)) )
     (display #\)) )
  (else
     (do ((part pureclause (cdr part)))
         ((null? part))
       (print-lit (car part))
       (if (not (null? (cdr part)))
           (display " | ") ))
     (display #\.) )))


(define (print-lit lit)
  (if *list-output*
      (begin
        (display #\()
        (print-pred (car lit))
        (if (not (null? (cdr lit))) (display #\space))
        (do ((part (cdr lit) (cdr part)))
            ((null? part))
          (print-term (car part))
          (if (not (null? (cdr part))) (display #\space)) )
        (display #\)) )
      (begin
        (print-pred (car lit))
        (display #\()
        (do ((part (cdr lit) (cdr part)))
            ((null? part))
          (print-term (car part))
          (if (not (null? (cdr part)))
              (if *comma-output*
                  (display #\,)
                  (display #\space)) ))
        (display #\)) )))


(define (print-term term)
  (cond
    ((composite? term)
       (if *list-output*
           (begin
             (display #\()
             (print-fun (car term))
             (if (not (null? (cdr term))) (display #\space))
             (if (eq? *if-fun-nr* (logand *fnsymb-mask* (car term)))
                 (begin (print-lit (cadr term))
                        (display #\space)
                        (set! term (cdr term)) ))
             (do ((part (cdr term) (cdr part)))
                 ((null? part))
               (print-term (car part))
               (if (not (null? (cdr part))) (display #\space)) )
             (display #\)) )
           (begin
             (print-fun (car term))
             (display #\()
             (if (eq? *if-fun-nr* (logand *fnsymb-mask* (car term)))
                 (begin (print-lit (cadr term))
                        (if *comma-output*
                            (display #\,)
                            (display #\space) )
                        (set! term (cdr term)) ))
             (do ((part (cdr term) (cdr part)))
                 ((null? part))
               (print-term (car part))
               (if (not (null? (cdr part)))
                   (if *comma-output*
                       (display #\,)
                       (display #\space)) ))
             (display #\)) )))
    ((var? term)
       (print-var term) )
    (else
       (print-const term) )))

(define (print-pred prednum)
    (if (not-zero? (logand *sgn-mask* prednum)) (display #\-))
    (display (car (rassq (logand *predsymb-mask* prednum)
                         *pred-num-table*))))

(define (print-fun funnum)
  (display (car (rassq (logand *fnsymb-mask* funnum) *fun-num-table*))) )

(define (print-const constnum)
  (display (car (rassq constnum *const-num-table*))) )

; *nr-var-table* is for printing only and has no connections to the
; *var-nr-table* used for numerifying vars in input clauses

(define *nr-var-table*
  '((1 . x)(2 . y)(3 . z)(4 . u)(5 . v)(6 . w)) )

(define (print-var varnr)
  (let ((yflg #f))
    (if (not (zero? (logand *y-var-mask* varnr)))
        (begin (set! yflg #t)(set! varnr (logxor *y-var-mask* varnr))) )
    (set! varnr (logand *var-nr-mask* varnr))
    (let ((try (assq varnr *nr-var-table*)))
      (if yflg (display #\'))
      (if try
          (if *prolog-style-variables*
              (print-prolog-var (cdr try))
              (display (cdr try)) )
          (begin
             (if *prolog-style-variables*
                 (display #\X)
                 (display #\x) )
             (display (- varnr 6))) ))))

(define (print-prolog-var symb)
  (let ((str '()))
    (if (not (symbol? symb))
        (display symb)
        (begin
          (set! str (symbol->string symb))
          (display (char-upcase (string-ref str 0)))
          (display (substring str 1 (string-length str))) ))))

(define (print-list-as-term lst)
  (if (and (list? lst)
           (not (null? lst)) )
      (begin
        (display (car lst))
        (display "(")
        (do ((part (cdr lst) (cdr part)))
            ((null? part)
               (display ")") )
          (print-list-as-term (car part))
          (if (not (null? (cdr part)))
              (display ",")) ))
      (display lst) ))



;----------------------- making discr-tree readable ------------------
; *discr-tree* is a discrimination tree for literals.
; It includes all unit-clause literals and also all ground literals
; occurring in nonunit clauses. The top level contains only
; POSITIVE predsymbs.
;
; structure:
;
; each node-level is built of three sublists for funs, consts and vars:
; (funs consts . vars), and in case there are no consts nor vars
; it is just (funs).
;
; level_i:  ( ((f1 level_i+1) (f2 level_i+1) ...)    funs
;             ((c1 level_i+1) (c2 level_i+1) ...)    consts
;             (x1 level_i+1) (x2 level_i+1) ... )  vars
;
; the last level on each branch is a leaf:
; (positive-lit negative-lit full-unit-cl)
; where the nonexistant ones of the first two are $f and the
; full-unit-cl may be either negative or positive or nonexistent.
;
; leaf is the standard tree level.
; (positive-lit negative-lit full-unit-cl)
; where the nonexistant ones of the first two are '() and the
; full-unit-cl may be either negative or positive or nonexistent.

(define (show-discr-tree tree litflag)
  (pretty-print
    (if (and (pair? tree) (not (and (null? (car tree)) (null? (cdr tree)))))
        (if (tree-debug-leaf? tree)
            (make-readable-discr-leaf tree)
            (make-readable-discr-tree tree litflag) )
        tree)))

(define (make-readable-discr-tree tree litflag)
  (let
    ((el1
       (map
        (lambda(x)
           (let* ((tmp '())
                  (funnum (car x))
                  (fn (if litflag
                         (car (rassq funnum ;(logand *predsymb-mask* funnum)
                                     *pred-num-table*))
                         (car (rassq funnum ;(logand *fnsymb-mask* funnum)
                                     *fun-num-table*)) )))
             (if (tree-debug-leaf? (cdr x))
                 (set! tmp (make-readable-discr-leaf (cdr x)))
                 (set! tmp (make-readable-discr-tree (cdr x) #f)) )
             (cons fn tmp) ))
        (car tree) ))
     (el2
      (if (and (not (null? (cdr tree)))
               (pair? (cdr tree)) )
        (map
         (lambda(x)
           (let* ((tmp '())
                  (funnum (car x))
                  (fn (car (rassq funnum *const-num-table*))) )
             (if (tree-debug-leaf? (cdr x))
                 (set! tmp (make-readable-discr-leaf (cdr x)))
                 (set! tmp (make-readable-discr-tree (cdr x) #f)) )
             (cons fn tmp) ))
         (cadr tree) )
        '()))
     (rst (if (not (null? (cdr tree)))
              (map (lambda(x)
                     (let* ((tmp '())
                            (funnum (car x))
                            (fn (string->symbol
                                  (string-append "v"
                                    (number->string (logand *var-nr-mask*
                                                            funnum))))))
                       (if (tree-debug-leaf? (cdr x))
                           (set! tmp (make-readable-discr-leaf (cdr x)))
                           (set! tmp (make-readable-discr-tree (cdr x) #f)) )
                       (cons fn tmp) ))
                   (cddr tree) )
              '())) )
    (cons el1
          (if (or (not (null? el2))
                  (not (null? rst)) )
              (cons el2 rst)
              '() ))))



(define (tree-debug-leaf? tree)
  (or (null? tree)
      (and (pair? (car tree)) ; first el in tree (leaf)
           (not (pair? (caar tree))) )
      (and (pair? (cdr tree))
           (pair? (cadr tree))
           (not (pair? (caadr tree))) )
      (and (pair? (cdr tree))
           (pair? (cddr tree))
           (pair? (car (caddr tree))) )))

(define (make-readable-discr-leaf leaf)
  (if *nonunit-discr-flag*
      leaf
      (let ((el1 (if (pair? (car leaf))
                     (make-readable-lit (car leaf))
                     '()))
            (el2 (if (and (not (null? (cdr leaf)))
                          (pair? (cadr leaf)) )
                     (make-readable-lit (cadr leaf))
                     '()))
            (el3 (if (and (not (null? (cdr leaf)))
                          (not (null? (cddr leaf))) )
                 (make-readable-fullcl (caddr leaf))
                 '() )))
        (cons el1
              (if (or (not (null? el2))
                      (not (null? el3)) )
                  (cons el2
                    (if (not (null? el3))
                        (list el3)
                        '() ))
                  '() )))))


(define (make-readable-fullcl fullcl)
  (cons (list (caar fullcl) (cadar fullcl)) ; (car fullcl) for full one
        (map make-readable-lit (cdr fullcl)) ))

(define (make-readable-cl cl)
  (map make-readable-lit cl) )

(define (make-readable-lit lit)
 (let ((funnum (car lit)))
  (cons  (if (not-zero? (logand *sgn-mask* funnum))
             (string->symbol
              (string-append "-"
                 (symbol->string (car (rassq (logand *predsymb-mask* funnum)
                                             *pred-num-table*)))))
              (car (rassq (logand *predsymb-mask* funnum)
                          *pred-num-table*)))
        (if (eq? *lambda-fun-nr* (pure-fun funnum))
            (cons (make-readable-term (cadr lit))
                  (map make-readable-term (cddr lit)) )
            (map make-readable-term (cdr lit)) ))))


(define (make-readable-term term)
  (if (pair? term)
      (cons (car (rassq (logand *fnsymb-mask* (car term))
                        *fun-num-table*))
            (if (eq? *lambda-fun-nr* (pure-fun (car term)))
                (cons (make-readable-term (cadr term))
                      (map make-readable-term (cddr term)) )
                (map make-readable-term (cdr term)) ))
      (if (var? term)
          (string->symbol
            (string-append "v" (number->string  (logand *var-nr-mask* term))))
          (car (rassq term *const-num-table*)) )))


; ---------------------- debugging-oriented printers -------------------


(define (debug-print-clause cl)
  (newline)
  (let* ((info (car cl))
         (pure (cdr cl))
         (num (car info)) )
    (display (vector
              (shiftright (logand *cl-maxvar-mask* num) *cl-maxvar-shift*)
              (shiftright (logand *cl-len-mask* num) *cl-len-shift*)
              (shiftright (logand *ground-mask* num) 16)
              (shiftright (logand *term-depth-mask* num) 11)
              (shiftright (logand *term-size-mask* num) 5)
              (logand *const-number-mask* num) ))
    (display #\space)
    (display (number->string (cadr info) 2))
    (display #\space)
    (newline)
    (display (map (lambda(el)
                     (if (not (pair? el))
                         el
                         (cons (vector
                                (get-litinfo-classnr (car el))
                                (get-litinfo-ordernr (car el))
                                (zero? (logand *litinf-reductionrulelit-mask*
                                               (car el)))
                                (zero? (logand *litinf-rightimplrulelit-mask*
                                               (car el)))
                                (zero? (logand *litinf-leftorrulelit-mask*
                                               (car el)))
                                (zero? (logand *litinf-rulelit-mask*
                                               (car el)))
                                (zero? (logand *litinf-noparafromright-mask*
                                               (car el)))
                                (zero? (logand *litinf-noparafromleft-mask*
                                               (car el)))
                                (zero? (logand *litinf-noparainto-mask*
                                               (car el)))
                                (zero? (logand *litinf-isanswer-mask*
                                               (car el)))
                                (zero? (logand *litinf-nofactor-mask*
                                               (car el)))
                                (zero? (logand *litinf-noresolve-mask*
                                               (car el))))
                               (cdr el) )))
                  (caddr info) ))
    (newline)
    (display (cdddr info))
    (debug-print-pure-clause pure) ))


(define (debug-print-pure-clause cl)
  (newline)
  (display #\{)
  (for-each (lambda(x)(debug-print-lit x)(display #\space)) cl)
  (display #\})
  (newline) )

(define (debug-print-lit term)
  (display #\[)
  (debug-print-pred (car term))
  (if (not (null? (cdr term))) (display #\space))
  (do ((part (cdr term) (cdr part)))
      ((null? part))
    (debug-print-term (car part))
    (if (not (null? (cdr part))) (display #\space)) )
  (display #\]) )

(define (debug-print-term term)
  (cond
    ((composite? term)
       (display #\()
       (debug-print-fun (car term))
       (if (not (null? (cdr term))) (display #\space))
       (do ((part (cdr term) (cdr part)))
           ((null? part))
         (debug-print-term (car part))
         (if (not (null? (cdr part))) (display #\space)) )
       (display #\)) )
    ((var? term)
       (debug-print-var term) )
    (else
       (debug-print-const term) )))


(define (debug-print-pred funnum)
  (if (not-zero? (logand *sgn-mask* funnum)) (display #\-))
  (display
   (vector (car (rassq (logand *predsymb-mask* funnum) *pred-num-table*))
           (shiftright (logand *maxvar-mask* funnum) 18)
           (shiftright (logand *repeated-vars-mask* funnum) 17)
           (shiftright (logand *ground-mask* funnum) 16)
           (shiftright (logand *term-depth-mask* funnum) 11)
           (shiftright (logand *term-size-mask* funnum) 5)
           (logand *const-number-mask* funnum) )))

(define (debug-print-fun funnum)
  (display
   (vector (car (rassq (logand *fnsymb-mask* funnum) *fun-num-table*))
           (shiftright (logand *maxvar-mask* funnum) 18)
           (shiftright (logand *repeated-vars-mask* funnum) 17)
           (shiftright (logand *ground-mask* funnum) 16)
           (shiftright (logand *term-depth-mask* funnum) 11)
           (shiftright (logand *term-size-mask* funnum) 5)
           (logand *const-number-mask* funnum) )))


(define (debug-print-const constnum)
  (if (eq? constnum *unassigned*)
      (display "*unassigned*")
      (display
       (car (rassq constnum *const-num-table*)) )))

(define (debug-print-var varnr)
  (let ((yflg #f)
        (zflg #f) )
    (if (not (zero? (logand *y-var-mask* varnr)))
        (begin (set! yflg #t)(set! varnr (logxor *y-var-mask* varnr))) )
   (if (not (zero? (logand *z-var-mask* varnr)))
        (begin (set! zflg #t)(set! varnr (logxor *z-var-mask* varnr))) )
    (display
     (string-append
      (number->string
       (shiftright (logand *single-occ-lit-var-mask* varnr) 15) )
      (number->string
       (shiftright (logand *no-occ-otherlit-var-mask* varnr) 14) )
      (number->string
       (shiftright (logand *no-occ-folllit-var-mask* varnr) 13) )
      (let ((try (assq (logand *var-nr-mask* varnr)
                       *nr-var-table*)))
        (if try
            (if yflg
                (string-append "'" (symbol->string (cdr try)))
                (if zflg
                   (string-append "''" (symbol->string (cdr try)))
                   (symbol->string (cdr try)) ))

            (if yflg
                (string-append "'x" (number->string
                                     (- (logand  *var-nr-mask* varnr)
                                        6)))
                (if zflg
                        (string-append "''x" (number->string
                                     (- (logand  *var-nr-mask* varnr)
                                        6)))
                        (string-append "x" (number->string
                                            (- (logand  *var-nr-mask* varnr)
                                               6)))))))))))


(define (debug-print-xvars var)
  (newline)(display "xvars and immediate refs and touched/not:")(newline)
  (do ()
      ((zero? var))
    (debug-print-var var)(display " : ")
    (debug-print-term (var-ref var))(display " : ")
    (display (touched-var? var))(newline)
    ; (if (not (eq? (var-ref var) *unassigned*)) (break))
    (set! var (- var 1)) ))


;(define (debug-print-xvars-spec var)
;  ;(newline)(display "xvars and immediate refs and touched/not:")(newline)
;  (do ()
;      ((zero? var))
;    ;(debug-print-var var)(display " : ")
;    ;(debug-print-term (var-ref var))(display " : ")
;    ;(display (touched-var? var))(newline)
;    (if (not (eq? (var-ref var) *unassigned*)) (break))
;    (set! var (- var 1)) ))

(define (debug-print-yvars var)
  (newline)(display "yvars and immediate refs and touched/not:")(newline)
  (set! var (logior *y-var-mask* var)) ; it is first made an yvar
  (do ()
      ((eq? *max-x-var* var)) ; a zero for yvars
    (debug-print-var var)(display " : ")
    (debug-print-term (var-ref var))(display " : ")
    (display (touched-var? var))(newline)
    (set! var (- var 1)) ))

(define (debug-print-zvars var)
  (newline)(display "zvars and immediate refs:")(newline)
  (set! var (logior *z-var-mask* var)) ; it is first made an yvar
  (do ()
      ((eq? *max-y-var* var)) ; a zero for zvars
    (debug-print-var var)(display " : ")
    (debug-print-term (var-ref var))(newline)
    (set! var (- var 1)) ))

(define (report-error str . obj)
  (newline)
  (display "ERROR: ")(display str)(display " ")
  (for-each display obj)
  (newline)
  (quit) )


; NB! Argument display-info-level not used

(define *internal-time-units-per-second* 100)
(define *total-time-start* 0)
(define *total-time-given* #f)
(define *one-pass-time* 0)
(define *strat-list* '())

(define *given-checkpoint-keptnrs* '())
(define *checkpoint-keptnrs* '())
(define *last-got-time* 0)
(define *use-real-time* #f)
(define *run-forever-flag* #t)

(define  (get-time)
  (if *use-real-time*
      (get-internal-real-time)
      (get-internal-run-time) ))

;(define (get-check-time-aux x)
;  (display ": ")  (display x)(display " ")
;  (display *kept-cl-nr*) (display "; ")
;  (get-check-time) )

(define (get-check-time x)
  (let ((tmp 0))
    (if (null? *given-checkpoint-keptnrs*)
        (begin
          (set! *last-got-time* (get-time))
          *last-got-time* )
        (if (and (>= *kept-cl-nr* (car *given-checkpoint-keptnrs*))
                 (eqv? x (cadr *given-checkpoint-keptnrs*)) )
            (begin
              ; (newline)
              ; (display "given checkpoint used: ")
              ; (display *given-checkpoint-keptnrs*)
              ; (newline)
              (set! *last-got-time* (caddr *given-checkpoint-keptnrs*))
              (set! *given-checkpoint-keptnrs*
                    (cdddr *given-checkpoint-keptnrs*) )
              *last-got-time*)
            (begin
              (set! *last-got-time* 0)
              0)))))

(define  (store-time-checkpoint! x)
   (set! *checkpoint-keptnrs*
         (cons  *last-got-time*
                (cons x
                      (cons *kept-cl-nr*
                            *checkpoint-keptnrs*)))) )
   ; (newline)
   ; (display "!!!!!!!! ")
   ; (display *checkpoint-keptnrs*)
   ; (display " !!!!!!!! ")
   ; (newline))


(define (prove raw-input stop-real-time display-info-level)
  (init-stats)
  (init-globals)
  (make-symbs-lex-order-vector)
  (make-symbs-kb-order-vector)
  (set! *internal-time-units-per-second*  internal-time-units-per-second)
  (use-input-settings raw-input)
  (set! *total-time-start* (get-time))

  ; - - - - - initial printing - - - - - -

  (if (>= *display-info-level* 30)
      (begin
        (if *auto-mode-flag*
            (begin
              (display "Using automatic strategy selection.")
              (newline) )
            (begin
              (display "Using a user-selected strategy.")
              (newline) ))))

  (if *timelimit*
      (begin
        (if (>= *display-info-level* 30)
            (begin
              (display "Time limit in seconds: ")
              (display *timelimit*)
              (newline) ))
        (set! *total-time-given*
              (*  *internal-time-units-per-second*
                  (string->number *timelimit*) )))
      (begin
        (if (>= *display-info-level* 30)
            (begin
              (if *auto-mode-flag*
                  (begin
                    (display "Time limit in seconds: ")
                    (display *default-time-limit*) )
                  (begin
                    (display "Time limit in seconds: none")
                    ;(display *timelimit*)
                    ))
              (newline) ))
        (set! *total-time-given* #f) ))


  ; - - - - - - initialisation done - - - - - - -
  ; (turn-on-demodulation)
  ; (set! *paramodulation-strategy-flag* #t)
  (set! *overdouble-discr-flag* #f)
  (set! *nonunit-discr-flag* #f)
  (let* ((compl-lit-demodulators '())
          (demodulators (assq *demodulator-list-symbol* raw-input))
         (lit-demodulators (assq *lit-demodulator-list-symbol* raw-input))
         (sos (assq *sos-list-symbol* raw-input))
         (usable (assq *usable-list-symbol* raw-input))
         (tmp '()) )
    (set! *stop-res-real-time* stop-real-time)
    ;(set! *display-info-level* display-info-level)
    ;(set! *display-info-char-flag* #f)
    ;(set! *display-debug-info-flag* #f)
    ;(if *timelimit*
        ;(begin
        ;  (set! *timelimit* (string->number *timelimit*)) ))
    (if demodulators
        (begin (set! demodulators (cdr demodulators))
               (set! *demodulate-flag* #t)
               (set! *new-demodulators-flag* #t)
               (set! *demodulate-constants-flag* #t)
               (if *display-debug-info-flag*
                   (begin
                     (newline)(display "input demodulators: ")(newline)
                     (for-each (lambda(x)(display x)(newline))
                               demodulators)))) )
   ;     (begin (set! *demodulate-flag* #f)
   ;            (set! *new-demodulators-flag* #f)
   ;            (set! *demodulate-constants-flag* #f) )
    (if lit-demodulators
        (begin (set! lit-demodulators (cdr lit-demodulators))
               (set! *demodulate-lit-flag* #t)
               (set! *new-demodulators-flag* #t) ;;;; !!!!! ?????? !!!!!!
               (if *display-debug-info-flag*
                   (begin
                     (newline)(display "input literal demodulators: ")(newline)
                     (for-each (lambda(x)(display x)(newline))
                               lit-demodulators) )))
        (begin (set! *demodulate-lit-flag* #f)) )
    (if sos
       (begin (set! sos (cdr sos))
              (if *display-debug-info-flag*
                  (begin
                    (newline)(display "input sos: ")(newline)
                    (for-each (lambda(x)(display x)(newline)) sos) ))))
    (if usable
       (begin (set! usable (cdr usable))
              (if *display-debug-info-flag*
                  (begin
                    (newline)(display "input usable: ")(newline)
                    (for-each (lambda(x)(display x)(newline)) usable) ))))

    ; if sos is empty, make usable sos.

    (cond
     ((and sos (not usable)))
     ((and usable (not sos))
        (set! sos usable)
        (set! usable #f)))

    (set! *sos-clauses-list* '())
    (set! *usable-clauses-list* '())
    (set! *subsumption-list* (list (list 0))) ; a unique '((0))
    (set! *unit-alist* '())
    (set! *discr-tree* (list '())) ; a unique (list '())
    (set! *nonunit-discr-tree* (list '())) ; a unique (list '())
    (set! *discr-tree-demod-terms* (list '()))
    (set! *discr-tree-demod-atoms* (list '()))
    (set! *kept-cl-nr* 0)

    (init-numerify-clause-list)
    (if demodulators
        (begin
          (if *type-theory-stratflag*
              (set! demodulators
                    (map (lambda(x) (add-typet-demod-ans x))
                         demodulators)))
          (set! demodulators (map numerify-clause demodulators)) ))


    (if lit-demodulators
        (set! lit-demodulators (map numerify-clause lit-demodulators)) )
    (if usable (set! usable (map numerify-clause usable)))
    (if sos (set! sos (map numerify-clause sos)))


    ;(newline)
    ;(display "Splitting test:")
    ;(newline)
    ;(for-each
    ;  (lambda(x)
    ;    (newline)
    ;    (display "orig clause: ")(newline)
    ;    (print-pure-clause x)
    ;    (newline)
    ;    (display x)
    ;    (newline)
    ;    (display (split-full-clause (cons '() x))) )
    ;   sos)
    ;(quit)

    ;if usable is empty (all the stuff in sos),
    ; move all but fully negative clauses to usable

    (if (and (not usable)
             *auto-mode-flag*)
        (begin
          (set! tmp '())
          (set! usable '())
          (do ((part sos (cdr part)))
              ((null? part)
                 (set! sos (reverse tmp)) )
            (if (negative-cl? (car part))
                (set! tmp (cons (car part) tmp))
                (set! usable (cons (car part) usable)) ))))

    (if *type-theory-stratflag*
      (begin
        (set! *subst-fun-nr* (cdr (assq *subst-fun-symbol* *fun-num-table*)))
        (set! *subst-prim-fun-nr*
              (cdr (assq *subst-prim-fun-symbol* *fun-num-table*)) )
        (set! *lambda-fun-nr* (cdr (assq *lambda-fun-symbol* *fun-num-table*)))
        (set! *lambda-dummypar-nr*
              (cdr (assq *lambda-dummypar-symbol* *const-num-table*)) )))


    ;(newline)(newline)
    ;(display *pred-num-table*)
    ;(if (>= *display-info-level* 40)
    ;        (begin
    ;          (newline)
    ;          (for-each (lambda(x)(display x)(newline))demodulators)
    ;          (for-each (lambda(x)(display x)(newline))sos) ))

    (fill-predsymb-bit-vector!) ; data for this is built in numerify-clause

    ;(newline)(newline)
    ;(display *pred-num-table*)
    ;(for-each (lambda(x)(display x)(newline))sos)

   ; - - - - - - - - - clauses are numerified - - - - - - - - - -

    ;(set! tmp
    ;          (prove-make-initial-clauses
    ;           demodulators lit-demodulators compl-lit-demodulators
    ;           usable sos) )
    ; (set! usable (car tmp))
    ;(set! sos (cadr tmp))
    ;(analyse-input usable sos)
    ;(if *auto-mode-flag* (select-initial-strategy))
    (set! *strat-list* '((hyper 3)(binary 4)))
    (set! *no-passes-started* #t)
    (prove-all-passes
      usable sos  demodulators lit-demodulators compl-lit-demodulators)))


; prove-make-initial-clauses takes numerified clause lists
; and builds a full input set of clauses, returning a list
; (usable sos)


(define (prove-make-initial-clauses usable sos
          demodulators lit-demodulators compl-lit-demodulators)
    ; usable unit equality clauses created of demodulators are not
    ; demodulated themselves.
    (if demodulators
        (begin
         (set! demodulators (map
                               (lambda(x)(make-input-clause x #f))
                               demodulators))
         (for-each modify-to-demod-inf! demodulators)
         (for-each intern-fullclause demodulators)
         (for-each
           (lambda(x)
             (set! *cur-clause*
                   (cons (car x) (map rename-x-to-z-term (cdr x))) )
             (set! *cur-lit* (cadr *cur-clause*))
             (insert-discr-tree!
               *discr-tree-demod-terms* (eq-lit-arg1 *cur-lit*) #f #t #t #f))
           (if *type-theory-stratflag*
               (map remove-typet-demod-ans demodulators)
               demodulators) )))
  ; now lit-demodulators are going to be term-demodulated while being built.
    (if lit-demodulators
        (begin
         ; make complementary clauses
         (set! compl-lit-demodulators
               (map (lambda(x)
                      (map (lambda(y)
                             (cons (logxor *sgn-mask* (car y)) (cdr y)))
                           x))
                    lit-demodulators))
         (set! lit-demodulators (map
                                 (lambda(x)(make-input-clause x #f))
                                 lit-demodulators))
         (set! compl-lit-demodulators
               (map (lambda(x)(make-input-clause x #f))
                    compl-lit-demodulators))
         (for-each intern-fullclause lit-demodulators)
         (for-each intern-fullclause compl-lit-demodulators)
         (for-each
           (lambda(x)
             ; rename for z-vars
             (set! *cur-clause*
                   (cons (car x) (map rename-x-to-z-term (cdr x))) )
             ; put the positive lit first
             (if (neg-fun? (caadr *cur-clause*))
                 (set! *cur-clause*
                       (list (car *cur-clause*)
                             (caddr *cur-clause*) (cadr *cur-clause*) )))
             (set! *cur-lit* (cadr *cur-clause*))
             (insert-discr-tree!
               *discr-tree-demod-atoms* *cur-lit* #f #t #t #f))
           lit-demodulators)))
    ; now sos is going to be demodulated while being built.
    ; Even constants and ground terms are demodulated.
    (if *display-debug-info-flag*
        (begin
          (newline)
          (display "starting to make-input-clause on sos:")
          (newline)(newline)
          (for-each (lambda(x)(display x)(newline)
                           (debug-print-pure-clause x)(newline))
                    sos)))
    (if sos
        (begin (if usable
                   (set! usable (map
                                  (lambda(x)(make-input-clause x #f))
                                  usable))
                   (set! usable '()) )
               ; clauses with nr less than *last-usable-cl-nr*
               ; are in the "usable" set - for set of support strategy
               (set! *last-usable-cl-nr* *kept-cl-nr*)
               (set! sos (map
                           (lambda(x)(make-input-clause x 1))
                           sos))
               (if (and (>= *display-info-level* 40)
                        *display-debug-info-flag*)
                   (begin
                     (newline)
                     (display "finished make-input-clause:")
                     (newline)(newline)
                     (for-each (lambda(x)(display x)(newline)
                                      (debug-print-clause x)(newline))
                               (append usable sos))) )
               (if *type-theory-stratflag*
                          (begin
                            (set! *ans-input-sos* (append demodulators sos))
                     (set! sos (remove-ans-from-clauses sos)) ))
               (for-each intern-fullclause usable)
               (for-each intern-fullclause sos) )
        (begin (newline)
               (display "Error: no sos-list given.")
               (newline)
               (quit) ))
    ; now there is no need any more for demodulating constants or
    ; assuming new demodulators.
    ;(newline)
    ;(display "finished intern-fullclause on sos:")
    ;(newline)(newline)
    ;(for-each (lambda(x)(display x)(newline)
        ;             (debug-print-clause x)(newline))
         ;     sos)
    ; (set! *new-demodulators-flag* #f)
    (set! *demodulate-constants-flag* #f)
    (if (not demodulators) (set! demodulators '()))
    (if (not sos) (set! sos '()))
    (if (not lit-demodulators) (set! lit-demodulators '()))
    (if (not compl-lit-demodulators) (set! compl-lit-demodulators '()))
;    (set! *sos-clauses-list*
;          (append demodulators lit-demodulators compl-lit-demodulators sos))
    ; the following is fine for hyperresolution:
    (if *type-theory-stratflag*
        (set! *sos-clauses-list*
              (append (remove-ans-from-clauses demodulators)
                      compl-lit-demodulators usable sos))
        ; create the initial clause lists
        (begin
          (set! *usable-clauses-list* usable)
          (set! *sos-clauses-list*
                ; (append demodulators compl-lit-demodulators sos)
                sos) ))
    (list usable sos demodulators lit-demodulators compl-lit-demodulators) )


(define (prove-show-input-clauses usable sos
           demodulators lit-demodulators compl-lit-demodulators)
    (if (and (>= *display-info-level* 40) (not (null? demodulators)))
        (begin
          (newline)(display "demodulators: ")(newline)
          (for-each (lambda(x)(print-pure-clause (cdr x))(newline))
                    demodulators)
          (if *display-debug-info-flag*
              (begin
                (newline)(display "demodulators in debug form: ")(newline)
                (for-each (lambda(x)(debug-print-clause x)(newline))
                          demodulators) ))))
    (if (and (>= *display-info-level* 40)  (not (null? lit-demodulators)))
        (begin
          (newline)(display "literal demodulators: ")(newline)
          (for-each (lambda(x)(print-pure-clause (cdr x))(newline))
                    lit-demodulators)
          (if *display-debug-info-flag*
              (begin
                (newline)(display "literal demodulators in debug form: ")
                (newline)
                (for-each (lambda(x)(debug-print-clause x)(newline))
                          lit-demodulators) ))))
    (if (and (>= *display-info-level* 40)  (not (null? usable)))
        (begin
          (newline)(display "clauses in usable: ")(newline)
          (for-each (lambda(x)(print-pure-clause (cdr x))(newline))
                    usable)
          (if *display-debug-info-flag*
              (begin
                (newline)(display "usable in debug form: ")
                (newline)
                (for-each (lambda(x)(debug-print-clause x)(newline))
                          usable) ))))
    (if (and (>= *display-info-level* 40) (not (null? sos)))
        (begin
          (newline)(display "clauses in sos: ")(newline)
          (for-each (lambda(x)(print-pure-clause (cdr x))(newline))
                    sos)
          (if *display-debug-info-flag*
              (begin
                (newline)(display "sos in debug form: ")
                (newline)
                (for-each (lambda(x)(debug-print-clause x)(newline))
                          sos) )))) )


(define *no-passes-started* #t)

(define *all-passes-time* 0)

(define *strat-selected* 0)
(define *strat-given-time* 0)
(define *strat-start-time* 0)
(define *strat-passed-time* 0)
(define *strat-left-time* 0)

(define *all-passes-selected-clauses* '())

; prove-all-passes loops over several prover passes.
; it takes numerified clause lists as input.
; for auto mode it initially computes the strategy list, then
; calls the loop over strategies

(define (prove-all-passes usable sos
           demodulators lit-demodulators compl-lit-demodulators)

  (newline)
  (display "prove-all-passes started")
  (newline)

  (let ((tmp '())
        (res #f) )
    (set! *all-passes-selected-clauses* '())
    (set! *strat-list*
          '((hyper 5 #t)
            (binary 10 #t)
            ; (binary-hyper-sos 20 #t) ; *binary-hyper-sos-strategy*
            ; (hyper-order 10 #f)
            (binary-unit 10 #t)
            (binary-unit 10 #f)
            (binary-order 10 #f)
            (binary-nameorder 10 #f)

            (hyper-order 30 #f)
            (binary 30 #t)
            (binary 50 #f) ))
    (cond
     ((not *auto-mode-flag*)
       (if *total-time-given*
           (set! *one-pass-time* *total-time-given*)
           (set! *one-pass-time* 1000000) )
       (set! res
             (prove-one-pass usable sos
                 demodulators lit-demodulators compl-lit-demodulators))
       (get-check-time 20)
       (store-time-checkpoint! 20)
       res)
     (else     ; automatic case
       (if *total-time-given*
               (set! *all-passes-time*
                         (quotient *total-time-given*
                                        *internal-time-units-per-second*))
               (set! *all-passes-time* *default-time-limit*) )
       ; build a strategy list to be looped over
       (build-strat-list  usable sos
           demodulators lit-demodulators compl-lit-demodulators)
       ; loop over strategies, doing one or more passes for each
       (set! res
            (prove-all-passes-loop-strategies usable sos
               demodulators lit-demodulators compl-lit-demodulators) )
       res))))



; prove-all-passes-loop-strategies loops over all strats in a list:
; it assumes the strategy list has been calculated and is
; available as a value of *strat-list*


(define (prove-all-passes-loop-strategies usable sos
           demodulators lit-demodulators compl-lit-demodulators)

 ;(newline)
 ;(display "prove-all-passes-loop-strategies called with *strat-list*: ")
 ;(display *strat-list*)
 ;(newline)

 (let ((tmp '())
       (res #f) )
    (do ()
        ((or res ; *foundflag*
             (and *sat-strategy-flag* *sat-found-flag*)
             (null? *strat-list*)
             (and *total-time-given*
                  (> (- (get-check-time 10) *total-time-start*)
                      *total-time-given*)))
           (if (and (not *foundflag*)
                         (not (null? *strat-list*)) )
                (store-time-checkpoint! 10) )
           (get-check-time 30)
           (store-time-checkpoint! 30)
           res)

       ; take a first strat in the list

       (set! *strat-selected* (car *strat-list*))
       (set! *strat-given-time* (* *internal-time-units-per-second*
                                                    (cadr *strat-selected*) ))
       (set! *strat-start-time* (get-check-time 40))
       (store-time-checkpoint! 40)

       ;(newline)
       ;(display "*strat-list* in prove-all-passes-loop-strategies loop:")
       ;(newline)
       ;(display *strat-list*)
       ;(newline)

       (cond
         ((eq? (car *strat-selected*) 'split) ;splitcase
            (set! *strat-list* (cdr *strat-list*))
            (set! res
                 (prove-all-passes-split-strategy usable sos
                      demodulators lit-demodulators compl-lit-demodulators
                      *strat-selected*)))
         (else ; normal, non-split case
            (set! res
              (prove-all-passes-single-strategy usable sos
                demodulators lit-demodulators compl-lit-demodulators) ))))))

; prove-all-passes-split-strategy gets in addition to the
; ordinary usable and sos also splitinfo:
; the selected element of *strat-list* headed by 'split
; and containing:
;  - split marker symbol
;  - time given for the split attempt
;  - clause in sos which is split
;  - splittings of the clause which is split
;  - strategy list to be used for all splits

(define *split-search-active-flag* #f)

(define (prove-all-passes-split-strategy usable sos
                   demodulators lit-demodulators compl-lit-demodulators
                   splitinfo)

  ;(newline)
  ;(display "prove-all-passes-split-strategy  called with splitinfo: ")
  ;(display splitinfo)
  ;(newline)

  (let ((origstratlist (copy-tree *strat-list*)) ; save original stratlist
        (savedstratlist '())
        (trynr 0)
        (tmp '())
        (tmpstrat '())
        (res '())
        (splittime '())
        (splitclause '())
        (splitparts '())
        (splitstrats '())
        (splitsos '())
        (okflag #t)
        (failflag #f) )
    (set! *split-search-active-flag* #t)
    ; splittime is the time given to the full split attempt
    ; for all splits summed
    (set! splittime (car (cdr splitinfo)))
    ; splitclause is the selected clause in sos which is split
    (set! splitclause (car (cddr splitinfo)))
    ; splitparts contains all N splittings of the clause
    (set! splitparts (car (cdddr splitinfo)))
    ; splitstrats is the list of strategies to be passed
    (set! splitstrats (car (cddddr splitinfo)))

    (if (>= *display-info-level* 20)
        (begin
          (newline)
          (display "Starting a split proof attempt with ")
          (display (length splitparts))
          (display " components.")
          (newline) ))

    (do ((splitslst splitparts (cdr splitslst)))
        ((or (null? splitslst) failflag)
           ; no more active split search
           (set! *split-search-active-flag* #f)
           ; first restore original stratlist
           (set! *strat-list* origstratlist)
           (if (>= *display-info-level* 30)
               (begin
                 (newline)
                 (display "Split attempt finished with ")
                 (if failflag
                     (display "FAILURE.")
                     (display "SUCCESS.") )
                 (newline) ))
           (if failflag
               (begin
                 (clear-space-for-split)
                 #f)
               (begin
                 ; (set! *foundflag* #t)
                 (clear-space-for-split)
                 #t)))

     (set! trynr (+ 1 trynr))
     (if (>= *display-info-level* 30)
         (begin
          (newline)
          (display "Split component ")
          (display trynr)
          (display " started.")
          (newline) ))

      (clear-space-for-split)
      ; make-split-for-sos replaces the split clause in
      ; sos with a selected splitting from splitparts
      (set! splitsos (make-split-sos sos splitclause (car splitslst)))
      (set! *strat-list* splitstrats)
      ; increase times for split
      (set! tmpstrat '())
      ;(newline)
      ;(display "initial *strat-list*:")(display *strat-list*)
      ;(newline)
      (do ((part *strat-list* (cdr part)))
          ((null? part)
             (set! *strat-list* (reverse tmpstrat)) )
        (if (null? (cdr part))
            (set! tmpstrat
               (cons
                (cons (caar part)
                      (cons (* 10 (cadar part))
                            (cddar part) ))
                tmpstrat))
            (set! tmpstrat
               (cons
                (cons (caar part)
                      (cons (* 2 (cadar part))
                            (cddar part) ))
                tmpstrat))))
      ;(newline)
      ;(display "new *strat-list*:")(display *strat-list*)
      ;(newline)
      (set! *strat-selected* (car *strat-list*))

      ;(newline)
      ;(display "ABOUT TO CALL IN-SPLIT PROVE-ALL-PASSES-LOOP-STRATEGIES")
      ;(newline)
      ;(display "WITH *STRAT-LIST*: ")
      ;(display *strat-list*)
      ;(newline)

      (set! *foundflag* #f)
      (set! savedstratlist *strat-list*)

      (set! okflag
           (prove-all-passes-loop-strategies usable splitsos
              demodulators lit-demodulators compl-lit-demodulators))

      ;(set! okflag (prove-all-passes-single-strategy  usable splitsos
      ;                   demodulators lit-demodulators compl-lit-demodulators))

      ;(newline)(display "COMPUTED okflag: ")(display okflag)(newline)

      (set! *foundflag* #f)
      (set! *strat-list* savedstratlist)
      (set! *strat-list* (cdr *strat-list*))

      ;(newline)
      ;(display "IN-SPLIT PROVE-ALL-PASSES RETURNED: ")
      ;(display okflag)

      (if (not okflag)
          (set! failflag #t) ) )))


; clear-space-for-split removes old clauses and
; discrimination tree before each split attempt and
; after all splits have been done:
; this is crucial for correctness

(define (clear-space-for-split)
  (let ()
    (set! *sos-clauses-list* '())
    (set! *usable-clauses-list* '())
    (set! *subsumption-list* (list (list 0))) ; a unique '((0))
    (set! *unit-alist* '())
    (set! *discr-tree* (list '())) ; a unique (list '())
    (set! *nonunit-discr-tree* (list '())) ; a unique (list '())
    (set! *discr-tree-demod-terms* (list '()))
    (set! *discr-tree-demod-atoms* (list '()))

    (set! *discr-tree-list* '())
    (set! *unit-alist-list* '())
    (set! *sos-clauses-list-list* '())
    (set! *usable-clauses-list-list* '()) ))


(define (make-split-sos sos splitclause splitpart)
  (let ((tmp '())
        (res '())
        (foundflag #f) )

    (if #t ; (> *display-info-level* 30)
        (begin
          (newline)
          (display "START OF PROOFPART")
          (newline)
          (display "Making new sos for split:")(newline)
          (display "Original clause to be split: ")(newline)
          (print-pure-clause  splitclause)(newline)
          (display "Split part used next: ")
          (print-pure-clause splitpart)
          (newline)
          (display "END OF PROOFPART") ))

    (do ((sospart sos (cdr sospart)))
        ((null? sospart))
      (if (and (not foundflag)
               (equal? (car sospart) splitclause) )
          (begin
            (set! res (cons splitpart res))
            (set! foundflag #t) )
          (set! res (cons (car sospart) res)) ))
    ;(display "computed sos: ")(display res)(newline)
    res))

; prove-all-passes-single-strategy will return the
; result:
; #t iff proved
; #f iff not proved

(define (prove-all-passes-single-strategy  usable sos
                  demodulators lit-demodulators compl-lit-demodulators)

  ;(newline)
  ;(display "prove-all-passes-single-strategy  called with strat-list: ")
  ;(display *strat-list*)
  ;(newline)

  (let ((iterate-depth-flag #t)
        (tmp '())
        (tmpusable '())
        (tmpsos '()) )
   ; ----- pseudosplitting -----
    (do ((part sos (cdr part)))
        ((null? part))
      (if *sat-strategy-flag*
          (set! tmp (cons '() '()))
          (set! tmp (pseudo-split-input-clause (cons '() (car part)))) )

      ;(display (car part))
      ;(set! tmp (split-flat-clause-into-components (cons '() (car part))))
      ;(newline)
      ;(display "Split-flat-clause-into-components results in:")
      ;(newline)
      ;(for-each (lambda(y)(display y)(newline)) tmp)

      (if (not (null? (cdr tmp)))
          (begin
            (if (> *display-info-level* 30)
                (begin
                  (newline)
                  (display "Orig sos clause for pseudosplit: ")(newline)
                  (print-pure-clause (car part))
                  (newline)
                  (display "Pseudo-split-input-clause results in: ")
                  (newline)
                  (for-each (lambda(y)
                             ;(display y)(newline)
                             (print-pure-clause (cdr y))
                             (newline))
                            tmp)))
            (set! tmpsos (append (map (lambda(x)(cdr x)) tmp) sos)) )
          (begin
            (set! tmpsos (cons (car part) tmpsos)) )))
    (set! sos (reverse tmpsos))

    ;(do ((part usable (cdr part)))
    ;    ((null? part))
    ;  (set! tmp (pseudo-split-input-clause (cons '() (car part))))

      ;(display (car part))
      ;(set! tmp (split-flat-clause-into-components (cons '() (car part))))
      ;(newline)
      ;(display "Split-flat-clause-into-components results in:")
      ;(newline)
      ;(for-each (lambda(y)(display y)(newline)) tmp)

    ;  (if (not (null? (cdr tmp)))
    ;      (begin
    ;
    ;        (newline)
    ;        (display "Orig usable clause for pseudosplit: ")(newline)
    ;        (print-pure-clause (car part))
    ;        (newline)
    ;        (display "Pseudo-split-input-clause results in: ")
    ;        (newline)
    ;        (for-each (lambda(y)
    ;               ;(display y)(newline)
    ;               (print-pure-clause (cdr y))
    ;               (newline))
    ;            tmp)
    ;
    ;        (set! tmpusable (append (map (lambda(x)(cdr x)) tmp) usable)) )
    ;      (begin
    ;        (set! tmpusable (cons (car part) tmpusable)) )))
    ;(set! usable (reverse tmpusable))

    ;(abort)
    ; ---- pseudosplitting ends ------

    (do ()
        ((not iterate-depth-flag)
               ; compute the result to return
            (if *foundflag*
                (begin
                  #t)
                (begin
                  #f)) )
       ; clear strategical flags

      (init-one-pass-strategical-flags)
      (pick-strat-from-list)
      (pick-extra-limits)
      ; (pick-midsearch-limits)

      ;(newline)
      ;(display "USABLE: ")
      ;(newline)
      ;(for-each
      ;  (lambda(x)
      ;    (print-pure-clause x)(newline)
      ;    (display x)(newline) )
      ;  usable)
      ;(newline)
      ;(display "SOS: ")
      ;(newline)
      ;(for-each
      ;  (lambda(x)
      ;    (print-pure-clause x)(newline)
      ;    (display x)(newline) )
      ;  sos)


      ; do one pass
      (prove-one-pass usable sos
                      demodulators lit-demodulators compl-lit-demodulators)
      ; check if pass was stopped too early
      (set! *strat-list* (cdr *strat-list*))
      (set! *strat-passed-time*
            (- (get-check-time 50) *strat-start-time*) )
      (store-time-checkpoint! 50)
       ;(display " 1 ")
      (set! *strat-left-time* (- *strat-given-time* *strat-passed-time*))
      ;(display " 2 ")
      ;(newline)
      ;(display "strat-given-time: ")
      ;(display strat-given-time)
      ;(newline)
      ;(display "strat-start-time: ")
      ;(display strat-start-time)
      ;(newline)
      ;(display "strat-left-time: ")
      ;(display strat-left-time)
      ;(newline)
      ;(display "strat-passed-time: ")
      ;(display strat-passed-time)
      ;(newline)


      (if (and (not *foundflag*)
               (not *sat-strategy-flag*)
                   (strat-poses-depth-limit? *strat-selected*)
                   (> *strat-left-time* (* 1 *internal-time-units-per-second*))
                   (> *strat-left-time* *strat-passed-time*)
                   (<= (strat-depth-limit *strat-selected*) 20) )
          (begin (set! *strat-list*
                       (cons
                         (set-strat-time
                          (increase-strat-depth-limit *strat-selected*)
                          (quotient *strat-left-time*
                                    *internal-time-units-per-second*))
                         *strat-list*))
                 (set! *strat-selected* (car *strat-list*))
                 ;(newline)
                 ;(display "*strat-list*: ")
                 ;(display *strat-list*)
                 ;(newline)

                 (set! iterate-depth-flag #t) )
          (set! iterate-depth-flag #f) ))))


(define (strat-poses-depth-limit? strat)

  ;(newline)
  ;(display "strat-poses-depth-limit? on: ")
  ;(display strat)
  ;(newline)

  ; example: (hyper-order 18 #f 2 5) (hyper 9 #t)

  ;(display " 4 ")

  (and (not (eq? 'split (car strat)))
       (not (null? (cdr strat)))
       (not (null? (cddr strat)))
       (not (null? (cdddr strat)))
       (< 0 (car (cdddr strat))) ))


(define (strat-depth-limit strat)
  ; example: (hyper-order 18 #f 2 5) (hyper 9 #t)

  ;(display " 4 ")
  (if (and (not (null? (cdr strat)))
           (not (null? (cddr strat)))
           (not (null? (cdddr strat))) )
       (car (cdddr strat))
       0))

(define (set-strat-time strat time)
  (cons (car strat)
        (cons time
              (cddr strat))) )


(define (increase-strat-depth-limit strat)
  ; example: (hyper-order 18 #f 2 5) (hyper 9 #t)
  ;(display " 5 ")

  (cons (car strat)
        (cons (cadr strat)
              (cons (caddr strat)
                    (cons (+ 1 (car (cdddr strat)))
                          (cdr (cdddr strat)) )))))


; pick-strat-from-list takes the first *strat-list* el
; and calls stratflag functions on it

(define (pick-strat-from-list)
  (let* ((el (car *strat-list*))
         (stratname (car el)) )
  ; pick a strategy from the given list
    (cond
     ((eq? (car el) 'binary-epr-order)
      (init-epr-order-strat) )
     ((eq? (car el) 'binary-posweight-kb-big-order)
      (init-posweight-ordered-binres-kb-big-strat) )
     ((eq? (car el) 'binary-posweight-kb-small-order)
      (init-posweight-ordered-binres-kb-small-strat) )

     ((eq? (car el) 'binary-posweight-order-kb-small)
      (init-posweight-ordered-binres-kb-small-strat) )
     ((eq? (car el) 'binary-posweight-order-kb-small-fulleq)
      (init-posweight-ordered-binres-kb-small-fulleq-strat) )

     ((eq? (car el) 'binary-posweight-lex-big-order)
      (init-posweight-ordered-binres-lex-big-strat) )
     ((eq? (car el) 'binary-posweight-lex-small-order)
      (init-posweight-ordered-binres-lex-small-strat) )
     ((eq? (car el) 'binary-posweight-firstpref-order)
      (init-posweight-ordered-firstpref-strat) )
     ((eq? (car el) 'binary-posweight-order)
      (init-posweight-ordered-binres-strat) )
     ((eq? (car el) 'hyper-order-uniteq)
      (init-ordered-hyper-uniteq-strat) )
     ((eq? (car el) 'binary-unit-uniteq)
      (init-unit-uniteq-strat) )
     ((eq? (car el) 'binary-bliksem)
      (init-binres-bliksem-strat) )
     ((eq? (car el) 'binary-rootpara)
      (init-binary-unit-depth-root-backdemod-alt) )
     ((eq? (car el) 'binary-nameorder-uniteq)
      (init-name-ordered-binres-uniteq-strat) )
     ((eq? (car el) 'binary-order-uniteq)
      (init-ordered-binres-uniteq-strat) )
     ((eq? (car el) 'binary-uniteq)
      (init-binres-uniteq-strat) )
     ((eq? (car el) 'binary-nameorder-sos)
       (init-name-ordered-binres-sos-strat) )
     ((eq? (car el) 'binary-order)
       (init-ordered-binres-strat) )
     ((eq? (car el) 'binary-order-sos)
       (init-ordered-binres-sos-strat) )
     ((eq? (car el) 'binary-posweightorder-sos)
       (init-posweight-ordered-binres-sos-strat) )
     ((eq? (car el)  'binary-order-fulleq)
       (init-ordered-binres-fulleq-strat) )
     ((eq? (car el)  'binary-weightorder)
       (init-weight-ordered-binres-strat) )
     ((eq? (car el)  'binary-weightorder-sos)
       (init-weight-ordered-binres-sos-strat) )

     ; unsat propositional

     ((eq? (car el) 'prop-unsat-nofun-strat)
        (init-prop-unsat-nofun-strat) )

     ; satisfiability ones

     ((eq? (car el) 'sat-hyper-strat)
        (init-sat-hyper-strat) )
     ((eq? (car el) 'sat-ordered-binres-strat)
        (init-sat-ordered-binres-strat) )
     ((eq? (car el) 'sat-posweight-order)
        (init-sat-posweight-ordered-binres-strat) )
     ((eq? (car el) 'sat-weight-ordered-binres-strat)
        (init-sat-weight-ordered-binres-strat) )
     ((eq? (car el) 'simple-model-strat)
       (init-simple-model-strat) )
     ((eq? (car el) 'mace-model-strat)
       (init-mace-model-strat) )

     ((eq? (car el) 'sem-model-strat)
       (init-sem-model-strat) )
     ((eq? (car el) 'sem-model-strat-0)
       (init-sem-model-strat)
       (set! *props-selection-strategy* 0) )
     ((eq? (car el) 'sem-model-strat-1)
       (init-sem-model-strat)
       (set! *props-selection-strategy* 1) )
     ((eq? (car el) 'sem-model-strat-2)
       (init-sem-model-strat)
       (set! *props-selection-strategy* 2) )
     ((eq? (car el) 'sem-model-strat-3)
       (init-sem-model-strat)
       (set! *props-selection-strategy* 3) )
     ((eq? (car el) 'sem-model-strat-4)
       (init-sem-model-strat)
       (set! *props-selection-strategy* 4) )
     ((eq? (car el) 'sem-model-strat-5)
       (init-sem-model-strat)
       (set! *props-selection-strategy* 5) )

     ((eq? (car el) 'mace-nofuns-model-strat)
       (init-mace-nofuns-model-strat) )
     ((eq? (car el) 'random-model-strat)
       (init-random-model-strat) )

     ; old ones

     ((eq? (car el) 'binary-hyper-sos)
      (init-binary-hyper-sos-strat) )
     ((eq? (car el) 'hyper)
      (init-hyper-strat) )
     ((eq? (car el) 'hyper-order)
      (init-ordered-hyper-strat) )
     ((eq? (car el) 'hyper-wos)
      (init-hyper-wos-strat) )
     ((eq? (car el) 'binary)
      (init-binres-strat) )
     ((eq? (car el) 'binary-unit)
      (init-unit-strat) )
     ((eq? (car el) 'binary-double)
      (init-double-strat) )
     ((eq? (car el) 'binary-order)
      (init-ordered-binres-strat) )
     ((eq? (car el) 'binary-nameorder)
      (init-name-ordered-binres-strat) )
     ((eq? (car el) 'binary-unit-depth)
      (init-binary-unit-depth) )
     ((eq? (car el) 'binary-unit-depth-sospref)
      (init-binary-unit-depth-sospref) )
     ((eq? (car el) 'binary-unit-depth-backdemod)
      (init-binary-unit-depth-backdemod) )
     ((eq? (car el) 'binary-unit-depth-sospref-alt)
      (init-binary-unit-depth-sospref-alt) )
     ((eq? (car el) 'binary-unit-depth-backdemod-alt)
      (init-binary-unit-depth-backdemod-alt) )
     ((eq? (car el) 'binary-unit-lex-backdemod)
      (init-binary-unit-lex-backdemod) )
     ((eq? (car el) 'binary-unit-lex)
       (init-binary-unit-lex) )
     ((eq? (car el) 'binary-unit-firstpref-backdemod)
      (init-binary-unit-firstpref-backdemod) )
     ((eq? (car el) 'binary-unit-firstpref)
       (init-binary-unit-firstpref) )
     (else
      (init-posweight-ordered-binres-kb-big-strat) ))


    ; pick time for a pass
    (set! el (cdr el))
    (set! *one-pass-time* (* *internal-time-units-per-second*
                             (car el) ))
    ; pick sos strategy used/not
    (set! el (cdr el))
    (set! *sos-strategy-flag* (car el))
    ; pick maxdepth and maxlength
    (set! el (cdr el))
    (if (not (null? el))
        (if (eq? stratname 'binary-bliksem)
            ; limit clause size
            (begin
               (set! *clause-finalkeepsize-limit* (car el))
               (set! *clause-keepsize-limit* 0)
               (set! el (cdr el)) )
            ; limit clause depth
            (begin
               (set! *clause-keepdepth-limit* (car el))
               (set! el (cdr el)) )))
    (if (not (null? el))
        (begin (set! *clause-keeplength-limit* (car el))
               (set! el (cdr el)) )) ))

    ;; ; remove first el from stratlist
    ;; (set! *strat-list* (cdr *strat-list*)) ))


(define (pick-extra-limits)
  (let ((tmp '()))
    (cond
       (#t #f)
       ((or *sat-strategy-flag*
            *complete-unsat-strategy*
            *complete-sat-strategy*)
           (set! *clause-keepsize-limit* 0)
           (set! *clause-finalkeepsize-limit* 0)
           (set! *clause-keepdepth-limit* 0)
           (set! *clause-keeplength-limit* 0) )
       ;((or (not (eq? *clause-keepsize-limit* 0))
       ;     (not (eq? *clause-keepdepth-limit* 0))
       ;     (not (eq? *clause-keeplength-limit* 0)) ))
       (else
           ;(set! *input-maxsize-nr* 0)
           ;(set! *input-maxdepth-nr* 0)
           ;(set! *input-avgdepth-nr* 0)
           ;(set! *input-avgsize-nr* 0)
          ; set max depth if not set already
          (if (eq? *clause-keepdepth-limit* 0)
              (begin
                (set! *clause-keepdepth-limit*
                      (* 20 *input-avgdepth-nr*) )
                (if (> (* 10 *input-maxdepth-nr*)
                       *clause-keepdepth-limit*)
                    (set! *clause-keepdepth-limit*
                          (* 10 *input-maxdepth-nr*) ))
                ;(set! *clause-keepdepth-limit* 5)
                (if (< *clause-keepdepth-limit* 50)
                    (set! *clause-keepdepth-limit* 50) )
                ))
          ; set max size if not set already
          (if (eq? *clause-keepsize-limit* 0)
              (begin
                (set! *clause-keepsize-limit*
                      (* 20 *input-avgsize-nr*) )
                (if (> (* 10 *input-maxsize-nr*)
                       *clause-keepsize-limit*)
                    (set! *clause-keepsize-limit*
                          (* 10 *input-maxsize-nr*) ))
                (if (< *clause-keepsize-limit*  200)
                    (set! *clause-keepsize-limit* 200) )
                ;(set! *clause-keepsize-limit* 30)
                ))
          ; set finalkeepsize if too big
          (if (> *clause-finalkeepsize-limit*
                 *clause-keepsize-limit*)
              (set! *clause-finalkeepsize-limit*
                    *clause-keepsize-limit*) )))))


;(define (pick-midsearch-limits)
;  (let ((tmp '()))
;    (cond
;       ((or *sat-strategy-flag*
;            *complete-unsat-strategy*
;            *complete-sat-strategy*)
;           (set! *clause-midsearch-keepsize-limit* 0) )
;       ;((or (not (eq? *clause-keepsize-limit* 0))
;       ;     (not (eq? *clause-keepdepth-limit* 0))
;       ;     (not (eq? *clause-keeplength-limit* 0)) ))
;       (else
;           (set! *clause-midsearch-keepsize-limit*
;                 (* 10 *input-avgsize-nr*) )
;           (if (> (* 5 *input-maxsize-nr*)
;                  *clause-midsearch-keepsize-limit*)
;               (set! *clause-midsearch-keepsize-limit*
;                     (* 5 *input-maxsize-nr*) ))))))


; build-strat-list is called for auto mode
; before any actual strategies are run.
; It determines the useful strategies
; and builds *strat-list*


(define (build-strat-list  usable sos
           demodulators lit-demodulators compl-lit-demodulators)
  (let ((tmp '())
        (origclauses '()) )
    (set! origclauses
         (list  usable sos
               demodulators lit-demodulators compl-lit-demodulators))
    ; (newline)(display "starting to make initial clauses")(newline)
    (set! tmp
          (prove-make-initial-clauses usable sos
              demodulators lit-demodulators compl-lit-demodulators))

    (set! usable (car tmp))
    (set! tmp (cdr tmp))
    (set! sos (car tmp))
    (set! tmp (cdr tmp))
    (set! demodulators (car tmp))
    (set! tmp (cdr tmp))
    (set! lit-demodulators (car tmp))
    (set! tmp (cdr tmp))
    (set! compl-lit-demodulators (car tmp))
    (set! tmp (cdr tmp))

    (set! *no-passes-started* #t)
    (if *no-passes-started*
        (begin
          (prove-show-input-clauses usable sos
                demodulators lit-demodulators compl-lit-demodulators)
          (analyse-input usable sos)
          (if *auto-mode-flag*
              (begin
                (select-initial-strategy)
                (if (and *splitting-allowed-flag*
                         (not *sat-strategy-flag*) )
                    (build-split-stratlist usable sos
                       demodulators lit-demodulators compl-lit-demodulators
                       origclauses))))
          (set! *no-passes-started* #f) )) ))


; split selection algorithm for selecting the best split:
;
; - the split clause with a largest amount of components is selected
; - if there are several split candidates with a largest amount
;   of components, then we select the one with the longest original
; - if the lengths of originals are same, we select one the original
;   of which has a largest amount of positive atoms
;
; deciding whether and how to use the split clause found (assuming
; we find one with S components and O original length):
;
; - we will not use a split clause if it contains over 10 components
; - if the original clause is at least twice longer than any other
;   then we use it
; - if the sos contains less than four clauses then we use it
; - otherwise we will not use the split clause
;
; deciding time and strategies for split
;
; - if split is used for a small clause set we give it half time
; - if split is used for a medium or large clause set we give it
;    1/3 of time
; - splitting strategies are run after all the non-split strategies
;
; in the future we will modify split component strategy times so
; that the first split always gets 1/2 of the remaining time overall
; and after it is successful, the times for other splits are computed again

(define (build-split-stratlist usable sos
           demodulators lit-demodulators compl-lit-demodulators
           origclauses)
  (let ((tmp '())
        (origsos '())
        (origstratlist *strat-list*)
        (simpleorigstratlist '())
        (origstrattime 0)
        (stratlist '())
        (splitflag #f)
        (splittime 0)
        (splittimequotient 1)
        (splitclause '())
        (splits '())
        (onesplit '())
        (onesplitlen 0)
        (res '())
        (maxsplitcomponents 0)
        (maxsplitnrvars 0)
        (splitnrvars 0)
        (maxoriglen 0)
        (maxposlit 0)
        (maxsplitoriglen 0)
        (maxsecondoriglen 0)
        (newnosplitlst '())
        (newsplitpartlst '())
        (soslen 0) )


    ;(newline)
    ;(display "build-split-stratlist called")
    ;(newline)
    ;(display "origclauses:")
    ;(display origclauses)
    ;(newline)
    ;(display "origstratlist: ")
    ;(newline)
    ;(display origstratlist)
    ;(newline)

    (if (and (not (null? origclauses))
             (not (null? (cdr origclauses))) )
        (set! origsos (cadr origclauses)) )

    ; get sos statistics
    (set! soslen (length origsos))
    (do ((sospart origsos (cdr sospart)))
        ((null? sospart))
      (set! tmp (length (car sospart)))
      (if (> tmp maxoriglen)
          (set! maxoriglen tmp)
          (if (> tmp maxsecondoriglen)
              (set! maxsecondoriglen tmp) )))

    ; loop over sos elems, checking for splits
    (for-each
      (lambda(x)

        ;(newline)
        ;(display "orig clause: ")(newline)
        ;(print-pure-clause x)
        ;(newline)
        ;(display x)

        (if (>= (length x) maxsplitcomponents)
            (set! onesplit (split-full-clause (cons '() x)))
            (set! onesplit '()) )

        ;(newline)
        ;(display "is split as: ")
        ;(newline)
        ;(display onesplit)

        (set! onesplitlen (length onesplit))
        (set! splitnrvars 0)
        (for-each
          (lambda(y)(set! splitnrvars (+ splitnrvars (length (car y)))))
          onesplit)

        (if (and (> onesplitlen 1)
                 (<= onesplitlen 10)
                 (or (>= (length x) (* 2 maxsecondoriglen))
                     (<= soslen 4) )
                 (or (> splitnrvars (+ 1 maxsplitnrvars))
                     (and (>= splitnrvars maxsplitnrvars)
                          (or (> onesplitlen maxsplitcomponents)
                              (and (eq? onesplitlen maxsplitcomponents)
                                   (> (length x) maxsplitoriglen) )))))
            (begin
              (set! splitflag #t)
              (set! splitclause x)
              (set! splits (map (lambda(x)(cdr x)) onesplit))
              (set! maxsplitcomponents onesplitlen)
              (set! maxsplitnrvars splitnrvars)
              (set! maxsplitoriglen (length x)) )))

       origsos)

    ; show whether splitting will be used

    (if (and splitflag
             (>= *display-info-level* 30))
        (begin
          (newline)
          (display "SOS clause ")(newline)
          (print-pure-clause splitclause)
          (newline)
          (display "was split for some strategies as: ")(newline)
          (for-each
            (lambda(x)
              (print-pure-clause x)
              (newline) )
             splits)))

    ; compute a new stratlist

    (if splitflag
        (begin
          (set! origstrattime 0)
          (for-each (lambda(x)
                      (set! origstrattime (+ origstrattime (cadr x))) )
                     origstratlist)
          ; more stress on splitting if only a splittable clause
          ; and less than 6 components in sos

          (if (and (eq? soslen 1)
                   (<= maxsplitcomponents 5))
              (set! splittimequotient 3)
              (set! splittimequotient 2) )

          ; simplify the strategy list for splits
          (if (and (pair? origstratlist)
                   (pair? (cdr origstratlist)) )
              (begin
                (set! simpleorigstratlist
                  (cons (car origstratlist)
                        (my-last-pair origstratlist) ))))

          (set! simpleorigstratlist origstratlist)

          (set! newnosplitlst
               (select-strategy-seconds
                  origstratlist
                  (quotient origstrattime splittimequotient) ))
          (set! newsplitpartlst
               (select-strategy-seconds
                  simpleorigstratlist
                  (quotient (- origstrattime
                               (quotient origstrattime splittimequotient))
                            maxsplitcomponents)))
          (set! res
            (append
              newnosplitlst
              (list
                (list 'split
                       splittime
                       splitclause
                       splits
                       newsplitpartlst))))
          (set! *strat-list* res) ))

    ;(set! *strat-list*
    ;      ; (cons '(binary 1 #f)
    ;            (append *strat-list* '((hyper 1 #f)) ));)

    (if (> *display-info-level* 30)
        (begin
         (newline)
         (display "new stratlist computed by split strat fun:")
         (newline)
         (for-each (lambda(x)(display x)(newline)) *strat-list*)
         (newline) )) ))


; prove-one-pass does one concrete prover pass with a certain
; strategy
;
; the main function below does the preparations:
; it calls the main proving function at the end,
; and the main proving function finally calls the
; after-proof bookkeeping function at its end

(define (prove-one-pass usable sos
           demodulators lit-demodulators compl-lit-demodulators)
  (let ((tmp '())
        (oldclauses '())
        (nr 0) )
    (if *combinedstrat-flag*
        (init-one-pass-globals-combinedstrategy)
        (init-one-pass-globals) )

    ; (newline)(display "starting to make initial clauses")(newline)
    (set! tmp
    (prove-make-initial-clauses usable sos
          demodulators lit-demodulators compl-lit-demodulators))

    (set! usable (car tmp))
    (set! tmp (cdr tmp))
    (set! sos (car tmp))
    (set! tmp (cdr tmp))
    (set! demodulators (car tmp))
    (set! tmp (cdr tmp))
    (set! lit-demodulators (car tmp))
    (set! tmp (cdr tmp))
    (set! compl-lit-demodulators (car tmp))
    (set! tmp (cdr tmp))

    ; (newline)(display "initial clauses made ok")(newline)

    ;(newline)
    ;(display "strat-list: ")(display *strat-list*)
    ;(newline)
    ;(newline)

    ; the following cond block adds selected clauses
    ; from previous passes to usable

    (cond
      ; case where one strategy is still present
      ; in the *strat-list*
      ((and (not *sat-strategy-flag*)
            (not (null? *all-passes-selected-clauses*))
            (not (null? *strat-list*))
            (null? (cdr *strat-list*)) )
        ;  one before the last
        (do ((part *all-passes-selected-clauses* (cdr part)))
            ((null? part))
         (if (and (pair? (car part))
                  (not (null? (cdar part)))
                  (null? (cddr (car part)))
                  (not (list-contains-clause (car part) usable)) )
             (set! usable (cons (car part) usable)) )))
     ; case where *strat-list* is emtpy
      ((and (not *sat-strategy-flag*)
            (not (null? *all-passes-selected-clauses*))
            (null? *strat-list*) )
        ; last one
        (do ((part *all-passes-selected-clauses* (cdr part)))
            ((null? part))
         (if (and (pair? (car part))
                  (not (null? (cdar part)))
                  (not (list-contains-clause (car part) usable)) )
             (set! usable (cons (car part) usable)) ))))

              ;(set! usable
                  ;     (append usable *all-passes-selected-clauses*) )

             ; (if  (> *display-info-level* 40)
             ;    (begin
             ;           (display "usable:  ")(newline)
             ;      (for-each (lambda(x)(print-full-clause x)(newline)) usable)
             ;      (newline) ))))

    ;(newline)(display "just added old clauses")(newline)

    ;     (prove-show-input-clauses usable sos
    ;            demodulators lit-demodulators compl-lit-demodulators)

    ;(newline)(display "starting to move usable to sos")(newline)


    ; if not sos-strategy, move all usable to sos

    (if (not *sos-strategy-flag*)
        (begin
          ; (set! sos (append usable sos)) ; old version
          (set! sos (append sos usable)) ; new version june 2003
          (set! usable '()) ))

    ; if hyper-strategy, make correct input-litinfo for nucleons

    (if *hyper-strategy-flag*
        (for-each
          (lambda(cl)
            (if (not (positive-cl? (cdr cl)))
                (set-car! (cddr (car cl))
                          (make-input-litinfolst (cdr cl)) )))
          (append usable sos) ))

    ;(newline)(display "starting to put positive to sos")(newline)
    ;(pretty-print sos)
    ;(pretty-print usable)

    ; for hyperres put positive clauses to sos, rest to usable

    (if (and *hyper-strategy-flag*
             (not *sos-strategy-flag*) )
        (do ((part sos (cdr part)))
            ((null? part)
               (set! usable (reverse tmp)) )
          (if (and (pair? (car part))
                   (positive-cl? (cdar part)) )
              (if (not (member (car part) sos))
                  (set! sos (cons (car part) sos)) )
              (if (not (member (car part) tmp))
                  (set! tmp (cons (car part) tmp)) ))))


    ;  (prove-show-input-clauses usable sos
    ;            demodulators lit-demodulators compl-lit-demodulators)

    ; if ordered strategy, make correct litinfos for all input clauses

    (if *order-sos-strategy-flag*
        (begin
          (for-each
            (lambda(cl)
              (set-car! (cddr (car cl))
                        (map cddr (clear-litorder-litinfo
                                    (map (lambda(x)(cons 0 (cons 0 x)))
                                         (caddr (car cl)) )
                                    (cdr cl) ))))
            sos)
          (if (or *order-strategy-flag*
                  *name-order-strategy-flag*
                  *exp-order-strategy-flag*
                  *epr-order-strategy-flag*
                  *firstlitpref-order-strategy-flag*
                  *posweight-order-strategy-flag*
                  *negweight-order-strategy-flag*
                  *ordered-paramodulation-flag*)
              (for-each
                (lambda(cl)
                  (set-car! (cddr (car cl))
                            (map cddr (order-prohibit-litinfos
                                       (map (lambda(x)(cons 0 (cons 0 x)))
                                            (caddr (car cl)) )
                                       (cdr cl) ))))
                usable))
          (set! sos (append sos usable))
          (set! usable '()) )
         (if (or *order-strategy-flag*
                 *name-order-strategy-flag*
                 *exp-order-strategy-flag*
                 *epr-order-strategy-flag*
                 *firstlitpref-order-strategy-flag*
                 *posweight-order-strategy-flag*
                 *negweight-order-strategy-flag*
                 *ordered-paramodulation-flag*)
             (for-each
              (lambda(cl)
                (set-car! (cddr (car cl))
                          (map cddr (order-prohibit-litinfos
                                     (map (lambda(x)(cons 0 (cons 0 x)))
                                          (caddr (car cl)) )
                                     (cdr cl) ))))
              (append usable sos) )))

    (show-one-pass-strategy)
    (if *type-theory-stratflag* (enumerate-sos-literals sos))
    (set! *sos-clauses-list* sos)
    (set! *usable-clauses-list* usable)

    ;(newline)(display "before actual proving")(newline)
    ;(show-derived)

      ;(newline)
      ;(display "BEFORE ACTUALLY PROVING:")
      ;(newline)
      ;(display "USABLE: ")
      ;(newline)
      ;(for-each
      ;  (lambda(x)
      ;    (print-clause x)(newline)
      ;    (display x)(newline) )
      ;  usable)
      ;(newline)
      ;(display "SOS: ")
      ;(newline)
      ;(for-each
      ;  (lambda(x)
      ;    (print-clause x)(newline)
      ;    (display x)(newline) )
      ;  sos)

    (prove-one-pass-actually-prove usable sos
           demodulators lit-demodulators compl-lit-demodulators
           tmp oldclauses nr) ))


; here happens actual proving:
; called only from prove-one-pass above

(define *combinedstrat-flag* #f)

(define (prove-one-pass-actually-prove usable sos
           demodulators lit-demodulators compl-lit-demodulators
           tmp oldclauses nr)
  (let ()

    ;(display "Given usable: ")
    ;(newline)
    ;(for-each (lambda(x)(print-clause x)(newline)(display x))
    ;                usable)
    ;(newline)
    ;(display "Given sos: ")
    ;(newline)
    ;(for-each (lambda(x)(print-clause x)(newline)(display x))
    ;                sos)
    ;(newline)


    (cond
       ((and *prop-unsat-strat* *simple-model-strat*)
          ; *all-passes-selected-clauses*
          (set! oldclauses
                (filter-map
                 (lambda(cl)
                   ; (newline)(print-pure-clause (cdr cl))
                   ; (display " : ")(display (fullcl-maxvar cl))
                   (if (and (pair? cl)
                            (<= (fullcl-maxvar cl) 3) )
                       cl
                       #f))
                  *all-passes-selected-clauses*))

          (newline)
          (set! tmp (prop-build-simple-model
                      (append oldclauses
                       (append *sos-clauses-list* *usable-clauses-list*) )))
          (cond
            (tmp
              (newline)
              (if (not *prop-clauses-generation-limited-flag*)
                  (begin (display "Clause set found to be **SATISFIABLE** (satisfiable)")(newline)
                         (set! *sat-found-flag* #t) ))
              (set! *sos-exhausted-flag* #t) )
            (*prop-time-ran-out*
              (display "prop search time ran out") )
            ((eq? tmp #f)
             (set! *foundflag* #t)
             (newline)
             (display "***GANDALF_FOUND_A_REFUTATION***") )))

       (*mace-nofuns-model-strat*
          (set! tmp (prop-build-nofuns-model
                       (append *sos-clauses-list* *usable-clauses-list*) ))
          (cond
              ((eq? tmp #t)
                 (newline)
                 (if (not *prop-clauses-generation-limited-flag*)
                     (begin
                       (display "Clause set found to be **SATISFIABLE** (satisfiable)")(newline)
                       (set! *sat-found-flag* #t)
                       (quit)
                       (set! *sos-exhausted-flag* #t) )
                     (set! *sos-exhausted-flag* #f) ))
              ((eq? tmp #f)
                 (newline)
                 (display "***GANDALF_FOUND_A_REFUTATION***")(newline)
                 (quit)
                 (set! *sos-exhausted-flag* #t) )
              (else
                 (newline)
                 (display "***** INCONCLUSIVE RESULT ********")
                 (set! *sos-exhausted-flag* #f) )))

       (*sem-model-strat*
          (set! tmp (props-build-model
                       (make-sat-inclauses
                          *sos-clauses-list* *usable-clauses-list*) ))
          ;(newline)
          ;(display "props-build-model returns: ")(display tmp)
          ;(newline)

          (if (eq? tmp #t)
              (begin
                 (newline)
                 (if (not *prop-clauses-generation-limited-flag*)
                     (begin
                       (display "Clause set found to be **SATISFIABLE** (satisfiable)")(newline)
                       (set! *sat-found-flag* #t) ))
                 (set! *sos-exhausted-flag* #t)) ))

       (*mace-model-strat*
          (set! tmp (prop-build-flatten-model
                       ;(make-sat-inclauses
                       ;   *sos-clauses-list* *usable-clauses-list*)
                       (append *sos-clauses-list* *usable-clauses-list*) ))
          (if tmp
              (begin
                 (newline)
                 (if (not *prop-clauses-generation-limited-flag*)
                     (begin
                       (display "Clause set found to be **SATISFIABLE** (satisfiable)")(newline)
                       (set! *sat-found-flag* #t) ))
                 (set! *sos-exhausted-flag* #t)) ))

       (*simple-model-strat*
          (set! tmp (prop-build-simple-model
                       (append *sos-clauses-list* *usable-clauses-list*) ))
          (if tmp
              (begin
                 (newline)
                 (if (not *prop-clauses-generation-limited-flag*)
                     (begin (display "Clause set found to be  **SATISFIABLE** (satisfiable)")(newline)
                            (set! *sat-found-flag* #t) ))
                 (set! *sos-exhausted-flag* #t)) ))

       (*random-model-strat*
          (set! tmp (prop-build-random-model
                       (append *sos-clauses-list* *usable-clauses-list*) ))
          (if tmp
              (begin
                 (newline)
                 (if (not *prop-clauses-generation-limited-flag*)
                     (begin (display "Clause set found to be **SATISFIABLE** (satisfiable)")(newline)
                            (set! *sat-found-flag* #t) ))
                 (set! *sos-exhausted-flag* #t)) ))

       (else
          ;(if *sat-strategy-flag* (turn-off-demodulation)) ; 25 march 2002
          ; combinedstrat will extend the input clauses by all old
          ; clauses
          (if *combinedstrat-flag*
            (begin

               ;(newline)
               ;(display "OLD CLAUSES JUST BEFORE ADDING:")(newline)
               ;(display "*sos-clauses-list-list*: ")(newline)
               ;(display *sos-clauses-list-list*)(newline)
               ;(display "*usable-clauses-list-list*:")(newline)
               ;(display *usable-clauses-list-list*)(newline)

              ; add old sos clauses to end of sos
              (if (and (pair? *sos-clauses-list-list*)
                       (pair? (car *sos-clauses-list-list*)) )
                  (set! *sos-clauses-list*
                    (append
                        *sos-clauses-list*
                        (car *sos-clauses-list-list*) )))
              ; add old usable clauses to end of sos
              (if (and (pair? *usable-clauses-list-list*)
                       (pair? (car *usable-clauses-list-list*))
                       (pair? (cdar *usable-clauses-list-list*)) )
                  (set! *sos-clauses-list*
                    (append
                        *sos-clauses-list*
                        (cdar *usable-clauses-list-list*) ))) ))
          ;(newline)
          ;(display "CLAUSES JUST BEFORE SLOW LOOP:")(newline)
          ;(display "*discr-tree*: ")(newline)
          ;(display *discr-tree*)(newline)
          ;(display "*unit-alist*:")(newline)
          ;(display *unit-alist*)(newline)
          ;(display "*sos-clauses-list*:")(newline)
          ;(display *sos-clauses-list*)(newline)
          ;(display "*usable-clauses-list*:")(newline)
          ;(display *usable-clauses-list*)(newline)
          ;(show-derived)
          ;(newline)

          (set! tmp (given-clause-slow-loop))

          ;(newline)(display "tmp: ")(display tmp)
          ;(newline)(display "*sat-strategy-flag*: ")(display *sat-strategy-flag*)
          ;(newline)(display "*complete-unsat-strategy*: ")(display *complete-unsat-strategy*)
          ;(newline)(display "*sos-exhausted-flag*: ")(display *sos-exhausted-flag*)
          ;(newline)

          (if (and tmp
                  (>= *display-info-level* 30) )
              (begin (newline)
                     (if (not (null? *answer-cl*))
                         (begin (display "answer-clause: ")
                                (print-pure-clause *answer-cl*)
                                (newline) ))
                      (if (not *prop-unsat-strat*)
                          (show-history *answer-history*)) ))

          (if (and *sat-strategy-flag*
                   *complete-unsat-strategy*
                   *sos-exhausted-flag*
                   (not tmp) )
              (begin
                 (newline)
                 (if (not *prop-clauses-generation-limited-flag*)
                     (begin (display "Clause set found to be **SATISFIABLE** (satisfiable)")(newline)
                            (set! *sat-found-flag* #t) ))
                 (set! *sos-exhausted-flag* #t) ))))

    ; - - - - do the stuff after proof attempt has finished - - -
    (cond
      (*combinedstrat-flag*
        ;(newline)
        ;(display "ONE STRAT IS FINISHED")
        ;(newline)
        (prove-one-pass-after-proving-combinedstrat
          usable sos
           demodulators lit-demodulators compl-lit-demodulators
           tmp oldclauses nr))
      (else
        (prove-one-pass-after-proving usable sos
           demodulators lit-demodulators compl-lit-demodulators
           tmp oldclauses nr) ))))

; prove-one-pass-after proving is only called from
; prove-one-pass-actually-prove above here


(define (prove-one-pass-after-proving usable sos
           demodulators lit-demodulators compl-lit-demodulators
           tmp oldclauses nr)
  (let ()
    (set! *discr-tree-list* (cons *discr-tree* *discr-tree-list*))
    (set! *unit-alist-list* (cons *unit-alist* *unit-alist-list*))
    ; for full proof showing do not keep old clauses at all!
    (if *full-proof-strategy-flag*
        (begin
          (set! *discr-tree-list* '())
          (set! *unit-alist-list* '())
          (set! *sos-clauses-list-list* '())
          (set! *usable-clauses-list-list* '()) ))
    ; should we keep all clauses?
    (if *sat-strategy-flag* (set! *keep-all-old-clauses-flag* #f))
    (if (and (not *sat-strategy-flag*)
              *keep-all-old-clauses-flag*)
        (begin
          (set! *sos-clauses-list-list*
                (cons *sos-clauses-list* *sos-clauses-list-list*) )
          (set! *usable-clauses-list-list*
                (cons *usable-clauses-list* *usable-clauses-list-list*) )))
    ; keep selected good clauses for later use
    (if (and ; (not *sat-strategy-flag*)
             (not *split-search-active-flag*)
             (not *foundflag*) )
        (if  *sat-strategy-flag*
            (begin
              (set! *all-passes-selected-clauses*
                  (append (select-best-clauses 20 20 10 10 *sos-clauses-list*)
                          *all-passes-selected-clauses*))
              (set! *all-passes-selected-clauses*
                  (append (select-best-clauses 20 20 10 10 *usable-clauses-list*)
                          *all-passes-selected-clauses*)))
            (begin
              (set! *all-passes-selected-clauses*
                  (append (select-best-clauses 60 60 30 20 *sos-clauses-list*)
                          *all-passes-selected-clauses*))
              (set! *all-passes-selected-clauses*
                  (append (select-best-clauses 40 40 30 20 *usable-clauses-list*)
                          *all-passes-selected-clauses*)))))
    (if (> (length  *all-passes-selected-clauses*) 5000)
        (set! *all-passes-selected-clauses* '()) )
    ; if half of clause kept resource used, throw away old units
    (set! nr 0)
    (for-each (lambda(alist)
                (for-each
                 (lambda(el) (set! nr (+ (length (cdr el)) nr)))
                 alist))
              *unit-alist-list*)
    (if (and
         (not *keep-all-old-clauses-flag*)
         (or (> (* 2 nr) *max-kept-clauses-nr*)
             (> (* 2 *kept-size-sum-nr*) *max-size-sum-nr*)))
        (begin
          (if (>= *display-info-level* 38)
              (begin (newline)
                     (display "old unit clauses discarded")
                     (newline) ))
          (set! *discr-tree-list* '())
          (set! *unit-alist-list* '())
          (set! *sos-clauses-list-list* '())
          (set! *usable-clauses-list-list* '())
          (set! *kept-size-sum-nr* 0) ))

    tmp))


(define (prove-one-pass-after-proving-combinedstrat usable sos
           demodulators lit-demodulators compl-lit-demodulators
           tmp oldclauses nr)

  ;(newline)
  ;(display "CLAUSES AT THE CURRENT MOMENT:")(newline)
  ;(display "*discr-tree*: ")(newline)
  ;(display *discr-tree*)(newline)
  ;(display "*unit-alist*:")(newline)
  ;(display *unit-alist*)(newline)
  ;(display "*sos-clauses-list*:")(newline)
  ;(display *sos-clauses-list*)(newline)
  ;(display "*usable-clauses-list*:")(newline)
  ;(display *usable-clauses-list*)(newline)
  ;(show-derived)
  ;(newline)

  (let ()
    (set! *discr-tree-list* (cons *discr-tree* *discr-tree-list*))
    (set! *unit-alist-list* (cons *unit-alist* *unit-alist-list*))
    (set! *sos-clauses-list-list*
          (cons *sos-clauses-list* *sos-clauses-list-list*) )
    (set! *usable-clauses-list-list*
          (cons *usable-clauses-list* *usable-clauses-list-list*) )
    ; keep selected good clauses for later use
    ;(if (and (not *sat-strategy-flag*)
    ;         (not *split-search-active-flag*)
    ;         (not *foundflag*) )
    ;  (begin
    ;        (set! *all-passes-selected-clauses*
    ;              (append (select-best-clauses 60 30 30 *sos-clauses-list*)
    ;                      *all-passes-selected-clauses*))
    ;        (set! *all-passes-selected-clauses*
    ;              (append (select-best-clauses 40 20 15 *usable-clauses-list*)
    ;                      *all-passes-selected-clauses*))))


    tmp))


;------------------------------------------------------------
;
;  making inclauses for sat
;
;-----------------------------------------------------------

(define (make-sat-inclauses sos usable)
  (let ((res '())
        (old '())
        (vars '())
        (tmp '())
        (count 0) )
    (set! res (append  *sos-clauses-list* *usable-clauses-list*))

    ;(newline)
    ;(display "sos:")
    ;(newline)
    ;(for-each (lambda(x)
    ;            (print-clause x)(newline) )
    ;          *sos-clauses-list*)
    ;(newline)
    ;(display "usable:")
    ;(newline)
    ;(for-each (lambda(x)
    ;            (print-clause x)(newline) )
    ;          *usable-clauses-list*)
    ;  (newline)
    ;(display "all-passes-selected-clauses:")
    ;(newline)
    ;(for-each (lambda(x)
    ;            (print-clause x)(newline) )
    ;          *all-passes-selected-clauses*)

    ;(newline)
    ;(display "Selecting old")
    ;(newline)

    (do ((part *all-passes-selected-clauses* (cdr part)))
        ((or (null? part) (> count 500)))
      (set! tmp (car part))
      (if (and (pair? tmp) (pair? (car tmp)) (pair? (cdr tmp))
               (not (eq? 0 (caar tmp))) )
          (begin
            (set! vars (logand *cl-maxvar-mask* (caar tmp)))
            (set! vars (shiftright vars *cl-maxvar-shift*))

            ;(newline)
            ;(display vars)(display " : ")(print-clause tmp)

            (if (< vars 4)
                (begin
                  (set! old (cons tmp old))
                  (set! count (+ 1 count))
                  ;(newline)
                  ;(display "OK")
                  ;(newline)
                  )
                (begin
                  ;(newline)
                  ;(display "NO")
                  ;(newline)
                  )))))
    (set! res (append res old))
    res))

;------------------------------------------------------------
;
;  helper functions to the proving main algorithms above
;
;-----------------------------------------------------------


(define (list-contains-clause cl lst)
  (let ((foundflag #f))
    (set! cl (cdr cl))
    (do ((part lst (cdr part)))
        ((or (null? part) foundflag)
           foundflag)
        (if (and (pair? (car part))
                 (pair? (cdar part))
                 (equal? cl (cdar part)) )
            (set! foundflag #t) ))))


(define (fullcl-maxvar cl)
   (shiftright (logand *cl-maxvar-mask* (caar cl))
               *cl-maxvar-shift*))

(define (select-best-clauses poseqnr posnr negnr dblnr clauselist)
  (let ((res '())
          (reseqpos '())
          (respos '())
          (resneg '())
          (resdbl '())
          (cl '())
          (oldcl '())
          (clvalue 0)
          (cllen 0)
          (clhist '())
          (oldvalue 0)
          (found-unit-pos-nr 0)
          (found-unit-neg-nr 0)
          (found-dbl-nr 0)
          (foundnr 0) )


    ;(newline)
    ;(display "select-best-clauses starts")
    ;(newline)

    (do ((lst clauselist (cdr lst)))
        ((null? lst)
         (set! res (append reseqpos (append respos (append resneg resdbl))))

         (if (> *display-info-level* 50)
            (begin
                 (newline)
                 (display "--------------------")
                 (newline)
                 (for-each (lambda(x)
                     (print-full-clause x)
                     (display " ")
                     (display (logand *term-depth-strength-mask* (caar x)))
                     (display " cnstnr: ")
                     (display (logand *const-number-mask* (caar x)))
                     (display " tdepth: ")
                     (display (fun-real-depth (caar x)))
                     (newline))
                    res)
                 (newline)
                 (display "--------------------")
                 (newline) ))

         res)
     (if (and (pair? lst)
              (pair? (car lst))
              (pair? (caar lst)) )
           (begin
              (set! cl (car lst))

              ;(newline)(print-full-clause cl)

              (set! clvalue (logand *term-depth-strength-mask* (caar cl)))
              (set! cllen (shiftright (logand *cl-len-mask* (caar cl)) *cl-len-shift*))
              (set! clhist (clinfo-history-sublist (car cl)))
              (if (pair? clhist) (set! clhist (car clhist)) (set! clhist '()))
              (cond
                ((zero? clvalue)) ; do not use that (back-subsumed)
                ((not (pair? clhist))) ; do not use that (erroneous hist)
                ((eq? 'input (car clhist))) ; do not use that (not derived)
                ((eqv? cllen 2)
                   (set! resdbl (insert-to-best-res cl clvalue resdbl dblnr)) )
                ((and (eqv? cllen 1)
                      (pos-fun? (caadr cl))
                      (eq? (pure-fun (caadr cl)) *eq-pred-nr*) )
                   (set! respos (insert-to-best-res cl clvalue reseqpos poseqnr)) )
                ((and (eqv? cllen 1) (pos-fun? (caadr cl)))
                   (set! respos (insert-to-best-res cl clvalue respos posnr)) )
                ((and (eqv? cllen 1) (neg-fun? (caadr cl)))
                   (set! resneg (insert-to-best-res cl clvalue resneg negnr)) )))))))



(define (insert-to-best-res cl clvalue lst nr)
  (let ((tmp '())
          (oldcl '())
          (oldclvalue 0)
          (inserted-flag #f) )

    ;(newline)
    ;(display "insert-to-best-res starts with clause ")
    ;(print-full-clause cl)(display " ")
    ;(display clvalue)(display " ")
    ;(display nr)
    ;(display lst)
    ;(newline)

    (do ((part lst (cdr part)))
        ((or (null? part) inserted-flag)

            ;(display "end")(newline)

            (set! tmp (length lst))
            (cond
                    ((zero? tmp)
                       (set! lst (cons cl '())) )
                    ((and (not inserted-flag) (< tmp nr))
                       (set-cdr! (my-last-pair lst) (cons cl '())) )
                    ((>= tmp nr)
                       (set! lst (remove-last! lst)) ))
            lst)

      (set! oldcl (car part))
      (set! oldclvalue (logand *term-depth-strength-mask* (caar oldcl)))

      ;(newline)
      ;(display clvalue)(display " ")(display oldclvalue)
      ;(newline)

      (if (<= clvalue oldclvalue)
          (begin
            (set! tmp oldcl)
            (set-car! part cl)
            (set-cdr! part (cons oldcl (cdr part)))
            (set! inserted-flag #t) )))))



(define (remove-last! lst)
  (if (null? lst)
      '()
      (if (null? (cdr lst))
          '()
          (do ((lstpart lst (cdr lstpart)))
              ((and (pair? (cdr lstpart))
                    (null? (cddr lstpart)) )
               (set-cdr! lstpart '())
               lst)))))



(define (show-one-pass-strategy)
  (if (> *display-info-level* 38)
      (begin
        (cond
         (*prop-unsat-strat*
           (newline)(display "using propositional strategy"))
         (*simple-model-strat*
           (newline)(display "using simple model building"))
         (*sem-model-strat*
           (newline)(display "using sem-type model building"))
         (*mace-nofuns-model-strat*
           (newline)(display "solving nofuns-problem with mace-type model building"))
         (*mace-model-strat*
           (newline)(display "using mace-type model building"))
         (*random-model-strat*
           (newline)(display "using random model building"))
         (*hyper-strategy-flag*
           (newline)(display "using hyperresolution"))
         (else
           (newline)(display "using binary resolution")) )
        (newline)
        (if *sat-strategy-flag*
            (begin
              (display "using satisfiability checking strategy")
              (newline) ))
        (if *bliksem-strategy-flag*
            (begin
              (display "using delayed iteration on clause size")
              (newline) ))
        (if *order-strategy-flag*
            (begin
              (display "using term-depth-order strategy")
              (newline) ))
        (if *posweight-order-strategy-flag*
            (begin
              (display "using first neg lit preferred strategy")
              (newline) ))
        (if *firstlitpref-order-strategy-flag*
            (begin
              (display "using weight-order strategy")
              (newline) ))
        (if *order-strategy-sos-flag*
            (begin
              (display "using sos version of order strategy")
              (newline) ))
        (if *name-order-strategy-flag*
            (begin
              (display "using pred-name-order strategy")
              (newline) ))
        (if *exp-order-strategy-flag*
            (begin
              (display "using experimental ordering strategy")
              (newline) ))
        (if *sos-strategy-flag*
            (begin
              (display "using sos strategy")
              (newline) )
            (begin
              (display "not using sos strategy")
              (newline) ))
        ; (newline)
        (if *para-pos-unit-strategy*
            (begin
              (display "using positive unit paramodulation strategy")
              (newline) ))
        (if *para-unit-strategy*
            (begin
              (display "using unit paramodulation strategy")
              (newline) ))
        (if *unit-strategy-flag*
            (begin
              (display "using unit strategy")
              (newline) ))
        (if *double-strategy-flag*
            (begin
              (display "using double strategy")
              (newline) ))
        (if *dynamic-demodulation-flag*
            (begin
              (display "using dynamic demodulation")
              (newline) ))
        (if *unit-equality-strat*
            (begin
              (display "using unit equality specials")
              (newline) ))
        (if *root-para-strat*
            (begin
              (display "limiting paramodulation to term root")
              (newline) ))
        (if *back-demodulate-flag*
            (begin
              (display "using back demodulation")
              (newline) ))

        (if *ordered-paramodulation-flag*
            (begin
              (display "using ordered paramodulation")
              (newline) ))
        (if *order-equality-kb-strat*
            (begin
              (display "using kb ordering for equality")
              (newline) ))
        (if *order-equality-lex-strat*
            (begin
              (display "using lex ordering for equality")
              (newline) ))
        (if *order-equality-firstpref-strat*
            (begin
              (display "using first arg depth ordering for equality")
              (newline) ))
        (if *lex-order-prefer-bigger-arity*
            (begin
              (display "preferring bigger arities for lex ordering")
              (newline) )
            (begin
              (display "preferring smaller arities for lex ordering")
              (newline) ))
        (if *demodulate-with-clauses-stratflag*
            (begin
              (display "using clause demodulation")
              (newline) ))
        (if (not (zero? *clause-keeplength-limit*))
            (begin
              (display "clause length limited to ")
              (display *clause-keeplength-limit*)
              (newline) ))
        (if (not (zero? *clause-keepdepth-limit*))
            (begin
              (if *bliksem-strategy-flag*
                  (begin
                    (display "clause size limited to ") )
                  (display "clause depth limited to ") )
              (display *clause-keepdepth-limit*)
              (newline) ))
        (if (not (zero? *clause-finalkeepsize-limit*))
            (begin
              (display "clause size limited (after delay) to ")
              (display *clause-finalkeepsize-limit*)
              (newline) ))
        (if (not (equal? *global-sos-preference-ratio* 5))
            (begin
              (display "initial sos preference ratio is ")
              (display *global-sos-preference-ratio*)
              (newline) ))
        (display "seconds given: ")
        (display (quotient *strat-given-time* *internal-time-units-per-second*))
        (newline)
        (newline) )))


(define (eq-lit-in-sos? sos)
  (some? (lambda(cl)
           (some? (lambda(lit)
                     (or (pos-eq-pred? (pure-unsigned-fun (car lit)))
                         (and *type-theory-stratflag*
                              (and (composite? (cadr lit))
                                   (pos-eq-fun? (pure-unsigned-fun (caadr lit))) ))))
                  (cdr cl) ))
         sos))


; add-typet-demod-ans adds a special ans-lit containing
; an application of idrefl constructor to the left eq arg;
;
; In addition, it wraps the literal into the special inhabited-predicate.

(define (add-typet-demod-ans cl)
  (let* ((eqlit (car cl))
         (eqtype (eq-lit-type eqlit))
         (eqarg1 (eq-lit-arg1 eqlit)) )
    (cons (list *ans-symbol* (list *id-refl-fun-symbol* eqtype eqarg1))
          (list (list *inhabited-symbol* (car cl))) )))

; NB! remove-typet-demod-ans does not correct the term size and term
; depth info of the resulting clause, although both decrease!!!

(define (remove-typet-demod-ans cl)
 (let* ((litlst
         (remove-if (lambda (y)
                      (not (zero? (logand *litinf-isanswer-mask*
                                          (car y) ))))
                    (caddr (car cl)) ))
        (newcl (remove-ans-literals (cdr cl)))
        (newlit (cadar newcl)) )
   (cons (cons (car (car cl))
               (cons (cadr (car cl))
                     (cons litlst (cdddr (car cl))) ))
         (list newlit) )))


; remove-ans-from-clauses removes ans-literals and corresponding
; litinfolst elements from all clauses in the given clause list

(define (remove-ans-from-clauses cllist)
  (map (lambda(x)
         (let ((litlst
                 (remove-if (lambda (y)
                              (not (zero? (logand *litinf-isanswer-mask*
                                                 (car y) ))))
                            (caddr (car x)) ))
               (cl (remove-ans-literals (cdr x))) )
           (cons (cons (car (car x))
                       (cons (cadr (car x))
                             (cons litlst (cdddr (car x))) ))
                 cl)))
       cllist))


; enumerate-sos-literals takes sos and enumerates all
; its literals, ie adds an unique number to the
; info-part of a literal.

(define *enumerate-sos-literals-nr* 0)

(define (enumerate-sos-literals sos)
  (set! *enumerate-sos-literals-nr* 0)
  (for-each
   (lambda(cl)
     (let ((litinfolst (caddar cl)))
       (set-car!
          (cddar cl)
          (map (lambda(litinf)
                 (set! *enumerate-sos-literals-nr*
                       (+ 1 *enumerate-sos-literals-nr*) )
                 (append litinf (list *enumerate-sos-literals-nr*)) )
               litinfolst))))
   sos))


;=================================================================
;
;                  strategy selection
;
;=================================================================

(define *input-lit-nr* 0)
(define *input-eqlit-nr* 0)
(define *input-cl-nr* 0)
(define *input-unitcl-nr* 0)
(define *input-negcl-nr* 0)
(define *input-poscl-nr* 0)
(define *input-mixcl-nr* 0)
(define *input-horncl-nr* 0)
(define *input-invhorncl-nr* 0)
(define *input-negdhorncl-nr* 0)
(define *input-wellorderedcl-nr* 0)
(define *input-somewhat-ordered-nr* 0)
(define *input-wellhyper-nr* 0)
(define *input-maxlen-nr* 0)
(define *input-maxlen-nonground-nr* 0)
(define *input-maxdepth-nr* 0)
(define *input-contains-eq* #f)
(define *input-negun-d* 0)
(define *input-posun-d* 0)
(define *sos-maxdepth-nr* 0)
(define *input-maxsize-nr* 0)
(define *input-maxdepth-nr* 0)
(define *input-avgdepth-nr* 0)
(define *input-avgsize-nr* 0)

(define (analyse-input usable sos)
  (set! *input-lit-nr* 0)
  (set! *input-eqlit-nr* 0)
  (set! *input-cl-nr* 0)
  (set! *input-unitcl-nr* 0)
  (set! *input-negcl-nr* 0)
  (set! *input-poscl-nr* 0)
  (set! *input-mixcl-nr* 0)
  (set! *input-horncl-nr* 0)
  (set! *input-invhorncl-nr* 0)
  (set! *input-negdhorncl-nr* 0)
  (set! *input-wellorderedcl-nr* 0)
  (set! *input-somewhat-ordered-nr* 0)
  (set! *input-wellhyper-nr* 0)
  (set! *input-maxlen-nr* 0)
  (set! *input-maxlen-nonground-nr* 0)
  (set! *input-maxdepth-nr* 0)
  (set! *sos-maxdepth-nr* 0)
  (set! *input-contains-eq* #f)
  (set! *input-negun-d* 0)
  (set! *input-posun-d* 0)
  (set! *input-maxsize-nr* 0)
  (set! *input-maxdepth-nr* 0)
  (set! *input-avgdepth-nr* 0)
  (set! *input-avgsize-nr* 0)
  (if (or (eq-lit-in-sos? sos)
          (eq-lit-in-sos? usable) )
      (set! *input-contains-eq* #t)
      (set! *input-contains-eq* #f) )
  (let ((tmp '())
        (cl '())
        (clinf '())
        (purecl '())
        (litinfolst '())
        (tmplitinfolst '())
        (prohibitlitnr 0)
        (newlitinf '())
        (litnr 0)
        (varlitnr 0)
        (ndepth 0)
        (pdepth 0)
        (depth 0)
        (poslitnr 0)
        (neglitnr 0)
        (cldepthsum 0)
        (clsizesum 0)
        (tmpd 0)
        (tmps 0) )

    (do ((part sos (cdr part)))
        ((null? part))
      (set! cl (car part))
      (set! clinf (car cl))
      (for-each (lambda(lit)
                  (if (> (fun-real-depth (car lit)) *sos-maxdepth-nr*)
                      (set! *sos-maxdepth-nr* (fun-real-depth (car lit))) ))
                (cdr cl) ))
    (do ((clauselstpart (append usable sos) (cdr clauselstpart)))
        ((null? clauselstpart)
           (set! *input-avgdepth-nr*
                (quotient *input-maxdepth-nr* *input-cl-nr*))
           (set! *input-avgsize-nr*
                (quotient *input-maxsize-nr* *input-cl-nr*)) )
      (set! cl (car clauselstpart))
      (set! clinf (car cl))
      (set! purecl (cdr cl))
      (set! litinfolst (caddr clinf))

     ; (newline)(display cl)(newline)

      (set! tmplitinfolst (map (lambda(x)(cons 0 (cons 0 x))) litinfolst))
      ; propositional stuff
      (set! *input-cl-nr* (+ 1 *input-cl-nr*))
      (set! litnr 0)
      (set! varlitnr 0)
      (set! neglitnr 0)
      (set! poslitnr 0)
      (set! ndepth 0)
      (set! pdepth 0)
      (for-each (lambda(lit)
                  (set! litnr (+ 1 litnr))
                  (set! *input-lit-nr* (+ 1 *input-lit-nr*))
                  (if (or (pos-eq-pred? (pure-fun (car lit)))
                          (pos-eq-pred? (logxor *sgn-mask*
                                                (pure-fun (car lit)) )))
                      (set! *input-eqlit-nr* (+ 1 *input-eqlit-nr*)) )
                  (if (fun-may-cont-vars? (car lit))
                      (set! varlitnr (+ 1 varlitnr)) )
                  (if (neg-fun? (car lit))
                      (begin
                         (set! neglitnr (+ 1 neglitnr))
                         (if (> (fun-real-depth (car lit)) ndepth)
                             (set! ndepth (fun-real-depth (car lit))) ))
                      (begin
                         (set! poslitnr (+ 1 poslitnr))
                         (if (> (fun-real-depth (car lit)) pdepth)
                             (set! pdepth (fun-real-depth (car lit))) ))))
                purecl)
      (if (> varlitnr  *input-maxlen-nonground-nr*)
          (set! *input-maxlen-nonground-nr* varlitnr) )
      (if (> ndepth pdepth)
          (set! depth ndepth)
          (set! depth pdepth) )
      (if (> depth *input-maxdepth-nr*)
          (set! *input-maxdepth-nr* depth) )
      (if (> litnr *input-maxlen-nr*)
          (set! *input-maxlen-nr* litnr) )
      (if (eq? litnr 1)
          (begin
            (if (and (neg-fun? (caar purecl))
                     (> (fun-real-depth (caar purecl)) *input-negun-d*) )
                (set! *input-negun-d* (fun-real-depth (caar purecl))) )
            (if (and (pos-fun? (caar purecl))
                     (> (fun-real-depth (caar purecl)) *input-posun-d*) )
                (set! *input-posun-d* (fun-real-depth (caar purecl))) )
            (set! *input-unitcl-nr* (+ 1 *input-unitcl-nr*)) ))
      (cond
        ((zero? poslitnr)
          (set! *input-negcl-nr* (+ 1 *input-negcl-nr*)) )
        ((zero? neglitnr)
          (set! *input-poscl-nr* (+ 1 *input-poscl-nr*)) )
        (else
          (set! *input-mixcl-nr* (+ 1 *input-mixcl-nr*)) ))
      (if (and (> litnr 1) (<= poslitnr 1))
          (begin
             (set! *input-horncl-nr* (+ 1 *input-horncl-nr*))
             (if (> pdepth ndepth)
                 (set! *input-negdhorncl-nr* (+ 1 *input-negdhorncl-nr*)) )))
      (if (and (> litnr 1)(<= neglitnr 1))
          (set! *input-invhorncl-nr* (+ 1 *input-invhorncl-nr*)) )
      (if (and (> litnr 1)
               (>= (* 2 neglitnr) poslitnr) )
          (set! *input-wellhyper-nr* (+ 1 *input-wellhyper-nr*)) )
      ; depth and weight calcs
      (set! tmpd (- (compute-term-depth-raw (cons 0 purecl)) 1))
      (set! tmps (- (compute-term-size-raw (cons 0 purecl)) 1))
      (set! cldepthsum (+ cldepthsum tmpd))
      (set! clsizesum (+ clsizesum tmps))
      (if (> tmpd *input-maxdepth-nr*)(set! *input-maxdepth-nr* tmpd))
      (if (> tmps *input-maxsize-nr*)(set! *input-maxsize-nr* tmps))

      ; ordering stuff
      (set! prohibitlitnr 0)
      (if (not (null? (cdr purecl)))
          (begin
            (set! tmp *order-strategy-flag*)
            (set! *order-strategy-flag* #t)
            (set! newlitinf (vdepth-order-litinfo-aux
                              tmplitinfolst purecl))
            (set! *order-strategy-flag* tmp)
            (for-each
             (lambda(inf)
               (if (not (litinf-binrespara? (caddr inf)))
                   (begin
                     (set! prohibitlitnr (+ 1 prohibitlitnr)) )))
             newlitinf)
            (if (>= (* 2 prohibitlitnr) litnr)
                (set! *input-wellorderedcl-nr*
                      (+ 1 *input-wellorderedcl-nr*) ))
            (if (> prohibitlitnr 0)
                (set! *input-somewhat-ordered-nr*
                      (+ 1 *input-somewhat-ordered-nr*) ))
            )))))


(define (compute-term-size-raw term)
  (let ((res 1))
    (do ((part (cdr term) (cdr part)))
        ((null? part) res)
      (if (pair? (car part))
          (set! res (+ res (compute-term-size-raw (car part))))
          (set! res (+ res 1)) ))))

(define (compute-term-depth-raw term)
  (let ((res 0)
        (tmp 0))
    (do ((part (cdr term) (cdr part)))
        ((null? part) res)
      (if (pair? (car part))
          (begin
            (set! tmp (+ 1 (compute-term-depth-raw (car part))))
            (if (> tmp res) (set! res tmp)) )))))


; select-initial-strategy selects the initial strategy, based on
; the data about input clause set computed by the analyse-input.

(define *hne-class-flag* #f)
(define *heq-class-flag* #f)

(define (select-initial-strategy)
  (let ((res '())
        (tmp '()) )
    (set! *hne-class-flag* #f)
    (set! *heq-class-flag* #f)
    (set! res (copy-tree (select-initial-strategy-aux)))
    ;(set! res
    ;      (map (lambda(x)
    ;             (cons (car x)
    ;                   (cons (- (quotient (* *all-passes-time* (cadr x)) 100) 1)
    ;                         (cddr x)) ))
    ;           res))
    (set! res (select-strategy-seconds res *all-passes-time*))
    (if #f ; *run-forever-flag*
        (begin
          (set! tmp (my-last-pair res))
          (set-car! (cdr (car tmp)) '100000) ))
    (if (>= *display-info-level* 30)
        (begin
            (newline)(display "strategies selected: ")(newline)
            (for-each (lambda(x)
                        (if (and (pair? x)
                                 (eq? 'split (car x)) )
                            (begin
                             (display 'split)
                             (newline) )
                            (begin
                             (display x)
                             (newline) )))
                       res)
            (newline) ))
    (set! *strat-list* res)
    (cond
      ((and *input-contains-eq* *paramodulation-allowed-flag*)
         (if (>= *display-info-level* 35)
           (begin (newline)(display "using equality")(newline)) )
         (set! *paramodulation-strategy-flag* #t)
         (turn-on-demodulation) )
      (else
         (if (>= *display-info-level* 35)
             (begin (newline)(display "not using equality")(newline)) )
         (set! *paramodulation-strategy-flag* #f)
         (turn-on-demodulation) ))
    res))


(define (select-strategy-seconds stratlst secs)
  (let ((givensecs 0)
        (oldsecs 0)
        (newsecs 0)
        (tmpsecs 0)
        (tmpres '())
        (timesum 0)
        (el '())
        (quotient1000 0))
    (for-each
      (lambda(x)(set! givensecs (+ (cadr x) givensecs)))
      stratlst)
    (do ((lstpart stratlst (cdr lstpart)))
        ((null? lstpart)
           (reverse tmpres) )
      (set! el (car lstpart))
      (set! tmpsecs (cadr el))
      (set! quotient1000 (quotient (* 1000 tmpsecs) givensecs))
      (set! newsecs (quotient (* quotient1000 secs) 1000))
      (if (< newsecs 1)
          (set! newsecs 1) )
      ;(display "newsecs: ")(display newsecs)(newline)
      ;(display "timesum: ")(display timesum)(newline)
      (if (null? (cdr lstpart))
          (begin
            (set! newsecs (- secs timesum))
            (if (< newsecs 1) (set! newsecs 1)) ))
      (set! timesum (+ timesum newsecs))
      (set! tmpres
            (cons (cons (car el)
                        (cons newsecs
                              (cddr el) ))
                  tmpres)) )))

(define (select-initial-strategy-aux)
  (let ((tmp '())
        (res '())
        (maxtime 150)
        (incrlen 0) )
    (if (>= *display-info-level* 50)
        (begin
          (newline)
          (display "========= stats concerning the input clauses: =========")
          (newline)
          (display "*input-lit-nr* ")(display *input-lit-nr*)(newline)
          (display "*input-eqlit-nr* ")(display *input-eqlit-nr*)(newline)
          (display "*input-cl-nr* ")(display *input-cl-nr*)(newline)
          (display "*input-unitcl-nr* ")(display *input-unitcl-nr*)(newline)
          (display "*input-negcl-nr* ")(display *input-negcl-nr*)(newline)
          (display "*input-poscl-nr* ")(display *input-poscl-nr*)(newline)
          (display "*input-mixcl-nr* ")(display *input-mixcl-nr*)(newline)
          (display "*input-horncl-nr* ")(display *input-horncl-nr*)(newline)
          (display "*input-invhorncl-nr* ")
          (display *input-invhorncl-nr*)(newline)
          (display "*input-negdhorncl-nr* ")
          (display *input-negdhorncl-nr*)(newline)
          (display "*input-wellorderedcl-nr* ")
          (display  *input-wellorderedcl-nr*)(newline)
          (display "*input-wellhyper-nr* ")
          (display  *input-wellhyper-nr*)(newline)
          (display "*input-somewhat-ordered-nr* ")
          (display  *input-somewhat-ordered-nr*)(newline)
          (display "*input-maxlen-nr* ")(display *input-maxlen-nr*)(newline)
          (display "*input-maxlen-nonground-nr* ") (display *input-maxlen-nonground-nr*)
          (newline)
          (display "*input-maxdepth-nr* ")(display *input-maxdepth-nr*)
                    (newline)
          (display "*input-contains-eq* ")
          (display *input-contains-eq*)(newline)
          (display "=======================")
          (newline) ))

    (set! incrlen (- (* 2 *input-maxlen-nr*) 1))
    ; selection starts
    (cond
     ; satisfiability is looked for
     (*sat-strategy-flag*
        (set! res (select-initial-strategy-sat incrlen)) )
     ; pure unit and eq present => unit equality strat
     ((and *input-contains-eq*
           (= *input-lit-nr* *input-eqlit-nr*)
           (= *input-cl-nr* *input-unitcl-nr*) )
        (set! res (select-initial-strategy-ueq incrlen)) )
     ; pure equality, nonunit
     ((and *input-contains-eq*
           (= *input-lit-nr* *input-eqlit-nr*) )
        (set! res (select-initial-strategy-peq incrlen)) )
     ; pure horn with eq
     ((and *input-contains-eq*
           (= *input-horncl-nr* (- *input-cl-nr* *input-unitcl-nr*)) )
        (set! res (select-initial-strategy-heq incrlen)) )
     ; pure horn without eq
     ((and (not *input-contains-eq*)
           (= *input-horncl-nr* (- *input-cl-nr* *input-unitcl-nr*)) )
        (set! res (select-initial-strategy-hne incrlen)) )
     ; nonhorn without eq
     ((not *input-contains-eq*)
        (set! res (select-initial-strategy-nne incrlen)) )
     ; all the rest => nonhorn with equality
     (else
        (set! res (select-initial-strategy-neq incrlen)) ))

    ;(if #t; *sat-strategy-flag*
    ;    ; sat case
    ;    (begin '())
    ;    ; normal case
    ;    (begin
    ;     (set! res
    ;              (map (lambda(x)
    ;                           (cons (car x)
    ;                               (cons (quotient (* 9 (cadr x)) 10)
    ;                                         (cddr x)) ))
    ;                    res))
    ;     (set! res (cons (car res) (cons '(binary-bliksem 10 #t 9 9) (cdr res))) )))

    ; split element in the strategy list contains:
    ;  - split marker symbol
    ;  - time given for the split attempt
    ;  - clause in sos which is split
    ;  - splittings of the clause which is split
    ;  - strategy list to be used for all splits

    ;(newline)
    ;(display "Splitting test in initial strategy selection:")
    ;(newline)
    ;(for-each
    ;  (lambda(x)
    ;    (newline)
    ;    (display "orig clause: ")(newline)
    ;    (print-pure-clause x)
    ;    (newline)
    ;    (display x)
    ;    (newline)
    ;    (display (split-full-clause (cons '() x))) )
    ;   sos)


    ;(set! res
    ;        '((split 5
    ;                 ()
    ;                 ()
    ;                 ((hyper 1 #f)
    ;                  (binary 2 #t) ))))

    ;(set! res
    ;    '((hyper 1 #f)
    ;      (binary 2 #t)) )

    ;(set! res
    ;                 `((hyper 2 #t ,*input-maxdepth-nr* ,incrlen)
    ;                   (hyper 2 #f ,*input-maxdepth-nr* ,incrlen)
    ;                   (hyper 2 #f)
    ;                   (binary-order 2 #f ,*input-maxdepth-nr* ,incrlen)
    ;                   (binary-posweight-order 2 #f ,*input-maxdepth-nr* ,incrlen)
    ;                   (binary-weightorder 2 #f ,(+ 2 *input-maxdepth-nr*) ,(+ 2 incrlen))
    ;                   (binary-negweightorder 2 #f)
    ;                   (binary-posweight-lex-big-order 2 #f)
    ;                   (binary-posweight-lex-small-order 2 #f)
    ;                   (binary-order-sos 2 #t ,*input-maxdepth-nr* ,incrlen)
    ;                   (binary-unit-uniteq 2 #f ,(+ 2 *input-maxdepth-nr*) ,(+ 2 incrlen))
    ;                   (binary-unit 2 #f)
    ;                   (binary 2 #t) ) )

    ;(set! res
    ;                 `((hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
    ;                   (binary-unit 2 #f ,*input-maxdepth-nr* ,incrlen)
    ;                   (binary-double 3 #f ,*input-maxdepth-nr* ,incrlen)
    ;                   (binary-double 3 #f)
    ;                   (binary-double 4 #t)
    ;                   (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
    ;                   (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
    ;                   (binary-posweight-order 20 #f)
    ;                   (binary-posweight-lex-big-order 5 #f)
    ;                   (binary-posweight-lex-small-order 2 #f)
    ;                   (binary-order-sos 10 #t)
    ;                   (binary-unit-uniteq 5 #f)
    ;                    (binary-weightorder 10 #f)
    ;                   (binary-order 10 #f)
    ;                   (hyper-order 6 #f)
    ;                   (binary 20 #t) ))

    ;(set! res
    ;          `( ;(random-model-strat 5 #f)
    ;             ;(simple-model-strat 20 #f)
    ;             ;(sat-hyper-strat 5 #f)
    ;              (binary-unit 5 #f)
    ;              (sat-ordered-binres-strat 5 #f)
    ;              ;(sat-ordered-binres-strat
    ;              (hyper 5 #f)
    ;              (binary-posweight-order 5 #f)
    ;              ;(sat-ordered-binres-strat 50 #f)
    ;              (sem-model-strat-1 2 #f)
    ;              (sem-model-strat-0 5 #f)
    ;              (sem-model-strat-2 5 #f)
    ;              (sem-model-strat-3 4 #f)
    ;              (sem-model-strat-4 4 #f)
    ;              (sem-model-strat-5 5 #f)
    ;              (sem-model-strat-1 20 #f)
    ;              (mace-nofuns-model-strat 20 #f) ))
    ;              (sat-posweight-order 2 #f)
    ;              (binary-epr-order 50 #f)
    ;              (mace-model-strat 50 #f) ))

    res))





(define (select-initial-strategy-sat incrlen)
  (let ((res '()))
    (if (>= *display-info-level* 30)
         (begin (newline)(display "detected problem class: sat")(newline) ))
    (cond
         ((<= *input-maxdepth-nr* 1)
            (if (>= *display-info-level* 30)
                (begin
                 (display "detected subclass: epr")(newline) ))
            (if *input-contains-eq*
                (set! res
                  `( (hyper-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                     (binary-unit 5 #f)
                     (binary-double 4 #f ,*input-maxdepth-nr* ,incrlen)
                     (binary-weightorder 8 #f)
                     (binary-posweight-order-kb-small-fulleq 2 #f)
                     (binary-posweight-order-kb-big 15 #f)
                     (binary 5 #t)
                     (binary-epr-order 5 #f)
                     (mace-nofuns-model-strat 20 #f)
                     (mace-model-strat 38 #f) ))
                (set! res
                  `( (hyper-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                     (binary-unit 5 #f)
                     (binary-double 4 #f ,*input-maxdepth-nr* ,incrlen)
                     (binary-weightorder 5 #f)
                     (binary-posweight-order 20 #f)
                     (binary 5 #t)
                     (binary-epr-order 5 #f)
                     (mace-nofuns-model-strat 18 #f)
                     (mace-model-strat 33 #f) ))))
        (*input-contains-eq*
             (if (>= *display-info-level* 30)
                (begin
                  (display "detected subclass: sat with eq")(newline) ))
             (set! res
              `( ;(hyper-order 10 #f ,*input-maxdepth-nr* ,incrlen)
                 ;(binary-weightorder-sos 10 #t)
                 (sem-model-strat-1 25 #f)
                 (sem-model-strat-3 12 #f)
                 (hyper 2 #f)
                 (binary-posweight-order 2 #f)
                 (binary-unit 2 #f)
                 (sat-posweight-order 3 #f)
                 ;(sem-model-strat-4 2 #f)
                 (sem-model-strat-1 3 #f)
                 (sem-model-strat-3 3 #f)
                 (sem-model-strat-5 9 #f)
                 (mace-model-strat 39 #f) )))
                 ; (sat-ordered-binres-strat 20 #f)
        (else
            (if (>= *display-info-level* 30)
                (begin
                 (display "detected subclass: sat without eq")(newline) ))
            (set! res
              `( ;(random-model-strat 5 #f)
                 ;(simple-model-strat 20 #f)
                 ;(sat-hyper-strat 5 #f)
                  (sem-model-strat-1 16 #f)
                  (sem-model-strat-3 17 #f)
                  (hyper 2 #f)
                  (binary-posweight-order 2 #f)
                  (binary-unit 2 #f)
                  (sat-ordered-binres-strat 4 #f)
                  (sat-posweight-order 3 #f)
                  (sem-model-strat-1 3 #f)
                  (sem-model-strat-3 3 #f)
                  (sem-model-strat-5 9 #f)
                  (mace-model-strat 40 #f) ))))

    res))

(define (select-initial-strategy-neq incrlen)
  (let ((res '()))
      (newline)(display "detected problem class: neq")(newline)
      (cond
         ((> *input-cl-nr* 100)
           (display "detected subclass: big")(newline)
                   ; heavy stress sos
                   (set! res
                     `((hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 2 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 3 #f)
                       (binary-double 3 #t)
                       (binary 15 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 10 #f)
                       (binary-posweight-lex-big-order 5 #f)
                       (binary-posweight-lex-small-order 2 #f)
                       (binary-order-sos 5 #t)
                       (binary-unit-uniteq 5 #f)
                       (binary-weightorder 5 #f)
                       (binary-weightorder-sos 3 #f)
                       (binary-order 5 #f)
                       (hyper-order 3 #f)
                       (binary 22 #t) )))
         ((> *input-cl-nr* 10)
           (display "detected subclass: medium")(newline)
                   ; stress sos
                   (set! res
                     `((hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 2 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 2 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 3 #f)
                       (binary-double 3 #t)
                       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 20 #f)
                       (binary-posweight-lex-big-order 5 #f)
                       (binary-posweight-lex-small-order 2 #f)
                       (binary-order-sos 10 #t)
                       (binary-unit-uniteq 5 #f)
                       (binary-weightorder 10 #f)
                       (binary-order 10 #f)
                       (hyper-order 6 #f)
                       (binary 20 #t) )))
         (else
          (display "detected subclass: small")(newline)
                   ; do not stress sos
                   (set! res
                     `((hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 2 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 2 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 3 #f)
                       (binary-double 2 #t)
                       (binary 3 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 23 #f)
                       (binary-posweight-lex-big-order 8 #f)
                       (binary-posweight-lex-small-order 3 #f)
                       (binary-order-sos 10 #t)
                       (binary-unit-uniteq 10 #f)
                       (binary-weightorder 12 #f)
                       (binary-order 5 #f)
                       (hyper-order 7 #f)
                       (binary 10 #t) ))))
   res))

(define (select-initial-strategy-nne incrlen)
  (let ((res '()))
      (newline)(display "detected problem class: nne")(newline)
      (cond
         ((> *input-cl-nr* 100)
           (display "detected subclass: big")(newline)
                   ; heavy stress sos
                   (set! res
                     `((hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 3 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary 15 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 9 #f)
                       (binary-order-sos 5 #t)
                       (binary-unit-uniteq 5 #f)
                       (binary-weightorder 5 #f)
                       (binary-weightorder-sos 5 #f)
                       (binary-order 5 #f)
                       (hyper-order 5 #f)
                       (binary 28 #t) )))
         ((> *input-cl-nr* 10)
           (display "detected subclass: medium")(newline)
                   ; stress sos
                   (set! res
                     `((hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 2 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 3 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 23 #f)
                       (binary-order-sos 10 #t)
                       (binary-unit-uniteq 5 #f)
                       (binary-weightorder 10 #f)
                       (binary-order 10 #f)
                       (hyper-order 8 #f)
                       (binary 19 #t) )))
         (else
          (display "detected subclass: small")(newline)
                   ; do not stress sos
                   (set! res
                     `((hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 2 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 3 #f)
                       (binary 3 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 33 #f)
                       (binary-order-sos 10 #t)
                       (binary-unit-uniteq 10 #f)
                       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-weightorder 13 #f)
                       (binary-order 5 #f)
                       (hyper-order 10 #f)
                       (binary 10 #t) ))))
   res))


(define (select-initial-strategy-heq incrlen)
   (let ((res '()))
     (newline)(display "detected problem class: heq")(newline)
     (set! *heq-class-flag* #t)
     (cond
         ((> *input-cl-nr* 100)
             (display "detected subclass: big")(newline)
             (cond
               ((> *input-maxlen-nonground-nr* 3)
                   (display "detected subclass: long")(newline)
                   (set! res
                       `( (hyper 5 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 3 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 5 #t)
                       (hyper 18 #f)
                       (binary-unit-uniteq 3 #f)
                       (binary-posweight-order 15 #f)
                       (binary-posweight-lex-big-order 5 #f)
                       (binary-posweight-lex-small-order 2 #f)
                       (binary-order 5 #f)
                       (binary-weightorder 4 #f)
                       (binary-posweight-order-sos 15 #t)
                       (binary-unit-sos 6 #t)
                       (binary 14 #t) )))
               (else
                   (display "detected subclass: short")(newline)
                   (set! res
                    `( (binary-posweight-order 9 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary 8 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 5 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 5 #f)
                       (binary-unit-uniteq 3 #f)
                       (binary-posweight-order 22 #f)
                       (binary-posweight-lex-big-order 6 #f)
                       (binary-posweight-lex-small-order 2 #f)
                       (binary-order 5 #f)
                       (binary-weightorder 4 #f)
                       (binary-unit 5 #f)
                       (binary 7 #t) ))) ))
         ((> *input-cl-nr* 10)
             (display "detected subclass: medium")(newline)
             (cond
               ((> *input-maxlen-nonground-nr* 3)
                   (display "detected subclass: long")(newline)
                   (set! res
                       `( (hyper 10 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary 5 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 5 #t)
                       (hyper 18 #f)
                       (binary-unit-uniteq 3 #f)
                       (binary-weightorder 4 #f)
                       (binary-posweight-order 12 #f)
                       (binary-posweight-lex-big-order 5 #f)
                       (binary-posweight-lex-small-order 2 #f)
                       (binary-order 5 #f)
                       (binary-unit 8 #f)
                       (binary 10 #t) )))
               (else
                   (display "detected subclass: short")(newline)
                   (set! res
                    `( (binary-posweight-order 10 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary 8 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 2 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 5 #f)
                       (binary-unit-uniteq 3 #f)
                       (binary-weightorder 4 #f)
                       (binary-posweight-order 28 #f)
                       (binary-posweight-lex-big-order 10 #f)
                       (binary-posweight-lex-small-order 2 #f)
                       (binary-order 5 #f)
                       (binary-unit 8 #f)
                       (binary 10 #t) ))) ))
          (else
             (display "detected subclass: small")(newline)
             (cond
               ((> *input-maxlen-nonground-nr* 3)
                   (display "detected subclass: long")(newline)
                   (set! res
                       `( (hyper 10 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-posweight-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (hyper 33 #f)
                          (binary-unit-uniteq 3 #f)
                          (binary-weightorder 4 #f)
                          (binary-posweight-order 20 #f)
                          (binary-posweight-lex-big-order 5 #f)
                          (binary-posweight-lex-small-order 2 #f)
                          (binary-order 5 #f)
                          (binary 4 #t)
                          (binary-unit 9 #f) )))
               (else
                   (display "detected subclass: short")(newline)
                   (set! res
                       `( (binary-unit-uniteq 5 #f)
                          (binary-posweight-order 20 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-posweight-order 40 #f)
                          (binary-posweight-lex-big-order 10 #f)
                          (binary-posweight-lex-small-order 2 #f)
                          (binary-weightorder 4 #f)
                          (hyper 5 #f)
                          (binary 4 #t)
                          (binary-order 5 #f)
                          (binary-unit 5 #f)  ))) )))
   res))

(define (select-initial-strategy-hne incrlen)
  (let ((res '()))
     (newline)(display "detected problem class: hne")(newline)
     (set! *hne-class-flag* #t)
     (cond
         ((> *input-cl-nr* 100)
             (display "detected subclass: big")(newline)
             (cond
               ((> *input-maxlen-nonground-nr* 3)
                   (display "detected subclass: long")(newline)
                   (set! res
                       `( (hyper 5 #t ,*input-maxdepth-nr* ,incrlen)
                          (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-double 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-posweight-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
                          (hyper 5 #t)
                          (hyper 23 #f)
                          (binary-posweight-order 15 #f)
                          (binary-order 5 #f)
                          (binary-weightorder 5 #f)
                          (binary-posweight-order-sos 15 #t)
                          (binary-unit-sos 8 #t)
                          (binary 12 #t) )))
               (else
                   (display "detected subclass: short")(newline)
                   (set! res
                    `( (hyper 5 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 10 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 5 #f)
                       (binary-posweight-order 28 #f)
                       (binary-weightorder 4 #f)
                       (binary-order 5 #f)
                       (binary-unit 5 #f)
                       (binary 13 #t) ))) ))
         ((> *input-cl-nr* 10)
             (display "detected subclass: medium")(newline)
             (cond
               ((> *input-maxlen-nonground-nr* 3)
                   (display "detected subclass: long")(newline)
                   (set! res
                    `( (hyper 10 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 5 #f)
                       (binary-posweight-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary 5 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 5 #t)
                       (hyper 28 #f)
                       (binary-posweight-order 15 #f)
                       (binary-weightorder 7 #f)
                       (binary-order 5 #f)
                       (binary-unit 8 #f)
                       (binary 7 #t) )))
               (else
                   (display "detected subclass: short")(newline)
                   (set! res
                    `( (hyper 2 #t ,*input-maxdepth-nr* ,incrlen)
                       (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-posweight-order 15 #f ,*input-maxdepth-nr* ,incrlen)
                       (binary-double 5 #f)
                       (binary 8 #t ,*input-maxdepth-nr* ,incrlen)
                       (hyper 5 #f)
                       (binary-posweight-order 30 #f)
                       (binary-weightorder 7 #f)
                       (binary-order 5 #f)
                       (binary-unit 8 #f)
                       (binary 10 #t) ))) ))
          (else
             (display "detected subclass: small")(newline)
             (cond
               ((> *input-maxlen-nonground-nr* 3)
                   (display "detected subclass: long")(newline)
                   (set! res
                       `( (hyper 10 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-unit 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-double 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-posweight-order 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-unit 11 #f)
                          (hyper 28 #f)
                          (binary-posweight-order 27 #f)
                          (binary-weightorder 5 #f)
                          (binary-order 5 #f)
                          (binary 4 #t)  )))
               (else
                   (display "detected subclass: short")(newline)
                   (set! res
                       `( (hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-unit 2 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-double 3 #f ,*input-maxdepth-nr* ,incrlen)
                          (hyper 5 #f)
                          (binary-unit 6 #f)
                          (binary-weightorder 7 #f)
                          (binary 3 #t)
                          (binary-order 5 #f)
                          (binary-posweight-order 19 #f ,*input-maxdepth-nr* ,incrlen)
                          (binary-posweight-order 47 #f)  ))) )))
   res))

(define (select-initial-strategy-peq incrlen)
  (let ((res '()))
     (newline)(display "detected problem class: peq")(newline)
     (cond
      (#t
            (set! res
             `((hyper 5 #f ,*input-maxdepth-nr* ,incrlen)
               (binary-unit 2 #f)
               (binary-unit-uniteq 2 #f)
               (binary-posweight-kb-big-order 10 #f ,*input-maxdepth-nr* ,incrlen)
               (binary-posweight-lex-big-order        5 #f ,*input-maxdepth-nr* ,incrlen)
               (binary 5 #t)
               (binary-posweight-kb-big-order 26 #f)
               (binary-posweight-lex-big-order 17 #f)
               (binary-posweight-firstpref-order 10 #f)
               (binary-order 5 #f)
               (binary-posweight-kb-small-order 8 #f)
               (binary-posweight-lex-small-order 5 #f) ))))

   res))

(define (select-initial-strategy-ueq incrlen)
  (let ((res '()))
     (newline)(display "detected problem class: ueq")(newline)
     (cond
       (#t
            (set! res
             `((binary-posweight-kb-big-order 10 #f ,*input-maxdepth-nr* ,incrlen)
               (binary-posweight-lex-big-order        5 #f ,*input-maxdepth-nr* ,incrlen)
               (binary 5 #t)
               (binary-posweight-kb-big-order 30 #f)
               (binary-posweight-lex-big-order 20 #f)
               (binary-posweight-firstpref-order 10 #f)
               (binary-posweight-kb-small-order 10 #f)
               (binary-posweight-lex-small-order 10 #f) ))))

   res))


; equality ordering selection principles:
;
; - default is knuth-bendix, bigger arity preferred
; - for posweight-order:
;   - binary-posweight-order-kb-big (default) is knuth-bendix, bigger arity preferred
;   - binary-posweight-order-kb-small is knuth-bendix, smaller arity preferred
;   - binary-posweight-order-lex-big (default) is lex, bigger arity preferred
;   - binary-posweight-order-lex-small is lex, smaller arity preferred
; - special unit-equality limits:
;   - hyper-uniteq is knuth-bendix, bigger arity preferred, units only
;   - binary-unit-uniteq is knuth-bendix, bigger arity preferred, units only
;   - binary-uniteq is knuth-bendix, bigger arity preferred, units only

;==================================================================
;
;               temporary debugging defs
;=================================================================


;(trace subsumed-pure-clause-aux? ground-lit-equal?
; fun-can-subsume? fun-may-cont-vars? restore-gcl-vars
;  clear-gcl-vars later-occurrences-var? subsumed-term-list-ordinary?)

;(trace subsumed-pure-clause-aux? fun-can-subsume? havealready-clause?
; funlist-can-subsume-aux?)

;(trace bigger-strength-aux? funlist-can-subsume-aux?
; subsumed-pure-clause-aux? fun-can-subsume? >-fun-ordering-aux?)

;(trace make-kept-clause full-unit-cutoff! full-unit-cutoff-aux!
; has-some-local-only-vars? has-only-local-only-vars? resolve-unify-cutoff
;  resolve-unify-cutoff-aux)

;(trace subsumed-clause?)
;(trace subsumed-term-list-ordinary?)

;(trace back-subsume)
;(trace havealready-clause?)
;(trace resolve-unify-ordinary)
;(trace resolve-unify-ordinary-aux)
;(trace resolve-unify-cutoff)
;(trace factor-unify-ordinary)
;(trace factor-unify-ordinary-aux)
;(trace subst-deco-cl)

;(trace subst-deco-term)
;(trace subst-deco-weight-term)

;(trace unit-handle-cl!)
;(trace make-clause-info)

;(trace make-kept-clause)
;(trace make-varinfo-lst!)
;(trace make-fulldeco-clause!)
;(trace make-kept-litinflist)

;(trace full-unit-cutoff!)
;(trace handle-derived-inf-cl-pair)
;(trace unit-cl?)
;(trace sort-clause)
;(trace factor)
;(trace factor-aux)


(define (break)
  (let ((flag #f))
    (do ()
        (flag)
      (newline)
      (display "|")
      (let ((tmp (eval (read))))
        (if (eq? tmp 'quit)
            (set! flag #t)
            (display tmp) ))
      (newline) )))

;(define (up) (load "resma.scm"))

;(require 'pretty-print)


;(define (clear-x-vars! var)
;  (do ()
;      ((zero? var)
;        (newline)(debug-print-xvars-spec 6)(newline))
;    (set-var-ref! var *unassigned*)
;    (set! var (- var 1)) ))

;(define (clear-x-vars-and-touched! var)
;  (do ()
;      ((zero? var)
;       (newline)(debug-print-xvars-spec 6)(newline))
;    (set-var-ref! var *unassigned*)
;    (set-var-touched! var #f)
;    (set! var (- var 1)) ))

; (define *qp-width* 1500)

(define (hprove x y z)(init-hyper-strat)(prove x y z))

(define (hfprove x y)(init-hyper-strat)(fprove x))

(define (init)
  (init-hyper-strat) )
;  (set! *neg-hyper-strategy-flag* #t))



;=======================================================================
;
;                      converting clauses with very many preds
;
;======================================================================


; - convert the clause set with many predicates and functions to
;   clause set, where the "excess" predicates and functions are
;   curried: instead of p(f(a)) you may get ganappred1(1,ganapfun1(1,a)).

(define *show-uncurry-input-predfuns* #t) ; to print uncurried or not
(define *absolute-max-curry-pred-nr* 120) ; so many are allowed in all
(define *absolute-max-curry-fun-nr* 120)  ; so many are allowed in all
(define *max-curry-pred-nr* 100)           ; computed: max nr of uncurried predsymbs
(define *max-curry-fun-nr* 200)            ; computed: max nr of uncurried funsymbs

; special-orig-curry-predsymbs may NOT be curried:
; equal and computable predicates and functions


; check-computable predicates checks whether the literal has
; a computable predicate and if yes, then whether it can be
; computed. It returns:
; #f : could not be computed
; 1  : literal evaluated to true
; -1 : literal evaluated to false

(define *special-curry-predsymbs* '())
(define *special-orig-curry-predsymbs*
   (list 'ans 'equal '-equal 'splitpred0 'splitpred1) )

; special-orig-curry-funsymbs may NOT be curried:
; arithmetic and computable functions

(define *special-curry-funsymbs* '())
(define *special-orig-curry-funsymbs* '())



(define *curried-pred-const-table* '())
(define *not-curried-pred-nr* 0) ; counting different nonspecial preds
(define *curried-pred-nr* 0) ; counting curried preds
(define *curry-pred-arity-table* '()) ; arities: actual predicates found in clauses
(define *not-curried-preds* '()) ; not curried: actual predicates found in clauses

(define *curried-fun-const-table* '())
(define *curried-fun-nr* 0)
(define *curry-fun-arity-table* '())
(define *not-curried-fun-nr* 0)
(define *not-curried-funs* '())

(define *curried-max-orig-pred-arity* 0)
(define *curried-max-orig-fun-arity* 0)

(define *curried-appred-arities* '())
(define *curried-apfun-arities*  '())

(define (curry-input-readylst  readylst)

    ;(newline)
    ;(pretty-print readylst)
    ;(newline)
    ;(display "readylst: ")(newline)
    ;(pretty-print readylst)

  (let ((sospart (assq 'sos readylst))
        (usablepart (assq 'usable readylst))
        (res '())
        (tmp '()) )
    (set! tmp
          (filter-map
             (lambda(x)
               (if (or (not (pair? x))
                       (not (memq (car x) '(sos usable))) )
                   x
                   #f))
             readylst))

;    (for-each (lambda(cl)
;                (for-each (lambda (lit)
;                            (if (and (pair? lit)
;                                     (> (length lit) *curried-max-orig-pred-arity*) )
;                                (set! *curried-max-orig-pred-arity* (length lit)) ))
;                          cl))
;              (append (cdr sospart) (cdr usablepart)) )


    (init-curry-clause-list)
    (set! *max-curry-pred-nr* (- *absolute-max-curry-pred-nr*
                                 (+ (length (make-set *special-curry-predsymbs*))
                                    (length *curried-appred-arities*) )))
    (set! *max-curry-fun-nr* (- *absolute-max-curry-fun-nr*
                                 (+ (length (make-set *special-curry-funsymbs*))
                                    (length *curried-apfun-arities*) )))
    (if sospart (set! sospart (curry-clause-list (cdr sospart))))
    (if usablepart (set! usablepart (curry-clause-list (cdr usablepart))))
    (set! res tmp)
    (if sospart (set! res (cons (cons 'sos sospart) res)))
    (if usablepart (set! res (cons (cons 'usable usablepart) res)))

    ;(newline)
    ;(display "readylst: ")
    ;(pretty-print readylst)
    ;(newline)
    ;(display "res: ")(newline)
    ;(pretty-print res)

    (if (and (< (+ *curried-pred-nr* *not-curried-pred-nr*) 115)
             (< (+ *curried-fun-nr* *not-curried-fun-nr*) 115) )
        (begin
          (set! res readylst) )
        (begin
          (newline)
          (display "Some predicates or functions were curried and renamed.")
          (newline) ))

    res))




(define (make-curried-special-predfuns-data)
 (set! *curried-appred-arities*
   '(($ganappred1 . 1)($ganappred2 . 2)($ganappred3 . 3)($ganappred4 . 4)($ganappred5 . 5)
     ($ganappred6 . 6)($ganappred7 . 7)($ganappred8 . 8)($ganappred9 . 9)($ganappred10 . 10)
     ($ganappred11 . 11)($ganappred12 . 12)($ganappred13 . 13)($ganappred14 . 14)($ganappred15 . 15)
     ($ganappred16 . 16)($ganappred17 . 17)($ganappred18 . 18)($ganappred19 . 19)($ganappred20 . 20)
     ($ganappred21 . 21)($ganappred22 . 22)($ganappred23 . 23)($ganappred24 . 24)($ganappred25 . 25)
     ($ganappred26 . 26)($ganappred27 . 27)($ganappred28 . 28)($ganappred29 . 29)($ganappred30 . 30)
     ($ganappred31 . 31)($ganappred32 . 32)($ganappred33 . 33)($ganappred34 . 34)($ganappred35 . 35)
     ($ganappred36 . 36)($ganappred37 . 37)($ganappred38 . 38)($ganappred39 . 39)($ganappred40 . 40)
     ($ganappred41 . 41)($ganappred42 . 42)($ganappred43 . 43)($ganappred44 . 44)($ganappred45 . 45)
     ($ganappred46 . 46)($ganappred47 . 47)($ganappred48 . 48)($ganappred49 . 49)($ganappred50 . 50)
     ($ganappred51 . 51)($ganappred52 . 52)($ganappred53 . 53)($ganappred54 . 54)($ganappred55 . 55)
     ($ganappred56 . 56)($ganappred57 . 57)($ganappred58 . 58)($ganappred59 . 59)($ganappred60 . 60)
     ($ganappred61 . 61)($ganappred62 . 62)($ganappred63 . 63)($ganappred64 . 64)($ganappred65 . 65)
     ($ganappred66 . 66)($ganappred67 . 67)($ganappred68 . 68)($ganappred69 . 69)($ganappred70 . 70)
      ))
 (set! *curried-apfun-arities*
   '(($ganapfun1 . 1)($ganapfun2 . 2)($ganapfun3 . 3)($ganapfun4 . 4)($ganapfun5 . 5)
     ($ganapfun6 . 6)($ganapfun7 . 7)($ganapfun8 . 8)($ganapfun9 . 9)($ganapfun10 . 10)
     ($ganapfun11 . 11)($ganapfun12 . 12)($ganapfun13 . 13)($ganapfun14 . 14)($ganapfun15 . 15)
     ($ganapfun16 . 16)($ganapfun17 . 17)($ganapfun18 . 18)($ganapfun19 . 19)($ganapfun20 . 20)
     ($ganapfun21 . 21)($ganapfun22 . 22)($ganapfun23 . 23)($ganapfun24 . 24)($ganapfun25 . 25)
     ($ganapfun26 . 26)($ganapfun27 . 27)($ganapfun28 . 28)($ganapfun29 . 29)($ganapfun30 . 30)
     ($ganapfun31 . 31)($ganapfun32 . 32)($ganapfun33 . 33)($ganapfun34 . 34)($ganapfun35 . 35)
     ($ganapfun36 . 36)($ganapfun37 . 37)($ganapfun38 . 38)($ganapfun39 . 39)($ganapfun40 . 40)
     ($ganapfun41 . 41)($ganapfun42 . 42)($ganapfun43 . 43)($ganapfun44 . 44)($ganapfun45 . 45)
     ($ganapfun46 . 46)($ganapfun47 . 47)($ganapfun48 . 48)($ganapfun49 . 49)($ganapfun50 . 50)
     ($ganapfun51 . 51)($ganapfun52 . 52)($ganapfun53 . 53)($ganapfun54 . 54)($ganapfun55 . 55)
     ($ganapfun56 . 56)($ganapfun57 . 57)($ganapfun58 . 58)($ganapfun59 . 59)($ganapfun60 . 60)
     ($ganapfun61 . 61)($ganapfun62 . 62)($ganapfun63 . 63)($ganapfun64 . 64)($ganapfun65 . 65)
     ($ganapfun66 . 66)($ganapfun67 . 67)($ganapfun68 . 68)($ganapfun69 . 69)($ganapfun70 . 70)
      )))

(define (init-curry-clause-list)
  (set! *special-curry-predsymbs*
        (copy-list
          (append
             '()
             *special-orig-curry-predsymbs*)))
  (set! *special-curry-funsymbs*
         (copy-list
           (append
              '()
              *special-orig-curry-funsymbs*)))
  (set! *curried-pred-const-table* '())
  (set! *curried-pred-nr* 0)
  (set! *curry-pred-arity-table* '())
  (set! *not-curried-preds* '())
  (set! *not-curried-pred-nr* 0)
  (set! *curried-fun-const-table* '())
  (set! *curried-fun-nr* 0)
  (set! *curry-fun-arity-table* '())
  (set! *not-curried-fun-nr* 0)
  (set! *not-curried-funs* '())
  (make-curried-special-predfuns-data) )


(define (curry-clause-list clauselst)
  (map (lambda(x)(curry-clause x))
       clauselst))

(define (curry-clause clause)
  (map (lambda(x)(curry-lit x))
       clause))

(define (curry-lit lit)
  ; (newline)(display "curry-lit: ")(display lit)(newline)
  (let ((predsymb '())
        (purepredsymb '())
        (predstr  '())
        (polarity '())
        (prednrplace '())
        (predarityplace '())
        (tmpres '())
        (arity 0)
        (appred '())
        (const '())
        (tmpconst '())
        (purepredstr '()) )
    ; locate predicate symbol
    (cond
      ((pair? lit)
         (set! predsymb (car lit)) )
      ((symbol? lit)
         (set! predsymb lit) )
      (else
         (set! predsymb #f) ))
    ; parse predicate symbol
    (set! predstr (symbol->string predsymb))
    (set! polarity (if (char=? #\- (string-ref predstr 0)) #f #t))
    (set! purepredstr
         (if polarity
             predstr
             (substring predstr 1 (string-length predstr)) ))
    (set! purepredsymb (string->symbol purepredstr))
    (set! predarityplace (assq purepredsymb *curry-pred-arity-table*))
    ; do the conversion, if necessary
    (cond
      ((not predsymb)    ; strange literal, do not change
         (set! tmpres lit) )
      ((memq purepredsymb *special-curry-predsymbs*) ; predicate may not be curried
         (set! tmpres lit) )
      ((not predarityplace) ; pred not yet handled, count it
         (if (>= *not-curried-pred-nr* *max-curry-pred-nr*)
             ; we must curry the pred
             (begin
                (set! *curried-pred-nr* (+ 1 *curried-pred-nr*))
                (set! arity (length (cdr lit)))
                (if (> arity 70)
                     (report-error "Too many predicates with too many arguments." #t) )
                (set! tmpconst (make-curry-const *curried-pred-nr*))
                (set! *curried-pred-const-table*
                      (cons (cons purepredsymb tmpconst)
                            *curried-pred-const-table*))
                 (set! *curry-pred-arity-table*
                      (cons (cons purepredsymb
                                  (length (cdr lit)) )
                             *curry-pred-arity-table*))
                (set! appred (car (rassq arity *curried-appred-arities*)))
                (set! tmpres
                      (cons (polarize-purepred-symb appred polarity)
                            (cons tmpconst
                                  (cdr lit)) )))
             ; predicate is not to be curried
             (begin
                (set! *not-curried-pred-nr* (+ 1 *not-curried-pred-nr*))
                (set! *not-curried-preds* (cons purepredsymb *not-curried-preds*))
                (set! *curry-pred-arity-table*
                      (cons (cons purepredsymb
                                  (length (cdr lit)) )
                             *curry-pred-arity-table*))
                (set! tmpres lit) )))
     (else ; pred is handled already
         (if (not (memq purepredsymb *not-curried-preds*))
              ; a known pred to be curried
             (begin
                (set! arity (length (cdr lit)))
                (set! const
                      (cdr (assq purepredsymb *curried-pred-const-table*)) )
                (if (> arity 70)
                     (report-error "Too many predicates with too many arguments." #t) )
                (set! appred (car (rassq arity *curried-appred-arities*)))
                (set! tmpres
                      (cons (polarize-purepred-symb appred polarity)
                            (cons const (cdr lit))) ))
             ; a known pred not to be curried
             (begin
                (set! tmpres lit) ))))
     ; return the result
     (if (pair? tmpres)
         (cons (car tmpres) (map (lambda(x)(curry-term x)) (cdr tmpres)))
         tmpres) ))


(define (polarize-purepred-symb symb polarity)
   (if polarity
       symb
       (string->symbol (string-append "-" (symbol->string symb))) ))


(define (curry-term term)
 (let  ((funsymb '())
        (purefunsymb '())
        (funnrplace '())
        (funarityplace '())
        (tmpres '())
        (arity 0)
        (tmpconst '())
        (apfun '())
        (const '()) )
  (if (not (pair? term))
   ; term is not a pair
   term
   ; term is a pair
   (begin
    ; do the recursive call
    (set! term (cons (car term)
                     (map (lambda(x)(curry-term x)) (cdr term)) ))
    ; locate fun symbol
    (set! funsymb (car term))
    (set! purefunsymb funsymb)
    (set! funarityplace (assq purefunsymb *curry-fun-arity-table*))
    ; do the conversion, if necessary
    (cond
      ((memq purefunsymb *special-curry-funsymbs*) ; fun may not be curried
         (set! tmpres term) )
      ((not funarityplace) ; fun not yet handled, count it
         (if (>= *not-curried-fun-nr* *max-curry-fun-nr*)
             ; we must curry the fun
             (begin
                (set! *curried-fun-nr* (+ 1 *curried-fun-nr*))
                (set! tmpconst (make-curry-const *curried-fun-nr*))
                (set! arity (length (cdr term)))
                (if (> arity 70)
                     (report-error "Too many functions with too many arguments." #t) )
                (set! *curried-fun-const-table*
                      (cons (cons purefunsymb tmpconst)
                            *curried-fun-const-table*))
                (set! *curry-fun-arity-table*
                       (cons (cons purefunsymb
                                  (length (cdr term)) )
                             *curry-fun-arity-table*))
                (set! apfun (car (rassq arity *curried-apfun-arities*)))
                (set! tmpres
                      (cons apfun
                            (cons tmpconst
                                  (cdr term)) )))
             ; fun is not to be curried
             (begin
                (set! *not-curried-fun-nr* (+ 1 *not-curried-fun-nr*))
                (set! *not-curried-funs* (cons purefunsymb *not-curried-funs*))
                (set! *curry-fun-arity-table*
                       (cons (cons purefunsymb
                                  (length (cdr term)) )
                             *curry-fun-arity-table*))
                (set! tmpres term) )))
     (else ; fun is handled already
         (if (not (memq purefunsymb *not-curried-funs*))
              ; a known fun to be curried
             (begin
                (set! arity (length (cdr term)))
                (set! const
                      (cdr (assq purefunsymb *curried-fun-const-table*)) )
                (if (> arity 70)
                     (report-error "Too many functions with too many arguments." #t) )
                (set! apfun (car (rassq arity *curried-apfun-arities*)))
                (set! tmpres
                      (cons apfun
                            (cons const (cdr term))) ))
             ; a known fun not to be curried
             (begin
                (set! tmpres term) ))))
     ; return the result
     tmpres))))

(define (make-curry-const nr)
  (string->symbol (string-append "$ganc" (number->string nr))) )

(define (make-set lst)
  (if (null? lst)
      '()
      (if (member (car lst) (cdr lst))
          (make-set (cdr lst))
          (cons (car lst) (make-set (cdr lst))) )))

;(define (need-to-curry-clause-list clauses)
;  (for-each (lambda(x)(need-to-curry-clause clause)) clauses) )



;=======================================================================
;
;       for sat only: converting clauses with ground terms
;
;======================================================================

(define *groundprocess-const-nr* 0)
(define *groundprocess-term-const-pairs* '())

(define (groundprocess-input-readylst readylst)

    ;(newline)
    ;(pretty-print readylst)
    ;(newline)
    ;(display "readylst: ")(newline)
    ;(pretty-print readylst)

  (let ((sospart (assq 'sos readylst))
        (usablepart (assq 'usable readylst))
        (res '())
        (tmp '())
        (sosaddedflag #f) )
    (set! *groundprocess-const-nr* 0)
    (set! *groundprocess-term-const-pairs* '())
    (set! tmp
          (filter-map
             (lambda(x)
               (if (or (not (pair? x))
                       (not (memq (car x) '(sos usable))) )
                   x
                   #f))
             readylst))

;    (for-each (lambda(cl)
;                (for-each (lambda (lit)
;                            (if (and (pair? lit)
;                                     (> (length lit) *curried-max-orig-pred-arity*) )
;                                (set! *curried-max-orig-pred-arity* (length lit)) ))
;                          cl))
;              (append (cdr sospart) (cdr usablepart)) )


    ;(init-groundprocess-clause-list)

    (if sospart (set! sospart (groundprocess-clause-list (cdr sospart))))
    (if usablepart (set! usablepart (groundprocess-clause-list (cdr usablepart))))
    (set! res tmp)

    ;(newline)
    ;(display "readylst: ")
    ;(pretty-print readylst)
    ;(newline)
    ;(display "res: ")(newline)
    ;(pretty-print res)

    (if (= *groundprocess-const-nr* 0)
        (begin
          (set! res readylst) )
        (begin

          (if sospart
              (begin
               (set! sosaddedflag #t)
               (set! sospart (append (groundprocess-make-equalities)
                                     sospart))
               (set! res (cons (cons 'sos sospart) res))))
          (if usablepart
              (begin
                (if (not sosaddedflag)
                    (set! usablepart (append (groundprocess-make-equalities)
                                     usablepart)))
                (set! res (cons (cons 'usable usablepart) res))))

          (newline)
          (display "Some ground terms were converted.")
          (newline) ))

    res))

(define (groundprocess-make-equalities)
   (let ((tmp '())
         (varsymb '()) )
     (set! tmp
           (map (lambda(x)
                  (list (list 'equal (car x) (cdr x))) )
                 *groundprocess-term-const-pairs*))
     (if *prolog-style-variables*
         (set! varsymb '^x)
         (set! varsymb 'x) )
     (set! tmp
          (cons (list (list 'equal varsymb varsymb)) tmp))
     tmp))

(define (groundprocess-clause-list clauselst)
  (map (lambda(x)(groundprocess-clause x))
       clauselst))

(define (groundprocess-clause clause)
  (map (lambda(x)(groundprocess-lit x))
       clause))


(define (groundprocess-lit lit)
  (let ((tmp '()))
    (cond
      ((not (pair? lit))
         lit)
      (else
         (set! tmp (map (lambda(x)(groundprocess-term x))
                        (cdr lit) ))
         (cons (car lit) tmp) ))))


(define (groundprocess-term term)
  (let ((tmp '())
        (newc '()) )
    (cond
      ((not (pair? term))
         term)
      (else
         (set! tmp (map (lambda(x)(groundprocess-term x))
                        (cdr term) ))
         (if (every? (lambda(y)
                       (and (not (pair? y))
                            (not (var-symbol? y)) ))
                     tmp)
             (begin
               (set! newc (assoc (cons (car term) tmp)
                                 *groundprocess-term-const-pairs*))
               (if newc
                   (cdr newc)
                   (begin
                     (set! newc (groundprocess-term-newconstant))
                     (set! *groundprocess-term-const-pairs*
                           (cons (cons (cons (car term) tmp) newc)
                                 *groundprocess-term-const-pairs*))
                     newc)))

             (cons (car term) tmp) )))))

(define (groundprocess-term-newconstant)
  (set! *groundprocess-const-nr* (+ 1 *groundprocess-const-nr*))
  (string->symbol
    (string-append "$gnew"
                   (number->string *groundprocess-const-nr*))) )


;=======================================================================
;
;                      batch version
;
;======================================================================

(define (abs-path-part path)
  (substring path
             0
             (- (string-length path) (string-length "gandalf"))  ))

(define (batch)
 (display (program-arguments))
 (fprove "tmp2.in") )


(define (batch)
 (if (null? (cdr (program-arguments)))
    (begin (display "No argument file name given: going interactive.")
           (newline)
           (abort)
           (quit))
    (begin ; (newline)
      (if (null? (cddr (program-arguments)))
          (begin
            (set! *timelimit* #f)
            (fprove (cadr (program-arguments))) )
          (begin
            (set! *timelimit*
                  (car (cddr (program-arguments))))
            (fprove (cadr (program-arguments))) ))
      (newline)
      (quit) )))

(define (show-old-units)
  (for-each
   (lambda(x)
     (display "--------------------------")
     (newline)
     (for-each
       (lambda (sublist)
        (for-each
         (lambda (el)
           (print-clause el)(newline) )
         (cdr sublist) ))
        x))
   *unit-alist-list*))

;(display "starting to run batch")
;(newline)

;(batch)

;(if (pair? x) (print-clause x))(newline))  *usable-clauses-list*)

; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))  *usable-clauses-list*)

; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))  *sos-clauses-list*)

;(trace check-discr-tree-part)
;(trace check-nonunit-discr-tree)
;(trace check-discr-tree-leaf)


;============================================================

;(set!  *prolog-style-variables* #t)
;(init-hyper-strat)

; (set! *timelimit* "10")

; (set! *qp-width* 1500)

;(trace hyper-contradiction?)
;(trace given-clause-slow-loop)
;(trace usable-clauses-fast-loop)
;(trace two-clause-binres-loop)
;(trace one-step-for-hyper)
;(trace resolve-unify-ordinary)
;(trace make-poslit-preferred!)
;(trace mixed-cl?)
;(trace make-all-resolvable!)

; (fprove "problems/luk3.prb" #f 60)
; (fprove "hyptest1.prb" #f 120)
; (fprove "problems/luk2.prb" #f 60)
; (fprove "problems/steam.in" #f 60)
;derived clause after 2-nd order deco:
;-p(i(x,i(i(i(i(y,z),i(i(u,v),i(v,i(u,v)))),i(w,i(i(u,v),i(v,i(u,v))))),i(x1,i(i(i(a,b),a),a))))) | -p(x1) | -p(x).
;(trace resolve-unify-cutoff)
;"CALLED" term-may-cont-vars? (2099296 24836 24834)
;"RETURNED" term-may-cont-vars? #t
;"CALLED" occur-check? 24580 (2099296 24836 24834)
; "CALLED" term-may-cont-vars? (2099296 24836 24834)
; "RETURNED" term-may-cont-vars? #t
; "CALLED" occur-check? 24580 (2099296 24836 24834)

;(trace otter-show-history-aux-aux)
;(trace otter-show-history-aux)
;(trace show-history-clnr)
;(trace show-history-find-clausepl)
;(set! *qp-width* 1500)
; (trace resolve-unify-ordinary)
; (trace paramodulate-into-term-aux)
; (trace paramodulate-into-term)


(define (show-derived)
  (newline)
  (display "*sos-clauses-list*:")(newline)
  (for-each (lambda(x)(if (pair? x)(print-clause x))(newline)) *sos-clauses-list*)
  (newline)
  (display "*usable-clauses-list*:")(newline)
  (for-each (lambda(x)(if (pair? x)(print-clause x))(newline)) *usable-clauses-list*) )

(define (go)
  ; (fprove "../Examples/Devcasc/LCL223-1.in"))
   (fprove "tst.in") )
  ;(fprove "../Examples/Devcasc/SYN827-1.in") )
  ; (fprove "../Examples/Devcasc/MGT029-1.in") )
  ; (fprove "../Examples/eqp1.in") )


(define (sgo)
  ;(fprove "../Examples/COL033-1.in"))
  ;(set! *prop-debug* #t)
  (fsat "tst.in") )
  ;(fsat "../Examples/Devcasc/LCL240-1.in") )
  ;(fsat "../Examples/Devcasc/GEO157-1.in") )
  ;(fsat "../Examples/MGT031-1.in") )

(define (q) (quit))

;-------------------------------------------------------------


(define (bgo)
  (let* ((filename "tst0.in")
         (iport (open-input-file filename))
        ;(iport (call-with-input-file filename string-port-handler))
         (lst '())
         (lst-end '())
         (res '())
         (pred '())
         (predstr '())
         (polarity '())
         (purepred '())
         (table '())
         (clauses 0)
         (literals 0)
         (clauselst '())
         (clause '())
         (nr 0)
         (place '()) )
    (set! lst (cons '() '()))
    (set! lst-end lst)
    (do ((obj (read iport) (read iport)))
        ((eof-object? obj)
          (close-input-port iport)
          (set! lst (cdr lst))
          (set! res lst) )
      (set-cdr! lst-end (cons obj '()))
      (set! lst-end (cdr lst-end)) )

    (newline)
    (display "list was read in, length: ")
    (display (length res))
    (newline)


   (set! table '())
   (set! nr 0)
   (set! clauses 0)
   (set! literals 0)
   (set! clauselst '())
   (set! clause '())
   (do ((part res (cdr part)))
       ((null? part))
     (set! pred (car part))
     ;(display pred)(newline)
     (cond
       ((eq? pred 'endofline)
          (set! clauselst (cons clause clauselst))
          (set! clause '())
          (set! clauses (+ 1 clauses)) )
       ((eq? pred '|))
       ((symbol? pred)
          (set! predstr (symbol->string pred))
          (set! polarity (if (char=? #\- (string-ref predstr 0)) #f #t))
          (set! purepred (if polarity pred
           (string->symbol (substring predstr 1 (string-length predstr))) ))
          (set! literals (+ 1 literals))
          (set! place (assq purepred table))
          (if place
              (if polarity
                  (set! clause (cons (cdr place) clause))
                  (set! clause (cons (- 0 (cdr place)) clause)) )
              (begin
                  (set! nr (+ 1 nr))
                  (set! table (cons (cons purepred nr) table))
                  (if polarity
                     (set! clause (cons nr clause))
                     (set! clause (cons (- 0 nr) clause)) ))))))

   (newline)
   (display "list analysed fully")
   (newline)
   (display "nr:")(display nr)
   (newline)
   (display "literals:")(display literals)
   (newline)
   (display "clauses:")(display clauses)
   (newline)

   (set! clauselst (reverse clauselst))

   (display "p cnf ")(display nr)(display " ")(display clauses)
   (newline)
   (do ((part clauselst (cdr part)))
       ((null? part))
     (do ((clpart (car part) (cdr clpart)))
         ((null? clpart)
            (display "0")
            (newline) )
       (display (car clpart))
       (display " ") )) ))




;==================================================================
;
;                    EOF of resio.scm
;
; It is necessary to also have resbase.scm and resmain.scm
;
;========================== END ===================================

