;======================================================================
;
;                          resmain.scm
;
;                   Gandalf version c-2.6 r1
;
;  Top-level, main loops, debugging, I/O, etc of the resolution
;  proof searcher.
;
;                  Copyright (c) Tanel Tammet
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 1, or (at your option)
; any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;
; The author can be reached at tammet@staff.ttu.ee, tanel@cs.ioc.ee or
; tammet@cs.chalmers.se
;
;
;=======================================================================


(define compile-inline
  '(litinf-left-para? litinf-right-para? litinf-parainto?
    pos-eq-pred? pos-eq-fun? litinf-binrespara?
    eq-lit-arg1 eq-lit-arg2 eq-lit-type may-be-demodulated?
    litinf-use make-binres-history))


;====================================================================
;
;                 main resolution loops
;
;====================================================================

; given-clause-slow-loop assumes that both *sos-clauses-list* and
; *usable-clauses-list* are given

(define *sos-preference-ratio* 5) ; the number of times that the ; RIGHT: 5
   ; smallest clause is preferred over the next-in-queue

(define *sos-preference-cur-ratio* 0) ; the local number of times of
   ; best-choice until the next choice of next-in-queue

(define *sos-preference-distance-ratio* 9)
;(define *sos-preference-distance-ratio* 3)

(define *sos-preference-cur-distance* 0)

(define *sos-preference-exp-ratio* 2)
(define *sos-preference-cur-exp* 0)

(define *given-clause-nr* 0) ; counting the number of given clauses so far

(define (display-kept-nr)
  (newline)
  (display "!!!!!!! ")
  (display *checkpoint-keptnrs* )
  (display "!!!!!!! ")
  (newline) )

(define *new-demodulators-list* '())

(define *slow-start-time* 0)
(define *slow-passed-time* 0)

;(define *given-cl-initial-count* 0)

(define (given-clause-slow-loop)
  ; (set! *slow-passed-time* 0)

  ;(set! *given-cl-initial-count* 0)

  (set! *sos-exhausted-flag* #f)
  (set! *last-best-weight* 0)
  (set! *last-best-clausesize* 0)
  (set! *foundflag* #f)
  (set! *sat-found-flag* #f)
  (set! *answer-cl* '())
  (set! *new-demodulators-list* '())
  (set! *answer-history* '())
  ;(set! *min-clause-length-in-sos* 1)
  ; (set! *given-clause-nr* 0)
  (set! *slow-start-time*  (get-check-time 0))
  (store-time-checkpoint! 0)
  (if (not *int-logic-flag*)
      (set! *sos-clauses-list-end* (my-last-pair *sos-clauses-list*)) )
  (set! *usable-clauses-list* (cons '() *usable-clauses-list*))
  (set! *usable-clauses-list-end* (my-last-pair *usable-clauses-list*))
  (clear-x-vars-and-touched! *max-nr-of-vars-in-clause*)
  (clear-y-vars-and-touched! *max-nr-of-vars-in-clause*)
  (clear-z-vars-and-touched! *max-nr-of-vars-in-clause*)
  (set! *one-step-hyper-phase* #f)

  (if (and
       *firstlitpref-order-strategy-flag*
       (or *heq-class-flag* *hne-class-flag*) )
      (set! *sos-preference-exp-ratio* 2)
      (set! *sos-preference-exp-ratio* 33) )

  (set! *sos-preference-cur-ratio* *sos-preference-ratio*)
  (set! *sos-preference-cur-distance* *sos-preference-distance-ratio*)
  (set! *sos-preference-cur-exp*  *sos-preference-exp-ratio*)
  (if *factors-stratflag* (factor-sos *sos-clauses-list*))
  (do ()
      ((or (null? *sos-clauses-list*)
           ;*stop-fast-loop*
           *foundflag*
           ;(and (not (= 0 *stop-res-real-time*))
           ;                 (>= (get-check-time) *stop-res-real-time*)) )
           (> (- (get-check-time 1) *slow-start-time*)
              *one-pass-time*))
         (if (and (not (null? *sos-clauses-list*))
                  (not *foundflag*) )
             (begin
               (set! *sos-exhausted-flag* #f)
               (store-time-checkpoint! 1)
               (if (> *display-info-level* 38)
                   (begin (newline)
                          (display "proof attempt stopped: time limit")
                          (newline) ))
               (set! *answer-history* 'not-enough-time) )
             (begin
                (set! *sos-exhausted-flag* #t)
                (if (and (null? *sos-clauses-list*)
                      (> *display-info-level* 38) )
                    (begin (newline)
                           (display "proof attempt stopped: sos exhausted")
                           (newline) ))))
         *foundflag*)
      ; time actions
      ; (set! *slow-passed-time* (- (get-check-time) *slow-start-time*))

      ;(newline)
      ;(display *one-pass-time*) (display " ")
      ;(display *slow-passed-time*)
      ;(newline)

      ;(if (and (not *sat-strategy-flag*)
      ;         (or *one-pass-time*
      ;                   (not (null? *given-checkpoint-keptnrs*)) )
      ;               ;(< *sos-preference-ratio* 100)
      ;               ;(< (* 100  (- (get-check-time 2) *slow-start-time*))
      ;               ;          (* 5 *one-pass-time*) ))
      ;          (begin
      ;            (store-time-checkpoint! 45)
      ;            (if (>= *display-info-level* 50)
      ;                (begin
      ;                  (newline)
      ;                  (display "using queue strategy")
      ;                  (newline) ))
      ;            (set! *sos-preference-ratio* 0) ))

      (if (and (not *sat-strategy-flag*)
               (or *one-pass-time*
                   (not (null? *given-checkpoint-keptnrs*)) )
               *unit-equality-strat*
               (< *sos-preference-ratio* 5)
               (> (* 100  (- (get-check-time 2) *slow-start-time*))
                  (* 20 *one-pass-time*) ))
          (begin
            (store-time-checkpoint! 55)
            (if (>= *display-info-level* 50)
                (begin
                  (newline)
                  (display "switching to mixed prefer-short strategy")
                  (newline) ))
            (set! *sos-preference-ratio* 5) ))

      (if (and (not *sat-strategy-flag*)
               (or *one-pass-time*
                   (not (null? *given-checkpoint-keptnrs*)) )
               *unit-equality-strat*
               (< *sos-preference-ratio* 100)
               (> (* 100  (- (get-check-time 2) *slow-start-time*))
                  (* 50 *one-pass-time*) ))
          (begin
            (store-time-checkpoint! 45)
            (if (>= *display-info-level* 50)
                (begin
                  (newline)
                  (display "switching to pure prefer-short strategy")
                  (newline) ))
            (set! *sos-preference-ratio* 100000) ))


      (if (and (not *sat-strategy-flag*)
               (or *one-pass-time*
                   (not (null? *given-checkpoint-keptnrs*)) )
               (not *unit-equality-strat*)
               *binary-hyper-sos-strategy*
               (not *hyper-strategy-flag*)
               (> (* 100  (- (get-check-time 2) *slow-start-time*))
                  (* 50 *one-pass-time*) ))
          (begin
            (store-time-checkpoint! 2)
            (if (>= *display-info-level* 50)
                (begin
                  (newline)
                  (display "switching to hyper-sos strategy")
                  (newline) ))
            (set! *unit-strategy-flag* #f)
            (set! *double-strategy-flag* #f)
            (set! *name-order-strategy-flag* #f)
            (set! *hyper-strategy-flag* #t)
            (set! *para-pos-unit-strategy* #t)
            (set! *hyper-strategy-flag* #t) ))


      (if (and (not *sat-strategy-flag*)
               (or *one-pass-time*
                   (not (null? *given-checkpoint-keptnrs*)) )
               (not *unit-equality-strat*)
               (or (not *unit-strategy-flag*)
                   (< *sos-preference-ratio* 100) )
               (not *binary-hyper-sos-strategy*)
               (not *hyper-contradiction-strategy*)
               (not *hyper-strategy-flag*)
               (> (* 100 (- (get-check-time 3) *slow-start-time*))
                  (* 50 *one-pass-time*) ))
          (begin
            (store-time-checkpoint! 3)
            (if (>= *display-info-level* 50)
                (begin
                  (newline)
                  (display "switching to unit restriction, pure short pref")
                  (newline) ))
            (set! *sos-preference-ratio* 100000)
            (set! *para-unit-strategy* #t)
            ;(set! *double-strategy-flag* #t) ))
            (set! *unit-strategy-flag* #t) ))

      (if (and (not *sat-strategy-flag*)
               *bliksem-strategy-flag*)
          (if (> *kept-clauses-nr* (+ 300 *one-pass-kept-clauses-firstnr*))
              ; now limit clause weights!
              (begin
                ;(newline)
                ;(display "now starting to limit clause weight to ")
                ;(display *clause-finalkeepsize-limit*)
                ;(newline)
                (set! *clause-keepsize-limit* *clause-finalkeepsize-limit*)) ))

      (if (and  (not *sat-strategy-flag*)
                (or *one-pass-time*
                   (not (null? *given-checkpoint-keptnrs*)) )
                (not *unit-equality-strat*)
                (not *binary-hyper-sos-strategy*)
                   (not *hyper-contradiction-strategy*)
                   (> (* 100 (- (get-check-time 4) *slow-start-time*))
                      (* 75  *one-pass-time*) ))
          (begin
            (store-time-checkpoint! 4)
            (if *hyper-strategy-flag*
                (begin
                  (if (>= *display-info-level* 50)
                      (begin
                        (newline)
                        (display "switching to backward hyper-analysis")
                        (newline) ))
                  (start-backward-hyper-analysis)
                  (set! *usable-clauses-list* (cons '() *usable-clauses-list*))
                  (set! *usable-clauses-list-end*
                        (my-last-pair *usable-clauses-list*))
                  (set! *sos-preference-cur-ratio* *sos-preference-ratio*) )
                (begin
                  (if (>= *display-info-level* 50)
                      (begin
                        (newline)
                        (display "switching to full hyper-contradiction-check")
                        (newline)  ))
                  (start-binary-hypercheck-strategy)
                  (if (>= *display-info-level* 50)
                      (begin
                        (newline)
                        (display
                         "continuing search with hyper-contradiction-check")
                        (newline) ))))))


          ;(and *one-pass-time*
          ;     (> (* 5 (- (get-check-time) *slow-start-time*))
          ;            (* 3 (* *one-pass-time* 60)) )) ; internal-time-units-per-second
          ;(begin
          ;  (newline)
          ;  (display "switching to unit restriction")
          ;  (newline)
          ;  (set! *one-pass-time* #f)
          ;  (set! *unit-strategy-flag* #t) ))

    (let* ((given-cl-place (select-given-cl-place *sos-clauses-list*))
           (given-cl (car given-cl-place))
           (tmp '()) )

      ; remove given-cl from sos
      (if (eq? given-cl-place *sos-clauses-list*)
          ; sos as a queue
          (set! *sos-clauses-list* (cdr *sos-clauses-list*))
          ; best-first
          (if (null? (cdr given-cl-place))
              ; last element selected
              (begin (remove-at-place! given-cl-place *sos-clauses-list*)
                     (set! *sos-clauses-list-end*
                           (my-last-pair *sos-clauses-list*) ))
              ; the selected el was not the last one
              (begin (remove-at-place! given-cl-place *sos-clauses-list*)) ))

      (if
        (and
           ; clause-strength is zero?
           (not (zero? (logand *term-depth-strength-mask*
                               (caar given-cl))))
           (not *foundflag*)
           ; back-demodulation
           (if (not *back-demodulate-flag*)
               #t
               (begin
                 (set! *special-insertion-place* (cons '() '())) ; dummy
                 (set! tmp (demodulate-clause given-cl))
                 (set! *special-insertion-place* '())
                 (if (not tmp)
                     #t
                     (set! given-cl (cdr tmp)) )
                 #t))
           ; special subsumption check, since back-subsumption restricted
           (or #f; *int-logic-flag*
               (not (given-cl-havealready? given-cl)) ))
        ; given-cl is usable
        (begin

          ; now process given clause

          (if (not (null? (cddr given-cl)))
              (set! given-cl (handle-ready-clause given-cl)) )

          (set! *given-clause-nr* (+ 1 *given-clause-nr*))
          ; append given-cl to the end of *usable-clauses-list*
          (set-cdr! *usable-clauses-list-end* (cons given-cl '()))
          (set! *usable-clauses-list-end* (cdr *usable-clauses-list-end*))
          ; insert given-clause to subsumption list
          (insert-to-subsumption-list!
            (shiftright (logand *cl-len-mask* (caar given-cl)) *cl-len-shift*)
            (logand *term-depth-strength-mask* (caar given-cl))
            given-cl)
          ; create an y-version of given-clause,
          ; precompute some data for given-clause

          (if (>= *display-info-level* 50)
              (begin
                (newline)
                (display "******* ")
                (display *given-clause-nr*)
                (display "-th given clause: ")
                (newline)
                ;(display given-cl)
                ;(newline)
                ; (print-pure-clause (cdr given-cl))
                (print-full-clause given-cl)
                (newline)
                (display "distance: ")
                (display (clinfo-distance (car given-cl)))
                (newline)
                (display "kept size sum: ")
                (display *kept-size-sum-nr*)
                (newline)


                ;(debug-print-xvars 3)(debug-print-yvars 3)
                ;(newline)
                ;(display "length of sos-clauses-list is: ")
                ;(display (length *sos-clauses-list*)) (newline)
                ;(for-each (lambda(x)(print-pure-clause (cdr x))(newline))
                ;          *sos-clauses-list*)
                ;(newline)

                (if *display-debug-info-flag*
                    (begin
                      (newline)(display "++++++ given clause in debug form: ")
                      (newline)
                      (debug-print-clause given-cl)(newline) ))))
          (set! *stop-fast-loop-flag* #f)
          (if (and *back-demodulate-flag*
                   (not (null? *new-demodulators-list*)) )
              (back-demodulate-usable) )
          (set! *new-demodulators-list* '())
          (if *int-logic-flag*
              (int-usable-clauses-fast-loop given-cl #t)
              (usable-clauses-fast-loop given-cl #t) ))

      (if *keep-all-old-clauses-flag*
          (set! *all-back-subsumed-clauses*
                (cons given-cl *all-back-subsumed-clauses*) )) ))) )



; in select-given-cl-place the best obvious candidates are perhaps
; *term-depth-strength-mask* and *term-size-strength-mask*
;

(define *last-best-weight* 0)
(define *last-best-clausesize* 0)

;(define *min-clause-length-in-sos* 999999)

(define (select-given-cl-place lst)
  ;(newline)
  ;(display "*sos-preference-cur-distance*: ")
  ;(display *sos-preference-cur-distance*)
  ;(newline)
  ;(newline)
  ;(display (map (lambda(x)(logand *cl-strength-mask* (caar x)))
  ;                lst))
  ; (newline)
  ;(set! *given-cl-initial-count* (+ 1 *given-cl-initial-count*))
  (if (null? lst)
      (set! lst (copy-tree '(((0))) ))) ; no sos, return dummy
  (if (<= *sos-preference-cur-ratio* 0)
          ; (< *given-cl-initial-count* (* 20 20))
      (begin (set! *sos-preference-cur-ratio* *sos-preference-ratio*)
             lst)
      (let ((bestpl lst)
            (hist '())
            (clinf '())
            (distanceflag #f)
            (bestdistance #f)
            (expflag #f)
            (thisweight 0)
            (bestweight (logand *term-size-strength-mask* (caaar lst))) )
        (if (<= *sos-preference-cur-distance* 0)
            (begin
               (set! *sos-preference-cur-distance*
                     *sos-preference-distance-ratio*)
               (set! distanceflag #t) ))
        (if (<= *sos-preference-cur-exp* 0)
            (begin
               (set! *sos-preference-cur-exp*
                     *sos-preference-exp-ratio*)
               (set! expflag #t) ))
        ;(set! *min-clause-length-in-sos* 999999)
        (set! *last-best-weight* 9999999)
        (set! *sos-preference-cur-ratio* (- *sos-preference-cur-ratio* 1))
        (set! *sos-preference-cur-distance*
             (- *sos-preference-cur-distance* 1))
        (set! *sos-preference-cur-exp* (- *sos-preference-cur-exp* 1))
        (if *split-strategy-flag*
            (do ((lstpart lst (cdr lstpart)))
                ((null? lstpart)
                 ;(display (car bestpl))
                 ;(newline)
                 (if (not (null? bestpl))
                     (set! *last-best-clausesize*
                          (compute-real-clausesize (cdar bestpl)) ))
                 (if (equal? *last-best-weight* 9999999)
                     (set! *last-best-weight* bestweight) )
                 bestpl)
              (if (and (< (logand *term-size-strength-mask*
                                 (cadddr (caar lstpart)) )
                             bestweight)
                            ;(begin
                            ;    (set! hist (car (clinfo-history-sublist
                            ;                      (caar lstpart))))
                            ;        (or (null? hist)
                            ;            (eq? 'input (car hist)) ))

                       (not (zero? (logand *term-size-strength-mask*
                                           (caaar lstpart)) )))
                  (begin (set! bestpl lstpart)
                         (set! *last-best-weight* bestweight)
                         (set! bestweight
                               (logand *term-size-strength-mask*
                                       (cadddr (caar lstpart)) )))))
            (do ((lstpart lst (cdr lstpart)))
                ((null? lstpart)

                 ;(display (logand *cl-strength-mask* (caaar bestpl)))
                 ;(newline)
                 ;(display (car bestpl))
                 ;(newline)
                 (if (not (null? bestpl))
                     (set! *last-best-clausesize*
                          (compute-real-clausesize (cdar bestpl)) ))
                 (if (equal? *last-best-weight* 9999999)
                     (set! *last-best-weight* bestweight) )
                 bestpl)
              ;(newline)
              ;(print-clause (car lstpart))
              ;(display (clinfo-distance (caar lstpart)))
              ;(if (< (length (cdar lstpart)) *min-clause-length-in-sos*)
              ;    (set! *min-clause-length-in-sos* (length (cdar lstpart))) )
              (set! clinf (caar lstpart))
              (if expflag
                  (begin
                    (set! thisweight
                         (logand *term-depth-strength-mask* (car clinf)) ))
                    ;(set! thisweight (length (car lstpart)))
                    ;(set! thisweight (shiftleft thisweight 7))
                    ;(set! thisweight (+ thisweight
                    ;                   (logand *term-size-strength-mask* (car clinf)) )))
                  (set! thisweight
                    (logand *term-size-strength-mask* (car clinf)) ) )

              (if (or
                     (and distanceflag
                          (not bestdistance)
                          (clinfo-distance clinf) )
                     (and (< thisweight
                             bestweight)
                          (not (zero? (logand *term-size-strength-mask*
                                                (car clinf)) ))
                          (or (not distanceflag)
                              (not bestdistance)
                              (clinfo-distance clinf) )))
                   ;(or
                   ;  (and distanceflag
                   ;       (not bestdistance)
                   ;       (clinfo-distance clinf) )
                   ;  (and (< (logand *term-size-strength-mask* (car clinf))
                   ;            bestweight)
                   ;         (not (zero? (logand *term-size-strength-mask*
                   ;                             (car clinf)) ))
                   ;         (or (not distanceflag)
                   ;             (not bestdistance)
                   ;             (clinfo-distance clinf) )))


                  (begin (set! bestpl lstpart)
                         (set! *last-best-weight*
                                 (logand *term-size-strength-mask* (car clinf)) )
                         (if *sos-preference-cur-distance*
                             (set! bestdistance (clinfo-distance clinf)) )
                         (set! bestweight
                               thisweight)
                               ;(logand *term-size-strength-mask*
                               ;        (caaar lstpart))
                          )))))))


(define (given-cl-havealready? given-cl)
  (let ((tmp #f)
        (res #f)
        (cl given-cl) )

    ; (display " given-cl-havealready? entered:")(newline)
    ; (debug-print-clause given-cl)   (newline)
    ; (print-clause given-cl)

    (set! *subsuming-clause* #f)
    ; prohibit subsumption with a given clause itself!
    (set! *prohibit-subs-clnr*
                (car (clinfo-clnumber-sublist (car given-cl))) )
    ;(if (not *classical-synth-stratflag*)
    ;    (set! cl (cons (car given-cal) (remove-ans-literals (cdr given-cl))))
    ;    (set! cl given-cl) )
    (if (null? (cddr given-cl))
        ; unit clause
        (begin
          (set! tmp (check-discr-tree *discr-tree* (cadr given-cl)))
          (if (and tmp (not (null? (cdr tmp))) (not (null? (cddr tmp)))
                       (eq? (logand *sgn-mask* (caadr given-cl))
                            (logand *sgn-mask* (caadr (caddr tmp))) ))
              (set! res (caddr tmp))
              (set! res #f) ))
        ; nonunit clause
        (begin
           (set! res
                 (some? (lambda(lit)

                          ;(display "about to check lit: ")
                          ;(print-lit lit)(newline)

                          (set! tmp (check-discr-tree *discr-tree* lit)) ; 25 mar 2002

                          ;(display "lit check tree finished")(newline)

                          (if (and tmp
                                (not (null? (cdr tmp)))
                                (not (null? (cddr tmp)))
                                (eq? (logand *sgn-mask* (car lit))
                                     (logand *sgn-mask* (caadr (caddr tmp))) ))
                              (caddr tmp)
                              #f))
                       (cdr given-cl) ))

           ; (display " checkpoint 1 reached")(newline)

           (if (not res)
               (set! res
                     (havealready-clause? (car given-cl) (cdr given-cl)) ))))

    (set!  *prohibit-subs-clnr* #f)
    (if res
        (begin
          (set! *back-subsumed-nr* (+ 1 *back-subsumed-nr*))
          (set-car! (car given-cl)
                    (logxor *term-depth-strength-mask*
                            (logior *term-depth-strength-mask*
                                    (caar given-cl) )))))
    (if (and res
             (>= *display-info-level* 50) )
        (begin
          (newline)
          (display "Given cl candidate ")
          (newline)
          (print-pure-clause (cdr given-cl))
          (display " is subsumed by ")
          (newline)
          (if *subsuming-clause*
              (print-pure-clause (cdr *subsuming-clause*))
              (print-pure-clause (cdr res)) )
          (newline) ))
    res))


; full-nucleus-flag is #t iff usable-clauses-flag has been
; given an unchanged given-cl. It is #f iff given-cl is
; a nucleus with some literals already cut off.

(define  *stop-fast-loop-flag* #f)

(define (usable-clauses-fast-loop given-cl full-nucleus-flag)

  ; (newline)(display "usable-clauses-fast-loop starts")(newline)

  (let* ((ycl (cdr given-cl))
         (yinf (car given-cl))
         (ylitinf (caddr yinf))
         (yclmaxvar (give-cl-maxvar yinf))
         (given-cl-is-nuc-flag #f)
         (given-cl-is-neg-flag #f)
         (given-cl-is-final-nuc-flag #f) ; a nuc with just one neg lit
         (given-final-nuc-premiss-lit '())
         (given-final-nuc-complfunpure '())
         (given-cl-contains-eq-flag #f)
         (given-is-unit-flag #f)
         (given-is-double-flag #f) )
    (set! *used-demodulators* '())
    (if full-nucleus-flag
        (set! ycl (map rename-x-to-y-term ycl))
        (set! ycl (map rename-x-to-y-term ycl)) )
        ; (for-each rename-x-to-y-term! ycl)   ; for efficiency NB! dangerous
    (if (unit-cl? ycl) (set! given-is-unit-flag #t))
    (if (double-cl? ycl) (set! given-is-double-flag #t))
    (if *hyper-strategy-flag*
        (if (if *neg-hyper-strategy-flag*
                (negative-cl? ycl)
                (positive-cl? ycl) )
            ; given clause is not a nucleus
            (begin (set! given-cl-is-nuc-flag #f)
                   (if (contains-eqlit-cl? ycl)
                       (set! given-cl-contains-eq-flag #t)
                       (set! given-cl-contains-eq-flag #f) ))
            ; given clause is a nucleus
            (begin
              (if (has-single-hyper-premiss? ycl)
                  (begin (set! given-cl-is-final-nuc-flag #t)
                         (set! given-final-nuc-premiss-lit
                               (first-hyper-premiss-lit ycl) )
                         (set! given-final-nuc-complfunpure
                               (logxor *sgn-mask*
                                  (pure-fun
                                    (car given-final-nuc-premiss-lit) )))))
              (if (negative-cl? ycl) (set! given-cl-is-neg-flag #t))
              (set! given-cl-is-nuc-flag #t) )))

    ;(newline)(display "maxvar for the given clause: ")
    ;(display yclmaxvar)

    ; do we have a hyperresolution nucleon with more than one premiss?
    (if (and *hyper-strategy-flag*
             given-cl-is-nuc-flag
             (not given-cl-is-final-nuc-flag) )
        ; ok, we have a hyperresolution nucleon with more than one premiss
        (let ((one-step-results
                (one-step-for-hyper (cons (car given-cl) ycl)) ))
          (do ((results-part one-step-results (cdr results-part)))
              ((or (null? results-part)  *stop-fast-loop-flag* *foundflag*))
            (set! *stop-fast-loop-flag* #f)
            (usable-clauses-fast-loop (car results-part) #f) )))

    ; use given-cl on *usable-clauses-list*??
    (if (or (not *hyper-strategy-flag*)
            (not given-cl-is-nuc-flag)
            given-cl-is-final-nuc-flag)
        ; ok, use given-cl on *usable-clauses-list
        (do ((usable-clauses-list-part
              (cdr *usable-clauses-list*)
              (cdr usable-clauses-list-part) ))
            ((or (null? usable-clauses-list-part)
                 *foundflag*
                 *stop-fast-loop-flag*))
          (let* ((usable-cl (car usable-clauses-list-part))
                 (xcl (cdr usable-cl))
                 (xinf (car usable-cl)) )
            (if (>= *display-info-level* 90)
                (begin
                  (debug-print-xvars 5)
                  (debug-print-yvars 5)
                  (newline)(display "...... usable clause in pure form: ")
                  (newline)
                  (print-pure-clause (cdr usable-cl))
                  (newline)
                  (if *display-debug-info-flag*
                      (begin
                        (display ",,,,,, usable  clause in debug form: ")
                        (newline)
                        (debug-print-clause usable-cl)(newline) ))))
            (if (not (zero? (logand *term-depth-strength-mask*
                                         (car xinf))))
                (if *hyper-strategy-flag*
                    ; hyperresolution
                    (if *neg-hyper-strategy-flag*
                        ; both pos and neg hyperresolution
                        (if (not (and given-cl-is-nuc-flag
                                      (not given-cl-is-neg-flag)
                                      (mixed-cl? xcl) ))
                            (two-clause-binres-loop
                             xcl ycl xinf yinf ylitinf yclmaxvar))
                        ; pure positive hyperresolution
                        (if given-cl-is-nuc-flag
                            ; a fully positive clause required
                            (if (positive-cl? xcl)
                                ; fully positive xcl
                                (if (null? (cdr xcl))
                                    (if (eq? given-final-nuc-complfunpure
                                             (pure-fun (caar xcl)) )
                                        (unit-and-final-nuc-res
                                         xcl ycl (car xcl)
                                         given-final-nuc-premiss-lit
                                         xinf yinf ylitinf yclmaxvar
                                         1 (car (caddr xinf))))
                                    (final-nuc-res
                                     xcl ycl given-final-nuc-premiss-lit
                                     xinf yinf ylitinf
                                     yclmaxvar
                                     given-final-nuc-complfunpure)))
                            ; a nucleus or a paramodulatable clause required
                            (if *type-theory-stratflag*
                                ; a type theory case
                                (if (positive-cl? xcl)
                                    ; a positive unit; may only paramodulate
                                    (if *paramodulation-strategy-flag*
                                        (two-posunits-typet-para
                                         xcl ycl xinf yinf ylitinf yclmaxvar))
                                    ; a nucleon; no paramodulation
                                    (two-clause-binres-loop
                                     xcl ycl xinf yinf ylitinf yclmaxvar) )
                                ; not a type theory case
                                (if given-cl-contains-eq-flag
                                    ; any clause is OK
                                    (two-clause-binres-loop
                                     xcl ycl xinf yinf ylitinf yclmaxvar)
                                    ; clause must contain neg or eq
                                    (if (contains-eqorneg-cl? xcl)
                                        (two-clause-binres-loop
                                           xcl ycl xinf yinf
                                           ylitinf yclmaxvar))))))
                    (if *neg-hyper-strategy-flag*
                        ; pure neg hyperresolution
                        (if (or (and given-cl-is-neg-flag
                                     (not (negative-cl? xcl)) )
                                (and (negative-cl? xcl)
                                     (not given-cl-is-neg-flag) ))
                            (two-clause-binres-loop
                             xcl ycl xinf yinf ylitinf yclmaxvar))
                        ; ordinary binary resolution
                        (if (or (and (not *unit-strategy-flag*)
                                     (not *double-strategy-flag*) )
                                given-is-unit-flag
                                (unit-cl? xcl)
                                (and *double-strategy-flag*
                                     (or given-is-double-flag
                                         (double-cl? xcl) )))

                                ;    (and (> *min-clause-length-in-sos* 1)
                                ;         (or (<= (length ycl)
                                ;                 *min-clause-length-in-sos*)
                                ;             (<= (length xcl)
                                ;                 *min-clause-length-in-sos*) ))
                            (begin
                              (two-clause-binres-loop
                               xcl ycl xinf yinf ylitinf yclmaxvar)
                              ; a type theory case
                              (if (and *type-theory-stratflag*
                                       *paramodulation-strategy-flag*
                                       given-is-unit-flag
                                       (unit-cl? xcl)
                                       (positive-cl? xcl)
                                       (positive-cl? ycl) )
                                  ; positive units; may only paramodulate
                                  (two-posunits-typet-para
                                    xcl ycl xinf yinf
                                    ylitinf yclmaxvar))))))))))))


; two-clause-binres-loop performs all allowed ordinary and
; paramodulation resolution
; steps between two clauses, a given clause (ycl) and another
; clause (xcl).
;
; ycl and xcl are pure clauses.
; ylitinf is assumed to be a list ((<flag1> ...) ... (<flagn> ...))
; where each <flagi> is #f iff literal i is not allowed to be
; resolved upon.

(define (two-clause-binres-loop xcl ycl xinf yinf ylitinf yclmaxvar)

  (if #f ; (> *given-clause-nr* 114)
  (begin
  (newline)(display "two-clause-binres-loop entered")(newline)

  ;(debug-print-xvars 3)(debug-print-yvars 3)
  (debug-print-clause (cons xinf xcl))
  (debug-print-clause (cons yinf ycl))
  (newline)(display xinf)(newline)
  (newline)(display yinf)(newline)
  ))

 (let ((xlitinf (caddr xinf))
       (history '())
       (distance '())
       (xhistsublist '())
       (yhistsublist '())
       (ans-inf #f) )
    (set! *used-demodulators* '())
    (do ((ypart ycl (cdr ypart))
         (ylitinfpart ylitinf (cdr ylitinfpart))
         (ylitnr 1 (+ 1 ylitnr)) )
        ((or (null? ypart)
             *foundflag*
             (and *posweight-order-strategy-flag*
                  (> *clause-keeplength-limit* 0)
                  (> (+ (length ycl) (length xcl))
                     (+ 1 *clause-keeplength-limit*) )))

         ;(newline) (display "two-clause-binres-loop exited")(newline)
         ) ;;; !!!!!!!! TMP DEBUGGING POS OF parenthesis !!!!!!!!!!!

      (let* ((ylit (car ypart))
             (yfun (car ylit))
             (ylitcurinf (car ylitinfpart)) )
        (if (litinf-binrespara? (car ylitcurinf))
          ; ylit allowed to be resolved and/or paramodulated upon or to
          (let ((compl-yfunpure (logxor *sgn-mask* (pure-fun yfun))))
            ; paramodulation stuff
            (if (and *paramodulation-strategy-flag*
                     (pos-eq-pred? (pure-fun yfun))
                     (or (not *para-unit-strategy*)
                         (and (unit-cl? ycl) (unit-cl? xcl) ))
                     (or (not *para-pos-unit-strategy*)
                         (and (unit-cl? ycl) (unit-cl? xcl)
                              (not (neg-fun?(car(first-nonans-lit xcl)))) )))
                (begin
                  ; (set! *param-tolit*)
                  (set! *used-demodulators* '())
                  (set! *param-eqlit* ylit)
                  (set! *param-tocl* xcl)
                  (set! *param-fromcl* ycl)
                  (set! *param-tofullinf* xinf)
                  (set! *param-fromfullinf* yinf)
                  (set! *param-left-arg* (eq-lit-arg1 ylit))
                  (set! *param-right-arg* (eq-lit-arg2 ylit))
                  (set! *param-toinf* xlitinf)
                  (set! *param-frominf* ylitinf)
                  (set! *param-from-is-y-flag* #t)
                  (set! *param-yclmaxvar* yclmaxvar)
                  (set! *param-xclmaxvar* (give-cl-maxvar xinf))
                  (set! *param-fromhist* (car (clinfo-history-sublist yinf)))
                  (set! *param-tohist* (car (clinfo-history-sublist xinf)))
                  (set! *param-fromlitnr* ylitnr)
                  ; (set! *param-tolitnr* xlitnr)
                  (set! *param-use-left-flag*
                        (or *root-para-strat* (litinf-left-para? (car ylitcurinf))))
                  (set! *param-use-right-flag*
                        (or *root-para-strat* (litinf-right-para? (car ylitcurinf)) )))
                (set! *param-eqlit* #f) ) ; flag for no param from ylit

            ;(newline)(display "ylit: ")(newline)
            ;(print-lit ylit)(newline)
            ;(display "litinfo: ")(display (caar ylitinfpart))(newline)
            ;(display "parabits: ")
            ;(display
            ;   (list
            ;      (litinf-left-para?
            ;             (caar ylitinfpart))
            ;      (litinf-right-para?
            ;             (caar ylitinfpart)) ))
            ;(newline)

            (do ((xpart xcl (cdr xpart))
                 (xlitinfpart xlitinf (cdr xlitinfpart))
                 (xlitnr 1 (+ 1 xlitnr)) )
                ((or (null? xpart) *stop-fast-loop-flag* *foundflag*))


              (let* ((xlit (car xpart))
                     (xfun (car xlit)) )
                (if (litinf-binrespara? (caar xlitinfpart))
                  ; xlit allowed to be resolved and/or paramodulated upon/to
                  (begin

                    ;(newline)(display "xlit: ")(newline)
                    ;(print-lit xlit)(newline)
                    ;(display "parabits: ")
                    ;(display
                    ;         (list
                    ;       (litinf-left-para?
                    ;               (caar xlitinfpart))
                    ;       (litinf-right-para?
                    ;               (caar xlitinfpart)) ))
                    ;(newline)

                      ; paramodulation into xlit?
                      (if (and *param-eqlit*
                               (litinf-parainto? (caar xlitinfpart))
                               (or *param-use-left-flag*
                                   *param-use-right-flag*))
                          (begin
                            (set! *param-tolit* xlit)
                            (set! *param-tolitnr* xlitnr)
                            (cond
                              ((not (pos-eq-pred? (pure-fun (car xlit))))
                                 (paramodulate-into-term xlit '() #t) )
                              ((and (litinf-left-para? (caar xlitinfpart))
                                    (not (litinf-right-para? (caar xlitinfpart))) )
                                 (paramodulate-into-term
                                    (cadr xlit) (cons 1 '()) #f) )
                              ((and (litinf-right-para? (caar xlitinfpart))
                                    (not (litinf-left-para? (caar xlitinfpart))) )
                                 (paramodulate-into-term
                                    (caddr xlit) (cons 2 '()) #f) )
                              (else
                                 (paramodulate-into-term xlit '() #t) ))))
                      ; paramodulation into ylit?
                      (if (and *paramodulation-strategy-flag*
                               (pos-eq-pred? (pure-fun xfun))
                               (litinf-parainto? (car ylitcurinf))
                               (or (not *para-unit-strategy*)
                                   (and (unit-cl? ycl) (unit-cl? xcl) ))
                               (or (not *para-pos-unit-strategy*)
                                   (and (unit-cl? ycl) (unit-cl? xcl)
                                        (not (neg-fun? yfun)) ))
                               (or (litinf-left-para? (caar xlitinfpart))
                                   (litinf-right-para? (caar xlitinfpart)) )
                               (not *foundflag*) )
                          (paramodulate-into-ycl ;simpleh
                             ylit xlit ycl xcl yinf xinf ylitinf xlitinf
                             (car xlitinfpart)(give-cl-maxvar xinf)
                             yclmaxvar
                             (car (clinfo-history-sublist yinf))
                             (car (clinfo-history-sublist xinf))
                             ylitnr xlitnr ylitcurinf))
                      ; binres stuff
                      (if (eq? compl-yfunpure (pure-fun xfun))
                          ; complementary predsymbs
                          (if (and
                                   ;(begin
                                   ;(debug-print-xvars 2)(debug-print-yvars 2)
                                   ;  #t)
                                   (resolve-unify-ordinary
                                     xfun yfun xlit ylit)
                                   (not *foundflag*))
                              ; resolution succeeded
                              (let*
                                ((unused (set! *used-demodulators* '()))
                                 (ytmp (make-rest-inf-cl
                                         ycl ylit (caddr yinf)))
                                 (xtmp (make-rest-inf-cl
                                         xcl xlit (caddr xinf)))
                                 (res  (make-raw-inf-cl-pair
                                         (cdr xtmp) (cdr ytmp)
                                         (car xtmp) (car ytmp) )))
                                ; make the conditional answer-inf if needed:
                                (if (and *classical-synth-stratflag*
                                        res
                                        (several-answer-literals? (cdr res)) )
                                    (let*
                                      ((condlit
                                         (if (and *xvars-set-flag*
                                                  (term-may-cont-vars? xlit) )
                                             (subst-deco-term-nodemodulate
                                               xlit #t)
                                             xlit)))
                                      (set! ans-inf
                                            (list condlit 1
                                                (give-cl-maxvar xinf)
                                                yclmaxvar))))
                                    ; the new 1-st level deco clause and
                                ; inflist has been built
                                ; set history
                                (set! xhistsublist (clinfo-history-sublist xinf))
                                (set! yhistsublist (clinfo-history-sublist yinf))
                                (if *hyper-strategy-flag*
                                    (set! history ;simpleh
                                          (make-hyper-history
                                           xcl ycl
                                           xinf yinf
                                           (car xhistsublist)
                                           (car yhistsublist)
                                           (car xlitinfpart) ylitcurinf
                                           xlitnr ylitnr #f))
                                    (set! history ;simpleh
                                          (make-binres-history
                                           xcl ycl
                                           xinf yinf
                                           (car xhistsublist)
                                           (car yhistsublist)
                                           xlitnr ylitnr #f)))
                                ;(debug-print-yvars yclmaxvar)
                                ;(debug-print-xvars (give-cl-maxvar xinf))
                                (if (>= *display-info-level* 70)
                                    (begin
                                      (newline)
                                      (display "raw cl by binary; parents: ")
                                      (newline)
                                      (print-pure-clause xcl)(newline)
                                      (print-pure-clause ycl)(newline)
                                      (print-pure-clause (cdr res))(newline) ))

                                ;clear all vars (subst-deco sets both!)
                                (clear-y-vars-and-touched! yclmaxvar)
                                (clear-x-vars-and-touched!
                                    (give-cl-maxvar xinf) )
                                ; (debug-print-xvars 1)(debug-print-yvars 1)
                                (set! distance
                                  (or (cadr xhistsublist)
                                      (cadr yhistsublist) ))
                                (set! res
                                      (handle-derived-inf-cl-pair
                                         res history #f ans-inf distance) )
                                (set! ans-inf #f) ; clear ans-inf again
                                (if (and res
                                         *factors-stratflag*
                                         (not *foundflag*) )
                                    (factor (car res) (cdr res)) ))
                              ; resolution failed; clear all set vars
                              (begin
                                ;(newline)(display "res failed ")
                                ;(display *xvars-set-flag*) (newline)
                                (if *yvars-set-flag*
                                    (clear-y-vars! yclmaxvar) )
                                (if *xvars-set-flag*
                                    (clear-x-vars!
                                       (give-cl-maxvar xinf) )))))))))))))))


; unit-and-final-nuc-res takes a clause xcl, a final nucleon ycl and its
; single neg lit yneglit and other standard stuff. It does all the
; resolution steps possible with the single premiss and literals in xcl

(define (final-nuc-res xcl ycl yneglit xinf yinf ylitinf yclmaxvar
                       given-final-nuc-complfunpure)

  ; (newline)(display "final-nuc-res started")(newline)

  (do ((xpart xcl (cdr xpart))
       (xlitinfpart (caddr xinf) (cdr xlitinfpart))
       (xlitnr 1 (+ 1 xlitnr)) )
      ((or (null? xpart)  *stop-fast-loop-flag* *foundflag*))
    (if (and (eq? given-final-nuc-complfunpure (pure-fun (caar xpart)))
             (litinf-binrespara? (caar xlitinfpart)) )
        (unit-and-final-nuc-res
          xcl ycl (car xpart) yneglit
          xinf yinf ylitinf yclmaxvar xlitnr (car xlitinfpart)))))

; unit-and-final-nuc-res takes a unit xcl, a final nucleon ycl and its
; single neg lit yneglit and other standard stuff. It does a single
; resolution step if possible.

(define (unit-and-final-nuc-res
           xcl ycl xlit yneglit xinf yinf ylitinf yclmaxvar xlitnr xcurinf)
  (if (resolve-unify-ordinary (car xlit) (car yneglit) xlit yneglit)
      ; resolution succeeded
      (let*
          ((unused (set! *used-demodulators* '()))
           (ytmp (make-rest-inf-cl
                  ycl yneglit (caddr yinf)))
           (xtmp (make-rest-inf-cl
                  xcl xlit (caddr xinf)))
           (res  (make-raw-inf-cl-pair
                  (cdr xtmp) (cdr ytmp)
                  (car xtmp) (car ytmp) ))
           (history '())
           (ycurnr 0)
           (ycurinf '()) )
        ; the new 1-st level deco clause and
        ; inflist has been built
        ; set history
        (do ((ypart ycl (cdr ypart))
             (ylitinfpart ylitinf (cdr ylitinfpart))
             (ylitnr 1 (+ 1 ylitnr)) )
            ((eq? (car ypart) yneglit)
              (set! ycurnr ylitnr)
              (set! ycurinf (car ylitinfpart)) ))
        (set! history ;simpleh
              (make-hyper-history
               xcl ycl
               xinf yinf
               (car(clinfo-history-sublist xinf))
               (car(clinfo-history-sublist yinf))
               xcurinf ycurinf
               xlitnr ycurnr #f))
        ;(debug-print-yvars yclmaxvar)
        ;(debug-print-xvars (give-cl-maxvar xinf))
        (if (>= *display-info-level* 70)
            (begin
              (newline)
              (display "raw cl by binary; parents: ")
              (newline)
              (print-pure-clause xcl)(newline)
              (print-pure-clause ycl)(newline)
              (print-pure-clause (cdr res))(newline) ))

        ;clear all vars (subst-deco sets both!)

        ;(newline)(display yclmaxvar)(display #\space)
        ;(display *max-x-var*)(display *y-var-mask*)
        ;(debug-print-yvars 1)

        (clear-y-vars-and-touched! yclmaxvar)

        ;(newline)(display yclmaxvar)
        ;(debug-print-yvars 1)

        (clear-x-vars-and-touched! (give-cl-maxvar xinf) )

        ;(newline)(display yclmaxvar)
        ;(debug-print-yvars 1)

        (set! res (handle-derived-inf-cl-pair res history #f #f #f))

        ;(newline)(display yclmaxvar)
        ;(debug-print-yvars 2)

        (if (and res *factors-stratflag* (not *foundflag*))
            (factor (car res) (cdr res)) ))
      ; resolution failed; clear all set vars
      (begin

        ;(newline)(display "res failed")(newline)

        (if *yvars-set-flag* (clear-y-vars! yclmaxvar) )
        (if *xvars-set-flag* (clear-x-vars! (give-cl-maxvar xinf))) )))


; two-posunits-typet-para takes two positive unit clauses. If at least
; one of them contains a term lead by eq-fun, it does paramodulation with
; that (these) eq-term.
;
; two-posunits-typet-para is meant for handling the equality Id
; for type theory as built-in.
;
; It is assumed that *paramodulation-strategy-flag* is #t


(define (two-posunits-typet-para xcl ycl xinf yinf ylitinf yclmaxvar)
 ;(newline)(display "two-posunits-typet-para entered")(newline)
  ;     (or (and (pair? (cadar xcl))
  ;                (pos-eq-fun? (car (cadar xcl))) )
  ;           (and (pair? (cadar ycl))
  ;                (pos-eq-fun? (car (cadar ycl))) ))
  (let* ((xlitinf (caddr xinf))
         (history '())
         (ylitnr 1)
         (ylit (car ycl))
         (ylitcurinf (car ylitinf))
         (xlitnr 1)
         (xlit (car xcl))
         (xlitcurinf (car xlitinf)) )
    ;para from ylit
    (if (and (composite? (cadr ylit))
             (pos-eq-fun? (pure-fun (caadr ylit)))
             (litinf-parainto? (car xlitcurinf))
             (or (litinf-left-para? (car ylitcurinf))
                 (litinf-right-para? (car ylitcurinf)) ))
        ; ylit allowed to be paramodulated from
        (begin
          (set! *param-tolit* xlit)
          (set! *param-eqlit* ylit)
          (set! *param-tocl* xcl)
          (set! *param-fromcl* ycl)
          (set! *param-left-arg* (eq-lit-arg1 (cadr ylit)))
          (set! *param-right-arg* (eq-lit-arg2 (cadr ylit)))
          (set! *param-toinf* xlitinf)
          (set! *param-frominf* ylitinf)
          (set! *param-from-is-y-flag* #t)
          (set! *param-yclmaxvar* yclmaxvar)
          (set! *param-xclmaxvar* (give-cl-maxvar xinf))
          (set! *param-fromhist* (car (clinfo-history-sublist yinf)))
          (set! *param-tohist* (car (clinfo-history-sublist xinf)))
          (set! *param-fromlitnr* ylitnr)
          (set! *param-tolitnr* xlitnr)
          (set! *param-use-left-flag*
                (litinf-left-para? (car ylitcurinf)))
          (set! *param-use-right-flag*
                (litinf-right-para? (car ylitcurinf)) )
              ;(newline)(display "ylit: ")(newline)
          ;(print-lit ylit)
          (paramodulate-into-term xlit '() #t) ))
    ; para from xlit
    (if (and (composite? (cadr xlit))
             (pos-eq-fun? (pure-fun (caadr xlit)))
             (litinf-parainto? (car ylitcurinf))
             (or (litinf-left-para? (car xlitcurinf))
                 (litinf-right-para? (car xlitcurinf)) ))
        ; ylit allowed to be paramodulated from
        (begin
          (set! *param-tolit* ylit)
          (set! *param-eqlit* xlit)
          (set! *param-tocl* ycl)
          (set! *param-fromcl* xcl)
          (set! *param-left-arg* (eq-lit-arg1 (cadr xlit)))
          (set! *param-right-arg* (eq-lit-arg2 (cadr xlit)))
          (set! *param-toinf* ylitinf)
          (set! *param-frominf* xlitinf)
          (set! *param-from-is-y-flag* #f)
          (set! *param-yclmaxvar* yclmaxvar)
          (set! *param-xclmaxvar* (give-cl-maxvar xinf))
          (set! *param-fromhist* (car (clinfo-history-sublist xinf)))
          (set! *param-tohist* (car (clinfo-history-sublist yinf)))
          (set! *param-fromlitnr* xlitnr)
          (set! *param-tolitnr* ylitnr)
          (set! *param-use-left-flag*
                (litinf-left-para? (car xlitcurinf)))
          (set! *param-use-right-flag*
                (litinf-right-para? (car xlitcurinf)) )
          (paramodulate-into-term ylit '() #t) ))))


; one-step-for-hyper takes a nucleon ycl and finds the
; premiss with the smallest number of unifiable (and giving
; non-removed results) electrons. The list of new nucleons
; obtained by cutting off at this premiss is returned.
;
; NB! in case unit-handling produces somewhere a non-nucleon,
; this is kept in the standard way and not included in the
; returned list.
;
; *unit-clash-flag* is #t iff immediate contradiction with units
; (any units - positive and negative) is looked for.

(define (one-step-for-hyper fullycl)
  ;(newline)
  ;(display "one-step-for-hyper called on fullycl: ")
  ;(print-clause fullycl)
  ;(newline)
  (set! *one-step-hyper-phase* #t)
  (let* ((min-nr 999999)
         (results '())
         (locresults '())
         (ylit '())
         (negfun '())
         (ycl (cdr fullycl))
         (yinf (car fullycl))
         (fullxcl '())
         (xcl '())
         (xinf '())
         (ylitinf (caddr yinf))
         (yclmaxvar (give-cl-maxvar yinf))
         (yfun '())
         (ynegpurefun '())
         (res '())
         (tmp '())
         (success-nr 0) )
    (set! *yvars-set-flag* #f)
    (set! *xvars-set-flag* #f)
    ; cycle over literals in the nucleon
    (do ((ypart ycl (cdr ypart))
         (ylitinfpart (caddr yinf) (cdr ylitinfpart))
         (ylitnr 1 (+ 1 ylitnr)) )
        ((or (null? ypart) *foundflag* (zero? min-nr))
           (set! *one-step-hyper-phase* #f)

           ; (display #\space)(display min-nr)(display #\space)

           results)
      (if (or *unit-clash-flag*
              (neg-fun? (caar ypart)) )
          (begin
            (set! success-nr 0)
            (set! ylit (car ypart))
            (set! yfun (car ylit))
            (set! ynegpurefun (logxor *sgn-mask* (pure-fun yfun)))
            ; cycle over all the clauses in the usable-list
            ; (for the unit clash check - all units)
            (do ((usable-list-part
                    (if *unit-clash-flag*
                        (begin
                          (set! tmp (assq ynegpurefun *unit-alist*))
                          (if tmp (cdr tmp) '()) )
                        (cdr *usable-clauses-list*) )
                    (cdr usable-list-part)))
                ((or (null? usable-list-part)
                     *foundflag*
                     *stop-fast-loop-flag*
                     (>= success-nr min-nr) ))
              (set! fullxcl (car usable-list-part))
              (set! xinf (car fullxcl))
              (set! xcl (cdr fullxcl))
              (if (and (not (zero? (logand *term-depth-mask* (car xinf))))
                       (or *unit-clash-flag*
                           (positive-cl? xcl) ))
                  ; cycle over literals in xcl
                  (begin
                    (do ((xclpart xcl (cdr xclpart))
                         (xlitinfpart (caddr xinf) (cdr xlitinfpart))
                         (xlitnr 1 (+ 1 xlitnr)) )
                        ((or (null? xclpart) *foundflag*))
                      (if (and (eq? (pure-fun (caar xclpart)) ynegpurefun)
                               (litinf-binrespara? (caar xlitinfpart))
                               (resolve-unify-ordinary
                                (caar xclpart) yfun (car xclpart) ylit))
                          ; resolution succeeded
                          (let* ((unused (set! *used-demodulators* '()))
                                 (ytmp (make-rest-inf-cl
                                        ycl ylit (caddr yinf)))
                                 (xtmp (make-rest-inf-cl
                                        xcl (car xclpart)
                                        (caddr xinf) ))
                                 (res  (make-raw-inf-cl-pair
                                        (cdr xtmp) (cdr ytmp)
                                        (car xtmp) (car ytmp) ))
                                 (history '()) )
                            ; the new 1-st level deco clause and
                            ; inflist has been built
                            ; set history
                            (set! history
                                  (if *unit-clash-flag*
                                      (make-binres-history ;simpleh
                                       xcl ycl
                                       xinf yinf
                                       (car(clinfo-history-sublist xinf))
                                       (car(clinfo-history-sublist yinf))
                                       xlitnr ylitnr #f)
                                      (make-hyper-history ;simpleh
                                       xcl ycl
                                       xinf yinf
                                       (car(clinfo-history-sublist xinf))
                                       (car(clinfo-history-sublist yinf))
                                       (car xlitinfpart) (car ylitinfpart)
                                       xlitnr ylitnr #f)))
                            ;(debug-print-yvars yclmaxvar)
                            ;(debug-print-xvars (give-cl-maxvar xinf))
                            (if (>= *display-info-level* 110)
                                (begin
                                  (newline)
                                  (display "vars after one-step res:")
                                  (newline)
                                  (debug-print-yvars yclmaxvar)
                                  (debug-print-xvars (give-cl-maxvar xinf))
                                  (newline)
                                  (display "raw prepcl by hyper: parents: ")
                                  (newline)
                                  (print-pure-clause xcl)(newline)
                                  (print-pure-clause ycl)(newline)
                                  (print-pure-clause (cdr res))(newline) ))
                            (clear-y-vars-and-touched! yclmaxvar)
                            (clear-x-vars-and-touched! (give-cl-maxvar xinf) )
                            (set! res (handle-derived-inf-cl-pair
                                        res history #f #f #f))
                            ; in case of success keep the resulting nucleon:
                            (if res
                                (begin
                                   (set! success-nr (+ 1 success-nr))
                                   (set! locresults (cons res locresults)) )))
                          ; resolution failed
                          (begin
                            ;(newline)(display "res failed")(newline)
                            (if *yvars-set-flag* (clear-y-vars! yclmaxvar) )
                            (if *xvars-set-flag*
                                (clear-x-vars!
                                  (give-cl-maxvar xinf) ))))))))
            ; all the clauses have been cycled over:
            ; now, if better than the best earlier, keep it:
            (if (< success-nr min-nr)
                (begin (set! min-nr success-nr)
                       (set! results locresults) )))))))


; the number of the literal in nucleus is not relevant: the
; one in electron is.

(define (make-hyper-history
          xcl ycl xinf yinf xhist yhist xlitcurinf ylitcurinf xlitnr ylitnr chain)
  (let* ((x '())
         (res '())
         (tmp  (make-hyper-history-aux
               xcl ycl xinf yinf xhist yhist xlitcurinf ylitcurinf xlitnr ylitnr)))
    ;(display "hyper-history-aux made ok")(display (cadr tmp))
    ;(if (and (eq? (car tmp) 'hyper)
    ;         (or (eq? (cadr tmp) 1569)
    ;                  (eq? (cadr tmp) 1207)
    ;                  (eq? (cadr tmp) 1275) ))
    ;    (begin
    ;       (display "1569 hyper-history: ")(newline)
    ;       (pretty-print tmp)(newline)  ))
    (set! chain #f)
    (if chain
        (set! tmp (append (cons 'chain (cdr chain)) tmp)) )
    (if (not (null? *used-demodulators*))
        (begin
          ;(display *used-demodulators*)(newline)
          (if *simple-history-stratflag*
             (set! tmp (append (make-demod-history *used-demodulators*) tmp))
             (begin
              (set! x (my-last-pair tmp))
              (set-cdr! x (list (make-demod-history *used-demodulators*))) ))
          (set! res tmp) )
        (set! res tmp) )
    res))

(define (make-hyper-history-aux
          xcl ycl xinf yinf xhist yhist xlitcurinf ylitcurinf xlitnr ylitnr)

  ;(newline)(display "starting to make hyper-history-aux")(newline)
  ;
  ;(display "xcl : ")(display xcl)(newline)
  ;(display "ycl : ")(display ycl)(newline)
  ;(display "xinf : ")(display xinf)(newline)
  ;(display "yinf : ")(display yinf)(newline)
  ;(display "xhist : ")(display xhist)(newline)
  ;(display "yhist : ")(display yhist)(newline)

  (if *simple-history-stratflag*
      (cond
        ((and (not (positive-cl? ycl)) (memq 'hyper yhist))
           ; ycl is a nucleus which has been resolved upon already
           ; we will use its history to build a new one
           (append yhist (list (clinfo-clnr xinf))) )
        ((and (not (positive-cl? xcl)) (memq 'hyper xhist))
           ; xcl is a nucleus which has been resolved upon already
           ; we will use its history to build a new one
           (append xhist (list (clinfo-clnr yinf))) )
        (else
           ; a fresh nucleus has been used
           (list 'hyper
                 (if (positive-cl? xcl) (clinfo-clnr yinf) (clinfo-clnr xinf))
                 (if (positive-cl? xcl) (clinfo-clnr xinf) (clinfo-clnr yinf)) )))
      (make-hyper-history-aux-complex
            xcl ycl xinf yinf xhist yhist xlitcurinf ylitcurinf xlitnr ylitnr)))


(define (make-hyper-history-aux-complex
          xcl ycl xinf yinf xhist yhist xlitcurinf ylitcurinf xlitnr ylitnr)
  (if (positive-cl? xcl)
      ; ycl is a nucleus
      (if (or (null? xhist)(null? yhist)
              (eq? 'binary (car yhist))
              (eq? 'binres (car yhist))
              (eq? 'binres_demod (car yhist))
              (eq? 'para (car yhist))
              (eq? 'factor (car yhist)) )
              ; (not (eq? 'hyper (car yhist)))
       (make-binres-history
             xcl ycl
             xinf yinf
             xhist
             yhist
             xlitnr ylitnr #f)
       (begin
        ;(newline)(display "starting an ycl-nucleus case")(newline)
        (if (zero? (get-litinfo-ordernr (car ylitcurinf)))
            ; no real ordernr
            (if (fresh-nucleus-hist? yhist)
                (list 'hyper 0 yhist (list 'literal xhist xlitnr))
                (begin
                  (set-cdr! (cddr yhist)
                            (cons (list 'literal xhist xlitnr)(cdddr yhist)) )
                  yhist))
            ; real ordernr
           (begin
            (replace-nth
             (if (fresh-nucleus-hist? yhist)
                 (cons 'hyper (cons 0 (cons yhist (make-null-neg-list ycl))))
                 yhist)
             (+ 3 (get-litinfo-ordernr (car ylitcurinf)))
             (list 'literal xhist xlitnr) ))))               )
      ; xcl is a nucleus
      (if (or (null? xhist)(null? yhist)
              (eq? 'binary (car xhist))
              (eq? 'binres (car xhist))
              (eq? 'binres_demod (car xhist))
              (eq? 'para (car xhist))
              (eq? 'factor (car xhist)) )
              ; (not (eq? 'hyper (car yhist)))
       (make-binres-history
             xcl ycl
             xinf yinf
             xhist
             yhist
             xlitnr ylitnr #f)
       (begin
        ;(newline)(display "starting an xcl-nucleus case")(newline)
        (if (zero? (get-litinfo-ordernr (car xlitcurinf)))
            ; no real ordernr
            (if (fresh-nucleus-hist? xhist)
                (list 'hyper 0 xhist (list 'literal yhist ylitnr))
                (begin
                  (set-cdr! (cddr xhist)
                            (cons (list 'literal yhist ylitnr)(cdddr xhist)) )
                  yhist))
            ; real ordernr
           (begin
            (replace-nth
             (if (fresh-nucleus-hist? xhist)
                 (cons 'hyper (cons 0 (cons xhist (make-null-neg-list xcl))))
                xhist)
             (+ 3 (get-litinfo-ordernr (car xlitcurinf)))
             (list 'literal yhist ylitnr) ))))))   )


(define (make-binres-history xcl ycl xinf yinf xhist yhist xlitnr ylitnr chain)
 (let ((res '()))
  ;(newline)(display "starting to make binres history:")(newline)
  ;(newline)(display "xcl: ")(display xcl)(newline)
  ;(newline)(display "ycl: ")(display ycl)(newline)
  ;(newline)(display "xinf: ")(display xinf)(newline)
  ;(newline)(display "yinf: ")(display yinf)(newline)

  (if *simple-history-stratflag*
      (begin
       (set! res
            (list 'binary
                  (list (clinfo-clnr xinf) xlitnr)
                  (list (clinfo-clnr yinf) ylitnr) ))
       ;(newline)(display "res: ")(display res)(newline)
       (if (not (null? *used-demodulators*))
           (set! res (append
                       (make-demod-history *used-demodulators*)
                       res)))
       ;(newline)(display "res: ")(display res)(newline)
       res)
      (if (null? *used-demodulators*)
          (list 'binres 0
                (list 'literal xhist xlitnr)
                (list 'literal yhist ylitnr) )
          (list  'binres 0
                (list 'literal xhist xlitnr)
                (list 'literal yhist ylitnr)
                (make-demod-history *used-demodulators*) )))))

(define (make-demod-history demodlst)
  (if *simple-history-stratflag*
      (cons 'demod (map (lambda(x)(clinfo-clnr (car x))) demodlst))
      (cons 'demod demodlst) ))

(define (make-chain-history-for-resolution xcl ycl xlitnr ylitnr)
   (let ((xlit (car (nth-cdr (- xlitnr 1) xcl)))
         (ylit (car (nth-cdr (- ylitnr 1) ycl))) )
     (if (eq? (pure-fun (car xlit))
              (logxor *sgn-mask* (pure-fun (car ylit))) )
          '()
          '() )))

(define (fresh-nucleus-hist? hist)
  (and (pair? hist)
       (eq? 'input (car hist)) ))

(define (make-null-neg-list nuc)
  (let ((res '()))
    (do ((nucpart nuc (cdr nucpart)))
        ((null? nucpart) res)
      (if (neg-fun? (caar nucpart)) (set! res (cons '() res))) )))


; paramodulate-into-ycl paramodulates from xlit into ylit.
;
; It takes:
;   tolit; a literal to be paramodulated into
;   eqlit : an equality atom
;   use-left-flag : if #t, paramodulate also from left arg of eqlit.
;   use-right-flag : if #t, paramodulate also from right arg of eqlit.
;   fromcl : whole pure fromcl
;   tocl: whole pure tocl
;   frominf: fromcl inf
;   curfrominf : inf of the eqlit literal
;   toinf: tocl inf
;   from-is-y-flag : #t iff the the fromcl is y-clause
;
; toinf, frominf are assumed to be a list ((<flag1> ...) ... (<flagn> ...))
; where each <flagi> is #f iff literal i is not allowed to be
; resolved upon.

(define (paramodulate-into-ycl
          tolit eqlit tocl fromcl tofullinf fromfullinf toinf frominf curfrominf
          xclmaxvar yclmaxvar yhist xhist ylitnr xlitnr ylitinf)

  ;(newline)(display "paramodulation-into-ycl entered")(newline)

  (let ((tmp-tolit *param-tolit*)
        (tmp-eqlit *param-eqlit*)
        (tmp-tocl *param-tocl*)
        (tmp-fromcl *param-fromcl*)
        (tmp-tofullinf *param-tofullinf*)
        (tmp-fromfullinf *param-fromfullinf*)
        (tmp-left-arg *param-left-arg*)
        (tmp-right-arg *param-right-arg*)
        (tmp-use-left-flag *param-use-left-flag*)
        (tmp-use-right-flag *param-use-right-flag*)
        (tmp-toinf *param-toinf*)
        (tmp-frominf *param-frominf*)
        (tmp-from-is-y-flag *param-from-is-y-flag*)
        (tmp-fromhist *param-fromhist*)
        (tmp-tohist *param-tohist*)
        (tmp-fromlitnr *param-fromlitnr*)
        (tmp-tolitnr *param-tolitnr*) )

  (set! *used-demodulators* '())
  (set! *param-tolit* tolit)
  (set! *param-eqlit* eqlit)
  (set! *param-tocl* tocl)
  (set! *param-fromcl* fromcl)
  (set! *param-tofullinf* tofullinf)
  (set! *param-fromfullinf* fromfullinf)
  (set! *param-left-arg* (eq-lit-arg1 eqlit))
  (set! *param-right-arg* (eq-lit-arg2 eqlit))
  (set! *param-use-left-flag* (litinf-left-para? (car curfrominf)))
  (set! *param-use-right-flag* (litinf-right-para? (car curfrominf)))
  (set! *param-toinf* toinf)
  (set! *param-frominf* frominf)
  (set! *param-xclmaxvar* xclmaxvar)
  (set! *param-yclmaxvar* yclmaxvar)
  (set! *param-from-is-y-flag* #f)
  (set! *param-fromhist* xhist)
  (set! *param-tohist* yhist)
  (set! *param-fromlitnr* xlitnr)
  (set! *param-tolitnr* ylitnr)

  (cond
    ((not (pos-eq-pred? (pure-fun (car tolit))))
      (paramodulate-into-term tolit '() #t) )
    ((and (litinf-left-para? (car ylitinf))
          (not (litinf-right-para? (car ylitinf))) )
      (paramodulate-into-term
          (cadr tolit) (cons 1 '()) #f) )
    ((and (litinf-right-para? (car ylitinf))
          (not (litinf-left-para? (car ylitinf))) )
      (paramodulate-into-term
          (caddr tolit) (cons 2 '()) #f) )
    (else
      (paramodulate-into-term tolit '() #t) ))

  ;(newline)(display "paramodulation-into-ycl exited")(newline)

  (set! *param-tolit* tmp-tolit)
  (set! *param-eqlit* tmp-eqlit)
  (set! *param-tocl* tmp-tocl)
  (set! *param-fromcl* tmp-fromcl)
  (set! *param-left-arg* tmp-left-arg)
  (set! *param-right-arg* tmp-right-arg)
  (set! *param-use-left-flag* tmp-use-left-flag)
  (set! *param-use-right-flag* tmp-use-right-flag)
  (set! *param-toinf* tmp-toinf)
  (set! *param-frominf* tmp-frominf)
  (set! *param-from-is-y-flag* tmp-from-is-y-flag)
  (set! *param-fromhist* tmp-fromhist)
  (set! *param-tohist* tmp-tohist)
  (set! *param-fromlitnr* tmp-fromlitnr)
  (set! *param-tolitnr* tmp-tolitnr) ))

(define (paramodulate-into-term term path litflag)

  ;(newline) (display "paramodulate-into-term called:")(newline)
  ;(display term)(newline)
  ;(debug-print-term term)(newline)
  ;(debug-print-term *param-left-arg*)(newline)
  ;(debug-print-term *param-right-arg*)(newline)
  ;(display path)
  ;(debug-print-xvars 6)
  ;(debug-print-yvars 6)
  ;(if litflag (begin (newline)(debug-print-xvars 6)(newline)))

  (cond
   ((composite? term)
      ; do the (left arg is to be unified) part
      (if *param-use-left-flag*
          (cond
            (litflag) ; the term is actually a literal, do nothing
            ((composite? *param-left-arg*)
               (if (eq? (pure-fun (car term))
                        (pure-fun (car *param-left-arg*)) )
                   (begin
                     (set! *compare-ground-litcdr-equal-flag* #t)
                     ;(display  (car term) ) (newline)
                     ;(display (car *param-left-arg*)) (newline)
                     ;(display term)(newline)
                     ;(display *param-left-arg*)(newline)


                     (if (if *param-from-is-y-flag*
                             (resolve-unify-ordinary
                               (car term) (car *param-left-arg*)
                               term *param-left-arg*)
                             (resolve-unify-ordinary
                               (car *param-left-arg*) (car term)
                               *param-left-arg* term))
                         (paramodulate-into-term-aux
                            path *param-right-arg* #t)
                         (begin
                           ;clear all vars
                           (set! *compare-ground-litcdr-equal-flag* #f)
                           (if *yvars-set-flag*
                               (clear-y-vars-and-touched! *param-yclmaxvar*) )
                           (if *xvars-set-flag*
                               (clear-x-vars-and-touched!
                                 *param-xclmaxvar*)))))))
            ((and *paramodulate-from-vars-stratflag*
                  (non-composite:-var? *param-left-arg*) )
               ; just set the var
               (set-var-ref! (var-nr *param-left-arg*) term)
               (if (x-var? *param-left-arg*)
                 (begin (set! *xvars-set-flag* #t)(set! *yvars-set-flag* #f))
                 (begin (set! *xvars-set-flag* #f)(set! *yvars-set-flag* #t)))
               (paramodulate-into-term-aux path *param-right-arg* #t) )))
       ; do the (right arg is to be unified) part
       (if *param-use-right-flag*
          (cond
            (litflag) ; the term is actually a literal, do nothing
            (*foundflag*)
            ((composite? *param-right-arg*)
               (if (eq? (pure-fun (car term))
                        (pure-fun (car *param-right-arg*)))
                   (begin
                     (set! *compare-ground-litcdr-equal-flag* #t)
                     (if (if *param-from-is-y-flag*
                             (resolve-unify-ordinary
                               (car term) (car *param-right-arg*)
                               term *param-right-arg*)
                             (resolve-unify-ordinary
                               (car *param-right-arg*) (car term)
                               *param-right-arg* term))
                         (paramodulate-into-term-aux
                            path *param-left-arg* #f)
                         (begin
                           ;clear all vars
                           (set! *compare-ground-litcdr-equal-flag* #f)
                           (if *yvars-set-flag*
                               (clear-y-vars-and-touched! *param-yclmaxvar*) )
                           (if *xvars-set-flag*
                               (clear-x-vars-and-touched!
                                 *param-xclmaxvar*) ))))))
            ((and *paramodulate-from-vars-stratflag*
                  (non-composite:-var? *param-right-arg*) )
               ; just set the var
               (set-var-ref! (var-nr *param-right-arg*) term)
               (if (x-var? *param-right-arg*)
                 (begin (set! *xvars-set-flag* #t)(set! *yvars-set-flag* #f))
                 (begin (set! *xvars-set-flag* #f)(set! *yvars-set-flag* #t)))
               (paramodulate-into-term-aux path *param-left-arg* #f) )))
       ; now go recursively into the term
       (if (not *foundflag*) ; (or (null? path) (null? (cdr path)))
           (do ((termpart (cdr term) (cdr termpart))
                (nr 1 (+ 1 nr)) )
               ((null? termpart))
            (paramodulate-into-term
              (car termpart) (cons nr path) #f) )))
   ; a noncomposite term
   ((non-composite:-var? term)) ; var case always fails
   ; a constant which is a curry-constant will fails
   ((>= term *mincurryconst*))
   ; term is a constant
   (else
       ; do the (left arg to be unified) part
       (if *param-use-left-flag*
           (cond
             ((and *paramodulate-from-vars-stratflag*
                   (var? *param-left-arg*))
              ; just set the var
              (set-var-ref! (var-nr *param-left-arg*) term)
              (if (x-var? *param-left-arg*)
                 (begin (set! *xvars-set-flag* #t)(set! *yvars-set-flag* #f))
                 (begin (set! *xvars-set-flag* #f)(set! *yvars-set-flag* #t)))
               (paramodulate-into-term-aux path *param-right-arg* #t) )
             ((eq? *param-left-arg* term)
               (set! *xvars-set-flag* #f)(set! *yvars-set-flag* #f)
               (paramodulate-into-term-aux path *param-right-arg* #t) )))
        ; do the (right arg to be unified) part
        (if *param-use-right-flag*
           (cond
             (*foundflag*)
             ((and *paramodulate-from-vars-stratflag*
                   (var? *param-right-arg*) )
              ; just set the var
               (set-var-ref! (var-nr *param-right-arg*) term)
               (if (x-var? *param-right-arg*)
                 (begin (set! *xvars-set-flag* #t)(set! *yvars-set-flag* #f))
                 (begin (set! *xvars-set-flag* #f)(set! *yvars-set-flag* #t)))
               (paramodulate-into-term-aux path *param-left-arg* #f) )
             ((eq? *param-right-arg* term)
               (set! *xvars-set-flag* #f)(set! *yvars-set-flag* #f)
               (paramodulate-into-term-aux path *param-left-arg* #f) ))))))



; paramodulate-into-term-aux is called in case of success to build,
; manipulate and store a new clause and to clear vars.

(define (paramodulate-into-term-aux path replterm left-used-flag)

  ;(newline)(display "paramodulate-into-term-aux called:")(newline)
  ;(display path)(newline)
  ;(display replterm)(newline)
  ;(display left-used-flag)(newline)
  ;(debug-print-term replterm)
  ;(display *xvars-set-flag*)(newline)(print *yvars-set-flag*)(newline)
  ; (begin (newline)(debug-print-xvars 6)(newline)(debug-print-yvars 6)(newline))

  (set! *compare-ground-litcdr-equal-flag* #f)
  (set! *used-demodulators* '())
  (let* ((newlit (subst-deco-replace-term
                    *param-tolit* (reverse path) replterm #t))
         (tmpto (make-rest-inf-cl
                   *param-tocl* *param-tolit* *param-toinf*))
         (tmpfrom (make-rest-inf-cl
                   *param-fromcl* *param-eqlit* *param-frominf*))
         (res '())
         (history '())
         (distance '())
         (ans-inf #f) )

   ;(newline)(display "subst-deco-replacement in param-aux finished:")
   ;(newline)
  ;(display *xvars-set-flag*)(newline)(print *yvars-set-flag*)(newline)
  ;(begin (newline)(debug-print-xvars 6)(newline)
        ; (debug-print-yvars 6)(newline) (debug-print-zvars 6)(newline))


   (if (not newlit)
     ; newlit not allowed
     #f
     ; successfully built newlit, not immediately restricted
     (begin
       (if *param-from-is-y-flag*
           (set! res
                 (make-raw-inf-cl-pair
                  (cdr tmpto) (cdr tmpfrom) (car tmpto) (car tmpfrom)))
           (set! res
                 (make-raw-inf-cl-pair
                  (cdr tmpfrom) (cdr tmpto) (car tmpfrom) (car tmpto))))
       ; make the conditional answer-inf if needed:
       (if (and *classical-synth-stratflag*
                res
                (several-answer-literals? (cdr res)) )
           (let*
               ((condlit
                 (if (or (and *param-from-is-y-flag* *xvars-set-flag*
                              (term-may-cont-vars? *param-tolit*) )
                         (and (not *param-from-is-y-flag*) *yvars-set-flag*
                              (term-may-cont-vars? *param-tolit*) ))
                     (subst-deco-term-nodemodulate *param-tolit* #t)
                     *param-tolit*)))
             (if *param-from-is-y-flag* ; to is x?
                 (set! ans-inf ; to is x
                       (list condlit 1 *param-xclmaxvar* *param-yclmaxvar*) )
                 (set! ans-inf ; to is y
                     (list condlit 1 *param-xclmaxvar* *param-yclmaxvar*) ))))
       (clear-y-vars-and-touched! *param-yclmaxvar*)
       (clear-x-vars-and-touched! *param-xclmaxvar*)
       (if (not res)
           #f
           (begin
             ; add a new inf of tolit and a ready tolit to res
             ; a substituted-into resulting lit has no restrictions:
             (set! res (cons (cons (list #t #f 0) ;;; !!!!!!! NB!!!!! new inf
                                   (car res) )
                             (cons newlit (cdr res)) ))
             (if *simple-history-stratflag*
                 (begin
                    ;(newline)(display "about to make para history")(newline)
                    ;(display *param-fromfullinf*)(newline)
                    ;(display *param-fromlitnr*)(newline)
                    ;(display *param-tofullinf*)(newline)
                    ;(display *param-tolitnr*)(newline)
                    ;(display left-used-flag)(newline)
                    ;(display path)(newline)
                    ;(display *used-demodulators*)(newline)
                    (set! history (list 'para
                                        (list (clinfo-clnr *param-fromfullinf*)
                                              *param-fromlitnr*)
                                        (list (clinfo-clnr *param-tofullinf*)
                                              *param-tolitnr*
                                              left-used-flag
                                              path)))
                    ;(newline)(display "history: ")(display history)(newline)
                    (if (not (null? *used-demodulators*))
                        (set! history (append (make-demod-history *used-demodulators*)
                                              history)) ))
                (if (null? *used-demodulators*)
                    (set! history
                       (list 'para 0
                             *param-fromhist* *param-tohist*
                             *param-fromlitnr* *param-tolitnr*
                             left-used-flag (reverse path)))
                    (set! history
                       (list 'para 0
                             *param-fromhist* *param-tohist*
                             *param-fromlitnr* *param-tolitnr*
                             left-used-flag (reverse path)
                             (cons 'demod *used-demodulators*) ))))

             ; the new 1-st level deco clause and
             ; inflist has been succesfully built

             ;(debug-print-yvars *param-yclmaxvar*)
             ;(debug-print-xvars *param-xclmaxvar*)
             ;(newline)
             (if (>= *display-info-level* 70)
                 (begin
                   (display "raw clause derived by param; ")
                   (display "to and from parents, fromlit:")
                   (newline)
                   (print-pure-clause *param-tocl*)(newline)
                   (print-pure-clause *param-fromcl*)(newline)
                   (display *param-tocl*)(newline)
                   (display *param-fromcl*)(newline)
                   (if *type-theory-stratflag*
                       (print-lit *param-eqlit*)
                       (print-lit *param-eqlit*) )
                   (newline)
                   (print-pure-clause (cdr res))(newline) ))
             (set! distance
               (or (clinfo-distance *param-tofullinf*)
                   (clinfo-distance *param-fromfullinf*) ))
             (set! res (handle-derived-inf-cl-pair
                            res history #f ans-inf distance))

             ;(newline)(display "paramodulate-into-term is to finish:")
             ;(newline)
             ;(display *param-yclmaxvar*)(newline)
             ;(display *param-xclmaxvar*)(newline)
             ;(debug-print-yvars 6)
             ;(debug-print-xvars 6)
             ;(newline)

             (if (and res *factors-stratflag* (not *foundflag*))
                 (factor (car res) (cdr res)) )))))))


; replace-subterm takes a term, a path in the term and a new subterm
; to be put to the place indicated by the path.
; A path is a list of nrs, (n1, ... nk) where each ni indicates the
; following cxr operation: ca(d1)...(dni)r)
;
; It builds a new term with newsub put to the right place.
;
;(define (replace-subterm term path newsub)
;  (if (null? path)
;      newsub
;      (let* ((res (cons '() '()))
;             (resend res) )
;        (do ((termpart term (cdr termpart))
;             (nrpart (car path) (- nrpart 1)) )
;            ((zero? nrpart)
;               (set-cdr! resend
;                 (cons (replace-subterm (car termpart) (cdr path) newsub)
;                       (cdr termpart) ))
;               (cdr res) )
;          (set-cdr! resend (cons (car termpart) '()))
;          (set! resend (cdr resend)) ))))


(define (give-cl-maxvar clinf)
  (let ((clmaxvar (shiftright (logand *cl-maxvar-mask* (car clinf))
                              *cl-maxvar-shift*)))
    (if (eq? clmaxvar *cl-maxvar*)
        *max-nr-of-vars-in-clause*
        clmaxvar)))


; mattias kirjutas jalle siia  moned sonad


; fac-gen-flag is #t iff pr is factor-generated
;
; ans-inf is #f in case no unique answer-generation necessary.
; else it has the following structure:
;
; (<cut-literal-under-substitution> <nr (1 or 2) of cutliteral clause>
;  <max-x-varnr> <max-y-varnr>)

(define (handle-derived-inf-cl-pair
             pr history fac-gen-flag ans-inf distance)

  (if (> *display-info-level* 100)
      (begin
  (newline)(display "handle-derived-inf-cl-pair entered with:")
  (display (car pr))
  (newline)
  (debug-print-pure-clause (cdr pr))
  (newline)
  (display "fac-gen-flag: ")(display fac-gen-flag)
  (newline)
  (display "distance: ")(display distance)
    ))

  ;(set! *used-demodulators* '())

  ; unique-answer producing section;
  ; make-unique-answer may generate a new clause
  ; (by calling handle-derived-inf-cl-pair on its own)
  (set! *derived-clauses-nr* (+ 1 *derived-clauses-nr*))
  (if ans-inf ; now we have a double answer-lit in cl
      (set! pr (make-unique-answer-pair pr history ans-inf)) )
  (let ((cl (cdr pr))
        (inf (car pr))
        (clauseinf '())
        (tmp-fullcl '()) )

    ;(newline)(display "cl part of pair: ")(newline)
    ;(print-pure-clause cl)(newline)
    ;(display "inf part of pair: ")(newline)
    ;(display inf)(newline)

    ; the standard handling continues
    (if (or (not cl)
            (and *classical-synth-stratflag*
                 (not ans-inf) ; if this is given, make-unique.. has taken care
                 (some? (lambda(x)
                          (and (answer-lit? x)(not (ans-computable-lit? x))) )
                        cl)))
      ; first-level restricted clause
      #f
      ; the clause was not restricted by 1-st level crit
      (if (empty-cl? cl)
        ; empty clause derived
        (begin
          (if (>= *display-info-level* 20)
              (begin
                (newline)(display "********* EMPTY CLAUSE DERIVED *********")
                (newline) ))
          (set! *answer-cl* cl)
          (set! *answer-history* history)
          (set! *foundflag* #t) )
        ; normal raw clause derived
        (let* ((tmp-pr '())
               (tmp-triple '()) )
          (set! tmp-pr (remove-double-lits!
                          inf cl *hyper-strategy-flag*))
          (if (or (removable-tautology-pr? tmp-pr)
                  (and *int-logic-flag* (int-prohibited-pr? tmp-pr)) )
              ; a removable tautology or a bad int clause
              #f
              ; the clause is not a tautology
              (begin
                ; a check by various user-supplied criteria
                (if (or (and *user-criterias-post-flag*
                            (not (clause-pass-user-crit-post?
                                   (car tmp-pr)(cdr tmp-pr) history)))
                        (not (clause-pass-user-limits-post?
                                    (cdr tmp-pr) )))
                    (set! tmp-pr #f) )
                ; a check and cutting-off using *discr-tree*
                ; NB! vars have to be clear here (discr-checking sets vars)

                (if tmp-pr
                    (set! tmp-triple (unit-handle-cl!
                                      (car tmp-pr)(cdr tmp-pr) history))
                    (set! tmp-triple #f) )
                ; NB! vars should be again clear here (discr-tree stuff
                ; continuously cleans up after itself)

                ;(newline)(display "checkpoint 0")(newline)
                ;(display "tmp-triple:")(display tmp-triple)(newline)


                ; test time

                ;(newline)
                ;(display "stop-fast-loop-flag: ")(display *stop-fast-loop-flag*)(newline)
                ;(display "(get-check-time 5): ")(display (get-check-time 5))(newline)
                ;(display "*slow-start-time*: ")(display *slow-start-time*)(newline)
                ;(display "*one-pass-time*: ")(display *one-pass-time*)(newline)

                (if  (and (not *stop-fast-loop-flag*)
                          (> (- (get-check-time 5) *slow-start-time*)
                             *one-pass-time*))
                     (begin

                       ;(display "*stop-fast-loop-flag* was now set to #t !!!")(newline)

                       (store-time-checkpoint! 5)
                       (set! *sos-exhausted-flag* #f)
                       (set! *stop-fast-loop-flag* #t) ))

                (if (> *display-info-level* 110)
                    (begin
                (if tmp-pr
                    (begin
                      (newline)(display "after exiting unit-handle-cl!:")
                      (display (car tmp-pr))(newline)
                      (debug-print-pure-clause (cdr tmp-pr))(newline) )) ))
                ; test user limits
                (if (and tmp-triple
                         (not (clause-pass-user-limits-post?
                               (cddr tmp-triple) )))
                    (begin
                      (set! tmp-triple #f)
                      (if (>= *display-info-level* 80)
                          (begin
                            (newline)
                            (display "removed for failing user limits")
                            (newline) ))))

                ;(newline)(display "checkpoint 10a")(newline)

                (cond
                 ((not tmp-triple) #f) ; subsumption by discr-tree
                 ; the following always fails
                 ((begin (set! history (car tmp-triple))
                         (set! tmp-pr (cdr tmp-triple))
                         #f))
                ((empty-cl? (cdr tmp-pr))
                    ; empty clause derived by unit-cutting with discr tree
                    (set! *answer-cl* (cdr tmp-pr))
                    (set! *answer-history* history)
                    (set! *foundflag* #t)
                    (if (>= *display-info-level* 20)
                        (begin
                         (newline)
                         (display "********* EMPTY CLAUSE DERIVED *********")
                         (newline) )))
                 ; compute clauseinf here and sort the clause;
                 ; following case always fails
                 ((begin

                     ;(newline)(display "checkpoint 12")(newline)

                    (set! tmp-pr
                      (sort-clause (map cons (car tmp-pr) (cdr tmp-pr)) #t) )
                    (set! tmp-pr
                      (cons (map car tmp-pr) (map cdr tmp-pr)) )
                    (set! clauseinf (make-clause-info
                                       (cdr tmp-pr) history distance))
                    (set! tmp-fullcl (cons clauseinf (cdr tmp-pr)))

                     ;(newline)(display "checkpoint 20")(newline)

                    #f))
                 ; for factor-generated only: attempt subsumption by
                 ; so far generated factors
                 ((and fac-gen-flag
                       (or (some? (lambda(x)
                                    (subsumed-clause? tmp-fullcl (cdr x)) )
                                   *locfactorpool*)
                           (some? (lambda(x)
                                    (subsumed-clause? tmp-fullcl (cdr x)) )
                                  *factorpool*)))
                    #f)
                 ; attempt regular subsumption for nonunit clauses;
                 ; if succeeds, do nothing
                 ((and (not fac-gen-flag) ; not factor-generated
                       (not (unit-cl? (cdr tmp-pr))) ; nonunit?
                       (not *handle-derived-inf-cl-pair-return-directly*)
                       ;(begin (newline)(display "checkpoint 30")(newline) #t)

                       (havealready-clause? clauseinf (cdr tmp-pr)) )
                   #f)
                 (else

                     ;(newline)(display "checkpoint 40")(newline)

                    ; For the hyperresolution case clause will be kept
                    ; only if fully positive. Otherwise it will
                    ; be further used as a nucleon.
                    ; Clause will be kept for non-factor-generated
                    ; non-nucleus case
                    ; and put to *locfactorspool* for factor-gen case
                  (let ((kept-cl (make-kept-clause clauseinf tmp-pr)))

                      (if (>= *display-info-level* 70)
                                (begin
                                  (newline)
                                  (display "derived clause nr ")
                                  (display (caddr (cddar kept-cl)))
                                  (display " after 2-nd order deco:")
                                  (newline)
                                  (print-clause kept-cl)(newline)
                                  (if *display-debug-info-flag*
                                      (begin (debug-print-clause kept-cl)
                                             (newline) )))
                                (if *display-info-char-flag*
                                    (display #\+) ))

                      ; full unit cutoff performed

                      (set! tmp-pr
                             (full-unit-cutoff! (car tmp-pr) kept-cl) )

                      ;(newline)(display "passed FULL UNIT CUTOFF")(newline)

                      (if *display-debug-info-flag*
                                (begin (debug-print-clause kept-cl)
                                             (newline) ))

                      (set! kept-cl (cdr tmp-pr))


                      ;(if *display-debug-info-flag*
                      ;          (begin (debug-print-clause kept-cl)
                      ;                       (newline) ))


                      ; ground lits and unit clauses inserted
                      (if (empty-cl? (cdr kept-cl))
                          (if (not ; do not stop when wrong answer clauses are produced
                                  (and *classical-synth-stratflag*
                                       (some? (lambda(x)
                                               (and (answer-lit? x)(not (ans-computable-lit? x))) )
                                            (cdr kept-cl) )))
                           (begin
                            ; empty clause derived by full unit-cutoff
                            (set! *answer-cl* (cdr kept-cl))
                            (set! *answer-history*
                               (car (clinfo-history-sublist (car kept-cl))) )
                            (set! *foundflag* #t)
                            (if (>= *display-info-level* 20)
                                (begin
                                  (newline)
                                  (display "**** EMPTY CLAUSE DERIVED ****")
                                  (newline) ))))
                          (cond
                            (*handle-derived-inf-cl-pair-return-directly*
                                tmp-pr)
                            ((not fac-gen-flag)
                              (handle-derived-inf-cl-pair-aux tmp-pr #f) )
                              ; factorization part
                            (else
                              (begin
                                 ; back-subsume factorpool and locfactorpool
                                 (set! *locfactorpool*
                                   (remove-if
                                     (lambda(x)
                                       (subsumed-clause? (cdr x) kept-cl) )
                                       *locfactorpool*))

                                 ;(newline)
                                 ;(display "factorpart one done")(newline)

                                 (set! *factorpool*
                                   (remove-if
                                     (lambda(x)
                                       (subsumed-clause? (cdr x) kept-cl) )
                                       *factorpool*))
                                 tmp-pr))))))))))))))

(define *handle-derived-inf-cl-pair-return-directly* #f)


(define (handle-ready-clause fullclause)
  (let* ((inf (car fullclause))
         (litinforig (caddr inf))
         (litinftmp '())
         (clhistsublist (clinfo-history-sublist inf))
         (history '())
         (cl '())
         (res1 '())
         (res2 '())
         (tmp '())
         (tmp2 '())
         (tmp-pr '())
         (tmp-triple '()) )

    ;(newline)(display "handling ready clause: ")(display fullclause)
    ;(newline)

    (set! litinftmp litinforig)
    (set! history (car clhistsublist))

  ;  (set! tmp *new-demodulators-flag*)
  ;  (set! *new-demodulators-flag* #t)
  ;  (set! res1
  ;        (handle-ready-clause-subst-deco-cl
  ;             (cdr fullclause) litinftmp))
  ;  (set! *new-demodulators-flag* tmp)
  ;  (set! history (car (clinfo-history-sublist inf)))
  ;  (if (not (null? *used-demodulators*))
  ;      (set! history
  ;           (append
  ;             (make-demod-history *used-demodulators*)
  ;             history)))


    (set! res1
          (map (lambda(x)
                 (list #t 1 (car x)) )
               litinftmp))
    (set! res1
          (cons res1
               (copy-list (cdr fullclause)) ))

    ;(newline)(display "res1 is: ")(display res1)
    ;(newline)

   ; (clear-x-vars-and-touched! *max-x-var*)
   ; (clear-y-vars-and-touched! *max-x-var*)

    (set! *handle-derived-inf-cl-pair-return-directly* #t)
    (set! tmp *discr-tree-subsume-stratflag*)
    (set! *discr-tree-subsume-stratflag* #f)
    (set! res2
        (handle-derived-inf-cl-pair
            res1 history #f #f (cadr clhistsublist) ))
    (set! *discr-tree-subsume-stratflag* tmp)
    (set! *handle-derived-inf-cl-pair-return-directly* #f)

    (if *foundflag*
        (begin
           (if (>= *display-info-level* 20)
               (begin
                 (set! res2 (cons '() fullclause))
                 (newline)
                 (display "**** EMPTY CLAUSE DERIVED ****")
                 (newline)
                 (display "By given clause simplification.")
                 (newline) )))
                 ; (quit)
        (begin
         #f
         ;(newline)(display "res2 is: ")(display res2)
         ;(newline)
          ))

    (set! tmp-pr res2)

        ; now comes a block for getting the clause to final shape

  (if (or (not res2)
          (eq? (length fullclause)
               (length (cdr res2)) ))
    fullclause
    (let* ((tmpres '())
           (tmpcl '())
           (kept-cl (cdr tmp-pr))
           (litinfolst
                  (if (or *order-strategy-flag*
                          *poslitpref-order-strategy-flag*
                          *name-order-strategy-flag*
                          *epr-order-strategy-flag*
                          *exp-order-strategy-flag*
                          *firstlitpref-order-strategy-flag*
                          *posweight-order-strategy-flag*
                          *ordered-paramodulation-flag*)
                      (map cddr (order-prohibit-litinfos
                                 (car tmp-pr) (cdr kept-cl) ))
                      (map cddr (car tmp-pr)) ))
                 (strengthint (caar kept-cl))
                 (len (shiftright
                       (logand *cl-len-mask* strengthint)
                       *cl-len-shift*))
                 (strength (logand *strength-mask* strengthint)) )
       ; for sos-order, clear all prohibition bits if
       ; distance is not #f (ie if we have a sos clause)
       (if (and *firstlitpref-order-strategy-flag*
                 *order-sos-strategy-flag*
                 (clinfo-distance (car kept-cl)) )
           (begin
              (set! litinfolst
                 (map cddr
                   (clear-litorder-litinfo
                      (car tmp-pr) (cdr kept-cl)))) ))
       ; keep original as a subsumed clause, if necessary
       (if *keep-all-old-clauses-flag*
          (set! *all-back-subsumed-clauses*
                (cons fullclause *all-back-subsumed-clauses*) ))

       ; insert a right litinfolst into the kept-cl infopart

       ;(newline)(display "litinfolst: ")(display litinfolst)
       ;(newline)

       (set-car! (cddar kept-cl) litinfolst)

       ;(newline)(display "kept-cl: ")(display kept-cl)
       ;(newline)

       (cond
         ((and (eq? (length (cdr fullclause))
                    (length (cdr kept-cl)) )
               (term-list-equal? (cdr fullclause) (cdr kept-cl)) )
            (set! kept-cl fullclause)
            ;(newline)
            ;(display "given clause unchanged")
            ;(newline)
            )
         (else
           #f
            ;(newline)
            ;(display "given clause converted")
            ;(newline)
            ))

       ;(newline)(display "handle-given-clause returns: ")(display kept-cl)
       ;(newline)

       kept-cl))))





;(define (too-many-kept? tmp-pr fac-gen-flag)
; (let ((tmp '()))
;  (newline)(display "too-many-kept: ");(display tmp-pr)
;  (newline)
;  (display (compute-real-clausesize (caddr tmp-pr)))
;  (newline)
;  (set! tmp (too-many-kept-real? tmp-pr fac-gen-flag))
;  (display tmp)
;  (newline)
;  tmp))


(define (too-many-kept? tmp-pr fac-gen-flag)
 (let ()
  ;(newline)(display "too-many-kept: ")
  ;(newline)(display *clause-keepsize-limit*)(newline)
  ;(display tmp-pr)
  ;(newline)
  ;(display (compute-real-clausesize (caddr tmp-pr)))
  ;(newline)
  (if  (or ; (> (compute-real-clausesize (caddr tmp-pr)) *clause-keepsize-limit*)
           (> *kept-size-sum-nr* *max-size-sum-nr*)
           (and (> (* 3 *kept-size-sum-nr*) *max-size-sum-nr*)
                (> (compute-real-clausesize (caddr tmp-pr))
                   ;(* 5 *input-maxsize-nr*)
                   (* 3 *last-best-clausesize*) ))
           (and (> (* 3 *kept-size-sum-nr*) (* 2 *max-size-sum-nr*))
                (> (compute-real-clausesize (caddr tmp-pr))
                   ;(* 3 *input-maxsize-nr*)
                   (* 2 *last-best-clausesize*) )
                (> (compute-real-clausesize (caddr tmp-pr))
                   (+ 10 *last-best-clausesize*)) )
           (and (> (* 5 *kept-size-sum-nr*) (* 4 *max-size-sum-nr*))
                (> (compute-real-clausesize (caddr tmp-pr))
                   (+ 10 *last-best-clausesize*)))
           (and
             (> *kept-clauses-nr*
                (- (+ *one-pass-kept-clauses-firstnr*
                      *max-kept-clauses-nr*)
                   3000))
             (not (and (< *kept-clauses-nr*
                          (- (+ *one-pass-kept-clauses-firstnr*
                                *max-kept-clauses-nr*)
                             1500))
                       (<= (logand *term-size-strength-mask*
                                   (caadr tmp-pr) )
                           *last-best-weight*)))
             (not (and (< *kept-clauses-nr*
                          (+ *one-pass-kept-clauses-firstnr*
                             *max-kept-clauses-nr*))
                       (< (logand *term-size-strength-mask*
                                  (caadr tmp-pr) )
                          *last-best-weight*))) ))
        (if *complete-unsat-strategy* #f #t)
        #f)))

; mattias kirjutas siia

; tmp-pr is a pair of litinfolst with extra data elements and a full clause
; where the info-part contains a wrong litinfolst. The new correct litinfolst
; is inserted here.

(define (handle-derived-inf-cl-pair-aux tmp-pr fac-gen-flag)

  ;(newline)(display "handle-derived-inf-cl-pair-aux entered:")
  ;(display tmp-pr)(display fac-gen-flag)(newline)

  ; check whether it is time to stop

  ;(if *display-debug-info-flag*
  ;             (begin
  ;                   (debug-print-clause (cdr tmp-pr))
  ;            (newline) ))

  ; (display *kept-clauses-nr*)(newline)

  (if (or (and *hyper-strategy-flag*
               (not fac-gen-flag)
               (not (positive-cl? (cddr tmp-pr))))
          *unit-clash-flag*)
      ; result is still a nucleus
      (let ((kept-cl (cdr tmp-pr))
            (litinfolst (map cddr (car tmp-pr))) )
        (set! *kept-mid-nucleus-nr* (+ 1 *kept-mid-nucleus-nr*))
        ; insert a right litinfolst into the kept-cl infopart
        (set-car! (cddar kept-cl) litinfolst)
        ; insert to discr-tree

        ;(newline)(display "starting to insert into DISCR TREE")(newline)

        (if  *unit-clash-flag*
            ;(and *unit-clash-flag*
            ;         (not (null? (cdr litinfolst))) )
            (insert-clause-discr-tree!
             *discr-tree* (cdr tmp-pr) (car tmp-pr) #f)
            (insert-clause-discr-tree!
             *discr-tree* (cdr tmp-pr) (car tmp-pr) #f) )

        ;(newline)(display "inserted OK into DISCR TREE")(newline)


        (if (>= *display-info-level* 100)
            (begin
              (newline)
              (display "remaining part of nucleon in pure form:")
              (newline)
              (print-pure-clause (cdr kept-cl))
              (newline)
              (debug-print-xvars 3)(debug-print-yvars 3)
              (if *display-debug-info-flag*
                  (begin
                    (newline)(display "kept clause in debug form:")
                    (newline)
                    (debug-print-clause kept-cl) )))
            (if *display-info-char-flag*
                (display #\$) ))
        (if *one-step-hyper-phase*
            kept-cl
            (begin (set! *stop-fast-loop-flag* #f)
                   (usable-clauses-fast-loop kept-cl #f)
                   #f))) ; #f prohibits attempts to factor the result

      ; the result is not still a nucleus
      ; save the derived clause to *sos-clauses-list*
      ; and *subsumption-list*. Do backsubsumption.
      (if (too-many-kept? tmp-pr fac-gen-flag)
          #f
          (let* ((tmpres '())
                 (tmpcl '())
                 (kept-cl (cdr tmp-pr))
                 (litinfolst
                  (if (or *order-strategy-flag*
                          *poslitpref-order-strategy-flag*
                          *name-order-strategy-flag*
                          *epr-order-strategy-flag*
                          *exp-order-strategy-flag*
                          *firstlitpref-order-strategy-flag*
                          *posweight-order-strategy-flag*
                          *ordered-paramodulation-flag*)
                      (map cddr (order-prohibit-litinfos
                                 (car tmp-pr) (cdr kept-cl) ))
                      (map cddr (car tmp-pr)) ))
                 (strengthint (caar kept-cl))
                 (len (shiftright
                       (logand *cl-len-mask* strengthint)
                       *cl-len-shift*))
                 (strength (logand *strength-mask* strengthint)) )

            ; for sos-order, clear all prohibition bits if
            ; distance is not #f (ie if we have a sos clause)
            (if (and *firstlitpref-order-strategy-flag*
                     *order-sos-strategy-flag*
                     (clinfo-distance (car kept-cl)) )
                (begin
                  ;(display "about to clear litinf:")(newline)
                  ;(print-clause kept-cl)
                  (set! litinfolst
                    (map cddr
                      (clear-litorder-litinfo
                         (car tmp-pr) (cdr kept-cl)))) ))

            ;(newline)(display "CHECKPOINT 1")(newline)

            (set! *kept-clauses-nr* (+ 1 *kept-clauses-nr*))
            (set! *kept-size-sum-nr*
                   (+ *kept-size-sum-nr*
                      (compute-real-clausesize (cdr kept-cl))))

            ; insert a right litinfolst into the kept-cl infopart
            ;(set! *kept-clauses-thispass-nr* (+ 1 *kept-clauses-thispass-nr*))
            (set-car! (cddar kept-cl) litinfolst)

            ;            (if (>= *display-info-level* 60)
            ;                (begin
            ;                  (newline)
            ;                  (newline)(display "kept clause 1 in pure form:")
            ;                  (newline)
            ;                  (print-pure-clause (cdr kept-cl))
            ;                 (newline)(display kept-cl)
            ;                  (newline)
            ;                  (if *display-debug-info-flag*
            ;                      (begin
            ;                        (newline)(display "kept clause 1 in debug form:")
            ;                        (newline)
            ;                        (debug-print-clause kept-cl) )))
            ;                (if *display-info-char-flag*
            ;                    (display #\+) ))


            ; insert to discr-tree


            ;(if *display-debug-info-flag*
                   ;  (begin  (display "before insert-clause-discr-tree:")
                   ;  (debug-print-clause kept-cl)
            ;  (newline) ))

            ;(newline)(display "strating to insert to CLAUSE DISCR TREE")(newline)

            (insert-clause-discr-tree!
             *discr-tree* (cdr tmp-pr) (car tmp-pr) #t)

            ;(newline)(display "inserted to CLAUSE DISCR TREE")(newline)

            ;(if *display-debug-info-flag*
                   ;  (begin  (display "after insert-clause-discr-tree:")
                   ;  (debug-print-clause kept-cl)
            ;  (newline) ))

            (if (eq? len 1) (insert-unit-alist kept-cl))


            ;(newline)(display "inserted to UNIT ALIST")(newline)

            ; if possible, insert into a clause demodulation tree
            ; CASC NEWEST TRY
            (if *demodulate-with-clauses-stratflag*
                (begin

                  ;(newline)(display "starting to demod-clause-test")(newline)

                  (set! tmpres (demodulation-clause-test (car tmp-pr) (cdr kept-cl)))

                  ;(newline)(display "ending  demod-clause-test")(newline)

                  ;(if tmpres
                  ;    (begin
                  ;      (newline)
                  ;      (display "new clause demod found: ")
                  ;      (print-clause kept-cl) ))

                  ;(display "demod-clause-test passed with: ")
                  ;(display tmpres)(newline)
                  (if tmpres
                    (for-each
                     (lambda (res lit)
                       (if res
                                 (begin

                                   ;(newline)
                                   ;(display "+")

                                   (set! tmpcl (cons lit (remove-eq lit (cdr kept-cl))))
                               (set! *cur-clause*
                                   ; (copy-list
                                     (cons (car kept-cl)
                                           (cdr (rename-x-to-z-term (cons 0 tmpcl))) )) ; )

                               ;(display "inside for-each we have *cur-clause*: ")
                               ;(debug-print-clause *cur-clause*)(newline)
                               ;(display " and kept-clause: ")
                               ;(debug-print-clause kept-cl)(newline)

                               (set! *cur-lit* (cadr *cur-clause*))

            ;(newline)(display "starting to insert into DISCR DEMODATOMS TREE")(newline)

            (insert-discr-tree!
                                   *discr-tree-demod-atoms* *cur-lit* #f #t #t #t)
                               ; *discr-tree-demod-atoms* lit #f #t #t #t)
            ;(newline)(display "inserted OK to DEMODATOMS TREE")(newline)

                            ;(display " kept-clause after insert-discr-tree: ")
                               ;(debug-print-clause kept-cl)(newline)

                               )))
                       tmpres
                       (cdr kept-cl) ))
                   (set! *cur-lit* '())
                   (set! *cur-clause* '())
                   (set! tmpcl '())
                   (set! tmpres '()) ))
                  ;(debug-print-xvars 5)
                  ;(debug-print-yvars 5)

            ;            (if (>= *display-info-level* 60)
            ;                (begin
            ;                  (newline)(display "kept clause 2 in pure form:")
            ;                  (newline)
            ;                  (print-pure-clause (cdr kept-cl))
            ;                  (newline)
            ;                  (if *display-debug-info-flag*
            ;                      (begin
            ;                        (newline)(display "kept clause 2 in debug form:")
            ;                        (newline)
            ;                        (debug-print-clause kept-cl) )))
            ;                (if *display-info-char-flag*
            ;                    (display #\+) ))

            ; back-subsume

            ;(if *display-debug-info-flag*
                   ;  (begin  (display "before back-subsume:")
                   ;  (debug-print-clause kept-cl)
            ;  (newline) ))

            ;(newline)(display "starting to do BACK SUBSUMPTION")(newline)

            (if *back-subsume-stratflag*
                (back-subsume (car kept-cl) (cdr kept-cl)) )

            ;(newline)(display "done BACK SUBSUMPTION")(newline)

            ;(if *display-debug-info-flag*
                   ;  (begin  (display "after back-subsume:")
                   ;  (debug-print-clause kept-cl)
            ;  (newline) ))

            (if (>= *display-info-level* 60)
                (begin
                  (newline)(display "kept clause in pure form:")
                  (newline)
                  ;(print-pure-clause (cdr kept-cl))
                  (print-full-clause kept-cl)

                  (newline)
                  (if *display-debug-info-flag*
                      (begin
                        (newline)(display "kept clause in debug form:")
                        (newline)
                        (display (cdr kept-cl))(newline)
                        (debug-print-clause kept-cl) )))
                (if *display-info-char-flag*
                    (display #\+) ))

            ; NOT inserted to back-subsumption-list here!
            ; (insert-to-subsumption-list! len strength kept-cl)

            ;(newline)(display "before insertion:")(newline)
            ; (for-each (lambda(x)(display x) (newline))
            ;       *sos-clauses-list*)
            ; (newline)
            (cond
             ((not (null? *special-insertion-place*))
                (set-cdr! *special-insertion-place*
                          (cons kept-cl
                                (cdr *special-insertion-place*) )))
            ((null? *sos-clauses-list*)
                (set! *sos-clauses-list* (list kept-cl))
                (set! *sos-clauses-list-end* *sos-clauses-list*) )
            (else
                (set-cdr! *sos-clauses-list-end* (cons kept-cl '()))
                (set! *sos-clauses-list-end*
                      (cdr *sos-clauses-list-end*) )))

            (if *dynamic-demodulation-flag*
                (attempt-to-make-a-demodulator kept-cl) )

            ;(newline)
            ;(display "inserted a new el to sos-clauses-list-end, length is: ")
            ;(display (length *sos-clauses-list*))(newline)
            ;(for-each (lambda(x)(display x) (newline))
            ;               *sos-clauses-list*)
            ;(newline)(newline)

            ; for intuitionistic logic: apply all the single-premiss rules
            (if *int-logic-flag* (use-all-single-premiss-rules kept-cl))

            ; for demodulation with clauses: apply all the demodulating clauses,
            ; resulting in max one clause to be kept

            (if *demodulate-with-clauses-stratflag*
                (demodulate-with-clauses kept-cl))

            ;(newline)(display "demodulated with clauses ok")(newline)

            ; test for immediate unit clash-contradiction
            (if (and *hyper-contradiction-strategy*
                     (not (unit-cl? (cdr kept-cl))) )
                (hyper-contradiction? kept-cl) )

            (if *one-step-hyper-phase*
                #f        ; non-nucleons will not be used further.
                tmp-pr)))))



(define (attempt-to-make-a-demodulator fullcl)
  (let* ((inf (car fullcl))
         (cl (cdr fullcl))
         (lit (car cl))
         (litinfnr (caar (caddr inf)))
         (eqlit '()) )

    ;(newline)(display "about to attempt to make a demodulator from: ")
    ;(print-clause fullcl)(newline)

    (if (and (null? (cdr cl))
             (eq? (pure-fun (car lit)) *eq-pred-nr*)
             (not (and
                   (not (zero? (logand *litinf-noparafromleft-mask*
                                       litinfnr)))
                   (not (zero? (logand *litinf-noparafromright-mask*
                                       litinfnr)))))
             (or (and
                   (not (zero? (logand *litinf-noparafromleft-mask* litinfnr)))
                   (or (not *order-equality-firstpref-strat*)
                       (terminating-demod-bigger? (caddr lit) (cadr lit)) ))
                 (and
                   (not (zero? (logand *litinf-noparafromright-mask* litinfnr)))
                      (or (not *order-equality-firstpref-strat*)
                       (terminating-demod-bigger? (cadr lit) (caddr lit)) ))))

        (begin
          (set! *new-demodulators-flag* #t)
          (if (not (zero? (logand *litinf-noparafromleft-mask* litinfnr)))
              (set! eqlit (cons (car lit)
                                (cons (caddr lit)
                                      (cons (cadr lit) (cdddr lit)) )))
              (set! eqlit lit) )
          (if (const? (cadr eqlit)) (set! *demodulate-constants-flag* #t))
          (set! *cur-lit* (rename-x-to-z-term eqlit))
          (set! *cur-clause* (list inf *cur-lit*))
          (if #t ; (>= *display-info-level* 60)
              (begin
                ;(newline)(display "equality produces a new demodulator:")
                ;(newline)
                ;(print-pure-clause (cdr *cur-clause*))
                ;(newline)
                ;(display *cur-clause*)
                (if *display-debug-info-flag*
                    (begin
                      (newline)(display "demodulator in debug form:")
                      (newline)
                      (debug-print-clause *cur-clause*) )))
              (if *display-info-char-flag*
                  (display #\+) ))
          (set! *new-demodulators-list*
                (cons *cur-clause* *new-demodulators-list*) )
          (set! *kept-new-demodulators-nr* (+ 1 *kept-new-demodulators-nr*))
          ;(newline)
          ;(show-discr-tree *discr-tree-demod-terms* #t)(newline)
          ;(display *cur-lit*)(newline)
          ;(display (eq-lit-arg1 *cur-lit*))(newline)
          (insert-discr-tree!
             *discr-tree-demod-terms* (eq-lit-arg1 *cur-lit*) #t #t #t #f)))))
          ;(show-discr-tree *discr-tree-demod-terms* #t)(newline)


(define (terminating-demod-bigger? lterm rterm)
   (let* ((ldepths (make-var-depth-info lterm))
          (rdepths (make-var-depth-info rterm))
          (tmp '()) )

      (if (and (not (null? ldepths)) (not (pair? (car ldepths))))
          (set! ldepths (cons ldepths '())) )
      (if (and (not (null? rdepths)) (not (pair? (car rdepths))))
          (set! rdepths (cons rdepths '())) )

      ;(newline)(display ldepths)(newline)
      ;(display rdepths)

      (or (null? rdepths)
          (vdepth->-varinfolit? (cons ldepths lterm) (cons rdepths rterm)) )))



; make-kept-clause is similar to make-input-clause. An important
; diff is that the cl make-kept-clause takes is 1-st-level decorated.
; NB! the clinfo it takes is PARTIAL (returned by make-clause-info)
; and gets augmented here.
; NB! cl is assumed to be sorted.
;
;
; inflist-cl-pair is a pair of inflist of literals and a pure clause.

(define (make-kept-clause clinfo inflist-cl-pair)

  ;(newline)(display "make-kept-clause entered: ")(newline)
  ;(display clinfo)
  ;(newline)
  ;(display inflist-cl-pair)(newline)

  (let ((cl (cdr inflist-cl-pair))
        (tmp '()))
    (set! *clause-var-nr* 0)
    (if (pure-clause-may-cont-vars? cl)
        (begin
          (set! cl (copy-list cl)) ; ??????????????
          ;create a list of lists (lit litrepvars litvars othervars follvars):
          (make-varinfo-lst! cl '(())) ; cl is destr replaced to be varinfo-lst
          (set! *new-var-names* '())
          (make-fulldeco-clause! cl) )) ; cl is destr replaced to be a clause
    ; cl is now the finished pure clause.
    ; *clause-var-nr* is now the max var nr in the final pure clause
    ; fill in missing details of the info-part.
    ; first, put in *clause-var-nr*
    (if (> *clause-var-nr* *cl-maxvar*)(set! *clause-var-nr* *cl-maxvar*))
    (set-car! clinfo
              (logior (shiftleft *clause-var-nr* *cl-maxvar-shift*)
                      (car clinfo) ))
    (set! *kept-cl-nr* (+ 1 *kept-cl-nr*)) ; make a new *kept-cl-nr*
    (set-car! (cddr clinfo)
              (make-kept-litinflist (car inflist-cl-pair) cl) )
    (set-car! (clinfo-clnumber-sublist clinfo) *kept-cl-nr*) ; clause nr
    (set! tmp (clinfo-history-sublist clinfo)) ; history
    (if (not *simple-history-stratflag*) ;simpleh
        (begin
           (set! tmp (clinfo-history-sublist clinfo)) ; history
           (if (not (pair? (car tmp)))
               (set-car! tmp (cons 'unknown (cons *kept-cl-nr* (car tmp))))
               (set-car! tmp (cons (caar tmp) (cons *kept-cl-nr* (cddar tmp)))) )))

; - - - -
;    (newline)
;    (do ((i 1 (+ 1 i)))
;        ((> i *clause-var-nr*))
;      (display i)(display ":")(display (var-ref i))(display " ") )
;    (newline)
; - - - -
    (cons clinfo cl) ))


;---------------------------------------------------------------------
;
; litinfo-lst elements are lists where the first element is
; a number with the bitwise encoding. Masks:
;
; (define *litinf-noresolve-mask* 1)
; (define *litinf-nofactor-mask* 2)
; (define *litinf-isanswer-mask* 4)
; (define *litinf-noparainto-mask* 8)
; (define *litinf-noparafromleft-mask* 16)
; (define *litinf-noparafromright-mask* 32)
;
;---------------------------------------------------------------------


(define (make-kept-litinflist litinf newcl) ; currently default everything OK
  (map cddr litinf) ) ; cut off first two temporary els

(define (litinf-binrespara? inf)  ; binres and/or paramodulation
  (not (eq? *litinf-nobinrespara-mask*
            (logand *litinf-nobinrespara-mask* inf) )))

(define (litinf-use? inf)
  (zero? (logand *litinf-nouse-mask* inf)) )


(define (litinf-left-para? inf)
  (zero? (logand *litinf-noparafromleft-mask* inf)) )

(define (litinf-right-para? inf)
  (zero? (logand *litinf-noparafromright-mask* inf)) )

(define (litinf-parainto? inf)
  (zero? (logand *litinf-noparainto-mask* inf)) )

; modify-to-demod-inf! takes a unit fullcl and modifies:
; inf is modified to allow only resolution, para into and para from left,
; but not para from right.

(define (modify-to-demod-inf! fullcl)
  (let* ((inflist (caddr (car fullcl)))
         (res (car inflist)) )
     (set! res (cons
                (logior 0 ; *litinf-noparainto-mask*
                        (logior *litinf-noparafromright-mask* (car res)) )
                (cdr res) ))
     (set-car! inflist res) ))


; make-clause-info takes a pure clause and a history and returns its infolist
; where the fifth el (clause nr) is
; intentionally left '() to be filled in later in case
; the clause is actually to be kept.
; NB! More important is that *clause-var-nr* is also always 0,
; and this has to be repaired later.

(define (make-clause-info cl history distance)
  (let ((strengthint 0)
        (len 0)
        (predsymbsint 0)
        (litinfolst (map (lambda(x)(list 0)) cl)) ; !!!!! ALL ops allowed
        (weight 0)
        (clinfo '())
        (minw 0)
        (tmp '()) )
    (do ((clpart cl (cdr clpart)))
        ((null? clpart))
      (let* ((lit (car clpart))
             (fun (car lit)) )
        (if (countable-pred? fun) ; answer-preds may be uncountable
            (begin
              (if (zero? strengthint)
                  (set! strengthint (logand *strength-mask* fun))
                  (set! strengthint (sum-fun-strengths fun strengthint)) )
              (if (not (zero? (logand (predsymb-bit fun) predsymbsint)))
                  (set! predsymbsint
                        (logior *cl-rep-predsymbs-mask* predsymbsint) ))
              (set! predsymbsint (logior (predsymb-bit fun) predsymbsint))
              (set! len (+ 1 len))
              (set! weight (+ weight (lit-weight lit))) ))))
    (if *split-strategy-flag*
        (set! weight strengthint) )
    (if (and *split-strategy-flag*
             (not (null? (cdr cl))) )
        (begin
           (set! weight strengthint)
           (set! tmp (map (lambda(x)(make-var-depth-info x)) cl))
           (if  (every? (lambda(el)
                         (and (not (null? el))
                             (every? (lambda(el2)
                                       (or (eq? el2 el)
                                           (every? (lambda(v)
                                                    (not (assq (car v) el2)) )
                                                   el)))
                                     tmp)))
                       tmp)
               (begin
                 (if (> *display-info-level* 70)
                     (begin
                       (newline)
                       (display "splittable clause: ")
                       (print-pure-clause cl)
                       (newline) ))
                 (do ((clpart cl (cdr clpart)))
                     ((null? clpart)
                      (set! weight minw) )
                   (cond ((zero? minw)
                          (set! minw (logand *strength-mask* (caar clpart))) )
                         ((> (logand *strength-mask* (caar clpart))
                             minw)
                          (set! minw (logand *strength-mask*
                                             (caar clpart) )))))))))
    (if (> len *cl-maxlen*) (set! len *cl-maxlen*))
    (set! strengthint
          (logior (shiftleft len *cl-len-shift*)
                  strengthint))
    (set! clinfo
          (list strengthint
                predsymbsint
                litinfolst
                weight
                '() ; it should be *kept-cl-nr*
                history
                distance))
    clinfo))

; intern-fullclause takes a full 2-nd level decorated clause
; and interns it, ie. adds it to the *discr-tree* and
; *subsumption-list* and destructively changes the clause
; itself to make the ground literals unique.

(define (intern-fullclause cl)

  ;(newline)(display "intern-fullclause entered with: ")(newline)
  ;(display cl)(newline)
  ;(print-clause cl)(newline)
  ;(newline)(display "discr-tree is: ")(newline)
  ;(show-discr-tree *discr-tree* #t)

  (insert-clause-discr-tree!
      *discr-tree* cl (map (lambda(x)(list #t)) (cdr cl)) #t)

  ;(newline)(display "intern-fullclause is to be exited with: ")(newline)
  ;(display cl)(newline)
  ;(print-clause cl)(newline)
  ;(newline)(display "discr-tree is: ")(newline)
  ;(show-discr-tree *discr-tree* #t)

  (let*
      ((strengthint (caar cl))
       (tmpcl '())
       (len (shiftright (logand *cl-len-mask* strengthint) *cl-len-shift*))
       (strength (logand *strength-mask* strengthint))
       (tmp-litinfolst (map (lambda(x)(list #t)) (cdr cl)))
       (tmpres '()) )

     (if *demodulate-with-clauses-stratflag*
        (begin
          ;(display "starting to demod-clause-test in intern-fullclause")
          (set! tmpres (demodulation-clause-test tmp-litinfolst (cdr cl)))
          ;(display "demod-clause-test passed in intern-fullclause")
          (if tmpres
            (for-each
             (lambda (res lit)
               (if res
                  (begin

                    ; (display "+")

                    (set! tmpcl (cons lit (remove-eq lit (cdr cl))))
                       (set! *cur-clause*
                             (cons (car cl)
                                   (cdr (rename-x-to-z-term (cons 0 tmpcl))) ))
                       (set! *cur-lit* (cadr *cur-clause*))
                       (insert-discr-tree!
                            ;*discr-tree-demod-atoms* lit #f #t #t #t))))
                            *discr-tree-demod-atoms* *cur-lit* #f #t #t #t))))
               tmpres
              (cdr cl) ))))

    (if (eq? len 1) (insert-unit-alist cl))
    (insert-to-subsumption-list! len strength cl) ))


;====================================================================
;
;                  demodulation functions
;
;===================================================================

; demodulation means replacing some terms or literals with their
; normal form according to a given complete term reduction system.
; We do normalization inside-out.
;
; Demodulation is done in cooperation with subst-deco-term.
; subst-deco-term calls demodulator functions while building a term.
; if demodulate-composite finds a demodulator, it saves all those
; values of vars in a left arg which might be demodulatable.
; Then it calls again subst-deco-term to build the result of
; demodulation step. In that case, however, subst-deco-term does not
; attempt demodulation. After the term is built, it is demodulated
; by another demodulator, demodulate-fully.
;
; It uses the flag *demodulate-constants-flag* to check whether term
; should be demodulated.
; It uses the flag *new-demodulators-flag* to check whether nonchanged
; sould be also demodulated.
;
; IMPORTANT: In case *demodulate-flag* = #f, all the other
; demodulation flags are also assumed to be #f.
; demodulate-composite takes a fully built 1st-order decorated term
; with fully demodulated argument terms and demodulates it fully.
;
; Also, we assume that demodulate-composite never takes a literal
; argument.

(define (back-demodulate-usable)
  (if (>= *display-info-level* 60)
      (begin
        (newline)
        (display "back-demodulate-usable started")
        (newline) ))
  (let ((tmp '())
        (given-cl '())
        (new '())
        (newlist '()) )
    (do ((part *usable-clauses-list* (cdr part)))
        ((null? part)
           (if (>= *display-info-level* 60)
               (begin
                 (newline)
                 (display "back-demodulate-usable ended")
                 (newline) ))
           (for-each (lambda(cl)
                       ; append cl to the end of *usable-clauses-list*
                       (set-cdr! *usable-clauses-list-end* (cons cl '()))
                       (set! *usable-clauses-list-end*
                             (cdr *usable-clauses-list-end*)) )
                     newlist))
      (set! given-cl (car part))
      (if (and (pair? given-cl)
               (not (zero? (logand *term-depth-strength-mask*
                               (caar given-cl) ))))
          (begin
            (set! tmp (demodulate-clause given-cl))
            (if tmp
                (begin
                  (set! new (cdr tmp))
                  (set! newlist (cons new newlist))
                  ; make old unusable
                  (set-car! (car given-cl)
                            (logxor *term-depth-strength-mask*
                                    (logior *term-depth-strength-mask*
                                            (caar given-cl) )))
                  ; insert new to subsumption list
                  (insert-to-subsumption-list!
                   (shiftright (logand *cl-len-mask* (caar new))
                               *cl-len-shift*)
                   (logand *term-depth-strength-mask* (caar new))
                   new))))))))


(define (demodulate-clause cl)
  (let ((inf (car cl))
        (purecl (cdr cl))
        (part '())
        (newlitinf '())
        (history '())
        (res '())
        (demods '())
        (new '()) )
    (set! *used-demodulators* '())
    (set! new  (map (lambda(lit)(demodulate-fully-lit lit)) purecl))
    (if (null? *used-demodulators*)
        #f
        (begin
           (set! demods *used-demodulators*)
           (set! newlitinf (map (lambda(inf)(cons #t (cons 1 inf)))
                                (caddr (car cl)) ))
           (set! history
                 (list 'backdemod
                       0
                       (list 'literal
                             (car (clinfo-history-sublist inf))
                             1)
                       (cons 'demod *used-demodulators*) ))
           (set! res
                 (handle-derived-inf-cl-pair (cons newlitinf new)
                                             history #f #f #f) )
           (if (and res
                    (>= *display-info-level* 60)
                    (not *foundflag*) )
               (begin
                 (newline)(display "back-demodulation parent: ")
                 (newline)
                 (print-pure-clause (cdr cl))
                 (newline)
                 (display "back-demodulation result: ")
                 (newline)
                 (print-pure-clause (cdr (cdr res)))
                 (newline)
                 (display "used demodulators: ")
                 (newline)
                 (for-each (lambda(x)
                             (print-pure-clause (cdr x))
                             (newline) )
                           demods)
                 (if *display-debug-info-flag*
                     (begin
                       (newline)(display "backdemod result in debug form:")
                       (newline)
                       (debug-print-clause (cdr res)) ))))
           (if *foundflag* (set! res #f))
           res))))





(define (demodulate-composite comp)
  ;(newline)(display "demodulate-composite starts: ")(display comp)(newline)
  (let ((tmp (demodulate-composite-aux10 comp)))
    ;(newline)(display "demodulate-composite ended: ")(display tmp)(newline)
    tmp))

(define (demodulate-composite-aux10 comp)
  (let ((checkres (check-discr-tree *discr-tree-demod-terms* comp)))
    ; check-discr-tree cleans vars up after itself
    (if (not checkres)
        ; no demodulators for this term
        (if (eq? *case-fun-nr* (pure-fun (car comp)))
            (subst-deco-case-term comp)
            comp)
        ; demodulator found
        (let* ((demodcl (caddr checkres))
               (demodlit (cadr demodcl))
               (maxdemodvar (give-cl-maxvar (car demodcl))) )

          ;(newline)(display "demodclause: ")(display demodcl)(newline)
          ;(abort)

          (if (not (memq demodcl *used-demodulators*))
              (set! *used-demodulators*
                    (cons demodcl *used-demodulators*)) )
          ; the following is necessary to set the z-vars in the first
          ; arg of demodlit
          (subsumed-term-list-ordinary?
             (cdr comp) (cdr (eq-lit-arg1 demodlit)) )
          ; we switch off demodulation for the next subst-deco-term
          (let ((tmp1 '())
                (tmp2 '()) )
            ; substitute into the second arg of eq-literal, while
            ; not doing demodulation
            (set! comp (subst-deco-term-nodemodulate-spec
                          (eq-lit-arg2 demodlit) ))
            ; save possibly demodulatable values of vars in the left arg
            (set! tmp2 '())
            (do ((nr (logior *z-var-mask* maxdemodvar) (- nr 1)))
                ((<= nr *max-y-var*))
              (set! tmp1 (var-ref nr))
              (if (and (composite? tmp1) (may-be-demodulated? tmp1))
                  (set! tmp2 (cons tmp1 tmp2)) ))
            ; clear zvars
            (clear-z-vars-and-touched! maxdemodvar)
            (set! *var-set-flag* #f)
            (set! *alt-var-set-flag* #f)
            (demodulate-fully comp tmp2 #f) )))))


; demodulate-equivalent-lit assumes that demodulation clauses have the following
; form: (<clinf> <demodatom> <newatom>)
; where <demodatom> is positive, but <newatom> is negative!

(define (demodulate-equivalent-lit lit)
  (let* ((atom (if (neg-fun? (car lit))
                   (cons (logxor *sgn-mask* (car lit)) (cdr lit))
                   lit))
         (checkres (check-discr-tree *discr-tree-demod-atoms* atom)) )
    ; check-discr-tree cleans vars up after itself
    (if (not checkres)
        ; no demodulators for this term
        lit
        ; demodulator found
        (let* ((demodcl (caddr checkres))
               (demodatom (cadr demodcl))
               (maxdemodvar (give-cl-maxvar (car demodcl))) )
          (if (not (memq demodcl *used-demodulators*))
                (set! *used-demodulators*
                    (cons demodcl *used-demodulators*)) )
          ; the following is necessary to set the z-vars in demodatom
          (subsumed-term-list-ordinary? (cdr atom) (cdr demodatom))
          ; we switch off demodulation for the next subst-deco-term
          (let ((tmp0 *demodulate-flag*)
                (tmp1 *demodulate-constants-flag*)
                (tmp2 *new-demodulators-flag*)
                (tmp3 *demodulate-lit-flag*)
                (new '()) )
            (set! *demodulate-flag* #f)
            (set! *demodulate-constants-flag* #f)
            (set! *new-demodulators-flag* #f)
            (set! *demodulate-lit-flag* #f)
            ; substitute into the second atom of demodcl, while
            ; temporarily prohibiting demodulation
            (set! new (subst-deco-term (caddr demodcl) #t))
            (if (eq? atom lit)
                ; lit was positive; make new also positive.
                (set! new (cons (logxor *sgn-mask* (car new)) (cdr new))) )
            (set! *demodulate-flag* tmp0)
            (set! *demodulate-constants-flag* tmp1)
            (set! *new-demodulators-flag* tmp2)
            (set! *demodulate-lit-flag* tmp3)
            ; save possibly demodulatable values of vars in the left arg
            (set! tmp2 '())
            (do ((nr (logior *z-var-mask* maxdemodvar) (- nr 1)))
                ((<= nr *max-y-var*))
              (set! tmp1 (var-ref nr))
              (if (and (composite? tmp1) (may-be-demodulated? tmp1))
                  (set! tmp2 (cons tmp1 tmp2)) ))
            ; clear zvars
            (clear-z-vars-and-touched! maxdemodvar)
            (set! *var-set-flag* #f)
            (set! *alt-var-set-flag* #f)
            (demodulate-fully new tmp2 #t) )))))



; demodulate-lit does not assume that the demod clause represents
; full equivalence - hence it keeps completeness only if the original
; nt-clause-demodulated clause is kept too.
; It  assumes that demodulation clauses have the following
; form: (<clinf> <demodatom> <newatom>)
; where <demodatom> and <newatom> have the original correct polarity!

(define (demodulate-lit lit)

  ;(newline)
  ;(display "demodulate-lit: ")(print-lit lit)
  ;(newline)

  (let* ((tmp '())
         (atom (cons (logxor *sgn-mask* (car lit)) (cdr lit))) ; invert literal !!
         (checkres (check-discr-tree-signed *discr-tree-demod-atoms* atom)) )
    ; check-discr-tree cleans vars up after itself
    ;(display "atom: ")(print-lit atom)(newline)
    ;(display "checkres: ")(display checkres)(newline)
    ;(if checkres
    ;   (begin (display "checkres first lit: ")(print-lit (cadr checkres))(newline) ))

    (if (not checkres)
        ; no demodulators for this term
        lit
        ; demodulator found
        (let* ((demodcl checkres)
               (demodatom (cadr demodcl))
               (maxdemodvar (give-cl-maxvar (car demodcl))) )

           ;(newline)
           ;(display "demodulate-lit success - orig lit, demodulator clause: ")(newline)
           ;(print-lit lit)(newline)
           ;(print-full-clause demodcl)
           ;(newline)

          (if (not (memq demodcl *used-demodulators*))
                (set! *used-demodulators*
                    (cons demodcl *used-demodulators*)) )


          ; the following is necessary to set the z-vars in demodatom
          (subsumed-term-list-ordinary? (cdr atom) (cdr demodatom))
          ; we switch off demodulation for the next subst-deco-term
          (let ((tmp0 *demodulate-flag*)
                (tmp1 *demodulate-constants-flag*)
                (tmp2 *new-demodulators-flag*)
                (tmp3 *demodulate-lit-flag*)
                (new '()) )
            (set! *demodulate-flag* #f)
            (set! *demodulate-constants-flag* #f)
            (set! *new-demodulators-flag* #f)
            (set! *demodulate-lit-flag* #f)
            ; substitute into the second atom of demodcl, while
            ; temporarily prohibiting demodulation
            (set! new (subst-deco-term (caddr demodcl) #t))
            (set! *demodulate-flag* tmp0)
            (set! *demodulate-constants-flag* tmp1)
            (set! *new-demodulators-flag* tmp2)
            (set! *demodulate-lit-flag* tmp3)
            ; save possibly demodulatable values of vars in the left arg
            ;(set! tmp2 '())
            ;(do ((nr (logior *z-var-mask* maxdemodvar) (- nr 1)))
            ;    ((<= nr *max-y-var*))
            ;  (set! tmp1 (var-ref nr))
            ;  (if (and (composite? tmp1) (may-be-demodulated? tmp1))
            ;      (set! tmp2 (cons tmp1 tmp2)) ))
            ; clear z-vars
            ;(clear-x-vars-and-touched! )
            ;(clear-y-vars-and-touched! )
            (clear-z-vars-and-touched! maxdemodvar)

            ;(newline)
            ;(display " lit demodulation result: ")(newline)
            ;(print-lit new)(newline)

            ;(display "*used-demodulators* :")(newline)
            ;(for-each (lambda(x)(print-full-clause x)(newline)) *used-demodulators*)

            (set! *var-set-flag* #f)
            (set! *alt-var-set-flag* #f)
            (set! tmp (demodulate-fully new '() #t))
            (if (and tmp (not (null? tmp)))
                tmp
                lit)))) ))


(define (demodulate-constant const)
 (set! *nonunit-discr-flag* #t)
 (let ((checkres
          (check-discr-tree-part *discr-tree-demod-terms* (list const) '())))
    ; check-discr-tree cleans vars up after itself
    (if (not checkres)
        ; no demodulators for this term
        const
        ; demodulator found
        (let* ((demodcl (caddr checkres))
               (demodlit (cadr demodcl)) )
          (eq-lit-arg2 demodlit) ))))


; demodulate-fully fully demodulates a ready-built 1-st order
; decorated term or lit. It DOES NOT assume that arguments are already fully
; demodulated.
;
; done-terms is a list of subterms which are known to be fully demodulated
;
; we assume that it never takes a literal argument

(define (demodulate-fully term done-terms litflag)
  (cond
    ((non-composite? term)
       (if (non-composite:-var? term)
           term
           (if *demodulate-constants-flag*
               (demodulate-constant term)
               term)))
    ((memq term done-terms)
       term)
    ((not (may-be-demodulated? term))
       term)
    (else
     (let* ((res (cons '() '()))
            (resend res)
            (str *ground-mask*) ) ; no args is just ground
       (do ((part (cdr term) (cdr part)))
           ((null? part)
               (set-car! res (create-new-fun-pos-object (car term) str))
               ; args of res are already demodulated (if possible).
               (if litflag
                   (demodulate-lit res)
                   (demodulate-composite res) ))
               ;(if (and *demodulate-flag* (not litflag))
               ;    (demodulate-composite res)
               ;    res)
         (let ((new (demodulate-fully (car part) done-terms #f)))
           (set-cdr! resend (cons new '()))
           (if (eq? resend res) ; the first arg at hand?
               ; now comes the case for the first arg
               (if (non-composite? new)
                   (if (non-composite:-var? new)
                       (set! str *var-strength*)
                       (set! str *const-strength*) )
                   (set! str (car new)) )
               ; now comes the case for the following args
               (if (non-composite? new)
                   (if (non-composite:-var? new)
                       (set! str (sum-fun-strengths *var-strength* str))
                       (set! str (sum-fun-strengths *const-strength* str)) )
                   (set! str (sum-fun-strengths (car new) str)) ))
           (set! resend (cdr resend)) ))))))


(define (may-be-demodulated? term)
   #t)


(define (eq-lit-arg1 lit)
  (cadr lit) )


(define (eq-lit-arg2 lit)
  (caddr lit) )

(define (eq-lit-type lit)
  (caddr (cdr lit)) )


(define (demodulate-fully-lit term)
     (let* ((res (cons '() '()))
            (resend res)
            (str *ground-mask*) ) ; no args is just ground
       (do ((part (cdr term) (cdr part)))
           ((null? part)
               (set-car! res (create-new-fun-pos-object (car term) str))
               ; args of res are already demodulated (if possible).
               res)
         (let ((new (demodulate-fully (car part) '() #f)))
           (set-cdr! resend (cons new '()))
           (if (eq? resend res) ; the first arg at hand?
               ; now comes the case for the first arg
               (if (non-composite? new)
                   (if (non-composite:-var? new)
                       (set! str *var-strength*)
                       (set! str *const-strength*) )
                   (set! str (car new)) )
               ; now comes the case for the following args
               (if (non-composite? new)
                   (if (non-composite:-var? new)
                       (set! str (sum-fun-strengths *var-strength* str))
                       (set! str (sum-fun-strengths *const-strength* str)) )
                   (set! str (sum-fun-strengths (car new) str)) ))
           (set! resend (cdr resend)) ))))


; demodulate-with-clauses takes a full decorated and ordered clause, assumes
; that it is already demodulated with term demodulators and attempts
; to demodulate the clause with clause demodulators.
; If that succeeds, it factorises and keeps a new fully built clause, otherwise
; it does and keeps nothing

(define (demodulate-with-clauses-ok fullcl) ; temporary dummy until next is OK
  #f)

(define (demodulate-with-clauses fullcl) ; do no use that!!
   ;(newline)
   ;(display "demodulate-with-clauses: ")
   ;(print-full-clause fullcl)
   ;(newline)
   ;(debug-print-xvars 5)
   ;(debug-print-yvars 5)

  (let ((res '())
        (tmp '())
        (infolst (caddr (car fullcl)))
        (oldhistory '())
        (history '())
        (ans-inf '())
        (foundflag #f) )
    (set! *used-demodulators* '())
    (set! *demodulate-lit-flag* #t)
    (set! res
      (map (lambda(lit info)
             ;(newline)
             ;(display "starting to demodulate-lit: ")
             ;(print-lit lit)(display " : ")(display info)
             ;(newline)
             (set! tmp (demodulate-lit lit))
             ;(display "res: ")(display tmp)
             (if (not (eq? tmp lit))
                 (begin
                    (set! foundflag #t)  ; we could demodulate this literal!
                    (cons (cons #t (cons 1 info)) tmp) )
                 (cons (cons #f (cons 1 info)) tmp) ))
           (cdr fullcl)
           infolst))
    (set! *demodulate-lit-flag* #f)
    ;(newline)(display foundflag)(newline)
    (if (not foundflag)
        #f ; exit if nothing demodulated
        (begin
              ; OK, successful demodulation

              ; (display "*")

              (set! infolst (map (lambda(x)(car x)) res))
              (set! res  ; NB! SLOW FULL COPY OF NOT DEMODULATED LITERALS
                 (map
                   (lambda(x)
                      (if (not (caar x))
                          (copy-nonground-term (cdr x))
                          (cdr x) ))
                   res))
              (if (>= *display-info-level* 70)
                (begin
                    (newline)
                    (display "raw cl by literal demodulation -  parent, demodulators, result: ")
                    (newline)
                    (print-clause fullcl)(newline)
                    (for-each (lambda(x)(print-full-clause x)(newline)) *used-demodulators*)
                    (print-clause (cons '() res)) (newline) ))

              ;clear all vars (subst-deco sets both!)
              ;(clear-y-vars-and-touched! yclmaxvar)
              ;(clear-x-vars-and-touched!(give-cl-maxvar (car fullcl)))
              ;(debug-print-xvars 1)(debug-print-yvars 1)

              ; create history

              (set! oldhistory (car (clinfo-history-sublist (car fullcl))))
              (if *simple-history-stratflag* ;simpleh
                  (set! history
                        (append (cons 'binarydemod
                                      (cdr (make-demod-history *used-demodulators*)) )
                                oldhistory))
                  (set! history
                     (list 'binres_demod 0 oldhistory
                          0 0
                              (make-demod-history *used-demodulators*) )))
              ; start handling the result
              (set! ans-inf #f)
              (set! res
                   (handle-derived-inf-cl-pair
                           (cons infolst res) history #f ans-inf #f) )
              ;(set! ans-inf #f) ; clear ans-inf again
              ;(debug-print-xvars 5)
          ;(debug-print-yvars 5)
          ;(debug-print-zvars 5)
              ))))


;===================================================================
;
;                  factorization functions
;
;===================================================================

; factor takes a clause and a corresponding infolist.
;
; The clause is assumed to be fully decorated and ordered.
;
; The info-list has a format:
; (<changed-flag> <nr> . <orig-info>)
; where
; <changed-flag> is #t or #f (dep. on whether the lit is unchanged)
; <nr> is 1 or 2 (dep on which premiss the lit came from) or #f
;      in which case it is always allowed to be factored upon
;      or #t in which case it is prohibited to be factored upon.
; <orig-info> is the corresponding infolist el of the premiss literal.
; <orig-info>-s of the infolist el are by default appended to each other.
;
; The general strategy:
; 0) Before starting factorizing a clause C, it is assumed to have been
;    checked by havealready?, fully decorated and ordered.
; 1) Produce all 1-st level factors of the clause C:
;    for each newly derived factor X check subsumption by already
;    derived factors and unit subsumption with doing also unit cutoff,
;    but no havealready? check yet.
;    If X is OK, decorate
;    and order it, getting X'. Do backsubsumption of already generated
;    factors by X'. Then try to subsume C by X'. If it succeeds,
;    throw away all so far generated 1-st order factors of C, cancel
;    factor generation of C, add X' to the beginning of factorpool.
;    Do NOT do havealready? check for a subsuming factor.
; 2) Once all the 1-st level factors of C are produced, add them
;    (locfactorpool) to the end of factorpool. Then take a next
;    element of factorpool. Check it for
;    havealready?, remove it from factorpool, save it, do backsumption
;    with it and go to step 1 with it.
; locfactorpool and factorpool are lists of info-fullclause pairs

(define *locfactorpool* '())
(define *factorpool* '())

; factor-sos factors all the clauses in sos.

(define (factor-sos soslist)
  (let ((cl '())
        (purecl '())
        (litinflist '())
        (newlitinflist '()) )
    (if (>= *display-info-level* 50)
        (begin (newline)
               ; (newline)
               ; (show-derived)
               (display "******* starting to factor input sos ******")
               (newline) ))
    (do ((sospart soslist (cdr sospart)))
        ((or (null? sospart)
             *foundflag*
             *stop-fast-loop-flag*)
           (if (>= *display-info-level* 50)
               (begin (newline)
                      (display "******* input sos has been factored ******")
                      ;(newline)
                      ;(show-derived)
                      (newline) )))
      (set! cl (car sospart))
      (set! purecl (cdr cl))
      (set! litinflist (caddar cl))

      ;(newline)
      ;(display "factoring: ")
      ;(print-clause cl)
      ;(newline)

      (if (or (not *hyper-strategy-flag*)
              (positive-cl? purecl) )
          (begin
            (set! newlitinflist
                  (map (lambda(inf)
                         (cons #t (cons (if *ordered-factor-flag*
                                            (not (zero?
                                                  (logand
                                                   *litinf-nofactor-mask*
                                                   (car inf) )))
                                            #f)
                                        inf)))
                       litinflist))
            (factor newlitinflist cl) )))))

; factor:
; fullcl is given by handle-derived-inf-cl-pair with a factor-flag #f,
; thus it is already checked for havealready?
;
; for hyperresolution ONLY fully-positive clauses are factored
; (fully negative for negative hyperresolution) and the literals
; originating from the same clause may be factored upon (actually
; it would suffice to allow it only for the ones from nucleon,
; to compensate for possibly no factoring of the input set.

(define (factor infolist fullcl)
  (if (and (not (null? (cddr fullcl)))
           (or (not *hyper-strategy-flag*)
               (or (and (not *neg-hyper-strategy-flag*)
                        (positive-cl? (cdr fullcl)) )
                   (and *neg-hyper-strategy-flag*
                        (negative-cl? (cdr fullcl)) ))))
   ; not unit
   (begin

    ;(newline)(display "factor entered with: ")(newline)
    ;(print-clause fullcl)(newline)
    ;(debug-print-clause fullcl)(newline)
    ;(display infolist)(newline)
    ; (debug-print-xvars 10)(debug-print-yvars 10)

    (set! *locfactorpool* '())
    (set! *factorpool* '())
    (set! *subs-factor-derived-flag* #f)
    (factor-aux infolist fullcl)
    (do ()
      ((or (null? *factorpool*) *foundflag* *stop-fast-loop-flag*))
         ; (newline)(display "factor exited with vars: ")(newline) )
         ; (debug-print-xvars 10)(debug-print-yvars 10)  )

      ; remove duplicate stuff from factorpool: addition 1997


      (if (not (null? (cdr *factorpool*)))
          (set! *factorpool*
                (filter-map
                 (lambda(cl1)
                   (if (member-if
                         (lambda(cl2)
                           (and (not (eq? cl1 cl2))
                                (subsumed-clause? (cdr cl1) (cdr cl2))) )
                         *factorpool*)
                       #f
                       cl1))
                 *factorpool*)) )

;      (if (not (null? (cdr *factorpool*)))
;          (begin
;            (newline)(display "factorpool:")(newline)
;            (for-each (lambda(x)(print-clause (cdr x))(newline))
;                      *factorpool*) ))

      (let* ((inf-fullcl-pr (car *factorpool*))
             (fullcl (cdr inf-fullcl-pr))
             (infolist (car inf-fullcl-pr)) )
        (set! *factorpool* (cdr *factorpool*))
        (if #f;  *subs-factor-derived-flag*
            (begin
              ; just save the clause

              (handle-derived-inf-cl-pair-aux inf-fullcl-pr #t)
              ; produce 1-st level factors of the clause
              (factor-aux infolist fullcl) )
            (if (or (null? (cddr fullcl))
                    (not (havealready-clause? (car fullcl) (cdr fullcl))) )
                (begin
                  ; just save the clause
                  (handle-derived-inf-cl-pair-aux inf-fullcl-pr #t)
                  ; produce 1-st level factors of the clause
                  (factor-aux infolist fullcl) ))))))))

(define (factor-aux infolist fullclause)

  ;(newline)(display "factor-aux entered with: ")(newline)
  ;(print-clause fullclause)(newline)
  ;(debug-print-clause fullclause)
  ;(display infolist)(newline)

  (let* ((clause (cdr fullclause))
         (clinf (car fullclause))
         (maxvar (give-cl-maxvar clinf))
         (history (car (clinfo-history-sublist (car fullclause))))
         (slow-list-ordnr 0) ) ; !!!!!!!!!!! TMP !!!!!!!!!!!!!!!!
    (set! *subs-factor-derived-flag* #f)
    (set! *locfactorpool* '())
    (do ((slow-cl-part clause (cdr slow-cl-part))
         (slow-info-part infolist (cdr slow-info-part))
         (slow-lit-ordnr 1 (+ 1 slow-lit-ordnr)) )
        ((null? slow-cl-part)

           ;(newline)(display "factor-aux exited")(newline)

           (set! *factorpool*
                 (append *factorpool* *locfactorpool*)) )
      (let* ((xlit (car slow-cl-part))
             (xlitinf (car slow-info-part))
             (xlitnr (cadr xlitinf)) )

         ;(newline)(display "xlit to be factored upon:")(newline)
         ;(print-lit xlit)(newline)(display xlitinf)(newline)

         (if (not (eq? #t xlitnr))
             (do ((cl-part (cdr slow-cl-part) (cdr cl-part))
                  (inf-part (cdr slow-info-part) (cdr inf-part))
                  (lit-ordnr 1 (+ 1 lit-ordnr)) )
                 ((null? cl-part))
               (let ((ylit (car cl-part))
                     (ylitinf (car inf-part)) )

                 ;(newline)(display "ylit to be factored upon:")(newline)
                 ;(print-lit ylit)(newline)(display ylitinf)(newline)

                 (if (or (not (eq? xlitnr (cadr ylitinf)))
                         (eq? xlitnr #f)
                         *hyper-strategy-flag*)
                     (begin

                       ;(newline)(display "starting to factor-unify:")
                       ;(display maxvar)
                       ;(newline)(debug-print-yvars 10)
                       ;(newline)(debug-print-xvars 10)
                       (set! *used-demodulators* '())
                       (if (factor-unify-ordinary ylit xlit)
                           (let* ((res (subst-deco-cl
                                        (remove-at-place cl-part clause)
                                        (merge-infos
                                          slow-info-part inf-part infolist)
                                        ; (remove-at-place inf-part infolist)
                                        #f #t)))
                             (clear-x-vars-and-touched! maxvar)
                             (set! *fac-vars-set* #f)
                             (set! *var-set-flag* #f)
                             (set! *alt-var-set-flag* #f)
                             (if *simple-history-stratflag* ;simpleh
                                 (begin
                                    (set! history (cons 'factor history)) )
                                 (if (null? *used-demodulators*)
                                     (set! history
                                          (list 'factor 0 history
                                             slow-list-ordnr lit-ordnr))
                                     (set! history
                                          (list 'factor 0 history
                                             slow-list-ordnr lit-ordnr
                                            (make-demod-history *used-demodulators*) ))))
                             ;(newline)(display "factor-unify successful:")
                             ;(newline)(debug-print-yvars 10)
                             ;(newline)(debug-print-xvars 10)

                             (set! res
                                (handle-derived-inf-cl-pair
                                   res history #t #f #f))

                             ; nb! res is now a info-fullclause pair
                             ; or #f, if it is not to be kept

                             ;(newline)(display "starting subsumed-clause:")
                             ;(newline)(debug-print-yvars 10)
                             ;(newline)(debug-print-xvars 10)

                             (cond
                               (*foundflag*
                                 ; empty clause derived
                                 (set! slow-cl-part *listofnull*)
                                 (set! cl-part *listofnull*) )
                               ((not res)) ; not to be kept
                               ; normal clause derived
                               ((subsumed-clause? fullclause (cdr res))
                                 ; a subsuming factor generated
                                 (set! *subs-factor-derived-flag* #t)
                                 (set! *fac-vars-set* #f)
                                 (set! *var-set-flag* #f)
                                 (set! *locfactorpool* '())
                                 (set! *factorpool* (cons res *factorpool*))
                                 (set! slow-cl-part *listofnull*)
                                 (set! cl-part *listofnull*) )
                               (else
                                 ; a nonsubsuming factor generated
                                 (set! *locfactorpool*
                                       (cons res *locfactorpool*) ))))

                           (begin
                             ;(newline) (display "factor-unify failed:")
                             ;(display *fac-vars-set*)
                             ;(display *var-set-flag*)
                             ;(display *alt-var-set-flag*)
                             ;(newline)(debug-print-yvars 10)
                             ;(newline)(debug-print-xvars 10)

                             (if (or *fac-vars-set* *var-set-flag*
                                     *alt-var-set-flag*)
                                 (begin
                                   (clear-x-vars-and-touched! maxvar)
                                   (set! *fac-vars-set* #f)
                                   (set! *var-set-flag* #f)
                                   (set! *alt-var-set-flag* #f) )))))))))))))

; merge-infos appends the ends (ie <orig-infos>-s) of the info at
; kept-infolist-place and removed-infolist-place. It then removes
; the info at removed-infolist-place and modifies the info at
; kept-infolist-place to contain the appended <orig-info>-s.

(define (merge-infos kept-infolist-place removed-infolist-place infolist)
  (let* ((new-orig (append (car kept-infolist-place)
                           (cddar removed-infolist-place) ))
         (res (cons '() '()))
         (resend res) )
    (do ((ipart infolist (cdr ipart)))
        ((null? ipart) (cdr res))
      (cond ((eq? ipart removed-infolist-place))
            ((eq? ipart kept-infolist-place)
               (set-cdr! resend (cons new-orig '()))
               (set! resend (cdr resend)) )
            (else
               (set-cdr! resend (cons (car ipart) '()))
               (set! resend (cdr resend)) )))))



;====================================================================
;
;              unique-answer-generation functions
;
;====================================================================

; we assume that if ans-inf is NOT #f, there is at least one answer-lit
; here

(define (make-unique-answer-pair pr history ans-inf)
  (let* ((alit1pl (member-if (lambda(x)(answer-lit? x)) (cdr pr)))
         (alit2pl (member-if (lambda(x)(answer-lit? x)) (cdr alit1pl))) )
    (cond
     ((not alit1pl) pr)
     ((not (ans-computable-lit? (car alit1pl))) '(() . #f))
     ((not alit2pl) pr)
     ((not (ans-computable-lit? (car alit2pl))) '(() . #f))
     ((term-list-equal-strong? (cdar alit1pl) (cdar alit2pl)) pr) ; rem later!
     (else
      (let* ((cl (cdr pr))
             (inf (car pr))
             (alit1inf '())
             (caseans '())
             (case-pr #f)
             (tmp-pr1 '())
             (tmp-pr '())
             (fact-pr #f) )
        ;- - - - - - - - - - - - - - - - - - - - - - - - - - - -
        ; creating the caseclause part
        ;- - - - - - - - - - - - - - - - - - - - - - - - - - - -
        (if (not (ans-computable-term? (car ans-inf) #t))
            (set! case-pr '(() . #f))
            (begin
              (set! tmp-pr1 (make-rest-inf-cl cl (car alit1pl) inf))
              (set! tmp-pr (make-rest-inf-cl
                            (cdr tmp-pr1) (car alit2pl) (car tmp-pr1) ))
              ; find the inf of anslit at alit1pl: alit1inf
              (do ((part cl (cdr part))
                   (infpart inf (cdr infpart)) )
                  ((eq? part alit1pl) (set! alit1inf (car infpart))) )
              ; construct the new answer-lit
              ; variables MUST be cleared!! CHECK DECO HERE!!!
              (set! caseans
                    (subst-deco-term-nodemodulate
                     (if (eq? (cadr alit1inf) (cadr ans-inf))
                         (list *ans-pred-nr*
                               (list *if-fun-nr* (car ans-inf)
                                     (cadar alit2pl) (cadar alit1pl) ))
                         (list *ans-pred-nr*
                               (list *if-fun-nr* (car ans-inf)
                                     (cadar alit1pl) (cadar alit2pl) )))
                     #t))
              (set! case-pr
                    (cons (cons alit1inf (car tmp-pr))
                          (cons caseans (cdr tmp-pr)) ))))
        ; - - - - - - - - - - - -- - - - - - - -- - - - - - - - -
        ; creating the factorized answers part
        ;- - - - - - - - - - - - - - - - - - - - - - - - - - -- -
        (if (factor-unify-ordinary (car alit1pl) (car alit2pl))
            (begin
              (set! fact-pr (subst-deco-cl
                                 (cdr tmp-pr1) (car tmp-pr1)
                                 #f *fac-vars-set*))))
        (if *fac-vars-set*
            (begin
              (clear-x-vars-and-touched! (caddr ans-inf))
              (clear-y-vars-and-touched! (cadddr ans-inf)) ))
        ; - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        ; handling the generated pairs
        ; - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        (cond ((and case-pr fact-pr)
                 (set! res (handle-derived-inf-cl-pair
                                  case-pr history #f #f #f))
                 fact-pr)
              (case-pr case-pr)
              (fact-pr fact-pr)
              (else '(() . #f)) ))))))


(define (several-answer-literals? purecl)
  (let ((nr 0))
    (do ((part purecl (cdr part)))
        ((or (null? part) (eq? nr 2)) (eq? nr 2))
      (if (answer-lit? (car part)) (set! nr (+ 1 nr))) )))

(define (ans-computable-lit? lit)
  (or (and (null? *ans-prohibited-funnrs*)
           (null? *ans-prohibited-prednrs*)
           (null? *ans-prohibited-constnrs*) )
      (ans-computable-term? (cadr lit) #f) ))

(define (ans-computable-term? x litflag)
  (if (pair? x)
      (if (eq? *if-fun-nr* (pure-fun (car x)))
          (and (ans-computable-term? (car x) #t)
               (ans-computable-term? (cadr x) #f)
               (ans-computable-term? (caddr x) #f) )
          (and (not (memq (pure-unsigned-fun (car x))
                          (if litflag
                              *ans-prohibited-prednrs*
                              *ans-prohibited-funnrs*)))
               (every? (lambda(y)(ans-computable-term? y #f))
                       (cdr x) )))
      (or (var? x)
          (not (memq x *ans-prohibited-constnrs*)) )))

;====================================================================
;
;              resolution-step functions
;
;====================================================================

; sort-clause sorts a pure clause according to the >-fun-ordering?
; if pair-flag is #t, then cdr-s of list els have to be considered.


(define (sort-clause clause pair-flag)
  (sort-clause-aux clause (length clause) pair-flag) )

(define (sort-clause-aux cl len pair-flag)
  (cond ((< len 2) cl)
        ((eq? len 2)
           (if (if pair-flag
                   (>-fun-ordering? (cadar (cdr cl)) (cadar cl))
                   (>-fun-ordering? (caadr cl) (caar cl)) )
               (list (cadr cl) (car cl))
               cl))
        (else
          (let* ((l1 (quotient len 2))
                 (l2 (- len l1))
                 (s (split-list cl l1)) )
            (sort-merge-clauses
               (sort-clause-aux (car s) l1 pair-flag)
               (sort-clause-aux (cdr s) l2 pair-flag)
               pair-flag)))))

(define (sort-merge-clauses cl1 cl2 pair-flag)
  (cond ((null? cl1) cl2)
        ((null? cl2) cl1)
        ((and pair-flag (>-fun-ordering? (cadar cl1) (cadar cl2)))
           (cons (car cl1)
                 (sort-merge-clauses (cdr cl1) cl2 #t) ))
        ((and (not pair-flag) (>-fun-ordering? (caar cl1) (caar cl2)))
           (cons (car cl1)
                 (sort-merge-clauses (cdr cl1) cl2 #f) ))
        (else
           (cons (car cl2)
                 (sort-merge-clauses cl1 (cdr cl2) pair-flag) ))))

;====================================================================
;
;  weighting, answer-literals, countable literals, predsymb bitlist
;
;====================================================================

; this lit-weight weighs all literals except answer-literals,
; which get a weight 0 in case *classical-synth-stratflag* is not set

(define (lit-weight lit)
  (if (and (not *classical-synth-stratflag*) (answer-pred? (car lit)))
      0
      (composite-weight lit) ))

(define (answer-pred? pred)
  (eq? *ans-pred-nr* (logand *signedfnsymb-mask* pred)) )

(define (countable-pred? pred)
  (if *classical-synth-stratflag*
      #t
      (not (answer-pred? pred)) ))

(define (answer-lit? lit)
  (eq? *ans-pred-nr* (logand *signedfnsymb-mask* (car lit))) )

; this composite-weight gives a size of the composite as its weight

(define (composite-weight comp)
  (let ((size (logand *term-size-mask* (car comp))))
    (if (eq? *term-size-mask* size) ; max (overfull) value?
        ; the encoded size was overfull
        (begin
          (set! size 1)
          (set! comp (cdr comp))
          (do ()
              ((null? comp) size)
            (let ((el (car comp)))
              (if (non-composite? el)
                  (set! size (+ 1 size))
                  (set! size (+ size (composite-weight el))))
              (set! comp (cdr comp)) )))
        ; the encoded size was not overfull
        (shiftright size *term-size-shift*) )))

(define (predsymb-bit pred)
  (vector-ref *predsymb-bit-vector*
              (shiftright pred *fnsymb-shift*) ))



;===================================================================
;
;              handling the derived clauses
;
;===================================================================

; make-raw-inf-cl-pair returns a pair of a newly generated
; 1-st level decorated clause and the info-list of its literals.
; The returned clause is not ordered. In case building the
; clause fails some 1-st level criterias, #f is returned.
;
; both xinf and yrestinf (both litinfolists) may be #f,
; in which case the infolist is generated in subst-deco-cl.
;
; as it uses subst-deco-cl, each el of the resulting litinfolist
; has the following modified form:
; (<changed-flag> <cl-nr-1-or-2> . <original>)
;
; It uses *xvars-set-flag* and *yvars-set-flag*


(define (make-raw-inf-cl-pair xcl ycl xinf yinf)
  (let ((tmp (make-raw-inf-cl-pair-aux xcl ycl xinf yinf)))
    ; (newline)(display "make-raw-inf-cl-pair ended: ")(display tmp)(newline)
    tmp))

(define (make-raw-inf-cl-pair-aux xcl ycl xinf yinf)

  ; (newline)(display "make-raw-inf-cl-pair started")(newline)
  ; first build up a new 1-st level deco new-yrest and new-yinf
  (let* ((new-yrest (subst-deco-cl ycl yinf 2 *yvars-set-flag*)))

    ;(newline)(display "new-yrest in pure form: ")(newline)
    ;(print-pure-clause (cdr new-yrest))(newline)
    ;(display "new-yinfo: ")(newline)
    ;(display (car new-yrest))(newline)

    ;(set! *used-demodulators* '())
    (if new-yrest
        ; new-yrest passed 1-st level criterias
        (let* ((new-xrest
                 (subst-deco-cl xcl xinf 1 *xvars-set-flag*) ))

          ;(newline)(display "new-xrest in pure form: ")(newline)
          ;(print-pure-clause (cdr new-xrest))(newline)
          ;(display "new-xinfo: ")(newline)
          ;(display (car new-xrest))(newline)

          ; (set! *used-demodulators* '())
          (if new-xrest
              ; new-xrest passed 1-st level criterias
              (cons (my-append! (car new-yrest) (car new-xrest))
                    (my-append! (cdr new-yrest) (cdr new-xrest)) )
              ; 1-st level criterias failed
              #f))
        #f)))





; make-rest-inf-cl takes a clause, lit in a clause and litinflist.
; It gives a clause and a litinflist where the lit and the corresponding
; inf is deleted. The result is a pair (<litinflist> . <cl>)
;
; litinflst may be given as #f
;
; lit MUST be in a clause (with eq? comparison)
;
; NB!!! The ordering of clause is changed (maybe it should be
; rewritten in the future?)

(define (make-rest-inf-cl cl lit inf)

  ;(newline)(display "make-rest-inf-cl enetred with cl, lit, inf:")(newline)
  ;(display cl)(newline)(display lit)(newline)(display inf)(newline)

  (if inf
      (if (null? (cdr cl))
          (cons '() '())
          (let ((passed-cl '())
                (passed-inf '()) )
            (do ((clrest cl (cdr clrest))
                 (infrest inf (cdr infrest)) )
                ((eq? (car clrest) lit)
                 (cons (my-append! passed-inf (cdr infrest))
                       (my-append! passed-cl (cdr clrest)) ))
              (set! passed-cl (cons (car clrest) passed-cl))
              (set! passed-inf (cons (car infrest) passed-inf)) )))
      (if (null? (cdr cl))
          (cons '() '())
          (let ((passed-cl '()))
            (do ((clrest cl (cdr clrest)))
                ((eq? (car clrest) lit)
                 (cons #f
                       (my-append! passed-cl (cdr clrest)) ))
              (set! passed-cl (cons (car clrest) passed-cl)) )))))

; subst-deco-cl takes a pure clause (part of one premiss)
; and applies the current
; substitution to it, building a 1-st level deco clause
; (where unchanged literals retain the old possibly wrong var-deco)
; and inf-list.
;
; It returns a pair (<info> . <cl>)
;
; If the built clause fails any 1-st level criterias,
; #f is returned.
; It uses var-set-flag (of the given premiss) and nr which is either
; 1 or 2 for binary resolution.
;
; vars-set-flag is true if some vars in the cl could be set
; in the current substitution.
;
; The info-list clinf may be given as #f, in which case it is handled
; as if it consisted entirely of ()-s. NB! In that case nr must be
; non-#f.
;
; The info-list has a format:
; (<changed-flag> <nr> . <orig-info>)
; where
; <changed-flag> is #t or #f (dep. on whether the lit is unchanged)
; <nr> is 1 or 2 (dep on which premiss the lit came from) or #f
;      in which case it is inherited from the original infolist
;      (useful for factorization)
; <orig-info> is the corresponding infolist el of the premiss literal.

(define (subst-deco-cl cl clinf nr vars-set-flag)
  ;(newline)(display "subst-deco-cl started: ")(newline)
  ;(display "  cl: ")(display cl)(newline)
  ;(display "  clinf: ")(display clinf)(newline)
  ;(display "  nr: ")(display nr)(newline)
  ;(display "  vars-set-flag: ")(display vars-set-flag)(newline)
  (if (null? cl)
      (cons '() '())
      (let* ((clres (cons '() '()))
             (clresend clres)
             (infres (cons '() '()))
             (infresend infres)
             (changed-flag #t)
             (lit '())
             (litinf '())
             (litinflstpart clinf) )

        ;(newline)(display "check1 in subst-deco-cl")(newline)

        (do ()
            ((null? cl)
               ;(newline)(display "subst-deco-cl ended! ")(newline)
               (if clres
                   (cons (cdr infres) (cdr clres))
                   #f))
          (set! lit (car cl))
          (if clinf (set! litinf (car clinf)) (set! litinf '()))

          ;(newline)(display "check10 in subst-deco-cl")(newline)

          (if (or (and vars-set-flag
                       (term-may-cont-vars? lit) )
                   *new-demodulators-flag*)
              (if (answer-pred? (car lit)) ; no demodulation of answers
                  (begin
                    (set! lit (subst-deco-term-nodemodulate lit #t))
                    (set! changed-flag #t) )
                   (begin
                    (set! lit (subst-deco-term lit #t))
                    (set! changed-flag #t) ))
              (begin
                (set! lit (copy-nonground-term lit))
                (set! changed-flag #f) ))

          ;(newline)(display "check20 in subst-deco-cl")(newline)

          (if lit
              (begin
                (set-cdr! clresend
                          (cons lit '()) )
                (set! clresend (cdr clresend))
                (if nr
                    (set-cdr! infresend
                              (cons (cons changed-flag (cons nr litinf))
                                    '() ))
                    (set-cdr! infresend
                              (cons (cons changed-flag (cdr litinf))
                                    '() )))
                (set! infresend (cdr infresend))
                (set! cl (cdr cl))
                (if clinf (set! clinf (cdr clinf))) )
              (begin
                (set! cl '())
                (set! clres #f) ))))))

; tautology-pure-cl? takes a pure clause (lit list) and returns #t or #f

(define (tautology-pure-cl? list)
  (if (or (null? list)
          (null? (cdr list)) )
      #f
      (if (member-polarity-lit #f (car list) (cdr list))
          #t
          (tautology-pure-cl? (cdr list)) )))

; remove-double-lits! removes double lits and corresponding
; infos from the list of lits and the list of infos it takes.
;
; if pos-only-flag is #t, remove only duplicates of positive literals
; it gives back a pair (new-inf-list .  new-lit-list).
; in case *neg-hyper-strategy-flag* is #t, remove duplicates of
; negative only, instead

(define (remove-double-lits! inf-list lit-list pos-only-flag)
  (if (null? lit-list)
      (cons inf-list lit-list)
      (let ((lit-list-part lit-list)
            (inf-list-part inf-list)
            (tmp '()) )
      (do ()
          ((null? (cdr lit-list-part))
             (if (and (or (not pos-only-flag)
                          (if (not *neg-hyper-strategy-flag*)
                              (not(neg-fun? (caar lit-list)))
                              (neg-fun? (caar lit-list)) ))
                      (member-polarity-lit #t (car lit-list) (cdr lit-list)) )
                 (cons (cdr inf-list) (cdr lit-list))
                 (cons inf-list lit-list) ))
        (set! tmp (cdr lit-list-part))
        (if (and (or (not pos-only-flag)
                     (if (not *neg-hyper-strategy-flag*)
                         (not(neg-fun? (caar tmp)))
                         (neg-fun? (caar tmp)) ))
                 (member-polarity-lit #t (car tmp) (cdr tmp)) )
            (begin
              (set-cdr! lit-list-part (cdr tmp))
              (set-cdr! inf-list-part (cddr inf-list-part)) )
            (begin
              (set! lit-list-part tmp)
              (set! inf-list-part (cdr inf-list-part)) ))))))

; unit-handle-cl! checks the discrimination tree for all of the changed
; literals in the clause. It checks whether:
; 1) the lit is subsumed by a unit clause; then the whole clause is subsumed;
;    #f is returned, *subsuming-clause* is set to the full subsumer
; 2) there is a complementary unit clause subsuming the lit; then
;    the lit is cut off (lit-list and inf-list are destructively changed).
;    NB! A new answer-literal may be introduced here!
; 3) the lit is ground and there is some equal lit (either directly
;    or complementary) derived earlier (that is, is it discr tree?);
;    then the ground lit or its cdr is destructively replaced by the
;    already-existing one. The *discr-tree* itself may also changed,
;    as the ground lit is inserted if it was not there before.
;    NB!!! In case the ground literal was there with NEITHER polarity,
;    nothing is inserted.
;
;  in case the clause is not subsumed, the triple
; (<history> . (<inf-list> . <lit-list>)) with possibly shorter or changed
; <history>, <inf-list> and <lit-list> are returned.
;
; any full unit clauses used for cutting off lits are put into
; the list *cut-clauses* (which is set to () at the beginning).
;

(define (unit-handle-cl! inf-list lit-list history)

  ;(newline)(display "unit-handle-cl! entered with:")(newline)
  ;(debug-print-pure-clause lit-list)(newline)
  ;(pretty-print history)
  ;(debug-print-yvars 5)
  ;(debug-print-xvars 5)
  ;(debug-print-zvars 5)
  ;(newline)

  (if (null? lit-list)
      (cons history (cons inf-list lit-list))
      (let* ((orig-lit-list (copy-list lit-list))
             (lit-list (cons '() lit-list))
             (inf-list (cons '() inf-list))
             (lit-list-part lit-list)
             (inf-list-part inf-list)
             (subsume-flag #f)
             (foreign-checkres-flag #f)
             (cut-flag #f) )
        (set! *cut-clauses* '())
        (set! *used-demodulators* '())
        (do ((litnr 1 (+ 1 litnr)))
          ((or (null? (cdr lit-list-part))
               subsume-flag)

             ;(newline)(display "unit-handling finished")(newline)

             (if subsume-flag
                 ; clause subsumed
                 (begin
                   (set! *forward-subsumed-unit-nr*
                         (+ 1 *forward-subsumed-unit-nr*) )
                   (cond ((>= *display-info-level* 80)
                            (newline)
                            (display
                              "the derived clause subsumed by unit:")
                            (newline)
                            (print-clause *subsuming-clause*)
                            (newline) )
                         (*display-info-char-flag*
                             (display "_") ))
                   #f)
                 (begin
                    (if (not (null? *cut-clauses*))
                        (set! *simple-unit-cutoff-nr*
                              (+ 1 *simple-unit-cutoff-nr*) ))
                    (if (and (not (null? *cut-clauses*))
                             (>= *display-info-level* 80) )
                        (begin (newline)
                               (display "discr tree used for unit cutoff:")
                               (newline)
                               (for-each (lambda(x)
                                           (print-clause x)(newline))
                                         *cut-clauses*)))
                    (cons history (cons (cdr inf-list) (cdr lit-list))) )))
        (set! cut-flag #f)
        (set! foreign-checkres-flag #f)
        (let* ((tmp-lit-lst (cdr lit-list-part))
               (lit (car tmp-lit-lst))
               (tmp-inf-list (cdr inf-list-part))
               (inf (car tmp-inf-list))
               (checkres #f) )
         ;(newline)
         ;(display "check-discr-tree to be entered with:")(newline)
         ;(print-lit lit)(newline)
         ;(show-discr-tree *discr-tree* #t)
         ;(newline)

         ; check for every literal, even unchanged ones, except
         ; answer-literals

          (if (answer-lit? lit)
              (set! checkres #f)
              (set! checkres (check-discr-tree *discr-tree* lit)) )

         ;(if (car inf) ; <changed-flag>
         ;     ; lit has been changed
         ;     (set! checkres (check-discr-tree *discr-tree* lit))
         ;     ; lit is unchanged, no checking necessary
         ;     (set! checkres #f) )

        ; (newline)
        ; (display "check-discr-tree exited with discr-tree remaining:")
        ; (newline)
        ; (show-discr-tree *discr-tree* #t)
        ; (newline)
        ; (newline)
        ; (display "check-discr-tree gave a checkres:")(newline)
        ; (display checkres)(newline)
        ; (newline)
        ; (show-discr-tree checkres #f)
         ;(read-char)

         ; if checking failed, try a number of alternative check-trees
         ; for cutoff only
         (if (not (answer-lit? lit))
             (do ((part *discr-tree-list* (cdr part)))
                 ((or (null? part)
                      checkres)
                   (if checkres (set! foreign-checkres-flag #t)) )
               (set! checkres (check-discr-tree (car part) lit))
;               (if checkres
;                       (begin
;                              (if (and
;                                   (not (null? (cdr checkres)))
;                                   (not (null? (cddr checkres))) )
;                                  (begin
;                                    (display "checkres: ")
;                                    (print-clause (caddr checkres))
;                                    (newline)
;                                    (print-lit lit)
;                                    (newline)
;                                    (if (not (null? (cddr lit-list)))
;                                        (print-pure-clause (cdr lit-list)) )
;                                    (newline) ))
;                              (newline) ))
               ; only cutoff allowed here, no subsumption
               (if (not (and checkres
                             (not (null? (cdr checkres)))
                             (not (null? (cddr checkres)))
                             (not (eq? (logand *sgn-mask* (car lit))
                                       (logand *sgn-mask*
                                               (caadr (caddr checkres)) )))))
                   (set! checkres #f) )))
                   ; (begin (display "old cutoff")(newline))

         ;(display checkres)
         ;(newline)
         ;(if (and checkres
         ;          (not (null? (cdr checkres)))
         ;          (not (null? (cddr checkres))) )
         ;    (begin (print-clause (caddr checkres))) )

         ; if checking failed and the pred was commutative, try again!
         (if (and (not checkres)
                  (commutative-fun? (car lit)) )
             (set! checkres
                   (check-discr-tree *discr-tree*
                                     (cons (car lit)
                                           (cons (caddr lit)
                                                 (cons (cadr lit)
                                                       (cdddr lit) ))))))
         (if checkres
              (let* ((rescdr (cdr checkres))
                     (rescddr (if (null? rescdr) '() (cdr rescdr))) )
                (if (null? rescddr)
                    ; the found one is a ground lit with no assoc unit clause
                    ; if it is changed and there only with the wrong polarity,
                    ; insert the new one with the right polarity.
                    (if (car inf) ; changed-flag
                        (if (neg-fun? (car lit))
                            ; the lit is negative
                            ; (and ground, as it is nonunit)
                            (if (or (null? rescdr)
                                    (null? (car rescdr)) )
                                ; no negative one in leaf
                                (begin (set-cdr! lit (cdar checkres))
                                       (set-cdr! checkres
                                                 (cons lit rescddr) ))
                                ; a negative one in leaf
                                (set-car! tmp-lit-lst (car rescdr)) )
                            ; the lit is positive
                            ; (and ground, as it is nonunit)
                            (if (null? (car checkres))
                                ; no positive one in leaf
                                (begin (set-cdr! lit (cdar rescdr))
                                       (set-car! checkres lit) )
                                ; a positive one in leaf
                                (set-car! tmp-lit-lst (car checkres)) )))
                    ; the found one has an associated unit clause
                    (let* ((leafcl (car rescddr))
                           (leafclpure (cdr leafcl))
                           (leafclinf (car leafcl)) )
                      (if (eq? (logand *sgn-mask* (car lit))
                               (logand *sgn-mask* (caar leafclpure)) )
                          ; same polarity unit; subsume if allowed!
                          (if (or *discr-tree-subsume-stratflag*
                                  (not (null? (cdr orig-lit-list))) )
                              (begin
                                (set! subsume-flag #t)
                                (set! *subsuming-clause* leafcl) ))
                          ; other polarity unit; cut off the lit!
                          ; special handling for hyperresolution
                          (begin
                            (if *hyper-strategy-flag*
                                ; for hyperresolution disallow cutting off
                                ; positive literals from nucleons
                                (if (or (neg-fun? (car lit))
                                        (positive-cl? orig-lit-list) )
                                    (begin
                                       (set! cut-flag #t)
                                       (set! *cut-clauses*
                                                   (cons leafcl *cut-clauses*) )))
                                (begin
                                  (set! cut-flag #t)
                                  (set! *cut-clauses*
                                        (cons leafcl *cut-clauses*) )))))))))

          ;(newline)
          ;(display "near the unit check loop end: ")
          ;(display "subsume-flag: ")(display subsume-flag)(newline)
          ;(display "cut-flag: ")(display cut-flag)(newline)
          ;(display "lit-list-part: ")(display lit-list-part) (newline)

          (if (not subsume-flag)
              (if cut-flag
                  (begin
                    (set-cdr! lit-list-part (cdr tmp-lit-lst))
                    (set-cdr! inf-list-part (cdr tmp-inf-list))
                    (set! history
                       (append (list 'cut (clinfo-clnr (caar *cut-clauses*)))
                               history)))
                  (begin
                    (set! lit-list-part tmp-lit-lst)
                    (set! inf-list-part tmp-inf-list) ))))))))


; full-unit-cutoff takes inflist and a full clause.
;
; It returns a pair of a new inflist and a full clause.
;
; any full unit clauses used for cutting off lits are put into
; the list *cut-clauses* (which is not set to () at the beginning!).



(define (full-unit-cutoff! inf-list fullcl)
  (let* ((clauseinf (car fullcl))
         (lit-list (cdr fullcl))
         (len 10000) ; a large never-reached integer
         (res '())
         (clhistsublist (clinfo-history-sublist clauseinf))
         (distance (cadr clhistsublist))
         (history (car clhistsublist)) )
    ; full cutoff with doubles
    (set! len (length lit-list))
    (if (and (not *int-logic-flag*)
             (or (not *hyper-strategy-flag*)
                 *simple-history-stratflag*
                 (positive-cl? lit-list)))
        (begin
          (set! res
                (full-unit-cutoff-double! clauseinf inf-list lit-list history))
          (set! history (car res))
          (set! inf-list (cadr res))
          (set! lit-list (cddr res)) ))
    (if (not (eq? len (length lit-list)))
        ; cutoff was a success
        (begin
          (set! *double-unit-cutoff-nr*
                (+ 1 *double-unit-cutoff-nr*) )
          (set! clauseinf (make-clause-info lit-list history distance))
          (set! fullcl
                (make-kept-clause clauseinf (cons inf-list lit-list)) )
          (set! clauseinf (car fullcl))
          (set! lit-list (cdr fullcl)) ))
    ; unit cutoff cycle starts
    (set! len 10000) ; large never-reached integer
    (do ()
        ((or (eq? len (length lit-list)) (empty-cl? lit-list))
           (for-each rename-y-to-x-term! lit-list)
           (cons inf-list (cons clauseinf lit-list)) )
      (set! len (length lit-list))
      (for-each  rename-x-to-y-term! lit-list)
      ; full cutoff with units
      (set! res (full-unit-cutoff-aux! clauseinf inf-list lit-list history))
      ;(display res)
      (set! history (car res))
      (set! inf-list (cadr res))
      (set! lit-list (cddr res))
      (cond ((empty-cl? lit-list) ; empty clause?
              (set! clauseinf (list 0 0 '() 0 0 history))
              (if (not (null? lit-list))
                  (begin
                    (set! clauseinf (make-clause-info lit-list history distance))
                          (set! fullcl
                      (make-kept-clause clauseinf (cons inf-list lit-list)) )
                    ; (print-pure-clause (cdr fullcl))
                    (set! lit-list (cdr fullcl))  )))
            ((eq? len (length lit-list))) ; clause unchanged?
            (else
              ; cutoff was a success! Build a new clause.
              (set! *full-unit-cutoff-nr*
                    (+ 1 *full-unit-cutoff-nr*) )
              (set! clauseinf (make-clause-info lit-list history distance))
              (set! fullcl
                    (make-kept-clause clauseinf (cons inf-list lit-list)) )
              (set! clauseinf (car fullcl))
              (set! lit-list (cdr fullcl)) )))))


; full-unit-cutoff-aux! performs a single pass over literals
; in a clause, cutting them off with units if possible


(define (full-unit-cutoff-aux! clauseinf inf-list lit-list history)

  ;(newline)
  ;(display "full-unit-cutoff! entered with :")
  ;(print-pure-clause lit-list)(newline)

  (if (or (empty-cl? lit-list)
          (null? *unit-alist*) )
      (cons history (cons inf-list lit-list))
      (let* ((orig-lit-list (copy-list lit-list))
             (lit-list-part lit-list)
             (inf-list-part inf-list)
             (yclmaxvar (give-cl-maxvar clauseinf))
             (lit-list (cons '() lit-list))
             (inf-list (cons '() inf-list))
             (lit-list-beg lit-list)
             (inf-list-beg inf-list)
             (has-only-local-only-vars-flag #f)
             (unit-flag (unit-cl? lit-list-part))
             (cut-flag #f) )
      ; loop over literals in the clause
         (do ((ylitnr 1 (+ 1 ylitnr)))
          ((null? lit-list-part)
             (if (and (not (null? *cut-clauses*))
                      (>= *display-info-level* 65) )
                 (begin (newline)
                        (display
                          "proper unit cutoff with the following unit(s):")
                        (newline)
                        (for-each (lambda(x)
                                    (print-clause x)(newline) )
                                  *cut-clauses*)))

             ;(newline)(display "full-unit-cutoff exits with:")
             ;(print-pure-clause (cdr lit-list-beg))(newline)

             (cons history (cons (cdr inf-list-beg) (cdr lit-list-beg))) )
        (let* ((ylit (car lit-list-part))
               (yinf (car inf-list-part))
               (discr-testclause '())
               (commylit '())
               (yfun (car ylit))
               (tmp '())
               (cutflag #f))

          ; normal unit cutoff

          (if (commutative-fun? yfun)
              (set! commylit (cons yfun
                                   (cons (caddr ylit)
                                         (cons (cadr ylit) (cdddr ylit)) ))))

          (if (or cutflag
                  (answer-lit? ylit)
                  (and *hyper-strategy-flag*
                       (not *simple-history-stratflag*) ; simpleh
                       ; for hyperresolution disallow cutting off
                       ; positive literals from nucleons
                       (not (neg-fun? yfun))
                       (not (positive-cl? orig-lit-list)) )
                  (not (has-some-local-only-vars? ylit)) )
              ; either answer-literal or a positive lit in a nucleus
              ; for hyperresolution or just
              ; not cuttable (since this would have been done by discr-tree)
              (begin
                (set! lit-list lit-list-part)
                (set! inf-list inf-list-part)
                (set! lit-list-part (cdr lit-list-part))
                (set! inf-list-part (cdr inf-list-part)) )
              ; loop over unit clauses, trying to cut
              (let* ((compl-yfunpure (logxor *sgn-mask* (pure-fun yfun)))
                     (cand-list (assq compl-yfunpure *unit-alist*)) )
                (if cand-list
                    (set! cand-list (cdr cand-list))
                    (set! cand-list '()) )
                (set! has-only-local-only-vars-flag
                      (or unit-flag
                          (has-only-local-only-vars?  ylit) ))
                (do ((part cand-list (cdr part)))
                    ((or (null? part) cutflag)
                       (if cutflag
                           ; cutter found
                           (begin
                             (set! lit-list-part (cdr lit-list-part))
                             (set! inf-list-part (cdr inf-list-part))
                             (set-cdr! lit-list lit-list-part)
                             (set-cdr! inf-list inf-list-part)
                             (if *simple-history-stratflag* ; simpleh
                                (begin
                                  (set! history
                                        (append
                                          (list 'slowcut (clinfo-clnr (caar *cut-clauses*)))
                                          history)) )
                                 (if (or *hyper-strategy-flag*
                                     ; *type-theory-stratflag*
                                     *int-logic-flag*)
                                    (set! history
                                      (make-hyper-history ;simpleh
                                           (cdar *cut-clauses*) orig-lit-list
                                           '() '()
                                           (car (clinfo-history-sublist
                                                  (caar *cut-clauses*) ))
                                           history
                                           (list (put-litinfo-ordernr 1 0))
                                           (cddr yinf)
                                           1 ylitnr #f))
                                   (set! history
                                      (make-binres-history ;simpleh
                                            (cdar *cut-clauses*) orig-lit-list
                                            '() '()
                                         (car(clinfo-history-sublist
                                              (caar *cut-clauses*) ))
                                         history
                                         1 ylitnr #f)))))
                           ; cutter not found
                           (begin
                             (set! lit-list lit-list-part)
                             (set! inf-list inf-list-part)
                             (set! lit-list-part (cdr lit-list-part))
                             (set! inf-list-part (cdr inf-list-part)) )))

                  ;(newline)(print-clause (car part))

                  (let* ((xlit (car (cdar part)))
                         (xfun (car xlit)) )
                    (if (and (or has-only-local-only-vars-flag
                                 (fun-may-cont-vars? xfun) )
                             (null? (cdr (cdar part))) ) ; has no ans-lit?
                        (begin

                          (if (>  *display-info-level* 110)
                              (begin
                          (newline)
                          (display "starting resolve-unify-cutoff with:")
                          (newline)
                          (debug-print-lit xlit)(newline)
                          (debug-print-lit ylit)(newline)
                          (newline)
                          (do ((i 1 (+ 1 i)))
                              ((> i 10))
                            (display i)(display ":")
                            (display (var-ref i))(display " ") )
                          (newline)
                          (do ((i (+ 1  *max-x-var*)  (+ 1 i)))
                              ((> i (+ 11 *max-x-var*)))
                            (display i)(display ":")
                            (display (var-ref i))(display " ") )
                          (newline) ))

                          (if (resolve-unify-cutoff xfun yfun xlit ylit)
                              (begin
                                (set! *cut-clauses*
                                      (cons (car part) *cut-clauses*) )
                                (set! cutflag #t) )
                              (if (not (null? commylit))
                                  (begin
                                    (if *yvars-set-flag*
                                        (clear-y-vars! yclmaxvar) )
                                    (if *xvars-set-flag*
                                        (clear-x-vars!
                                         (give-cl-maxvar (caar part)) ))
                                    (if (resolve-unify-cutoff
                                         xfun yfun xlit ylit)
                                        (begin
                                          (set! *cut-clauses*
                                                (cons (car part)
                                                      *cut-clauses*))
                                          (set! part *listofnull*) )))))
                          (if *yvars-set-flag* (clear-y-vars! yclmaxvar))
                          (if *xvars-set-flag*
                              (clear-x-vars!
                               (give-cl-maxvar (caar part)) )))))))))))))



; full-unit-cutoff-double! performs a single pass over literals
; in a clause, cutting them off with doubles if possible


(define (full-unit-cutoff-double! clauseinf inf-list lit-list history)

  ;(newline)
  ;(display "full-unit-cutoff-double! entered with :")
  ;(print-pure-clause lit-list)(newline)

  (if (or (empty-cl? lit-list)
          (null? (cdr lit-list)) ); not for units!
      (cons history (cons inf-list lit-list))
      (let* ((orig-lit-list (copy-list lit-list))
             (lit-list-part lit-list)
             (inf-list-part inf-list)
             (yclmaxvar (give-cl-maxvar clauseinf))
             (lit-list (cons '() lit-list))
             (inf-list (cons '() inf-list))
             (lit-list-beg lit-list)
             (inf-list-beg inf-list)
             (unit-flag (unit-cl? lit-list-part))
             (passed-lits '())
             (cut-flag #f) )
      ; loop over literals in the clause
         (do ((ylitnr 1 (+ 1 ylitnr)))
          ((null? lit-list-part)
             (if (and (not (null? *cut-clauses*))
                      (>= *display-info-level* 65) )
                 (begin (newline)
                        (display
                          "double cutoff with the following cls:")
                        (newline)
                        (for-each (lambda(x)
                                    (print-clause x)(newline) )
                                  *cut-clauses*)))

             ;(newline)(display "full-unit-cutoff exits with:")
             ;(print-pure-clause (cdr lit-list-beg))(newline)

             (cons history (cons (cdr inf-list-beg) (cdr lit-list-beg))) )
        (let* ((ylit (car lit-list-part))
               (yinf (car inf-list-part))
               (discr-testclause '())
               (yfun (car ylit))
               (tmp '())
               (cutflag #f))

          ; cutoff with double clauses in the nonunit discr tree

          (if (or (answer-lit? ylit)
                  (and *hyper-strategy-flag*
                       ; for hyperresolution disallow cutting off
                       ; positive literals from nucleons
                       (not (neg-fun? yfun))
                       (not (positive-cl? orig-lit-list)) ))
              ; either answer-literal or a positive lit in a nucleus
              ; for hyperresolution or just
              ; not cuttable (since this would have been done by discr-tree)
              (begin
                (set! passed-lits (cons ylit passed-lits))
                (set! lit-list lit-list-part)
                (set! inf-list inf-list-part)
                (set! lit-list-part (cdr lit-list-part))
                (set! inf-list-part (cdr inf-list-part)) )
              ; try to cut with doubles
              (begin
                (set! discr-testclause
                      (cons (cons (logxor *sgn-mask* (pure-fun yfun))
                                  (cdr ylit))
                          (cons
                           '()
                           (append passed-lits (cdr lit-list-part)) )))
                (set! *nonunit-discr-flag* #t)
                (set! *overdouble-discr-flag* #f)
                (if (null? (car *nonunit-discr-tree*))
                    (set! tmp #f)
                    (set! tmp (check-discr-tree-part
                               (cdaar *nonunit-discr-tree*)
                               discr-testclause
                               '() )))
                (set! *nonunit-discr-flag* #f)
                (set! *overdouble-discr-flag* #f)
                (cond
                 (tmp ; current literal cut off
                  (set! cutflag #t)
                  (set! *cut-clauses* (cons (car tmp) *cut-clauses*))
                  (set! lit-list-part (cdr lit-list-part))
                  (set! inf-list-part (cdr inf-list-part))
                  (set-cdr! lit-list lit-list-part)
                  (set-cdr! inf-list inf-list-part)
                  (if *simple-history-stratflag* ;simpleh
                      (set! history
                            (append
                                 (list 'binarycut (clinfo-clnr (caar *cut-clauses*)))
                                  history))
                     (if (or *hyper-strategy-flag*
                          ; *type-theory-stratflag*
                          *int-logic-flag*)
                        (set! history
                               (make-hyper-history ;simpleh
                                (cdar *cut-clauses*) orig-lit-list
                                '() '()
                                (car (clinfo-history-sublist
                                      (caar *cut-clauses*) ))
                                history
                                (list (put-litinfo-ordernr 1 0))
                                (cddr yinf)
                                1 ylitnr #f))
                         (set! history ;simpleh
                               (make-binres-history
                                (cdar *cut-clauses*) orig-lit-list
                                '() '()
                                (car(clinfo-history-sublist
                                     (caar *cut-clauses*) ))
                                history
                                1 ylitnr #f)))))
                 (else ; cutter not found
                  (set! passed-lits (cons ylit passed-lits))
                  (set! lit-list lit-list-part)
                  (set! inf-list inf-list-part)
                  ;(newline)
                  ;(display "----")
                  ;(newline)
                  ;(display part)
                  ;(newline)
                  ;(display lit-list-part)
                  ;(newline)
                  ;(display inf-list-part)
                  (set! lit-list-part (cdr lit-list-part))
                  (set! inf-list-part (cdr inf-list-part)) )))))))))


(define (has-some-local-only-vars? cterm)
  (let ((success-flag #f)
        (el '()) )
    (if (fun-may-cont-vars? (car cterm))
        (do ((part (cdr cterm) (cdr part)))
            ((or (null? part) success-flag)
               success-flag)
          (set! el (car part))
          (if (composite? el)
              (if (has-some-local-only-vars? el)
                  (set! success-flag #t) )
              (if (and (non-composite:-var? el)
                       (not (zero? (logand *no-occ-otherlit-var-mask* el))) )
                  (set! success-flag #t) )))
        #f)))

(define (has-only-local-only-vars? cterm)
  (let ((fail-flag #f)
        (el '()) )
    (if (fun-may-cont-vars? (car cterm))
        (do ((part (cdr cterm) (cdr part)))
            ((or (null? part) fail-flag)
               (not fail-flag) )
          (set! el (car part))
          (if (composite? el)
              (if (not (has-only-local-only-vars? el))
                  (set! fail-flag #t) )
              (if (and (non-composite:-var? el)
                       (zero? (logand *no-occ-otherlit-var-mask* el)) )
                  (set! fail-flag #t) )))
        #t)))

; member-polarity takes a samesgn/othersign flag, a lit and a lit list.

(define (member-polarity-lit samesgn-flag lit lst)
  (if (null? lst)
      #f
      (let* ((fun (if samesgn-flag
                      (logand *fun-except-varinf-mask* (car lit))
                      (logand *fun-except-varinf-mask*
                              (logxor *sgn-mask* (car lit)) )))
             (comm-flag (commutative-fun? fun))
             (res #f) )
        (do ()
            ((null? lst) res)
          (if (and (eq? fun
                        (logand *fun-except-varinf-mask* (caar lst)) )
                   (or (term-list-equal-strong? (cdr lit) (cdar lst))
                       (and comm-flag
                            (swapped-term-list-strong-equal?
                               (cdr lit)(cdar lst)) )))
              (begin
                (set! res lst)
                (set! lst '()) )
              (set! lst (cdr lst)) )))))


(define (swapped-term-list-strong-equal? x y)
  (term-list-equal-strong?
    (cons (cadr x) (cons (car x) (cddr x)))
    (cons (cadr y) (cons (car y) (cddr y))) ))


;(define (keep-clause fullcl)
;  (let* ((strengthint (caar fullcl))
;         (len (shiftright (logand *cl-len-mask* strengthint)
;                          *cl-len-shift*))
;         (strength (logand *strength-mask* strengthint)) )
;    (set! *sos-clauses-list* (cons fullcl *sos-clauses-list*))
;    (insert-to-subsumption-list! len strength fullcl) ))

; the following is a completely "ignorant" version of inserting
; the clause.

(define (insert-to-subsumption-list! len strength fullcl)

  ;(newline)(display "insert-to-subsumption-list! enetred with:")(newline)
  ;(display len)(newline)
  ;(display strength)(newline)
  ;(display fullcl) (newline)

  (let ((lenlist '())
        (place '()) )
    ; the first do-cycle finds a sublist lenlist with the right length
    ; if such one does not exist, it is created.
    (do ((part *subsumption-list* (cdr part)))
        ((or (null? (cdr part)) (>= (caadr part) len))
           (if (and (not (null? (cdr part)))
                    (= (caadr part) len) )
               (set! lenlist (cadr part))
               (begin
                 (set-cdr! part (cons (list len) (cdr part)))
                 (set! lenlist (cadr part)) ))))
    ; now lenlist is the right sublist.
    ; pass it until a clause with a smaller or equal strength is met.

    ;(display lenlist)

    (set! place (cdr lenlist))
    (do ()
        ((or (null? place)
             (<= (logand *strength-mask* (caar (car place)))
                 strength))
           (set-cdr! lenlist (cons fullcl place)) )
      (set! lenlist place)
      (set! place (cdr place)) )))

; the following takes a right place to insert fullcl and
; inserts it there.
; The right place is found beforehand by the subsumption algorithm.
; if place = '(), then lenlists is the right place to insert
; a new lenlist.
;
;(define (insert-to-subsumption-place! lenlists place len fullcl)
;  (if (null? place)
;      (set-cdr! lenlists (cons (list len fullcl) (cdr lenlists)))
;      (set-cdr! place (cons fullcl (cdr place))) ))

; insert-unit-alist inserts a unit fullclause to the
; alist of all units (with predicate symbs as keys).

(define (insert-unit-alist fullcl)
  (let* ((purepred (pure-fun (caadr fullcl)))
         (part (assq purepred *unit-alist*)) )
    (if part
        (set-cdr! part (cons fullcl (cdr part)))
        (set! *unit-alist*
              (cons (list purepred fullcl) *unit-alist*) ))))


;===================================================================
;
;                      partial discrimination tree
;
;==================================================================
;
;  Partial discrimination tree discriminates only upon the ground
;  components of the term, up to a certain predetermined depth.
;
;  The structure of the partial discr tree is similar to the standard
;  tree (difference: each var is just a 0):
;
; each node-level is built of three sublists for funs, consts and vars:
; (funs consts . vars), and in case there are no consts nor vars
; it is just (funs).
;
; level_i:  ( ((f1 level_i+1) (f2 level_i+1) ...)    funs
;             ((c1 level_i+1) (c2 level_i+1) ...)    consts
;             (0 level_i+1))  all vars
;
; in order to insert a literal to a partial discr-tree:
; call insert-partial-discr-tree! with:
; tree: full partial discr-tree
; term: literal
; funflag: #f (literal is not a fun)
; endflag #t (the single (thus the last) object)
; depth: max depth of indexing
;
; NB!!!!! all function symbs must be only 7-bit, since we use
; *predsymb-mask* here to not have a special treatment for
; literals/nonliteral-terms (preds must have the sgn turned down here!)
;
; and the following globals must be set (used by insert-discr-tree-leaf! ):
; *cur-lit*: inserted literal
; *cur-clause*: if you have a unit clause (containing inserted literal)
;               then the full inserted unit clause
;               else #f

(define (insert-partial-discr-tree! tree term funflag endflag depth)
  (if (non-composite? term)
      ; term is non-composite
      (let ((place '()))
        (cond (funflag
                 (set! term (logand *predsymb-mask* term))
                 (set! place (assq term (car tree)))
                 (if (not place)
                     (begin (set-car! tree
                                      (cons (list term '()) (car tree)) )
                            (set! place (caar tree)) )))
              ((non-composite:-var? term)
                 (set! term 0) ; NB! special var: 0
                 (if (null? (cdr tree)) (set-cdr! tree (list '())))
                 (set! tree (cdr tree))
                 (set! place (assq term (cdr tree)))
                 (if (not place)
                     (begin (set-cdr! tree
                                      (cons (list term '()) (cdr tree)) )
                            (set! place (cadr tree)) )))
              (else
                 (if (null? (cdr tree)) (set-cdr! tree (list '())))
                 (set! tree (cdr tree))
                 (set! place (assq term (car tree)))
                 (if (not place)
                     (begin (set-car! tree
                                      (cons (list term '()) (car tree)) )
                            (set! place (caar tree)) ))))
        (if endflag
            (insert-partial-discr-tree-leaf! (cdr place))
            (cdr place) ))
      ; term is composite
      (begin
        (set! tree (insert-partial-discr-tree!
                      tree (car term) #t #f depth))
        (set! term (cdr term))
        (do ()
            ((null? term)
               tree)
          (set! tree (insert-partial-discr-tree! tree (car term) #f
                                         (and endflag (null? (cdr term)))
                                         (- depth 1) ))
          (set! term (cdr term)) ))))


;===================================================================
;
;                      full discrimination tree
;
;==================================================================

; principles of unit/nonunit discrimination for clauses with
; answer-literals.
;
; 1) subsumption. Any clause containing answer-literal
;    subsumes another clause iff either
;    * it subsumes with ans-lit taken as an ordinary lit
;    * the flag *all-computable-flag* is #t
;    * the subsumed clause contains an answer-lit and the
;      answer-lit in the subsuming clause is ground.
;
; 2) cutting units with subsumption (complementary subsumption).
;    A unit clause containing an answer-literal
;    may be used for the unit cut with subsumption iff either
;    * the flag *all-computable-flag* is #t
;    * the clause with the lit cut off contains an ans-lit and the
;      cutting clause contains a ground ans-lit.

; check-discr-tree takes a full discr tree and a literal.
;
; it returns a leaf of the form: (<pos-l> <neg-l> <full-unit-cl>)
; or some first segment of it.
; <pos-l> and <neg-l> are normally assumed to be ground literals.
;
; if there are different possible leaves, one without <full-unit-cl>
; and one with <full-unit-cl>, then always the second one is returned.


(define (check-discr-tree tree lit)
  (let ((fun (car lit))
        (res #f) )
    (set! *ground-check-res* #f)
    (set! *nonunit-discr-flag* #f)
    (set! tree (assq (logand *predsymb-mask* fun) (car tree)))
    (if tree
        (set! res (check-discr-tree-part (cdr tree) (cdr lit) '()))
        (set! res #f) )
    (if (and (not res)
             (commutative-fun? fun) )
        (set! res
              (check-discr-tree-part
                (cdr tree)
                (cons (caddr lit) (cons (cadr lit) (cdddr lit)))
                '() )))
    (if res res        *ground-check-res*) ))



(define (check-discr-tree-signed tree lit)
  (let ((fun (car lit))
        (res #f) )
    (set! *ground-check-res* #f)
    (set! *nonunit-discr-flag* #f)
    (set! tree (assq (logand *fnsymb-mask* fun) (car tree)))
    (if tree
        (set! res (check-discr-tree-part (cdr tree) (cdr lit) '()))
        (set! res #f) )
    (if (and (not res)
             (commutative-fun? fun) )
        (set! res
              (check-discr-tree-part
                (cdr tree)
                (cons (caddr lit) (cons (cadr lit) (cdddr lit)))
                '() )))
    (if res res        *ground-check-res*) ))


; insert-clause-discr-tree! takes a tree, full clause,
; litinfolist (first el of each list there is a changed/unchanged flag)
; and insert-flag, which should be #t to enable insertion
;
; it inserts ground lits into tree AND if
; the insert-flag is #t, also the clause
;
; in any case it also changes destructively fullcl, replacing new (changed)
; ground lits with internalized ground lits.

(define (insert-clause-discr-tree! tree fullcl litinfolist insert-flag)

   ;(newline)(display "insert-clause-discr-tree! entered with:")
   ;(newline)(print-clause fullcl)(newline)

   (let ((cl (cdr fullcl)))
     (cond
        ((and (null? (cdr cl)) insert-flag)
             ; really unit clause and insertion allowed
           (set! *cur-lit* (car cl))
           (set! *cur-clause* fullcl)
           (insert-discr-tree! tree *cur-lit* #f #t #t #f)

           ;(newline)(display "*discr-tree*:")(newline)
           ;(show-discr-tree tree #t)(read-char)
           ;(newline)

           ; replace with the interned lit
           (set-car! cl *cur-lit*) )
        ((not insert-flag)
           ; uniquenize only
          (do ((clpart cl (cdr clpart))
                (infpart litinfolist (cdr infpart)) )
               ((null? clpart))
             (if (and (caar infpart) ; changed-flag is #t
                      (not (fun-may-cont-vars? (caar clpart)))  ; ground
                      (not (answer-pred? (caar clpart))) )
                 (begin
                   (set! *cur-lit* (car clpart))
                   (set! *cur-clause* #f)
                   (insert-discr-tree! tree *cur-lit* #f #t #t #f)

                   ;(newline)(display "*discr-tree*:")(newline)
                   ;(show-discr-tree tree #t)(read-char)
                   ;(newline)

                   ; replace with the interned lit
                   (set-car! clpart *cur-lit*)))))
        (else
         ; nonunit clause
           ;   (if (null? (cddr cl))  ; two-literal clause)
           (insert-discr-tree-nonunit! *nonunit-discr-tree* fullcl)
           ; handle literals for ground literal uniquenisation
           (do ((clpart cl (cdr clpart))
                (infpart litinfolist (cdr infpart)) )
               ((null? clpart))
             (if (and (caar infpart) ; changed-flag is #t
                      (not (fun-may-cont-vars? (caar clpart)))  ; ground
                      (not (answer-pred? (caar clpart))) )
                 (begin
                   (set! *cur-lit* (car clpart))
                   (set! *cur-clause* #f)
                   (insert-discr-tree! tree *cur-lit* #f #t #t #f)

                   ;(newline)(display "*discr-tree*:")(newline)
                   ;(show-discr-tree tree #t)(read-char)
                   ;(newline)

                   ; replace with the interned lit
                   (set-car! clpart *cur-lit*) )))))))



; in order to insert a literal to discr-tree:
; call insert-discr-tree! with:
; tree: full discr-tree
; term: literal
; funflag: #f (literal is not a fun)
; endflag #t (the single (thus the last) object)
; unitflag: #t if unit discr tree, #f if nonunit discr tree
; clauseflag: #f if term demod or unit clause, #t if clause demod
;
; NB!!!!!
; except iff clauseflag is true,
; all function symbs must be only 7-bit, since we use
; *predsymb-mask* here to not have a special treatment for
; literals/nonliteral-terms (preds must have the sgn turned down here!)
;
; and the following globals must be set (used by insert-discr-tree-leaf! ):
; *cur-lit*: inserted literal
; *cur-clause*: if you have a unit clause (containing inserted literal)
;               then the full inserted unit clause
;               else #f

(define (insert-discr-tree! tree term funflag endflag unitflag clauseflag)

  ;(newline)
  ;(display "insert-discr-tree! called with term, cur-clause and tree: ")
  ;(newline)
  ;(display term)(newline)
  ;(if (pair? *cur-clause*)
  ;   (debug-print-clause *cur-clause*)
  ;   (display *cur-clause*) )
  ;(newline)
  ;(if clauseflag (pretty-print tree))

  (if (non-composite? term)
      ; term is non-composite
      (let ((place '()))
        (cond (funflag
                 (if (and unitflag (not clauseflag))
                     (set! term (logand *predsymb-mask* term))
                     (set! term (logand *fnsymb-mask* term)) )
                 (set! place (assq term (car tree)))
                 (if (not place)
                     (begin (set-car! tree
                                      (cons (list term '()) (car tree)) )
                            (set! place (caar tree)) )))
              ((non-composite:-var? term)
                 (set! term (logand *var-nr-mask* term))
                 (if (null? (cdr tree)) (set-cdr! tree (list '())))
                 (set! tree (cdr tree))
                 (set! place (assq term (cdr tree)))
                 (if (not place)
                     (begin (set-cdr! tree
                                      (cons (list term '()) (cdr tree)) )
                            (set! place (cadr tree)) )))
              (else
                 (if (null? (cdr tree)) (set-cdr! tree (list '())))
                 (set! tree (cdr tree))
                 (set! place (assq term (car tree)))
                 (if (not place)
                     (begin (set-car! tree
                                      (cons (list term '()) (car tree)) )
                            (set! place (caar tree)) ))))
        (if endflag
            (cond
               (clauseflag
                 (set-cdr! place (cons '() '()))
                 (insert-discr-tree-clausedemod-leaf! (cdr place)) )
               (unitflag
                 (insert-discr-tree-leaf! (cdr place)) )
               (else
                 (insert-nonunit-discr-tree-leaf! (cdr place)) ))
            (cdr place) ))
      ; term is composite
      (begin
        ;(display "dlpozepa1")
        (set! tree (insert-discr-tree! tree (car term) #t #f unitflag clauseflag))
        ;(display "dlpozepa2")
        (set! term (cdr term))
        (do ()
            ((null? term)
               tree)
          (set! tree (insert-discr-tree! tree (car term) #f
                                         (and endflag (null? (cdr term)))
                                         unitflag
                                         clauseflag))
          (set! term (cdr term)) ))))

; insert-discr-tree-leaf! creates a leaf of the tree.
; it should be used ONLY if the checking fails.
;
; place is the standard tree level.
; (positive-lit negative-lit full-unit-cl)
; where the nonexistant ones of the first two are '() and the
; full-unit-cl may be either negative or positive or nonexistent.
;
; *cur-lit* is the inserted literal; it may be dest. changed or set!
; *cur-clause* is either the unit clause of *cur-lit* or #f.
;
; in case an equal (or complementary equal) lit is found, either
; *cur-lit* is set! to the old one or *cur-lit* is set-cdr! to the cdr of
; the (complementary) old one.

(define (insert-discr-tree-leaf! place)
  (if (zero? (logand *sgn-mask* (car *cur-lit*)))
      ; positive literal
      (begin
        (if (not (null? (car place)))
            ; there is a positive one already
            (set! *cur-lit* (car place)) )
        (if (null? (cdr place))
            ; no negative, no unit given in the leaf
            (begin
              (if *cur-clause*
                  (set-cdr! place (list '() *cur-clause*)) ))
            ; something is given
            (begin
              (if (not (null? (cadr place)))
                  ; negative is given: in case *cur-lit* is ground,
                  ; replace the cdr of the *cur-lit*
                  ; with the cdr of negative.
                  (if (not (fun-may-cont-vars? (car (cadr place))))
                      (set-cdr! *cur-lit* (cdr (cadr place)))  ))
              (if (and *cur-clause* (null? (cddr place)))
                  ; we have a unit, but no unit given
                  (set-cdr! (cdr place) (list *cur-clause*)) )))
        ; insert the new (possibly changed) positive
        (set-car! place *cur-lit*) )
      ; negative literal
      (begin
        (if (not (null? (car place)))
            ; there is a positive one: in case *cur-lit* is ground,
            ; replace the cdr of the *cur-lit* with the cdr of positive.
            (if (not (fun-may-cont-vars? (caar place)))
                (set-cdr! *cur-lit* (cdar place)) ))
        (if (null? (cdr place))
            ; no negative, no unit given in the leaf
            (begin
              (if *cur-clause*
                  (set-cdr! place (list *cur-lit* *cur-clause*))
                  (set-cdr! place (list *cur-lit*)) ))
            ; something is given
            (begin
              (if (null? (cadr place))
                  ; negative is not given
                  (if (or *cur-clause*
                          (not (fun-may-cont-vars? (car *cur-lit*))) )
                      (set-car! (cdr place) *cur-lit*) )
                  ; negative is given: replace the *cur-lit*
                  ; with the old negative.
                  (set! *cur-lit* (cadr place)) )
              (if (and *cur-clause* (null? (cddr place)))
                  ; we have a unit, but no unit given
                  (set-cdr! (cdr place) (list *cur-clause*)) ))))))



; insert-discr-tree-clausedemod-leaf! creates a leaf of the tree.
; it should be used ONLY if the subsumption fails.
;
; place is the standard tree level.
;
; *cur-clause* is the inserted clause

(define (insert-discr-tree-clausedemod-leaf! place)

   ;(newline)
   ;(display "insert-discr-tree-clausedemod-leaf! called with place and *cur-clause*: ")
   ;(newline)
   ;(display place)(newline)
   ;(debug-print-clause *cur-clause*)(newline)

   (set-car! place (car *cur-clause*))
   (set-cdr! place (cdr *cur-clause*))  )


; search by recursion down the elements of the term.
; it uses the flag *discr-depth-first* to determine whether
; depth-first or vars-first search should be used for composites.
;
; If it encounters a leaf with no unit clause, it sets *ground-check-res*
; to the leaf and proceeds with searching.

(define *ttt1* '())
(define *ttt2* '())
(define *ttt3* '())
(define *ttt4* '())

(define (check-discr-tree-part tree part cont)
  (set! *ttt1* tree)
  (set! *ttt2* part)
  (set! *ttt3* cont)
  ;(newline)(display "check-discr-tree-part entered with:")
  ;(show-discr-tree tree #t)(newline)
  ;(display part)(newline)
  ;(display cont)(newline)
  ;(display *nonunit-discr-flag*)
  ;(display *overdouble-discr-flag*) (newline)
  (if (null? part)
      (if (null? cont)
          (check-discr-tree-leaf tree)
          (if (null? (car cont))
              ; nonunit discr case - choice point ('() at first pos a marker)
              (if (null? (cdr cont))
                  #f
                  (or (check-discr-tree-part tree (list (cadr cont)) '())
                      (check-discr-tree-part
                         tree '() (cons '() (cddr cont))) ))
              ; standard case
              (check-discr-tree-part tree cont '()) ))
      (let ((el (car part))
            (vars (cdr tree)) ; NB! both the cons and varspart might be missing
            (res #f)
            (var #f)
            (val #f) )
        (if (not (null? vars))  ; varspart proper is taken now
            (set! vars (cdr vars)) )
        (if (non-composite? el)
            (if (non-composite:-var? el)
                ; el is a variable
                (begin
                  (set! el (logand *var-nr-mask* el))
                  (do ()
                      ((or (null? vars) res)
                         res)
                    (set! var (caar vars))
                    (set! val (var-ref var))
                    (if (unassigned-val? val)
                        (begin
                          (set-var-ref! var el)
                          (set! res
                             (check-discr-tree-part
                               (cdar vars) (cdr part) cont))
                          (set-var-ref! var val) )
                        (if (eq? val el)
                            (set! res
                                  (check-discr-tree-part
                                    (cdar vars) (cdr part) cont))))
                    (set! vars (cdr vars)) ))
                ; el is a constant
                (begin
                  (set! tree (cdr tree))
                  (if (null? tree)
                      (set! tree #f) ; no constants, no vars
                      (set! tree (assq el (car tree))) ) ; may be also vars
                  (if tree ; the right constant found
                      (set! tree
                         (check-discr-tree-part (cdr tree) (cdr part) cont)))
                  (if *nonunit-discr-flag*
                      ; nonunit discr tree case
                      (if (and tree (not (car tree)))
                      ; no double clause
                          (begin
                            (if *overdouble-discr-flag*
                                (set! *nonunit-res*
                                      (cons (cdr tree) *nonunit-res*)))
                            (set! tree #f) ))
                      ; unit discr tree case
                      (if (and tree (or (null? (cdr tree))
                                        (null? (cddr tree))))
                          ; a leaf found, but no unit clause
                          (begin (set! *ground-check-res* tree)
                                 (set! tree #f) )))
                  (if tree
                      tree
                      (do ()
                          ((or (null? vars) res)
                             res)
                        (set! var (caar vars))
                        (set! val (var-ref var))
                        (if (unassigned-val? val)
                            (begin
                              (set-var-ref! var el)
                              (set! res
                                    (check-discr-tree-part
                                      (cdar vars) (cdr part) cont))
                              (set-var-ref! var val) )
                            (if (eq? val el)
                                (set! res
                                      (check-discr-tree-part
                                        (cdar vars) (cdr part) cont))))
                        (set! vars (cdr vars)) ))))
            ; el is a composite
            (if (null? el)
             ; special case: a choice point, not a normal element
             (check-discr-tree-part tree '() part)
             ; normal element
             (begin
              (if *discr-depth-first* ; current default #t
                  (begin
                    (set! tree (assq (logand *fnsymb-mask* (car el))
                                     (car tree) ))
                    (if tree ; funsymb found
                        (let ((tmp '()))
                          (if (null? cont)
                              (set! tmp (cdr part))
                              (set! tmp (append (cdr part) cont)) )
                          (set! tree
                                (check-discr-tree-part
                                 (cdr tree) (cdr el) tmp)))))
                  (set! tree #f) )
              (if *nonunit-discr-flag*
                  ; nonunit discr tree case
                  (if (and tree (not (car tree)))
                      ; no double clause
                      (begin
                        (if *overdouble-discr-flag*
                            (set! *nonunit-res*
                                  (cons (cdr tree) *nonunit-res*)))
                        (set! tree #f) ))
                  ; unit discr tree case
                  (if (and tree (or (null? (cdr tree)) (null? (cddr tree))))
                      ; a leaf found, but no unit clause
                      (begin (set! *ground-check-res* tree)
                             (set! tree #f) )))
              (if tree
                  tree
                  (do ()
                      ((or (null? vars) res)
                         (if (or *discr-depth-first* res)
                             res
                             (begin
                               (set! tree
                                     (assq (logand *fnsymb-mask* (car el))
                                           (car tree) ))
                               (if tree
                                   (begin
                                     (if (null? cont)
                                         (set! cont (cdr part))
                                         (set! cont (append (cdr part)
                                                            cont)))
                                     (check-discr-tree-part
                                      (cdr tree) (cdr el) cont))))))
                    (set! var (caar vars))
                    (set! val (var-ref var))
                    (if (unassigned-val? val)
                        (begin
                          (set-var-ref! var el)
                          (set! res
                                (check-discr-tree-part
                                 (cdar vars) (cdr part) cont))
                          (set-var-ref! var val) )
                        (if *nonunit-discr-flag*
                            (if (term-equal-strong? val el)
                                (set! res
                                  (check-discr-tree-part
                                   (cdar vars) (cdr part) cont)))
                            (if (term-equal? val el)
                                (set! res
                                      (check-discr-tree-part
                                       (cdar vars) (cdr part) cont)))))
                    (set! vars (cdr vars)) ))))))))


;(define (check-discr-tree-leaf tree)
;  (if (and *overdouble-discr-flag* *nonunit-discr-flag*)
;      (if (and tree (not (car tree)))
;          (begin
;            (set! *nonunit-res*
;                  (cons (cdr tree) *nonunit-res*) )
;            #f)
;          tree)
;      (begin
;        (newline)
;        (display tree)
;        (newline)
;        (if (and (cdr tree) (cddr tree))
;            (display (car (clinfo-clnumber-sublist (car (caddr tree))))) )
;        (newline)
;        tree) ))

; *prohibit-subs-clnr* is the number of a clause which is
; prohibited for use: special case of subsuming the given clause.
; Normally *prohibit-subs-clnr* sould be #f.


(define (check-discr-tree-leaf tree)
  (if (and *prohibit-subs-clnr*
           (if *nonunit-discr-flag*
               (and (car tree)
                    (eq? (car (clinfo-clnumber-sublist (caar tree)))
                         *prohibit-subs-clnr*))
               (and
                (car tree)
                (not (null? (cdr tree)))
                (not (null? (cddr tree)))
                (eq? (car (clinfo-clnumber-sublist (car (caddr tree))))
                     *prohibit-subs-clnr*) )))
      #f
      (if *nonunit-discr-flag*
          (if (and tree (not (car tree)))
              (begin
                (if *overdouble-discr-flag*
                    (set! *nonunit-res*
                          (cons (cdr tree) *nonunit-res*) ))
                #f)
              tree)
          tree)))


;===================================================================
;
;              discrimination tree for nonunit stuff
;
;===================================================================

; insert-discr-tree-nonunit inserts a nonunit clause into a
; discrimination tree. It takes:
;
; tree - a special discrimination tree for nonunit clauses
; fullcl - full clause to be inserted


(define (insert-discr-tree-nonunit! tree fullcl)
  (let* ((info (car fullcl))
         (purecl (cdr fullcl))
         (object (list 0 (car purecl) (cadr purecl))) )
    ; object is a two-place pseudo-literal: 0(<lit1>,<lit2>)
    (set! *cur-clause* fullcl)
    ; (display "dlpozepa1")
    (insert-discr-tree! tree object #f #t #f #f) ))

; insert-nonunit-discr-tree-leaf! creates a leaf of the tree.
; it should be used ONLY if the subsumption fails.
;
; place is the standard tree level.
; (<full-double-cl> (nr . <full-cl1>) ... (nr . <full-cl-n>))
; where <full-double-cl> is either the full existing double clause or #f
; and <full-cl1> ... <full-cl-n> are longer clauses to be used
; as candidates, nr is a a possibly different timer-like counter
; used for avoiding double subsumption checks.
;
; *cur-clause* is the inserted clause: either double or long

(define (insert-nonunit-discr-tree-leaf! place)
   (if (null? (cddr (cdr *cur-clause*)))
       ; double clause
       (set-car! place *cur-clause*)
       ; longer clause
       (begin
         (set-car! place #f)
         (set-cdr! place (cons (cons 0 *cur-clause*) (cdr place))) )))


; check-nonunit-discr-tree takes:
; tree - a full nonunit discr tree,
; firstlit - a literal to be checked first,
; purecl - pure clause to be checked, incl. the firstlit.
;
; It creates a temporary object
; containing two literals from the checked clause:
; 0(<lit1>,<lit2>).
;
; it returns a leaf of the form:
; (<full-double-cl> <full-cl1> ... <full-cl-n>)
; where <full-double-cl> is either the full existing double clause or '()
; or it returns #f if nothing found
;
; if there are different possible leaves, one without <full-double-cl>
; and one with <full-double-cl>, then always the second one is returned.


(define (check-nonunit-discr-tree tree purecl)
  ; (newline)
  ; (display "check-nonunit-discr-tree entered with")
  ; (print-pure-clause purecl)
  ; (newline)
  (let* ((res #f)
         (tmp '())
         (passed '()) )
    (set! *nonunit-discr-flag* #t)
    (set! *nonunit-res* '())
    (cond
     ((null? (car tree))
        (set! res #f) )
     ((null? (cdr purecl))
        (set! res #f) )
     ((null? (cddr purecl))
        (set! *overdouble-discr-flag* #f)
        (set! res
              (or (check-discr-tree-part
                        (cdaar tree) purecl '())
                  (check-discr-tree-part
                     (cdaar tree) (list (cadr purecl) (car purecl)) '()) )))
     (else
        (set! *overdouble-discr-flag* #t)
        (do ((clpart purecl (cdr clpart)))
            ((or res (null? clpart)))
          (if (not (null? passed))
              (set! tmp
                    (cons (car clpart)
                          (cons '() (append passed (cdr clpart))) ))
              (set! tmp
                    (cons (car clpart) (cons '() (cdr clpart))) ))
          (set! res (check-discr-tree-part (cdaar tree) tmp '()))
          (set! passed (cons (car clpart) passed)) )))
    (set! *nonunit-discr-flag* #f)
    (set! *overdouble-discr-flag* #f)
;- - - - - - - - - - - - -
;    (newline)(display ". . . . . . . . . . . ")(newline)
;    ;(pretty-print *nonunit-res*)(newline)
;    (for-each
;      (lambda(x)
;         (for-each (lambda(y)(newline)(print-full-clause (cdr y))) x) )
;      *nonunit-res*)
;    (newline)(display ". . . . . . . . . . . ")(newline)
;- - - - - - - - - - - - -
    res))


;==================================================================
;
;    finding demodulation clauses and giving orientation back
;
;==================================================================

; demodulation-clause-test returns:
; #f  - not usable as demod clause
; (#t #f)   - left literal should be first
; (#f #t)   - right literal should be first

(define (demodulation-clause-test litinfolst purecl)
  ;(display "starting to work on test: ")
  ;(print-pure-clause purecl)
  ;(newline)
  ;(display litinfolst)
  (let ((reslst '())
        (smalllits '()) ; the set of prohibited lits
        (biglits '())   ; the set of not yet-looked-through lits
        ; the following is a list of pairs of vardepth-infos and literals:
        ; ((((x . 1)(y . 3)) p (g x) (h (g y)) x) ...)
        (newlitinfnr 0)
        (ordernr 0)
        (varinfo-purecl '()) )
    (if (or (null? purecl)
            (null? (cdr purecl))
            (not (null? (cddr purecl)))
            (member-ans-lit purecl)
            (every? (lambda(x)(not (term-may-cont-vars? x))) purecl) )
        #f ; choose only  clauses of length two with no answer-literals
       (begin
         (set! varinfo-purecl
              (map (lambda(x)(cons (make-var-depth-info x) x))
                    purecl) )
         (set! ordernr 0)
         (set! reslst
          (map
            (lambda(litinf varinfolit)
               ;(newline)
               ;(display litinf)
               ;(newline)
               ;(display varinfolit)
               ;(newline)
               (set! ordernr (+ 1 ordernr))
                  (if (or (answer-lit? (cdr varinfolit))
                       (some?
                         (lambda(x)
                             ;(newline)
                             ;(display x)
                             ;(newline)
                             ;(display varinfolit)
                             ;(newline)
                             ;(display (list
                             ;  (fun-may-cont-vars? (cadr x))
                             ;  (> (fun-depth (cadr x))
                             ;     (fun-depth (cadr varinfolit)) )
                             ;  (vdepth->-weak-varinfolit? x varinfolit) ))
                             ;(newline)
                             ;(cond)
                              ;(*posweight-order-strategy-flag*
                              ;  (and firstflag
                              ;       (not (answer-lit? (cdr varinfolit)))
                              ;       (not (neg-fun? (cadr varinfolit)))
                              ;       (vdepth-superset-varinfolit?
                              ;          (cadr varinfolit) (cadr x) )))
                              ;(*weight-order-strategy-flag*
                              ;  (and firstflag
                              ;       (not (answer-lit? (cdr varinfolit)))
                              ;       (vdepth-superset-varinfolit?
                              ;          (cadr varinfolit) (cadr x) )))
                              ;
                              ;(else)
                                (and
                                  (not (eq? x varinfolit))
                                       ;(vdepth-superset-varinfolit?
                                       ;     (cadr varinfolit) (cadr x) )
                                       ;(begin (display "ok") #F)
                                  (if (not (fun-may-cont-vars? (cadr varinfolit)))
                                       ; main is ground
                                       ;(if(or *firstlitpref-order-strategy-flag*
                                       ;        *weight-order-strategy-flag*
                                       ;        *posweight-order-strategy-flag*
                                       ;        *nameorder-strategy-flag*)
                                       ;   (or (fun-may-cont-vars? (cadr x))
                                       ;       (> (fun-depth (cadr x))
                                      ;                (fun-depth (cadr varinfolit)) )
                                      ;             (> ordernr 1) )
                                      ;   (or (fun-may-cont-vars? (cadr x))
                                      ;            (>  (fun-depth (cadr x))
                                      ;                (fun-depth (cadr varinfolit)) )
                                      ;            (>= (logand *term-size-mask* (cadr x))
                                      ;                 (logand *term-size-mask*
                                      ;                     (cadr varinfolit) ))))
                                      #f
                                      ; main contains vars
                                      (if (fun-may-cont-vars? (cadr x))
                                          ; both contain vars
                                          (and
                                             (> (fun-depth (cadr x))
                                                (fun-depth (cadr varinfolit)) )
                                             (vdepth->-varinfolit? x varinfolit) )
                                          ; main contans vars, smaller not.
                                          (or  (>  (fun-depth (cadr x))
                                                   (fun-depth (cadr varinfolit)) )
                                               (>= (logand *term-size-mask* (cadr x))
                                                   (logand *term-size-mask*
                                                           (cadr varinfolit) ))) ))))


                          varinfo-purecl))
               ; not allowed to be resolved upon
               #f
               ; allowed to be resolved upon
               #t))
               ;(set! firstflag #f)
            litinfolst
            varinfo-purecl))

         ; return the result

         ;(newline)
         ;(display "demod-cl-test: ")
         ;(print-pure-clause purecl)(display " : ")
         ;(display reslst)

         (cond
           ((and (car reslst) (cadr reslst)) '(#f #f))
           ((car reslst) reslst)
           ((cadr reslst) reslst)
           (else reslst) )))))



;==================================================================
;
;    ordering the equality and the clause, ie computing a right litinfolst
;
;==================================================================

; order-prohibit-litinfos is the top procedure. It takes
; the litinfolst and the purecl and returns a new
; litinfolst where some resolution-, factorization-
; and paramodulation-prohibiting bits may be set
; or cleared.

(define (order-prohibit-litinfos litinfolst purecl)

  ;(newline)(display "order-prohibit-litinfos entered:")
  ;(display litinfolst)(newline)(display purecl)
  ;(newline)
  ;(print-clause purecl)(newline)

  ;(newline)(display "CHECK X0")(newline)
  ; first order equalities
  (set! litinfolst (order-litinfo-eq-only litinfolst purecl))

  ;(newline)(display "Ordering strategy flags:")(newline)
  ;(display "*order-strategy-flag*:")(display *order-strategy-flag*)(newline)
  ;(display "*ordered-factor-flag*:")(display *ordered-factor-flag*)(newline)
  ;(display "*hyper-strategy-flag*:")(display *hyper-strategy-flag*)(newline)
  ;(display "*name-order-strategy-flag*:")(display *name-order-strategy-flag*)(newline)
  ;(display "*poslitpref-order-strategy-flag*:")(display *poslitpref-order-strategy-flag*)(newline)
  ;(display "*firstlitpref-order-strategy-flag*:")(display *firstlitpref-order-strategy-flag*)(newline)
  ;(display "*posweight-order-strategy-flag*:")(display *posweight-order-strategy-flag*)(newline)
  ;(display "*negweight-order-strategy-flag*:")(display *negweight-order-strategy-flag*)(newline)
  ;(display "*ordered-paramodulation-flag*:")(display *ordered-paramodulation-flag*)(newline)
  ;(display "*order-equality-kb-strat*:")(display *order-equality-kb-strat*)(newline)
  ;(display "*order-equality-lex-strat*:")(display *order-equality-lex-strat*)(newline)
  ;(display "*order-equality-firstpref-strat*:")(display *order-equality-firstpref-strat*)(newline)
  ;(display "*lex-order-prefer-bigger-arity*:")(display *lex-order-prefer-bigger-arity*)(newline)
  ;(newline)
  ; second, order literals
  (cond
;    (#t
;      (order-litinfo-eq-only litinfolst purecl) )
    (*hyper-strategy-flag*
      (if (positive-cl? purecl)
          (if *name-order-strategy-flag*
              (name-order-litinfo litinfolst purecl)
              (vdepth-order-litinfo litinfolst purecl) )
          litinfolst) )
         ; (map (lambda(litinf)
         ;         (cons (car litinf)
         ;               (cons (cadr litinf)
         ;                     (cons (logxor *litinf-nouse-mask*
         ;                                   (logior (caddr litinf)
         ;                                           *litinf-nouse-mask*))
         ;                           (cdddr litinf) ))))
         ;      litinfolst)
    (*name-order-strategy-flag*
     (name-order-litinfo litinfolst purecl) )
    (*epr-order-strategy-flag*
     (epr-order-litinfo litinfolst purecl) )
    (*poslitpref-order-strategy-flag*
     (poslit-order-litinfo litinfolst purecl) )
    (*firstlitpref-order-strategy-flag*
     (firstlitpref-order-litinfo litinfolst purecl) )
    (*posweight-order-strategy-flag*
     (posweight-order-litinfo litinfolst purecl) )
    (*negweight-order-strategy-flag*
     (negweight-order-litinfo litinfolst purecl) )
    (*exp-order-strategy-flag*
       ;(let* ((tmp1 (copy-tree litinfolst))
       ;       (tmp2 (copy-tree purecl))
       ;       (tmp (exp-order-litinfo litinfolst purecl))
       ;       (nr 0))
       ;  (do ((part tmp (cdr part)))
       ;      ((null? part))
       ;    (if (litinf-binrespara? (caddr (car part)))
       ;        (set! nr (+ 1 nr)) ))
       ;  (if (> nr 1)
       ;      (posweight-order-litinfo tmp1 tmp2)
       ;      tmp)))

      ;    (firstlitpref-order-litinfo litinfolst purecl) ))
      (exp-order-litinfo litinfolst purecl) )
    (else
     (vdepth-order-litinfo litinfolst purecl) )))


(define (vdepth-order-litinfo litinfolst purecl)
  (if *order-strategy-flag*
      (vdepth-order-litinfo-aux litinfolst purecl)
      (if *ordered-paramodulation-flag*
          (vdepth-order-litinfo-eq-only litinfolst purecl)
          litinfolst)))

; - - - - - - -  equality ordering - - - - - - - - - - - -


; order-litinfo-eq-only takes a litinfolst ((x y <infonr> ..) ...)
; and a pure clause and returns a new litinfolst with
; equality orderings set right, NO OTHER ORDERING IMPOSED ON LITERALS


(define (order-litinfo-eq-only litinfolst purecl)
  (let* ((newlitinfnr 0))
    (map (lambda(litinf lit)
           (if (or (eq? (pure-fun (car lit)) *eq-pred-nr*)
                   (and *type-theory-stratflag*
                        (composite? (cadr lit))
                        (eq? (pure-fun (caadr lit)) *eq-fun-nr*)
                        (not (neg-fun? (car lit))) ))
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 (set! newlitinfnr
                       (order-equality-litinfo newlitinfnr lit) )
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf)) )))
               litinf))
         litinfolst
         purecl)))



(define (order-equality-litinfo litinfonr lit)
  (cond
    (*order-equality-kb-strat*
      (order-equality-litinfo-kb litinfonr lit) )
    (*order-equality-lex-strat*
      (order-equality-litinfo-lex litinfonr lit) )
    (*order-equality-firstpref-strat*
      (order-equality-litinfo-firstterm litinfonr lit) )
    (else
      (order-equality-litinfo-depth litinfonr lit) )))


(define (prohibit-right-para litinfonr)
  (logior *litinf-noparafromright-mask* litinfonr) )

(define (prohibit-left-para litinfonr)
  (logior *litinf-noparafromleft-mask* litinfonr) )



; - - - - knuth-bendix ordering - - - - -


(define *order-equality-kb-strat* #t)

(define (order-equality-litinfo-kb litinfonr lit)
  (let* ((left (if *type-theory-stratflag* (cadadr lit) (cadr lit)))
         (right (if *type-theory-stratflag* (caddr (cadr lit)) (caddr lit)))
         (compres '()) )
    (set! compres (compare-terms-kb-order left right #f))
    (cond
     ((eq? compres 1)
       (prohibit-right-para litinfonr) )
     ((eq? compres -1)
       (prohibit-left-para litinfonr) )
     (else
       litinfonr))))





; compare-terms-kb-order returns:
; 1  : left is bigger
; -1 : right is bigger
; 0  : equal in comparison
; #f : incomparable
;
; algorithm basic:
;  0) terms may be equal if they have the same symbol count and
;    exactly same variable occurrences (same nr of each var)
;  1) comparison is done by symbol weight count
;  2) if symbol weight counts are equal, then use
;     the symbol counts of first args, then second, etc, lexicographically
;  3) if variables are present, then:
;       - if terms t and g both contain vars not present in other,
;         then terms are incomparable
;       - if term t contains all vars in g, then t may be bigger than
;             t if symbol count is bigger,
;       - else t is incomparable to g
; 4) we give weights to constant and fun symbs. However,
;    we assume that fun symbs have always a bigger weight
;    than constants

; leftvars is a list of pairs of vars in left with the occurrence nr
; rightvars is the same for right term

(define (compare-terms-kb-order left right followvarsflag)
  (let* ((leftvars '())
         (rightvars '())
         (leftsum 0)
         (rightsum 0)
         (leftres '())
         (rightres '())
         (varres '())
         (tmp '())
         (stopflag #f) )
    (cond
     ((eq? left right)  0) ; syntactically equal terms are equal
     ((var? left)
        (cond
           ((var? right) #f) ; different vars x and y are incomparable
           ((non-composite? right) #f); x and c are incomparable
           ; now right is a noncomposite term: x < t(..x..)
           ((var-inside-term? left right followvarsflag) -1)
           (else #f) ))
     ((var? right)
         (cond
           ((non-composite? left) #f); c and x are incomparable
           ; now right is a noncomposite term: t(..x..) > x
           ((var-inside-term? right left followvarsflag) 1)
           (else #f) ))
     ; now left and right cannot be vars
     ((non-composite? left)
         (cond
           ((non-composite? right)
              (compare-const-kb-order left right) )
           ; right is a composite
           (else -1) )); t(..) > c
     ; now left is a composite
     ((non-composite? right) ; right is a composite
           1); t(..) < c
     ; now both left and right are composite: compute counts and vars
     ((begin
        (set! leftres (cons 0 '()))
        (set! rightres (cons 0 '()))
        (count-symbs-vars! left leftres followvarsflag)
        (count-symbs-vars! right rightres followvarsflag)
        (set! varres
              (compare-terms-varcount-subset
                 (cdr leftres) (cdr rightres)) )
        #f))
     ; now counts and vars have been computed
     ((not varres)
        #f) ; variables do not allow ordering
     ; now variables allow ordering in some way
     ((and (>= varres 0)
           (or (> (car leftres) (car rightres))
               (and (eq? (car leftres) (car rightres))
                    (> (fun-lex-order-value (car left))
                       (fun-lex-order-value (car right)) ))))
        1) ; left bigger by size or funsymb order
     ((and (<= varres 0)
           (or (> (car rightres) (car leftres))
               (and (eq? (car leftres) (car rightres))
                    (> (fun-lex-order-value (car right))
                       (fun-lex-order-value (car left)) ))))
       -1) ; right bigger by size or funsymb order
     ((or  (not (eq? (car leftres) (car rightres)))
           (not (eq? (fun-lex-order-value (car left))
                     (fun-lex-order-value (car right)) )))
        #f)   ; incomparable (sizes and funsymbs orders not equal)
     ; now terms have equal size and equally lex-ordered fun symbs
     ; and vars allow some ordering: do lex check
     (else
       (set! tmp #f)
       (set! stopflag #f)
       (do ((lpart (cdr left) (cdr lpart))
            (rpart (cdr right) (cdr rpart)) )
           ((or (null? lpart)
                (null? rpart)
                stopflag)
              ; now we have all the info about lex order:
              ; get the full terms ordering if possible
              (cond
                ((eq? tmp 0)
                     (if (and (null? lpart)(null? rpart)
                              (eq? varres 0) )
                         0
                         #f) )
                ((eq? tmp 1)
                   (if (>= varres 0)
                       1
                       #f) )
                ((eq? tmp -1)
                   (if (<= varres 0)
                       -1
                       #f) )
                (else #f) ))
        (set! tmp
              (compare-terms-kb-order
                 (car lpart) (car rpart) followvarsflag) )
        (if (or (not tmp) (eq? 1 tmp) (eq? -1 tmp))
            (set! stopflag #t) ) )))))


(define *const-kb-order-vector* '())
(define *fun-kb-order-vector* '())

(define (make-symbs-kb-order-vector)
  (let ((x 0))
    (set! *const-kb-order-vector* (make-vector 10000))
    (set! *fun-kb-order-vector* (make-vector 300))
    (do ((x 0 (+ 1 x)))
        ((>= x 10000))
     (vector-set! *const-kb-order-vector* x 10) )
    (do ((x 0 (+ 1 x)))
        ((>= x 300))
     (vector-set! *fun-kb-order-vector* x 11) )))


(define (const-kb-order-value x)
   ; (newline)(display "const-kb-order-value: ")(display x)(newline)
  (let ((tmp (- x *minconst*)))
    (if (< tmp 10000)
        (vector-ref *const-kb-order-vector* tmp)
        10)))

(define (fun-kb-order-value x)
  ; (newline)(display "const-kb-order-value: ")(display x)
  ;(display " shifted: ")
  ;(display (shiftright (pure-fun x) *fnsymb-shift*))(newline)
  (vector-ref *fun-kb-order-vector* (shiftright (pure-fun x) *fnsymb-shift*)) )

(define (compare-const-kb-order left right)
  (let ((lval (const-kb-order-value left))
        (rval (const-kb-order-value right)) )
    (cond
      ((> lval rval) 1)
      ((> rval lval) -1)
      (else
        (compare-symbs-lex-order left right) ))))

(define (var-inside-term? var term followvarsflag)
  (if (not (composite? term))
      (if (non-composite:-var? term)
          (eq? (var-nr var) (var-nr term))
          #f)
      (and (fun-may-cont-vars? (car term))
           (some? (lambda(x)(var-inside-term? var x followvarsflag))
                  (cdr term) ))))


(define (compare-terms-varcount-subset leftvars rightvars)
  (let ((lsubflag #f)
        (rsubflag #f)
        (tmp '()) )
   (set! lsubflag
        (every? (lambda(var)
                    (set! tmp (assq (car var) rightvars))
                    (and tmp (>= (cdr tmp) (cdr var))) )
                leftvars))
   (set! rsubflag
        (every? (lambda(var)
                    (set! tmp (assq (car var) leftvars))
                    (and tmp (>= (cdr tmp) (cdr var))) )
                rightvars))
   (cond
     ((and lsubflag rsubflag) 0)
     (lsubflag -1)
     (rsubflag 1)
     (else #f) )))


(define (count-symbs-vars! term res followvarsflag)
  ; (display "count-symbs-vars!: ")(display term)(display res)(newline)
  (let ((tmp '()))
    (cond
      ((non-composite? term)
         (if (non-composite:-var? term)
             (begin
               (set! tmp (assq (var-nr term) (cdr res)))
               (if tmp
                   (set-cdr! tmp (+ 1 (cdr tmp)))
                   (set-cdr! res (cons (cons (var-nr term) 1) (cdr res))) ))
             (set-car! res (+ (const-kb-order-value term) (car res))) ))
      (else
        (set-car! res (+ (fun-kb-order-value (car term)) (car res)))
        (for-each (lambda(x)(count-symbs-vars! x res followvarsflag))
                  (cdr term) )))))


; - - - - lex ordering - - - - -

(define (order-equality-litinfo-lex-testing litinfonr lit)
  (let ((tmp '()))
    (set! tmp (order-equality-litinfo-lex-real litinfonr lit))
    (newline)(display "order-equality-litinfo-lex ended")(newline)
    tmp))

(define (order-equality-litinfo-lex litinfonr lit)

  ;(newline)(display "order-equality-litinfo-lex called on: ")(newline)
  ;(display litinfonr)(newline)
  ;(display lit)(newline)
  ;(print-lit lit)

  (let* ((left (if *type-theory-stratflag* (cadadr lit) (cadr lit)))
         (right (if *type-theory-stratflag* (caddr (cadr lit)) (caddr lit)))
         (leftvars '())
         (compres '())
         (rightvars '()) )

    ;(newline)
    ;(display "lex ordering called on eq args: ")(newline)
    ;(print-term left)(newline)
    ;(print-term right)(newline)

    (set! leftvars (make-var-depth-info left))
    (set! rightvars (make-var-depth-info right))

    ;(newline)
    ;(display "leftvars: ")(display leftvars)(newline)
    ;(display "rightvars: ")(display rightvars)(newline)

    (cond
     ((every? (lambda(vd)(assq (car vd) leftvars))
              rightvars)
       ; all rightvars are present in leftvars
       ; ie left could be made bigger than right
       (set! compres (compare-terms-lex-order left right #f))
       (cond
         ((eq? compres 1)
             (prohibit-right-para litinfonr) )
         ((eq? compres -1)
             (if (every? (lambda(vd)(assq (car vd) rightvars))
                         leftvars)
                 (prohibit-left-para litinfonr)
                 litinfonr))
         (else
             litinfonr)))
     ((every? (lambda(vd)(assq (car vd) rightvars))
              leftvars)
       ; all leftvars are present in rightvars
       ; ie right could be made bigger than left
       (set! compres (compare-terms-lex-order left right #f))
       (cond
         ((eq? compres -1)
             (prohibit-left-para litinfonr) )
         (else
             litinfonr)))
     (else
         litinfonr) )))


; compare-terms-lex-order returns:
; 1  : left is bigger
; -1 : right is bigger
; 0  : equal in comparison
; #f : incomparable
;
; algorithm basic:
;  0) terms are compared by leading fun symbols
;  1) if these are equal, terms are compared lexicographically:
;     first args, second args, etc etc
;  2) variable and non-variable are incomparable unless one contains another
;  3) composite is bigger than constant
;  4) we give weights to constant and fun symbs. However,
;    we assume that fun symbs have always a bigger weight
;    than constants
;  5) NB! a smaller term MAY NOT HAVE VARS NOT IN BIGGER

; leftvars is a list of pairs of vars in left with the occurrence nr
; rightvars is the same for right term

(define (compare-terms-lex-order-old left right followvarsflag)

  ;(newline)
  ;(display "lex on:")(newline)
  ;(display left)(newline)
  ;(display right)(newline)
  ;(display followvarsflag)(newline)


  (let* ((tmp '())
         (tmp2 '())
         (lres '())
         (rres '())
         (stopflag #f) )
    (cond
     ((eq? left right)  0) ; syntactically equal terms are equal
     ((var? left)
        (if (and (composite? right)
                 (occurs-term-for-order? right left followvarsflag) )
            -1
            #f))
     ((var? right)
        (if (and (composite? left)
                 (occurs-term-for-order? left right followvarsflag) )
            1
            #f))
     ; now left and right cannot be vars
     ((non-composite? left)
         (cond
           ((non-composite? right)
              (compare-symbs-lex-order left right) )
           ; right is a composite
           (else -1) )); t(..) > c
     ; now left is a composite
     ((non-composite? right) ; right is a composite
           1); t(..) < c
     ; now both left and right are composite
     ((begin  ; precompute by-arg comparisons with the main term
       ;(newline)(display "check Y0")(newline)
       (set! lres
             (some? (lambda(x)
                     (set! tmp (compare-terms-lex-order
                                  x right followvarsflag))
                     (or (eq? tmp 0)(eq? tmp 1)) )
                    (cdr left) ))
       ;(newline)(display "check Y1")(newline)
       (set! rres
             (some? (lambda(x)
                     (set! tmp (compare-terms-lex-order
                                  x left followvarsflag))
                     (or (eq? tmp 0)(eq? tmp 1)) )
                    (cdr right) ))
        (newline)(display "check Y2")(newline)
       #f))
     ; do three checks after by-arg comparison computation
     ((and lres rres) 0)
     ;((and lres rres) #f)
     (lres 1)
     (rres -1)
     ; now starts the main lex comparison of composites
     ((< (fun-lex-order-value (car left)) ; left less than right
         (fun-lex-order-value (car right)) )
         ;(newline)(display "check Y3")(newline)
        (if (every?
              (lambda(arg)
                (set! tmp (compare-terms-lex-order arg right followvarsflag))
                (eq? tmp -1) )
              (cdr left) )
            -1
            #f))
     ((> (fun-lex-order-value (car left)) ; left bigger than right
         (fun-lex-order-value (car right)) )
          ;(newline)(display "check Y4")(newline)
        (if (every?
              (lambda(arg)
                (set! tmp (compare-terms-lex-order arg left followvarsflag))
                (eq? tmp -1) )
              (cdr right) )
            1
            #f))
     ; now functions symbols are equal: recurse lexicographically
     (else
       (set! tmp #f)
       (set! tmp2 #f)
       (set! stopflag #f)
        ;(newline)(display "check Y5")(newline)
       ; loop, comparing args lexicographically
       (do ((lpart (cdr left) (cdr lpart))
            (rpart (cdr right) (cdr rpart)) )
           ((or (null? lpart)
                (null? rpart)
                stopflag)
               ;(newline)(display "check Y6")(newline)
              tmp)
        (set! tmp
              (compare-terms-lex-order
                 (car lpart) (car rpart) followvarsflag) )
        (if (or (not tmp) (eq? 1 tmp) (eq? -1 tmp))
            (set! stopflag #t) ))
       ; loop, comparing full term with args of the second term
        (cond
          ((not tmp) #f)
          ((eq? tmp 0) 0)
          ((eq? tmp -1)
              ;(newline)(display "check Y7")(newline)
             (if (every?
                    (lambda(arg)
                       (set! tmp (compare-terms-lex-order arg right followvarsflag))
                       (eq? tmp -1) )
                    (cddr left) )
                 -1
                 #f))
          ((eq? tmp 1)
               ;(newline)(display "check Y8")(newline)
             (if (every?
                    (lambda(arg)
                       (set! tmp (compare-terms-lex-order arg left followvarsflag))
                       (eq? tmp -1) )
                    ;(cddr right)
                     (cdr right) )
                 1
                 #f)) )))))

; ---------------


(define (compare-terms-lex-order left right followvarsflag)
  (let ((tmp #f))
    (cond
      ((compare-terms-lex-equal left right followvarsflag)
           0)
      ((compare-terms-lex-bigger left right followvarsflag)
           1)
      ((compare-terms-lex-bigger right left followvarsflag)
           -1)
      (else
           #f))))


(define (compare-terms-lex-equal left right followvarsflag)
  (equal-terms-for-order? left right followvarsflag) )


(define (compare-terms-lex-bigger left right followvarsflag)
  ;(newline)
  ;(display "lex on:")(newline)
  ;(display left)(newline)
  ;(display right)(newline)
  ;(display followvarsflag)(newline)


  (let* ((tmp '())
         (lres '())
         (stopflag #f) )
    (cond
     ((eq? left right) #f) ; syntactically equal terms are equal
     ((var? left) #f)
     ((var? right)
        (if (and (composite? left)
                 (occurs-term-for-order? left right followvarsflag) )
            #t
            #f))
     ; now left and right cannot be vars
     ((non-composite? left)
         (cond
           ((non-composite? right)
              (if (eq? (compare-symbs-lex-order left right) 1)
                   #t
                   #f))
           ; right is a composite
           (else #f) )); t(..) > c
     ; now left is a composite
     ((non-composite? right) ; right is a noncomposite
           #t); t(..) > c
     ; now both left and right are composite
     ((begin  ; precompute by-arg comparisons with the main term
       ;(newline)(display "check Y0")(newline)
       (set! lres
             (some? (lambda(x)
                      (or (compare-terms-lex-equal
                              x right followvarsflag)
                          (compare-terms-lex-bigger
                               x right followvarsflag) ))
                    (cdr left) ))
       #f))
     (lres #t)
     ; now starts the main lex comparison of composites
     ((> (fun-lex-order-value (car left)) ; left bigger than right
         (fun-lex-order-value (car right)) )
          ;(newline)(display "check Y4")(newline)
        (if (every?
              (lambda(arg)
                (compare-terms-lex-bigger left arg followvarsflag) )
              (cdr right) )
            #t
            #f))
     ((> (fun-lex-order-value (car right)) ; right bigger than left
         (fun-lex-order-value (car left)) )
       #f)
     ; now functions symbols are equal: recurse lexicographically
     (else
       (set! tmp #f)
       (set! stopflag #f)
        ;(newline)(display "check Y5")(newline)
       ; loop, comparing args lexicographically
       (do ((lpart (cdr left) (cdr lpart))
            (rpart (cdr right) (cdr rpart)) )
           ((or (null? lpart)
                (null? rpart)
                stopflag)
               ;(newline)(display "check Y6")(newline)
              tmp)
         (if (not (compare-terms-lex-equal
                    (car lpart) (car rpart) followvarsflag))
             (begin
               (set! stopflag #t)
               (if (compare-terms-lex-bigger
                      (car lpart) (car rpart) followvarsflag)
                   (set! tmp #t)
                   (set! tmp #f) ))))
        ; loop, comparing full term with args of the second term
        (cond
          ((not tmp) #f)
          (else
               ;(newline)(display "check Y8")(newline)
             (if (every?
                    (lambda(arg)
                       (compare-terms-lex-bigger
                         left arg followvarsflag))
                    (cddr right) )
                  #t
                  #f)))) )))


;----------------











; occurs-term-for-order? checks if rterm properly occurs inside lterm.
; if yes, it returns #t, else #f

(define (occurs-term-for-order? lterm rterm followvarsflag)
  (let* ((stopflag #f) )
    (cond
     ((eq? lterm rterm)  #f)
     ((or (non-composite? lterm)
          (and (not (fun-may-cont-vars? (car lterm)))
               (or (var? rterm)
                   (and (composite? rterm)
                        (>= (fun-depth-strength (car rterm))
                            (fun-depth-strength (car lterm)) )))))
        #f)
     (else
       (set! stopflag #f)
       (do ((lpart (cdr lterm) (cdr lpart)) )
           ((or (null? lpart)
                stopflag)
              stopflag)
         (if (or  (equal-terms-for-order?
                      (car lpart) rterm followvarsflag)
                  (occurs-term-for-order?
                      (car lpart) rterm followvarsflag) )
             (set! stopflag #t) ))) )))



(define (equal-terms-for-order? left right followvarsflag)
  (let* ((stopflag #f) )
    (cond
     ((eq? left right)  #t)
     ((var? left)
         (and (var? right)
              (eq? (var-nr left) (var-nr right)) ))
     ((or (non-composite? left) (non-composite? right)) #f)
     ((and (not followvarsflag)
           (not (eq? (car left) (car right))) )
        #f)
     ((and followvarsflag
           (if (fun-may-cont-vars? (car left))
               ; left is potentially non-ground
               (if (fun-may-cont-vars? (car right))
                   ; both may contain vars
                   #f
                   ; left may contain vars,  right is ground
                   ; only left can grow bigger through subst
                   (> (fun-depth-strength (car left))
                      (fun-depth-strength (car right)) ))
               ; left is surely ground
               (if (fun-may-cont-vars? (car right))
                   ; only right may contain vars
                   (> (fun-depth-strength (car right))
                      (fun-depth-strength (car left)) )
                   ; both are ground
                   (not (eq? (car left) (car right))) )))
          #f)
     ((eq? (pure-fun (car left)) (pure-fun (car right)))
       (set! stopflag #f)
       (do ((lpart (cdr left) (cdr lpart))
            (rpart (cdr right) (cdr rpart)) )
           ((or (null? lpart)
                (null? rpart)
                stopflag)
              (not stopflag) )
         (if (not (equal-terms-for-order?
                      (car lpart) (car rpart) followvarsflag))
             (set! stopflag #t) )))
     (else #f) )))


(define *const-lex-order-vector* '())
(define *fun-lex-order-vector* '())

(define *lex-order-prefer-bigger-arity* #f)

(define (make-symbs-lex-order-vector)
  (let ((x 0)
        (count 0)
        (constlst '())
        (funlst '()) )
    (set! *const-lex-order-vector* (make-vector 10000))
    (set! *fun-lex-order-vector* (make-vector 300))
    ; *const-num-table*: ((b . 65537) (a . 65536)) etc
    ; *fun-num-table:  ((g . 4194304) (f . 2097152)) etc: use (logsright n *fnsymb-shift*)
    ; *fun-num-arity-table:  ((4194304 . 2) (2097152 . 1)) etc
    ; create initial lists
    (set! constlst (map (lambda(x)(cdr x)) *const-num-table*))
    (set! funlst (map (lambda(x)(car x)) *fun-num-arity-table*))
    ; sort lists
    (if *lex-order-prefer-bigger-arity*
        (begin
          ; (set! constlst (reverse constlst))
          (set! funlst (sort-funlist-lex funlst)) )
        (begin
          (set! constlst (reverse constlst))
          (set! funlst (reverse (sort-funlist-lex funlst))) ))
    ;(newline)
    ;(display "funlst: ")(display funlst)(newline)
    ;(display "fun-num: ")(display *fun-num-table*)(newline)
    ;(display "fun-arities: ")(display *fun-num-arity-table*)(newline)
    ; fill vector for certainty
    (do ((x 0 (+ 1 x)))
        ((>= x 10000))
     (vector-set! *const-lex-order-vector* x (+ 100 x)) )
    (do ((x 0 (+ 1 x)))
        ((>= x 300))
     (vector-set! *fun-lex-order-vector* x (+ 1000 x)) )
     ; use lists to fill right values for consts and funs
     (set! count 100)
     (for-each (lambda(x)
                 (set! count(+ 1 count))
                 (if (< (- x *minconst*) 10000)
                     (vector-set!
                        *const-lex-order-vector*
                        (- x *minconst*)
                        count) ))
                (reverse constlst) )
     (set! count 1000)
     (for-each (lambda(x)
                 (set! count(+ 1 count))
                 (vector-set!
                    *const-lex-order-vector*
                    (shiftright (pure-fun x) *fnsymb-shift*)
                    count))
               (reverse funlst) )))



(define (sort-funlist-lex lst)
  (if (> (length lst) 1)
      (sort-funlist-lex-aux lst (length lst))
      lst))

(define (sort-funlist-lex-aux lst len)
  (cond ((< len 2) lst)
        ((eq? len 2)
           (if (bigger-fun-for-lex-order? (car lst) (cadr lst))
               (list (car lst) (cadr lst))
               (list (cadr lst) (car lst)) ))
        (else
          (let* ((l1 (quotient len 2))
                 (l2 (- len l1))
                 (s (split-list lst l1)) )
            (sort-merge-lex-funlists
               (sort-funlist-lex-aux (car s) l1)
               (sort-funlist-lex-aux (cdr s) l2) )))))

(define (sort-merge-lex-funlists cl1 cl2)
  (cond ((null? cl1) cl2)
        ((null? cl2) cl1)
        ((bigger-fun-for-lex-order? (car cl1) (car cl2))
           (cons (car cl1)
                 (sort-merge-lex-funlists (cdr cl1) cl2) ))
        (else
           (cons (car cl2)
                 (sort-merge-lex-funlists cl1 (cdr cl2)) ))))


(define (bigger-fun-for-lex-order? f1 f2)
   (or (> (cdr (assq (pure-fun f1) *fun-num-arity-table*))
          (cdr (assq (pure-fun f2) *fun-num-arity-table*)) )
       (and
          (= (cdr (assq (pure-fun f1) *fun-num-arity-table*))
             (cdr (assq (pure-fun f2) *fun-num-arity-table*)) )
          (> (pure-fun f1) (pure-fun f2)) )))



(define (const-lex-order-value x)
   ; (newline)(display "const-lex-order-value: ")(display x)(newline)
  (let ((tmp (- x *minconst*)))
    (if (< tmp 10000)
        (vector-ref *const-lex-order-vector* tmp)
        1)))

(define (fun-lex-order-value x)
  ; (newline)(display "const-lex-order-value: ")(display x)
  ;(display " shifted: ")
  ;(display (shiftright (pure-fun x) *fnsymb-shift*))(newline)
  (vector-ref *fun-lex-order-vector* (shiftright (pure-fun x) *fnsymb-shift*)) )


(define (compare-symbs-lex-order left right)
   (let ((lval (const-lex-order-value left))
         (rval (const-lex-order-value right)) )
     (cond
       ((eq? lval rval) 0)
       ((> lval rval) 1)
       ((< lval rval) -1)
       (else #f) )))



; order-equality-litinfo-kb takes a litinfolst ((x y <infonr> ..) ...)
; and a pure clause and returns a new litinfolst with
; resolution and factoring bits set correctly for ordering
; (ie determining which literals are allowed to be resolved upon)
; NB! bits are changed only for prohibited lits.


(define (kb-order-litinfo-aux litinfolst purecl )
  (let* ((smalllits '()) ; the set of prohibited lits
         (biglits '())   ; the set of not yet-looked-through lits
         ; the following is a list of pairs of vardepth-infos and literals:
         ; ((((x . 1)(y . 3)) p (g x) (h (g y)) x) ...)
         (newlitinfnr 0)
         (varinfo-purecl
           (map (lambda(x)
                  (cons (make-var-depth-info x) x))
                 purecl) ))
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (some? (lambda(x)
                            ;(and (not (eq? x varinfolit))
                            ;         (or *order-strategy-flag*
                            ;             (eq? (pure-fun (cadr x)) *eq-pred-nr*) )
                            ;         (vdepth->-varinfolit? x varinfolit) )
                             (and (not (eq? x varinfolit))
                                  (or *order-strategy-flag*
                                      (eq? (pure-fun (cadr x)) *eq-pred-nr*) )
                                  (if (not (fun-may-cont-vars? (cadr varinfolit)))
                                      (and (not (fun-may-cont-vars? (cadr x)))
                                           (> (fun-depth (cadr x))
                                              (fun-depth (cadr varinfolit)) ))
                                      (and (fun-may-cont-vars? (cadr x))
                                           (> (fun-depth (cadr x))
                                              (fun-depth (cadr varinfolit)) )
                                           (if *sat-strategy-flag*
                                               (vdepth->-varinfolit?
                                                  x varinfolit)
                                               (vdepth->-weak-varinfolit?
                                                 x varinfolit) )))))
                          varinfo-purecl))
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))


(define *tmp-var-occurrences-table* '())

(define (make-kb-order-term-info term)
  (let ((sum 0))
    (set! *tmp-var-occurrences-table* '())
    (if (composite? term)
        (set! sum (make-kb-order-term-info-aux term 1 0))
        (if (var? term)
            (set! *tmp-var-occurrences-table* (cons (var-nr term) 1))
            (set! sum (get-const-term-lex term)) ))
    (cons sum *tmp-var-occurrences-table*) ))

(define (make-kb-order-term-info-aux origterm depth sum)
  (let ((tmp '())
        (term '()) )
    (do ((part (cdr origterm) (cdr part)))
        ((null? part) (+ (get-fun-term-lex term) sum))
      (set! term (car part))
      (if (non-composite? term)
          ; term is non-composite
          (if (non-composite:-var? term)
              (begin (set! tmp (assq (var-nr term) *tmp-var-occurrences-table*))
                     (if tmp
                         (set-cdr! tmp (+ 1 (cdr tmp)))
                         (set! *tmp-var-occurrences-table*
                               (cons (cons (var-nr term) 1)
                                     *tmp-var-occurrences-table*))))
              (set! sum (+ sum (get-const-term-lex term))) )
          ; term is composite
          (if (fun-may-cont-vars? (car term))
              (make-var-depth-occurrences-aux term (+ 1 depth)) )))))


(define *lex-symbol-orderlst* '())

(define (lex-bigger? left right)
  (let* ((foundflag #f)
         (res #f)
         (lp '())
         (rp '()) )
    (cond
     ((and (non-composite? left) (non-composite? right))
        (cond ((var? left) #f)
              ((var? right) #f)
              ((lex-symbol-bigger? left right) #t)
              ((lex-symbol-bigger? right left) #f)
              ((eq? left right) #f)
              (else
                (make-lex-symbol-bigger! left right)
                #t)))
     ((non-composite? right)
        ; left is now composite
        (or (not (var? right))
            (lex-occur-check? right left) ))
     ((non-composite? left)
        ; right is now composite
        #f)
     ; now both left and right are composite
     ((eq? (pure-fun (car left)) (pure-fun (car right))) ; same heads?
        (do ((lp (cdr left) (cdr lp))
             (rp (cdr right) (cdr rp)) )
            ((or (null? lp) foundflag)
              (and res (lex-bigger-than-subterms? left right)) )
          (cond
            ((lex-bigger? (car lp) (car rp))
               (set! foundflag #t)
               (set! res #t) )
            ; ((not (lex-bigger? (car rp) (car lp))))
            ((or (term-equal? (car lp) (car rp))
                 (and (var? (car rp))
                      (composite? (car lp))
                      (lex-occur-check? (car rp) (car lp)) )))
            (else
                 (set! foundflag #t)
               (set! res #f) ))))
     ; now both left and right are composite with different heads
     ((and (> (length right) (length left))
           (lex-symbol-bigger? (car left) (car right))
           (lex-bigger-than-subterms? left right) )
        ; (make-lex-symbol-bigger! (car left) (car right))
        ; (lex-bigger-than-subterms? left right)
             #t)
     ;((> (length left) (length right))
     ;   ; (make-lex-symbol-bigger! (car right) (car left))
     ;         #f)
     ((lex-symbol-bigger? (car left) (car right))
        (lex-bigger-than-subterms? left right) )
     ((lex-symbol-bigger? (car right) (car left))
        #f)
        ; (lex-subterm-bigger-than left right)
     ; try to make symbol bigger
     (else
        (make-lex-symbol-bigger! (car left) (car right))
        (lex-bigger-than-subterms? left right) ))))

(define (lex-occur-check? var composite)
  (set! composite (cdr composite))
  (do ()
      ((null? composite)
         (eq? var #t) ) ; var is used as a succ/failure flag here
      (let ((el (car composite)))
        (if (non-composite? el)
            (if (non-composite:-var? el)
                (begin
                   (if (or (and (var? el) (eq? var (var-nr el)))
                           (and (composite? el)
                                (term-may-cont-vars? el)
                                (lex-occur-check? var el) ))
                       (begin (set! composite '()) ; signal end
                              (set! var #t) )      ; signal: occurs
                       (set! composite (cdr composite)) ))
                (set! composite (cdr composite)) )
            (if (and (term-may-cont-vars? el)
                     (lex-occur-check? var el) )
                (begin (set! composite '()) ; signal end
                       (set! var #t) )      ; signal: occurs
                (set! composite (cdr composite)) )))))

(define (lex-bigger-than-subterms? left right)
  (or (non-composite? right)
      (every? (lambda(x) (lex-bigger? left x))
              (cdr right) )))

(define (lex-subterm-bigger-than left right)
  (and (not (non-composite? left))
       (some? (lambda(x)(lex-bigger? x right)) (cdr left)) ))

(define (lex-symbol-bigger? left right)
  (let ((tmp (memq (pure-fun right) *lex-symbol-orderlst*)))
    (cond
      ((not tmp) #f)
      ((memq (pure-fun left) (cdr tmp)) #t)
      (else #f))))

(define (make-lex-symbol-bigger! left right)
  (set! left (pure-fun left))
  (set! right (pure-fun right))
  (if (not (memq left *lex-symbol-orderlst*))
      (set! *lex-symbol-orderlst* (append *lex-symbol-orderlst* (list left))) )
  (if (not (memq right *lex-symbol-orderlst*))
      (set! *lex-symbol-orderlst* (cons right *lex-symbol-orderlst*)) ))


; - - - - firstterm-ordering - - - - -

; order-equality-litinfo-firstterm prefers terms where the first arg is deep

(define (order-equality-litinfo-firstterm litinfonr lit)
  (let* ((left (if *type-theory-stratflag* (cadadr lit) (cadr lit)))
         (right (if *type-theory-stratflag* (caddr (cadr lit)) (caddr lit)))
         (leftvars (make-var-depth-info left))
         (rightvars (make-var-depth-info right)) )
    (cond
     ((and (non-composite? left) (non-composite? right))
        (cond ((var? left)
                 (if (var? right)
                     (prohibit-left-para
                        (prohibit-right-para litinfonr) )
                     (prohibit-right-para litinfonr) ))
              ((var? right) (prohibit-left-para litinfonr))
              (else litinfonr) ))
     ((non-composite? right)
         ;left is now composite
        (if (or (not (var? right))
                (assq (var-nr right) leftvars) )
            (prohibit-right-para litinfonr)
            litinfonr))
     ((non-composite? left)
        ; right is now composite
        (if (or (not (var? left))
                (assq (var-nr left) rightvars) )
            (prohibit-left-para litinfonr)
            litinfonr))
     ; now both left and right are composite
     ((and (vdepth-superset-varinfolit?
            (cons leftvars left) (cons rightvars right) )
           (first-term-depth-bigger? left right) )

        ;(newline)
        ;(display "right prohibited (l,r): ")
        ;(print-term left)(newline)
        ;(print-term right)(newline)

        (prohibit-right-para litinfonr) )
     ((and (vdepth-superset-varinfolit?
            (cons rightvars right) (cons leftvars left) )
           (first-term-depth-bigger? right left) )


        ;(newline)
        ;(display "left prohibited (l,r): ")
        ;(print-term left)(newline)
        ;(print-term right)(newline)

        (prohibit-left-para litinfonr) )
     (else litinfonr) )))

; xterm and yterm must be composite

(define (first-term-depth-bigger? xterm yterm)
  (let ()
    (and (>= (fun-depth (car xterm)) (fun-depth (car yterm)))
         (> (first-term-depth xterm)
            (first-term-depth yterm) ))))


(define (first-term-depth term)
    (if (non-composite? term)
        0
        (+ 1 (first-term-depth (cadr term))) ))


; - - - - depth-ordering - - - - -


(define (order-equality-litinfo-depth litinfonr lit)
  (let* ((left (if *type-theory-stratflag* (cadadr lit) (cadr lit)))
         (right (if *type-theory-stratflag* (caddr (cadr lit)) (caddr lit)))
         (leftvars (make-var-depth-info left))
         (rightvars (make-var-depth-info right)) )
    (cond
     ((and (non-composite? left) (non-composite? right))
        (cond ((var? left)
                 (if (var? right)
                     (prohibit-left-para
                        (prohibit-right-para litinfonr) )
                     (prohibit-right-para litinfonr) ))
              ((var? right) (prohibit-left-para litinfonr))
              (else litinfonr) ))
     ((non-composite? right)
        ; left is now composite
        (if (or (not (var? right))
                (assq (var-nr right) leftvars) )
            (prohibit-right-para litinfonr)
            litinfonr))
     ((non-composite? left)
        ; right is now composite
        (if (or (not (var? left))
                (assq (var-nr left) rightvars) )
            (prohibit-left-para litinfonr)
            litinfonr))
     ; now both left and right are composite
     ((or (and (vdepth->-varinfolit?
                (cons leftvars left) (cons rightvars right) )
               (> (fun-depth (car left)) (fun-depth (car right))) )
          (and (null? rightvars)
               (> (fun-depth (car left)) (fun-depth (car right))) ))
        (prohibit-right-para litinfonr) )
     ((or (and (vdepth->-varinfolit?
                 (cons rightvars right) (cons leftvars left) )
               (> (fun-depth (car right)) (fun-depth (car left))))
          (and (null? leftvars)
               (> (fun-depth (car right)) (fun-depth (car left))) ))
        (prohibit-left-para litinfonr) )
     (else litinfonr) )))

; - - - - variable-depth-ordering - - - - -

; vdepth-order-litinfo takes a litinfolst ((x y <infonr> ..) ...)
; and a pure clause and returns a new litinfolst with
; resolution and factoring bits set correctly for ordering
; (ie determining which literals are allowed to be resolved upon)
; NB! bits are changed only for prohibited lits.


(define (vdepth-order-litinfo-aux litinfolst purecl )
  ;(newline)(display "ohoo")
  (let* ((smalllits '()) ; the set of prohibited lits
         (biglits '())   ; the set of not yet-looked-through lits
         ; the following is a list of pairs of vardepth-infos and literals:
         ; ((((x . 1)(y . 3)) p (g x) (h (g y)) x) ...)
         (newlitinfnr 0)
         (varinfo-purecl
           (map (lambda(x)
                  (cons (make-var-depth-info x) x))
                 purecl) ))
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (some? (lambda(x)
                            ;(and (not (eq? x varinfolit))
                            ;         (or *order-strategy-flag*
                            ;             (eq? (pure-fun (cadr x)) *eq-pred-nr*) )
                            ;         (vdepth->-varinfolit? x varinfolit) )
                             (and (not (eq? x varinfolit))
                                  (or *order-strategy-flag*
                                      (eq? (pure-fun (cadr x)) *eq-pred-nr*) )
                                  (if (not (fun-may-cont-vars? (cadr varinfolit)))
                                      (and (not (fun-may-cont-vars? (cadr x)))
                                           (if *complete-unsat-strategy*
                                               (> (fun-depth (cadr x))
                                                  (fun-depth (cadr varinfolit)) )
                                               (> (fun-depth-size (cadr x))
                                                  (fun-depth-size (cadr varinfolit)) )))
                                      (and (fun-may-cont-vars? (cadr x))
                                           (if *complete-unsat-strategy*
                                               (> (fun-depth (cadr x))
                                                  (fun-depth (cadr varinfolit)) )
                                               (> (fun-depth-size (cadr x))
                                                  (fun-depth-size (cadr varinfolit)) ))
                                           (if *sat-strategy-flag*
                                               (vdepth->-varinfolit?
                                                  x varinfolit)
                                               (vdepth->-weak-varinfolit?
                                                 x varinfolit) )))))
                          varinfo-purecl))
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))



(define (vdepth-order-weak-litinfo-aux litinfolst purecl )
  (let* ((smalllits '()) ; the set of prohibited lits
         (biglits '())   ; the set of not yet-looked-through lits
         ; the following is a list of pairs of vardepth-infos and literals:
         ; ((((x . 1)(y . 3)) p (g x) (h (g y)) x) ...)
         (newlitinfnr 0)
         (varinfo-purecl
           (map (lambda(x)
                  (cons (make-var-depth-info x) x))
                  purecl) ))
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (some? (lambda(x)
                            (and (not (eq? x varinfolit))
                                 (or *order-strategy-flag*
                                     (eq? (pure-fun (cadr x)) *eq-pred-nr*) )
                                 (if (ground? (cadr varinfolit))
                                     (and (ground? (cadr x))
                                          (>-depth? (cadr x) (cadr varinfolit)) )
                                     (and (non-ground (cadr x))
                                          (vdepth->-weak-varinfolit? x varinfolit) ))))
                          varinfo-purecl))
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))


(define (vdepth-order-litinfo-eq-only litinfolst purecl)
  (let* ((newlitinfnr 0))
    (map (lambda(litinf lit)
           (if (or (eq? (pure-fun (car lit)) *eq-pred-nr*)
                   (and *type-theory-stratflag*
                        (composite? (cadr lit))
                        (eq? (pure-fun (caadr lit)) *eq-fun-nr*)
                        (not (neg-fun? (car lit))) ))
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 (set! newlitinfnr
                       (order-equality-litinfo newlitinfnr lit) )
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf)) )))
               litinf))
         litinfolst
         purecl)))

; - - - - - - - order literals in clause only - - - - - -

; name-order-litinfo orders literals by the name of the
; predicate


(define (name-order-litinfo litinfolst purecl )
  (let* (; the following is a list of pairs of infos and literals:
         ; (#t p (g x) (h (g y)) x) ...
         (varinfo-purecl
           (map (lambda(x)(cons #t x)) purecl) )
         (newlitinfnr 0) )
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (some? (lambda(x)
                            (and (not (eq? x varinfolit))
                                 (> (pure-fun (cadr varinfolit))
                                    (pure-fun (cadr x)) )))
                          varinfo-purecl))
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))


; epr-order-litinfo orders literals suitably for epr


(define (epr-order-litinfo litinfolst purecl )

  ;(newline)
  ;(display "epr-order called")
  ;(newline)

  (let* (; the following is a list of pairs of infos and literals:
         ; (#t p (g x) (h (g y)) x) ...
         (varinfo-purecl
           (map (lambda(x)(cons #t x)) purecl) )
         (newlitinfnr 0) )
    ;(newline)(display      varinfo-purecl)(newline)
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (some? (lambda(x)
                            (epr-order-litinfo-aux varinfolit x) )
                           varinfo-purecl))
                   ;(some? (lambda(x)
                   ;         (and (not (eq? x varinfolit))
                   ;              (> (pure-fun (cadr varinfolit))
                   ;                 (pure-fun (cadr x)) )))
                   ;       varinfo-purecl))
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))


(define (epr-order-litinfo-aux ailit bilit)
 (let ((alen (length (cdr ailit)))
       (blen (length (cdr bilit))) )
   (and (not (eq? ailit bilit))
        (or (< alen blen)
            (and
              (eq? alen blen)
              (or (and (not (neg-fun? (cadr ailit)))
                       (neg-fun? (cadr bilit)) )
                  (and (not (and (neg-fun? (cadr ailit))
                                 (not (neg-fun? (cadr bilit))) ))
                       (or (< (pure-fun (cadr ailit))
                              (pure-fun (cadr bilit)) )
                           (and (eq? (pure-fun (cadr ailit))
                                     (pure-fun (cadr bilit)) )
                                (const? (caddr ailit))
                                (const? (caddr bilit))
                                (< (caddr ailit)
                                   (caddr bilit) ))))))))))




(define (poslit-order-litinfo litinfolst purecl )
  (let* (; the following is a list of pairs of infos and literals:
         ; (#t p (g x) (h (g y)) x) ...
         (varinfo-purecl
           (map (lambda(x)(cons #t x)) purecl) )
         (mix-flag (mixed-cl? purecl))
         (newlitinfnr 0) )
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (and mix-flag
                        (neg-fun? (cadr varinfolit)) ))
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))


(define (firstlitpref-order-litinfo litinfolst purecl)
  (let* (; the following is a list of pairs of infos and literals:
         ; (#t p (g x) (h (g y)) x) ...
         (varinfo-purecl
           (map (lambda(x)(cons #t x)) purecl) )
         (newlitinfnr 0)
         (first-flag #t) )
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (not first-flag) )
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 (set! first-flag #f)
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))


; posweight-order-litinfo allows to resolve only upon
; the first negative literal (if negative are present)
; and if no negative ones are present, it allows
; to resolve upon all the positive ones
; it also orders equality
; Present in 3 series, new for 2.4

(define (posweight-order-litinfo litinfolst purecl)
  (let* (; the following is a list of pairs of infos and literals:
         ; (#t p (g x) (h (g y)) x) ...
         (varinfo-purecl
           (map (lambda(x)(cons #t x)) purecl) )
         (newlitinfnr 0)
         (hasnegflag (some? (lambda(x)(neg-fun? (cadr x))) varinfo-purecl))
         (negallowedflag #f) )
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (and hasnegflag
                        (not (neg-fun? (cadr varinfolit))) )
                   negallowedflag)
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 (if (neg-fun? (cadr varinfolit))
                     (set! negallowedflag #t) )
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))


; negweight-order-litinfo allows to resolve only upon
; the first positive literal (if positive are present)
; and if no positive ones are present, it allows
; to resolve upon all the negative ones
; it also orders equality
; New for 2.4

(define (negweight-order-litinfo litinfolst purecl)
 ;(display "negweight-order-litinfo called")(newline)
 (let* (; the following is a list of pairs of infos and literals:
        ; (#t p (g x) (h (g y)) x) ...
        (varinfo-purecl
          (map (lambda(x)(cons #t x)) purecl) )
        (newlitinfnr 0)
        (hasposflag (some? (lambda(x)(not (neg-fun? (cadr x)))) varinfo-purecl))
        (posallowedflag #f)
        (res '())
        (tmporderflag *order-strategy-flag*) )
  ;(display "hasposflag:")(display hasposflag)(newline)
  (cond
   (hasposflag
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (and hasposflag
                        (neg-fun? (cadr varinfolit)) )
                   posallowedflag)
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 (if (not (neg-fun? (cadr varinfolit)))
                     (set! posallowedflag #t) )
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl))
   (else
      (set! tmporderflag *order-strategy-flag*)
      (set! *order-strategy-flag* #t)
      (set! res (vdepth-order-litinfo litinfolst purecl))
      (set! *order-strategy-flag* tmporderflag)
      res))))


; experimental order for 2.6


(define (exp-order-litinfo-suva litinfolst purecl )
  ;(newline)(display "ohoo")
  (let* ((smalllits '()) ; the set of prohibited lits
         (biglits '())   ; the set of not yet-looked-through lits
         ; the following is a list of pairs of vardepth-infos and literals:
         ; ((((x . 1)(y . 3)) p (g x) (h (g y)) x) ...)
         (newlitinfnr 0)
         (varinfo-purecl
           (map (lambda(x)
                  (cons (make-var-depth-info x) x))
                 purecl) ))
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (some? (lambda(x)
                            ;(and (not (eq? x varinfolit))
                            ;         (or *order-strategy-flag*
                            ;             (eq? (pure-fun (cadr x)) *eq-pred-nr*) )
                            ;         (vdepth->-varinfolit? x varinfolit) )
                             (and (not (eq? x varinfolit))
                                  (if (not (fun-may-cont-vars? (cadr varinfolit)))
                                      (if (not (fun-may-cont-vars? (cadr x)))
                                          (> (fun-depth-size (cadr x))
                                             (fun-depth-size (cadr varinfolit))  )
                                          (>= (fun-depth-size (cadr x))
                                              (fun-depth-size (cadr varinfolit))  ))
                                      (and (fun-may-cont-vars? (cadr x))
                                           (or (and (vdepth-superset-varinfolit?
                                                          x  varinfolit)
                                                    (not (vdepth-superset-varinfolit?
                                                          varinfolit x)))
                                               (> (fun-depth-size (cadr x))
                                                  (fun-depth-size (cadr varinfolit)) )))
                                                      )))
                                               ;(> (fun-depth-size (cadr x))
                                               ;   (fun-depth-size (cadr varinfolit)) ))

                                           ;(if *sat-strategy-flag*
                                           ;    (vdepth->-varinfolit?
                                           ;       x varinfolit)
                                           ;    (vdepth->-weak-varinfolit?
                                           ;      x varinfolit) )))))
                          varinfo-purecl))
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))



(define (exp-order-litinfo litinfolst purecl )

  ;(newline)
  ;(display "exp-order called")
  ;(newline)

  (let* (; the following is a list of pairs of infos and literals:
         ; (#t p (g x) (h (g y)) x) ...
         (varinfo-purecl
           (map (lambda(x)(cons #t x)) purecl) )
         (newlitinfnr 0) )
    ;(newline)(display      varinfo-purecl)(newline)
    (map (lambda(litinf varinfolit)
           (if (or (answer-lit? (cdr varinfolit))
                   (some? (lambda(x)
                            (exp-order-litinfo-aux (cdr varinfolit) (cdr x)) )
                           varinfo-purecl))
                   ;(some? (lambda(x)
                   ;         (and (not (eq? x varinfolit))
                   ;              (> (pure-fun (cadr varinfolit))
                   ;                 (pure-fun (cadr x)) )))
                   ;       varinfo-purecl))
               ; not allowed to be resolved upon: set prohibition bits
               (begin (set-car! varinfolit '())
                      (cons (car litinf)
                            (cons (cadr litinf)
                                  (cons
                                   (logior (caddr litinf) *litinf-nouse-mask*)
                                   (cdddr litinf) ))))
                 ; allowed to be resolved upon: clear prohibition bits
               (begin
                 ; clear all prohibition bits
                 (set! newlitinfnr
                       (logxor *litinf-nouse-mask*
                               (logior (caddr litinf) *litinf-nouse-mask*)))
                 ; order the equality
                 (if (and *ordered-paramodulation-flag*
                          (or (eq? (pure-fun (cadr varinfolit)) *eq-pred-nr*)
                              (and *type-theory-stratflag*
                                   (composite? (caddr varinfolit))
                                   (eq? (pure-fun (caaddr varinfolit))
                                        *eq-fun-nr*)
                                   (not (neg-fun? (cadr varinfolit))) )))
                     (set! newlitinfnr
                           (order-equality-litinfo
                              newlitinfnr (cdr varinfolit) )))
                 ; return the full litinf
                 (cons (car litinf)
                       (cons (cadr litinf)
                             (cons newlitinfnr (cdddr litinf) ))))))
         litinfolst
         varinfo-purecl)))


(define (exp-order-litinfo-aux ailit bilit)
 ;(newline)
 ;(display "exp-order-litinfo-aux: ")
 ;(print-lit ailit)
 ;(print-lit bilit)
 ;(newline)
 (let ((tmp '()))
   (cond
     ((eq? ailit bilit)
        #f)
     ((begin (set! tmp (compare-terms-kb-order ailit bilit #f))
        #f))
     ;((begin (set! tmp (compare-terms-lex-order ailit bilit #f))
     ;   #f))
     ((eq? tmp -1)
        #t)
     (else
        #f))))


(define (clear-litorder-litinfo litinfolst purecl)
  (let* ((newlitinfnr 0))
    (map (lambda(litinf lit)
           (set! newlitinfnr
                 (logxor *litinf-nouse-mask*
                         (logior (caddr litinf) *litinf-nouse-mask*)))
           (if (or (eq? (pure-fun (car lit)) *eq-pred-nr*)
                   (and *type-theory-stratflag*
                        (composite? (cadr lit))
                        (eq? (pure-fun (caadr lit)) *eq-fun-nr*)
                        (not (neg-fun? (car lit))) ))
               (set! newlitinfnr
                     (order-equality-litinfo newlitinfnr lit) ))

           (cons (car litinf)
                 (cons (cadr litinf)
                       (cons newlitinfnr (cdddr litinf)) )))
         litinfolst
         purecl)))

(define (vdepth->-varinfolit? alit blit)
  (and (not (null? (car blit)))
       (every? (lambda(pr)
                 (let ((tmp (assq (car pr) (car alit))))
                   (and tmp
                        (> (cdr tmp) (cdr pr)))))
               (car blit) )))

(define (vdepth->-weak-varinfolit? alit blit)
  (and (not (null? (car blit)))
       (every? (lambda(pr)
                 (let ((tmp (assq (car pr) (car alit))))
                   (and tmp
                        (>= (cdr tmp) (cdr pr)))))
               (car blit) )
       (some? (lambda(pr)
                 (let ((tmp (assq (car pr) (car alit))))
                   (and tmp
                        (> (cdr tmp) (cdr pr)))))
               (car blit) )))


(define (vdepth-superset-varinfolit? alit blit)
  (or (null? (car blit))
      (every? (lambda(pr)
                 (assq (car pr) (car alit)) )
              (car blit) )))

(define *tmp-var-depth-table* '())

(define (make-var-depth-info term)
  (set! *tmp-var-depth-table* '())
  (if (and (composite? term) (term-may-cont-vars? term)
           (not (answer-lit? term)) )
      (make-var-depth-info-aux term 1)
      (if (var? term)
          (set! *tmp-var-depth-table* (list (cons (var-nr term) 0))) ))
  *tmp-var-depth-table*)

(define (make-var-depth-info-aux term depth)
  (let ((tmp '()))
    (do ((part (cdr term) (cdr part)))
        ((null? part))
      (set! term (car part))
      (if (non-composite? term)
          (if (non-composite:-var? term)
              (begin (set! tmp (assq (var-nr term) *tmp-var-depth-table*))
                     (if tmp
                         (if (> depth (cdr tmp)) (set-cdr! tmp depth))
                         (set! *tmp-var-depth-table*
                               (cons (cons (var-nr term) depth)
                                     *tmp-var-depth-table*)))))
          (if (fun-may-cont-vars? (car term))
              (make-var-depth-info-aux term (+ 1 depth)) )))))

;===================================================================
;
;            showing and manipulating the ready-built history
;
;===================================================================

; history is the sixth element of the clinfo, with clnumber being fifth
;
; unknown history:
; (unknown <newnr>)
;
; history structure for hyperresolution:
;
; all input clauses with clnumber <clnr> have the following history:
; (input <clnr>)
;
; all clauses (electrons) with <clnr> derived by hyperresolution:
; (hyper <newnr> <from-cl> <el1> ... <eln>)
;    where <newnr> is the number of the currently derived clause; for
;      clauses which are NOT saved, the <newnr> is always 0.
;    and <from-cl> is the history of the from-cl (normally (input <clnr>))
;    and <el1> ... <eln> have the form: (literal <el-hist> <el-cllitnr>)
;    and <el-hist> is the history of the electron clause
;    and <el-cllitnr> is the nr of the literal in the electron clause
;    in the standard (not necessary input) order.
;    original input order of the negative literals in the nucleon <from-cl>
; the temporary unsaved results of hyperresolution have not-yet-cut <eli>
; as ()
;
; all clauses derived by binary resolution:
; (binres <newnr> <el1> <el2>)
;    where parts have the same form as in case of hyperresolution.
;
; all clauses derived by factoring:
; (factor <newnr> <from-cl> <nr1> <nr2>)
;    where <nr1> and <nr2> are numbers of literals factored upon in the
;    standard order
;
; all clauses derived by paramodulation:
; (para <newnr> <fromhist> <tohist> <fromlitnr> <tolitnr>
;       <left-used-flag> <path>)


; show-history prints out the clauses in the history together
; with the info how and from what are they derived from.
; It builds no new clauses.

(define *simple-history-stratflag* #t)
(define *shown-clausenrs* '()) ; a list of clausenrs already displayed
(define *otter-prepared-history* '())
(define *simple-prepared-history* '())
(define *simple-history-clausenr-stack* '())
(define *all-history-clauses-list* '())

(define (show-history history)
  (if *simple-history-stratflag*
     (show-history-simple history)
     (show-history-complex history) ))

;  - - - - - -  simple history building - - - - -

(define (show-history-simple history)
  (let ((propclnumbers '()))
   ; show checkpoints
   (if (and (not (null? *checkpoint-keptnrs*))
           (>= *display-info-level* 20) )
      (begin
        (newline)
        (display "timer checkpoints: ")
        (print-list-as-term (cons 'c (reverse *checkpoint-keptnrs*)))
        (newline)(newline) ))
   ; use all clause lists accessible
   (for-each (lambda(lst)
              (set! *sos-clauses-list* (append lst *sos-clauses-list*)) )
             *sos-clauses-list-list*)
   (if (not (null? *usable-clauses-list*))
       (set! *usable-clauses-list* (cdr *usable-clauses-list*)) )
   (for-each (lambda(lst)
              (set! *usable-clauses-list*
               (append
                (if (null? lst) '() (cdr lst))
                *usable-clauses-list*)) )
              *usable-clauses-list-list*)
   (set! *usable-clauses-list* (cons '() *usable-clauses-list*))
   (set! *all-history-clauses-list*
         (append *all-back-subsumed-clauses*
                (append *sos-clauses-list*
                        (cdr *usable-clauses-list*))))
   ; now we will just have to use the  *all-history-clauses-list*

   ; init globals to empty lists
   (set! *otter-prepared-history* '())
   (set! *shown-clausenrs* '())
   (set! *simple-history-clausenr-stack* '())

   ;(newline)(pretty-print *all-history-clauses-list*)
   ; now we will prepare *simple-prepared-history*

   (if (and (pair? history)
            (eq? (car history) 'propositional-from-input)
            (not (null? *unsprop-history*)) )
       (begin
         (set! propclnumbers
               (show-history-simple-propclauses *unsprop-history*) )
         (set! *simple-history-clausenr-stack* propclnumbers) ))

   (show-history-simple-onehistory history)
   (set! *simple-prepared-history* (list (list (+ 1 *kept-cl-nr*) history '())))

   ;(newline)(display "initial *simple-history-clausenr-stack*: ")
   ;(display *simple-history-clausenr-stack*)(newline)

   (show-history-simple-stackprocess)

   ; now we will print out the simple history
   (newline)
   (display "START OF PROOF")
   (show-history-simple-printout *simple-prepared-history*)
   (newline)
   (display "END OF PROOF")
   (newline)
   (set! *display-info-level* 40)
   (newline)
   (display "Proof found by the following strategy:")
   (newline)
   (show-one-pass-strategy) ))



(define (show-history-simple-printout prepared-history)
  (let ((sorted-history
             (sort-by-firstel prepared-history) )
        (histelem '()) )
    (do ((histpart sorted-history (cdr histpart)))
        ((null? histpart))
      (set! histelem (car histpart))

      ;(newline)
      ;(display histelem)

      (newline)
      (display (car histelem))
      (display " [")
      (show-history-simple-printparents
         (show-history-simple-split-proofsteps (cadr histelem)) #f)
      (display "] ")
      (cond
         ((pair? (caddr histelem))
           (print-pure-clause (caddr histelem)) )
         ((eq? '? (caddr histelem))
           (display '?) )
         ((null? (caddr histelem))
           (display 'contradiction) )
         (else
           (display '??) )))))

(define (show-history-simple-split-proofsteps history)
   (let ((res '())
         (resel '()) )
      (do ((histpart history (cdr histpart)))
          ((null? histpart)
             (set! res (append (reverse resel) res))
             res)
        (cond ((symbol? (car histpart))
                 (set! res (append (reverse resel) res))
                 (set! resel (list (car histpart))) )
              (else
                 (set! resel (cons (car histpart) resel)) )))))

(define (show-history-simple-printparents lst inputflag)
  ;(newline)(display "show-history-simple-printparents called: ")(display lst)(newline)
  (let ()
    (cond
     ((null? lst))
     ((and (eq? (car lst) 'input)
           (every? (lambda(x)(not (symbol? x))) (cdr lst))
           (not inputflag) )
        #f) ; display nothing
     ((and (eq? 'para (car lst)) ; paramodulation
           (pair? (cadr lst))
           (pair? (caddr lst)) )
        (display "para:")
        (show-history-simple-printparents-aux (cadr lst))
        (if (pair? (cddr (caddr lst)))
            (if (caddr (caddr lst))
                (display ".1")
                (display ".2") ))
        (display ",")
        (show-history-simple-printparents-aux
           (list (car (caddr lst)) (cadr (caddr lst))) )
        (display ".")
        (show-history-simple-printparents-aux (reverse (cadddr (caddr lst))))
        (if (pair? (cdddr lst))
            (display ",") )
        (show-history-simple-printparents (cdddr lst) inputflag))
     ((symbol? (car lst)) ; symbol, ie some proof operation
        (display (car lst))
        (if (not (null? (cdr lst)))
            (display ":") )
        (show-history-simple-printparents (cdr lst) inputflag))
     ((and (pair? (car lst)) ; a pair: an argument to proof op, clause and lit nrs
           (number? (caar lst))
           (equal? (cadar lst) 1)
           (null? (cddar lst)) )
        (display (caar lst))
        (if (not (null? (cdr lst)))
            (display ",") )
        (show-history-simple-printparents (cdr lst) inputflag))
     ((pair? (car lst)) ; a pair: an argument to proof op, clause and lit nrs
        (show-history-simple-printparents-aux (car lst))
        (if (not (null? (cdr lst)))
            (display ",") )
        (show-history-simple-printparents (cdr lst) inputflag))
     (else ;  an argument to proof op, just clause
        (display (car lst))
        (if (not (null? (cdr lst)))
            (display ",") )
        (show-history-simple-printparents (cdr lst) inputflag) ))))


(define (show-history-simple-printparents-aux y)
  (if (not (null? y))
      (begin
        (display (car y))
        (if (not (null? (cdr y)))
            (display "."))
        (show-history-simple-printparents-aux (cdr y)) )))


(define (show-history-simple-onehistory history)
  (let ((histelem '())
        (orighistory history) )
    ; (newline)(display "handling one history: ")(display history)(newline)
    ; run over one history, push unworked clausenrs to stack
    (do ((histpart history (cdr histpart)))
        ((null? histpart))
      (set! histelem (car histpart))
      ; (newline)(display "histelem: ")(display histelem)(newline)
      (cond ((number? histelem)
                (or (member histelem *simple-history-clausenr-stack*)
                    (member histelem *shown-clausenrs*)
                    (set! *simple-history-clausenr-stack*
                         (cons histelem *simple-history-clausenr-stack*) )))
            ((and (pair? histelem)
                  (symbol? (car histelem))
                  (pair? (cdr histelem))
                  (number? (cadr histelem)) )
                (or (member (cadr histelem) *simple-history-clausenr-stack*)
                    (member (cadr histelem) *shown-clausenrs*)
                    (set! *simple-history-clausenr-stack*
                          (cons (cadr histelem) *simple-history-clausenr-stack*) )))
            ((and (pair? histelem)
                  (number? (car histelem)) )
                (or (member (car histelem) *simple-history-clausenr-stack*)
                    (member (car histelem) *shown-clausenrs*)
                    (set! *simple-history-clausenr-stack*
                          (cons (car histelem) *simple-history-clausenr-stack*) )))))))

(define (show-history-simple-stackprocess)
    ; run over clausenr stack until it becomes empty

  (let ((stackelem '())
        (stackelem-clause '()) )
    (do ()
        ((null? *simple-history-clausenr-stack*))
      (set! stackelem (car *simple-history-clausenr-stack*))
      (set! *simple-history-clausenr-stack* (cdr *simple-history-clausenr-stack*))
      ; (newline)(display "handling stackelem: ")(display stackelem)(newline)
      (set! *shown-clausenrs* (cons stackelem *shown-clausenrs*))
      (set! stackelem-clause (find-simple-history-clause stackelem))
      (if (not stackelem-clause)
          (set! *simple-prepared-history*
                (cons (list stackelem '(?) '?)
                      *simple-prepared-history*))
          (begin
            (show-history-simple-onehistory
                (car (clinfo-history-sublist (car stackelem-clause))) )
            (set! *simple-prepared-history*
                 (cons (list stackelem
                             (car (clinfo-history-sublist (car stackelem-clause)))
                             (cdr stackelem-clause) )
                       *simple-prepared-history*)))))))



(define (find-simple-history-clause stackelem)
 ;(newline)(display "starting to find stackelem: ")(display stackelem)(newline)
  (let ((endflag #f)
        (foundclause #f) )
    (do ((clauselstpart *all-history-clauses-list* (cdr clauselstpart)))
        ((or (null? clauselstpart) endflag foundclause)
             ;(newline)(display "foundclause: ")(display foundclause)(newline)
             foundclause)
      ;(display (caar clauselstpart))
      ;(display (clinfo-clnr (caar clauselstpart)))
      ;(newline)
      (if (and (pair? (car clauselstpart))
               (eq? (clinfo-clnr (caar clauselstpart)) stackelem) )
          (begin
            (set! foundclause (car clauselstpart)) )))))


; - - - old-style (complex) history showing - - - - - - -


(define (show-history-complex history)
  ;(set! *shown-clausenrs* '())
  ;(set! *otter-prepared-history* '())
  ;(show-history-aux history 0)
  ;(newline)
  (if (and (not (null? *checkpoint-keptnrs*))
           (>= *display-info-level* 20) )
      (begin
        (newline)
        (display "timer checkpoints: ")
        (print-list-as-term (cons 'c (reverse *checkpoint-keptnrs*)))
        (newline)(newline) ))
  (set! *otter-prepared-history* '())
  (set! *shown-clausenrs* '())
  (if (and (pair? history)
           (or (eq? 'hyper (car history))
               (eq? 'factor (car history))
               (eq? 'binres_demod (car history))
               (eq? 'binres (car history)) )
           (zero? (cadr history)) )
      (set! history
            (cons (car history)
                  (cons (+ 1 *kept-cl-nr*)
                        (cddr history) ))))
  (otter-show-history-aux history #t)


  ;(display "after otter-show-history: ")(newline)
  ;(for-each (lambda(x)(display (but-last x)) (newline))
  ;            *otter-prepared-history*)
  ;(pretty-print *otter-prepared-history*)

  (set! *otter-prepared-history*
        (sort-by-firstel  *otter-prepared-history*))
  (for-each (lambda(x)
              (newline)
              (display (car x))
              (display " [")
              (otter-show-history-printparents (but-last (cdr x)))
              (display "] ")
              (cond
               ((pair? (car (my-last-pair x)))
                  (print-clause (car (my-last-pair x))) )
               ((eq? '? (car (my-last-pair x)))
                  (display '?) )
               (else
                  (display "contradiction.") )))
            *otter-prepared-history*)
  (newline) (newline) )


(define (otter-show-history-printparents lst)
  (let ()
    (cond
     ((null? lst))
     ((symbol? (car lst))
        (display (car lst))
        (if (not (null? (cdr lst)))
            (display ",") )
        (otter-show-history-printparents (cdr lst)) )
     ((pair? (car lst))
        (otter-show-history-printparents-aux (car lst))
        (if (not (null? (cdr lst)))
            (display ",") )
        (otter-show-history-printparents (cdr lst)) )
     (else
        (display (car lst))
        (if (not (null? (cdr lst)))
            (display ",") )
        (otter-show-history-printparents (cdr lst)) ))))


(define (otter-show-history-printparents-aux y)
  (if (not (null? y))
      (begin
        (display (car y))
        (if (not (null? (cdr y)))
            (display "."))
        (otter-show-history-printparents-aux (cdr y)) )))

(define (otter-show-history-aux history pos)
  (let* ((old-history history)
         (constr (car history))
         (dernr (cadr history))
         (premissnrs '())
         (clausepl '())
         (clause '())
         (res '())
         (b1 '())
         (b2 '())
         (newres '())
         (demods '())
         (tmp '()) )
    ; filter out demods
    (do ((part history (cdr part)))
        ((null? part))
      (if (or (not (pair? (car part)))
              (not (eq? 'demod (caar part))) )
          (set! tmp (cons (car part) tmp)) ))
    (set! history (reverse tmp))


    ;(display "otter-show-history-aux: filter done")(newline)
    ;(display "pos: ")(display pos)(display " history: ")(newline)
    ;(display history)(newline)
    ;(pretty-print history)
    ;(newline)

    (set! res (cons dernr res))
    ; parent nrs for factor
    (if (or (eq? 'factor constr) (eq? 'binres_demod constr))
        (set! b1 (cadr (caddr history))) ) ; parent nr
    ; parent nrs for binary
    (if (eq? 'binres constr)
        (begin
          (set! b1 (cadr (cadr (caddr history)))) ; left parent nr
          (set! b2 (cadr (cadr (cadddr history)))) )) ; right parent nr
    (if (eq? 'hyper constr)
        (set! b1 (cadr (cadr (cadddr history))) )) ; first electr nr
    (cond
;     ((and (number? (cadr history))
;           (< *one-pass-kept-clauses-firstnr* (cadr history))
;           (begin
;             (set! tmp (show-history-find-clausepl (cadr history)))
;             (and tmp
;                  (eq? 'old (car tmp)) )))

     ((and (or (eq? 'factor constr)(eq? 'binres_demod constr))
           (or (zero? b1)
               (>= b1 *one-pass-kept-clauses-firstnr*) )
           (not (show-history-find-clausepl b1)) )
        ; process parent with a new nr

       ; (display "otter-show-history-aux: starting to proc parent")(newline)

       (set! tmp
            (otter-show-history-aux
             (append (list (caaddr history) dernr)
                     (cddr (caddr history)) )
             pos))
       ; add cutoff data
       (set! res (append (list ; (+ 1 (car (cdddr history)))
                               ; (+ 1 (car (cdr (cdddr history))))
                               'factor_s)
                         (cdr tmp) ))
       ; add derived clause
       (set! res (cons (car tmp) res)) )
     ((and (eq? 'hyper constr)
           (or (zero? b1)
               (>= b1 *one-pass-kept-clauses-firstnr*) )
           (not (show-history-find-clausepl b1)) )
        ; do the cutoff stuff
        ; process first electron with new nr

        ; (display "otter-show-history-aux: starting to proc electron")(newline)

        (set! tmp
              (otter-show-history-aux
                 (append (list (caadr (cadddr history)) dernr)
                       (cddr (cadr (cadddr history))) )
                  pos ))
         ; add cutoff data
        (set! res (append (list (cadr (caddr history)) 'hyper_s)
                          (cdr tmp) ))
         ; add derived clause
        (set! res (cons (car tmp) res))
        ; process the cutter
        (otter-show-history-aux  (caddr history) #f) )
     ((and (eq? 'binres constr)
           (or (and (or (zero? b1)
                        (>= b1 *one-pass-kept-clauses-firstnr*) )
                    (not (show-history-find-clausepl b1)) )
               (and (or (zero? b2)
                        (>= b2 *one-pass-kept-clauses-firstnr*) )
                    (not (show-history-find-clausepl b2)) )))


        ; (display "otter-show-history-aux: starting cutoff stuff")(newline)

         ; do the cutoff stuff
        (cond
            ((not (show-history-find-clausepl b1))
               ; process left(zero) parent with a new nr
               (set! tmp
                 (otter-show-history-aux
                   (append (list (caadr (caddr history)) dernr)
                           (cddr (cadr (caddr history))) )
                   pos ))
               ; add cutoff data
               (set! res (append (list b2 'binary_s) (cdr tmp)))
               ; aresdd derived clause
               (set! res (cons (car tmp) res))
               ; process the cutter
               (otter-show-history-aux
                  (cadr (cadddr history)) #f ))
            ((not (show-history-find-clausepl b2))
               ; process right(zero) parent with a new nr
               (set! tmp
                 (otter-show-history-aux
                   (append (list (caadr (cadddr history)) dernr)
                           (cddr (cadr (cadddr history))) )
                   pos ))
               ; add cutoff data
               (set! res (append (list b1 'binary_s) (cdr tmp)))
               ; add derived clause
               (set! res (cons (car tmp) res))
               ; process the cutter
               (otter-show-history-aux
                  (cadr (caddr history)) #f ))))
     ((and (number? dernr)
           (< dernr *one-pass-kept-clauses-firstnr*) )


    ; (display "otter-show-history-aux: number? dernr")(newline)

        (if (not (show-history-find-clausepl dernr))
            (begin
              (set! res (cons '? res))
              (set! res (cons '? res)) )
            (begin
              (if (not (eq? 'input constr))
                  (set! res (cons '? (cons constr res))) )
              (set! res (cons (otter-show-history-aux-aux  dernr '() '())
                              res)) )))
     ((eq? 'binres constr)


        ; (display "otter-show-history-aux: constr is binres")(newline)

        (set! res (cons 'binary res))
        ; derivation rule and premisses:
        (set! res
              (cons
               (if (not (eqv? 1 (caddr (caddr history))))
                   (list  (show-history-clnr (caddr history))
                          (caddr (caddr history)) )
                   (show-history-clnr (caddr history)) )
               res))
        (set! res
              (cons
                (if (not (eqv? 1 (caddr (cadddr history))))
                    (list  (show-history-clnr (cadddr history))
                           (caddr (cadddr history)) )
                    (show-history-clnr (cadddr history)) )
               res))
        (set! res
              (cons
               (otter-show-history-aux-aux
                  dernr history (list (caddr history) (caddr (cdr history))) )
               res)) )
     ((eq? 'hyper constr)
        (set! res (cons 'hyper res))
        ; derivation rule and premisses:
        (set! res (cons (show-history-clnr (caddr history)) res))
        (for-each (lambda(x)
                    (set! res (cons (show-history-clnr x) res))
                    (if (and (pair? x)
                             (eq? 'literal (car x))
                             (not (eq? 1 (caddr x))) )
                         (set! res (cons (list (caddr x)) res)) ))
                  (cdddr history) )
        (set! res
              (cons (otter-show-history-aux-aux dernr history (cddr history))
                    res)))
     ((eq? 'para constr)
        (set! res
          (append
            (list
               (show-history-clnr (caddr (cdr history)))
              (cons (show-history-clnr (caddr history))
                    (cdddr (cdr history)) )
              'para)
            res))
        (set! res
              (cons
               (otter-show-history-aux-aux
                dernr history (list (caddr history) (caddr (cdr history))) )
               res)))
     ((or (eq? 'factor constr) (eq? 'binres_demod constr))
        (set! res
          (append
             (list
              ;(+ 1 (cadddr history))
              ;(+ 1 (cadddr (cdr history)))
              (show-history-clnr (caddr history))
              constr)
             res))
        (set! res
              (cons
               (otter-show-history-aux-aux
                dernr history (list (caddr history)) )
              res)))
     ((eq? 'input constr)
        (set! clausepl (show-history-find-clausepl dernr))
        ; showing the derived clause
        (if (and clausepl
                 (eq? 'old (car clausepl)) )
            (set! clausepl (cdr clausepl)) )
        (if clausepl
            (begin
              (set! res  (cons (car clausepl) res))
              (set! *shown-clausenrs* (cons dernr *shown-clausenrs*)) )))
     ((eq? 'backdemod constr)
        (set! res
          (append
             (list
              ;(+ 1 (cadddr history))
              ;(+ 1 (cadddr (cdr history)))
              (show-history-clnr (caddr history))
              'backdemod)
             res))
        (set! res
              (cons
               (otter-show-history-aux-aux
                dernr history (list (caddr history)) )
              res)))

;        (set! res (cons 'backdemod res))
;        ; derivation rule and premisses:
;        (set! res (cons (show-history-clnr (caddr history)) res))
;        (set! res
;              (cons
;               (otter-show-history-aux-aux
;                  dernr history (list (caddr history)) )
;               res))
;        (set! clausepl (show-history-find-clausepl dernr))
;        (if (and clausepl
;                 (eq? 'old (car clausepl)) )
;            (set! clausepl (cdr clausepl)) )
;        (if clausepl
;            (begin
;              (set! res  (cons (car clausepl) res))
;              (set! *shown-clausenrs* (cons dernr *shown-clausenrs*)) )))
     ((eq? 'literal constr)
        (otter-show-history-aux (cadr history) pos)
        (set! res #f) )
     ((number? pos)
            (newline)(display-spaces pos) )
     (else
       (report-error "unknown history-constructor: " constr)))


   ; (display "otter-show-history-aux: near end case")(newline)

    (if res
        (begin
          ; treat demods
          (do ((part old-history (cdr part)))
              ((null? part))
            ;(newline)(display "part: ")(display part)(newline)
            (if (and (pair? (car part)) (eq? 'demod (caar part)))
                (set! demods (append (cdar part) demods)) ))

         ;(display "otter-show-history-aux: demods: ")
         ;(display demods)
         ;(newline)

          (if (not (null? demods))
              (begin
                (set! res
                      (cons (car res)
                            (append
                             (map (lambda(x)
                                    (car (clinfo-clnumber-sublist (car x))))
                                  demods)
                             (cons 'demod (cdr res)) ))) ))

                 ; (display "demods2: ")(display res)(newline)

                (for-each
                 (lambda(x)
                   ;(display
                   ; (car (clinfo-history-sublist (car x))) )
                   (otter-show-history-aux
                    (car (clinfo-history-sublist (car x))) pos))
                 demods)

          ; OK, continue
          (set! tmp (assq (car (my-last-pair res)) *otter-prepared-history*))
          (cond
           ((not tmp)
            (set! *otter-prepared-history*
                  (cons (reverse res) *otter-prepared-history*) ))
           ((< (length tmp) (length res))
            (set-cdr! tmp (cdr (reverse res))) ))))
    res))


(define (sort-by-firstel lst)
  (let ((res '()))
    (do ((part lst (cdr part)))
        ((null? part)
         res)
      (cond
       ((null? res) (set! res (list (car part))))
       ((< (caar part) (caar res))
          (set! res (cons (car part) res)) )
       (else
         (insert-to-firstel-sorted! (car part) res) )))))

(define (insert-to-firstel-sorted! el lst)
  (if (or (null? (cdr lst))
          (< (car el) (caadr lst)) )
      (set-cdr! lst (cons el (cdr lst)))
      (insert-to-firstel-sorted! el (cdr lst)) ))

(define (otter-show-history-aux-aux dernr history elements)
  (let ((tmp '())
        (res '())
        (clausepl '())
        (old-flag #f) )
    ; actually derived clause:
    (set! clausepl (show-history-find-clausepl dernr))
    (if clausepl
        (begin
          (set! *shown-clausenrs* (cons dernr *shown-clausenrs*))
        ;  (if (eq? 'old (car clausepl))
        ;      (begin
        ;        (set! clausepl (cdr clausepl))
        ;        (set! old-flag #t) ))
          (if (not (pair? (car clausepl)))
              (set! res #f)
              (set! res (car clausepl)) )))
    ; parents:
    (if (not old-flag)
        (for-each (lambda(x)
                    (set! tmp (show-history-clnr x))
                    (if (not (member tmp *shown-clausenrs*))
                        (begin
                          (set! clausepl (show-history-find-clausepl tmp))
                          (if clausepl
                              (otter-show-history-aux
                               (car (clinfo-history-sublist
                                     (caar clausepl))) #t)
                              (otter-show-history-aux x #t) )))) ;
                  elements))
    res))


(define (show-history-aux history pos)
  (let* ((constr (car history))
         (dernr (cadr history))
         (premissnrs '())
         (clausepl '())
         (clause '())
         (tmp '()) )
    (cond
     ((eq? 'hyper constr)
        (newline)(display-spaces pos)
        (display "by ")(display constr)(display " from ")
        ; derivation rule and premisses:
        (display (show-history-clnr (caddr history)))
        (display " with ")
        (for-each (lambda(x)
                    (display (show-history-clnr x))
                    (if (and (pair? x)
                             (eq? 'literal (car x))
                             (not (eq? 1 (caddr x))) )
                      (begin (display #\()(display (caddr x))(display #\))) )
                    (display #\space) )
                  (cdddr history) )
        (show-history-aux-aux dernr history (cddr history)) )
     ((eq? 'binres constr)
        (newline)(display-spaces pos)
        (display "by binary from ")
        (display (show-history-clnr (caddr history)))
        (if (> (caddr (caddr history)) 1)
            (begin (display #\() (display (caddr (caddr history)))
                   (display #\)) ))
        (display " and ")
        (display (show-history-clnr (cadddr history)))
        (if (> (caddr (cadddr history)) 1)
            (begin (display #\() (display (caddr (cadddr history)))
                   (display #\)) ))
        (show-history-aux-aux
           dernr history (list (caddr history) (caddr (cdr history))) ))
     ((eq? 'para constr)
        (newline)(display-spaces pos)
        (display "by para from ")
        (display (show-history-clnr (caddr history)))
        (display " into ")
        (display (show-history-clnr (caddr (cdr history))))
        (display " at place ")
        (display (cdddr (cdr history)))
        (show-history-aux-aux
           dernr history (list (caddr history) (caddr (cdr history))) ))
     ((eq? 'factor constr)
        (newline)(display-spaces pos)
        (display "by factor from ")
        (display (show-history-clnr (caddr history)))
        (display " literals ")
        (display (+ 1 (cadddr history)))
        (display #\space)
        (display (+ 1 (cadddr (cdr history))))
        (show-history-aux-aux dernr history (list (caddr history))) )
      ((eq? 'binres_demod constr)
        (newline)(display-spaces pos)
        (display "by binary demod from ")
        (display (show-history-clnr (caddr history)))
        (show-history-aux-aux dernr history (list (caddr history))) )
     ((eq? 'input constr)
        (newline)(display-spaces pos)
        (display "input: ")(newline)
        (set! clausepl (show-history-find-clausepl dernr))
        ; showing the derived clause
        (display dernr)(display ": ")
        (if clausepl
            (begin
              (if *int-logic-flag*
                  (print-full-int-clause (car clausepl))
                  (print-full-clause (car clausepl)) )
              (set! *shown-clausenrs* (cons dernr *shown-clausenrs*)) )))
     ((eq? 'literal constr)
        (show-history-aux (cadr history) 0) )
     (else
        (newline)(display-spaces pos)
       (report-error "unknown history-constructor: " constr) ))))


(define (show-history-aux-aux dernr history elements)
  (let ((tmp '())
        (clausepl '()) )
    (display " we get:")(newline)(display dernr)(display ": ")
    ; actually derived clause:
    (set! clausepl (show-history-find-clausepl dernr))
    (if clausepl
        (begin
          (if *int-logic-flag*
              (print-full-int-clause (car clausepl))
              (print-full-clause (car clausepl)) )
          (set! *shown-clausenrs* (cons dernr *shown-clausenrs*)) ))
    ; parents:
    (for-each (lambda(x)
                (set! tmp (show-history-clnr x))
                (if (not (member tmp *shown-clausenrs*))
                    (begin
                      (set! clausepl (show-history-find-clausepl tmp))
                      (if clausepl
                          (show-history-aux
                            (car (clinfo-history-sublist (caar clausepl))) 0)
                          (show-history-aux x 0) )))) ;
              elements)))


(define *single-reduction-rule-clauses-list* '())
(define *single-rule-clauses-list* '())
(define *mult-rule-clauses-list* '())


(define (show-history-find-clausepl  dernr)
  (if (or (not (number? dernr)) (zero? dernr))
      #f
      (let ((clausepl #f))
        (if (not clausepl)
            (set! clausepl (member-if
                            (lambda(x)(= dernr (caddr (cddr (car x)))))
                                  *sos-clauses-list*)))
        (if (and (not clausepl) (not (null? *usable-clauses-list*)))
            (set! clausepl (member-if
                            (lambda(x)(= dernr (caddr (cddr (car x)))) )
                            (cdr *usable-clauses-list*) )))
        (if (and (not clausepl) (not (null? *unit-alist*)))
            (set! clausepl
                  (some? (lambda(y)
                           (member-if
                            (lambda(x)(= dernr (caddr (cddr (car x)))))
                            (cdr y) ))
                         *unit-alist*)))
        (if (and (not clausepl) (not (null? *unit-alist-list*)))
            (set! clausepl
                  (some?
                   (lambda(x)
                       (some? (lambda(y)
                           (member-if
                            (lambda(x)(= dernr (caddr (cddr (car x)))))
                            (cdr y) ))
                              x))
                   *unit-alist-list*)))
        (if (and (not clausepl) (not (null? *sos-clauses-list-list*)))
            (set! clausepl
                  (some?
                   (lambda(x)
                     (member-if
                      (lambda(y)
                        (and (pair? y)
                             (= dernr (caddr (cddr (car y)))) ))
                      x))
                  *sos-clauses-list-list*)))
        (if (and (not clausepl) (not (null? *usable-clauses-list-list*)))
            (set! clausepl
                  (some?
                   (lambda(x)
                     (member-if
                      (lambda(y)
                        (and (pair? y)
                             (= dernr (caddr (cddr (car y)))) ))
                      x))
                  *usable-clauses-list-list*)))
        (if (and *int-logic-flag* (not clausepl))
            (set! clausepl (member-if
                             (lambda(x)(= dernr (caddr (cddr (car x)))))
                             *mult-rule-clauses-list*)))
        (if (and *int-logic-flag* (not clausepl))
            (set! clausepl (member-if
                             (lambda(x)(= dernr (caddr (cddr (car x)))))
                             *single-rule-clauses-list*)))
        (if (and *int-logic-flag* (not clausepl))
            (set! clausepl (member-if
                             (lambda(x)(= dernr (caddr (cddr (car x)))))
                             *single-reduction-rule-clauses-list*)))
        ;(if (and clausepl
        ;         (< dernr *one-pass-kept-clauses-firstnr*) )
        ;    (set! clausepl (cons 'old clausepl)) ) ; derived in old pass
        clausepl)))


(define (show-history-clnr premiss)
  (cond ((not (pair? premiss))
           '?)
        ((eq? (car premiss) 'literal)
           (cadr (cadr premiss)) )
        ((eq? (car premiss) 'input)
           (cadr premiss) )
        ((eq? (car premiss) 'hyper)
           (cadr premiss) )
        ((eq? (car premiss) 'binres)
           (cadr premiss) )
        ((eq? (car premiss) 'factor)
           (cadr premiss) )
        ((eq? (car premiss) 'binres_demod)
           (cadr premiss) )
        ((eq? (car premiss) 'para)
           (cadr premiss) )
        (else '??) ))

(define (display-spaces nr)
  (do ((i 0 (+ 1 i)))
      ((>= i nr))
    (display #\space) ))





; in history->answer input-sos is the sos used for building the
; history and input-ans-sos is the same sos with the typet ans-literals
; added

(define (history->answer history)
  (clear-x-vars-and-touched! *max-nr-of-vars-in-clause*)
  (clear-y-vars-and-touched! *max-nr-of-vars-in-clause*)
  (clear-z-vars-and-touched! *max-nr-of-vars-in-clause*)
  (let ((res (history->answer-step history)))
    (clear-x-vars-and-touched! *max-nr-of-vars-in-clause*)
    (clear-y-vars-and-touched! *max-nr-of-vars-in-clause*)
    (clear-z-vars-and-touched! *max-nr-of-vars-in-clause*)
    (make-readable-cl (cdr res)) ))


(define (history->answer-step history)
  (let ((constr (car history))
        (dernr (cadr history))
        (orig-cl '())
        (ans-cl '()) )
    (cond
      ((eq? 'input constr)
         (set! ans-cl (input-ans-clause dernr))
         ans-cl)
      ((eq? 'hyper constr)
         (let* ((ans-nucleon (history->answer-step (caddr history)))
                (litlst (caddr (car ans-nucleon)))
                (res ans-nucleon)
                (tmp '())
                (arg '()) )
           (do ((argspart (cdddr history) (cdr argspart))
                (litnr 1 (+ 1 litnr)) )
               ((null? argspart)
                  res)
                  ;(cons (car res)
                  ;        (map rename-y-to-x-term (cdr res)) )
             (set! res (cons (car res)
                             (map rename-x-to-y-term (cdr res)) ))
             (set! arg (history->answer-step (cadr (car argspart))))
             (set! res (history->answer-binstep
                         (cdr arg) (cdr res) (car arg) (car res)
                         (caddr (car argspart)) litnr history #t '())))))
      ((eq? 'para constr)
         (let* ((fromcl (history->answer-step (car (cddr history))))
                (tocl (history->answer-step (cadr (cddr history))))
                (fromlitnr (caddr (cddr history)))
                (tolitnr (caddr (cdr (cddr history))))
                (left-used-flag (caddr (cddr (cddr history))))
                (path (caddr (cdddr (cddr history))))
                (frompurecl (map rename-x-to-y-term (cdr fromcl)))
                (res '()) )
           (set! res (history->answer-binstep
                        (cdr tocl) frompurecl (car tocl) (car fromcl)
                        tolitnr fromlitnr history
                        left-used-flag path))
           res)))))


(define (input-ans-clause dernr)
  (let ((place (member-if (lambda(x)
                            (eq? (car (clinfo-clnumber-sublist (car x)))
                                 dernr))
                           *ans-input-sos*)))
    (if (not place)
        (report-error "input clause not found: " dernr)
        (car place) )))


; history->answer-binstep gives a fully built clause resulting
; from one binary resolution or paramodulation step.
; hyper: nucleus is always ycl.
; para: fromcl is always ycl.

(define (history->answer-binstep xcl ycl xinf yinf
                                 xlitnr ylitnr derhistory
                                 left-used-flag path)
 ;(newline)(display "two-clause-binres-loop entered")(newline)
  (let ((xlitinf (caddr xinf))
        (ylitinf (caddr yinf))
        (yclmaxvar (give-cl-maxvar yinf))
        (ypart '())
        (ylitinfpart '())
        (xpart '())
        (xlitinfpart '())
        (subterm '())
        (tmpres '()) )
    (do ((lypart ycl (cdr lypart))
         (lylitinfpart ylitinf (cdr lylitinfpart))
         (litnr 1 (+ 1 litnr)) )
        ((if (eq? 'hyper (car derhistory))
             (eq? (get-litinfo-ordernr (caar lylitinfpart)) ylitnr)
             (and (eq? litnr ylitnr) (not (answer-lit? (car lypart)))))
          (set! ypart lypart)
          (set! ylitinfpart lylitinfpart) )
      (if (answer-lit?  (car lypart)) (set! litnr (- litnr 1))) )
    (do ((lxpart xcl (cdr lxpart))
         (lxlitinfpart xlitinf (cdr lxlitinfpart))
         (litnr 1 (+ 1 litnr)) )
        ((and (eq? litnr xlitnr) (not (answer-lit? (car lxpart))))
           (set! xpart lxpart)
           (set! xlitinfpart lxlitinfpart) )
      (if (answer-lit?  (car lxpart)) (set! litnr (- litnr 1))) )
    (let* ((ylit (car ypart))
           (yfun (car ylit))
           (xlit (car xpart))
           (xfun (car xlit))
           (ylitcurinf (car ylitinfpart))
           (compl-yfunpure (logxor *sgn-mask* (pure-fun yfun))) )
      (cond
       ((eq? 'para (car derhistory))
        ; paramodulation stuff. here always para from ycl.
          (set! *param-tolit* xlit)
          (set! *param-eqlit* ylit)
          (set! *param-tocl* xcl)
          (set! *param-fromcl* ycl)
          (set! *param-fromfullinf* yinf)
          (set! *param-tofullinf* xinf)
          (set! *param-left-arg* (eq-lit-arg1 (cadr ylit)))
          (set! *param-right-arg* (eq-lit-arg2 (cadr ylit)))
          (set! *param-toinf* xlitinf)
          (set! *param-frominf* ylitinf)
          (set! *param-from-is-y-flag* #t)
          (set! *param-yclmaxvar* yclmaxvar)
          (set! *param-xclmaxvar* (give-cl-maxvar xinf))
          (set! *param-fromhist* '())
          (set! *param-tohist* '())
          (set! *param-fromlitnr* ylitnr)
          (set! *param-tolitnr* xlitnr)
          (set! *param-use-left-flag* left-used-flag)
          ; paramodulation into xlit
          ; always para from the left arg (by exchanging args if needed):
          (if (not *param-use-left-flag*)
              (begin (set! *param-left-arg* (eq-lit-arg2 (cadr ylit)))
                     (set! *param-right-arg* (eq-lit-arg1 (cadr ylit))) ))
          ; here and in the following we assume that path is not in reverse.
          ; unify the terms:
          (set! subterm (subterm-at-path *param-tolit* path))
          (cond
            ((composite? *param-left-arg*)
                (set! *compare-ground-litcdr-equal-flag* #t)
               (if (or (not (eq? (pure-fun (car subterm))
                                 (pure-fun (car *param-left-arg*)) ))
                       (not (resolve-unify-ordinary
                                 (car subterm) (car *param-left-arg*)
                               subterm *param-left-arg*)))
                   (report-error "cannot unify for paramodulation:"
                                 (list subterm *param-left-arg*) ))
               (set! *compare-ground-litcdr-equal-flag* #f) )
            ((non-composite:-var? *param-left-arg*)
               ; just set the var (it is always yvar)
               (set-var-ref! (var-nr *param-left-arg*) subterm)
               (set! *xvars-set-flag* #f)(set! *yvars-set-flag* #t) )
            ((not (eq? *param-left-arg* subterm))
               (report-error "cannot unify for paramodulation:"
                                  (list subterm *param-left-arg*)))
            (else
              (set! *xvars-set-flag* #f)(set! *yvars-set-flag* #f) ))
          ; construct the clauses with substitutions and replacements applied:
          (set! tmpres (paramodulate-into-term-aux-typet
                          path *param-right-arg* left-used-flag))
          tmpres)
       ; now the hyperresolution case
       ((eq? 'hyper (car derhistory))
          ;(newline)(display "xlit: ")(newline)
          ;(print-lit xlit)(newline)
          (set! *compare-ground-litcdr-equal-flag* #t)
          (if (not (and (eq? compl-yfunpure (pure-fun xfun))
                        (resolve-unify-ordinary
                           xfun yfun xlit ylit) ))
              (report-error
                "unification during answer-construction failed:"
                 (list xlit ylit)))
          (set! *compare-ground-litcdr-equal-flag* #f)
          (let* ((ytmp (make-rest-inf-cl
                          ycl ylit (caddr yinf)))
                 (xtmp (make-rest-inf-cl
                          xcl xlit (caddr xinf)))
                 (res  (make-raw-inf-cl-pair-typet-hyper
                         (car ylitinfpart) #t
                         (cdr xtmp) (cdr ytmp)
                         (car xtmp) (car ytmp) )))
            ; the new 1-st level deco clause and
            ; inflist has been built
            ; set history
            ;(debug-print-yvars yclmaxvar)
            ;(debug-print-xvars (give-cl-maxvar xinf))
            ;(newline)
            ;(display "raw cl by binary; parents: ")
            ;(newline)
            ;(print-pure-clause xcl)(newline)
            ;(print-pure-clause ycl)(newline)
            ;(print-pure-clause (cdr res))(newline)
            ;clear all vars (subst-deco sets both!)

            (clear-y-vars-and-touched! yclmaxvar)
            (clear-x-vars-and-touched! (give-cl-maxvar xinf) )
            (set! res (history->ans-handle-derived-inf-cl-pair
                        res (list 'dummy 0) #f))
            res))))))

(define (history->ans-handle-derived-inf-cl-pair pr history fac-gen-flag)
  ;(newline)(display "handle-derived-inf-cl-pair entered with:")
  ;(display (car pr))(newline)
  ;(display (cdr pr))(newline)
  ;(debug-print-pure-clause (cdr pr))(newline)

  (let ((cl (cdr pr))
        (inf (car pr))
        (clauseinf '())
        (tmp-pr '())
        (kept-cl '())
        (litinfolst '()) )

    ;(newline)(display "cl part of pair: ")(newline)
    ;(print-pure-clause cl)(newline)
    ;(display "inf part of pair: ")(newline)
    ;(display inf)(newline)

    (set! tmp-pr (remove-double-lits! inf cl *hyper-strategy-flag*))
    (set! tmp-pr
          (sort-clause (map cons (car tmp-pr) (cdr tmp-pr)) #t) )
    (set! tmp-pr
          (cons (map car tmp-pr) (map cdr tmp-pr)) )
    (set! clauseinf (make-clause-info (cdr tmp-pr) history #f))
    (set! kept-cl (make-kept-clause clauseinf tmp-pr))
    (set! litinfolst (map cddr (car tmp-pr)))

    ; (newline)
    ; (display "derived clause after 2-nd order deco:")
    ; (newline)
    ;(display kept-cl)(newline)
    ;(debug-print-clause kept-cl)(newline)
    ;(print-clause kept-cl)(newline)
    ;(debug-print-clause kept-cl)(newline)

    ; insert a right litinfolst into the kept-cl infopart
    (set-car! (cddar kept-cl) litinfolst)

    ;(newline)(display "kept clause in pure form:")
    ;(newline)
    ;(print-pure-clause (cdr kept-cl))
    ;(newline)
    ;(newline)(display "kept clause in debug form:")
    ;(newline)
    ;(debug-print-clause kept-cl)

    kept-cl))


; make-raw-inf-cl-pair-typet takes:
;       litinfel: a litinfolst element corresponding to the cut lit of the nuc
;       nuc-is-y-flag: #t iff ycl is nucleon, else #f
;  and all the same things that make-raw-inf-cl-pair takes.
;
; If there are two answer-literals, it merges them into one:
; an answer-literal of a non-nucleon is inserted into the answer-literal
; of the nucleon


(define (make-raw-inf-cl-pair-typet-hyper
                litinfel nuc-is-y-flag xcl ycl xinf yinf)
  ; first build up a new 1-st level deco new-yrest and new-yinf
  (let* ((new-yrest (subst-deco-cl ycl yinf 2 *yvars-set-flag*)))

    ;(newline)(display "new-yrest in pure form: ")(newline)
    ;(print-pure-clause (cdr new-yrest))(newline)
    ;(display "new-yinfo: ")(newline)
    ;(display (car new-yrest))(newline)

    (set! *used-demodulators* '())
    (if new-yrest
        ; new-yrest passed 1-st level criterias
        (let* ((new-xrest
                 (subst-deco-cl xcl xinf 1 *xvars-set-flag*) ))

          ;(newline)(display "new-xrest in pure form: ")(newline)
          ;(print-pure-clause (cdr new-xrest))(newline)
          ;(display "new-xinfo: ")(newline)
          ;(display (car new-xrest))(newline)

          (set! *used-demodulators* '())
          (if new-xrest
              ; new-xrest passed 1-st level criterias,
              ; handle the answer-lit manipulation
              (let* ((nuc-ans-lit
                       (member-ans-lit
                          (if nuc-is-y-flag (cdr new-yrest) (cdr new-xrest)) ))
                     (elec-ans-lit
                       (member-ans-lit
                           (if nuc-is-y-flag (cdr new-xrest)(cdr new-yrest)) ))
                     (new-nuc-ans-lit '()) )
                 (if (and nuc-ans-lit elec-ans-lit)
                     ; both exist
                     (begin
                       (set! new-nuc-ans-lit
                             (replace-subterm
                               (car nuc-ans-lit)
                               (list 1 (get-litinfo-ordernr (car litinfel)))
                               (cadar elec-ans-lit) ))
                       (if nuc-is-y-flag
                           (begin
                             (set! new-yrest
                                   (cons (car new-yrest)
                                         (replace-one-eq
                                           (car nuc-ans-lit) new-nuc-ans-lit
                                           (cdr new-yrest) )))
                             (set! new-xrest
                                   (make-rest-inf-cl
                                       (cdr new-xrest) (car elec-ans-lit)
                                       (car new-xrest) )))
                           (begin
                             (set! new-xrest
                                   (cons (car new-xrest)
                                         (replace-one-eq
                                           (car nuc-ans-lit) new-nuc-ans-lit
                                           (cdr new-xrest) )))
                             (set! new-yrest
                                   (make-rest-inf-cl
                                       (cdr new-yrest) (car elec-ans-lit)
                                       (car new-yrest) ))))))
                 (cons (my-append! (car new-yrest) (car new-xrest))
                       (my-append! (cdr new-yrest) (cdr new-xrest)) ))
              ; 1-st level criterias failed
              #f))
        #f)))


; paramodulate-into-term-aux-typet is called in case of success to build,
; manipulate and store a new clause and to clear vars.
; in this typet version path is not reversed.
;
; it does not use make-raw-inf-cl-pair.


(define (paramodulate-into-term-aux-typet path replterm left-used-flag)

  ;(print *xvars-set-flag*)(newline)(print *yvars-set-flag*)(newline)
  ;(begin (newline)(debug-print-xvars 6)(newline)
  ; (debug-print-yvars 6)(newline))

  (set! *compare-ground-litcdr-equal-flag* #f)
  (let* ((newtype (subst-deco-term (eq-lit-type (cadr *param-eqlit*)) #f))
         (lambdalit (subst-deco-replace-term
                       *param-tolit* path *lambda-dummypar-nr* #t))
         (lambdabody
           (map (lambda(x)
                   (if (eq? x *param-tolit*)
                       (cadr x)
                       (cadr (subst-deco-term x #t))) )
                (remove-ans-literals
                   (replace-one-eq *param-tolit* lambdalit *param-tocl*)) ))
         ; (lambdabody (list lambdalit)) ; !!!!!!!! NB!!! Unit tocl assumed!
         (newleft (subst-deco-term *param-left-arg* #f))
         (newright (subst-deco-term *param-right-arg* #f))
         (newlit (subst-deco-replace-term
                    *param-tolit* path replterm #t))
         ; remove both to- and from-literals and corresp litinfos:
         (tmpto1 (make-rest-inf-cl
                   *param-tocl* *param-tolit* *param-toinf*))
         (tmpfrom1 (make-rest-inf-cl
                   *param-fromcl* *param-eqlit* *param-frominf*))
         ; find sublists starting with ans-literals and ans-literals
         ; and substitute into the literals:
         (to-ans-lit-lst (member-ans-lit (cdr tmpto1)))
         (from-ans-lit-lst (member-ans-lit (cdr tmpfrom1)))
         (to-ans-lit (if to-ans-lit-lst
                         (subst-deco-term (car to-ans-lit-lst) #t) #f))
         (from-ans-lit (if from-ans-lit-lst
                           (subst-deco-term (car from-ans-lit-lst) #t) #f))
         ; remove ans-literals and corresp litinfos:
         (tmpto (if to-ans-lit-lst
                    (make-rest-inf-cl
                       (cdr tmpto1) (car to-ans-lit-lst) (car tmpto1) )
                    tmpto1))
         (tmpfrom (if from-ans-lit-lst
                      (make-rest-inf-cl
                       (cdr tmpfrom1) (car from-ans-lit-lst) (car tmpfrom1) )
                    tmpfrom1))
         ; make a union and substitute varvals into it:
         ; res1 contains answer-literals,
         ; res does not.
         (res1 (make-raw-inf-cl-pair
                 (cdr tmpto1) (cdr tmpfrom1)
                 (car tmpto1) (car tmpfrom1) ))
         (res  (make-raw-inf-cl-pair
                 (cdr tmpto) (cdr tmpfrom)
                 (car tmpto) (car tmpfrom) ))
         (subst-applic '())
         (new-ans-lit '()) )

    (clear-y-vars-and-touched! *param-yclmaxvar*)
    (clear-x-vars-and-touched! *param-xclmaxvar*)
    ; if there is no from-ans-lit, then the from-eq is
    ; a demodulator used for paramodulation. In that case do not
    ; treat the step as paramodulation: treat is just as a simple
    ; substitution into vars.
    ; if there is no to-ans-lit but there is a from-ans-lit, then
    ; we are in trouble since we have a paramodulation from a non-demodulator
    ; into a demodulator. It is temporarily prohibited.
    ; here we assume both from-cl and to-cl to be unit (except answer-lits)
    (if (not (and to-ans-lit from-ans-lit))
        ; a substitution-to-vars and demodulation step,
        ; thus just add a newlit and keep answer-literals if any.
        (set! res (cons (cons (cons #t (cons #f (list 0)))
                              (car res1) )
                        (cons newlit (cdr res1)) ))
        ; a real construct-typet-subst step:
        (begin
          (set! subst-applic
                (deco-leading-fun
                 (list
                  *ans-pred-nr*
                  (deco-leading-fun
                   (list (if *param-use-left-flag*
                              *subst-fun-nr* *subst-prim-fun-nr*)
                         newtype
                         (deco-leading-fun
                          (cons *lambda-fun-nr*
                                (cons *lambda-dummypar-nr*
                                      lambdabody)))
                         newleft
                         newright
                         (cadr to-ans-lit)
                         (cadr from-ans-lit) )))))
          ; add a new inf of tolit and a ready tolit to res:
          (set! res (cons (cons (cons #t (cons #t (list 0)))
                                (car res) )
                          (cons newlit (cdr res)) ))

          ;(newline)(display res)(newline)

          ; add answer-lit and its info:
          (set! res (cons (cons (cons #t (cons #t (list *answer-litinf*)))
                                (car res) )
                          (cons subst-applic (cdr res)) ))))
    ;res is built.
    ;(newline)(display "asnwer-building para step gave a raw clause:")(newline)
    ;(print-pure-clause (cdr res)) (newline)
    ;(newline)(display res)(newline)

    (set! res (history->ans-handle-derived-inf-cl-pair
                 res (list 'dummy 0) #f))
    res))


;======================= EOF resmain.scm ====================================
